//---------------------------------------------------------------------------
//
//  FILE NAME:          SCSConnection.cc
//
//  DEVELOPMENT:        CXCDS, User Infrastructure
//
//  DESCRIPTION:
//
//
//---------------------------------------------------------------------------
// Copyright (c) 2006, Smithsonian Astrophysical Observatory
//---------------------------------------------------------------------------



#include <unistd.h>
#include <string.h>
#include <sys/select.h>
#include <iostream>
#include <signal.h>
#include <iostream>
#include <sstream>
#include <errno.h>

#include "SCSConnection.hh"
#include "SCSMessage.hh"
#include "SCSLog.hh"

using namespace std;

SCSConnection::SCSConnection(int inSocket)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::SCSConnection(int inSocket) enter",2);
#endif 

  connectionSocket=-1;
  if (inSocket >=0 )
    connectionSocket=inSocket;
  writeReadyFlag=false;
  readReadyFlag=false;

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::SCSConnection(int inSocket) exit",2);
#endif 
};


SCSConnection::~SCSConnection()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::~SCSConnection() enter",2);
#endif 
 

  CloseProtocol();

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::~SCSConnection() exit",2);
#endif 
};


//
// Pre: Connection is initialized( it contains an open socket)
//
//
// Post: data in inMessage is send through a socket.
//       Returns 1 on success , 0 on timeout, -1 error.
//       If timeout is negative write will block untill message 
//       is sent. 
//
//
// Desc:
//

int SCSConnection::SendMessage(long inTimeoutUs,const SCSMessage& inMessage)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::SendMessage(long inTimeoutUs,const SCSMessage& inMessage) enter",2);
#endif 

  int stat=0;


  if ((stat=IsWriteReady(inTimeoutUs)) <=0 )
    {
#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"SCSConnection::SendMessage(long inTimeoutUs,const SCSMessage& inMessage) exit "<<stat;
      SCSLog::Write(tmpStr.str(),2);
#endif 
      return stat;
    };
  
  stat = SendMessage(inMessage);
  
  
#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSConnection::SendMessage(long inTimeoutUs,const SCSMessage& inMessage) exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif
  
  return stat;

};


//
// Pre: Connection is initialized( it contains an open socket)
//
//
//
// Post: data is read from the socket and put in outMessage 
//       Returns 1 on success , 0 on timeout, -1 error.
//       If timeout is negative write will block untill message 
//       is received. 
//
//
// Desc:
//


int SCSConnection::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::ReceiveMessage(long inTimeoutUs,const SCSMessage& inMessage) enter",2);
#endif

  int stat=0;


  if ((stat=IsReadReady(inTimeoutUs)) <=0 )
    {
#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"SCSConnection::ReceiveMessage(long inTimeoutUs,const SCSMessage& inMessage) exit "<<stat;
      SCSLog::Write(tmpStr.str(),2);
#endif
      return stat;
    };

  stat = ReceiveMessage(outMessage);

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSConnection::ReceiveMessage(long inTimeoutUs,const SCSMessage& inMessage) exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif
  return stat;
};


//
// Pre: 
//
//
// Post: Closes local socket.
//       
//
// Desc:
//


void SCSConnection::Close()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::Close() enter",2);
#endif

  if ( connectionSocket > 0 )
    {
      close(connectionSocket);
      connectionSocket=-1;
      
      writeReadyFlag=false;
      readReadyFlag=false;

    };
  
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::Close() exit",2);
#endif

};


//
// Pre: Socket is open
//
//
// Post: Sends close message to the other end of connection and 
//       closes local socket.
//       
//
// Desc:
//


int SCSConnection::CloseProtocol()
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::CloseProtocol() enter",2);
#endif

  int stat=1;

  if ( connectionSocket >= 0 )
    {
      SCSMessage tmpMessage;
      tmpMessage.Set(0,0,SCSMessage::PROTOCOL_CLOSE);
      stat=SendMessage(0,tmpMessage);


      close(connectionSocket);
      connectionSocket=-1;

      writeReadyFlag=false;
      readReadyFlag=false;

    };

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSConnection::CloseProtocol() exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif

  return stat;

};


//
// Pre: Socket is write ready
//
//
// Post: Message is sent through write ready socket
//       Returns 1 on success , -1 otherwise
//
// Desc:
//

int SCSConnection::SendMessage(const SCSMessage& inMessage)
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) enter",2);
#endif

  long doneSize=0;
  long tmpSize=(inMessage.size+1);
  long tmpSize1=sizeof(char)*(inMessage.size+1);
  char* tmpBuf= new char[SCSMessage::SCS_MESSAGE_HEADER_SIZE];
  struct sigaction sigPipeAction;
  struct sigaction oldSigPipeAction;


  if ( tmpSize > SCSMessage::SCS_MAX_MESSAGE_SIZE )
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSConnection::SendMessage - message too long",errno,1);
#endif
#ifdef SCSDEBUG
      SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) exit -1",2);
#endif
      delete[] tmpBuf;
      return -1;
    };



  sigPipeAction.sa_handler=SIG_IGN;

  // Block SIGPIPE so we get error status on write if socket is broken
  
  sigaction(SIGPIPE,&sigPipeAction,&oldSigPipeAction);



  //  memset(tmpBuf,0,SCSMessage::SCS_MESSAGE_HEADER_SIZE);
  //  memcpy(tmpBuf,&tmpSize,sizeof(long));

   if ( PackMsgSize(tmpBuf,SCSMessage::SCS_MESSAGE_HEADER_SIZE,tmpSize ) <=0 ) 
     {
#ifdef SCSDEBUG
       SCSLog::Write("ERROR : SCSConnection::PackMsgSize - failed",errno,1);
#endif
       delete[] tmpBuf;
       writeReadyFlag=false;
       sigaction(SIGPIPE,&oldSigPipeAction,0);
       
#ifdef SCSDEBUG
       SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) exit -1",2);
#endif
       return -1;  
     };


  // Write message size

  while(doneSize != SCSMessage::SCS_MESSAGE_HEADER_SIZE)
    {
      tmpSize=write(connectionSocket,tmpBuf,SCSMessage::SCS_MESSAGE_HEADER_SIZE-doneSize);
      if ( tmpSize <=0 )
	{
#ifdef SCSDEBUG
	  if( tmpSize < 0 )
	    SCSLog::Write("ERROR : SCSConnection::SendMessage - write failed",errno,1);
#endif
	  delete[] tmpBuf;
	  writeReadyFlag=false;
	  sigaction(SIGPIPE,&oldSigPipeAction,0);

#ifdef SCSDEBUG
	  SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) exit -1",2);
#endif
	  return -1;
	};
      doneSize+=tmpSize;
    };
  
  delete[] tmpBuf;

  doneSize=tmpSize=0;
  

  // Write data

  while(doneSize != tmpSize1)
    {
      tmpSize=write(connectionSocket,inMessage.data+doneSize,tmpSize1-doneSize);
      if ( tmpSize <=0 ) 
	{
#ifdef SCSDEBUG
	  if( tmpSize < 0 )
	    SCSLog::Write("ERROR : SCSConnection::SendMessage - write failed",errno,1);
#endif
	  writeReadyFlag=false;
	  sigaction(SIGPIPE,&oldSigPipeAction,0);
#ifdef SCSDEBUG
	  SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) exit -1",2);
#endif
	  return -1;
	};

      doneSize+=tmpSize;
    };

  writeReadyFlag=false;
  sigaction(SIGPIPE,&oldSigPipeAction,0);

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::SendMessage(const SCSMessage& inMessage) exit 1",2);
#endif
  return 1;

};



//
// Pre: Socket is read ready
//
//
// Post: Message is received through read ready socket
//       Returns 1 on success , -1 otherwise
//
// Desc:
//


int SCSConnection::ReceiveMessage(SCSMessage& outMessage)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::ReceiveMessage(const SCSMessage& inMessage) enter",2);
#endif

  long doneSize=0;
  long tmpSize=0;
  char* tmpBuf= new char[SCSMessage::SCS_MESSAGE_HEADER_SIZE];
  struct sigaction sigPipeAction;
  struct sigaction oldSigPipeAction;

  sigPipeAction.sa_handler=SIG_IGN;


  // Block SIGPIPE so we receive error status on read if socket is broken

  sigaction(SIGPIPE,&sigPipeAction,&oldSigPipeAction);

  outMessage.Clear();

  memset(tmpBuf,0,SCSMessage::SCS_MESSAGE_HEADER_SIZE);


  // Read size of message

  while(doneSize != SCSMessage::SCS_MESSAGE_HEADER_SIZE)
    {
      tmpSize=read(connectionSocket,tmpBuf+doneSize, SCSMessage::SCS_MESSAGE_HEADER_SIZE-doneSize);
      if ( tmpSize <=0 )
	{
#ifdef SCSDEBUG
	  if( tmpSize < 0 )
	    SCSLog::Write("ERROR : SCSConnection::ReceiveMessage - read failed",errno,1);
#endif
	  delete[] tmpBuf; 
	  readReadyFlag=false;
	  sigaction(SIGPIPE,&oldSigPipeAction,0);
#ifdef SCSDEBUG
	  SCSLog::Write("SCSConnection::ReceiveMessage(const SCSMessage& inMessage) exit -1",2);
#endif
	  return -1;
	};
      doneSize+=tmpSize;
    };
  
  
  if ( UnpackMsgSize(tmpBuf,SCSMessage::SCS_MESSAGE_HEADER_SIZE,&outMessage.size ) <=0 ) 
    {
#ifdef SCSDEBUG
      if( tmpSize < 0 )
	SCSLog::Write("ERROR : SCSConnection::UnpackMsgSize -  failed",errno,1);
#endif
      readReadyFlag=false;
      sigaction(SIGPIPE,&oldSigPipeAction,0);
      delete[] tmpBuf;

#ifdef SCSDEBUG
      SCSLog::Write("SCSConnection::ReceiveMessage(const SCSMessage& inMessage) exit -1",2);
#endif

      return -1;
    };
  
  /*
  if ( sizeof(long) < SCSMessage::SCS_MESSAGE_HEADER_SIZE )
    tmpSize= sizeof(long);
  else
    tmpSize=SCSMessage::SCS_MESSAGE_HEADER_SIZE;

  outMessage.size=*(long*)tmpBuf;
  //  memcpy(&outMessage.size,tmpBuf,tmpSize);
  */

  delete[] tmpBuf;

  if (outMessage.data)
     delete[] outMessage.data;
  outMessage.data = new char[outMessage.size];


  doneSize=tmpSize=0;
  

  // Read data

  while(doneSize != outMessage.size)
    {
      tmpSize=read(connectionSocket,outMessage.data+doneSize,outMessage.size-doneSize);
      if ( tmpSize <=0 )
	{
#ifdef SCSDEBUG
	  if( tmpSize < 0 )
	    SCSLog::Write("ERROR : SCSConnection::ReceiveMessage - read failed",errno,1);
#endif
	  readReadyFlag=false;
	  sigaction(SIGPIPE,&oldSigPipeAction,0);
#ifdef SCSDEBUG
	  SCSLog::Write("SCSConnection::ReceiveMessage(const SCSMessage& inMessage) exit -1",2);
#endif
	  return -1;
	};

      doneSize+=tmpSize;
    };

  outMessage.size--;
  
  readReadyFlag=false;
  sigaction(SIGPIPE,&oldSigPipeAction,0);

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::ReceiveMessage(const SCSMessage& inMessage) exit 1",2);
#endif
  
  return 1;

};


//
// Pre: 
//
//
// Post: Check if connection is read ready within specified timeout.
//
//
// Desc:
//

int SCSConnection::IsReadReady(long inTimeoutUs)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::IsReadReady(long inTimeoutUs) enter",2);
#endif


  if ( connectionSocket <0 )
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSConnection::IsReadReady failed - connection closed",1);
#endif 


#ifdef SCSDEBUG
      SCSLog::Write("SCSConnection::IsReadReady(long inTimeoutUs) exit -1",2);
#endif
    return -1;
    };

  if ( readReadyFlag)
    return 1;

  fd_set rmask;
  struct timeval timeout;
  int stat=0;

  FD_ZERO(&rmask);
  FD_SET(connectionSocket,&rmask);
  
  if ( inTimeoutUs<0)
    {
      stat=select(connectionSocket+1,&rmask,(fd_set*)0,(fd_set*)0,0);
    }
  else
    {
      timeout.tv_sec=inTimeoutUs/1000000;
      timeout.tv_usec=inTimeoutUs-timeout.tv_sec*1000000;
      
      stat=select(connectionSocket+1,&rmask,(fd_set*)0,(fd_set*)0,&timeout);

    };
  
#ifdef SCSDEBUG
  if ( stat < 0 )
    SCSLog::Write("ERROR : SCSConnection::IsReadReady select failed",errno,1);
#endif 
  if ( stat > 0 )
    readReadyFlag=true;
  
#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSConnection::IsReadReady(long inTimeoutUs) exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif
  return stat;
};


//
// Pre:
//
//
// Post: Check if connection is read ready within specified timeout.
//
//
// Desc:
//


int SCSConnection::IsWriteReady(long inTimeoutUs)
{ 

#ifdef SCSDEBUG
  SCSLog::Write("SCSConnection::IsWriteReady(long inTimeoutUs) enter",2);
#endif

  if ( connectionSocket <0 )
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSConnection::IsWriteReady failed - connection closed",1);
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSConnection::IsWriteReady(long inTimeoutUs) exit -1",2);
#endif
      return -1;
    };


  if ( writeReadyFlag)
    {
#ifdef SCSDEBUG
      SCSLog::Write("SCSConnection::IsWriteReady(long inTimeoutUs) exit 1",2);
#endif
      return 1;
    };

  fd_set wmask;
  struct timeval timeout;
  int stat=0;


  FD_ZERO(&wmask);
  FD_SET(connectionSocket,&wmask);
  
  if ( inTimeoutUs<0)
    {
      stat=select(connectionSocket+1,(fd_set*)0,&wmask,(fd_set*)0,0);
    }
  else
    {
      timeout.tv_sec=inTimeoutUs/1000000;
      timeout.tv_usec=inTimeoutUs-timeout.tv_sec*1000000;
      
      stat=select(connectionSocket+1,(fd_set*)0,&wmask,(fd_set*)0,&timeout);
    };
  
#ifdef SCSDEBUG
  if ( stat < 0 )
    SCSLog::Write("ERROR : SCSConnection::IsReadReady select failed",errno,1);
#endif 

  if ( stat > 0 )
    writeReadyFlag=true;

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSConnection::IsWriteReady(long inTimeoutUs) exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif
  
  return stat;
};


  
//
// Pre: 
//
//
// Post: Return 1 if host is little endian,
//       0 otherwise
//
// Desc:
//

int SCSConnection::Endian()
{
  short int val = 0x0001;
  char *bptr = (char *) &val;
  
  if ( bptr[0] == 1 )
    {
      return 1;
    };

  return 0;
  
};



//
// Pre: 
//
//
// Post: 
//
//
// Desc:
//

int  SCSConnection::PackMsgSize(char* outBuffer,int inBufferSize,long inMessageSize)
{

  if ( (! outBuffer) || (!inMessageSize) || (inBufferSize < 0 ))
    {
      return -1;
    };

  char* bptr= (char*)&inMessageSize;


  memset(outBuffer,0,inBufferSize);

  if ( ! Endian() )
    {
      int sindx=sizeof(long)-1;

      for(int indx=0;indx<inBufferSize;indx++)
	{
	  outBuffer[indx]=bptr[sindx-indx];
	};
    }
  else
    {
      for(int indx=0;indx<inBufferSize;indx++)
	{
	  outBuffer[indx]=bptr[indx];
	}; 
    };






  return 1;
};



//
// Pre: 
//
//
// Post: 
//
//
// Desc:
//

int  SCSConnection::UnpackMsgSize(char* inBuffer,int inBufferSize,long* outMessageSize)
{
  if ( (! inBuffer) || (! outMessageSize) || (inBufferSize < 0))
    {
      return -1;
    };


  char* bptr= (char*)outMessageSize;

  (*outMessageSize)=0;

  if ( ! Endian() )
    {
      int sindx=sizeof(long)-1;


      for(int indx=0;indx<inBufferSize;indx++)
	{
	  bptr[sindx-indx]=inBuffer[indx];
	};
    }
  else
    {
      for(int indx=0;indx<inBufferSize;indx++)
	{
	  bptr[indx]=inBuffer[indx];
	};
    };





  return 1;

};





int  SCSConnection::Test()
{

  SCSConnection tmpConnection(10);
  int stat =1;
  SCSMessage tmpMessage;






  cout <<" Testing SCSConnection::PackMsgSize(char* outBuffer,unsigned int inBufferSize,long inMessageSize)...";
  
  char* tmpBuf0 = new char[SCSMessage::SCS_MESSAGE_HEADER_SIZE];
  long tmpSize=7;

  
  if ( tmpConnection.PackMsgSize(tmpBuf0,SCSMessage::SCS_MESSAGE_HEADER_SIZE,tmpSize) <=0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else 
    {

      short int tmpv = tmpBuf0[0];
      if ( tmpv  != 7)
	{
	  stat = 0;
	  cout<<"FAILED"; 
	}
      else
	{
	  cout<<"OK";
	};

    }
  delete[] tmpBuf0;
  cout<<endl;
  
  




  cout <<" Testing SCSConnection::UnpackMsgSize(char* inBuffer,int inBufferSize,long* outMessageSize)...";


  char* tmpBuf1= new char[SCSMessage::SCS_MESSAGE_HEADER_SIZE];

  long tmpSize1=3;


  tmpConnection.PackMsgSize(tmpBuf1,SCSMessage::SCS_MESSAGE_HEADER_SIZE,tmpSize1);

  tmpSize=0;

  
  if ( tmpConnection.UnpackMsgSize(tmpBuf1,SCSMessage::SCS_MESSAGE_HEADER_SIZE,&tmpSize1) <=0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else 
    {

      if ( tmpSize1 != 3 )
	{
	  stat = 0;
	  cout<<"FAILED"; 
	}
      else
	{
	  cout<<"OK";
	}
    };
  
  delete[] tmpBuf1;
  cout<<endl;
  








  cout <<" Testing SCSConnection::SendMessage(long inTimeoutUs,const SCSMessage& inMessage)...";
  
  if ( tmpConnection.SendMessage(0,tmpMessage) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;
  
  
  
  
  cout <<" Testing SCSConnection::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage)...";
  
  if ( tmpConnection.ReceiveMessage(0,tmpMessage) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;



  cout <<" Testing SCSConnection::IsReadReady(long inTimeoutUs)...";
  
  if ( tmpConnection.IsReadReady(0) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;

  

  cout <<" Testing SCSConnection::IsWriteReady(long inTimeoutUs)...";
  
  if ( tmpConnection.IsWriteReady(0) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;


  
  cout <<" Testing SCSConnection::SendMessage(const SCSMessage& inMessage)...";
  
  if ( tmpConnection.SendMessage(tmpMessage) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;
  
  
  
  
  cout <<" Testing SCSConnection::ReceiveMessage(SCSMessage& outMessage)...";
  
  if ( tmpConnection.ReceiveMessage(tmpMessage) >0 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;



  cout <<" Testing SCSConnection::GetSocketId()...";
  
  if ( tmpConnection.GetSocketId() !=10 )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;



  cout <<" Testing SCSConnection::SetReadReady(bool inFlag)...";
  
  tmpConnection.SetReadReady(true);

  if (! tmpConnection.readReadyFlag )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;


  
  cout <<" Testing SCSConnection::SetWriteReady(bool inFlag)...";
  
  tmpConnection.SetWriteReady(true);

  if (! tmpConnection.writeReadyFlag )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;


  cout <<" Testing SCSConnection::CloseProtocol()...";
  
  if ( !tmpConnection.CloseProtocol() )
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;



  cout <<" Testing SCSConnection::Close()...";
  
  tmpConnection.Close();
  if (( tmpConnection.connectionSocket>=0) 
      ||(tmpConnection.readReadyFlag)
      || (tmpConnection.writeReadyFlag))
    {
      stat = 0;
      cout<<"FAILED";
    }
  else
    cout<<"OK";
  
  cout<<endl;

  return stat;

};
