//---------------------------------------------------------------------------
//
//  FILE NAME:          SCSServer.cc
//
//  DEVELOPMENT:        CXCDS, User Infrastructure
//
//  DESCRIPTION:        Simple Clinet/Server implementation
//
//
//---------------------------------------------------------------------------
// Copyright (c) 2006, Smithsonian Astrophysical Observatory
//---------------------------------------------------------------------------
 
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <iostream>
#include <sstream>
#include <errno.h>

#include "SCSMessage.hh"
#include "SCSConnection.hh"
#include "SCSServer.hh"
#include "SCSLog.hh"

// index var to  keep track of all connected clients

long SCSServer::globalConnectionId=1;

SCSServer::SCSServer()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::SCSServer() enter",2);
#endif

  serverSocket=-1;

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::SCSServer() exit",2);
#endif
};


SCSServer::~SCSServer()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::~SCSServer() enter",2);
#endif 

  Shutdown();

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::~SCSServer() exit",2);
#endif 
};
  
//
// Pre: Servr is not running
//
//
// Post: Server is started using INET sockets on inServerPort port
//       and only inConnectionsLimit number of connections
//       are allowed at any given time
//       Returns 1 on success , -1 on failure
//   
// Desc: A listenning socket is opened and connections are accepted
//       on that socket.  Once connection from client occurs all the
//       necessary info is stored in a hash table and client is refered to
//       using entry in that hash table
//
//

int SCSServer::Start(long inConnectionsLimit,int inServerPort)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) enter",2);
#endif 
  

  if ( serverSocket >=0 )
    {
#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - on port '"<<inServerPort<<"' failed - server already running";
      SCSLog::Write(tmpStr.str(),1);
#endif
      
#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit",2);
#endif 
      return -1;
    };

  connectionsLimit=inConnectionsLimit;

  // Create a socket

  if ( (serverSocket=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))<0 )
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSServer::Start - socket create failed",errno,1); 
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit -1",2);
#endif 
      return -1;
    };

  int yes=1;
  if (setsockopt(serverSocket,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) 
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSServer::Start - setsockopt failed",errno,1); 
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit -1",2);
#endif 
      return -1;

    } 
  
  struct sockaddr_in serverAddr;
  
  memset(&serverAddr,0,sizeof(serverAddr));
  serverAddr.sin_family=AF_INET;
  serverAddr.sin_addr.s_addr=htonl(INADDR_ANY);
  serverAddr.sin_port=htons(inServerPort);
  
  // bind socket

  if ( bind(serverSocket,(struct sockaddr*)&serverAddr,sizeof(serverAddr))< 0 )
    {
#ifdef SCSDEBUG
      int tmpErr=errno;
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - bind on port "<<inServerPort<<" failed";
      SCSLog::Write(tmpStr.str(),tmpErr,1);

#endif 

      close(serverSocket);
      serverSocket=-1;
#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit -1",2);
#endif 
      return -1;
    };
  
  // listen on socket

  if ( listen(serverSocket,10)< 0 )
    {
#ifdef SCSDEBUG
      int tmpErr=errno;
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - listen on socket "<<serverSocket<<" failed";
      SCSLog::Write(tmpStr.str(),tmpErr,1);

#endif 
      close(serverSocket);
      serverSocket=-1;
#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit -1",2);
#endif 
      return -1;
    }
  domainType = AF_INET;

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::Start(long inConnectionsLimit,int inServerPort) exit 1",2);
#endif 
  
  return 1;
};


//
// Pre: Server is not running
//
//
// Post: Server is started using UNIX(local) sockets using inSocketPath
//       for clients to connect to 
//       and only inConnectionsLimit number of connections
//       are allowed at any given time
//       Returns 1 on success , -1 on failure
//
//
// Desc: A listenning socket is opened and connections are accepted
//       on that socket.  Once connection from client occurs all the
//       necessary info is stored in a hash table and client is refered to
//       using entry in that hash table
//
//
//
  

int SCSServer::Start(long inConnectionsLimit,const string& inSocketPath)
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) enter",2);
#endif 

  if ( serverSocket >=0 )
    {
#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - on socket "<<inSocketPath<<" failed - server already running";
      SCSLog::Write(tmpStr.str(),1);
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) exit -1",2);
#endif 
      return -1;
    };

  connectionsLimit=inConnectionsLimit;

  unlink(inSocketPath.c_str());
  

  // Create socket

  if ( (serverSocket=socket(AF_UNIX,SOCK_STREAM,0))<0 )
    {
#ifdef SCSDEBUG
      
      SCSLog::Write("ERROR : SCSServer::Start - socket create failed",errno,1); 
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) exit -1",2);
#endif 
      

      return -1;
    };
  
  struct sockaddr_un serverAddr;
  
  memset(&serverAddr,0,sizeof(serverAddr));
  serverAddr.sun_family=AF_UNIX;
  strcpy(serverAddr.sun_path,inSocketPath.c_str());

  // bind socket

  if ( bind(serverSocket,(struct sockaddr*)&serverAddr,sizeof(serverAddr))< 0 )
    {
#ifdef SCSDEBUG
      int tmpErr=errno;
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - bind on socket "<<inSocketPath<<" failed";
      SCSLog::Write(tmpStr.str(),tmpErr,1);

#endif 
    
      unlink(socketPath.c_str());
      close(serverSocket);
      serverSocket=-1;

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) exit -1",2);
#endif 
      return -1;
    };

  // listen on the socket

  if ( listen(serverSocket,10)< 0 )
    {
#ifdef SCSDEBUG
      int tmpErr=errno;
      ostringstream tmpStr;
      tmpStr<<"ERROR : SCSServer::Start - listen on socket '"<<serverSocket<<"' failed";
      SCSLog::Write(tmpStr.str(),tmpErr,1);

#endif 
      unlink(socketPath.c_str());
      close(serverSocket);
      serverSocket=-1;

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) exit -1",2);
#endif 
      return -1;
    }
  domainType = AF_UNIX;
  socketPath=inSocketPath;
  
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::Start(long inConnectionsLimit,const string& inSocketPath) exit 1",2);
#endif 

  return 1;
};


//
// Pre: Server is running
//
//
// Post: Server is shut down and all client connections are closed
//
//
// Desc: All connected clients receive shutdown message and connections
//       are closed.
//
//
  


int SCSServer::Shutdown()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::Shutdown() enter",2);
#endif 

  if ( serverSocket>=0 )
    {
      // Close all connections
      CloseConnection();
      // Close server socket
      close(serverSocket);
      // Delete server socket
      if ( domainType == AF_UNIX)
	unlink(socketPath.c_str());
      serverSocket=-1;

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Shutdown() exit 1",2);
#endif 


      return 1;
    };
#ifdef SCSDEBUG
  SCSLog::Write("ERROR : SCSServer::Shutdown failed - server not running",1); 
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::Shutdown() exit -1",2);
#endif 
  return -1;
};


//
// Pre: Server is running
//
//
// Post: Attempt is made to read data from any available connection
//       into outMessage with inTimeoutUs useconds timeout.  On success
//       connection id is returned from which data has been read, on failure
//       0 is returned if timeout occured -1 otherwise. If new connections
//       are pending they are initialized and also considered for reading
//       If timeout is negative blocking read occures until socket is read
//       ready on any connection.
//
// Desc: Server checks all connections and reads data from first available
//       connection using round robin technique.
//
//
//
   
long SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage)
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) enter",2);
#endif 


  if ( serverSocket < 0 )
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSServer::ReceiveMessage failed - server not running",1); 
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) exit -1",2);
#endif 

      return -1;
    };
  

  int stat;


  while(1)
    {

      // Check for connection ready for reading
      // If any available exit loop

      if (! GetReadReadyCount())
	{
	  // Check if any connection is ready for reading
	  // or new connection arrived

	  stat=CheckRead(inTimeoutUs);
	  if ( stat <=0)
	    {
#ifdef SCSDEBUG
	      ostringstream tmpStr;
	      tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) exit "<<stat;
	      SCSLog::Write(tmpStr.str(),2);
#endif 
	      return stat;
	    };

	  // New connection arrived
	  if (newConnectionReady)
	    {

	      // Add it 

	      stat=AddNewConnection();
	      if ( stat <=0)
		{
#ifdef SCSDEBUG
		  ostringstream tmpStr;
		  tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) exit "<<stat;
		  SCSLog::Write(tmpStr.str(),2);
#endif 
		  return stat;
		};
	    }
	  else
	    break;
	}
      else
	break;

    }


  // Process connections

  long tmpId = PickReadConnection();
 

  if (tmpId <= 0 )
    {
#ifdef SCSDEBUG
       ostringstream tmpStr;
       tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) exit "<<tmpId;
      SCSLog::Write(tmpStr.str(),2);
#endif 
      return tmpId;
    };
  stat=connections[tmpId]->ReceiveMessage(outMessage);

  // Check if message was result of closure by client

  if ( stat<=0)
    {
      outMessage.Set(0,0,SCSMessage::PROTOCOL_CLOSE);
    };

  if (outMessage.Type() == SCSMessage::PROTOCOL_CLOSE)
    {
#ifdef SCSDEBUG 
      ostringstream tmpStr;
      tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) closing connection "<<tmpId;
      SCSLog::Write(tmpStr.str(),2);
#endif 
      connections[tmpId]->Close();
      delete connections[tmpId];
      connections.erase(tmpId);
    };

#ifdef SCSDEBUG 
  ostringstream tmpStr;
  tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage) exit "<<tmpId;
  SCSLog::Write(tmpStr.str(),2);
#endif 
  
  return tmpId;
 
};

//
// Pre: Server is running
//
//
// Post: Attempt is made to read data from inConnectionId connectionn
//       into outMessage with inTimeoutUs useconds timeout.  On success
//       connection id is returned from which data has been read, on failure
//       0 is returned if timeout occured -1 otherwise. 
//       If timeout is negative blocking read occures until socket is read
//       ready.
//
//
// Desc: Only specified connection is considered for reading
//
//
//
  
int SCSServer::ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage)
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage) enter",2);
#endif 
  
  map<long,SCSConnection*>::iterator tmpIter=connections.find(inConnectionId);
  
  if ( tmpIter != connections.end())
    {
      int stat =tmpIter->second->ReceiveMessage(inTimeoutUs,outMessage);
#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"SCSServer::ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage) exit "<<stat;
      SCSLog::Write(tmpStr.str(),2);
#endif 
      return stat;
    };

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"ERROR : SCSServer::ReceiveMessage failed - connection "<<inConnectionId<<" not present";
  SCSLog::Write(tmpStr.str(),1);
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage) exit -1",2);
#endif 

  return -1;

};


//
// Pre: Server is running
//
//
// Post: Attempt is made to send data in inMessage to inConnectionId 
//       connectionn with inTimeoutUs useconds timeout.  On success
//       connection id is returned from which data has been read, on failure
//       0 is returned if timeout occured -1 otherwise. 
//       If timeout is negative blocking read occures until socket is write
//       ready.
//
//
// Desc: Data in inMessage is send to connection inConnectionId
//
//
//
  


int SCSServer::SendMessage(long inTimeoutUs,long inConnectionId,const SCSMessage& inMessage)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::SendMessage(long inTimeoutUs,long inConnectionId,const SCSMessage& inMessage) enter",2);
#endif 

  map<long,SCSConnection*>::iterator tmpIter=connections.find(inConnectionId);
  
  if ( tmpIter != connections.end())
    {
      int stat =tmpIter->second->SendMessage(inTimeoutUs,inMessage);
#ifdef SCSDEBUG
        ostringstream tmpStr;
	tmpStr<<"SCSServer::SendMessage(long inTimeoutUs,long inConnectionId,const SCSMessage& inMessage) exit "<<stat;
	SCSLog::Write(tmpStr.str(),2);
#endif 
      return stat;
    };

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"ERROR : SCSServer::SendMessage failed - connection "<<inConnectionId<<" not present";
  SCSLog::Write(tmpStr.str(),1);
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::SendMessage(long inTimeoutUs,long inConnectionId,const SCSMessage& inMessage) exit -1",2);
#endif 

  return -1;
};


//
// Pre: Server is running
//
//
// Post: Connection inConnectionId is closed. Client receives message
//       informing about connection closure. Returns 1 on success, -1
//       on failure.
//
// Desc: Connection inConnectionId is deleted from the list of 
//       connected clients
//
//
//
  


int SCSServer::CloseConnection(long inConnectionId)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CloseConnection(long inConnectionId) enter",2);
#endif 
  map<long,SCSConnection*>::iterator tmpIter=connections.find(inConnectionId);
  
  if ( tmpIter != connections.end())
    {
      delete tmpIter->second;
      connections.erase(inConnectionId);
#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::CloseConnection(long inConnectionId) exit 1",2);
#endif 
      return 1;
    };

#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"ERROR : SCSServer::CloseConnection failed - connection "<<inConnectionId<<" not present";
    SCSLog::Write(tmpStr.str(),1);
#endif 

#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::CloseConnection(long inConnectionId) exit -1",2);
#endif 

  return -1; 
};


//
// Pre: Server is running
//
//
// Post: All connections are closed. Returns 1 on success, -1 on failure.
//
//
// Desc: All connections are deleted 
//
//
//
  


int SCSServer::CloseConnection()
{

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CloseConnection() enter",2);
#endif 
  
  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      delete tmpIter->second;
    };

  connections.clear();

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CloseConnection() exit 1",2);
#endif 
  return 1;
};



//
// Pre: Server is running
//
//
// Post: Puts ids of all open connections into outConnectionIds list
//
//
// Desc: Intarnal ids assigned to each connection are put into 
//       outConnecitonIds.  These ids can be used to select particular
//       connections for reading/writing of data.
//
//
//


void SCSServer::GetConnectionIds(list<long>& outConnectionIds)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::GetConnectionIds(list<long>& outConnectionIds) enter",2);
#endif 

  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      outConnectionIds.push_back(tmpIter->first);
    };

 #ifdef SCSDEBUG
  SCSLog::Write("SCSServer::GetConnectionIds(list<long>& outConnectionIds) exit",2);
#endif  
};



//
// Pre: Server is running, new connection is available on server socket
//
//
// Post: New connection is created and added to hash table. Returns 1 
//       on success, -1 on failure.
//
//
// Desc: Depending on how server was started a new connection is opened
//       and initialized.
//
//
//


int SCSServer::AddNewConnection()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::AddNewConnection() enter",2);
#endif 

  int newsocket;

  // accept connections from server socket

  if ( domainType == AF_INET )
    {
      struct sockaddr_in remoteSocket;
      socklen_t addrLen=sizeof(remoteSocket);

      newsocket=accept(serverSocket,(struct sockaddr*)&remoteSocket,&addrLen);
      if (newsocket < 0 )
	{
#ifdef SCSDEBUG
	  SCSLog::Write("ERROR : SCSServer::AddNewConnection - accept failed",errno,1);
#endif 

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  return -1;
	};
    }
  else if ( domainType == AF_UNIX)
    {
      struct sockaddr_un remoteSocket;
      socklen_t addrLen=sizeof(remoteSocket);
      struct stat statInfo;
      time_t staleTime;

      newsocket=accept(serverSocket,(struct sockaddr*)&remoteSocket,&addrLen);
      if (newsocket < 0 )
	{
#ifdef SCSDEBUG
	  SCSLog::Write("ERROR : SCSServer::AddNewConnection - accept failed",errno,1);
#endif 

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  
	  return -1;
	};  
      addrLen -=sizeof(remoteSocket.sun_family);

#ifdef __APPLE__
        addrLen -=sizeof(remoteSocket.sun_len);
#endif

      remoteSocket.sun_path[addrLen]=0;

      if ( stat(remoteSocket.sun_path,&statInfo) < 0 )
	{
#ifdef SCSDEBUG
	  int tmpErr=errno;
	  ostringstream tmpStr;
	  tmpStr<<"ERROR : SCSServer::AddNewConnection - stat failed on socket '"<<remoteSocket.sun_path<<"'";
	  SCSLog::Write(tmpStr.str(),tmpErr,1);
#endif 
	  close(newsocket);

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  return -1;
	};

      if ( (statInfo.st_mode & (S_IRWXG | S_IRWXO)) ||
	   (statInfo.st_mode & S_IRWXU) != S_IRWXU)
	{
#ifdef SCSDEBUG
	  ostringstream tmpStr;
	  tmpStr<<"ERROR : SCSServer::AddNewConnection - wrong permissions on socket '"<<remoteSocket.sun_path<<"'";
	  SCSLog::Write(tmpStr.str(),1);
#endif 
	  close(newsocket);

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  
	  return -1;
	};

      staleTime = time(0)-30;
      if ( statInfo.st_atime< staleTime ||
	   statInfo.st_ctime < staleTime ||
	   statInfo.st_mtime < staleTime )
	{
#ifdef SCSDEBUG
	  ostringstream tmpStr;
	  tmpStr<<"ERROR : SCSServer::AddNewConnection - stale socket '"<<remoteSocket.sun_path<<"'";
	  SCSLog::Write(tmpStr.str(),1);
#endif 
	  close(newsocket);

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  
	  return -1;
	};
      
      unlink(remoteSocket.sun_path);
      
    }
  else
    {
#ifdef SCSDEBUG
      SCSLog::Write("ERROR : SCSServer::AddNewConnection - failed - unknown domain type",1);
#endif
 
#ifdef SCSDEBUG
      SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
      

      return -1;
    };
      ;

  // put new connection into hash table


  if ( domainType == AF_UNIX)
    {
      SCSMessage tmpMessage;
      SCSConnection *tmpConnection = new SCSConnection(newsocket);


      tmpMessage.Set(0,0,SCSMessage::PROTOCOL_ACK);
      if ( tmpConnection->SendMessage(1000000,tmpMessage)<=0 )
	{
#ifdef SCSDEBUG
	  SCSLog::Write("ERROR : SCSServer::AddNewConnection - ACK send to client failed",1);
#endif 
	  delete tmpConnection;

#ifdef SCSDEBUG
	  SCSLog::Write("SCSServer::AddNewConnection() exit -1",2);
#endif 
	  

	  return -1;
	}
      else
	connections[globalConnectionId]=tmpConnection;

    }
  else
    {
      connections[globalConnectionId]=new SCSConnection(newsocket);      
    };

  // increase id counter

  globalConnectionId++;

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::AddNewConnection() exit 1",2);
#endif 

  return 1;
};


//
// Pre: Server is running
//
//
// Post: All the connections and receiving socket are checked if they 
//       are ready to be read.  Appropriate data members are set to indicate
//       this status. On success 1 is returned , 0 on timeout
//       (no connection was ready within inTimeoutUs useconds), -1 on error
//
// Desc: All connections are checked for read status as well as server 
//       receiving socket which receives new connections.  
//
//
//




int SCSServer::CheckRead(long inTimeoutUs)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CheckRead(long inTimeoutUs) enter",2);
#endif 

  fd_set rmask;
  struct timeval timeout;
  int stat;
  int maxSock=0;

  FD_ZERO(&rmask);

  
  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      FD_SET(tmpIter->second->GetSocketId(),&rmask);
      if ( tmpIter->second->GetSocketId() > maxSock)
	maxSock=tmpIter->second->GetSocketId();
    };
  
  if ( connections.size() < connectionsLimit)
    {
      FD_SET(serverSocket,&rmask);
      if ( serverSocket > maxSock)
	maxSock=serverSocket;
    };

  

  if ( inTimeoutUs<0)
    {
      stat =  select(maxSock+1,&rmask,(fd_set*)0,(fd_set*)0,0);
    }
  else
    {
      
      timeout.tv_sec=inTimeoutUs/1000000;
      timeout.tv_usec=inTimeoutUs-timeout.tv_sec*1000000;
      
      stat = select(maxSock+1,&rmask,(fd_set*)0,(fd_set*)0,&timeout);
    }


  if ( stat <=0 )
    {
#ifdef SCSDEBUG
      if ( stat < 0 )
	SCSLog::Write("ERROR : SCSServer::CheckRead select failed",errno,1);
#endif 

#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"SCSServer::CheckRead(long inTimeoutUs) exit "<<stat;
      SCSLog::Write(tmpStr.str(),2);
#endif  
      return stat;
    };


  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      if ( FD_ISSET(tmpIter->second->GetSocketId(),&rmask))
	tmpIter->second->SetReadReady(true);
      else
	tmpIter->second->SetReadReady(false);

    };

  newConnectionReady=false;
  if ( connections.size() < connectionsLimit)
    {
      if ( FD_ISSET(serverSocket,&rmask))
	newConnectionReady=true;
    };
  
#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSServer::CheckRead(long inTimeoutUs) exit "<<stat;
  SCSLog::Write(tmpStr.str(),2);
#endif 

  return stat;
  
};

//
// Pre: Server is running
//
//
// Post: All the connections are checked if they 
//       are ready to be written to.  Appropriate data members 
//       are set to indicate
//       this status. On success 1 is returned , 0 on timeout
//       (no connection was ready within inTimeoutUs useconds), -1 on error
//
// Desc: All connections are checked for write status.  Server 
//       receiving socket which receives new connections is not checked.  
//
//



int SCSServer::CheckWrite(long inTimeoutUs)
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CheckWrite(long inTimeoutUs) enter",2);
#endif 

  
  fd_set wmask;
  struct timeval timeout;
  int stat;
  int maxSock=0;


    FD_ZERO(&wmask);
  
  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
        FD_SET(tmpIter->second->GetSocketId(),&wmask);
      if ( tmpIter->second->GetSocketId() > maxSock)
	maxSock=tmpIter->second->GetSocketId();
    };
  
  if ( inTimeoutUs<0)
    {
      stat =  select(maxSock+1,(fd_set*)0,&wmask,(fd_set*)0,0);
    }
  else
    {
      
      timeout.tv_sec=inTimeoutUs/1000000;
      timeout.tv_usec=inTimeoutUs-timeout.tv_sec*1000000;
      
      stat = select(maxSock+1,(fd_set*)0,&wmask,(fd_set*)0,&timeout);
    }


  if ( stat <=0 )
    {
#ifdef SCSDEBUG
      if ( stat < 0 )
	SCSLog::Write("ERROR : SCSServer::CheckRead select failed",errno,1);
#endif 

#ifdef SCSDEBUG
      ostringstream tmpStr;
      tmpStr<<"SCSServer::CheckWrite(long inTimeoutUs) exit "<<stat;
      SCSLog::Write(tmpStr.str(),2);
#endif 
      return stat;
    };

  
  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      if ( FD_ISSET(tmpIter->second->GetSocketId(),&wmask))
	tmpIter->second->SetWriteReady(true);
      else
	tmpIter->second->SetWriteReady(false);


    };

#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::CheckWrite(long inTimeoutUs) exit 1",2);
#endif 
  return 1;
  
};


//
// Pre: Server is running
//
//
// Post: Number of connections that are read ready is returned
//
//
// Desc:
//
//
//


long SCSServer::GetReadReadyCount()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::GetReadReadyCount() enter",2);
#endif 

  long tmpCount=0;
  

  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
	if (tmpIter->second->readReadyFlag)
	  tmpCount++;
    };
  
#ifdef SCSDEBUG
  ostringstream tmpStr;
  tmpStr<<"SCSServer::GetReadReadyCount() exit "<<tmpCount;
  SCSLog::Write(tmpStr.str(),2);
#endif 

  return tmpCount;

};


//
// Pre: Server is running
//
//
// Post: First available for reading connection is picked
//       Returns connectionId or 0 if no connections are available
//
// Desc:
//
//
//


long SCSServer::PickReadConnection()
{
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::PickReadConnection() enter",2);
#endif 


  for( map<long,SCSConnection*>::iterator tmpIter=connections.begin();tmpIter!=connections.end();tmpIter++)
    {
      if (tmpIter->second->readReadyFlag )
	{
#ifdef SCSDEBUG
	  ostringstream tmpStr;
	  tmpStr<<"SCSServer::PickReadConnection() exit "<<tmpIter->first;
	  SCSLog::Write(tmpStr.str(),2);
#endif 
	  return tmpIter->first;
	};
    };
  
#ifdef SCSDEBUG
  SCSLog::Write("SCSServer::PickReadConnection() exit 0",2);
#endif 
  return 0;

};



int SCSServer::Test()
{

  SCSServer tmpServer;
  int status =1;
  ostringstream tmpOStr;
  tmpOStr<<"/tmp/testserver"<<getpid();
  cout <<" Testing SCSServer::Start(long inConnectionsLimit,int inServerPort)...";
  
  if ( tmpServer.Start(0,10000) <=0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      tmpServer.Shutdown();
      cout<<"OK";
    };

  cout <<endl;



  cout <<" Testing SCSServer::Start(long inConnectionsLimit,const string& inSocketPath)...";
  
  if ( tmpServer.Start(0,tmpOStr.str()) <=0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      tmpServer.Shutdown();
      cout<<"OK";
    };

  cout <<endl;


  cout <<" Testing SCSServer::Shutdown()...";
  
  if ( tmpServer.Shutdown() >0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      if ( tmpServer.serverSocket != -1) 
	{
	  status = 0;
	  cout<<"FAILED";
	}
      else
	{
	  struct stat statInfo;
	  if ( stat(tmpOStr.str().c_str(),&statInfo) > 0 )
	    {
	      status = 0;
	      cout<<"FAILED"; 
	    }
	  else
	    cout<<"OK";
	};
    };

  cout <<endl;


  
  cout <<" Testing SCSServer::ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage)...";
  
  SCSMessage tmpMessage;
 if ( tmpServer.Start(0,tmpOStr.str()) <=0 )
    {
      status = 0;
      cout<<"FAILED";
    }
 else  if ( tmpServer.ReceiveMessage(0,tmpMessage) > 0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };

  cout <<endl;

  

  cout <<" Testing SCSServer::ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage)...";
  

  if ( tmpServer.ReceiveMessage(0,1,tmpMessage) > 0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };

  cout <<endl;



  cout <<" Testing SCSServer::SendMessage(long inTimeoutUs,long inConnectionId,SCSMessage& inMessage)...";
  

  if ( tmpServer.SendMessage(0,1,tmpMessage) > 0 )
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };

  cout <<endl;


  cout <<" Testing SCSServer::CloseConnection(long inConnectionId)...";
    
    
  if ( tmpServer.CloseConnection(1)>0)
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };

  cout <<endl;


  cout <<" Testing SCSServer::CloseConnection()...";
    
    
  if ( tmpServer.CloseConnection() != 1)
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };

  cout <<endl;




  cout <<" Testing SCSServer::GetConnectionIds(list<long>& outConnectionIds)...";
  
  list<long> tmpList;
  tmpServer.GetConnectionIds(tmpList);
  if ( tmpList.size())
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;
  


  cout <<" Testing SCSServer::GetConnectionsCount()...";
  
  if (tmpServer.GetConnectionsCount())
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;
  

  cout <<" Testing SCSServer::AddNewConnection()...";
  tmpServer.Shutdown()  ;
  if (tmpServer.AddNewConnection()>0)
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;


  cout <<" Testing SCSServer::CheckRead(long inTimeoutUs)...";
  
  if (tmpServer.CheckRead(0)>0)
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;


  cout <<" Testing SCSServer::CheckWrite(long inTimeoutUs)...";
  
  if (tmpServer.CheckWrite(0)>0)
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;



  cout <<" Testing SCSServer::GetReadReadyCount()...";
  
  if (tmpServer.GetReadReadyCount())
    {
      status = 0;
      cout<<"FAILED";
    }
  else
    {
      cout<<"OK";
    };
  
  cout <<endl;


  return status;
  
};
