//---------------------------------------------------------------------------
//
//  FILE NAME:          SCSServer.hh
//
//  DEVELOPMENT:        CXCDS, User Infrastructure
//
//  DESCRIPTION:
//
//
//---------------------------------------------------------------------------
// Copyright (c) 2006, Smithsonian Astrophysical Observatory
//---------------------------------------------------------------------------
 
#ifndef _SCSSERVER_HH_
#define _SCSSERVER_HH_


#include <map>
#include <list>
#include <sys/socket.h>

using namespace std;

class SCSMessage;
class SCSConnection;

class SCSServer
{

public:
  SCSServer();
  virtual ~SCSServer();
 
  //
  // Pre: Servr is not running
  //
  //
  // Post: Server is started using INET sockets on inServerPort port
  //       and only inConnectionsLimit number of connections
  //       are allowed at any given time
  //       Returns 1 on success , -1 on failure
  //   
  // Desc: A listenning socket is opened and connections are accepted
  //       on that socket.  Once connection from client occurs all the
  //       necessary info is stored in a hash table and client is refered to
  //       using entry in that hash table
  //
  //
  
  int Start(long inConnectionsLimit,int inServerPort);

  //
  // Pre: Server is not running
  //
  //
  // Post: Server is started using UNIX(local) sockets using inSocketPath
  //       for clients to connect to 
  //       and only inConnectionsLimit number of connections
  //       are allowed at any given time
  //       Returns 1 on success , -1 on failure
  //
  //
  // Desc: A listenning socket is opened and connections are accepted
  //       on that socket.  Once connection from client occurs all the
  //       necessary info is stored in a hash table and client is refered to
  //       using entry in that hash table
  //
  //
  //
  

  int Start(long inConnectionsLimit,const string& inSocketPath);

  //
  // Pre: Server is running
  //
  //
  // Post: Server is shut down and all client connections are closed
  //
  //
  // Desc: All connected clients receive shutdown message and connections
  //       are closed.
  //
  //
  

  int Shutdown();

  //
  // Pre: Server is running
  //
  //
  // Post: Attempt is made to read data from any available connection
  //       into outMessage with inTimeoutUs useconds timeout.  On success
  //       connection id is returned from which data has been read, on failure
  //       0 is returned if timeout occured -1 otherwise. If new connections
  //       are pending they are initialized and also considered for reading
  //       If timeout is negative blocking read occures until socket is read 
  //       ready on any connection.      
  //
  // Desc: Server checks all connections and reads data from first available
  //       connection using round robin technique.
  //
  //
  //
  


  long ReceiveMessage(long inTimeoutUs,SCSMessage& outMessage);


  //
  // Pre: Server is running
  //
  //
  // Post: Attempt is made to read data from inConnectionId connectionn
  //       into outMessage with inTimeoutUs useconds timeout.  On success
  //       connection id is returned from which data has been read, on failure
  //       0 is returned if timeout occured -1 otherwise. 
  //       If timeout is negative blocking read occures until data arrives.
  //      
  //
  //
  // Desc: Only specified connection is considered for reading
  //
  //
  //
  

  int ReceiveMessage(long inTimeoutUs,long inConnectionId,SCSMessage& outMessage);

  //
  // Pre: Server is running
  //
  //
  // Post: Attempt is made to send data in inMessage to inConnectionId 
  //       connectionn with inTimeoutUs useconds timeout.  On success
  //       connection id is returned from which data has been read, on failure
  //       0 is returned if timeout occured -1 otherwise. 
  //       If timeout is negative blocking write occures until socket is write
  //       ready.
  //      
  //
  //
  // Desc: Data in inMessage is send to connection inConnectionId
  //
  //
  //
  

  int SendMessage(long inTimeoutUs,long inConnectionId,const SCSMessage& inMessage);


  //
  // Pre: Server is running
  //
  //
  // Post: Connection inConnectionId is closed. Client receives message
  //       informing about connection closure. Returns 1 on success, -1
  //       on failure.If timeout is negative blocking write occures until 
  //       socket is write ready.
  //
  // Desc: Connection inConnectionId is deleted from the list of 
  //       connected clients
  //
  //
  //
  



  int CloseConnection(long inConnectionId);

  
  //
  // Pre: Server is running
  //
  //
  // Post: All connections are closed. Returns 1 on success, -1 on failure.
  //
  //
  // Desc: All connections are deleted 
  //
  //
  //
  



  int CloseConnection();


  //
  // Pre: Server is running
  //
  //
  // Post: Puts ids of all open connections into outConnectionIds list
  //
  //
  // Desc: Intarnal ids assigned to each connection are put into 
  //       outConnecitonIds.  These ids can be used to select particular
  //       connections for reading/writing of data.
  //
  //
  //
  


  void GetConnectionIds(list<long>& outConnectionIds);


  //
  // Pre: Server is running
  //
  //
  // Post: Returns number of open connections at the time when this method
  //       is called.
  //
  // Desc:
  //
  //
  //
  
  long GetConnectionsCount(){return connections.size();};

  //
  // Pre: 
  //
  //
  // Post: 
  //       
  //
  // Desc:
  //
  //
  //
  
  static int Test();



protected:


private:

  //
  // Pre: Server is running, new connection is available on server socket
  //
  //
  // Post: New connection is created and added to hash table. Returns 1 
  //       on success, -1 on failure.
  //
  //
  // Desc: Depending on how server was started a new connection is opened
  //       and initialized.
  //
  //
  //

  int AddNewConnection();

  //
  // Pre: Server is running
  //
  //
  // Post: All the connections and receiving socket are checked if they 
  //       are ready to be read.  Appropriate data members are set to indicate
  //       this status. On success 1 is returned , 0 on timeout
  //       (no connection was ready within inTimeoutUs useconds), -1 on error
  //
  // Desc: All connections are checked for read status as well as server 
  //       receiving socket which receives new connections.  
  //
  //
  //

  int CheckRead(long inTimeoutUs);

  //
  // Pre: Server is running
  //
  //
  // Post: All the connections are checked if they 
  //       are ready to be written to.  Appropriate data members 
  //       are set to indicate
  //       this status. On success 1 is returned , 0 on timeout
  //       (no connection was ready within inTimeoutUs useconds), -1 on error
  //
  // Desc: All connections are checked for write status.  Server 
  //       receiving socket which receives new connections is not checked.  
  //
  //


  int CheckWrite(long inTimeoutUs);

  //
  // Pre: Server is running
  //
  //
  // Post: Number of connections that are read ready is returned
  //
  //
  // Desc:
  //
  //
  //

  long GetReadReadyCount();

  //
  // Pre: Server is running
  //
  //
  // Post: First available for reading connection is picked.
  //       Returns connectionId or 0 if no connections are available
  //
  // Desc:
  //
  //
  //


  long PickReadConnection();  


  // holds actual connections
  
  map<long,SCSConnection*> connections;

  // upper limit on number of connections

  long connectionsLimit;

  // server socket for connections

  int serverSocket;

  // flag indicating if any new connections are available

  bool newConnectionReady;

  // static connection id index that is increased to keep connection ids
  // unique.

  static long globalConnectionId;

  // socket type 

  int domainType;

  // socket path for UNIX socket

  string socketPath;

};

#endif
