//---------------------------------------------------------------------------
//
//  FILE NAME:          Chips2.cc
//
//  DEVELOPMENT:        CXCDS, User Infrastructure
//
//  DESCRIPTION:        Chips2 Application code
//
//
//---------------------------------------------------------------------------
// Copyright (c) 2004, Smithsonian Astrophysical Observatory
//---------------------------------------------------------------------------

#include "ChipsRemote.hh"
#include "ChipsParser.hh"

#include <readline/readline.h>
#include <readline/history.h>

#include <iostream>
#include <stdio.h>
#include <string>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

using namespace std;

static struct sigaction sigint_act;
static void sigint_hdlr(int sig)
{
   sig = errno;                                 // save errno
   sigaction(SIGINT,&sigint_act,NULL);          // restore handler
   write(STDERR_FILENO,"\n",1);                 // POSIX signal-safe output
   errno = sig;                                 // restore errno
}


// This routine, removes any preceeding whitespace and returns a value of
// true, if the strings are, ignoring case, equivalent upto the specified
// number of non-whitespace characters.

bool equivalent_substring(const char *aa, const char *bb, unsigned int len)
{
   bool same = false;
   if (aa && bb && len)
   {
      const char* c1 = aa;
      const char* c2 = bb;
      int l1, l2, cnt = len;

      // remove preceeding whitespace
      while(isspace(*c1)) c1++;
      while(isspace(*c2)) c2++;

      l1 = strlen(c1);
      l2 = strlen(c2);

      // if either string is shorter than the expected text,
      // then both must be the same length to be equivalent
      if ((l1 < len) || (l2 < len))
      {
         if (l1 == l2)
            cnt = l1;
         else
            return false;
      }

      // check lowercase values of strings
      same = true;
      while(same && cnt--)
      {
         if (tolower(*c1) != tolower(*c2))  same = false;
         c1++;
         c2++;
      }
   }
   return (same);
}




// routine to handle instructions. Made a separate function to allow
// recursion when loading a script.
int process_line(char* instring, ChipsRemote* sys, ChipsParser* parser, 
   bool feedback, bool echoscript, bool ignore)
{
   int err = 0; 

   if (!instring) return -1; 

   // skip preceeding whitespace 
   char* intext = &instring[0]; 
   while(isspace(*intext)) intext++;

   if (intext && (intext[0] != '\0'))
   {
      // remove any trailing whitespace
      int ln = strlen(intext) -1;
      while( (ln > 0) && (isspace(intext[ln])))  intext[ln--] = '\0';

      // since we are tracking the commandline history in the application
      // rather than in the parser, we should handle the 'history read' and
      // 'history write' commands here
      if (equivalent_substring(intext, "write", 5))
      {
         char* hw = intext;
         while(isspace(*hw)) hw++;
         hw+= 5; // adjust past 'write'  
         while(isspace(*hw)) hw++;

         // write history to specified file- omit 'history commands'
         write_history(hw);
      }
      else if (equivalent_substring(intext, "use", 3))
      {
         FILE *fp = NULL;
         char *line;
         char* hr = intext;
         while(isspace(*hr)) hr++;
         hr+= 3;  // adjust past 'use' text  
         while(isspace(*hr)) hr++;

         // load script request  
         if (!(fp = fopen(hr,"r"))) 
         {
            // unable to open the specified history file
            err = ERR_FEATURE_DNE;

            if (feedback)
               cerr << " ERROR: Unable to 'use' the file:" << hr << endl;
         }
         else
         {
            char buf[1025];
            int check =0;

            // execute the commands in the read request 
            while ((line = fgets(buf,1024,fp)) && (ignore || (err == 0)))
            {
               if (!check)
               {
                  if (!strcmp(line, "chips2-state"))  
                  {
                     err = ERR_STATEFILE_USE;   
                     if (feedback)   cerr << get_error_message(err) << endl;
		     break; 
                  }
                  check = 1; 
               }
               if (*line && (line[0] != '\0'))
               {
                  // strip off newline character
                  int ln = strlen(line);
                  if (ln && (line[ln-1] == '\n'))  line[ln-1] = '\0';

                  // ignore blank lines 
                  if (strlen(line))
                  {
                     if (echoscript)  cerr << " script line: " << line << endl;
                     err = process_line(line,sys, parser, feedback, 
                                        echoscript, ignore);
                  }
               }
            }
            fclose(fp);
         }
      }
      else
      {
         // store command in history
         add_history(intext);
        
         if (intext[0] == '$')
         {
            // run the specified shell escape or command
            pclose(popen(&intext[1], "w"));
         }
         else if (intext[0] != '\0')
         {
            // call parser if input text is not blank
            err = parser->parser(intext);

            // if unsuccessful, display error text
            if (feedback && (err != ERR_STATUS_GOOD))
            {
               const char* msg = get_error_message(err);
               if (msg)  cerr << msg << endl;
            }
         }
         if ((strlen(intext) == 4) && 
             (equivalent_substring(intext, "quit", 4) ||
              equivalent_substring(intext, "exit", 4)))      err = -1;
      }
   }
   return (err);
}


void get_prefs(ChipsRemote *sys, string* prompt, bool* feedback, 
   bool* echo, bool* ignore) 
{
   if (sys)
   {
      char* pref;

      if (prompt)
      {
         // set up prompt- use preference value if set
         if ((pref = sys->get_preference("chips.prompt")) != NULL)
         {
            *prompt = pref;
            free(pref);
         }
         else
            *prompt = "chips2>";
         prompt->append(" ");
      }
      if (feedback)
      {
         if ((pref = sys->get_preference("chips.feedback")) != NULL)
         {
            *feedback = (strcmp(pref, "true") == 0);
            free(pref);
         }
      }
      if (echo)
      {
         if ((pref = sys->get_preference("chips.echoscript")) != NULL)
         {
            *echo = (strcmp(pref, "true") == 0);
            free(pref);
         }
      }
      if (ignore)
      {
         if ((pref = sys->get_preference("chips.scripterrs")) != NULL)
         {
            *ignore = (strcmp(pref, "true") == 0);
            free(pref);
         }
      }
   }
}

int main(int argc, char *argv[])
{
   bool feedback = true, echoscript = false, ignore = true, done = false, 
        autoserver = true;
   int  err = ERR_STATUS_GOOD; 
   char* intext = NULL;
   ChipsRemote *sys = NULL;
   ChipsParser *parser;
   string server = "";
   string usefile = ""; 
   char *arg;

   // add signal handler to catch <control><c>
   if (isatty(STDIN_FILENO)) 
   {
      sigemptyset(&sigint_act.sa_mask);
      sigint_act.sa_handler = sigint_hdlr;
      sigint_act.sa_flags   = 0;
      sigaction(SIGINT,&sigint_act,NULL);
   }

   argv++;
   argc--;

   // To run regression tests set default to starting server
   // autoserver = true;

   while (argc > 0)
   {
      arg = argv[0];
      if (!strcmp(arg, "-a"))  
      {
         // run application with a new 'automatically created' server
         autoserver = true;
      }
      else if (!strcmp(arg, "-c"))
      {
         if (argc > 1)
         {
            // run application connected to specified existing server 
            autoserver = false;
	    server = argv[1];
            argv++;
            argc--;
         }
	 else
	 {
            cerr << " ERROR: -c option requires server id argument" << endl;
            exit(1); 
         }
      }
      else if (!strcmp(arg, "-f"))
      {
         if (argc > 1)
         {
	    // run using the specified file name 
            usefile = "use "; 
            usefile.append(argv[1]);
            argv++;
            argc--;
	 }
	 else
	 {
            cerr << " ERROR: -f option requires script file name" << endl;
            exit(1); 
         }
      }
      else if (!strcmp(arg, "-s"))
      {
         if (argc>1) 
         {
            // run as background server only- no application
	    string command="chipsServer -d -s ";
	    command += (string)argv[1];
	     exit (system(command.c_str()) < 0); 
         }
	 else
	 {
            cerr << " ERROR: -s option requires server socket id" << endl;
            exit(1); 
         }
      }
      else if (!strcmp(arg, "-n"))
      {
         // run application without a server connection
         autoserver = false;
      }
      else
      {
         cerr << " ERROR: Unknown or unsupported commandline option- " << 
            argv[0] << endl;
         exit(1); 
      }
      argc--; 
      argv++;
   }

   if (autoserver)
      sys = new ChipsRemote(true, true);
   else if (server.length())
{
      sys = new ChipsRemote(server, true);
//string foo = sys->get_server_id();
//cout << " Connected to server: " << foo << endl;
}
   else
      sys = new ChipsRemote();

   if (sys)
   {
      string prompt;
      int ln;

      // initialize parser
      parser = new ChipsParser(sys);

      // set up prompt- use preference value if set 
      get_prefs(sys, &prompt, &feedback, &echoscript, &ignore); 

      if ((ln = usefile.length()) > 0)
      {
         char* val = (char*)malloc((ln+1)*sizeof(char));
         if (val)
         {
            strcpy(val, usefile.c_str());
            err = process_line(val, sys, parser, feedback, 
                               echoscript, ignore);
            free(val);
         }
         exit(err); 
      }

      // get user commands 
      while(!done)
      {
         // get line of input
         intext = readline(prompt.c_str());

         // handle the line
         err = process_line(intext, sys, parser, feedback, echoscript, ignore);
 
         done = (err == -1);
         free(intext);

         // set up prompt- use preference value if set 
         get_prefs(sys, &prompt, &feedback, &echoscript, &ignore); 
      }
      delete sys; 
   }
   else  
   {
      cerr << get_error_message(ERR_LIBRARY_INIT) << endl;
      exit(1); 
   }
   exit(0); 
}

