// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsAdjustRelsizeCommand.hh"
#include "ChipsDefaults.hh"
#include "ChipsStateStream.hh"
#include "ChipsWindow.hh"
#include "ChipsFrame.hh"

extern ChipsDefaults* theDefaults; 

ChipsAdjustRelsizeCommand::ChipsAdjustRelsizeCommand(ChipsObject *new_co, int cnt,
   int dimension, float* ovals, float osum)
{
   co = new_co;
   weight=1;
   oldv = newv = NULL; 
   orelsum = nrelsum = 0; 
   if (cnt > 0)
   {
      orelsum = osum;
      int rr; 
      ocount = cnt;
      dimn = dimension; 
      if (ovals) 
      {
         oldv = new float[ocount]; 
	 for (rr = cnt; rr--; )   oldv[rr] = ovals[rr];
      }
   }
}

ChipsAdjustRelsizeCommand::~ChipsAdjustRelsizeCommand()
{
   if (oldv)  delete []oldv; 
   if (newv)  delete []newv;
}


ChipsAdjustRelsizeCommand::ChipsAdjustRelsizeCommand(ChipsObject* cob, ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

int ChipsAdjustRelsizeCommand::execute ()
{
   int rval = ERR_STATUS_GOOD;
   if (co && (co->get_object_type() == chips_frame))
   {
      rval = dynamic_cast<ChipsFrame *>(co)->reset_relsizes(ncount, dimn, newv, nrelsum);
      if (rval == ERR_STATUS_GOOD) 
        dynamic_cast<ChipsWindow *>(co->get_parent())->update_display();
   }
   return rval;
}

int ChipsAdjustRelsizeCommand::unexecute ()
{
   int rval = ERR_STATUS_GOOD;
   if (co && (co->get_object_type () == chips_frame))
   {
      rval = dynamic_cast<ChipsFrame *>(co)->reset_relsizes(ocount, dimn, oldv, orelsum);
      if (rval == ERR_STATUS_GOOD) 
        dynamic_cast<ChipsWindow *>(co->get_parent())->update_display();
   }
   return rval;
}

void ChipsAdjustRelsizeCommand::serialize(ChipsStateStream& out)
{
   int val = 23;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_int(&dimn);
   out.write_int(&ocount);
   out.write_int(&ncount);
   out.write_float(&orelsum);
   out.write_float(&nrelsum);
   if (ocount > 0)  out.write_float(oldv, ocount);
   if (ncount > 0)  out.write_float(newv, ncount);
}

void ChipsAdjustRelsizeCommand::deserialize(ChipsStateStream& in)
{
   int rr; 
   in.read_int(&ref);
   in.read_int(&block_ref);
   int usedbl = 0; 
   oldv = newv = NULL; 
   in.read_int(&dimn);
   in.read_int(&ocount);
   in.read_int(&ncount);
   if (in.get_major_version() < 3 ||
      (in.get_major_version() == 3 && in.get_minor_version() < 16))
   {
      usedbl = 1;
      double dd;
      in.read_double(&dd);
      orelsum = (float)dd; 
      in.read_double(&dd);
      nrelsum = (float)dd; 
   }
   else 
   {
      in.read_float(&orelsum);
      in.read_float(&nrelsum);
   }
   if (ocount > 0)
   {
      oldv = new float[ocount];
      if (usedbl)
      {
         double * tmpd = new double[ocount];
         in.read_double(tmpd, ocount);
         for (rr = 0; rr < ocount; rr++)   oldv[rr] = tmpd[rr];
         delete[] tmpd; 
      } 
      else 
      {
         in.read_float(oldv, ncount);
      } 
   }
   if (ncount > 0)
   {
      newv = new float[ncount];
      if  (usedbl)
      {
         double * tmpd = new double[ncount];
         in.read_double(tmpd, ncount);
         for (rr = 0; rr < ncount; rr++)   newv[rr] = tmpd[rr];
         delete[] tmpd;  
      }
      else 
      {  
         in.read_float(newv, ncount);
      }
   }
}

void ChipsAdjustRelsizeCommand::print ()
{
   cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"AdjustRelsize" << endl;
};


void ChipsAdjustRelsizeCommand::store_new_values(int cnt, float* arr, float sum)
{
   if (newv && ncount)
   {
      delete[] newv;
      newv = NULL;
      ncount = 0;
   }
   if (cnt > 0)
   {
      newv = new float[cnt]; 
      ncount = cnt; 
   }
   if (arr && newv)
   {
      for (int rr = ncount; rr--; ) newv[rr] = arr[rr]; 
      nrelsum = sum; 
   }
}


