// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsAxis.hh"
#include "ChipsFontList.hh"
#include "ChipsPlot.hh"

#include <math.h>
#include <float.h>
#include <stdio.h>
#include <sstream>
#include <algorithm>

#include "ChipsCmdManager.hh"
#include "ChipsAxisTickCommand.hh"
#include "ChipsAxisAlgorithm.hh"
#include "ChipsCartesianAxisAlgorithm.hh"
#include "ChipsAxisAlgorithmFactory.hh"
#include "ChipsTransformFactory.hh"
#include "Transform.hh"


extern ChipsDefaults* theDefaults; 
extern ChipsFontList* theFonts;
extern ChipsCmdManager* cmd_manager;

ChipsAxis::ChipsAxis(ChipsObject* par, ChipsCoordinate *inCoord,
   COORD_SYS coord_sys, ChipsAxisAlgorithmFactory::TransformType type,
   string val):ChipsObject(par, val)
{
   initialized = false;
   id = val; 
   orig_pos[0] = ChipsVector (0.0, 0.0);
   orig_pos[1] = ChipsVector (1.0, 1.0);
   norm_pos[0] = orig_pos[0];
   norm_pos[1] = orig_pos[1];
   
   axis_alg = ChipsAxisAlgorithmFactory::create (type, X_AXIS, this);
   initialize();
   post_initialize (inCoord, coord_sys);

}

ChipsAxis::ChipsAxis(ChipsObject* par, ChipsCoordinate *inCoord, COORD_SYS coord_sys, ChipsStateStream& in):ChipsObject(par, "")
{
   initialized = false;
   orig_pos[0] = ChipsVector (0.0, 0.0);
   orig_pos[1] = ChipsVector (1.0, 1.0);
   

   axis_alg = ChipsAxisAlgorithmFactory::create (ChipsAxisAlgorithmFactory::CARTESIAN, X_AXIS, this);
   initialize();


   deserialize(in);
   norm_pos[0] = orig_pos[0];
   norm_pos[1] = orig_pos[1];

   if (inCoord)  coord = inCoord;
   if (coord)
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_system == WINDOW_NORM) || (coord_system == PIXEL) ||
          (coord_system == PIXEL_REAL))
         coord = coord->get_coord(FRAME_NORM, "");
      else
         coord = coord->get_coord(coord_system, data_system_id);

   }


   post_initialize (coord, coord_system);

}

ChipsAxis::ChipsAxis(ChipsObject* par, ChipsCoordinate *inCoord,
   COORD_SYS coord_sys, double inMinXPos, double inMinYPos,
   double inMaxXPos, double inMaxYPos, int inOrientation,
   ChipsAxisAlgorithmFactory::TransformType type,
   string val):ChipsObject(par, val)
{
   initialized = false;
   id = val;
   orig_pos[0] = ChipsVector (inMinXPos, inMinYPos);
   orig_pos[1] = ChipsVector (inMaxXPos, inMaxYPos);
   norm_pos[0] = orig_pos[0];
   norm_pos[1] = orig_pos[1];

   axis_alg = ChipsAxisAlgorithmFactory::create (type, inOrientation, this);
   initialize();

   post_initialize (inCoord, coord_sys);
}


ChipsAxis::ChipsAxis(ChipsObject* par, ChipsCoordinate *inCoord, 
   COORD_SYS coord_sys, double inMinXPos, double inMinYPos, 
   double inMaxXPos, double inMaxYPos, int inOrientation, double inmin, 
   double inmax, ChipsAxisAlgorithmFactory::TransformType type, 
   string val):ChipsObject(par, val)
{
   double tmp_min = inmin;
   double tmp_max = inmax;
   initialized = false;
   id = val;
   orig_pos[0] = ChipsVector (inMinXPos, inMinYPos);
   orig_pos[1] = ChipsVector (inMaxXPos, inMaxYPos);
   norm_pos[0] = orig_pos[0];
   norm_pos[1] = orig_pos[1];

   axis_alg = ChipsAxisAlgorithmFactory::create (type, inOrientation, this);
   initialize();

   post_initialize (inCoord, coord_sys);

   if (inmin == inmax)
   {
      if (inmin == 0)
      {
         tmp_min = -1.0;
         tmp_max = 1.0;
      }
      else
      {
         tmp_min = inmin* 0.95;
         tmp_max = inmax * 1.05;
      }
   }
   axis_alg->set_absolute_minmax (tmp_min, tmp_max);
/*   rval = update_tick_limits ();
*/
   if (reverse)   change_reverse();
}


ChipsAxis::~ChipsAxis()
{
   if (tick_text)
   {
      for (unsigned int ii = 0; ii < ttext_cnt; ii++)
         delete_tree(tick_text[ii]);
      delete[] tick_text;
   }
   if (coord)
      coord->remove (this);

   if (axis_alg)
      delete axis_alg;
}

void ChipsAxis::object_delete ()
{
   delete this;
}

void ChipsAxis::initialize()
{

   reverse_set = 0; 
   title_flip = false; 
   label_vertalignment=0.0;
   label_horizalignment=0.5;
   temp_label = ""; 
   ticklabfontfile = "DEFAULT_TIMES";
   alterid = id;
   tick_text = NULL;
   ttext_cnt = 0;
   use_pad = true;
   border_axis = false;

   if (theDefaults)
   {
      string temp;
      int    tmpi; 
      temp = theDefaults->get_axiscolor();
      if (set_color(temp) != 0)   set_color("yellow");
      temp = theDefaults->get_majortickcolor();
      if (set_major_color(temp) != 0)   set_major_color("white");
      temp = theDefaults->get_minortickcolor();
      if (set_minor_color(temp) != 0)   set_minor_color("white");
      set_thickness(theDefaults->get_axisthickness()); 
      temp = theDefaults->get_tickformat();
      if (set_tickformat(temp) != 0)  set_tickformat("%3d"); 
      temp = theDefaults->get_majortickmode();
      if (set_major_mode(temp) != 0)   set_major_mode("count");  
      temp = theDefaults->get_minortickmode();
      if (set_minor_mode(temp) != 0)   set_minor_mode("nice");  
      if (set_major_int(theDefaults->get_majortickinterval()) != 0)
         set_major_int(10);
      if (set_minor_int(theDefaults->get_minortickinterval()) != 0)
         set_minor_int(2);
      set_major_num(theDefaults->get_majortickcount()); 
      set_minor_num(theDefaults->get_minortickcount()); 
      if (set_major_thickness(theDefaults->get_majortickthickness()) != 0)
         set_major_thickness(2);  
      if (set_minor_thickness(theDefaults->get_minortickthickness()) != 0)
         set_minor_thickness(1);  
      if (set_major_tickstyle(theDefaults->get_majortickstyle()) != 0)
         set_major_tickstyle(chips_inside);
      if (set_minor_tickstyle(theDefaults->get_minortickstyle()) != 0)
         set_minor_tickstyle(chips_inside);
      if (set_ticklabelstyle(theDefaults->get_ticklabelstyle()) != 0)
         set_ticklabelstyle(chips_outside);
      tmpi = theDefaults->get_majorticklength(); 
      if (tmpi >= 0)  set_major_len((int)tmpi);
         else         set_major_len(4);
      tmpi = theDefaults->get_minorticklength(); 
      if (tmpi >= 0)  set_minor_len((int)tmpi);
         else         set_minor_len(4);
      if (theDefaults->get_majortickvisible()) show_major();  
         else                                  hide_major();
      if (theDefaults->get_minortickvisible()) show_minor();  
         else                                  hide_minor();
      offset_par = theDefaults->get_axisparalleloffset();
      offset_perp = theDefaults->get_axisperpendicularoffset();
      pad = theDefaults->get_axispad();
      label_size = theDefaults->get_axlabsize();
      label_color = theDefaults->get_axlabcolor();
      label_font = theDefaults->get_axlabfont();
      label_fontstyle = theDefaults->get_axlabfontstyle(); 
      automin = theDefaults->get_axisautomin();
      automax = theDefaults->get_axisautomax();
      set_minor_grid_visible(theDefaults->get_minorgridvisible());
      set_major_grid_visible(theDefaults->get_majorgridvisible());
      temp = theDefaults->get_majorgridcolor();
      if (set_major_grid_color(temp) != 0)   set_major_grid_color("white");
      temp = theDefaults->get_minorgridcolor();
      if (set_minor_grid_color(temp) != 0)   set_minor_grid_color("white");
      if (set_major_grid_thickness(theDefaults->get_majorgridthickness()) != 0)
         set_major_grid_thickness(1);
      if (set_minor_grid_thickness(theDefaults->get_minorgridthickness()) != 0)
         set_minor_grid_thickness(1);
      if (set_major_grid_style(theDefaults->get_majorgridstyle()) != 0)
         set_major_grid_style(chips_shortdash);
      if (set_minor_grid_style(theDefaults->get_minorgridstyle()) != 0)
         set_minor_grid_style(chips_dot);
      temp = theDefaults->get_ticklabelcolor();
      if (set_ticklabel_color(temp) != 0)   set_ticklabel_color("white");
      ticklaboffset = theDefaults->get_ticklabeloffset();
      ticklabangle = theDefaults->get_ticklabelangle();
      ticklabsize = theDefaults->get_ticklabelsize();
      ticklabfontstyle = theDefaults->get_ticklabelfontstyle();
      temp = theDefaults->get_ticklabelfont();
      set_ticklabel_font(temp);
      depth = theDefaults->get_axisdepth();
      ticklabvertalignment = theDefaults->get_ticklabelvalign();
      ticklabhorizalignment = theDefaults->get_ticklabelhalign();
      if (axis_alg)
      {
         if (axis_alg->get_orientation() == X_AXIS)
           label_angle = theDefaults->get_axlabxangle();
         else
           label_angle = theDefaults->get_axlabyangle();
      }
   }
   else
   {
      set_color("yellow"); 
      set_major_color("white"); 
      set_minor_color("white"); 
      set_tickformat("%3d"); 
      set_thickness(2); 
      set_major_mode("count"); 
      set_minor_mode("nice"); 
      set_major_int(10);
      set_minor_int(2);
      set_major_num(4);
      set_minor_num(4);
      set_major_len(4);
      set_minor_len(2);
      set_major_thickness(2);  
      set_minor_thickness(1);  
      set_major_tickstyle(chips_inside);
      set_minor_tickstyle(chips_inside);
      set_ticklabelstyle(chips_outside);
      set_minor_grid_visible(false);
      set_major_grid_visible(false);
      set_major_grid_color("white");
      set_major_grid_style(chips_shortdash);
      set_major_grid_thickness(1);
      set_minor_grid_color("white");
      set_minor_grid_style(chips_dot);
      set_minor_grid_thickness(1);
      show_major(); 
      show_minor();
      offset_par = 0.0; 
      offset_perp = 20.0; 
      label_size = 10;
      label_color = "yellow";
      label_font = "helvetica";
      label_fontstyle = chips_normal; 
      automin = automax = true; 
      pad = 0.05; 
      set_ticklabel_color("white");
      ticklaboffset = 2;
      ticklabangle = 0.0;
      ticklabsize = 10;
      ticklabfont = "helvetica";
      ticklabfontstyle = chips_normal;
      ticklabvertalignment = 0.0;
      ticklabhorizalignment = 0.5;
      depth = V_USE_DEFAULT_DEPTH;
      if (axis_alg && axis_alg->get_orientation() == Y_AXIS)
         label_angle = 90.0;
      else
         label_angle = 0.0;
   }
   if (depth == V_USE_DEFAULT_DEPTH)  depth = ((ChipsPlot*)parent)->get_depth();
   labvis = true;
   ticklabvis = true;
   nx1 = ny1 = 0.0;
   nx2 = ny2 = 1.0;
   label = NULL; 
   deserialize_label = 0; 
   reverse = false;
   axis_num = 0;
   sexagesimal = false;
   initialized = true;
   shared_ds = false;

   // since major grids are used to represent minor grids at major tick 
   // positions when major are off but minor are visible, we have to
   // set default 'use' grid values if they weren't initialized 
   if ((grid_min_vis && !grid_maj_vis))
   {
      grid_use_style = grid_min_style;
      grid_use_clr = grid_min_clr;
      grid_use_thick = grid_min_thick;
   }

   // set properties for corner miters 
   ll_corner = ur_corner = chips_miter;
   ll_corner_width = ll_corner_width = thickness; 
   alignment = chips_centered; 

   // for now, populate the arbitrary tick locations with some
   // default values- user can reset them to their preferences
   double arbpos [] = {0};
   axis_alg->set_arbitrary_positions (arbpos, NULL, 1);
   object_type = chips_axis;

// NEW FOR DEPTH
   grid_start_pos = 0;
   grid_stop_pos = 0;

   coord = NULL;

}

void ChipsAxis::post_initialize (ChipsCoordinate *inCoord, COORD_SYS coord_sys)
{
   // Cant process if there's no chips coordinate
   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);

      // convert the data to logical coordinates if necessary
      if (coord_sys == DATA)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (orig_pos, orig_pos, 2, DATA, "", DATA, data_system_id);
         }

      }


   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);

   if (!status)
      status = calculate_norm_coordinate ();

   // Set the axis direction
   invert_labels();
   set_ticklabel_valign(ticklabvertalignment);
   set_ticklabel_halign(ticklabhorizalignment);
}


void ChipsAxis::synchronize(const ChipsAxis& other)
{
   ChipsPlot *parent_plot;
   automax    = other.automax;  
   automin    = other.automin;  
   pad        = other.pad;
   reverse    = other.reverse;
   use_pad    = other.use_pad;

   int alg_type = ChipsAxisAlgorithmFactory::get_type (other.axis_alg);
   int curr_alg_type = 0;

   if (axis_alg)
   {
      curr_alg_type = ChipsAxisAlgorithmFactory::get_type (axis_alg);
      delete axis_alg;
   }

   axis_alg = ChipsAxisAlgorithmFactory::create (alg_type, other.axis_alg->get_orientation (), this, other.axis_alg);
 
   axis_alg->set_axis (this);

   tickformat = other.tickformat; 


   // if a no data systems exist with this axis copy in the data systems from the
   // bound axis
   if ((alg_type != curr_alg_type) && !this->border_axis)
   {
      this->data_coord_ids = other.data_coord_ids;
      parent_plot = dynamic_cast <ChipsPlot *> (this->parent);
      if (parent_plot)
         parent_plot->update_axis_coordinates (this, this->data_coord_ids);
   }
   else if (this->border_axis)
   {
      // just copy the data system for border axes
      ChipsDataSystem *ds = this->get_data_system ();
      const ChipsDataSystem *new_ds;

      if (other.data_coord_ids.empty ())
         new_ds = NULL;
      else
         new_ds = dynamic_cast <ChipsDataSystem *> (coord->get_coord (DATA, *(other.data_coord_ids.begin ())));
      if (ds && new_ds)
      {
         string xaxis_id = ds->get_xaxis_id ();
         string yaxis_id = ds->get_yaxis_id ();
     
         *ds = *new_ds;
         ds->set_xaxis_id (xaxis_id);
         ds->set_yaxis_id (yaxis_id);
      }
   }
      
   if (initialized)
   {

      update_tick_positions ();
//      if (axis_alg->get_scaletype () == V_SCALE_LOG)   log_scale(); 
//         else                         linear_scale();
      change_scale();
      change_minor_mode ();
      change_minor_num ();
      change_tick_format(); 
      change_range ();
   }

}


int ChipsAxis::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot) ||
       (obj == chips_axis))
   {
      int mask = theDefaults->get_allpropmask();

      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();

         // set plot properties
         rval =  set_property((void*)clr.c_str(), SET_AXIS_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_MAJOR_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_MINOR_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_LABEL_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_TICKLABEL_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_MAJGRD_COLOR);
         rval |= set_property((void*)clr.c_str(), SET_AXIS_MINGRD_COLOR);
      }
      else if ((mask & PREF_ALL_FONT) == PREF_ALL_FONT)
      {
         string newfont = theDefaults->get_allfont();
         rval =  set_property((void*)newfont.c_str(), SET_AXIS_LABEL_FONT);
         rval |=  set_property((void*)newfont.c_str(), SET_AXIS_TICKLABEL_FONT);
      }
      else if ((mask & PREF_ALL_FONTSTYLE) == PREF_ALL_FONTSTYLE)
      {
         V_FONTSTYLE_TYPE newfontstyle = theDefaults->get_allfontstyle();
         string style = v_fontstyle_to_string(newfontstyle);
         rval =  set_property((void*)style.c_str(), SET_AXIS_LABEL_FONTSTYLE);
         rval |=  set_property((void*)style.c_str(), SET_AXIS_TICKLABEL_FNTSTL);
      }
      else if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
      {
         float thick = theDefaults->get_allthickness();
         rval = set_property(static_cast<void*>(&thick), SET_AXIS_THICKNESS);
         rval |= set_property(static_cast<void*>(&thick), SET_AXIS_MAJ_THICKNESS);
         rval |= set_property(static_cast<void*>(&thick), SET_AXIS_MIN_THICKNESS);
         rval |= set_property(static_cast<void*>(&thick), SET_AXIS_MAJGRD_THICK);
         rval |= set_property(static_cast<void*>(&thick), SET_AXIS_MINGRD_THICK);
      }
      else if ((mask & PREF_ALL_COUNT) == PREF_ALL_COUNT)
      {
         int newcount = theDefaults->get_allcount();
         rval = set_property(static_cast<void*>(&newcount), SET_AXIS_MAJOR_NUM);
         rval |= set_property(static_cast<void*>(&newcount), SET_AXIS_MINOR_NUM);
      }
      else if ((mask & PREF_ALL_LENGTH) == PREF_ALL_LENGTH)
      {
         int newlength = theDefaults->get_alllength();
         rval = set_property(static_cast<void*>(&newlength), SET_AXIS_MAJORLENGTH);
         rval |= set_property(static_cast<void*>(&newlength), SET_AXIS_MINORLENGTH);
      }
      else if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         rval = set_property(static_cast<void*>(&vis), SET_AXIS_VISIBLE); 
      }
      else if ((mask & PREF_ALL_ANGLE) == PREF_ALL_ANGLE)
      {
         float newangle = theDefaults->get_allangle();
         rval = set_property(static_cast<void*>(&newangle), SET_AXIS_LABEL_ANGLE);
         rval |= set_property(static_cast<void*>(&newangle), SET_AXIS_TICKLABEL_ANGLE);
      }
      else if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
	 int dep = theDefaults->get_alldepth();
         rval = set_property(static_cast<void*>(&dep), SET_AXIS_DEPTH);
      }
   }
   return rval;
}

int ChipsAxis::set_property(void* in_val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   string save_current; 
   if (parent && (parent->get_object_type() == chips_plot)) 
   {
      if (axis_alg->get_orientation () == X_AXIS)
         save_current = ((ChipsPlot *)parent)->get_current_xaxis_id();
      else
         save_current = ((ChipsPlot *)parent)->get_current_yaxis_id();
   }

   switch(property)
   {
      case SET_AXIS_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_color(), clr, save_current);
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;  
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJOR_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_major_color(), clr, save_current);
         if ((rval = set_major_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec) 
               if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINOR_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_minor_color(), clr, save_current);
         if ((rval = set_minor_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            thickness, *fval, save_current);
         if ((rval = set_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJ_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            major_thickness, *fval, save_current);
         if ((rval = set_major_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MIN_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            minor_thickness, *fval, save_current);
         if ((rval = set_minor_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TEXT:
      {
	 string val = (const char*)in_val; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_text(), val, save_current);
         if ((rval = set_label_text(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_XLAB_TXT:
      {
         if (axis_alg->get_orientation ()  == X_AXIS)
         { 
            string val = (const char*)in_val;
            if (rec)  cmd = new ChipsSetCommand(this, property,
               get_label_text(), val, save_current);
            if ((rval = set_label_text(val)) == ERR_STATUS_GOOD)
            {
               if (rec)
                 if (!add_command (cmd, rval))  delete cmd;
            }
            else if (cmd)  delete cmd;
         }
      }
      break;
      case SET_AXIS_YLAB_TXT:
      {
         if (axis_alg->get_orientation ()  == Y_AXIS)
         {
            string val = (const char*)in_val;
            if (rec)  cmd = new ChipsSetCommand(this, property,
               get_label_text(), val, save_current);
            if ((rval = set_label_text(val)) == ERR_STATUS_GOOD)
            {
               if (rec)
                 if (!add_command (cmd, rval))  delete cmd;
            }
            else if (cmd)  delete cmd;
         }
      }
      break;
      case SET_AXIS_LABEL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_angle(), *fval, save_current);
         if ((rval = set_label_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_XLAB_ANG:
      {
         if (axis_alg->get_orientation ()  == X_AXIS)
         {
            float* fval = static_cast<float*>(in_val);
            if (rec)  cmd = new ChipsSetCommand(this, property,
               get_label_angle(), *fval, save_current);
            if ((rval = set_label_angle(*fval)) == ERR_STATUS_GOOD)
            {
               if (rec)
                 if (!add_command (cmd, rval))  delete cmd;
            }
            else if (cmd)  delete cmd;
         }
      }
      break;
      case SET_AXIS_YLAB_ANG:
      {
         if (axis_alg->get_orientation ()  == Y_AXIS)
         {
            float* fval = static_cast<float*>(in_val);
            if (rec)  cmd = new ChipsSetCommand(this, property,
               get_label_angle(), *fval, save_current);
            if ((rval = set_label_angle(*fval)) == ERR_STATUS_GOOD)
            {
               if (rec)
                 if (!add_command (cmd, rval))  delete cmd;
            }
            else if (cmd)  delete cmd;
         }
      }
      break;
      case SET_AXIS_LABEL_HALIGN:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_halign(), *fval, save_current);
         if ((rval = set_label_halign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_LABEL_VALIGN:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_valign(), *fval, save_current);
         if ((rval = set_label_valign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKSTYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)major_tickstyle; 
         if (rec)  cmd = new ChipsSetCommand(this, SET_AXIS_MAJ_STYLE,
            oval, *ival, save_current);
         if ((rval = set_major_tickstyle((V_TICKSTYLE_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
	    oval = (int)minor_tickstyle; 
            if (rec)  cmd = new ChipsSetCommand(this, SET_AXIS_MIN_STYLE,
               oval, *ival, save_current);
            if ((rval = set_minor_tickstyle((V_TICKSTYLE_TYPE)*ival)) == 
	       ERR_STATUS_GOOD)
            {
               if (rec)
                  if (!add_command (cmd, rval))  delete cmd;
            }
            else if (cmd)  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJ_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)major_tickstyle; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_major_tickstyle((V_TICKSTYLE_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MIN_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)minor_tickstyle; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_minor_tickstyle((V_TICKSTYLE_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKFORMAT:
      {
	 string val = (const char*)in_val; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_tickformat(), val, save_current);
         if ((rval = set_tickformat(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
            cmd_manager->update_mask (STATE_CHNG_AXS_FMT);
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)ticklabstyle; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_ticklabelstyle((V_TICKSTYLE_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJOR_NUM:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            axis_alg->get_major_num (), *ival, save_current);
         if ((rval = set_major_num(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
{
              if (!add_command (cmd, rval))  delete cmd;
            rval = set_property((void*)"count", SET_AXIS_MAJ_MODE,rec);
}
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINOR_NUM:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            axis_alg->get_minor_num (), *ival, save_current);
         if ((rval = set_minor_num(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJORINT:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_major_int(), *dval, save_current);
         if ((rval = set_major_int(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
{
              if (!add_command (cmd, rval))  delete cmd;
            rval = set_property((void*)"interval", SET_AXIS_MAJ_MODE,rec);
}
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINORINT:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            axis_alg->get_minor_interval (), *dval, save_current);
         if ((rval = set_minor_int(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_LOG_MIN:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_log_minimum (), *dval, save_current);
         if ((rval = set_log_minimum(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_PAD:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            pad, *dval, save_current);
         if ((rval = set_pad(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_MAJORLENGTH:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            major_len, *ival, save_current);
         if ((rval = set_major_len(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINORLENGTH:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            minor_len, *ival, save_current);
         if ((rval = set_minor_len(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_AXIS_MAJ_MODE:
      {
	 string val = (const char*)in_val; 
	 string oval = v_tickmode_to_string(get_major_mode());
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_major_mode(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_MIN_MODE:
      {
	 string val = (const char*)in_val; 
	 string oval = v_tickmode_to_string(get_minor_mode());
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_minor_mode(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_VISIBLE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            (int)visible, *ival, save_current);
         if (*ival == 1)
         {
            if (initialized)  rval = show();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (*ival == 0)
         {
            if (initialized)  rval = hide();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else 
	 {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "axis visibility", *ival,0, 1);
	    if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
      }
      break;
      case SET_AXIS_AUTOMIN:
      {
         bool* bval = static_cast<bool*>(in_val);
	 bool oval = automin;
         set_automin(*bval);
         if (rec)
	 {
	    cmd = new ChipsSetCommand(this, property,
              oval, *bval, save_current);
            if (!add_command (cmd, rval))  delete cmd;
         }
      }
      break;
      case SET_AXIS_AUTOMAX:
      {
         bool* bval = static_cast<bool*>(in_val);
	 bool oval = automax;
         set_automax(*bval);
         if (rec)
	 {
	    cmd = new ChipsSetCommand(this, property,
              oval, *bval, save_current);
            if (!add_command (cmd, rval))  delete cmd;
         }
      }
      break;
      case SET_AXIS_DEPTH:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int newval;
         if ((*ival) == V_USE_DEFAULT_DEPTH)    
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, property,
              (int)depth, newval, save_current);
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      case SET_AXIS_REVERSE:
      {
         bool bval = !reverse;
         reverse_set = 1; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
              reverse, bval, save_current);
         if ((rval = reverse_axis()) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break; 
      case SET_AXIS_SCALE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              axis_alg->get_scaletype (), *ival, save_current);
         if ((*ival) == V_SCALE_LOG)
            rval = log_scale();
          else  
            rval = linear_scale();
         if (rval == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break;
      case SET_AXIS_SCALABLE:
      {
         bool* bval = static_cast<bool*>(in_val);
	 bool oval = axis_alg->get_scalable ();
         axis_alg->set_scalable(*bval);
         if (rec)
	 {
	    cmd = new ChipsSetCommand(this, property,
              oval, *bval, save_current);
            if (!add_command (cmd, rval))  delete cmd;
         }
      }
      break; 
      case SET_AXIS_MAJ_VIS:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = major_vis;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show_major();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide_major();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "axis major tick visibility", *ival, 0, 1);
            if (cmd)
               delete cmd;
         }
      }
      break; 
      case SET_AXIS_MIN_VIS:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = minor_vis;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show_minor();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide_minor();
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "axis minor tick visibility", *ival, 0, 1);
            if (cmd)
               delete cmd;
         }
      }
      break; 
      case SET_AXIS_OFFSET_PAR:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            offset_par, *fval, save_current);
         if ((rval = set_parallel_offset(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_OFFSET_PERP:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            offset_perp, *fval, save_current);
         if ((rval = set_perpendicular_offset(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_LABEL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_color(), clr, save_current);
         if ((rval = set_label_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_ticklabel_color(), clr, save_current);
         if ((rval = set_ticklabel_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_OFFSET:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            ticklaboffset, *ival, save_current);
         if ((rval = set_ticklabel_offset(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            ticklabsize, *ival, save_current);
         if ((rval = set_ticklabel_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            ticklabangle, *fval, save_current);
         if ((rval = set_ticklabel_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_HALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = ticklabhorizalignment;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *fval, save_current);
         if ((rval = set_ticklabel_halign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_VALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = ticklabvertalignment;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *fval, save_current);
         if ((rval = set_ticklabel_valign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_VIS:
      {
         bool* bval = static_cast<bool*>(in_val);
         bool oval = ticklabvis;
         if (rec)  cmd = new ChipsSetCommand(this, property,
              oval, *bval, save_current);
         if ((rval = set_ticklabel_visible(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_FONT:
      {
         string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_ticklabel_font(), val, save_current);
         if ((rval = set_ticklabel_font(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_TICKLABEL_FNTSTL:
      {
         string val = (const char*)in_val;
	 string oval = v_fontstyle_to_string(get_ticklabel_fontstyle());
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_ticklabel_fontstyle(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_LABEL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = get_label_size();
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_label_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_LABEL_FONT:
      {
         string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_label_font(), val, save_current);
         if ((rval = set_label_font(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_LABEL_VIS:
      {
         bool* bval = static_cast<bool*>(in_val);
         bool oval = labvis;
         if (rec)  cmd = new ChipsSetCommand(this, property,
              oval, *bval, save_current);
         if ((rval = set_label_visible(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break;
      case SET_AXIS_LABEL_FONTSTYLE:
      {
         string val = (const char*)in_val;
	 string oval = v_fontstyle_to_string(get_label_fontstyle());
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_label_fontstyle(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJGRD_VIS:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              grid_maj_vis, *bval, save_current);
         if ((rval = set_major_grid_visible(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break;
      case SET_AXIS_MAJGRD_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_major_grid_color(), clr, save_current);
         if ((rval = set_major_grid_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_MAJGRD_THICK:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            grid_use_thick, *fval, save_current);
         if ((rval = set_major_grid_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MAJGRD_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)grid_use_style;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_major_grid_style((V_LINE_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_MINGRD_VIS:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              grid_min_vis, *bval, save_current);
         if ((rval = set_minor_grid_visible(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)
            delete cmd;
      }
      break;
      case SET_AXIS_MINGRD_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_minor_grid_color(), clr, save_current);
         if ((rval = set_minor_grid_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINGRD_THICK:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            grid_min_thick, *fval, save_current);
         if ((rval = set_minor_grid_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_MINGRD_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int)grid_min_style;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_minor_grid_style((V_LINE_TYPE)*ival)) ==
	    ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      case SET_AXIS_LLCORNER:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)ll_corner;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_ll_corner((V_CORNERSTYLE_TYPE)*ival)) ==
            ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_URCORNER:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)ur_corner;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_ur_corner((V_CORNERSTYLE_TYPE)*ival)) ==
            ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_LLCORNWID:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            ll_corner_width, *dval, save_current);
         if ((rval = set_ll_corner_width(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break;
      case SET_AXIS_URCORNWID:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            ur_corner_width, *dval, save_current);
         if ((rval = set_ur_corner_width(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
         }
         else if (cmd)  delete cmd;
      }
      break; 
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);
   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_AXS);


   return rval;
}


int ChipsAxis::get_property(void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_AXIS_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_AXIS_MAJOR_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_major_color();
      }
      break;
      case SET_AXIS_MINOR_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_minor_color();
      }
      break;
      case SET_AXIS_THICKNESS:
      {
         float *fv = (float*)out_val;
         *fv = get_thickness();
      }
      break;
      case SET_AXIS_MAJ_THICKNESS:
      {
         float *fv = (float*)out_val;
         *fv = get_major_thickness();
      }
      break;
      case SET_AXIS_MIN_THICKNESS:
      {
         float *fv = (float*)out_val;
         *fv = get_minor_thickness();
      }
      break;
      case SET_AXIS_LABEL_ANGLE:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_label_angle();
      }
      break;
      case SET_AXIS_XLAB_ANG:
      {
         if (axis_alg->get_orientation ()  == X_AXIS)
         {
            float* fv = static_cast<float*>(out_val);
            *fv = get_label_angle();
         }
      }
      break;
      case SET_AXIS_YLAB_ANG:
      {
         if (axis_alg->get_orientation ()  == Y_AXIS)
         {
            float* fv = static_cast<float*>(out_val);
            *fv = get_label_angle();
         }
      }
      break;
      case SET_AXIS_LABEL_HALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_label_halign();
      }
      break;
      case SET_AXIS_LABEL_VALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_label_valign();
      }
      break;
      case SET_AXIS_TEXT:
      {
         string *sval =  (string*)out_val;
         *sval = get_label_text();
      }
      break;
      case SET_AXIS_XLAB_TXT:
      {
         if (axis_alg->get_orientation() == X_AXIS)
         {
            string *sval = (string*)out_val;
            *sval = get_label_text();
         }
      }
      break;
      case SET_AXIS_YLAB_TXT:
      {
         if (axis_alg->get_orientation() == Y_AXIS)
         {
            string *sval =  (string*)out_val;
            *sval = get_label_text();
         }
      }
      break;
      case SET_AXIS_DEPTH:
      {
         int *iv = (int*)out_val;
         *iv = get_depth();
      }
      break;
      case SET_AXIS_TICKSTYLE:
      {
         string *sval =  (string*)out_val;
         if (major_tickstyle == minor_tickstyle)
            *sval = v_tickstyle_to_string(major_tickstyle);
         else
            *sval = "custom"; 
      }
      break;
      case SET_AXIS_MAJ_STYLE:
      {
         string *sval =  (string*)out_val;
         *sval = v_tickstyle_to_string(get_major_tickstyle());
      }
      break;
      case SET_AXIS_MIN_STYLE:
      {
         string *sval =  (string*)out_val;
         *sval = v_tickstyle_to_string(get_minor_tickstyle());
      }
      break;
      case SET_AXIS_TICKFORMAT:
      {
         string *sval =  (string*)out_val;
         *sval = get_tickformat();
      }
      break;
      case SET_AXIS_TICKLABEL_STYLE:
      {
         string *sval =  (string*)out_val;
         *sval = v_tickstyle_to_string(get_ticklabelstyle());
      }
      break;
      case SET_AXIS_MAJOR_NUM:
      {
         int *iv = (int*)out_val;
         *iv = get_major_num();
      }
      break;
      case SET_AXIS_MINOR_NUM:
      {
         int *iv = (int*)out_val;
         *iv = get_minor_num();
      }
      break;
      case SET_AXIS_MAJORINT:
      {
         double* dv = static_cast<double*>(out_val);
         *dv = get_major_int();
      }
      break;
      case SET_AXIS_MINORINT:
      {
         double* dv = static_cast<double*>(out_val);
         *dv = get_minor_int();
      }
      break;
      case SET_AXIS_PAD:
      {
         double* dv = static_cast<double*>(out_val);
         *dv = get_pad();
      }
      break;
      case SET_AXIS_LOG_MIN:
      {
         double* dv = static_cast<double*>(out_val);
         *dv = get_log_minimum();
      }
      break;
      case SET_AXIS_MAJORLENGTH:
      {
         int *iv = (int*)out_val;
         *iv = get_major_len();
      }
      break;
      case SET_AXIS_MINORLENGTH:
      {
         int *iv = (int*)out_val;
         *iv = get_minor_len();
      }
      break;
      case SET_AXIS_MAJ_MODE:
      {
         string *sval =  (string*)out_val;
         *sval = v_tickmode_to_string(get_major_mode());
      }
      break;
      case SET_AXIS_MIN_MODE:
      {
         string *sval =  (string*)out_val;
         *sval = v_tickmode_to_string(get_minor_mode());
      }
      break;
      case SET_AXIS_OFFSET_PAR:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_parallel_offset();
      }
      break;
      case SET_AXIS_OFFSET_PERP:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_perpendicular_offset();
      }
      break;
      case SET_AXIS_LABEL_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_label_color();
      }
      break;
      case SET_AXIS_LABEL_VIS:
      {
         int *iv = (int*)out_val;
         *iv = get_label_visible();
      }
      break;
      case SET_AXIS_TICKLABEL_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_ticklabel_color();
      }
      break;
      case SET_AXIS_TICKLABEL_OFFSET:
      {
         int *iv = (int*)out_val;
         *iv = get_ticklabel_offset();
      }
      break;
      case SET_AXIS_TICKLABEL_SIZE:
      {
         int *iv = (int*)out_val;
         *iv = get_ticklabel_size();
      }
      break;
      case SET_AXIS_TICKLABEL_ANGLE:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_ticklabel_angle();
      }
      break;
      case SET_AXIS_TICKLABEL_HALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_ticklabel_halign();
      }
      break;
      case SET_AXIS_TICKLABEL_VALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_ticklabel_valign();
      }
      break;
      case SET_AXIS_TICKLABEL_VIS:
      {
         int *iv = (int*)out_val;
         *iv = get_ticklabel_visible();
      }
      break;
      case SET_AXIS_TICKLABEL_FONT:
      {
         string *sval =  (string*)out_val;
         *sval = get_ticklabel_font();
      }
      break;
      case SET_AXIS_TICKLABEL_FNTSTL:
      {
         string *sval =  (string*)out_val;
         *sval = v_fontstyle_to_string(get_ticklabel_fontstyle());
      }
      break; 
      case SET_AXIS_LABEL_SIZE:
      {
         int *iv = (int*)out_val;
         *iv = get_label_size();
      }
      break;
      case SET_AXIS_LABEL_FONT:
      {
         string *sval =  (string*)out_val;
         *sval = get_label_font();
      }
      break;
      case SET_AXIS_LABEL_FONTSTYLE:
      {
         string *sval =  (string*)out_val;
         *sval = v_fontstyle_to_string(get_label_fontstyle());
      }
      break;
      case SET_AXIS_MAJGRD_VIS:
      {
         int *iv = (int*)out_val;
         *iv = get_major_grid_visible();
      }
      break;
      case SET_AXIS_MAJGRD_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_major_grid_color();
      }
      break;
      case SET_AXIS_MAJGRD_THICK:
      {
         float *fv = (float*)out_val;
         *fv = get_major_grid_thickness();
      }
      break;
      case SET_AXIS_MAJGRD_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_major_grid_style();
      }
      break;
      case SET_AXIS_MINGRD_VIS:
      {
         int *iv = (int*)out_val;
         *iv = get_minor_grid_visible();
      }
      break;
      case SET_AXIS_MINGRD_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_minor_grid_color();
      }
      break;
      case SET_AXIS_MINGRD_THICK:
      {
         float *fv = (float*)out_val;
         *fv = get_minor_grid_thickness();
      }
      break;
      case SET_AXIS_MINGRD_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_minor_grid_style();
      }
      break;
      case SET_AXIS_AUTOMIN:
      {
         int *iv = (int*)out_val;
         *iv = get_automin();
      }
      break;
      case SET_AXIS_AUTOMAX:
      {
         int *iv = (int*)out_val;
         *iv = get_automax();
      }
      break;
      case SET_AXIS_MAJ_VIS:
      {
         int *iv = (int*)out_val;
         *iv = major_visible();
      }
      break;
      case SET_AXIS_MIN_VIS:
      {
         int *iv = (int*)out_val;
         *iv = minor_visible();
      }
      break;
      case SET_AXIS_VISIBLE:
      {
         int *iv = (int*)out_val;
         *iv = is_visible();
      }
      break;
      case SET_AXIS_SCALE:  // fallthrough intended
      case GET_AXIS_SCALE:
      {
         int *iv = (int*)out_val;
         *iv = get_scaletype();
      }
      break;
      case SET_AXIS_REVERSE:
      {
         int *iv = (int*)out_val;
         *iv = get_reverse();
      }
      break;
      case GET_AXIS_RANGE:
      {
         double *dv =  (double*)out_val;
         dv[0] = this->get_min ();
         dv[1] = this->get_max ();
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }
   return rval;
}


int   ChipsAxis::set_log_minimum(double val)
{
   int rval = ERR_STATUS_GOOD;
   ChipsCartesianAxisAlgorithm *cart_alg;
   cart_alg = dynamic_cast <ChipsCartesianAxisAlgorithm *> (axis_alg);
   if (cart_alg)
   {
      if (val > 0.0)
      {
         cart_alg->set_log_minimum (val);
   
         if (initialized)  rval = change_log_minimum();
      }
      else 
         cmd_manager->add_err (rval=ERR_NO_FLT_NEGATIVE,id, 0, val);
   }
   else 
      rval = ERR_STATUS_BAD;
   return rval;
}


int   ChipsAxis::set_pad(double val)
{
   int rval = ERR_STATUS_GOOD;

   if (val >= 0.0)
   {
      pad = val;
/*
      if (initialized && (automin || automax))
         update_tick_limits ();
*/
   }
   else
      cmd_manager->add_err (rval=ERR_NO_FLT_NEGATIVE,id, 0, val);
   return rval;
}


void ChipsAxis::set_automin(bool val)
{
   if (val != automin)
   {
      automin = val;
/*
      if (initialized)
         // for modes that utilize pad- recalculate padding
         update_tick_limits ();
*/
   }
}


void ChipsAxis::set_automax(bool val)
{
   if (val != automax)
   {
      automax = val;
/*
      if (initialized)
      {
         // for modes that utilize pad- recalculate padding
         update_tick_limits ();
      }
*/
   }
}


int ChipsAxis::set_tick_positions(unsigned int arraysize, double* pos,
   string* vals, bool rec)
{
   int rval = ERR_STATUS_BAD;

   if ((arraysize>0) && pos)
   {
      rval = ERR_STATUS_GOOD;
      ChipsAxisTickCommand* cmd = new ChipsAxisTickCommand(this, 
         axis_alg->get_arbitrary_count (), 
         axis_alg->get_arbitrary_positions (),
         axis_alg->get_arbitrary_text (),
         arraysize, pos, vals);
      axis_alg->set_arbitrary_positions (pos, vals, arraysize);

      if (initialized && (rval == ERR_STATUS_GOOD)) 
      {
/***
         if (rec)
            if (!add_command (cmd, rval))  delete cmd;
         if (major_mode == chips_arbitrary)  
            rval = set_major_mode(chips_arbitrary);
***/
         if (rec)
         {
            if (!add_command (cmd, rval)) delete cmd;
            rval = set_property((void*)"arbitrary", SET_AXIS_MAJ_MODE,rec);
         }
         
         if (!rval)
            update_tick_positions ();
      }
      if (!rec)  delete cmd;
   }
   else 
      cmd_manager->add_err ("No arbitrary tick position specified.",id, 0);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_AXS);


   return rval;
}

void ChipsAxis::get_tick_positions(unsigned int &arraysize, double** pos,
                                 string** text)
{
   double *tick_pos;
   string *tick_text;

   if (!pos || !text)
      return;

   tick_pos = axis_alg->get_tick_positions (true);
   arraysize = axis_alg->get_tick_text_count ();

   if (arraysize > 0)
   {
      tick_text = axis_alg->get_tick_text ();

      *pos = new double[arraysize];
      *text = new string[arraysize];

      for (int ii=arraysize; ii--;)
      {
         (*pos)[ii] = tick_pos[ii];
         (*text)[ii] = tick_text[ii];
      }
   }
   else
   {
      arraysize = axis_alg->get_tick_count ();
      *pos = new double[arraysize];
      *text = new string[arraysize];

      for (int ii=arraysize; ii--;)
      {
         (*pos)[ii] = tick_pos[ii];
         axis_alg->format_point (tickformat, tick_pos[ii], (*text)[ii]);
      }
   }
}

void ChipsAxis::get_arbitrary_tick_positions(unsigned int &arraysize, double** pos,
                                 string** text)
{
   double *tick_pos;
   string *tick_text;

   if (!pos || !text)
      return;

   tick_pos = axis_alg->get_arbitrary_positions ();
   tick_text = axis_alg->get_arbitrary_text ();
   arraysize = axis_alg->get_arbitrary_count ();

   *pos = new double[arraysize];
   *text = new string[arraysize];

   for (int ii=arraysize; ii--;)
   {
      (*pos)[ii] = tick_pos[ii];
      if (tick_text)
         (*text)[ii] = tick_text[ii];
      else
         axis_alg->format_point (tickformat, tick_pos[ii], (*text)[ii]);
   }

}


string  ChipsAxis::get_label_text()
{
   if (label)
      return label->get_text();
   else
      return temp_label;
}

string  ChipsAxis::get_label_color()
{
   if (label)
      return label->get_color();
   else
      return temp_label;
}


int ChipsAxis::get_label_visible()
{
   if (label)
      return label->is_visible();
   else
      return labvis;
}


string  ChipsAxis::get_label_font()
{
   if (label)
      return label->get_font();
   else
      return temp_label;
}


V_FONTSTYLE_TYPE ChipsAxis::get_label_fontstyle()
{
   if (label)
      return label->get_fontstyle();
   else
      return label_fontstyle;
}


int ChipsAxis::get_label_size()
{
   if (label)
      return label->get_size();
   else
      return label_size;
}


int  ChipsAxis::set_label_text(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
      label->set_text(val);
   else if (val.length())
      temp_label = val; 
   else  
      cmd_manager->add_err ("Axis label text should have a length greater than zero", rval=ERR_INT_MIN,id, 0);
   return rval;
}


int  ChipsAxis::set_label_angle(float val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
      label->set_angle(val);
   else 
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "angle");
   return rval;
}


float ChipsAxis::get_label_angle()
{
   float fval = 0;
   if (label)   fval = label->get_angle();
   return fval;
}


int  ChipsAxis::set_label_halign(float val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if (val == AUTO_ALIGN)
         label->set_horizalignment(0.5);
      else
         label->set_horizalignment(val);
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "halign");
   return rval;
}


float ChipsAxis::get_label_halign()
{
   float fval = 0;
   if (label)   fval = label->get_horizalignment();
   return fval;
}


int  ChipsAxis::set_label_valign(float val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if (val == AUTO_ALIGN)
         label->set_vertalignment(0.0);
      else
         label->set_vertalignment(val);
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "valign");
   return rval;
}


float ChipsAxis::get_label_valign()
{
   float fval = 0;
   if (label)   fval = label->get_vertalignment();
   return fval;
}


int ChipsAxis::set_label_visible(bool viz)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if (viz)  rval = label->show();
      else      rval = label->hide(); 
      labvis = viz;
      if (initialized)  rval = change_title_visible();
   }
   else   
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "visible");

   return rval;
}


int ChipsAxis::set_label_color(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      rval = label->set_color(val);
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "color");

   return rval;
}

int ChipsAxis::set_label_font(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      rval = label->set_font(val);
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "font");

   return rval;
}


int ChipsAxis::set_label_size(int sz)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if ((rval = label->set_size(sz)) == ERR_STATUS_GOOD)
         label_size = sz;
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "size");

   return rval;
}


int ChipsAxis::set_label_fontstyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if ((rval = c2_string_to_fontstyle(val.c_str(), &label_fontstyle))
          == ERR_STATUS_GOOD)
      {
         rval = label->set_fontstyle(label_fontstyle);
      }
   }
   else
      cmd_manager->add_err(rval=ERR_AXIS_LABELS_DNE, id, -1, "fontstyle");
   
   return rval;
}


int ChipsAxis::set_parallel_offset(float in_offset)
{
   int rval = ERR_STATUS_GOOD;

   if (in_offset != offset_par)
   {
      if ((in_offset >= V_MIN_PAR_OFFSET)  && (in_offset <= V_MAX_PAR_OFFSET))
      { 
         offset_par = in_offset;
         if (initialized)  rval = change_parallel_offset();
      }
      else
         cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis parallel offset", in_offset, V_MIN_PAR_OFFSET, V_MAX_PAR_OFFSET);
 
   }
   return rval; 
}


int ChipsAxis::set_perpendicular_offset(float in_offset)
{
   int rval = ERR_STATUS_GOOD;

   if (in_offset != offset_perp)
   {
      if ((in_offset >= V_MIN_PERP_OFFSET)  && (in_offset <= V_MAX_PERP_OFFSET))
      {
         offset_perp = in_offset;
         if (initialized)  rval = change_perpendicular_offset();
      }
      else 
         cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis perpendicular offset", in_offset, (float)V_MIN_PERP_OFFSET, (float)V_MAX_PERP_OFFSET);
   }
   return rval; 
}



int ChipsAxis::set_title_flip(bool inFlag)
{
  
  int rval = ERR_STATUS_GOOD;
  
  if (inFlag != title_flip)
    {
      title_flip=inFlag;
      if (initialized)  rval = change_title_flip();
   }
   return rval;  

};



int ChipsAxis::adjust_logmin(double minval, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if ((minval > 0.0) && (axis_alg->get_absolute_min (V_SCALE_LOG) != minval))
   {
      automin_used = false;
      ChipsLimitsCommand* cmd = NULL;
      if (rec)
         cmd = new ChipsLimitsCommand(this, 2, axis_alg->get_absolute_min (V_SCALE_LINEAR), axis_alg->get_absolute_min (V_SCALE_LOG), minval);
      axis_alg->set_absolute_min (minval, V_SCALE_LOG);
      if (initialized && (axis_alg->get_scaletype () == V_SCALE_LOG))
      {
/*         rval = update_tick_limits ();
*/
	 if (rval == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
	 }

         if (rval != ERR_STATUS_GOOD && cmd)
	    delete cmd;
      }
      else if (cmd)  delete cmd; 
   } 
   return rval; 
}


int ChipsAxis::adjust_autologmin(double minval, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (automin && (minval > 0.0) && (minval < axis_alg->get_absolute_min (V_SCALE_LOG)))
   {
      ChipsLimitsCommand* cmd = NULL;
      if (rec)
         cmd = new ChipsLimitsCommand(this, 1, axis_alg->get_absolute_min (V_SCALE_LINEAR), axis_alg->get_absolute_min (V_SCALE_LOG), minval);
      axis_alg->set_absolute_min (minval);
      automin_used = true;
      if (initialized && (axis_alg->get_scaletype () == V_SCALE_LOG))
      {
/*         rval = update_tick_limits ();
*/
	 if (rval == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
	 }
	 if (rval != ERR_STATUS_GOOD && cmd)
            delete cmd;
      }
      else if (cmd)   delete cmd;
   }
   return rval;
}


int ChipsAxis::adjust_min(double minval, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (axis_alg->get_min () != minval)
   {
      ChipsLimitsCommand* cmd = NULL;
      automin_used = false;
      if (rec)
         cmd = new ChipsLimitsCommand(this, 4, axis_alg->get_absolute_min (V_SCALE_LINEAR), axis_alg->get_absolute_min (V_SCALE_LOG), minval);
      axis_alg->set_absolute_min (minval);
      if (initialized)
      {
/*         rval = update_tick_limits ();
*/
	 if (rval == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
	 }
	 if (rval != ERR_STATUS_GOOD && cmd)
            delete cmd;
      } 
      else if (cmd)  delete cmd;
   } 
   return rval; 
}


int ChipsAxis::adjust_max(double maxval, bool rec)
{
   int rval = ERR_STATUS_GOOD; 
   ChipsLimitsCommand* cmd = NULL;
   if (axis_alg->get_max () != maxval)
   {
      automax_used = false;
      if (rec)
         cmd = new ChipsLimitsCommand(this, 8, 
                axis_alg->get_absolute_max (V_SCALE_LINEAR), 
                axis_alg->get_absolute_max (V_SCALE_LOG),
                maxval);
      axis_alg->set_absolute_max (maxval);
      if (initialized)
      {
/*         rval = update_tick_limits ();
*/
	 if (rval == ERR_STATUS_GOOD)
         {
            if (rec)
              if (!add_command (cmd, rval))  delete cmd;
	 }
	 if ((rval != ERR_STATUS_GOOD) && rec)
            delete cmd;
      }
      else if (rec)  delete cmd;
   }
   return rval;
}
  
int ChipsAxis::adjust_minmax(double minval, double maxval, int autos, bool rec, bool transformed)
{
   int rval = ERR_STATUS_GOOD;
   bool changed = false; 
   bool reversed = false;
   bool delete_cmd = true;

   reversed = axis_alg->fix_minmax (minval, maxval, transformed);

   int of = AUTOLIM_DEF;
   if (automax)  of |= AUTOMAX_ON; 
   if (automin)  of |= AUTOMIN_ON; 
   if (!(automin || automax))  of = AUTOLIM_OFF;
   switch(autos)
   {
      case AUTOLIM_DEF:     break;
      case AUTOLIM_OFF:     automin = automax = false;  break;
      case AUTOMIN_ON:      automin = true;  automax = false;  break;
      case AUTOMAX_ON:      automin = false; automax = true;   break; 
      case AUTOLIM_ON:      automin = automax = true;   break; 
      default:                  autos = 0;  break;
   }   
   automin_used = automin;
   automax_used = automax;

   ChipsLimitsCommand* cmd = new ChipsLimitsCommand(this, 16, 
       axis_alg->get_absolute_min (V_SCALE_LINEAR, transformed), 
       axis_alg->get_absolute_min (V_SCALE_LOG, transformed),
       axis_alg->get_absolute_max (V_SCALE_LINEAR, transformed),
       axis_alg->get_absolute_max (V_SCALE_LOG, transformed),
       of, autos, minval, maxval, transformed);

   // if automin/max settings have changed- padding may need to be adjusted
   changed = (autos != of);

   if ((axis_alg->get_absolute_min (V_SCALE_LINEAR, transformed) != minval) ||
       (axis_alg->get_absolute_max (V_SCALE_LINEAR, transformed) != maxval))
      changed = true; 
   axis_alg->set_absolute_minmax (minval, maxval, V_SCALE_LINEAR, transformed);
   if (minval > 0) axis_alg->set_absolute_min (minval, V_SCALE_LOG, transformed);
   if (maxval > 0) axis_alg->set_absolute_max (maxval, V_SCALE_LOG, transformed);

   if (changed && initialized)
   {
/*      update_tick_limits ();
*/
      if (rval == ERR_STATUS_GOOD)
      {
         if (rec && add_command (cmd, rval))
            delete_cmd = false;
      }
   }

   if (delete_cmd && cmd)
      delete cmd;
   return rval; 
}

int ChipsAxis::adjust_from_aspect(double minval, double maxval)
{
   bool changed;
   
   if ((minval != axis_alg->get_min ()) || (axis_alg->get_absolute_min () != minval))
   {
      axis_alg->set_absolute_min (minval, V_SCALE_LINEAR);
      if (minval > 0) axis_alg->set_absolute_min (minval, V_SCALE_LOG);
      changed = true;
      automin_used = false;
   }
   if ((maxval != axis_alg->get_max ()) || (axis_alg->get_absolute_max () != maxval))
   {
      axis_alg->set_absolute_max (maxval);
      changed = true;
      automax_used = false;
   }

/*   rval = update_tick_limits ();
*/
   if (reverse)   change_reverse();

   return ERR_STATUS_GOOD;
}

int ChipsAxis::adjust_autominmax(double minval, double maxval, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   bool changed = false;

   if (minval > maxval)
   {
      // adjust to ensure min is less than max
      double tmp = maxval;
      maxval = minval;
      minval = tmp;
   }
   else if (minval == maxval && minval == axis_alg->get_min () && maxval == axis_alg->get_max ())
   {
      // adjust to ensure range is not 1 point
      if (minval != 0)
      {
         minval *= 0.95;
         maxval *= 1.05;
      }
      else
      {
         minval = -1.0;
         maxval = 1.0;
      }
   }

   int of = AUTOLIM_DEF;
   if (automax)  of |= AUTOMAX_ON; 
   if (automin)  of |= AUTOMIN_ON; 
   if (!(automin || automax))  of = AUTOLIM_OFF;
   ChipsLimitsCommand* cmd = new ChipsLimitsCommand(this, 32, 
       axis_alg->get_absolute_min (V_SCALE_LINEAR),  
       axis_alg->get_absolute_min (V_SCALE_LOG),
       axis_alg->get_absolute_max (V_SCALE_LINEAR),
       axis_alg->get_absolute_max (V_SCALE_LOG),
       of, of, minval, maxval, true);

if ((minval != axis_alg->get_min ()) || (axis_alg->get_absolute_min () != minval))
   {
      axis_alg->set_absolute_min (minval, V_SCALE_LINEAR);
      if (minval > 0) axis_alg->set_absolute_min (minval, V_SCALE_LOG);
      changed = true;
      automin_used = false;
   }
   if ((maxval != axis_alg->get_max ()) || (axis_alg->get_absolute_max () != maxval))
   {
      axis_alg->set_absolute_max (maxval);
      changed = true;
      automax_used = false;
   }

   if (automin)
   {
      if (axis_alg->get_absolute_min (V_SCALE_LINEAR) != minval)
      {
         axis_alg->set_absolute_min (minval, V_SCALE_LINEAR);
         if (minval > 0) axis_alg->set_absolute_min (minval, V_SCALE_LOG);
         changed = true;
         automin_used = false;
      }
   }
   if (automax)
   {
      if (axis_alg->get_absolute_max (V_SCALE_LINEAR) != minval)
      {
         axis_alg->set_absolute_max (maxval, V_SCALE_LINEAR);
         if (maxval > 0) axis_alg->set_absolute_max (maxval, V_SCALE_LOG);
         changed = true;
         automin_used = false;
      }
   }

   if (changed && initialized)
   {
/*      rval = update_tick_limits ();
*/
      if (rval == ERR_STATUS_GOOD)
      {
         if (rec)
           if (!add_command (cmd, rval))  delete cmd;
      }

      if (rval != ERR_STATUS_GOOD)  delete cmd;
   }
   else  delete cmd;
   return rval;
}


void ChipsAxis::check_minmax (bool &reversed)
{
   double tmp_min = axis_alg->get_min ();
   double tmp_max = axis_alg->get_max ();
   reversed = axis_alg->fix_minmax (tmp_min, tmp_max, true);

}

int ChipsAxis::reverse_axis()
{
   int rval = ERR_STATUS_GOOD;
   reverse = !reverse;
   if (initialized)  rval = change_reverse();


   return rval;
}


int ChipsAxis::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  change_color();
   return rval;
}


int ChipsAxis::set_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   // need to determine thickness of line
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis thickness", val, V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      thickness = val;
      if (initialized)  rval = change_thickness();
   }
   return rval;
}


int ChipsAxis::set_major_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   // need to determine thickness of major ticks
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis major tick thickness", val, V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      major_thickness = val;
      if (initialized)  rval = change_major_thickness();
   }
   return rval;
}


int ChipsAxis::set_minor_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   // need to determine thickness of minor ticks
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis minor tick thickness", val, V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      minor_thickness = val;
      if (initialized)  rval = change_minor_thickness();
   }
   return rval;
}



int ChipsAxis::set_major_mode(string inmode)
{
   V_TICKMODE_TYPE tval;
   int rval;
   if ((rval = c2_string_to_tickmode(inmode.c_str(), &tval)) == ERR_STATUS_GOOD)
      rval = set_major_mode(tval);
   return rval;
}

int ChipsAxis::set_major_mode(V_TICKMODE_TYPE inmode)
{
   int rval = ERR_STATUS_GOOD;
   rval = axis_alg->set_major_mode (inmode);
   if (initialized && !rval)
   {
      if (inmode == chips_arbitrary)
      {
         if (axis_alg->get_arbitrary_count () == 0)
            cmd_manager->add_err(rval=ERR_ARBTICKPOS_DNE, id, 0);
      }
/*      update_tick_limits ();
*/
      if (rval == ERR_STATUS_GOOD)
         rval = change_major_mode();
   }
   return rval;
}


int ChipsAxis::set_minor_mode(string inmode)
{
   V_TICKMODE_TYPE tval;
   int rval;

   if ((rval = c2_string_to_tickmode(inmode.c_str(), &tval)) == ERR_STATUS_GOOD)
   {
      if ((tval == chips_count) || (tval == chips_interval) || 
          (tval == chips_nice))
      {
          rval = axis_alg->set_minor_mode (tval);
      }
      if (initialized && !rval)  rval = change_minor_mode();
   }
   return rval;
}



int ChipsAxis::set_major_tickstyle(V_TICKSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_tickstyle_type(val))
   {
      major_tickstyle = val; 
      set_use_major_tickstyle();
      if (initialized)  rval = change_major_style(); 
   }
   else
   {
      stringstream errstr;
      errstr << "The axis major tickstyle value (" << val << ") must be one of [inside|outside|centered]";
      cmd_manager->add_err(errstr.str ());
   }


   return rval;
}


int ChipsAxis::set_minor_tickstyle(V_TICKSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_tickstyle_type(val))
   {
      minor_tickstyle = val;
      set_use_minor_tickstyle();
      if (initialized)  rval = change_minor_style();
   }
   else
   {
      stringstream errstr;
      errstr << "The axis minor tickstyle value (" << val << ") must be one of [chips_inside|chips_outside|chips_centered]";
      cmd_manager->add_err(errstr.str (), id, 0);
   }
   return rval;
}


int ChipsAxis::set_ticklabelstyle(V_TICKSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val == chips_inside) || (val == chips_outside))
   {
      ticklabstyle = val;
      if (initialized)
      {
         set_ticklabel_halign(ticklabhorizalignment);
         set_ticklabel_valign(ticklabvertalignment);
         rval = change_ticklabel_style();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "tick label style", (int)val, "[chips_inside, chips_outside]");
   return rval;
}


int  ChipsAxis::set_ticklabel_angle(float val)
{
   int rval = ERR_STATUS_GOOD;

   ticklabangle = val;
   if (initialized)  rval = change_ticklabel_angle();
   return rval;
}


float ChipsAxis::get_ticklabel_angle()
{
   return ticklabangle;
}


int  ChipsAxis::set_ticklabel_halign(float val)
{
   int rval = ERR_STATUS_GOOD;
   use_halign = ticklabhorizalignment = val;
   if (ticklabhorizalignment == AUTO_ALIGN)
   {
      if (axis_alg->get_orientation ()  == X_AXIS)
         use_halign = 0.5; 
      else if ((title_flip && (ticklabstyle == chips_inside)) || 
                (!title_flip && (ticklabstyle == chips_outside))) 
          use_halign = 0.0;
      else 
          use_halign = 1.0;
   }
   if (initialized)  rval = change_ticklabel_alignment();
   return rval;
}


float ChipsAxis::get_ticklabel_halign(bool computed)
{
   if (computed)
      return use_halign;

   return ticklabhorizalignment;
}


int  ChipsAxis::set_ticklabel_valign(float val)
{
   int rval = ERR_STATUS_GOOD;
   use_valign = ticklabvertalignment = val;
   if (ticklabvertalignment == AUTO_ALIGN) 
   {
      if (axis_alg->get_orientation ()  != X_AXIS)
         use_valign = 0.5;
       else if ((title_flip && (ticklabstyle == chips_inside)) || 
                (!title_flip && (ticklabstyle == chips_outside))) 
          use_valign = 1.0;
      else 
          use_valign = 0.0;
   }
   if (initialized)  rval = change_ticklabel_alignment();
   return rval;
}


float ChipsAxis::get_ticklabel_valign(bool computed)
{
   if (computed)
      return use_valign;
   return ticklabvertalignment;
}


int ChipsAxis::set_ticklabel_color(string val)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = ticklabcolor.set_foreground_color(val)) == ERR_STATUS_GOOD)
   {
      if (initialized)
      {
         update_tick_text();
	 change_ticklabel_color();
      }
   }
   return rval;
}

string  ChipsAxis::get_ticklabel_color()
{
   return ticklabcolor.get_name(); 
}


int ChipsAxis::set_ticklabel_font(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (val.length())
   {
      if (theFonts && theFonts->font_exists(val))
      {
         string tmpfont = theFonts->get_path(val, ticklabfontstyle); 
         if (tmpfont.length())
         {
            ticklabfontfile = tmpfont;
            ticklabfont = val;
            if (initialized)
	    {
               update_tick_text(); 
	       rval = change_ticklabel_font();
            }
         }
      }
      else
         cmd_manager->add_err(rval=ERR_STR_SUBSET, id, 0, "font", val.c_str (), "helvetica|courier|times|greek|user loaded fonts");
   }
   else
      cmd_manager->add_err(rval=ERR_STR_SUBSET, id, 0, "font", val.c_str (), "helvetica|courier|times|greek|user loaded fonts");

   return rval;
}


string  ChipsAxis::get_ticklabel_font()
{
   return ticklabfont;
}

int ChipsAxis::set_ticklabel_size(int sz)
{
   int rval = ERR_STATUS_GOOD;

   if (sz >= 0)
   {
      ticklabsize = sz;
      if (initialized)
      {
         update_tick_text ();
         rval = change_ticklabel_size();
      }
   }
   else
      cmd_manager->add_err(rval=ERR_NO_INT_NEGATIVE, id, 0, sz);
   return rval;
}


int ChipsAxis::set_ticklabel_visible(bool viz)
{
   int rval = ERR_STATUS_GOOD;
   ticklabvis = viz;
   if (initialized)  rval = change_ticklabel_visible();
   return rval;
}


int ChipsAxis::get_ticklabel_size()
{
   return ticklabsize;
}


int ChipsAxis::set_ticklabel_fontstyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (val.length())
   {
      V_FONTSTYLE_TYPE tmp;
      if ((rval = c2_string_to_fontstyle(val.c_str(), &tmp)) == ERR_STATUS_GOOD)
      {
         ticklabfontstyle = tmp; 
         set_ticklabel_font(ticklabfont); 
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_STR_SUBSET, id, 0, "fontstyle", val.c_str (), "normal|bold|italic|bolditalic");

   return rval;
}


V_FONTSTYLE_TYPE ChipsAxis::get_ticklabel_fontstyle()
{
   return ticklabfontstyle;
}


int ChipsAxis::set_ticklabel_offset(int sz)
{
   int rval = ERR_STATUS_GOOD;

   if (sz >= 0)
   {
      ticklaboffset = sz;
      if (initialized)  rval = change_ticklabel_offset();
   }
   else
      cmd_manager->add_err(rval=ERR_NO_INT_NEGATIVE, id, 0, sz);
   
   return rval;
}


int ChipsAxis::get_ticklabel_offset()
{
   return ticklaboffset;
}


int ChipsAxis::set_minor_int(double val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > 0.0)
   {
      if (val <= axis_alg->get_major_interval ())
      {
         axis_alg->set_minor_interval (val);
         if (initialized)  rval = change_minor_interval();
      }
      else
         cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "axis minor tick interval", val, 0.0, axis_alg->get_major_interval ());
   }
   else 
      cmd_manager->add_err (rval=ERR_NO_FLT_NEGATIVE,id, 0, val);

   return rval; 
}


int ChipsAxis::set_major_int(double val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > 0.0)
   {
      axis_alg->set_major_interval (val);
/*
      if (initialized)  
      {
         if (axis_alg->get_major_mode () == chips_interval)
         {
            rval = update_tick_limits ();
         } 
      }
*/
   }
   else
      cmd_manager->add_err (rval=ERR_NO_FLT_NEGATIVE,id, 0, val);
   
   return rval; 
}

int ChipsAxis::set_tickformat(string val)
{
   int rval = ERR_STATUS_GOOD;

   if (val.length())
   {
      if (tickformat != val)
      {
         bool sg = is_sexigesimal_format(tickformat); 
         tickformat = val;
         axis_alg->reset_string_labels (tickformat);
         update_tick_text ();

         if (initialized)
         {
            update_tick_positions();
            rval = change_tick_format();
            if (is_sexigesimal_format(tickformat) != sg)
            {
               // changed between ra/dec and numeric- check grid positions
               set_major_grid_visible(grid_maj_vis);
               set_minor_grid_visible(grid_min_vis);
            } 
         }
      }
   }
   else
      cmd_manager->add_err ("No tickformat was specified. The format should be from the subset count|interval|nice|limits|arbitrary]",id, 0);
   return rval;
}


// adjust number of major ticks 
int ChipsAxis::set_major_num(int val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > 1) 
   {
      axis_alg->set_major_num (val);

      if (initialized)
      {
         // if we're calculating tick locations, adjust them accordingly 
/*         if (val == chips_count)
         {
            rval = update_tick_limits ();
         }
*/
	 if (rval == ERR_STATUS_GOOD)
            rval = change_major_num(); 
      }
   }
   else
      cmd_manager->add_err (rval=ERR_INT_GTE,id,0,"axis major num", val, 2);
   return rval;
}


// adjust number of minor ticks
int ChipsAxis::set_minor_num(int val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > -1 && val < 101) 
   {
      axis_alg->set_minor_num (val);

      // minor_int may need to be adjusted
      if (initialized)  rval = change_minor_num();
   }
   else
      cmd_manager->add_err (rval=ERR_NO_INT_NEGATIVE,id, 0, val);
   return rval;
}


// set major tick color
int ChipsAxis::set_major_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = major_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_major_color();
   return rval;
}


// set minor tick color
int ChipsAxis::set_minor_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = minor_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_minor_color();
   return rval;
}



int ChipsAxis::set_major_len(int val)
{
   int rval = ERR_STATUS_GOOD;

   if (val >= 0)
   {
      major_len = val;
      if (initialized)  rval =  change_major_length();
   }
   else
      cmd_manager->add_err (rval=ERR_NO_INT_NEGATIVE,id, 0, val);
   return rval;
}

int ChipsAxis::set_minor_len(int val)
{
   int rval = ERR_STATUS_GOOD;

   if (val >= 0)
   {
      minor_len = val;
      if (initialized)  rval = change_minor_length();
   }
   else
      cmd_manager->add_err (rval=ERR_NO_INT_NEGATIVE,id, 0, val);
   return rval;
}


int ChipsAxis::set_major_grid_visible(bool val)
{ 
   if ((grid_use_vis = val) == true)
   {
      grid_maj_vis = true; 
      grid_maj_clr.set_foreground_color(grid_use_clr.get_name());
      grid_maj_style = grid_use_style;
      grid_maj_thick = grid_use_thick;
   }
   else if ((grid_maj_vis = grid_min_vis) == true)  
   {
      grid_maj_clr.set_foreground_color(grid_min_clr.get_name());
      grid_maj_style = grid_min_style;
      grid_maj_thick = grid_min_thick;
   }

   if (grid_maj_vis == true)
   {
      int line_count = 0;
      int slices = 0;
      double ***grid = axis_alg->calculate_grid (line_count, slices, 0);
      set_major_grid (grid, slices, line_count);
   }
   
   return change_grid_property();
}

// set major grid color
int ChipsAxis::set_major_grid_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = grid_use_clr.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
   {
      if (grid_use_vis)  
         rval = grid_maj_clr.set_foreground_color(in_color); 
      if ((rval == ERR_STATUS_GOOD) && initialized)  
         rval = change_grid_property();
   }
   return rval;
}

// set major grid thickness
int ChipsAxis::set_major_grid_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "major grid thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      grid_use_thick = val;
      if (grid_use_vis)   grid_maj_thick = val;
      if (initialized)   rval = change_grid_property();
   }
   return rval;
}

// set major grid style
int ChipsAxis::set_major_grid_style(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;
   if (valid_line_type(val))
   {
      grid_use_style = val;
      if (grid_use_vis)   grid_maj_style = val;
       else  grid_maj_style = chips_noline;
      if (initialized)   rval = change_grid_property();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val,"chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");

   return rval;
}

int ChipsAxis::set_minor_grid_visible(bool val)
{ 
   if ((grid_min_vis = val) == true)
   {
      if (!grid_use_vis) 
      {
         grid_maj_vis = grid_min_vis;
         grid_maj_clr.set_foreground_color(grid_min_clr.get_name());
         grid_maj_style = grid_min_style;
         grid_maj_thick = grid_min_thick;
      }
   }
   else 
      grid_maj_vis = grid_use_vis;

   if (grid_min_vis == true)
   {
      int line_count = 0;
      int slices = 0;
      double ***grid = axis_alg->calculate_grid (line_count, slices, 1);
      set_minor_grid (grid, slices, line_count);
   
      if (grid_maj_vis == true)
      {
         int line_count = 0;
         int slices = 0;
         double ***grid = axis_alg->calculate_grid (line_count, slices, 0);
         set_major_grid (grid, slices, line_count);
      }
   }

   return change_grid_property();
}

// set minor grid color
int ChipsAxis::set_minor_grid_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = grid_min_clr.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
   {
      if (grid_min_vis && !grid_use_vis)
         rval = grid_maj_clr.set_foreground_color(in_color);
      if ((rval == ERR_STATUS_GOOD) && initialized)  
         rval = change_grid_property();
   }
   return rval;
}

// set minor grid thickness
int ChipsAxis::set_minor_grid_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "minor grid thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      grid_min_thick = val;
      if (grid_min_vis && !grid_use_vis) grid_maj_thick = val;   
      if (initialized)   rval = change_grid_property();
   }
   return rval;
}

// set minor grid style
int ChipsAxis::set_minor_grid_style(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_line_type(val))
   {
      grid_min_style = val;
      if (grid_min_vis && !grid_use_vis) grid_maj_style = val;   
      if (initialized)   rval = change_grid_property();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val,"chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");
   return rval;
}

int ChipsAxis::set_grid_position (double start_pos, double stop_pos)
{
   grid_start_pos = start_pos;
   grid_stop_pos = stop_pos;
   return change_grid_property();
}

int ChipsAxis::log_scale()
{
   int rval = ERR_AXIS_UNSCALABLE;
   if (axis_alg->get_scalable ())
   {
      axis_alg->set_scaletype (V_SCALE_LOG);
      if (initialized)
      { 
         rval = change_scale(); 
/*         rval = update_tick_limits ();
*/
      }
   }

   if (rval == ERR_AXIS_UNSCALABLE)
      cmd_manager->add_err(rval=ERR_AXIS_UNSCALABLE, id, 0, "log");

   return rval; 
}


int ChipsAxis::linear_scale()
{
   int rval = ERR_AXIS_UNSCALABLE; 

   if (axis_alg->get_scalable ())
   {
      axis_alg->set_scaletype (V_SCALE_LINEAR); 
      if (initialized)
      {
         rval = change_scale(); 
/*         rval = update_tick_limits ();
*/
      }
   }
   if (rval == ERR_AXIS_UNSCALABLE)
      cmd_manager->add_err(rval=ERR_AXIS_UNSCALABLE, id, 0, "linear");
   return rval; 
}

int ChipsAxis::update_tick_limits ()
{
   // the algorithm may adjust its copies of the parameters -- reset them
   // back to axis default value
   int rval = axis_alg->calculate_tick_limits(automin, automax, (use_pad ? pad : 0));
   return rval;
}

int ChipsAxis::update_tick_positions ()
{
   int rval = ERR_STATUS_GOOD;
   bool ticks_updated = false;

   // the limits should be set ... but make sure they are
   update_tick_limits ();


   // the algorithm may adjust its copies of the parameters -- reset them
   // back to axis default value
   rval = axis_alg->update_tick_positions(ticks_updated);

   if (rval != ERR_STATUS_GOOD)
      return rval;

   axis_alg->reset_string_labels (tickformat);
   update_tick_text ();

   if (ticks_updated)
   {
      int slices=0;
      int line_count=0;
      double ***grid;

      change_minor_mode ();
      switch (axis_alg->get_minor_mode (true))
      {
         case chips_nice:
         case chips_count:
            change_minor_num ();
         break;
         case chips_interval:
            change_minor_interval ();
         break;
         case chips_arbitrary:
         {
         }
         break;
         default:
         break;
      }

      if (grid_maj_vis)
      {
         grid = axis_alg->calculate_grid (line_count, slices, 0);
         set_major_grid (grid, slices, line_count);
      }
      if (grid_min_vis)
      {
         grid = axis_alg->calculate_grid (line_count, slices, 1);
         set_minor_grid (grid, slices, line_count);
      }

      change_range ();
   }

   return rval;
}

void ChipsAxis::update_tick_text ()
{

   string *text = axis_alg->get_tick_text ();
   int count = axis_alg->get_tick_text_count ();

   // tick text has changed- recompute tick_text
   for (unsigned int rr = 0; rr < ttext_cnt; rr++)
      delete_tree(tick_text[rr]);
   if (tick_text) delete[] tick_text;
   ttext_cnt = count;
   tick_text = NULL;

   if (count > 0)
   {
      ttext_cnt = count;
      tick_text = new ELEM_P_T[count];
      for (int rr = 0; rr < count; rr++)
         format_text(text[rr], ticklabfont, ticklabfontstyle,
                  ticklabcolor.get_color(), ticklabsize,
                  &tick_text[rr]);

      change_tick_format();
   }
      
}

int ChipsAxis::move_object (V_SUBOBJECT_TYPE obj, COORD_SYS coord_sys, const string &coord_id, double xx, double yy, int mode)
{

   int rval = ERR_STATUS_GOOD;

   switch (obj)
   {
      case chips_axis_label:
      {
         ChipsVector pos1 = norm_pos[0];
         ChipsVector pos2 (xx, yy);
         ChipsVector result (pos1);
         float tmp_offset_perp = 0;
         float tmp_offset_par = 0;
         int par_index = (axis_alg->get_orientation () == X_AXIS) ? 0 : 1;
         int perp_index = (axis_alg->get_orientation () == X_AXIS) ? 1 : 0;
         int flip = (this->title_flip) ? 1 : -1;
         int orientation_flip = (axis_alg->get_orientation () == X_AXIS) ? 1 : -1;
        

         coord->convert_cs_to_cs (&pos2, &pos2, 1, coord_sys, coord_id, PIXEL_REAL, "");

         if (mode == V_MOVE_DELTA)
         {
            pos1[par_index] = (this->offset_par-.5) * (norm_pos[0][par_index] - norm_pos[1][par_index]) + norm_pos[0][par_index];

            coord->convert_cs_to_cs (&pos1, &pos1, 1, FRAME_NORM, "", PIXEL_REAL, "");

            pos1[perp_index] += (this->offset_perp * orientation_flip);
            pos1 += pos2;

            tmp_offset_perp = this->offset_perp + (pos2[perp_index] * flip * orientation_flip);
            coord->convert_cs_to_cs(&pos1,&pos1,1,PIXEL_REAL,"",FRAME_NORM,"");

            tmp_offset_par = pos1[par_index];
/*
            if (dbl_equal((norm_pos[0][par_index] - norm_pos[1][par_index]), 0))
               tmp_offset_par = 0;
            else
               tmp_offset_par = .5+(pos1[par_index]-norm_pos[0][par_index])/(norm_pos[0][par_index] - norm_pos[1][par_index]);
*/

         }
         else
         {
            ChipsVector pix_pos1 (pos1);
            ChipsVector pix_pos2 (pos2);

            pos1[par_index] = .5 * (norm_pos[1][par_index] - norm_pos[0][par_index]) + norm_pos[0][par_index];

            coord->convert_cs_to_cs (&pos1, &pix_pos1, 1, FRAME_NORM, "", PIXEL_REAL, "");
            coord->convert_cs_to_cs(&pos2,&pos2,1,PIXEL_REAL,"",FRAME_NORM,"");

            if (((flip*orientation_flip == -1) && (ticklabstyle == chips_outside)) ||
                ((flip*orientation_flip == 1) && (ticklabstyle != chips_outside)))
            {
               tmp_offset_perp =  pix_pos1[perp_index] - pix_pos2[perp_index];
               tmp_offset_par = pos1[par_index] - pos2[par_index];
            }
            else
            {
               tmp_offset_perp =  pix_pos2[perp_index] - pix_pos1[perp_index];
               tmp_offset_par = pos2[par_index] - pos1[par_index];
            }

         }

         if (dbl_equal((norm_pos[0][par_index] - norm_pos[1][par_index]), 0))
            tmp_offset_par = 0;
         else
            tmp_offset_par = .5+(tmp_offset_par-norm_pos[0][par_index])/(norm_pos[0][par_index] - norm_pos[1][par_index]);


         this->set_property (&tmp_offset_par, SET_AXIS_OFFSET_PAR);
         this->set_property (&tmp_offset_perp, SET_AXIS_OFFSET_PERP);
      }
      break;
      default:
         rval = ERR_CANT_MOVE;
   }
   return rval;
}

int ChipsAxis::resize_object(V_SUBOBJECT_TYPE subobj, COORD_SYS coord_sys, const string &coord_id, double *bounds)
{
   int rval = ERR_STATUS_GOOD;

   switch (subobj)
   {
      case chips_axis_label:
         rval = this->label->resize (coord_sys, coord_id, bounds);
      break;
      default:
         rval = ERR_FEATURE_DNE;
   }

   return rval;
}

int ChipsAxis::rotate_object(V_SUBOBJECT_TYPE subobj, float angle)
{
   int rval = ERR_STATUS_GOOD;

   switch (subobj)
   {
      case chips_axis_label:
         rval = this->set_property (&angle, SET_AXIS_LABEL_ANGLE);
      break;
      default:
         rval = ERR_FEATURE_DNE;
   }

   return rval;
}

int ChipsAxis::select_object(V_SUBOBJECT_TYPE subobj, bool selected)
{
   int rval = ERR_STATUS_GOOD;

   switch (subobj)
   {
      case chips_axis_label:
      {
         if (selected) this->label->select ();
           else this->label->unselect ();
      }
      break;
      default:
         rval = ERR_FEATURE_DNE;
   }

   return rval;

}



int ChipsAxis::move(COORD_SYS coord_sys, const string &coord_id, double xx,
   double yy, int mode, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsVector tmp_n[2];
   ChipsVector tmp_org[2];
   ChipsMoveCommand* cmd = NULL;
   string csid; 

   tmp_n[0] = ChipsVector (0.0, 0.0);
   tmp_n[1] = ChipsVector (0.0, 0.0);
   tmp_org[0] = ChipsVector (0.0, 0.0);
   tmp_org[1] = ChipsVector (0.0, 0.0);

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (coord_sys == COORD_NONE)  coord_sys = coord_system;
   if ((coord_sys == DATA)  && (coord_id.length() == 0))
      csid = coord->get_id ();
   else
      csid = coord_id;

   if (rec)
   {
      string save_current;
      if (parent && (parent->get_object_type() == chips_plot))
      {
         if (axis_alg->get_orientation () == X_AXIS)
            save_current = ((ChipsPlot *)parent)->get_current_xaxis_id();
         else
            save_current = ((ChipsPlot *)parent)->get_current_yaxis_id();
      }
      cmd = new ChipsMoveCommand(this, orig_pos[0][0], orig_pos[0][1], xx, yy, coord_sys,
                                 coord_id, mode, save_current);
   }
   if (mode == V_MOVE_DELTA)
   {
      // Only allow the axis to move in one direction
      if (axis_alg->get_orientation () == X_AXIS)
         xx = 0;
      else
         yy = 0;

      rval = coord->convert_cs_to_cs(&orig_pos[0], &tmp_org[0], 2, coord_system,
				     data_system_id, coord_sys, csid);
      if (!rval){
         for (int ii=0; ii<2; ii++){
            tmp_org[ii][0] += xx;
            tmp_org[ii][1] += yy;
         }
         rval = coord->convert_cs_to_cs(&tmp_org[0], &tmp_n[0], 2, coord_sys,
                                  csid, coord_system, data_system_id);
      }

   }
   else{
      ChipsVector len = orig_pos[1] - orig_pos[0];

      
      tmp_org[0][0] = xx;
      tmp_org[0][1] = yy;

      rval = coord->convert_cs_to_cs(&tmp_org[0], &tmp_n[0], 1, coord_sys,
                                  csid, coord_system, data_system_id);

      // Only allow the axis to move in one direction
      if (axis_alg->get_orientation () == X_AXIS)
         tmp_n[0][0] = orig_pos[0][0];
      else
         tmp_n[0][1] = orig_pos[0][1];

      tmp_n[1] = tmp_n[0] + len;
   }
  
   if (rval == ERR_NEG_LOG_VAL)
      cmd_manager->add_err("Cannot move an object to a position less than or equal to zero when in log scale.", rval=ERR_EDGE_BOUNDARY, id, 0);

   if (rval == ERR_STATUS_GOOD)
   {
      orig_pos[0] = tmp_n[0];
      orig_pos[1] = tmp_n[1];

      rval = calculate_norm_coordinate ();
      if (!rval)
      {
         invert_labels();
         if ((rval = change_position()) == ERR_STATUS_GOOD)
         {
            set_ticklabelstyle(ticklabstyle);
            change_title_flip();
            change_major_style();
            rval = change_scale_factor();
         }
      }
   }

   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if (!add_command (cmd, rval))  
         delete cmd;
   }
   else if (rec)
      delete(cmd);
   update_tick_positions ();
   return rval;
}


int ChipsAxis::reset_position ()
{
   int rval = ERR_STATUS_GOOD;
   ChipsCoordinate *plot_coord = coord->get_coord (PLOT_NORM, "");
   ChipsVector plot_bounds[2];

   // This ensures that the axes stay locked to the plot boundaries
   if (plot_coord)
   {
      plot_bounds[0] = plot_coord->get_lower_bounds ();
      plot_bounds[1] = plot_coord->get_upper_bounds ();
      set_plot_location (plot_bounds[0][0], plot_bounds[0][1], plot_bounds[1][0], 
                         plot_bounds[1][1]);
   }

   rval = calculate_norm_coordinate ();
   if (!rval)
   {
      invert_labels();
      if ((rval = change_position()) == ERR_STATUS_GOOD)
      {
         set_ticklabelstyle(ticklabstyle);
         change_title_flip(); 
         change_major_style();
         rval = change_scale_factor();
      }
   }
   return rval; 
}

int ChipsAxis::adjust_default_colors()
{
   if (get_color() == "default")  change_color();
   if (label && (get_label_color() == "default"))  label->change_color();
   if (get_ticklabel_color() == "default")      change_ticklabel_color();
   if (get_major_color() == "default")  change_major_color();
   if (get_minor_color() == "default")  change_minor_color();
   if ((get_major_grid_color() == "default") ||
       (get_minor_grid_color() == "default"))  change_grid_property();
   return ERR_STATUS_GOOD; 
}


// do we need to invert the axis label or ticklabels?
void ChipsAxis::invert_labels()
{
   ChipsVector tmp_pnt (.5,.5);
   if (!coord)
      return;

   coord->convert_cs_to_cs (&tmp_pnt, &tmp_pnt, 1,
                                                  PLOT_NORM, "", FRAME_NORM, "");

   if (axis_alg->get_orientation () == X_AXIS)
   {
      title_flip = (norm_pos[0][1] > tmp_pnt[1]); 
   }
   else
   {
      title_flip = (norm_pos[0][0] <= tmp_pnt[0]);
   }
   set_use_major_tickstyle();
   set_use_minor_tickstyle();
}


void ChipsAxis::set_use_major_tickstyle()
{
   if (title_flip)
   {
      if (major_tickstyle == chips_inside)
         use_major_tickstyle = chips_outside;
      else if (major_tickstyle == chips_outside)
         use_major_tickstyle = chips_inside;
      else 
         use_major_tickstyle = major_tickstyle;
   }
   else 
      use_major_tickstyle = major_tickstyle;
}

void ChipsAxis::set_use_minor_tickstyle()
{
   if (title_flip)
   {
      if (minor_tickstyle == chips_inside)
         use_minor_tickstyle = chips_outside;
      else if (minor_tickstyle == chips_outside)
         use_minor_tickstyle = chips_inside;
      else   
         use_minor_tickstyle = minor_tickstyle;
   }
   else
      use_minor_tickstyle = minor_tickstyle;
}

int ChipsAxis::set_alignment(V_TICKSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;
   alignment = val;
   if (initialized)  rval = change_alignment();
   return rval;  
}


int ChipsAxis::change_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_corner()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_alignment()
{  
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_position()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_range()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_reverse()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_scale()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_scale_factor()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_tick_format()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_log_minimum()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_interval()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_length()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_mode()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_num()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_style()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_major_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_interval()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_length()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_mode()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_num()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_style()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_style()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_font()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_fontstyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_angle()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_size()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_ticklabel_alignment()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_ticklabel_offset()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::change_minor_visible()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_parallel_offset()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_perpendicular_offset()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_title_flip()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_title_visible()
{
   return ERR_FEATURE_DNE;
}


int ChipsAxis::change_grid_property()
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::set_major_grid (double ***grid, int slices, int lines)
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::set_minor_grid (double ***grid, int slices, int lines)
{
   return ERR_FEATURE_DNE;
}

int ChipsAxis::set_plot_location (double x1, double y1, double x2, double y2)
{
   ChipsVector tmp_pos (x1,y1);
   ChipsVector tmp_pos2 (x2,y2);
   int rval = ERR_STATUS_GOOD;
   // set the ends of the axis to correct position
   if (coord)
   {
      rval = coord->convert_cs_to_cs (&tmp_pos, &tmp_pos, 1, FRAME_NORM, "",
                                coord_system, data_system_id);
      rval = coord->convert_cs_to_cs (&tmp_pos2, &tmp_pos2, 1, FRAME_NORM, "",
                                coord_system, data_system_id);
   }

   if (axis_alg->get_orientation ()  == X_AXIS)
   {
      orig_pos[0][0] = tmp_pos[0];
      orig_pos[1][0] = tmp_pos2[0];
      norm_pos[0][0] = x1;
      norm_pos[1][0] = x2;
   }
   else
   {
      orig_pos[0][1] = tmp_pos[1];
      orig_pos[1][1] = tmp_pos2[1];
      norm_pos[0][1] = y1;
      norm_pos[1][1] = y2;
   }

   if (!rval)
      rval = change_position (); 

   return rval;
}


void ChipsAxis::serialize(ChipsStateStream& out)
{
   int sz;
   bool bval;

   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   sz = (int)coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);
   out.write_double(&(orig_pos[0][0]));
   out.write_double(&(orig_pos[0][1]));
   out.write_double(&(orig_pos[1][0]));
   out.write_double(&(orig_pos[1][1]));

   if (out.get_minor_version() >= 6)
   {
     
      sz = static_cast <int> (ChipsAxisAlgorithmFactory::get_type (axis_alg));
      out.write_int (&sz);
      axis_alg->serialize (out);
      out.write_double(&pad);
      out.write_double(&label_angle);
      out.write_double(&use_halign);
      out.write_double(&use_valign);
      out.write_float(&offset_par);
      out.write_float(&offset_perp);
      out.write_bool(&title_flip);
      out.write_bool(&major_vis);
      out.write_bool(&minor_vis);
      out.write_bool(&reverse);
      out.write_bool(&automin);
      out.write_bool(&automax);
      
      out.write_bool(&automin_used);
      out.write_bool(&automax_used);
      
      out.write_float(&thickness);
      out.write_float(&major_thickness);
      out.write_float(&minor_thickness);
      out.write_int(&major_len);
      out.write_int(&minor_len);
      sz = (int)major_tickstyle;
      out.write_int(&sz);
      sz = (int)minor_tickstyle;
      out.write_int(&sz);
      out.write_string(&tickformat);
      out.write_bool(&shared_ds);

      bval = (ticklabhorizalignment == AUTO_ALIGN);
      out.write_bool (&bval);
      bval = (ticklabvertalignment == AUTO_ALIGN);
      out.write_bool (&bval);

   }
   else
   {
      double dval;
      int    ival;

      dval = axis_alg->get_absolute_min (V_SCALE_LINEAR);
      out.write_double(&dval);
      dval = axis_alg->get_absolute_max (V_SCALE_LINEAR);
      out.write_double(&dval);
      dval = axis_alg->get_absolute_min (V_SCALE_LOG);
      out.write_double(&dval);
      dval = axis_alg->get_absolute_max (V_SCALE_LOG);
      out.write_double(&dval);

      dval = axis_alg->get_major_interval ();
      out.write_double(&dval);
      dval = axis_alg->get_minor_interval ();
      out.write_double(&dval);

      dval = get_log_minimum ();
      out.write_double(&dval);

      out.write_double(&pad);
      out.write_double(&label_angle);
      out.write_double(&use_halign);
      out.write_double(&use_valign);
      out.write_float(&offset_par);
      out.write_float(&offset_perp);
      out.write_bool(&title_flip);
      out.write_bool(&major_vis);
      out.write_bool(&minor_vis);
      out.write_bool(&reverse);
      out.write_bool(&automin);
      out.write_bool(&automax);
      if (out.get_minor_version() > 4)
      {
         out.write_bool(&automin_used);
         out.write_bool(&automax_used);
      }
      out.write_int(&depth);
      ival = axis_alg->get_scaletype ();
      out.write_int(&ival); 
      ival = axis_alg->get_major_num ();
      out.write_int(&ival);
      ival = axis_alg->get_minor_num ();
      out.write_int(&ival);
      ival = axis_alg->get_tick_count ();
      out.write_int(&ival);
      ival = axis_alg->get_arbitrary_count ();
      out.write_int(&ival);
      out.write_float(&thickness);
      out.write_float(&major_thickness);
      out.write_float(&minor_thickness);
      out.write_int(&major_len);
      out.write_int(&minor_len);
      sz = (int)major_tickstyle;
      out.write_int(&sz);
      sz = (int)minor_tickstyle;
      out.write_int(&sz);
      out.write_double(axis_alg->get_tick_positions (false), axis_alg->get_tick_count ());
      out.write_double(axis_alg->get_arbitrary_positions (), axis_alg->get_arbitrary_count ());

      bool texists = (axis_alg->get_arbitrary_text () != NULL); 
      out.write_bool(&texists);
      if (texists)
      {
         int zz;
         string *arbtext = axis_alg->get_arbitrary_text ();
         for (zz = axis_alg->get_arbitrary_count (); zz--; )   out.write_string(&arbtext[zz]);
      }
   
      out.write_string(&tickformat); 
      sz = (int)axis_alg->get_major_mode (); 
      out.write_int(&sz);
   }
   if (((out.get_major_version() == 3) && (out.get_minor_version() >= 7)) ||
       (out.get_major_version() > 3))
      out.write_bool (&use_pad);
   if (((out.get_major_version() == 3) && (out.get_minor_version() >= 8)) ||
       (out.get_major_version() > 3))
      out.write_bool (&border_axis);
   out.write_bool(&ticklabvis);
   out.write_bool(&labvis);

   // tick label props
   out.write_string(&ticklabfont);
   out.write_int(&ticklaboffset);
   out.write_float(&ticklabangle);
   out.write_int(&ticklabsize);
   sz = (int)ticklabfontstyle;
   out.write_int(&sz);
   if (out.get_minor_version() >= 15)
   {
      sz = (int)ticklabstyle;
      out.write_int(&sz);
      out.write_double(&ticklabhorizalignment);
   }
   else
   {
      out.write_double(&ticklabvertalignment);
   }
   out.write_double(&ticklabvertalignment);
   string tc = get_ticklabel_color();
   out.write_string(&tc);
   
   // axis|majortick|minortick colors 
   string col, col2;
   col = get_color();
   out.write_string(&col);
   col2 = get_major_color();
   out.write_string(&col2);
   col2 = get_minor_color();
   out.write_string(&col2);

   // grid info
   col = get_major_grid_color();
   out.write_string(&col);
   col2 = get_minor_grid_color();
   out.write_string(&col2);
   sz = (int)grid_use_style;
   out.write_int(&sz);
   out.write_float(&grid_use_thick);
   out.write_bool(&grid_maj_vis);
   out.write_bool(&grid_use_vis);
   sz = (int)grid_min_style;
   out.write_int(&sz);
   out.write_float(&grid_min_thick);
   out.write_bool(&grid_min_vis);

   out.write_double(&ll_corner_width);
   out.write_double(&ur_corner_width);
   sz = (int) ll_corner; 
   out.write_int(&sz);
   sz = (int) ur_corner; 
   out.write_int(&sz);
   sz = (int) alignment; 
   out.write_int(&sz);

   out.write_double (&grid_start_pos);
   out.write_double (&grid_stop_pos);

   if (out.get_minor_version() > 3)
   {
      out.write_bool(&visible);
   }

   if (label)
   {
      sz = 1;
      out.write_int(&sz);
      label->serialize(out);
   }
   else
   {
      sz = 0;
      out.write_int(&sz);
   }

}
                                                                                        
void ChipsAxis::deserialize(ChipsStateStream& in)
{
   int len;
   bool bval;

   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);

   double d1, d2, d3, d4; 
   in.read_double(&d1);
   in.read_double(&d2);
   in.read_double(&d3);
   in.read_double(&d4);
   orig_pos[0][0] = d1;
   orig_pos[0][1] = d2;
   orig_pos[1][0] = d3;
   orig_pos[1][1] = d4;

   if (in.get_minor_version() >= 6)
   {
      int ival;
      int orientation = 0;

      if (in.get_minor_version () == 6)
         in.read_int(&orientation);
      in.read_int (&ival);
      if (in.get_minor_version () == 6)
         ival = ChipsAxisAlgorithmFactory::CARTESIAN;
      if (axis_alg)
         delete axis_alg;
      axis_alg = ChipsAxisAlgorithmFactory::create (static_cast <ChipsAxisAlgorithmFactory::TransformType> (ival), orientation, this);
      axis_alg->deserialize (in);
      in.read_double(&pad);
      in.read_double(&label_angle);
      in.read_double(&use_halign);
      in.read_double(&use_valign);
      in.read_float(&offset_par);
      in.read_float(&offset_perp);
      in.read_bool(&title_flip);
      in.read_bool(&major_vis);
      in.read_bool(&minor_vis);
      in.read_bool(&reverse);
      in.read_bool(&automin);
      in.read_bool(&automax);
      in.read_bool(&automin_used);
      in.read_bool(&automax_used);
      in.read_float(&thickness);
      in.read_float(&major_thickness);
      in.read_float(&minor_thickness);
      in.read_int(&major_len);
      in.read_int(&minor_len);
      in.read_int(&len);
      major_tickstyle = (V_TICKSTYLE_TYPE)len;
      in.read_int(&len);
      minor_tickstyle = (V_TICKSTYLE_TYPE)len;
      in.read_string(&tickformat);
      if (in.get_minor_version() >= 7)
         in.read_bool (&shared_ds);

      if (in.get_minor_version () >= 10)
      {
         in.read_bool (&bval);
         ticklabhorizalignment = (bval) ? AUTO_ALIGN : use_halign;
         in.read_bool (&bval);
         ticklabvertalignment = (bval) ? AUTO_ALIGN : use_valign;
      }

   }
   else
   {
      double dval;
      int ival;
      int tickcnt, arbcnt;
      int orientation;

      in.read_int(&orientation);

      if (axis_alg)
         delete axis_alg;
      axis_alg = ChipsAxisAlgorithmFactory::create (ChipsAxisAlgorithmFactory::CARTESIAN, orientation, this);

      in.read_double(&dval);
      in.read_double(&dval);
      
      in.read_double(&dval);
      axis_alg->set_absolute_min (dval, V_SCALE_LINEAR);
      in.read_double(&dval);
      axis_alg->set_absolute_max (dval, V_SCALE_LINEAR);
      in.read_double(&dval);
      axis_alg->set_absolute_min (dval, V_SCALE_LOG);
      in.read_double(&dval);
      axis_alg->set_absolute_max (dval, V_SCALE_LOG);

      in.read_double(&dval);
      axis_alg->set_major_interval (dval);
      in.read_double(&dval);
      axis_alg->set_minor_interval (dval);

      in.read_double(&dval);
      set_log_minimum (dval);

      in.read_double(&pad);
      in.read_double(&label_angle);
      in.read_double(&use_halign); 
      in.read_double(&use_valign);
      in.read_float(&offset_par);
      in.read_float(&offset_perp);
      in.read_bool(&title_flip);
      in.read_bool(&major_vis);
      in.read_bool(&minor_vis);
      in.read_bool(&reverse);
      in.read_bool(&automin);
      in.read_bool(&automax);
      if (in.get_minor_version() > 4)
      {
         in.read_bool(&automin_used);
         in.read_bool(&automax_used);
      }
      in.read_int(&depth);
      in.read_int(&ival); 
      axis_alg->set_scaletype (ival);
      in.read_int(&ival);
      axis_alg->set_major_num (ival);
      in.read_int(&ival);
      axis_alg->set_minor_num (ival);
      in.read_int(&tickcnt);
      in.read_int(&arbcnt);

      in.read_float(&thickness);
      in.read_float(&major_thickness);
      in.read_float(&minor_thickness);
      in.read_int(&major_len);
      in.read_int(&minor_len);
      in.read_int(&len);
      major_tickstyle = (V_TICKSTYLE_TYPE)len;
      in.read_int(&len);
      minor_tickstyle = (V_TICKSTYLE_TYPE)len;

      if (tickcnt)
      {
         double *tickpos =  new double[tickcnt];
         in.read_double(tickpos, tickcnt); 
         axis_alg->set_tick_positions (tickpos, tickcnt);
      }

      double *arbpos=NULL;
      string *arbtext=NULL;
      if (arbcnt)
      {
         arbpos = new double[arbcnt];
         in.read_double(arbpos, arbcnt); 
      }

      bool texists;
      in.read_bool(&texists);
      if (texists)
      {
         int zz;
         arbtext = new string[arbcnt];
         for (zz = arbcnt; zz--; )   in.read_string(&arbtext[zz]);
      }

      if (arbcnt)
      {
         axis_alg->set_arbitrary_positions (arbpos, arbtext, arbcnt);
         delete []arbpos;
         if (arbtext)
            delete []arbtext;
      }
      in.read_string(&tickformat);
      in.read_int(&len);

      // it's possible that the mode is not supported by the axis
      // algorithm in which case use try limits
      if (axis_alg->set_major_mode ((V_TICKMODE_TYPE)len) != ERR_STATUS_GOOD)
         axis_alg->set_major_mode (chips_limits);
      

   }

   if (((in.get_major_version() == 3) && (in.get_minor_version() >= 7)) ||
       (in.get_major_version() > 3))
      in.read_bool (&use_pad);
   if (((in.get_major_version() == 3) && (in.get_minor_version() >= 8)) ||
       (in.get_major_version() > 3))
      in.read_bool (&border_axis);

   in.read_bool(&ticklabvis);
   in.read_bool(&labvis);
   
   string col;

   in.read_string(&ticklabfont);
   in.read_int(&ticklaboffset);
   in.read_float(&ticklabangle);
   in.read_int(&ticklabsize);
   in.read_int(&len);
   ticklabfontstyle = (V_FONTSTYLE_TYPE)len;
   if (in.get_minor_version() >= 15)
   {
      in.read_int(&len);
      ticklabstyle = (V_TICKSTYLE_TYPE)len; 
      in.read_double(&ticklabhorizalignment);
   }
   else 
   {
      in.read_double(&ticklabvertalignment);
   }
   in.read_double(&ticklabvertalignment);
   in.read_string(&col);
   if (col.length())   set_ticklabel_color(col);

   in.read_string(&col);
   if (col.length())   set_color(col);
   in.read_string(&col);
   if (col.length())   set_major_color(col);
   in.read_string(&col);
   if (col.length())   set_minor_color(col);

   // grid info
   in.read_string(&col);
   if (col.length())   set_major_grid_color(col);
   in.read_string(&col);
   if (col.length())   set_minor_grid_color(col);
   in.read_int(&len);
   grid_use_style =  (V_LINE_TYPE)len;
   in.read_float(&grid_use_thick);
   in.read_bool(&grid_maj_vis);
   in.read_bool(&grid_use_vis);
   set_major_grid_visible(grid_use_vis);
   in.read_int(&len);
   grid_min_style =  (V_LINE_TYPE)len;
   in.read_float(&grid_min_thick);
   in.read_bool(&grid_min_vis);
   set_minor_grid_visible(grid_min_vis);

   in.read_double(&ll_corner_width);
   in.read_double(&ur_corner_width);
   in.read_int(&len);
   ll_corner = (V_CORNERSTYLE_TYPE)len;
   in.read_int(&len);
   ur_corner = (V_CORNERSTYLE_TYPE)len;
   in.read_int(&len);
   alignment = (V_TICKSTYLE_TYPE)len;

   in.read_double (&grid_start_pos);
   in.read_double (&grid_stop_pos);

   if (in.get_minor_version() > 3)
   {
      in.read_bool(&visible);
   }

   in.read_int(&len);
   deserialize_label = len;  // deserialize label in derived Axis constuctor?
/*   update_tick_limits ();
*/

}

int ChipsAxis::show()
{
   visible = true;
   return change_visible ();
}
int ChipsAxis::hide()
{
   visible = false;
   return change_visible ();
}


// change style of joint where axes meet 
int ChipsAxis::set_corners(V_CORNERSTYLE_TYPE val, double lv, double rv)
{
   int rval = ERR_STATUS_GOOD;
   if (valid_cornerstyle_type(val) == false)
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "axis corner style", (int)val,"chips_miter|chips_bevel|chips_fill|chips_notch");
   else if ((lv < 0) || (rv < 0))
      cmd_manager->add_err(rval=ERR_DBL_MIN, id, 0, "axis corner width", 0);
   else
   {
      ll_corner = ur_corner = val;
      ll_corner_width = lv;
      ur_corner_width = rv;
      if (initialized)  rval =  change_corner();
   }
   return rval;
}

int ChipsAxis::set_ll_corner(V_CORNERSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD; 
                                                
   if (valid_cornerstyle_type(val))
   {
      ll_corner = val;
      if (initialized)  rval =  change_corner(); 
   }
   else 
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "axis corner style", (int)val,"chips_miter|chips_bevel|chips_fill|chips_notch");
   return rval; 
}

int ChipsAxis::set_ur_corner(V_CORNERSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD; 
   if (valid_cornerstyle_type(val))
   {
      ur_corner = val;
      if (initialized)  rval =  change_corner(); 
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "axis corner style", (int)val,"chips_miter|chips_bevel|chips_fill|chips_notch");
   return rval; 
}


int ChipsAxis::set_ll_corner_width(double val)
{
   int rval = ERR_STATUS_GOOD; 
   if (val < 0)
      cmd_manager->add_err(rval=ERR_DBL_MIN, id, 0, "axis corner width", 0);
   else 
      ll_corner_width = val;
   if (initialized)  rval =  change_corner(); 
   return rval; 
}


int ChipsAxis::set_ur_corner_width(double val)
{
   int rval = ERR_STATUS_GOOD; 
   if (val < 0)
      cmd_manager->add_err(rval=ERR_DBL_MIN, id, 0, "axis corner width", 0);
   else 
      ur_corner_width = val;
   if (initialized)  rval =  change_corner(); 
   return rval; 
}

// pack up all data values used for axis synch- incase of bind undo 
int ChipsAxis::compile_synch_data(double** d_p, int** i_p, string** s_p)
{
   if (d_p && i_p && s_p) 
   {
      int dp_size = 9 + axis_alg->get_arbitrary_count () + axis_alg->get_tick_count ();
      int ip_size = 17;

      double* dp = new double[dp_size];
      int* ip = new int[ip_size];
      string* sp = NULL; 

      ip[0] = ip_size; // ip size
      ip[1] = dp_size; // dp size
      ip[2] = 0; // sp_size


      dp[0] = axis_alg->get_min ();
      dp[1] = axis_alg->get_max ();
      dp[2] = axis_alg->get_absolute_min (V_SCALE_LOG);
      dp[3] = axis_alg->get_absolute_max (V_SCALE_LOG);
      dp[4] = axis_alg->get_absolute_min (V_SCALE_LINEAR);
      dp[5] = axis_alg->get_absolute_max (V_SCALE_LINEAR);
      dp[6] = axis_alg->get_major_interval ();
      dp[7] = axis_alg->get_minor_interval ();
      dp[8] = pad;
      ip[3] = axis_alg->get_major_mode (); 
      ip[4] = axis_alg->get_minor_mode ();
      ip[5] = automax;
      ip[6] = automin;
      ip[7] = reverse;
      ip[8] = axis_alg->get_major_num ();
      ip[9] = axis_alg->get_minor_num ();
      ip[10] = 0;            // GARBAGE VALUE
      ip[11] = axis_alg->get_scaletype ();
      ip[12] = axis_alg->get_tick_count (); 
      ip[13] = axis_alg->get_arbitrary_count ();
      unsigned int rr, dc = 9;
      double *tickpos = axis_alg->get_tick_positions (false);
      unsigned int tickcnt = axis_alg->get_tick_count ();
      double * arbpos = axis_alg->get_arbitrary_positions ();
      string *arbtext = axis_alg->get_arbitrary_text ();
      unsigned int arbcnt = axis_alg->get_arbitrary_count ();

      for (rr = 0; rr < tickcnt; rr++)   dp[dc++] = tickpos[rr];
      for (rr = 0; rr < arbcnt; rr++)  dp[dc++] = arbpos[rr]; 
      if (arbcnt && arbtext)
      {
         ip[14] = arbcnt; 
         sp = new string[arbcnt + 1];
         sp[0] = tickformat; 
         for (rr = 0; rr < arbcnt; rr++)  sp[rr+1] = arbtext[rr];
         ip[2] = arbcnt+1;
      }
      else
      {
         ip[14] = 0; 
         sp = new string[1];
         sp[0] = tickformat; 
         ip[2] = 1;
      }

      ip[15] = ChipsAxisAlgorithmFactory::get_type (axis_alg);
      ip[16] = use_pad;

      *d_p = dp; 
      *i_p = ip;
      *s_p = sp; 
      return ERR_STATUS_GOOD; 
   }
   return ERR_STATUS_BAD; 
}


// revert 'synched' values back to specified values (for bind undo)
int ChipsAxis::unsync_data(double* d_p, int* i_p, string* s_p)
{
   if (d_p && i_p && axis_alg)
   {
      unsigned int dc = 9; 
    
      int ip_size = i_p[0];
      int sp_size = i_p[2];
      int base_sp_size = sp_size;
      int base_ip_size = 17;


      if (ip_size != base_ip_size)
         base_sp_size = i_p[base_ip_size+1];
      

/*      axis_alg->set_minmax (d_p[0],d_p[1]);
*/

      if (i_p[15] != (int)ChipsAxisAlgorithmFactory::get_type(axis_alg))
      {
         ChipsAxisAlgorithm *tmp_axis_alg = ChipsAxisAlgorithmFactory::create (i_p[15], axis_alg->get_orientation (), this, axis_alg);
         delete axis_alg;

         axis_alg = tmp_axis_alg;
         axis_alg->set_axis (this);
      }
      
      axis_alg->set_absolute_minmax (d_p[2], d_p[3], V_SCALE_LOG);
      axis_alg->set_absolute_minmax (d_p[4], d_p[5], V_SCALE_LINEAR);
      axis_alg->set_major_interval (d_p[6]);
      axis_alg->set_minor_interval (d_p[7]);
      pad       = d_p[8];
      axis_alg->set_major_mode ((V_TICKMODE_TYPE) i_p[3]); 
      axis_alg->set_minor_mode ((V_TICKMODE_TYPE) i_p[4]);
      automax   = i_p[5];
      automin   = i_p[6];
      reverse   = i_p[7];
      axis_alg->set_minor_num (i_p[8]);
      axis_alg->set_major_num (i_p[9]);
      axis_alg->set_scaletype (i_p[11]);
      axis_alg->set_tick_positions (&d_p[dc],  i_p[12]);
      dc += i_p[12];
      if (i_p[14] != 0)
         axis_alg->set_arbitrary_positions (&d_p[dc], &s_p[1], i_p[14]);
      else
         axis_alg->set_arbitrary_positions (&d_p[dc], NULL, i_p[14]);


      if (this->border_axis)
      {
         // border axes information is not saved in a synch so simply create
         // a new ds from scratch
         ChipsDataSystem *ds = this->get_data_system ();
         if (ds)
         {
            const ChipsDataSystem *new_ds;

            new_ds = new ChipsDataSystem (ds->get_parent (), ChipsVector (0,0), ChipsVector (1,1), ds->get_xaxis_id (), ds->get_yaxis_id ());
            *ds = *new_ds;
            delete new_ds;
         }
      }
      else
      {
         int coord_index = base_sp_size;
         int coord_count = 0;
         for (int ii=base_ip_size+2; ii<ip_size; ii+=2)
         {
            if (i_p[ii] == this->get_ref_cnt ())
            {
               coord_count = i_p[ii+1];
               break;
            }
            coord_index += i_p[ii+1];
         }
        

         data_coord_ids.clear ();
         if (coord_count > 0)
         {
            for (int ii=0; ii<coord_count; ii++)
               data_coord_ids.push_back (s_p[ii+coord_index]);

            ChipsPlot *parent_plot = dynamic_cast <ChipsPlot *> (this->parent);
            if (parent_plot)
               parent_plot->update_axis_coordinates (this, this->data_coord_ids);
         }
      }

      use_pad = i_p[16];

      // redraw the axis 
      if (initialized)
      {
         update_tick_positions ();
         if (axis_alg->get_scaletype () == V_SCALE_LOG)   log_scale();
            else                         linear_scale();
         change_tick_format();
         change_range ();

      }
      return ERR_STATUS_GOOD; 
   }
   return ERR_STATUS_BAD;
}

int ChipsAxis::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}

int ChipsAxis::calculate_norm_coordinate ()
{
   int rval = ERR_STATUS_GOOD;
   if (coord)
      rval = coord->convert_cs_to_cs (orig_pos, norm_pos, 2, coord_system,
                            data_system_id, FRAME_NORM, "");

   return rval;
}

bool ChipsAxis::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED) outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}

int ChipsAxis::reset_values()
{
   string tmp_s;
   double tmp_d;
   float  tmp_f;
   int    tmp_i, rval = ERR_STATUS_GOOD;
   bool   tmp_b;

   if (theDefaults)   tmp_s = theDefaults->get_axiscolor();
   else               tmp_s = "yellow";
   if (tmp_s != get_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_COLOR);
   if (theDefaults)   tmp_s = theDefaults->get_majortickcolor();
   else               tmp_s = "white";
   if (tmp_s != get_major_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MAJOR_COLOR);
   if (theDefaults)   tmp_s = theDefaults->get_minortickcolor();
   else               tmp_s = "white";
   if (tmp_s != get_minor_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MINOR_COLOR);
   if (theDefaults)   tmp_f = theDefaults->get_axisthickness();
   else               tmp_f = 2;
   if (tmp_f != thickness)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_THICKNESS);
   if (theDefaults)   tmp_f = theDefaults->get_majortickthickness();
   else               tmp_f = 2; 
   if (tmp_f != major_thickness)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_MAJ_THICKNESS);
   if (theDefaults)   tmp_f = theDefaults->get_minortickthickness();
   else               tmp_f = 1; 
   if (tmp_f != minor_thickness)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_MIN_THICKNESS);
   if (theDefaults)   tmp_b = theDefaults->get_majorgridvisible();
   else               tmp_b = false; 
   if (tmp_b != grid_maj_vis)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_MAJGRD_VIS);
   if (theDefaults)   tmp_b = theDefaults->get_minorgridvisible();
   else               tmp_b = false; 
   if (tmp_b != grid_min_vis)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_MINGRD_VIS);
   tmp_s.clear();
   if (axis_alg->get_orientation () == X_AXIS)
   {
      if (theDefaults) tmp_s = theDefaults->get_axisxlabel(); 
      else             tmp_s = "";
   } 
   if (get_label_text() != tmp_s)
   {
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_TEXT);
   }
   if (axis_alg->get_orientation () == X_AXIS)
   {
      if (theDefaults)   tmp_f = theDefaults->get_axlabxangle();
      else               tmp_f = 0.0;
   } else {
      if (theDefaults)   tmp_f = theDefaults->get_axlabyangle();
      else               tmp_f = 90.0;
   }
   if (get_label_angle() != tmp_f)
   {
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_LABEL_ANGLE);
   }
   if (get_label_halign() != 0.5)
   {
      tmp_f = 0.5;
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_LABEL_HALIGN);
   }
   if (get_label_valign() != 0.0)
   {
      tmp_f = 0.0;
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_LABEL_VALIGN);
   }
   if (theDefaults)   tmp_i = theDefaults->get_majortickstyle();
   else               tmp_i = chips_inside;
   if (major_tickstyle != tmp_i)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MAJ_STYLE); 
   if (theDefaults)   tmp_i = theDefaults->get_minortickstyle();
   else               tmp_i = chips_inside;
   if (minor_tickstyle != tmp_i)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MIN_STYLE); 
   if (theDefaults)   tmp_s = theDefaults->get_tickformat();
   else               tmp_s = "%3d";
   if (tmp_s != get_tickformat())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_TICKFORMAT);
   if (theDefaults)   tmp_i = theDefaults->get_ticklabelstyle();
   else               tmp_i = chips_outside;
   if (ticklabstyle != tmp_i)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_TICKLABEL_STYLE);
   if (theDefaults)   tmp_i = theDefaults->get_majortickcount();
   else               tmp_i = 4;
   if (get_major_num () != tmp_i)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MAJOR_NUM);
   if (theDefaults)   tmp_i = theDefaults->get_minortickcount();
   else               tmp_i = 4;
   if (get_minor_num () != tmp_i)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MINOR_NUM);
   if (theDefaults)   tmp_d = theDefaults->get_majortickinterval();
   else               tmp_d = 10;
   if (get_major_int () != tmp_d)
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_MAJORINT);
   if (theDefaults)   tmp_d = theDefaults->get_minortickinterval();
   else               tmp_d = 2;
   if (get_minor_int () != tmp_d)
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_MINORINT);
   tmp_d = 0.01;
   if (get_log_minimum () != tmp_d)
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_LOG_MIN);
   if (theDefaults)   tmp_d = theDefaults->get_axispad();
   else               tmp_d = 0.05;
   if (pad != tmp_d)
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_PAD);
   if (theDefaults)   tmp_i = theDefaults->get_majorticklength();
   else               tmp_i = 4;
   if (tmp_i != major_len)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MAJORLENGTH);
   if (theDefaults)   tmp_i = theDefaults->get_minorticklength();
   else               tmp_i = 2;
   if (tmp_i != minor_len)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MINORLENGTH);
   if (theDefaults)   tmp_s = theDefaults->get_majortickmode();
   else               tmp_s = "count";
   if (tmp_s != v_tickmode_to_string(get_major_mode()))
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MAJ_MODE);
   if (theDefaults)   tmp_s = theDefaults->get_minortickmode();
   else               tmp_s = "count";
   if (tmp_s != v_tickmode_to_string(get_minor_mode()))
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MIN_MODE);
   if (theDefaults)   tmp_b = theDefaults->get_axisautomax();
   else               tmp_b = true;
   if (tmp_b != automax)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_AUTOMAX);
   if (theDefaults)   tmp_b = theDefaults->get_axisautomin();
   else               tmp_b = true;
   if (tmp_b != automin)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_AUTOMIN);
   if (reverse)
   {
      tmp_b = false;
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_REVERSE);
   }
   if (axis_alg->get_scaletype () != V_SCALE_LINEAR)
   {  
      tmp_i = V_SCALE_LINEAR;
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_SCALE);
   }
   if (!axis_alg->get_scalable ())
   {
      tmp_b = true;
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_SCALABLE);
   }
   if (theDefaults)   tmp_b = theDefaults->get_majortickvisible();
   else               tmp_b = true; 
   if (tmp_b != major_vis)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_MAJ_VIS);
   if (theDefaults)   tmp_b = theDefaults->get_minortickvisible();
   else               tmp_b = true; 
   if (tmp_b != minor_vis)
      rval = set_property(static_cast<void*>(&tmp_b), SET_AXIS_MIN_VIS);
   if (theDefaults)   tmp_f = theDefaults->get_axisparalleloffset();
   else               tmp_f = 0.0;
   if (tmp_f != offset_par)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_OFFSET_PAR);
   if (theDefaults)   tmp_f = theDefaults->get_axisperpendicularoffset();
   else               tmp_f = 20.0;
   if (tmp_f != offset_perp)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_OFFSET_PERP);
   if (theDefaults)   tmp_s = theDefaults->get_axlabcolor();
   else               tmp_s = "yellow";
   if (tmp_s != get_label_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_LABEL_COLOR);
   if (theDefaults)   tmp_s = theDefaults->get_ticklabelcolor();
   else               tmp_s = "white";
   if (tmp_s != get_ticklabel_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_TICKLABEL_COLOR);
   if (theDefaults)   tmp_i = theDefaults->get_ticklabeloffset();
   else               tmp_i = 2;
   if (tmp_i != ticklaboffset)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_TICKLABEL_OFFSET);
   if (theDefaults)   tmp_i = theDefaults->get_ticklabelsize();
   else               tmp_i = 10;
   if (tmp_i != ticklabsize)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_TICKLABEL_SIZE);
   if (theDefaults)   tmp_f = theDefaults->get_ticklabelangle();
   else               tmp_f = 0.0;
   if (tmp_i != ticklabangle)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_TICKLABEL_ANGLE);
   if (theDefaults)   tmp_i = theDefaults->get_axlabsize();
   else               tmp_i = 10;
   if (tmp_i != get_label_size())
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_LABEL_SIZE);
   if (theDefaults)   tmp_s = theDefaults->get_axlabfont();
   else               tmp_s = "helvetica";
   if (tmp_s != get_label_font())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_LABEL_FONT);
   if (theDefaults)   tmp_i = theDefaults->get_axlabfontstyle();
   else               tmp_i  = chips_normal;
   if (tmp_i != get_label_fontstyle())
   {
      tmp_s = v_fontstyle_to_string((V_FONTSTYLE_TYPE)tmp_i);
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_LABEL_FONTSTYLE);
   }
   if (theDefaults)   tmp_f = theDefaults->get_axisthickness();
   else               tmp_f = 2;
   if (tmp_f != thickness)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_THICKNESS);
   if (theDefaults)   tmp_s = theDefaults->get_majorgridcolor();
   else               tmp_s = "white";
   if (tmp_s != get_major_grid_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MAJGRD_COLOR);
   if (theDefaults)   tmp_s = theDefaults->get_minorgridcolor();
   else               tmp_s = "white";
   if (tmp_s != get_minor_grid_color())
      rval = set_property((void*)tmp_s.c_str(), SET_AXIS_MINGRD_COLOR);
   if (theDefaults)   tmp_i = theDefaults->get_majorgridstyle();
   else               tmp_i = chips_shortdash;
   if (tmp_i != grid_use_style)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MAJGRD_STYLE);
   if (theDefaults)   tmp_i = theDefaults->get_minorgridstyle();
   else               tmp_i = chips_dot;
   if (tmp_i != grid_min_style)
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_MINGRD_STYLE);
   if (theDefaults)   tmp_f = theDefaults->get_majorgridthickness();
   else               tmp_f = 1;
   if (tmp_f != grid_use_thick)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_MAJGRD_THICK);
   if (theDefaults)   tmp_f = theDefaults->get_minorgridthickness();
   else               tmp_f = 1;
   if (tmp_f != grid_min_thick)
      rval = set_property(static_cast<void*>(&tmp_f), SET_AXIS_MINGRD_THICK);
   if (ll_corner != chips_miter)
   {
      tmp_i = chips_miter; 
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_LLCORNER);
   }
   if (ur_corner != chips_miter)
   {
      tmp_i = chips_miter; 
      rval = set_property(static_cast<void*>(&tmp_i), SET_AXIS_URCORNER);
   }
   if (ll_corner_width != thickness) 
   {
      tmp_d = thickness;
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_LLCORNWID);
   }
   if (ur_corner_width != thickness) 
   {
      tmp_d = thickness;
      rval = set_property(static_cast<void*>(&tmp_d), SET_AXIS_URCORNWID);
   }

   return rval; 
}

bool ChipsAxis::get_scalable ()
{
   return axis_alg->get_scalable ();
}

int ChipsAxis::get_scaletype ()
{
   return axis_alg->get_scaletype ();
}

int ChipsAxis::set_data_coord_id (const string &data_id)
{
   ChipsAxisAlgorithmFactory::TransformType alg_type = ChipsAxisAlgorithmFactory::get_type (axis_alg);

   if (alg_type == ChipsAxisAlgorithmFactory::WCSTAN)
   {
      // ***** FILL IN ERRORS *********
      if (!data_coord_ids.empty ())
      {
         if (*data_coord_ids.begin () == data_id)
            return ERR_STATUS_GOOD;
         return ERR_STATUS_BAD;
      }

      // ALSO NEED TO CHECK FOR SAME TRANFORMS??
   }
   
   list<string>::iterator ii= find(data_coord_ids.begin (), data_coord_ids.end (), data_id);

   if (ii == data_coord_ids.end ())
      data_coord_ids.push_back (data_id);
   return ERR_STATUS_GOOD;
}

int ChipsAxis::remove_data_coord_id (const string &id)
{
   list<string>::iterator ii= find(data_coord_ids.begin (), data_coord_ids.end (), id);

   if (ii != data_coord_ids.end ())
      data_coord_ids.erase (ii);
   else
      return ERR_STATUS_BAD;

   return ERR_STATUS_GOOD;
}


string ChipsAxis::get_data_coord_id (int index)
{
   list<string>::iterator ii=data_coord_ids.begin ();
   int count=0;

   for (; ii!=data_coord_ids.end(); ii++)
   {
      if (count == index)
         return *ii;
   }

   return "";
}

void ChipsAxis::get_data_coord_id (list<string> &ids)
{
   ids = this->data_coord_ids;
}



void ChipsAxis::get_linear_range(double& outMin,double& outMax, bool transformed)
{
   outMin = (this->reverse) ? this->axis_alg->get_absolute_max (V_SCALE_LINEAR, transformed) : 
                              this->axis_alg->get_absolute_min (V_SCALE_LINEAR, transformed);
   outMax = (this->reverse) ? this->axis_alg->get_absolute_min (V_SCALE_LINEAR, transformed) :
                              this->axis_alg->get_absolute_max (V_SCALE_LINEAR, transformed);
}

double ChipsAxis::get_log_minimum()
{
   double minimum = 0.1;
   ChipsCartesianAxisAlgorithm *cart_alg = dynamic_cast <ChipsCartesianAxisAlgorithm *> (axis_alg);

   if (cart_alg)
      minimum = cart_alg->get_log_minimum ();
   
   return minimum;
}

double ChipsAxis::get_min (bool transformed)
{
   if (reverse)
      return axis_alg->get_max (transformed);
   return axis_alg->get_min (transformed);
}

double ChipsAxis::get_max (bool transformed)
{
   if (reverse)
      return axis_alg->get_min (transformed);
   return axis_alg->get_max (transformed);
}


double ChipsAxis::get_absolute_min (bool transformed)
{
   return axis_alg->get_absolute_min (-1, transformed);
}
                  
double ChipsAxis::get_absolute_max (bool transformed)
{
   return axis_alg->get_absolute_max (-1, transformed);
}

bool ChipsAxis::get_transform_valid ()
{
   return axis_alg->get_transform_valid ();
}

void ChipsAxis::transform_tick_limits ()
{
   return axis_alg->transform_tick_limits ();
}


double ChipsAxis::get_minor_int()
{ 
   return axis_alg->get_minor_interval (); 
}

double ChipsAxis::get_major_int()
{
   return axis_alg->get_major_interval ();
}


V_TICKMODE_TYPE ChipsAxis::get_major_mode()
{
   return axis_alg->get_major_mode ();
}


V_TICKMODE_TYPE ChipsAxis::get_minor_mode()
{
   return axis_alg->get_minor_mode ();
};


int ChipsAxis::get_major_num() 
{
   return axis_alg->get_major_num ();
}

int ChipsAxis::get_minor_num()
{
   return axis_alg->get_minor_num ();
}

int ChipsAxis::get_orientation ()
{
   return axis_alg->get_orientation ();
}

double ChipsAxis::get_norm_pos()
{
   return (axis_alg->get_orientation () == X_AXIS)?norm_pos[0][1]:norm_pos[0][0];
}

double ChipsAxis::get_orig_pos()
{
   if (axis_alg->get_orientation () == Y_AXIS)
      return orig_pos[0][0];
   else
      return orig_pos[0][1];
}

void ChipsAxis::serialize_ascii(ChipsStateStream& out)
{
   string cmd, atts = build_axis_property_string();
  
   ChipsVector tmp_pos[2];

   tmp_pos[0] = orig_pos[0];
   tmp_pos[1] = orig_pos[1];

   // convert the data to wcs coordinates if necessary
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
      {
         data_system_id = ds_coord->get_id_no_wcs ();
         ds_coord->convert_cs_to_cs (tmp_pos, tmp_pos, 2, DATA, data_system_id, DATA, "");
      }
   }
   if (axis_alg->get_orientation () == X_AXIS)
      out << "add_axis(X_AXIS," << tmp_pos[0][1] << ", ";
   else
      out << "add_axis(Y_AXIS," << tmp_pos[0][0] << ", ";
   if (reverse && !reverse_set)
      out << axis_alg->get_absolute_max (V_SCALE_LINEAR) << ", " << axis_alg->get_absolute_min (V_SCALE_LINEAR);
   else
      out << axis_alg->get_absolute_min (V_SCALE_LINEAR) << ", " << axis_alg->get_absolute_max (V_SCALE_LINEAR);
   out << ", \"id=" << id << " " << atts << "\")" << endl;
   cmd = build_axis_scale_string();
   if (cmd.length())  out << cmd << endl;
}

string ChipsAxis::build_axis_property_string(int set_flag)
{
   // set_flag = 0 -> string is for add_axis, otherwise for set_axis
   string tmp, props;
   double dtmp; 
   int tmpi;
   char buf[80];
   const char* cc;

   if (major_vis != theDefaults->get_majortickvisible())
   {
      if (props.length())  props.append(" ");
      if (major_vis)       props.append("majortick.visible=1");
         else              props.append("majortick.visible=0");
   }
   if (minor_vis != theDefaults->get_minortickvisible())
   {
      if (props.length())  props.append(" ");
      if (minor_vis)       props.append("minortick.visible=1");
         else              props.append("minortick.visible=0");
   }
   if (grid_maj_vis != theDefaults->get_majorgridvisible())
   {
      if (props.length())  props.append(" ");
      if (grid_maj_vis)    props.append("majorgrid.visible=1");
         else              props.append("majorgrid.visible=0");
   }
   if (grid_min_vis != theDefaults->get_minorgridvisible())
   {
      if (props.length())  props.append(" ");
      if (grid_min_vis)    props.append("minorgrid.visible=1");
         else              props.append("minorgrid.visible=0");
   }
   if (set_flag)
   {
      // tick label behavior is different if we're a border axis- default is
      // off for borders and on for regular axes
      if (id == "bx1" || id == "bx2" || id == "by1" || id == "by2")
      {
         if (ticklabvis)
         {
            if (props.length())  props.append(" ");
            props.append("ticklabel.visible=true");
         }
      }
      else if (!ticklabvis)
      {
         if (props.length())  props.append(" ");
         props.append("ticklabel.visible=false");
      }
   }
   if (tickformat != theDefaults->get_tickformat())
   {
      if (props.length())  props.append(" ");
      props.append("tickformat=");
      props.append(tickformat);
   }
   if (ticklabstyle != theDefaults->get_ticklabelstyle())
   {
      if ((cc = v_tickstyle_to_string(ticklabstyle)) != NULL)
      {
         if (props.length())  props.append(" ");
         props.append("ticklabel.style=");
         props.append(cc);
      }
   }

   tmp = get_ticklabel_color();
   if (theDefaults->get_ticklabelcolor() != tmp)
   {
      if (props.length()) props.append(" ");
      props.append("ticklabel.color=");
      props.append(tmp);
   }
   if (ticklabsize != theDefaults->get_ticklabelsize())
   {
      sprintf(buf, "ticklabel.size=%1d", ticklabsize);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   dtmp = theDefaults->get_ticklabelangle();
   if (ticklabangle > dtmp + DBL_EPSILON  || 
       ticklabangle < dtmp - DBL_EPSILON)
   {
      sprintf(buf, "ticklabel.angle=%2.1f", ticklabangle);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   if (theDefaults->get_ticklabelfont() != ticklabfont)
   {
      if (props.length()) props.append(" ");
      props.append("ticklabel.font=");
      props.append(ticklabfont);
   }
   if (theDefaults->get_ticklabelfontstyle() != ticklabfontstyle &&
       ((cc = v_fontstyle_to_string(ticklabfontstyle))!=NULL))
   {
      if (props.length()) props.append(" ");
      props.append("ticklabel.fontstyle=");
      props.append(cc);
   }
   dtmp = theDefaults->get_ticklabelvalign();
   if (ticklabvertalignment > dtmp + DBL_EPSILON  ||
       ticklabvertalignment < dtmp - DBL_EPSILON)
   {
      sprintf(buf, "ticklabel.valign=%2.1f", ticklabvertalignment);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   dtmp = theDefaults->get_ticklabelhalign();
   if (ticklabhorizalignment > dtmp + DBL_EPSILON  ||
       ticklabhorizalignment < dtmp - DBL_EPSILON)
   {
      sprintf(buf, "ticklabel.halign=%2.1f", ticklabhorizalignment);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   if (ticklaboffset != theDefaults->get_ticklabeloffset())
   {
      sprintf(buf, "ticklabel.offset=%1d", ticklaboffset);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   tmp = get_label_text();
   if (tmp.find('\\') != string::npos)
   {
      // need to escape backslash in text
      string::size_type pos=tmp.rfind("\\");
      while (pos != string::npos)
      {
         tmp.replace(pos, 1, "\\\\");
         if (pos > 0) pos=tmp.rfind("\\",pos-1);
            else      pos = string::npos;
      }
      tmpi = 1;       // escaped backslash- we'll need to quote label later   
   }
   else  tmpi = 0;    // no backslashes found  
   if (tmp.find('"') != string::npos)
   {
      // need to escape quotes in text
      string::size_type pos=tmp.rfind("\"");
      while (pos != string::npos)
      {
         tmp.replace(pos, 1, "\\\"");
         pos=tmp.rfind("\"",pos);
      }
   }
   if (tmp.length())
   {
      if (props.length()) props.append(" ");
      if (axis_alg->get_orientation () == X_AXIS)  props.append("x.label.text=\\\"");
         else                props.append("y.label.text=\\\"");
      props.append(tmp);
      props.append("\\\"");
      tmp = get_label_color();
      if (theDefaults->get_axlabcolor() != tmp)
      {
         props.append(" label.color=");
         props.append(tmp);
      }
      tmpi = get_label_size();
      if (tmpi != theDefaults->get_axlabsize())
      {
         sprintf(buf, " label.size=%1d", tmpi);
         props.append(buf);
      }
      float tmpf = get_label_angle();
      if (tmpf != theDefaults->get_labelangle())
      {
         if (axis_alg->get_orientation () == X_AXIS)
            sprintf(buf, " x.label.angle=%2.1f", tmpf);
         else 
            sprintf(buf, " y.label.angle=%2.1f", tmpf);
         props.append(buf);
      }
      tmp = get_label_font();
      if (theDefaults->get_axlabfont() != tmp)
      {
         props.append(" label.font=");
         props.append(tmp);
      }
      V_FONTSTYLE_TYPE vfs = get_label_fontstyle();
      if ((theDefaults->get_axlabfontstyle() != vfs) &&
          ((cc = v_fontstyle_to_string(vfs))!=NULL))
      {
         props.append(" label.fontstyle=");
         props.append(cc);
      }
      if ((tmpf = get_label_valign()) != theDefaults->get_axlabvalign())
      {
         sprintf(buf, " label.valign=%2.1f", tmpf);
         props.append(buf);
      }
      if ((tmpf = get_label_halign()) != theDefaults->get_axlabhalign())
      {
         sprintf(buf, " label.halign=%2.1f", tmpf);
         props.append(buf);
      }
      if (!get_label_visible())   props.append(" label.visible=0");
   }
   /** since we explicitly use the limits command later, these values
    ** don't need to be written out  
   if (automin != theDefaults->get_axisautomin())
   {
      if (props.length()) props.append(" ");
      if (automin)  props.append("automin=true");
         else       props.append("automin=false");
   }
   if (automax != theDefaults->get_axisautomax())
   {
      if (props.length()) props.append(" ");
      if (automax)  props.append("automax=true");
         else       props.append("automax=false");
   }
   **/
   tmp = get_color();
   if (tmp != theDefaults->get_axiscolor())
   {
      if (props.length()) props.append(" ");
      props.append("color=");
      props.append(tmp);
   }
   tmp = get_major_color();
   if (tmp != theDefaults->get_majortickcolor())
   {
      if (props.length()) props.append(" ");
      props.append("majortick.color=");
      props.append(tmp);
   }
   tmp = get_minor_color();
   if (tmp != theDefaults->get_minortickcolor())
   {
      if (props.length()) props.append(" ");
      props.append("minortick.color=");
      props.append(tmp);
   }
   tmp = get_major_grid_color();
   if (tmp != theDefaults->get_majorgridcolor())
   {
      if (props.length()) props.append(" ");
      props.append("majorgrid.color=");
      props.append(tmp);
   }
   tmp = get_minor_grid_color();
   if (tmp != theDefaults->get_minorgridcolor())
   {
      if (props.length()) props.append(" ");
      props.append("minorgrid.color=");
      props.append(tmp);
   }
   if (grid_use_thick != theDefaults->get_majorgridthickness())
   {
      if (props.length()) props.append(" ");
      if (grid_use_thick != 0.5)
      {
         sprintf(buf,"majorgrid.thickness=%-1.0f", grid_use_thick);
         props.append(buf);
      }
      else
         props.append("majorgrid.thickness=0.5");
   }
   if (grid_min_thick != theDefaults->get_minorgridthickness())
   {
      if (props.length()) props.append(" ");
      if (grid_use_thick != 0.5)
      {
         sprintf(buf,"minorgrid.thickness=%-1.0f", grid_min_thick);
         props.append(buf);
      }
      else
         props.append("minorgrid.thickness=0.5");
   }
   if (theDefaults->get_majorgridstyle() != grid_use_style &&
       ((cc = v_line_to_string(grid_use_style))!=NULL))
   {
      if (props.length()) props.append(" ");
      props.append("majorgrid.style=");
      props.append(cc);
   }
   if (theDefaults->get_minorgridstyle() != grid_min_style &&
       ((cc = v_line_to_string(grid_min_style))!=NULL))
   {
      if (props.length()) props.append(" ");
      props.append("minorgrid.style=");
      props.append(cc);
   }
   if (major_tickstyle != theDefaults->get_majortickstyle() &&
       (cc = v_tickstyle_to_string(major_tickstyle))!=NULL)
   {
      if (props.length()) props.append(" ");
      props.append("majortick.style=");
      props.append(cc);
   }
   if (minor_tickstyle != theDefaults->get_minortickstyle() &&
       (cc = v_tickstyle_to_string(minor_tickstyle))!=NULL)
   {
      if (props.length()) props.append(" ");
      props.append("minortick.style=");
      props.append(cc);
   }
   if (major_len != theDefaults->get_majorticklength())
   {
      sprintf(buf,"majortick.length=%1d", major_len);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   if (minor_len != theDefaults->get_minorticklength())
   {
      sprintf(buf,"minortick.length=%1d", minor_len);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   if (major_thickness != theDefaults->get_majortickthickness())
   {
      if (props.length()) props.append(" ");
      if (major_thickness != 0.5)
      {
         sprintf(buf,"majortick.thickness=%-1.0f", major_thickness);
         props.append(buf);
      }
      else
         props.append("majortick.thickness=0.5");
   }
   if (minor_thickness != theDefaults->get_minortickthickness())
   {
      if (props.length()) props.append(" ");
      if (minor_thickness != 0.5)
      {
         sprintf(buf,"minortick.thickness=%-1.0f", minor_thickness);
         props.append(buf);
      }
      else
         props.append("minortick.thickness=0.5");
   }
   if (thickness != theDefaults->get_axisthickness())
   {
      if (props.length()) props.append(" ");
      if (thickness != 0.5)
      {
         sprintf(buf,"thickness=%-1.0f", thickness);
         props.append(buf);
      }
      else
         props.append("thickness=0.5");
   }

   if (offset_par != theDefaults->get_axisparalleloffset())
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"offset.parallel=%3.2f", offset_par);
      props.append(buf);
   }
   if (offset_perp != theDefaults->get_axisperpendicularoffset())
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"offset.perpendicular=%3.2f", offset_perp);
      props.append(buf);
   }
   if (pad != theDefaults->get_axispad())
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"pad=%3.2f", pad);
      props.append(buf);
   }
   if ((cc = v_tickmode_to_string(get_major_mode())) !=
       theDefaults->get_majortickmode())
   {
      if (props.length()) props.append(" ");
      props.append("majortick.mode=");
      props.append(cc);
   }
   if ((cc = v_tickmode_to_string(get_minor_mode())) !=
       theDefaults->get_minortickmode())
   {
      if (props.length()) props.append(" ");
      props.append("minortick.mode=");
      props.append(cc);
   }
   if ((axis_alg->get_major_num () != (int)theDefaults->get_majortickcount()) &&
       get_major_mode() == chips_count)
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"majortick.count=%1d", axis_alg->get_major_num ());
      props.append(buf);
   }
   if ((axis_alg->get_minor_num () != (int)theDefaults->get_minortickcount()) &&
       get_minor_mode() == chips_count)
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"minortick.count=%1d", axis_alg->get_minor_num ());
      props.append(buf);
   }
   if ((get_major_int () != theDefaults->get_majortickinterval()) &&
       get_major_mode() == chips_interval)
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"majortick.interval=%f", get_major_int ());
      props.append(buf);
   }
   if ((axis_alg->get_minor_interval () != theDefaults->get_minortickinterval()) && 
      get_minor_mode() == chips_interval)
   {
      if (props.length()) props.append(" ");
      sprintf(buf,"minortick.interval=%f", axis_alg->get_minor_interval ());
      props.append(buf);
   }
   tmpi = theDefaults->get_axisdepth();
   if (depth != tmpi &&
       (tmpi != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (props.length()) props.append(" ");
      props.append(buf);
   }
   return props;
}

string ChipsAxis::build_axis_scale_string()
{
   string cmd = "";
   if (axis_alg->get_scaletype () == V_SCALE_LOG)
   {
      if (axis_alg->get_orientation () == X_AXIS) cmd.append("log_scale(X_AXIS)");
         else               cmd.append("log_scale(Y_AXIS)");
   }
   if (reverse_set && reverse)
   {
      if (cmd.length()) cmd.append("\n");
      cmd.append("reverse_axes(\"");
      cmd.append(id);
      if (axis_alg->get_orientation () == X_AXIS) cmd.append("\", X_AXIS)");
         else               cmd.append("\", Y_AXIS)");
   }
   
   double *arbpos = axis_alg->get_arbitrary_positions ();
   string *arbt   = axis_alg->get_arbitrary_text(); 
   unsigned int arbcnt = axis_alg->get_arbitrary_count ();
   if (axis_alg->get_major_mode () == chips_arbitrary && arbpos && arbcnt)
   {
      if (cmd.length()) cmd.append("\n");
      cmd.append("set_arbitrary_tick_positions(\"" + id + "\",[");
      for (unsigned int xx = 0; xx < arbcnt; xx++)
      {
         stringstream ostr;
         ostr << arbpos[xx];
         cmd.append(ostr.str());
         if (xx < arbcnt-1) cmd.append(",");
      }
      cmd.append("]");
      if (arbt) 
      {
         cmd.append(",[");
         for (unsigned int xx = 0; xx < arbcnt; xx++)
         {
            cmd.append("\"");
            cmd.append(arbt[xx]);
            if (xx < arbcnt-1) cmd.append("\",");
         }
         cmd.append("\"]");
      }
      cmd.append(")");
   }
   return cmd;
}

string ChipsAxis::build_set_axis_string(int setatts)
{
   string atts = "", scale = "", cmd = "";
   if (setatts)
   {
      int border = (id == "bx1" || id == "bx2" || id == "by1" || id == "by2");
      atts = build_axis_property_string(1);
      if (atts.length())
      {
         string tmpbase;
         cmd =  "set_axis(\"";
         // since this method is called when the axes are created by default,
         // we replace the axis id w/ 'ax1'|'ay1' to avoid issues with axes 
         // created after make plot
         if (border)
            cmd.append(id);
         else if (axis_alg->get_orientation () == X_AXIS)
         {
            alterid = theDefaults->get_axisxstem();
            alterid.append("1");
            cmd.append(alterid);
         }
         else  
         {
            alterid = theDefaults->get_axisystem();
            alterid.append("1");
            cmd.append(alterid);
         }
         cmd.append("\",\"");
         cmd.append(atts);
         cmd.append("\")");
      }
      if (!border)
      {
         scale = build_axis_scale_string();
         if (scale.length())
         {
            if (cmd.length())   cmd.append("\n");
            cmd.append(scale);
         }
      }
   }
   return cmd;
}

void ChipsAxis::set_algorithm_type (ChipsAxisAlgorithmFactory::TransformType type)
{
 
   ChipsAxisAlgorithmFactory::TransformType alg_type = ChipsAxisAlgorithmFactory::get_type (axis_alg);
   ChipsAxisAlgorithm *tmp_alg = axis_alg;

   if (type == alg_type)
      return;

   axis_alg = ChipsAxisAlgorithmFactory::create (type, tmp_alg->get_orientation (), this, tmp_alg);

   delete tmp_alg;
  
}

void ChipsAxis::set_algorithm_type (Transform *transform)
{
    set_algorithm_type (ChipsAxisAlgorithmFactory::get_algorithm_type (transform));
}


bool ChipsAxis::check_transform_compatibility (Transform *transform)
{
   return ChipsAxisAlgorithmFactory::check_transform_compatibility (transform, this->axis_alg);
}

ChipsAxisAlgorithmFactory::TransformType ChipsAxis::get_algorithm_type ()
{
   return (ChipsAxisAlgorithmFactory::get_type (axis_alg));
}


ChipsDataSystem *ChipsAxis::get_data_system (int index)
{
   string ds_id = "";
   if (this->data_coord_ids.empty ())
      return NULL;
   else
      ds_id = this->get_data_coord_id (index);

   if (ds_id.length () == 0)
      return NULL;

   return dynamic_cast <ChipsDataSystem *> (coord->get_coord (DATA, ds_id));

}

ChipsDataSystem *ChipsAxis::get_data_system (const string &ds_id)
{
   if (this->data_coord_ids.empty ())
      return NULL;

   if (ds_id.length () == 0)
      return NULL;

   return dynamic_cast <ChipsDataSystem *> (coord->get_coord (DATA, ds_id));
}

int ChipsAxis::validate_limits (const ChipsVector &coord, bool transformed)
{
   int rval = ERR_STATUS_GOOD;
   bool valid = true;
   ChipsDataSystem *ds;
   list<string>::iterator ds_id_iter = data_coord_ids.begin ();
   string ds_id;
   for (; valid && ds_id_iter!=data_coord_ids.end(); ds_id_iter++)
   {
      ds = this->get_data_system (*ds_id_iter);
      if (transformed)
         ds_id = *ds_id_iter;
      else
         ds_id = ds->get_id_no_wcs ();

      valid =  axis_alg->validate_coordinates (coord, ds_id);
   }

   if (!valid)
   {
      stringstream err_msg;
      rval = ERR_STATUS_BAD;
      err_msg << "Unable to set limits due to invalid coordinate " << coord[0] << "," << coord[1];

      cmd_manager->add_err(err_msg.str (),0,"axis");
   }

   return rval;

}




int ChipsAxis::validate_limits (double xmin, double xmax, double ymin, 
                                 double ymax, int dim, bool transformed)
{
   int rval = ERR_STATUS_GOOD;
   bool valid = true;
   ChipsDataSystem *ds;
   ChipsVector min_coord;
   ChipsVector max_coord;
   list<string>::iterator ds_id_iter = data_coord_ids.begin ();
   string ds_id;
   for (; valid && ds_id_iter!=data_coord_ids.end(); ds_id_iter++)
   {
      ds = this->get_data_system (*ds_id_iter);  
      if (transformed)
      {
         min_coord = ds->get_lower_bounds ("");
         max_coord = ds->get_upper_bounds ("");
      }
      else
      {
         min_coord = ds->get_lower_bounds ();
         max_coord = ds->get_upper_bounds ();
      }
      if ((dim & X_AXIS) == X_AXIS)
      {
         min_coord[0] = xmin;
         max_coord[0] = xmax;
      }
      if ((dim & Y_AXIS) == Y_AXIS)
      {
         min_coord[1] = ymin;
         max_coord[1] = ymax;
      }

      if (transformed)
         ds_id = *ds_id_iter;
      else
         ds_id = ds->get_id_no_wcs ();
         
      valid =  axis_alg->validate_coordinates (min_coord, ds_id);
      if (valid)
         valid =  axis_alg->validate_coordinates (max_coord, ds_id);
   }

   if (!valid)
   {
      stringstream err_msg;
      rval = ERR_STATUS_BAD;
      err_msg << "Unable to set limits due to invalid coordinates " << min_coord[0] << "," << min_coord[1] << " : " << max_coord[0] << "," << max_coord[1];

      cmd_manager->add_err(err_msg.str (),0,"axis");
   }

   return rval;

}

int ChipsAxis::format_axis_point (double pnt, string &formatted_pnt)
{
   this->axis_alg->format_point (this->tickformat, pnt, formatted_pnt);

   return ERR_STATUS_GOOD;
}

bool ChipsAxis::is_sexigesimal_format(string val)
{
   if (val.find("ra") != string::npos)   return true;
   if (val.find("RA") != string::npos)   return true;
   if (val.find("dec") != string::npos)  return true;
   if (val.find("DEC") != string::npos)  return true;
   return false;
}
