// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSAXISALGORITHM_HH
#define CHIPSAXISALGORITHM_HH

#include <string>
#include "ChipsDefinesP.hh"
#include "ChipsMatrix.hh"
#include "ChipsStateStream.hh"

class ChipsDataSystem;
class ChipsAxis;

using namespace std;

class ChipsAxisAlgorithm
{
  public:
    
   ChipsAxisAlgorithm (int orientation, ChipsAxis *axis);
   ChipsAxisAlgorithm (ChipsAxisAlgorithm &copy);
   virtual ~ChipsAxisAlgorithm ();
 
   // needed to copy base information between two unlike algorithms
   void operator = (ChipsAxisAlgorithm &copy);

   void               set_axis (ChipsAxis *axis) {this->axis = axis;};

   virtual double     get_min (bool transformed=true) = 0;
   virtual double     get_max (bool transformed=true) = 0;

   virtual double     get_absolute_min (int scaletype=-1, bool transformed=true) = 0;
   virtual int        set_absolute_min (double val, int scaletype=-1, bool transformed=true) = 0;
   virtual double     get_absolute_max (int scaletype=-1, bool transformed=true) = 0;
   virtual int        set_absolute_max (double val, int scaletype=-1, bool transformed=true) = 0;
   virtual int        set_absolute_minmax (double min, double max, int scaletype=-1, bool transformed=true) = 0;

   virtual void       set_scalable(bool val) = 0;
   bool               get_scalable() {return this->scalable;};
   virtual void       set_scaletype (int val) = 0;
   int                get_scaletype () {return this->scaletype;};

   virtual double*    get_tick_positions (bool transformed) = 0;
   virtual void       set_tick_positions (double *ticks, int count) = 0;
   unsigned int       get_tick_count () {return this->tickcnt;};

   string*            get_arbitrary_text () {return this->arbtext;};
   double*            get_arbitrary_positions () {return this->arbpos;};
   virtual void       set_arbitrary_positions (double *positions, string *text, int count) = 0;
   unsigned int       get_arbitrary_count () {return this->arbcnt;};

   bool               get_transform_valid () {return this->transform_valid;};
   virtual void       transform_tick_limits () = 0;

   virtual void       format_point (const string &tick_format, double pnt, string &formatted_pnt) = 0;
/*
   virtual Transform* process_transform (Transform *transform) = 0;
*/  

   string*            get_tick_text () {return this->tick_text;};
   int                get_tick_text_count () {return this->ttext_cnt;};

   double*            get_minortick_positions () {return this->minortickpos;};
   int                get_minortick_count ()    {return this->minortickcnt;};

   V_TICKMODE_TYPE    get_minor_mode (bool adjusted=false) 
                        {return (adjusted) ? this->minor_mode_adjusted : this->minor_mode;};
   virtual int        set_minor_mode (V_TICKMODE_TYPE mode) = 0;
   V_TICKMODE_TYPE    get_major_mode (bool adjusted=false) 
                        {return (adjusted) ? this->major_mode_adjusted : this->major_mode;};
   virtual int        set_major_mode (V_TICKMODE_TYPE mode) = 0;


   int                get_minor_num (bool adjusted=false) 
                        {return (adjusted) ? this->minor_num_adjusted : this->minor_num;};
   void               set_minor_num (int num);
   int                get_major_num () {return this->major_num;};
   void               set_major_num (int num);

   double             get_minor_interval () {return this->minor_int;};
   void               set_minor_interval (double interval);
   double             get_major_interval () {return this->major_int;};
   void               set_major_interval (double interval);

   bool               get_modified () {return this->modified;};

   int                get_orientation () {return this->orientation;};
   void               set_orientation (int orient);

   virtual int        calculate_tick_limits(bool automin, bool automax,
                                  double pad)=0;

   virtual int        update_tick_positions(bool &out_updates) = 0;
   virtual void       reset_string_labels(const string& tickformat) = 0;
   virtual double *** calculate_grid (int &out_line_count, int &out_slices, int type) = 0;

   virtual void       serialize (ChipsStateStream &out) = 0;
   virtual void       deserialize (ChipsStateStream &in) = 0;

   virtual ChipsAxisAlgorithm * clone () = 0;

   virtual bool       fix_minmax (double &min, double &max, bool transformed) = 0;
   virtual bool       validate_coordinates (const ChipsVector &coord, const string &ds_id) = 0;

   

  protected:
   virtual void       reset_adjusted_parameters () = 0;

   bool       scalable;
   int        scaletype;

   double*    tickpos;       // nice tick locations
   double*    arbpos;        // arbritrary tick locations
   string*    arbtext;       // arbitrary text labels
   string*  tick_text;       // unparsed latext tick text
   unsigned int ttext_cnt;   // number of text elements

   unsigned int tickcnt;     // number of nice ticks
   unsigned int arbcnt;      // number of arbitrary ticks

   double*     minortickpos;  // arbitrary minor tick locations
   int         minortickcnt; // number of arbitrary ticks

   V_TICKMODE_TYPE  major_mode;  // position by interval, count, nice
   V_TICKMODE_TYPE  major_mode_adjusted;  // position by interval, count, nice
   V_TICKMODE_TYPE  minor_mode;  // position by interval, count, nice
   V_TICKMODE_TYPE  minor_mode_adjusted;  // position by interval, count, nice
   int minor_num;            // total number of minor ticks to display
   int minor_num_adjusted;   // total number of minor ticks to display
   int major_num;            // total number of major ticks to display
   double     major_int;     // interval between major ticks
   double     minor_int;     // interval between minor ticks

   bool       arbitrary_set; // flag to indicate if the tickpos == arbpos
   bool       modified;      // flag to indicate whether any property has been modified
   int        orientation;   // the orientation of the axis

   bool       transform_valid; // flag to indicate if the transformed min/max is valid

   ChipsAxis *axis;          // axis associated with this alogrithm

};

#endif
   


   
