#include "ChipsAxisAlgorithmFactory.hh"
#include "ChipsCartesianAxisAlgorithm.hh"
#include "ChipsWCSTANAxisAlgorithm.hh"
#include "ChipsAxis.hh"
#include "Transform.hh"
#include "ChipsTransformFactory.hh"

ChipsAxisAlgorithm *ChipsAxisAlgorithmFactory::create (int type, 
                                             int orientation,
                                             ChipsAxis *axis,
                                             ChipsAxisAlgorithm *copy)
{
   ChipsAxisAlgorithm *alg;

   switch (type)
   {
      case ChipsAxisAlgorithmFactory::CARTESIAN:
      {
         if (copy)
         {
            // if a copy exists and its of the same type as the output
            // simply clone the copy
            if (dynamic_cast <ChipsCartesianAxisAlgorithm *>(copy) != NULL)
            {
               alg = copy->clone ();
               alg->set_axis (axis);
            }
            // otherwise create new ouput and copy the base information
            else
            {
               alg = new ChipsCartesianAxisAlgorithm (orientation, axis);
               *alg = *copy;
            }
         }
         else
            alg = new ChipsCartesianAxisAlgorithm (orientation, axis);
      }
      break;
      case ChipsAxisAlgorithmFactory::WCSTAN:
      {
         if (copy)
         {
            // if a copy exists and its of the same type as the output
            // simply clone the copy
            if (dynamic_cast <ChipsWCSTANAxisAlgorithm *>(copy) != NULL)
            {
               alg = copy->clone ();
               alg->set_axis (axis);
            }
            // otherwise create new ouput and copy the base information
            else
            {
               alg = new ChipsWCSTANAxisAlgorithm (orientation, axis);
               *alg = *copy;
            }
         }
         else
            alg = new ChipsWCSTANAxisAlgorithm (orientation, axis);
      }
      break;
      default:
         alg = NULL;
      break;
   }

   return alg;
}

ChipsAxisAlgorithmFactory::TransformType ChipsAxisAlgorithmFactory::get_type (ChipsAxisAlgorithm *alg)
{
   TransformType type = ChipsAxisAlgorithmFactory::UNKNOWN;

   if (dynamic_cast <ChipsCartesianAxisAlgorithm *>(alg))
      type = ChipsAxisAlgorithmFactory::CARTESIAN;
   else if (dynamic_cast <ChipsWCSTANAxisAlgorithm *>(alg))
      type = ChipsAxisAlgorithmFactory::WCSTAN;

   return type;
}

ChipsAxisAlgorithmFactory::TransformType ChipsAxisAlgorithmFactory::get_algorithm_type (Transform *transform)
{
   ChipsTransformFactory tf;
   ChipsTransformFactory::TransformType type = tf.get_transform_type (transform);

   if (type == ChipsTransformFactory::WCSTAN)
      return (ChipsAxisAlgorithmFactory::WCSTAN);
   else if ((type == ChipsTransformFactory::LINEAR2D) || !transform)
      return (ChipsAxisAlgorithmFactory::CARTESIAN);

   return ChipsAxisAlgorithmFactory::UNKNOWN;
}

bool ChipsAxisAlgorithmFactory::check_transform_compatibility (Transform *transform, ChipsAxisAlgorithm *axis_alg)
{
   ChipsTransformFactory tf;
   ChipsAxisAlgorithmFactory::TransformType alg_type = ChipsAxisAlgorithmFactory::get_type (axis_alg);
   ChipsTransformFactory::TransformType type = tf.get_transform_type (transform);

   if ((alg_type == ChipsAxisAlgorithmFactory::WCSTAN) && (type == ChipsTransformFactory::WCSTAN))
      return true;
   else if ((alg_type == ChipsAxisAlgorithmFactory::CARTESIAN) && ((type == ChipsTransformFactory::LINEAR2D)))
      return true;

   // if no transform is supplied then we assume compatibility with 
   // all algorithms
   if (!transform)
      return true;

   return false;
}

