// 
//  Copyright (C) 2005-2009,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsAxisBind.hh"
#include "ChipsAxisGraph.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"

#include <stack>
#include <queue>

#include <stdio.h>

extern ChipsCmdManager* cmd_manager;

ChipsAxisBind::ChipsAxisBind()
{
   bindings.clear();
   graph = new ChipsAxisGraph();

   record_elem.updated_axes.clear ();
   access_plot = NULL;
   accessed.clear();
}

ChipsAxisBind::~ChipsAxisBind()
{
   list<AXISBIND_ELEM_P_T>::iterator ii;

   // clean up graph 
   delete graph;
   graph = NULL;

   // delete all bindings
   for(ii=bindings.begin(); ii != bindings.end(); ++ii)
   {
      delete(*ii);
   }
   bindings.clear();
}

int ChipsAxisBind::add_aspect_ratio(ChipsAxis* a1, ChipsAxis*a2, double xx,
    double yy)
{
   int rval = ERR_STATUS_GOOD;

   if (a1 && a2)
   {
      if (a1->get_orientation() == a2->get_orientation())
         return ERR_STATUS_BAD; // aspect ratio must be cross-dim 

      graph->remove_edge(a1, a2);   // remove entry if it already exists 
      if (!graph->find_node(a1))  graph->add_node(a1);
      if (!graph->find_node(a2))  graph->add_node(a2);
      graph->add_edge(a1, a2, 1, xx, yy);   // add new entry  
      if (graph->check_for_cycle())
      { 
         graph->remove_edge(a1, a2);
         return ERR_CYC_DEP_ASP;  // cyclic dep 
      } 
   }
   return rval; 
}

int ChipsAxisBind::remove_aspect_ratio(ChipsAxis* a1, ChipsAxis*a2)
{
   if (a1 && a2 && graph->find_node(a1) && graph->find_node(a2)) 
   {
      graph->remove_edge(a1, a2);
      return ERR_STATUS_GOOD; 
   }
   return ERR_STATUS_BAD; 
}

int ChipsAxisBind::add_binding(ChipsAxis* a1, ChipsAxis* a2)
{
   int rval = ERR_STATUS_GOOD;

   if (a1 && a2)
   {
      if (a1->get_orientation() != a2->get_orientation()) 
         return ERR_CROSSDIM_BIND; // no longer support cross-dim bind

      // add entry to dependency graph 
      if (!graph->find_node(a1))  graph->add_node(a1);
      if (!graph->find_node(a2))  graph->add_node(a2);
      graph->add_edge(a1, a2, 2); 
      if (graph->check_for_cycle()) 
      {
         graph->remove_edge(a1, a2); 
         return  ERR_CYC_DEP_BIND;   // cyclic dep 
      }     

      if (!check_binding(a1, a2))
      {
         AXISBIND_ELEM_P_T ap = new AXISBIND_ELEM_T;
         if (ap) 
	 {
	    ap->axis1 = a1;
	    ap->axis2 = a2;
	    bindings.push_back(ap); 
	 }
	 else rval = ERR_STATUS_BAD; //alloc failed  
      }
      else rval = ERR_DUP_BINDING;   // duplicate 
   }
   else   rval = ERR_STATUS_BAD;     // insufficient # axes 
   return rval; 
}

int ChipsAxisBind::add_axis(ChipsAxis* a1)
{
   if (!a1)  return ERR_STATUS_BAD;     // no axis
   if (graph->find_node(a1))  return ERR_STATUS_BAD;  // axis already in graph
   graph->add_node(a1);
   return ERR_STATUS_GOOD;
}

int ChipsAxisBind::delete_binding(ChipsAxis* a1, ChipsAxis* a2)
{
   int rval = ERR_STATUS_GOOD;

   if (a1 && a2)
   {
      graph->remove_edge(a1, a2); 

      AXISBIND_ELEM_P_T ap;
      for(list<AXISBIND_ELEM_P_T>::iterator ii=bindings.begin();
          (ii != bindings.end()); ++ii)
      {
         if ((((*ii)->axis1 == a1) && ((*ii)->axis2 == a2)) ||
             (((*ii)->axis1 == a2) && ((*ii)->axis2 == a1)))
         {
            ap = *ii;
            delete(ap);
            ii = bindings.erase(ii);
         }
      }

      
      if ((a1->get_shared_ds ()) || (a2->get_shared_ds ()))
      {
         list<ChipsAxis*>::iterator ii;
         list <ChipsAxis*> tl;
         ChipsPlot *plot;

         if (a1->get_shared_ds ())
            tl = get_axes(a1);
         else
            tl = get_axes(a2);

         for(ii=tl.begin(); ii != tl.end(); ++ii)
         {
            plot = dynamic_cast <ChipsPlot *>((*ii)->get_parent ());
            plot->clone_data_systems (*ii);
            (*ii)->set_shared_ds (false);
         }
      }
   }
   else   rval = ERR_STATUS_BAD; 
   return rval; 
}


int ChipsAxisBind::remove_axis(ChipsAxis* a1)
{
   int rval = ERR_STATUS_GOOD;
   if (a1)
   {
      graph->remove_node(a1);

      AXISBIND_ELEM_P_T ap;
      list<AXISBIND_ELEM_P_T>::iterator ii=bindings.begin();
      while (ii != bindings.end())
      {
         if (((*ii)->axis1 == a1) || ((*ii)->axis2 == a1))
         {
            ap = *ii;
            delete(ap);
            ii = bindings.erase(ii);
         }
         else ++ii; 
      }
   }
   else   rval = ERR_STATUS_BAD; 
   return rval; 
}

int ChipsAxisBind::synch_axes(ChipsAxis* ax1, ChipsAxis* ax2, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (ax1 && ax2)
   {
      list<ChipsAxis*>::iterator ii;
      list <ChipsAxis*> tl = get_axes(ax2);
      for(ii=tl.begin(); ii != tl.end(); ++ii)
         if ((*ii) != ax1)   (*ii)->synchronize(*ax1);

      update_connected_nodes (ax1, ax1->get_absolute_min (), ax1->get_absolute_max (), true, AUTOLIM_DEF, 1, rec);
   }
   else   rval = ERR_STATUS_BAD; 
   return rval; 
}

int ChipsAxisBind::compile_synch_data (ChipsAxis *ax, double** d_p, int** i_p, string** s_p)
{
   int dp_size, ip_size, sp_size;
   int *ip;
   string *sp;
   int coord_size = 0;
   int rr=0;
   

   if (!ax)
      return ERR_STATUS_BAD;

   list<ChipsAxis*>::iterator ii;
   list <ChipsAxis*> tl = get_axes(ax);
   list <string> *coord_ids = new list<string>[tl.size ()];
   list <string>::iterator coord_iter;


   // compile generic properties for this axis
   ax->compile_synch_data(d_p, &ip, &sp);
   ip_size = ip[0];
   dp_size = ip[1];
   sp_size = ip[2];


   // create space for each axes ref id
   *i_p = new int[ip_size + (tl.size ()*2) + 1+2];
   (*i_p)[ip_size] = tl.size ();
   for (rr=0; rr<ip_size; ++rr)
      (*i_p)[rr] = ip[rr];

   delete []ip;
   ip = *i_p;

   // copy the original sizes
   ip[ip_size] = ip_size;
   ip[ip_size+1] = sp_size;

   // set the new ip_size
   ip[0] = ip_size + (tl.size ()*2) + 1 +2;

   // collect all the data systems and reference ids
   for(rr=0,ii=tl.begin(); ii != tl.end(); ++ii,++rr)
   {
      (*ii)->get_data_coord_id (coord_ids[rr]);
      ip[rr*2+ip_size+2] = (int)(*ii)->get_ref_cnt ();
      ip[rr*2+1+ip_size+2] = coord_ids[rr].size ();
      coord_size += coord_ids[rr].size ();
   }

   // add the data coord ids to the string array
   *s_p = new string[sp_size + coord_size];
   ip[2] = sp_size + coord_size;
   for (rr=0; rr<sp_size; rr++)
      (*s_p)[rr] = sp[rr];

   rr=sp_size;
   for (int jj=0; jj<tl.size (); jj++)
   {
      for (coord_iter=coord_ids[jj].begin(); coord_iter!=coord_ids[jj].end(); coord_iter++)
         (*s_p)[rr++] = *coord_iter;
   }

   delete []sp;

   return ERR_STATUS_GOOD;
}

int ChipsAxisBind::check_valid_binding(ChipsAxis* a1, ChipsAxis* a2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsAxis *tmp_ax = NULL;

   if (a1 && a2)
   {
      list<ChipsAxis*>::iterator ii;
      list <ChipsAxis*> tl = get_axes(a2);
      
      // check if the axis algorithms match 
      if (a1->get_algorithm_type () != a2->get_algorithm_type ())
      {

         if (a2->get_algorithm_type () != ChipsAxisAlgorithmFactory::CARTESIAN)
         {
            rval=ERR_STATUS_BAD;
            cmd_manager->add_err("Cannot bind a WCS axis to a Cartesian axis. It must be Cartesian to WCS.",rval, a2->get_id (),0);
         }   
         else if (aspect_ratio_set (a2))
         {
            rval=ERR_STATUS_BAD;
            cmd_manager->add_err("Cannot bind different axis types which have aspect ratios set.",rval, a2->get_id (),0);
         }
         else
         {
            ChipsDataSystem *ds;
            list <string> data_ids;
            list <string>::iterator data_ids_iter;

            // if they don't check to see that none of data systems have
            // objects already associated with them
            for(ii=tl.begin(); ii != tl.end(); ++ii)
            {
               data_ids.clear ();
               tmp_ax = *ii;
               tmp_ax->get_data_coord_id (data_ids);
   
               data_ids_iter = data_ids.begin ();
               for (; data_ids_iter!=data_ids.end(); data_ids_iter++)
               {
                  ds = tmp_ax->get_data_system (*data_ids_iter);
                  if (ds->get_num_objs () > 0)
                     cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,tmp_ax->get_id (),0,"object");
               }
            }
         }
      }
   }
   else
     rval=ERR_OBJ_DNE;

   return rval;

}



int ChipsAxisBind::check_binding(ChipsAxis* a1, ChipsAxis* a2, bool cross)
{
   int found = 0;
   list<AXISBIND_ELEM_P_T>::iterator ii;
   AXISBIND_ELEM_P_T ap = NULL;

   if (a1 && a2) 
   {
      for(ii=bindings.begin(); (!found) && (ii != bindings.end()); ++ii)
      {
	 ap = *ii;
         if (!cross)
            found = (((ap->axis1 == a1) && (ap->axis2 == a2)) ||
                     ((ap->axis1 == a2) && (ap->axis2 == a1)));
         else 
            found = (((ap->axis1 == a1) && (ap->axis2 == a2) && 
                     (ap->axis1->get_parent() != ap->axis2->get_parent())) ||
                     ((ap->axis1 == a2) && (ap->axis2 == a1) && 
                     (ap->axis1->get_parent() != ap->axis2->get_parent())));
      }
   }
   else if (a1 || a2) 
   {
      ChipsAxis* tmp;
      if (a1)   tmp = a1;
         else   tmp = a2;

      for(ii=bindings.begin(); (!found) && (ii != bindings.end()); ++ii)
      {
	 ap = *ii;
         if (!cross)
            found = ((ap->axis1 == tmp) || (ap->axis2 == tmp));
         else 
            found = (((ap->axis1 == tmp) || (ap->axis2 == tmp)) &&
                     (ap->axis1->get_parent() != ap->axis2->get_parent()));
      }
   }
   return found;
}

list <ChipsAxis*> ChipsAxisBind::get_connections(ChipsAxis* ax)
{
   list <ChipsAxis*> alist;
   alist.clear();
   if (ax)
   {
      vector<ChipsAxis*> vec = graph->list_attached_vertices_dfs(ax);
      vector<ChipsAxis*>::iterator vi;
      for (vi = vec.begin(); vi != vec.end(); vi++)   alist.push_back(*vi);
   }
   return alist;
}

bool ChipsAxisBind::aspect_ratio_set(ChipsAxis* ax)
{
   double max; 
   bool val = false; 
   if (!ax)  return false;
   list <ChipsAxis*> axes = get_connections(ax); 
   list <ChipsAxis*>::iterator ai;

   for (ai = axes.begin(); ai != axes.end() && !val; ai++) 
   {
      max = 0;
      graph->get_edge_max((*ai), max); 
      if (max)  val = true; 
   }
   return val; 
}

int ChipsAxisBind::update_connected_nodes(ChipsAxis* ax, double min, double max,
   bool transformed, int autolim, int center, bool rec, double *altmin, double *altmax, string altid)
{
   if (!ax)  return ERR_STATUS_BAD;

   ChipsDataSystem *ds = NULL;

   queue <ChipsAxis*> Q;
   vector<ChipsAxis*> vec = graph->list_attached_vertices_bfs(ax);

   ChipsAxis *an = NULL;
   ChipsAxis *cur = ax;
   ChipsPlot *plot = dynamic_cast <ChipsPlot *>(ax->get_parent ());
   double cur_range[2] = {min,max};
   double an_range[2] = {-1,1};
   double d1, d2;

   // perform the initial update
   if (altid.size () > 0)
   {
      if (altmin)
         an_range[0] = *altmin;
      if (altmax)
         an_range[1] = *altmax;
      an = plot->get_xaxis (altid);
      if (!an)
         an = plot->get_yaxis (altid);
   
      if (!(accessed.count(an) && accessed[an]))
      {

         double sc = plot->get_scale_factor(cur->get_orientation());
         double dist, ctr, pixw=1, pixh=1;

         if ((ax->get_orientation () == X_AXIS) && (an->get_orientation () == Y_AXIS))
            ds = plot->get_datasystem (ax, an);
         else if ((ax->get_orientation () == Y_AXIS) && (an->get_orientation () == X_AXIS))
            ds = plot->get_datasystem (an, ax);

         if (ds)
         {
            pixw = ds->get_unit_pix_w();
            pixh = ds->get_unit_pix_h();
         }

         if (transformed)
         {
            if (an->get_orientation () == X_AXIS)
               compute_data_range (ds, an_range, cur_range, 1, an_range, cur_range);
            else
               compute_data_range (ds, cur_range, an_range, 1, cur_range, an_range);
         }

         // check if the two axes are connected
         if (!graph->get_edge_values(an, ax, d1, d2))
         {
            if (cur->get_orientation () == X_AXIS)
               dist = fabs (((cur_range[1]-cur_range[0])/pixw * pixh)/ 2.0 * sc);
            else
               dist = fabs (((cur_range[1]-cur_range[0])/pixh * pixw)/ 2.0 * sc);

            ctr = (an_range[1] - an_range[0])/2.0 + an_range[0];
            if ((dist > ctr - FLT_EPSILON) && (dist < ctr + FLT_EPSILON))
               dist = ctr; // reduce round off
            an_range[0] = ctr - dist;
            an_range[1] = ctr + dist;
         }
         else 
         {
            // add this axis graph to the list of axes which need to
            // be adjusted
            vector<ChipsAxis*> vec2 = graph->list_attached_vertices_bfs(an); 
            vec.resize (vec.size () + vec2.size ());
             
            for (int ii=0; ii<vec2.size (); ii++)
               vec.push_back (vec2[ii]);
         }

         an->adjust_minmax(an_range[0], an_range[1], autolim, rec, false);

         accessed[an] = 1;

         transformed = false;
      }
   }


   // update the current axis and modify the limits
   if (!(accessed.count(cur) && accessed[cur]))
   {
      cur->adjust_minmax(cur_range[0], cur_range[1], autolim, rec, transformed);
      accessed[cur] = 1;
   
      if (cur->get_orientation () == X_AXIS)
         plot->update_data_coord_limits (cur, cur_range[0], cur_range[1], an, an_range[0], an_range[1], transformed);
      else
         plot->update_data_coord_limits (an, an_range[0], an_range[1], cur, cur_range[0], cur_range[1], transformed);
   }

   if (this->record_elem.track_updates)
   {
      this->record_elem.updates_recorded |= 1;
      this->record_elem.updated_axes[plot->get_id ()].push_back (cur->get_id());
      if (an)
         this->record_elem.updated_axes[plot->get_id ()].push_back (an->get_id());

   }

   // go through the graph and update all aspect ratios
   for (int ii=1; ii<vec.size (); ii++)
   {
      bool tmp_transformed = transformed;
      an = vec[ii];
      if (!graph->get_edge_values(an, cur, d1, d2))
      {
         if (accessed.count(an) && accessed[an])
         {
            Q.push (an);
            continue;
         }

         plot = dynamic_cast <ChipsPlot *>(an->get_parent ());

         if (an->get_orientation() != cur->get_orientation())
         {
            // aspect
            double mn;

            if (center)
            {
               an->get_linear_range(an_range[0], an_range[1], true);
               tmp_transformed = true;
            }
            else if ((altid.size () == 0) && altmin && altmax)
            {
               an_range[0] = *altmin;
               an_range[1] = *altmax;
            }
            else
            {
               double tmp_min, tmp_max;
               if (plot->find_axis_min_max(an,
                  an->get_orientation(), &tmp_min, &tmp_max, &mn,
                  &an_range[0], &an_range[1]) != ERR_STATUS_GOOD)
               {
                  an_range[0] = an->get_absolute_min(true);
                  an_range[1] = an->get_absolute_max(true);
               }
               tmp_transformed = true;
            }
            
            if (an->get_orientation () == X_AXIS)
               ds = plot->get_datasystem(an,cur);
            else
               ds = plot->get_datasystem(cur,an);
            if (ds)
            {
               double sc = plot->get_scale_factor(cur->get_orientation());
               double dist, ctr,
                      pixw = ds->get_unit_pix_w(),
                      pixh = ds->get_unit_pix_h();
               if (cur->get_orientation () == X_AXIS)
               {
                  cur_range[0] = cur->get_absolute_min (tmp_transformed);
                  cur_range[1] = cur->get_absolute_max (tmp_transformed);

                  if (tmp_transformed)
                     compute_data_range (ds, cur_range, an_range, 1, cur_range, an_range);

                  dist = fabs (((cur_range[1]-cur_range[0])/pixw * pixh)/ 2.0 * sc);
               }
               else
               {
                  cur_range[0] = cur->get_absolute_min (tmp_transformed);
                  cur_range[1] = cur->get_absolute_max (tmp_transformed);
                  if (tmp_transformed)
                     compute_data_range (ds, an_range, cur_range, 1, an_range, cur_range);
                  dist = fabs (((cur_range[1]-cur_range[0])/pixh * pixw)/ 2.0 * sc);
               }
               ctr = (an_range[1] - an_range[0])/2.0 + an_range[0];
               if ((dist > ctr - FLT_EPSILON) && (dist < ctr + FLT_EPSILON))
                  dist = ctr; // reduce round off
               an_range[0] = ctr - dist;
               an_range[1] = ctr + dist;

               an->adjust_minmax(an_range[0], an_range[1], autolim, rec, false);

               if (an->get_orientation () == X_AXIS)
                  plot->update_data_coord_limits (an, an_range[0], an_range[1], cur, cur_range[0], cur_range[1], false);
               else
                  plot->update_data_coord_limits (cur, cur_range[0], cur_range[1], an, an_range[0], an_range[1], false);
            }
         }
         else
         {
            cur->get_linear_range(cur_range[0], cur_range[1], true);
            an->adjust_minmax(cur_range[0], cur_range[1], autolim, rec, true);

            if (an->get_orientation () == X_AXIS)
               plot->update_data_coord_limits (an, cur_range[0], cur_range[1], NULL, -1, 1, true);
            else
               plot->update_data_coord_limits (NULL, -1, 1, an, cur_range[0], cur_range[1], true);

         }
         Q.push (an);
         accessed[an] = 1;
         if (this->record_elem.track_updates)
         {
            this->record_elem.updates_recorded |= 1;
            this->record_elem.updated_axes[plot->get_id ()].push_back (an->get_id());
         }
      }
      else if (!Q.empty ())
      {
         cur = Q.front ();
         Q.pop ();
         --ii; // redo the last axis
      }
      else
         cur = an;
   }

   return ERR_STATUS_GOOD;
}

int ChipsAxisBind::update_connected_node_properties(ChipsAxis* ax, bool rec)
{
   if (!ax)  return ERR_STATUS_BAD;
   vector<ChipsAxis*> vec = graph->list_attached_vertices_dfs(ax);
   vector<ChipsAxis*>::iterator vi;
   ChipsAxis *an;
   int finished = 0;

   stack<ChipsAxis*>curnode;
   ChipsAxis *cur;
   double d1, d2;
   for (vi = vec.begin(); vi != vec.end() && !finished; vi++)
   {
      an = (*vi);
      if (accessed.count(an) && accessed[an]) 
      {
         finished = 1;    // already handled this node- we're done 
      }
      else if (curnode.empty())
      {
         // empty stack- first element
//         an->update_tick_positions();
         curnode.push(an); 

         if (this->record_elem.track_updates)
         {
            this->record_elem.updates_recorded |= 1;
            this->record_elem.updated_axes[an->get_parent()->get_id ()].push_back (an->get_id());
         }
      }
      else 
      {
         int done = 0;
         while (!curnode.empty() && !done)
         {
            cur = curnode.top();
            if (!graph->get_edge_values(an, cur, d1, d2))
            {
               // found edge 
//               an->update_tick_positions();
               curnode.push(an); 
               done = true; 
               if (this->record_elem.track_updates)
               {
                  this->record_elem.updates_recorded |= 1;
                  this->record_elem.updated_axes[an->get_parent()->get_id ()].push_back (an->get_id());
               }
            } 
            else    curnode.pop(); 
         } 
      }
      accessed[an] = 1;
   }
   return ERR_STATUS_GOOD;
}

list <ChipsAxis*> ChipsAxisBind::get_axes(ChipsAxis* ax)
{
   list <ChipsAxis*> alist;
   alist.clear();
   if (ax)
   {
      vector<ChipsAxis*> vec = graph->list_attached_vertices_dfs(ax);
      vector<ChipsAxis*>::iterator vi;
      list<AXISBIND_ELEM_P_T>::iterator ii;
      list <ChipsAxis*> checklist;
      list <ChipsAxis*>::iterator axisp;
      int  cc, iglen;
      int* ignore;

      // once an entry has been placed on our checklist, we can
      // ignore it when checking other potential bound axes
      iglen = bindings.size();
      ignore = new int[iglen];
      for (cc = iglen; cc--; )  ignore[cc] = 0;

      // the requested axis is the initial element in the checklist
      checklist.push_back(ax);

      while(!checklist.empty()) 
      {
         axisp = checklist.begin();
	 cc = 0;
         for(ii=bindings.begin(); ii != bindings.end(); ++ii)
         {
            if (!ignore[cc])
            {
	       // if either axis contains the check item, add the other
	       // axis to the checklist and ignore it during future checks 
               if (((*ii)->axis1 == *axisp))
               {
                  checklist.push_back((*ii)->axis2);
                  ignore[cc] = 1;
               }
               else if (((*ii)->axis2 == *axisp))
               {
                  checklist.push_back((*ii)->axis1);
                  ignore[cc] = 1;
               }
            }
            cc++;
         } 
         alist.push_back(*axisp);
         checklist.pop_front();
      }
      delete[] ignore;
   }
   return alist;
}

void ChipsAxisBind::serialize(ChipsStateStream& out)
{
   int sz = bindings.size();
   out.write_int(&sz);
   out.write_int(&sz);

   list<AXISBIND_ELEM_P_T>::iterator ii;
   ChipsAxis* axp;
   string id;

   for(ii=bindings.begin(); ii != bindings.end(); ++ii)
   {
      axp = (*ii)->axis1;
      id = axp->get_parent()->get_id();
      out.write_string(&id);
      id = axp->get_id();
      out.write_string(&id);
      axp = (*ii)->axis2;
      id = axp->get_parent()->get_id();
      out.write_string(&id);
      id = axp->get_id();
      out.write_string(&id);
   }

   this->graph->serialize (out);
}

void ChipsAxisBind::deserialize (ChipsFrame *cf, ChipsStateStream &out)
{
   this->graph->deserialize (cf, out);
}

void ChipsAxisBind::serialize_ascii(ChipsStateStream& out)
{
   list<AXISBIND_ELEM_P_T>::iterator ii;
   ChipsAxis* axp;
   string p1, p2, a1, a2;
   bool wrote = 0;

   for(ii=bindings.begin(); ii != bindings.end(); ++ii)
   {
      axp = (*ii)->axis1;
      p1 = axp->get_parent()->get_id();
      a1 = axp->get_id();
      axp = (*ii)->axis2;
      p2 = axp->get_parent()->get_id();
      a2 = axp->get_id();

      if (!((a1 == "by1" && a2 == "by2" && p1 == p2) ||
          (a1 == "by2" && a2 == "by1" && p1 == p2) ||
          (a1 == "bx2" && a2 == "bx1" && p1 == p2) ||
          (a1 == "ax1" && a2 == "bx2" && p1 == p2) ||
          (a1 == "ay1" && a2 == "by2" && p1 == p2) ||
          (a2 == "bx2" && a1 == "bx1" && p1 == p2)))
      {
         out << "bind_axes(\""<<p1<<"\",\""<<a1<<"\",\""<<p2<<"\",\""<<a2<<"\")"<<endl;
         wrote = 1;
      }
   }

   // ensure all axes limits have been written out
   wrote |= serialize_all_nodes(out);

   if (wrote)  out << endl;  // add blank line if we wrote a bind
}


// return string containing lines of bound frame pairs
string ChipsAxisBind:: get_bindings_list(ChipsAxis* ax)
{
   listings.clear(); 
   if (ax)
   {  // return bindings for the provided axis
      list<ChipsAxis*>::iterator ii;
      list <ChipsAxis*> tl = get_axes(ax);
      string front = "    ";
      front.append(ax->get_parent()->get_id());
      front.append(":");  
      front.append(ax->get_id());
      front.append("   ->   ");
      for(ii=tl.begin(); ii != tl.end(); ++ii)
         if ((*ii) != ax)  
         {
            listings.append(front);
            listings.append((*ii)->get_parent()->get_id());
            listings.append(":");
            listings.append((*ii)->get_id());
            listings.append("\n");
         } 
   }
   else 
   {  // return all bindings 
      list<AXISBIND_ELEM_P_T>::iterator ii;
      ChipsAxis *ax1, *ax2;
      for(ii=bindings.begin(); ii != bindings.end(); ++ii)
      {
         ax1 = (*ii)->axis1;
         ax2 = (*ii)->axis2;
         listings.append("    ");
         listings.append(ax1->get_parent()->get_id());
         listings.append(":");
         listings.append(ax1->get_id());
         listings.append("   ->   ");
         listings.append(ax2->get_parent()->get_id());
         listings.append(":");
         listings.append(ax2->get_id());
         listings.append("\n");
      }
   }
   return listings;
}

int ChipsAxisBind::record_updates (bool flag)
{
   this->record_elem.track_updates = flag;
   
   // if tracking updates then clear the update list
   if (this->record_elem.track_updates)
   {
      record_elem.updated_axes.clear ();
      record_elem.updates_recorded = 0;
   }

   return ERR_STATUS_GOOD;
}

int ChipsAxisBind::get_recorded_plots (list <string> &plot_ids)
{
   map <string, list <string> >::iterator ii;

   for (ii=this->record_elem.updated_axes.begin (); ii!=this->record_elem.updated_axes.end (); ii++)
      plot_ids.push_back (ii->first);

   return ERR_STATUS_GOOD;
}

int ChipsAxisBind::get_recorded_axes (string plot_id, list <string> &axis_ids)
{
   list <string>::iterator ii;

   // return if the plot id is not in the update list
   if (this->record_elem.updated_axes.find (plot_id) == this->record_elem.updated_axes.end ())
      return ERR_STATUS_BAD;

   axis_ids = this->record_elem.updated_axes[plot_id];

   return ERR_STATUS_GOOD;

}

int ChipsAxisBind::adjust_axis_minmax (ChipsAxis *a1, double min, double max, bool transformed, int autos, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (!a1)
      return ERR_STATUS_BAD;

   rval = a1->adjust_minmax(min, max, autos, rec, transformed);

   if (!rval && this->record_elem.track_updates) 
   {
      this->record_elem.updates_recorded |= 1;
      this->record_elem.updated_axes[a1->get_parent()->get_id ()].push_back (a1->get_id());
   }

   return rval;
}


int ChipsAxisBind::adjust_axis_min_or_max (ChipsAxis *a1, double val, int flag, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsAxis *ax = NULL;

   if (!a1)
      return ERR_STATUS_BAD;;

   list <ChipsAxis *>::iterator ii;
   list <ChipsAxis *> bound_list = this->get_axes (a1);

   for (ii=bound_list.begin (); ii!=bound_list.end () && rval == ERR_STATUS_GOOD; ii++)
   {

      ax = *ii;
      switch (flag)
      {
         case 0: // flag == 0 => adjust just min
            rval = ax->adjust_min(val, rec);
         break;
         case 1: // flag == 1 => adjust just max
            rval = ax->adjust_max(val, rec);
         break;
         case 2: // flag == 2 => adjust log min
            rval = ax->adjust_logmin(val, rec);
         break;
      }
 
      if (!rval && this->record_elem.track_updates)
      {
         this->record_elem.updates_recorded |= 1;
         this->record_elem.updated_axes[a1->get_parent()->get_id ()].push_back (a1->get_id());
      }
   }
 
   return rval;    
}

int ChipsAxisBind::update_property (list <ChipsAxis *> &axis_list, void *val, int property, bool rec)
{
 
   int rval=ERR_STATUS_GOOD;

   map <ChipsAxis *, bool> found_axes;
   map <ChipsAxis *, bool>::iterator kk;
   list <ChipsAxis *>::iterator ii;
   list <ChipsAxis *>::iterator jj;
   list <ChipsAxis *> bound_list;

   // go through the list of axes and find a list of unique bindings
   for(ii=axis_list.begin (); ii != axis_list.end (); ii++)
   {
      // if the axis has not been and the list to the map
      if (found_axes.count (*ii) == 0)
      {
         bound_list = this->get_axes (*ii);
         for (jj=bound_list.begin (); jj!=bound_list.end (); jj++)
            found_axes[*jj] = true;
      }
   }

   for (kk=found_axes.begin (); kk!=found_axes.end (); kk++)
   {
      rval |= kk->first->set_property(val, property, rec);
      if (this->record_elem.track_updates)
      {
         this->record_elem.updates_recorded |= 2;
         this->record_elem.property_recorded = property;
         this->record_elem.updated_axes[kk->first->get_parent()->get_id ()].push_back (kk->first->get_id());
      }
   }

   return rval;

  
}

int ChipsAxisBind::update_property (ChipsAxis *a1, void *val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (!a1)
      return ERR_STATUS_BAD;

   list <ChipsAxis*> tl = this->get_axes(a1);
   list <ChipsAxis*>::iterator jj;
   for(jj=tl.begin(); jj != tl.end(); ++jj)
   {
      rval |= (*jj)->set_property(val, property, rec);
      if (this->record_elem.track_updates)
      {
         this->record_elem.updates_recorded |= 2;
         this->record_elem.property_recorded = property;
         this->record_elem.updated_axes[(*jj)->get_parent()->get_id ()].push_back ((*jj)->get_id());
      }
   }

   return rval;
}

int ChipsAxisBind::update_all_nodes(int dim, double min, double max,
   bool transformed, int autolim, int center, bool rec, double *altmin, 
   double *altmax)
{
   if ((dim & XY_AXIS) == 0)  return ERR_STATUS_BAD;

   vector<ChipsAxis*> vec_all = graph->list_all_vertices();
   vector<ChipsAxis*> vec;
   vector<ChipsAxis*>::iterator vi;
   
   for (vi = vec_all.begin(); vi != vec_all.end(); vi++)
   {
       // if the axis is in specified dimension, is not in the list, and
       // is a child of the accessed plot (or no access_plot specified)- reset
       if ((((*vi)->get_orientation() & dim) == dim) && !accessed.count((*vi)) &&
          (!access_plot || ((*vi)->get_parent() == access_plot)))
           accessed[(*vi)] = 0;  
   }

   for (map<ChipsAxis*, int>::iterator mi = accessed.begin(); mi != accessed.end(); mi++)
      if (mi->second == 0) 
      {
         // update axes which have not already been set 
         update_connected_nodes(mi->first, min, max, transformed, autolim, center, rec, altmin, altmax);
      }
   return ERR_STATUS_GOOD;
}

void ChipsAxisBind::reset_access_list(ChipsPlot* cplot)
{
   access_plot = cplot;
   accessed.clear(); 
}


void ChipsAxisBind::compute_data_range (ChipsDataSystem *ds, double xrange[], double yrange[], int direction, double out_xrange[], double out_yrange[])
{
   ChipsVector min_pnt(xrange[0], yrange[0]);
   ChipsVector max_pnt(xrange[1], yrange[1]);
   ChipsVector tmp_pnt (3);

   if (!ds)
   {
      if (out_xrange)
      {
          out_xrange[0] =  xrange[0];
          out_xrange[1] =  xrange[1];
      }
      if (out_yrange)
      {
          out_yrange[1] =  yrange[1];
          out_yrange[0] =  yrange[0];
      }
   }
   else
   {

      string ds_id = ds->get_id_no_wcs ();
      if (direction == 0)
      {
         ds->convert_cs_to_cs (&min_pnt, &tmp_pnt, 1, DATA, ds_id,  DATA, "");

         if (out_xrange)    out_xrange[0] =  tmp_pnt[0];
         if (out_yrange)    out_yrange[0] =  tmp_pnt[1];
         ds->convert_cs_to_cs (&max_pnt, &tmp_pnt, 1, DATA, ds_id, DATA, "");
         if (out_xrange)    out_xrange[1] =  tmp_pnt[0];
         if (out_yrange)    out_yrange[1] =  tmp_pnt[1];
      }
      else
      {
         ds->convert_cs_to_cs (&min_pnt, &tmp_pnt, 1, DATA, "", DATA, ds_id);
         if (out_xrange)    out_xrange[0] =  tmp_pnt[0];
         if (out_yrange)    out_yrange[0] =  tmp_pnt[1];
         ds->convert_cs_to_cs (&max_pnt, &tmp_pnt, 1, DATA, "", DATA, ds_id); 
         if (out_xrange)    out_xrange[1] =  tmp_pnt[0];
         if (out_yrange)    out_yrange[1] =  tmp_pnt[1];
      }
   }
}

int ChipsAxisBind::serialize_all_nodes(ChipsStateStream& out)
{
   vector<ChipsAxis*> vec_all = graph->list_all_vertices();
   ChipsAxis* ax;
   vector<ChipsAxis*>::iterator vi;
   map<ChipsAxis*, int>used;
   map<string, int>plotids;
   string tmpid, parid, myid;
   int wrote = 0; 
 
   for (vi = vec_all.begin(); vi != vec_all.end(); vi++)
   {
       // if the axis is in specified dimension, is not in the list, and
       // is a child of the accessed plot (or no access_plot specified)- reset
       if (!used.count((*vi))) 
       {
          tmpid = (*vi)->get_id();
          if (tmpid=="bx1" || tmpid =="by1" || tmpid == "bx2" || tmpid == "by2")
             used[(*vi)] = 1;
          else
             used[(*vi)] = 0;
          tmpid = ((*vi)->get_parent())->get_id();
          if (!plotids.count(tmpid))
             plotids[tmpid] = 0; 
       }
   }

   double xmn, xmx, bmn, bmx;
   ChipsAxis *curx=NULL, *cury=NULL;
   int amn = 0, amx = 0;

   // iterate through the plots in sequence to minimize calling current_plot
   for (map<string, int>::iterator pi=plotids.begin(); pi != plotids.end();pi++)
   {
      for (map<ChipsAxis*,int>::iterator mi=used.begin(); mi!=used.end();mi++)
         if (mi->second == 0) 
         {
            ax = mi->first;
            myid = ax->get_id(); 
            parid = (ax->get_parent())->get_id();
            if (ax->get_orientation () == X_AXIS)
               curx = ax;
            else
               cury = ax; 
            if (parid == pi->first && !ax->get_wrote_limits())
            {
               // determine if we're bound to another non-border axis
               int bcnt = 0, auto_ok = 0; 
               list <ChipsAxis*> axes = get_connections(ax); 
               list <ChipsAxis*>::iterator ai;
               for (ai = axes.begin(); ai != axes.end(); ai++) 
               {
                  tmpid = (*ai)->get_id();
                  if (tmpid != "bx1" && tmpid != "bx2" && tmpid != "by1" &&
                      tmpid != "by2" && (*ai) != ax &&
                      (*ai)->get_orientation() == ax->get_orientation())
                     bcnt++; // found a non-border bind pair 
               } 

               // only deal with axis for plots with objects 
               if (((ChipsPlot*)(ax->get_parent()))->get_plot_object_count())
               {
                  if (((ChipsPlot*)(ax->get_parent()))->get_xaxis_cnt() == 1 &&
                      ((ChipsPlot*)(ax->get_parent()))->get_yaxis_cnt() == 1) 
                     auto_ok = 1;
                  if (plotids.size() > 1 && plotids[parid] == 0)
                  {
                     out << "current_plot(\""<<pi->first<<"\")"<<endl;
                     plotids[parid] = 1;
                  }
                  amn = ax->get_automin_used();
                  amx = ax->get_automax_used();
                  ax->get_range(xmn, xmx);

                  if (!auto_ok && curx && cury && 
                      curx->get_parent() == cury->get_parent()) 
                  {
                     // plot could have objects, but if we have multiple 
                     // axes, we need to ensure data system has curves/histo/
                     // contour/image in order to use auto
                      ChipsDataSystem* ds = 
                        ((ChipsPlot*)(curx->get_parent()))->get_datasystem(
                         curx, cury); 
                      if (ds && ds->get_num_objs()) 
                      {
                         auto_ok = 1; 
                      }
                  }
                  
                  string alid = ax->get_alter_id();
                  if (alid.length() == 0) alid = ax->get_id();
                  out << "current_axis(\""<<alid<<"\")"<<endl; 
                  if (ax->get_orientation() == X_AXIS)  
                     out << "limits(X_AXIS,";
                  else if (ax->get_orientation() == Y_AXIS) 
                     out << "limits(Y_AXIS,";
                  if (amn && auto_ok && !bcnt) out << "AUTO, ";
                     else  out << xmn << ", ";
                  if (amx && auto_ok && !bcnt) out << "AUTO)" << endl;
                     else  out << xmx << ")";
                  if (bcnt) 
                  {
                     if (amn && amx) 
                        out << "   # explicitly set ambiguous AUTO AUTO";
                     else if (amn)  
                        out << "   # explicitly set ambiguous AUTO min";  
                     else if (amx)   
                        out << "   # explicitly set ambiguous AUTO max";  
                  }
                  out << endl;
                  wrote = 1; 
               }

               // remove all bound axes from list 
               for (ai = axes.begin(); ai != axes.end(); ai++) 
               {
                  tmpid = (*ai)->get_id();
                  if (tmpid == "bx1" || tmpid == "bx2" || tmpid == "by1" ||
                      tmpid == "by2")
                  {
                     used[(*ai)] = 1;  // ignore border axes
                  }
                  else if ((*ai) != ax)
                  {
                     (*ai)->get_range(bmn, bmx);
                     if (bmn > xmn - FLT_EPSILON && bmn < xmn + FLT_EPSILON &&  
                         bmx > xmx - FLT_EPSILON && bmx < xmx + FLT_EPSILON)
                     {
                        // should be handled by auto case
                        used[(*ai)] = 1;
                     } 
                     else if (((*ai)->get_parent()) == (ax->get_parent())) 
                     {
                        // handle bound axes in same plot- if in different 
                        // plot let it wait til we iterate over that plot
                        used[(*ai)] = 1;
                        out << "current_axis(\""<<ax->get_alter_id()<<"\")"<<endl;
                        if ((*ai)->get_orientation() == X_AXIS)
                           out << "limits(X_AXIS,";
                        else if ((*ai)->get_orientation() == Y_AXIS)
                           out << "limits(Y_AXIS,";
                        out << bmn <<", " << bmx << ")" << endl;         
                     }
                  } 
               }
            }
         }
   }

   // reset limits flags for all axes
   for (vi = vec_all.begin(); vi != vec_all.end(); vi++)
      (*vi)->set_wrote_limits(0);  

   return wrote; 
}


int ChipsAxisBind::find_axis_min_max(ChipsAxis* axis, int dim,
               double* min, double* max, double* lmin, bool apply_transform)
{
   // Go through each of bound the axes and find the plot's min/max values

   // collect all the different plots into a single list
   list <ChipsAxis *> axis_list = this->get_axes (axis);
   list <ChipsAxis *>::iterator axis_iter;
   ChipsPlot *plot;
   double tmin, tmax;
   int rval = ERR_STATUS_GOOD;
   bool init=true;
   for(axis_iter=axis_list.begin (); !rval && axis_iter != axis_list.end (); axis_iter++)
   {
      plot =  dynamic_cast <ChipsPlot *>((*axis_iter)->get_parent ());
      rval = plot->find_axis_min_max ((*axis_iter), dim, min, max, lmin, &tmin, &tmax, init);
      if (rval == ERR_STATUS_GOOD)
         init = false;
   }

   if (apply_transform)
   {
      *min = tmin;
      *max = tmax;
   }

   return rval;
}
         
  

