// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsAxisGraph.hh"
#include "ChipsAxis.hh"
#include "ChipsFrame.hh"
#include "ChipsPlot.hh"

#include <queue>

ChipsAxisEdge::ChipsAxisEdge(ChipsAxisNode* from, ChipsAxisNode* to, int typ,
   double val1, double val2)
{
   source = from;
   dest = to; 
   type = typ; 
   v1 = val1;
   v2 = val2; 
}

ChipsAxisEdge::~ChipsAxisEdge()
{
}

ChipsAxisNode::ChipsAxisNode(ChipsAxis *ax)
{
   axis = ax; 
   if (axis)
   {
      id = axis->get_id(); 
      id += "_";
      id += axis->get_parent ()->get_id ();
   }
   status = 0;
   edges.clear(); 
}

void ChipsAxisNode::add_edge(ChipsAxisNode *node, int type, double val1, 
   double val2)
{
   // don't add an edge if it already exists
   if (!check_edge(node))
   { 
      ChipsAxisEdge connect(this, node, type, val1, val2);
      edges.push_back(connect);
   }
}

int ChipsAxisNode::check_edge(ChipsAxisNode *node)
{
  vector<ChipsAxisEdge>::iterator vi;
  int found = 0;
   for(vi=edges.begin(); vi != edges.end() && !found; vi++)
      if ((*vi).get_source() == this && (*vi).get_dest() == node) 
	 found++;
   return found; 
}

int ChipsAxisNode::get_edge_values(ChipsAxisNode *node, double &v1, 
   double &v2)
{
   vector<ChipsAxisEdge>::iterator vi;
   int look = 1;
   for(vi=edges.begin(); vi != edges.end() && look; vi++)
      if ((*vi).get_source() == this && (*vi).get_dest() == node)
      {
         v1 = (*vi).get_value1(); 
         v2 = (*vi).get_value2(); 
         look = 0; 
      }
   return look; 
}

void ChipsAxisNode::get_max_edge_value(double &v1)
{
   vector<ChipsAxisEdge>::iterator vi;
   double tmp, max = 0; 
   for(vi=edges.begin(); vi != edges.end(); vi++)
      if ((tmp = (*vi).get_value1()) > max) max = tmp;
   v1 = max;
}

void ChipsAxisNode::remove_edge(ChipsAxisNode *node)
{
   vector<ChipsAxisEdge>::iterator vi, fi;
   int found = 0;  
   for(vi=edges.begin(); vi != edges.end() && !found; vi++)
      if ((*vi).get_source() == this && (*vi).get_dest() == node) 
      {
	 fi = vi;
	 found = 1;
      }
   if (found) edges.erase(fi);
}

void ChipsAxisNode::list_edges()
{
   cout <<" node : " << id << endl;
   for(unsigned int ee=0 ; ee < edges.size() ; ee++)
   {
      ChipsAxisEdge edge = edges[ee]; 
      cout << "  " << id << "("<<status<<") -> " << edge.get_dest()->get_id() << endl; 
   }
}

int ChipsAxisNode::traverse_edges(int val)
{
   int found = 0;
   status = val;
   ChipsAxisNode* tmpnode;
   for(unsigned int ee=0 ; ee < edges.size() ; ee++)
   {
      ChipsAxisEdge edge = edges[ee];
      tmpnode = edge.get_dest();
      if (tmpnode->get_status() == 0)
	 found |= tmpnode->traverse_edges(val+1);
      else if (tmpnode->get_status() != status-1)
	 found |= 1;
   }
   return found;
}

void ChipsAxisNode::list_associated_axes_dfs(vector<ChipsAxis*> &vec)
{
   status = 1;
   ChipsAxisNode* tmpnode;
   vec.push_back(axis); 
   for(unsigned int ee=0 ; ee < edges.size() ; ee++)
   {
      ChipsAxisEdge edge = edges[ee];
      tmpnode = edge.get_dest();
      if (tmpnode->get_status() == 0)
         tmpnode->list_associated_axes_dfs(vec);
   }
}

void ChipsAxisNode::serialize (ChipsStateStream& out)
{
   ChipsAxisNode *edge_node;
   string tmp_id;
   int count = this->edges.size ();
   int type;
   int status;
   double value;

   out.write_string(&id);
   tmp_id = this->axis->get_parent()->get_id();
   out.write_string(&tmp_id);
   tmp_id = this->axis->get_id();
   out.write_string(&tmp_id);

   out.write_int (&count);
   this->status = 1;

   for(unsigned int ee=0 ; ee < edges.size() ; ee++)
   {
      ChipsAxisEdge edge = edges[ee];
      if (edge.get_source () == this)
         edge_node = edge.get_dest ();
      else
         edge_node = edge.get_source ();

      // write out the status so when deserializing the 
      // node knows whether to continue processing or not
      status = edge_node->get_status ();
      out.write_int (&status);

      // if the edge node hasn't been serialized serialize it
      if (edge_node->get_status () == 0)
         edge_node->serialize (out);

      // the edge node status is 2 when all it's edges have 
      // been written
      status = edge_node->get_status ();
      out.write_int (&status);
      if (edge_node->get_status () == 2)
      {
         type = edge.get_type ();
         out.write_int (&type);
         value = edge.get_value1 ();
         out.write_double (&value);
         value = edge.get_value2 ();
         out.write_double (&value);
      }
   }
 
   this->status = 2;
}

void ChipsAxisNode::deserialize (ChipsFrame *cf,
                                ChipsStateStream& in,
                                vector<ChipsAxisNode*> &out_nodes)
{

   ChipsPlot *plot;
   ChipsAxis *axis;
   ChipsAxisNode *node = NULL;
   string tmp_id;
   int count = 0;
   int type;
   int status;
   double value1, value2;


   in.read_string (&id);

   in.read_string(&tmp_id);
   plot = cf->get_plot (tmp_id);
   in.read_string(&tmp_id);
   if (plot) 
   {
      if ((this->axis = plot->get_xaxis (tmp_id)) == NULL)
         this->axis = plot->get_yaxis (tmp_id);
   } 
   in.read_int (&count);


   for(int ee=0 ; ee < count ; ee++)
   {
      // read in the status of the next node
      // if it was not zero when writing out then it was 
      // already deserialized
      in.read_int (&status);

      if (status == 0)
      {
         node = new ChipsAxisNode (NULL);
         out_nodes.push_back (node);

         node->deserialize (cf, in, out_nodes);
      }
      
      // reread in the status of the next node (as it could have changed)
      // if it's less than 2 then read in the edge information
      in.read_int (&status);
      if (status == 2)
      {
         in.read_int (&type);
         in.read_double (&value1);
         in.read_double (&value2);
         this->add_edge (node, type, value1, value2);
         node->add_edge (this, type, value2, value1);
      }
   }
}


ChipsAxisGraph::ChipsAxisGraph()
{
   nodes.clear();
}

ChipsAxisGraph::~ChipsAxisGraph()
{
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)
      delete nodes[nn];
   nodes.clear();
}

void ChipsAxisGraph::print()
{
   for(unsigned int i=0 ; i < nodes.size() ; i++)
      nodes[i]->list_edges();
}

vector<ChipsAxis*> ChipsAxisGraph::list_attached_vertices_dfs(ChipsAxis* entry)
{
   ChipsAxisNode* tmp = NULL; 
   vector<ChipsAxis*> vec; 
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)
      if (nodes[nn]->get_axis() == entry)  tmp = nodes[nn];
   vec.clear(); 
   if (tmp)
   {
      tmp->list_associated_axes_dfs(vec);
      for(unsigned int nn=0 ; nn < nodes.size(); nn++)  // reset node status
         nodes[nn]->set_status(0);
   }
   return vec; 
}

vector<ChipsAxis*> ChipsAxisGraph::list_attached_vertices_bfs(ChipsAxis* entry)
{

   ChipsAxisNode* tmpnode = NULL;
   queue<ChipsAxisNode *> Q;

   vector<ChipsAxis*> vec;
   vector<ChipsAxisEdge> edges;
   
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)
   {
      nodes[nn]->set_status(0);
      if (nodes[nn]->get_axis() == entry)  tmpnode = nodes[nn];
   }

   vec.clear();

   if (tmpnode)
   {
      Q.push (tmpnode);
      vec.push_back(tmpnode->get_axis ());
      tmpnode->set_status(1);

      while (!Q.empty ())
      {
         tmpnode = Q.front ();
         edges = tmpnode->get_edges ();
         for(unsigned int ee=0 ; ee < edges.size() ; ee++)
         {
            tmpnode = edges[ee].get_dest ();
            if (tmpnode->get_status() == 0)
            {
               vec.push_back (tmpnode->get_axis ());
               tmpnode->set_status (1);
               Q.push (tmpnode);
            }
         }

         Q.pop ();
      }
   }

   for(unsigned int nn=0 ; nn < nodes.size(); nn++)  // reset node status
      nodes[nn]->set_status(0);

   return vec;
}


vector<ChipsAxis*> ChipsAxisGraph::list_all_vertices()
{
   vector<ChipsAxis*> vec;
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)
      vec.push_back(nodes[nn]->get_axis());
   return vec; 
}

void ChipsAxisGraph::add_node(ChipsAxis* ax)
{
   if (!ax) return; 
  ChipsAxisNode* axnode = new ChipsAxisNode(ax); 
  nodes.push_back(axnode);
}

ChipsAxisNode* ChipsAxisGraph::find_node(ChipsAxis* axis)
{
   ChipsAxisNode* node = NULL;
   int sz = nodes.size(); 

   for(int nn=0 ; nn < sz && !node; nn++)
      if (nodes[nn]->get_axis() == axis)   node = nodes[nn];
   return node; 
} 

void ChipsAxisGraph::remove_node(ChipsAxis* axis)
{
   vector<ChipsAxisNode*>::iterator vi, fi;
   int found = 0;
   for(vi=nodes.begin(); vi != nodes.end() && !found; vi++)
      if ((*vi)->get_axis() == axis) 
      {
         fi = vi;
         found = 1;
      }
   if (found) 
   {
      for(vi=nodes.begin(); vi != nodes.end(); vi++)
         (*vi)->remove_edge((*fi));  // remove edges containing axis 
      nodes.erase(fi);               // remove axis from list of nodes
   }
} 

bool ChipsAxisGraph::check_for_cycle()
{
   bool found = 0; 
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)  // reset node status
      nodes[nn]->set_status(0);
   for(unsigned int nn=0 ; nn < nodes.size() && !found ; nn++)
      if (nodes[nn]->get_status() == 0)   
         found |= nodes[nn]->traverse_edges(1); 
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)  // reset node status
      nodes[nn]->set_status(0);
   return found; 
}

int ChipsAxisGraph::add_edge(ChipsAxis *ax1, ChipsAxis* ax2, int type, double val1, double val2)
{
   ChipsAxisNode* n1 = find_node(ax1);
   ChipsAxisNode* n2 = find_node(ax2);

   if (!(n1 && n2)) return ERR_STATUS_BAD;
 
   n1->add_edge(n2, type, val1, val2);
   n2->add_edge(n1, type, val2, val1); 
   return ERR_STATUS_GOOD; 
}

void ChipsAxisGraph::remove_edge(ChipsAxis *ax1, ChipsAxis* ax2)
{
   ChipsAxisNode* n1 = find_node(ax1);
   ChipsAxisNode* n2 = find_node(ax2);

   if (!(n1 && n2)) return;

   n1->remove_edge(n2);
   n2->remove_edge(n1);
}

int ChipsAxisGraph::get_edge_values(ChipsAxis* ax1, ChipsAxis* ax2,
   double &v1, double&v2)
{
   if (!(ax1 && ax2))  return ERR_STATUS_BAD;  // null axes passed in

   ChipsAxisNode* n1 = find_node(ax1);
   ChipsAxisNode* n2 = find_node(ax2);

   if (!(n1 && n2))  return ERR_STATUS_BAD;   // nodes do not exist

   return n1->get_edge_values(n2, v1, v2);
}

int ChipsAxisGraph::get_edge_max(ChipsAxis* ax1, double &val)
{
   if (!ax1)  return ERR_STATUS_BAD;  // null axes passed in
   ChipsAxisNode* n1 = find_node(ax1);
   if (!n1 )  return ERR_STATUS_BAD;   // nodes do not exist
   n1->get_max_edge_value(val);
   return ERR_STATUS_GOOD;
}
 
void ChipsAxisGraph::serialize (ChipsStateStream& out)
{
   int count;
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)  // reset node status
      nodes[nn]->set_status(0);

   // go through each node and write the axis and the edge axis
   count = nodes.size ();
   out.write_int (&count);
   for(unsigned int nn=0 ; nn < nodes.size(); nn++)
   {
      
      if (nodes[nn]->get_status () == 0)
         nodes[nn]->serialize (out);
   }

}

void ChipsAxisGraph::deserialize (ChipsFrame *frame, ChipsStateStream& in)
{
   int count;

   in.read_int (&count);
   nodes.clear ();

   while (nodes.size () < count)
   {
      ChipsAxisNode *node = new ChipsAxisNode (NULL);
      this->nodes.push_back(node);
      node->deserialize (frame, in, this->nodes);
   }   


}
