// 
//  Copyright (C) 2009-2010  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsCartesianAxisAlgorithm.hh"
#include "ChipsDefines.hh"
#include "ChipsDataSystem.hh"
#include "Transform.hh"
#include "ChipsAxis.hh"
#include "ChipsDefinesP.hh"
#include "ChipsMatrix.hh"


#include <math.h>
#include <float.h>

ChipsCartesianAxisAlgorithm::ChipsCartesianAxisAlgorithm (int orientation, ChipsAxis *axis) : ChipsAxisAlgorithm (orientation,axis)
{
   this->padded_min = -1;
   this->padded_max = 1;
   this->lin_min = 0;
   this->lin_max = 1;
   this->log_min = .1;
   this->log_max = 1;

   this->log_minimum = .01;

   this->last_pad = 0;
   this->last_automin = false;
   this->last_automax = false;

   this->transform_applied = false;
   this->untransformed_tickpos = NULL;
}

ChipsCartesianAxisAlgorithm::ChipsCartesianAxisAlgorithm (ChipsCartesianAxisAlgorithm &copy) : ChipsAxisAlgorithm (copy)
{
   this->padded_min = copy.padded_min;
   this->padded_max = copy.padded_max;
   this->lin_min = copy.lin_min;
   this->lin_max = copy.lin_max;
   this->log_min = copy.log_min;
   this->log_max = copy.log_max;

   this->log_minimum = copy.log_minimum;

   this->arbitrary_set = false;  // force setting
   this->last_pad = copy.last_pad;
   this->last_automin = copy.last_automin;
   this->last_automax = copy.last_automax;
 
   this->transform_applied = copy.transform_applied;
   if (this->transform_applied && copy.untransformed_tickpos)
   {
      this->untransformed_tickpos = new double[this->tickcnt];
      for (int ii=this->tickcnt; ii--;)
         this->untransformed_tickpos[ii] = copy.untransformed_tickpos[ii];
   }


}

ChipsCartesianAxisAlgorithm::~ChipsCartesianAxisAlgorithm ()
{
   if (this->untransformed_tickpos && this->transform_applied)
   {
      delete []this->untransformed_tickpos;
      this->untransformed_tickpos = NULL;
   }
}

ChipsCartesianAxisAlgorithm * ChipsCartesianAxisAlgorithm::clone ()
{
   return new ChipsCartesianAxisAlgorithm (*this);
}

void ChipsCartesianAxisAlgorithm::set_scalable(bool val)
{
   if (this->scalable != val)
      this->modified=true;
   this->scalable = val;
}

void ChipsCartesianAxisAlgorithm::set_scaletype (int val)
{
   if (this->scaletype != val)
      this->modified=true;
   this->scaletype = val;
}

void ChipsCartesianAxisAlgorithm::set_tick_positions (double *ticks, int count)
{
   this->modified = true;

   if (this->tickpos)
      delete []this->tickpos;

   this->tickpos = NULL;
   this->tickcnt = 0;
   this->arbitrary_set = false;

   if (!ticks)
      return;

   this->tickpos = new double[count];
   this->tickcnt = count;
   for (int ii=count; ii--;)
      this->tickpos[ii] = ticks[ii];

   if (this->use_transform ())
   {
      if (this->untransformed_tickpos && this->transform_applied)
         delete []this->untransformed_tickpos;

      this->untransformed_tickpos = new double[this->tickcnt];
      for (int ii=this->tickcnt; ii--;)
         this->untransformed_tickpos[ii] = this->tickpos[ii];
      this->remove_transform (this->untransformed_tickpos, this->tickcnt);
      this->transform_applied = true;
   }
   else
   {

      if (this->untransformed_tickpos && this->transform_applied)
         delete []this->untransformed_tickpos;

      this->untransformed_tickpos = this->tickpos;
      this->transform_applied = false;
   }
}

void ChipsCartesianAxisAlgorithm::set_arbitrary_positions (double *positions, string *text, int count)
{
   this->modified = true;

   if (this->arbpos)
   {
      delete []this->arbpos;
      this->arbpos = NULL;
   }

   if (this->arbtext)
   {
      delete []this->arbtext;
      this->arbtext = NULL;
   }

   this->arbcnt = 0;

   if (!positions)
      return;

   if (text)
   {
      this->arbtext = new string[count];
      for (int ii = count; ii--;)
         this->arbtext[ii] = text[ii];
   }
   this->arbcnt = count;
   this->arbpos = new double[count];
   for (int ii = count; ii--;)
      this->arbpos[ii] = positions[ii];
   this->arbitrary_set = false;
}

void ChipsCartesianAxisAlgorithm::set_log_minimum (double val)
{
   if (val > 0)
   {
      if ((scaletype == V_SCALE_LOG) && (this->log_minimum != val))
         this->modified = true;

      this->log_minimum = val;
   }
}

int ChipsCartesianAxisAlgorithm::set_absolute_min (double val, int scaletype, bool transformed)
{
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   // transform the data coordinates if a transform exists
   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {

         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector min_coord = dsCoord->get_lower_bounds ();

         min_coord[index] = val;
         dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, tmp_id, DATA, "");
         val = min_coord[index];
      }
   }

   // if the scale type is -1 set everything; otherwise only set the specified
   // scale type.
   if ((scaletype == V_SCALE_LINEAR) ||(scaletype == -1))
   {
      if (this->lin_min != val)
         this->modified = true;
      this->lin_min = val;
   }
   if ((scaletype == V_SCALE_LOG) || (scaletype == -1))
   {
      if (val <=0)
      {
         if (this->log_min != log_minimum)
            this->modified = true;
         this->log_min = this->log_minimum;
      }
      else
      {
         if (this->log_min != val)
            this->modified = true;
         this->log_min = val;
      }
   }

   return ERR_STATUS_GOOD;
}


double ChipsCartesianAxisAlgorithm::get_min (bool transformed)
{
   double val = this->padded_min;
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {

         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector min_coord = dsCoord->get_lower_bounds ();

         min_coord[index] = this->padded_min;
         dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, "", DATA, tmp_id);
         val = min_coord[index];
      }
   }

   return val;
}

double ChipsCartesianAxisAlgorithm::get_absolute_min (int scaletype, bool transformed)
{
   double val;
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   if (scaletype == -1)
      scaletype = this->scaletype;

   if (scaletype == V_SCALE_LOG) 
      val = this->log_min;
   else
      val =this->lin_min;

   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {
         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector min_coord = dsCoord->get_lower_bounds ();

         min_coord[index] = val;
         dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, "", DATA, tmp_id);
         val = min_coord[index];
      }
   }

   return val;
}

int ChipsCartesianAxisAlgorithm::set_absolute_max (double val, int scaletype, bool transformed)
{ 
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {
         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector max_coord = dsCoord->get_upper_bounds ();

         max_coord[index] = val;
         dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, tmp_id, DATA, "");
         val = max_coord[index];
      }
   }


   // if the scale type is -1 set everything; otherwise only set the specified
   // scale type.
   if ((scaletype == V_SCALE_LINEAR) ||(scaletype == -1))
   {
      if (this->lin_max != val)
         this->modified = true;
      this->lin_max = val;
   }
   if ((scaletype == V_SCALE_LOG) || (scaletype == -1))
   {
      if (val <=0)
      {
         if (this->log_max != log_minimum)
            this->modified = true;
         this->log_max = this->log_minimum;
      }
      else
      {
         if (this->log_max != val)
            this->modified = true;
         this->log_max = val;
      }
   }

   return ERR_STATUS_GOOD;
}

double ChipsCartesianAxisAlgorithm::get_max (bool transformed)
{
   double val = this->padded_max;
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {
         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector max_coord = dsCoord->get_upper_bounds ();

         max_coord[index] = val;
         dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, "", DATA, tmp_id);
         val = max_coord[index];
      }
   }

   return val;
}

double ChipsCartesianAxisAlgorithm::get_absolute_max (int scaletype, bool transformed)
{
   double val;
   int index = (this->orientation == X_AXIS) ? 0 : 1;

   if (scaletype == -1)
      scaletype = this->scaletype;

   if (scaletype == V_SCALE_LOG)
      val = this->log_max;
   else
      val = this->lin_max;

   if (!transformed)
   {
      ChipsDataSystem *dsCoord = this->axis->get_data_system ();
      if ((dsCoord != NULL) && (dsCoord->get_default_transform_id () != ""))
      {

         string tmp_id = dsCoord->get_id_no_wcs ();
         ChipsVector max_coord = dsCoord->get_upper_bounds ();

         max_coord[index] = val;
         dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, "", DATA, tmp_id);
         val = max_coord[index];
      }
   }

   return val;
}


int ChipsCartesianAxisAlgorithm::set_absolute_minmax (double min, double max, int scaletype, bool transformed)
{
   this->set_absolute_min (min, scaletype, transformed);
   this->set_absolute_max (max, scaletype, transformed);

   return ERR_STATUS_GOOD;
}


int ChipsCartesianAxisAlgorithm::update_tick_positions(bool &out_updated)
                                           
{
   int rval = ERR_STATUS_GOOD;

   out_updated = false;

   if (!this->modified)
      return rval;

   this->reset_adjusted_parameters ();

   if (this->scaletype == V_SCALE_LOG)
   {
      switch (this->major_mode)
      {
         case chips_nice:
            rval = this->log_nice_tick_positions ();
         break;
         case chips_limits:
            rval = this->log_limit_tick_positions ();
         break;
         case chips_interval:
            rval = this->log_interval_tick_positions ();
         break;
         case chips_count:
            rval = this->log_count_tick_positions ();
         break;
         case chips_arbitrary:
            rval = this->log_arbitrary_tick_positions ();
         break;
         default:
            rval = ERR_STATUS_BAD;
         break;
      }
   }
   else
   {
      switch (this->major_mode)
      {
         case chips_nice:
            rval = this->linear_nice_tick_positions ();
         break;
         case chips_limits:
            rval = this->linear_limit_tick_positions ();
         break;
         case chips_interval:
            rval = this->linear_interval_tick_positions ();
         break;
         case chips_count:
            rval = this->linear_count_tick_positions ();
         break;
         case chips_arbitrary:
            rval = this->linear_arbitrary_tick_positions ();
         break;
         default:
            rval = ERR_STATUS_BAD;
         break;
      }
   }

   // the algorithms were rerun so assume they were updated
   out_updated = true;

   // all parameters were used to calculated; reset modified flag
   this->modified = false;

   return rval;
}


int ChipsCartesianAxisAlgorithm::calculate_tick_limits(bool automin, bool automax, double pad)
{
   int rval = ERR_STATUS_GOOD;

   if (!this->modified && (this->last_pad == pad) &&
        (this->last_automax == automax) &&
        (this->last_automin == automin))
      return rval;

   this->last_pad = pad;
   this->last_automax = automax;
   this->last_automin = automin;

   this->reset_adjusted_parameters ();

   if (this->scaletype == V_SCALE_LOG)
   {
      switch (this->major_mode)
      {
         case chips_nice:
            rval = this->log_nice_tick_limits ();
         break;
         case chips_limits:
            rval = this->log_limit_tick_limits (automin, automax, pad);
         break;
         case chips_interval:
            rval = this->log_interval_tick_limits (automin, automax, pad);
         break;
         case chips_count:
            rval = this->log_count_tick_limits ();
         break;
         case chips_arbitrary:
            rval = this->log_arbitrary_tick_limits (automin, automax, pad);
         break;
         default:
            rval = ERR_STATUS_BAD;
         break;
      }
   }
   else
   {
      switch (this->major_mode)
      {
         case chips_nice:
            rval = this->linear_nice_tick_limits ();
         break;
         case chips_limits:
            rval = this->linear_limit_tick_limits (automin, automax, pad);
         break;
         case chips_interval:
            rval = this->linear_interval_tick_limits (automin, automax, pad);
         break;
         case chips_count:
            rval = this->linear_count_tick_limits (automin, automax, pad);
         break;
         case chips_arbitrary:
            rval = this->linear_arbitrary_tick_limits (automin, automax, pad);
         break;
         default:
            rval = ERR_STATUS_BAD;
         break;
      }
   }

   // make sure modified is true for update positions
   this->modified = true;

   return rval;
}

int ChipsCartesianAxisAlgorithm::linear_nice_tick_limits ()
{
   double xpmin, xpmax, dist, base;

   scale1(this->lin_min, this->lin_max, 5, &xpmin, &xpmax, &dist, &base);

   // set number of minor ticks- if the unit base is 2, setup
   // minor tick intervals of 1/4, otherwise use 1/5
   if (this->minor_mode == chips_nice)
   {
      if (base == 2.0)     this->minor_num_adjusted = 3;
         else              this->minor_num_adjusted = 4;
   }

   this->padded_min = xpmin;
   this->padded_max = xpmax;

   this->modified = true;

   return ERR_STATUS_GOOD;
}

int ChipsCartesianAxisAlgorithm::linear_nice_tick_positions()
{
   double xpmin, xpmax, dist, base;
   int rval = ERR_STATUS_GOOD;


   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   scale1(this->lin_min, this->lin_max, 5, &xpmin, &xpmax, &dist, &base);

   // set number of minor ticks- if the unit base is 2, setup
   // minor tick intervals of 1/4, otherwise use 1/5
   if (this->minor_mode == chips_nice)
   {
      if (base == 2.0)     this->minor_num_adjusted = 3;
         else              this->minor_num_adjusted = 4;
   }
   if (xpmax > xpmin)   this->tickcnt = (unsigned int)((xpmax - xpmin) / dist + 1.5);
      else              this->tickcnt = 1;

   // make sure we have a reasonable number of ticks
   if ((this->tickcnt < 0) || (this->tickcnt > 500))
   {
      // setup new tick locations
      this->tickcnt = 2;
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         this->tickpos[0] = xpmin;
         this->tickpos[1] = xpmax;
         if (this->minor_mode == chips_nice)
            this->minor_mode_adjusted = chips_count;
      }
   }
   else
   {
      // setup new tick locations
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         if ((dist > DBL_EPSILON) && (xpmin < DBL_EPSILON) &&
             (xpmin > -1*DBL_EPSILON))
            this->tickpos[0] = 0;
         else
            this->tickpos[0] = xpmin;
         for (rr = 1; rr < tickcnt; rr++)
         {
            this->tickpos[rr] = this->tickpos[rr-1] + dist;
            if ((dist > DBL_EPSILON) && (this->tickpos[rr] < DBL_EPSILON) &&
               (this->tickpos[rr] > -1*DBL_EPSILON))
               this->tickpos[rr] = 0;
         }
         if (this->minor_mode == chips_nice)
            this->minor_mode_adjusted = chips_count;
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {

         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }

   }

   return rval;
}


int ChipsCartesianAxisAlgorithm::linear_count_tick_limits (bool automin, bool automax, double pad)
{
   double xpmin, xpmax, dist;

   if ((this->major_num > 1) && (this->major_num <  V_SIZE_MAX_VAL))
   {
      scale2(this->lin_min, this->lin_max, this->major_num-1, &xpmin, &xpmax, &dist);
      this->padded_min = xpmin;
      this->padded_max = xpmin +  (this->major_num-1) * dist;
   }
   else
      return this->linear_limit_tick_limits(automin, automax, pad);

   return ERR_STATUS_GOOD;
}

int ChipsCartesianAxisAlgorithm::linear_count_tick_positions ()
{
   double xpmin, xpmax, dist;
   int rval = ERR_STATUS_GOOD;

   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   scale2(this->lin_min, this->lin_max, this->major_num-1, &xpmin, &xpmax, &dist);

   // if the count is at least 2 and less than a defined number of
   // ticks (nominally 100) then use count mode
   if ((this->major_num > 1) && (this->major_num <  V_SIZE_MAX_VAL))
   {
      this->tickcnt = this->major_num;
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         if ((dist > DBL_EPSILON) && (xpmin < DBL_EPSILON) &&
             (xpmin > -1*DBL_EPSILON))
            this->tickpos[0] = 0;
         else
            this->tickpos[0] = xpmin;
         for (rr = 1; rr < this->tickcnt; rr++)
         {
            this->tickpos[rr] = this->tickpos[rr-1] + dist;
            if ((dist > DBL_EPSILON) && (this->tickpos[rr] < DBL_EPSILON) &&
                (this->tickpos[rr] > -1*DBL_EPSILON))
               this->tickpos[rr] = 0;
         }
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {

         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }
   }
   if (!this->tickpos)
      // too high a count or allocation failed- try limit mode
      return  this->linear_limit_tick_positions();
 
   return rval;
}


int ChipsCartesianAxisAlgorithm::linear_limit_tick_limits(bool automin, bool automax, double pad)
{
   double padmin, padmax, xpmin, xpmax, dist, base;

   // adjust by pad factor
   if (automin)
   {
      double tmp = (this->lin_max - this->lin_min)*pad;
      padmin = this->lin_min - tmp;
   }
   else
      padmin = this->lin_min;
   if (automax)
   {
      double tmp = (this->lin_max - this->lin_min)*pad;
      padmax = this->lin_max + tmp;
   }
   else
      padmax = this->lin_max;

   scale1(padmin, padmax, 5, &xpmin, &xpmax, &dist, &base);

   this->padded_min = padmin;
   this->padded_max = padmax;

   return ERR_STATUS_GOOD;
}




int ChipsCartesianAxisAlgorithm::linear_limit_tick_positions()
{
   double xpmin, xpmax, dist, base;
   int rval = ERR_STATUS_GOOD;

   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   scale1(this->padded_min, this->padded_max, 5, &xpmin, &xpmax, &dist, &base);

   // set number of minor ticks- if the unit base is 2, setup
   // minor tick intervals of 1/4, otherwise use 1/5
   // only calculate the minor mode when the minor tick mode is nice
   if (this->minor_mode == chips_nice)
   {
      if (base == 2.0)     this->minor_num_adjusted = 3;
         else              this->minor_num_adjusted = 4;
   }
   if (xpmax > xpmin)   this->tickcnt = (unsigned int)((xpmax - xpmin) / dist + 1.5);
      else              this->tickcnt = 1;

   // make sure we have a reasonable number of ticks
   if ((this->tickcnt <= 0) || (this->tickcnt > 500))
   {
      this->tickcnt = 2;
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         this->tickpos[0] = xpmin;
         this->tickpos[1] = xpmax;

         if (this->minor_mode == chips_nice)
            this->minor_mode_adjusted = chips_count;
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {

         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }

   }
   else
   {
      // setup new tick locations
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         if ((dist > DBL_EPSILON) && (xpmin < DBL_EPSILON) &&
             (xpmin > -1*DBL_EPSILON))
            this->tickpos[0] = 0;
         else
            this->tickpos[0] = xpmin;
         for (rr = 1; rr < tickcnt; rr++)
         {
            this->tickpos[rr] = this->tickpos[rr-1] + dist;
            if ((dist > DBL_EPSILON) && (this->tickpos[rr] < DBL_EPSILON) &&
                (this->tickpos[rr] > -1*DBL_EPSILON))
               this->tickpos[rr] = 0;
         }
         if (this->minor_mode == chips_nice)
            this->minor_mode_adjusted = chips_count;
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {

         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }

   }
   return rval;
}


int ChipsCartesianAxisAlgorithm::linear_arbitrary_tick_limits(bool automin, bool automax, double pad)
{
   if (automin)
   {
      double tmp = (this->lin_max - this->lin_min)*pad;
      this->padded_min = this->lin_min - tmp;
   }
   else
      this->padded_min = this->lin_min;
   if (automax)
   {
      double tmp = (this->lin_max - this->lin_min)*pad;
      this->padded_max = this->lin_max + tmp;
   }
   else
      this->padded_max = this->lin_max;

   return ERR_STATUS_GOOD;
}



int ChipsCartesianAxisAlgorithm::linear_arbitrary_tick_positions()
{
   int rval = ERR_STATUS_GOOD;

   if (!this->arbitrary_set)
   {
      this->set_tick_positions (this->arbpos, this->arbcnt);
      this->arbitrary_set = true;
   }

   return rval;
}

int ChipsCartesianAxisAlgorithm::linear_interval_tick_limits(bool automin, bool automax, double pad)
{
   double numticks = fabs(this->lin_max - this->lin_min)/this->major_int;

   // if out interval is 0 or is too small for the data range (more than
   // a defined number of ticks (nominally 100) then use limits mode
   if ((this->major_int > 0) && (numticks < V_SIZE_MAX_VAL))
   {
      // adjust limits to include pad factor
      if (automin)
      {
         double tmp = (this->lin_max - this->lin_min)*pad;
         this->padded_min = this->lin_min - tmp;
      }
      else
         this->padded_min = this->lin_min;
      if (automax)
      {
         double tmp = (this->lin_max - this->lin_min)*pad;
         this->padded_max = this->lin_max + tmp;
      }
      else
         this->padded_max = this->lin_max;
   }
   else
      // bad interval or allocation failed- try limit mode
      return this->linear_limit_tick_limits(automin, automax, pad);

   return ERR_STATUS_GOOD;
}

int ChipsCartesianAxisAlgorithm::linear_interval_tick_positions()
{
   int rval = ERR_STATUS_GOOD;

   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   // adjust limits to include pad factor
   double numticks = fabs(this->lin_max - this->lin_min)/this->major_int;

   // if out interval is 0 or is too small for the data range (more than
   // a defined number of ticks (nominally 100) then use limits mode
   if ((this->major_int > 0) && (numticks < V_SIZE_MAX_VAL))
   {
      // pad tickcnt to extend past range to get minor ticks
      this->tickcnt = (int)(numticks + 3);
      this->tickpos = new double[tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         if (this->padded_min >= 0)
            this->tickpos[0] = ((int)(this->lin_min/this->major_int)) * this->major_int;
         else
            // subtract 1 so that the floored value is less than the
            // min value
            this->tickpos[0] = ((int)(this->lin_min/this->major_int) - 1) * this->major_int;

         if ((this->major_int > DBL_EPSILON) && (this->tickpos[0] < DBL_EPSILON) &&
             (this->tickpos[0] > -1*DBL_EPSILON))
            this->tickpos[0] = 0;
         for (rr = 1; rr < this->tickcnt; rr++)
         {
            this->tickpos[rr] = this->tickpos[rr-1] + this->major_int;
               if ((this->major_int > DBL_EPSILON) && (this->tickpos[rr] < DBL_EPSILON) &&
                   (this->tickpos[rr] > -1*DBL_EPSILON))
                  this->tickpos[rr] = 0;
         }
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }

   }
   if (!this->tickpos)
      // bad interval or allocation failed- try limit mode
      return this->linear_limit_tick_positions();

   return rval;
}


int ChipsCartesianAxisAlgorithm::log_nice_tick_limits()
{
   double limloga, limlogb, dist, min_init = -5, max_init = 1;

   // compute log data range
   min_init = log10(this->log_min);
   max_init = log10(this->log_max);

   // use linear scale1 algorithm to compute log axis limits
   scale1(min_init, max_init, 5, &limloga, &limlogb, &dist);


   // compute and store axis limits
   this->padded_min = pow(10.0, (double)limloga);
   this->padded_max = pow(10.0, (double)limlogb);

   return ERR_STATUS_GOOD;
}


int ChipsCartesianAxisAlgorithm::log_nice_tick_positions()
{
   double limloga, limlogb, dist, min_init = -5, max_init = 1;
   int rval = ERR_STATUS_GOOD;

   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   // compute log data range
   min_init = log10(this->log_min);
   max_init = log10(this->log_max);

   this->fix_minmax (min_init, max_init, false);

   // use linear scale1 algorithm to compute log axis limits
   scale1(min_init, max_init, 5, &limloga, &limlogb, &dist);

   // set tickcnt to account for an extra tick before and after range to allow
   // minor tick calculations to occur correctly
   if (limlogb > limloga)  this->tickcnt = (unsigned int)((limlogb - limloga) + 3);
      else                 this->tickcnt = 3;

   // compute new tick positions
   this->tickpos = new double[tickcnt];
   if (this->tickpos)
   {
      unsigned int rr;
      int adj;
      adj = (int)limloga -1;

      // compute major tick locations as powers of 10
      for (rr = 0; rr < this->tickcnt; rr++)
         this->tickpos[rr] = pow(10.0, adj++);

      if (this->minor_mode == chips_nice)
      {
         this->minor_mode_adjusted = chips_count;
         this->minor_num_adjusted = 8;  // set number of minor ticks
      }
   }

   if (this->use_transform ())
   {
      if (this->untransformed_tickpos && this->transform_applied)
         delete []this->untransformed_tickpos;

      this->untransformed_tickpos = new double[this->tickcnt];
      for (int ii=this->tickcnt; ii--;)
         this->untransformed_tickpos[ii] = this->tickpos[ii];
      this->remove_transform (this->untransformed_tickpos, this->tickcnt);
      this->transform_applied = true;
   }
   else
   {
      if (this->untransformed_tickpos && this->transform_applied)
         delete []this->untransformed_tickpos;

      this->untransformed_tickpos = this->tickpos;
      this->transform_applied = false;
   }

   return rval;
}

int ChipsCartesianAxisAlgorithm::log_limit_tick_limits(bool automin, bool automax, double pad)
{
   double limloga, limlogb, dist, min_init = -5, max_init = 1;
   double delta, delta_lo, delta_hi, tmin, tmax;

   // clear out the tick positions
   this->set_tick_positions (NULL, 0);

   tmin = this->log_min;
   tmax = this->log_max;

   // adjust by pad factor
   delta = (log10(this->log_max) - log10(this->log_min)) * pad;
   if (automin)
      delta_lo = pow(10.0, delta);
   else
      delta_lo = 1.0;
   if (automax)
      delta_hi = pow(10.0, delta);
   else
      delta_hi = 1.0;

   if (delta_lo != 0)
      tmin = this->log_min / delta_lo;
   tmax = this->log_max * delta_hi;

   // compute log data range
   min_init = log10(tmin);
   max_init = log10(tmax);

   // use linear scale1 algorithm to compute log axis limits
   scale1(min_init, max_init, 5, &limloga, &limlogb, &dist);

   // store axis limits
   this->padded_min = tmin;
   this->padded_max = tmax;

   this->fix_minmax (this->padded_min, this->padded_max, false);

   return ERR_STATUS_GOOD;
}



int ChipsCartesianAxisAlgorithm::log_limit_tick_positions()
{
   double limloga, limlogb, dist, min_init = -5, max_init = 1;
   int rval = ERR_STATUS_GOOD;

   // compute log data range
   min_init = log10(this->padded_min);
   max_init = log10(this->padded_max);

   // use linear scale1 algorithm to compute log axis limits
   scale1(min_init, max_init, 5, &limloga, &limlogb, &dist);

   // set tickcnt to account for an extra tick before and after range to allow
   // minor tick calculations to occur correctly
   if (limlogb > limloga)  this->tickcnt = (unsigned int)((limlogb - limloga) + 3);
      else                 this->tickcnt = 3;


   // compute new tick positions
   this->tickpos = new double[tickcnt];
   if (this->tickpos)
   {
      unsigned int rr;
      int adj = (int)limloga -1;

      // compute major tick locations as powers of 10
      for (rr = 0; rr < this->tickcnt; rr++)
         this->tickpos[rr] = pow(10.0, adj++);

      if (this->minor_mode == chips_nice)
      {
         this->minor_mode_adjusted = chips_count;
         this->minor_num_adjusted = 8;  // set number of minor ticks
      }
   }

   if (this->use_transform ())
   {
      if (this->untransformed_tickpos && this->transform_applied)
         delete []this->untransformed_tickpos;

      this->untransformed_tickpos = new double[this->tickcnt];
      for (int ii=this->tickcnt; ii--;)
         this->untransformed_tickpos[ii] = this->tickpos[ii];
      this->remove_transform (this->untransformed_tickpos, this->tickcnt);
      this->transform_applied = true;
   }
   else
   {
      if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

      this->untransformed_tickpos = this->tickpos;
      this->transform_applied = false;
   }

   return rval;
}

int ChipsCartesianAxisAlgorithm::log_count_tick_limits()
{
   double base;

   if (major_num != 0)
   {
      base = rint(pow( 10.0, log10(this->log_max-this->log_min) / this->major_num));
      if (base <= 1.0) base = 2;

      // find tick range
      if (this->log_min < this->log_max)
      {
         int el = 1, eh;

         if (this->log_min <= base)
         {
            while (pow(base, el) > this->log_min) el--;
         }
         else
         {
            while (pow(base, el) < this->log_min) el++;
            el--;
         }
         eh = el;
         while(pow(base, eh) < this->log_max) eh++;
         this->tickcnt = eh - el + 1;

         this->padded_min =  pow(base, el);
         this->padded_max = pow(base, eh);
      }
   }

   return ERR_STATUS_GOOD;
}




int ChipsCartesianAxisAlgorithm::log_count_tick_positions()
{
   int rval = ERR_STATUS_GOOD;
   double base;

   if (major_num != 0)
   {
      base = rint(pow( 10.0, log10(this->log_max-this->log_min) / this->major_num));
      if (base <= 1.0) base = 2;

      // find tick range
      if (this->log_min < this->log_max)
      {
         int el = 1, eh;

         // clear out the tick positions
         this->set_tick_positions (NULL, 0);

         if (this->log_min <= base)
         {
            while (pow(base, el) > this->log_min) el--;
         }
         else
         {
            while (pow(base, el) < this->log_min) el++;
            el--;
         }
         eh = el;
         while(pow(base, eh) < this->log_max) eh++;
         this->tickcnt = eh - el + 1;

         // setup new tick locations
         this->tickpos = new double [tickcnt];
         if (this->tickpos)
         {
            unsigned int rr;
            for (rr = 0; rr < this->tickcnt; rr++)
               this->tickpos[rr] = pow(base, el++);
         }

         if (this->minor_mode == chips_nice)
         {
            this->minor_mode_adjusted = chips_count;
            if (base >= 2 && base <101) this->minor_num_adjusted = (int)(base - 2);
               else this->minor_num_adjusted = 0;
         }

         if (this->use_transform ())
         {
            if (this->untransformed_tickpos && this->transform_applied)
               delete []this->untransformed_tickpos;

            this->untransformed_tickpos = new double[this->tickcnt];
            for (int ii=this->tickcnt; ii--;)
               this->untransformed_tickpos[ii] = this->tickpos[ii];
            this->remove_transform (this->untransformed_tickpos, this->tickcnt);
            this->transform_applied = true;
         }
         else
         {
            if (this->untransformed_tickpos && this->transform_applied)
               delete []this->untransformed_tickpos;

            this->untransformed_tickpos = this->tickpos;
            this->transform_applied = false;
         }
      }
   }

   return rval;
}

int ChipsCartesianAxisAlgorithm::log_interval_tick_limits(bool automin, bool automax, double pad)
{
   double adjust = pow(10.0, ((log10(this->log_max) - log10(this->log_min)) * pad));
   this->padded_min = this->log_min;
   this->padded_max = this->log_max;

   this->fix_minmax (this->padded_min, this->padded_max, false);

   if (automin && (adjust != 0)) this->padded_min /= adjust;
   if (automax) this->padded_max *= adjust;

   return ERR_STATUS_GOOD;
}



int ChipsCartesianAxisAlgorithm::log_interval_tick_positions()
{
   int rval = ERR_STATUS_GOOD;

   // for now, if the interval is 1 or less, use 10
   if (major_int <= 1)  major_int = 10;

   // find tick range
   if (this->padded_min < this->padded_max)
   {
      int el = 1, eh;

      // clear out the tick positions
      this->set_tick_positions (NULL, 0);

      // find appropriate exponents for range
      if (this->log_min <= this->major_int)
      {
         while (pow(this->major_int, el) > this->padded_min) el--;
      }
      else
      {
         while (pow(this->major_int, el) < this->padded_min) el++;
         el--;
      }
      eh = el;
      while(pow(this->major_int, eh) < this->padded_max) eh++;
      this->tickcnt = eh - el + 1;

      // setup new tick locations
      this->tickpos = new double[tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         for (rr = 0; rr < this->tickcnt; rr++)
            this->tickpos[rr] = pow(this->major_int, el++);
      }
      else if ((this->tickpos = new double[2]) != NULL)
      {
         // original malloc failed- set minimal array
         this->tickcnt = 2;
         this->tickpos[0] = this->padded_min;
         this->tickpos[1] = this->padded_max;
      }

      if (this->minor_mode == chips_nice)
      {
         this->minor_mode_adjusted = chips_count;
         if (this->major_int < 101)
            this->minor_num_adjusted = (int)(this->major_int - 2);  // set number of minor ticks
      }

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }

   }
   return rval;
}

int ChipsCartesianAxisAlgorithm::log_arbitrary_tick_limits(bool automin, bool
 automax, double pad)
{
   double adjust = pow(10.0, ((log10(this->log_max) - log10(this->log_min)) * pad));
   this->padded_min = this->log_min;
   this->padded_max = this->log_max;

   this->fix_minmax (this->padded_min, this->padded_max, false);

   if (automin && (adjust != 0)) this->padded_min /= adjust;
   if (automax) this->padded_max *= adjust;

   return ERR_STATUS_GOOD;
}

int ChipsCartesianAxisAlgorithm::log_arbitrary_tick_positions()
{

   if (!this->arbitrary_set)
   {
      this->set_tick_positions (this->arbpos, this->arbcnt);
      this->arbitrary_set = true;
   }

   return ERR_STATUS_GOOD;
}



void ChipsCartesianAxisAlgorithm::reset_string_labels (const string& tickformat)
{
   unsigned int rr;

   if ((major_mode == chips_arbitrary) && this->arbtext)
   {
      if (this->tick_text != NULL)
         // tick text has changed- recompute tick_text
         delete[] this->tick_text;
      this->ttext_cnt = this->arbcnt;
      this->tick_text = new string[this->arbcnt];
      for (unsigned int ii=0; ii<this->arbcnt; ii++)
         this->tick_text[ii] = this->arbtext[ii];
      
   }
   else if (strchr(tickformat.c_str(), 'z') || strchr(tickformat.c_str(), 'Z'))
   {
      // format using 3.2 x10^2 instead of 3.2e+02 (%z=%e, %Z=%g)
      char tmpfmt[80];
      char tmpval[80];
      bool neg;
      strcpy(tmpfmt, tickformat.c_str());
      char* cpos = strchr(tmpfmt, 'z');
      if (!cpos)
      {
         cpos = strchr(tmpfmt, 'Z');
         if (cpos) *cpos = 'g';
      }
      else   *cpos = 'e';

      if (this->tick_text) delete []this->tick_text;
      this->tick_text = new string[this->tickcnt];
      this->ttext_cnt = tickcnt;

      unsigned int set_cnt = 0;
      if (tickformat == "%0.0z")
      {
         for (rr = 0; rr < this->tickcnt; rr++)
         {
            sprintf(tmpval, "%.3e", this->tickpos[rr]);
            if (strstr(tmpval, "1.000e") == &tmpval[0])  set_cnt++;
               else if (strstr(tmpval, "-1.000e") == &tmpval[0])  set_cnt++;
                  else if (strstr(tmpval, "0.000e") == &tmpval[0])  set_cnt++;
                     else if (this->tickpos[rr]< get_min() || this->tickpos[rr] > this->get_max())
                        set_cnt++;
         }
         if (set_cnt == this->tickcnt) strcpy(tmpfmt, "%.3e");
      }
      for (rr = 0; rr < this->tickcnt; rr++)
      {
         sprintf(tmpval, tmpfmt, this->tickpos[rr]);
         cpos = strchr(tmpval, 'e');
         if (cpos)
         {
            *cpos = '\0';
            // handle %0.0z as only 10^x
            if (set_cnt != this->tickcnt || this->tickpos[rr] != 0)
            {
               if (set_cnt != this->tickcnt)
               {
                  this->tick_text[rr] = tmpval;
                  this->tick_text[rr].append(" \\times 10^{");
               }
               else if (this->tickpos[rr] > 0)
                  this->tick_text[rr] = ("10^{");
               else
                  this->tick_text[rr] = ("-10^{");
               cpos++;
               neg =  (*cpos == '-');
               cpos++;
               while(*cpos == '0') cpos++;
               if (neg) this->tick_text[rr].append("-");
               if (*cpos == '\0')  this->tick_text[rr].append("0");
               else               this->tick_text[rr].append(cpos);
               this->tick_text[rr].append("}");
            }
            else   this->tick_text[rr] = "0";
         }
         else   this->tick_text[rr] = tmpval;
      }
   }
/********** TEMPORARY CODE *********/
   else if (strstr(tickformat.c_str(),"ra") || strstr(tickformat.c_str(),"RA"))
   {
      // using sexagesimal (HMS) format for right ascension (ra)
      sexagesimal_ra_ticks(tickformat);
   }
   else if (strstr(tickformat.c_str(),"dec") || strstr(tickformat.c_str(),"DEC"))
   {
      // using sexagesimal (DMS) format for declination (dec)
      sexagesimal_dec_ticks(tickformat);
   }
/*********** END TEMPORARY CODE **********/
   else
   {
      char tmpval[80];

      // don't use string labels
      if (this->tick_text && this->ttext_cnt)
         delete[] this->tick_text;
      if (!this->transform_applied)
      {
         this->tick_text = NULL;
         this->ttext_cnt = 0;
      }
      else
      {
         this->tick_text = new string[this->tickcnt];
         this->ttext_cnt = tickcnt;

         for (rr = 0; rr < this->tickcnt; rr++)
         {
            sprintf(tmpval, tickformat.c_str (), this->tickpos[rr]);
            this->tick_text[rr] = tmpval;
         }
      }
   }
}


void ChipsCartesianAxisAlgorithm::serialize (ChipsStateStream &out)
{
   bool bval;
   int ival;

   out.write_bool (&this->scalable);
   out.write_int (&this->scaletype);

   out.write_int ((int *)&this->ttext_cnt);
   out.write_int ((int *)&this->tickcnt);
   out.write_int ((int *)&this->arbcnt);

   if (this->tick_text)
   {
      for (unsigned int ii=this->ttext_cnt; ii--;)
         out.write_string (&this->tick_text[ii]);
   }
   if (this->tickpos)
      out.write_double (this->tickpos, (int)this->tickcnt);
   out.write_bool (&this->transform_applied);
   if (this->transform_applied)
      out.write_double (this->untransformed_tickpos, (int)this->tickcnt);
   
   if (this->arbpos)
      out.write_double (this->arbpos, (int)this->arbcnt);

   bval = (this->arbtext != NULL);
   out.write_bool (&bval);

   if (this->arbtext)
   {
      for (unsigned int ii=this->arbcnt; ii--;)
         out.write_string (&this->arbtext[ii]);
   }

   ival = this->major_mode;
   out.write_int (&ival);
   ival = this->minor_mode;
   out.write_int (&ival);
   ival = this->minor_mode_adjusted;
   out.write_int (&ival);
   out.write_int (&this->minor_num);
   out.write_int (&this->minor_num_adjusted);
   out.write_int (&this->major_num);
   out.write_double (&this->major_int);
   out.write_double (&this->minor_int);

   out.write_double (&this->padded_min);
   out.write_double (&this->padded_max);
   out.write_double (&this->lin_min);
   out.write_double (&this->lin_max);
   out.write_double (&this->log_min);
   out.write_double (&this->log_max);

   out.write_double (&this->log_minimum);

   out.write_bool (&this->arbitrary_set);

   out.write_double (&this->last_pad);
   out.write_bool (&this->last_automin);
   out.write_bool (&this->last_automax);

   out.write_int (&this->orientation);
}

void ChipsCartesianAxisAlgorithm::deserialize (ChipsStateStream &in)
{
   bool bval;
   int ival;

   in.read_bool (&this->scalable);
   in.read_int (&this->scaletype);

   in.read_int ((int *)&this->ttext_cnt);
   in.read_int ((int *)&this->tickcnt);
   in.read_int ((int *)&this->arbcnt);

   if (this->tick_text)
   {
      delete []this->tick_text;
      this->tick_text = NULL;
   }

   if (this->ttext_cnt)
   {
      this->tick_text = new string[this->ttext_cnt];
      for (unsigned int ii=this->ttext_cnt; ii--;)
         in.read_string (&this->tick_text[ii]);
   }

   if (this->tickpos)
   {
      delete []this->tickpos;
      this->tickpos = NULL;
   }
 
   if (this->tickcnt)
   {
      this->tickpos = new double[this->tickcnt];
      in.read_double (this->tickpos, (int)this->tickcnt);
   }

   if (in.get_minor_version () > 8)
   {
      if (this->untransformed_tickpos)
         delete []this->untransformed_tickpos;

      in.read_bool (&this->transform_applied);
      if (this->transform_applied && (this->tickcnt > 0))
      {
         this->untransformed_tickpos = new double[this->tickcnt];
         in.read_double (this->tickpos, (int)this->tickcnt);
      }
      else
         this->untransformed_tickpos = this->tickpos;
    
   }
   else
   {
      if (this->use_transform () && (this->tickcnt > 0))
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      } 
   }


   if (this->arbpos)
   {
      delete []this->arbpos;
      this->arbpos = NULL;
   }

   if (this->arbcnt)
   {
      this->arbpos = new double[this->arbcnt];
      in.read_double (this->arbpos, (int)this->arbcnt);
   }

   in.read_bool (&bval);

   if (this->arbtext)
   {
      delete []this->arbtext;
      this->arbtext = NULL;
   }

   if (bval)
   {
      this->arbtext = new string[this->arbcnt];
      for (int ii=this->arbcnt; ii--;)
         in.read_string (&this->arbtext[ii]);
   }

   in.read_int (&ival);
   this->major_mode = static_cast <V_TICKMODE_TYPE>(ival);
   this->major_mode_adjusted = this->major_mode;
   in.read_int (&ival);
   this->minor_mode = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&ival);
   this->minor_mode_adjusted = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&this->minor_num);
   in.read_int (&this->minor_num_adjusted);
   in.read_int (&this->major_num);
   in.read_double (&this->major_int);
   in.read_double (&this->minor_int);

   in.read_double (&this->padded_min);
   in.read_double (&this->padded_max);
   in.read_double (&this->lin_min);
   in.read_double (&this->lin_max);

   in.read_double (&this->log_min);
   in.read_double (&this->log_max);

   in.read_double (&this->log_minimum);

   in.read_bool (&this->arbitrary_set);

   in.read_double (&this->last_pad);
   in.read_bool (&this->last_automin);
   in.read_bool (&this->last_automax);

   if (in.get_minor_version () > 6)
      in.read_int (&this->orientation);
}





// SCALE1 algorithm (C.R. Lewart, CACM, Oct 73, v16 n10)
// finds a new range to nicely divide data into approx n linear intervals
void ChipsCartesianAxisAlgorithm::scale1(double inmin, double inmax, 
                     int slices, double* outmin, double* outmax, 
                     double* dist, double* base)
{
   double interval, scaled, fm, baseval;
   int nal, m;

   if (inmax == inmin)  ++inmax;
   if (!slices) slices = 5;     // avoid div by 0- use 5 as desired num ticks

   interval = fabs((inmax-inmin)/(double)slices);
   nal = (int) log10(interval);
   if (interval < 1) --nal;

   // scale interval between 1-10 and find value- e
   if ((scaled = interval/pow(10.0, nal)) < 1.41421356237) baseval = 1.0;
      else if (scaled < 3.16227766017)                     baseval = 2.0;
         else  if (scaled < 7.07106781187)                 baseval = 5.0;
            else                                           baseval = 10;
   if (base)  *base = baseval;

   if ((*dist = copysign(baseval * pow(10.0, nal), (inmax - inmin))) != 0)
   {
      // compute interval size
      fm = inmin/(*dist);
      if (fm < 0.0)   m = (int)fm - 1;
         else   m = (int)fm;
      if (fabs((double)m+1.0-fm) < 0.0000000000000001)  ++m;

      // find new minimum and maximum
      *outmin = (*dist) * (double)m;
      fm = inmax/(*dist);
      if (fm < -1.0)  m = (int)fm;
         else m = (int)fm + 1;
      if (fabs(fm+1.0-(double)m) < 0.0000000000000001)  --m;
      *outmax = (*dist) * (double)m;

      // adjust for roundoff
      if (*outmin > inmin)   *outmin = inmin;
      if (*outmax < inmax)   *outmax = inmax;
   }
   else
   {
      *outmin = inmin;
      *outmax = inmax;
   }
}

void ChipsCartesianAxisAlgorithm::scale2(double inmin, double inmax, 
                     int slices, double* outmin, double* outmax, 
                     double* dist)
{
   double interval, scaled, fm, del = 0.0000000000000001;
   double vint[5] = { 1.0, 2.0, 5.0, 10.0, 20.0 };
   int nal, m1, m2, nx, np, done = 0, ii = 1;

   interval = fabs((inmax-inmin)/(double)slices);
   nal = (int) log10(interval);
   if (interval < 1)   --nal;
   scaled = interval/pow(10.0, nal);
   while (!done && (ii < 6))
   {
      if ((scaled < vint[ii - 1] + del) || (ii > 3))
      {
         *dist = vint[ii - 1] * pow(10.0, nal);
         m1 = (int)(fm = inmin/(*dist));
         if (fabs((double)m1 + 1.0 - fm) < del) ++m1;
         *outmin = *dist * m1;
         fm = inmax/(*dist);
         m2 = (int)(fm + 1);
         if (fabs(fm + 1.0 -(double)m2) < del)  --m2;
         *outmax = (*dist) * m2;
         if ((np = m2 - m1) <= slices) done = 1;
      }
      ++ii;
   }
   nx = (slices - np) / 2;
   *outmin -= (double) nx * (*dist);
   *outmax = *outmin + (double)slices * (*dist);
   if (*outmin > inmin)   *outmin = inmin;
   if (*outmax < inmax)   *outmax = inmax;
}

void ChipsCartesianAxisAlgorithm::reset_adjusted_parameters ()
{
   this->minor_mode_adjusted = this->minor_mode;
   this->minor_num_adjusted = this->minor_num;
   this->major_mode_adjusted = this->major_mode;
}


/**************** TEMPORARY CODE ***********************/
int ChipsCartesianAxisAlgorithm::compute_sexagesimal_locations(int dec)
{
//   int rval = ERR_STATUS_GOOD;
   if (this->tickcnt == 0)
   {
//      cmd_manager->add_err("Cannot set tick format to sexagesimal if there are no tick marks.", id, 1);
      return ERR_STATUS_BAD;
   }

   double step, range =  this->tickpos[this->tickcnt -1] - this->tickpos[0];
   unsigned int rr;

   this->modified = true;

   // since ra is divided by 15 for hours, multiply dec range by 15 so
   // we can use the step/minor tick calculations for dec minutes/seconds
   if (dec)    range = range * 15;

   if (range > 30)     // hours  (range in degrees/15 = hours)
   {
      if (dec)  // dec is in degrees so use base 10 steps
      {
         if (range < 75)          // under 5 degrees -> step = 1 degree
            step = 15;
         else if (range < 195)    // under 13 degrees -> step = 2 degrees
            step = 30;
         else if (range < 300)    // under 20 degrees -> step = 5 degrees
            step = 75;
         else if (range < 1125)   // under 75 degrees -> step = 10 degrees
            step = 150;
         else                     // otherwise  step = 25 degrees
            step = 375;
         if (range < 195)         // minor ticks to 3 or 4 depending on step
            this->minor_num_adjusted = 3;
         else
            this->minor_num_adjusted = 4;
      }
      else      // ra is in hours so use steps of 1,2,4, or 6
      {
         this->minor_num_adjusted = 3;
         if (range < 75)          // under 5 hours -> step = 1 hour
            step = 15;
         else if (range < 150)    // under 10 hours -> step = 2 hour
            step = 30;
         else if (range < 225)    // under 18 hours -> step = 4 hour
            step = 60;
         else                     // otherwise  step = 6 hours
         {
            step = 90;
            this->minor_num_adjusted = 5;
         }
         range = fmod(range, 360); // limit hours to 0 - 360 degrees
      }
   }
   else if (range >= 1)    // minutes (1 degree = 4 minutes)
   {
      if (range > 22.5)        // over 1.5 hours  -> step = 30 min
      {
         step = 7.5;
         this->minor_num_adjusted = 2;
      }
      else if (range > 11.25)  // over 45 minutes -> step = 15 min
      {
         step = 3.75;
         this->minor_num_adjusted = 2;
      }
      else if (range > 7.5)    // over 30 minutes -> step = 10 min
      {
         step = 2.5;
         this->minor_num_adjusted = 4;
      }
      else if (range > 3.75)   // over 15 minutes -> step =  5 min
      {
         step = 1.25;
         this->minor_num_adjusted = 4;
      }
      else if (range > 1.25)   // over 5 minutes ->  step =  2 min
      {
         step = 0.5;
         this->minor_num_adjusted = 3;
      }
      else                 // otherwise   step = 1 min
      {
         step = 0.25;
         this->minor_num_adjusted = 3;
      }
   }
   else if (range > 0.0167)   // seconds
   {
      if (range > 0.3750)       // over 1.5 minutes  -> step = 30 sec
      {
        step = 0.125;
        this->minor_num_adjusted = 2;
      }
      else if (range > 0.1875)  // over 45 seconds -> step = 15 sec
      {
        step = 0.0625;
        this->minor_num_adjusted = 2;
      }
      else if (range > 0.1250)  // over 30 seconds -> step = 10 sec
      {
        step = 0.041666667;
        this->minor_num_adjusted = 4;
      }
      else if (range > 0.0625)   // over 15 seconds -> step =  5 sec
      {
        step = 0.020833333;
        this->minor_num_adjusted = 4;
      }
      else if (range > 0.020833333)  // over 5 seconds ->  step =  2 sec
      {
        step = 0.008333333;
        this->minor_num_adjusted = 3;
      }
      else                 // otherwise   step = 1 sec
      {
        step = 0.004166667;
        this->minor_num_adjusted = 3;
      }
   }
   else
   {
//      cmd_manager->add_err(rval=ERR_DBL_MIN, id, 0, "minimum axis data value", 0.0167);
      return ERR_DBL_MIN;
   }

   if (dec)   step /= 15.0;    // remove scaling used for dec

   if ((this->major_mode != chips_arbitrary) && (this->scaletype != V_SCALE_LOG))
   {


   if (step)
   {
      // recompute loccations for tick labels
      double lowtick = ((int)(this->tickpos[0]/step - 1)) * step;
      this->tickcnt = (int)((this->tickpos[this->tickcnt -1] - lowtick)/step) + 4;
      delete []this->tickpos;
      this->tickpos = new double[this->tickcnt];
      for (rr = 0; rr < this->tickcnt; rr++)  this->tickpos[rr] = lowtick + step*rr;

      if (this->use_transform ())
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = new double[this->tickcnt];
         for (int ii=this->tickcnt; ii--;)
            this->untransformed_tickpos[ii] = this->tickpos[ii];
         this->remove_transform (this->untransformed_tickpos, this->tickcnt);
         this->transform_applied = true;
      }
      else
      {
         if (this->untransformed_tickpos && this->transform_applied)
            delete []this->untransformed_tickpos;

         this->untransformed_tickpos = this->tickpos;
         this->transform_applied = false;
      }
   }
   else   return ERR_STATUS_BAD;

   }
   return ERR_STATUS_GOOD;
}


int ChipsCartesianAxisAlgorithm::sexagesimal_ra_ticks(const string &tickformat)
{
   // converts tick positions to sexagesimal format for ra
   if (this->tickcnt == 0)
   {
//      cmd_manager->add_err("Cannot set tick format to sexagesimal if there are no tick marks.", id, 1);
      return ERR_STATUS_BAD;
   }

   // set 'nice' tick locations for non-user specified tick locations
   // for now- simply plot log scale values at their decimal degree locations
   if ((this->major_mode != chips_arbitrary) && (this->scaletype != V_SCALE_LOG))
      this->compute_sexagesimal_locations(0);

   // create labels
   string hmfmt;
   double tpos, hd, md, ss;
   unsigned int rr;
   char tmpfmt[80], tmpval[80], tmpspec[80];
   char *fmt = tmpspec;

   strcpy(tmpspec, tickformat.c_str());
   char* cpos = strstr(tmpspec, "ra");
   if (!cpos)   cpos = strstr(tmpspec, "RA");
   *cpos = '\0';
   if (tmpspec[0] == '%') fmt++;
   if (!strlen(fmt))  strcpy(fmt, "02.0");
   if (strstr(tickformat.c_str(), "ra1"))
   {
      hmfmt = "%02d:%02d:";
      sprintf(tmpfmt, "%%%sf", fmt);
   }
   else if (strstr(tickformat.c_str(), "ra2"))
   {
      hmfmt="%02d %02d ";
      sprintf(tmpfmt, "%%%sf", fmt);
   }
   else if (strstr(tickformat.c_str(), "RA3"))
   {
      hmfmt="%02dH %02dM ";
      sprintf(tmpfmt, "%%%sfS", fmt);
   }
   else if (strstr(tickformat.c_str(), "ra3"))
   {
      hmfmt = "%02dh %02dm ";
      sprintf(tmpfmt, "%%%sfs", fmt);
   }
   else if (strstr(tickformat.c_str(), "RA4"))
   {
      hmfmt = "%02d^H %02d^M ";
      sprintf(tmpfmt, "%%%sf^S", fmt);
   }
   else
   {
      hmfmt = "%02d^h %02d^m ";
      sprintf(tmpfmt, "%%%sf^s", fmt);
   }

   if (this->tick_text) delete []this->tick_text;
   this->tick_text = new string[this->tickcnt];
   this->ttext_cnt = this->tickcnt;

   char* sixty_p;
   for (rr = 0; rr < this->tickcnt; rr++)
   {
      if ((tpos = fmod(this->tickpos[rr], 360.0)) < 0) tpos+=360;
      ss = tpos/15.0;
      ss = modf(ss, &hd);
      hd = fmod(hd, 24);       // wrap at 24 hours
      ss = modf(ss*60.0 + FLT_EPSILON, &md);
      ss = ss * 60 + 0.00137;

      // account for roundup to 60 seconds
      sprintf(tmpspec, tmpfmt, ss);
      if ((sixty_p = strstr(tmpspec,"60")) != NULL)
      {
         *sixty_p = '0';
         md++;
      }
      // account for roundup to 60 minutes;
      sprintf(tmpval,hmfmt.c_str(), (int)hd, (int)md);
      if ((sixty_p = strstr(tmpval,"60")) != NULL)
      {
         // account for 23h to 00h
         if (hd == 23)  sprintf(tmpval,hmfmt.c_str(), 0, 0);
            else        sprintf(tmpval,hmfmt.c_str(), (int)hd+1, 0);
      }
      this->tick_text[rr] = tmpval;
      this->tick_text[rr].append(tmpspec);
   }
   return ERR_STATUS_GOOD;
}


int ChipsCartesianAxisAlgorithm::sexagesimal_dec_ticks(const string &tickformat)
{
   // converts tick positions to sexagesimal format for dec
   if (this->tickcnt == 0)
   {
//      cmd_manager->add_err("Cannot set tick format to sexagesimal if there are no tick marks.", id, 1);
      return ERR_STATUS_BAD;
   }

   // set 'nice' tick locations for non-user specified tick locations
   // for now- simply plot log scale values at their decimal degree locations
   if ((this->major_mode != chips_arbitrary) && (this->scaletype != V_SCALE_LOG))
      this->compute_sexagesimal_locations(1);

   // create labels
   double hd, md, ss;
   unsigned int swapsign, rr;
   char tmpfmt[80], tmpval[80], tmpspec[80];
   char *fmt = tmpspec;

   strcpy(tmpspec, tickformat.c_str());
   char* cpos = strstr(tmpspec, "dec");
   if (!cpos)   cpos = strstr(tmpspec, "DEC");
   *cpos = '\0';
   if (tmpspec[0] == '%') fmt++;
   if (!strlen(fmt))  strcpy(fmt, "02.0");
   if (strstr(tickformat.c_str(), "dec1"))
      sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);
   else if (strstr(tickformat.c_str(), "dec2"))
      sprintf(tmpfmt, "%%+02d %%02d %%%sf", fmt);
   else if (strstr(tickformat.c_str(), "DEC3"))
      sprintf(tmpfmt, "%%+02dD %%02dM %%%sfS", fmt);
   else if (strstr(tickformat.c_str(), "dec3"))
      sprintf(tmpfmt, "%%+02dd %%02dm %%%sfs", fmt);
   else
      sprintf(tmpfmt, "%%+02d^o %%02d' %%%sf\"", fmt);

   if (tick_text) 
      delete []this->tick_text;
   this->tick_text = new string[tickcnt];
   this->ttext_cnt = this->tickcnt;

   for (rr = 0; rr < this->tickcnt; rr++)
   {
      ss = modf(this->tickpos[rr], &hd);
      if (ss >= 0)
         ss = modf(ss*60.0 + FLT_EPSILON, &md);
      else
         ss = modf(ss*60.0 - FLT_EPSILON, &md);
      ss *= 60;
      // handle degrees between -1:0 with a - prefix infront of degrees
      if ((swapsign = ((hd <= 0) && ((md < 0) || (ss < 0)))))
      {
         if (md < 0) md *= -1;
         if (ss < 0) ss *= -1;
      }
      // handle minutes field set to +/-60 due to precision issues
      if ((md == 60) || (md == -60))
      {
         md = 0;
         hd++;
      }

      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss);
      this->tick_text[rr] = tmpval;
      if (swapsign)  this->tick_text[rr][0] = '-';
   }
   return ERR_STATUS_GOOD;
}

bool ChipsCartesianAxisAlgorithm::fix_minmax (double &minval, double &maxval, bool transformed)
{
   bool reversed;
   if (minval > maxval)
   {
      // adjust to ensure min is less than max
      double tmp = maxval;
      maxval = minval;
      minval = tmp;

      reversed = true;
   }
   else
   {
      reversed = false;
//      if (minval == maxval && minval == this->padded_min && maxval == this->padded_max)
      if (flt_equal (minval, maxval) && flt_equal (minval, this->padded_min) && flt_equal (maxval, this->padded_max))
      {
         // adjust to ensure range is not 1 point
         if (minval != 0)
         {
            minval *= 0.95;
            maxval *= 1.05;
         }
         else
         {
            minval = -1.0;
            maxval = 1.0;
         }
      }
   }

   return reversed;
}

double ***ChipsCartesianAxisAlgorithm::calculate_grid (int &out_line_count, int &out_slices, int tick_type)
{
   out_line_count = 0;
   out_slices = 0;

   return NULL;
}

bool ChipsCartesianAxisAlgorithm::validate_coordinates (const ChipsVector &coord, const string &ds_id)
{
/*   ChipsDataSystem *ds = this->axis->get_data_system (ds_id);
   return ds->validate_coordinates (coord, ds_id);
*/
   return true;
}

void ChipsCartesianAxisAlgorithm::remove_transform (double *data, int data_size)
{

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   Transform *transform = NULL;
   if (dsCoord != NULL)
      transform = dsCoord->get_transform (dsCoord->get_default_transform_id ());
   if (transform)
   {

      string tmp_id = dsCoord->get_id_no_wcs ();
      ChipsVector min_coord, max_coord;
      double **matrix = transform->get_transform_matrix ();
      int index = (this->orientation == RA_AXIS) ? 0 : 1;
      int opp_index = (index == 0) ? 1 : 0;
      double opp_value;

      get_axis_position (false, min_coord, max_coord);
      opp_value = min_coord[opp_index];

      for (int ii=0; ii<data_size; ii++)
         data[ii] = (data[ii] - (opp_value *matrix[opp_index][index]) - matrix[2][index])/matrix[index][index]; 

   }
}

void ChipsCartesianAxisAlgorithm::apply_transform (double *data, int data_size)
{
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   Transform *transform = NULL;
   if (dsCoord != NULL)
      transform = dsCoord->get_transform (dsCoord->get_default_transform_id ());
   if (transform)
   {

      string tmp_id = dsCoord->get_id_no_wcs ();
      ChipsVector min_coord, max_coord;
      double **matrix = transform->get_transform_matrix ();
      int index = (this->orientation == RA_AXIS) ? 0 : 1;
      int opp_index = (index == 0) ? 1 : 0;
      double opp_value;

      get_axis_position (false, min_coord, max_coord);
      opp_value = min_coord[opp_index];

      for (int ii=0; ii<data_size; ii++)
         data[ii] = (data[ii]*matrix[index][index]) + (opp_value*matrix[opp_index][index]) + matrix[opp_index][index];

   }

}

void ChipsCartesianAxisAlgorithm::get_axis_position (bool use_transform, ChipsVector &min_coord, ChipsVector &max_coord)
{
   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   int opp_index = (index == 0) ? 1 : 0;

   min_coord = ChipsVector (0,0);
   max_coord = ChipsVector (1,1);

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   if (dsCoord == NULL)
      return;

   string tmp_id = "";

   if (!use_transform)
      tmp_id = dsCoord->get_id_no_wcs ();


   // convert from frame to plot to get the correct position of the plot
   min_coord[opp_index] = this->axis->get_norm_pos ();
   max_coord[opp_index] = this->axis->get_norm_pos ();

   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               FRAME_NORM, "", PLOT_NORM, "");
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               FRAME_NORM, "", PLOT_NORM, "");

   // get the data coordinates from the plot
   min_coord[index] = 0;
   max_coord[index] = 1;
   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               PLOT_NORM, "", DATA, tmp_id);
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               PLOT_NORM, "", DATA, tmp_id);
}

double* ChipsCartesianAxisAlgorithm::get_tick_positions (bool transformed)
{
   if (!transformed)
      return this->untransformed_tickpos;
   return this->tickpos;
}

bool ChipsCartesianAxisAlgorithm::use_transform ()
{
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   Transform *transform;
   string transform_id;
   bool rval = false;
   double **matrix;

   if (dsCoord)
   {
      transform_id = dsCoord->get_default_transform_id ();
      transform = dsCoord->get_transform (transform_id);

      if (transform)
      {
         // ignore the identity matrix as a transform
         matrix = transform->get_transform_matrix ();
         if (matrix)
         {
            for (int ii=3; ii--;)
               for (int jj=3; jj--;)
                  if (((ii != jj) && !flt_equal(matrix[ii][jj],0)) ||
                      ((ii == jj) && !flt_equal(matrix[ii][jj],1)))
                  {
                     rval = true;
                     break;
                  }
                  
         }
         else
            rval = true;
      }
   }
   return rval;
}

void ChipsCartesianAxisAlgorithm::format_point (const string &tickformat, double pnt, string &formatted_pnt)
{
   char tmp_pnt[80];
/*   // for arbitrary points return the interpolated enumerated value
   if (major_mode == chips_arbitrary)
   {
      sprintf (tmp_pnt, "%f", pnt); 
      formatted_pnt = tmp_pnt;
   }
*/

   // format z/Z for using 3.2 x10^2 instead of 3.2e+02 (%z=%e, %Z=%g)
   // don't support z or Z -- use alternative
   if (strchr(tickformat.c_str(), 'z'))
   {
      string tmp_format = tickformat;
      tmp_format[tickformat.length ()-1] = 'e';
      sprintf (tmp_pnt, tmp_format.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
   else if (strchr(tickformat.c_str(), 'Z'))
   {
      string tmp_format = tickformat;
      tmp_format[tickformat.length ()-1] = 'g';
      sprintf (tmp_pnt, tmp_format.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
   else if (strstr(tickformat.c_str(),"ra") || strstr(tickformat.c_str(),"RA"))
   {

      string hmfmt;
      double tpos, hd, md, ss;
      char tmpfmt[80], tmpval[80], tmpspec[80];
      char *fmt = tmpspec;
      char* sixty_p;

      strcpy(tmpspec, tickformat.c_str());
      char* cpos = strstr(tmpspec, "ra");
      if (!cpos)   cpos = strstr(tmpspec, "RA");
      *cpos = '\0';
      if (tmpspec[0] == '%') fmt++;
      if (!strlen(fmt))  strcpy(fmt, "02.0");

      // using sexagesimal (HMS) format for right ascension (ra)
/*      if (strstr(tickformat.c_str(), "ra1"))
      {
         hmfmt = "%02d:%02d:";
         sprintf(tmpfmt, "%%%sf", fmt);
      }
      else if (strstr(tickformat.c_str(), "ra2"))
      {
         hmfmt="%02d %02d ";
         sprintf(tmpfmt, "%%%sf", fmt);
      }
      else if (strstr(tickformat.c_str(), "RA3"))
      {
         hmfmt="%02dH %02dM ";
         sprintf(tmpfmt, "%%%sfS", fmt);
      }
      else if (strstr(tickformat.c_str(), "ra3"))
      {
         hmfmt = "%02dh %02dm ";
         sprintf(tmpfmt, "%%%sfs", fmt);
      }
      else if (strstr(tickformat.c_str(), "RA4"))
      {
         hmfmt = "%02d^H %02d^M ";
         sprintf(tmpfmt, "%%%sf^S", fmt);
      }
      else
      {
         hmfmt = "%02d^h %02d^m ";
         sprintf(tmpfmt, "%%%sf^s", fmt);
      }
*/
      // always use : deliminated format
      hmfmt = "%02d:%02d:";
      sprintf(tmpfmt, "%%%sf", fmt);

      if ((tpos = fmod(pnt, 360.0)) < 0) tpos+=360;
      ss = tpos/15.0;
      ss = modf(ss, &hd);
      hd = fmod(hd, 24);       // wrap at 24 hours
      ss = modf(ss*60.0 + FLT_EPSILON, &md);
      ss = ss * 60 + 0.00137;

      // account for roundup to 60 seconds
      sprintf(tmpspec, tmpfmt, ss);
      if ((sixty_p = strstr(tmpspec,"60")) != NULL)
      {
         *sixty_p = '0';
         md++;
      }
      // account for roundup to 60 minutes;
      sprintf(tmpval,hmfmt.c_str(), (int)hd, (int)md);
      if ((sixty_p = strstr(tmpval,"60")) != NULL)
      {
         // account for 23h to 00h
         if (hd == 23)  sprintf(tmpval,hmfmt.c_str(), 0, 0);
            else        sprintf(tmpval,hmfmt.c_str(), (int)hd+1, 0);
      }
      formatted_pnt = tmpval;
      formatted_pnt.append(tmpspec);
   }
   else if (strstr(tickformat.c_str(),"dec") || strstr(tickformat.c_str(),"DEC"))
   {
      // using sexagesimal (DMS) format for declination (dec)

      double hd, md, ss;
      unsigned int swapsign;
      char tmpfmt[80], tmpval[80], tmpspec[80];
      char *fmt = tmpspec;

      strcpy(tmpspec, tickformat.c_str());
      char* cpos = strstr(tmpspec, "dec");
      if (!cpos)   cpos = strstr(tmpspec, "DEC");
      *cpos = '\0';
      if (tmpspec[0] == '%') fmt++;
      if (!strlen(fmt))  strcpy(fmt, "02.0");
/*      if (strstr(tickformat.c_str(), "dec1"))
         sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);
      else if (strstr(tickformat.c_str(), "dec2"))
         sprintf(tmpfmt, "%%+02d %%02d %%%sf", fmt);
      else if (strstr(tickformat.c_str(), "DEC3"))
         sprintf(tmpfmt, "%%+02dD %%02dM %%%sfS", fmt);
      else if (strstr(tickformat.c_str(), "dec3"))
         sprintf(tmpfmt, "%%+02dd %%02dm %%%sfs", fmt);
      else
         sprintf(tmpfmt, "%%+02d^o %%02d' %%%sf\"", fmt);
*/
      // always use : deliminated format
      sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);

      ss = modf(pnt, &hd);
      if (ss >= 0)
         ss = modf(ss*60.0 + FLT_EPSILON, &md);
      else
         ss = modf(ss*60.0 - FLT_EPSILON, &md);
      ss *= 60;
      // handle degrees between -1:0 with a - prefix infront of degrees
      if ((swapsign = ((hd <= 0) && ((md < 0) || (ss < 0)))))
      {
         if (md < 0) md *= -1;
         if (ss < 0) ss *= -1;
      }
      // handle minutes field set to +/-60 due to precision issues
      if ((md == 60) || (md == -60))
      {
         md = 0;
         hd++;
      }

      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss);
      formatted_pnt = tmpval;
      if (swapsign)  formatted_pnt[0] = '-';
   }
   else
   {
      sprintf(tmp_pnt, tickformat.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
}

int ChipsCartesianAxisAlgorithm::set_major_mode (V_TICKMODE_TYPE mode)
{
   if (this->major_mode != mode)
      this->modified=true;
   this->major_mode=mode;

   return ERR_STATUS_GOOD;
}

int ChipsCartesianAxisAlgorithm::set_minor_mode (V_TICKMODE_TYPE mode)
{
   if (this->minor_mode != mode)
      this->modified=true;
   this->minor_mode=mode;

   return ERR_STATUS_GOOD;
}



