// 
//  Copyright (C) 2006,2007,2011-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsCmdManager.hh"
#include "ChipsEngine.hh"
#include "ChipsAdjustRelsizeCommand.hh"
#include "ChipsAxisTickCommand.hh"
#include "ChipsFillCommand.hh"
#include "ChipsPrefCommand.hh"
#include "ChipsRepoCommand.hh"
#include "ChipsContourLevelCommand.hh"
#include "ChipsDepthUpdateCommand.hh"
#include "ChipsDepthVisCommand.hh"
#include "ChipsDepthArrangeCommand.hh"
#include "ChipsDepthCollapseCommand.hh"
#include "ChipsDeleteCommand.hh"
#include "ChipsImageChannelCommand.hh"
#include "ChipsTransformCommand.hh"
#include "ChipsDataAspectCommand.hh"
#include "ChipsZoomCommand.hh"
#include "ChipsPanCommand.hh"
#include "ChipsError.hh"

#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>

bool ChipsCmdManager::errmap_sort_check(list<ChipsError*>::const_reference first, list<ChipsError*>::const_reference second)
{
   return (first->get_text() < second->get_text());
}

bool ChipsCmdManager::errmap_dup_check(list<ChipsError*>::const_reference first, list<ChipsError*>::const_reference second)
{
   return ((first->get_text() == second->get_text()) &&
            (first->get_type() == second->get_type())); 
}

ChipsCmdManager::ChipsCmdManager (ChipsEngine *eng)
{
   ignore_state = false; 
   state_mask = STATE_CHNG_CLEAR; 
   lock = false;
   curr_cmd = commands.end ();
   ref = 0;
   blockref = 0; 
   tot_cmd_weight = 0;
   cm_eng = eng;
   curr_state = undo_states.end ();
   autocommit = true; 
   errmap.clear();
   def_errtype = 0;
   err_verbosity = 1;
   err_cnt = warn_cnt = 0; 
}

ChipsCmdManager::~ChipsCmdManager ()
{
   list<ChipsCommand *>::iterator ii;
   for (ii=commands.begin (); ii!=commands.end (); ii++)
      delete (*ii);
   list <pair<string, int> >::iterator jj;
   for (jj=undo_states.begin (); jj!=undo_states.end (); jj++)
      remove (jj->first.c_str());
}

int ChipsCmdManager::add_cmd (ChipsCommand *cmd)
{
   int retVal = ERR_STATUS_GOOD;
   if (lock)   return ERR_CMD_LOCKED;

   // by default, autocommit is set to true to push the command onto
   // the undo stack and clear the redo stack. If autocommit is false,
   // put the command in the hold buffer. This allows us to undo a
   // failed command w/o clearing the redo stack
   if (!autocommit)
   {
      if (cmd)  
         hold.push_back(cmd);
      else
         retVal = ERR_STATUS_BAD;
      return retVal;
   }

   list<ChipsCommand *>::iterator ii=commands.end ();
   int curr_block_ref = 0;
   if (!commands.empty ()){
      ii--;
      if (!block_ref_cnts.empty ())
         curr_block_ref = block_ref_cnts.back ();
   }

   // remove any commands after the current command
   while (curr_cmd != ii){
      tot_cmd_weight -= (*ii)->get_weight ();
      // if the commands ref equals the current block
      // ref then pull that block from the block_ref_cnts list
      // and close out the block
      if ((*ii)->get_ref_id () == curr_block_ref){
         block_ref_cnts.pop_back ();
	 if (!block_ref_cnts.empty ())
            curr_block_ref = block_ref_cnts.back ();
         else
            curr_block_ref = 0;

         // close out the block 
         blockref--;
      }
      (*ii)->clear ();
      delete (*ii);
      commands.pop_back ();
      ii = commands.end ();
      if (!commands.empty ())   ii--;
   }


   // Remove tags that come after this command
   remove_tag_by_iterator (curr_state, ref);

   if (!undo_states.empty () && (curr_state != undo_states.end ()))
   {
      while (curr_state != undo_states.begin ())
      {
         // if we pop off the state- remove associated tags
         remove_tag_by_iterator(undo_states.begin());
         remove (undo_states.front ().first.c_str ());
         undo_states.pop_front ();
      }


      // If total command weight exceends the max weight
      // the clear the commands. This means its the last command
      // in the state was executed, but the state is full. Since
      // the state is already written no need to re-write it
      // which would happen in the following command sequence.
      if (tot_cmd_weight > CMD_MNGER_MAX_WEIGHT)
         clear_cmds ();
      else
      {

         // Replace the tags with this state with end as its about
         // to be removed.
         replace_tag_iterator (undo_states.begin(), undo_states.end ());

      
         // Pop the current state off the list and remove the file
         // this file needs to be rewritten. 
         remove (undo_states.front ().first.c_str ());
         undo_states.pop_front ();
      }


      // Set the current state to end to so that we know we're no
      // longer using any state
      curr_state = undo_states.end ();
   }

   if (tot_cmd_weight > CMD_MNGER_MAX_WEIGHT){
      // cannot write to state if in block. if the
      // total command weight exceeds the max while 
      // in a block the state will be saved after the
      // block is closed
      if (!in_block ()){
         if ((retVal = commands_to_file ()) != ERR_STATUS_GOOD)
            return retVal;
         clear_cmds ();
      }
   }
   ref++;

   // Store of the first ref cnt of this block
   if (save_ref){
      block_ref_cnts.push_back (ref);
      save_ref = false;
   }


   tot_cmd_weight += cmd->get_weight ();

   cmd->set_ref_id (ref);
   cmd->set_block_id (blockref);
   commands.push_back (cmd);
   if (curr_cmd == commands.end ())
      curr_cmd = commands.begin ();
   else
      curr_cmd++;

   return retVal;
}

int ChipsCmdManager::open_block()
{
   if (lock)   return ERR_CMD_LOCKED;

   blockref++;

   // Store of the first ref cnt of this block
   //block_ref_cnts.push_back (ref+1);
   save_ref = true;

   return ERR_STATUS_GOOD; 
}

int ChipsCmdManager::close_block()
{
   int rval = ERR_STATUS_GOOD; 

   if (lock)   return ERR_STATUS_GOOD;
   if (!in_block ())   return ERR_STATUS_BAD;

   blockref--;

   // Collapse any nested blocks within this block
   // to be part of this block. Also reset all 
   // ref id to a single ref id. This helps with
   // undo and redo.
   int block_ref_cnt = -1;

   if (curr_cmd != commands.end())
   {
      list <ChipsCommand *>::iterator cmd_iter = curr_cmd;
      int cur_block = (*curr_cmd)->get_block_id();

      // The reference id for each block is held in the
      // block_ref_cnts list. If current block is
      // greater than the blockref then this is a block which
      // has commands to be collapsed. Pull the block reference
      // id for this block from the list.
      if (!block_ref_cnts.empty () && (cur_block > blockref)){
         block_ref_cnt = block_ref_cnts.back ();
         block_ref_cnts.pop_back ();
      }

      while(cur_block > blockref)
      {
         (*cmd_iter)->set_block_id (blockref);
	 (*cmd_iter)->set_ref_id (block_ref_cnt);
         if (cmd_iter != commands.begin())
            cmd_iter--;
         else
            break;
         cur_block = (*cmd_iter)->get_block_id();
      }

      // Put the blocks reference id back into the list if
      // the reference id was pulled from the list (block_ref_cnt
      // would not equal -1), there are parent blocks (blockref > 0),
      // and this is either the first command
      if ((blockref > 0) && (block_ref_cnt != -1)){
         if ((cmd_iter == commands.begin()) || 
             (cur_block < blockref))
               block_ref_cnts.push_back (block_ref_cnt);
      }
   }

   // Reset the ref count to the current command. 
   if (block_ref_cnt != -1)   ref = block_ref_cnt;
   
   // If no longer in a block then its ok to write cache to state file
   if (!in_block() &&  ((tot_cmd_weight > CMD_MNGER_MAX_WEIGHT)))
   {
      // write state on close of block...
      if ((rval = commands_to_file ()) != ERR_STATUS_GOOD)
         return rval;
      clear_cmds (true);
      tot_cmd_weight = 0;
   }

   return rval; 
}


int ChipsCmdManager::execute_cmd (int cnt)
{
   int num_cmds, pos, rr, retVal = ERR_STATUS_GOOD;
   int curr_cnt = cnt;
   bool old_lock = lock;


   list<ChipsCommand *>::iterator ii = commands.end ();
   if (commands.empty ())
   {
      add_err(ERR_NO_REDO);
      return ERR_NO_REDO;
   }
   if (cnt < 1) 
   {
      add_err(ERR_ZERO_CNT);
      return ERR_ZERO_CNT;  
   }

   cm_eng->disable_redraw();

   pos = tot_cmd_cnt (curr_cmd);

   /* Check if there's enough commands in the command list otherwise
      find a state file to use */
   ii--;
   num_cmds = tot_cmd_cnt (ii);
   if ((curr_cmd == ii) ||(curr_cnt+pos > num_cmds)){
      if (undo_states.empty () || (curr_state  == undo_states.begin ()) ||
	  curr_state == undo_states.end ()){
	 lock = true;
         lock = old_lock;
         cm_eng->enable_redraw();
         if (cnt > 1) 
         {
            add_err(ERR_UNDO_COUNT,"",0,cnt,"redo");
            return ERR_UNDO_COUNT;
         }
         else
         {
            add_err(ERR_NO_REDO);
            return ERR_NO_REDO;
         }
      }
      curr_cnt -= (num_cmds - pos);

      list<pair <string, int> >::iterator jj = curr_state;
      jj--;
      while ((curr_cnt > jj->second) && (jj != undo_states.begin ())){
         curr_cnt -= jj->second;
         jj--;
      }
      if (curr_cnt > jj->second){
         lock = true;
         cm_eng->enable_redraw();
         lock = old_lock;
         if (cnt > 1)
         {
            add_err(ERR_UNDO_COUNT,"",0,cnt,"redo");
            return ERR_UNDO_COUNT;
         }
         else
         {
            add_err(ERR_NO_REDO);
            return ERR_NO_REDO;
         }
      }
      cnt = curr_cnt;
      curr_state = jj;

      // Load the new state
      cm_eng->load_state (jj->first, V_LOAD_ALL);

      // Back off all the commands that were done in this state
      if (curr_cmd != commands.end ())
      {
         for (;curr_cmd != commands.begin (); curr_cmd--)
            (*curr_cmd)->unexecute ();
         (*curr_cmd)->unexecute ();

         // curr_cmd will be at the beginning but the beginning will be undone
         // reset it to the end so that the section below will know to start
         // at the beginning
         curr_cmd=commands.end ();
      }
   }
 
   // iterate through count steps
   while(cnt--) 
   {
      int done = 0, xx = 0; 

      // handle blocks (compound commands)
      while(!done)
      {
         if (curr_cmd == commands.end())
            curr_cmd = commands.begin();
         else
            curr_cmd++;
         if (xx++ == 0)
         {
            rr = (*curr_cmd)->get_ref_id();
            retVal = (*curr_cmd)->execute();
         }
         else if (commands.empty() || ((*curr_cmd) == NULL) ||
                 (curr_cmd == commands.end()) ||
                 ((*curr_cmd)->get_ref_id() != rr))
         {
            done = 1;
            curr_cmd--; 
         }
         else
            retVal = (*curr_cmd)->execute();
      }
      if (curr_cmd != commands.end ())
         ref = (*curr_cmd)->get_ref_id();
      else
	 ref = 0;
   }

   lock = true;
   cm_eng->enable_redraw();
   cm_eng->redraw_window("all");
   lock = old_lock;
   return retVal;
}


int ChipsCmdManager::execute_cmd (string usetag)
{
   int retVal = ERR_STATUS_GOOD;
   TAG_ELEM_P_T tag_p;
   bool old_lock = lock;

   if (commands.empty ())
   {
      add_err(ERR_NO_REDO);
      return ERR_NO_REDO;
   }
   if ((tag_p = find_tag(usetag)) == NULL)
   {
      add_err(ERR_TAG_DNE);
      return ERR_TAG_DNE;    // tag does not exist
   }
   // check if tag is current or stored
   if (tag_p->iter == curr_state)
   {
      if (ref >= tag_p->ref)  
      {
         add_err(ERR_BAD_TAG_REDO);
         return ERR_BAD_TAG_REDO;
      }
   }
   // this is an odd corner case that can popup if the last action was
   // written to file and there was a tag after that but no other commands
   else if (tag_p->iter == undo_states.end ())
   {
      cm_eng->disable_redraw();

      // if the current state is the not the first state load the
      // first a state file
      if (curr_state != undo_states.begin ())
      {
         curr_state = undo_states.begin ();

         cm_eng->load_state (curr_state->first, V_LOAD_ALL);
      }

      // move the tag from the begining of the empty state to the end of the 
      // first state. this allows for easier redo/undo capabilities in the 
      // future
      list <ChipsCommand *>::iterator tmp_cmd = commands.end ();
      tmp_cmd--;
      tag_p->iter = undo_states.begin ();
      tag_p->ref = (*tmp_cmd)->get_ref_id ();

   }
   else
   {
      list<pair <string, int> >::iterator jj = curr_state;
      int done = 0, found = 0; 
      while (!(found || done))
      {
         if (tag_p->iter == jj)   found = 1; 
         else if (jj == undo_states.begin())   done = 1;  
         else   jj--; 
      }
      if (!found)  
      {
         add_err(ERR_BAD_TAG_REDO);
         return ERR_BAD_TAG_REDO;
      }
      else
      {
         curr_state = tag_p->iter;

         cm_eng->disable_redraw();
         cm_eng->load_state (curr_state->first, V_LOAD_ALL);

         // Back off all the commands that were done in this state
         if (curr_cmd != commands.end ())
         {
            for (;curr_cmd != commands.begin (); curr_cmd--)
               (*curr_cmd)->unexecute ();
            (*curr_cmd)->unexecute ();

            // curr_cmd will be at the beginning but the beginning will be 
            // undone reset it to the end so that the section below will
            // know to start at the beginning
            curr_cmd=commands.end ();
         }
      }
   }

   cm_eng->disable_redraw();

   int  done = 0;
   while(!done)
   {
      if (curr_cmd == commands.end())
         curr_cmd = commands.begin();
      else
         curr_cmd++;
      if (commands.empty() || ((*curr_cmd) == NULL) ||
               (curr_cmd == commands.end()) ||
              ((*curr_cmd)->get_ref_id() > tag_p->ref))
      {
         done = 1;
         curr_cmd--;
      }
      else
         retVal = (*curr_cmd)->execute();
   }
   if (curr_cmd != commands.end ())
      ref = (*curr_cmd)->get_ref_id();
   else
      ref = 0;

   lock = true;
   lock = old_lock;
   cm_eng->enable_redraw();
   cm_eng->redraw_window("all");
   return retVal;
}


int ChipsCmdManager::unexecute_cmd (int cnt)
{
   int rr, retVal = ERR_STATUS_GOOD, done = 0;
   int num_cmds = tot_cmd_cnt (curr_cmd);
   int curr_cnt = cnt;
   bool old_lock = lock;
   bool state_exists = false;
   rr = num_cmds;

   if (cnt < 1) 
   {
      add_err(ERR_ZERO_CNT);
      return ERR_ZERO_CNT;  
   }

   // Find out if another state exists
   list<pair <string, int> >::iterator ii = curr_state;
   if (!undo_states.empty ()){
      // If the list isn't empty then another state probably exists
      state_exists = true;
      if (curr_state != undo_states.end ()){
         ii = curr_state;
         ii++;
	 // If the next state is end then no more states exists
	 if (ii == undo_states.end ())
            state_exists = false;
      }
   }

   // If the cnt is greater than the number of commands in this state or
   // the cnt equals the number of commands in this state and there's another
   // state to process then use the next undo state. 
   if ((cnt > num_cmds) || ((cnt == num_cmds) && state_exists)){
      if (undo_states.empty ())
      {
         if (cnt > 1)
         {
            add_err(ERR_UNDO_COUNT,"",0,cnt,"undo");
            return ERR_UNDO_COUNT;
         }
         else
         {
            add_err(ERR_NO_UNDO);
            return ERR_NO_UNDO;
         }
      }
      if (curr_cmd != commands.end ())
	 curr_cnt -= num_cmds;

      // If the current state is at the end it mean that we're not
      // actually in a loaded undo state. Traverse back to the 
      // beginning of the list to find the next viable undo state
      if (curr_state == undo_states.end ())
         ii = undo_states.begin ();

      // Find the next state that has the number commands
      while ((ii != undo_states.end ()) && (curr_cnt > ii->second)){
         curr_cnt -= ii->second;
	 ii++;
      }
      if (ii == undo_states.end ())
      {
         if (cnt > 1)
         {
            add_err(ERR_UNDO_COUNT,"",0,cnt,"undo");
            return ERR_UNDO_COUNT;
         }
         else
         {
            add_err(ERR_NO_UNDO);
            return ERR_NO_UNDO;
         }
      }
      cm_eng->disable_redraw();

      if (curr_state == undo_states.end ()){
	 /* Redo everything so that the whole state is available 
	    This might not be the best way to handle this another
	    approach is to write the state to file every time an
	    undo sequence is started */
	 list<ChipsCommand *>::iterator jj = curr_cmd;

	 // Write the current state to file so we can get back to it
	 // if the user starts doing redo's
         if (!commands.empty())
            retVal = commands_to_file ();

	 clear_cmds ();
	 if (retVal != ERR_STATUS_GOOD){
            lock = true;
            cm_eng->enable_redraw();
	    lock = old_lock;
	    return retVal; 
	 }
      }

      // When the state is loaded one command is undone simply by starting at
      // the next command
      cnt = curr_cnt; 
      curr_state = ii;

      // Always load state. This is necessary because if an object was
      // deleted it is explicitly deleted from the system when the state
      // is written out.
      cm_eng->load_state (ii->first, V_LOAD_ALL);

      // Redo all commands to make sure that the current command is at
      // the very end of the state
      if (curr_cmd ==commands.end ())
         curr_cmd = commands.begin ();
      else
         curr_cmd++;

      for (; curr_cmd!=commands.end (); curr_cmd++)
         (*curr_cmd)->execute ();
      curr_cmd--;

   }   

   // iterate through cnt undo's 

   cm_eng->disable_redraw();

   while(cnt--) 
   {
      rr = (*curr_cmd)->get_ref_id();
      done = 0; 
      while(!done)
      {
         if ((curr_cmd == commands.end ()) || commands.empty () ||
             ((*curr_cmd)->get_ref_id() != rr)) 
            done = 1;
         else 
         {
            if ((retVal = (*curr_cmd)->unexecute()) == ERR_STATUS_GOOD)
            {
               if (curr_cmd == commands.begin ())
	       {
                  curr_cmd = commands.end ();
	          done = 1;  
	       }
               else
                  curr_cmd--;
            }
	    else  done = 1; 
         }
      }
      if (curr_cmd != commands.end ())
         ref=(*curr_cmd)->get_ref_id();
      else
	 ref = 0;
   }
   lock = true;
   cm_eng->enable_redraw();
   cm_eng->redraw_window("all");
   lock = old_lock;
   return retVal;
}


int ChipsCmdManager::unexecute_cmd (string usetag)
{
   int retVal = ERR_STATUS_GOOD;
   TAG_ELEM_P_T tag_p;
   bool old_lock = lock;

   if ((tag_p = find_tag(usetag)) == NULL)
   {
      add_err(ERR_TAG_DNE);
      return ERR_TAG_DNE;    // tag does not exist 
   } 
   // check if tag is current or stored
   if (tag_p->iter == curr_state)
   {
      if (ref <= tag_p->ref)   return ERR_BAD_TAG_UNDO;
   }
   else 
   {
      // tag state must be before current for redo
      bool found = false; 
      list <pair<string, int> >::iterator ll = curr_state;
      list <pair<string, int> >::iterator last = undo_states.end();
      last--; 

      if (curr_state == undo_states.end ())
         ll = undo_states.begin ();

      while ((ll != undo_states.end()) && !found)
      {
         if (ll == tag_p->iter)  found = 1;  
         ll++;
      }
      if ((!found) &&  (tag_p->iter != last)) 
      {
         add_err(ERR_BAD_TAG_UNDO);
	 return ERR_BAD_TAG_UNDO;
      }

      cm_eng->disable_redraw();

      if (curr_state == undo_states.end ()){

         // Write the current state to file so we can get back to it
         // if the user starts doing redo's
         if (!commands.empty ())
            retVal = commands_to_file ();

         clear_cmds ();
         if (retVal != ERR_STATUS_GOOD){
	    lock = true;
            lock = old_lock;
            cm_eng->enable_redraw();
            return retVal;
         }
      }

      // reload stored file
      curr_state = tag_p->iter; 
      cm_eng->load_state (tag_p->iter->first, V_LOAD_ALL);

      // redo all commands so that the current command starts at the end
      // of the state
      if (curr_cmd ==commands.end ())
         curr_cmd = commands.begin ();
      else
         curr_cmd++;

      for (; curr_cmd!=commands.end (); curr_cmd++)
         (*curr_cmd)->execute ();
      curr_cmd--;

   }

   cm_eng->disable_redraw();

   // iterate through cnt undo's
   if (curr_cmd == commands.end())   curr_cmd = commands.begin();

   list<ChipsCommand *>::iterator jj = curr_cmd;
   int done = 0;

   while(!done)
   {
      if ((curr_cmd == commands.end ()) || commands.empty () ||
          ((*curr_cmd)->get_ref_id() == tag_p->ref))
      {
         done = 1;
      }
      else
      {
         if ((retVal = (*curr_cmd)->unexecute()) == ERR_STATUS_GOOD)
         {
            if (curr_cmd == commands.begin ())
            {
               curr_cmd = commands.end ();
               done = 1;
            }
            else
               curr_cmd--;
         }
         else  done = 1;
      }
   }
   if (curr_cmd != commands.end ())
      ref=(*curr_cmd)->get_ref_id();
   else
      ref = 0;

   lock = true;
   lock = old_lock;
   cm_eng->enable_redraw();
   cm_eng->redraw_window("all");
   return retVal;
}


void ChipsCmdManager::print ()
{
   list<ChipsCommand *>::iterator ii;
   for (ii=commands.begin (); ii!=commands.end (); ii++)
      (*ii)->print ();
   cerr << "tot command weight: " << tot_cmd_weight << endl;
   if (curr_cmd != commands.end())
      (*curr_cmd)->print ();
}


int ChipsCmdManager::serialize(ChipsStateStream& out)
{
   int sz = 0, retVal = ERR_STATUS_GOOD;
   list<ChipsCommand *>::iterator ii;
   out.write_bool(&compound);
   // write out blockref. when saving state blockref should always be zero
   // since you cant be in a block and write to file. However its possible
   // that the block ref is not zero since a block could be opened for
   // for the current state. Explicitly set the block ref to be zero
   out.write_int(&sz);
   out.write_int(&ref);
   out.write_checkpoint();
   sz = commands.size(); 
   out.write_int(&sz);
   for (ii=commands.begin (); ii!=commands.end (); ii++)
   {
      (*ii)->serialize(out);
   }

   int ref_id = -1;
   if (curr_cmd != commands.end ())
      ref_id = (*curr_cmd)->get_ref_id ();
   out.write_int (&ref_id); 

   return retVal;
}

int ChipsCmdManager::deserialize(ChipsEngine* eng, ChipsStateStream& in)
{
   int retVal = ERR_STATUS_GOOD;
   int val, len, typ, rr;
   ChipsCommand* cm; 
   V_OBJECT_TYPE otype;
   ChipsObject* obj; 
   string id;
   ChipsId cid;
   int obj_ref;

   if (!eng) return ERR_STATUS_BAD;

   // remove existing undo/redo commands 
   list<ChipsCommand *>::iterator ii;
   for (ii=commands.begin (); ii!=commands.end (); ii++)
      delete (*ii);
   commands.clear();
   curr_cmd = commands.end ();
   tot_cmd_weight = 0;

   in.read_bool(&compound);
   in.read_int(&blockref);
   in.read_int(&ref);
   if (in.read_checkpoint() != ERR_STATUS_GOOD) 
   {
      add_err(ERR_STATEFILE_INC);
      return ERR_STATEFILE_INC;
   }
   // repopulate command list
   in.read_int(&len);
   for (rr =len; rr--; )
   {
      in.read_int(&val);
      in.read_int(&typ);
      otype = (V_OBJECT_TYPE)typ; 
      if (otype != chips_not_object)  
      {
         in.read_int (&obj_ref);
         cid.set_object_ref_id (otype, static_cast <unsigned int>(obj_ref));
      }
      switch(otype)
      {
         case chips_colorbar:
           in.read_string(&id);
           cid.set_colorbar_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
         break;
	 case chips_region:
	   in.read_string(&id);
           cid.set_region_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
	 break; 
	 case chips_point:
	   in.read_string(&id);
           cid.set_point_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
	 break; 
	 case chips_line:
	   in.read_string(&id);
           cid.set_line_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
	 break; 
	 case chips_label:
	   in.read_string(&id);
           cid.set_label_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
	 break;
	 case chips_axis:      // fallthrough intended
	   in.read_string(&id);
           cid.set_axis_id(id);
	 case chips_plot:      // fallthrough intended
	   in.read_string(&id);
           cid.set_plot_id(id);
           if (otype != chips_plot)
           {
              in.read_int (&obj_ref);
              cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           }
         case chips_frame:     // fallthrough intended
	   in.read_string(&id);
           cid.set_frame_id(id);
           if (otype != chips_frame)
           {
              in.read_int (&obj_ref);
              cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           }
         case chips_window:
	   in.read_string(&id);
           cid.set_window_id(id);
           if (otype != chips_window)
           {
              in.read_int (&obj_ref);
              cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
           }
	 break; 
         case chips_curve:  
           in.read_string(&id);
           cid.set_curve_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
         break;
         case chips_image:  
           in.read_string(&id);
           cid.set_image_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
         break;
         case chips_contour:  
           in.read_string(&id);
           cid.set_contour_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
         break;
         case chips_histogram:  
           in.read_string(&id);
           cid.set_histogram_id(id);
           in.read_string(&id);
           cid.set_plot_id(id);
           in.read_int (&obj_ref);
           cid.set_plot_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_frame_id(id);
           in.read_int (&obj_ref);
           cid.set_frame_ref_id (static_cast <unsigned int>(obj_ref));
           in.read_string(&id);
           cid.set_window_id(id);
           in.read_int (&obj_ref);
           cid.set_window_ref_id (static_cast <unsigned int>(obj_ref));
         break;
         default:
	 break; 
      }
      // The add/remove transform function has no object associated with it
      if ((otype != chips_not_object) && ((obj = eng->get_object(&cid, otype, true)) != NULL))
      {
         switch(val)
         {
            case 1: 
	       cm = new ChipsCreateCommand(obj, in); 
            break; 
            case 2: 
	       cm = new ChipsDeleteCommand(obj, in); 
            break; 
            case 3: 
	       cm = new ChipsSetCommand(obj, in); 
            break; 
	    case 4:
               cm = new ChipsLimitsCommand(obj, in);
            break;
	    case 5:
               cm = new ChipsMoveCommand(obj, in);
            break;
            case 6:
               cm = new ChipsTileCommand(obj, in);
            break; 
            case 7:
               cm = new ChipsSplitCommand(obj, in);
            break; 
            case 8:
               cm = new ChipsSetSplitCommand(obj, in);
            break; 
            case 9:
               cm = new ChipsCurrentCommand(obj, in);
            break; 
	    case 10:
            break;
	    case 11:
            break;
	    case 12:
	       cm = new ChipsBindCommand(obj, in);
	    break;
	    case 13:
	       cm = new ChipsRepoCommand(obj, in);
	    break;
            case 15:
               cm = new ChipsDepthVisCommand(obj, in);
            break;
            case 16:
               cm = new ChipsDepthArrangeCommand(obj, in);
            break;
            case 17:
               cm = new ChipsDepthCollapseCommand(obj, in);
            break;
            case 18:
               cm = new ChipsTransformCommand(obj, in);
            break;
            case 19:
               cm = new ChipsDepthUpdateCommand(obj, in);
            break;
            case 20:
               cm = new ChipsContourLevelCommand(obj, in);
            break;
            case 21:
               cm = new ChipsAxisTickCommand(obj, in);
            break;
            case 23: 
               cm = new ChipsAdjustRelsizeCommand(obj, in);
            break; 
            case 24:
               cm = new ChipsDataAspectCommand (obj, in);
            break;
            case 25:
               cm = new ChipsZoomCommand (obj, in);
            break;
            case 26:
               cm = new ChipsPanCommand (obj, in);
            break;
            case 27:
               cm = new ChipsImageChannelCommand(obj, in);
            break;
            case 28:
               cm = new ChipsDefDispColorCommand(obj, in);
            break;
	    default:
	       cm = NULL; 
	    break; 
         }
      }
      else
      {
	 // handle commands without objects (such as pref commands)
         switch(val)
         {
            case 14:
               cm = new ChipsPrefCommand(in);
            break;
            case 22:
               cm = new ChipsFillCommand(in);
            break;
	    default:
	       cm = NULL; 
	    break; 
         }
      }
      if (cm)
      {
	 commands.push_back (cm);
	 tot_cmd_weight += cm->get_weight ();
      }
   }

   int ref_id = -1;
   in.read_int (&ref_id);

   if (ref_id == -1)
      curr_cmd = commands.end ();
   else
   {
      list<ChipsCommand *>::iterator ii = commands.end ();
      while (ii != commands.begin ())
      {
         ii--;
         if ((*ii)->get_ref_id () == ref_id)
         {
            curr_cmd = ii;
            break;
         }
      }
   }
   return retVal;
}
int ChipsCmdManager::commands_to_file ()
{
   int ret_val = ERR_STATUS_GOOD;
   char *buf;

// The following should be replaced by a common function from ChipsDefine
// Since the same exact task is done by the chips server

   string filename = "/tmp/.chips";
   struct passwd *pw = getpwuid(getuid());
   if (pw && pw->pw_name[0] != '\0')
   {
      filename.append(".");
      filename.append(pw->pw_name);
   }

   if (access(filename.c_str(), W_OK) != 0)
   {
      mkdir(filename.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
      if (access(filename.c_str(), W_OK) != 0)  filename = "/tmp";
   }
   filename.append("/");

// This is the end  of what needs to be replaced
   
   filename.append ("chips-XXXXXX");

   buf = new char [filename.length ()+1];
   strcpy (buf, filename.c_str ());
   if (!mktemp (buf))
      return ERR_STATUS_BAD;
   pair <string, int> p;
   p.first = buf;
   list<ChipsCommand *>::iterator ii = commands.end ();
   if (commands.empty ())
      return ERR_STATUS_BAD;
   ii--;
   p.second = tot_cmd_cnt (ii);
   undo_states.push_front (p);
   ofstream out(buf, ios::out|ios::binary);

   // save state- set clobber override to true
   cm_eng->save_state (p.first, true, true);

   // update the tag array
   list<TAG_ELEM_P_T>::iterator tt;
   for(tt = tags.begin(); tt != tags.end(); ++tt)
   {
      if ((*tt)->iter == undo_states.end())
         (*tt)->iter = undo_states.begin();
   }

   delete []buf;

   return ret_val;
}


TAG_ELEM_P_T ChipsCmdManager::find_tag(string val)
{
   TAG_ELEM_P_T rval = NULL;
 
   if (val.length())
   {
      list<TAG_ELEM_P_T>::iterator tt;
      for(tt = tags.begin(); tt != tags.end(); ++tt)
      {
         if ((*tt)->tag == val)
            rval = (*tt);
      }
   } 
   return rval; 
}


int ChipsCmdManager::add_tag(string val)
{
   int rval = ERR_STATUS_BAD;
   if (val.length())
   {
      TAG_ELEM_P_T tagged = find_tag(val);

      if (tagged)
         rval = ERR_TAG_USED;   // tag already in use
      else 
      {
         if ((tagged = new TAG_ELEM_T) != NULL)
         { 
            tagged->tag = val;
            if (curr_cmd != commands.end())
               tagged->ref = (*curr_cmd)->get_ref_id();
            else
               tagged->ref = 0;
            
            tagged->iter = curr_state;

	    // Group tags together that belong to the same state
            list<TAG_ELEM_P_T>::iterator ii=tags.end();
	    while ((ii-- != tags.begin ()) && ((*ii)->iter == curr_state));

            if ((ii != tags.end ()) &&(*ii)->iter == curr_state)
            {
               ii++;
               tags.insert (ii,tagged);
	    }
            else
               tags.push_back(tagged); 
            rval = ERR_STATUS_GOOD; 
         }
      }
   }
   return rval; 
}


int ChipsCmdManager::remove_tag(string val) 
{
   int rval = ERR_TAG_DNE; 
   if (val.length())
   {
      TAG_ELEM_P_T deltag;
      for(list<TAG_ELEM_P_T>::iterator ii=tags.begin();
          (ii != tags.end()); ++ii)
      {
         if ((*ii)->tag == val)
         {
            deltag = *ii;
            delete(deltag);
            ii = tags.erase(ii);
            rval = ERR_STATUS_GOOD;
         }
      }
   }
   else   rval = ERR_STATUS_BAD; 
   return rval; 
}


int ChipsCmdManager::remove_tag_by_iterator(
   list<pair <string, int> >::iterator val, int tag_ref)
{
   int rval = ERR_TAG_DNE;
   bool done = false;
   bool found = false;
   TAG_ELEM_P_T deltag;
   list<TAG_ELEM_P_T>::iterator ii=tags.end();
   while (!done && !tags.empty() && (ii != tags.begin ()))
   {
      ii--;
      // remove everything from this state after the reference
      // the default reference is -1
      if (((*ii)->iter == val) && ((*ii)->ref > tag_ref))
      {
         found = true;
         deltag = *ii;
         delete(deltag);
	 ii = tags.erase(ii);
         rval = ERR_STATUS_GOOD;
      }
      // since tags are grouped by iterators once all the tags
      // were removed from this iterator then we're done
      else if (found) done = true; 
   }
   return rval;
}

int ChipsCmdManager::replace_tag_iterator(
   list<pair <string, int> >::iterator val, 
   list<pair <string, int> >::iterator new_val)
{
   int rval = ERR_TAG_DNE;
   bool done = false;
   bool found = false;
   list<TAG_ELEM_P_T>::iterator ii=tags.end();
   while (!done && !tags.empty() && (ii != tags.begin ()))
   {
      ii--;
      // remove everything from this state after the reference
      // the default reference is -1
      if ((*ii)->iter == val)
      {
         found = true;
	 (*ii)->iter = new_val;
         rval = ERR_STATUS_GOOD;
      }
      // since tags are grouped by iterators once all the tags
      // were removed from this iterator then we're done
      else if (found) done = true;
   }
   return rval;
}



int ChipsCmdManager::print_tags()
{
   list<TAG_ELEM_P_T>::iterator tt;
   for(tt = tags.begin(); tt != tags.end(); ++tt)
   {
      if ((*tt)->iter != undo_states.end())
         cout << "    tag: " << (*tt)->tag << " "<< (*tt)->ref << ":" << (*tt)->iter->first << endl; 
      else 
         cout << "    tag: " << (*tt)->tag << " "<< (*tt)->ref << ":end" << endl; 
   }
   return 0; 
}


void ChipsCmdManager::clear_cmds (bool force_del)
{
  list<ChipsCommand *>::iterator ii;
  for (ii=commands.begin (); ii!=commands.end (); ii++)
  {
     ChipsDeleteCommand* cdc = dynamic_cast<ChipsDeleteCommand*>(*ii);
     if (cdc)   cdc->clear_objects();
     delete (*ii);
  }
   commands.clear ();
   curr_cmd = commands.end ();
   ref=0;
   tot_cmd_weight = 0;
}


int ChipsCmdManager::clear_hold_buffer()
{
   list<ChipsCommand *>::reverse_iterator ri;
   for (ri=hold.rbegin (); ri!=hold.rend (); ri++)
   {
      (*ri)->unexecute();
      delete (*ri);
   }
   hold.clear ();

   return ERR_STATUS_GOOD; 
}


int ChipsCmdManager::commit_hold_buffer()
{
   int sts, retval = ERR_STATUS_GOOD;
   bool autoc = autocommit;
   autocommit = true; 
   list<ChipsCommand *>::iterator ii;
   for (ii=hold.begin (); ii!=hold.end (); ii++)
   {
      sts = add_cmd(*ii);
      if (sts != ERR_STATUS_GOOD) retval = sts; 
   }
   hold.clear ();
   autocommit = autoc;
   return retval;  
}


int ChipsCmdManager::tot_cmd_cnt (list<ChipsCommand *>::iterator cmds)
{
   int num_cmds = 1; // Start one for the current command
   int prev_ref_id;
   list<ChipsCommand *>::iterator ii;

   if (cmds == commands.end())
      return 0;

   // Do get the number commands up to the 'cmds'
   prev_ref_id = -1;
   for (ii=commands.begin (); ii!=cmds; ii++)
   {
      if ((*ii)->get_ref_id () != prev_ref_id)
	 num_cmds++;
      prev_ref_id = (*ii)->get_ref_id ();
   }

   // If the command is not part of a block then the previous
   // ref id should not equal the command iterator's ref id
   // If it is part of a block then the command has been counted
   // twice. Once when num_cmds was initialized and again during
   // the for loop. Subtract off one from num_cmds to account for
   // this.
   if ((*ii)->get_ref_id () == prev_ref_id)
      num_cmds--;
   return num_cmds;
}


void ChipsCmdManager::get_undo_counts(int &ucnt, int &rcnt)
{
   int found=0; 

   ucnt = rcnt = 0; 

   // handle state files
   list <pair<string, int> >::reverse_iterator ri;
   for (ri=undo_states.rbegin (); ri!=undo_states.rend (); ri++)
   {
      if (ri->first == curr_state->first)
      {
         found = 1;
      }
      else if (found)
      {
         rcnt += ri->second; 
      }
      else  
      {
         ucnt += ri->second; 
      } 
   }
   // handle current command list
   list<ChipsCommand *>::iterator ii = commands.end();
   int tot = tot_cmd_cnt (curr_cmd);
   ucnt += tot;
   rcnt = rcnt + tot_cmd_cnt(--ii) - tot; 
}


void ChipsCmdManager::add_err(string txt, string oid, int typ)
{
   if (typ == -1)  typ = def_errtype; 
   ChipsError* err = new ChipsError(oid, 0, txt, typ); 
   if (typ == 0) err_cnt++;
      else warn_cnt++; 
   errmap.push_back(err);
}

void ChipsCmdManager::add_err(string txt, int errcode, string oid, int typ)
{
   if (typ == -1)  typ = def_errtype; 
   ChipsError* err = new ChipsError(oid, errcode, txt, typ); 
   if (typ == 0) err_cnt++;
      else warn_cnt++; 
   errmap.push_back(err);
}

void ChipsCmdManager::add_err(int errcode, string oid, int typ, ...)
{
   if (typ == -1)  typ = def_errtype; 
   va_list  ap;

   va_start(ap, typ);
   ChipsError* err = new ChipsError(oid, errcode, typ, ap); 
   va_end(ap);

   if (typ == 0) err_cnt++;
      else warn_cnt++; 
   errmap.push_back(err);
}

void ChipsCmdManager::remove_last_err()
{
   if (!errmap.empty())
   {
      ChipsError* berr  = errmap.back();
      if (berr)
      {
         if (tagged_err && (berr == tagged_err))
            tagged_err = NULL;   
         if (berr->get_type() == 0)  err_cnt--;
            else  warn_cnt--;
         errmap.pop_back();
         delete berr; 
      } 
   }
}

void ChipsCmdManager::remove_errs_from_tagged()
{
   if (tagged_err)
   {
      list<ChipsError *>::iterator emi, first;
      ChipsError* curr_p;
      int erase = 0; 
      for (emi = errmap.begin(); emi != errmap.end(); emi++) 
      {
         curr_p = (*emi);
         if (curr_p == tagged_err)
         {
            erase = 1; 
            first = emi; 
         }
         if (erase)
         {
            delete curr_p;
            if (curr_p->get_type() == 0)  err_cnt--;
               else   warn_cnt--;
         }
      }
      if (erase)  errmap.erase(first, errmap.end()); 
   }
}


string ChipsCmdManager::get_errs(int mode)
{
   static string errs;
   list<ChipsError *>::iterator emi;

   // if we have a low verbosity- remove errors with duplicate text
   list<ChipsError*>errmap2(errmap); 
   if (err_verbosity < 3)
   {
      errmap2.sort(&errmap_sort_check);
      errmap2.unique(&errmap_dup_check);
   }

   errs.clear();

   if (mode == 2)   // return count of warnings only  
   {
      for (emi = errmap2.begin(); emi != errmap2.end(); ++emi)
         if ((*emi)->get_type() == 1)  
            errs.append((*emi)->get_message(err_verbosity));
   }
   else if (mode == 1)   // return count of errs only 
   {
      for (emi = errmap2.begin(); emi != errmap2.end(); ++emi)
         if ((*emi)->get_type() == 0) 
            errs.append((*emi)->get_message(err_verbosity));
   }
   else                  // return count of errs and warnings 
   {
      for (emi = errmap2.begin(); emi != errmap2.end(); ++emi)
         if ((err_verbosity > 1) || ((*emi)->get_type() == 0))
         errs.append((*emi)->get_message(err_verbosity));
   }
   errmap2.clear();
   return errs; 
}

int ChipsCmdManager::get_err_cnt(int mode)
{
   int val = 0; 
   if  (mode == 2)        // return count of warnings only
      val = warn_cnt;
   else if (mode == 1)
      val = err_cnt;
   else
      val = err_cnt + warn_cnt; 
   return val; 
}

void ChipsCmdManager::clear_errs(int val)
{
   if (val == CLEAR_ALL_ERRS)
   {
      // remove all errors/warnings
      if (!errmap.empty())
      {
         list<ChipsError *>::iterator emi;
         for (emi = errmap.begin(); emi != errmap.end(); ++emi)
            delete (*emi);
         errmap.clear(); 
      }
      err_cnt = warn_cnt = 0; 
   }
   else if ((val > 0) && (val < ERR_MAX_NUM)) 
   {
      // remove specific errs
      if (!errmap.empty())
      {
         list<ChipsError *>::iterator emi;
         emi = errmap.begin();
         while (emi != errmap.end())
         {
            if ((*emi)->get_id() == val)
            {
               if ((*emi)->get_type() == 0)      err_cnt--;
               else if ((*emi)->get_type() == 1) warn_cnt--;
               delete (*emi);
               emi = errmap.erase(emi);
            }
            else  ++emi;
         }
      }
   }
   else if ((val == CLEAR_DNE_ERRS))
   {
      // remove object doen't exist type errs
      if (!errmap.empty())
      {
         list<ChipsError *>::iterator emi;
         emi = errmap.begin();
         while (emi != errmap.end())
         {
            if (err_type_object_DNE((*emi)->get_id()))
            {
               if ((*emi)->get_type() == 0)      err_cnt--;
               else if ((*emi)->get_type() == 1) warn_cnt--;
               delete (*emi);
               emi = errmap.erase(emi);
            }
            else  ++emi; 
         }
      }
   }
}


void ChipsCmdManager::add_frame_id_to_errs(string id)
{
   if (!errmap.empty())
   {
      list<ChipsError *>::iterator emi;
      for (emi = errmap.begin(); emi != errmap.end(); ++emi)
         (*emi)->set_frame(id);
   }
}

void ChipsCmdManager::add_plot_id_to_errs(string id)
{
   if (!errmap.empty())
   {
      list<ChipsError *>::iterator emi;
      for (emi = errmap.begin(); emi != errmap.end(); ++emi)
         (*emi)->set_plot(id);
   }
}

void ChipsCmdManager::add_window_id_to_errs(string id)
{
   if (!errmap.empty())
   {
      list<ChipsError *>::iterator emi;
      for (emi = errmap.begin(); emi != errmap.end(); ++emi)
         (*emi)->set_window(id);
   }
}

void ChipsCmdManager::add_object_id_to_errs(string id)
{
   if (!errmap.empty())
   {
      list<ChipsError *>::iterator emi;
      for (emi = errmap.begin(); emi != errmap.end(); ++emi)
         (*emi)->set_object(id);
   }
}

void ChipsCmdManager::set_default_err_type(int type)
{
   // if value is 1, issue warning otherwise issue error
   def_errtype = (type == 1);
}


int ChipsCmdManager::get_err_types()
{
   // returns the types of messages stored in the error list
   // 0=none 1=errs only  2=warnings only    3=errs and warnings
   int val = 0;
   if (err_cnt)  val |= 0x01;
   if (warn_cnt) val |= 0x02;
   return val; 
}

bool ChipsCmdManager::check_mask(unsigned long mask)
{
   return (state_mask & mask);
}

void ChipsCmdManager::update_mask(unsigned long mask)
{
   // calling routines can set ignore_state to avoid
   // cyclic issues with using state_mask  
   if (!ignore_state)  state_mask |= mask;
}

void ChipsCmdManager::reset_mask(unsigned long mask)
{
   if (!ignore_state)
      reset_state_mask |= mask;
}

void ChipsCmdManager::clear_mask()
{
   if (!ignore_state)  state_mask = STATE_CHNG_CLEAR;
}

void ChipsCmdManager::clear_reset_mask()
{
   if (!ignore_state)  reset_state_mask = STATE_CHNG_CLEAR;
}


void ChipsCmdManager::apply_resets ()
{
   if (!ignore_state)
   {
      state_mask &= ~reset_state_mask;
      clear_reset_mask ();
   }
}






