// 
//  Copyright (C) 2006,2007,2011-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSCMDMANAGER_HH
#define CHIPSCMDMANAGER_HH

#include "ChipsCommand.hh"
#include "ChipsEngine.hh"
#include "ChipsStateStream.hh"

#include <stdarg.h>

class ChipsError;

#define   STATE_CHNG_CLEAR       0x00000000
#define   STATE_CHNG_XLIMS       0x00000001 
#define   STATE_CHNG_YLIMS       0x00000002
#define   STATE_CHNG_COORD       0x00000004
#define   STATE_CHNG_LIMS        (STATE_CHNG_XLIMS|STATE_CHNG_YLIMS)
#define   STATE_CHNG_AXS_FMT     0x00000008
#define   STATE_CHNG_ADD_PLT     0x00000010
#define   STATE_CHNG_ADD_XAXS    0x00000020
#define   STATE_CHNG_ADD_YAXS    0x00000040
#define   STATE_CHNG_ADD_IMG     0x00000080
#define   STATE_CHNG_ADD_CRV     0x00000100
#define   STATE_CHNG_ADD_CNTR    0x00000200
#define   STATE_CHNG_ADD_HSTG    0x00000400
#define   STATE_CHNG_ADD_LINE    0x00000800
#define   STATE_CHNG_ADD_LBL     0x00001000
#define   STATE_CHNG_ADD_PNT     0x00002000
#define   STATE_CHNG_ADD_RGN     0x00004000
#define   STATE_CHNG_ADD_CLBAR   0x00008000
#define   STATE_CHNG_DATA_OBJ    (STATE_CHNG_ADD_CRV|STATE_CHNG_ADD_CNTR| \
                                  STATE_CHNG_ADD_HSTG|STATE_CHNG_ADD_IMG)
#define   STATE_CHNG_PROP_PLT    0x00010000
#define   STATE_CHNG_PROP_AXS   0x00020000
#define   STATE_CHNG_PROP_IMG    0x00080000
#define   STATE_CHNG_PROP_CRV    0x00100000
#define   STATE_CHNG_PROP_CNTR   0x00200000
#define   STATE_CHNG_PROP_HSTG   0x00400000
#define   STATE_CHNG_PROP_LINE   0x00800000
#define   STATE_CHNG_PROP_LBL    0x01000000
#define   STATE_CHNG_PROP_PNT    0x02000000
#define   STATE_CHNG_PROP_RGN    0x04000000
#define   STATE_CHNG_PROP_CLBAR  0x08000000
#define   STATE_CHNG_PROP_FRM    0x10000000
#define   STATE_CHNG_PROP_WIN    0x20000000
#define   STATE_CHNG_CURRENCY    0x40000000
#define   STATE_CHNG_FONTS       0x80000000
#define   STATE_CHNG_PROP_DEF    (STATE_CHNG_PROP_PLT|STATE_CHNG_PROP_AXS| \
                                  STATE_CHNG_PROP_IMG|STATE_CHNG_PROP_CRV| \
                                  STATE_CHNG_PROP_CNTR|STATE_CHNG_PROP_HSTG| \
                                  STATE_CHNG_PROP_LBL|STATE_CHNG_PROP_PNT| \
                                  STATE_CHNG_PROP_RGN|STATE_CHNG_PROP_CLBAR| \
                                  STATE_CHNG_PROP_FRM|STATE_CHNG_PROP_WIN| \
                                  STATE_CHNG_PROP_LINE)
#define   STATE_CHNG_ALL         0xFFFFFFFF

typedef struct 
{
   string tag;
   int    ref;
   list<pair <string, int> >::iterator iter;
} TAG_ELEM_T, *TAG_ELEM_P_T;


class ChipsCmdManager
{
   public:
      ChipsCmdManager (ChipsEngine *eng);
      ~ChipsCmdManager ();
      int  add_cmd (ChipsCommand *cmd);
      int  set_autocommit(bool val=true)  { autocommit = val; return ERR_STATUS_GOOD;}
      bool get_autocommit()  { return autocommit; }
      int  clear_hold_buffer(); 
      int  commit_hold_buffer(); 
      void set_cmd_lock (bool flag) {lock = flag;};
      bool cmd_lock () {return lock;};
      void print ();

      int  open_block();
      int  close_block();
      bool in_block()           { return (blockref > 0); } 
      bool get_block_depth()    { return blockref; } 
      int  execute_cmd (int cnt=1);
      int  execute_cmd (string usetag);
      int  unexecute_cmd (int cnt=1);
      int  unexecute_cmd (string usetag);

      int  serialize(ChipsStateStream& out);
      int  deserialize(ChipsEngine* eng, ChipsStateStream& in);
      int  commands_to_file ();

      int  add_tag(string val);
      int  remove_tag(string val); 
      int  remove_tag_by_iterator(list<pair <string, int> >::iterator val, int tag_ref = -1);
      int  replace_tag_iterator(list<pair <string, int> >::iterator val, list<pair <string, int> >::iterator new_val);

      void clear_cmds (bool force_del=true);
      int  tot_cmd_cnt (list<ChipsCommand *>::iterator cmds);
      void get_undo_counts(int &undo, int &redo);

      // error handling 
      void   add_err(string txt, string oid="", int typ=-1);
      void   add_err(string txt, int errcode, string oid="", int typ=-1);
      void   add_err(int errcode, string oid="", int typ=-1, ...);
      void   add_frame_id_to_errs(string id);
      void   add_plot_id_to_errs(string id);
      void   add_object_id_to_errs(string id);
      void   add_window_id_to_errs(string id);
      void   remove_errs_from_tagged(); 
      void   remove_last_err(); 
      void   set_default_err_type(int type=0);
      string get_errs(int mode=0);     // 0=all  1=errs only  2=warnings only
      int    get_err_cnt(int mode=0);  // 0=all  1=errs only  2=warnings only
      int    get_err_types();          // 0=none 1=errs only  2=warnings only    3=errs and warnings 
      void   clear_errs(int val=CLEAR_ALL_ERRS);  
      void   set_err_verbosity(int val)  { if (val > -1) err_verbosity = val; }
      int    get_err_verbosity()         { return err_verbosity; }

      bool   get_ignore_state()  { return ignore_state; } 
      void   set_ignore_state(bool val)  { ignore_state = val; }
      void   update_mask(unsigned long mask);
      void   reset_mask(unsigned long mask=STATE_CHNG_ALL);
      void   apply_resets ();
      unsigned long get_mask()   { return state_mask; }
      void   clear_mask();
      void   clear_reset_mask ();
      bool   check_mask(unsigned long mask); 

   protected:
      list <ChipsCommand *> commands;
      list <ChipsCommand *>::iterator curr_cmd;
      list <pair<string, int> >undo_states;
      list <pair<string, int> >::iterator curr_state;
      list <TAG_ELEM_P_T> tags; 

      TAG_ELEM_P_T find_tag(string tag); 
      int  print_tags();

      bool lock;
      int tot_cmd_weight;
      ChipsEngine *cm_eng;

  private:
      static bool errmap_sort_check (list<ChipsError*>::const_reference first,
                                     list<ChipsError*>::const_reference second);
      static bool errmap_dup_check (list<ChipsError*>::const_reference first,
                                    list<ChipsError*>::const_reference second);

      bool   compound; 
      bool   autocommit;  // should commands be added to stack or put on hold
      bool   save_ref;    // save the ref for the first command in a block
      int    blockref;
      int    ref;
      list <ChipsCommand *> hold; 
      list <int> block_ref_cnts;  // ref of first command in each open block

      list <ChipsError*> errmap;    // current error messages 
      ChipsError* tagged_err; // last error inserted
      int         def_errtype;      // default issue 0=error 1=warning
      int         err_verbosity;    // error verbosity level
      int         err_cnt;          // current number of errs
      int         warn_cnt;         // current number of warnings
      unsigned long state_mask;     // mask to flag state changes 
      bool        ignore_state;     // when true, don't update state mask
      int         reset_state_mask;  // a list of masks to be reset
};

#endif
