// 
//  Copyright (C) 2006,2007,2012-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsCommand.hh"
#include "ChipsErrs.hh"
#include "ChipsEngine.hh"
#include "ChipsWindow.hh"
#include "ChipsFrame.hh"
#include "ChipsLine.hh"
#include "ChipsPoint.hh"
#include "ChipsRegion.hh"
#include "ChipsColorBar.hh"
#include "ChipsStateStream.hh"
#include "ChipsPlot.hh"

void ChipsCommand::serialize_object(ChipsStateStream& out)
{
   int val;
   int obj_ref = 0;
   int number_of_parents = 0;
   string myid;
   V_OBJECT_TYPE obtype; 
   ChipsObject* ob_p;
   obtype = co->get_object_type();
   val = (int)obtype;
   out.write_int(&val);
   obj_ref = static_cast <int> (co->get_ref_cnt ());
   out.write_int (&obj_ref);

   myid = co->get_id();
   out.write_string(&myid);
   switch(obtype)
   {
      case chips_window:
         number_of_parents = 0;
      break; 
      case chips_frame:
         number_of_parents = 1;
      break; 
      case chips_plot:
         number_of_parents = 2;
      break; 
      case chips_label:        // fallthrough intended
      case chips_line:         // fallthrough intended
      case chips_curve:        // fallthrough intended
      case chips_image:        // fallthrough intended
      case chips_contour:      // fallthrough intended
      case chips_histogram:    // fallthrough intended
      case chips_point:        // fallthrough intended 
      case chips_region:       // fallthrough intended 
      case chips_colorbar:     // fallthrough intended 
      case chips_axis:
         number_of_parents = 3;
      break; 
      default:
      break; 
   }

   ob_p = co->get_parent ();
   for (int ii=0; ii<number_of_parents; ii++)
   {
      if (ob_p)
      {
         myid = ob_p->get_id();
         obj_ref = (int)(ob_p->get_ref_cnt ());
         ob_p = ob_p->get_parent();
      }
      else
      {
         myid = "\0";
         obj_ref = 0;
         ob_p = NULL;
      }
      out.write_string(&myid);
      out.write_int (&obj_ref);
   }

}

ChipsCreateCommand::ChipsCreateCommand(ChipsObject* cob, bool chng_cnt, const string &curr)
{
   co = cob;
   weight = 1;
   current = curr; 
   change_cnt = chng_cnt;
}

ChipsCreateCommand::ChipsCreateCommand(ChipsObject* cob, 
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}


ChipsCreateCommand::ChipsCreateCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight = 1;
}

ChipsCreateCommand::~ChipsCreateCommand()
{
}


int ChipsCreateCommand::execute () 
{
   if (!co)
      return ERR_STATUS_BAD;
   if (co->get_parent ())
   {
      return co->get_parent ()->reference_object(co, change_cnt);
   }
   else
   {
      return co->reference_object(co, change_cnt);
   }
}

int ChipsCreateCommand::unexecute ()
{
   if (!co)
      return ERR_STATUS_BAD;
   if (co->get_parent ())
   {
      return co->get_parent ()->dereference_object(co, change_cnt, current);
   }
   else
   {
      return co->dereference_object(co, change_cnt, current);
   }
}

int ChipsCreateCommand::clear ()
{
   if (!co)
      return ERR_STATUS_BAD;

   if (co->get_parent ())
      co->get_parent ()->clear_cmd (co);
   else
      co->clear_cmd (co);
   co = NULL;
   return ERR_STATUS_BAD;
}

void ChipsCreateCommand::print () 
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Create " << v_object_to_string (co->get_object_type ()) << endl;
};


void ChipsCreateCommand::serialize(ChipsStateStream& out)
{
   int val = 1;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_string(&current);
   out.write_bool(&change_cnt);
}


void ChipsCreateCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_string(&current);
   in.read_bool(&change_cnt);
}


ChipsSetCommand::ChipsSetCommand(ChipsObject* cob, ChipsStateStream& in)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   deserialize(in);
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsStateStream& in)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = NULL;
   deserialize(in);
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property, string oldv, 
                                 string newv, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   oldval = oldv;
   newval = newv;  
   current = oldcurr;
   type = 0; // string type
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property, bool oldv,
                                 bool newv, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   oval_b = oldv;
   nval_b = newv;
   current = oldcurr;
   type = 1; // bool type
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property, int oldv,
                                 int newv, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   oval_i = oldv;
   nval_i = newv;
   current = oldcurr;
   type = 2; // integer type
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property, double oldv,
                                 double newv, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   oval_d = oldv;
   nval_d = newv;
   current = oldcurr;
   type = 3; // double type
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property, float oldv,
                                 float newv, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   oval_f = oldv;
   nval_f = newv;
   current = oldcurr;
   type = 4; // float type
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property,
   unsigned char* oldv,  unsigned char *newv, int asize, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   if ((sz = asize) < 1)  sz = 1; 
   {
      oval_uc = new unsigned char[sz];
      nval_uc = new unsigned char[sz];
      if (oldv)  memcpy(oval_uc, oldv, sz);  
         else    memset(oval_uc, 0, sz); 
      if (newv)  memcpy(nval_uc, newv, sz);  
         else    memset(nval_uc, 0, sz); 
   }
   current = oldcurr;
   type = 5;  // unsigned char array
   weight = 1;
}

ChipsSetCommand::ChipsSetCommand(ChipsObject *cob, int property,
   double* oldv,  double* newv, int asize, string oldcurr)
{
   oval_uc = nval_uc = NULL; 
   oval_da = nval_da = NULL; 
   co = cob;
   prop = property;
   if ((sz = asize) < 1)  sz = 1; 
   {
      oval_da = new double[sz];
      nval_da = new double[sz];
      for (int cc = sz; cc--; )
      {  
         if (oldv)  oval_da[cc] = oldv[cc];
	    else    oval_da[cc] = 0;
         if (newv)  nval_da[cc] = newv[cc];
	    else    nval_da[cc] = 0;
      }
   }
   current = oldcurr;
   type = 6;  // unsigned char array
   weight = 1;
}


ChipsSetCommand::~ChipsSetCommand()
{
   if (nval_uc)   delete[] nval_uc; 
   nval_uc = NULL; 
   if (oval_uc)   delete[] oval_uc; 
   oval_uc = NULL; 
   if (oval_da)   delete[] oval_da; 
   oval_da = NULL; 
   if (nval_da)   delete[] nval_da; 
   nval_da = NULL; 
}


int ChipsSetCommand::execute ()
{
   int rval; 
   if (co)
   {
      switch(co->get_object_type())
      {
         case chips_axis:
	 {
	    ChipsAxis* ca = (ChipsAxis*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ca->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ca->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = ca->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = ca->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = ca->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = ca->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6:
                  rval = ca->set_property(static_cast<void*>(nval_da), prop,0);
	       break; 
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
               string tid = ca->get_id();
               if ((tid != "bx1") && (tid != "bx2") && (tid != "by1") && (tid != "by2")) 
               {
	          ChipsId my_cid;
	          if (ca->get_orientation() == X_AXIS)
	          {
	             my_cid.set_xaxis_id(ca->get_id()); 
	             ((ChipsPlot*)(co->get_parent()))->set_current_xaxis(&my_cid);
	          }
	          else
	          {
	             my_cid.set_yaxis_id(ca->get_id()); 
	             ((ChipsPlot*)(co->get_parent()))->set_current_yaxis(&my_cid);
	          }
               }
               ((ChipsPlot*)(co->get_parent()))->update_property (ca->get_id (), prop);
	    }
	 }
	 break; 
         case chips_curve:
	 {
	    ChipsCurve* cv = (ChipsCurve*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cv->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cv->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cv->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cv->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cv->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cv->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cv->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_curve_id(cv->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_curve(&my_cid, false);
	    }
	 }
	 break; 
         case chips_image:
	 {
	    ChipsImage* ci = (ChipsImage*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ci->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ci->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = ci->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = ci->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = ci->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = ci->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ci->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_image_id(ci->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_image(&my_cid, false);
	    }
	 }
	 break; 
         case chips_colorbar:
         {
            ChipsColorBar* cb = (ChipsColorBar*)co;
            switch(type)
            {
               case 0:
                  rval = cb->set_property((void*)newval.c_str(), prop, 0);
               break;
               case 1:
                  rval = cb->set_property(static_cast<void*>(&nval_b), prop,0);
               break;
               case 2:
                  rval = cb->set_property(static_cast<void*>(&nval_i), prop,0);
               break;
               case 3:
                  rval = cb->set_property(static_cast<void*>(&nval_d), prop,0);
               break;
               case 4:
                  rval = cb->set_property(static_cast<void*>(&nval_f), prop,0);
               break;
               case 5:
                  rval = cb->set_property(static_cast<void*>(nval_uc), prop,0);
               break;
               case 6:
                  rval = cb->set_property(static_cast<void*>(nval_da), prop,0);
               break;
               default:
                  rval = ERR_STATUS_BAD;
               break;
            }
            if (rval == ERR_STATUS_GOOD)
            {
               ChipsId my_cid;
               my_cid.set_colorbar_id(cb->get_id());
               ((ChipsPlot*)(co->get_parent()))->set_current_colorbar(&my_cid, false);
            }
         }
         break;
         case chips_contour: 
	 {
	    ChipsContour* ct = (ChipsContour*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ct->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ct->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = ct->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = ct->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = ct->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = ct->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ct->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_contour_id(ct->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_contour(&my_cid, false);
	    }
	 }
	 break; 
         case chips_histogram: 
	 {
	    ChipsHistogram* ch = (ChipsHistogram*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ch->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ch->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = ch->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = ch->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = ch->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = ch->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ch->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_histogram_id(ch->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_histogram(&my_cid, false);
	    }
	 }
	 break;
         case chips_label:
	 {
	    ChipsLabel* cl = (ChipsLabel*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cl->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cl->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cl->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cl->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cl->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cl->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cl->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_label_id(cl->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_label(&my_cid, false);
	    }
	 }
	 break; 
         case chips_line:
	 {
	    ChipsLine* cl = (ChipsLine*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cl->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cl->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cl->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cl->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cl->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cl->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cl->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_line_id(cl->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_line(&my_cid,false);
	    }
	 }
	 break; 
         case chips_point:
	 {
	    ChipsPoint* cp = (ChipsPoint*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cp->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cp->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cp->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cp->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cp->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cp->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cp->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_point_id(cp->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_point(&my_cid,false);
	    }
	 }
         break; 
	 case chips_region:
	 {
	    ChipsRegion* cr = (ChipsRegion*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cr->set_property((void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cr->set_property(static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cr->set_property(static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cr->set_property(static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cr->set_property(static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cr->set_property(static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cr->set_property(static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_region_id(cr->get_id()); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_region(&my_cid,false);
	    }
	 }
	 break; 
         case chips_legend:
         break; 
         case chips_plot:
	 {
	    ChipsPlot* cp = (ChipsPlot*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cp->set_property(NULL, (void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cp->set_property(NULL, static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cp->set_property(NULL, static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cp->set_property(NULL, static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cp->set_property(NULL, static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cp->set_property(NULL,static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cp->set_property(NULL,static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       string my_id = ((ChipsPlot*)co)->get_id(); 
	       ((ChipsFrame*)(co->get_parent()))->set_current_plot(my_id);
	    }
	 }
         break; 
         case chips_frame:
	 {
	    ChipsFrame* cf = (ChipsFrame*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cf->set_property(NULL, (void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cf->set_property(NULL, static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cf->set_property(NULL, static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cf->set_property(NULL, static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cf->set_property(NULL, static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cf->set_property(NULL, static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cf->set_property(NULL, static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
	    if (rval == ERR_STATUS_GOOD) ((ChipsWindow*)(co->get_parent()))->redraw();
	 }
	 break; 
         case chips_window:
	 {
	    ChipsWindow* cw = (ChipsWindow*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cw->set_property(NULL, (void*)newval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cw->set_property(NULL, static_cast<void*>(&nval_b), prop,0);
	       break;
	       case 2: 
                  rval = cw->set_property(NULL, static_cast<void*>(&nval_i), prop,0);
	       break;
	       case 3: 
                  rval = cw->set_property(NULL, static_cast<void*>(&nval_d), prop,0);
	       break;
	       case 4: 
                  rval = cw->set_property(NULL, static_cast<void*>(&nval_f), prop,0);
	       break;
	       case 5: 
                  rval = cw->set_property(NULL, static_cast<void*>(nval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cw->set_property(NULL, static_cast<void*>(nval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_window_id(cw->get_id()); 
	       ((ChipsEngine*)(cw->get_engine()))->set_current(&my_cid, chips_window, false);
	    }
	 }
	 break; 
         default:
            rval = ERR_STATUS_BAD;
         break; 
      }
   }
   else   rval = ERR_STATUS_BAD;
   return rval;
}

int ChipsSetCommand::unexecute ()
{
   int rval; 
   if (co)
   {
      switch(co->get_object_type())
      {
         case chips_axis:
	 {
	    ChipsAxis* ca = (ChipsAxis*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ca->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ca->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = ca->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = ca->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = ca->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = ca->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ca->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       if (ca->get_orientation() == X_AXIS)
	       {
	          my_cid.set_xaxis_id(current); 
	          ((ChipsPlot*)(co->get_parent()))->set_current_xaxis(&my_cid);
	       }
	       else
	       {
	          my_cid.set_yaxis_id(current); 
	          ((ChipsPlot*)(co->get_parent()))->set_current_yaxis(&my_cid);
	       }
               ((ChipsPlot*)(co->get_parent()))->update_property (ca->get_id (), prop);
	    }
	 }
         break;
         case chips_curve:
	 {
	    ChipsCurve* cv = (ChipsCurve*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cv->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cv->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cv->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cv->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cv->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cv->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cv->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_curve_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_curve(&my_cid, false);
	    }
	 }
	 break; 
         case chips_image:
	 {
	    ChipsImage* ci = (ChipsImage*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ci->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ci->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = ci->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = ci->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = ci->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = ci->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ci->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_image_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_image(&my_cid, false);
	    }
	 }
	 break; 
         case chips_colorbar:
         {
            ChipsColorBar* cb = (ChipsColorBar*)co;
            switch(type)
            {
               case 0:
                  rval = cb->set_property((void*)oldval.c_str(), prop, 0);
               break;
               case 1:
                  rval = cb->set_property(static_cast<void*>(&oval_b), prop,0);
               break;
               case 2:
                  rval = cb->set_property(static_cast<void*>(&oval_i), prop,0);
               break;
               case 3:
                  rval = cb->set_property(static_cast<void*>(&oval_d), prop,0);
               break;
               case 4:
                  rval = cb->set_property(static_cast<void*>(&oval_f), prop,0);
               break;
               case 5:
                  rval = cb->set_property(static_cast<void*>(oval_uc), prop,0);
               break;
               case 6:
                  rval = cb->set_property(static_cast<void*>(oval_da), prop,0);
               break;
               default:
                  rval = ERR_STATUS_BAD;
               break;
            }
            if (rval == ERR_STATUS_GOOD)
            {
               ChipsId my_cid;
               my_cid.set_colorbar_id(current);
               ((ChipsPlot*)(co->get_parent()))->set_current_colorbar(&my_cid, false);
            }
         }
         break;
         case chips_contour: 
	 {
	    ChipsContour* ct = (ChipsContour*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ct->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ct->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = ct->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = ct->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = ct->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = ct->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ct->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_contour_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_contour(&my_cid, false);
	    }
	 }
	 break; 
         case chips_histogram: 
	 {
	    ChipsHistogram* ch = (ChipsHistogram*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = ch->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = ch->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = ch->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = ch->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = ch->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = ch->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = ch->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_histogram_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_histogram(&my_cid, false);
	    }
	 }
	 break; 
         case chips_label:
	 {
	    ChipsLabel* cl = (ChipsLabel*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cl->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cl->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cl->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cl->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cl->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cl->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cl->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_label_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_label(&my_cid, false);
	    }
	 }
         break; 
         case chips_legend:
	 break; 
         case chips_line:
	 {
	    ChipsLine* cl = (ChipsLine*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cl->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cl->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cl->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cl->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cl->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cl->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cl->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_line_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_line(&my_cid,false);
	    }
	 }
	 break; 
         case chips_point:
	 {
	    ChipsPoint* cp = (ChipsPoint*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cp->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cp->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cp->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cp->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cp->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cp->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cp->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_point_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_point(&my_cid, false);
	    }
	 }
         break; 
	 case chips_region:
	 {
	    ChipsRegion* cr = (ChipsRegion*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cr->set_property((void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cr->set_property(static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cr->set_property(static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cr->set_property(static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cr->set_property(static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cr->set_property(static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cr->set_property(static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_region_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_region(&my_cid, false);
	    }
	 }
	 break; 
         case chips_window:
	 {
	    ChipsWindow* cw = (ChipsWindow*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cw->set_property(NULL, (void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cw->set_property(NULL, static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cw->set_property(NULL, static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cw->set_property(NULL, static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cw->set_property(NULL, static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cw->set_property(NULL, static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cw->set_property(NULL, static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	    {
	       ChipsId my_cid;
	       my_cid.set_window_id(current); 
	       ((ChipsEngine*)(cw->get_engine()))->set_current(&my_cid, chips_window, false);
	    }
	 }
	 break; 
         case chips_plot:
	 {
	    ChipsPlot* cp = (ChipsPlot*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cp->set_property(NULL, (void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cp->set_property(NULL, static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cp->set_property(NULL, static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cp->set_property(NULL, static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cp->set_property(NULL, static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cp->set_property(NULL, static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cp->set_property(NULL, static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
	       ((ChipsFrame*)(co->get_parent()))->set_current_plot(current);
	 }
         break; 
         case chips_frame:
	 {
	    ChipsFrame* cf = (ChipsFrame*)co; 
	    switch(type)
	    {
	       case 0: 
                  rval = cf->set_property(NULL, (void*)oldval.c_str(), prop, 0);
	       break; 
	       case 1: 
                  rval = cf->set_property(NULL, static_cast<void*>(&oval_b), prop,0);
	       break;
	       case 2: 
                  rval = cf->set_property(NULL, static_cast<void*>(&oval_i), prop,0);
	       break;
	       case 3: 
                  rval = cf->set_property(NULL, static_cast<void*>(&oval_d), prop,0);
	       break;
	       case 4: 
                  rval = cf->set_property(NULL, static_cast<void*>(&oval_f), prop,0);
	       break;
	       case 5: 
                  rval = cf->set_property(NULL, static_cast<void*>(oval_uc), prop,0);
	       break;
	       case 6: 
                  rval = cf->set_property(NULL, static_cast<void*>(oval_da), prop,0);
	       break;
               default:
                  rval = ERR_STATUS_BAD; 
	       break;
	    }
            if (rval == ERR_STATUS_GOOD) 
            {
               ((ChipsWindow*)(co->get_parent()))->redraw();
	       ((ChipsWindow*)(co->get_parent()))->set_current(current);
            }
	 }
         break; 
         default:
            rval = ERR_STATUS_BAD;
         break; 
      }
   }
   else   rval = ERR_STATUS_BAD;
   return rval;
}

void ChipsSetCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Set " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsSetCommand::serialize(ChipsStateStream& out)
{
   int val = 3;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref); 
   out.write_int(&block_ref);

   out.write_int(&prop); 
   out.write_int(&type); 
   switch(type)
   {
      case 0: //string
	 out.write_string(&oldval);
	 out.write_string(&newval);
      break;
      case 1: // bool
         out.write_bool(&oval_b);
         out.write_bool(&nval_b);
      break;
      case 2: // int
         out.write_int(&oval_i);
         out.write_int(&nval_i);
      break;
      case 3: // double
         out.write_double(&oval_d);
         out.write_double(&nval_d);
      break;
      case 4: // float
         out.write_float(&oval_f);
         out.write_float(&nval_f);
      break;
      case 5: // unsigned char array 
	 out.write_int(&sz);
         out.write_char((char*)&oval_uc[0], sz);
         out.write_char((char*)&nval_uc[0], sz);
      break;
      case 6: // double array 
         out.write_int(&sz);
	 out.write_double(oval_da, sz); 
	 out.write_double(nval_da, sz); 
      break; 
      default:
      break; 
   }
   out.write_string(&current); 
}

void ChipsSetCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_int(&prop);
   in.read_int(&type);
   switch(type)
   {
      case 0: 
      {
	 in.read_string(&oldval);
	 in.read_string(&newval);
      }
      break; 
      case 1:
         in.read_bool(&oval_b);
         in.read_bool(&nval_b);
      break;
      case 2:
         in.read_int(&oval_i);
         in.read_int(&nval_i);
      break;
      case 3:
         in.read_double(&oval_d);
         in.read_double(&nval_d);
      break;
      case 4:
         in.read_float(&oval_f);
         in.read_float(&nval_f);
      break;
      case 5: // unsigned char array 
      {
	 in.read_int(&sz);
	 if (sz > 0)
	 {
            oval_uc = new unsigned char[sz];
            nval_uc = new unsigned char[sz];
            in.read_char((char*)&oval_uc[0], sz);
            in.read_char((char*)&nval_uc[0], sz);
         }
      }
      break; 
      case 6: // double array
      {
         in.read_int(&sz);
	 if (sz > 0)
	 {
	    oval_da = new double[sz];
	    nval_da = new double[sz];
	    in.read_double(oval_da, sz); 
	    in.read_double(nval_da, sz); 
	 }
      }
      break; 
      default:
      break; 
   }
   in.read_string(&current); 
}

ChipsLimitsCommand::ChipsLimitsCommand(ChipsObject* cob, 
   int flag, double olin, double olog, double nval)
{
   co = cob;
   o_logmin = o_logmax = o_linmin = o_linmax = n_min = n_max = 0.0;
   o_autoflag = n_autoflag = 0; 
   weight=1;
   transformed = true;
   switch ((lim_type = flag))
   {
      case 1:
      case 2:
      case 4:
         o_logmin = olog;
         o_linmin = olin;
         n_min    = nval;
      break; 
      case 8:
         o_logmax = olog;
         o_linmax = olin;
         n_max    = nval;
      break; 
      default: 
      break; 
   }
}

ChipsLimitsCommand::ChipsLimitsCommand(ChipsObject* cob,
   int flag, double olinmin, double ologmin, double olinmax,
   double ologmax, int oflag, int nflag, double nmin, double nmax,
   bool transformed)
{
   co = cob;
   weight=1;
   this->transformed = transformed;
   switch ((lim_type = flag))
   {
      case 16:
      case 32:
         o_logmin   = ologmin;
         o_linmin   = olinmin;
         o_logmax   = ologmax;
         o_linmax   = olinmax;
         n_min      = nmin;
         n_max      = nmax;
         o_autoflag = oflag; 
         n_autoflag = nflag; 
      break;
      default: 
      break;
   }
}

ChipsLimitsCommand::ChipsLimitsCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}


ChipsLimitsCommand::ChipsLimitsCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight=1;
}


ChipsLimitsCommand::~ChipsLimitsCommand()
{
}


int ChipsLimitsCommand::execute()
{
   ChipsAxis* ca = (ChipsAxis*)co;
   int rval = ERR_STATUS_BAD; 
   if (ca)
   {
      switch (lim_type)
      {
         case 1:
            rval = ca->adjust_autologmin(n_min, 0);
         break; 
         case 2:
            rval = ca->adjust_logmin(n_min, 0);
         break; 
         case 4:
            rval = ca->adjust_min(n_min, 0);
         break; 
         case 8:
            rval = ca->adjust_max(n_max, 0);
         break;
         case 16:
            rval = ca->adjust_minmax(n_min, n_max, n_autoflag, 0, transformed);
         break;
         case 32:
            rval = ca->adjust_autominmax(n_min, n_max, 0);
         break;
         default: 
         break;
      }
      ((ChipsPlot *)(ca->get_parent ()))->update_limits(ca->get_id ());

   }
   return rval;
}

int ChipsLimitsCommand::unexecute ()
{
   ChipsAxis* ca = (ChipsAxis*)co;
   int rval = ERR_STATUS_BAD; 
   if (ca)
   {
      switch (lim_type)
      {
         case 1:
            rval = ca->adjust_autologmin(o_logmin, 0);
         break; 
         case 2:
            rval = ca->adjust_logmin(o_logmin, 0);
         break; 
         case 4:
            rval = ca->adjust_min(o_linmin, 0);
         break; 
         case 8:
            rval = ca->adjust_max(o_linmax, 0);
         break;
         case 32:
	    // since automin/max doesn't work if our undo range is 
	    // a subset of the current range use the minmax case
         case 16:
            if (ca->get_scaletype() == V_SCALE_LINEAR)
               rval = ca->adjust_minmax(o_linmin, o_linmax, o_autoflag, 0, transformed);
            else
	       rval = ca->adjust_minmax(o_logmin, o_logmax, o_autoflag, 0, transformed);
         break;
         default: 
         break;
      }
      ((ChipsPlot *)(ca->get_parent ()))->update_limits(ca->get_id ());

   }
   return rval;
}

void ChipsLimitsCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Limits " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsLimitsCommand::serialize(ChipsStateStream& out)
{
   int val = 4;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_int(&lim_type);
   out.write_int(&o_autoflag);
   out.write_int(&n_autoflag);
   out.write_double(&o_logmin); 
   out.write_double(&o_linmin); 
   out.write_double(&o_logmax); 
   out.write_double(&o_linmax); 
   out.write_double(&n_min); 
   out.write_double(&n_max); 
   out.write_bool(&transformed);
}

void ChipsLimitsCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_int(&lim_type);
   in.read_int(&o_autoflag);
   in.read_int(&n_autoflag);
   in.read_double(&o_logmin); 
   in.read_double(&o_linmin); 
   in.read_double(&o_logmax); 
   in.read_double(&o_linmax); 
   in.read_double(&n_min); 
   in.read_double(&n_max); 
   if (in.get_minor_version () < 8)
      transformed = true;
   else
      in.read_bool (&transformed);
}



ChipsMoveCommand::ChipsMoveCommand(ChipsObject* cob,  double ox, double oy,
   double nx, double ny, COORD_SYS ncs, string ncstr, int nm, string oldcurr)
{
   co = cob;
   weight=1;
   o_x = ox;
   o_y = oy;
   n_x = nx;
   n_y = ny;
   n_mode = nm;
   n_coord = ncs;
   n_coordid = ncstr; 
   current = oldcurr;
}

ChipsMoveCommand::ChipsMoveCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}


ChipsMoveCommand::ChipsMoveCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight=1;
}


ChipsMoveCommand::~ChipsMoveCommand()
{
}


int ChipsMoveCommand::execute()
{
   int rval = ERR_STATUS_BAD; 
   if (co)
   {
      switch(co->get_object_type())
      {
         case chips_frame:
         {
            ChipsFrame* cf = (ChipsFrame*)co;
	    COORD_SYS cs = (COORD_SYS)n_coord; 
            rval = cf->move(cs, n_coordid, n_x, n_y, n_mode, false); 
	    if (rval == ERR_STATUS_GOOD) 
	       ((ChipsWindow*)(cf->get_parent()))->update_display();
         }
         break; 
         case chips_plot:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cp->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
         case chips_axis:
         {
            ChipsAxis* ca = (ChipsAxis*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = ca->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
         case chips_colorbar:
         {
            ChipsColorBar* cb = (ChipsColorBar*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cb->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
         case chips_label:
         {
            ChipsLabel* cl = (ChipsLabel*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cl->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
         case chips_line:
         {
            ChipsLine* cl = (ChipsLine*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cl->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
         case chips_point:
         {
            ChipsPoint* cp = (ChipsPoint*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cp->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
	 break; 
	 case chips_region:
         {
            ChipsRegion* cr = (ChipsRegion*)co;
            COORD_SYS cs = (COORD_SYS)n_coord;
            rval = cr->move(cs, n_coordid, n_x, n_y, n_mode, false);
         }
      	 break; 
         default: 
         break; 
      }
   }
   return rval;
}

int ChipsMoveCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      switch(co->get_object_type())
      {
         case chips_frame:
         {
            ChipsFrame* cf = (ChipsFrame*)co;
            rval = cf->move(COORD_NONE, "", o_x, o_y, V_MOVE_ABSOLUTE, false); 
	    if (rval == ERR_STATUS_GOOD) 
	    {
	       ((ChipsWindow*)(cf->get_parent()))->update_display();
               ((ChipsWindow*)(co->get_parent()))->set_current(current);
            } 
         }
         break;
         case chips_plot:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            rval = cp->move(COORD_NONE, "", o_x, o_y, V_MOVE_ABSOLUTE, false);
         }
         break; 
         case chips_axis:
         {
            ChipsAxis* ca = (ChipsAxis*)co;
            rval = ca->move(ca->get_coord_system (), ca->get_data_system_id (), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
            {
               ChipsId my_cid;
               if (ca->get_orientation() == X_AXIS)
               {
                  my_cid.set_xaxis_id(current);
                  ((ChipsPlot*)(co->get_parent()))->set_current_xaxis(&my_cid);
               }
               else
               {
                  my_cid.set_yaxis_id(current);
                  ((ChipsPlot*)(co->get_parent()))->set_current_yaxis(&my_cid);
               }
            }
         }
         break; 
         case chips_colorbar:
         {
            ChipsColorBar* cb = (ChipsColorBar*)co;
            rval = cb->move(cb->get_coord_system (), cb->get_data_system_id (), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
            {
               ChipsId my_cid;
               my_cid.set_colorbar_id(current);
               ((ChipsPlot*)(co->get_parent()))->set_current_colorbar(&my_cid, false);
            }
         }
         break;
         case chips_label:
         {
            ChipsLabel* cl = (ChipsLabel*)co;
            rval = cl->move(cl->get_coord_system (), cl->get_data_system_id(), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
	    {
	       ChipsId my_cid;
	       my_cid.set_label_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_label(&my_cid, false);
	    }
         }
         break; 
         case chips_line:
         {
            ChipsLine* cl = (ChipsLine*)co;
            rval = cl->move(cl->get_coord_system (), cl->get_data_system_id (), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
	    {
	       ChipsId my_cid;
	       my_cid.set_line_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_line(&my_cid,false);
	    }
         }
         break; 
         case chips_point:
         {
            ChipsPoint* cp = (ChipsPoint*)co;
            rval = cp->move(cp->get_coord_system (), cp->get_data_system_id (), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
	    {
	       ChipsId my_cid;
	       my_cid.set_point_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_point(&my_cid, false);
	    }
         }
         break; 
	 case chips_region:
         {
            ChipsRegion* cr = (ChipsRegion*)co;
            rval = cr->move(cr->get_coord_system (), cr->get_data_system_id (), o_x, o_y, V_MOVE_ABSOLUTE, false);
            if (rval == ERR_STATUS_GOOD)
	    {
	       ChipsId my_cid;
	       my_cid.set_region_id(current); 
	       ((ChipsPlot*)(co->get_parent()))->set_current_region(&my_cid,false);
	    }
         }
         break; 
         default:
         break;
      }
   }
   return rval;
}

void ChipsMoveCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Move " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsMoveCommand::serialize(ChipsStateStream& out)
{
   int val = 5;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_double(&o_x); 
   out.write_double(&o_y); 
   out.write_double(&n_x); 
   out.write_double(&n_y); 
   out.write_int(&n_mode); 
   out.write_int(&n_coord); 
   out.write_string(&n_coordid);
   out.write_string(&current);
}

void ChipsMoveCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_double(&o_x); 
   in.read_double(&o_y); 
   in.read_double(&n_x); 
   in.read_double(&n_y); 
   in.read_int(&n_mode); 
   in.read_int(&n_coord); 
   in.read_string(&n_coordid);
   in.read_string(&current);
}



ChipsTileCommand::ChipsTileCommand(ChipsObject *new_co, int ofmode, int otcols,
   int otrows, string of, int nfmode, int ntcols, int ntrows, string nf, 
   string ocw)
{
   co = new_co;
   weight=1;
   o_mode = ofmode;
   o_rows = otrows;
   o_cols = otcols;
   n_mode = nfmode;
   n_rows = ntrows;
   n_cols = ntcols;
   o_frame = of;
   n_frame = nf;
   currw  = ocw;
}

ChipsTileCommand::ChipsTileCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

ChipsTileCommand::~ChipsTileCommand()
{
}

int ChipsTileCommand::execute()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsWindow* cw = (ChipsWindow*)co;
      switch(n_mode)
      {
         case 0:  // placed
         {
            cw->set_col_row_vals(n_cols, n_rows);
            rval  =cw->set_framemode(0, false);
         }
         break;
         case 1:  // single
         {
            cw->set_col_row_vals(n_cols, n_rows);
            rval = cw->single_frame(n_frame, false);
         }
         break;
         case 2:  // tiled 
            rval = cw->tile(n_cols, n_rows, false); 
         break;
         default: 
         break;
      }
   }
   return rval;
}

int ChipsTileCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsWindow* cw = (ChipsWindow*)co;
      switch(o_mode)
      {
         case 0:  // placed
         { 
            cw->set_col_row_vals(o_cols, o_rows);
            rval  =cw->set_framemode(0, false);
         }
         break;
         case 1:  // single
         {
            cw->set_col_row_vals(o_cols, o_rows);
            rval = cw->single_frame(o_frame, false);
         }
         break;
         case 2:  // tiled 
            rval = cw->tile(o_cols, o_rows, false); 
            if (rval == ERR_STATUS_GOOD)
            {
               ChipsId my_cid;
               my_cid.set_window_id(currw);
               ((ChipsEngine*)(cw->get_engine()))->set_current(&my_cid, chips_window, false);
            }
         break;
         default:
         break;
      }
   }
   return rval;
}

void ChipsTileCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Tile " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsTileCommand::serialize(ChipsStateStream& out)
{
   int val = 6;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);
   out.write_int(&o_mode);
   out.write_int(&o_rows);
   out.write_int(&o_cols);
   out.write_int(&n_mode);
   out.write_int(&n_rows);
   out.write_int(&n_cols);
   out.write_string(&o_frame);
   out.write_string(&n_frame);
   out.write_string(&currw);
}

void ChipsTileCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_int(&o_mode);
   in.read_int(&o_rows);
   in.read_int(&o_cols);
   in.read_int(&n_mode);
   in.read_int(&n_rows);
   in.read_int(&n_cols);
   in.read_string(&o_frame);
   in.read_string(&n_frame);
   in.read_string(&currw);
}


ChipsSplitCommand::ChipsSplitCommand(ChipsObject *new_co, 
   float ollx, float olly, float ourx, float oury, bool osp,
   float nllx, float nlly, float nurx, float nury, bool nsp, int flag)
{
   co = new_co;
   weight=1;
 
   o_llx = ollx;
   o_lly = olly;
   o_urx = ourx;
   o_ury = oury;
   o_sp  = osp;
   n_llx = nllx;
   n_lly = nlly;
   n_urx = nurx;
   n_ury = nury;
   n_sp = nsp;
   typ = flag; 
}

ChipsSplitCommand::ChipsSplitCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

ChipsSplitCommand::~ChipsSplitCommand()
{
}

int ChipsSplitCommand::execute()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cl = (ChipsFrame*)co;
      if (typ == 0)
      {
//WLM         rval = cl->adjust_object_defs(n_llx, n_lly, n_urx, n_ury, false);
//WLM         cl->set_in_split(n_sp); 
      }
      else
      {
         rval = cl->reposition_plot(NULL, n_llx, n_lly, n_urx, n_ury, false);
      }
   }
   return rval;
}

int ChipsSplitCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cl = (ChipsFrame*)co;
      if (typ == 0) 
      {
//WLM         rval = cl->adjust_object_defs(o_llx, o_lly, o_urx, o_ury, false);
//WLM         cl->set_in_split(o_sp); 
      }
      else 
      {
         rval = cl->reposition_plot(NULL, o_llx, o_lly, o_urx, o_ury, false);
      }
   }
   return rval;
}

void ChipsSplitCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Split " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsSplitCommand::serialize(ChipsStateStream& out)
{
   int val = 7;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);


   out.write_float(&o_llx);
   out.write_float(&o_lly);
   out.write_float(&o_urx);
   out.write_float(&o_ury);
   out.write_bool(&o_sp);
   out.write_float(&n_llx);
   out.write_float(&n_lly);
   out.write_float(&n_urx);
   out.write_float(&n_ury);
   out.write_bool(&n_sp);
   out.write_int(&typ);
}

void ChipsSplitCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);


   in.read_float(&o_llx);
   in.read_float(&o_lly);
   in.read_float(&o_urx);
   in.read_float(&o_ury);
   in.read_bool(&o_sp);
   in.read_float(&n_llx);
   in.read_float(&n_lly);
   in.read_float(&n_urx);
   in.read_float(&n_ury);
   in.read_bool(&n_sp);
   in.read_int(&typ);

}


ChipsSetSplitCommand::ChipsSetSplitCommand(ChipsObject *new_co, int orient, 
   int cnt, int* vals, int cols, int rows, float xg, float yg, string* lids,
   string oldcurr)
{
   int rr; 
   co = new_co;
   weight=1;
   o_orient = orient;
   o_lids = n_lids = NULL; 
   o_vals = n_vals = NULL;
   o_cols = cols; 
   o_rows = rows;
   o_lcnt = 0;
   if (((o_cnt = cnt) > 0) && vals) 
   {
      o_vals = new int[cnt];
      for (rr = cnt; rr--; ) o_lcnt += (o_vals[rr] = vals[rr]);   
   }
   else
   {
      o_lcnt = o_cols * o_rows;  
      o_cnt = 0;
   } 
   
   if (lids)
   {
      o_lids = new string[o_lcnt]; 
      for (rr = o_lcnt; rr--; ) o_lids[rr] = lids[rr]; 
   }
   o_xg = xg;
   o_yg = yg;
   o_curr = oldcurr; 
}


ChipsSetSplitCommand::ChipsSetSplitCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}


ChipsSetSplitCommand::~ChipsSetSplitCommand()
{
   if (o_vals) delete [] o_vals; 
   if (n_vals) delete [] n_vals; 
   if (o_lids) delete [] o_lids;
   if (n_lids) delete [] n_lids;
}


void ChipsSetSplitCommand::set_new_values(int norient, int ncnt, int* nvals,
   int ncols, int nrows, float nxg, float nyg, string* nlids, string ncurr)
{
   int rr; 
   n_orient = norient;
   n_cols = ncols;
   n_rows = nrows;
   n_lcnt = 0; 
   if (((n_cnt = ncnt) > 0) && nvals)
   {
      n_vals = new int[ncnt];
      for (rr = ncnt; rr--; ) n_lcnt += (n_vals[rr] = nvals[rr]);
   }
   else  n_lcnt = n_cols * nrows;
   if (nlids)
   {
      n_lids = new string[n_lcnt];
      for (rr = n_lcnt; rr--; ) n_lids[rr] = nlids[rr];
   }
   n_xg = nxg;
   n_yg = nyg;
   n_curr = ncurr; 
}


int ChipsSetSplitCommand::execute()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cf = (ChipsFrame*)co; 
      if (n_cnt != 0) 
         rval = cf->reset_split_variables(n_orient, n_cnt, n_vals, 0, 0, n_xg, n_yg, n_lids); 
      else
         rval = cf->reset_split_variables(0, 0, NULL, n_cols, n_rows, n_xg, n_yg, n_lids); 
      ChipsId my_cid;
      my_cid.set_plot_id(n_curr); 
      cf->set_current(&my_cid, chips_plot, false);
   }
   return rval;
}

int ChipsSetSplitCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cf = (ChipsFrame*)co; 
      if (n_cnt != 0) 
         rval = cf->reset_split_variables(o_orient, o_cnt, o_vals, 0, 0, o_xg, o_yg, o_lids);
      else
         rval = cf->reset_split_variables(0, 0, NULL, o_cols, o_rows, o_xg, o_yg, o_lids);

      ChipsId my_cid;
      my_cid.set_plot_id(o_curr); 
      cf->set_current(&my_cid, chips_plot, false);
   }
   return rval;
}


void ChipsSetSplitCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"SetSplit " << v_object_to_string (co->get_object_type ()) << endl;
};


void ChipsSetSplitCommand::serialize(ChipsStateStream& out)
{
   int val = 8;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   if (out.get_major_version() < 3 ||
      (out.get_major_version() == 3 && out.get_minor_version() < 16))
   {
      double dd = (double)o_xg; 
      out.write_double(&dd);
      dd = (double)o_yg;
      out.write_double(&dd);
      dd = (double)n_xg;
      out.write_double(&dd);
      dd = (double)n_yg;
      out.write_double(&dd);
   }
   else 
   {
      out.write_float(&o_xg);
      out.write_float(&o_yg);
      out.write_float(&n_xg);
      out.write_float(&n_yg);
   }
   out.write_int(&o_orient);
   out.write_int(&o_cnt);
   if (o_cnt) 
   {
      out.write_int(o_vals, o_cnt); 
   }
   out.write_int(&o_cols);
   out.write_int(&o_rows);
   out.write_int(&n_orient);
   out.write_int(&n_cnt);
   if (n_cnt) 
   {
      out.write_int(n_vals, n_cnt);
   }
   out.write_int(&n_cols);
   out.write_int(&n_rows);
   if (o_lids)
   {
      out.write_int(&o_lcnt); 
      for (val = o_lcnt; val--; )  out.write_string(&o_lids[val]);  
   }
   else 
   {
      val = 0; 
      out.write_int(&val); 
   }
   if (n_lids)
   {
      out.write_int(&n_lcnt); 
      for (val = n_lcnt; val--;)  out.write_string(&n_lids[val]);  
   }
   else 
   {
      val = 0; 
      out.write_int(&val); 
   }
   out.write_string(&o_curr); 
   out.write_string(&n_curr); 
}


void ChipsSetSplitCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   if (in.get_major_version() < 3 ||
      (in.get_major_version() == 3 && in.get_minor_version() < 16))
   {
      double dd;
      in.read_double(&dd);
      o_xg = (float)dd;
      in.read_double(&dd);
      o_yg = (float)dd;
      in.read_double(&dd);
      n_xg = (float)dd;
      in.read_double(&dd);
      n_yg = (float)dd;
   }
   else
   {
      in.read_float(&o_xg);
      in.read_float(&o_yg);
      in.read_float(&n_xg);
      in.read_float(&n_yg);
   }
   in.read_int(&o_orient);
   in.read_int(&o_cnt);
   if (o_cnt) 
   {
      o_vals = new int[o_cnt]; 
      in.read_int(o_vals, o_cnt);
   }
   else 
      o_vals = NULL;
   in.read_int(&o_cols);
   in.read_int(&o_rows);
   in.read_int(&n_orient);
   in.read_int(&n_cnt);
   if (n_cnt) 
   {
      n_vals = new int[n_cnt]; 
      in.read_int(n_vals, n_cnt);
   }
   else
      n_vals = NULL;
   in.read_int(&n_cols);
   in.read_int(&n_rows);

   in.read_int(&o_lcnt);
   if (o_lcnt > 0)
   {
      int rr; 
      o_lids = new string[o_lcnt];
      for (rr = o_lcnt; rr--; )  in.read_string(&o_lids[rr]);  
   }
   else 
      o_lids = NULL;
   in.read_int(&n_lcnt);
   if (n_lcnt > 0)
   {
      int rr; 
      n_lids = new string[n_lcnt];
      for (rr = n_lcnt; rr--; )  in.read_string(&n_lids[rr]);  
   }
   else 
      n_lids = NULL;
   in.read_string(&o_curr); 
   in.read_string(&n_curr); 
}




ChipsCurrentCommand::ChipsCurrentCommand(ChipsObject* cob, string oval, 
   string nval, V_OBJECT_TYPE ob)
{
   co = cob;
   weight=1;
   obtype = ob; 
   o_curr = oval; 
   n_curr = nval; 
}

ChipsCurrentCommand::ChipsCurrentCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

ChipsCurrentCommand::ChipsCurrentCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight=1;
}


ChipsCurrentCommand::~ChipsCurrentCommand()
{
}


int ChipsCurrentCommand::execute ()
{
   int rval = ERR_STATUS_BAD; 
   if (co)
   {
      ChipsId cid;
      switch(obtype)
      {
         case chips_axis:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_axis_id(n_curr);
            rval = cp->set_current_axis(&cid, false);
         }
         break;  
         case chips_colorbar:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_colorbar_id(n_curr);
            rval = cp->set_current_colorbar(&cid, false);
         }
         break;
         case chips_curve:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_curve_id(n_curr);
            rval = cp->set_current_curve(&cid, false);
         }
         break; 
         case chips_image:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_image_id(n_curr);
            rval = cp->set_current_image(&cid, false);
         }
         break; 
         case chips_contour:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_contour_id(n_curr);
            rval = cp->set_current_contour(&cid, false);
         }
         break;
         case chips_histogram:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_histogram_id(n_curr);
            rval = cp->set_current_histogram(&cid, false);
         }
         break;
         case chips_label:
         {
            ChipsPlot* cl = (ChipsPlot*)co;
            cid.set_label_id(n_curr);
            rval = cl->set_current_label(&cid, false);
         }
         break;
         case chips_line:
         {
            ChipsPlot* cl = (ChipsPlot*)co;
            cid.set_line_id(n_curr);
            rval = cl->set_current_line(&cid, false);
         }
         break;
         case chips_point:
         {
            ChipsPlot* cl = (ChipsPlot*)co;
            cid.set_point_id(n_curr);
            rval = cl->set_current_point(&cid, false);
         }
         break;
         case chips_region:
         {
            ChipsPlot* cl = (ChipsPlot*)co;
            cid.set_region_id(n_curr);
            rval = cl->set_current_region(&cid, false);
         }
         break;
         case chips_window:
         {
            ChipsWindow* cw = (ChipsWindow*)co;
            cid.set_window_id(n_curr); 
            rval = ((ChipsEngine*)(cw->get_engine()))->set_current(&cid,
                                                           obtype, false);
         }
         break;
         case chips_plot:
         {
            ChipsFrame* cl = (ChipsFrame*)co;
            cid.set_plot_id(n_curr);
            rval = cl->set_current(&cid, obtype, false);
         }
         break;
         case chips_frame:
         {
            ChipsWindow* cw = (ChipsWindow*)co;
            cid.set_frame_id(n_curr); 
            rval = cw->set_current(&cid, obtype, false); 
         }
         break;
         default:
            rval = ERR_STATUS_BAD;
         break; 
      }
   }
   return rval;
}
int ChipsCurrentCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD; 
   if (co)
   {
      ChipsId cid;
      switch(obtype)
      {
         case chips_axis:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_axis_id(o_curr);
            rval = cp->set_current_axis(&cid, false);
         }
         break;  
         case chips_colorbar:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_colorbar_id(o_curr);
            rval = cp->set_current_colorbar(&cid, false);
         }
         break;
         case chips_curve:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_curve_id(o_curr);
            rval = cp->set_current_curve(&cid, false);
         }
         break; 
         case chips_image:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_image_id(o_curr);
            rval = cp->set_current_image(&cid, false);
         }
         break; 
         case chips_contour:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_contour_id(o_curr);
            rval = cp->set_current_contour(&cid, false);
         }
         break;
         case chips_histogram:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_histogram_id(o_curr);
            rval = cp->set_current_histogram(&cid, false);
         }
         break;
         case chips_label:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_label_id(o_curr);
            rval = cp->set_current_label(&cid, false);
         }
         break;
         case chips_line:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_line_id(o_curr);
            rval = cp->set_current_line(&cid, false);
         }
         break;
         case chips_point:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_point_id(o_curr);
            rval = cp->set_current_point(&cid, false);
         }
         break;
         case chips_region:
         {
            ChipsPlot* cp = (ChipsPlot*)co;
            cid.set_region_id(o_curr);
            rval = cp->set_current_region(&cid, false);
         }
         break;
         case chips_window:
         {
            ChipsWindow* cw = (ChipsWindow*)co;
            cid.set_window_id(o_curr); 
            rval = ((ChipsEngine*)(cw->get_engine()))->set_current(&cid,
                                                                   obtype, false);
         }
         break;
         case chips_plot:
         {
            ChipsFrame* cl = (ChipsFrame*)co;
            cid.set_plot_id(o_curr);
            rval = cl->set_current(&cid, obtype, false);
         }
         break;
         case chips_frame:
         {
            ChipsWindow* cw = (ChipsWindow*)co;
            cid.set_frame_id(o_curr); 
            rval = cw->set_current(&cid, obtype, false); 
         }
         break;
         default:
            rval = ERR_STATUS_BAD;
         break; 
      }
   }
   return rval;
}

void ChipsCurrentCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Current " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsCurrentCommand::serialize(ChipsStateStream& out)
{
   int val = 9;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   val = (int) obtype;
   out.write_int(&val);
   out.write_string(&o_curr);
   out.write_string(&n_curr);
}

void ChipsCurrentCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);

   int val;
   in.read_int(&val);
   obtype = (V_OBJECT_TYPE)val; 
   in.read_string(&o_curr);
   in.read_string(&n_curr);
}


ChipsBindCommand::ChipsBindCommand(ChipsObject* cob, string plt1, string ax1, 
                       string plt2, string ax2, double* dp, int* ip, string* sp,
                       bool bind_flag)
{
   co = cob;
   weight=1;
   plot1 = plt1;
   plot2 = plt2;
   axis1 = ax1;
   axis2 = ax2; 
   osp = NULL;
   odp = NULL;
   oip = NULL;
   scnt = dcnt = 0;
   bind = bind_flag;

   if (dp && ip)
   {
      icnt = ip[0];
      int rr, tcnt = ip[12];       // tcnt = number of tick position entries 
      int acnt = ip[13];          // acnt = number of arbitrary tick positions 
      scnt = ip[2];               // scnt
      dcnt = ip[1];               // dcnt
      if (tcnt < 0) tcnt = 0;
      if (acnt < 0) acnt = 0;
      if ((odp = new double[dcnt]) != NULL)
         for (rr = 0; rr < dcnt; rr++)   odp[rr] = dp[rr]; 
      if ((oip = new int[icnt]) != NULL)  
         for (rr = 0; rr < icnt; rr++)   oip[rr] = ip[rr]; 
      if (sp)  
      {
         if ((osp = new string[scnt]) != NULL)  
            for (rr = 0; rr < scnt; rr++) osp[rr] = sp[rr]; 
      }
   }
}

ChipsBindCommand::ChipsBindCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

ChipsBindCommand::ChipsBindCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight=1;
}

ChipsBindCommand::~ChipsBindCommand()
{
   if (odp) delete [] odp;
   if (oip) delete [] oip;
   if (osp) delete [] osp;
}

int ChipsBindCommand::execute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cf = (ChipsFrame*)co;
      ChipsId temp;
      temp.set_plot_id(plot1);
      temp.set_axis_id(axis1);
      if (bind)
         rval = cf->bind_axes(&temp, plot2, axis2, false);
      else
         rval = cf->unbind_axes(&temp, plot2, axis2, false);
   }
   return rval;
}

int ChipsBindCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame* cf = (ChipsFrame*)co;
      ChipsId temp;
      temp.set_plot_id(plot1);
      temp.set_axis_id(axis1);
      ChipsPlot* myplot = cf->get_plot(plot2);
      ChipsAxis* myaxis = NULL;

      if (myplot)
      {
         myaxis = myplot->get_xaxis(axis2);
         if (myaxis == NULL)  myaxis = myplot->get_yaxis(axis2);
      }
      
      if (bind)
         rval = cf->unbind_axes(&temp, plot2, axis2, false);
      else
         rval = cf->bind_axes(&temp, plot2, axis2, false);

      // now return axis2 back to its original state
      if (myaxis)
      {
         ChipsAxisBind *bind_table = cf->get_bind_table ();
         list <ChipsAxis *> bound_axes;
         list <ChipsAxis *>::iterator iter;
         list <string> axis_list;
         int cursc = myaxis->get_scaletype();

         myaxis->unsync_data(odp, oip, osp);
         axis_list.push_back (myaxis->get_id ());

         // go through all bound axes and unsync them from this axis
         if (bind_table)
         {
            // record any changes in the axes
            bind_table->record_updates (true);
            bind_table->reset_access_list();

            bound_axes = bind_table->get_axes (myaxis);
            for (iter = bound_axes.begin (); iter != bound_axes.end (); iter++)
            {
               (*iter)->unsync_data(odp, oip, osp);
               axis_list.push_back ((*iter)->get_id ());
            }

         
            bind_table->update_connected_nodes (myaxis, myaxis->get_absolute_min (), myaxis->get_absolute_max (), true, AUTOLIM_DEF, 0, false);

            cf->check_plot_updates ();
            bind_table->record_updates (false);
            bind_table->reset_access_list();
         }

         // if the scaletype changed- reset scale in plot so coords update
         if (cursc != myaxis->get_scaletype())
         {
            ChipsId mycid;
            mycid.set_plot_id(plot2);
            mycid.set_axis_id(axis2);
            int ior = myaxis->get_orientation();
            if (cursc == V_SCALE_LOG) {
               myplot->set_property(&mycid,static_cast<void*>(&ior),SET_PLOT_LINEARSCALE,0);
            } else {
               myplot->set_property(&mycid,static_cast<void*>(&ior),SET_PLOT_LOGSCALE,0);
            }
         }
      }
   }
   return rval;
}


void ChipsBindCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Bind " << v_object_to_string (co->get_object_type ()) << endl;
};

void ChipsBindCommand::serialize(ChipsStateStream& out)
{
   int val = 12;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);


   out.write_string(&plot1);
   out.write_string(&plot2);
   out.write_string(&axis1);
   out.write_string(&axis2);
   out.write_int(&dcnt);
   if (dcnt)  out.write_double(odp, dcnt);
   out.write_int(&icnt);
   if (icnt)  out.write_int(oip, icnt);
   out.write_int(&scnt);
   if (scnt)  for (val = scnt; val--;)  out.write_string(&osp[val]);
   out.write_bool(&bind);
}

void ChipsBindCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);

   in.read_int(&block_ref);


   in.read_string(&plot1);
   in.read_string(&plot2);
   in.read_string(&axis1);
   in.read_string(&axis2);
   in.read_int(&dcnt);
   if (dcnt)
   {
      odp = new double[dcnt]; 
      in.read_double(odp, dcnt);
   }
   else odp = NULL; 
   in.read_int(&icnt);
   if (icnt)
   {
      oip = new int[icnt]; 
      in.read_int(oip, icnt);
   }
   else icnt = 0; 
   in.read_int(&scnt);
   if (scnt)
   {
      osp = new string[scnt]; 
      for (int rr = scnt; rr--; )  in.read_string(&osp[rr]);
   }
   else osp = NULL;  
   in.read_bool (&bind);

   if (in.get_minor_version () < 8)
   {
      int *tmp_ip = new int[icnt];
      for (int rr=0; rr<icnt; rr++)
         tmp_ip[rr] = oip[rr];
      delete []oip;
      oip = new int[icnt+3];
      oip[0] = icnt;
      oip[1] = dcnt;
      oip[2] = scnt;
      for (int rr=0; rr<icnt; rr++)
         oip[rr+3] = tmp_ip[rr];
      delete []tmp_ip;
   }

}



ChipsDefDispColorCommand::ChipsDefDispColorCommand(ChipsObject* cob)
{
   co = cob;
   weight=1;
}

ChipsDefDispColorCommand::ChipsDefDispColorCommand(ChipsObject* cob,
   ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

ChipsDefDispColorCommand::ChipsDefDispColorCommand(ChipsStateStream& in)
{
   co = NULL;
   deserialize(in);
   weight=1;
}


ChipsDefDispColorCommand::~ChipsDefDispColorCommand()
{
}


int ChipsDefDispColorCommand::execute ()
{
   if (co)
   {
      ChipsWindow* cw = (ChipsWindow*)co;
      ((ChipsEngine*)(cw->get_engine()))->update_default_display_colors();
   }
   return ERR_STATUS_GOOD;
}
int ChipsDefDispColorCommand::unexecute ()
{
   execute();     // run same code as for execute 
   return ERR_STATUS_GOOD;
}

void ChipsDefDispColorCommand::print ()
{
   if (co)
      cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"DefDispColor " << endl;
};

void ChipsDefDispColorCommand::serialize(ChipsStateStream& out)
{
   int val = 28;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);
}

void ChipsDefDispColorCommand::deserialize(ChipsStateStream& in)
{
   in.read_int(&ref);
   in.read_int(&block_ref);
}


