// 
//  Copyright (C) 2006,2007,2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSCOMMAND_HH
#define CHIPSCOMMAND_HH

#include "ChipsObject.hh"

class ChipsStateStream; 

class ChipsCommand
{
   public:
      virtual ~ChipsCommand () {};
      virtual int execute () = 0;
      virtual int unexecute () =0;
      virtual int clear () = 0;
      virtual void print () = 0;
      virtual void serialize(ChipsStateStream& out) = 0; 
      virtual void deserialize(ChipsStateStream& in) = 0; 
      void serialize_object(ChipsStateStream& out); 
      void set_ref_id(int val)  { ref = val;  }
      int  get_ref_id()         { return ref; } 
      int  get_weight()         { return weight; }
      void set_block_id (int val) { block_ref = val; };
      int get_block_id ()       { return block_ref; };
   protected:
      int ref; 
      int block_ref;
      int weight;
      ChipsObject *co;
};

class ChipsCreateCommand : public ChipsCommand
{
   public:
      ChipsCreateCommand (ChipsObject *new_co, bool chng_cnt, const string &curr="");
      ChipsCreateCommand (ChipsObject *new_co, ChipsStateStream& in);
      ChipsCreateCommand(ChipsStateStream& in);
      virtual ~ChipsCreateCommand();
      virtual int execute ();
      virtual int unexecute ();
      virtual int clear ();
      virtual void print ();
      virtual void serialize(ChipsStateStream& out); 
      virtual void deserialize(ChipsStateStream& in); 
   protected:
      string current; 
      bool change_cnt;
};

class ChipsSetCommand : public ChipsCommand
{
   public:
      ChipsSetCommand(ChipsObject *cob, int property, string oldv, 
                      string newv, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, bool oldv, 
                      bool newv, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, int oldv, 
                      int newv, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, float oldv, 
                      float newv, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, double oldv, 
                      double newv, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, unsigned char* oldv,
                      unsigned char* newv, int alen, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, int property, double* oldv, double* newv,
                      int alen, string oldcurr);
      ChipsSetCommand(ChipsObject *cob, ChipsStateStream& in);
      ChipsSetCommand(ChipsStateStream& in);
      virtual ~ChipsSetCommand();
      virtual int execute ();
      virtual int unexecute ();
      virtual int clear () {return ERR_STATUS_GOOD;};
      virtual void print ();
      virtual void serialize(ChipsStateStream& out); 
      virtual void deserialize(ChipsStateStream& in); 
   protected:
      int prop;
      int type;     // 0=string 1=bool 2=int 3=double 4=float 5=uchar 6=double[] 
      string oldval;
      string newval;
      string current;
      double oval_d;
      double nval_d;
      float  oval_f;
      float  nval_f;
      int    oval_i;
      int    nval_i; 
      bool   oval_b;
      bool   nval_b;
      unsigned char* oval_uc;
      unsigned char* nval_uc;
      double* oval_da; 
      double* nval_da; 
      int    sz; 
};


class ChipsLimitsCommand : public ChipsCommand
{
   public:
      ChipsLimitsCommand(ChipsObject *new_co, int flag, double olin, 
                         double olog, double nval);
      ChipsLimitsCommand(ChipsObject *new_co, int flag, double olinmin, 
                         double ologmin, double olinmax, double ologmax,
                         int oflag, int nflag, double nmin, double nmax,
                         bool transformed);
      ChipsLimitsCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsLimitsCommand(ChipsStateStream& in);
      virtual ~ChipsLimitsCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      double  n_min;
      double  n_max; 
      double  o_logmin;
      double  o_logmax; 
      double  o_linmin;
      double  o_linmax; 
      int     lim_type; 
      int     o_autoflag; 
      int     n_autoflag; 
      bool    transformed;
};


class ChipsMoveCommand : public ChipsCommand
{
   public:
      ChipsMoveCommand(ChipsObject *new_co, double ox, double oy, double nx, double ny,
                       COORD_SYS ncs, string ncstr, int nm, string current); 
      ChipsMoveCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsMoveCommand(ChipsStateStream& in);
      virtual ~ChipsMoveCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      double  o_x;
      double  o_y;
      double  n_x;
      double  n_y;
      int     n_mode;
      int     n_coord;
      string  n_coordid;
      string  current; 
};

class ChipsTileCommand : public ChipsCommand
{
   public:
      ChipsTileCommand(ChipsObject *new_co, int ofmode, int otcols, int otrows, 
                       string of, int nfmode, int ntcols, int ntrows, string nf,
                       string cw);
      ChipsTileCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsTileCommand(ChipsStateStream& in);
      virtual ~ChipsTileCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      int     o_mode;
      int     o_rows;
      int     o_cols;
      int     n_mode;
      int     n_rows;
      int     n_cols;
      string  o_frame;
      string  n_frame;
      string  currw;
};


class ChipsSplitCommand : public ChipsCommand
{
   public:
      ChipsSplitCommand(ChipsObject *new_co, float ollx, float olly, float ourx,
                      float oury, bool osp, float nllx, float nlly, float nurx,
                      float nury, bool nsp, int flag);
      ChipsSplitCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsSplitCommand(ChipsStateStream& in);
      virtual ~ChipsSplitCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      float   o_llx;
      float   o_lly; 
      float   o_urx;
      float   o_ury;
      bool    o_sp;
      float   n_llx;
      float   n_lly; 
      float   n_urx;
      float   n_ury;
      bool    n_sp;
      int     typ; 
};


class ChipsSetSplitCommand : public ChipsCommand
{
   public:
      ChipsSetSplitCommand(ChipsObject *new_co, int orient, int cnt, int* vals,
                          int cols, int rows, float xg, float yg, 
                          string* lids, string oldcur);
      ChipsSetSplitCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsSetSplitCommand(ChipsStateStream& in);
      void set_new_values(int norient, int ncnt, int* nvals, int ncols, 
                          int nrows, float nxg, float nyg, string* nlids,
                          string newcur); 
      virtual ~ChipsSetSplitCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      string* o_lids;
      string* n_lids; 
      string  o_curr; 
      string  n_curr; 
      float   o_xg;
      float   o_yg;
      float   n_xg;
      float   n_yg;
      int     o_orient;
      int     o_cnt;
      int*    o_vals;
      int     o_cols;
      int     o_rows;
      int     n_orient;
      int     n_cnt;
      int*    n_vals;
      int     n_cols;
      int     n_rows;
      int     o_lcnt; 
      int     n_lcnt; 
};


class ChipsBindCommand : public ChipsCommand
{
   public:
      ChipsBindCommand(ChipsObject* co, string plt1, string ax1, 
                       string plt2, string ax2, double* dp,
                       int* ip, string* sp, bool bind_flag);
      ChipsBindCommand(ChipsObject* co,  ChipsStateStream& in);
      ChipsBindCommand(ChipsStateStream& in);
            virtual ~ChipsBindCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      string  plot1;
      string  plot2;
      string  axis1;  
      string  axis2; 
      int     dcnt; // number of double elements 
      int     icnt; // number of int elements 
      int     scnt; // number of string elements 
      double* odp;  // pre-bind/synch axis double values
      int*    oip;  // pre-bind/synch axis int values
      string* osp;  // pre-bind/synch axis string values
      bool    bind; // flag whether this is a bind or unbind command
};


class ChipsCurrentCommand : public ChipsCommand
{
   public:
      ChipsCurrentCommand(ChipsObject *new_co, string oldv, string nval,
                          V_OBJECT_TYPE ob);
      ChipsCurrentCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsCurrentCommand(ChipsStateStream& in);
      virtual ~ChipsCurrentCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
   protected:
      string o_curr;
      string n_curr;
      V_OBJECT_TYPE obtype; 
};

class ChipsDefDispColorCommand : public ChipsCommand
{
   public:
      ChipsDefDispColorCommand(ChipsObject *new_co);
      ChipsDefDispColorCommand(ChipsObject *new_co, ChipsStateStream& in);
      ChipsDefDispColorCommand(ChipsStateStream& in);
      virtual ~ChipsDefDispColorCommand();
      virtual int execute();
      virtual int unexecute();
      virtual int clear() {return ERR_STATUS_GOOD;};
      virtual void print();
      virtual void serialize(ChipsStateStream& out);
      virtual void deserialize(ChipsStateStream& in);
};


#endif
