// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsContour.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"
#include "ChipsContourLevelCommand.hh"
#include "ChipsTransformFactory.hh"
#include "ChipsTransformCommand.hh"
#include "Transform.hh"
#include <stdio.h>
#include <math.h>

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;

ChipsContour::ChipsContour(ChipsObject* par, unsigned long cnt, double* x, 
   double* y, double* z, string transfrm_id, string val):ChipsObject(par, val)
{
   initialized = false;
   id = val; 
   num_pts = cnt; 

   if (((xx = new double[cnt]) != NULL) && ((yy = new double[cnt]) != NULL) &&
       ((zz = new double[cnt]) != NULL))
   {
      int rr;
      for (rr = cnt; rr--; )
      {
         xx[rr] = x[rr];
         yy[rr] = y[rr];
         zz[rr] = z[rr];
      }
   }

   transform = NULL;
   transform_id = transfrm_id;

   initialize();
   calculate_limits();
   set_mode(mode);
}


ChipsContour::ChipsContour(ChipsObject* par, unsigned long cnt, double* x, 
   double* y, double* z, unsigned int numlev, double* levs, string transfrm_id,
   string val): ChipsObject(par, val)
{
   initialized = false;
   id = val;
   num_pts = cnt; 
   if (((xx = new double[cnt]) != NULL) && ((yy = new double[cnt]) != NULL) &&
       ((zz = new double[cnt]) != NULL))
   {
      int rr;
      for (rr = cnt; rr--; )
      {
         xx[rr] = x[rr];
         yy[rr] = y[rr];
         zz[rr] = z[rr];
      }
   }
   num_levels = numlev;
   initialize();

   transform = NULL;
   transform_id = transfrm_id;

   calculate_limits();
   if (numlev && levs)
   {

      num_levels = numlev;

      if (arb_levels)
         delete []arb_levels;
      if (levels)
         delete []levels;
      
      int rr; 
      arb_num = num_levels; 
      arb_levels = new double[arb_num];
      levels = new double[arb_num];
      if (arb_levels && levels) 
      {
         for (rr = arb_num; rr--; ) 
            arb_levels[rr] = levels[rr] = levs[rr]; 
         mode = chips_arbitrary; 
      }
   }

   set_mode(mode);
}


ChipsContour::ChipsContour(ChipsObject* par, ChipsStateStream& in):ChipsObject(par,"")
{
   initialized = false;

   transform = NULL;
   transform_id = "";

   initialize();
   deserialize(in);
   calculate_limits();
}


ChipsContour::~ChipsContour()
{
  if (xx)
   {
      delete [] xx;
      xx = NULL;
   }
   if (yy)
   {
      delete [] yy;
      yy = NULL;
   }
   if (zz)
   {
      delete [] zz;
      zz = NULL;
   }
   if (levels)
      delete []levels;
   if (arb_levels)
      delete []arb_levels;
   if (coord)
      coord->remove(this);

   if (transform)
      delete transform;
}

void ChipsContour::object_delete ()
{
   delete this;
}


void ChipsContour::initialize()
{
   string tmp_color; 

   mode = chips_nice; 
   interval = 10; 
   transform_applied = false;
   if (theDefaults)
   {
      vector<double> tmp_levels;
      style = theDefaults->get_contourstyle();
      thickness = theDefaults->get_contourthickness();
      tmp_color = theDefaults->get_contourcolor();
      mode = theDefaults->get_contourmode();
      interval = theDefaults->get_contourinterval();
      algorithm = theDefaults->get_contouralgorithm();
      if (set_color(tmp_color) != 0)
      {
         tmp_color = "cyan";
         set_color(tmp_color);
      }
      theDefaults->get_contourlevels(tmp_levels);
      if (tmp_levels.size () > 0)
      {
         num_levels = tmp_levels.size ();
         arb_num = tmp_levels.size ();
         levels = new double[tmp_levels.size ()];
         arb_levels = new double[tmp_levels.size ()];
         for (unsigned int ii=0; ii<tmp_levels.size (); ii++)
            arb_levels[ii] = levels[ii] = tmp_levels[ii];
      }
      else
      {
         levels = NULL;
         arb_levels = NULL;
         arb_num = 0;
      }
      count = theDefaults->get_contournumlevels(); 
      depth = theDefaults->get_contourdepth();
      meta = theDefaults->get_meta();
   }
   else 
   {
      style = chips_dot;
      thickness = 1; 
      tmp_color = "cyan";
      set_color(tmp_color); 
      count = 5;
      algorithm = chips_standard; 
      depth = V_USE_DEFAULT_DEPTH;
      levels = NULL;
      arb_levels = NULL;
      meta = "";
   }
   if (depth == V_USE_DEFAULT_DEPTH)  depth = ((ChipsPlot*)parent)->get_depth();
   object_type = chips_contour;
   ticks = false;
   tick_dir = 0;
   set_tick_color(tmp_color);
   tick_len = 3;
   tick_thick = 1; 
   label = false;
   set_label_color(tmp_color);
   label_size = 14;
   label_font = "helvetica";
   label_style = chips_normal;
   contour_num = 0;
   initialized = true;
}


void ChipsContour::calculate_limits()
{
   // get data ranges
   int rr, ndx = num_pts -1;
   xlmin = xmin = xmax = xx[ndx]; 
   ylmin = ymin = ymax = yy[ndx]; 
   zlmin = zmin = zmax = zz[ndx]; 
   for (rr = ndx; rr--; )
   {
      if (xx[rr] < xmin)  xmin = xx[rr];
         else if (xx[rr] > xmax)  xmax = xx[rr];
      if (yy[rr] < ymin)  ymin = yy[rr];
         else if (yy[rr] > ymax)  ymax = yy[rr];
      if (zz[rr] < zmin)  zmin = zz[rr];
         else if (zz[rr] > zmax)  zmax = zz[rr];
      // store smallest positive value as log minimum
      if ((xx[rr] > 0) && ((xx[rr] < xlmin) || (xlmin < 0)))   xlmin = xx[rr];
      if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin < 0)))   ylmin = yy[rr];
      if ((zz[rr] > 0) && ((zz[rr] < zlmin) || (zlmin < 0)))   zlmin = zz[rr];
   }
   if (xlmin < 0) xlmin = 0; 
   if (ylmin < 0) ylmin = 0; 
   if (zlmin < 0) zlmin = 0; 

   // check for regular grid 
   regular = regular_grid(); 

   // create a default set of arbitrary mode values 
   if (arb_num == 0)
   {
      arb_num = 7; 
      if ((arb_levels = new double[arb_num]) != NULL)
      {
         arb_levels[0] = 10;
         arb_levels[1] = 50;
         arb_levels[2] = 100;
         arb_levels[3] = 500;
         arb_levels[4] = 1000;
         arb_levels[5] = 5000;
         arb_levels[6] = 10000;
      }
   }
}


int ChipsContour::set_transform (Transform *trans, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   rval = ERR_FEATURE_DNE;
   return rval;
}

int ChipsContour::apply_transform (bool rec)
{
   int rval = ERR_STATUS_GOOD;
   rval = ERR_FEATURE_DNE;

   return rval;
}

int ChipsContour::apply_transform_inverse (bool rec)
{
   int rval = ERR_STATUS_GOOD;
   rval = ERR_FEATURE_DNE;

   return rval;
}

int ChipsContour::get_transform (Transform **trans)
{
   int rval = ERR_STATUS_GOOD;

   // keep for backwards compatibility with state files
   
   if (!trans)
      return ERR_STATUS_BAD;

   if (transform)
      *trans = ChipsTransformFactory::copy_transform (transform);
   else
      *trans = static_cast <Transform *>(0);

   return rval;
}



// This recalculates the ranges after a coordinate system is updated
int ChipsContour::reset_position ()
{
   int rval = ERR_STATUS_GOOD;
   ChipsDataSystem *data_sys;
   if (!coord || (coord_system != DATA))
      return ERR_STATUS_BAD;

   data_sys = static_cast <ChipsDataSystem *>(coord);

   if (!transform_applied)
   {
      if (data_sys)
      {
         Transform *world_transform = data_sys->get_transform (transform_id);
         Transform *logical_transform = data_sys->get_transform (data_sys->get_default_transform_id ());

         // convert from logical to world

         if (world_transform != logical_transform)
         {
            if (world_transform)
               add_transform (world_transform);
            if (logical_transform)
               add_transform (logical_transform, true);
         }
      }

      transform_applied = true;
         
   }
   change_range (data_sys->get_lower_bounds (), data_sys->get_upper_bounds ());

   return rval;
}



int ChipsContour::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   int mask = theDefaults->get_allpropmask();

   if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
   {
       string clr = theDefaults->get_allcolor();
       rval = set_property((void*)clr.c_str(), SET_CONTOUR_COLOR);
   }
   if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
   {
      float newthick = theDefaults->get_allthickness();
      rval = set_property(static_cast<void*>(&newthick), SET_CONTOUR_THICKNESS);
      // adjust contour tick mark thickness when ticks are supported
      // rval |= set_tick_thickness(newthick);
   }
   if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
   {
      int vis = theDefaults->get_allvisible();
      rval = set_property(static_cast<void*>(&vis), SET_CONTOUR_VISIBLE);
   }
   if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
   {
      int dep = theDefaults->get_alldepth();
      rval = set_property(static_cast<void*>(&dep), SET_CONTOUR_DEPTH);
   }
   return rval;
}


int ChipsContour::set_style(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_line_type(val))
   {
      style = val;
      if (initialized) change_style();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val,"chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");
   return rval;
}

int ChipsContour::set_style(string val)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = c2_string_to_line(val.c_str(), &style, "line style")) == 0)
      if (initialized) change_style();
   return rval;
}


int ChipsContour::set_algorithm(V_CONTOUR_TYPE val)
{
   int rval = ERR_STATUS_GOOD;
   if (!valid_contour_type(val))
      cmd_manager->add_err(rval=ERR_BAD_CONTOUR_ALG,id,0,val);
   else if (algorithm != val)
   {
      algorithm = val; 
      if (initialized)  rval = change_algorithm();
   }
   return rval;
}


int ChipsContour::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}


int ChipsContour::set_interval(double val)
{
   int rval = ERR_STATUS_GOOD;
   if (interval != val)
   {
      interval = val;
      if (initialized && ((rval = compute_levels()) == ERR_STATUS_GOOD))
         rval = change_levels();
   }
   return rval;
}

int ChipsContour::set_numlevels(int val)
{
   int rval = ERR_STATUS_GOOD;
   if (count != val)
   {
      count = val;
      if (initialized && ((rval = compute_levels()) == ERR_STATUS_GOOD))
         rval = change_levels();
   }
   return rval;
}


int ChipsContour::set_mode(string val)
{
   int rval = ERR_STATUS_GOOD;

   V_TICKMODE_TYPE tt;
   if ((rval = string_to_v_tickmode(val.c_str(), &tt)) == ERR_STATUS_GOOD)
      rval = set_mode(tt);
   else
      cmd_manager->add_err(rval, id, 0, val.c_str());
   return rval;
}


int ChipsContour::set_mode(int val)
{
   int rval = ERR_STATUS_GOOD;
   mode = (V_TICKMODE_TYPE)val;
   if ((rval = compute_levels()) == ERR_STATUS_GOOD)  rval = change_levels();
   return rval;
}


int ChipsContour::compute_levels()
{
   int rval = ERR_STATUS_GOOD;

   // set levels array based on mode 
   switch(mode)
   {
      case chips_limits:  // fallthorugh intended 
      case chips_nice:
      {
         double xpmin, xpmax, dist, base;
         scale1(zmin, zmax, 5, &xpmin, &xpmax, &dist, &base);

         if (num_levels && levels)
	 {
	    delete [] levels;
	    levels = NULL;
	 } 
         if (xpmax > xpmin)   
            num_levels = (unsigned int)((xpmax - xpmin) / dist + 1);
         else 
            num_levels = 1;
         if ((levels = new double[num_levels]) != NULL)
         {
            unsigned int rr;
            levels[0] = xpmin; 
            for (rr = 1; rr < num_levels; rr++) 
               levels[rr] = levels[rr -1] + dist;
         } 
         else   cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      break;
      case chips_arbitrary:
      if (arb_levels && arb_num)
      {
         if (num_levels && levels) 
	 {
	    delete [] levels;
	    levels = NULL;
	 } 
         num_levels = arb_num; 
         if ((levels = new double[num_levels]) != NULL)
         {
            unsigned int rr;
            for (rr = num_levels; rr--; ) levels[rr] = arb_levels[rr];
         }  
         else   cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      break;
      case chips_count:
      {
         double xpmin, xpmax, dist;

         if ((count > 0) && (count < 501)) 
         {
            scale2(zmin, zmax, count, &xpmin, &xpmax, &dist);

            if (num_levels && levels) 
	    {
	       delete [] levels;
	       levels = NULL;
	    } 
            num_levels = count; 
            if ((levels = new double[num_levels]) != NULL)
            {
               unsigned int rr;
               levels[0] = xpmin; 
               for (rr = 1; rr < num_levels; rr++) 
                  levels[rr] = levels[rr -1] + dist;
            } 
            else   cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
         }
         else
         {
            mode = chips_limits;
            if (compute_levels() == ERR_STATUS_GOOD)
            {
               cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, 
                                    "contour number of levels",count,1,500);
            }
            mode = chips_count;  
         }
      }
      break;
      case chips_interval:
      if (interval > 0)
      {
         if (num_levels && levels)
	 {
	    delete [] levels;
	    levels = NULL;
	    num_levels = 0; 
	 } 
         double start = ((int)(zmin/ interval) + 1)*interval; 
         num_levels = (unsigned int)((zmax - zmin)/interval + 1);  
         if ((num_levels > 0) && (num_levels < 501))
         {
            if ((levels = new double[num_levels]) != NULL)
            {
               unsigned int rr;
               levels[0] = start; 
               for (rr = 1; rr < num_levels; rr++) 
                  levels[rr] = levels[rr -1] + interval;
            } 
            else   cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
         }
         else
         {
            mode = chips_limits;
            if (compute_levels() == ERR_STATUS_GOOD)
            {
               cmd_manager->add_err(rval=ERR_INT_MODE_RANGE, id, 0, "contour",
                                    interval);
            }
            mode = chips_interval; 
         }
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "contours's algorithm", (int)mode,"chips_standard|chips_marching");
      break; 
   }
   return rval;
}


int ChipsContour::set_levels(unsigned int sz, double* levs, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   bool change_mode = false;
   if (sz  && levs)
   {
      ChipsContourLevelCommand *cmd = new ChipsContourLevelCommand(this, 
         arb_num, arb_levels, sz, levs);
      if (arb_num && arb_levels)
      {
         // check for differences in the levels
         if (arb_num != sz)
            change_mode = true;
         for (int rr = arb_num; !change_mode && rr--; )
            if (arb_levels[rr] != levs[rr])
               change_mode = true;
         delete [] arb_levels;
      }
      if (mode != chips_arbitrary)
         change_mode = true;

      arb_num = sz;
      if ((arb_levels = new double[arb_num]) != NULL)
      {
         int rr;  
         for (rr = arb_num; rr--; )   arb_levels[rr] = levs[rr];
         if (rec)
         {
            // only change the mode if something has changed
            if (change_mode)
               rval = set_property((void*)"arbitrary", SET_CONTOUR_MODE,rec);
            if (!add_command (cmd, rval)) delete cmd;
         }
         else if ((mode == chips_arbitrary) && change_mode)
            // this needs to be done if the user does a redo. The set
            // property command gets called before this functions
            // therefore set_mode isn't called with the level changes
            set_mode (chips_arbitrary);

      }
      else   cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      if (!rec) delete cmd;
   }
   else   
   {
      cmd_manager->add_err("No contour levels were provided", id, 0);
      rval = ERR_DATA_RANGE;
   }
   return rval; 
} 


int ChipsContour::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_color();
   return rval;
}

int ChipsContour::set_label_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = label_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_label_property();
   return rval;
}

int ChipsContour::set_tick_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = tick_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_tick_property();
   return rval;
}


int ChipsContour::set_tick_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "contour tick thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      tick_thick = val;
      if (initialized)   rval = change_tick_property();
   }
   return rval;
}


int ChipsContour::set_tick_direction(int val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val == 0 ) || (val == 1))
   {
      tick_dir = val;
      if (initialized)   rval = change_tick_property();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "contour tick direction", val, 0, 1);

   return rval;
}


int ChipsContour::set_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "line thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      thickness = val;
      if (initialized) change_thickness();
   }
   return rval;
}


int ChipsContour::set_label_size(int val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val >= V_SIZE_MIN_VAL) && (val <= 50))
   {
      label_size = val;
      if (initialized) change_label_property();
   }
   else
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "error thickness", val,
                                  V_SIZE_MIN_VAL, 50);

   return rval;
}


int ChipsContour::set_label_font(string val)
{
   int rval = ERR_STATUS_BAD;

   if (val.length())
   {
      label_font = val;
      if (initialized)   rval = change_label_property();
   }
   else
      cmd_manager->add_err(rval=ERR_STR_SUBSET, id, 0, "font", val.c_str (), "helvetica|courier|times|greek|user loaded fonts");
   return rval;
}


int ChipsContour::set_label_fontstyle(string val)
{
   int rval = ERR_STATUS_BAD;

   V_FONTSTYLE_TYPE fs; 
   if ((rval = c2_string_to_fontstyle(val.c_str(), &fs)) == ERR_STATUS_GOOD)
   {
      label_style = fs;
      if (initialized)   rval = change_label_property();
   }
   else
      cmd_manager->add_err(rval=ERR_STR_SUBSET, id, 0, "fontstyle", val.c_str (), "normal|bold|italic|bolditalic");
   return rval;
}


int ChipsContour::set_tick_length(int val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val >= V_SIZE_MIN_VAL) && (val <= V_SIZE_MAX_VAL))
   {
      tick_len = val;
      if (initialized)  rval = change_tick_property();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "contour tick length", val,
                                  V_SIZE_MIN_VAL, V_SIZE_MAX_VAL);

   return rval;
}


int ChipsContour::set_tick_visibility(bool val)
{
   int rval = ERR_STATUS_GOOD;

   ticks = val;
   if (initialized)  rval = change_tick_property();
   return rval;
}


int ChipsContour::set_label_visibility(bool val)
{
   int rval = ERR_STATUS_GOOD;

   label = val;
   if (initialized)  rval = change_label_property();
   return rval;
}



int ChipsContour::get_x_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && xx)
      {
         *cnt = num_pts; 
         dat = new double[num_pts];
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory 
            for (ii = num_pts; ii--; )  dat[ii] = xx[ii]; 
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval; 
}


int ChipsContour::get_y_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && yy)
      {
         *cnt = num_pts; 
         dat = new double[num_pts];
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = yy[ii];
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval; 
}


int ChipsContour::get_z_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && zz)
      {
         *cnt = num_pts;
         dat = new double[num_pts];
         if (dat)
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = zz[ii];
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}


int ChipsContour::set_property(void* in_val, int in_property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   string save_current;
   if (parent && (parent->get_object_type() == chips_plot))
      save_current = ((ChipsPlot *)parent)->get_current_contour();

   switch(in_property)
   {
      case SET_CONTOUR_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_color(), clr, save_current);
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_STYLE:
      {
         string val = (const char*)in_val;
         string oval = v_line_to_string(style);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, save_current);
         if ((rval = set_style(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_ALGORITHM:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            (int)algorithm, *ival, save_current);
         if ((rval = set_algorithm((V_CONTOUR_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            thickness, *fval, save_current);
         if ((rval = set_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_MODE:
      {
         string val = (const char*)in_val;
         string oval = v_tickmode_to_string(mode);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, save_current);
         if ((rval = set_mode(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_TICK_VIS:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_tick_visibility((bool)*ival);  
      }
      break;
      case SET_CONTOUR_TICK_COLOR:
      {
         string clr = (const char*)in_val;
         set_tick_color(clr);
      }
      break;
      case SET_CONTOUR_TICK_THICK:
      {
         float* fval = static_cast<float*>(in_val);
         rval = set_tick_thickness(*fval);
      }
      break;
      case SET_CONTOUR_TICK_LENGTH:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_tick_length(*ival);
      }
      break;
      case SET_CONTOUR_TICK_DIR:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_tick_direction(*ival);
      }
      break;
      case SET_CONTOUR_LAB_VIS:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_label_visibility((bool)*ival);  
      }
      break;
      case SET_CONTOUR_LAB_COLOR:
      {
         string clr = (const char*)in_val;
         set_label_color(clr);
      }
      break;
      case SET_CONTOUR_LAB_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            label_size, *ival, save_current);
         if ((rval = set_label_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_LAB_FONT:
      {
         string font = (const char*)in_val;
         rval = set_label_font(font);
      }
      break;
      case SET_CONTOUR_LAB_FONTSTYLE:
      {
         string styl = (const char*)in_val;
         rval = set_label_fontstyle(styl);
      }
      break;
      case SET_CONTOUR_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "contour visibility", *ival, 0, 1);
            delete cmd;
         }
      }
      break;
      case SET_CONTOUR_INTERVAL:
      {
         double* dval = static_cast<double*>(in_val);
         double oval = interval;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *dval, save_current);
         if ((rval = set_interval(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
            {
               if (!add_command (cmd, rval))  delete cmd;
               // only set the mode when something has changed
               if ((oval != *dval) || (this->mode != chips_interval))
                  rval = set_property((void*)"interval", SET_CONTOUR_MODE,rec);
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_CONTOUR_NUMLEVELS:
      {
         int* ival = static_cast<int*>(in_val);
         int current_numlevels = get_numlevels ();
         V_TICKMODE_TYPE current_mode = get_mode ();
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            count, *ival, save_current);
         if ((rval = set_numlevels(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
            {
               if (!add_command (cmd, rval))  delete cmd;

               // only set the mode when something has changed
               if ((current_numlevels != *ival) || (current_mode != chips_count))
                  rval = set_property((void*)"count", SET_CONTOUR_MODE,rec);
            }
         }
         else  delete cmd;
      }
      break;
      case SET_CONTOUR_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH) 
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, in_property,
               oval, newval, save_current);
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, in_property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_CNTR);

   return rval;
}


// SCALE1 algorithm (C.R. Lewart, CACM, Oct 73, v16 n10)
// finds a new range to nicely divide data into approx n linear intervals
void ChipsContour::scale1(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist, double *base)
{
   double interv, scaled, fm, baseval, del = 0.0000000000000001;
   int nal, m;

   if (inmax == inmin)  ++inmax;
   if (!slices) slices = 5;     // avoid div by 0- use 5 as desired num ticks

   interv = fabs((inmax-inmin)/(double)slices);
   nal = (int) log10(interv);
   if (interv < 1) --nal;

   // scale interval between 1-10 and find value- e
   if ((scaled = interv/pow(10.0, nal)) < 1.41421356327)    baseval = 1.0;
      else if (scaled < 3.1622766017)                       baseval = 2.0;
         else  if (scaled < 7.07106781187)                  baseval = 5.0;
            else                                            baseval = 10;
   if (base)  *base = baseval;

   if ((*dist = copysign(baseval * pow(10.0, nal), (inmax - inmin))) != 0)
   {
      // compute interval size
      fm = inmin/(*dist);
      if (fm < 0.0)   m = (int)fm - 1;
         else   m = (int)fm;
      if (fabs((double)m+1.0-fm) < del)  ++m;

      // find new minimum and maximum
      *outmin = (*dist) * (double)m;
      fm = inmax/(*dist);
      if (fm < -1.0)  m = (int)fm;
         else m = (int)fm + 1;
      if (fabs(fm+1.0-(double)m) < del)  --m;
      *outmax = (*dist) * (double)m;

      // adjust for roundoff
      if (*outmin > inmin)   *outmin = inmin;
      if (*outmax < inmax)   *outmax = inmax;
   }
   else
   {
      *outmin = inmin;
      *outmax = inmax;
   }
}

// SCALE2 algorithm (C.R. Lewart, CACM, Oct 73, v16 n10)
// finds a new range to nicely divide data into exactly n linear intervals
void ChipsContour::scale2(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist)
{
   double interv, scaled, fm, del = 0.0000000000000001;
   double vint[5] = { 1.0, 2.0, 5.0, 10.0, 20.0 };
   int nal, m1, m2, nx, np, done = 0, ii = 1;

   interv = fabs((inmax-inmin)/(double)slices);
   nal = (int) log10(interv);
   if (interv < 1)   --nal;
   scaled = interv/pow(10.0, nal);
   while (!done && (ii < 6))
   {
      if ((scaled < vint[ii - 1] + del) || (ii > 3))
      {
         *dist = vint[ii - 1] * pow(10.0, nal);
         m1 = (int)(fm = inmin/(*dist));
         if (fabs((double)m1 + 1.0 - fm) < del) ++m1;
         *outmin = *dist * m1;
         fm = inmax/(*dist);
         m2 = (int)(fm + 1);
         if (fabs(fm + 1.0 -(double)m2) < del)  --m2;
         *outmax = (*dist) * m2;
         if ((np = m2 - m1) <= slices) done = 1;
      }
      ++ii;
   }
   nx = (slices - np) / 2;
   *outmin -= (double) nx * (*dist);
   *outmax = *outmin + (double)slices * (*dist);
   if (*outmin > inmin)   *outmin = inmin;
   if (*outmax < inmax)   *outmax = inmax;
}


// check if data forms a regular grid - N x M array where all N axis entries
// are equi-distant and all M axis entries are equi-distant and no missing 
// points in the already sorted data array.  
bool ChipsContour::regular_grid()
{
   bool valid = false;

   if (xx && yy && zz && (num_pts > 1))
   {
      unsigned int rr = 1;

      // need to handle either row or column precedence... 
      if ((x_spacing = xx[1] - xx[0]) > 0)
      {
         x_gridsize = (int)((xmax - xmin) / x_spacing + 1.5);

         if (((unsigned int)x_gridsize < num_pts) && 
             ((y_spacing = yy[x_gridsize] - yy[0]) > 0))
         {
            y_gridsize = (int)((ymax - ymin) / y_spacing + 1.5);

            // make sure we have the right number of points 
            if ((x_gridsize * y_gridsize) == (int)num_pts)
            {
               int rw = 0; 
               rr = 0;
               valid = true;
               double dist = yy[0];
               // verify y positions are regular
               while (valid && (rr < num_pts))
               {
		  valid = approx_equal(yy[rr], dist, .001);
                  if (rr && (((rr+1)%x_gridsize) == 0) && ((rr+1) < num_pts))
                  {
                     rw++;
                     dist = yy[rr+1] - yy[rr];
		     if (approx_equal(dist, y_spacing, .001))
                        dist = yy[rr+1];
                     else 
		        dist = yy[rr] + y_spacing;
                  }
                  rr++;
               }
               rr = 1;
               dist = xx[0];
               // verify x positions are regular
               while (valid && (rr < num_pts))
               {
		  valid =  approx_equal((xx[rr] - dist), x_spacing, .001);
                  if (((rr+1) % x_gridsize) == 0) dist = xx[0] - x_spacing;
                     else dist = xx[rr]; 
                  rr++;
               }
            }
         }
      }
      else if ((y_spacing = yy[1] - yy[0]) > 0)
      {
         y_gridsize = (int)((ymax - ymin) / y_spacing + 1.5);

         if (((unsigned int)y_gridsize < num_pts) &&
             ((x_spacing = xx[y_gridsize] - xx[0]) > 0))
         {
            x_gridsize = (int)((xmax - xmin) / x_spacing + 1.5);

            // make sure we have the right number of points
            if ((unsigned int)(y_gridsize * x_gridsize) == num_pts)
            {
               int rw = 0;
               rr = 0;
               valid = true;
               double dist = xx[0];
               // verify x positions are regular
               while (valid && (rr < num_pts))
               {
		  valid = approx_equal(xx[rr], dist, .001);
                  // adjust expected distance value if new row
                  if (rr && (((rr+1)%y_gridsize) == 0))
                  {
                     rw++;
                     dist = xx[rr+1] - xx[rr];
		     if (approx_equal(dist, x_spacing, .001))
                        dist = xx[rr+1];
                     else
		        dist = xx[rr] + x_spacing;
                  }
                  rr++;
               }
               rr = 1;
               dist = yy[0];
               // verify y positions are regular
               while (valid && (rr < num_pts))
               {
		  valid = approx_equal((yy[rr] - dist), y_spacing, .001);
                  if (((rr+1) % y_gridsize) == 0) dist = yy[0] - y_spacing;
                     else dist = yy[rr];
                  rr++;
               }
            }
         }
      }
   }
   return valid;
}


int ChipsContour::change_tick_property()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_label_property()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_style()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_levels()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::change_interval()
{
   int rval = ERR_STATUS_GOOD;

   // workaround- til middle layer implements change_interval
   if ((rval = change_levels()) == ERR_STATUS_GOOD)
        rval = change_color();
   return rval;
}

int ChipsContour::update_data ()
{
   return ERR_FEATURE_DNE;
}

int ChipsContour::set_plot_location (double x1, double y1, double x2, double y2)
{
   return ERR_FEATURE_DNE;
}  


string ChipsContour::get_xaxis()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA))
   {
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_xaxis_id ();
   } 
   return "";
}

string ChipsContour::get_yaxis()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA))
   {
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_yaxis_id ();
   } 
   return "";
}


int ChipsContour::change_algorithm()
{
   return 0;
}


int ChipsContour::change_range (const ChipsVector &low_range,const  ChipsVector &high_range)
{
   return ERR_FEATURE_DNE; 
};

int ChipsContour::add_transform (Transform *transform, bool inverse)
{
   return ERR_FEATURE_DNE;
};

int ChipsContour::remove_transforms ()
{
   return ERR_FEATURE_DNE;
};


int ChipsContour::get_levels (vector<double> &levs)
{
   return ERR_FEATURE_DNE;
};



void ChipsContour::serialize(ChipsStateStream& out)
{
   string tmp; 
   int sz;
   
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   
   // style, color, thickness
   sz = (int)style;
   out.write_int(&sz);
   tmp = get_color();
   out.write_string(&tmp);
   out.write_float(&thickness);
   sz = (int)algorithm;
   out.write_int(&sz);
   out.write_int(&depth);

   // tick stuff
   out.write_bool(&regular);
   out.write_bool(&initialized);
   out.write_bool(&ticks);
   out.write_int(&tick_dir);
   out.write_int(&tick_len);
   tmp = tick_color.get_name();
   out.write_string(&tmp);
   out.write_float(&tick_thick);
   sz = (int) mode;
   out.write_int(&sz);

   out.write_int((int*)&num_pts);
   out.write_double(xx, (int)num_pts);
   out.write_double(yy, (int)num_pts);
   out.write_double(zz, (int)num_pts);

   out.write_int((int*)&num_levels);
   out.write_double(levels, (int)num_levels);

   out.write_int((int*)&arb_num);
   out.write_double(arb_levels, (int)arb_num);
   out.write_double(&interval);
   out.write_int(&count); 
   out.write_bool (&visible);
   out.write_string (&data_system_id);
   sz = (int)coord_system;
   out.write_int (&sz);

   bool trans_exist;

   if (out.get_minor_version() == 0)
   {

      if (transform)
      {
         trans_exist = true;
         out.write_bool (&trans_exist);
         ChipsTransformFactory::serialize (transform, out);
      }
      else
      {
         trans_exist = false;
         out.write_bool (&trans_exist);
      }
   }
   else
      out.write_string (&transform_id);
}

void ChipsContour::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];
   const char* ccp;

   tmp = get_color();
   if (theDefaults->get_contourcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("color=");
      att.append(tmp);
   }
   if (theDefaults->get_contourstyle() != style)
   {
      if (att.length()) att.append(" ");
      att.append("style=");
      att.append(v_line_to_string(style));
   }
   if (theDefaults->get_contourthickness() != thickness)
   {
      if (att.length()) att.append(" ");
      if (thickness != 0.5)
      {
         sprintf(buf,"thickness=%-1.0f", thickness);
         att.append(buf);
      }
      else
         att.append("thickness=0.5");
   }
   if ((algorithm != theDefaults->get_contouralgorithm()) &&
       ((ccp = v_contour_to_string(algorithm)) != NULL))
   {
      if (att.length()) att.append(" ");
      att.append("algorithm=");
      att.append(ccp);
   }
   int dep = theDefaults->get_contourdepth();
   if (depth != dep &&
       (dep != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   string dfname = out.use_tempname();
   string mfile = "";
   string transform_desc = "";

   if (meta.length())
   {
      mfile = dfname;
      mfile.append("=\"");
      mfile.append(meta);
      mfile.append("\"");
      out << "add_contour("<<dfname;
   }
   else
   {
      string data_sys_id, coord_transform_id;
      string transform_fname = "";
      string transform_obj = "";
      ChipsDataSystem *data_sys;
      string tmpfile = out.get_filename();
      string base_filename = tmpfile;
      sprintf(buf, ".Data%d_contour", out.use_tempfnum());
      base_filename.append(buf);


      // determine transform name
      data_sys = static_cast <ChipsDataSystem *>(coord);
      data_sys_id = data_sys->get_id_no_wcs ();

      coord_transform_id = data_sys->get_default_transform_id ();

      // if the data system has no default transform is the same as
      // the incoming transform then the position is already in data
      // coordinates
      if (transform_id.size () > 0)
      {
         // set the transform file name
         transform_fname = base_filename + "_transform.txt";

         // write the transform to a ascii file
         fstream transform_file(transform_fname.c_str(), ios::out);
         Transform *transform = data_sys->get_transform (transform_id);
         ChipsTransformFactory::serialize_ascii (transform, transform_file);

         transform_obj = dfname + ".transform";
         transform_desc = dfname + ".transform = #  insert transform here - " + transform_fname + "\n";

      }

      tmpfile = base_filename+".bin";
      mfile = dfname;
      mfile.append(" = DataObj()  # data- ");
      mfile.append(tmpfile);
      mfile.append("\n");
      mfile.append(dfname);
      mfile.append(".x = #  insert data array here \n");
      mfile.append(dfname);
      mfile.append(".y = #  insert data array here \n");
      mfile.append(dfname);
      mfile.append(".z = #  insert data array here \n");

      ofstream ff(tmpfile.c_str(), ios::out|ios::binary);
      if (ff)
      {
         // table data format:
         //  [type (0= table) ][ numcols         ]
         //  [number of rows  ][ colname buff size]
         //  [ char array '\0' separated col names ]
         //  [ double array col1 data .. numcols]
         char buffer[1024];
         long bpos = 0, ncols = 0;
         string colnames = "";
         ff.write((char*)&ncols, sizeof(long));               // table format
         if (xx)    { strcpy(&buffer[bpos], "x\0"); bpos+=2; ncols++; }
         if (yy)    { strcpy(&buffer[bpos], "y\0"); bpos+=2; ncols++; }
         if (zz)    { strcpy(&buffer[bpos], "z\0"); bpos+=2; ncols++; }
         // number of doubles in buffer- add 1 less than sizeof double for
         // double word aligning and subtract 1 for extra \0 insertion
         bpos = (bpos + sizeof(double) - 2) / sizeof(double);
         ff.write((char*)&ncols, sizeof(long));               // num cols
         ff.write((char*)&num_pts, sizeof(unsigned long));    // num rows
         ff.write((char*)&bpos, sizeof(long));                // colname sz
         ff.write((char*)buffer, bpos* sizeof(double));       // colnames

         if (xx)    ff.write((char*)xx, num_pts* sizeof(double));
         if (yy)    ff.write((char*)yy, num_pts* sizeof(double));
         if (zz)    ff.write((char*)zz, num_pts* sizeof(double));
         ff.close();
      }

      out << "add_contour("<<dfname<<".x, "<<dfname<<".y, "<<dfname<<".z";
   
      if (transform_obj.length () > 0) out << ", " << transform_obj;
   }
   if (att.length()) out <<",'"<<att<<"'";
   out <<")"<<endl;

   // add the transform last
   mfile.append (transform_desc);
   out.append_to_template(mfile);

   if (num_levels && levels)
   {
      out<< "set_contour('levels=[";
      for (unsigned int rr = 0; rr < num_levels; rr++)
      {
         stringstream ostr;
         ostr << levels[rr];
         out << ostr.str();
         if (rr < num_levels-1)  out << ",";
      }
      out <<"]')"<<endl;
   }
}

void ChipsContour::deserialize(ChipsStateStream& in)
{
   string axid, cname;
   int len;

   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_int(&len);
   style = (V_LINE_TYPE)len;
   in.read_string(&cname);
   set_color(cname); 
   in.read_float(&thickness);
   in.read_int(&len);
   algorithm =  (V_CONTOUR_TYPE)len;
   in.read_int(&depth);
   in.read_bool(&regular); 
   in.read_bool(&initialized);
   in.read_bool(&ticks); 
   in.read_int(&tick_dir); 
   in.read_int(&tick_len); 
   in.read_string(&cname);
   set_tick_color(cname); 
   in.read_float(&tick_thick);
   in.read_int(&len);
   mode = (V_TICKMODE_TYPE)len;

   in.read_int((int*)&num_pts);
   xx = new double[num_pts];
   yy = new double[num_pts];
   zz = new double[num_pts];
   in.read_double(xx, num_pts); 
   in.read_double(yy, num_pts); 
   in.read_double(zz, num_pts); 

   in.read_int((int*)&num_levels);
   levels = new double[num_levels];
   in.read_double(levels, (int)num_levels);
   in.read_int((int*)&arb_num);
   arb_levels = new double[arb_num];
   in.read_double(arb_levels, (int)arb_num);

   in.read_double(&interval); 
   in.read_int(&count);
   in.read_bool (&visible);
   in.read_string (&data_system_id);
   in.read_int (&len);
   coord_system = (COORD_SYS)len;

   if (in.get_minor_version() == 0)
   {
      bool trans_exist;
      in.read_bool (&trans_exist);
      if (trans_exist)
      {
         transform = ChipsTransformFactory::deserialize (in);
      }
   }
   else
      in.read_string (&transform_id);
}

int ChipsContour::show()
{
  int rval = ERR_STATUS_GOOD;
  visible = true;
  rval = change_visible ();
  return rval;
};

int ChipsContour::hide()
{
  int rval = ERR_STATUS_GOOD;
  visible = false;
  rval = change_visible ();
  return rval;
};


bool ChipsContour::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED) outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}

double ChipsContour::get_xmin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, "");
      }
      return out[0];
   }
   else
      return xmin;
}

double ChipsContour::get_xmax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, "");
      }
      return out[0];
   }
   else
      return xmax;

}

double ChipsContour::get_ymin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, "");
      }
      return out[1];
   }
   else
      return ymin;

}

double ChipsContour::get_ymax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, "");
      }

      return out[1];
   }
   else
      return ymax;
}

void ChipsContour::set_transform_id (const string &id)
{
   this->transform_id = id;
   this->transform_applied = false;
}

