// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSCONTOUR_HH
#define CHIPSCONTOUR_HH

#include <string>
#include <stdio.h>
#include <vector>
#include "ChipsObject.hh"
#include "ChipsDefaults.hh"
#include "ChipsAxis.hh"
#include "ChipsColor.hh"
#include "ChipsStateStream.hh"
#include "ChipsCommand.hh"

using namespace std;

class Transform;

class ChipsContour: public ChipsObject
{
  public:

   ChipsContour(ChipsObject* par, unsigned long cnt, double* x, double* y, 
             double* z, string transfrm_id, string cid);
   ChipsContour(ChipsObject* par, unsigned long cnt, double* x, double* y, 
             double* z, unsigned int numlev, double *levels, string transfrm_id,
             string cid);
   ChipsContour(ChipsObject* par, ChipsStateStream& in);

   void       object_delete ();

   void       initialize(); 

   int        set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj);
   int        set_property(void* in_val, int property, bool rec=true);

   string     get_color()        { return color.get_name(); }
   int        set_color(string in_color);
   float      get_thickness()    { return thickness; }
   int        set_thickness(float val); 
   V_LINE_TYPE get_style()        { return style; }
   int        set_style(V_LINE_TYPE val); 
   int        set_style(string val);
   int        set_mode(string val); 
   int        set_mode(int val); 
   V_TICKMODE_TYPE get_mode()     { return mode; } 
   int        set_levels(unsigned int sz, double* levs, bool rec=true);
   int        set_algorithm(V_CONTOUR_TYPE val);
   V_CONTOUR_TYPE get_algorithm()  { return algorithm; }
   int        set_tick_visibility(bool val);
   int        set_tick_color(string val);
   int        set_tick_length(int val); 
   int        set_tick_thickness(float val); 
   int        set_tick_direction(int val); 

   int        set_label_visibility(bool val);
   int        set_label_color(string val);
   int        set_label_size(int val);
   int        set_label_font(string val);
   int        set_label_fontstyle(string val);


   unsigned long get_cnt()               { return num_pts; }
   double     get_xmin(bool transformed);
   double     get_xmax(bool transformed);
   double     get_ymin(bool transformed);
   double     get_ymax(bool transformed);
   double     get_zmin()                 { return zmin; }
   double     get_zmax()                 { return zmax; }
   double     get_xlmin()                { return xlmin;}
   double     get_ylmin()                { return ylmin;}
   double     get_zlmin()                { return zlmin;}

   // grid related methods 
   double     get_x_spacing()            { return x_spacing;  }
   double     get_y_spacing()            { return y_spacing;  }
   int        get_x_gridsize()           { return x_gridsize; }
   int        get_y_gridsize()           { return y_gridsize; }
   bool       is_regular()               { return regular;    } 

   int        get_x_data(double* dat, unsigned long* cnt);
   int        get_y_data(double* dat, unsigned long* cnt);
   int        get_z_data(double* dat, unsigned long* cnt);

   string     get_xaxis();
   string     get_yaxis();

   void       set_contour_num (int val) {contour_num = val;};
   int        get_contour_num () {return contour_num;};
   int        set_interval(double val);
   double     get_interval() {return interval;};
   int        set_numlevels(int val); 
   int        get_numlevels() { return count; }


   // transform functions
   int set_transform (Transform *trans, bool rec=true);
   int apply_transform (bool rec=true);
   int apply_transform_inverse (bool rec=true);
   void set_transform_id (const string &id);


   // keep for backwards compatibility with chips 4.1
   int get_transform (Transform **trans);
   
//   virtual int apply_transform ();
   virtual int reset_position ();
   virtual int  set_depth(int newdepth);

   virtual int change_algorithm();
   virtual int change_levels(); 
   virtual int change_tick_property();
   virtual int change_label_property(); 
   virtual int change_thickness(); 
   virtual int change_style(); 
   virtual int change_color(); 
   virtual int change_visible();
   virtual int change_interval();
   virtual int change_depth();
   virtual int update_data ();
   virtual int set_plot_location (double x1, double y1, double x2, double y2);
   virtual int change_range (const ChipsVector &low_range,const  ChipsVector &high_range);
   virtual int get_levels (vector<double> &levs);


   virtual void serialize(ChipsStateStream& out);
   virtual void serialize_ascii(ChipsStateStream& out);
   virtual void deserialize(ChipsStateStream& in);

   virtual int show();
   virtual int hide();
   
  protected:

   virtual ~ChipsContour();

   // methods to compute 'nice' intervals (C.R. Lewart, CACM, Oct 73,v16 n10)
   void       scale1(double min, double max, int n, double* outmin,
                     double* outmax, double* dist, double* base=NULL);
   void       scale2(double min, double max, int n, double* outmin,
                     double* outmax, double* dist);

   int        compute_levels(); 
   void       calculate_limits();

   // check if the input data is a regular grid 
   bool       regular_grid(); 

   virtual int add_transform (Transform *transform, bool inverse=false);
   virtual int remove_transforms ();


   unsigned long num_pts;    // number of points in curve 
   unsigned long x_pts;      // number of points in x 
   unsigned long y_pts;      // number of points in y 
   double*    xx;            // x column data
   double*    yy;            // y column data
   double*    zz;            // z column data

   double     crpix[2];      // ref point along axis 
   double     crval[2];      // value at ref point
   double     crdelt[2];     // delta from ref point
 
   int        x_gridsize;    // x dimension grid size 
   int        y_gridsize;    // y dimension grid size 
   double     x_spacing;     // distance between consecutive x points 
   double     y_spacing;     // distance between consecutive y points 

   double     xmin;          // x minimum value 
   double     xmax;          // x maximum value 
   double     ymin;          // y minimum value 
   double     ymax;          // y maximum value 
   double     zmin;          // z minimum value 
   double     zmax;          // z maximum value 
   double     xlmin;         // x minimum value for log scale 
   double     ylmin;         // x minimum value for log scale 
   double     zlmin;         // x minimum value for log scale 

   unsigned int num_levels;  // number of contour levels
   double*    levels;        // array of level values 

   unsigned int arb_num;     // number of contour levels for arbitrary mode
   double*    arb_levels;    // array of level values for arbitrary mode 
   double     interval;      // interval to use for interval mode 
   int        count;         // count to use for count mode

   ChipsAxis*  xaxis;        // associated x axis 
   ChipsAxis*  yaxis;        // associated x axis 

   bool       regular;       // regular or irregular data 
   bool       initialized;   // have we initialized everything
   bool       ticks;         // draw down gradient tick marks 
   int        tick_dir;      // tick direction 0 = down gradient, 1 = up
   int        tick_len;      // length of tickmark
   ChipsColor tick_color;    // color of tickmark
   float      tick_thick;    // thickness of tickmark 

   bool       label;         // do we want to label contours
   int        label_size;    // size of contour labels 
   ChipsColor label_color;   // contour label color
   string     label_font;    // font of label
   V_FONTSTYLE_TYPE label_style;  // normal, bold, italic
 
   V_TICKMODE_TYPE mode;     // contour mode (equivalent to tickmode)    
   V_CONTOUR_TYPE  algorithm; // algorithm to use for contouring
   ChipsColor  def_color;    // default contour level color 
   V_LINE_TYPE def_style;    // default contour level style
   float       def_thickness;// default contour level line thickness

   // for now use scalars but these eventually should be arrays 
   // so that we can support colors for independent levels 
   ChipsColor  color;        // colors for levels 
   V_LINE_TYPE style;        // styles for levels 
   float       thickness;    // thickness value for levels 
   int        contour_num;       // the contour's order 0th, 1st, 2nd, etc.

   Transform *transform; // Keep for backwards compatibility for state files

   string     transform_id;  // the transform id associated with this image
   string     meta;          // meta data for contour (ie. file info)
   bool transform_applied;

  private:

   bool add_command (ChipsCommand *inCmd, int &outStatus);

};


#endif 
