// 
//  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsCurve.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"
#include "ChipsTransformFactory.hh"

#include <stdio.h>
#include <config.h>

#ifdef HAVE_IEEEFP_H
#include <ieeefp.h>        // finite defined in ieeefp.h (solaris)
#else
#include <math.h>          // finite() defined here      (linux, OSX)
#endif

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;

ChipsCurve::ChipsCurve(ChipsObject* par, unsigned long cnt, double* x, 
   double* y, const string &transform_id, string val, bool compute):ChipsObject(par, val)
{
   unsigned int rr; 
   initialized = false;
   id = val;
   num_pts = 0; 
   xx = yy = xup = xdown = yup = ydown = xlim = ylim = NULL;
   if (x) xx = (double*)malloc(cnt*sizeof(double));
   if (y) yy = (double*)malloc(cnt*sizeof(double));
   if (x && y)
   {
      for (rr = 0; rr < cnt; rr++)
      {
         if (finite(x[rr]) && finite(y[rr]))
         {
            xx[num_pts] = x[rr]; 
            yy[num_pts++] = y[rr];
         }
      }
   }
   else if (x)
   {
      for (rr = 0; rr < cnt; rr++)
         if (finite(x[rr]))   xx[num_pts++] = x[rr];
   }
   else if (y)
   {
      for (rr = 0; rr < cnt; rr++)
         if (finite(y[rr]))   yy[num_pts++] = y[rr];
   }

   this->transform_id = transform_id;
   
   initialize();
   if (compute)  calculate_limits();
}


ChipsCurve::ChipsCurve(ChipsObject* par, ChipsStateStream& in):ChipsObject(par, "")
{
   initialized = false;
   num_pts = 0;

   xx = yy = xup = xdown = yup = ydown = xlim = ylim = NULL;
   initialize();
   deserialize(in);
   calculate_limits();
}

ChipsCurve::ChipsCurve(ChipsObject* par, unsigned long cnt, double* x, 
   double* y, double* yd, double* yu, double* xd, double* xu,
   double* ylims, double* xlims, const string &transform_id, string val, 
   bool compute): ChipsObject(par, val)
{
   initialized = false;
   id = val;
   num_pts = 0; 
  
   // allocate memory and copy data
   xx = yy = xup = xdown = yup = ydown = ylim = xlim = NULL; 
   if (x) xx = (double*)malloc(cnt*sizeof(double));
   if (y) yy = (double*)malloc(cnt*sizeof(double));
   if (xu) xup = (double*)malloc(cnt*sizeof(double));
   if (xd) xdown = (double*)malloc(cnt*sizeof(double));
   if (yu || yd) yup = (double*)malloc(cnt*sizeof(double));
   if (yd) ydown = (double*)malloc(cnt*sizeof(double));
   if (ylims) ylim = (double*)malloc(cnt*sizeof(double));
   if (xlims) xlim = (double*)malloc(cnt*sizeof(double));
   if ((xx || (x == NULL)) && (yy || (y == NULL)) && 
       (xup || (xu == NULL)) && (xdown || (xd == NULL)) &&
       (yup || ((yu == NULL) && (yd == NULL))) && (ydown || (yd == NULL)))
   {
      int rr;
      num_pts = cnt;
      for (rr = cnt; rr--; )
      {
         if (x) xx[rr] = x[rr];
         if (y) yy[rr] = y[rr];
         // for now, treat NaNs in errs as zero
         if (xu)  xup[rr] = (finite(xu[rr])) ? xu[rr] : 0.0;
         if (xd)  xdown[rr] = (finite(xd[rr])) ? xd[rr] : 0.0;
         if (yu)
            yup[rr] = (finite(yu[rr])) ? yu[rr] : 0.0;
         else if (yd)
            yup[rr] = (finite(yd[rr])) ? yd[rr] : 0.0;
         if (yd)  ydown[rr] = (finite(yd[rr])) ? yd[rr] : 0.0;
         if (ylim)  ylim[rr] = (finite(ylims[rr])) ? ylims[rr] : 0.0;
         if (xlim)  xlim[rr] = (finite(xlims[rr])) ? xlims[rr] : 0.0;
      }
   }

   this->transform_id = transform_id;

   initialize();
   if (compute)  calculate_limits();
}


ChipsCurve::~ChipsCurve()
{
   if (xx)
   {
      free(xx);
      xx = NULL;
   }
   if (yy)
   {
      free(yy);
      yy = NULL;
   }
   if (xup)
   {
      free(xup);
      xup = NULL;
   }
   if (xdown)
   {
      free(xdown);
      xdown = NULL;
   }
   if (yup)
   {
      free(yup);
      yup = NULL;
   }
   if (ydown)
   {
      free(ydown);
      ydown = NULL;
   }
   list<CHIPS_RANGE_P>::iterator ii;
   for(ii=indices.begin(); ii != indices.end(); ++ii)
      delete((*ii));
   
   if (coord)
      coord->remove (this);
}

void ChipsCurve::object_delete ()
{
   delete this;
}


void ChipsCurve::initialize()
{
   if (!initialized)
   {
      int sz; 
      V_BADDATAMODE_TYPE bmode; 
      if (theDefaults)
      {
         string temp;
   
         symbolstyle = theDefaults->get_curvesymbolstyle();
         sz = theDefaults->get_curvesymbolsize();
         symbolangle = theDefaults->get_curvesymbolangle();
         symbolfilled = theDefaults->curvesymbol_filled();
         temp = theDefaults->get_curvesymbolcolor();
         if (set_symbolcolor(temp) != 0)   set_symbolcolor("blue");
   
         linestyle   = theDefaults->get_curvelinestyle();
         linethickness = theDefaults->get_curvelinethickness();
         temp = theDefaults->get_curvelinecolor();
         if (set_linecolor(temp) != 0)   set_linecolor("white");
         errup = ((yup || ylim) && theDefaults->get_errup()); 
         errdown = ((ydown || ylim) && theDefaults->get_errdown()); 
         errleft = ((xdown || xlim)  && theDefaults->get_errleft()); 
         errright = ((xup || xlim)  &&  theDefaults->get_errright()); 
         errstyle = theDefaults->get_errstyle(); 
         errthickness = theDefaults->get_errthickness(); 
         errcaplen = theDefaults->get_errcaplength(); 
         limoverride = theDefaults->get_limitoverride();
         limlength = theDefaults->get_limitlength();
         temp = theDefaults->get_errcolor();
         bmode = theDefaults->get_curvebaddatamode(); 
	 depth = theDefaults->get_curvedepth();
         meta = theDefaults->get_meta();
         if (set_errcolor(temp) != 0)   set_errcolor("red");
      }
      else 
      {
         symbolstyle = chips_cross;
         sz = 3; 
         symbolangle = 0.0;
         set_symbolcolor("blue");
         symbolfilled = true; 
         linestyle = chips_solid;
         set_linecolor("white");
         linethickness = 1; 
         errstyle = chips_bar;
         errthickness = 1; 
         errup = (yup || ylim);
         errdown = (ydown|| ylim);
         errleft = (xdown|| xlim);
         errright = (xup || xlim);
         errcaplen = 10; 
         set_errcolor("red");
         bmode = chips_omit; 
         limoverride = 1;
         limlength = 0.05;
         depth = V_USE_DEFAULT_DEPTH;
         meta = ""; 
      }
      // if depth is default (0) - get 'current' depth from depth table
      if (depth==V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
      display_mask = REG_INTERVAL;
      set_baddatamode(bmode);

      num_disc = 0; 
      set_symbolsize(sz);

      curve_num = 0;
      object_type = chips_curve;

      transform_applied = false;
      initialized = true;
   }
}

void ChipsCurve::calculate_limits()
{
   // determine the limits as well as good/bad intervals for the data
   if (xx && yy)
   {
      unsigned int rr;
      
      if (finite(yy[0]) && finite(xx[0]))
      {
         if (ydown)
	 {
            ylmin = ymin = ymax = yy[0] -ydown[0];
	    if (ylmin <= 0)  ylmin = yy[0];
	    if ((ylmin <= 0) && (yup && ((yy[0]+yup[0])>0))) ylmin = yy[0] + yup[0];
	 }
         else
            ylmin = ymin = ymax = yy[0];
         if (xdown)
	 {
            xlmin = xmin = xmax = xx[0] - xdown[0];
	    if (xlmin <= 0)  xlmin = xx[0];
	    if ((xlmin <= 0) && (xup && ((xx[0]+xup[0])>0))) xlmin = xx[0] + xup[0];
	 }
         else
            xlmin = xmin = xmax = xx[0];
      }
      else
      {
         xlmin = xmin = ylmin = ymin = FLT_MAX;
	 xmax  =  ymax = FLT_MIN;
      }

      // remove bad data segment lists- since we'll recalculate them 
      list<CHIPS_RANGE_P>::iterator ii;
      for(ii=indices.begin(); ii != indices.end(); ++ii)
         delete((*ii));
      indices.clear();

      // calculate min/max of data- if errs provided, use them in calculations
      if (xup || xdown || yup || ydown)
      {
         double tmp; 
         CHIPS_RANGE_P seg = NULL;
         int badval;

         // create an index list which contains subscripts of valid/nan intervals
         for (rr = 0; rr < num_pts; rr++)
         {
            badval = REG_INTERVAL;
	    if (finite(xx[rr]) && finite(yy[rr]))
	    {
               if (xdown)
               {
                  if ((tmp=(xx[rr]-xdown[rr])) > 0)  
		  {
		     if ((tmp < xlmin) || (xlmin <= 0))  xlmin = tmp;
		  }
		  else if (xx[rr] > 0) 
		  {
		     if ((xx[rr] < xlmin) || (xlmin <= 0))  xlmin = xx[rr];
		  }
		  if (tmp < xmin)  xmin = tmp;
               }
               else
               {
                  if ((xx[rr] > 0) && ((xx[rr] < xlmin) || (xlmin <= 0.0)))
                     xlmin = xx[rr];
                  if (xx[rr] < xmin)  xmin = xx[rr];
               }
               if (xup)
               {
                  if ((xx[rr] + xup[rr]) > xmax)  xmax = xx[rr] + xup[rr]; 
               }
               else
               {
                  if (xx[rr] > xmax)  xmax = xx[rr]; 
               }
               if (ydown)
               {
                  if ((tmp=(yy[rr]-ydown[rr])) > 0) 
		  { 
                     if ((tmp < ylmin) || (ylmin <= 0))  ylmin = tmp;
		  }
		  else if (yy[rr] > 0) 
		  {
		     if ((yy[rr] < ylmin) || (ylmin <= 0))  ylmin = yy[rr];
		  }
                  if (tmp < ymin)  ymin = yy[rr] - ydown[rr];
               }
               else
               {
                  if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin <= 0)))
                     ylmin = yy[rr];
                  if (yy[rr] < ymin)  ymin = yy[rr];
               }
               if (yup)
               {
                  if ((yy[rr] + yup[rr]) > ymax)  ymax = yy[rr] + yup[rr];
               }
               else
               {
                  if (yy[rr] > ymax)  ymax = yy[rr];
               }
            }
            else  badval = NAN_INTERVAL;  
               
            if (seg)
            {
               // the mode (valid/nan) has changed- store off the indices
               if (seg->mask != badval)
               {
                  if (badval == NAN_INTERVAL)
                  {
                     seg->hi = rr - 1;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr - 1;
                     seg->mask = NAN_INTERVAL;
                     seg->inclusive = 0;
                  }
                  else
                  {
                     seg->hi = rr;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr;
                     seg->mask = REG_INTERVAL;
                     seg->inclusive = 1;
                  }
               }
            }
            else if (badval != NAN_INTERVAL)
            {
               seg = new CHIPS_RANGE;
               seg->lo = rr;
               seg->mask = REG_INTERVAL;
               seg->inclusive = 1;
            }
         }
         if (seg)
         {
            seg->hi = num_pts -1;
            indices.push_back(seg);
         }
      }
      else
      {
         CHIPS_RANGE_P seg = NULL;
         int badval;

         for (rr = 0; rr < num_pts; rr++)
         {
            badval = REG_INTERVAL;

	    if (finite(xx[rr]) && finite(yy[rr]))
	    {
               if (xx[rr] < xmin)  xmin = xx[rr]; 
               if (xx[rr] > xmax)  xmax = xx[rr]; 
               if (yy[rr] < ymin)  ymin = yy[rr]; 
               if (yy[rr] > ymax)  ymax = yy[rr]; 
               // store smallest positive value as log minimum
               if ((xx[rr] > 0) && ((xx[rr] < xlmin) || (xlmin <=0)))
                  xlmin = xx[rr];
               if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin <=0))) 
                  ylmin = yy[rr];
            }
            else  badval = NAN_INTERVAL;

            if (seg)
            {
               if (seg->mask != badval)
               {
                  // the mode (valid/nan) has changed- store off the indices
                  if (badval == NAN_INTERVAL)
                  {
                     seg->hi = rr - 1;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr - 1;
                     seg->mask = NAN_INTERVAL;
                     seg->inclusive = 0;
                  }
                  else
                  {
                     seg->hi = rr;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr;
                     seg->mask = REG_INTERVAL;
                     seg->inclusive = 1;
                  }
               }
            }
            else if (badval != NAN_INTERVAL)
            {
               seg = new CHIPS_RANGE;
               seg->lo = rr;
               seg->mask = REG_INTERVAL;
               seg->inclusive = 1;
            }
         }
         if (seg)
         {
            seg->hi = num_pts -1;
            indices.push_back(seg);
         }
      } 
   }
   if ((xlmin == FLT_MAX) && (xmin == FLT_MAX) && (xmax == FLT_MIN))
   {
      xlmin = xmin = xmax = 0; 
   }
   if ((ylmin == FLT_MAX) && (ymin == FLT_MAX) && (ymax == FLT_MIN))
   {
      ylmin = ymin = ymax = 0; 
   }
}


string ChipsCurve::get_xaxis ()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA)){
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_xaxis_id ();
   }

   return "";
}


string ChipsCurve::get_yaxis ()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA)){
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_yaxis_id ();
   }

   return "";
}


// This recalculates the ranges after a coordinate system is updated
int ChipsCurve::reset_position ()
{
   int rval = ERR_STATUS_GOOD;
   ChipsDataSystem *data_sys;
   if (!coord || (coord_system != DATA))
      return ERR_STATUS_BAD;

   data_sys = static_cast <ChipsDataSystem *>(coord);

   if (!transform_applied)
   {
      string coord_transform_id = data_sys->get_default_transform_id  ();
      string data_sys_id = data_sys->get_id_no_wcs ();
      if (coord_transform_id.size () > 0)
      {
         ChipsVector *coords = new ChipsVector[num_pts];

         for (unsigned int ii=0; ii<num_pts; ii++)
            coords[ii] = ChipsVector (xx[ii], yy[ii]);

         // convert the points from local coord to default coord
         if ((transform_id.size () > 0) && (coord_transform_id != transform_id))
         {
            string tmp_coord_world = data_sys->get_id (transform_id);
            coord->convert_cs_to_cs (coords, coords, num_pts, DATA, data_sys_id, DATA, tmp_coord_world);
            coord->convert_cs_to_cs (coords, coords, num_pts, DATA, "", DATA, data_sys_id);

         }
         // go through each of the data points and adjust
         // apply the transforms inverse
         else if (transform_id.size () == 0)
         {
            coord->convert_cs_to_cs (coords, coords, num_pts, DATA, "", DATA, data_sys_id);        
            transform_id =  coord_transform_id;
         }

 
         for (unsigned int ii=0; ii<num_pts; ii++)
         {
            xx[ii] = coords[ii][0];
            yy[ii] = coords[ii][1];
         }

         delete []coords;

         calculate_limits();
         update_data ();

      }
      transform_applied = true;
   }


   rval = change_scale (data_sys->get_log_flag (X_AXIS), data_sys->get_log_flag (Y_AXIS));
   if (!rval)
      change_range (data_sys->get_lower_bounds (), data_sys->get_upper_bounds ());

   return rval;
}


int ChipsCurve::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   int mask = theDefaults->get_allpropmask();

   // apply the specified value to all appropriate attributes
   if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
   {
      string clr = theDefaults->get_allcolor();
      rval = set_property((void*)clr.c_str(), SET_CURVE_SYMBOL_COLOR);
      rval |= set_property((void*)clr.c_str(), SET_CURVE_LINE_COLOR);
      rval |= set_property((void*)clr.c_str(), SET_CURVE_ERR_COLOR);
   }
   if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
   {
      float thick = theDefaults->get_allthickness();
      rval = set_property(static_cast<void*>(&thick), SET_CURVE_LINE_THICKNESS);
      rval |= set_property(static_cast<void*>(&thick), SET_CURVE_ERR_THICKNESS);
   }
   if ((mask & PREF_ALL_ANGLE) == PREF_ALL_ANGLE)
   {
      float ang = theDefaults->get_allangle();
      rval = set_property(static_cast<void*>(&ang), SET_CURVE_SYMBOL_ANGLE);
   }
   if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
   {
      int vis = theDefaults->get_allvisible();
      rval = set_property(static_cast<void*>(&vis), SET_CURVE_VISIBLE);
   }
   if ((mask & PREF_ALL_FILL) == PREF_ALL_FILL)
   {
      bool bval = theDefaults->get_allfill(); 
      rval = set_property(static_cast<void*>(&bval), SET_CURVE_SYMBOL_FILL);
   }
   if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
   {
      int dep = theDefaults->get_alldepth();
      rval = set_property(static_cast<void*>(&dep), SET_CURVE_DEPTH);
   }
   return rval;
}


int ChipsCurve::set_property(void* in_val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   string save_current;
   if (parent && (parent->get_object_type() == chips_plot))
      save_current = ((ChipsPlot *)parent)->get_current_curve(); 

   switch(property)
   {
      case SET_CURVE_SYMBOL_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)symbolstyle;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_symbolstyle((V_SYMBOL_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_LINE_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)linestyle;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_linestyle((V_LINE_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_SYMBOL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_symbolcolor(), clr, save_current);
         if ((rval = set_symbolcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_SYMBOL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            symbolsize_i, *ival, save_current);
         if ((rval = set_symbolsize(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_SYMBOL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            symbolangle, *fval, save_current);
         if ((rval = set_symbolangle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_SYMBOL_FILL:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              symbolfilled, *bval, save_current);
         if ((rval = set_symbolfill(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_LINE_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_linecolor(), clr, save_current);
         if ((rval = set_linecolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_LINE_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            linethickness, *fval, save_current);
         if ((rval = set_linethickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_ERR_UP:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errup, *bval, save_current);
         if ((rval = set_errup(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_ERR_DOWN:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errdown, *bval, save_current);
         if ((rval = set_errdown(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_ERR_LEFT:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errleft, *bval, save_current);
         if ((rval = set_errleft(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_ERR_RIGHT:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errright, *bval, save_current);
         if ((rval = set_errright(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_ERR_STYLE:
      {
         string val = (const char*)in_val;
         string oval = v_errstyle_to_string(errstyle); 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_errstyle(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_ERR_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_errcolor(), clr, save_current);
         if ((rval = set_errcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_ERR_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            errthickness, *fval, save_current);
         if ((rval = set_errthickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_LIMLENGTH:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            limlength, *dval, save_current);
         if ((rval = set_limitslength(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_ERR_CAPLEN:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            errcaplen, *ival, save_current);
         if ((rval = set_errcaplength(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_LIMOVERRIDE:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              limoverride, *bval, save_current);
         if ((rval = set_limitsoverride(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_CURVE_BADDATA:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)badmode;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_baddatamode((V_BADDATAMODE_TYPE)*ival)) ==
             ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_CURVE_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "curve visibility", *ival, 0, 1);
            delete cmd;
         }
      }
      break;
      case SET_CURVE_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, property,
               oval, newval, save_current);
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_CRV);


   return rval;
}


int ChipsCurve::set_linestyle(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine style of line
   if (valid_line_type(val))
   {
      linestyle = val;
      if (initialized) change_linestyle();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val,"chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");

   return rval;
}

int ChipsCurve::set_linestyle(string val)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = c2_string_to_line(val.c_str(), &linestyle, "line style")) == 0)
      if (initialized) change_linestyle();
   return rval;
}


int ChipsCurve::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}


int ChipsCurve::set_linecolor(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = linecolor.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  change_linecolor();
   return rval;
}


int ChipsCurve::set_symbolangle(float val)
{
   if (symbolangle != val)
   {
      symbolangle = val;
      if (initialized) change_symbolstyle();
   }
   return ERR_STATUS_GOOD;
}


int ChipsCurve::set_symbolstyle(V_SYMBOL_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to verify symbol type is valid
   if (valid_symbol_type(val))
   {
      symbolstyle = val;
      if (initialized) change_symbolstyle();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "symbol style", (int)val,"chips_circle|chips_cross|chips_diamond|chips_none|chips_plus|chips_square|chips_point_type|chips_uptriangle|chips_downtriangle");

   return rval;
}


int ChipsCurve::set_symbolstyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if ((rval = c2_string_to_symbol(val.c_str(), &symbolstyle, "symbol style")) == ERR_STATUS_GOOD)
   {
      if (initialized) change_symbolstyle();
   }
   return rval;
}


int ChipsCurve::set_symbolcolor(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = symbolcolor.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  change_symbolcolor();
   return rval;
}

int ChipsCurve::set_symbolfill(bool in_fill)
{
   int rval = ERR_STATUS_GOOD;
   symbolfilled = in_fill;

   if (initialized) change_symbolstyle();

   return rval;
}


int ChipsCurve::set_linethickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine thickness of line
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "line thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      linethickness = val;
      if (initialized) change_linethickness();
   }
   return rval;
}


int ChipsCurve::set_symbolsize(int val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val >= V_SIZE_MIN_VAL) && (val <= V_SIZE_MAX_VAL))
   {
      symbolsize_i = val;
      symbolsize = symbolsize_i * 2.0;
      if (initialized) change_symbolsize();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "symbol size", val,
                                  V_SIZE_MIN_VAL, V_SIZE_MAX_VAL);

   return rval;
}


int ChipsCurve::set_errthickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine thickness of errorbars
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "error thickness", val,
                                  V_THICK_MIN_VAL, V_THICK_MAX_VAL);
   else
   {
      errthickness = val;
      if (initialized) change_errthickness();
   }
   return rval;
}


int ChipsCurve::set_errcolor(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = errcolor.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  change_errcolor();
   return rval;
}


int ChipsCurve::set_errup(bool val)
{
   errup = val;
   if (initialized) change_errup();
   return ERR_STATUS_GOOD;
}


int ChipsCurve::set_errdown(bool val)
{
   errdown = val;
   if (initialized) change_errdown();
   return ERR_STATUS_GOOD;
}

int ChipsCurve::set_errleft(bool val)
{
   errleft = val;
   if (initialized) change_errleft();
   return ERR_STATUS_GOOD;
}

int ChipsCurve::set_errright(bool val)
{
   errright = val;
   if (initialized) change_errright();
   return ERR_STATUS_GOOD;
}

int ChipsCurve::set_errcaplength(int val)
{
   int rval = ERR_STATUS_GOOD;

   if (val < V_SIZE_MIN_VAL || val > V_SIZE_MAX_VAL)
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "error cap length", val,
                           V_SIZE_MIN_VAL, V_SIZE_MAX_VAL);
   else 
   {
      errcaplen = val;
      if (initialized) change_errstyle();
   }
   return rval;
}

int ChipsCurve::set_limitslength(double val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val > 0.0) && (val < 1.0))
   {
      limlength = val;
      if (initialized) change_limitslength();
   }
   else
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "limits length", val,
                                  0.0, 1.0);
   return rval;
}


int ChipsCurve::set_limitsoverride(bool val)
{
   limoverride = val;

   if (initialized)   change_limitsoverride();
   return ERR_STATUS_GOOD; 
}


int ChipsCurve::set_errstyle(string in_style)
{
   int rval = ERR_STATUS_GOOD;
   V_ERRSTYLE_TYPE typ;

   if ((rval = string_to_v_errstyle(in_style.c_str(), &typ)) == ERR_STATUS_GOOD)
   {
      errstyle = typ;
      if (initialized) change_errstyle();
   }
   return rval;
}


int ChipsCurve::set_errstyle(V_ERRSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to verify err style is valid
   if (valid_errstyle_type(val))
   {
      errstyle = val;
      if (initialized) change_errstyle();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "error style", (int)val,"chips_line|chips_capped");
   return rval;
}


int ChipsCurve::set_baddatamode(V_BADDATAMODE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to verify err style is valid
   if (valid_baddatamode_type(val))
   {
      badmode = val;
      if (badmode == chips_omit)
         display_mask = REG_INTERVAL; 
      else 
         display_mask = REG_INTERVAL | NAN_INTERVAL; 
      if (initialized) change_visible();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "bad data handling modes", (int)val,"chips_ignore|chips_omit");
   return rval;
}


int ChipsCurve::get_x_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && xx)
      {
         *cnt = num_pts; 
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory 
            for (ii = num_pts; ii--; )  dat[ii] = xx[ii]; 
         }
         else 
            cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } else
      cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval; 
}


int ChipsCurve::get_y_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && yy)
      {
         *cnt = num_pts; 
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = yy[ii];
         }
         else
            cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } 
   else
      cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);   

   return rval; 
}


int ChipsCurve::get_xup_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && xup)
      {
         *cnt = num_pts;
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = xup[ii];
         }
         else cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}


int ChipsCurve::get_xdown_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && xdown)
      {
         *cnt = num_pts;
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat)
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = xdown[ii];
         }
         else cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}


int ChipsCurve::get_yup_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && yup)
      {
         *cnt = num_pts;
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat)
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = yup[ii];
         }
         else cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}

int ChipsCurve::get_ydown_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && ydown)
      {
         *cnt = num_pts;
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat)
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = ydown[ii];
         }
         else cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}



void ChipsCurve::show_errs()
{
   if (xdown) set_errleft(true); 
   if (xup)   set_errright(true);
   if (yup)   set_errup(true);
   if (ydown) set_errdown(true);
}


void ChipsCurve::hide_errs()
{
   set_errleft(false); 
   set_errright(false);
   set_errup(false);
   set_errdown(false);
}


int ChipsCurve::render_err()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_linecolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_linethickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_linestyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_symbolcolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_symbolsize()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_symbolfilled()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_symbolstyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errcolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errstyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errup()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errdown()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errleft()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errright()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_errthickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_limitslength()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_limitsoverride()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::change_depth()
{
   return ERR_FEATURE_DNE;
}


int ChipsCurve::update_data ()
{
   return ERR_FEATURE_DNE;
}

int ChipsCurve::set_plot_location (double x1, double y1, double x2, double y2)
{
   return ERR_FEATURE_DNE;
}


void ChipsCurve::serialize(ChipsStateStream& out)
{
   string tmp; 
   int sz; 
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   sz = (int)linestyle;
   out.write_int(&sz);
   tmp = get_linecolor();
   out.write_string(&tmp);
   out.write_float(&linethickness);
   sz = (int)badmode;
   out.write_int(&sz);
   sz = (int)symbolstyle;
   out.write_int(&sz);
   tmp = get_symbolcolor();
   out.write_string(&tmp);
   out.write_float(&symbolangle);
   out.write_int(&symbolsize_i);
   out.write_int(&depth);
   out.write_bool(&symbolfilled);
   out.write_int((int*)&num_pts);
   out.write_double(xx, num_pts);
   out.write_double(yy, num_pts);

   // output err data
   unsigned char ub = 0;
   if (ylim)   ub |= 0x020;
   if (xlim)   ub |= 0x010;
   if (xup)   ub |= 0x08;
   if (xdown) ub |= 0x04;
   if (yup)   ub |= 0x02;
   if (ydown) ub |= 0x01;
   out.write((char*)&ub, sizeof(ub));
   if (xup)   out.write_double(xup, num_pts);
   if (xdown) out.write_double(xdown, num_pts);
   if (yup)   out.write_double(yup, num_pts);
   if (ydown) out.write_double(ydown, num_pts);
   if (ylim)  out.write_double(ylim, num_pts);
   if (xlim)  out.write_double(xlim, num_pts);
   out.write_bool(&errup);
   out.write_bool(&errdown);
   out.write_bool(&errleft);
   out.write_bool(&errright);
   sz = (int)errstyle;
   out.write_int(&sz);
   out.write_float(&errthickness);
   if (out.get_minor_version() > 2)
      out.write_int(&errcaplen);
   tmp = get_errcolor();
   out.write_string(&tmp);
   out.write_double(&limlength);
   out.write_bool(&limoverride);
   out.write_bool (&visible);
   out.write_string (&data_system_id);
   sz = (int)coord_system;
   out.write_int (&sz);

   out.write_string (&transform_id);
}


void ChipsCurve::deserialize(ChipsStateStream& in)
{
   string axid, cname;
   int len;

   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_int(&len);
   linestyle = (V_LINE_TYPE)len;
   in.read_string(&cname);
   set_linecolor(cname); 
   in.read_float(&linethickness);
   in.read_int(&len);
   set_baddatamode((V_BADDATAMODE_TYPE) len);
   in.read_int(&len);
   symbolstyle = (V_SYMBOL_TYPE)len;
   in.read_string(&cname);
   set_symbolcolor(cname); 
   in.read_float(&symbolangle);
   in.read_int(&symbolsize_i);
   symbolsize = symbolsize_i * 2.0;
   in.read_int(&depth);
   in.read_bool(&symbolfilled);

   in.read_int((int*)&num_pts);
   xx = (double*)malloc(num_pts*sizeof(double));
   yy = (double*)malloc(num_pts*sizeof(double));
   in.read_double(xx, (int)num_pts);
   in.read_double(yy, (int)num_pts);

   // output err data
   unsigned char ub;
   in.read((char*)&ub, sizeof(ub));
   if ((ub & 0x08) == 0x08)
   {
      xup = (double*)malloc(num_pts*sizeof(double));
      in.read_double(xup, num_pts);
   }
   else  xup = NULL;
   if ((ub & 0x04) == 0x04)
   {
      xdown = (double*)malloc(num_pts*sizeof(double));
      in.read_double(xdown, num_pts);
   }
   else  xdown = NULL;
   if ((ub & 0x02) == 0x02)
   {
      yup = (double*)malloc(num_pts*sizeof(double));
      in.read_double(yup, num_pts);
   }
   else  yup = NULL;
   if ((ub & 0x01) == 0x01)
   {
      ydown = (double*)malloc(num_pts*sizeof(double));
      in.read_double(ydown, num_pts);
   }
   else  ydown = NULL;
   if ((ub & 0x020) == 0x020)
   {
      ylim = (double*)malloc(num_pts*sizeof(double));
      in.read_double(ylim, num_pts);
   }
   else  ylim = NULL;
   if ((ub & 0x010) == 0x010)
   {
      xlim = (double*)malloc(num_pts*sizeof(double));
      in.read_double(xlim, num_pts);
   }
   else  xlim = NULL;
   in.read_bool(&errup);
   in.read_bool(&errdown);
   in.read_bool(&errleft);
   in.read_bool(&errright);
   in.read_int(&len);
   errstyle = (V_ERRSTYLE_TYPE)len;
   in.read_float(&errthickness);
   if (in.get_minor_version() > 2)
      in.read_int(&errcaplen);
   in.read_string(&cname);
   set_errcolor(cname); 
   in.read_double(&limlength);
   in.read_bool(&limoverride);
   in.read_bool (&visible);
   in.read_string (&data_system_id);
   in.read_int (&len);
   coord_system = (COORD_SYS)len;

   if (in.get_minor_version () > 8)
      in.read_string (&transform_id);
}

int ChipsCurve::show()
{
  int rval = ERR_STATUS_GOOD;
  visible = true;
  rval = change_visible ();

  return rval;
};

int ChipsCurve::hide()
{
  int rval = ERR_STATUS_GOOD;
  visible = false;
  rval = change_visible ();
  
  return rval;
};

bool ChipsCurve::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED) outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}


double ChipsCurve::get_xmin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }
      return out[0];
   }
   else
      return xmin;
}

double ChipsCurve::get_xmax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }
      return out[0];
   }
   else
      return xmax;

}

double ChipsCurve::get_ymin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }
      return out[1];
   }
   else
      return ymin;

}

double ChipsCurve::get_ymax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }

      return out[1];
   }
   else
      return ymax;
}

void ChipsCurve::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80]; 
   const char *cp; 
   string transform_desc;
   string transform_obj;

   tmp = get_linecolor(); 
   if (theDefaults->get_curvelinecolor() != tmp)
   {
      if (att.length()) att.append(" "); 
      att.append("line.color=");
      att.append(tmp);
   }
   if (theDefaults->get_curvelinestyle() != linestyle) 
   {
      if (att.length()) att.append(" ");
      att.append("line.style=");
      att.append(v_line_to_string(linestyle));
   }
   if (theDefaults->get_curvelinethickness() != linethickness)
   {
      if (att.length()) att.append(" ");
      if (linethickness != 0.5)
      {
         sprintf(buf,"line.thickness=%-1.0f", linethickness); 
         att.append(buf); 
      }
      else 
         att.append("line.thickness=0.5");
   }
   tmp = get_symbolcolor(); 
   if (theDefaults->get_curvesymbolcolor() != tmp)
   {
      if (att.length()) att.append(" "); 
      att.append("symbol.color=");
      att.append(tmp);
   }
   if (theDefaults->get_curvesymbolstyle() != symbolstyle) 
   {
      if (att.length()) att.append(" ");
      att.append("symbol.style=");
      att.append(v_symbol_to_string(symbolstyle));
   }
   if (theDefaults->get_curvesymbolsize() != symbolsize_i)
   {
      if (att.length()) att.append(" ");
         sprintf(buf,"symbol.size=%d", symbolsize_i); 
         att.append(buf); 
   }
   tmp = get_symbolcolor(); 
   if (symbolfilled != theDefaults->curvesymbol_filled())
   {
      if (att.length()) att.append(" ");
      if (symbolfilled)  att.append("symbol.fill=true"); 
         else            att.append("symbol.fill=false"); 
   }
   if (symbolangle != theDefaults->get_curvesymbolangle())
   {
      if (att.length()) att.append(" ");
      sprintf(buf,"symbol.angle=%2.1f", symbolangle); 
      att.append(buf); 
   }
   if (xup||xdown||yup||ydown||ylim||xlim)
   {
      tmp = get_errcolor(); 
      if (theDefaults->get_errcolor() != tmp)
      {
         if (att.length()) att.append(" "); 
         att.append("err.color=");
         att.append(tmp);
      }
      if (theDefaults->get_errthickness() != errthickness)
      {
         if (att.length()) att.append(" ");
         if (errthickness != 0.5)
         {
            sprintf(buf,"err.thickness=%-1.0f", errthickness);
            att.append(buf);
         }
         else
            att.append("err.thickness=0.5");
      }
   }
   if (badmode != theDefaults->get_curvebaddatamode() && 
       (cp = v_baddatamode_to_string(badmode)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("baddata=");
      att.append(cp); 
   }
   if (xlim || ylim)
   {
      if (limoverride != theDefaults->get_limitoverride())
      {
         if (att.length()) att.append(" ");
         if (limoverride)  att.append("limit.override=true");
            else           att.append("limit.override=false");
      }
      if (limlength != theDefaults->get_limitlength())
      {
         if (att.length()) att.append(" ");
         sprintf(buf, "limit.length=%2.1f", limlength);
         att.append(buf);
      }
   }
   if (yup || ylim) 
   {
      if (errup != theDefaults->get_errup())
      {
         if (att.length()) att.append(" ");
         if (errup) att.append("err.up=true"); 
            else    att.append("err.up=false"); 
      }
   }
   if (ydown || ylim)
   {
      if (errdown != theDefaults->get_errdown())
      {
         if (att.length()) att.append(" ");
         if (errdown) att.append("err.down=true");
            else    att.append("err.down=false");
      }
   }
   if (xup || xlim)
   {
      if (errright != theDefaults->get_errright())
      {
         if (att.length()) att.append(" ");
         if (errright) att.append("err.right=true");
            else    att.append("err.right=false");
      }
   }
   if (xdown || xlim)
   {
      if (errleft != theDefaults->get_errleft())
      {
         if (att.length()) att.append(" ");
         if (errleft) att.append("err.left=true");
            else    att.append("err.left=false");
      }
   }
   int dep = theDefaults->get_curvedepth();
   if (depth != dep &&
       (dep != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   string dfname = out.use_tempname();
   string mfile = "";
   
   if (meta.length())
   {
      int quoted = (meta.find("'") == 0 && meta.rfind("'") == meta.length()-1);
      mfile = dfname;
      mfile.append("=");
      if (!quoted)  mfile.append("\"");
      mfile.append(meta);
      if (!quoted)  mfile.append("\""); 
      out << "add_curve("<<dfname;
   }
   else 
   {
      string base_filename = out.get_filename();
      sprintf(buf, ".Data%d_curve", out.use_tempfnum());
      base_filename.append(buf);
      string tmpfile (base_filename+".bin");
      mfile = dfname;
      mfile.append(" = DataObj()  # data- ");
      mfile.append(tmpfile);
      mfile.append("\n");
      if (xx)
      {
          mfile.append(dfname);
          mfile.append(".xx    = #  insert data array here \n");
      }
      if (yy)
      {
          mfile.append(dfname);
          mfile.append(".yy    = #  insert data array here \n");
      }
      if (ydown)
      {
          mfile.append(dfname);
          mfile.append(".ydown = #  insert data array here \n");
      }
      if (yup)
      {
          mfile.append(dfname);
          mfile.append(".yup   = #  insert data array here \n");
      }
      if (xdown)
      {
          mfile.append(dfname);
          mfile.append(".xdown = #  insert data array here \n");
      }
      if (xup)
      {
          mfile.append(dfname);
          mfile.append(".xup   = #  insert data array here \n");
      }
      if (ylim)
      {
          mfile.append(dfname);
          mfile.append(".ylim  = #  insert data array here \n");
      }
      if (xlim)
      {
          mfile.append(dfname);
          mfile.append(".xlim  = #  insert data array here \n");
      }

      // binary dump of data 
      ofstream ff(tmpfile.c_str(), ios::out|ios::binary);
      if (ff)
      {
         // table data format:
         //  [type (0= table) ][ numcols         ]
         //  [number of rows  ][ colname buff size] 
         //  [ char array '\0' separated col names ]
         //  [ double array col1 data .. numcols]   
         char buffer[1024];
         long bpos = 0; 
         string colnames = ""; 
         if (xx) { strcpy(&buffer[bpos], "xx\0"); bpos+=3;  }
         if (yy) { strcpy(&buffer[bpos], "yy\0"); bpos+=3;  }
         if (yup) { strcpy(&buffer[bpos], "yup\0"); bpos+=4;  }
         if (ydown) { strcpy(&buffer[bpos], "ydown\0"); bpos+=6;  }
         if (xup) { strcpy(&buffer[bpos], "xup\0"); bpos+=4;  }
         if (xdown) { strcpy(&buffer[bpos], "xdown\0"); bpos+=6;  }
         if (ylim) { strcpy(&buffer[bpos], "ylim\0"); bpos+=5;  }
         if (xlim) { strcpy(&buffer[bpos], "xlim\0"); bpos+=5;  }
         // number of long's in buffer
         bpos = (bpos + sizeof(double) - 1) / sizeof(double); 

         long ncols = 0;   // 0 represents table data 
         ff.write((char*)&ncols, sizeof(long)); 
         ncols = (xx!=NULL) + (yy!=NULL) + (yup!=NULL) + (ydown!=NULL) +
                      (xup!=NULL) + (xdown!=NULL) + (ylim!=NULL) + (xlim!=NULL);
         ff.write((char*)&ncols, sizeof(long)); 
         ff.write((char*)&num_pts, sizeof(unsigned long)); 
         ff.write((char*)&bpos, sizeof(long));
         ff.write((char*)buffer, bpos* sizeof(double));

         ChipsDataSystem * data_sys = static_cast <ChipsDataSystem *>(coord);
         if (transform_id.size () > 0)
         {
            ChipsTransformFactory tf;

            // set the transform file name
            string transform_fname = base_filename + "_transform.txt";

            // write the transform to a ascii file
            fstream transform_file(transform_fname.c_str(), ios::out);
            Transform *transform = data_sys->get_transform (transform_id);
            tf.serialize_ascii (transform, transform_file);

            transform_obj = dfname + ".transform";
            transform_desc = dfname + ".transform = #  insert transform here - " + transform_fname + "\n";

         }
         else
            this->transform_id = data_sys->get_default_transform_id ();

         if (xx)    ff.write((char*)xx, num_pts* sizeof(double));
         if (yy)    ff.write((char*)yy, num_pts* sizeof(double));
         if (yup)   ff.write((char*)yup, num_pts* sizeof(double));
         if (ydown) ff.write((char*)ydown, num_pts* sizeof(double));
         if (xup)   ff.write((char*)xup, num_pts* sizeof(double));
         if (xdown) ff.write((char*)xdown, num_pts* sizeof(double));
         if (ylim)  ff.write((char*)ylim, num_pts* sizeof(double));
         if (xlim ) ff.write((char*)xlim, num_pts* sizeof(double));
         ff.close(); 
      }
      
      out << "add_curve("<<dfname<<".xx, "<<dfname<<".yy";
      if (xup || xdown || yup || ydown)
      {
         out << ",["; 
         if (ydown) out<<dfname<<".ydown,";
            else    out<<"None,";
         if (yup) out<<dfname<<".yup,";
            else    out<<"None,";
         if (xdown) out<<dfname<<".xdown,";
            else    out<<"None,";
         if (xup) out<<dfname<<".xup,";
            else    out<<"None,";
         out << "]"; 
         if (xlim && ylim)  out <<",["<<dfname<<".ylim, "<<dfname<<".xlim]";
         else if (xlim)     out <<",[None,"<<dfname<<".xlim]"; 
         else if (ylim)     out <<",["<<dfname<<".ylim]"; 
      }
      else 
      {
         if (xlim || ylim)
         {
            // no errors but limits
            if (xlim && ylim)  out <<",None,["<<dfname<<".ylim, "<<dfname<<".xlim]";
            else if (xlim)     out <<",None,[None,"<<dfname<<".xlim]"; 
            else if (ylim)     out <<",None,["<<dfname<<".ylim]"; 
         }
      }

      if (transform_obj.length () > 0) out << ", " << transform_obj;
   }
   if (att.length()) out <<",'"<<att<<"'";
   out<<")"<<endl;
   mfile += transform_desc;
   out.append_to_template(mfile);
}

void ChipsCurve::set_transform_id (const string &transform_id)
{
   // cant reset the transform id if it's already been set
   // since we overwrite the original positions
   if (this->transform_id == "")
   {
      this->transform_id = transform_id;
      this->transform_applied = false;
   }
}

