// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSCURVE_HH
#define CHIPSCURVE_HH

#include <string>
#include <stdio.h>
#include <stdlib.h>
#include "ChipsObject.hh"
#include "ChipsDefaults.hh"
#include "ChipsColor.hh"
#include "ChipsStateStream.hh"
#include "ChipsId.hh"
#include "ChipsCommand.hh"

using namespace std;

#define REG_INTERVAL     0x001 
#define NAN_INTERVAL     0x010 

typedef struct chips_range
{
   unsigned long  lo;         // lower index 
   unsigned long  hi;         // upper index 
   short          inclusive;  // 1 = all indices from lo-hi, 0 = single segment
   unsigned short mask;       // mask indicating when to display 
} CHIPS_RANGE, *CHIPS_RANGE_P;


class ChipsCurve: public ChipsObject
{
  public:

   ChipsCurve(ChipsObject* par, unsigned long cnt, double* x, double* y,
             const string &transform_id, string cid, bool compute=true);
   ChipsCurve(ChipsObject* par, unsigned long cnt, double* x, double* y, 
              double* ydown, double* yup, double* xdown, double* xup,
              double* ylims, double* xlims, const string &transform_id,
              string cid, bool compute=true);
   ChipsCurve(ChipsObject* par, ChipsStateStream& in);

   void       object_delete ();

   virtual void initialize(); 

   int        set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj);
   int        set_property(void* in_val, int property, bool rec=true);

   string     get_linecolor()        { return linecolor.get_name(); }
   int        set_linecolor(string in_color);
   float      get_linethickness()    { return linethickness; }
   int        set_linethickness(float val); 
   V_LINE_TYPE get_linestyle()        { return linestyle; }
   int        set_linestyle(V_LINE_TYPE val); 
   int        set_linestyle(string val);

   float      get_symbolangle()    { return symbolangle; }
   int        set_symbolangle(float val);
   string     get_symbolcolor()        { return symbolcolor.get_name(); }
   int        set_symbolcolor(string in_color);
   int        get_symbolsize()    { return symbolsize_i; }
   int        set_symbolsize(int val); 
   V_SYMBOL_TYPE get_symbolstyle()        { return symbolstyle; }
   int        set_symbolstyle(V_SYMBOL_TYPE val); 
   int        set_symbolstyle(string val); 
   int        set_symbolfill(bool in_val);
   bool       get_symbolfill()  { return symbolfilled; }

   int        set_errcaplength(int val);
   int        get_errcaplength()    { return errcaplen; } 
   string     get_errcolor()        { return errcolor.get_name(); }
   int        set_errcolor(string in_color);
   float      get_errthickness()    { return errthickness; }
   int        set_errthickness(float val); 
   V_ERRSTYLE_TYPE get_errstyle()             { return errstyle; } 
   int        set_errstyle(V_ERRSTYLE_TYPE val); 
   int        set_errstyle(string val); 
   bool       get_errup()                  { return errup; }
   int        set_errup(bool val); 
   bool       get_errdown()                { return errdown; }
   int        set_errdown(bool val); 
   bool       get_errleft()                { return errleft; }
   int        set_errleft(bool val); 
   bool       get_errright()               { return errright; }
   int        set_errright(bool val); 
   int        set_baddatamode(V_BADDATAMODE_TYPE val);
   V_BADDATAMODE_TYPE get_baddatamode()  { return badmode; }

   unsigned long get_cnt()               { return num_pts; }
   unsigned long get_discarded_cnt()     { return num_disc; }
   int        set_limitslength(double len);
   double     get_limitslength()         { return limlength; }
   int        set_limitsoverride(bool val);
   bool       get_limitsoverride()       { return limoverride; }

   int        get_x_data(double* dat, unsigned long* cnt);
   int        get_y_data(double* dat, unsigned long* cnt);
   int        get_xup_data(double* dat, unsigned long* cnt);
   int        get_yup_data(double* dat, unsigned long* cnt);
   int        get_xdown_data(double* dat, unsigned long* cnt);
   int        get_ydown_data(double* dat, unsigned long* cnt);

   void       show_errs();
   void       hide_errs();
   virtual int show();
   virtual int hide();

   double     get_xlmin()                { return xlmin;}
   double     get_xmin(bool transformed);
   double     get_xmax(bool transformed);
   double     get_ylmin()                { return ylmin;}
   double     get_ymin(bool transformed);
   double     get_ymax(bool transformed);

   string     get_xaxis();
   string     get_yaxis();
   string     get_meta()  { return meta; }


   void       set_curve_num (int val) {curve_num = val;};
   int        get_curve_num () {return curve_num;};

   void       set_transform_id (const string &transform_id);


   virtual int reset_position ();
   virtual int  set_depth(int newdepth);


   virtual int render_err();
   virtual int change_range (const ChipsVector &low_range, const ChipsVector &high_range)
      { return ERR_FEATURE_DNE; }
   virtual int change_scale (int inXScale, int inYScale)
      { return ERR_FEATURE_DNE; }
   virtual int change_linethickness(); 
   virtual int change_linestyle(); 
   virtual int change_linecolor(); 
   virtual int change_symbolsize(); 
   virtual int change_symbolstyle(); 
   virtual int change_symbolcolor(); 
   virtual int change_symbolfilled(); 
   virtual int change_errcolor(); 
   virtual int change_errstyle(); 
   virtual int change_errup(); 
   virtual int change_errdown(); 
   virtual int change_errleft(); 
   virtual int change_errright(); 
   virtual int change_errthickness(); 
   virtual int change_visible ();
   virtual int change_limitslength();
   virtual int change_limitsoverride();
   virtual int change_depth();
   virtual int update_data ();
   virtual int set_plot_location (double x1, double y1, double x2, double y2);

   virtual void serialize(ChipsStateStream& out);
   virtual void serialize_ascii(ChipsStateStream& out);
   virtual void deserialize(ChipsStateStream& in);
   
  protected:

   virtual ~ChipsCurve();

   void calculate_limits();

   unsigned long num_pts;    // number of points in curve 
   unsigned long num_disc;   // number of discarded/ignored points
   double*    xx;            // x column data
   double*    yy;            // y column data
   double*    xup;           // x up error
   double*    xdown;         // x down error
   double*    yup;           // y up error 
   double*    ydown;         // y down error
   double*    ylim;          // y upper/lower limits
   double*    xlim;          // x upper/lower limits

   double     xlmin;         // min log x (x > 0.0) value in dataset 
   double     xmin;          // min x value in dataset 
   double     ylmin;         // min log y (y > 0.0) value in dataset 
   double     ymin;          // min y value in dataset 
   double     xmax;          // max x value in dataset 
   double     ymax;          // max y value in dataset 


   list <CHIPS_RANGE_P>indices;  // index table for curve segments
   V_BADDATAMODE_TYPE badmode;   // bad data handling mode
   unsigned short display_mask;  // which segments to display

   bool       initialized;   // have we initialized everything
   bool       errup;         // display up error bar?
   bool       errdown;       // display down error bar?
   bool       errleft;       // display left error bar?
   bool       errright;      // display right error bar?
   V_ERRSTYLE_TYPE errstyle; // error bar style- capped or line 
   ChipsColor errcolor;      // error bar color 
   float      errthickness;  // error bar thickness
   double     limlength;     // length for limits arrow shafts 
   bool       limoverride;   // u/l limits overrides both up and down errs?
   int        errcaplen;     // error bar cap length 

   ChipsColor linecolor;     // color of line 
   V_LINE_TYPE linestyle;    // style of line
   float      linethickness; // thickness of line 

   ChipsColor symbolcolor;   // color of symbol 
   V_SYMBOL_TYPE symbolstyle; 
   int        symbolsize_i;  // size of symbol
   double     symbolsize;    // size of symbol
   float      symbolangle;   // angle of rotation
   bool       symbolfilled;  // solid or hollow point

   int        curve_num;     // the curve's order 0th, 1st, 2nd, etc.
   bool       transform_applied; // transform applied to points
   string     meta;          // meta data for curve (ie. file info)
   string     transform_id;  // the transform id

  private:
   bool add_command (ChipsCommand *inCmd, int &outStatus);
};


#endif 
