// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <math.h>
#include "ChipsDataSystem.hh"
#include "ChipsTransformFactory.hh"
#include <algorithm>
#include <sstream>

unsigned long ChipsDataSystem::global_ref_id = 0;

ChipsDataSystem::ChipsDataSystem (ChipsCoordinate *par):ChipsCoordinate (par)
{
   aspect_use = false;
   aspect_w = aspect_h = 1.0;
   resizing = 0; 
   ratio = "";
   transform_ref = 0;
   data_min = ChipsVector (3);
   data_max = ChipsVector (3);
   modified = true;

   log_flag = new bool [data_min.get_size ()];
   for (int ii=0; ii<3; ii++)
      log_flag[ii] = false;

   default_transform = "";

   this->new_id ();

   this->old_id = "";

//   data_transform.set_size (3,3);
//   data_transform.identity ();
   if (this->parent)
      this->parent->add_coord_sys (this);

}

ChipsDataSystem::ChipsDataSystem (ChipsCoordinate *par,
                const ChipsVector &ds_data_min, const ChipsVector  &ds_data_max,
                const string &ax_id, const string &ay_id):ChipsCoordinate (par)
{
   data_min = ds_data_min;
   data_max = ds_data_max;
   log_flag = new bool [data_min.get_size ()];
   for (int ii=0; ii<data_min.get_size (); ii++)
      log_flag[ii] = false;
   xaxis_id =ax_id;
   yaxis_id =ay_id;
   aspect_use = false;
   aspect_w = aspect_h = 1.0;
   ratio = "";
   resizing = 0;
   transform_ref = 0;

   default_transform = "";
   modified = true;

   this->new_id ();
   this->old_id = "";

//   data_transform.set_size (3,3);
//   data_transform.identity ();
   if (this->parent)
      this->parent->add_coord_sys (this);

}

ChipsDataSystem::ChipsDataSystem (const ChipsDataSystem &ds):ChipsCoordinate (ds)
{
   map <string, Transform *>::iterator iter;

   this->parent = ds.parent;
   this->id = ds.id;
   this->data_min = ds.data_min;
   this->data_max = ds.data_max;
   this->xaxis_id = ds.xaxis_id;
   this->yaxis_id = ds.yaxis_id;
   for (int ii=0; ii<data_min.get_size (); ii++)
      log_flag[ii] = ds.log_flag[ii];

//   this->data_transform = ds.data_transform;
   this->chips_objects = ds.chips_objects;
//   this->transform_stk = ds.transform_stk;
   this->aspect_use = ds.aspect_use;
   this->aspect_w = ds.aspect_w;
   this->aspect_h = ds.aspect_h;
   this->ratio = ds.ratio;
   this->resizing = ds.resizing;
   this->transform_ref = ds.transform_ref;
   this->modified = ds.modified;
   this->ref_id = ds.ref_id;
   this->old_id = ds.old_id;

   this->transform_map.clear ();
   for (iter=transform_map.begin (); iter!=transform_map.end (); iter++)
      this->transform_map[iter->first] = ChipsTransformFactory::copy_transform (iter->second);
   this->default_transform = ds.default_transform;
   this->transform_ref = ds.transform_ref;
   
}

ChipsDataSystem::ChipsDataSystem (ChipsStateStream &in, ChipsCoordinate *par):ChipsCoordinate (par)
{
   modified = true;
   data_min = ChipsVector (3);
   data_max = ChipsVector (3);

//   data_transform.set_size (3,3);
//   data_transform.identity ();

   log_flag = new bool [data_min.get_size ()];
   for (int ii=0; ii<3; ii++)
      log_flag[ii] = false;


   this->deserialize (in);
}


ChipsDataSystem::~ChipsDataSystem ()
{
   delete []log_flag;
   if (parent)
      parent->remove_coord_sys(this->id);

   map <string, Transform *>::iterator iter;
   for (iter=transform_map.begin (); iter!=transform_map.end (); iter++)
      delete iter->second;

   transform_map.clear ();
}


void ChipsDataSystem::operator = (const ChipsDataSystem &ds)
{
   if (this == &ds)
      return;

   map <string, Transform *>::iterator iter;

   this->data_min = ds.data_min;
   this->data_max = ds.data_max;
   this->xaxis_id = ds.xaxis_id;
   this->yaxis_id = ds.yaxis_id;

   for (int ii=0; ii<data_min.get_size (); ii++)
      log_flag[ii] = ds.log_flag[ii];

//   this->data_transform = ds.data_transform;
//   this->chips_objects = ds.chips_objects;
//   this->transform_stk = ds.transform_stk;
   this->aspect_use = ds.aspect_use;
   this->aspect_w = ds.aspect_w;
   this->aspect_h = ds.aspect_h;
   this->ratio = ds.ratio;
   this->resizing = ds.resizing;

   for (iter=this->transform_map.begin (); iter!=this->transform_map.end (); iter++)
      delete iter->second;
   this->transform_map.clear ();
   map <string, Transform *> ds_transform_map = ds.transform_map;
   for (iter=ds_transform_map.begin (); iter!=ds_transform_map.end (); iter++)
      this->transform_map[iter->first] = ChipsTransformFactory::copy_transform (iter->second);
   this->default_transform = ds.default_transform;
   this->transform_ref = ds.transform_ref;
   this->modified = ds.modified;

}

int ChipsDataSystem::convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id)
{
   int rval = ERR_STATUS_GOOD;
   bool use_logs = 0;
   string wcs_to_id = default_transform;
   string wcs_from_id = default_transform;
   string tmp_from;
   string tmp_to;
   unsigned int pos;

   if (!in || !out)
      return ERR_STATUS_BAD;

   if (size <= 0)
      return ERR_STATUS_BAD;

   // parse out the wcs ids, wcs is appended onto the data sys
   // id with an _wcs
   pos = from_id.rfind ('_');
   if ((pos == (unsigned int)-1) || (pos >= (from_id.size()-1)))
      tmp_from = from_id;
   else
   {
      // check to see if the wcs id is really part of this class
      // if it isn't assume the there was no wcs id
      wcs_from_id = from_id.substr (pos+1);

      if (wcs_from_id == "none")
      {
         tmp_from = from_id.substr (0,pos);
         wcs_from_id = "";
      }
      else if (transform_map.count (wcs_from_id) > 0)
         tmp_from = from_id.substr (0,pos);
      else
      {
         tmp_from = from_id;
         wcs_from_id = default_transform;
      }
   }

   pos = to_id.rfind ('_');
   if ((pos == (unsigned int)-1) || (pos >= (to_id.size()-1)))
      tmp_to = to_id;
   else
   {
      // check to see if the wcs id is really part of this class
      // if it isn't assume the there was no wcs id
      wcs_to_id = to_id.substr (pos+1);

      if (wcs_to_id == "none")
      {
         tmp_to = to_id.substr (0,pos);
         wcs_to_id = "";
      }
      else if (transform_map.count (wcs_to_id) > 0)
         tmp_to = to_id.substr (0,pos);
      else
      {
         tmp_to = to_id;
         wcs_to_id = default_transform;
      }
   }    


   if ((from_cs == DATA) && (tmp_from.length () > 0) && (id != tmp_from))
      return ERR_STATUS_BAD;

   if ((to_cs == DATA) && (tmp_to.length () > 0) && (id != tmp_to))
      return ERR_STATUS_BAD;

   for (int ii=0; !use_logs && ii<data_min.get_size (); ii++)
      use_logs = log_flag[ii];

   switch (from_cs){
      case WINDOW_NORM: // Fallthrough intended 
      case PIXEL:       // Fallthrough intended
      case PIXEL_REAL:  // Fallthrough intended
      case FRAME_NORM:  // Fallthrough intended
      case PLOT_NORM:
      {
         if (!parent)
            rval = ERR_STATUS_BAD;

	 if (!rval && ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM) ||
	     (to_cs == PLOT_NORM))){
            return parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
                                                to_cs, to_id);
	 }
	 else if (!rval)
	    rval = parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
                                                PLOT_NORM, parent->get_id());
         
 	 if (rval != ERR_STATUS_GOOD)
	    return rval;

	 ChipsMatrix calc_matrix;
         Transform *transform = NULL;

         if (wcs_to_id != "") 
            transform = transform_map[wcs_to_id];

	 if (use_logs){
            ChipsVector tmp_min = data_min;
            ChipsVector tmp_max = data_max;
	    for (int ii=0; ii<data_min.get_size (); ii++){
               if (log_flag[ii]){
                  if ((data_min[ii] <= 0) || (data_max[ii] <= 0))
                     return ERR_NEG_LOG_VAL;
                  data_min[ii] = log10(data_min[ii]);
                  data_max[ii] = log10(data_max[ii]);
               }
            }

            endpoint_matrix (calc_matrix);
            data_min = tmp_min;
            data_max = tmp_max;
         }
         else
            endpoint_matrix (calc_matrix);

	 for (int ii=size; ii--; ){
	    if ((out[ii].get_size () != calc_matrix.get_row()) &&
                (out[ii].get_size () != calc_matrix.get_col()))
               return ERR_BAD_TRANSFORM_SIZE;
            if (use_logs){
               out[ii] = calc_matrix * out[ii];
	       for (int jj=in->get_size (); jj--;){
                  if (log_flag[jj])
                     out[ii][jj] = pow (10,out[ii][jj]);
                  else
                     out[ii][jj] = out[ii][jj];
               }

            }
            else
            {
               out[ii] = calc_matrix * out[ii];

               // if a transform exists apply it
               if (transform)
               {
                  double *tmp = (double *)transform->apply ((void *)(&out[ii][0]),dmDOUBLE,1);
                  double val = out[ii][out[ii].get_size ()-1];
                  for (int jj=0; jj<out[ii].get_size ()-1; jj++)
                     out[ii][jj] = tmp[jj];
                  out[ii][out[ii].get_size ()-1] = val;
                  delete []tmp;
               }

            }
         }
      }
      break;
      case DATA:
      {
         if ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM) ||
	     (to_cs == PLOT_NORM)){

	    if ((data_max[0] == data_min[0]) || 
	        (data_max[1] == data_min[1]))
               return ERR_COORD_DATA_LIM_EQ;

            // This works for simple cases, but need to test
	    // for complex cases. Projections?? 
	    ChipsMatrix calc_matrix;
            Transform *transform = NULL;

            if (wcs_from_id != "")
               transform = transform_map[wcs_from_id];

            if (use_logs){
               ChipsVector tmp_min = data_min;
               ChipsVector tmp_max = data_max;
	       for (int ii=0; ii<data_min.get_size (); ii++){
                  if (log_flag[ii]){
		     if ((data_min[ii] <= 0) || (data_max[ii] <= 0))
                        return ERR_NEG_LOG_VAL;
                     data_min[ii] = log10(data_min[ii]);
                     data_max[ii] = log10(data_max[ii]);
                  }
               }
               endpoint_matrix (calc_matrix);
               data_min = tmp_min;
               data_max = tmp_max;
            }
            else
                endpoint_matrix (calc_matrix);

            if (!calc_matrix.invertable ())
               return ERR_STATUS_BAD;
	    calc_matrix = calc_matrix.inverse ();


            for (int ii=size; ii--; ){
	       if ((in[ii].get_size () != calc_matrix.get_row()) &&
                   (in[ii].get_size () != calc_matrix.get_col()))
                  return ERR_BAD_TRANSFORM_SIZE;
	       if (use_logs){
		  // Initialize out to be the same size as in
		  out[ii] = in[ii];
	          for (int jj=in->get_size (); jj--;){
                     if (log_flag[jj]){
		        if (in[ii][jj] > 0)
                           out[ii][jj] = log10 (in[ii][jj]);
			else
			   return ERR_NEG_LOG_VAL;
                     }
		     else
			out[ii][jj] = in[ii][jj];
		  }
		  out[ii] = calc_matrix * out[ii];
	       }
	       else
               {
                  out[ii] = in[ii];
                  if (transform)
                  {
                     double *tmp = (double *)transform->invert ((void *)(&out[ii][0]),dmDOUBLE,1);
                     double val = out[ii][out[ii].get_size ()-1];
                     for (int jj=0; jj<out[ii].get_size ()-1; jj++)
                        out[ii][jj] = tmp[jj];
                     out[ii][out[ii].get_size ()-1] = val;
                     delete []tmp;
                  }

                  out[ii] = calc_matrix * out[ii];
               }
            }
	    if (parent && to_cs != PLOT_NORM)
	       rval = parent->convert_cs_to_cs (out, out, size, PLOT_NORM, 
						parent->get_id(), to_cs, tmp_to);
	    else if (parent)
	       rval = ERR_STATUS_GOOD;
            else
               rval = ERR_STATUS_BAD;
         }
	 else if (to_cs == DATA){
            Transform *transform_in = NULL;
            Transform *transform_out = NULL;

            if (wcs_from_id != "")
               transform_in = transform_map[wcs_from_id];

            if (wcs_to_id != "")
               transform_out = transform_map[wcs_to_id];

	    // Currently the tmp_to must equal the id of coord. It may
            // worth it in the future to have more flexibility
	    if ((id == tmp_to) || (tmp_to.length () == 0)){
	       for (int ii=size; ii--;)
               {
                  out[ii] = in[ii];

                  if (transform_in)
                  {
                     double *tmp = (double *)transform_in->invert ((void *)(&out[ii][0]),dmDOUBLE,1);
                     for (int jj=0; jj<out[ii].get_size ()-1; jj++)
                        out[ii][jj] = tmp[jj];
                     out[ii][out[ii].get_size ()-1] = in[ii][out[ii].get_size ()-1];
                     delete []tmp;
                  }

                  if (transform_out)
                  {
                     double *tmp = (double *)transform_out->apply ((void *)(&out[ii][0]),dmDOUBLE,1);

                     for (int jj=0; jj<out[ii].get_size ()-1; jj++)
                        out[ii][jj] = tmp[jj];
                     out[ii][out[ii].get_size ()-1] = in[ii][out[ii].get_size ()-1];
                     delete []tmp;
                  }

               }
               rval =  ERR_STATUS_GOOD;
	    }
/*	    else if (parent && (to_cs == DATA))
	       rval = parent->convert_cs_to_cs (in, out, size, PLOT_NORM, 
							id, to_cs, to_id);
*/
            else
               rval = ERR_STATUS_BAD;
         }
      }
      break;
      default: rval = ERR_STATUS_BAD;
   }
   return rval;
}

ChipsCoordinate * ChipsDataSystem::get_coord (COORD_SYS coord_sys, const string &id)
{
   ChipsCoordinate *coord;
   if (parent && ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) || 
       (coord_sys == PIXEL_REAL) || (coord_sys == FRAME_NORM) || 
       (coord_sys == PLOT_NORM)))
      coord = parent->get_coord(coord_sys, id);
   else if (coord_sys == DATA){
      if ((id.length() == 0) || (strip_transform_id(id) == this->id))
         coord = this;
      else
         coord = parent->get_coord(coord_sys, id);
   }
   return coord;
}  


int ChipsDataSystem::insert (ChipsObject *co, bool use_co_info)
{
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   int rval = ERR_STATUS_BAD;
   if (co == NULL)
      return ERR_STATUS_BAD;

   // Store object local forces the object to remain
   // in this coordinate systems list. It overrides
   // the coordinate objects coord system
   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
      case PLOT_NORM:
         if (parent)
	    rval = parent->insert (co, use_co_info);
      break;
      case COORD_NONE:
      case DATA:
      {
         if ((coord_id.length () == 0) || (strip_transform_id (coord_id) == id)){
            chips_objects.push_back (co);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval = ERR_COORD_DS_DNE;
      }
      break;
   }

   return ERR_STATUS_GOOD;
}


int ChipsDataSystem::remove (ChipsObject *co, bool use_co_info)
{
   int rval = ERR_STATUS_BAD;
   COORD_SYS coord_sys;
   string coord_id = "";
   if (co == NULL)
      return ERR_STATUS_BAD;

   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);
      

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
      case PLOT_NORM:
         if (parent)
            rval = parent->remove (co, use_co_info);
      break;
      case COORD_NONE:
      case DATA:
      {
         list <ChipsObject *>::iterator ii = find (chips_objects.begin (), 
						chips_objects.end (), co);
         if (ii != chips_objects.end ()){
            chips_objects.erase (ii);
            rval = ERR_STATUS_GOOD;
	 }
      }
      break;
   }
   return rval;
}


bool ChipsDataSystem::coord_exists (COORD_SYS coord_sys, const string &id)
{
   bool rval = false;
   if (parent && ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
       (coord_sys == PIXEL_REAL) || (coord_sys == FRAME_NORM) || 
       (coord_sys == PLOT_NORM)))
      rval = parent->coord_exists(coord_sys, id);
   else if (coord_sys == DATA){
      if ((id.length() == 0) || (strip_transform_id(id) == this->id))
         rval = true;
   }
   return rval;
}


int ChipsDataSystem::modify_limits (ChipsVector limit, int minmax_flag, const string &transform_id)
{
   int rval = ERR_STATUS_GOOD;
   bool changed = false;
   ChipsVector out_limit (limit.get_size ());


   // convert the dmin and dmax to data from the appropriate 
   // transform
   if ((transform_id != "none") && (this->default_transform.length () > 0))
   {
      string tmp_data_id = id + "_none";
      string tmp_transform_id = id;

      if (transform_id.length () != 0)
      {
         tmp_transform_id += "_";
         tmp_transform_id += transform_id;
      }

      this->convert_cs_to_cs (&limit, &out_limit, 1, DATA, tmp_transform_id, DATA, tmp_data_id);
   }
   else
      out_limit = limit;     


   // min = 0 and max = 1
   if ((minmax_flag == 0) && (this->data_min != out_limit))
   {
      changed = true;
      this->data_min = out_limit;
   }
   else if ((minmax_flag == 1) && (this->data_max != out_limit))
   {
      changed = true;
      this->data_max = out_limit;
   }

   if (changed)
   { 
      compute_pixels_per_unit(); 
/*      rval = update_obj_positions ();
*/
   }
   return rval;
}

int ChipsDataSystem::modify_limits (const ChipsVector &min_limit, const ChipsVector &max_limit, const string &transform_id)
{
   int rval = ERR_STATUS_GOOD;
   ChipsVector out_limit[2];

   out_limit[0] = min_limit;
   out_limit[1] = max_limit;

   // convert the dmin and dmax to data from the appropriate
   // transform
   if ((transform_id != "none") && (this->default_transform.length () > 0))
   {
      string tmp_data_id = id + "_none";
      string tmp_transform_id = id;

      if (transform_id.length () != 0)
      {
         tmp_transform_id += "_";
         tmp_transform_id += transform_id;
      }

      this->convert_cs_to_cs (out_limit, out_limit, 2, DATA, tmp_transform_id, DATA, tmp_data_id);
   }


   // min = 0 and max = 1
   if (this->data_min != out_limit[0])
   {
      modified = true;
      this->data_min = out_limit[0];
   }
   if (this->data_max != out_limit[1])
   {
      modified = true;
      this->data_max = out_limit[1];
   }

   if (modified)
   {
      compute_pixels_per_unit();
/*      rval = update_obj_positions ();
*/
   }
   return rval;
}


int ChipsDataSystem::compute_adjust(int axis, double &min, double& max,
   double scaling)
{
   double lo = min, hi = max, val = 0.0;
   if (scaling == 0.0) scaling = 1.0; 
   if (aspect_use && unit_pix_w && unit_pix_h && !resizing)
   {
      if (axis == X_AXIS)
         val =((this->data_max[0]-this->data_min[0])/unit_pix_h * unit_pix_w) /
	    2.0 * scaling;
      else 
         val =((this->data_max[1]-this->data_min[1])/unit_pix_w * unit_pix_h) /
	    2.0 * scaling;
      min =(hi-lo)/2 + lo - val;
      max =(hi-lo)/2 + lo + val;
   }
   return static_cast <int>(val); 
}

void ChipsDataSystem::set_log_flag (bool flag, int axis)
{
   if ((axis == X_AXIS) && (log_flag[0] != flag))
   {
      log_flag[0] = flag;
/*      if ((data_min[0] > 0) && (data_max[0] > 0))
         update_obj_positions();
*/
   }
   else if ((axis == Y_AXIS) && (log_flag[1] != flag))
   {
      log_flag[1] = flag;
/*      if ((data_min[1] > 0) && (data_max[1] > 0))
         update_obj_positions();
*/
   }
   modified = true;
}

bool ChipsDataSystem::get_log_flag (int axis)
{
   if (axis == X_AXIS)
      return log_flag[0];
   if (axis == Y_AXIS)
      return log_flag[1];

   return false;
}

void ChipsDataSystem::endpoint_matrix (ChipsMatrix &mtrx)
{
   int size = data_max.get_size ();
   mtrx.set_size(size,size);
   mtrx.identity ();
   for (int ii=0; ii<size; ii++){
      mtrx[ii][ii] = data_max[ii] - data_min[ii];
      mtrx[ii][size-1] = data_min[ii];
   }
}


int ChipsDataSystem::deserialize (ChipsStateStream &in)
{
   int sz = 0;
   long lval = 0;
   int rval = ERR_STATUS_GOOD;

   if (in.get_minor_version() > 6)
   {
      in.read_long ((long *)&(this->global_ref_id));
      in.read_long ((long *)&(this->ref_id));
      in.read_string (&this->id);
   }
   else
   {
      this->ref_id = this->global_ref_id;
      this->global_ref_id++;
      in.read_string (&this->old_id);
      this->create_id (this->id);
   }

   in.read_bool (&visible);
   in.read_int (&sz);
   if (sz > 0){
      in.read_double (&data_min[0], sz);
      in.read_double (&data_max[0], sz);
   }
   else 
      rval = ERR_STATEFILE_COR;

   if (!rval){
      in.read_int (&sz);
      if (sz)
         in.read_bool (log_flag,sz);

      in.read_string (&xaxis_id);
      in.read_string (&yaxis_id);

      // read in axis reference which isn't needed
      if (in.get_minor_version() < 7)
         in.read_int (&sz);
   }

   if (in.get_minor_version() > 0)
   {
      Transform *transform;
      string transform_id;

      in.read_bool(&aspect_use);
      in.read_double(&aspect_w);
      in.read_double(&aspect_h);
      in.read_string(&ratio);
      in.read_double (&unit_pix_w);
      in.read_double (&unit_pix_h);

      if (in.get_minor_version() > 1)
      {
         in.read_long (&lval);
         transform_ref = lval;

         in.read_string (&default_transform);
         in.read_int (&sz);

         for (int ii=0; ii<sz; ii++)
         {
            in.read_string (&transform_id);
            transform = ChipsTransformFactory::deserialize (in);
            transform_map[transform_id] = transform;
         }
      }
   }
   else
   {
      aspect_use = false;
      aspect_w = 0;
      aspect_h = 0; 
      ratio = "";
      unit_pix_w = 0;
      unit_pix_h = 0;
      if (in.get_minor_version() > 1)
      {
         transform_ref = 0;
         default_transform = "";
      }
   }

   return rval;
};


void ChipsDataSystem::serialize (ChipsStateStream &out)
{
   long lval =0;
   int sz = 0;

   if (out.get_minor_version() > 6)
   {
      out.write_long ((long *)&(this->global_ref_id));
      out.write_long ((long *)&(this->ref_id));
   }

   out.write_string (&id);
   out.write_bool (&visible);
   sz = data_min.get_size ();
   out.write_int (&sz);
   if (sz > 0){
      out.write_double (&data_min[0], sz);
      out.write_double (&data_max[0], sz);
   }

   if (log_flag){
      sz = 3;
      out.write_int (&sz);
      out.write_bool (log_flag,3);

   }
   else{
      sz = 0;
      out.write_int (&sz);
   } 
	
   out.write_string (&xaxis_id);
   out.write_string (&yaxis_id);

   if (out.get_minor_version() > 0)
   {
      map <string, Transform *>::iterator iter;
      string transform_id;
      Transform *transform;

      out.write_bool(&aspect_use);
      out.write_double(&aspect_w);
      out.write_double(&aspect_h);
      out.write_string(&ratio);
      out.write_double (&unit_pix_w);
      out.write_double (&unit_pix_h);

      if (out.get_minor_version() > 1)
      {
         lval = transform_ref;
         out.write_long (&lval);

         out.write_string (&default_transform);

         sz = transform_map.size ();
         out.write_int (&sz);

         for (iter=transform_map.begin (); iter != transform_map.end (); iter++)
         {
            transform_id = iter->first;
            transform = iter->second;
            out.write_string (&transform_id);
            ChipsTransformFactory::serialize (transform, out);
         }
      }
   }

};


int ChipsDataSystem::set_aspect_ratio(string val)
{
   int rval = ERR_STATUS_GOOD; 
   float daw, dah;

   modified = true;
   if (val.length() == 0)
   {
      aspect_use = 0; 
      ratio.clear();
   }
   else if ((rval = string_to_ratio_values(val.c_str(), daw, dah)) ==
       ERR_STATUS_GOOD)
   {
      aspect_w = daw;
      aspect_h = dah; 
      aspect_use = 1;
      ratio = val; 
      compute_pixels_per_unit(); 
   }
   return rval;
}

void ChipsDataSystem::compute_pixels_per_unit()
{
   ChipsVector out0 (0.0, 0.0);
   ChipsVector out1 (0.0, 0.0);
   this->convert_cs_to_cs(&data_min, &out0, 1, DATA, this->get_id_no_wcs(), PIXEL_REAL, "");
   this->convert_cs_to_cs(&data_max, &out1, 1, DATA, this->get_id_no_wcs(), PIXEL_REAL, "");
   unit_pix_w = (out1[0] - out0[0]) / (data_max[0] - data_min[0]); 
   if (aspect_w == 0) aspect_w = 1; 
   unit_pix_h = unit_pix_w * (aspect_w / aspect_h); 
}

string ChipsDataSystem::add_transform (Transform *transform)
{
   ostringstream transform_id;
   string return_id;
   Transform *tmp_transform = NULL;

   // check that this transform isn't already in the system
   map <string, Transform *>::iterator iter;

   // see if the transform is already in the system. if so return that
   // transform id
   for (iter=this->transform_map.begin (); iter!=this->transform_map.end (); iter++)
   {
      if (ChipsTransformFactory::compare (iter->second, transform))
         return iter->first;
   }
  

   transform_id << "wcs" << this->transform_ref++;
   return_id = transform_id.str ();
   this->transform_map[return_id] = ChipsTransformFactory::copy_transform (transform);

   // set the default transform if it hasn't been set yet
   if (this->default_transform.length () == 0)
   {
      // make sure the rotation is zero 
      // ** This really belongs in the AxisAlgorithm code 
      // ** We should not for all data systems to have no rotation
      double *rotation = (double *)ChipsTransformFactory::get_parameter_copy (transform,"rotation");
      if (rotation && (*rotation != 0))
      {
         transform_id.str ("");
         tmp_transform = ChipsTransformFactory::copy_transform (transform);
         if (rotation)
         {
            *rotation = 0;
            ChipsTransformFactory::set_parameter<double> (tmp_transform,"rotation", rotation, 1);
         }
       
         transform_id << "wcs" << this->transform_ref++;
         this->transform_map[transform_id.str ()] = tmp_transform;  

      }

      if (rotation)
         delete []rotation;

      this->default_transform = transform_id.str ();
      modified = true;
   }
   return return_id;
}

void ChipsDataSystem::remove_transform (const string &id)
{
   map <string, Transform *>::iterator iter;

   iter = this->transform_map.find (id);
   if (iter != this->transform_map.end ()) 
   {
      delete iter->second;
      transform_map.erase (iter);
   }

   if (id == this->default_transform)
   {
      modified = true;
      if (transform_map.empty ())
         this->default_transform = "";
      else
         this->default_transform = transform_map.begin ()->first;
   }
    
}

Transform *ChipsDataSystem::get_transform (const string &id)
{
   string transform_id;
   map <string, Transform *>::iterator iter;

   // First check if the id is a full data system id with
   // transform at the end. If it's not assume that it's
   // purely the transform id.

   if (id == this->id)
      transform_id = this->default_transform;
   else
   {
      transform_id = this->strip_coord_id (id);

      if (transform_id.length () == 0)
         transform_id = id;
   }
   
   
   iter = this->transform_map.find (transform_id);
   if (iter != this->transform_map.end ())
      return iter->second;
   return NULL;
}

void ChipsDataSystem::set_transform (const string &id, Transform *transform)
{
   // if the transform is null remove it
   if (transform == NULL)
      remove_transform (id);
   else if (id.length () == 0)
      add_transform (transform);
   else
      transform_map[id] = ChipsTransformFactory::copy_transform (transform);
}

void ChipsDataSystem::set_default_transform_id (const string &transform_id)
{
   map <string, Transform *>::iterator iter;

   iter = this->transform_map.find (transform_id);
   if (iter != this->transform_map.end ())
      this->default_transform = transform_id;

}



string ChipsDataSystem::strip_transform_id (const string &data_id)
{
   string out_id = data_id;
   int pos = 0;

   // ***** NEED TO FIX THIS FOR BACKWARDS COMPATIBILITY?? ********

   pos = data_id.find ('_', 0);

   if (pos > 0)
      out_id = data_id.substr (0,pos);

   return out_id;
}

string ChipsDataSystem::strip_coord_id (const string &data_id)
{
   string out_id = "";
   int pos = 0;

   // ***** NEED TO FIX THIS FOR BACKWARDS COMPATIBILITY?? ********

   pos = data_id.find ('_', 0);

   if (pos > 0)
      out_id = data_id.substr (pos+1,data_id.length ());

   return out_id;
}


ChipsVector ChipsDataSystem::get_lower_bounds (const string &transform_id)
{
   ChipsVector min_coord (data_min);
   string tmp_id = this->id;
   if (transform_id.length () != 0)
   {
      tmp_id += "_";
      tmp_id += transform_id;
   }

   this->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, this->get_id_no_wcs(), DATA, tmp_id);

   return min_coord;
}

ChipsVector ChipsDataSystem::get_upper_bounds (const string &transform_id)
{
   ChipsVector max_coord (data_max);
   string tmp_id = this->id;
   if (transform_id.length () != 0)
   {
      tmp_id += "_";
      tmp_id += transform_id;
   }
   this->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, this->get_id_no_wcs(), DATA, tmp_id);

   return max_coord;
}

int ChipsDataSystem::update_obj_positions (bool update_all)
{
   modified = false;
   return ChipsCoordinate::update_obj_positions (update_all);
}

void ChipsDataSystem::create_id (string &curr_id)
{
   stringstream tmp_id;

   if (this->parent)
      this->parent->create_id (curr_id);

   tmp_id << curr_id << "." << this->ref_id;
   curr_id = tmp_id.str ();
}

void ChipsDataSystem::new_id ()
{
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->id = "";
   create_id (this->id);
}

/*
bool ChipsDataSystem::validate_coordinates (const ChipsVector &coord, const string &transform_id, ChipsVector *x_wrap_around, ChipsVector *y_wrap_around)
{
   ChipsVector tmp_coord (coord);
   ChipsVector comp_coord (coord);
   double epsilon = 10e-7;
   bool valid = true;

   string tmp_transform_id = this->strip_coord_id (transform_id);

   if (tmp_transform_id.length () == 0)
      tmp_transform_id = tmp_transform_id;

   // convert the dmin and dmax to data from the appropriate
   // transform
   if ((tmp_transform_id != "none") && (this->default_transform.length () > 0))
   {
      string tmp_data_id = id + "_none";

      if (transform_id.length () != 0)
         tmp_transform_id = this->id;
      else if (tmp_transform_id == transform_id)
      {
         tmp_transform_id += "_";
         tmp_transform_id += transform_id;
      }

      this->convert_cs_to_cs (&coord, &tmp_coord, 1, DATA, tmp_transform_id, DATA, tmp_data_id);
      this->convert_cs_to_cs (&tmp_coord, &comp_coord, 1, DATA, tmp_data_id, DATA, tmp_transform_id);
      

      for (int ii=0; valid && ii<coord.get_size (); ii++)
      {
         if (x_wrap_around)
         { 
            if (comp_coord[0] < (*x_wrap_around)[0])
               comp_coord[0] = (*x_wrap_around)[1] - ((*x_wrap_around)[0] - comp_coord[0]);
            else if (((comp_coord[0] > (*x_wrap_around)[1])) || (fabs(comp_coord[0] - (*x_wrap_around)[1]) < epsilon))
               comp_coord[0] = (*x_wrap_around)[0] + (comp_coord[0] - (*x_wrap_around)[1]);
         }
         if (y_wrap_around)
         {
            if (comp_coord[1] < (*y_wrap_around)[0])
               comp_coord[1] = (*y_wrap_around)[1] - ((*y_wrap_around)[0] - comp_coord[1]);
            else if (((comp_coord[0] > (*y_wrap_around)[1])) || (fabs(comp_coord[0] - (*y_wrap_around)[1]) < epsilon))
               comp_coord[1] = (*y_wrap_around)[0] + (comp_coord[1] - (*y_wrap_around)[1]);
         }

          
         if (fabs (comp_coord[ii] - coord[ii]) > epsilon)
            valid = false;
      }
   }

   return valid;

}
*/

string ChipsDataSystem::get_id (const string &transform_id) 
{ 
   return (transform_id.length () > 0) ? this->id+"_"+transform_id : this->id;
}

