// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSDATASYSTEM_HH
#define CHIPSDATASYSTEM_HH


#include <string>
#include <map>
#include "ChipsObject.hh"
#include "ChipsDefines.hh"
#include "ChipsMatrix.hh"
#include "ChipsCoordinate.hh"
#include "Transform.hh"

/**************************************************************
*
*    Data System
*
*    This is a collection of information the describes the
*    data coordinate systems which relates to coorinate
*    systems in a plot area. This contains information
*    about what the data limits are and in frame normalized
*    coordinates where the plot position is. It also contains
*    a list of objects associating with this coordinate system.
*
****************************************************************/
class ChipsDataSystem: public ChipsCoordinate
{
   public:

      // Constructors
      ChipsDataSystem (ChipsCoordinate *par);
      ChipsDataSystem (ChipsCoordinate *par,
                const ChipsVector &ds_data_min, const ChipsVector  &ds_data_max,
                const string &ax_id, const string &ay_id);
      ChipsDataSystem (const ChipsDataSystem &);
      ChipsDataSystem (const string &ds_id, const ChipsVector &ds_data_min, const ChipsVector &ds_data_max,
                     const string &ax_id, const string &ay_id);
      ChipsDataSystem (ChipsStateStream &in, ChipsCoordinate *par);

      virtual ~ChipsDataSystem ();

      // Overloaded operators
      void operator = (const ChipsDataSystem &);

      virtual int convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id);
      virtual ChipsCoordinate *get_coord (COORD_SYS coord_sys, const string &id="");
      virtual int insert (ChipsObject *co, bool use_co_info=true);
      virtual int remove (ChipsObject *co, bool use_co_info=true);
      virtual bool coord_exists (COORD_SYS coord_sys, const string &id);
      virtual int update_obj_positions (bool update_all=true);

      // Update the data limits and move objects accordingly
      int modify_limits (ChipsVector limit, int minmax_flag, const string & transform_id);
      int modify_limits (const ChipsVector &min_limit, const ChipsVector &max_limit, const string & transform_id);


      // transform functions
      string add_transform (Transform *transform);
      void remove_transform (const string &id);
      string get_default_transform_id () { return this->default_transform; };
      void set_default_transform_id (const string &transform_id);
      void set_transform (const string &id, Transform *transform);

      string get_id (const string &transform_id);
      string get_id () { return this->id;};
      void new_id ();
      string get_id_no_wcs () { return this->id+"_none";};
      Transform *get_transform (const string &id);

      // Set and Get functions
      void set_log_flag (bool flag, int axis);
      bool get_log_flag (int axis);
      string get_xaxis_id () {return xaxis_id;};
      string get_yaxis_id () {return yaxis_id;};
      void set_xaxis_id (const string &id) {xaxis_id = id;};
      void set_yaxis_id (const string &id) {yaxis_id = id;};

      virtual ChipsVector get_lower_bounds (){return data_min;};
      virtual ChipsVector get_upper_bounds (){return data_max;};
      ChipsVector get_lower_bounds (const string &transform_id);
      ChipsVector get_upper_bounds (const string &transform_id);
      virtual COORD_SYS   get_bounds_coord_sys () {return DATA;};

      // create an id from all parent information
      virtual void create_id (string &curr_id);

      string get_old_id () {return this->old_id;};

      virtual int deserialize (ChipsStateStream &in);
      virtual void serialize (ChipsStateStream &out);

      // These functions should be deprecated. Use get_lower/upper_bounds
      double get_x_min () {return data_min[0];};
      double get_x_max () {return data_max[0];};
      double get_y_min () {return data_min[1];};
      double get_y_max () {return data_max[1];};
      double get_unit_pix_w ()  { return unit_pix_w; }
      double get_unit_pix_h ()  { return unit_pix_h; }
      double get_aspect_width() { return aspect_w; }
      double get_aspect_height() { return aspect_h; }

      int set_aspect_ratio(string val);
      string get_aspect_ratio()  { return ratio; }  
      bool use_ratio()           { return aspect_use; }
      int compute_adjust(int axis, double &min, double &max, double scaling=1.0);
      void  set_resizing(int val)  { resizing = val; }   
      int   get_resizing()     { return resizing; }

      bool  get_modified ()    { return modified; };

/*      bool  validate_coordinates (const ChipsVector &coord, const string &transform_id, ChipsVector *x_wrap_around=NULL, ChipsVector *y_wrap_around=NULL);
*/
   protected:
      string strip_transform_id (const string &data_id);
      string strip_coord_id (const string &data_id);
      void compute_pixels_per_unit();
      void endpoint_matrix (ChipsMatrix &mtrx);

      ChipsVector data_min;
      ChipsVector data_max;

      string ratio;
      double aspect_w; 
      double aspect_h; 
      bool   aspect_use; 

      double unit_pix_w;    // average pixels per width unit;
      double unit_pix_h;    // average pixels per height unit;

      bool *log_flag;                      //data is in log scale

      // axis id's are necessary in coordinate info command
      string xaxis_id;                      //x axis id used to create the system
      string yaxis_id;                      //y axis id used to create the system

      int resizing; 

      unsigned long transform_ref;

      string default_transform;             // the default transform id
      map <string, Transform *> transform_map;   // map of transforms known to this data system
      int refcnt;                           // count for the id

      bool modified;

      unsigned long ref_id;               // reference id
      static unsigned long global_ref_id; // reference id
      string old_id;                      // the old id tag for backwards compat.


};


#endif
