// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <sstream>

#include "ChipsDefines.hh"
#include "ChipsColor.hh"


const char* ChipsSymbolStyles[10] = {
   "none", "cross", "diamond", "downtriangle", "circle", "plus", "square", 
   "uptriangle", "point", "arrow" }; 

const char* ChipsLineStyles[9] = {
   "noline", "solid", "dot", "dotlongdash", "dotshortdash", "longdash", 
   "shortdash", "shortdashlongdash" };

const char* ChipsFillStyles[17] = {
   "nofill", "solidfill", "updiagonal", "downdiagonal", "horizontal", "vertical",
   "crisscross", "brick", "grid", "hexagon", "polkadot", "zigzag", "wave",
   "flower", "userfill1", "userfill2", "userfill3"}; 

const char* ChipsRGBAStyles[11] = {
   "red", "green", "blue", "alpha"};

const char* ChipsColorMapStyles[11] = {
   "red", "green", "blue", "grayscale", "rainbow", "hsv", "heat", "cool",
   "usercmap1", "usercmap2", "usercmap3"};

const char* ChipsInterpolationStyles[3] = {
   "neighbor", "bilinear", "bicubic"};

const char* ChipsUnitTypes[4] = {
   "cm", "mm", "inch", "pixels" }; 

const char* ChipsFontStyles[4] = {
   "helvetica", "courier", "times", "greek" };

const char* ChipsContourStyles[2] = {
   "standard", "marching" };

const char* ChipsFontStyleStyles[4] = {
   "normal", "bold", "italic", "bolditalic" };

const char* ChipsErrStyleStyles[2] = {
   "line", "capped" };

const char* ChipsBadDataModeStyles[2] = {
   "ignore", "omit" };

const char* ChipsAxisStyleStyles[3] = {
   "open", "boxed", "closed" };

const char* ChipsCornerStyles[4] = {
   "miter", "bevel", "filled", "notch" };

const char* ChipsTickModeStyles[5] = {
   "count", "interval", "nice", "limits", "arbitrary" };

const char* ChipsTickStyleStyles[3] = {
   "inside", "outside", "centered" };

const char* ChipsHAlignStyles[4] = {
   "left", "center", "right", "auto" };

const char* ChipsVAlignStyles[4] = {
   "base", "center", "top", "auto" };

const char* ChipsVPrintfileType[5] = {
   "default", "rgb", "cmyk", "grayscale", "cmyksep" };

const char* ChipsVPagesizeType[7] = {
   "default", "letter", "legal", "executive", "A3", "A4", "A5" };


const char* ChipsObjectTypes[17] = {
   "non-object", 
   "object", "window", "frame", "image", "plot", "curve", "histogram",
   "contour", "axis", "label", "line", "legend", "point", "region", 
   "colorbar","coordsys" };

const char* ChipsLineExtensionTypes[4] = {"none", "start", "end", "both"};

const char * ChipsCoordSysTypes[7] = {
   "none", "pixel", "frame normalized", "data", "plot normalized",
   "window normalized", "real pixel" };

const char * ChipsCoordSysEnums[7] = {"COORD_NONE", "PIXEL", "FRAME_NORM", 
   "DATA", "PLOT_NORM", "WINDOW_NORM", "PIXEL_REAL"};


const char* v_symbol_to_string(V_SYMBOL_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_arrow:          // fallthrough intended
      case chips_circle:         // fallthrough intended
      case chips_cross:          // fallthrough intended
      case chips_diamond:        // fallthrough intended
      case chips_downtriangle:   // fallthrough intended
      case chips_none:           // fallthrough intended
      case chips_plus:           // fallthrough intended
      case chips_square:         // fallthrough intended
      case chips_uptriangle:   
      case chips_point_type:
         ans = ChipsSymbolStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}

const char* v_line_to_string(V_LINE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_solid:          // fallthrough intended
      case chips_dot:            // fallthrough intended
      case chips_dotlongdash:    // fallthrough intended
      case chips_dotshortdash:   // fallthrough intended
      case chips_longdash:       // fallthrough intended
      case chips_shortdash:      // fallthrough intended
      case chips_shortdashlongdash:  // fallthrough intended
      case chips_noline:  
         ans = ChipsLineStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}

const char* v_fill_to_string(V_FILL_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_nofill:         // fallthrough intended
      case chips_solidfill:      // fallthrough intended
      case chips_updiagonal:     // fallthrough intended
      case chips_downdiagonal:   // fallthrough intended
      case chips_horizontal:     // fallthrough intended
      case chips_vertical:       // fallthrough intended
      case chips_crisscross:     // fallthrough intended
      case chips_brick:          // fallthrough intended
      case chips_grid:           // fallthrough intended
      case chips_hexagon:        // fallthrough intended
      case chips_polkadot:       // fallthrough intended
      case chips_zigzag:         // fallthrough intended
      case chips_wave:           // fallthrough intended
      case chips_flower:         // fallthrough intended
      case chips_userfill1:      // fallthrough intended
      case chips_userfill2:      // fallthrough intended
      case chips_userfill3:  
         ans = ChipsFillStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}

const char* v_colormap_to_string(V_COLORMAP_TYPE val)
{
   const char *ans = NULL;
   switch(val)
   {
      case chips_red:            // fallthrough intended
      case chips_green:          // fallthrough intended
      case chips_blue:           // fallthrough intended
      case chips_grayscale:      // fallthrough intended
      case chips_rainbow:        // fallthrough intended
      case chips_hsv:            // fallthrough intended
      case chips_heat:           // fallthrough intended
      case chips_cool:           // fallthrough intended
      case chips_usercmap1:      // fallthrough intended
      case chips_usercmap2:      // fallthrough intended
      case chips_usercmap3:  
         ans = ChipsColorMapStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}

const char* v_rgba_to_string(V_RGBA_TYPE val)
{
   const char *ans = NULL;
   switch(val)
   {
      case chips_red:            // fallthrough intended
      case chips_green:          // fallthrough intended
      case chips_blue:           // fallthrough intended
      case chips_alpha:
         ans = ChipsRGBAStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_interpolation_to_string(V_INTERPOLATION_TYPE val)
{
   const char *ans = NULL;
   switch(val)
   {
      case chips_neighbor:       // fallthrough intended
      case chips_bilinear:       // fallthrough intended
      case chips_bicubic:
         ans = ChipsInterpolationStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}

const char* v_units_to_string(V_UNITS_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_cm:             // fallthrough intended
      case chips_mm:             // fallthrough intended
      case chips_inch:           // fallthrough intended
      case chips_pixels:         // fallthrough intended
         ans = ChipsUnitTypes[val-1]; // -1 accounts for chips_nounit
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_font_to_string(V_FONT_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_helvetica:          // fallthrough intended
      case chips_courier:        // fallthrough intended
      case chips_times:
      case chips_greek:
         ans = ChipsFontStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_contour_to_string(V_CONTOUR_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_standard:       // fallthrough intended
      case chips_marching:
         ans = ChipsContourStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_fontstyle_to_string(V_FONTSTYLE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_normal:         // fallthrough intended
      case chips_bold:           // fallthrough intended
      case chips_italic:         // fallthrough intended
      case chips_bolditalic:
         ans = ChipsFontStyleStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_errstyle_to_string(V_ERRSTYLE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_bar:             // fallthrough intended
      case chips_cap:
         ans = ChipsErrStyleStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_baddatamode_to_string(V_BADDATAMODE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_ignore:             // fallthrough intended
      case chips_omit:
         ans = ChipsBadDataModeStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_axisstyle_to_string(V_AXISSTYLE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_open:             // fallthrough intended
      case chips_box:              // fallthrough intended
      case chips_closed:
         ans = ChipsAxisStyleStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_cornerstyle_to_string(V_CORNERSTYLE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_miter:             // fallthrough intended
      case chips_bevel:             // fallthrough intended
      case chips_fill:              // fallthrough intended
      case chips_notch:
         ans = ChipsCornerStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_tickmode_to_string(V_TICKMODE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_count:     // fallthrough intended
      case chips_interval:  // fallthorugh intended
      case chips_nice:      // fallthorugh intended 
      case chips_limits:    // fallthorugh intended
      case chips_arbitrary: 
         ans = ChipsTickModeStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_tickstyle_to_string(V_TICKSTYLE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_inside:   // fallthrough intended
      case chips_outside:  // fallthorugh intended
      case chips_centered:
         ans = ChipsTickStyleStyles[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_printfile_to_string(V_PRINTFILE_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_nofiletype: // fallthrough intended
      case chips_rgb:        // fallthrough intended
      case chips_cmyk:       // fallthrough intended
      case chips_grayscale:  // fallthrough intended
      case chips_cmyksep:
         ans = ChipsVPrintfileType[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* voffset_to_string(float val)
{
   const char *ans = NULL;

   if (approx_equal(val, 0.0))
      ans = ChipsVAlignStyles[0];
   else if (approx_equal(val, 0.5))
      ans = ChipsVAlignStyles[1];
   else if (approx_equal(val, 1.0))
      ans = ChipsVAlignStyles[2];
   else if (approx_equal(val, AUTO_ALIGN))
      ans = ChipsVAlignStyles[3];
   return ans;
}


const char* hoffset_to_string(float val)
{
   const char *ans = NULL;

   if (approx_equal(val, 0.0))
      ans = ChipsHAlignStyles[0];
   else if (approx_equal(val, 0.5))
      ans = ChipsHAlignStyles[1];
   else if (approx_equal(val, 1.0))
      ans = ChipsHAlignStyles[2];
   else if (approx_equal(val, AUTO_ALIGN))
      ans = ChipsHAlignStyles[3];
   return ans;
}


const char* v_pagesize_to_string(float w, float h, V_UNITS_TYPE u)
{
   const char *ans = NULL;
   switch (u){
      case chips_mm:
      {
         if ((w == 148) && (h == 210))
            ans = ChipsVPagesizeType[6];
         else if ((w == 210) && (h == 297))
            ans = ChipsVPagesizeType[5];
         if ((w == 297) && (h == 420))
            ans = ChipsVPagesizeType[4];
      }
      break;
      case chips_inch:
      {
         if ((w == 8.5) && (h == 11))
            ans = ChipsVPagesizeType[1];
         if ((w == 8.5) && (h == 14))
            ans = ChipsVPagesizeType[2];
         if ((w == 7.5) && (h == 10))
            ans = ChipsVPagesizeType[3];
      }
      break;
      default:
      break; 
   }
   return ans;
}

const char* v_object_to_string(V_OBJECT_TYPE obj)
{
   const char *ans = NULL;
   switch (obj)
   {
      case chips_window:        // fallthrough intended:
      case chips_frame:         // fallthrough intended:
      case chips_image:         // fallthrough intended:
      case chips_plot:          // fallthrough intended:
      case chips_curve:         // fallthrough intended:
      case chips_histogram:     // fallthrough intended:
      case chips_contour:       // fallthrough intended:
      case chips_axis:          // fallthrough intended:
      case chips_label:         // fallthrough intended:
      case chips_line:          // fallthrough intended:
      case chips_legend:        // fallthrough intended:
      case chips_point:         // fallthrough intended:
      case chips_region:        // fallthrough intended:
      case chips_colorbar:      // fallthrough intended:
      case chips_coordinate_system:  
         ans = ChipsObjectTypes[obj];
      break;
      default:
      break; 
   }
   return ans;
}


const char* v_line_extension_to_string (V_LINE_EXTENSION_TYPE val)
{
   const char *ans = NULL;

   switch(val)
   {
      case chips_extend_none:   // fallthrough intended
      case chips_extend_start:  // fallthrough intended
      case chips_extend_end:    // fallthrough intended
      case chips_extend_both:   // fallthrough intended
         ans = ChipsLineExtensionTypes[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_coord_sys_to_string (COORD_SYS val)
{
   const char *ans = NULL;

   switch(val)
   {
      case COORD_NONE:   // fallthrough intended
      case PIXEL:        // fallthrough intended
      case FRAME_NORM:   // fallthrough intended
      case DATA:         // fallthrough intended
      case PLOT_NORM:    // fallthrough intended
      case WINDOW_NORM:  // fallthrough intended
      case PIXEL_REAL:   // fallthrough intended

         ans = ChipsCoordSysTypes[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


const char* v_coord_sys_to_enum_string (COORD_SYS val)
{
   const char *ans = NULL;

   switch(val)
   {
      case COORD_NONE:   // fallthrough intended
      case PIXEL:        // fallthrough intended
      case FRAME_NORM:   // fallthrough intended
      case DATA:         // fallthrough intended
      case PLOT_NORM:    // fallthrough intended
      case WINDOW_NORM:  // fallthrough intended
      case PIXEL_REAL:   // fallthrough intended
         ans = ChipsCoordSysEnums[val];
      break;
      default:
         // do nothing
      break;
   }
   return ans;
}


int string_to_v_line(const char* val, V_LINE_TYPE* line)
{
   int rval = ERR_STATUS_GOOD;

   if (val && line)
   {
      if (!chips_strcmp(val, "none") || !chips_strcmp(val, "noline") || 
               !chips_strcmp(val, "0"))
         *line = chips_noline;
      else if (!chips_strcmp(val, "solid") || !chips_strcmp(val, "1")) 
         *line = chips_solid;
      else if (!chips_strcmp(val, "dot") || !chips_strcmp(val, "2"))
         *line = chips_dot;
      else if (!chips_strcmp(val, "dotlongdash") || !chips_strcmp(val, "3")) 
         *line = chips_dotlongdash;
      else if (!chips_strcmp(val, "dotshortdash") || !chips_strcmp(val, "4"))
         *line = chips_dotshortdash;
      else if (!chips_strcmp(val, "longdash") || !chips_strcmp(val, "5"))
         *line = chips_longdash;
      else if (!chips_strcmp(val, "shortdash") || !chips_strcmp(val, "6"))
         *line = chips_shortdash;
      else if (!chips_strcmp(val,"shortdashlongdash") || !chips_strcmp(val,"7"))
         *line = chips_shortdashlongdash;
      else
      {
         rval = ERR_BAD_LINESTYLE;
         *line = (V_LINE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified 
 
   return rval; 
}


int string_to_v_fill(const char* val, V_FILL_TYPE* fill)
{
   int rval = ERR_STATUS_GOOD;

   if (val && fill)
   {
      if (!chips_strcmp(val, "nofill") || !chips_strcmp(val, "none") || 
               !chips_strcmp(val, "0"))
         *fill = chips_nofill;
      else if (!chips_strcmp(val, "solid") || !chips_strcmp(val, "solidfill") || 
               !chips_strcmp(val, "1"))
         *fill = chips_solidfill;
      else if (!chips_strcmp(val, "updiagonal") ||
            !chips_strcmp(val, "updia") || !chips_strcmp(val, "2"))
         *fill = chips_updiagonal;
      else if (!chips_strcmp(val, "downdiagonal") ||
            !chips_strcmp(val, "downdia") || !chips_strcmp(val, "3"))
         *fill = chips_downdiagonal;
      else if (!chips_strcmp(val, "horizontal") || !chips_strcmp(val, "4"))
         *fill = chips_horizontal;
      else if (!chips_strcmp(val, "vertical") || !chips_strcmp(val, "5"))
         *fill = chips_vertical;
      else if (!chips_strcmp(val, "crisscross") || !chips_strcmp(val, "6"))
         *fill = chips_crisscross;
      else if (!chips_strcmp(val, "brick") || !chips_strcmp(val, "7"))
         *fill = chips_brick;
      else if (!chips_strcmp(val, "grid") || !chips_strcmp(val, "8"))
         *fill = chips_grid;
      else if (!chips_strcmp(val, "hexagon") || !chips_strcmp(val, "9"))
         *fill = chips_hexagon;
      else if (!chips_strcmp(val, "polkadot") || !chips_strcmp(val, "10"))
         *fill = chips_polkadot;
      else if (!chips_strcmp(val, "zigzag") || !chips_strcmp(val, "11"))
         *fill = chips_zigzag;
      else if (!chips_strcmp(val, "wave") || !chips_strcmp(val, "12"))
         *fill = chips_wave;
      else if (!chips_strcmp(val, "flower") || !chips_strcmp(val, "13"))
         *fill = chips_flower;
      else if (!chips_strcmp(val, "userfill1") || !chips_strcmp(val, "14"))
         *fill = chips_userfill1;
      else if (!chips_strcmp(val, "userfill2") || !chips_strcmp(val, "15"))
         *fill = chips_userfill2;
      else if (!chips_strcmp(val, "userfill3") || !chips_strcmp(val, "16"))
         *fill = chips_userfill3;
      else
      {
         rval = ERR_BAD_FILLSTYLE;
         *fill = (V_FILL_TYPE)-1;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified 
 
   return rval; 
}


int string_to_v_colormap(const char* val, V_COLORMAP_TYPE* cmap)
{
   int rval = ERR_STATUS_GOOD;

   if (val && cmap)
   {
      if (!chips_strcmp(val, "red") || !chips_strcmp(val, "0"))
         *cmap = (V_COLORMAP_TYPE)(chips_red);
      else if (!chips_strcmp(val, "green") || !chips_strcmp(val, "1"))
         *cmap = (V_COLORMAP_TYPE)(chips_green);
      else if (!chips_strcmp(val, "blue") || !chips_strcmp(val, "2"))
         *cmap = (V_COLORMAP_TYPE)(chips_blue);
      else if (!chips_strcmp(val, "grayscale") || !chips_strcmp(val, "3") || 
               !chips_strcmp(val, "greyscale"))
         *cmap = (V_COLORMAP_TYPE)(chips_grayscale);
      else if (!chips_strcmp(val, "rainbow") || !chips_strcmp(val, "4"))
         *cmap = chips_rainbow;
      else if (!chips_strcmp(val, "hsv") || !chips_strcmp(val, "5"))
         *cmap = chips_hsv;
      else if (!chips_strcmp(val, "heat") || !chips_strcmp(val, "6"))
         *cmap = chips_heat;
      else if (!chips_strcmp(val, "cool") || !chips_strcmp(val, "7"))
         *cmap = chips_cool;
      else if (!chips_strcmp(val, "usercmap1") || !chips_strcmp(val, "8"))
         *cmap = chips_usercmap1;
      else if (!chips_strcmp(val, "usercmap2") || !chips_strcmp(val, "9"))
         *cmap = chips_usercmap2;
      else if (!chips_strcmp(val, "usercmap3") || !chips_strcmp(val, "10"))
         *cmap = chips_usercmap3;
      else
      {
         *cmap = chips_cmap_size;
         rval = ERR_BAD_CMAP;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified 
   return rval; 
}

int string_to_v_rgba(const char* val, V_RGBA_TYPE* cmap)
{
   int rval = ERR_STATUS_GOOD;

   if (val && cmap)
   {
      if (!chips_strcmp(val, "red") || !chips_strcmp(val, "0"))
         *cmap = chips_red;
      else if (!chips_strcmp(val, "green") || !chips_strcmp(val, "1"))
         *cmap = chips_green;
      else if (!chips_strcmp(val, "blue") || !chips_strcmp(val, "2"))
         *cmap = chips_blue;
      else if (!chips_strcmp(val, "alpha") || !chips_strcmp(val, "3"))
         *cmap = chips_alpha;
      else 
      {
         rval = ERR_BAD_RGBA;
         *cmap = (V_RGBA_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified
   return rval;
}




int string_to_v_interpolation(const char* val, V_INTERPOLATION_TYPE* cmap)
{
   int rval = ERR_STATUS_GOOD;

   if (val && cmap)
   {
      if (!chips_strcmp(val, "neighbor") || !chips_strcmp(val, "0"))
         *cmap = chips_neighbor;
      else if (!chips_strcmp(val, "bilinear") || !chips_strcmp(val, "1"))
         *cmap = chips_bilinear;
      else if (!chips_strcmp(val, "bicubic") || !chips_strcmp(val, "2"))
         *cmap = chips_bicubic;
      else 
      {
         rval = ERR_BAD_IMG_INTERP;
         *cmap = (V_INTERPOLATION_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified 
   return rval; 
}

int string_to_hoffset(const char* val, float* pos)
{
   int rval = ERR_STATUS_GOOD;

   if (val && pos)
   {
      if (!chips_strcmp(val, "left"))              *pos = 0.0;
      else if (!chips_strcmp(val, "center"))       *pos = 0.5;
      else if (!chips_strcmp(val, "right"))        *pos = 1.0;
      else if (!chips_strcmp(val, "auto"))         *pos = AUTO_ALIGN;
      else
      {
         char* ptr = NULL;
         *pos = strtod(val, &ptr);
         if (*ptr)   rval = ERR_INVALID_NUMERIC_ARG;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified
   return rval;
}


int string_to_voffset(const char* val, float* pos)
{
   int rval = ERR_STATUS_GOOD;

   if (val && pos)
   {
      if (!chips_strcmp(val, "base"))              *pos = 0.0;
      else if (!chips_strcmp(val, "center"))       *pos = 0.5;
      else if (!chips_strcmp(val, "top"))          *pos = 1.0;
      else if (!chips_strcmp(val, "auto"))         *pos = AUTO_ALIGN;
      else
      {
         char* ptr = NULL;
         *pos = strtod(val, &ptr);
         if (*ptr)   rval = ERR_INVALID_NUMERIC_ARG;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified
   return rval;
}


int string_to_v_symbol(const char* val, V_SYMBOL_TYPE* sym)
{
   int rval = ERR_STATUS_GOOD;

   if (val && sym)
   {
      if (!chips_strcmp(val, "none") || !chips_strcmp(val, "0"))
         *sym = chips_none; 
      else if (!chips_strcmp(val, "cross") || !chips_strcmp(val, "1"))
         *sym = chips_cross; 
      else if (!chips_strcmp(val, "diamond") || !chips_strcmp(val, "2"))
         *sym = chips_diamond; 
      else if (!chips_strcmp(val, "downtriangle") || !chips_strcmp(val, "3"))
         *sym = chips_downtriangle; 
      else if (!chips_strcmp(val, "circle") || !chips_strcmp(val, "4"))
         *sym = chips_circle; 
      else if (!chips_strcmp(val, "plus") || !chips_strcmp(val, "5"))
         *sym = chips_plus; 
      else if (!chips_strcmp(val, "square") || !chips_strcmp(val, "6"))
         *sym = chips_square; 
      else if (!chips_strcmp(val, "uptriangle") || !chips_strcmp(val, "7"))
         *sym = chips_uptriangle; 
      else if (!chips_strcmp(val, "point") || !chips_strcmp(val, "8")) 
         *sym = chips_point_type;
      else if (!chips_strcmp(val, "arrow") || !chips_strcmp(val, "9")) 
         *sym = chips_arrow;
      else
      {
         rval = ERR_BAD_SYMBOL;
         *sym = (V_SYMBOL_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified 

   return rval;
}


int string_to_v_units(const char* val, V_UNITS_TYPE* unt)
{
   int rval = ERR_STATUS_GOOD;

   if (val && unt)
   {
      if (!chips_strcmp(val, "cm") || !chips_strcmp(val, "centimeter") ||
          !chips_strcmp(val, "1"))
         *unt = chips_cm;
      else if (!chips_strcmp(val, "mm") || !chips_strcmp(val, "millimeter") ||
               !chips_strcmp(val, "2"))
         *unt = chips_mm;
      else if (!chips_strcmp(val, "inches") || !chips_strcmp(val, "inch") ||
               !chips_strcmp(val, "in") || !chips_strcmp(val, "3"))
         *unt = chips_inch;
      else if (!chips_strcmp(val, "pixels") || !chips_strcmp(val, "pix") ||
               !chips_strcmp(val, "4"))
         *unt = chips_pixels;
      else 
      {
         rval = ERR_BAD_UNITS;
         *unt = (V_UNITS_TYPE)-1;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_font(const char* val, V_FONT_TYPE* fnt)
{
   int rval = ERR_STATUS_GOOD;

   if (val && fnt)
   {
      if (!chips_strcmp(val, "helvetica") || !chips_strcmp(val, "0"))
         *fnt = chips_helvetica;
      else if (!chips_strcmp(val, "courier") || !chips_strcmp(val, "1"))
         *fnt = chips_courier;
      else if (!chips_strcmp(val, "times") || !chips_strcmp(val, "2"))
         *fnt = chips_times;
      else if (!chips_strcmp(val, "greek") || !chips_strcmp(val, "3"))
         *fnt = chips_greek;
      else 
      {
         rval = ERR_BAD_FONT;
         *fnt = (V_FONT_TYPE)-1;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}



int string_to_v_contour(const char* val, V_CONTOUR_TYPE* alg)
{
   int rval = ERR_STATUS_GOOD;

   if (val && alg)
   {
      if (!chips_strcmp(val, "standard") || !chips_strcmp(val, "0"))
         *alg = chips_standard;
      else if (!chips_strcmp(val, "marching") || !chips_strcmp(val, "1"))
         *alg = chips_marching;
      else 
      {
         rval = ERR_BAD_CONTOUR_ALG;
         *alg = (V_CONTOUR_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}



int string_to_v_fontstyle(const char* val, V_FONTSTYLE_TYPE* styl)
{
   int rval = ERR_STATUS_GOOD;

   if (val && styl)
   {
      if (!chips_strcmp(val, "normal") || !chips_strcmp(val, "0"))
         *styl= chips_normal;
      else if (!chips_strcmp(val, "bold") || !chips_strcmp(val, "1"))
         *styl = chips_bold;
      else if (!chips_strcmp(val, "italic") || !chips_strcmp(val, "2"))
         *styl = chips_italic;
      else if (!chips_strcmp(val, "bolditalic") || !chips_strcmp(val, "3"))
         *styl = chips_bolditalic;
      else rval = ERR_BAD_FONTSTYLE;
   }
   else
   {
      rval = ERR_RETURN_PTR_NULL;   // no value specified
      *styl = (V_FONTSTYLE_TYPE)-1;
   }
   return rval;
}


int string_to_v_errstyle(const char* val, V_ERRSTYLE_TYPE* styl)
{
  int rval = ERR_STATUS_GOOD;

   if (val && styl)
   {
      if (!chips_strcmp(val, "line") || !chips_strcmp(val, "bar") ||
          !chips_strcmp(val, "0")) 
         *styl = chips_bar;
      else if (!chips_strcmp(val, "cap") || !chips_strcmp(val, "capped") ||
               !chips_strcmp(val, "1"))
         *styl = chips_cap;
      else rval = ERR_BAD_ERRSTYLE;
   }
   else
   {
      rval = ERR_RETURN_PTR_NULL;   // no value specified
      *styl = (V_ERRSTYLE_TYPE)-1;
   }
   return rval;
}


int string_to_v_baddatamode(const char* val, V_BADDATAMODE_TYPE* styl)
{
  int rval = ERR_STATUS_GOOD;

   if (val && styl)
   {
      if (!chips_strcmp(val, "ignore") || !chips_strcmp(val, "0"))
         *styl = chips_ignore;
      else if (!chips_strcmp(val, "omit") || !chips_strcmp(val, "1"))
         *styl = chips_omit;
      else
      {
         rval = ERR_BAD_BADDATASTYLE;
         *styl = (V_BADDATAMODE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_axisstyle(const char* val, V_AXISSTYLE_TYPE* styl)
{
  int rval = ERR_STATUS_GOOD;

   if (val && styl)
   {
      if (!chips_strcmp(val, "open") || !chips_strcmp(val, "0")) 
         *styl = chips_open;
      else if (!chips_strcmp(val, "box") || !chips_strcmp(val, "boxed") ||
               !chips_strcmp(val, "1"))
         *styl = chips_box;
      else if (!chips_strcmp(val, "close") || !chips_strcmp(val, "closed") ||
               !chips_strcmp(val, "2"))
         *styl = chips_closed;
      else 
      {
         rval = ERR_BAD_PLOTSTYLE;
         *styl = (V_AXISSTYLE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_cornerstyle(const char* val, V_CORNERSTYLE_TYPE* styl)
{
  int rval = ERR_STATUS_GOOD;

   if (val && styl)
   {
      if (!chips_strcmp(val, "miter") || !chips_strcmp(val, "0"))
         *styl = chips_miter;
      else if (!chips_strcmp(val, "bevel") || !chips_strcmp(val, "1"))
         *styl = chips_bevel;
      else if (!chips_strcmp(val, "fill") || !chips_strcmp(val, "filled") ||
               !chips_strcmp(val, "2"))
         *styl = chips_fill;
      else if (!chips_strcmp(val, "notch") || !chips_strcmp(val, "notched") ||
               !chips_strcmp(val, "3"))
         *styl = chips_notch;
      else
      {
         rval = ERR_BAD_CORNERSTYLE;
         *styl = (V_CORNERSTYLE_TYPE) -1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_tickmode(const char* val, V_TICKMODE_TYPE* mode)
{
  int rval = ERR_STATUS_GOOD;

   if (val && mode)
   {
      if (!chips_strcmp(val, "count") || !chips_strcmp(val, "0"))        
         *mode = chips_count;
      else if (!chips_strcmp(val, "interval") || !chips_strcmp(val, "1"))    
         *mode = chips_interval;
      else if (!chips_strcmp(val, "nice") || !chips_strcmp(val, "2"))    
         *mode = chips_nice;
      else if (!chips_strcmp(val, "limits") || !chips_strcmp(val, "3"))
         *mode = chips_limits;
      else if (!chips_strcmp(val, "arbitrary") || !chips_strcmp(val, "4"))
         *mode = chips_arbitrary;
      else 
      {
         rval = ERR_BAD_TICKMODE;
         *mode = (V_TICKMODE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_tickstyle(const char* val, V_TICKSTYLE_TYPE* mode)
{
  int rval = ERR_STATUS_GOOD;

   if (val && mode)
   {
      if (!chips_strcmp(val, "inside") || !chips_strcmp(val, "0"))
         *mode = chips_inside;
      else if (!chips_strcmp(val, "outside") || !chips_strcmp(val, "1"))
         *mode = chips_outside;
      else if (!chips_strcmp(val, "center") || !chips_strcmp(val, "2") ||
               !chips_strcmp(val, "centered"))
         *mode = chips_centered;
      else
      {
         rval = ERR_BAD_TICKSTYLE;
         *mode = (V_TICKSTYLE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_printfile(const char* val, V_PRINTFILE_TYPE* type)
{ 
  int rval = ERR_STATUS_GOOD;

   if (val && type)
   {
      if (!chips_strcmp(val, "rgb") || !chips_strcmp(val, "1"))
         *type = chips_rgb;
      else if (!chips_strcmp(val, "cmyk") || !chips_strcmp(val, "2"))
         *type = chips_cmyk;
      else if (!chips_strcmp(val, "greyscale") || !chips_strcmp(val, "3") ||
               !chips_strcmp(val, "grayscale"))
         *type = chips_grayscale;
      else if (!chips_strcmp(val, "cmyksep") || !chips_strcmp(val, "4"))
         *type = chips_cmyksep;
      else 
      {
         rval = ERR_BAD_COLORSCHEME;
         *type = (V_PRINTFILE_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_object(const char* val, V_OBJECT_TYPE* type)
{
  int rval = ERR_STATUS_GOOD;

   if (val && type)
   {
      if (!chips_strcmp(val, "window") || !chips_strcmp(val, "2"))
         *type = chips_window;
      else if (!chips_strcmp(val, "frame") || !chips_strcmp(val, "3"))
         *type = chips_frame;
      else if (!chips_strcmp(val, "image") || !chips_strcmp(val, "5"))
         *type = chips_image;
      else if (!chips_strcmp(val, "plot") || !chips_strcmp(val, "6"))
         *type = chips_plot;
      else if (!chips_strcmp(val, "curve") || !chips_strcmp(val, "7"))
         *type = chips_curve;
      else if (!chips_strcmp(val, "histogram") || !chips_strcmp(val, "8"))
         *type = chips_histogram;
      else if (!chips_strcmp(val, "contour") || !chips_strcmp(val, "9"))
         *type = chips_contour;
      else if (!chips_strcmp(val, "axis") || !chips_strcmp(val, "10"))
         *type = chips_axis;
      else if (!chips_strcmp(val, "label") || !chips_strcmp(val, "11"))
         *type = chips_label;
      else if (!chips_strcmp(val, "line") || !chips_strcmp(val, "12"))
         *type = chips_line;
      else if (!chips_strcmp(val, "legend") || !chips_strcmp(val, "13"))
         *type = chips_legend;
      else if (!chips_strcmp(val, "point") || !chips_strcmp(val, "14"))
         *type = chips_point;
      else if (!chips_strcmp(val, "region") || !chips_strcmp(val, "15"))
         *type = chips_region;
      else if (!chips_strcmp(val, "colorbar") || !chips_strcmp(val, "16"))
         *type = chips_colorbar;
      else if (!chips_strcmp(val, "coordsys") || !chips_strcmp(val, "17"))
         *type = chips_coordinate_system;
      else 
      {
         rval = ERR_BAD_OBJECT;
         *type = (V_OBJECT_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_line_extension (const char* val, V_LINE_EXTENSION_TYPE *type)
{
   int rval = ERR_STATUS_GOOD;

   if (val && type)
   {
      if (!chips_strcmp(val, "none") || !chips_strcmp(val, "0"))
         *type = chips_extend_none;
      else if (!chips_strcmp(val, "start") || !chips_strcmp(val, "1"))
         *type = chips_extend_start;
      else if (!chips_strcmp(val, "end") || !chips_strcmp(val, "2"))
         *type = chips_extend_end;
      else if (!chips_strcmp(val, "both") || !chips_strcmp(val, "3"))
         *type = chips_extend_both;
      else 
      {
         rval = ERR_BAD_LINEEXTENSION;
         *type = (V_LINE_EXTENSION_TYPE)-1; 
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_v_coord_sys (const char* val, COORD_SYS *type)
{
   int rval = ERR_STATUS_GOOD;

   if (val && type)
   {
      if (!chips_strcmp(val, "none") || !chips_strcmp(val, "coord_none") || 
          !chips_strcmp(val, "0"))
         *type = COORD_NONE;
      else if (!chips_strcmp(val, "pixel") || !chips_strcmp(val, "1"))
         *type = PIXEL;
      else if (!chips_strcmp(val, "frame normalized") || 
               !chips_strcmp(val, "FRAME_NORM") || !chips_strcmp(val, "2"))
         *type = FRAME_NORM;
      else if (!chips_strcmp(val, "data") || !chips_strcmp(val, "3"))
         *type = DATA;
      else if (!chips_strcmp(val, "plot normalized") || 
               !chips_strcmp(val, "plot_norm") || !chips_strcmp(val, "4"))
         *type = PLOT_NORM;
      else if (!chips_strcmp(val, "window normalized") || 
               !chips_strcmp(val, "window_norm") || !chips_strcmp(val, "5"))
         *type = WINDOW_NORM;
      else if (!chips_strcmp(val, "real pixel") || 
               !chips_strcmp(val, "pixel_real") || !chips_strcmp(val, "6"))
         *type = PIXEL_REAL;
      else 
      {
         rval = ERR_STATUS_BAD;
         *type = (COORD_SYS)-1;
      }
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


int string_to_pagesize(const char* val, float &w, float &h, V_UNITS_TYPE &u)
{
   int rval = ERR_STATUS_GOOD;

   if (val)
   {
      if (!chips_strcmp (val, "letter")){
         w = 8.5;
         h = 11;
         u = chips_inch;
      }
      else if (!chips_strcmp (val, "A5")){
         w = 148;
         h = 210;
         u = chips_mm;
      }
      else if (!chips_strcmp (val, "A4")){
         w = 210;
         h = 297;
         u = chips_mm;
      }
      else if (!chips_strcmp (val, "A3")){
         w = 297;
         h = 420;
         u = chips_mm;
      }
      else if (!chips_strcmp (val, "legal")){
         w = 8.5;
         h = 14;
         u = chips_inch;
      }
      else if (!chips_strcmp (val, "executive")){
         w = 7.5;
         h = 10;
         u = chips_inch;
      }
      else rval = ERR_BAD_PAGESIZE;
   }
   else   rval = ERR_RETURN_PTR_NULL;   // no value specified

   return rval;
}


bool valid_line_type(V_LINE_TYPE val)
{
   bool rval; 

   switch(val)
   {
      case chips_solid:          // fallthrough intended
      case chips_dot:            // fallthrough intended
      case chips_dotlongdash:    // fallthrough intended
      case chips_dotshortdash:   // fallthrough intended
      case chips_longdash:       // fallthrough intended
      case chips_shortdash:      // fallthrough intended
      case chips_shortdashlongdash:  // fallthrough intended
      case chips_noline:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}

bool valid_fill_type(V_FILL_TYPE val)
{
   bool rval; 

   switch(val)
   {
      case chips_nofill:         // fallthrough intended
      case chips_solidfill:      // fallthrough intended
      case chips_updiagonal:     // fallthrough intended
      case chips_downdiagonal:   // fallthrough intended
      case chips_horizontal:     // fallthrough intended
      case chips_vertical:       // fallthrough intended
      case chips_crisscross:     // fallthrough intended
      case chips_brick:          // fallthrough intended
      case chips_grid:           // fallthrough intended
      case chips_hexagon:        // fallthrough intended
      case chips_polkadot:       // fallthrough intended
      case chips_zigzag:         // fallthrough intended
      case chips_wave:           // fallthrough intended
      case chips_flower:         // fallthrough intended
      case chips_userfill1:      // fallthrough intended
      case chips_userfill2:      // fallthrough intended
      case chips_userfill3:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}

bool valid_colormap_type(V_COLORMAP_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_red:            // fallthrough intended
      case chips_green:          // fallthrough intended
      case chips_blue:           // fallthrough intended
      case chips_grayscale:      // fallthrough intended
      case chips_rainbow:        // fallthrough intended
      case chips_hsv:            // fallthrough intended
      case chips_heat:           // fallthrough intended
      case chips_cool:           // fallthrough intended
      case chips_usercmap1:      // fallthrough intended
      case chips_usercmap2:      // fallthrough intended
      case chips_usercmap3:  
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_rgba_type(V_RGBA_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_red:            // fallthrough intended
      case chips_green:          // fallthrough intended
      case chips_blue:           // fallthrough intended
      case chips_alpha:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}

bool valid_interpolation_type(V_INTERPOLATION_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_neighbor:     // fallthrough intended
      case chips_bilinear:     // fallthrough intended
      case chips_bicubic:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_symbol_type(V_SYMBOL_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_circle:         // fallthrough intended
      case chips_cross:          // fallthrough intended
      case chips_diamond:        // fallthrough intended
      case chips_downtriangle:   // fallthrough intended
      case chips_none:           // fallthrough intended
      case chips_plus:           // fallthrough intended
      case chips_square:         // fallthrough intended
      case chips_uptriangle:     // fallthrough intended
      case chips_arrow:          // fallthrough intended
      case chips_point_type:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_units_type(V_UNITS_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_cm:             // fallthrough intended
      case chips_mm:             // fallthrough intended
      case chips_inch:           // fallthrough intended
      case chips_pixels:         // fallthrough intended
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_font_type(V_FONT_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_helvetica:           // fallthrough intended
      case chips_courier:             // fallthrough intended
      case chips_times:               // fallthrough intended
      case chips_greek:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_contour_type(V_CONTOUR_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_standard:        // fallthrough intended
      case chips_marching:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}



bool valid_fontstyle_type(V_FONTSTYLE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_normal:          // fallthrough intended
      case chips_bold:            // fallthrough intended
      case chips_italic:            // fallthrough intended
      case chips_bolditalic:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_errstyle_type(V_ERRSTYLE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_bar:          // fallthrough intended
      case chips_cap:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}

bool valid_baddatamode_type(V_BADDATAMODE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_ignore:          // fallthrough intended
      case chips_omit:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_axisstyle_type(V_AXISSTYLE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_open:          // fallthrough intended
      case chips_box:           // fallthrough intended
      case chips_closed:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_cornerstyle_type(V_CORNERSTYLE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_miter:          // fallthrough intended
      case chips_bevel:          // fallthrough intended
      case chips_fill:           // fallthrough intended
      case chips_notch:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_tickmode_type(V_TICKMODE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_count:          // fallthrough intended
      case chips_interval:       // fallthrough intended
      case chips_nice:           // fallthrough intended
      case chips_limits:         // fallthrough intended
      case chips_arbitrary: 
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_tickstyle_type(V_TICKSTYLE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_inside:          // fallthrough intended
      case chips_outside:         // fallthrough intended
      case chips_centered:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}

bool valid_printfile_type(V_PRINTFILE_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_rgb:          // fallthrough intended
      case chips_cmyk:         // fallthrough intended
      case chips_grayscale:    // fallthrough intended
      case chips_cmyksep:     
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_object_type(V_OBJECT_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_window:        // fallthrough intended:
      case chips_frame:         // fallthrough intended:
      case chips_image:         // fallthrough intended:
      case chips_plot:          // fallthrough intended:
      case chips_curve:         // fallthrough intended:
      case chips_histogram:     // fallthrough intended:
      case chips_contour:       // fallthrough intended:
      case chips_axis:          // fallthrough intended:
      case chips_label:         // fallthrough intended:
      case chips_line:          // fallthrough intended:
      case chips_legend:        // fallthrough intended:
      case chips_point:         // fallthrough intended:
      case chips_region:        // fallthrough intended:
      case chips_colorbar:      // fallthrough intended:
      case chips_coordinate_system:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_line_extension_type (V_LINE_EXTENSION_TYPE val)
{
   bool rval;

   switch(val)
   {
      case chips_extend_none:  // fallthrough intended
      case chips_extend_start: // fallthrough intended
      case chips_extend_end:   // fallthrough intended
      case chips_extend_both:  // fallthrough intended
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_coord_sys_type (COORD_SYS val)
{
   bool rval;

   switch(val)
   {
      case COORD_NONE:  // fallthrough intended
      case PIXEL:       // fallthrough intended
      case FRAME_NORM:  // fallthrough intended
      case DATA:        // fallthrough intended
      case PLOT_NORM:   // fallthrough intended
      case WINDOW_NORM: // fallthrough intended
      case PIXEL_REAL:  // fallthrough intended
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_shuffle_type(V_SHUFFLE_TYPE val)
{
   bool rval;
   switch(val)
   {
      case chips_forward:  // fallthrough intended
      case chips_backward: // fallthrough intended
      case chips_front:    // fallthrough intended
      case chips_back:
         rval = true;
      break;
      default:
         rval = false;
      break;
   }
   return rval;
}


bool valid_color(string in_color)
{
   ChipsColor clr;
   return (clr.valid_color(in_color) == ERR_STATUS_GOOD);
}


int string_to_pickmask(const char* val, unsigned int* mask)
{
   int rval = ERR_STATUS_GOOD; 
   unsigned int omask = 0;
  
   if (val && mask)
   {
      int ll = strlen(val);
      char *tmp = new char[ll+1];
      if (tmp)
      {
         strcpy(tmp, val);
         char* entry;

         entry = strtok(tmp, "| ");
         while(entry && (rval == ERR_STATUS_GOOD))
         {
            if (!chips_strcmp(entry, "btn1_down")) omask |= BTN1_DOWN;   
            else if (!chips_strcmp(entry, "btn2_down")) omask |= BTN2_DOWN;   
            else if (!chips_strcmp(entry, "btn3_down")) omask |= BTN3_DOWN;   
            else if (!chips_strcmp(entry, "btns_down")) omask |= BTNS_DOWN;   
            else if (!chips_strcmp(entry, "btn1_up")) omask |= BTN1_UP;   
            else if (!chips_strcmp(entry, "btn2_up")) omask |= BTN2_UP;   
            else if (!chips_strcmp(entry, "btn3_up")) omask |= BTN3_UP;   
            else if (!chips_strcmp(entry, "btns_up")) omask |= BTNS_UP;   
            else if (!chips_strcmp(entry, "key_press")) omask |= KEY_PRESS;
            else if (!chips_strcmp(entry, "key_release")) omask |= KEY_RELEASE;
            else rval = ERR_BAD_PICKMASK;
            entry = strtok(NULL, "| "); 
         }

         delete []tmp;
      }
   }
   // if valid mask specified- return it 
   if (omask != 0)
   {
      if (!rval) *mask = omask;  
   }
   else if (rval == ERR_STATUS_GOOD) 
      rval = ERR_NO_PICK_MASK;  // no mask specified 

   return rval; 
}

const char*  pickmask_to_string(unsigned int mask)
{
   static string maskstr;

   maskstr = "";
   if ((mask & BTNS_DOWN) == BTNS_DOWN) 
      maskstr = "BTNS_DOWN";
   else 
   {
      if ((mask & BTN1_DOWN) == BTN1_DOWN) 
         maskstr = "BTN1_DOWN";
      if ((mask & BTN2_DOWN) == BTN2_DOWN) 
      {
         if (maskstr.length())   maskstr.append("|BTN2_DOWN"); 
         else maskstr = "BTN2_DOWN";
      }
      if ((mask & BTN3_DOWN) == BTN3_DOWN) 
      {
         if (maskstr.length())   maskstr.append("|BTN3_DOWN"); 
         else maskstr = "BTN3_DOWN";
      }
   } 
   if ((mask & BTNS_UP) == BTNS_UP) 
   {
      if (maskstr.length())   maskstr.append("|BTNS_UP"); 
      else maskstr = "BTNS_UP";
   }
   else 
   {
      if ((mask & BTN1_UP) == BTN1_UP)
      {
         if (maskstr.length())   maskstr.append("|BTN1_UP");
         else maskstr = "BTN1_UP";
      }
      if ((mask & BTN2_UP) == BTN2_UP)
      {
         if (maskstr.length())   maskstr.append("|BTN2_UP");
         else maskstr = "BTN2_UP";
      }
      if ((mask & BTN3_UP) == BTN3_UP)
      {
         if (maskstr.length())   maskstr.append("|BTN3_UP");
         else maskstr = "BTN3_UP";
      }
   }
   if ((mask & KEY_PRESS) == KEY_PRESS)
   {
      if (maskstr.length())   maskstr.append("|KEY_PRESS");
      else maskstr = "KEY_PRESS";
   }
   if ((mask & KEY_RELEASE) == KEY_RELEASE)
   {
      if (maskstr.length())   maskstr.append("|KEY_RELEASE");
      else maskstr = "KEY_RELEASE";
   }
   return maskstr.c_str(); 
}


int chips_strcmp(const char *aa,  /* i: input string */
                 const char *bb)  /* i: input string */
{
   int  ii,     /* loop variable */
        xx_len, /* lengths of strings */
        zz = 1; /* comparison status */
   if (aa && bb)
   {
      if ((ii = xx_len = strlen(aa)) == strlen(bb))
      {
         zz = 0;
         while (ii-- && !zz)
            if (tolower(aa[ii]) != tolower(bb[ii]))   zz = 1;
      }
   }
   return zz;
}


bool approx_equal(double d1, double d2, double relerr, double maxerr)
{
   double err;

   if (d1 == d2)
      err = 0;
   else if (fabs (d1 - d2) < maxerr)  // check for close to zero
      err = 0;
   else if (fabs(d1) > fabs(d2))
      err = fabs((d1 - d2) / d1);
   else
      err = fabs((d1 - d2) / d2);
   return (err <= relerr);
}

int string_to_ratio_values(const char* val, float &w, float &h)
{
   string::size_type pos;
   string vval = val;
   if ((pos = vval.find(":")) == string::npos)
      return ERR_STATUS_BAD;     // no semi-colon 
   if (pos == 0)  
      return ERR_STATUS_BAD;     // semi-colon before width 
   
   float tmp1, tmp2;
   istringstream v1(vval.substr(0,pos));
   istringstream v2(vval.substr(pos+1));
   v1 >> tmp1;
   v2 >> tmp2;
   if (tmp1 <= 0 || v1.fail())
      return ERR_STATUS_BAD;     // non-positive width
   else if (tmp2 <= 0 || v2.fail())
      return ERR_STATUS_BAD;     // non-positive height

   w = tmp1; 
   h = tmp2;
   return ERR_STATUS_GOOD; 
}

