// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsDepthArrangeCommand.hh"
#include "ChipsStateStream.hh"
#include "ChipsFrame.hh"
#include "ChipsPlot.hh"
#include "ChipsId.hh"

ChipsDepthArrangeCommand::ChipsDepthArrangeCommand(ChipsObject *cob, 
                                int inPos, int inTotalPos, int inDirection, 
                                V_OBJECT_TYPE inObj, ChipsId *inCid)
{
   co = cob;
   weight = 1;
   pos = inPos;
   total_pos = inTotalPos;
   direction = inDirection;
   obj = inObj;
   if (inCid)
      cid = new ChipsId (*inCid);
   else
      cid = new ChipsId;
}

ChipsDepthArrangeCommand::~ChipsDepthArrangeCommand()
{
   delete cid;
}


ChipsDepthArrangeCommand::ChipsDepthArrangeCommand(ChipsObject *cob, ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

int ChipsDepthArrangeCommand::execute ()
{
   int rval = ERR_STATUS_BAD;
   if (direction == -1)  return ERR_STATUS_GOOD;  // shuffle no-op
   if (co)
   {
      int obj_type = co->get_object_type ();
      if (obj_type == chips_frame)
      {
         ChipsFrame *cf = dynamic_cast <ChipsFrame *>(co);
         rval = cf->rearrange_depth (cid, obj, direction, false);
      }
      else if (obj_type == chips_plot)
      {
         ChipsPlot *cp = dynamic_cast <ChipsPlot *>(co);
         rval = cp->rearrange_object (cid, obj, direction, false);
      }

   }
   return rval;
}

int ChipsDepthArrangeCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co)
   {
      ChipsFrame *cf = NULL;
      ChipsPlot *cp = NULL;
      int obj_type = co->get_object_type ();

      if (obj_type == chips_frame)
         cf = dynamic_cast <ChipsFrame *>(co);
      else if (obj_type == chips_plot)
         cp = dynamic_cast <ChipsPlot *>(co);

      switch (direction)
      {
         case chips_forward:
         {
            if (cf)
               rval = cf->rearrange_depth (cid, obj,chips_backward,false);
            else if (cp)
               rval = cp->rearrange_object (cid, obj,chips_backward,false);
         }
         break;
         case chips_backward:
            if (cf)
               rval = cf->rearrange_depth (cid, obj,chips_forward,false);
            else if (cp)
               rval = cp->rearrange_object (cid, obj,chips_forward,false);
         break;
         case chips_front:
         {
            rval = ERR_STATUS_GOOD;
            // Step forwad or backward to get the object back to
            // it's previous position in the list. To help optimize
            // this see if the original position was more or less
            // then half way in the. Move to the correct position
            // accordingly.
            if (pos > (total_pos/2))
            {
               int tmp_pos = total_pos - pos;
               if (cf)
               {
                  for (int ii=0; !rval && ii<tmp_pos; ii++)
                     rval = cf->rearrange_depth (cid, obj, chips_backward, false);
               }
               else if (cp)
               {
                  for (int ii=0; !rval && ii<tmp_pos; ii++)
                     rval = cp->rearrange_object (cid, obj, chips_backward, false);
               }
               else rval = ERR_STATUS_BAD;

            }
            else
            {
               if (cf)
               {
                  rval = cf->rearrange_depth (cid, obj, chips_back, false);
                  for (int ii=1; !rval && ii<pos; ii++)
                     rval = cf->rearrange_depth (cid, obj, chips_forward, false);
               }
               else if (cp)
               {
                  rval = cp->rearrange_object (cid, obj, chips_back, false);
                  for (int ii=1; !rval && ii<pos; ii++)
                     rval = cp->rearrange_object (cid, obj, chips_forward, false);
               }
               else rval = ERR_STATUS_BAD;
            }
         }
         break;
         case chips_back:
         {
            rval = ERR_STATUS_GOOD;
            // Step forwad or backward to get the object back to
            // it's previous position in the list. To help optimize
            // this see if the original position was more or less
            // then half way in the. Move to the correct position
            // accordingly.
            if (pos > (total_pos/2))
            {
               if (cf)
               {
                  int tmp_pos = total_pos - pos;
                  rval = cf->rearrange_depth (cid, obj, chips_front, false);
                  for (int ii=0; !rval && ii<tmp_pos; ii++)
                     rval = cf->rearrange_depth (cid, obj, chips_backward, false);
               }
               else if (cp)
               {
                  int tmp_pos = total_pos - pos;
                  rval = cp->rearrange_object (cid, obj, chips_front, false);
                  for (int ii=0; !rval && ii<tmp_pos; ii++)
                     cp->rearrange_object (cid, obj, chips_backward, false);
               }
               else rval = ERR_STATUS_BAD;
            }
            else
            {
               if (cf)
               {
                  for (int ii=1; !rval && ii<pos; ii++)
                     rval = cf->rearrange_depth (cid, obj, chips_forward, false);
               }
               else if (cp)
               {
                  for (int ii=1; !rval && ii<pos; ii++)
                     rval = cp->rearrange_object (cid, obj, chips_forward, false);
               }
               else rval = ERR_STATUS_BAD;
            }
         }
         break;
         case -1:
            // no op - ie. a shuffle front when already at the front
            rval = ERR_STATUS_GOOD; 
         break; 
         default:
            // unknown direction- do nothing- rval is already set 
         break; 
      }
   }
   return rval;
}

void ChipsDepthArrangeCommand::serialize(ChipsStateStream& out)
{
   int val = 16;
   // Standard serialize output for a command
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   // Specific output for depth arrange command
   out.write_int (&pos);
   out.write_int (&total_pos);
   out.write_int (&direction);

   val = static_cast <int>(obj);
   out.write_int (&val);
   cid->serialize (out);
}

void ChipsDepthArrangeCommand::deserialize(ChipsStateStream& in)
{
   int val;
   // Standard serialize input for a command
   in.read_int(&ref);
   in.read_int(&block_ref);

   // Specific input for depth vis command
   in.read_int (&pos);
   in.read_int (&total_pos);
   in.read_int (&direction);
   in.read_int (&val);
   obj = static_cast <V_OBJECT_TYPE> (val);

   cid = new ChipsId (in);
}

void ChipsDepthArrangeCommand::print ()
{
   cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Depth Arrange" << endl;
};



