// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsDepthCollapseCommand.hh"
#include "ChipsStateStream.hh"
#include "ChipsFrame.hh"

ChipsDepthCollapseCommand::ChipsDepthCollapseCommand(ChipsObject *obj, int inDepth, int *inSrcDepth, int inSrcCnt)
{
   co = obj;
   weight = 1;
   depth = inDepth;
   src_cnt = inSrcCnt;
   src_depths = NULL;

   if (src_cnt > 0)
   {
      src_depths = new int [src_cnt];
      for (int ii=0; ii<src_cnt; ii++)
         src_depths[ii] = inSrcDepth[ii];
   }
}

ChipsDepthCollapseCommand::~ChipsDepthCollapseCommand()
{
   if (src_depths)
      delete []src_depths;
}


ChipsDepthCollapseCommand::ChipsDepthCollapseCommand(ChipsObject *cob, ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

int ChipsDepthCollapseCommand::execute ()
{
   int rval = ERR_STATUS_BAD;
   if (co && (co->get_object_type () == chips_frame) && src_depths)
   {
      ChipsFrame *cf = dynamic_cast <ChipsFrame *>(co);
      
      rval = cf->collapse_depths (src_depths, src_cnt, depth, false);

   }
   return rval;
}

int ChipsDepthCollapseCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   if (co && (co->get_object_type () == chips_frame))
   {
      ChipsFrame *cf = dynamic_cast <ChipsFrame *>(co);
      rval = cf->uncollapse_depths (depth_table, depth);

   }
   return rval;

}

void ChipsDepthCollapseCommand::serialize(ChipsStateStream& out)
{
   int val = 17;
   multimap<int, int>::iterator ii;
   list<int>::iterator jj;

   // Standard serialize output for a command
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   // Specific output for depth arrange command
   out.write_int (&depth);

   val = depth_table.size ();
   out.write_int (&val);
   for (ii=depth_table.begin (); ii!=depth_table.end (); ii++)
   {
      val = ii->first;
      out.write_int (&val);
      val = ii->second;
      out.write_int (&val);
   }

   // The source list has to be serialized and deserialized
   // seperate from the depth table because its order matters
   out.write_int (&src_cnt);
   out.write_int (src_depths, src_cnt);

}

void ChipsDepthCollapseCommand::deserialize(ChipsStateStream& in)
{
   int val1,val2;
   int sz;
   // Standard serialize input for a command
   in.read_int(&ref);
   in.read_int(&block_ref);

   // Specific input for depth vis command
   in.read_int (&depth);

   in.read_int (&sz);

   for (int ii=0; ii<sz; ii++)
   {
      
      in.read_int (&val1);
      in.read_int (&val2);
      insert_to_table (val1, val2);
   }

   // The source list has to be serialized and deserialized 
   // seperate from the depth table because its order matters
   in.read_int (&src_cnt);
   src_depths = new int [src_cnt];
   in.read_int (src_depths, src_cnt);

}

void ChipsDepthCollapseCommand::print ()
{
   cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Depth Collapse" << endl;
};

void ChipsDepthCollapseCommand::insert_to_table (int obj_depth, int ref_cnt)
{
   depth_table.insert (pair <const int, int> (obj_depth, ref_cnt));
}



