// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsDepthControl.hh"
#include "ChipsStateStream.hh"
#include "ChipsPlot.hh"
#include "ChipsObject.hh"
#include "ChipsFrame.hh"
#include "ChipsErrs.hh"
#include "ChipsDepthCollapseCommand.hh"

#include <limits.h> 
#include <stdio.h>
#include <cmath>
#include <sstream>

// sort objects based on their reference count
bool object_sort (list<ChipsObject *>::const_reference object1, list<ChipsObject *>::const_reference object2);

bool object_sort (list<ChipsObject *>::const_reference object1, list<ChipsObject *>::const_reference object2)
{
   if (!object1)
      return true;
   if (!object2)
      return false;
   if (object1->get_ref_cnt () < object2->get_ref_cnt ())
      return true;
   return false;
}


ChipsDepthControl::ChipsDepthControl(int val)
{
   add_cnt = 0; 
   start_depth = INT_MAX; 
   shuffle_cnt = 0;
   if (val > 0)  default_depth = val;
   else          default_depth = V_DEFAULT_DEPTH;
}


ChipsDepthControl::ChipsDepthControl(ChipsFrame* frm, ChipsStateStream& in)
{
   depth_vis_map::iterator ii;
   int ival, sz = vis_table.size();
   bool bval; 
   shuffle_cnt = 0;

   // load depth visibility table
   in.read_int(&sz);
   in.read_int(&ival);
   if (ival == sz)
      for (int rr = 0; rr < sz; rr++) 
      {
         in.read_int(&ival);
         in.read_bool(&bval);
         if (!vis_table.count(ival))   vis_table[ival] = bval; 
      } 
}


ChipsDepthControl::~ChipsDepthControl()
{

}


int ChipsDepthControl::add_object(ChipsObject* obj)
{
   if (obj)
   {
      int depth = obj->get_depth();
      // maintain lowest depth that needs to be sorted
      if (depth < start_depth)  start_depth = depth; 
      obj->set_vis_override(vis_table[depth]);
      obj_table.insert(pair<const int, ChipsObject*>(depth, obj));
      obj->change_depth();
      obj->change_visible();
      add_cnt++; 
      sort_single_item();
   }
   return 0; 
}

int ChipsDepthControl::remove_object(ChipsObject* obj)
{
   if (obj)
   {
      depth_obj_map::iterator mm;
      int cnt, dep = obj->get_depth();
      if ((cnt = obj_table.count(dep)) > 0)
      {
         mm = obj_table.find(dep);
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
            if (((*mm).second->get_id() == obj->get_id()) &&
                ((*mm).second->get_ref_cnt() == obj->get_ref_cnt()))
            {
               obj_table.erase(mm);
               break;
            }
      }
   }
   return 0; 
}


// adjust the position of the specified object within the current depth.
// objects can be moved to the front or back of the queue as well as 
// shuffled forward or backward by one object.
int ChipsDepthControl::rearrange_object(ChipsObject* obj, int manner)
{
   if (!obj)  return ERR_STATUS_BAD; 
   depth_obj_map::iterator mm, m1;
   int cnt, dep = obj->get_depth();
   if ((cnt = obj_table.count(dep)) < 1)  return ERR_STATUS_BAD;
   m1 = mm = obj_table.find(dep);
   add_cnt = 0;
   start_depth = dep;
   switch(manner) 
   {
      case chips_front:
      {
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
            if (((*mm).second->get_id() == obj->get_id()) &&
                ((*mm).second->get_ref_cnt() == obj->get_ref_cnt()))
	    {
               obj_table.erase(mm);
	       ndx = cnt; 
               break;
	    }
         add_object(obj);
      }
      break; 
      case chips_back:
      {
         list <ChipsObject*> obs;
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
         {
            if (((*mm).second->get_id() != obj->get_id()) ||
                ((*mm).second->get_ref_cnt() != obj->get_ref_cnt()))
               obs.push_back((*mm).second);
         }
         obj_table.erase(m1, mm);
         add_object(obj);
         for(list<ChipsObject*>::iterator ll= obs.begin(); ll!=obs.end(); ++ll)
            add_object((*ll));
         obs.clear(); 
      }
      break;
      case chips_forward:
      {
         list <ChipsObject*> obs;
         int found = 0;
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
         {
            if (((*mm).second->get_id() == obj->get_id()) &&
                ((*mm).second->get_ref_cnt() == obj->get_ref_cnt()))
            {
               found = 1;
               m1 = mm;
            }
            else if (found)
            {
               obs.push_back((*mm).second);
            }
         }
         if (!obs.empty())  // don't do anything if already in front 
         {
            obj_table.erase(m1, mm);
            list<ChipsObject*>::iterator ll= obs.begin();
            add_object((*ll));
            add_object(obj);
            for(++ll; ll!=obs.end(); ++ll)
               add_object((*ll));
            obs.clear();
         }
         else return ERR_STATUS_BAD; 
      }
      break;
      case chips_backward:
      {
         list <ChipsObject*> obs;
         int found = -1;
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
         {
            if (((*mm).second->get_id() == obj->get_id()) &&
                ((*mm).second->get_ref_cnt() == obj->get_ref_cnt()))
               found = ndx;
            else 
               obs.push_back((*mm).second);
         }
         if (found >0)
         {
            // object was in list and not backmost position
            obj_table.erase(m1, mm);
	    int pos = 1;
            for(list<ChipsObject*>::iterator ll=obs.begin();ll!=obs.end();++ll,++pos)
            {
               if (pos == found)  add_object(obj); 
               add_object((*ll));
            }
            obs.clear();
         } 
         else 
         {
            obs.clear();
            return ERR_STATUS_BAD;
         }
      }
      break;
      default:
         return ERR_FEATURE_DNE; 
      break; 
   }
   shuffle_cnt++;
   return 0;
}


int ChipsDepthControl::display_depth(int depth)
{
   depth_obj_map::iterator ii;
   int cnt;

   vis_table[depth] = false;
   if ((cnt = obj_table.count(depth)) != 0)
   {
      ii = obj_table.find(depth);
      // if we display a depth that contains axes, the plot borders may need
      // to be hidden: store impacted plots and check when finished
      map<string, ChipsPlot*>checkplots;
      ChipsPlot* cpt;
      for (int ndx = 0; ndx < cnt; ++ndx, ++ii)
      {
         (*ii).second->set_vis_override(false);
         (*ii).second->change_visible();
         if ((*ii).second->get_object_type() == chips_axis)
         {
            cpt = (ChipsPlot*) (*ii).second->get_parent();
            checkplots.insert(pair<string, ChipsPlot*>(cpt->get_id(),cpt));
         }
      }
      map<string, ChipsPlot*>::iterator ip;
      for(ip = checkplots.begin(); ip != checkplots.end(); ++ip)
         (*ip).second->adjust_border_visibility();
   }
   return 0;
}
 
int ChipsDepthControl::hide_depth(int depth)
{
   depth_obj_map::iterator ii;
   int cnt;

   vis_table[depth] = true;
   if ((cnt = obj_table.count(depth)) != 0)
   {
      ii = obj_table.find(depth);

      // if we hide a depth that contains axes, the plot borders may need
      // to be redisplayed: store impacted plots and check when finished 
      map<string, ChipsPlot*>checkplots;
      ChipsPlot* cpt;
      for (int ndx = 0; ndx < cnt; ++ndx, ++ii)
      {
         (*ii).second->set_vis_override(true);
         (*ii).second->change_visible();
         if ((*ii).second->get_object_type() == chips_axis)
	 {
	    cpt = (ChipsPlot*) (*ii).second->get_parent();
	    checkplots.insert(pair<string, ChipsPlot*>(cpt->get_id(),cpt));
	 }
      }
      map<string, ChipsPlot*>::iterator ip;
      for(ip = checkplots.begin(); ip != checkplots.end(); ++ip)
         (*ip).second->adjust_border_visibility();
   }
   return 0;
}
 

char* ChipsDepthControl::list_depths(int depth)
{
   string temp;
   char* ptr = NULL;
   if (!vis_table.empty())
   {
      depth_vis_map::iterator dh;
      depth_obj_map::iterator mm;
      char level[80];
      int dep;
     
      temp.clear();
      if (depth == 0)
      {
         for (dh = vis_table.begin(); dh != vis_table.end(); dh++)
         {
            dep = (*dh).first;
            if ((*dh).second)
               sprintf(level, "    Depth %2d  (hidden)\n", dep);
            else
               sprintf(level, "    Depth %2d\n", dep);
            temp.append(level);
            temp.append(list_at_depth(dep)); 
         }
      }
      else if (vis_table.count(depth) > 0)
      {
         if (vis_table[depth])
            sprintf(level, "    Depth %2d  (hidden)\n", depth);
         else
            sprintf(level, "    Depth %2d\n", depth);
         temp.append(level);
         temp.append(list_at_depth(depth));
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = new char[ll];
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


int ChipsDepthControl::reset_addition_count(int* start)
{
   // returns the number of additions since last reset and passes back the
   // lowest modified depth level. Both counters are reset 
   int val = add_cnt;
   if (start)  *start = start_depth; 
   start_depth = INT_MAX; 
   add_cnt = 0;
   return val; 
}

int ChipsDepthControl::sort_check()
{
   if (!add_cnt)  
      return ERR_STATUS_GOOD;
   else if (add_cnt == 1)
      sort_single_item();
   else 
      sort_depth(start_depth);
   // reset counters 
   start_depth = INT_MAX;
   add_cnt = 0;
   return ERR_STATUS_GOOD;
}


int ChipsDepthControl::collapse_depths(int* sources, int src_cnt, int dest, 
                                  ChipsDepthCollapseCommand* cmd)
{
   if (!sources || (src_cnt < 1))  return ERR_STATUS_BAD; 
   depth_obj_map::iterator ii, i1;
   int cnt = 0; 
   add_cnt = 0;
   start_depth = dest;

   // collapsing a source into itself could pose a problem. If the dest is
   // in the source list, store objects to a list before moving the to the
   // dest.
   for (int rr = 0; rr < src_cnt; rr++) 
      if (sources[rr] == dest) cnt = 1; 
      else if (sources[rr] < start_depth)  start_depth = sources[rr];
 
   if (cnt == 0)
      for (int rr = 0; rr < src_cnt; rr++) 
      {
         if ((cnt = obj_table.count(sources[rr])) != 0)
         {
            i1 = ii = obj_table.find(sources[rr]);
            ChipsObject* obj; 
            for (int ndx = 0; ndx < cnt; ++ndx, ++ii)
            {
               obj = (*ii).second;
               // Update the command with the objects original depths
               if (cmd != NULL)
                  cmd->insert_to_table (obj->get_depth(), obj->get_ref_cnt());
               obj->set_depth(dest);  
               add_object(obj);
            }
            obj_table.erase(i1, ii);
         }
      }   
   else 
   {
      // take all objects out of the source depths
      list <ChipsObject*> obs;
      for (int rr = 0; rr < src_cnt; rr++) 
      {
         if ((cnt = obj_table.count(sources[rr])) != 0)
         {
            i1 = ii = obj_table.find(sources[rr]);
            ChipsObject* obj; 
            for (int ndx = 0; ndx < cnt; ++ndx, ++ii)
            {
               obj = (*ii).second;

               // Update the command with the objects original depths
               if (cmd != NULL)
                  cmd->insert_to_table (obj->get_depth (), obj->get_ref_cnt ());

               obs.push_back(obj);
            }
            obj_table.erase(i1, ii);
         }
      }   
      // add all items to the destination depth 
      for(list<ChipsObject*>::iterator ll= obs.begin(); ll != obs.end(); ++ll)
      {
         (*ll)->set_depth(dest);
         add_object((*ll));
      }
      obs.clear();
   }

   return ERR_STATUS_GOOD; 
}

int ChipsDepthControl::uncollapse_depths(multimap<int, int> &dest_table, int src_depth)
{
   int rval = ERR_STATUS_GOOD;
   int cnt;
   depth_obj_map::iterator obj_table_iter, start_table_iter;
   multimap<int, int>::iterator dest_table_iter;
  
   if ((cnt = obj_table.count(src_depth)) != 0)
   {
      start_table_iter = obj_table.find(src_depth);
      ChipsObject* obj;

      dest_table_iter = dest_table.begin ();
      for (; dest_table_iter != dest_table.end (); ++dest_table_iter)
      {
         obj_table_iter = start_table_iter;
         for (int ii=0; ii<cnt; ++ii, ++obj_table_iter)
         {
            
            obj = obj_table_iter->second;
            if (obj->get_ref_cnt () == dest_table_iter->second)
            {
               obj->set_depth(dest_table_iter->first);
               add_object(obj);

               if ((obj_table_iter == start_table_iter) &&
                   (start_table_iter != obj_table.end ()))
                  start_table_iter++;

               obj_table.erase(obj_table_iter);

               cnt--;

               // this gets out the loop
               ii = cnt;
            }
         }
      }
   }

   return rval;
}

int ChipsDepthControl::set_default_depth(int val)
{
   if (val < 1)   return  ERR_STATUS_BAD; 
   
   default_depth = val; 
   return ERR_STATUS_GOOD;
}

bool ChipsDepthControl::get_depth_visible (int depth)
{
   if (depth == 0)
      depth = default_depth;

   return vis_table[depth];
}

int ChipsDepthControl::position_in_depth (int obj_ref_cnt, int obj_depth)
{
   int cnt;
   int ndx=0;
   bool done = false;
   if ((cnt = obj_table.count(obj_depth)) > 0)
   {
      depth_obj_map::iterator mm  = obj_table.find(obj_depth);
      for (ndx = 0; !done && (ndx < cnt); ++ndx, ++mm)
      {
         if (mm->second->get_ref_cnt () == obj_ref_cnt)
         {
            done = true;
//            ndx++;
         }
      }
   }
   return ndx;

}

int ChipsDepthControl::total_objects_in_depth (int depth)
{
   return obj_table.count(depth);

}


int ChipsDepthControl::sort_depth (int start_lvl)
{
   return ERR_FEATURE_DNE; 
}

int ChipsDepthControl::sort_single_item()
{
   return ERR_FEATURE_DNE; 
}

int ChipsDepthControl::sort_last_item()
{
   return ERR_FEATURE_DNE; 
}

string ChipsDepthControl::list_at_depth(int dep)
{
   static string listings;
   listings.clear();
   int cnt; 
   if ((cnt = obj_table.count(dep)) > 0)
   {
      depth_obj_map::iterator mm  = obj_table.find(dep);
      V_OBJECT_TYPE obtyp;
      for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
      {
         listings.append("      ");
         obtyp = (*mm).second->get_object_type();
         listings.append(v_object_to_string(obtyp));
         listings.append(" [");
         listings.append((*mm).second->get_id());
         if (obtyp == chips_curve || obtyp == chips_axis || obtyp == chips_image ||
             obtyp == chips_histogram || obtyp == chips_contour ||
             ((obtyp == chips_label) && 
             ((*mm).second->get_parent()->get_object_type() == chips_plot)))
         {
            listings.append(" (");
            listings.append((*mm).second->get_parent()->get_id());
            listings.append(")");
         }
         listings.append("]\n");
      }
   }
   else listings = "";
   return listings; 
}

void ChipsDepthControl::serialize(ChipsStateStream& out)
{
   depth_vis_map::iterator ii;
   ChipsObject *obj, *obj2;
   int cnt, dep, ival, sz = vis_table.size();
   string sval;
   bool bval; 

   out.write_int(&sz);
   out.write_int(&sz);
   for (ii = vis_table.begin(); ii != vis_table.end(); ii++)
   {
      ival = (*ii).first;
      bval = (*ii).second;
      out.write_int(&ival);
      out.write_bool(&bval);
   } 

   out.write_int(&sz);
   depth_obj_map::iterator mm;
   for (ii = vis_table.begin(); ii != vis_table.end(); ii++)
   {
      dep = (*ii).first;
      cnt = obj_table.count(dep);
      out.write_int(&cnt);
      if (cnt > 0)
      {
         mm = obj_table.find(dep);
         for (int ndx = 0; ndx < cnt; ++ndx, ++mm)
	 {
	    // store off  [obj type][(plot id) obj id][depth]
	    obj = (*mm).second;
	    ival = (int) obj->get_object_type();
	    out.write_int(&ival);
	    // store parent id
	    if ((obj2 = obj->get_parent()) != NULL)
	       sval = obj2->get_id();
	    else
	       sval = "";
	    out.write_string(&sval);
            sval = obj->get_id();
	    out.write_string(&sval);
            out.write_int(&dep);
         }
      }
   }
   out.write_int (&default_depth);
}

int ChipsDepthControl::deserialize_obj_table(ChipsFrame* frm, 
   ChipsStateStream& in)
{
   int cnt, cnt2, ival, dep;
   ChipsId cid;
   ChipsObject *obj;
   string s1, s2;
   V_OBJECT_TYPE otype; 
   in.read_int(&cnt);
   // iterate over each depth
   for (int rr =0; rr < cnt; rr++)
   {
      in.read_int(&cnt2);
      // iterate over each entry at this depth 
      for (int cc =0; cc < cnt2; cc++)
      {
         in.read_int(&ival);
	 otype = (V_OBJECT_TYPE) ival; 
	 s2.clear();

         // maintain backwards compatiblity
         if (((in.get_major_version() == 3) && (in.get_minor_version () < 4)) ||
             (in.get_major_version() < 3))
         {
	    switch(ival)
            {
               case chips_axis:             // fallthrough intended 
               case chips_image:            // fallthrough intended 
               case chips_curve:            // fallthrough intended 
               case chips_contour:          // fallthrough intended 
               case chips_histogram:       
                  in.read_string(&s2);
                  cid.reset_plot_id(s2);
                  in.read_string(&s1);
               break;
               case chips_label:
                  // read the parent id, but don't set the ChipsId value since
                  // we only need the value if the label isn't in the frame's list
                  in.read_string(&s2);
                  in.read_string(&s1);
               break;
               default:
                  in.read_string(&s1);
               break;
            }
         }
         else
         {
            // get plot id
            in.read_string(&s2);
            cid.reset_plot_id(s2);
            in.read_string(&s1);
         }
         in.read_int(&dep);

         cid.set_object_id(otype, s1);
         if ((obj = frm->get_object(&cid, otype)) != NULL)  
	    add_object(obj);
         else if (otype == chips_label)
	 {
	    // plot title- s1 should be <plot id>_title
            if (s2.length())
            {
	       cid.reset_plot_id(s2);
	       ChipsPlot* pt = (ChipsPlot*) frm->get_object(&cid, chips_plot);
               if (pt && ((obj = pt->get_title()) != NULL))
                  if (obj->get_id() == s1)   add_object(obj); 
	    }
	 }
      }
   }
   in.read_int (&default_depth);
   return ERR_STATUS_GOOD;
}

void ChipsDepthControl::serialize_ascii(ChipsStateStream& out, string plid)
{

   depth_obj_map::iterator depth_obj_start;
   depth_obj_map::iterator depth_obj_end;
   depth_obj_map::iterator depth_obj_iter;
   list <ChipsObject *> object_list;
   list <ChipsObject *>::iterator object_list_iter;

   list <string> *command_list = NULL;
   list <string>::iterator command_list_iter;
    
   int obj_cnt = 0;
   string current_plot_id = "";
   string parent_id = "";

   ChipsObject *obj;
   string direction = "";
 
   int end_pos = 0;
   int begin_pos = 0;

   // go through each depth in the table an organize the depth
   depth_obj_end = obj_table.begin ();
   while (depth_obj_end != obj_table.end ())
   { 
      int index = 0;

      depth_obj_iter = depth_obj_start = depth_obj_end;
      obj_cnt = obj_table.count (depth_obj_iter->first);

      // create an array of commands for each object
      if (obj_cnt > 0)
         command_list = new list<string>[obj_cnt];
    
      // create a list of all objects at this depth  
      object_list.clear ();
      for (int ii=0; ii<obj_cnt; ii++, depth_obj_iter++)
         object_list.push_back (depth_obj_iter->second);

      depth_obj_end = depth_obj_iter;
      depth_obj_iter = depth_obj_start;

      // sort the objects based on when they were added to the system
      object_list.sort (object_sort);


      // set the front and back of the list first
      depth_obj_iter = depth_obj_start;
      object_list_iter = object_list.begin ();
      if (depth_obj_iter->second != (*object_list_iter))
      {
         obj = depth_obj_iter->second;

         // *** TEMPORARY CODE *** This prevents shuffle from being applied to
         // plot titles.
         if (((ChipsPlot *)(obj->get_parent ()))->get_title() != obj)
         {
            if (obj->get_parent () != NULL)
               parent_id = obj->get_parent ()->get_id ();

            if (current_plot_id != parent_id)
            {
               command_list[index].push_back ("\ncurrent_plot(\""+parent_id+"\")");
               current_plot_id = parent_id;
            }
            direction = "";
            direction.append("shuffle(\"");
            direction.append(obj->get_id());
            direction.append("\",chips_");
            direction.append(v_object_to_string(obj->get_object_type()));
            direction.append(", chips_back)");
            command_list[index].push_back (direction);

            for (object_list_iter = object_list.begin (); object_list_iter != object_list.end () && *object_list_iter != obj; object_list_iter++);
            if (object_list_iter != object_list.end ())
               object_list.erase (object_list_iter);
            object_list.push_front (obj);

            index++;
         }
      }

      depth_obj_iter = depth_obj_end;
      object_list_iter = object_list.end ();
      object_list_iter--;
      depth_obj_iter--;
      if (depth_obj_iter->second != (*object_list_iter))
      {
         obj = depth_obj_iter->second;

         // *** TEMPORARY CODE *** This prevents shuffle from being applied to
         // plot titles.
         if (((ChipsPlot *)(obj->get_parent ()))->get_title() != obj)
         {

            if (obj->get_parent () != NULL)
               parent_id = obj->get_parent ()->get_id ();

            if (current_plot_id != parent_id)
            {
               command_list[index].push_back ("\ncurrent_plot(\""+parent_id+"\")");
               current_plot_id = parent_id;
            }

            direction = "";
            direction.append("shuffle(\"");
            direction.append(obj->get_id());
            direction.append("\",chips_");
            direction.append(v_object_to_string(obj->get_object_type()));
            direction.append(", chips_front)");
            command_list[index].push_back (direction);

            for (object_list_iter = object_list.begin (); object_list_iter != object_list.end () && *object_list_iter != obj; object_list_iter++);
            if (object_list_iter != object_list.end ())
               object_list.erase (object_list_iter);
            object_list.push_back (obj);

            index++;
         }
      }
     

      // go through the list and see where the objects belong
      end_pos = 0;
      for (depth_obj_iter = depth_obj_start; end_pos<obj_cnt; end_pos++, depth_obj_iter++)
      {
         obj = depth_obj_iter->second;

         // *** TEMPORARY CODE *** This prevents shuffle from being applied to
         // plot titles. 
         if (((ChipsPlot *)(obj->get_parent ()))->get_title() == obj)
            continue;

         // find the object in the sorted list
         begin_pos =0;
         for (object_list_iter = object_list.begin (); begin_pos < obj_cnt; begin_pos++, object_list_iter++)
         {
            if (*object_list_iter == obj)
               break;
         }

         parent_id  = obj->get_parent ()->get_id ();
         
         // if the end position is the same as current position continue
         if (end_pos == begin_pos)
            continue;
         else if (end_pos > begin_pos)
         {
            if (current_plot_id != parent_id)
            {
               command_list[index].push_back ("\ncurrent_plot(\""+parent_id+"\")");
               current_plot_id = parent_id;
            } 
            // if the end position falls after the start position go forward

            // which is faster going to the front and back tracking or
            // just going forward
            if ((end_pos - begin_pos) < (obj_cnt - end_pos))
            {
               for (int ii=begin_pos; ii<end_pos; ii++)
               {
                  direction = "";
                  direction.append("shuffle(\"");
                  direction.append(obj->get_id());
                  direction.append("\",chips_");
                  direction.append(v_object_to_string(obj->get_object_type()));
                  direction.append(", chips_forward)");
                  command_list[index].push_back (direction);
               }
            }
            else
            {
               
               direction = "";
               direction.append("shuffle(\"");
               direction.append(obj->get_id());
               direction.append("\",chips_");
               direction.append(v_object_to_string(obj->get_object_type()));
               direction.append(", chips_front)");
               command_list[index].push_back (direction);
               for (int ii=obj_cnt-1; ii>end_pos; ii--)
               {
                  direction = "";
                  direction.append("shuffle(\"");
                  direction.append(obj->get_id());
                  direction.append("\",chips_");
                  direction.append(v_object_to_string(obj->get_object_type()));
                  direction.append(", chips_backward)");
                  command_list[index].push_back (direction);
               }
            }

            // update the list with the new position
            object_list_iter = object_list.erase (object_list_iter);
            for (int ii=begin_pos; ii!=end_pos; ii++, object_list_iter++);
            object_list.insert (object_list_iter, 1, obj);

            index++;
         }
         else
         {
            if (current_plot_id != parent_id)
            {
               command_list[index].push_back ("\ncurrent_plot(\""+parent_id+"\")");
               current_plot_id = parent_id;
            }

            // if the end position falls before the start position go forward

            // which is faster going to the back and forward tracking or
            // just going backward
            if ((begin_pos - end_pos) <= end_pos)
            {
               for (int ii=begin_pos; ii>end_pos; ii--)
               {
                  direction = "";
                  direction.append("shuffle(\"");
                  direction.append(obj->get_id());
                  direction.append("\",chips_");
                  direction.append(v_object_to_string(obj->get_object_type()));
                  direction.append(", chips_backward)");
                  command_list[index].push_back (direction);
               }
            }
            else
            {

               direction = "";
               direction.append("shuffle(\"");
               direction.append(obj->get_id());
               direction.append("\",chips_");
               direction.append(v_object_to_string(obj->get_object_type()));
               direction.append(", chips_back)");
               command_list[index].push_back (direction);
               for (int ii=0; ii<end_pos; ii++)
               {
                  direction = "";
                  direction.append("shuffle(\"");
                  direction.append(obj->get_id());
                  direction.append("\",chips_");
                  direction.append(v_object_to_string(obj->get_object_type()));
                  direction.append(", chips_forward)");
                  command_list[index].push_back (direction);
               }
            }

            // update the list with the new position
            object_list_iter = object_list.erase (object_list_iter);
            for (int ii=begin_pos; ii != end_pos; ii--, object_list_iter--);
            object_list.insert (object_list_iter, 1, obj);

            index++;

         }
      }

      // write out the command list -- this list leaves room for more optimization
      for (int ii=0; ii<obj_cnt; ii++)
      {
         command_list_iter = command_list[ii].begin ();
         for (; command_list_iter != command_list[ii].end (); command_list_iter++)
            out << (*command_list_iter) << endl;
      }

      delete []command_list;
   }

   // hide depths if necessary
   for (depth_vis_map::iterator dh=vis_table.begin();dh!=vis_table.end();dh++)
      if ((*dh).second) out << "hide_depth("<<(*dh).first<<")"<<endl;

   
}

