// 
//  Copyright (C) 2004-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSENGINE_HH
#define CHIPSENGINE_HH

#include <string>
#include <list>
#include <iostream>
#include <vector>
#include <stdio.h>

#include "ChipsWindow.hh"
#include "ChipsId.hh"
#include "ChipsMatrix.hh"

#include "Transform.hh"


using namespace std;


class ChipsEngine
{
  public:

   ChipsEngine(bool load_prefs=true);
   virtual ~ChipsEngine();

   // current indicator operations
   int     set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec=true, 
                       bool setmask=true);
   string  current_window()   { return current; }
   char*   get_current_frame(ChipsId* cid);

   // system preferences 
   int     load_preferences(string in_file = "");
   int     save_preferences(string in_file = "");
   int     set_preference(string pref, string value);
   int     override_preferences(string pref, string object);
   int     end_override_preferences ();
   char*   get_preference(string pref);
   int     set_cascading_property(ChipsId* cid, V_OBJECT_TYPE start,
                                  string property,  string value);

   int     highlight_object(ChipsId* cid, V_OBJECT_TYPE obj);
   int     move(ChipsId* cid, V_OBJECT_TYPE obj,  double xx, double yy, int mode);
   int     move(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double xx, double yy, int mode);
   int     resize(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *resize_data);
   int     rotate(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle);
   int     select(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected);
   int     get_annotation_position (const ChipsId* cid, V_OBJECT_TYPE obj, ChipsVector &pos);
   int     change_annotation_coordinate (const ChipsId* cid, V_OBJECT_TYPE obj, COORD_SYS coordsys, const string& coordid);

   int     get_object_count(ChipsId* cid, V_OBJECT_TYPE obj, int &out_val);
   int     get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj, COORD_SYS& 
                                out_val, string& out_id);
   int     get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num);
   int     set_line_stipple(V_LINE_TYPE line, int val); 
   int     get_line_stipple(V_LINE_TYPE line, int &out_val); 

   int     setup_pick (const ChipsId *cid, int count, unsigned int &inout_mask, ChipsId &out_cid);

   // undo/redo
   int     undo(int depth=1);
   int     undo(string tag);
   int     redo(int depth=1);
   int     redo(string tag);
   int     set_undo_tag(string tag);
   int     clear_undo_tag(string tag);
   int     load_state(string fname, unsigned int mask=V_LOAD_ALL|V_CLEAR_ALL);
   int     save_state(string fname, bool override=false, 
                      bool override_block=false, int mode=0);
   int     open_undo_block();
   int     close_undo_block();
   int     open_undo_buffer();
   int     close_undo_buffer();
   int     discard_undo_buffer();

   int     load_fill(string in_file, int slot=chips_userfill1);
   int     load_colormap(long len, double* r, double* g, 
			 double* b, double* alpha, int slot=chips_usercmap1);
   int     load_colormap(string in_file, int slot=chips_usercmap1);
   
   ChipsObject* get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                            bool include_deref=false); 

   // window operations
   int     add_window(ChipsId* cid, string wid = "", bool reset_errs = true);
   int     add_window(ChipsId* cid, float ww, float hh, string format, 
                      string wid="", bool reset_errs = true); 
   int     delete_window(ChipsId* cid,  bool force_del=false); 
   int     iconify_window(ChipsId* cid);
   int     deiconify_window(ChipsId* cid);
   int     display_grid(ChipsId* cid, int scale=V_GRID_SCALE_DEFAULT);
   int     hide_grid(ChipsId* cid);
   int     set_grid_color(ChipsId* cid, string color);
   int     display_layout(ChipsId* cid);
   int     hide_layout(ChipsId* cid);
   int     set_redraw(bool val); 
   int     append_title_text (const ChipsId *cid, string text);
 
   // frame operations 
   int     add_frame(ChipsId* cid, string fid = "", bool reset_errs = true);
   int     add_frame(ChipsId* cid, float x1, float y1, float x2, float y2,
                     string val = "", bool reset_errs = true);
   int     delete_frame(ChipsId* cid);
   int     set_framemode(ChipsId* cid, int val);
   int     tile(ChipsId* cid, unsigned int cols, unsigned int rows);
   int     row_tile(ChipsId* cid, unsigned int rows);
   int     col_tile(ChipsId* cid, unsigned int cols);
   int     prev_frame(ChipsId* cid);
   int     next_frame(ChipsId* cid);
   int     reposition_frame(ChipsId* cid, float x1,
                            float y1, float x2, float y2);
   int     view_single_frame(ChipsId* cid);
   int     view_placed_frame(ChipsId* cid);
   int     arrange_frames(ChipsId* cid, int cnt, string* fids);
/*   char*   get_frame_order(ChipsId* cid);
*/
   int     get_frame_order (const ChipsId* cid, list<string> &out_order);
   int     blink_frames(ChipsId* cid, string fid2, float duration=3.0);
   int     stop_blink(ChipsId* cid);
   int     display_depth(ChipsId* cid, int depth);
   int     hide_depth(ChipsId* cid, int depth);
   int     collapse_depths(ChipsId* cid, int* srcs, int scnt, int dest);
   int     rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner);

   int     set_property(ChipsId* cid, void* in_val, int in_property);
   int     get_property(ChipsId* cid, void* out_val, int out_property);
   int     adjust_grid_xgap(ChipsId* cid, float xgap);
   int     adjust_grid_ygap(ChipsId* cid, float ygap);
   int     adjust_grid_xrelsize(ChipsId* cid, int col, float ratio);
   int     adjust_grid_yrelsize(ChipsId* cid, int row, float ratio);
   int     adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* ratios);
   int     adjust_grid_yrelsizes(ChipsId* cid, int rnt, float* ratios);
   int     adjust_grid_gaps(ChipsId* cid, float xgap, float ygap);
   int     grid_objects(ChipsId* cid, int cols, int rows, float xgap, 
                        float ygap, int usage = 0, int lcnt =0,
                        string* lids = NULL);
   int     grid_objects(ChipsId* cid, int orient, int cnt, int* vals, 
                        float xgap, float ygap, int usage = 0, int lcnt = 0,
                        string* lids = NULL);
   int     swap_object_positions(ChipsId* cid, string lid2);

   // plot operations
   int     add_plot(ChipsId* cid, string plid = "");
   int     add_plot(ChipsId* cid, float px1, float py1, float px2, float py2,
                    string plid = "");
   int     delete_plot(ChipsId* cid);
   int     clear_plot(ChipsId* cid);
   int     reposition_plot(ChipsId* cid, float x1,
                           float y1, float x2, float y2);
   int     limits(ChipsId* cid, int dim, double xmin, double xmax,
                     double ymin, double ymax);
   int     limits(ChipsId* cid, int dim, double min, double max);
   int     limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                  V_AUTOLIMIT_TYPE automax);
   int     limits(ChipsId* cid, int dim, double min,
                  V_AUTOLIMIT_TYPE automax);
   int     limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                  double max);
   int     limits(ChipsId* cid, int dim, V_OBJECT_TYPE obj,  ChipsId* ccid);
   
   int     zoom (ChipsId* cid, float scale); 
   int     panto (ChipsId* cid, double x, double y);
   int     pan (ChipsId* cid, double x1, double y1, double x2, double y2);

   // deprecated commands
   int     pan (ChipsId* cid);
   int     zoom (ChipsId* cid);


   // curve operations
   int     add_curve(ChipsId* cid, long cnt, double* x, double* y,
                     double* ydown, double* yup, double* xdown, double* xup, 
                     double* ylims, double* xlims, Transform *transform,
                     string cvid = "");
   int     delete_curve(ChipsId* cid);

   // contour operations 
   int     add_contour(ChipsId* cid, long cnt, double* x, double* y,
                     double* z, unsigned int numlev, double* levels, 
                     Transform *trans=NULL, string ctid = "");
   int     add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
                       unsigned int numlev, double* levels, Transform *trans=NULL, 
                       string attlist= "");
   int     delete_contour(ChipsId* cid);
   int     set_contour_levels(ChipsId* cid, unsigned int arraysize,
                              double* pos);
   int     get_contour_levels(ChipsId* cid, vector<double> &pos);


   // histogram operations
   int     add_histogram(ChipsId* cid, long cnt, 
                         double* bin_low, double* bin_high, double* counts, 
                         double* yd= NULL, double* yu = NULL, 
                         string hid = "");
   int     delete_histogram(ChipsId* cid);

   // axis  operations
   int     add_axis(ChipsId* cid, int dim, float pos, float minx, float maxx,
              float miny, float maxy, Transform *transform, string aid="");
   int     delete_axis(ChipsId* cid);
   int     set_arbitrary_tick_positions(ChipsId* cid, unsigned int arraysize,
                                        double* pos, string* vals=NULL);
   int     bind_axes(ChipsId* cid, const string lid, const string aid);
   int     unbind_axes(ChipsId* cid, const string lid, const string aid);

   int     format_axis_point (const ChipsId* cid, double pnt, string &formatted_pnt);

   // image operations
   int     add_image(ChipsId* cid, string filename,
                     Transform *transform, string iid = "");
   int     add_image(ChipsId* cid, string filename, float ix1, float iy1,
                     float ix2, float iy2, Transform *transform, string iid = "");
   int     add_image(ChipsId* cid, double* rvals, double* gvals, 
                     double* bvals, double* avals, int xaxis, int yaxis, 
                     Transform *transform, string iid="");
   int     delete_image(ChipsId* cid);
   int        set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel);
   int        remove_image_channel(ChipsId* cid, int channel);

   // colorbar operations
   int     add_colorbar(ChipsId* cid, double xpos, double ypos, string obj_id="",
              V_OBJECT_TYPE obj=chips_not_object, string tid = "");
   int     delete_colorbar(ChipsId* cid);

   // label operations
   int     add_label(ChipsId* cid, double xpos, double ypos,
                     string text, string tid = "");
   int     delete_label(ChipsId* cid);

   // line operations 
   int     add_line(ChipsId* cid, double startx, double starty,
                    double endx, double endy, string tid = "");
   int     add_line(ChipsId* cid, double* xx, double* yy, 
                    unsigned int count, string tid = "");
   int     delete_line(ChipsId* cid);

   // point operations
   int     add_point(ChipsId* cid,  double xx, double yy,
                     string ptid = "");
   int     add_point(ChipsId* cidi, double xx, double yy, double zz,
                     string ptid = "");
   int     delete_point(ChipsId* cid);

   // legend operations
   int     add_legend(ChipsId* cid, const char** curve_ids,
                      const char** curve_labels, string lgid = "");
   int     delete_legend(ChipsId* cid);

   // region operations
   int     add_region(ChipsId* cid, double* xx, double* yy, 
                    unsigned int count, string tid = "");
   int     add_region(ChipsId* cid, int sides, double xx, double yy, 
                    double radius, string tid = "");
   int     delete_region(ChipsId* cid);

   // print operations
   int     print_window(ChipsId* cid, string outfile, string prefs);
   int     print_frame(ChipsId* cid, string outfile);
   int     print_image (ChipsId* cid, string outfile, string prefs);

   // font operations
   int     load_font(string fontname, string fontpath, V_FONTSTYLE_TYPE style);
   virtual bool verify_fontfile_exists(string fontpath); 


   // info commands
   char*   info(ChipsId* cid, V_OBJECT_TYPE obj=chips_window, bool descend=true,
                bool full=true); 
   char*   info_coordinate(ChipsId* cid);
   char*   info_depth(ChipsId* cid, int depth=0);
   char*   info_bound_axes(ChipsId* cid);
   char*   list_current(); 

// coordinate commands
   int     convert_coordinate (ChipsVector &in_val, ChipsVector &out_val,
                               ChipsId *cid1, ChipsId *cid2, int axis);
   int     convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, 
                               ChipsId *cid2, int axis);

   bool    check_coordinate_exists (ChipsId *cid1);
   int     get_current_coord (ChipsId *cid);
   int     get_coord_object_count(ChipsId *cid, int objtype = chips_not_object);

   // 'all' methods 
   int     delete_all_windows(ChipsId* cid=NULL, bool force_del=false);
   int     delete_all_frames(ChipsId* cid);
   int     delete_all_plots(ChipsId* cid);
   int     delete_all_images(ChipsId* cid);

   void    disable_redraw()  {  redraw_block = true; }
   void    enable_redraw()   {  redraw_block = false; }
   bool    redraw_blocked()  {  return redraw_block; }
   void    redraw_window(string wid = "");
   void    update_default_display_colors(); 

   // virtual functinos
   virtual ChipsWindow* new_window(string window_id);
   virtual ChipsWindow* new_window(ChipsStateStream& in);
   virtual ChipsWindow* new_window(float width, float height,
                                 string format="inch",
                                 string window_id="win");
   virtual int reference_object (ChipsObject *co, bool change_cnt, const string &curr = "");
   virtual int dereference_object (ChipsObject *co, bool change_cnt, const string &curr="");
   virtual int clear_cmd (ChipsObject *co);
   char* create_window_id(); 

   char* get_errors(int mode=0); 
   int   add_client_error(int errcode, string errtext);
   int   clear_errors();
   int   set_error_verbosity(int val);
   int   get_error_verbosity(int &out_val);
   int   get_error_types(int &out_val);
   int   get_error_count(int type, int &out_val);
   bool  get_ignore_state(); 
   void  set_ignore_state(bool value); 

  protected:
   int get_windows (const ChipsId *cid, list <ChipsWindow *>&out_windows, bool allow_all);

   ChipsWindow* get_window(string wid);

  private:
   list <ChipsWindow*>window;  // list of windows
   map <unsigned int, ChipsWindow *> dwindow; // list of derefenced windows
   string  current;            // current window
   int     window_cnt;         // reference count for window id creation
   unsigned int buffer_depth;  // nested undo buffer depth 
   bool    redraw_block;       // force disable redraw of windows
};

#endif 
