// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSFRAME_HH
#define CHIPSFRAME_HH

#include <list>
#include <map>
#include <string>
#include <stdio.h>
#include <vector>

#include "ChipsObject.hh"
#include "ChipsId.hh"
#include "ChipsFrameSystem.hh"
#include "ChipsColor.hh"

#include "Transform.hh"

class ChipsDepthControl; 
class ChipsStateStream; 
class ChipsImage;
class ChipsPlot;

using namespace std;


class ChipsFrame: public ChipsObject
{
  public:

   ChipsFrame(ChipsObject* par, string val);
   ChipsFrame(ChipsObject* par, ChipsStateStream& in);
   ChipsFrame(ChipsObject* par, float x1, float y1, float x2, float y2, 
             string val);

   virtual void object_delete ();

   // methods for split interactions
   int        adjust_grid_xgap(float xg) { return adjust_grid_gaps(xg, ygap);} 
   int        adjust_grid_ygap(float yg) { return adjust_grid_gaps(xgap, yg);} 
   int        adjust_grid_gaps(float xg, float yg); 
   int        adjust_grid_xrelsize(int entry, float size, bool rec=true); 
   int        adjust_grid_yrelsize(int entry, float size, bool rec=true); 
   int        adjust_grid_xrelsizes(int cnt, float* sizes, bool rec=true); 
   int        adjust_grid_yrelsizes(int cnt, float* sizes, bool rec=true); 
   int        grid_objects(int cols, int rows, float xg, float yg,
                           int usage = 1, int pcnt = 0, string* plids = NULL); 
   int        grid_objects(int orient, int cnt, int* vals, float xg, 
                           float yg, int usage = 1, int pcnt = 0,
                           string* plids = NULL);

   int        reset_relsizes(int count, int dimn, float *vals, float relsum);
   int        reset_split_variables(int orient, int cnt, int* vals, int cols,
                                    int rows, float xg, float yg, 
                                    string* plids=NULL);
   int        swap_object_positions(string id1, string id2);

   // methods for plot interactions 
   int        add_plot(ChipsId* cid);
   int        add_plot(ChipsId* cid, float px1, float py1, float px2, 
                       float py2);
   int        reposition_plot(const ChipsId* cid, float px1, float py1,
                              float px2, float py2, bool rec=true);
   int        reposition_plot_area(ChipsPlot* plt, float px1, float py1,
                                   float px2, float py2, bool offscreen=false,
				   bool rec=true);
   int        delete_plot(ChipsId* cid, bool force_del=false);
   string     get_current_plot()  { return curr_plot; }
   int        set_current_plot(const string &val);
   ChipsPlot* get_plot(const string& val="");

   int        limits(ChipsId* cid, int dim, double xmin, double xmax, 
                     double ymin, double ymax);
   int        limits(ChipsId* cid, int dim, double min, double max);
   int        limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                     V_AUTOLIMIT_TYPE automax);
   int        limits(ChipsId* cid, int dim, double min,
                     V_AUTOLIMIT_TYPE automax);
   int        limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                     double max);
   int        zoom (ChipsId* cid, float scale, bool rec=true);
   int        panto (ChipsId* cid, double x, double y);
   int        pan (ChipsId* cid, double x1, double y1, double x2, double y2, bool rec=true);

   // methods for curve interactions
   int        add_curve(ChipsId* cid, long cnt, double* x, double* y,
                        double* ydown, double* yup, double* xdown, 
                        double* xup, double* ylims, double* xlims,
                        Transform *tranform);
   int        delete_curve(ChipsId* cid);

   // methods for contour interactions
   int        add_contour(ChipsId* cid, long cnt, double* x, double* y,
                          double* z, unsigned int numlev, double* levels, 
                          Transform *trans);
   int        add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
                          unsigned int numlev, double* levels, Transform *trans);
   int        delete_contour(ChipsId* cid);
   int        set_contour_levels(ChipsId* cid, unsigned int arraysize,
                                 double* pos);
   int        get_contour_levels(ChipsId* cid, vector<double> &pos);

   // histogram operations
   int        add_histogram(ChipsId* cid, long cnt,
                            double* bin_low, double* bin_high, double* counts,
                            double* yd = NULL, double* yu = NULL);
   int        delete_histogram(ChipsId* cid);

   // methods for axis interactions
   int        add_axis(ChipsId* cid, int dim, float pos, float minx, float maxx, float miny, float maxy, Transform *transform);
   int        delete_axis(ChipsId* cid);
   int        set_arbitrary_tick_positions(ChipsId* cid, unsigned int arraysize,
                                           double* pos, string* vals=NULL);
   int        bind_axes(ChipsId* cid, const string lid, const string aid,
                        bool rec = true);
   int        unbind_axes(ChipsId* cid, const string lid, const string aid,
                        bool rec = true);
   int        format_axis_point (const ChipsId* cid, double pnt, string &formatted_pnt);

   // methods for image interactions
   int        add_image(ChipsId* cid, string filename, Transform *transform);
   int        add_image(ChipsId* cid, string filename, float ix1, 
                      float iy1, float ix2, float iy2, Transform *transform);
   int        add_image(ChipsId* cid, double* rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, Transform *transform);
   int        delete_image(ChipsId* cid);
   int        set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel);
   int        remove_image_channel(ChipsId* cid, int channel);
   int        print_image (ChipsId *cid, const string &filename, int fmt);

   int        add_label(ChipsId* cid, double xpos, double ypos, string text);
   int        delete_label(ChipsId* cid);

   int        add_line(ChipsId* cid, double startx, double starty, double endx,
                       double endy);
   int        add_line(ChipsId* cid, double* xx, double* yy,unsigned int count);
   int        delete_line(ChipsId* cid);

   int        add_colorbar(ChipsId* cid, double xpos, double ypos,
                            string obj_id="", V_OBJECT_TYPE obj=chips_not_object);
   int        delete_colorbar(ChipsId* cid);

   int        add_point(ChipsId* cid, double xx, double yy);
   int        add_point(ChipsId* cid, double xx, double yy, double zz);
   int        delete_point(ChipsId* cid);

   int        add_region(ChipsId* cid, double* xx, double* yy,
                       unsigned int count);
   int        add_region(ChipsId* cid, int sides, double xx, double yy,
                         double radius);
   int        delete_region(ChipsId* cid);

   // methods for legend object interactions
   int        add_legend(ChipsId* cid, const char** curve_ids,
                         const char** curve_labels, string lgid = "");
   int        delete_legend(ChipsId* cid);

   // methods for coordinate system
   int        convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, ChipsId *cid2, 
                               int axis);
   int        get_current_coord (ChipsId *cid);
   ChipsCoordinate *find_coordinate_system (ChipsId &cid, int &err_status);


   int        move_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double xx, double yy, int mode);
   int        resize_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *bounds);
   int        rotate_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle);
   int        select_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected);

   int        get_annotation_position (const ChipsId* cid, V_OBJECT_TYPE obj, ChipsVector &pos);
   int        change_annotation_coordinate (const ChipsId* cid, V_OBJECT_TYPE obj, COORD_SYS coordsys, const string& coordid);



   int        get_object_count(ChipsId* cid, V_OBJECT_TYPE obj, int& out_val);
   int        get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj, 
                                   COORD_SYS& out_val, string& out_id);
   int        get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num);
   int        set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj);
   int        set_property(ChipsId* cid, void* in_val, int in_property,
                           bool rec=true);
   int        get_property(ChipsId* cid, void* out_val, int out_property);

   string     get_color()    { return color.get_name(); }
   int        set_color(string in_color);
   string     get_border_color()    { return border_color.get_name(); }
   int        set_border_color(string in_color);
   int        adjust_default_colors(); 
   
   int        update_print_scale (double scale);

   int        set_border_visibility(bool val); 
   bool       get_border_visibility()     { return display_border; }

   char*      info(ChipsId* cid, V_OBJECT_TYPE obj=chips_plot, 
                   bool descend=true, bool current_obj=false, bool full=true);
   char*      info_coordinate (ChipsId* cid); 
   char*      info_bound_axes(ChipsId* cid); 
   char*      info_depth(ChipsId* cid, int depth=0);
   char*      list_current();
   int        set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec=true);

   int        adjust_pos_by_delta(float dx, float dy);
   int        reposition(float x1, float y1, float x2, float y2, bool rec=true);
   int        set_display_location(int mode, int x1, int y1, int x2, int y2); 
   void       keep_aspect(bool val)  { preserve_aspect = val; } 
   bool       keep_aspect()   { return preserve_aspect; }

   int        set_transparency(bool val);
   bool       get_transparency()   { return !opaque; }
   void       get_bb_norm_pos(float &out_x1, float &out_y1, 
                              float &out_x2, float &out_y2);
   ChipsAxisBind*   get_bind_table()   { return bind_table; };
   double           get_scale_factor() { return scale_factor; }
   int        highlight_object(ChipsId* cid, V_OBJECT_TYPE obj, bool vis);

   virtual int get_depth();
   virtual int move(COORD_SYS coord_sys, const string &coord_id, double xx,
                    double yy, int mode, bool rec=true);

   virtual int set_order (int order_num)    { return ERR_FEATURE_DNE; }
   virtual int change_order()               { return ERR_FEATURE_DNE; }
   virtual int change_position()            { return ERR_FEATURE_DNE; }
   virtual int change_color()               { return ERR_FEATURE_DNE; }
   virtual int change_border_visibility()   { return ERR_FEATURE_DNE; }
   virtual int change_border_color()        { return ERR_FEATURE_DNE; }
   virtual int change_transparency()        { return ERR_FEATURE_DNE; }
   virtual int change_visible ()            { return ERR_FEATURE_DNE; }
   virtual int blink_frames()               { return ERR_FEATURE_DNE; }
   virtual int redraw()                     { return ERR_FEATURE_DNE; }
   virtual int show ();
   virtual int hide ();
   virtual int force_show ();  // only show the frame if visible is true

   virtual ChipsPlot* new_plot(string plid="");
   virtual ChipsPlot* new_plot(ChipsStateStream& in);
   virtual ChipsPlot* new_plot(float posx1, float posy1, float posx2,
                               float posy2, string plid="");

   virtual int reference_object (ChipsObject *co, bool change_cnt, const string &curr="");
   virtual int dereference_object (ChipsObject *co, bool change_cnt, const string &curr="");
   virtual int clear_cmd (ChipsObject *co);

   virtual void serialize(ChipsStateStream& out);
   void         serialize(ChipsStateStream& out, bool write_coord);
   void         serialize_ascii(ChipsStateStream& out, int keep=1);
   void         serialize_objects(ChipsStateStream& out);
   virtual void deserialize(ChipsStateStream& in);
   int          deserialize_bind_table(ChipsStateStream& in);
   int          deserialize_plot(ChipsStateStream& in);
   int          deserialize_objects(ChipsStateStream& in);
   virtual int  deserialize_depth_table(ChipsStateStream& in);

   void       set_frame_num (int val) {frame_num = val;};
   int        get_frame_num () const {return frame_num;};

   ChipsObject* get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                            bool include_deref=false);
   int        delete_object (const string &id, V_OBJECT_TYPE otype);

   int        display_depth(int depth, bool rec=true);
   int        hide_depth(int depth, bool rec=true);
   int        collapse_depths(int* sources, int src_cnt, int dest, bool rec=true);
   int        uncollapse_depths(multimap<int, int> &dest_table, int src_depth);
   int        rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner, bool rec=true);
   int        update_depth_table(V_OBJECT_TYPE obj, string oid, int how,
                                 int pos = 0);
   int        clear_plot (ChipsId *cid);

   int        check_plot_updates ();

   virtual void enable_picklimits (bool flag);

   void disable_all_interaction ();
   void enable_all_interaction ();

   // depth code
   int        add_object_at_depth(ChipsObject* obj);
   int        add_object_to_depth_table(ChipsObject* ob); 
   int        remove_object_from_depth_table(ChipsObject* ob); 

   ChipsDepthControl* get_depth_table () {return this->depth_table;};

  protected:

   virtual ~ChipsFrame();

   virtual void change_mode (int mode) {};

   int        dereference_plot (ChipsPlot *cp);

   int        compute_object_pos(); 
   int        arrange_objects(int usage, int lcnt, string* lids);  
   char*      create_plot_id();

   string     get_default_plot_id();

   int get_plots (const ChipsId *cid, list <ChipsPlot *>&out_plots, bool allow_all);

   void       disable_plot_interaction (const string &plot_id);
   void       enable_plot_interaction (const string &plot_id);

   ChipsAxisBind* bind_table;
   ChipsDepthControl* depth_table; 

   void       initialize(bool create_coord=true);  // set defaults

   map <string, ChipsPlot*>   plot;  
   map <unsigned int, ChipsPlot*>   dplot;
   int        plot_cnt;      // reference count for plot id creation
   string     curr_plot;     // id of current plot 
   string     asf_plot;      // last plot used in  ascii serialization 
   bool       preserve_aspect; 
   bool       opaque;       // is frame background opaque or transparent? 
   bool       scale;        // scale labels, etc... 
   bool       display_border;  // display frame border 
   double     scale_factor; // scale factor for labels, etc... 
   ChipsColor color;        // color of frame
   ChipsColor border_color;  // color of frame border

   // split variables 
   float      leftmargin;   // margin for 'split' 
   float      rightmargin;  // margin for 'split' 
   float      topmargin;    // margin for 'split' 
   float      bottommargin; // margin for 'split' 
   float      xgap;         // size of x split gap
   float      ygap;         // size of x split gap
   float      xrelsum;      // sum of x rel size components
   float      yrelsum;      // sum of y rel size components
   float*     spleft;       // left  x split position array  
   float*     spbottom;     // lower y split position array 
   float*     spright;      // right x split position array  
   float*     sptop;        // upper y split position array 
   float*     spxrel;       // relative x array
   float*     spyrel;       // relative x array
   int        spcnt;        // count of split layers 
   int        sprows;       // number of split rows (0 = use sparr)
   int        spcols;       // number of split cols (0 = use sparr)
   int*       sparr;        // array of split counts
   string*    spplot;       // list of split plots 

   int        bbx1;         // Window coord in pixel 
   int        bbx2;         // Window coord in pixel
   int        bby1;         // Window coord in pixel
   int        bby2;         // Window coord in pixel
   int        def_depth;    // initial default depth 
   float      init_height;  // initial window height 
   float      init_width;   // initial window width 

   double      bb_norm_x1;   // Bounding box normalize
   double      bb_norm_x2;   // Bounding box normalize
   double      bb_norm_y1;   // Bounding box normalize
   double      bb_norm_y2;   // Bounding box normalize

   int        frame_num;       // this frames order 0th, 1st, 2nd, etc.
   bool       deserialize_coord; // do we need to deserialize coord
};


#endif 
