// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsFrameSystem.hh"
#include "ChipsPlotSystem.hh"
#include <algorithm>
#include <sstream>

unsigned long ChipsFrameSystem::global_ref_id = 0;

ChipsFrameSystem::ChipsFrameSystem (ChipsCoordinate *par):ChipsCoordinate (par)
{
   this->parent =  par;
   min_pos = ChipsVector (3);
   max_pos = ChipsVector (3);
   range = ChipsVector (3);
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
   this->old_id = "";

   this->parent->add_coord_sys (this);

}

ChipsFrameSystem::ChipsFrameSystem (ChipsCoordinate *par,
		const ChipsVector &pmin, const ChipsVector &pmax):ChipsCoordinate (par)
{
   this->min_pos = pmin;
   this->max_pos = pmax;
   this->range = this->max_pos - this->min_pos;
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
   this->old_id = "";

   this->parent->add_coord_sys (this);
}


ChipsFrameSystem::ChipsFrameSystem (const ChipsFrameSystem &fs):ChipsCoordinate (fs)
{
   this->parent = fs.parent;
   this->id = fs.id;
   this->min_pos = fs.min_pos;
   this->max_pos = fs.max_pos;
   this->range = fs.range;
   this->plot_systems = fs.plot_systems;
   this->chips_objects = fs.chips_objects;
   this->ref_id = fs.ref_id;
   this->old_id = fs.old_id;
}

ChipsFrameSystem::ChipsFrameSystem (ChipsStateStream &in, ChipsCoordinate *par):ChipsCoordinate (par)
{
   this->deserialize (in);
}


ChipsFrameSystem::~ChipsFrameSystem ()
{
   if (parent)
      parent->remove_coord_sys (id);
   if (!plot_systems.empty ())
   {
      map<string, ChipsPlotSystem *>::iterator ii;
      for (ii = plot_systems.begin (); ii!=plot_systems.end (); ii++)
         ii->second->remove_parent ();
      plot_systems.clear ();
   }

}


void ChipsFrameSystem::operator = (const ChipsFrameSystem &fs)
{
   if (this == &fs)
      return;

   this->min_pos = fs.min_pos;
   this->max_pos = fs.max_pos;
   this->range = fs.range;
   this->plot_systems = fs.plot_systems;
   this->chips_objects = fs.chips_objects;
}

int ChipsFrameSystem::convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id)
{
   int rval = ERR_STATUS_GOOD;
   map <string, ChipsPlotSystem *>::iterator from_ps;
   map <string, ChipsPlotSystem *>::iterator to_ps;
   ChipsCoordinate *to_ds;

   if (!in || !out)
      return ERR_STATUS_BAD;

   if (size <= 0)
      return ERR_STATUS_BAD;

   if ((from_cs == FRAME_NORM) && (from_id.length () > 0) && (id != from_id))
      return ERR_STATUS_BAD;

   if ((to_cs == FRAME_NORM) && (to_id.length () > 0) && (id != to_id))
      return ERR_STATUS_BAD;

   if (to_cs == PLOT_NORM){
      to_ps = plot_systems.find (to_id);
      if (to_ps == plot_systems.end ())
         return ERR_COORD_DNE;
   }

   if (to_cs == DATA){
      to_ds = get_coord (DATA, to_id);
      if (!to_ds)
         return ERR_COORD_DNE;
   }

   if ((range[0] == 0) || (range[1] == 0))
      return ERR_COORD_RANGE_ZERO;
   
   switch (from_cs){
      case WINDOW_NORM: /* Fallthrough intended */
      case PIXEL:
      case PIXEL_REAL:
      {
         if (!parent)
            rval = ERR_STATUS_BAD;

	 if (!rval && ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) || (to_cs == PIXEL_REAL))){
            return parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
						to_cs, to_id);
            
	 }
	 else if (!rval)
	    rval = parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
                                                WINDOW_NORM, parent->get_id ());

	 if (rval != ERR_STATUS_GOOD)
	    return rval;

	 for (int ii=size; ii--;){
	    // if the size is one use the dimension to determine
            // which axis to convert
            if (out[ii].get_size () < 2){
	        return ERR_STATUS_BAD;
	    }
            else{
		out[ii][0] = (out[ii][0]-min_pos[0])/range[0];
                out[ii][1] = (out[ii][1]-min_pos[1])/range[1];
	    }
         }

         if (to_cs == PLOT_NORM){
            rval = to_ps->second->convert_cs_to_cs (out, out, size, FRAME_NORM, id,
                                                to_cs, to_id);
         }
   
         if (to_cs == DATA){
            rval = to_ds->convert_cs_to_cs (out, out, size, FRAME_NORM, id,
                                                to_cs, to_id);
         }

      }
      break;
      case FRAME_NORM:

         if (!rval && ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) || (to_cs == PIXEL_REAL))){
            for (int ii=size; ii--;){
	       out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
               if (out[ii].get_size () < 2)
	          return ERR_STATUS_BAD;
               else{
                  out[ii][0] = in[ii][0]*range[0] + min_pos[0];
                  out[ii][1] = in[ii][1]*range[1] + min_pos[1];
	       }
            }
            rval =  ERR_STATUS_GOOD;
            if (!parent)
               rval = ERR_STATUS_BAD;
	    if (!rval && ((to_cs == PIXEL) || (to_cs == PIXEL_REAL)))
 	       rval = parent->convert_cs_to_cs (out, out, size, WINDOW_NORM, 
						parent->get_id(), to_cs, to_id);
         }
	 else if (to_cs == FRAME_NORM){
	    for (int ii=size; ii--;)
               out[ii] = in[ii];
            rval =  ERR_STATUS_GOOD;
	 }
	 else if (to_cs == PLOT_NORM){
	    rval = to_ps->second->convert_cs_to_cs (in, out, size, FRAME_NORM, 
							id, to_cs, to_id);
         }
         else if (to_cs == DATA){
            rval = to_ds->convert_cs_to_cs (in, out, size, FRAME_NORM,
                                                        id, to_cs, to_id);
         }

      break;
      case PLOT_NORM:
         from_ps = plot_systems.find (from_id);
         if (from_ps == plot_systems.end ())
            return ERR_COORD_DNE;
	 if ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM) ||
	     (to_cs == PLOT_NORM)){
	    rval = from_ps->second->convert_cs_to_cs (in, out, size, from_cs,
                                                from_id, FRAME_NORM, id);
	    if (rval != ERR_STATUS_GOOD)
	       return rval;

	    if ((to_cs == WINDOW_NORM) ||
                (to_cs == PIXEL) || (to_cs == PIXEL_REAL))
	       rval = convert_cs_to_cs (out, out, size, FRAME_NORM,
                                                id, to_cs, to_id);
	    else if (to_cs == PLOT_NORM)
               rval = to_ps->second->convert_cs_to_cs (out, out, size, 
						FRAME_NORM, id, to_cs, to_id);
            else if (to_cs == DATA)
               rval = to_ds->convert_cs_to_cs (out, out, size,
                                                FRAME_NORM, id, to_cs, to_id);
         }
	 else 
	    rval = ERR_STATUS_BAD;
      break;
      case DATA:
      {
         ChipsCoordinate *from_ds = get_coord (DATA, from_id);
         if (!from_ds)
            return ERR_COORD_DNE;
         if ((to_cs == WINDOW_NORM) ||
             (to_cs == PIXEL) ||
             (to_cs == PIXEL_REAL) ||
             (to_cs == FRAME_NORM) ||
             (to_cs == PLOT_NORM)){
            rval = from_ds->convert_cs_to_cs (in, out, size, from_cs,
                                                from_id, FRAME_NORM, id);
            if (rval != ERR_STATUS_GOOD)
               return rval;

            if ((to_cs == WINDOW_NORM) ||
                (to_cs == PIXEL) || (to_cs == PIXEL_REAL))
               rval = convert_cs_to_cs (out, out, size, FRAME_NORM,
                                                id, to_cs, to_id);
            else if (to_cs == PLOT_NORM)
               rval = to_ps->second->convert_cs_to_cs (out, out, size,
                                                FRAME_NORM, id, to_cs, to_id);
            else if (to_cs == DATA)
               rval = to_ds->convert_cs_to_cs (out, out, size,
                                                FRAME_NORM, id, to_cs, to_id);
         }
         else
            rval = ERR_STATUS_BAD;
      }
      break;
      default: rval = ERR_STATUS_BAD;
   }
   return rval;
}

int ChipsFrameSystem::add_coord_sys (ChipsCoordinate *cs)
{
   if (cs->get_id().length () == 0)
      return ERR_STATUS_BAD;

   if (plot_systems.find (cs->get_id()) != plot_systems.end ())
      return ERR_STATUS_BAD;

   plot_systems[cs->get_id()] = (ChipsPlotSystem *)cs;
   cs->set_parent (this);

   return ERR_STATUS_GOOD;
}


int ChipsFrameSystem::remove_coord_sys (const string &id)
{
   map <string, ChipsPlotSystem *>::iterator ps;

   if ((ps = plot_systems.find (id)) == plot_systems.end ())
      return ERR_STATUS_BAD;

   ps->second->remove_parent ();
//   delete ps->second;
   plot_systems.erase(ps);
   return ERR_STATUS_GOOD;
}


ChipsCoordinate *ChipsFrameSystem::get_coord (COORD_SYS coord_sys, const string &coord_id)
{
   ChipsCoordinate *coord = NULL;
   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      {
         if (parent)
            coord = parent->get_coord(coord_sys, coord_id);
      }
      break;
      case COORD_NONE:
      case FRAME_NORM:
      {
         if ((coord_id.length () == 0) || (coord_id == this->id) || (coord_id == this->old_id))
            coord = this;
      }
      break;
      case PLOT_NORM:
      case DATA:
      {

         map <string, ChipsPlotSystem *>::iterator ps;

         int pos = coord_id.find_first_of (".", 0);
         pos = coord_id.find_first_of (".", pos+1);
         pos = coord_id.find_first_of (".", pos+1);

         string tmp_id = coord_id.substr (0, pos);

         if ((ps = plot_systems.find (tmp_id)) != plot_systems.end ())
            coord = ps->second;

         if ((coord_sys == DATA) && coord)
            coord = coord->get_coord (coord_sys, coord_id);
         else if ((coord_sys == PLOT_NORM) && !coord)
         {
            // check the frame old ids for backwards compatibility
            for (ps = plot_systems.begin (); ps != plot_systems.end (); ps++)
               if (coord_id == ps->second->get_old_id ())
                  coord = ps->second;
         }
      }
      break;
      default:
      break; 
   }
   return coord;
}


int ChipsFrameSystem::insert (ChipsObject *co,  bool use_co_info)
{
   int rval = ERR_STATUS_BAD;
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   if (co == NULL)
      return ERR_STATUS_BAD;

   // Store object local forces the object to remain
   // in this coordinate systems list. It overrides
   // the coordinate objects coord system
   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      {
         if (parent)
            rval = parent->insert (co, use_co_info);
      }
      break;
      case COORD_NONE:
      case FRAME_NORM:
         if ((coord_id.length () == 0) || (coord_id == id)){
            chips_objects.push_back (co);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval = ERR_COORD_DNE;
      break;
      case PLOT_NORM:
      {
         map <string, ChipsPlotSystem *>::iterator ps;

         if ((ps = plot_systems.find (coord_id)) != plot_systems.end ())
            rval = ps->second->insert (co, use_co_info);
      }
      break;
   }

   return rval;
}


int ChipsFrameSystem::remove (ChipsObject *co,  bool use_co_info)
{
   int rval = ERR_STATUS_BAD;
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   if (co == NULL)
      return ERR_STATUS_BAD;

   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
         if (parent)
            rval = parent->remove (co, use_co_info);
      break;
      case COORD_NONE:
      case FRAME_NORM:
      {
         list <ChipsObject *>::iterator ii = find (chips_objects.begin (),
                                                chips_objects.end (), co);
         if (ii != chips_objects.end ()){
            chips_objects.erase (ii);
            rval = ERR_STATUS_GOOD;
         }
      }
      break;
      case PLOT_NORM:
      {
         map <string, ChipsPlotSystem *>::iterator ps;

         if ((ps = plot_systems.find (coord_id)) != plot_systems.end ())
            rval = ps->second->insert (co, use_co_info);
      }
      break;
   }


   return rval;
}


bool ChipsFrameSystem::coord_exists (COORD_SYS coord_sys, const string &id)
{
   bool rval = false;
   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
         if (parent)
            rval = parent->coord_exists(coord_sys, id);
      break;
      case COORD_NONE:
      case FRAME_NORM:
         rval = true;
      break;
      case PLOT_NORM:
      case DATA:
      {
         ChipsCoordinate *tmp_coord = this->get_coord (coord_sys, id);
         rval = (tmp_coord != NULL);
      }
      break;
   }

   return rval;
}


void ChipsFrameSystem::get_coord_list (list <string> &outList)
{
   map <string, ChipsPlotSystem *>::iterator ii;

   for (ii = plot_systems.begin (); ii != plot_systems.end (); ++ii)
      outList.push_back (ii->first);

}



void ChipsFrameSystem::set_min_pos (const ChipsVector &min) 
{
   min_pos = min;
   range = max_pos - min_pos;
   update_obj_positions ();
   
};

void ChipsFrameSystem::set_max_pos (const ChipsVector &max) 
{
   max_pos = max;
   range = max_pos - min_pos;
   update_obj_positions ();

};

void ChipsFrameSystem::set_pos (const ChipsVector &min, const ChipsVector &max)
{
   min_pos = min;
   max_pos = max;
   range = max_pos - min_pos;

   update_obj_positions ();

};

int ChipsFrameSystem::update_obj_positions (bool update_all)
{
   int rval = ERR_STATUS_GOOD;

   list <ChipsObject *>::iterator ii;
   for (ii = chips_objects.begin (); ii != chips_objects.end (); ++ii){
       rval = (*ii)->reset_position ();
       if (rval != ERR_STATUS_GOOD)
          return rval;
   }

   if (update_all){
      map <string, ChipsPlotSystem *>::iterator iter;
      for (iter = plot_systems.begin (); iter != plot_systems.end (); iter++){
         if ((rval=iter->second->update_obj_positions(update_all)) != ERR_STATUS_GOOD)
            break;
      }
   }

   return rval;
};


void ChipsFrameSystem::serialize (ChipsStateStream &out)
{ 
   map <string, ChipsPlotSystem *>::iterator ii;
   int sz = 0;

   if (out.get_minor_version() > 6)
   {
      out.write_long ((long *)&(this->global_ref_id));
      out.write_long ((long *)&(this->ref_id));
   }
   
   out.write_string (&id);
   out.write_bool (&visible);
   sz = min_pos.get_size ();
   out.write_int (&sz);
   out.write_double (&min_pos[0], min_pos.get_size ());
   out.write_double (&max_pos[0], max_pos.get_size ());

   out.write_checkpoint ();

   sz = plot_systems.size ();
   out.write_int (&sz);

   for (ii=plot_systems.begin (); ii!=plot_systems.end (); ii++)
      ii->second->serialize (out);
}

int ChipsFrameSystem::deserialize (ChipsStateStream &in)
{
   int sz = 0;
   ChipsPlotSystem *plot_sys;
   int rval = ERR_STATUS_GOOD;

   if (in.get_minor_version() > 6)
   {
      in.read_long ((long *)&(this->global_ref_id));
      in.read_long ((long *)&(this->ref_id));
      in.read_string (&id);
   }
   else
   {
      this->ref_id = this->global_ref_id;
      this->global_ref_id++;
      in.read_string (&old_id);
      this->create_id (this->id);
   }

   in.read_bool (&visible);
   in.read_int (&sz);
   if (sz > 0){
      this->min_pos = ChipsVector (sz);
      this->max_pos = ChipsVector (sz);

      in.read_double (&min_pos[0], sz);
      in.read_double (&max_pos[0], sz);

      this->range = this->max_pos - this->min_pos;
   }

   // check for corruption
   if (in.read_checkpoint () != ERR_STATUS_GOOD)
      rval = ERR_STATEFILE_INC;
   else{
      in.read_int (&sz);

      for (int ii=0; !rval && ii<sz; ii++){
         plot_sys = new ChipsPlotSystem (in, this);
         if (plot_sys)
            plot_systems[plot_sys->get_id ()] = plot_sys;
         else
            rval = ERR_STATEFILE_COR;
      }
   }

   return rval;
}

void ChipsFrameSystem::create_id (string &curr_id)
{
   stringstream tmp_id;

   if (this->parent)
      this->parent->create_id (curr_id);

   tmp_id << curr_id << "." << this->ref_id;
   curr_id = tmp_id.str ();
}

