// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSFRAMESYSTEM_HH
#define CHIPSFRAMESYSTEM_HH

#include <list>
#include <string>
#include <map>
#include "ChipsObject.hh"
#include "ChipsDefines.hh"
#include "ChipsMatrix.hh"
#include "ChipsCoordinate.hh"

class ChipsPlotSystem;

/**************************************************************
*
*    Frame System
*
*    This is a collection of information the describes the
*    frame coordinate systems This contains information
*    about where the frame boundaries are in pixel coordinates.
*    It also contains a list of objects associating with this
*    coordinate system.
*
****************************************************************/
class ChipsFrameSystem: public ChipsCoordinate
{
   public:
      /* Constructors */
      ChipsFrameSystem (ChipsCoordinate *par);
      ChipsFrameSystem (ChipsCoordinate *par,
                const ChipsVector &pmin, const ChipsVector &pmax);
      ChipsFrameSystem (const ChipsFrameSystem &);
      ChipsFrameSystem (ChipsStateStream &in, ChipsCoordinate *par);

      virtual ~ChipsFrameSystem ();
      
      /* Overloaded operators */
      void operator = (const ChipsFrameSystem &);

      virtual int convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id);
      virtual int add_coord_sys (ChipsCoordinate *);
      virtual int remove_coord_sys (const string &id);
      virtual ChipsCoordinate *get_coord (COORD_SYS coord_sys, const string &id="");
      virtual bool coord_exists (COORD_SYS coord_sys, const string &id);
      virtual void get_coord_list (list <string> &outList);

      virtual int insert (ChipsObject *co, bool use_co_info=true);
      virtual int remove (ChipsObject *co, bool use_co_info=true);
      virtual int update_obj_positions (bool update_all=true);

      void set_min_pos (const ChipsVector &min);
      void set_max_pos (const ChipsVector &max);
      void set_pos (const ChipsVector &min, const ChipsVector &max);
      virtual ChipsVector get_lower_bounds (){return min_pos;};
      virtual ChipsVector get_upper_bounds (){return max_pos;};
      virtual COORD_SYS   get_bounds_coord_sys () {return WINDOW_NORM;};

      // create an id from all parent information
      virtual void create_id (string &curr_id);

      string get_old_id () {return this->old_id;};

      virtual int deserialize (ChipsStateStream &in);
      virtual void serialize (ChipsStateStream &out);

      /* These two function should be deprecated. Use get_lower/upper_bounds. */
      ChipsVector get_min_pos () {return min_pos;};
      ChipsVector get_max_pos () {return max_pos;};

//      int change_plot_id (const string &old_id, const string &new_id);


   protected:
      map <string, ChipsPlotSystem *> plot_systems;
      ChipsVector min_pos;                //Min x/y window norm coordinate
      ChipsVector max_pos;                //Max x/y window norm coordinate
      ChipsVector range;                  //window norm range
      unsigned long ref_id;               // reference id
      static unsigned long global_ref_id; // global reference id
      string old_id;                      // the old id tag for backwards compat.
};

#endif

