// 
//  Copyright (C) 2005-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsHistogram.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"
#include "ChipsFillPattern.hh"

#include <stdio.h>
#include <config.h>

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;
extern ChipsFillPattern *fill_patterns;

#ifdef HAVE_IEEEFP_H
#include <ieeefp.h>        // finite defined in ieeefp.h (solaris)
#else
#include <math.h>          // finite() defined here      (linux, OSX)
#endif

ChipsHistogram::ChipsHistogram(ChipsObject* par, unsigned long cnt, 
   double* bin_lo, double* bin_hi, double* bincnts, string hid):
   ChipsCurve(par, cnt, bin_lo, bincnts, hid, false) 
{
   initialized = false;

   binlo = (double*)malloc(num_pts*sizeof(double));
   binhi = (double*)malloc(num_pts*sizeof(double));
   if (!bin_lo)  xx = (double*)malloc(num_pts*sizeof(double));
   if (binlo && binhi && xx)
   {
      int rr;
      for (rr = num_pts; rr--;)
      {
         if (bin_lo) binlo[rr] = bin_lo[rr]; 
         if (bin_hi) binhi[rr] = bin_hi[rr]; 
      }
   } 
   initialize();
   calculate_limits();
}


ChipsHistogram::ChipsHistogram(ChipsObject* par, ChipsStateStream& in):ChipsCurve(par, in)
{
   deserialize(in);
   // since we're deserializing all values we can skip calling
   // initialize() and calculate_limits()
   object_type = chips_histogram;
   initialized = true;
}


ChipsHistogram::ChipsHistogram(ChipsObject* par, unsigned long cnt, 
   double* bin_lo, double* bin_hi, double* bincnts, double* yd, double* yu, 
   string hid):
   ChipsCurve(par, cnt, bin_lo, bincnts, yd, yu, NULL, NULL, NULL, NULL, 
   "", hid, false) 
{
   initialized = false;
   binlo = (double*)malloc(num_pts*sizeof(double));
   binhi = (double*)malloc(num_pts*sizeof(double));
   if (!(lo_supplied = (bin_lo != NULL)))
      xx = (double*)malloc(num_pts*sizeof(double));
   hi_supplied = (bin_hi != NULL);
   if (binlo && binhi && xx)
   {
      int rr;
      if (lo_supplied && hi_supplied) 
         for (rr = num_pts; rr--;)
         {
            binlo[rr] = bin_lo[rr];
            binhi[rr] = bin_hi[rr];
         }
      else if (lo_supplied)
         for (rr = num_pts; rr--;)   binlo[rr] = bin_lo[rr];
      else if (hi_supplied)
         for (rr = num_pts; rr--;)   binhi[rr] = bin_hi[rr];
   }
   initialize();
   calculate_limits();
}


ChipsHistogram::~ChipsHistogram()
{
   if (binlo)
   {
      free(binlo); 
      binlo = NULL;
   }
   if (binhi)
   {
      free(binhi); 
      binhi = NULL;
   }
}

void ChipsHistogram::object_delete ()
{
   delete this;
}


void ChipsHistogram::initialize()
{
   int sz; 
   if (theDefaults)
   {
      string temp;

      symbolstyle = theDefaults->get_histogramsymbolstyle();
      sz = theDefaults->get_histogramsymbolsize();
      symbolangle = theDefaults->get_histogramsymbolangle();
      symbolfilled = theDefaults->histogramsymbol_filled();
      temp = theDefaults->get_histogramsymbolcolor();
      if (set_symbolcolor(temp) != 0)   set_symbolcolor("blue");

      linestyle   = theDefaults->get_histogramlinestyle();
      linethickness = theDefaults->get_histogramlinethickness();
      temp = theDefaults->get_histogramlinecolor();
      if (set_linecolor(temp) != 0)   set_linecolor("white");
      drop_line = theDefaults->get_histogramdropline(); 

      temp = theDefaults->get_histogramfillcolor();
      if (set_fillcolor(temp) != 0)   set_fillcolor("white");
      fillopacity = theDefaults->get_histogramfillopacity();
      fillstyle  = theDefaults->get_histogramfillstyle();

      errup = (yup && theDefaults->get_histogramerrup());
      errdown = (ydown && theDefaults->get_histogramerrdown());
      errstyle = theDefaults->get_histogramerrstyle();
      errthickness = theDefaults->get_histogramerrthickness();
      errcaplen = theDefaults->get_histogramerrcaplength(); 
      temp = theDefaults->get_histogramerrcolor();
      if (set_errcolor(temp) != 0)   set_errcolor("red");
      depth = theDefaults->get_histogramdepth();
      meta = theDefaults->get_meta();
   }
   else 
   {
      symbolstyle = chips_none;
      sz = 3; 
      symbolangle = 0.0;
      set_symbolcolor("cyan");
      symbolfilled = false; 
      linestyle = chips_solid;
      set_linecolor("cyan");
      linethickness = 1; 
      drop_line = false; 
      set_fillcolor ("cyan");
      fillopacity = 1.0;
      fillstyle = chips_solidfill; 
      errup = (yup != NULL);
      errdown = (ydown != NULL);
      errstyle = chips_bar;
      errthickness = 1;
      errcaplen = 10; 
      set_errcolor("red"); 
      depth = V_USE_DEFAULT_DEPTH;
      meta = "";
   }
   if (depth == V_USE_DEFAULT_DEPTH)  depth = ((ChipsPlot*)parent)->get_depth();
   display_mask = REG_INTERVAL;
   set_symbolsize(sz);
   force_drop = false; 
   object_type = chips_histogram;
   curve_num = 0;
   set_fillstyle(fillstyle);
   initialized = true;
}

void ChipsHistogram::calculate_limits()
{
   double tmp; 
   CHIPS_RANGE_P seg = NULL;
   int badval, rr; 
   if (!(lo_supplied || hi_supplied))
   {
      // set bin data - since no data was provided for the bin positions, 
      // simply set the bins uniformly with a value of 1 
      if (finite(yy[0]))
      { 
         if (ydown)
            ylmin = ymin = yy[0] - ydown[0];
         else 
            ylmin = ymin = yy[0];
         if (yup)
            ymax = yy[0] + yup[0];
         else
            ymax = yy[0];
      }
      else
      {
         ylmin = ymin = FLT_MAX;
         ymax = -1*FLT_MAX;
      }

      for (rr = 0; rr < num_pts; rr++)
      {
         binlo[rr] = rr + 0.5; 
         xx[rr]    = rr + 1; 
         binhi[rr] = rr + 1.5; 
         badval = REG_INTERVAL;

         // compute y range- include err bars if present 
         if (finite(yy[rr]))
         {
            if (ydown)
            {
               if (((tmp=(yy[rr]-ydown[rr])) > 0) && 
                   ((tmp < ylmin) || (ylmin <= 0)))
                  ylmin = tmp;
               if (tmp < ymin)  ymin = yy[rr] - ydown[rr];
            }
            else
            {
               if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin <= 0)))
                  ylmin = yy[rr];
               if (yy[rr] < ymin)  ymin = yy[rr];
            }
            if (yup)
            {
               if ((yy[rr]+yup[rr]) > ymax) ymax = yy[rr] + yup[rr];
            }
            else
            {
               if (yy[rr] > ymax)  ymax = yy[rr];
            }
         }
         else  badval = NAN_INTERVAL;

         if (seg)
         {
            if (seg->mask != badval)
            {
               if (badval == NAN_INTERVAL)
               {
                  seg->hi = rr - 1;
                  indices.push_back(seg);
                  seg = new CHIPS_RANGE;
                  seg->lo = rr - 1;
                  seg->mask = NAN_INTERVAL;
                  seg->inclusive = 0;
               }
               else
               {
                  seg->hi = rr;
                  indices.push_back(seg);
                  seg = new CHIPS_RANGE;
                  seg->lo = rr;
                  seg->mask = REG_INTERVAL;
                  seg->inclusive = 1;
               }
            }
         }
         else if (badval != NAN_INTERVAL)
         {
            seg = new CHIPS_RANGE;
            seg->lo = rr;
            seg->mask = REG_INTERVAL;
            seg->inclusive = 1;
         }
      }
      if (seg)
      {
         seg->hi = num_pts -1;
         indices.push_back(seg);
      }
      xmin = xlmin = 0.5;
      xmax = num_pts + 0.5;  
   }
   else if (lo_supplied && hi_supplied)
   {
      // both bins are set- calculate min/max
      double mn, mx; 

      if (finite(yy[0])) 
      {
         if (ydown)
            ylmin = ymin = yy[0] - ydown[0];
         else 
            ylmin = ymin = yy[0];
         if (yup)
            ymax = yy[0] + yup[0];
         else
            ymax = yy[0];
      }
      else
      {
         ylmin = ymin = FLT_MAX;
         ymax = -1*FLT_MAX;
      }
      if (finite(binlo[0]) && finite(binhi[0]))
      {
         xlmin = xmin = xmax = binlo[0]; 
         xx[0] = binlo[0] + (binhi[0] - binlo[0])/2.0;
         if (finite(yy[0]))
	 {
            seg = new CHIPS_RANGE;
            seg->lo = 0;
            seg->mask = REG_INTERVAL;
            seg->inclusive = 1;
	 }
      }
      else 
      {
         xlmin = xmin = FLT_MAX;
         xmax = -1*FLT_MAX;
      } 
      for (rr = 0; rr < num_pts; rr++)
      {
         badval = REG_INTERVAL;
         if (finite(binlo[rr]) && finite(binhi[rr]) && finite(yy[rr]))
         {
            // should we ensure that binlo < binhi?
            if (binlo[rr] > binhi[rr])
            {
               mn = binhi[rr];
               mx = binlo[rr]; 
            }
            else
            {
               mx = binhi[rr];
               mn = binlo[rr]; 
            }
            // set indicator if we need to force the drawing of 
            // drop lines- consecutive bins do not touch 
            if (rr > 0 && binlo[rr] != binhi[rr-1]) force_drop = true;

            // compute x range 
            xx[rr] = (mn + mx)/2;
            if (mn > 0)
            {
               if (xlmin < 0)   xlmin = mn;
                  else if (mn < xlmin)   xlmin = mn;
            }
            if (mn < xmin)  xmin = mn;
            if (mx > xmax)  xmax = mx;

            // compute y range- include err bars if present 
            if (ydown)
            {
               if (((tmp=(yy[rr]-ydown[rr])) > 0) && 
                   ((tmp < ylmin) || (ylmin <= 0)))
                  ylmin = tmp;
               if (tmp < ymin)  ymin = yy[rr] - ydown[rr];
            }
            else
            {
               if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin <= 0)))
                  ylmin = yy[rr];
               if (yy[rr] < ymin)  ymin = yy[rr];
            }
            if (yup)
            {
               if ((yy[rr] + yup[rr]) > ymax)  ymax = yy[rr] + yup[rr];
            }
            else
            {
               if (yy[rr] > ymax)  ymax = yy[rr];
            }
         } 
         else  badval = NAN_INTERVAL;

         if (seg && rr > 0)
         {
            if (seg->mask != badval)
            {
               if (badval == NAN_INTERVAL)
               {
                  seg->hi = rr - 1;
                  indices.push_back(seg);
                  seg = new CHIPS_RANGE;
                  seg->lo = rr - 1;
                  seg->mask = NAN_INTERVAL;
                  seg->inclusive = 0;
               }
               else
               {
                  seg->hi = rr;
                  indices.push_back(seg);
                  seg = new CHIPS_RANGE;
                  seg->lo = rr;
                  seg->mask = REG_INTERVAL;
                  seg->inclusive = 1;
               }
            }
         }
         else if (badval != NAN_INTERVAL)
         {
            seg = new CHIPS_RANGE;
            seg->lo = rr;
            seg->mask = REG_INTERVAL;
            seg->inclusive = 1;
         }
      }
      if (seg)
      {
         seg->hi = num_pts -1;
         indices.push_back(seg);
      }
   }
   else
   {
      double *tmpbin;
   
      if (lo_supplied)
      {
         tmpbin = binlo;  
         binlo = (double*)malloc(num_pts*sizeof(double));
      }
      else
      {
         tmpbin = binhi;
         binhi = (double*)malloc(num_pts*sizeof(double));
      }
      if (binlo && binhi && tmpbin)
      {
         double dist, mn, mx;
         for (rr = 0; rr < num_pts-1; rr++)
         {
            if (finite(tmpbin[rr+1]) && finite(tmpbin[rr])) 
            {
               dist = (tmpbin[rr+1] - tmpbin[rr])/2.0;
               binhi[rr] = binlo[rr+1] = tmpbin[rr] + dist; 
            }
            else if (finite(tmpbin[rr+1])) 
	    {
	       if ((rr < num_pts-2) && finite(tmpbin[rr+2]))
	          binhi[rr] = binlo[rr+1] = tmpbin[rr+1] - 
		     (tmpbin[rr+2] - tmpbin[rr+1])/2.0;
	       else
                  binhi[rr] = binlo[rr+1] = tmpbin[rr+1]; 
	    }
            else
	    {
	       if ((rr > 0) && finite(tmpbin[rr-1]))
	          binhi[rr] = binlo[rr+1] = tmpbin[rr] + 
		     (tmpbin[rr] - tmpbin[rr-1])/2.0;
	       else
                  binhi[rr] = binlo[rr+1] = tmpbin[rr]; 
            }
         }
         if (finite(tmpbin[0]) && finite(tmpbin[1]))
         {
            binlo[0] = tmpbin[0] - (tmpbin[1] - tmpbin[0])/ 2.0;  
         }
         else if (finite(tmpbin[0]))
            binlo[0] = tmpbin[0]; 
         else 
            binlo[0] = tmpbin[1]; 
         if (finite(tmpbin[rr]))
            binhi[rr] = tmpbin[rr] + dist; 
         else 
            binhi[rr] = tmpbin[rr]; 

         if (finite(yy[0]))
         {
            if (ydown)
               ylmin = ymin = yy[0] - ydown[0];
            else 
               ylmin = ymin = yy[0];
            if (yup)
               ymax = yy[0] + yup[0];
            else
               ymax = yy[0];
         }
         else
         {
            ylmin = ymin = FLT_MAX;
            ymax = -1*FLT_MAX;
         }

         xlmin = xmin = FLT_MAX;
         xmax = -1*FLT_MAX;
         for (rr = 0; rr < num_pts; rr++)
         {
            badval = REG_INTERVAL;

            // should we ensure that binlo < binhi?
            if (finite(binlo[rr]) && finite(binhi[rr]) && finite(yy[rr]) &&
	        finite(tmpbin[rr]))
            {
               if (binlo[rr] > binhi[rr])
               {
                  mn = binhi[rr];
                  mx = binlo[rr];
               }
               else
               {
                  mx = binhi[rr];
                  mn = binlo[rr];
               }
   
               // set symbol position 
               xx[rr] = tmpbin[rr];

               if (mn > 0)
               {
                  if (xlmin < 0)   xlmin = mn;
                     else if (mn < xlmin)   xlmin = mn;
               }
               if (mn < xmin)  xmin = mn;
               if (mx > xmax)  xmax = mx;

               // compute y range- include err bars if present 
               if (ydown)
               {
                  if (((tmp=(yy[rr]-ydown[rr])) > 0) && 
                      ((tmp < ylmin) || (ylmin <= 0)))
                     ylmin = tmp;
                  if (tmp < ymin)  ymin = yy[rr] - ydown[rr];
               }
               else
               {
                  if ((yy[rr] > 0) && ((yy[rr] < ylmin) || (ylmin <= 0)))
                     ylmin = yy[rr];
                  if (yy[rr] < ymin)  ymin = yy[rr];
               }
               if (yup)
               {
                  if ((yy[rr] + yup[rr]) > ymax)  ymax = yy[rr] + yup[rr];
               }
               else
               {
                  if (yy[rr] > ymax)  ymax = yy[rr];
               }
            }
            else  badval = NAN_INTERVAL;

            if (seg)
            {
               if (seg->mask != badval)
               {
                  if (badval == NAN_INTERVAL)
                  {
                     seg->hi = rr - 1;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr - 1;
                     seg->mask = NAN_INTERVAL;
                     seg->inclusive = 0;
                  }
                  else
                  {
                     seg->hi = rr;
                     indices.push_back(seg);
                     seg = new CHIPS_RANGE;
                     seg->lo = rr;
                     seg->mask = REG_INTERVAL;
                     seg->inclusive = 1;
                  }
               }
            }
            else if (badval != NAN_INTERVAL)
            {
               seg = new CHIPS_RANGE;
               seg->lo = rr;
               seg->mask = REG_INTERVAL;
               seg->inclusive = 1;
            }
         }
         if (seg)
         {
            seg->hi = num_pts -1;
            indices.push_back(seg);
         }
         free(tmpbin);
      }
   }
}


// This recalculates the ranges after a coordinate system is updated
int ChipsHistogram::reset_position ()
{
   int rval = ERR_STATUS_GOOD;
   ChipsDataSystem *data_sys;
   if (!coord || (coord_system != DATA))
      return ERR_STATUS_BAD;

   data_sys = static_cast <ChipsDataSystem *>(coord);

   rval = change_scale (data_sys->get_log_flag (X_AXIS), data_sys->get_log_flag (Y_AXIS));
   if (!rval)
      change_range (data_sys->get_lower_bounds (), data_sys->get_upper_bounds ());

   return rval;
}


int ChipsHistogram::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}



int ChipsHistogram::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   int mask = theDefaults->get_allpropmask();

   if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
   {
      string clr = theDefaults->get_allcolor();
      rval = set_property((void*)clr.c_str(), SET_HISTO_SYMBOL_COLOR);
      rval |= set_property((void*)clr.c_str(), SET_HISTO_LINE_COLOR);
      rval |= set_property((void*)clr.c_str(), SET_HISTO_FILL_COLOR);
      rval |= set_property((void*)clr.c_str(), SET_HISTO_ERR_COLOR);
   }
   if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
   {
      float thick = theDefaults->get_allthickness();
      rval = set_property(static_cast<void*>(&thick), SET_HISTO_LINE_THICKNESS);
      rval |= set_property(static_cast<void*>(&thick), SET_HISTO_ERR_THICKNESS);
   }
   if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
   {
      int vis = theDefaults->get_allvisible();
      rval = set_property(static_cast<void*>(&vis), SET_HISTO_VISIBLE);
   }
   if ((mask & PREF_ALL_FILL) == PREF_ALL_FILL)
   {
      bool bval = theDefaults->get_allfill(); 
      rval = set_property(static_cast<void*>(&bval), SET_HISTO_SYMBOL_FILL);
   }
   if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
   {
      int dep = theDefaults->get_alldepth();
      rval = set_property(static_cast<void*>(&dep), SET_HISTO_DEPTH);
   }
   return rval;
}


int ChipsHistogram::set_property(void* in_val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   string save_current;
   if (parent && (parent->get_object_type() == chips_plot))
      save_current = ((ChipsPlot *)parent)->get_current_histogram(); 

   switch(property)
   {
      case SET_HISTO_SYMBOL_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)symbolstyle;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_symbolstyle((V_SYMBOL_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_SYMBOL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            symbolangle, *fval, save_current);
         if ((rval = set_symbolangle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_LINE_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)linestyle;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_linestyle((V_LINE_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_FILL_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)fillstyle;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((rval = set_fillstyle((V_FILL_TYPE)*ival, rec)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
         
      }
      break;
      case SET_HISTO_FILL_STYLE2:
      {
         unsigned char* uc = (unsigned char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
	              fillpattern, uc, 128,
                      ((ChipsPlot *)parent)->get_current_histogram());
         if ((rval = set_fillstyle(uc)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
	 else  delete cmd;
      }
      break;
      case SET_HISTO_SYMBOL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_symbolcolor(), clr, save_current);
         if ((rval = set_symbolcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_SYMBOL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            symbolsize_i, *ival, save_current);
         if ((rval = set_symbolsize(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_SYMBOL_FILL:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              symbolfilled, *bval, save_current);
         if ((rval = set_symbolfill(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_HISTO_LINE_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_linecolor(), clr, save_current);
         if ((rval = set_linecolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_LINE_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            linethickness, *fval, save_current);
         if ((rval = set_linethickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
     case SET_HISTO_ERR_CAPLEN:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            errcaplen, *ival, save_current);
         if ((rval = set_errcaplength(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_ERR_UP:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errup, *bval, save_current);
         if ((rval = set_errup(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_HISTO_ERR_DOWN:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              errdown, *bval, save_current);
         if ((rval = set_errdown(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_HISTO_DROPLINE:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
              drop_line, *bval, save_current);
         if ((rval = set_dropline(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
            delete cmd;
      }
      break;
      case SET_HISTO_ERR_STYLE:
      {
         string val = (const char*)in_val;
         string oval = v_errstyle_to_string(errstyle); 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, val, save_current);
         if ((rval = set_errstyle(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_ERR_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_errcolor(), clr, save_current);
         if ((rval = set_errcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_ERR_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            errthickness, *fval, save_current);
         if ((rval = set_errthickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "curve visibility", *ival, 0, 1);
            delete cmd;
         }
      }
      break;
      case SET_HISTO_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this,property,oval,newval,save_current);
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      case SET_HISTO_FILL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_fillcolor(), clr, save_current);
         if ((rval = set_fillcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_HISTO_FILL_OPACITY:
      {
         double* dval = static_cast<double *>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            get_fillopacity (), *dval, save_current);
         if ((rval = set_fillopacity(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;

  
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_HSTG);

   return rval;
}


int ChipsHistogram::get_counts_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && yy)
      {
         *cnt = num_pts; 
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory 
            for (ii = num_pts; ii--; )  dat[ii] = yy[ii]; 
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval; 
}


int ChipsHistogram::get_binlo_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && binlo)
      {
         *cnt = num_pts; 
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = binlo[ii];
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0; 
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval; 
}


int ChipsHistogram::get_binhi_data(double* dat, unsigned long* cnt)
{
   int rval = ERR_STATUS_GOOD;

   if (dat && cnt)
   {
      if (num_pts && binhi)
      {
         *cnt = num_pts;
         dat = (double*) malloc((num_pts)*sizeof(double));
         if (dat) 
         {
            unsigned long ii;

            // copy data into memory
            for (ii = num_pts; ii--; )  dat[ii] = binhi[ii];
         }
         else  cmd_manager->add_err(rval=ERR_ALLOC_FAIL, id, 0);
      }
      else
      {
         dat = NULL;
         *cnt = 0;
      }
   } else  cmd_manager->add_err(rval=ERR_RETURN_PTR_NULL, id, 0);

   return rval;
}


int ChipsHistogram::set_dropline(bool val)
{
   drop_line = val;
   if (initialized) change_dropline();
   return ERR_STATUS_GOOD;
}


int ChipsHistogram::set_fillcolor (string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = fillcolor.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  change_fillcolor();
   return rval;
}


int ChipsHistogram::set_fillopacity(double in_val)
{
   int rval = ERR_STATUS_GOOD;
   if ((in_val < 0.0) || (in_val > 1.0))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "opacity", in_val,
                                  0.0, 1.0);
   else
   {
      fillopacity = in_val;
      if (initialized)  rval = change_fillopacity();
   }
   return rval;
}


int ChipsHistogram::set_fillstyle(V_FILL_TYPE val, bool record)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_fill_type(val))
   {
      record = record && ((val >= chips_userfill1 && val <= chips_userfill3) || (fillstyle >= chips_userfill1 && fillstyle <= chips_userfill2));
      fillstyle = val;
      set_property((void*)fill_patterns->get_pattern(fillstyle), SET_HISTO_FILL_STYLE2,record);
      if (initialized)  change_fillstyle();
   }
   else
      cmd_manager->add_err(rval=ERR_BAD_FILLSTYLE2, id, 0, (int)val);
   return rval;
}


int ChipsHistogram::set_fillstyle(unsigned char* bpattern)
{
   int rval = ERR_STATUS_GOOD;

   if (bpattern)
   {
      memcpy(fillpattern, bpattern, 128);
      if (initialized)  change_fillstyle();
   }
   else
      rval = ERR_BAD_FILLSTYLE;
   return rval;
}


int ChipsHistogram::change_linecolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_linethickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_linestyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_symbolcolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_symbolsize()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_symbolfilled()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_symbolstyle()
{
   return ERR_FEATURE_DNE;
}


int ChipsHistogram::change_dropline()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_fillcolor ()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_fillopacity ()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_fillstyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::set_plot_location (double x1, double y1, double x2, double y2)
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_range (const ChipsVector &low_range, const ChipsVector&high_range)
{
   return ERR_FEATURE_DNE;
}

int ChipsHistogram::change_scale (int inXScale, int inYScale)
{
   return ERR_FEATURE_DNE;
}

void ChipsHistogram::serialize(ChipsStateStream& out)
{
   string cname;

   ChipsCurve::serialize(out);

   out.write_bool(&force_drop);
   out.write_bool(&drop_line);
   out.write_bool(&lo_supplied);
   out.write_bool(&hi_supplied);
   out.write_int((int*)&num_pts);
   out.write_double(binlo, (int)num_pts);
   out.write_double(binhi, (int)num_pts);
   out.write_double(&ylmin);
   out.write_double(&ymin);
   out.write_double(&ymax);
   out.write_double(&xlmin);
   out.write_double(&xmin);
   out.write_double(&xmax);
   bool bb = 0; 
   cname = get_fillcolor ();
   out.write_string (&cname);
   out.write_bool (&bb);
   out.write_double (&fillopacity);
   int sz = (int)fillstyle;
   out.write_int(&sz);
   out.write_char((char*)fillpattern, 128);
}

void ChipsHistogram::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];

   tmp = get_linecolor();
   if (theDefaults->get_histogramlinecolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("line.color=");
      att.append(tmp);
   }
   tmp = get_fillcolor();
   if (theDefaults->get_histogramfillcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("fill.color=");
      att.append(tmp);
   }
   if (theDefaults->get_histogramlinestyle() != linestyle)
   {
      if (att.length()) att.append(" ");
      att.append("line.style=");
      att.append(v_line_to_string(linestyle));
   }
   if (theDefaults->get_histogramfillstyle() != fillstyle)
   {
      if (att.length()) att.append(" ");
      att.append("fill.style=");
      att.append(v_fill_to_string(fillstyle));
   }
   if (theDefaults->get_histogramlinethickness() != linethickness)
   {
      if (att.length()) att.append(" ");
      if (linethickness != 0.5)
      {
         sprintf(buf,"line.thickness=%-1.0f", linethickness);
         att.append(buf);
      }
      else
         att.append("line.thickness=0.5");
   }
   tmp = get_symbolcolor();
   if (theDefaults->get_histogramsymbolcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("symbol.color=");
      att.append(tmp);
   }
   if (theDefaults->get_histogramsymbolstyle() != symbolstyle)
   {
      if (att.length()) att.append(" ");
      att.append("symbol.style=");
      att.append(v_symbol_to_string(symbolstyle));
   }
   if (theDefaults->get_histogramsymbolsize() != symbolsize_i)
   {
      if (att.length()) att.append(" ");
         sprintf(buf,"symbol.size=%d", symbolsize_i);
         att.append(buf);
   }
   if (symbolfilled != theDefaults->histogramsymbol_filled())
   {
      if (att.length()) att.append(" ");
      if (symbolfilled)  att.append("symbol.fill=true");
         else            att.append("symbol.fill=false");
   }
   if (symbolangle != theDefaults->get_histogramsymbolangle())
   {
      if (att.length()) att.append(" ");
      sprintf(buf,"symbol.angle=%2.1f", symbolangle);
      att.append(buf);
   }
   if (drop_line != theDefaults->get_histogramdropline())
   {
      if (att.length()) att.append(" ");
      if (drop_line)   att.append("dropline=on");
         else          att.append("dropline=off");
   }
   if (yup)
   {
      if (errup != theDefaults->get_histogramerrup())
      {
         if (att.length()) att.append(" ");
         if (errup) att.append("err.up=true");
            else    att.append("err.up=false");
      }
   }
   if (ydown)
   {
      if (errdown != theDefaults->get_histogramerrdown())
      {
         if (att.length()) att.append(" ");
         if (errdown) att.append("err.down=true");
            else    att.append("err.down=false");
      }
   }
   tmp = get_errcolor();
   if ((yup || ydown) && theDefaults->get_histogramerrcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("err.color=");
      att.append(tmp);
   }
   if (theDefaults->get_histogramfillopacity() != fillopacity)
   {
      if (att.length()) att.append(" ");
      sprintf(buf, "fill.opacity=%2.1f", fillopacity);
      att.append(buf);
   }
   int dep = theDefaults->get_histogramdepth();
   if (depth != dep &&
       (dep != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   string dfname = out.use_tempname();
   string mfile = "";

   if (meta.length())
   {
      int quoted = (meta.find("'") == 0 && meta.rfind("'") == meta.length()-1);
      mfile = dfname;
      mfile.append("=");
      if (!quoted)  mfile.append("\"");
      mfile.append(meta);
      if (!quoted)  mfile.append("\"");
      out << "add_histogram("<<dfname;
   }
   else
   {
      string tmpfile = out.get_filename();
      sprintf(buf, ".Data%d_histogram.bin", out.use_tempfnum());
      tmpfile.append(buf);
      mfile = dfname;
      mfile.append(" = DataObj()  # data- ");
      mfile.append(tmpfile);
      mfile.append("\n");
      if (binlo)
      {
         mfile.append(dfname);
         mfile.append(".binlo = #  insert data array here \n");
      }
      if (binhi)
      {
         mfile.append(dfname);
         mfile.append(".binhi = #  insert data array here \n");
      }
      if (yy)
      {
         mfile.append(dfname);
         mfile.append(".cnts  = #  insert data array here \n");
      }
      if (ydown)
      {
         mfile.append(dfname);
         mfile.append(".ydown = #  insert data array here \n");
      }
      if (yup)
      {
         mfile.append(dfname);
         mfile.append(".yup   = #  insert data array here \n");
      }

      ofstream ff(tmpfile.c_str(), ios::out|ios::binary);
      if (ff)
      {
         // table data format:
         //  [type (0= table) ][ numcols         ]
         //  [number of rows  ][ colname buff size]
         //  [ char array '\0' separated col names ]
         //  [ double array col1 data .. numcols]
         char buffer[1024];
         long bpos = 0, ncols = 0;
         string colnames = "";
         ff.write((char*)&ncols, sizeof(long));               // table format
         if (binlo) { strcpy(&buffer[bpos], "binlo\0"); bpos+=6; ncols++; }
         if (binhi) { strcpy(&buffer[bpos], "binhi\0"); bpos+=6; ncols++; }
         if (yy)    { strcpy(&buffer[bpos], "cnts\0"); bpos+=5; ncols++; }
         if (yup)   { strcpy(&buffer[bpos], "yup\0"); bpos+=4; ncols++; }
         if (ydown) { strcpy(&buffer[bpos], "ydown\0"); bpos+=6; ncols++; }
         // number of doubles in buffer- add 1 less than sizeof double for
         // double word aligning and subtract 1 for extra \0 insertion  
         bpos = (bpos + sizeof(double) - 2) / sizeof(double);
         ff.write((char*)&ncols, sizeof(long));               // num cols
         ff.write((char*)&num_pts, sizeof(unsigned long));    // num rows
         ff.write((char*)&bpos, sizeof(long));                // colname sz
         ff.write((char*)buffer, bpos* sizeof(double));       // colnames

         if (binlo) ff.write((char*)binlo, num_pts* sizeof(double));
         if (binhi) ff.write((char*)binhi, num_pts* sizeof(double));
         if (yy)    ff.write((char*)yy, num_pts* sizeof(double));
         if (yup)   ff.write((char*)yup, num_pts* sizeof(double));
         if (ydown) ff.write((char*)ydown, num_pts* sizeof(double));
         ff.close();
      }

      out << "add_histogram("<<dfname<<".binlo, "<<dfname<<".binhi, "<<dfname<<".cnts";
      if (yup && ydown) out <<","<<dfname<<".ydown,"<<dfname<<".yup";
      else if (ydown)   out <<","<<dfname<<".ydown,None";
      else if (yup)     out <<",None,"<<dfname<<".yup";
   }
   if (att.length()) out <<",'"<<att<<"'";
   out <<")"<<endl;
   out.append_to_template(mfile);
}

void ChipsHistogram::deserialize(ChipsStateStream& in)
{
   string cname;

   in.read_bool(&force_drop);
   in.read_bool(&drop_line);
   in.read_bool(&lo_supplied);
   in.read_bool(&hi_supplied);
   in.read_int((int*)&num_pts);
   binlo = (double*)malloc(num_pts*sizeof(double));
   binhi = (double*)malloc(num_pts*sizeof(double));
   in.read_double(binlo, (int)num_pts);
   in.read_double(binhi, (int)num_pts);
   in.read_double(&ylmin);
   in.read_double(&ymin);
   in.read_double(&ymax);
   in.read_double(&xlmin);
   in.read_double(&xmin);
   in.read_double(&xmax);
   bool bb; 
   in.read_string (&cname);
   set_fillcolor (cname);
   in.read_bool (&bb);
   in.read_double (&fillopacity);
   int len;
   in.read_int(&len);
   fillstyle = (V_FILL_TYPE)len;
   in.read_char((char*)fillpattern, 128);
}

int ChipsHistogram::show()
{
  int rval = ERR_STATUS_GOOD;
  visible = true;
  rval = change_visible ();
  return rval;
};

int ChipsHistogram::hide()
{
  int rval = ERR_STATUS_GOOD;
  visible = false;
  rval = change_visible ();
  return rval;
};


bool ChipsHistogram::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED) outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}



