// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsId.hh"
#include "ChipsImage.hh"
#include "ChipsCommand.hh"
#include "ChipsPlot.hh"
#include "ChipsDataSystem.hh"
#include "ChipsCmdManager.hh"
#include "ChipsDefines.hh"
#include "ChipsColorMapTable.hh"
#include "ChipsImageChannelCommand.hh"
#include "ChipsTransformFactory.hh"

#include <sstream>
#include <stdio.h>

extern ChipsColorMapTable *colormap_table;
extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;

ChipsImage::ChipsImage(ChipsObject* par, string filename, string transfrm_id,
   string val):ChipsObject(par, val)
{
/*   id = val; 
   initialize(true); 
   fname = filename; 

   colormap_id = colormap_table->add_colormap ();
*/
}


ChipsImage::ChipsImage(ChipsObject* par, string filename, 
   float x1n, float y1n, float x2n, float y2n, string transfrm_id,
   string val): ChipsObject(par, val)
{
/*   id = val; 
   initialize(true); 
   nx1 = x1n;
   ny1 = y1n;
   nx2 = x2n;
   ny2 = y2n;
   if (nx1 < 0.0) nx1 = 0.0;
   if (nx1 > 1.0) nx1 = 1.0;
   if (ny1 < 0.0) ny1 = 0.0;
   if (ny1 > 1.0) ny1 = 1.0;
   if (nx2 < 0.0) nx2 = 0.0;
   if (nx2 > 1.0) nx2 = 1.0;
   if (ny2 < 0.0) ny2 = 0.0;
   if (ny2 > 1.0) ny2 = 1.0;
   fname = filename; 
*/
}

ChipsImage::ChipsImage(ChipsObject* par, int xaxis, int yaxis, 
   double* rvals, double* gvals, double *bvals, double *avals,
   string transfrm_id, string val): ChipsObject(par, val)
{
   id = val;
   initialize(true);
   fname = "";
   dim[0] = xaxis;
   dim[1] = yaxis;

   unrotated_xmin = xmin = .5;
   unrotated_xmax = xmax = dim[0]+.5;
   unrotated_ymin = ymin = .5;
   unrotated_ymax = ymax = dim[1]+.5;

   position_set = false;
   transform_id = transfrm_id;

}

ChipsImage::ChipsImage(ChipsObject* par, ChipsStateStream& in):ChipsObject(par, "")
{
   initialized = false;
   position_set = false;
   initialize(false);
   // deserialize run by derived class
}


void ChipsImage::initialize(bool load_colormap)
{
   nx1 = ny1 = 0.0;
   nx2 = ny2 = 1.0;
   current = ""; 
   object_type = chips_image;
   flipped_x = false;
   flipped_y = false;
   rotation = 0.0;

   print_width = 0;
   print_height = 0;

   if (load_colormap)
   {
      colormap_id = colormap_table->add_colormap ();
      colormap_table->set_invert_colormap (colormap_id, false);
   }
   else
      colormap_id = (unsigned short)-1;


   if (theDefaults)
   {
      depth = theDefaults->get_image_depth();
      interpolation = theDefaults->get_image_interpolation ();
      normalize_channels = theDefaults->get_image_scale_channels ();
      if (load_colormap)
      {
         double alpha_range[2];
         double threshold[2];
         colormap_table->set_colormap_type (colormap_id, theDefaults->get_image_colormap());
         
         alpha_range[0] = theDefaults->get_image_alpha_lower ();
         alpha_range[1] = theDefaults->get_image_alpha_upper ();
         if (alpha_range[0] != -1)
            colormap_table->set_colormap_alpha_range (colormap_id, alpha_range);

         threshold[0] = theDefaults->get_image_threshold_lower ();
         threshold[1] = theDefaults->get_image_threshold_upper ();
         if (threshold[0] || threshold[1])
            colormap_table->set_colormap_scalar_threshold (colormap_id, threshold);

         colormap_table->set_invert_colormap(colormap_id, theDefaults->get_image_invert_colormap ());
         colormap_table->set_colormap_size (colormap_id, theDefaults->get_image_colormap_size ());
         colormap_table->set_colormap_interpolate (colormap_id, theDefaults->get_image_colormap_interpolate ());
      }
      meta = theDefaults->get_meta();
   }
   else
   {
      depth = 100;
      if (load_colormap)
         colormap_table->set_colormap_type(colormap_id,
                                           ((V_COLORMAP_TYPE)chips_grayscale));
      interpolation = chips_neighbor; 
      normalize_channels = false;
      meta = "";
   }
   if (depth==V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
   initialized = 1; 
}

ChipsImage::~ChipsImage()
{
   if (coord)
      coord->remove(this);
}

void ChipsImage::object_delete ()
{
   delete this;
}


int ChipsImage::reset_position ()
{
   return calculate_position (1);
}

int ChipsImage::calculate_position (double scale)
{
   int rval = ERR_STATUS_GOOD;

   double scale_x = 1;
   double scale_y = 1;
   double extent_scale_x = 1;
   double extent_scale_y = 1;

   int extent_x = 0;
   int extent_y = 0;

   bool reverse_x = false;
   bool reverse_y = false;

   ChipsDataSystem *data_sys;
   ChipsVector lower_bounds;
   ChipsVector upper_bounds;

   ChipsVector scale_lower_bounds;
   ChipsVector scale_upper_bounds;

   ChipsVector tmp_lower_bounds(3);
   ChipsVector tmp_upper_bounds(3);

   ChipsMatrix rotation_matrix (3,3);

   string data_sys_id;

   bool clipped = false;

   if (!coord || (coord_system != DATA))
      return ERR_STATUS_BAD;

   data_sys = static_cast <ChipsDataSystem *>(coord);
   data_sys_id = data_sys->get_id_no_wcs ();

   if (data_sys->get_log_flag (X_AXIS) || data_sys->get_log_flag(Y_AXIS))
      return ERR_STATUS_BAD;

   // calculate the position in data coordinates
   if (!position_set)
   {
      this->reproject_image ();
      position_set = true;
   }


   // position the image with respect to the plot
   lower_bounds = data_sys->get_lower_bounds ();
   upper_bounds = data_sys->get_upper_bounds ();

   // reverse the axis if necessary
   reverse_x = lower_bounds[0] > upper_bounds[0];
   reverse_y = lower_bounds[1] > upper_bounds[1];
   if (reverse_x && !flipped_x)
   {
      flip_axis (X_AXIS);
      flipped_x = true;
   }
   else if (!reverse_x && flipped_x)
   {
      flip_axis (X_AXIS);
      flipped_x = false;
   }

   if (reverse_y && !flipped_y)
   {
      flip_axis (Y_AXIS);
      flipped_y = true;
   }
   else if (!reverse_y && flipped_y)
   {
      flip_axis (Y_AXIS);
      flipped_y = false;
   }


   if (reverse_x)
   {
      // if the lower bound is less than zero than the entire image is clipped
      if ((lower_bounds[0] < xmin) || (upper_bounds[0] > xmax))
      {
         clip_image (true);
         clipped = true;
      }
      else 
      {
         if (lower_bounds[0] > xmax)
            lower_bounds[0] = xmax;
         if (upper_bounds[0] < xmin)
            upper_bounds[0] = xmin;
      }
   }
   else
   {
      if ((upper_bounds[0] < xmin) || (lower_bounds[0] > xmax))
      {
         clip_image (true);
         clipped = true;
      }
      else
      {
         if (lower_bounds[0] < xmin)
            lower_bounds[0] = xmin;
         if (upper_bounds[0] > xmax)
            upper_bounds[0] = xmax;
      }

   }


   if (reverse_y)
   {
      // if the lower bound is less than zero than the entire image is clipped
      if ((lower_bounds[1] < this->ymin) || (upper_bounds[1] > this->ymax))
      {
         clip_image (true);
         clipped = true;
      }
      else
      {
         if (lower_bounds[1] > this->ymax)
            lower_bounds[1] = this->ymax;
         if (upper_bounds[1] < this->ymin)
            upper_bounds[1] = this->ymin;
      }
   }
   else
   {
      // if the lower bound is less than zero than the entire image is clipped
      if ((upper_bounds[1] < this->ymin) || (lower_bounds[1] > this->ymax))
      {
         clip_image (true);
         clipped = true;
      }
      else
      {
         if (lower_bounds[1] < this->ymin)
            lower_bounds[1] = this->ymin;
         if (upper_bounds[1] > this->ymax)
            upper_bounds[1] = this->ymax; 
      }
   }


   // if the entire image is clipped there's no point in
   // calculating any of the positions
   if (!clipped)
   {

      data_sys->convert_cs_to_cs (&lower_bounds, &tmp_lower_bounds, 1, DATA, data_sys_id, FRAME_NORM, "");

      change_position (tmp_lower_bounds[0], tmp_lower_bounds[1]);


      // scale the image
      // scale the image
      scale_lower_bounds = ChipsVector (this->unrotated_xmin,this->unrotated_ymin);
      scale_upper_bounds = ChipsVector (this->unrotated_xmax,this->unrotated_ymax);

      data_sys->convert_cs_to_cs (&scale_lower_bounds, &tmp_lower_bounds, 1, DATA, data_sys_id, PIXEL_REAL, "");
      data_sys->convert_cs_to_cs (&scale_upper_bounds, &tmp_upper_bounds, 1, DATA, data_sys_id, PIXEL_REAL, "");

      scale_x = scale * fabs(tmp_upper_bounds[0] - tmp_lower_bounds[0]) / dim[0];
      scale_y = scale * fabs(tmp_upper_bounds[1] - tmp_lower_bounds[1]) / dim[1];

      change_scale (scale_x, scale_y);

      // change the extent to clip the image to the limits

      scale_lower_bounds = ChipsVector (this->xmin,this->ymin);
      scale_upper_bounds = ChipsVector (this->xmax,this->ymax);

      data_sys->convert_cs_to_cs (&scale_lower_bounds, &tmp_lower_bounds, 1, DATA, data_sys_id, PIXEL_REAL, "");
      data_sys->convert_cs_to_cs (&scale_upper_bounds, &tmp_upper_bounds, 1, DATA, data_sys_id, PIXEL_REAL, "");

      extent_scale_x = scale * fabs(tmp_upper_bounds[0] - tmp_lower_bounds[0]) / (this->xmax-this->xmin);
      extent_scale_y = scale * fabs(tmp_upper_bounds[1] - tmp_lower_bounds[1]) / (this->ymax-this->ymin);


      tmp_lower_bounds = lower_bounds;

      if (reverse_x)
         tmp_lower_bounds[0] = this->xmax - lower_bounds[0] + 0.5;
      if (reverse_y)
         tmp_lower_bounds[1] = this->ymax - lower_bounds[1] + 0.5;

      extent_x = (int)(fabs(upper_bounds[0]-lower_bounds[0])*extent_scale_x + .5);
      extent_y = (int)(fabs(upper_bounds[1]-lower_bounds[1])*extent_scale_y + .5);

      if ((extent_x <= 0) || (extent_y <= 0))
         clip_image (true);
      else
      { 
         clip_image (false);
         change_origin (tmp_lower_bounds[0], tmp_lower_bounds[1], extent_scale_x, extent_scale_y);
         change_extent (0,0, extent_x, extent_y);
      }

      if (rotation != 0) 
         change_rotation (rotation);
   }

   return rval;
}


int ChipsImage::change_visible()
{
   return ERR_FEATURE_DNE;
}


int ChipsImage::show ()
{
  visible = true;
  return change_visible ();
}

int ChipsImage::hide ()
{
  visible = false;
  return change_visible ();
}

void ChipsImage::serialize(ChipsStateStream& out)
{
   string tmp;
   int sz;
   short sval;
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   out.write_int(&type); 
   out.write_int(dim, 3); 
   sz = dim[0] * dim[1];
   out.write_string(&current);
   out.write_string(&fname);
   out.write_string(&data_system_id);
   sz = (int)coord_system;
   out.write_int (&sz);

   out.write_string (&transform_id);

   sval = static_cast<short>(colormap_id);
   out.write_short (&sval);
   out.write_int(&interpolation);
   out.write_bool (&visible);

   out.write_int (&print_width);
   out.write_int (&print_height);

   out.write_double (&xmin);
   out.write_double (&xmax);
   out.write_double (&ymin);
   out.write_double (&ymax);

   out.write_double (&unrotated_xmin);
   out.write_double (&unrotated_xmax);
   out.write_double (&unrotated_ymin);
   out.write_double (&unrotated_ymax);

}

void ChipsImage::serialize_ascii(ChipsStateStream& out)
{
   string tmp;
   stringstream att;
   stringstream buf;

   att.precision (5);

   V_COLORMAP_TYPE ic = colormap_table->get_colormap_type(colormap_id);
   if (ic != theDefaults->get_image_colormap())
   {
      const char *val = v_colormap_to_string(ic);
      if (val)
         att << "colormap=" << val;
   }
   if (interpolation != theDefaults->get_image_interpolation())
   {
      const char * ic = v_interpolation_to_string((V_INTERPOLATION_TYPE)interpolation);
      if (ic)
      {
         if (att.str().length()) att << " ";
         att << "interpolation=" << ic;
      }
   }
   // add image depth if not default
   int dep = theDefaults->get_image_depth();
   if (depth != dep &&
       (dep!=V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      if (att.str().length()) att << " ";
      att <<"depth="<<depth;
   }

   if (colormap_table && (get_mode () == 0))
   {
      double thresh[2];
      colormap_table->get_colormap_scalar_threshold (colormap_id, thresh);

      if (thresh[0] != theDefaults->get_image_threshold_lower() ||
          thresh[1] != theDefaults->get_image_threshold_upper())
      {
         if (att.str().length()) att << " ";
         att << "threshold=["<<thresh[0]<<","<<thresh[1]<<"]";
      }

      double alpha[2];
      colormap_table->get_colormap_alpha_range (colormap_id, alpha);

      if (alpha[0] != theDefaults->get_image_alpha_lower() ||
          alpha[1] != theDefaults->get_image_alpha_upper())
      {
         
         if (att.str().length()) att << " ";
         att << "alpha=["<<alpha[0]<<","<<alpha[1] <<"]"; 
      }
      
      bool invert_colormap = get_invert_colormap ();
      if (invert_colormap != theDefaults->get_image_invert_colormap())
      {
         if (att.str().length()) att << " ";
         if (invert_colormap) att << "invert_colormap=true";
            else att << "invert_colormap=false";
      }

      int colormap_size = get_colormap_size ();
      if (colormap_size != theDefaults->get_image_colormap_size())
      {
         if (att.str().length()) att << " ";
         att << "colormap_size=" << colormap_size;
      }

      bool interpolate_colormap = get_colormap_interpolate ();
      if (interpolate_colormap != theDefaults->get_image_colormap_interpolate())
      {
         if (att.str().length()) att << " ";
         if (interpolate_colormap) att << "interpolate_colormap=true";
            else att << "interpolate_colormap=false";
      }

   }
   else if (get_mode () == 1)
   {
      bool scale_channels = get_normalize_channels ();
      if (scale_channels != theDefaults->get_image_scale_channels())
      {
         if (att.str().length()) att << " ";
         if (scale_channels) att <<"scale_channels=true";
            else att << "scale_channels=false";
      }

   }

   // add image id
   if (att.str().length()) att << " ";
   att << "id=" << id;

   string dfname = out.use_tempname();
   string mfile = "";
   string dataobject_desc = "";
   string transform_desc = "";

   if (meta.length())
   {
      mfile = dfname;
      mfile.append("=\"");
      mfile.append(meta);
      mfile.append("\"");
      out << "add_image("<<dfname;
   }
   else
   {
      bool add_comma = false;
      string data_sys_id, coord_transform_id;
      string transform_fname = "";
      string transform_obj = "";
      ChipsDataSystem *data_sys;

      dataobject_desc = dfname;
      dataobject_desc.append(" = DataObj()  # image data- ");
      string base_filename;

      buf.clear ();
      buf << out.get_filename() << ".Data" << out.use_tempfnum () <<"_image";
      base_filename = buf.str ();

      // determine transform name
      data_sys = static_cast <ChipsDataSystem *>(coord);
      data_sys_id = data_sys->get_id_no_wcs ();

      coord_transform_id = data_sys->get_default_transform_id ();

      // if the data system has no default transform is the same as
      // the incoming transform then the position is already in data
      // coordinates
      if (transform_id.size () > 0)
      {
         ChipsTransformFactory tf;

         // set the transform file name
         transform_fname = base_filename + "_transform.txt";

         // write the transform to a ascii file
         fstream transform_file(transform_fname.c_str(), ios::out);
         Transform *transform = data_sys->get_transform (transform_id);
         tf.serialize_ascii (transform, transform_file);

         transform_obj = dfname + ".transform";
         transform_desc = dfname + ".transform = #  insert transform here - " + transform_fname + "\n";
          
      }

  
      int num_channels = get_mode () ? 4 : 1;
      for (int ii=0; ii<num_channels; ii++)
      {
         double* dp = (this->get_data (ii, true));
         
         
         if (dp)
         {
            string tmpfile = out.get_filename();
            buf.clear ();
            buf.str ("");
            if (num_channels > 1)
               buf << base_filename << ii << ".bin";
            else
               buf << base_filename <<".bin";

            tmpfile = buf.str ();


            if (add_comma)
               dataobject_desc.append (",");
            add_comma = true;
            dataobject_desc.append(tmpfile);

            
            mfile.append(dfname);
            mfile.append(".channel");
            if (ii == 0)
               mfile.append("0");
            else if (ii == 1)
               mfile.append("1");
            else if (ii == 2)
               mfile.append("2");
            else if (ii == 3)
               mfile.append("3");

            mfile.append (" = #  insert image array here \n");

            // dump image
            ofstream ff(tmpfile.c_str(), ios::out|ios::binary);
            if (ff)
            {
               // image data format:
               //  [type (1 = image)][ num of axis (2)   ]
               //  [axis 1 dim      ][ axis 2 dim        ]
               //  [ double array col1 data .. dim1 * dim 2]
               long tmpax1 = 1, tmpax2 = 2;
               ff.write((char*)&tmpax1, sizeof(long));   // image format (1)
               ff.write((char*)&tmpax2, sizeof(long));   // num axes (2)
               tmpax1 = dim[1];
               ff.write((char*)&tmpax1, sizeof(long));   // x axis dim
               tmpax2 = dim[0];
               ff.write((char*)&tmpax2, sizeof(long));   // y axis dim

               int sz = dim[0] * dim[1];
         
               ff.write((char*)dp, sizeof(double)*sz);
            }
            ff.close ();
         }
         if (ii == 0)
            out<<"add_image(";

         if (dp)
            out<<dfname<<".channel"<<ii<<", ";
         else
            out<<"None, ";
      }
      out <<this->dim[0]<<", "<<this->dim[1];

      if (transform_obj.length () > 0) out << ", " << transform_obj;
   }

   if (att.str().length()) out <<",'"<<att.str()<<"'";
   out <<")"<<endl;
   mfile = dataobject_desc + "\n" + mfile + transform_desc;
   if (mfile.length())   out.append_to_template(mfile);

}


void ChipsImage::deserialize(ChipsStateStream& in)
{
   int len;
   short sval;
   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_int(&type); 
   in.read_int(dim, 3); 
   in.read_string(&current);
   in.read_string(&fname);
   in.read_string (&data_system_id);
   in.read_int (&len);
   coord_system = (COORD_SYS)len;
  
   in.read_string (&transform_id);

   in.read_short (&sval);
   colormap_id = sval;
   in.read_int(&interpolation);
   in.read_bool (&visible);

   in.read_int (&print_width);
   in.read_int (&print_height);

   in.read_double (&xmin);
   in.read_double (&xmax);
   in.read_double (&ymin);
   in.read_double (&ymax);

   if (in.get_minor_version () >= 10)
   {
      in.read_double (&unrotated_xmin);
      in.read_double (&unrotated_xmax);
      in.read_double (&unrotated_ymin);
      in.read_double (&unrotated_ymax);
   }
   else
   {
      unrotated_xmin = xmin;
      unrotated_xmax = xmax;
      unrotated_ymin = ymin;
      unrotated_ymax = ymax;
   }


   position_set = true;

}

int ChipsImage::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_normalize_channels ()
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::set_plot_location (double x1, double y1, double x2, double y2)
{
/*   ChipsDataSystem *data_sys;
   ChipsVector lower_bounds;

   if (!coord || (coord_system != DATA))
      return ERR_STATUS_BAD;

   data_sys = static_cast <ChipsDataSystem *>(coord);
   lower_bounds = data_sys->get_lower_bounds ();


   data_sys->convert_cs_to_cs (&lower_bounds, &lower_bounds, 1, DATA, data_sys->get_id (), FRAME_NORM, "");

   change_position (lower_bounds[0], lower_bounds[1]);

   return ERR_STATUS_GOOD;
*/
   return this->reset_position ();
}


string ChipsImage::get_xaxis ()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA))
   {
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_xaxis_id ();
   }
   return "";
}


string ChipsImage::get_yaxis ()
{
   ChipsDataSystem *data_sys;
   if (coord && (coord_system == DATA))
   {
      data_sys = static_cast <ChipsDataSystem *>(coord);
      return data_sys->get_yaxis_id ();
   }
   return "";
}

int ChipsImage::add_channel (double* vals, int xaxis, int yaxis, int channel, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsImageChannelCommand *cmd;
   double *current_data;

   if (!valid_rgba_type(static_cast<V_RGBA_TYPE>(channel)))
      cmd_manager->add_err(rval=ERR_BAD_RGBA, id, 0, channel);
      
   // 0 = pseudo color 1 = true color
   else if (get_mode () == 0)
      cmd_manager->add_err(rval=ERR_IMAGE_PC_MODE, id, 0, "set_image_channel");

   else if (xaxis * yaxis != dim[0] * dim[1])
      cmd_manager->add_err(rval=ERR_BAD_CHANNEL_DIM, id, 0, xaxis*yaxis, 
                           dim[0]*dim[1], dim[0], dim[1]);
   else
   {
      // normalize the channel to be between 0 and 3
      channel = channel - chips_red;


      current_data = this->get_data (channel, true);
      if (rec)  
         cmd = new ChipsImageChannelCommand (this, current_data,
                                           vals, channel, this->dim, true);
      add_component (vals, channel);
      if (rec && !add_command (cmd, rval))
         delete cmd;
   }

   return rval;
}

int ChipsImage::remove_channel (int channel, bool rec)
{
   ChipsImageChannelCommand *cmd;
   double *current_data;
   int rval = ERR_STATUS_GOOD;

   if (!valid_rgba_type(static_cast<V_RGBA_TYPE>(channel)))
      cmd_manager->add_err(rval=ERR_BAD_RGBA, id, 0, channel);

   // 0 = pseudo color 1 = true color
   else if (get_mode () == 0)
      cmd_manager->add_err(rval=ERR_IMAGE_PC_MODE, id, 0, "remove_image_channel");

   else
   {
      // normalize the channel to be between 0 and 3
      channel = channel - chips_red;

      current_data = this->get_data (channel, true);

      if (current_data == NULL)
         cmd_manager->add_err(rval=ERR_DEL_NULL_CHANNEL, id, 0, channel);

      else
      {
         if (rec)
            cmd = new ChipsImageChannelCommand (this, current_data,
                                           NULL, channel, this->dim, false);
         remove_component (channel);
         if (rec && !add_command (cmd, rval))
            delete cmd;
      }

   }
   return rval;
}


int ChipsImage::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   int mask = theDefaults->get_allpropmask();
   if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
   {
      int vis = theDefaults->get_allvisible();
      rval = set_property(static_cast<void*>(&vis), SET_IMAGE_VISIBLE);
   }
   if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
   {
      int dep = theDefaults->get_alldepth();
      rval = set_property(static_cast<void*>(&dep), SET_IMAGE_DEPTH);
   }
   return rval;
}

int ChipsImage::set_property(void* in_val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   string save_current;
   if (parent && (parent->get_object_type() == chips_plot))
      save_current = ((ChipsPlot *)parent)->get_current_image();

   switch(property)
   {
      case SET_IMAGE_CMAP:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            colormap_table->get_colormap_type(colormap_id), *ival, save_current);
         if ((rval = set_colormap((V_COLORMAP_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_INV_CMAP:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            colormap_table->get_invert_colormap (colormap_id), *bval, save_current);
         if ((rval = set_invert_colormap(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_CMAP_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            colormap_table->get_colormap_size (colormap_id), *ival, save_current);
         if ((rval = set_colormap_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_CMAP_INTERPOLATE:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            colormap_table->get_colormap_interpolate (colormap_id), *bval, save_current);
         if ((rval = set_colormap_interpolate(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_INTERPOLATION:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            interpolation, *ival, save_current);
         if ((rval = set_interpolation((V_INTERPOLATION_TYPE)*ival)) == 
	    ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_SCALE_CHANNELS:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            normalize_channels, *bval, save_current);
         if ((rval = set_normalize_channels(*bval)) ==
            ERR_STATUS_GOOD)
         {
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_IMAGE_ALPHA:
      {
         double* dval = static_cast<double*>(in_val);
	 if (rec)
         {
            double alpha[2];
            colormap_table->get_colormap_alpha_range (colormap_id, alpha);
	    cmd = new ChipsSetCommand(this, property, alpha, dval, 2,
	                 save_current);
         }
	 if ((rval = set_alpha(dval[0], dval[1])) == ERR_STATUS_GOOD)
	 {
	    if (rec && !add_command (cmd, rval))  delete cmd;
	 }
	 else  delete cmd;
      }
      break; 
      case SET_IMAGE_THRESHOLD:
      {
         double* dval = static_cast<double*>(in_val);
	 if (rec)
         {
            double thresh[2];
            colormap_table->get_colormap_scalar_threshold (colormap_id, thresh);
            cmd = new ChipsSetCommand(this, property, thresh,
	                  dval, 2, save_current);
         }
	 if ((rval = set_threshold(dval[0], dval[1])) == ERR_STATUS_GOOD)
	 {
	    if (rec && !add_command (cmd, rval))  delete cmd;
	 }
	 else  delete cmd;
      }
      break; 
      case SET_IMAGE_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, property,
              oval, newval, save_current);
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break; 
      case SET_IMAGE_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, save_current);
         if ((*ival) == 1)
         {
            show();
            if (rec && !add_command (cmd, rval))  delete cmd;
	 }
	 else if ((*ival) == 0)
         {
            hide();
            if (rec && !add_command (cmd, rval))  delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "image visibility", *ival, 0, 1);
            delete cmd;
         }
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)   cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_IMG);

   return rval;
}

int ChipsImage::set_normalize_channels (bool val)
{
   int rval = ERR_STATUS_GOOD;
   normalize_channels = val;
   if (initialized)
   {
      // 0 = pseudo color 1 = true color
      if (get_mode () == 0)
         cmd_manager->add_err(rval=ERR_IMAGE_PC_MODE, id, 0, "set_image_scale_channel");
      else
         change_normalize_channels ();
   }
   return rval;
}

int ChipsImage::set_colormap(V_COLORMAP_TYPE val)
{
   int rval = ERR_STATUS_GOOD;
   if (valid_colormap_type(val))
   {
      if (initialized)
         colormap_table->set_colormap_type (colormap_id, val);
   }
   else
      cmd_manager->add_err(rval=ERR_BAD_CMAP, id, 0, (int)val);
   return rval; 
}

int ChipsImage::set_invert_colormap(bool val)
{
   if (initialized)
      colormap_table->set_invert_colormap (colormap_id, val);
   return ERR_STATUS_GOOD;
}

int ChipsImage::set_colormap_size(int val)
{
   int rval = ERR_STATUS_GOOD;
   if (initialized)
   {
      if (val > 0)
         colormap_table->set_colormap_size (colormap_id, val);
      else 
         cmd_manager->add_err(rval=ERR_INT_MIN, id, 0, "colormap_size",
                              (int)val, 0);
   }
   return rval;
}

int ChipsImage::set_colormap_interpolate(bool val)
{
   if (initialized)
      colormap_table->set_colormap_interpolate (colormap_id, val);
   return ERR_STATUS_GOOD;
}


int ChipsImage::set_interpolation(V_INTERPOLATION_TYPE val)
{
   int rval = ERR_STATUS_GOOD;
   if (valid_interpolation_type(val))
   {
      interpolation = val;
      if (initialized)  change_interpolation();
   }
   else 
      cmd_manager->add_err(rval = ERR_BAD_IMG_INTERP,id,0,val);
   return rval;
}


int ChipsImage::set_alpha(double min, double max)
{
   if (max == -1) max = min; 
   if (min < 0 || min > 1.0)
   {
      cmd_manager->add_err(ERR_DBL_MINMAX,id,0,"alpha",min,0.0,1.0); 
      return ERR_DBL_MINMAX;
   }
   if (max < 0 || max > 1.0)
   {
      cmd_manager->add_err(ERR_DBL_MINMAX,id,0,"alpha",max,0.0,1.0); 
      return ERR_DBL_MINMAX;
   }
   if (initialized)
   {
      double alpha[] = {min, max};
      colormap_table->set_colormap_alpha_range (colormap_id, alpha);
   }
   return ERR_STATUS_GOOD;
}

int ChipsImage::set_threshold(double min, double max)
{
   if (min == max && min != 0)
   {
      ostringstream errmsg;
      errmsg << "The threshold must be specified as [min,max] where min (" << min <<") is less than max (" << max << ") or min and max are both zero." << ends;
      cmd_manager->add_err(errmsg.str(), ERR_DMIN_GT_DMAX);
      return ERR_DMIN_GT_DMAX;
   }
   else if (min > max)
   {
      cmd_manager->add_err(ERR_DMIN_GT_DMAX,id,0,min,max);
      return ERR_DMIN_GT_DMAX;
   }
   if (initialized)
   {
      double thresh[] = {min, max};

      // if the user supplied 0 for min and max, use the data extent
      if (min == 0 && max == 0)
         colormap_table->get_colormap_scalar_range(colormap_id, thresh);

      colormap_table->set_colormap_scalar_threshold (colormap_id, thresh);
   }
   return ERR_STATUS_GOOD;
}

int ChipsImage::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}

bool ChipsImage::get_invert_colormap()
{
   return colormap_table->get_invert_colormap (colormap_id);
}

int ChipsImage::get_colormap_size()
{
   return colormap_table->get_colormap_size (colormap_id);
}

bool ChipsImage::get_colormap_interpolate()
{
   return colormap_table->get_colormap_interpolate (colormap_id);
}


V_COLORMAP_TYPE ChipsImage::get_colormap()
{
   return colormap_table->get_colormap_type (colormap_id);
}

double ChipsImage::get_alpha_min()
{
   double alpha[2];
   colormap_table->get_colormap_alpha_range (colormap_id, alpha);
   return alpha[0];
}

double ChipsImage::get_alpha_max()
{
   double alpha[2];
   colormap_table->get_colormap_alpha_range (colormap_id, alpha);
   return alpha[1];
}

double ChipsImage::get_threshold_min()
{
   double thresh[2];
   colormap_table->get_colormap_scalar_threshold (colormap_id, thresh);
   return thresh[0];
}

double ChipsImage::get_threshold_max()
{
   double thresh[2];
   colormap_table->get_colormap_scalar_threshold (colormap_id, thresh);
   return thresh[1];
}

double ChipsImage::get_xmin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }
      return out[0];
   }
   else
      return xmin;
}

double ChipsImage::get_xmax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }
      return out[0];
   }
   else
      return xmax;

}

double ChipsImage::get_ymin(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmin, ymin);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }

      return out[1];
   }
   else
      return ymin;

}

double ChipsImage::get_ymax(bool transformed)
{
   if (transformed)
   {
      ChipsVector out(xmax, ymax);
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *> (coord);
      if (ds_coord)
      {
         string image_id = ds_coord->get_id_no_wcs ();
         coord->convert_cs_to_cs (&out, &out, 1, DATA, image_id, DATA, ds_coord->get_id (this->transform_id));
      }

      return out[1];
   }
   else
      return ymax;
}

double ChipsImage::get_xlmin()
{
   return 1;
}

double ChipsImage::get_ylmin()
{
   return 1;
}

int ChipsImage::update_print_scale (double scale)
{
   return calculate_position (scale);
}

int ChipsImage::print (string outfile, int fmt)
{
   int rval = ERR_STATUS_GOOD;
   int print_units = chips_pixels;

   if (theDefaults)
   {
      print_width = theDefaults->get_pagewidth();
      print_height = theDefaults->get_pageheight();
      print_units = theDefaults->get_pageunits();
      if (print_units == chips_nounit)
         print_units = chips_pixels;

      if (print_units != chips_pixels)
      {
         cmd_manager->add_err("The page units must be in pixels", id, 0);
         rval = ERR_STATUS_BAD;
      }   
   }

   if (!rval)
   {
      switch (fmt)
      {
         case 0: 
            rval = ERR_FEATURE_DNE;
            cmd_manager->add_err("Postscript is currently not a supported format for print_image", ERR_FEATURE_DNE,id,0);
         break;
         case 1:
            rval = ERR_FEATURE_DNE;
            cmd_manager->add_err("Encapsulated postcript is currently not a supported format for print_image", ERR_FEATURE_DNE,id,0);
         break;
         case 2:
            rval = ERR_FEATURE_DNE;
            cmd_manager->add_err("PDF is currently not a supported format for print_image", ERR_FEATURE_DNE,id,0);
         break;
         default:
         break;
      }
   }
      

   if (!rval)
      rval = this->doprint (outfile, fmt);

   return rval;
}

int ChipsImage::change_rotation (double angle)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_interpolation()
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_position (double x, double y)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_origin (double x, double y, double scaleX, double scaleY)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_extent (int lower_x, int lower_y, int upper_x, int upper_y)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::change_scale (double scaleX, double scaleY)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::flip_axis (int axis)
{
   return ERR_FEATURE_DNE;
}

/*
int ChipsImage::change_mode ()
{
   return ERR_FEATURE_DNE;
}
*/


int ChipsImage::add_component (double* vals, int channel)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::remove_component (int channel)
{
   return ERR_FEATURE_DNE;
}

int ChipsImage::doprint (string output, int fmt)
{
   return ERR_FEATURE_DNE;
}

double *ChipsImage::get_data (int channel, bool ignore_zero)
{
   return static_cast <double *>(0);
}

int ChipsImage::get_mode ()
{
   return 0;
}

int ChipsImage::clip_image (bool clip)
{
   return ERR_FEATURE_DNE;
}

bool ChipsImage::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED) outStatus = ERR_STATUS_GOOD;
          success = false;
   }
   return success;
}

void ChipsImage::set_transform_id (const string &transform_id)
{
   this->transform_id = transform_id;
   this->unrotated_xmin = this->xmin = .5;
   this->unrotated_xmax = this->xmax = dim[0]+.5;
   this->unrotated_ymin = this->ymin = .5;
   this->unrotated_ymax = this->ymax = dim[1]+.5;
   this->rotation = 0;
   this->position_set = false;
}

void ChipsImage::reproject_image ()
{
   ChipsDataSystem *data_sys = static_cast <ChipsDataSystem *>(coord);
   string default_transform_id = data_sys->get_default_transform_id ();
   Transform *image_transform = data_sys->get_transform (this->transform_id);
   Transform *default_transform = data_sys->get_transform (default_transform_id);
   bool calculate_rotation = false;
   double *default_rotation = NULL;
   double *default_scale = NULL;
   bool reflect_x = false;
   bool reflect_y = false;


   if (!default_transform || (default_transform_id == this->transform_id))
      return;


   this->rotation = 0;

   default_rotation = (double *)ChipsTransformFactory::get_parameter_copy (default_transform,"rotation");
   if (default_rotation)
   {
      this->rotation += *default_rotation;
      calculate_rotation = true;
      delete []default_rotation;
   }

   default_scale = (double *)ChipsTransformFactory::get_parameter_copy (default_transform,"scale");
   if (default_scale)
   {
      if (default_scale[0] < 0)
         reflect_x = true;
      if (default_scale[1] < 0)
         reflect_y = true;
      delete []default_scale;
   }


   // if the data system has no default transform is the same as
   // the incoming transform then the position is already in data
   // coordinates
   if (image_transform)
   {
      double corner[4][2] = {{this->xmin, this->ymin},
                              {this->xmin, this->ymax},
                              {this->xmax, this->ymax},
                              {this->xmax, this->ymin}};

      double *output;
      double *img_scale;

      if (calculate_rotation)
      {
         double *img_rotation = (double *)ChipsTransformFactory::get_parameter_copy (image_transform,"rotation");
         if (img_rotation)
         {
            this->rotation = *img_rotation - this->rotation;
            delete []img_rotation;
         }
      }

      img_scale = (double *)ChipsTransformFactory::get_parameter_copy (image_transform,"scale");
      if (img_scale)
      {
         // if the scales are opposite then it's a reflection
         if (img_scale[0] < 0)
            reflect_x = !reflect_x;
         if (img_scale[1] < 0)
            reflect_y = !reflect_y;

         delete []img_scale;
      }


      for (int ii=0; ii<4; ii++)
      {
         // convert from logical to world and from world
         output = (double *)image_transform->apply ((void*)(&corner[ii][0]), dmDOUBLE, 1);
      
         corner[ii][0] = output[0];
         corner[ii][1] = output[1];
         delete []output;

         // convert from world back to logical using the default data                          // coordinates
         output = (double *)default_transform->invert ((void*)(&corner[ii][0]), dmDOUBLE, 1);
         corner[ii][0] = output[0];
         corner[ii][1] = output[1];
         delete []output;

         if (ii==0)
         {
            xmax = xmin = corner[ii][0];
            ymax = ymin = corner[ii][1];
         }
         else
         {
            if (xmin > corner[ii][0])
               xmin = corner[ii][0];
            if (xmax < corner[ii][0])
               xmax = corner[ii][0];
            if (ymin > corner[ii][1])
               ymin = corner[ii][1];
            if (ymax < corner[ii][1])
               ymax = corner[ii][1];
         }
      }

      if (reflect_x)
      {
//         xmin = xmax;
         flipped_x = true;
      }
      if (reflect_y)
      {
//         ymin = ymax;
         flipped_y = true;
      }

      if (this->rotation != 0)
      {
         ChipsMatrix rotation_matrix (3,3);
         ChipsVector local_min (corner[0][0], corner[0][1]);
         ChipsVector local_max (corner[2][0], corner[2][1]);

         // rotate about zero

         rotation_matrix.identity ();

         // rotate the image
         rotation_matrix[0][0] = cos (this->rotation);
         rotation_matrix[0][1] = -sin (this->rotation);
         rotation_matrix[1][0] = sin (this->rotation);
         rotation_matrix[1][1] = cos (this->rotation);

         local_min = rotation_matrix * local_min;
         local_max = rotation_matrix * local_max;

         if (local_min[0] < local_max[0])
         {
            unrotated_xmin = local_min[0];
            unrotated_xmax = local_max[0];
         }
         else
         {
            unrotated_xmax = local_min[0];
            unrotated_xmin = local_max[0];
         }
         if (local_min[1] < local_max[1])
         {
            unrotated_ymin = local_min[1];
            unrotated_ymax = local_max[1];
         }
         else
         {
            unrotated_ymax = local_min[1];
            unrotated_ymin = local_max[1];
         }
      }
      else
      {
         unrotated_xmin = xmin;
         unrotated_xmax = xmax;
         unrotated_ymin = ymin;
         unrotated_ymax = ymax;
      }

   }
 
   this->transform_id = default_transform_id;

}
