// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSIMAGE_HH
#define CHIPSIMAGE_HH

#include <string>
#include <iostream>
#include <stdio.h>
#include "ChipsObject.hh"
#include "ChipsDefaults.hh"
#include "Transform.hh"

using namespace std;

class ChipsCommand;
class ChipsId;

class ChipsImage: public ChipsObject
{
  public:

   ChipsImage(ChipsObject* par, string filename, string transfrm_id, string val);
   ChipsImage(ChipsObject* par, string filename, 
             float x1, float x2, float y1, float y2, string transfrm_id, string val);
   ChipsImage(ChipsObject* par, int xaxis, int yaxis, double* rvals, double* gvals, double *bvals, double *avals, string transfrm_id, string val);
   ChipsImage(ChipsObject* par, ChipsStateStream& in);

   virtual ~ChipsImage();

   int        add_channel (ChipsImage &img, int channel);


   double     get_xmin(bool transformed);
   double     get_xmax(bool transformed);
   double     get_ymin(bool transformed);
   double     get_ymax(bool transformed);
   double     get_xlmin();
   double     get_ylmin();

   string     get_xaxis();
   string     get_yaxis();

   int        add_channel (double* vals, int xaxis, int yaxis, int channel, bool rec=true);
   int        remove_channel (int channel, bool rec=true);

   int        set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj);
   int        set_property(void* in_val, int property, bool rec=true);

   int        set_invert_colormap(bool val);
   bool       get_invert_colormap();
   int        set_colormap(V_COLORMAP_TYPE val); 
   V_COLORMAP_TYPE get_colormap();
   int        get_interpolation()        { return interpolation; }
   int        set_interpolation(V_INTERPOLATION_TYPE val); 
   int        set_alpha(double min, double max=-1);
   double     get_alpha_min();
   double     get_alpha_max();
   int        set_threshold(double min, double max);
   double     get_threshold_min();
   double     get_threshold_max();
   int        set_colormap_size (int size);
   int        get_colormap_size ();
   int        set_colormap_interpolate (bool interpolate);
   bool       get_colormap_interpolate ();
   int        set_depth(int newdepth);
   int        set_normalize_channels (bool val);
   bool       get_normalize_channels () { return normalize_channels; };

   int        print (string outfile, int fmt);
   int        update_print_scale (double scale);

   void       set_transform_id (const string &transform_id);

   unsigned int get_colormap_id () {return colormap_id;};

   virtual int reset_position ();

   virtual int change_visible();
   virtual int change_rotation (double angle);
   virtual int change_interpolation();
   virtual int change_position (double x, double y);
   virtual int change_origin (double x, double y, double scaleX, double scaleY);
   virtual int change_extent (int lower_x, int lower_y, int upper_x, int upper_y);
   virtual int change_scale (double scaleX, double scaleY=0);
/*   virtual int change_mode ();
*/
   virtual int change_normalize_channels ();
   virtual int flip_axis(int axis);
   virtual int add_component (double* vals, int channel);
   virtual int remove_component (int channel);
   virtual int clip_image (bool clip);


   virtual int change_depth();
   virtual int set_plot_location (double x1, double y1, double x2, double y2);

   virtual void object_delete ();
   virtual int show ();
   virtual int hide ();

   void       set_image_num (int val) {image_num = val;};
   int        get_image_num () {return image_num;};

   virtual void serialize(ChipsStateStream& out);
   virtual void serialize_ascii(ChipsStateStream& out);
   virtual void deserialize(ChipsStateStream& in);

   virtual double *get_data (int channel, bool ignore_zero=false);

   virtual int doprint (string outfile, int fmt);


  protected:

   virtual int     get_mode ();
/*   virtual void    set_data (double **data, int xaxis, int yaxis);
*/
   void       initialize(bool load_colormap);  // default initialization 
   int        calculate_position (double scale);
   void       reproject_image ();
/*   double     nearest_neighbor (double x, double y, double *data, 
                                int xdim, int ydim);
*/
        

   string     current;       // current curve 
   string     fname;         // name of file to image 
   int        type;          // type of image 
   int        dim[3];        // image dimensions 
   int        image_num;     // this labels order 0th, 1st, 2nd, etc.
   int        interpolation; // 0 = nearest neighbor, 1 = bilinear 2=bicubic:w
   bool       initialized;   // have we initialized everything
   int        mode;          // true color or psuedo color mode
   bool       flipped_x;     // indicates image flipped in x direction
   bool       flipped_y;     // indicates image flipped in y direction

   bool       normalize_channels; // normalize each channel individually

   double     xmin;          // the min x position in data coordinates
   double     xmax;          // the max x position in data coordinates
   double     ymin;          // the min y position in data coordinates
   double     ymax;          // the max y position in data coordinates
   bool       position_set;  // indicate when the positions have been set
   double     rotation;      // the rotation of the image

   double     unrotated_xmin; // the min x position unrotated
   double     unrotated_xmax; // the max x position unrotated
   double     unrotated_ymin; // the min y position unrotated
   double     unrotated_ymax; // the max y position unrotated


   string     transform_id;  // the transform id associated with this image
   string     meta;          // meta info for image (ie. filename)

   unsigned short colormap_id; // reference id for the colormap

   int        print_width;
   int        print_height;

  private:
   bool add_command (ChipsCommand *inCmd, int &outStatus);
};


#endif 
