// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsImageChannelCommand.hh"
#include "ChipsStateStream.hh"
#include "ChipsImage.hh"

ChipsImageChannelCommand::ChipsImageChannelCommand(ChipsImage* obj, 
   double *ovals, double *nvals, int channel, int *dim, bool add_channel)
{
   int size = 0;
   co = (ChipsObject*)obj;
   weight = 1;

   this->dim[0] = dim[0];
   this->dim[1] = dim[1];
   this->add_channel = add_channel;
   this->channel = channel;

   size = this->dim[0]*this->dim[1];

   this->diff_exists = false;
   if ((ovals == NULL) && (nvals == NULL))
      this->diff_vals = NULL;
   else
   {
      this->diff_vals = new double[size];
      if ((ovals == NULL) && (nvals != NULL))
         for (int rr = size; rr--; )
            this->diff_vals[rr] = nvals[rr];
      else if ((ovals != NULL) && (nvals == NULL))
         for (int rr = size; rr--; )
            this->diff_vals[rr] = ovals[rr];
      else
      {
         this->diff_exists = true;
         for (int rr = size; rr--; )
            this->diff_vals[rr] = nvals[rr]-ovals[rr];
      }
   }
}

ChipsImageChannelCommand::~ChipsImageChannelCommand()
{
   if (this->diff_vals)
      delete[] this->diff_vals;
}


ChipsImageChannelCommand::ChipsImageChannelCommand(ChipsObject *cob, ChipsStateStream& in)
{
   co = cob;
   deserialize(in);
   weight=1;
}

int ChipsImageChannelCommand::execute ()
{
   int rval = ERR_STATUS_BAD;
   double *data;
   double *new_vals;

   if (!this->diff_vals)
      rval = ERR_STATUS_GOOD;

   if (this->co && this->diff_vals)
   {
      ChipsImage* img = dynamic_cast <ChipsImage *>(co);
      if (this->add_channel)
      {
         // get the old channel from the image
         data = img->get_data (this->channel, true);

         // if there is no data on this channel just add in
         // the diff values
         if (data == NULL)
         {
            rval = img->add_channel(this->diff_vals, this->dim[0], this->dim[1], this->channel, false);
         } 
         else
         {
            // if there is data on the channel calculate the difference
            new_vals = new double[this->dim[0]*this->dim[1]];
            for (int rr = this->dim[0]*this->dim[1]; rr--; )
               new_vals[rr] = this->diff_vals[rr] + data[rr];
            rval = img->add_channel(new_vals, this->dim[0], this->dim[1], this->channel, false);
            delete []new_vals;
         }
      }
      else
         rval = img->remove_channel(this->channel, false);
   }
   return rval;
}

int ChipsImageChannelCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   double *data;
   double *old_vals;

   if (!this->diff_vals)
      rval = ERR_STATUS_GOOD;

   if (this->co && this->diff_vals)
   {
      ChipsImage* img = dynamic_cast <ChipsImage *>(co);

      if (this->add_channel)
      {
         // if a diff doesn't exist the old data must be null.
         // simply delete the channel
         if (!diff_exists)
            rval = img->remove_channel(this->channel, false);
         else
         {

            // if a diff does exists then old data must be another
            // channel -- add back the channel
            data = img->get_data (this->channel, true);
            old_vals = new double[this->dim[0]*this->dim[1]];
            for (int rr = this->dim[0]*this->dim[1]; rr--; )
               old_vals[rr] = data[rr] - this->diff_vals[rr];
            rval = img->add_channel(old_vals, this->dim[0], this->dim[1], this->channel, false);
            delete []old_vals;

         }
      }
      else
         rval = img->add_channel(this->diff_vals, this->dim[0], this->dim[1], this->channel, false);
   }
   return rval;
}

void ChipsImageChannelCommand::serialize(ChipsStateStream& out)
{
   int val = 27;
   bool diff_vals_set = (diff_vals != NULL);
   // Standard serialize output for a command
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_int (this->dim, 2);
   out.write_bool (&(this->add_channel));
   out.write_int (&(this->channel)); 
   out.write_bool (&diff_vals_set);
   
   if (diff_vals_set)
      out.write_double (this->diff_vals, this->dim[0]*this->dim[1]);
   out.write_bool (&(this->diff_exists));
}

void ChipsImageChannelCommand::deserialize(ChipsStateStream& in)
{
   bool diff_vals_set;
   // Standard serialize input for a command
   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_int (this->dim, 2);
   in.read_bool (&(this->add_channel));
   in.read_int (&(this->channel));
   in.read_bool (&diff_vals_set);

   if (diff_vals_set)
      in.read_double (this->diff_vals, this->dim[0]*this->dim[1]);
   else
      this->diff_vals = NULL;
   in.read_bool (&(this->diff_exists));

}

void ChipsImageChannelCommand::print ()
{
   cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Image Level" << endl;
};



