// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsPlot.hh"
#include "ChipsLabel.hh"
#include "ChipsFontList.hh"
#include "ChipsCmdManager.hh"
#include "ChipsDataSystem.hh"
#include "ChipsCommand.hh"

#include <stdio.h>

extern ChipsDefaults* theDefaults; 
extern ChipsFontList* theFonts;
extern ChipsCmdManager* cmd_manager;

ChipsLabel::ChipsLabel(ChipsObject* par, string lbid, 
                       ChipsCoordinate *inCoord, COORD_SYS coord_sys,
		       double inXPos, double inYPos,
		       string val):ChipsAnnotation(par, lbid)
{
   id = lbid; 
   text = val;  
   bbox = 0; 
   initialize(); 
   orig = ChipsVector (inXPos, inYPos);

   format_text(text, font, style, get_color(), size, &format);

   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;

      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) || 
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);

      // convert the data to logical coordinates if necessary
      if (coord_sys == DATA)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (&orig, &orig, 1, DATA, "", DATA, data_system_id);
         }

      }

   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, v_coord_sys_to_string(coord_sys));


}

ChipsLabel::ChipsLabel(ChipsObject* par, ChipsCoordinate *inCoord, COORD_SYS coord_sys, ChipsStateStream& in):ChipsAnnotation(par,"")
{
   initialize();

   deserialize(in);

   if (inCoord)  coord = inCoord;
   if (coord)
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_system == WINDOW_NORM) || (coord_system == PIXEL) ||
          (coord_system == PIXEL_REAL))
         coord = coord->get_coord(FRAME_NORM, "");
      else
         coord = coord->get_coord(coord_system, data_system_id);
      if (coord)
      {
         data_system_id = coord->get_id ();
         status = coord->insert (this);

         // older state files don't record the correct data_system_id
         // ensure that we get one without wcs transform
         if (coord_system == DATA)
         {
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
            if (ds_coord->get_default_transform_id () != "")
               data_system_id = ds_coord->get_id_no_wcs ();
         }

      }
   }


   format_text(text, font, style, get_color(), size, &format);
}

ChipsLabel::~ChipsLabel()
{
   if (coord)
      coord->remove (this);
   delete_tree(format);   // cleanup format tree
   format = NULL;
}

void ChipsLabel::object_delete ()
{
   delete this;
}


void ChipsLabel::initialize()
{
   bbox = 0;
   if (theDefaults)
   {
      string temp;
      temp       = theDefaults->get_labelfont(); 
      if (theFonts && theFonts->font_exists(temp))
         font = temp;
      else 
      {
         font = "helvetica";
         cmd_manager->add_err(ERR_FONT_DNE, id, 1, temp.c_str(), font.c_str());
      }
      style      = theDefaults->get_labelfontstyle(); 
      size       = theDefaults->get_labelsize(); 
      temp       = theDefaults->get_labelcolor();
      vert_alignment  = theDefaults->get_labelvalign();
      horiz_alignment = theDefaults->get_labelhalign();
      if (set_color(temp))  set_color("white");
      angle      = theDefaults->get_labelangle();
      depth      = theDefaults->get_labeldepth();
   }
   else
   {
      size = 1;
      font = "helvetica";
      style = chips_normal;
      vert_alignment = 0.0;
      horiz_alignment = 0.0;
      angle = 0;
      depth = V_USE_DEFAULT_DEPTH;
   }
   if (depth == V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
   format = NULL;
   initialized = true;
   object_type = chips_label;
   label_num = 0; 
   coord = NULL;
 
}


int ChipsLabel::set_cascading_property(ChipsId*, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot) || (obj == chips_label))
   {
      int mask = theDefaults->get_allpropmask();

      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();
         rval = set_property((void*)clr.c_str(), SET_LABEL_COLOR);
      }
      else if ((mask & PREF_ALL_FONT) == PREF_ALL_FONT)
      {
         string newfont = theDefaults->get_allfont();
         rval =  set_property((void*)newfont.c_str(), SET_LABEL_FONT);
      }
      else if ((mask & PREF_ALL_FONTSTYLE) == PREF_ALL_FONTSTYLE)
      {
         V_FONTSTYLE_TYPE fstyle = theDefaults->get_allfontstyle();
	 string nfs = v_fontstyle_to_string(fstyle); 
         rval = set_property((void*)nfs.c_str(), SET_LABEL_STYLE);
      }
      else if ((mask & PREF_ALL_ANGLE) == PREF_ALL_ANGLE)
      {
         float ang = theDefaults->get_allangle();
         rval = set_property(static_cast<void*>(&ang), SET_LABEL_ANGLE);
      }
      else if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         rval = set_property(static_cast<void*>(&vis), SET_LABEL_VISIBLE);
      }
      else if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
         int dep = theDefaults->get_alldepth();
         rval = set_property(static_cast<void*>(&dep), SET_LABEL_DEPTH);
      }
   }
   return rval;
}


int ChipsLabel::set_property(void* in_val, int property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;

   switch(property)
   {
      case SET_LABEL_TEXT:
      {
         string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            text, val, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_text(val)) == ERR_STATUS_GOOD)
         {
            if (rec)  
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_FONT:
      {
	 string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            font, val, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_font(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_STYLE:
      {
         V_FONTSTYLE_TYPE ival;
         if ((rval = c2_string_to_fontstyle((char*)in_val, &ival, 
             "label's fontstyle")) == ERR_STATUS_GOOD)
	 {
            if (rec)
	    {
	       string nv = (char*)in_val; 
               string ov = v_fontstyle_to_string(style);
	       cmd = new ChipsSetCommand(this, property, ov, nv,
                  ((ChipsPlot*)parent)->get_current_label());
	    }
            if ((rval = set_fontstyle(ival)) == ERR_STATUS_GOOD)
            {
               if (rec)
                  if (!add_command(cmd, rval))   delete cmd;
            }
            else  delete cmd;
         }
      }
      break;
      case SET_LABEL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            size, *ival, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_BBOX:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            size, *ival, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_bounding_box(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, property,
            angle, *fval, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_HALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = horiz_alignment;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *fval, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_horizalignment(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_VALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = vert_alignment;
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *fval, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_vertalignment(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_COLOR:
      {
         string clr = (const char*)in_val;
	 if (rec)  cmd = new ChipsSetCommand(this, property,
            get_color(), clr, ((ChipsPlot*)parent)->get_current_label());
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LABEL_VISIBLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = visible; 
         if (rec)  cmd = new ChipsSetCommand(this, property,
            oval, *ival, ((ChipsPlot*)parent)->get_current_label());
	 if ((*ival) == 1) 
	 {
	    show();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
	 else if ((*ival) == 0) 
	 {
	    hide();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  
	 {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "label visibility", *ival, 0, 1);
            delete cmd;
         }
      }
      break; 
      case SET_LABEL_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
	 int newval, oval = depth; 
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {  cmd = new ChipsSetCommand(this, property,
            oval, newval, ((ChipsPlot*)parent)->get_current_label());
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break; 
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);
   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_LBL);


   return rval;
}


int ChipsLabel::get_property(void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_LABEL_FONT:
      {
         string *sval =  (string*)out_val;
         *sval = get_font(); 
      }
      break;
      case SET_LABEL_TEXT:
      {
         string *sval =  (string*)out_val;
         *sval = get_text();
      }
      break;
      case SET_LABEL_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_LABEL_SIZE:
      {
         int *iv = (int*)out_val;
         *iv = get_size();
      }
      break; 
      case SET_LABEL_DEPTH:
      {
         int *iv = (int*)out_val;
         *iv = get_depth();
      }
      break; 
      case SET_LABEL_STYLE:
      {
         string *sval =  (string*)out_val;
         *sval = v_fontstyle_to_string(get_fontstyle());
      }
      break;
      case SET_LABEL_ANGLE:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_angle();
      }
      break; 
      case SET_LABEL_HALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_horizalignment();
      }
      break; 
      case SET_LABEL_VALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_vertalignment();
      }
      break; 
      case SET_LABEL_VISIBLE:
      {
         int *ival = (int*)out_val;
         *ival = is_visible();
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   return rval;
}


int ChipsLabel::set_text(string val) 
{
   if (val != text)
   {
      text = val;
      delete_tree(format);   // cleanup old format tree
      format_text(text, font, style, get_color(), size, &format);
      if (initialized)  change_text();
   }
   return ERR_STATUS_GOOD;
}


int ChipsLabel::set_font(string in_font)
{
   if (theFonts && theFonts->font_exists(in_font))
   {
      font = in_font;  
   }
   else 
   {
      string tmpfont = theDefaults->get_labelfont();
      if (theFonts && theFonts->font_exists(tmpfont))
         font = tmpfont;
      else 
         font = "helvetica"; 

      cmd_manager->add_err(ERR_FONT_DNE, id, 0, in_font.c_str(), font.c_str()); 
//      return ERR_FONT_DNE; 
   }
   delete_tree(format);   // cleanup old format tree
   format_text(text, font, style, get_color(), size, &format);
   if (initialized)  change_font();
   return ERR_STATUS_GOOD;
}


int ChipsLabel::set_fontstyle(V_FONTSTYLE_TYPE in_style)
{
   int rval = ERR_STATUS_GOOD;

   // validate specified font
   if (valid_fontstyle_type(in_style))
   {
      style = in_style;
      delete_tree(format);   // cleanup old format tree
      format_text(text, font, style, get_color(), size, &format);
      if (initialized)  change_font();
   }
   else
     cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "font style", (int)in_style, "chips_normal|chips_bold|chips_italic|chips_bolditalic");

   return rval;
}


int ChipsLabel::set_size(int val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine size range of fonts 
   if ((val < V_SIZE_MIN_VAL) || (val > V_SIZE_MAX_VAL))
   {   
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "label size", val, 
                                  V_SIZE_MIN_VAL, V_SIZE_MAX_VAL);
   }
   else
   {
      size = val;
      delete_tree(format);   // cleanup old format tree
      format_text(text, font, style, get_color(), size, &format);
      if (initialized)  change_size(); 
   }
   return rval;
}

int ChipsLabel::set_bounding_box(int val)
{
   bbox = val;
   change_bbox(); 
   return ERR_STATUS_GOOD;
}

int ChipsLabel::set_position (const ChipsVector &pos)
{
   orig = pos;

   return reset_position ();
}


int ChipsLabel::get_position (ChipsVector &pos)
{
   int rval  = ERR_STATUS_GOOD;
   pos = orig;
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
      {
         string ds_id = ds_coord->get_id (ds_coord->get_default_transform_id ());
         rval = ds_coord->convert_cs_to_cs (&orig, &pos, 1, DATA, data_system_id, DATA, ds_id);
      }
   }
   
   return rval;
}

int ChipsLabel::change_coordinate (COORD_SYS coordsys, const string& coordid)
{

   ChipsCoordinate *new_coord;
   int rval = ERR_STATUS_GOOD;

   // This is temporarily hard coded until a better method of searching
   // down the coordinate system tree is implemented. Otherwise objects
   // in window system must be kept in the frame coord
   if ((coordsys == WINDOW_NORM) || (coordsys == PIXEL) ||
       (coordsys == PIXEL_REAL))
      new_coord = coord->get_coord (FRAME_NORM, "");
   else
      new_coord = coord->get_coord (coordsys, coordid);

   if (!new_coord)
   {
      cmd_manager->add_err(ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coordsys));
      return ERR_COORD_DNE;
   }
   coord->convert_cs_to_cs (&orig, &orig, 1, coord_system, data_system_id, coordsys, coordid);

   rval = coord->remove (this);

   if (!rval)
      rval = new_coord->insert (this, false);

   if (!rval)
   {
      this->set_chips_coord (new_coord);
      this->set_coord_system (coordsys, coordid);

      this->change_position ();

      cmd_manager->update_mask (STATE_CHNG_PROP_LBL);
   }

   return ERR_STATUS_GOOD;
}




int ChipsLabel::reset_position ()
{
   int rval = ERR_STATUS_GOOD;

   if ((rval == ERR_STATUS_GOOD) && initialized)
   {
      if ((rval = change_position ()) == ERR_STATUS_GOOD)
         rval = change_scale_factor();
   }
   return rval;
}


int ChipsLabel::move(COORD_SYS coord_sys, const string &coord_id, double xx,
   double yy, int mode, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsMoveCommand* cmd = NULL;
   ChipsVector tmp_n (0.0, 0.0);
   ChipsVector tmp_org (0.0, 0.0);

   string csid = coord_id; 

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (coord_sys == COORD_NONE)
   {
      coord_sys = coord_system;
      csid = coord->get_id ();
   }

   if (!coord->coord_exists (coord_sys, csid))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), csid.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(DATA));
      return ERR_COORD_DNE;
   }

   if (rec)  cmd = new ChipsMoveCommand(this, orig[0], orig[1], xx, yy, coord_sys,
      coord_id, mode, ((ChipsPlot*)parent)->get_current_label());

   if (mode != V_MOVE_DELTA)
      tmp_org = ChipsVector(xx, yy);
   else
   {
      // convert offset and (0,0) to frame norm to get offset
      rval = coord->convert_cs_to_cs(&orig, &tmp_org, 1, coord_system,
				     data_system_id, coord_sys, csid);

      if (rval == ERR_STATUS_GOOD){
         // adjust norm pos by offset
         tmp_org[0] += xx;
         tmp_org[1] += yy;
      }
   }

   // convert new pos to frame norm
   rval = coord->convert_cs_to_cs(&tmp_org, &tmp_org, 1, coord_sys,
                                     csid, coord_system, data_system_id);

   if (rval == ERR_NEG_LOG_VAL)
      cmd_manager->add_err("Cannot move an object to a position less than or equal to zero when in log scale.", rval=ERR_EDGE_BOUNDARY, id, 0);

   if (!rval){
      orig = tmp_org;
      if ((rval == ERR_STATUS_GOOD) && initialized)
      {
         if ((rval = change_position ()) == ERR_STATUS_GOOD)
            rval = change_scale_factor();
      }

   }
   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if (!add_command(cmd, rval))   delete cmd; 
   }
   else if (rec)
      delete(cmd);

   if (!rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_LBL);

   return rval;
}


int ChipsLabel::resize(COORD_SYS coord_sys, const string &coord_id, 
                           double *bounds, bool rec)
{

   ChipsCoordinate *tmp_coord = coord;
   ChipsVector lower (bounds[0], bounds[1]);
   ChipsVector upper (bounds[2], bounds[3]);
   int rval;
   bool lock = cmd_manager->cmd_lock ();
   double boundingBox[4];

   if (!rec)
      cmd_manager->set_cmd_lock (true);

   if (!tmp_coord)
      tmp_coord = this->parent->get_chips_coord ();

   if (!tmp_coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (!tmp_coord->coord_exists (coord_sys, coord_id))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), coord_id.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(DATA));
      return ERR_COORD_DNE;
   }

   // convert to pixel real which is what compute_resize uses
   rval = tmp_coord->convert_cs_to_cs(&lower, &lower, 1, coord_sys,
                                     coord_id, PIXEL_REAL, "");
   rval = tmp_coord->convert_cs_to_cs(&upper, &upper, 1, coord_sys,
                                     coord_id, PIXEL_REAL, "");

   if (!rec)
      cmd_manager->set_cmd_lock (true);
   else
      cmd_manager->open_block ();

 
   boundingBox[0] = lower[0];
   boundingBox[1] = lower[1];
   boundingBox[2] = upper[0];
   boundingBox[3] = upper[1];

   rval = this->compute_resize (boundingBox);
   if (!rec)
      cmd_manager->set_cmd_lock (lock);
   else
      cmd_manager->close_block ();

   return rval;


}


int ChipsLabel::set_vertalignment(double val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > -10.0 && val < 10.0)
      vert_alignment=val;
   else 
      cmd_manager->add_err(rval=ERR_DBL_MINMAX, id, 0, "label's vertical alignment", 
                               val, -10.0, 10.0);
   if (initialized)  change_alignment();
   return rval;
}


int ChipsLabel::set_horizalignment(double val)
{
   int rval = ERR_STATUS_GOOD;

   if (val > -10.0 && val < 10.0)
      horiz_alignment=val;
   else 
      cmd_manager->add_err(rval=ERR_DBL_MINMAX, id, 0, "label's horizontal alignment",
                               val, -10.0, 10.0);
   if (initialized)  change_alignment();
   return rval;
}


int ChipsLabel::redraw()
{
   return 1;
}

int ChipsLabel::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
   {
      if (initialized)
      {
         delete_tree(format);   // cleanup old format tree
         format_text(text, font, style, get_color(), size, &format);
         rval = change_color();
      }
   }
   return rval;
}


int ChipsLabel::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}

int ChipsLabel::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsLabel::change_text()
{
   return 1;
}

int ChipsLabel::change_font()
{
   return 1;
}

int ChipsLabel::change_size()
{
   return 1;
}

int ChipsLabel::change_position()
{
   return 1;
}

int ChipsLabel::change_scale_factor()
{
   return 1;
}

int  ChipsLabel::change_alignment()
{
  return 1;
};

int  ChipsLabel::change_visible()
{
  return ERR_FEATURE_DNE;
};

int  ChipsLabel::change_bbox()
{
  return ERR_FEATURE_DNE;
};

int ChipsLabel::compute_resize( double *bounds) 

{
  return ERR_FEATURE_DNE;
}


void ChipsLabel::serialize(ChipsStateStream& out)
{
   string tmp;
   int sz;
  
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   out.write_string(&text);
   out.write_int(&size);
   out.write_string(&font);
   sz = (int)style;
   out.write_int(&sz);
   out.write_double(&vert_alignment);
   out.write_double(&horiz_alignment);
   out.write_int (&depth);
   sz = orig.get_size ();
   out.write_int (&sz);
   for (int ii=0; ii<sz; ii++)
      out.write_double(&(orig[ii]));

   tmp = get_color();
   out.write_string(&tmp);
   out.write_float(&angle);

   sz = (int) coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);
   out.write_bool(&visible);
}

void ChipsLabel::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];
   const char *cp;

   tmp = get_color();
   if (theDefaults->get_labelcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("color=");
      att.append(tmp);
   }
   if (size != theDefaults->get_labelsize())
   {
      sprintf(buf, "size=%1d", size);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (angle != theDefaults->get_labelangle())
   {
      sprintf(buf, "angle=%2.1f", angle);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   int dep = theDefaults->get_labeldepth();
   if (depth != dep &&
       (dep!=V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (theDefaults->get_labelfont() != font)
   {
      if (att.length()) att.append(" ");
      att.append("font=");
      att.append(font);
   }
   if (theDefaults->get_labelfontstyle() != style &&
       ((cp = v_fontstyle_to_string(style))!=NULL))
   {
      if (att.length()) att.append(" ");
      att.append("fontstyle=");
      att.append(cp);
   }
   if (vert_alignment != theDefaults->get_labelvalign())
   {
      sprintf(buf, "valign=%2.1f", vert_alignment);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (horiz_alignment != theDefaults->get_labelhalign())
   {
      sprintf(buf, "halign=%2.1f", horiz_alignment);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (coord_system != DATA &&
       (cp = v_coord_sys_to_enum_string(coord_system)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("coordsys=");
      att.append(cp);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   // convert the data to logical coordinates if necessary
   ChipsVector tmp_pos = orig;
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
      {
         data_system_id = ds_coord->get_id_no_wcs ();
         ds_coord->convert_cs_to_cs (&orig, &tmp_pos, 1, DATA, data_system_id, DATA, "");
      }
   }
   string tmpval = text;
   if (text.find('\\') != string::npos)
   {
      // need to escape backslash in text
      string::size_type pos=tmpval.rfind("\\");
      while (pos != string::npos)
      {
         tmpval.replace(pos, 1, "\\\\");
         if (pos > 0) pos=tmpval.rfind("\\",pos-1);
            else      pos = string::npos; 
      }
   }
   if (text.find('"') != string::npos)
   {
      // need to escape quotes in text 
      string::size_type pos=tmpval.rfind("\"");
      while (pos != string::npos)
      {
         tmpval.replace(pos, 1, "\\\"");
         pos=tmpval.rfind("\"",pos);
      }
   }
   out <<"add_label("<<tmp_pos[0]<<", "<<tmp_pos[1]<<", \""<<tmpval<<"\"";
   if (att.length()) out <<", \""<<att<<"\"";
   out <<")"<< endl;
}

void ChipsLabel::deserialize(ChipsStateStream& in)
{
   int len;
   string tmp;
   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_string(&text);
   in.read_int(&size);
   in.read_string(&font);
   in.read_int(&len);
   style = (V_FONTSTYLE_TYPE)len;
   in.read_double(&vert_alignment);
   in.read_double(&horiz_alignment);
   in.read_int(&depth);

   in.read_int((int*)&len);
   orig = ChipsVector (len);
   for (int ii=0; ii<orig.get_size (); ii++)
      in.read_double(&(orig[ii]));
   in.read_string(&tmp);
   set_color(tmp);
   in.read_float(&angle);
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);
   in.read_bool(&visible);
}

int ChipsLabel::show ()
{
  visible = true;
  return change_visible ();
}

int ChipsLabel::hide ()
{
  visible = false;
  return change_visible ();
}

bool ChipsLabel::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED)
         outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}
