// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsLine.hh"
#include "ChipsPlot.hh"

#include "ChipsCmdManager.hh"
#include "ChipsCommand.hh"

#include <stdio.h>
#include <cmath>

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;


ChipsLine::ChipsLine(ChipsObject* par, ChipsCoordinate *inCoord,
             COORD_SYS coord_sys, double startx, double starty,
             double endx, double endy, string lnid):ChipsAnnotation(par,
             lnid)
{
   id = lnid; 
   initialize();

   pnts = new ChipsVector[2];
   num_pnts = 2;

   if (pnts)
   {
      pnts[0] = ChipsVector (startx, starty);
      pnts[1] = ChipsVector (endx, endy);
      endpoints[0] = pnts[0];
      endpoints[1] = pnts[1];
   }

   // Cant process if there's no chips coordinate
   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);

      if ((coord_sys == DATA) && coord)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (pnts, pnts, num_pnts, DATA, "", DATA, data_system_id);
            endpoints[0] = pnts[0];
            endpoints[1] = pnts[1];
         }
      }


      // calculate the end points incase they extend
      calculate_end_points ();
   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);
}


ChipsLine::ChipsLine(ChipsObject* par, ChipsCoordinate *inCoord, 
   COORD_SYS coord_sys, ChipsStateStream& in):ChipsAnnotation(par, "")
{
   initialize();
   deserialize(in);

   if (inCoord)  coord = inCoord;
   if (coord)
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_system == WINDOW_NORM) || (coord_system == PIXEL) ||
          (coord_system == PIXEL_REAL))
         coord = coord->get_coord(FRAME_NORM, "");
      else
         coord = coord->get_coord(coord_system, data_system_id);
      if (coord)
      {
         data_system_id = coord->get_id ();
         status = coord->insert (this);

         // older state files don't record the correct data_system_id
         // ensure that we get one without wcs transform
         if (coord_system == DATA)
         {
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
            if (ds_coord->get_default_transform_id () != "")
               data_system_id = ds_coord->get_id_no_wcs ();
         }
      }
   }

   endpoints[0] = pnts[0];
   endpoints[1] = pnts[num_pnts-1];

   // calculate the end points incase they extend
   calculate_end_points ();

}


ChipsLine::ChipsLine(ChipsObject* par, ChipsCoordinate *inCoord, 
             COORD_SYS coord_sys, double* xcoords, double* ycoords, 
             unsigned int cnt, string lnid):ChipsAnnotation(par, lnid)
{
   unsigned int rr; 

   id = lnid;
   initialize();
   num_pnts = cnt;

   if (xcoords && ycoords)
   {
      pnts = new ChipsVector[cnt];
   }
   else   num_pnts = 0;

   for (rr = num_pnts; rr--; )
   {
      pnts[rr] = ChipsVector (xcoords[rr], ycoords[rr]);
   }

   endpoints[0] = pnts[0];
   endpoints[1] = pnts[num_pnts-1];


   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);

      if ((coord_sys == DATA) && coord)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (pnts, pnts, num_pnts, DATA, "", DATA, data_system_id);
            endpoints[0] = pnts[0];
            endpoints[1] = pnts[num_pnts-1];
         }
      }

      // calculate the end points incase they extend
      calculate_end_points ();
   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);

}

ChipsLine::ChipsLine (const ChipsLine &inLine):ChipsAnnotation (inLine)
{

   if (pnts)
      for (int ii=0; ii<inLine.pnts[0].get_size (); ii++)
         pnts[ii] = inLine.pnts[ii];
   endpoints[0] = inLine.endpoints[0];
   endpoints[1] = inLine.endpoints[1];
   num_pnts = inLine.num_pnts;
   thickness = inLine.thickness;
   linestyle = inLine.linestyle;
   line_num = inLine.line_num;
   extend_line = inLine.extend_line;
}



ChipsLine::~ChipsLine()
{
   if (pnts)
      delete []pnts;

   if (coord)
      coord->remove (this);
}

void ChipsLine::object_delete ()
{
   delete (this);
}


void ChipsLine::initialize()
{
   if (theDefaults)
   {
      string temp; 

      linestyle  = theDefaults->get_linestyle();
      thickness  = theDefaults->get_linethickness();
      temp       = theDefaults->get_linecolor();
      depth      = theDefaults->get_linedepth();
      extend_line = theDefaults->get_lineextend();

      if (set_color(temp))  set_color("white");
   }
   else
   {
      linestyle  = chips_solid;
      set_color("white");
      thickness  = 1;
      depth = V_USE_DEFAULT_DEPTH;
      extend_line = chips_extend_none;
   }
   if (depth == V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
   object_type = chips_line;
   angle = 0;
   initialized = true;
   line_num = 0;
   coord = NULL;
}

int ChipsLine::change_coordinate (COORD_SYS coordsys, const string& coordid)
{

   ChipsCoordinate *new_coord;
   int rval = ERR_STATUS_GOOD;

   // This is temporarily hard coded until a better method of searching
   // down the coordinate system tree is implemented. Otherwise objects
   // in window system must be kept in the frame coord
   if ((coordsys == WINDOW_NORM) || (coordsys == PIXEL) ||
       (coordsys == PIXEL_REAL))
      new_coord = coord->get_coord (FRAME_NORM, "");
   else
      new_coord = coord->get_coord (coordsys, coordid);

   if (!new_coord)
   {
      cmd_manager->add_err(ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coordsys));
      return ERR_COORD_DNE;
   }

   coord->convert_cs_to_cs (pnts, pnts, num_pnts, coord_system, data_system_id, coordsys, coordid);
   coord->convert_cs_to_cs (&endpoints[0], &endpoints[0], 2, coord_system, data_system_id, coordsys, coordid);

   rval = coord->remove (this);

   if (!rval)
      rval = new_coord->insert (this, false);

   if (!rval)
   {
      this->set_chips_coord (new_coord);
      this->set_coord_system (coordsys, coordid);
      cmd_manager->update_mask (STATE_CHNG_PROP_LINE);
   }

   if (!rval)
      rval = this->update_clipping ();

   return rval;
}

int ChipsLine::get_position (ChipsVector &pos)
{
   int rval = ERR_STATUS_GOOD;
   pos = pnts[0];

   // convert the data to logical coordinates if necessary
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
         rval = ds_coord->convert_cs_to_cs (&pos, &pos, 1, DATA, data_system_id, DATA, "");
   }

   return rval;
}



int ChipsLine::move(COORD_SYS coord_sys, const string &coord_id, double xx,
                   double yy, int mode, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsMoveCommand* cmd = NULL;
   string csid = coord_id;
 
   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }
  
   if (coord_sys == COORD_NONE)
   {
      coord_sys = coord_system;
      csid = coord->get_id ();
   }

   if (!coord->coord_exists (coord_sys, csid))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), csid.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coord_sys));

      return ERR_COORD_DNE;
   }

   // Reset the end points back to normal
   pnts[0] = endpoints[0];
   pnts[num_pnts-1] = endpoints[1];


   // convert each point 
   ChipsVector tmp_n(pnts[0].get_size ());
   ChipsVector tmp_org(pnts[0].get_size ());

   ChipsVector off(pnts[0]);
   unsigned int rr;

   if (rec)  cmd = new ChipsMoveCommand(this, off[0], off[1], xx, yy, coord_sys,
      coord_id, mode, ((ChipsPlot *)parent)->get_current_line());

   if (mode != V_MOVE_DELTA){
      tmp_n = ChipsVector (xx, yy);
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, coord_sys,
                                     csid, coord_system, data_system_id);

   }
   else
   {
      // convert offset and (0,0) to frame norm to get offset
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, coord_sys,
                                     csid, FRAME_NORM, "");

      tmp_n = ChipsVector (xx, yy);
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_n, 1, coord_sys,
                                     csid, FRAME_NORM, "");
      tmp_org[tmp_org.get_size()-1] = 0;
      tmp_n -= tmp_org;
      // convert original pos to frame norm and add offset
      rval = coord->convert_cs_to_cs(&off, &tmp_org, 1, coord_system,
                                     data_system_id, FRAME_NORM, "");
      tmp_org[tmp_org.get_size()-1] = 0;
      tmp_n += tmp_org;
      // convert adjusted frame norm pos to orignal coord sys
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, FRAME_NORM, "",
                                     coord_system, data_system_id);

   }

   if (rval == ERR_NEG_LOG_VAL)
      cmd_manager->add_err("Cannot move an object to a position less than or equal to zero when in log scale.", rval=ERR_EDGE_BOUNDARY, id, 0);
   else if (rval == ERR_STATUS_GOOD)
   {
      ChipsVector translation(tmp_org-off);
      for (rr = 0; rr < num_pnts; rr++)
         pnts[rr] += translation;
   }

   

   // copy back the new endpoints
   endpoints[0] = pnts[0];
   endpoints[1] = pnts[num_pnts-1];

   if (!rval)
      rval = change_position ();
   if ((extend_line != chips_extend_none) && !rval)
      rval = calculate_end_points ();

   // Update the data back points actually changed position
   // This is not the most efficient way of doing this as it
   // involves recreating the polygon from scratch
   if (rval == ERR_STATUS_GOOD)  rval = update_data ();
   

   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if (!add_command(cmd, rval))   delete cmd;
   }
   else if (rec)
      delete(cmd);

   if (!rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_LINE);

   return rval;
}


int ChipsLine::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot) || (obj == chips_line))
   {
      int mask = theDefaults->get_allpropmask();

      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();
	 rval = set_property((void*)clr.c_str(), SET_LINE_COLOR);
      }
      else if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
      {
         float vis = theDefaults->get_allthickness();
         rval = set_property(static_cast<void*>(&vis), SET_LINE_THICKNESS);
      }
      else if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         rval = set_property(static_cast<void*>(&vis), SET_LINE_VISIBLE); 
      }
      if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
         int dep = theDefaults->get_alldepth();
         rval = set_property(static_cast<void*>(&dep), SET_LINE_DEPTH);
      }
   }
   return rval;
}


int ChipsLine::set_property(void* in_val, int in_property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;

   switch(in_property)
   {
      case SET_LINE_STYLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            (int)linestyle, *ival, ((ChipsPlot *)parent)->get_current_line());
         if ((rval = set_linestyle((V_LINE_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
	 }
	 else  delete cmd;
      }
      break;
      case SET_LINE_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            thickness, *fval, ((ChipsPlot*)parent)->get_current_line());
         if ((rval = set_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
	 }
	 else  delete cmd;
      }
      break;
      case SET_LINE_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_color(), clr, ((ChipsPlot*)parent)->get_current_line());
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_LINE_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
	    oval, *ival, ((ChipsPlot*)parent)->get_current_line());
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "line visibility", *ival,0, 1);
            delete cmd;
         }
      }
      break;
      case SET_LINE_DEPTH:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int newval;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, in_property,
               (int)depth, newval, ((ChipsPlot *)parent)->get_current_line());
            ((ChipsPlot *)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
               if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot *)parent)->add_object_to_depth_table(this);
      }
      break; 
      case SET_LINE_EXTEND:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            extend_line, *ival, ((ChipsPlot *)parent)->get_current_line());

         if ((rval = set_extend((V_LINE_EXTENSION_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;

      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, in_property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_LINE);


   return rval;
}


int ChipsLine::get_property(void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_LINE_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_linestyle();
      }
      break;
      case SET_LINE_THICKNESS:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_thickness();
      }
      break;
      case SET_LINE_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_LINE_VISIBLE:
      {
         int *ival = (int*)out_val;
         *ival = is_visible();
      }
      break; 
      case SET_LINE_DEPTH:
      {
         int *ival = (int*)out_val;
         *ival = get_depth();
      }
      break; 
      case SET_LINE_EXTEND:
      {
         int *ival = (int*)out_val;
         *ival = get_extend ();
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }
   return rval;
}


int ChipsLine::set_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine thickness of line
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "line thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      thickness = val;
      change_thickness(); 
   }
   return rval;
}

int ChipsLine::set_linestyle(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine thickness of line
   if (valid_line_type(val))
   {
      linestyle = val;
      change_linestyle();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val, "chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");
   return rval;
}

int ChipsLine::set_linestyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if ((rval = c2_string_to_line(val.c_str(), &linestyle, "line style")) == ERR_STATUS_GOOD)
      change_linestyle();
   return rval;
}

int ChipsLine::set_extend(V_LINE_EXTENSION_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine thickness of line
   if (valid_line_extension_type(val))
   {
      if (extend_line != val)
      {
         extend_line = val;
         pnts[0] = endpoints[0];
         pnts[num_pnts -1] = endpoints[1];

         rval = calculate_end_points ();

         // WARNING: This redraws the line from scratch may want to
         // investigate a faster way of doing this
         if (!rval)
            rval = update_data ();
      }
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line extend", (int)val, "chips_extend_none|chips_extend_start|chips_extend_end|chips_extend_both");
   return rval;
}


int ChipsLine::set_extend(string val)
{
   int rval = ERR_STATUS_GOOD;
   V_LINE_EXTENSION_TYPE tmp_val;

   if ((rval = c2_string_to_line_extension(val.c_str(), &tmp_val, "line extend")) == ERR_STATUS_GOOD)
      if (extend_line != tmp_val)
      {
         extend_line = tmp_val;
         pnts[0] = endpoints[0];
         pnts[num_pnts -1] = endpoints[1];

         rval = calculate_end_points ();

         // WARNING: This redraws the line from scratch may want to
         // investigate a faster way of doing this
         if (!rval)
            rval = update_data ();
      }

   return rval;
}


int ChipsLine::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}

int ChipsLine::reset_position ()
{
   int rval;
   rval = change_position ();

   if ((extend_line != chips_extend_none) && !rval)
   {
      rval = calculate_end_points ();

      // WARNING: This redraws the line from scratch may want to
      // investigate a faster way of doing this
      if (!rval)
         rval = update_data ();
   }

   return rval;
}

int ChipsLine::redraw()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::change_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::change_linestyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::change_position()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::update_data()
{
   return ERR_FEATURE_DNE;
}

int ChipsLine::update_clipping ()
{
   return ERR_FEATURE_DNE;
}


void ChipsLine::serialize(ChipsStateStream& out)
{
   string tmp;
   int sz;
   out.write_string(&id);
   sz = (int)ref_cnt;
   out.write_int(&sz);
   tmp = get_color();
   out.write_string(&tmp);
   sz = (int) coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);

   out.write_int((int*)&num_pnts);
   out.write_int((int*)&num_pnts);
   if (pnts)
      sz = pnts[0].get_size ();
   else
      sz = 0;
   out.write_int (&sz);
   for (unsigned int ii=0; ii<num_pnts; ii++)
      for (int jj=0; jj<sz; jj++)
         out.write_double(&(pnts[ii][jj]));
   out.write_float(&thickness);
   out.write_int(&depth);
   sz = (int)linestyle;
   out.write_int(&sz);
   out.write_bool (&visible);
   out.write_int (&extend_line);
}

void ChipsLine::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];
   const char *cp;

   tmp = get_color();
   if (theDefaults->get_linecolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("color=");
      att.append(tmp);
   }
   if (thickness != theDefaults->get_linethickness())
   {
      if (att.length()) att.append(" ");
      if (thickness != 0.5)
      {
         sprintf(buf,"line.thickness=%-1.0f", thickness);
         att.append(buf);
      }
      else
         att.append("thickness=0.5");
   }
   int dep = theDefaults->get_linedepth();
   if (depth != dep &&
       (dep != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if  (extend_line != theDefaults->get_lineextend() &&
        (cp = v_line_extension_to_string((V_LINE_EXTENSION_TYPE)extend_line)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("extend=");
      att.append(cp);
   }
   if (theDefaults->get_linestyle() != linestyle &&
       ((cp = v_line_to_string(linestyle))!=NULL))
   {
      if (att.length()) att.append(" ");
      att.append("style=");
      att.append(cp);
   }
   if (coord_system != DATA &&
       (cp = v_coord_sys_to_enum_string(coord_system)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("coordsys=");
      att.append(cp);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   ChipsDataSystem *ds_coord = NULL;
   ChipsVector *tmp_pnts = pnts;
   bool delete_tmp_pnts = false;

   // convert the data to logical coordinates if necessary
   if ((coord_system == DATA) && coord)
      ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
   if (ds_coord && (ds_coord->get_default_transform_id () != ""))
   {
      tmp_pnts = new ChipsVector[num_pnts];
      ds_coord->convert_cs_to_cs (pnts, tmp_pnts, num_pnts, DATA, data_system_id, DATA, "" );
      delete_tmp_pnts = true;
   }

   out <<"add_line("; 
   if (num_pnts == 2 && tmp_pnts)
   {
      out <<tmp_pnts[0][0]<<", "<<tmp_pnts[0][1]<<", "<<tmp_pnts[1][0]<<", "<<tmp_pnts[1][1];
   }
   else
   {
      unsigned int ii; 
      out << "[";
      for (ii = 0; ii < num_pnts; ii++)
      {
         out << tmp_pnts[ii][0];
         if (ii < num_pnts-1)  out << ", ";
      }
      out << "],[";
      for (ii = 0; ii < num_pnts; ii++)
      {
         out << tmp_pnts[ii][1];
         if (ii < num_pnts-1)  out << ", ";
      }
      out <<"]";
      if (delete_tmp_pnts)   delete []tmp_pnts;
   }
   if (att.length()) out <<", \""<<att<<"\"";
   out <<")"<< endl;
}

void ChipsLine::deserialize(ChipsStateStream& in)
{
   string tmp;
   int len;

   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = (unsigned int) len;
   in.read_string(&tmp);
   set_color(tmp);
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);

   in.read_int((int*)&num_pnts);
   in.read_int((int*)&len);

   if ((unsigned int)len == num_pnts) 
   {
      in.read_int((int*)&len);
      pnts = new ChipsVector[num_pnts];
      for (unsigned int ii=0; ii<num_pnts; ii++)
      {
	 pnts[ii] = ChipsVector (len);
	 for (int jj=0; jj<len; jj++)
 	    in.read_double(&(pnts[ii][jj]));
      }
   }
   in.read_float(&thickness);
   in.read_int(&depth);
   in.read_int(&len);
   linestyle = (V_LINE_TYPE)len;
   in.read_bool (&visible);
   in.read_int (&extend_line);
}

int ChipsLine::show ()
{
  visible = true;
  return change_visible ();
}

int ChipsLine::hide ()
{
  visible = false;
  return change_visible ();
}

int ChipsLine::calculate_end_points ()
{

   ChipsVector  *parametric_eq;
   int curr_indx;
   int prev_indx;
   int endpnt_indx;
   int rval = ERR_STATUS_GOOD;
   bool done = false;
   bool point_set = false;
   ChipsVector low_bound;
   ChipsVector up_bound;
   bool use_upbound = false;
   double t;
   double tmp_pos = 0;
   int pnt_size;
   bool *log_scale;

   if (num_pnts < 2)
   {
     cmd_manager->add_err(ERR_STATUS_BAD, id, 0);
      return ERR_STATUS_BAD;
   }

   if (!(extend_line & chips_extend_start) &&
       !(extend_line & chips_extend_end))
      return ERR_STATUS_GOOD;

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   pnt_size = pnts[0].get_size () -1;

   log_scale = new bool [pnt_size];
   for (int ii=0; ii<pnt_size; ii++)
      log_scale [ii] = false;

   if (coord_system == DATA)
   {
      log_scale [0] = static_cast <ChipsDataSystem *>(coord)->get_log_flag (X_AXIS);
      log_scale [1] = static_cast <ChipsDataSystem *>(coord)->get_log_flag (Y_AXIS);
   }

   low_bound = coord->get_lower_bounds ();
   up_bound = coord->get_upper_bounds ();


   // convert the bounds to the current coordinate system
   string coord_id = "";
   if (coord->get_bounds_coord_sys () == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      coord_id = ds_coord->get_id_no_wcs ();
   }
   coord->convert_cs_to_cs (&low_bound, &low_bound, 1, 
                  coord->get_bounds_coord_sys (), coord_id,
                  coord_system, data_system_id);
   coord->convert_cs_to_cs (&up_bound, &up_bound, 1,
                  coord->get_bounds_coord_sys (), coord_id,
                  coord_system, data_system_id);
  
   for (int ii=0; ii<pnt_size; ii++)
   {
      // swap bounds such that the lower bounds are less than
      // the upper bounds
      if (low_bound[ii] > up_bound[ii])
      {
         double tmp_bound = low_bound[ii];
         low_bound[ii] = up_bound[ii];
         up_bound[ii] = tmp_bound;
      }
        
      // convert bounds to log scale if necessary 
      if (log_scale[ii] && (low_bound[ii] > 0) && (up_bound[ii]>0))
      {
         low_bound[ii] = log10(low_bound[ii]);
         up_bound[ii] = log10(up_bound[ii]);
      }
   }

   parametric_eq = new ChipsVector [pnt_size];

   // Calculate start points

   if (extend_line & chips_extend_start)
   {
      // go from point1 to point zero
      curr_indx = 0;
      prev_indx = 1;
      endpnt_indx = 0;
   }
   else if (extend_line & chips_extend_end)
   {
      // go from second to last point to last point
      curr_indx = num_pnts-1;
      prev_indx = num_pnts-2;
      endpnt_indx = 1;
   }


   // loop through twice if extending in both directions
   for (int kk=0; kk<2 && !done; kk++)
   {
      
      for (int ii=0; ii<pnt_size; ii++)
      {
         parametric_eq[ii] = ChipsVector (2);
 
         // check if log scale. 
         if (log_scale[ii])
         {
            double tmp_pnt;
            double tmp_endpoint;
       
            if (pnts[prev_indx][ii] < DBL_EPSILON)
               tmp_pnt = low_bound[ii];
            else
               tmp_pnt = log10(pnts[prev_indx][ii]);

            if (endpoints[endpnt_indx][ii] < DBL_EPSILON)
               tmp_endpoint = low_bound[ii];
            else
               tmp_endpoint = log10 (endpoints[endpnt_indx][ii]);

            if ((pnts[prev_indx][ii] >= DBL_EPSILON) ||
                (endpoints[endpnt_indx][ii] >= DBL_EPSILON))
            {
               parametric_eq[ii][0] = tmp_pnt;
               parametric_eq[ii][1] = tmp_endpoint - tmp_pnt;
            }
         }
         else
         {
            parametric_eq[ii][0] = pnts[prev_indx][ii];
            parametric_eq[ii][1] = (endpoints[endpnt_indx][ii] - pnts[prev_indx][ii]); 
         }
      }

      point_set = false;

      // Check that all points are valid in log scale. If an
      // invalid point is found skip solving the equation as
      // it will produce bad results
      for (int ii=0; ii<pnt_size; ii++)
         if (log_scale [ii] && 
             ((endpoints[endpnt_indx][ii] < DBL_EPSILON) &&
              (pnts[prev_indx][ii] < DBL_EPSILON)))
           point_set = true;

      // loop throw and see if the line crosses the x bounds then
      // check the y bounds. Break when the points are set
      for (int ii=0; ii<pnt_size && !point_set; ii++)
      {

         // For horizontal and vertical lines there will be only
         // one axis for the equation can be solved the as the
         // other axis is constant. Also don't solve for log scale
         // points which are less than or equal to zero these are
         // invalid points
         if (parametric_eq[ii][1] != 0)
         {
            double t_low, t_up;

            t_low = (low_bound[ii] - parametric_eq[ii][0])/parametric_eq[ii][1];
            t_up = (up_bound[ii] - parametric_eq[ii][0])/parametric_eq[ii][1];
            if (t_low < t_up)
            {
               t = t_up;
               use_upbound = true;
            }
            else
            {
               t = t_low;
               use_upbound = false;
            }

            // make sure the intersection is after the line segment
            // this also handles if the intersection is in the middle
	    // of the line segment
            if (t >= (1-DBL_EPSILON))
            {
               point_set = true;
               // go through all points and find a new position
               // continue to loop while all points are set
               for (int jj=0; jj<pnt_size; jj++)
               {
                  if (jj != ii)
                     tmp_pos = parametric_eq[jj][0] + parametric_eq[jj][1]*t;
                  else
                     tmp_pos = (use_upbound) ? up_bound[ii] : low_bound[ii];
                  // if point is within the boundaries set the vector
                  if ((tmp_pos >= low_bound[jj]) && (tmp_pos <= up_bound[jj]))
                     pnts[curr_indx][jj] = (log_scale[jj]) ? pow (10.0, tmp_pos) : tmp_pos;
                  else
                     point_set = false;
               }    
            }
         }
      }

      // if the point did not get set reset it back to
      // its original value as parts may have been changed
      if (!point_set)
      {
         pnts[curr_indx] = endpoints[endpnt_indx];
      }

      // only loop once for lines of two points and single extensions
      done = (extend_line != chips_extend_both);

      // go from second to last point to the last point
      curr_indx = num_pnts-1; 
      prev_indx = num_pnts-2;
      endpnt_indx = 1;
                                                       
   }

   delete []log_scale;
   delete []parametric_eq;

   return rval;
  
}

bool ChipsLine::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED)
         outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}




