// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsMatrix.hh"
#include "ChipsDefinesP.hh"
#include <iostream>
#include <cmath>
#include <float.h>

ChipsVector::ChipsVector ()
{
   this->elements = NULL;
   this->size = 0;
}

ChipsVector::ChipsVector (double x, double y)
{
   elements = new double[3];
   this->elements[0] = x;
   this->elements[1] = y;
   this->elements[2] = 1.0; //homogeneous value
   this->size = 3;
}

ChipsVector::ChipsVector (double x, double y, double z)
{
   elements = new double[4];
   this->elements[0] = x;
   this->elements[1] = y;
   this->elements[2] = z;
   this->elements[3] = 1.0; //homogeneous value
   this->size = 4;
}

ChipsVector::ChipsVector (int size)
{
   if (size <= 0){
      elements = NULL;
      return;
   }
   elements = new double[size];
   for (int ii=size; ii--;)
      elements[ii] = 0;
   elements[(int)size-1] = 1;
   this->size = size;
}


ChipsVector::ChipsVector (const ChipsVector &v)
{
   this->elements = new double [v.size];
   for (int ii=0; ii<v.size; ii++)
      this->elements[ii] = v.elements[ii];
   this->size = v.size;
}

ChipsVector::~ChipsVector ()
{
   if (this->elements)
      delete [] (this->elements);
}

void ChipsVector::operator = (const ChipsVector &v)
{
   if (this->elements && (this->size != v.size)){
      delete [] (this->elements);
      this->elements = NULL;
   }

   this->size = v.size;
  
   if (!this->elements) 
      this->elements = new double[this->size];

   for (int ii=this->size; ii--;)
      this->elements[ii] = v.elements[ii];
}


void ChipsVector::operator = (double elem)
{
   if (!this->elements)
      return;

   for (int ii=this->size; ii--;)
      this->elements[ii] = elem;
}

void ChipsVector::operator = (double *elem)
{
   if ((!this->elements) || !elem)
      return;

   for (int ii=this->size; ii--;)
      this->elements[ii] = elem[ii];
}


ChipsVector ChipsVector::operator + (const ChipsVector &v)
{
   ChipsVector out (this->size);
   if (v.get_size () != this->get_size ())
      return *this;

   for (int ii=this->size; ii--;)
      out[ii] = this->elements[ii] + v[ii];

   return out;
}

ChipsVector ChipsVector::operator - (const ChipsVector &v)
{
   ChipsVector out (this->size);
   if (v.get_size () != this->get_size ())
      return *this;

   for (int ii=this->size; ii--;)
      out[ii] = this->elements[ii] - v[ii];

   return out;
}

ChipsVector ChipsVector::operator * (double scale)
{
   ChipsVector out (this->size);
   for (int ii=this->size; ii--;)
      out[ii] = this->elements[ii]*scale;

   return out;
}


bool ChipsVector::operator == (const ChipsVector &v)
{
   bool flag = true;
   if (this->size != v.size){
      return false;
   }

   if (!this->elements)
      return false;

   for (int ii=this->size; ii--;)
      flag = (flag && dbl_equal(this->elements[ii],v.elements[ii]));

   return flag;

}


bool ChipsVector::operator != (const ChipsVector &v)
{
   bool flag = true;
   if (this->size != v.size){
      return true;
   }

   if (!this->elements)
      return true;

   for (int ii=this->size; ii--;)
      flag = (flag && (dbl_equal (this->elements[ii],v.elements[ii])));

   return !flag;

}

void ChipsVector::print ()
{
   for (int ii=0; ii<size; ii++)
      cerr <<  this->elements[ii] << " ";
   cerr << endl;
}
ChipsMatrix::ChipsMatrix ()
{
   this->elements = NULL;
   this->row = 0;
   this->col = 0;
}

ChipsMatrix::ChipsMatrix (int r, int c)
{
   /* Only allow square matrix for now */
   if (r <= c){
      this->row = r;
      this->col = r;
   }
   else{
      this->row = c;
      this->col = c;
   }

   elements = new double* [this->row];
   for (int ii=0; ii<this->row; ii++){
      this->elements[ii] = new double[this->col];
   }

   for (int ii=0; ii<this->row; ii++)
      for (int jj=0; jj<this->col; jj++)
         this->elements[ii][jj] = 0;
}

ChipsMatrix::ChipsMatrix (const ChipsMatrix &m)
{
   this->row = m.row;
   this->col = m.col;

   (this->elements) = new double* [this->row];
   for (int ii=0; ii<this->row; ii++)
      this->elements[ii] = new double[this->col];

   for (int ii=row; ii--;)
      for (int jj=col; jj--;)
         this->elements[ii][jj] = m.elements[ii][jj];
}

ChipsMatrix::~ChipsMatrix ()
{
   if (elements){
      for (int ii=this->row; ii--;)
         delete [] this->elements[ii];
      delete [] (this->elements);
   }
}

void ChipsMatrix::operator = (const ChipsMatrix &m)
{
   if (this == &m)
      return;

   if (elements){
      if ((this->row != m.row) && (this->col != m.col)){
         for (int ii=this->row; ii--;)
            delete [] this->elements[ii];
         delete [] (this->elements);

	 this->row = m.row;
    	 this->col = m.col;
	 (this->elements) = new double* [this->row];
         for (int ii=0; ii<this->row; ii++)
            this->elements[ii] = new double[this->col];
      }
   }
   else{
      this->row = m.row;
      this->col = m.col;
      (this->elements) = new double* [this->row];
      for (int ii=0; ii<this->row; ii++)
         this->elements[ii] = new double[this->col];
   }

/*   if ((this->row != m.row) && (this->col != m.col))
      return;
*/

   for (int ii=row; ii--;)
      for (int jj=col; jj--;)
         this->elements[ii][jj] = m.elements[ii][jj];

}

ChipsVector ChipsMatrix::operator * (const ChipsVector &v)
{
   ChipsVector v_out(v.get_size());
   int num_cols  = v.get_size (); 

   v_out = 0.0;
   for (int rr = 0; rr < this->row; rr++)
      for (int cc = 0; cc < num_cols; cc++)
	 v_out[rr] += (this->elements[rr][cc] * v[cc]);

   return v_out;
}

ChipsMatrix ChipsMatrix::operator * (const ChipsMatrix &m)
{
   ChipsMatrix m_out(this->row, m.get_col ());

   if (this->col != m.get_row ()){
      m_out.identity ();
      return m_out;
   }

   for (int a = 0; a < this->row; a++)
      for (int c = 0; c < m.get_col(); c++)
	 for (int b=0; b < this->col; b++)
            m_out[a][c] += (this->elements[a][b] * m[b][c]);

   return m_out;
}


void ChipsMatrix::identity ()
{
   for (int rr=this->row; rr--;)
      for (int cc=this->col; cc--;)
	 elements[rr][cc] = (rr == cc) ?  1 : 0;
}

double ChipsMatrix::determinant ()
{
   if (this->row == 2) 
      return (elements[0][0]*elements[1][1] - elements[0][1]*elements[1][0]);
   else if (this->row == 3)
      return ((elements[0][0]*elements[1][1]*elements[2][2] + 
	      elements[0][1]*elements[1][2]*elements[2][0] +
	      elements[0][2]*elements[1][0]*elements[2][1]) -
	      (elements[0][2]*elements[1][1]*elements[2][0] +
	      elements[0][1]*elements[1][0]*elements[2][2] +
	      elements[0][0]*elements[1][2]*elements[2][1]));

   ChipsMatrix m (this->row-1, this->col-1);
   double det = 0;
   int r,c;
   for (int ii=0; ii<this->col; ii++){

      r=0; c=0;
      /*populate the new matrix with cofactors */
      for (int jj=1; jj<this->row; jj++, r++){
	 c = 0;
	 for (int kk=0; kk<this->col; kk++){
	    if (kk != ii){
               m[r][c++] = this->elements[jj][kk];
            }
	 }
      }

      if (ii%2 == 0)
	 det += (elements[0][ii]*m.determinant ());
      else
	 det -= (elements[0][ii]*m.determinant ());
   }
   return det;
}

ChipsMatrix ChipsMatrix::inverse ()
{
   ChipsMatrix m (this->row, this->col);
   double det = 0;

   if (this->row == 2){
      det = determinant ();
      m[0][0] = elements[1][1]/det;
      m[0][1] = -elements[0][1]/det;
      m[1][0] = -elements[1][0]/det;
      m[1][1] = elements[0][0]/det;
      return m;
   }
   else if (this->row == 3){
      det = determinant ();
      m[0][0] = (elements[1][1]*elements[2][2]-elements[1][2]*elements[2][1])/det;
      m[0][1] = (elements[0][2]*elements[2][1]-elements[0][1]*elements[2][2])/det;
      m[0][2] = (elements[0][1]*elements[1][2]-elements[0][2]*elements[1][1])/det;
      m[1][0] = (elements[1][2]*elements[2][0]-elements[1][0]*elements[2][2])/det;
      m[1][1] = (elements[0][0]*elements[2][2]-elements[0][2]*elements[2][0])/det;
      m[1][2] = (elements[0][2]*elements[1][0]-elements[0][0]*elements[1][2])/det;
      m[2][0] = (elements[1][0]*elements[2][1]-elements[1][1]*elements[2][0])/det;
      m[2][1] = (elements[0][1]*elements[2][0]-elements[0][0]*elements[2][1])/det;
      m[2][2] = (elements[0][0]*elements[1][1]-elements[0][1]*elements[1][0])/det;

      return m;
   }

   /* using gauss-jordan elmination */
   double max;
   int maxrow;
   int cc, rr;

   m.identity ();

   for (cc = 0; cc < this->col; cc++){

      /* Through out this function cc and rr are almost
	 always the same (except when traversing the loop) 
	 rr is the pivot row we're working on */

      rr = cc;
      max = elements [rr][cc];
      maxrow = cc;
      for (rr=cc; rr<this->row; rr++){
         if (max < elements[rr][cc]){ 
 	    max = elements[rr][cc];
 	    maxrow = rr;
         }
      }

      /* singular return identity */
      if (dbl_equal (max,0)){
	 m.identity ();
	 return m;
      }

      rr = cc;

      /* swap rows */
      double tmp_elem;
      for (int c=0; c<this->col; c++){
	 tmp_elem = this->elements[rr][c];
	 this->elements[rr][c] = this->elements[maxrow][c];
	 this->elements[maxrow][c] = tmp_elem;
	 tmp_elem = m[rr][c];
	 m[rr][c] = m[maxrow][c];
	 m[maxrow][c] = tmp_elem;
      }

      /* set diagnol or row to 1 */
      tmp_elem = this->elements[rr][cc];
      for (int c=0; c<this->col; c++){
         this->elements[rr][c] /= tmp_elem;
         m[rr][c] /= tmp_elem;
      }

      /* set elements under and above the diagnol to 0 */
      for (int r=0; r<this->row; r++){
	 tmp_elem = this->elements[r][cc];
         for (int c=0; c<this->col; c++){
	    if (r != rr){
	       this->elements[r][c] -= (this->elements[rr][c] * tmp_elem);
	       m[r][c] -= (m[rr][c] * tmp_elem);
	    }
	 }
      }
   }

   return m;
}

void ChipsMatrix::set_size (int r, int c)
{
   if (elements){
      for (int ii=this->row; ii--;)
         delete [] this->elements[ii];
      delete [] (this->elements);
   }

   this->row = r;
   this->col = c;

   (this->elements) = new double* [this->row];
   for (int ii=0; ii<this->row; ii++)
      this->elements[ii] = new double[this->col];

   for (int ii=this->row; ii--;)
      for (int jj=this->col; jj--;)
         elements[ii][jj] = 0;
}


int ChipsMatrix::set_elements (double *elem)
{
   if (!elements)
      return ERR_STATUS_BAD;

   for (int ii=this->row; ii--;)
      for (int jj=this->col; jj--;)
         elements[ii][jj] = elem[jj+ii*this->col];

   return ERR_STATUS_GOOD;
}

int ChipsMatrix::set_elements (double **elem)
{
   if (!elements)
      return ERR_STATUS_BAD;

   for (int ii=this->row; ii--;)
      for (int jj=this->col; jj--;)
         elements[ii][jj] = elem[ii][jj];

   return ERR_STATUS_GOOD;
}


void ChipsMatrix::print ()
{
   for (int ii=0; ii<this->row; ii++){
      for (int jj=0; jj<this->col; jj++)
         cerr << "  " <<  elements[ii][jj] << " ";
      cerr << endl;
   }
}


