// 
//  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSOBJECT_HH
#define CHIPSOBJECT_HH

class ChipsObject;
class ChipsAxisBind;
class ChipsCmdManager;

#include <string>
#include <iostream>
#include <stdio.h>
#include "ChipsDefinesP.hh"
#include "ChipsCoordinate.hh"
#include "ChipsStateStream.hh"
#include "ChipsId.hh"


using namespace std;


class ChipsObject
{
  public:

   ChipsObject(ChipsObject* par, string val);
   ChipsObject (const ChipsObject &co);

   // Delete's the current object. Safer than calling the destructor
   virtual void object_delete (); 

   // method to get the inherited class name 
   virtual V_OBJECT_TYPE     get_object_type () { return object_type;};

   // methods for id interactions
   string     get_id()  { return id; }
   int        set_id(string val)  { id = val; return ERR_STATUS_GOOD; }
   int        get_ref_cnt () {return ref_cnt;};

   // fill in a chips id based on the object
   void  populate_chips_id (ChipsId &cid); 

   static unsigned int get_cur_ref_cnt () {return cur_ref_cnt;};
   static void set_cur_ref_cnt (unsigned int val) { cur_ref_cnt = val;};


   // methods for position interactions
   int         get_norm_pos(float &x1, float &y1, float &x2, float &y2);
   virtual int reset_position () {return  ERR_FEATURE_DNE;};
   virtual int move(COORD_SYS, const string&, double, double, int, bool)
                   { return ERR_FEATURE_DNE; }

   // methods for object visibility  
   virtual int show()        { visible = true; return ERR_STATUS_GOOD; }
   virtual int hide()        { visible = false; return ERR_STATUS_GOOD; }
   virtual int force_hide ();  // hide the object without changing visible flag
   virtual int force_show ();  // show the object without changing visible flag
   bool        set_vis_override(bool val); 
   virtual bool is_visible();
   virtual int change_visible() {return ERR_FEATURE_DNE;}

   ChipsObject* get_parent () {return parent;};
 
   // methods for coordinate system 
   void             set_chips_coord (ChipsCoordinate *cord) 
                                 {this->coord = cord;};
   ChipsCoordinate* get_chips_coord () {return coord;};
   bool             coord_system_set () {return !(coord_system == COORD_NONE);};
   COORD_SYS        get_coord_system () {return coord_system;};
   string           get_data_system_id (){return data_system_id;};
   void             get_coord_system (COORD_SYS &cs, string &coord_id) 
                                 {cs = coord_system; coord_id = data_system_id;};
   void             set_coord_system (COORD_SYS cs, const string &coord_id)                                         {coord_system = cs; data_system_id = coord_id;};

   // methods for undo/redo
   virtual int reference_object (ChipsObject*, bool change_cnt, const string &curr="") 
                                 {return ERR_STATUS_BAD;};
   virtual int dereference_object (ChipsObject*, bool change_cnt, const string &curr="") 
                                 {return ERR_STATUS_BAD;};
   virtual int clear_cmd (ChipsObject*) {return ERR_STATUS_BAD;};
   virtual void serialize (ChipsStateStream& out) {};
   virtual void serialize_ascii(ChipsStateStream& out) {};
   virtual void deserialize (ChipsStateStream& in) {};

   // misc methods
   int         get_status() { return status; }

   virtual int get_depth()  { return depth; }
   virtual int set_depth(int newdepth);
   virtual int change_depth() {return ERR_FEATURE_DNE;}

   // gui methods
   virtual void select () {};
   virtual void unselect () {};
   virtual void disable_interaction () {};
   virtual void enable_interaction () {};
   virtual bool is_interaction_blocked ();
   void block_interaction () {this->interaction_blocked = true;};
   void unblock_interaction () {this->interaction_blocked = false;};

   virtual ChipsAxisBind*   get_bind_table();
   virtual double           get_scale_factor();

  protected:

   virtual ~ChipsObject();

   string     id;         // object identifier
   ChipsObject* parent;    // parent object in hierarchy
   V_OBJECT_TYPE     object_type;  // name of the objects class

   double      nx1;        // frame normalized x1 position
   double      ny1;        // frame normalized y1 position
   double      nx2;        // frame normalized x2 position
   double      ny2;        // frame normalized y2 position

   COORD_SYS coord_system;   // The current coordinate system
   string data_system_id;    // Data system id if in data system
   ChipsCoordinate* coord;

   bool       visible;    // object visible? -- set by system
   int        status;     // failure status indicator 
   int        depth;    // depth level for current object 
   bool       vis_override;  // object's visibility been overridden (ie. depth)

   unsigned int ref_cnt;
   static unsigned int cur_ref_cnt;

   bool interaction_blocked; // whether interaction is can be enabled

};


#endif 
