// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsPlotSystem.hh"
#include "ChipsFrameSystem.hh"
#include "ChipsDataSystem.hh"

#include <algorithm>
#include <stdlib.h>
#include <sstream>

unsigned long ChipsPlotSystem::global_ref_id = 0;

ChipsPlotSystem::ChipsPlotSystem (ChipsCoordinate *par):ChipsCoordinate (par)
{
   min_position = ChipsVector (3);
   max_position = ChipsVector (3);
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
   parent->add_coord_sys (this);
   remove_data = true;
   this->old_id = "";

}

ChipsPlotSystem::ChipsPlotSystem (ChipsCoordinate *par,
			const ChipsVector &min, const ChipsVector &max):ChipsCoordinate (par)
{
   min_position = min;
   max_position = max;
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
   parent->add_coord_sys (this);
   remove_data = true;
   this->old_id = "";
}


ChipsPlotSystem::ChipsPlotSystem (const ChipsPlotSystem &ps):ChipsCoordinate (ps)
{
   this->parent = ps.parent;
   this->id = ps.id;
   this->min_position = ps.min_position;
   this->max_position = ps.max_position;
   this->data_systems = ps.data_systems;
   this->chips_objects = ps.chips_objects;
   this->remove_data = ps.remove_data;
   this->ref_id = ps.ref_id;
   this->old_id = ps.old_id;
}

ChipsPlotSystem::ChipsPlotSystem (ChipsStateStream &in, ChipsCoordinate *par):ChipsCoordinate (par)
{
   min_position = ChipsVector (3);
   max_position = ChipsVector (3);
   remove_data = true;

   this->deserialize (in);
}


ChipsPlotSystem::~ChipsPlotSystem ()
{
   if (parent)
      parent->remove_coord_sys (id);

   if (!data_systems.empty ())
   {
      map<string, ChipsDataSystem *>::iterator ii;
      for (ii = data_systems.begin (); ii!=data_systems.end (); ii++)
         ii->second->remove_parent ();
      data_systems.clear ();
   }

}


void ChipsPlotSystem::operator = (const ChipsPlotSystem &ps)
{
   if (this == &ps)
      return;

   this->min_position = ps.min_position;
   this->max_position = ps.max_position;
   this->data_systems = ps.data_systems;
   this->chips_objects = ps.chips_objects;
   this->remove_data = ps.remove_data;

}

int ChipsPlotSystem::convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id)
{
   int rval = ERR_STATUS_GOOD;
   map <string, ChipsDataSystem *>::iterator from_ds;
   map <string, ChipsDataSystem *>::iterator to_ds;
   ChipsVector range;

   if (!in || !out)
      return ERR_STATUS_BAD;

   if (size <= 0)
      return ERR_STATUS_BAD;

   if ((from_cs == PLOT_NORM) && (from_id.length () > 0) && (id != from_id))
      return ERR_STATUS_BAD;

   if ((to_cs == PLOT_NORM) && (to_id.length () > 0) && (id != to_id))
      return ERR_STATUS_BAD;

   if (to_cs == DATA){
      to_ds = data_systems.find (strip_transform_id(to_id));
      if (to_ds == data_systems.end ())
         return ERR_COORD_DS_DNE;
   }

   switch (from_cs){
      case WINDOW_NORM: // Fallthrough intended 
      case PIXEL:       // Fallthrough intended
      case PIXEL_REAL:
      case FRAME_NORM:
      {
         if (!parent)
            rval = ERR_STATUS_BAD;

	 if (!rval && ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM))){
            return parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
                                                to_cs, to_id);
	 }
	 else if (!rval)
	    rval = parent->convert_cs_to_cs (in, out, size, from_cs, from_id,
                                                FRAME_NORM, parent->get_id());
         
 	 if (rval != ERR_STATUS_GOOD)
	    return rval;

         for (int ii=0; ii<size; ii++)
            for (int jj=out[ii].get_size (); jj--;)
            {
               if (!dbl_equal (max_position[jj]-min_position[jj], 0))
                  out[ii][jj] = (out[ii][jj] - min_position[jj])/(max_position[jj]-min_position[jj]);
            }
         if (to_cs == DATA){
            rval = to_ds->second->convert_cs_to_cs (out, out, size, PLOT_NORM, id,
                                                to_cs, to_id);
         }
      }
      break;
      case PLOT_NORM:
         if ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM)){

            for (int ii=size; ii--;)
            {
               out[ii] = in[ii];
               for (int jj=in[ii].get_size (); jj--;)
               {
                  if (!dbl_equal (max_position[jj]-min_position[jj], 0))
                     out[ii][jj] = (in[ii][jj] * (max_position[jj] - min_position[jj])) + min_position[jj];
               }

            }

	    if (parent && to_cs != FRAME_NORM)
	       rval = parent->convert_cs_to_cs (out, out, size, FRAME_NORM, 
						parent->get_id(), to_cs, to_id);
	    else if (parent)
	       rval = ERR_STATUS_GOOD;
            else
               rval = ERR_STATUS_BAD;
         }
	 else if (to_cs == PLOT_NORM){
	    for (int ii=size; ii--;)
               out[ii] = in[ii];
            rval =  ERR_STATUS_GOOD;
	 }
	 else if (to_cs == DATA){
	    rval = to_ds->second->convert_cs_to_cs (in, out, size, PLOT_NORM, 
							id, to_cs, to_id);
         }
      break;
      case DATA:
         from_ds = data_systems.find (strip_transform_id (from_id));
         if (from_ds == data_systems.end ())
            return ERR_STATUS_BAD;
	 if ((to_cs == WINDOW_NORM) || 
	     (to_cs == PIXEL) ||
	     (to_cs == PIXEL_REAL) ||
	     (to_cs == FRAME_NORM) ||
	     (to_cs == PLOT_NORM) ||
	     (to_cs == DATA)){
	    rval = from_ds->second->convert_cs_to_cs (in, out, size, from_cs,
                                                from_id, PLOT_NORM, id);
	    if (rval != ERR_STATUS_GOOD)
	       return rval;

	    if ((to_cs == WINDOW_NORM) ||
                (to_cs == PIXEL) ||
		(to_cs == PIXEL_REAL) ||
		(to_cs == FRAME_NORM))
	       rval = convert_cs_to_cs (out, out, size, PLOT_NORM,
                                                id, to_cs, to_id);
	    else if (to_cs == DATA)
               rval = to_ds->second->convert_cs_to_cs (out, out, size, 
						PLOT_NORM, id, to_cs, to_id);
         }
	 else 
	    rval = ERR_STATUS_BAD;
      break;
      default: rval = ERR_STATUS_BAD;
   }
   return rval;
}

int ChipsPlotSystem::add_coord_sys (ChipsCoordinate *cs)
{
   if (cs->get_id().length () == 0)
      return ERR_STATUS_BAD;

   if (data_systems.find (cs->get_id()) != data_systems.end ())
      return ERR_STATUS_BAD;

   data_systems[cs->get_id()] = (ChipsDataSystem *)cs;
   cs->set_parent (this);

   return ERR_STATUS_GOOD;
}


int ChipsPlotSystem::remove_coord_sys (const string &id)
{
   map <string, ChipsDataSystem *>::iterator ds;

   if (!remove_data)
      return ERR_STATUS_GOOD;

   if ((ds = data_systems.find (strip_transform_id(id))) == data_systems.end ())
      return ERR_STATUS_BAD;

   ds->second->remove_parent ();
//   delete ds->second; 
   data_systems.erase(ds);
   return ERR_STATUS_GOOD;
}


ChipsCoordinate *ChipsPlotSystem::get_coord (COORD_SYS coord_sys, const string &coord_id)
{
   ChipsCoordinate *coord = NULL;
   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
      {
         if (parent)
            coord = parent->get_coord(coord_sys, coord_id); 
      }
      break;
      case COORD_NONE:
      case PLOT_NORM:
      {
         if ((coord_id.length () == 0) || (coord_id == this->id) || (coord_id == this->old_id))
	    coord = this;
         else if (parent)
         {
            int pos = coord_id.find_first_of (".", 0);
            pos = coord_id.find_first_of (".", pos+1);
            pos = coord_id.find_first_of (".", pos+1);
            string tmp_id = coord_id.substr (0, pos);

            if (this->id != tmp_id)
               coord = parent->get_coord(coord_sys, coord_id);
         }
          
      }
      break;
      case DATA:
      {
         map <string, ChipsDataSystem *>::iterator ds;

         if ((ds = data_systems.find (strip_transform_id(coord_id))) != data_systems.end ())
            coord = ds->second;
         else
         {
            // check the data old ids for backwards compatibility
            for (ds = data_systems.begin (); ds != data_systems.end (); ds++)
            {
               string old_id = ds->second->get_old_id ();
               if (strip_transform_id(coord_id,true) == old_id)
                  coord = ds->second;
            }
         }
         
         if (!coord)
         {
            int pos = coord_id.find_first_of (".", 0);
            string tmp_id;
            if (coord_id.length () > pos + 1)
               pos = coord_id.find_first_of (".", pos+1);
            if (coord_id.length () > pos + 1)
               pos = coord_id.find_first_of (".", pos+1);
            if (coord_id.length () > pos)
               tmp_id = coord_id.substr (0, pos);
 
            if (this->id != tmp_id)
               coord = parent->get_coord(coord_sys, coord_id);

         }
      }
      break;
   }

   return coord;

}

ChipsDataSystem *ChipsPlotSystem::get_data_coord (COORD_SYS coord_sys, const string &xaxis_id, const string &yaxis_id)
{
   map <string, ChipsDataSystem *>::iterator ds; 
   ChipsDataSystem *ds_coord;
   for (ds=data_systems.begin (); ds!=data_systems.end (); ds++)
   {
      ds_coord = ds->second;
      if ((ds_coord->get_xaxis_id () == xaxis_id) && (ds_coord->get_yaxis_id () == yaxis_id))
         return ds_coord;
   }
   return NULL;
}

int ChipsPlotSystem::insert (ChipsObject *co, bool use_co_info)
{
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   int rval = ERR_STATUS_BAD;
   if (co == NULL)
      return ERR_STATUS_BAD;

   // Store object local forces the object to remain
   // in this coordinate systems list. It overrides
   // the coordinate objects coord system
   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
      { 
         if (parent)
            rval = parent->insert (co, use_co_info);
      }
      break;
      case COORD_NONE:
      case PLOT_NORM:
	 if ((coord_id.length () == 0) || (coord_id == this->id)){
            chips_objects.push_back (co);
	    rval = ERR_STATUS_GOOD;
	 }
	 else
	    rval = ERR_COORD_DNE;
      break;
      case DATA:
      {
         map <string, ChipsDataSystem *>::iterator ds;

         if ((ds = data_systems.find (strip_transform_id(coord_id))) != data_systems.end ())
            rval = ds->second->insert (co, use_co_info);
	 else
	    rval = ERR_COORD_DS_DNE;
      }
      break;
   }


   return rval;
}


int ChipsPlotSystem::remove (ChipsObject *co, bool use_co_info)
{
   int rval = ERR_STATUS_BAD;
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   if (co == NULL)
      return ERR_STATUS_BAD;

   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
         if (parent)
            rval = parent->remove (co, use_co_info);
      break;
      case COORD_NONE:
      case PLOT_NORM:
      {
	 list <ChipsObject *>::iterator ii = find (chips_objects.begin (),
                                                chips_objects.end (), co);
         if (ii != chips_objects.end ()){
            chips_objects.erase (ii);
            rval = ERR_STATUS_GOOD;
         }
      }
      break;
      case DATA:
      {
         map <string, ChipsDataSystem *>::iterator ds;

         if ((ds = data_systems.find (strip_transform_id(coord_id))) != data_systems.end ())
            rval = ds->second->remove (co, use_co_info);
         else
            rval = ERR_COORD_DS_DNE;
      }
   }

   return rval;
}


bool ChipsPlotSystem::coord_exists (COORD_SYS coord_sys, const string &id)
{
   bool rval = false;

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case FRAME_NORM:
         if (parent)
            rval = parent->coord_exists(coord_sys, id);
      break;
      case COORD_NONE:
      case PLOT_NORM:
	 rval = true;
      break;
      case DATA:
      {
//         if ((id.length () != 0) && (data_systems.find (id) != data_systems.end ()))
         if (data_systems.find (strip_transform_id (id)) != data_systems.end ())
            rval = true;
      }
   }
   return rval;
}


void ChipsPlotSystem::get_coord_list (list <string> &outList)
{
   map <string, ChipsDataSystem *>::iterator ii;

   for (ii = data_systems.begin (); ii != data_systems.end (); ++ii)
      outList.push_back (ii->first);

}



int ChipsPlotSystem::modify_position (double position_min, double position_max, int axis)
{
   int rval = ERR_STATUS_GOOD;

   if (axis == X_AXIS){
      min_position[0] = position_min;
      max_position[0] = position_max;
   }
   else if (axis == Y_AXIS){

      min_position[1] = position_min;
      max_position[1] = position_max;
   }

   rval = update_obj_positions ();

   return rval;
}


int ChipsPlotSystem::update_obj_positions (bool update_all)
{
   int rval = ERR_STATUS_GOOD;

   list <ChipsObject *>::iterator ii;
   for (ii = chips_objects.begin (); ii != chips_objects.end (); ++ii){
       rval = (*ii)->reset_position ();
       if (rval != ERR_STATUS_GOOD)
          return rval;
   }

   if (update_all){
      map <string, ChipsDataSystem *>::iterator iter;
      for (iter = data_systems.begin (); iter != data_systems.end (); iter++){
         if ((rval=iter->second->update_obj_positions(update_all)) != ERR_STATUS_GOOD)
	    break;
      }
   }

   return rval;
};

void ChipsPlotSystem::delete_all_data_systems ()
{
   map <string, ChipsDataSystem *>::iterator iter = data_systems.begin ();
   remove_data = false;
   for (iter = data_systems.begin (); iter != data_systems.end (); iter++){
      delete (iter->second);
   }
   data_systems.clear ();
   remove_data = true;
}

int ChipsPlotSystem::deserialize (ChipsStateStream &in)
{
   int sz = 0;
   int rval = ERR_STATUS_GOOD;
   ChipsDataSystem *data_sys;

   if (in.get_minor_version() > 6)
   {
      in.read_long ((long *)&this->global_ref_id);
      in.read_long ((long *)&this->ref_id);
      in.read_string (&id);
   }
   else
   {
      this->ref_id = this->global_ref_id;
      this->global_ref_id++;
      in.read_string (&old_id);
      this->create_id (this->id);
   }

   in.read_bool (&visible);

   min_position = ChipsVector (3);
   max_position = ChipsVector (3);

   in.read_double (&min_position[0], 3);
   in.read_double (&max_position[0], 3);

   in.read_bool (&remove_data);

   // check for corruption
   if (in.read_checkpoint () != ERR_STATUS_GOOD)
      rval = ERR_STATEFILE_INC;
   else{

      in.read_int (&sz);

      for (int ii = 0; !rval && ii<sz; ii++){
         data_sys = new ChipsDataSystem (in, this);
         if (data_sys)
            data_systems[data_sys->get_id ()] = data_sys;
         else
            rval = ERR_STATEFILE_COR;
      }

/*      in.read_int (&sz);

      for (int ii = 0; !rval && ii<sz; ii++){
         data_sys = new ChipsDataSystem (in, this);
         if (data_sys)
            deref_data_systems[data_sys->get_id ()] = data_sys;
         else
            rval = ERR_STATEFILE_COR;
      }
*/

   }

   return rval;

};


void ChipsPlotSystem::serialize (ChipsStateStream &out)
{
   int sz = 0;

   map <string, ChipsDataSystem *>::iterator ii;

   if (out.get_minor_version() > 6)
   {
      out.write_long ((long *)&(this->global_ref_id));
      out.write_long ((long *)&(this->ref_id));
   }

   out.write_string (&id);
   out.write_bool (&visible);
   out.write_double (&min_position[0], min_position.get_size ());
   out.write_double (&max_position[0], min_position.get_size ());

   out.write_bool (&remove_data);

   out.write_checkpoint ();

   sz = data_systems.size ();
   out.write_int (&sz);

   for (ii = data_systems.begin (); ii != data_systems.end (); ii++)
      ii->second->serialize (out);
};


string ChipsPlotSystem::strip_transform_id (const string &data_id, bool old)
{
   string out_id = data_id;
   int pos = 0;
   
   if (!old)
   {
      // there should be at least one _ between plot id and axes. 
      // if there's a second then anything after it is wcs info
      pos = data_id.find ('_', 0);
      
      if (pos > 0)
         out_id = data_id.substr (0,pos);
   }
   else
   {
      pos = data_id.find ('_');

      // there should be at least one _ between plot id and axes.
      // if there's a second then anything after it is wcs info
      pos = data_id.find ('_', pos+1);

      if (pos > 0)
         out_id = data_id.substr (0,pos);
   }

   return out_id;
}

void ChipsPlotSystem::create_id (string &curr_id)
{
   stringstream tmp_id;

   if (this->parent)
      this->parent->create_id (curr_id);

   tmp_id << curr_id << "." << this->ref_id;
   curr_id = tmp_id.str ();
}

