// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsPoint.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"
#include "ChipsCommand.hh"

#include <stdio.h>

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;


ChipsPoint::ChipsPoint(ChipsObject* par, ChipsCoordinate *inCoord,
			COORD_SYS coord_sys, double xx, double yy,
			string ptid):ChipsAnnotation(par, ptid)
{
   id = ptid; 
   pnt = ChipsVector (xx, yy);
   initialize();

   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);

      // convert the data to logical coordinates if necessary
      if (coord_sys == DATA)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (&pnt, &pnt, 1, DATA, "", DATA, data_system_id);
         }

      }

   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);
}


ChipsPoint::ChipsPoint(ChipsObject* par, ChipsCoordinate *inCoord, 
   COORD_SYS coord_sys, ChipsStateStream& in):ChipsAnnotation(par, "")
{
   object_type = chips_point;
   deserialize(in);

   if (inCoord)  coord = inCoord;
   if (coord)
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_system == WINDOW_NORM) || (coord_system == PIXEL) ||
          (coord_system == PIXEL_REAL))
         coord = coord->get_coord(FRAME_NORM, "");
      else
         coord = coord->get_coord(coord_system, data_system_id);
      if (coord)
      {
         data_system_id = coord->get_id ();
         status = coord->insert (this);

         // older state files don't record the correct data_system_id
         // ensure that we get one without wcs transform
         if (coord_system == DATA)
         {
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
            if (ds_coord->get_default_transform_id () != "")
               data_system_id = ds_coord->get_id_no_wcs ();
         }
      }
   }
   initialized = true;
}


ChipsPoint::ChipsPoint(ChipsObject* par, ChipsCoordinate *inCoord,
			COORD_SYS coord_sys, double xx, double yy,
   			double zz, string ptid):ChipsAnnotation(par, ptid)
{
   id = ptid;
   pnt = ChipsVector (xx, yy, zz);
   initialize();

   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);
   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);

}


ChipsPoint::~ChipsPoint()
{
   if (coord)
      coord->remove (this);
}


void ChipsPoint::object_delete ()
{
   delete (this);
}


void ChipsPoint::initialize()
{
   if (theDefaults)
   {
      string temp;
      style   = theDefaults->get_pointstyle();
      filled  = theDefaults->point_filled();
      size    = theDefaults->get_pointsize();
      temp    = theDefaults->get_pointcolor(); 
      angle   = theDefaults->get_pointangle(); 
      depth   = theDefaults->get_pointdepth();
      if (set_color(temp))  set_color("white"); 
   }
   else
   {
      style   = chips_diamond; 
      filled  = true; 
      set_color("white"); 
      size    = 1;
      angle = 0; 
      depth = V_USE_DEFAULT_DEPTH;
   }
   if (depth == V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
   object_type = chips_point;
   point_num = 0;
   initialized = true;
   coord = NULL;
}


int ChipsPoint::move(COORD_SYS coord_sys, const string &coord_id, double xx,
                   double yy, int mode, bool rec) 
{
   int rval = ERR_STATUS_GOOD;
   ChipsVector tmp_n (0.0, 0.0);
   ChipsVector tmp_org (0.0, 0.0);
   ChipsMoveCommand* cmd = NULL;
   string csid = coord_id;

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (coord_sys == COORD_NONE)
   {
      coord_sys = coord_system;
      csid = coord->get_id ();
   }

   if (!coord->coord_exists (coord_sys, csid))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), csid.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coord_sys));
      return ERR_COORD_DNE;
   }
   if (rec)  cmd = new ChipsMoveCommand(this, pnt[0], pnt[1], xx, yy, coord_sys,
      coord_id, mode, ((ChipsPlot*)parent)->get_current_point());

   if (mode != V_MOVE_DELTA)
      tmp_org = ChipsVector(xx, yy);
   else
   {

      // convert offset and (0,0) to frame norm to get offset
      rval = coord->convert_cs_to_cs(&pnt, &tmp_org, 1, coord_system, 
				     data_system_id, coord_sys, csid);

      if (rval == ERR_STATUS_GOOD){
         // adjust norm pos by offset
         tmp_org[0] += xx;
         tmp_org[1] += yy;
      }
   }

   // convert new pos to frame norm
   rval = coord->convert_cs_to_cs(&tmp_org, &tmp_org, 1, coord_sys,
                                     csid, coord_system, data_system_id); 

   if (rval == ERR_NEG_LOG_VAL)
      cmd_manager->add_err("Cannot move an object to a position less than or equal to zero when in log scale.", rval=ERR_EDGE_BOUNDARY, id, 0);

   if (rval == ERR_STATUS_GOOD)
   {
      pnt = tmp_org;
      if ((rval == ERR_STATUS_GOOD) && initialized)
         rval = change_position ();
   }

   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if (!add_command(cmd, rval))   delete cmd;
   }
   else if (rec)
      delete(cmd);

   if (!rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_PNT);

   return rval;
}

int ChipsPoint::resize(COORD_SYS coord_sys, const string &coord_id,
                           double *bounds, bool rec)
{

   ChipsVector lower (bounds[0], bounds[1]);
   ChipsVector upper (bounds[2], bounds[3]);
   int rval;
   bool lock = cmd_manager->cmd_lock ();
   double boundingBox[4];

   if (!rec)
      cmd_manager->set_cmd_lock (true);

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (!coord->coord_exists (coord_sys, coord_id))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), coord_id.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(DATA));
      return ERR_COORD_DNE;
   }

   // convert to pixel real which is what compute_resize uses
   rval = coord->convert_cs_to_cs(&lower, &lower, 1, coord_sys,
                                     coord_id, PIXEL_REAL, "");
   rval = coord->convert_cs_to_cs(&upper, &upper, 1, coord_sys,
                                     coord_id, PIXEL_REAL, "");

   if (!rec)
      cmd_manager->set_cmd_lock (true);
   else
      cmd_manager->open_block ();


   boundingBox[0] = lower[0];
   boundingBox[1] = lower[1];
   boundingBox[2] = upper[0];
   boundingBox[3] = upper[1];

   rval = this->compute_resize (boundingBox);
   if (!rec)
      cmd_manager->set_cmd_lock (lock);
   else
      cmd_manager->close_block ();

   return rval;

}

int ChipsPoint::change_coordinate (COORD_SYS coordsys, const string& coordid)
{

   ChipsCoordinate *new_coord;
   int rval = ERR_STATUS_GOOD;
   
   // This is temporarily hard coded until a better method of searching
   // down the coordinate system tree is implemented. Otherwise objects
   // in window system must be kept in the frame coord
   if ((coordsys == WINDOW_NORM) || (coordsys == PIXEL) ||
       (coordsys == PIXEL_REAL))
      new_coord = coord->get_coord (FRAME_NORM, "");
   else
      new_coord = coord->get_coord (coordsys, coordid);

   if (!new_coord)
   {
      cmd_manager->add_err(ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coordsys));
      return ERR_COORD_DNE;
   }
    
   coord->convert_cs_to_cs (&pnt, &pnt, 1, coord_system, data_system_id, coordsys, coordid);

   rval = coord->remove (this);

   if (!rval)
      rval = new_coord->insert (this, false);

   if (!rval)
   {
      this->set_chips_coord (new_coord);
      this->set_coord_system (coordsys, coordid);

      this->change_position ();

      cmd_manager->update_mask (STATE_CHNG_PROP_PNT);
   }
   
   return ERR_STATUS_GOOD;
}

int ChipsPoint::get_position (ChipsVector &pos)
{
   int rval = ERR_STATUS_GOOD;
   pos = pnt;

   // convert the data to logical coordinates if necessary
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
         rval = ds_coord->convert_cs_to_cs (&pos, &pos, 1, DATA, data_system_id, DATA, "");
   }

   return rval;
}


int ChipsPoint::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot) || (obj == chips_point))
   {
      int mask = theDefaults->get_allpropmask();

      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();
         rval = set_property((void*)clr.c_str(), SET_POINT_SYMBOL_COLOR);
      }
      if ((mask & PREF_ALL_ANGLE) == PREF_ALL_ANGLE)
      {
         float ang = theDefaults->get_allangle();
         rval = set_property(static_cast<void*>(&ang), SET_POINT_SYMBOL_ANGLE);
      }
      if ((mask & PREF_ALL_FILL) == PREF_ALL_FILL)
      {
         bool fil = theDefaults->get_allfill();
         rval = set_property(static_cast<void*>(&fil), SET_POINT_SYMBOL_FILL);
      }
      if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         rval = set_property(static_cast<void*>(&vis), SET_POINT_VISIBLE);
      }
      if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
         int dep = theDefaults->get_alldepth();
         rval = set_property(static_cast<void*>(&dep), SET_POINT_DEPTH);
      }
   }
   return rval;
}


int ChipsPoint::set_property(void* in_val, int in_property, bool rec) 
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;

   switch(in_property)
   {
      case SET_POINT_SYMBOL_FILL:
      {
         bool* bval = static_cast<bool*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            filled, *bval, ((ChipsPlot*)parent)->get_current_point());
         if ((rval = set_filled(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_POINT_SYMBOL_STYLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            (int)style, *ival, ((ChipsPlot*)parent)->get_current_point());
         if ((rval = set_symbolstyle((V_SYMBOL_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_POINT_SYMBOL_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            size, *ival, ((ChipsPlot*)parent)->get_current_point());
         if ((rval = set_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_POINT_SYMBOL_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            angle, *fval, ((ChipsPlot*)parent)->get_current_point());
         if ((rval = set_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_POINT_SYMBOL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
	    get_color(), clr, ((ChipsPlot*)parent)->get_current_point());
	 if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_POINT_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
             oval, *ival, ((ChipsPlot*)parent)->get_current_point());
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else
         {
            cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "point visibility", *ival,0, 1);
            delete cmd;
         }
      }
      break;
      case SET_POINT_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival;
         if (rec)
         {
            cmd = new ChipsSetCommand(this, in_property,
               oval, newval, ((ChipsPlot*)parent)->get_current_point());
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
               if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, in_property);
      break;
   }
   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

   if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_PNT);

   return rval;
}

int ChipsPoint::get_property(void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_POINT_SYMBOL_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_POINT_SYMBOL_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_symbolstyle();
      }
      break;
      case SET_POINT_SYMBOL_SIZE:
      {
         int *iv = (int*)out_val;
         *iv = get_size();
      }
      break;
      case SET_POINT_SYMBOL_ANGLE:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_angle();
      }
      break;
      case SET_POINT_SYMBOL_FILL:
      {
         int *iv = (int*)out_val;
         *iv = get_filled();
      }
      break;
      case SET_POINT_VISIBLE:
      {
         int *iv = (int*)out_val;
         *iv = is_visible();
      }
      break;
      case SET_POINT_DEPTH:
      {
         int *iv = (int*)out_val;
         *iv = get_depth();
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }

   return rval;
}


int ChipsPoint::set_size(int val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine size of point
   if ((val < V_SIZE_MIN_VAL) || (val > V_SIZE_MAX_VAL))
      cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "point size", val,
                                  V_SIZE_MIN_VAL, V_SIZE_MAX_VAL);
   else
   {
      size = val;
      change_size(); 
   }
   return rval;
}


int ChipsPoint::set_symbolstyle(V_SYMBOL_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to verify symbol type is valid
   if (valid_symbol_type(val))
   {
      style = val;
      change_style();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "point style", (int)val, "chips_circle|chips_cross|chips_diamond|chips_none|chips_plus|chips_square|chips_point_type|chips_uptriangle|chips_downtriangle");

   return rval;
}

int ChipsPoint::set_filled(bool in_filled)
{
   filled = in_filled; 
   change_filled(); 
   return 0;
}

int ChipsPoint::reset_position ()
{
   return change_position ();

}

int ChipsPoint::redraw()
{
   cout << " ChipsPoint - virtual redraw() " << endl; 
   return 1;
}

int ChipsPoint::change_size()
{
   cout << " ChipsPoint - virtual change_size() " << endl; 
   return 1;
}

int ChipsPoint::change_style()
{
   cout << " ChipsPoint - virtual change_style() " << endl; 
   return 1;
}

int ChipsPoint::change_filled()
{
   cout << " ChipsPoint - virtual change_filled() " << endl; 
   return 1;
}

int ChipsPoint::change_position()
{
   cout << " ChipsPoint - virtual change_position() " << endl; 
   return 1;
}

int ChipsPoint::change_visible()
{
   return ERR_FEATURE_DNE;
}

int ChipsPoint::compute_resize( double *bounds)

{
  return ERR_FEATURE_DNE;
}



void ChipsPoint::serialize(ChipsStateStream& out)
{
   string tmp;
   int sz;
   
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   tmp = get_color();
   out.write_string(&tmp);
   out.write_float(&angle);
   sz = (int) coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);
   sz = pnt.get_size ();
   out.write_int (&sz);
   for (int ii=0; ii<sz; ii++)
      out.write_double(&(pnt[ii]));
   out.write_int(&size);
   out.write_int(&depth);
   out.write_bool(&filled);
   sz = (int)style;
   out.write_int(&sz);
   out.write_bool (&visible);
}

void ChipsPoint::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];
   const char *cp;

   tmp = get_color();
   if (theDefaults->get_pointcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("color=");
      att.append(tmp);
   }
   if (size != theDefaults->get_pointsize())
   {
      sprintf(buf, "size=%1d", size);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (angle != theDefaults->get_pointangle())
   {
      sprintf(buf, "angle=%2.1f", angle);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   int dep = theDefaults->get_pointdepth();
   if (depth != dep &&
       (dep != V_USE_DEFAULT_DEPTH || depth!=((ChipsPlot*)parent)->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (filled != theDefaults->point_filled())
   {
      if (att.length()) att.append(" ");
      if (filled)  att.append("fill=true");
         else      att.append("fill=false");
   }
   if (theDefaults->get_pointstyle() != style &&
       ((cp = v_symbol_to_string(style))!=NULL))
   {
      if (att.length()) att.append(" ");
      att.append("style=");
      att.append(cp);
   }
   if (coord_system != DATA &&
       (cp = v_coord_sys_to_enum_string(coord_system)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("coordsys=");
      att.append(cp);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);

   
   ChipsVector tmp_pnt = pnt;
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
      {
         data_system_id = ds_coord->get_id_no_wcs ();
         ds_coord->convert_cs_to_cs (&pnt, &tmp_pnt, 1, DATA, data_system_id, DATA, "");
      }
   }
   out <<"add_point("<<tmp_pnt[0]<<", "<<tmp_pnt[1];
   if (att.length()) out <<", \""<<att<<"\"";
   out <<")"<< endl;
}

void ChipsPoint::deserialize(ChipsStateStream& in)
{
   int len;
   string tmp;
   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = (unsigned int) len;
   in.read_string(&tmp);
   set_color(tmp);
   in.read_float(&angle);
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);
   in.read_int((int*)&len);
   pnt = ChipsVector (len);
   for (int ii=0; ii<len; ii++)
      in.read_double(&(pnt[ii]));
   in.read_int(&size);
   in.read_int(&depth);
   in.read_bool(&filled);
   in.read_int(&len);
   style = (V_SYMBOL_TYPE)len;
   in.read_bool (&visible);
}

int ChipsPoint::show ()
{
  visible = true;
  return change_visible ();
}

int ChipsPoint::hide ()
{
  visible = false;
  return change_visible ();
}

int ChipsPoint::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}

int ChipsPoint::change_depth()
{
   return ERR_FEATURE_DNE;
}

bool ChipsPoint::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED)
         outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}




