// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsPolygon.hh"
#include "ChipsCmdManager.hh"
#include "ChipsDataSystem.hh"


#include <stdio.h>
#include <math.h>

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;


// add polygon region specified as series of x,y points 
ChipsPolygon::ChipsPolygon(ChipsObject* par, ChipsCoordinate *inCoord,
   COORD_SYS coord_sys, double* xcoords, double* ycoords, unsigned int cnt,
   string val):ChipsRegion(par, inCoord, coord_sys, val)
{
   int rr;
   id = val; 

   if (xcoords && ycoords)
   {
      num_pnts = cnt;
      if ((xcoords[0] == xcoords[cnt -1]) && (ycoords[0] == ycoords[cnt -1]))
         pnts = new ChipsVector[num_pnts];
      else 
      {
         // last point was omitted, so add the first point to close the polyon
         pnts = new ChipsVector[++num_pnts];
         pnts[cnt] = ChipsVector(xcoords[0], ycoords[0]);
      }

      for (rr = cnt; rr--; )
         pnts[rr] = ChipsVector (xcoords[rr], ycoords[rr]);


   }
   else   num_pnts = 0;

   // convert the data to logical coordinates if necessary
   if ((coord_sys == DATA) && coord)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
         ds_coord->convert_cs_to_cs (pnts, pnts, num_pnts, DATA, "", DATA, ds_coord->get_id_no_wcs ());
      data_system_id = ds_coord->get_id_no_wcs ();
         
   }

   if (num_pnts > 0)
   {
      lower_bounds = pnts[0];
      upper_bounds = pnts[0];
   }

   for (rr = num_pnts; rr--; )
   {
      for (int jj=0; jj<pnts[rr].get_size (); jj++)
      {
         if (pnts[rr][jj] < lower_bounds[jj])
            lower_bounds[jj] = pnts[rr][jj];
         if (pnts[rr][jj] > upper_bounds[jj])
            upper_bounds[jj] = pnts[rr][jj];
      }
   }


   if (status != ERR_STATUS_GOOD)
      return;

   if (check_degenerate()) num_pnts = 0;
}


// add regular N-gon region specified by number of sides, center point, and radius
ChipsPolygon::ChipsPolygon(ChipsObject* par, ChipsCoordinate *inCoord,
   COORD_SYS coord_sys, int sides, double xcoord, double ycoord,
   double radius, string val):ChipsRegion(par, inCoord, coord_sys, val)
{
   int rr;
   id = val; 
   num_pnts = sides + 1;
   if (sides > 2) 
   {
      double ang, del_ang, offset;

      center = ChipsVector (xcoord, ycoord);
      center_set = true;
      radinp = radius;

      // convert the data to logical coordinates if necessary
      if ((coord_sys == DATA) && coord)
      {

         ChipsVector tmp_radius(xcoord+radius, ycoord+radius);
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            data_system_id = ds_coord->get_id_no_wcs ();
            ds_coord->convert_cs_to_cs (&center, &center, 1, DATA, "", DATA, data_system_id);
            ds_coord->convert_cs_to_cs (&tmp_radius, &tmp_radius, 1, DATA, "", DATA, data_system_id);

//            radius = sqrt (((center[0] - tmp_radius[0])*(center[0] - tmp_radius[0])) + ((center[1] - tmp_radius[1])*(center[1] - tmp_radius[1])));
            radius = fabs(center[0] - tmp_radius[0]);
         }
      }


      pnts = new ChipsVector[num_pnts];

      del_ang = 2.0 * V_PI / (double)sides; 

      if (sides == 4)      offset = 45;  // align square parallel to x axis
      else if (sides % 2)  offset = 90;  // align odd N-gons parallel to x axis
      else                 offset = 0; 
      offset = (angle + offset) / 180.0 * V_PI; 

      for (rr = 0; rr < sides; rr++)
      {
         ang = (rr) * del_ang + offset;
         pnts[rr] = ChipsVector (radius * cos(ang) + center[0], 
                                 radius * sin(ang) + center[1]);

         if (rr == 0)
         {
            lower_bounds = pnts[0];
            upper_bounds = pnts[0];
         }
         else
         {
            for (int jj=0; jj<pnts[rr].get_size (); jj++)
            {
               if (pnts[rr][jj] < lower_bounds[jj])
                  lower_bounds[jj] = pnts[rr][jj];
               if (pnts[rr][jj] > upper_bounds[jj])
                  upper_bounds[jj] = pnts[rr][jj];
            }
         }
      } 
      pnts[sides] = ChipsVector(pnts[0][0], pnts[0][1]);

   }
   else   num_pnts = 0;

   if (check_degenerate()) num_pnts = 0;
}


ChipsPolygon::ChipsPolygon(ChipsObject* par, ChipsCoordinate *inCoord, 
   COORD_SYS coord_sys, ChipsStateStream& in):ChipsRegion(par,in)
{
   object_type = chips_region;
   deserialize(in);

   if (inCoord)  coord = inCoord;
   if (coord)
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_system == WINDOW_NORM) || (coord_system == PIXEL) ||
          (coord_system == PIXEL_REAL))
         coord = coord->get_coord(FRAME_NORM, "");
      else
         coord = coord->get_coord(coord_system, data_system_id);

      if (coord)
      {
         data_system_id = coord->get_id ();
         status = coord->insert (this);

         // older state files don't record the correct data_system_id
         // ensure that we get one without wcs transform
         if (coord_system == DATA)
         {
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
            if (ds_coord->get_default_transform_id () != "")
               data_system_id = ds_coord->get_id_no_wcs ();
         }
      }
   }
   initialized = true;
}


ChipsPolygon::~ChipsPolygon()
{
}

void ChipsPolygon::serialize(ChipsStateStream& out)
{
   string tmp;
   int sz;
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   tmp = get_color();
   out.write_string(&tmp);
   tmp = get_fillcolor();
   out.write_string(&tmp);
   sz = (int) coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);

   out.write_int((int*)&num_pnts);
   out.write_int((int*)&num_pnts);
   if (pnts)
      sz = pnts[0].get_size ();
   else
      sz = 0;
   out.write_int (&sz);
   for (unsigned int ii=0; ii<num_pnts; ii++)
      for (int jj=0; jj<sz; jj++)
         out.write_double(&(pnts[ii][jj]));
   out.write_float(&thickness);
   sz = (int)linestyle;
   out.write_int(&sz);
   sz = (int)fillstyle;
   out.write_int(&sz);
   out.write_char((char*)fillpattern, 128);
   out.write_double(&opacity);
   out.write_float(&angle);
   out.write_int(&depth);
   out.write_bool(&source);
   out.write_bool(&include);
   out.write_bool(&center_set);
   if (center_set)
   {
      sz = center.get_size ();
      out.write_int (&sz);
      for (int ii=0; ii<sz; ii++)
         out.write_double(&(center[ii]));

      out.write_double (&radinp);
   }
   out.write_bool (&visible);
}

void ChipsPolygon::serialize_ascii(ChipsStateStream& out)
{
   string att = "", tmp;
   char buf[80];
   const char *cp;

   tmp = get_fillcolor();
   if (theDefaults->get_regionfillcolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("fill.color=");
      att.append(tmp);
   }
   tmp = get_color();
   if (theDefaults->get_regionfillstyle() != fillstyle)
   {
      if ((cp = v_fill_to_string(fillstyle)) != NULL)
      {
         if (att.length()) att.append(" ");
         att.append("fill.style=");
         att.append(cp);
      }
   }
   if (theDefaults->get_regionedgestyle() != linestyle)
   {
      if ((cp = v_line_to_string(linestyle)) != NULL)
      {
         if (att.length()) att.append(" ");
         att.append("edge.style=");
         att.append(cp);
      }
   }
   if (theDefaults->get_regionedgecolor() != tmp)
   {
      if (att.length()) att.append(" ");
      att.append("edge.color=");
      att.append(tmp);
   }
   if (theDefaults->get_regionopacity() != opacity)
   {
      if (att.length()) att.append(" ");
      sprintf(buf, "opacity=%2.1f", opacity);
      att.append(buf);
   }
   if (thickness != theDefaults->get_regionedgethickness())
   {
      if (att.length()) att.append(" ");
      if (thickness != 0.5)
      {
         sprintf(buf,"edge.thickness=%-1.0f", thickness);
         att.append(buf);
      }
      else
         att.append("edge.thickness=0.5");
   }
   int dep = theDefaults->get_pointdepth();
   if (depth != dep &&
       (dep!=V_USE_DEFAULT_DEPTH || depth!=parent->get_depth()))
   {
      sprintf(buf, "depth=%d", depth);
      if (att.length()) att.append(" ");
      att.append(buf);
   }
   if (coord_system != DATA &&
       (cp = v_coord_sys_to_enum_string(coord_system)) != NULL)
   {
      if (att.length()) att.append(" ");
      att.append("coordsys=");
      att.append(cp);
   }
   if (att.length()) att.append(" ");
   att.append("id=");
   att.append(id);
   if (center_set)
   {
      double tmp_radius = radinp;
      ChipsVector tmp_center = center;
      ChipsVector tmp_pnt = pnts[0];

      // convert the data to logical coordinates if necessary
      if ((coord_system == DATA) && coord)
      {
         ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
         if (ds_coord->get_default_transform_id () != "")
         {
            ds_coord->convert_cs_to_cs (&center, &tmp_center, 1, DATA, data_system_id, DATA, "");

            ds_coord->convert_cs_to_cs (&tmp_pnt, &tmp_pnt, 1, DATA, data_system_id, DATA, "");

         }
      }

      // if the original radius is not available try to calculate the radius
      if (tmp_radius == -1)
         tmp_radius = sqrt (((tmp_center[0] - tmp_pnt[0])*(tmp_center[0] - tmp_pnt[0])) + ((tmp_center[1] - tmp_pnt[1])*(tmp_center[1] - tmp_pnt[1])));

      // regular n-gon
      out << "add_region("<< num_pnts-1<<","<<tmp_center[0]<<","<<tmp_center[1]<<","<<tmp_radius;
      if (att.length()) out <<",\""<<att<<"\"";
      out <<")"<<endl;
   }
   else
   {
      ChipsDataSystem *ds_coord = NULL;
      ChipsVector *tmp_pnts = pnts;
      bool delete_tmp_pnts = false;
      // convert the data to logical coordinates if necessary
      if ((coord_system == DATA) && coord)
         ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord && (ds_coord->get_default_transform_id () != ""))
      {
         tmp_pnts = new ChipsVector[num_pnts];
         ds_coord->convert_cs_to_cs (pnts, tmp_pnts, num_pnts, DATA, data_system_id, DATA, "" );
         delete_tmp_pnts = true;
      }
      
      // list of points
      unsigned int zz;
      out << "add_region([";
      for (zz = 0; zz < num_pnts-2; zz++) out << tmp_pnts[zz][0] << ",";
      out << tmp_pnts[num_pnts-2][0] <<"],[";
      for (zz = 0; zz < num_pnts-2; zz++) out << tmp_pnts[zz][1] << ",";
      out << tmp_pnts[num_pnts-2][1] <<"]";
      if (att.length()) out <<",\""<<att<<"\"";
      out <<")"<< endl;
      
      if (delete_tmp_pnts)
         delete []tmp_pnts;
   }
}

void ChipsPolygon::deserialize(ChipsStateStream& in)
{
   string tmp;
   int len;

   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = (unsigned int) len;
   in.read_string(&tmp);
   set_color(tmp);
   in.read_string(&tmp);
   set_fillcolor(tmp);
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);
   in.read_int((int*)&num_pnts);
   in.read_int((int*)&len);

   if ((unsigned int)len == num_pnts)
   {
      in.read_int((int*)&len);
      pnts = new ChipsVector[num_pnts];
      for (unsigned int ii=0; ii<num_pnts; ii++)
      {
         pnts[ii] = ChipsVector (len);
         for (int jj=0; jj<len; jj++)
            in.read_double(&(pnts[ii][jj]));

         if (ii == 0)
         {
            lower_bounds = pnts[0];
            upper_bounds = pnts[0];
         }
         else
         {
            for (int jj=0; jj<pnts[ii].get_size (); jj++)
            {
               if (pnts[ii][jj] < lower_bounds[jj])
                  lower_bounds[jj] = pnts[ii][jj];
               if (pnts[ii][jj] > upper_bounds[jj])
                  upper_bounds[jj] = pnts[ii][jj];
            }
         }

      }
   }
   in.read_float(&thickness);
   in.read_int(&len);
   linestyle = (V_LINE_TYPE)len;
   in.read_int(&len);
   fillstyle = (V_FILL_TYPE)len;
   in.read_char((char*)fillpattern, 128);
   in.read_double(&opacity);
   in.read_float(&angle);
   in.read_int(&depth);
   in.read_bool(&source);
   in.read_bool(&include);
   in.read_bool(&center_set);
   if (center_set)
   {
      in.read_int (&len);
      center = ChipsVector (len);
      for (int ii=0; ii<len; ii++)
         in.read_double(&(center[ii]));

      if (((in.get_major_version() == 3) && (in.get_minor_version () >= 4)) ||
             (in.get_major_version() > 3))
         in.read_double (&radinp);
      else
         radinp = -1;
   }
   in.read_bool (&visible);
}



