// 
//  Copyright (C) 2004,2005,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsRegion.hh"
#include "ChipsPlot.hh"
#include "ChipsCmdManager.hh"
#include "ChipsCommand.hh"
#include "ChipsFillPattern.hh"


#include <stdio.h>
#include <math.h>

extern ChipsFillPattern *fill_patterns;
extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;


ChipsRegion::ChipsRegion(ChipsObject* par, ChipsCoordinate *inCoord, 
                COORD_SYS coord_sys, string val):ChipsObject(par, val)
{
   initialized = false;
   initialize();

   if (inCoord)
   {
      coord = inCoord;
      coord_system = coord_sys;
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         data_system_id = "";
      else
         data_system_id = coord->get_id ();
      status = coord->insert (this);
   }
   else
      cmd_manager->add_err(status=ERR_COORD_DNE, id, 1, coord_sys);

}


ChipsRegion::ChipsRegion(ChipsObject* par, ChipsStateStream& in):ChipsObject(par,"")
{
   object_type = chips_region;
   angle = 0;
}


ChipsRegion::~ChipsRegion()
{
   delete []pnts;

   if (coord)
      coord->remove (this);
}

void ChipsRegion::object_delete ()
{
   delete (this);
}

void ChipsRegion::initialize()
{
   if (theDefaults)
   {
      string temp;
      linestyle  = theDefaults->get_linestyle();
      temp       = theDefaults->get_regionedgecolor();
      if (set_color(temp))  set_color("green");
      temp       = theDefaults->get_regionfillcolor();
      if (set_fillcolor(temp))  set_fillcolor("green");
      thickness  = theDefaults->get_regionedgethickness();
      linestyle  = theDefaults->get_regionedgestyle(); 
      fillstyle  = theDefaults->get_regionfillstyle(); 
      angle      = theDefaults->get_regionangle();
      opacity    = theDefaults->get_regionopacity();
      depth      = theDefaults->get_regiondepth();

   }
   else 
   {
      set_color("green"); 
      set_fillcolor("green"); 
      thickness = 1; 
      linestyle = chips_solid; 
      fillstyle = chips_solidfill; 
      angle = 0; 
      opacity = 0.5; 
      depth = V_USE_DEFAULT_DEPTH;
   }
   if (depth == V_USE_DEFAULT_DEPTH) depth = ((ChipsPlot*)parent)->get_depth();
   label = NULL; 
   object_type = chips_region;
   initialized = true;
   coord = NULL;
   center = ChipsVector (0,0);
   center_set = false;

   set_fillstyle(fillstyle); 
}

int ChipsRegion::change_coordinate (COORD_SYS coordsys, const string& coordid)
{
  ChipsCoordinate *new_coord;
   int rval = ERR_STATUS_GOOD;

   // This is temporarily hard coded until a better method of searching
   // down the coordinate system tree is implemented. Otherwise objects
   // in window system must be kept in the frame coord
   if ((coordsys == WINDOW_NORM) || (coordsys == PIXEL) ||
       (coordsys == PIXEL_REAL))
      new_coord = coord->get_coord (FRAME_NORM, "");
   else
      new_coord = coord->get_coord (coordsys, coordid);

   if (!new_coord)
   {
      cmd_manager->add_err(ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coordsys));
      return ERR_COORD_DNE;
   }

   coord->convert_cs_to_cs (pnts, pnts, num_pnts, coord_system, data_system_id, coordsys, coordid);
   coord->convert_cs_to_cs (&center, &center, 1, coord_system, data_system_id, coordsys, coordid);

   coord->convert_cs_to_cs (&lower_bounds, &lower_bounds, 1, coord_system, data_system_id, coordsys, coordid);
   coord->convert_cs_to_cs (&upper_bounds, &upper_bounds, 1, coord_system, data_system_id, coordsys, coordid);


   rval = coord->remove (this);

   if (!rval)
      rval = new_coord->insert (this, false);

   if (!rval)
   {
      this->set_chips_coord (new_coord);
      this->set_coord_system (coordsys, coordid);
      cmd_manager->update_mask (STATE_CHNG_PROP_RGN);
   }

   if (!rval)
      rval = this->update_clipping ();

   return rval;

}

int ChipsRegion::get_position (ChipsVector &pos)
{
   if (center_set)
      pos = center;
   else
      pos = pnts[0];

   // convert the data to logical coordinates if necessary
   if (coord_system == DATA)
   {
      ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>(coord);
      if (ds_coord->get_default_transform_id () != "")
         ds_coord->convert_cs_to_cs (&pos, &pos, 1, DATA, data_system_id, DATA, "");
   }

   return ERR_STATUS_GOOD;
}


int ChipsRegion::move(COORD_SYS coord_sys, const string &coord_id, double xx,
                   double yy, int mode, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsMoveCommand* cmd = NULL;
   string csid = coord_id;

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }

   if (coord_sys == COORD_NONE)
   {
      coord_sys = coord_system;
      csid = coord->get_id ();
   }

   if (!coord->coord_exists (coord_sys, csid))
{
      if (coord_sys == DATA)
         cmd_manager->add_err(rval=ERR_COORD_DS_DNE, id, 0, v_coord_sys_to_string(DATA), csid.c_str ());
      else
         cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string(coord_sys));

      return ERR_COORD_DNE;
   }


   // convert each point 
   ChipsVector tmp_n(3);
   ChipsVector tmp_org(3);

   ChipsVector off;
   int rr;

   // If the center is set move based off the center of the region
   // otherwise move based off the first point in the region definition
   if (center_set)
      off = center;
   else
      off = pnts[0];

   if (rec)  cmd = new ChipsMoveCommand(this, off[0], off[1], xx, yy, coord_sys,
      coord_id, mode, ((ChipsPlot*)parent)->get_current_region());

   if (mode != V_MOVE_DELTA)
   {
      tmp_n = ChipsVector (xx, yy);
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, coord_sys,
                                     csid, coord_system, data_system_id);
   }
   else
   {
      // convert offset and (0,0) to frame norm to get offset 
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, coord_sys,
                                     csid, FRAME_NORM, "");
      tmp_n = ChipsVector (xx, yy);
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_n, 1, coord_sys,
                                     csid, FRAME_NORM, "");
      tmp_org[tmp_org.get_size()-1] = 0;
      tmp_n -= tmp_org;
      // convert original pos to frame norm and add offset  
      rval = coord->convert_cs_to_cs(&off, &tmp_org, 1, coord_system, 
				     data_system_id, FRAME_NORM, "");
      tmp_org[tmp_org.get_size()-1] = 0;
      tmp_n += tmp_org; 
      // convert adjusted frame norm pos to orignal coord sys
      rval = coord->convert_cs_to_cs(&tmp_n, &tmp_org, 1, FRAME_NORM, "",
				     coord_system, data_system_id);
   }

   if (rval == ERR_STATUS_GOOD)
   {
      ChipsVector translation(tmp_org-off);
      for (rr = 0; rr < num_pnts; rr++)
         pnts[rr] += translation;
      lower_bounds += translation;
      upper_bounds += translation;
      if (center_set)  center = tmp_org;


      // Update the data back points actually changed position
      // This is not the most efficient way of doing this as it
      // involves recreating the polygon from scratch
      rval = update_data ();
   }


   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if (!add_command(cmd, rval))   delete cmd;
   }
   else if (rec)
      delete(cmd);

   if (!rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_RGN);

   return rval;
}


int ChipsRegion::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot) || (obj == chips_region))
   {
      int mask = theDefaults->get_allpropmask();
      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();
         rval = set_property((void*)clr.c_str(), SET_REGN_EDGE_COLOR);
         rval = set_property((void*)clr.c_str(), SET_REGN_FILL_COLOR);
      }
      if ((mask & PREF_ALL_THICKNESS) == PREF_ALL_THICKNESS)
      {
         float vis = theDefaults->get_allthickness();
         rval = set_property(static_cast<void*>(&vis), SET_REGN_EDGE_THICKNESS);
      }
      if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         rval = set_property(static_cast<void*>(&vis), SET_REGN_VISIBLE);
      }
      if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
         int dep = theDefaults->get_alldepth();
         rval = set_property(static_cast<void*>(&dep), SET_REGN_DEPTH);
      }
   }
   return rval;
}

int ChipsRegion::set_property(void* in_val, int in_property, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;

   switch(in_property)
   {
      case SET_REGN_EDGE_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_color(), clr, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_FILL_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_fillcolor(), clr, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_fillcolor(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_EDGE_STYLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            (int)linestyle, *ival, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_linestyle((V_LINE_TYPE)*ival)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_FILL_STYLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            (int)fillstyle, *ival, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_fillstyle((V_FILL_TYPE)*ival, rec)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_FILL_STYLE2:
      {
	 unsigned char* uc = (unsigned char*)in_val; 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            fillpattern, uc, 128, 
	    ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_fillstyle(uc)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_EDGE_THICKNESS:
      {
         float* fval  =  static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            thickness, *fval, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_thickness(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            angle, *fval, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break; 
      case SET_REGN_OPACITY:
      {
         double* dval = static_cast<double*>(in_val);
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            opacity, *dval, ((ChipsPlot*)parent)->get_current_region());
         if ((rval = set_opacity(*dval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if (!add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
      }
      break;
      case SET_REGN_VISIBLE:
      {
         int*   ival  =  static_cast<int*>(in_val);
	 int oval = visible;
	 if (rec)  cmd = new ChipsSetCommand(this, in_property,
	            oval, *ival, ((ChipsPlot*)parent)->get_current_region());
	 if ((*ival) == 1)
	 {
	    show();
	    if (rec)
               if (!add_command(cmd, rval))   delete cmd;
	 }
	 else if ((*ival) == 0)
	 {
	    hide();
	    if (rec)
               if (!add_command(cmd, rval))   delete cmd;
	 }
	 else
	 {
	    cmd_manager->add_err(rval=ERR_INT_MINMAX, id, 0, "region visibility", *ival,0, 1);
	    delete cmd;
	 }
      }
      break; 
      case SET_REGN_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
	 int newval, oval = depth;
         if ((*ival) == V_USE_DEFAULT_DEPTH)
            newval = ((ChipsPlot*)parent)->get_depth();
         else
            newval = *ival; 

	 if (rec)
         {
            cmd = new ChipsSetCommand(this, in_property,
	       oval, newval, ((ChipsPlot*)parent)->get_current_region());
            ((ChipsPlot*)parent)->remove_object_from_depth_table(this);
         }
         if ((rval = set_depth(newval)) == ERR_STATUS_GOOD)
         {
            if (rec && !add_command(cmd, rval))   delete cmd;
         }
         else  delete cmd;
         if (rec) ((ChipsPlot*)parent)->add_object_to_depth_table(this);
      }
      break;
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, in_property);
      break;
   }

   // need to add object id for errors
   if (cmd_manager && rval)
      cmd_manager->add_object_id_to_errs(id);

  if (cmd_manager && !rval)
      cmd_manager->update_mask (STATE_CHNG_PROP_RGN);

   return rval;
}

int ChipsRegion::get_property(void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_REGN_EDGE_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_REGN_EDGE_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_linestyle();
      }
      break; 
      case SET_REGN_FILL_STYLE:
      {
         int *iv = (int*)out_val;
         *iv = get_fillstyle();
      }
      break; 
      case SET_REGN_EDGE_THICKNESS:
      {
         float *fv = (float*)out_val;
         *fv = get_thickness();
      }
      break;
      case SET_REGN_FILL_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_fillcolor();
      }
      break;
      case SET_REGN_ANGLE:
      {
         float *fv = (float*)out_val;
         *fv = get_angle();
      }
      break; 
      case SET_REGN_OPACITY:
      {
         double *dv = (double*)out_val;
         *dv = get_opacity();
      }
      break; 
      case SET_REGN_DEPTH:
      {
         int *iv = (int*)out_val;
         *iv = get_depth();
      }
      break; 
      case SET_REGN_VISIBLE: 
      {
         int *ival = (int*)out_val;
         *ival = is_visible();
      }
      break; 
      default:
         cmd_manager->add_err (rval=ERR_UNKNOWN_PROPERTY,id, 0, property);
      break;
   }
   return rval;
}

int ChipsRegion::reset_position ()
{
   return change_position ();
}


int ChipsRegion::set_angle(float in_angle)
{
   float old_angle = angle;

   // convert angle (in degrees) to range 0-360 
   angle = fmod(in_angle, (float)360.0); 
   if (angle < 0.0)  angle = 360.0 - angle;

   // if angle has changed, call virtual
//   if (old_angle != angle)   change_rotation(); 
   if (old_angle != angle)
   {
      ChipsVector center;
      ChipsMatrix rotation(3,3);
      float cos_theta = cos ((angle - old_angle) * V_PI/180.0);
      float sin_theta = sin ((angle - old_angle) * V_PI/180.0);

      center = (this->upper_bounds + this->lower_bounds) * .5;

      rotation.identity ();
      rotation[0][0] = cos_theta;
      rotation[0][1] = -sin_theta;
      rotation[1][0] = sin_theta;
      rotation[1][1] = cos_theta;

      rotation[0][2] = -center[0]*cos_theta + center[1]*sin_theta + center[0];
      rotation[1][2] = -center[0]*sin_theta - center[1]*cos_theta + center[1];

      for (int ii=num_pnts; ii--;)
         pnts[ii] = rotation * pnts[ii];

      lower_bounds = pnts[0];
      upper_bounds = pnts[0];

      for (int ii=num_pnts; ii--;)
      {
         for (int jj=pnts[0].get_size (); jj--;)
         {
            if (pnts[ii][jj] < lower_bounds[jj])
               lower_bounds[jj] = pnts[ii][jj];
            if (pnts[ii][jj] > upper_bounds[jj])
               upper_bounds[jj] = pnts[ii][jj];
         }
      }

      update_data ();

//      change_rotation ();

//      change_position ();

   }

   return 1; 
}

string ChipsRegion::get_label_text()
{
   static string blank = "";
   if (label)
   {
      return label->get_text();
   }
   else return blank; 
}

int ChipsRegion::set_label_text(string in_val)
{
   int rval = ERR_STATUS_GOOD;

   if (label)
   {
      if (in_val != label->get_text())
      {
         label->set_text(in_val.c_str()); 
      }
   }
   else
   {
      // Don't have a label- should we create one or err out?
   }
   return rval; 
} 


int ChipsRegion::set_opacity(double in_val)
{
   int rval = ERR_STATUS_GOOD;
   if ((in_val < 0.0) || (in_val > 1.0))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "opacity", in_val,
                                  0.0, 1.0);
   else 
   {
      opacity = in_val; 
      if (initialized)  rval = change_opacity();
   }
   return rval;
}


int ChipsRegion::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_color();
   return rval; 
}


int ChipsRegion::set_fillcolor(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = fillcolor.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (initialized)  rval = change_fillcolor();
   return rval; 
}


int ChipsRegion::set_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX, id, 0, "thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      thickness = val;
      if (initialized)  change_thickness();
   }
   return rval;
}

int ChipsRegion::set_linestyle(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_line_type(val))
   {
      linestyle = val;
      if (initialized)  change_linestyle();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_SUBSET, id, 0, "line style", (int)val,"chips_solid|chips_dot|chips_noline|chips_longdash|chips_shortdash|chips_dotlongdash|chips_dotshortdash|chips_shortdashlongdash");

   return rval;
}

int ChipsRegion::set_linestyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if ((rval = c2_string_to_line (val.c_str(), &linestyle,
       "line style")) == ERR_STATUS_GOOD)
   {
      if (initialized)  change_linestyle();
   }

   return rval;
}

int ChipsRegion::set_fillstyle(V_FILL_TYPE val, bool record)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_fill_type(val))
   {
     record = record && ((val >= chips_userfill1 && val <= chips_userfill3) || 
        (fillstyle >= chips_userfill1 && fillstyle <= chips_userfill2)); 
      fillstyle = val;
     set_property((void*)fill_patterns->get_pattern(fillstyle), SET_REGN_FILL_STYLE2,record);
   }
   else
      cmd_manager->add_err(rval=ERR_BAD_FILLSTYLE2, id, 0, (int)val);
   return rval;
}

int ChipsRegion::set_fillstyle(unsigned char* bpattern)
{
   int rval = ERR_STATUS_GOOD;

   if (bpattern)
   {
      memcpy(fillpattern, bpattern, 128);
      if (initialized)  change_fillstyle();
   }
   else
      rval = ERR_BAD_FILLSTYLE; 
   return rval;
}

int ChipsRegion::set_fillstyle(string val)
{
   int rval = ERR_STATUS_GOOD;

   if ((rval = c2_string_to_fill (val.c_str(), &fillstyle,
       "fill style")) == ERR_STATUS_GOOD)
   {
     set_property((void*)fill_patterns->get_pattern(fillstyle), SET_REGN_FILL_STYLE2,0);
   }
   return rval;
}


bool ChipsRegion::check_segment_intersection(
   ChipsVector& pnta, ChipsVector& pntb, ChipsVector& pntc, ChipsVector& pntd)
{
   double t1,t2, delta;
   double xba,yba,xdc,ydc,xca,yca;
   bool intersect = false;

   xba = pntb[0] - pnta[0];    yba = pntb[1] - pnta[1];
   xdc = pntd[0] - pntc[0];    ydc = pntd[1] - pntc[1];
   xca = pntc[0] - pnta[0];    yca = pntc[1] - pnta[1];
   t1=xca*ydc-yca*xdc;
   t2=xca*yba-yca*xba;

   if ((delta = xba*ydc-yba*xdc) != 0)
   {
      double u = t1/delta,  v = t2/delta;

      /* lines intersect (excluding intersecting end points) */
      if ( u<=1 && u>=0 && v<=1 && v>=0 )
         intersect =
            !((pnta[0] == pntc[0] && pnta[1] == pntc[1]) ||
              (pnta[0] == pntd[0] && pnta[1] == pntd[1]) ||
              (pntb[0] == pntc[0] && pntb[1] == pntc[1]) ||
              (pntb[0] == pntd[0] && pntb[1] == pntd[1]));
   }
   else if ((t1!=0) && (t2!=0))
   {
      /* line are parallel. */
      intersect = false;
   }
   else
   {
      /* lines are collinear */
      double a,b,c,d;
      int index = (pnta[0] == pntb[0]);
      a = (pnta[index] < pntb[index])? pnta[index] : pntb[index];
      b = (pnta[index] > pntb[index])? pnta[index] : pntb[index];
      c = (pntc[index] < pntd[index])? pntc[index] : pntd[index];
      d = (pntc[index] > pntd[index])? pntc[index] : pntd[index];
      intersect = (((d > a) && (d < b)) || ((c > a) && (c < b)));
   }
   return intersect;
}


bool ChipsRegion::check_degenerate()
{
   bool bval = false;

   // check for intersecting edges 
   for (int rr = 0; rr < num_pnts - 1; rr++)
   {
      for (int cc = rr + 1; cc < num_pnts - 1; cc++)
      {
	 // check for intersecting segments or 
	 // vertices containing more than two edges
         if ((check_segment_intersection(pnts[rr], pnts[rr+1], 
	      pnts[cc], pnts[cc+1])) ||
	     ((pnts[rr][0] == pnts[cc][0]) && (pnts[rr][1] == pnts[cc][1]) &&
              ((rr > 0) || (cc != num_pnts -1))))
         {
            bval = true;
            cc = rr = num_pnts; 
            cmd_manager->add_err(status=ERR_REGION_CREATE_FAIL, id, 0);
         }
      }
   }
   return bval; 
}


int ChipsRegion::set_depth(int newdepth)
{
   int rval = ERR_STATUS_GOOD;
   if (newdepth < 0)
      cmd_manager->add_err(rval=ERR_NEG_DEPTH,id,0,newdepth);
   else
   {
      depth = newdepth;
      if (initialized)  rval = change_depth();
   }
   return rval;
}


void ChipsRegion::serialize(ChipsStateStream& out)
{
   out.write_string(&id);
   out.write_bool (&visible);
}

void ChipsRegion::deserialize(ChipsStateStream& in)
{
   in.read_string(&id);
   in.read_bool (&visible);
}

int ChipsRegion::change_depth()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_opacity()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_fillcolor()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_rotation()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_thickness()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_linestyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_fillstyle()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_position()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::change_visible()
{
   return ERR_FEATURE_DNE;
}


int ChipsRegion::redraw()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::show ()
{
  visible = true;
  return change_visible ();
}

int ChipsRegion::hide ()
{
  visible = false;
  return change_visible ();
}


int ChipsRegion::update_data()
{
   return ERR_FEATURE_DNE;
}

int ChipsRegion::update_clipping ()
{
   return ERR_FEATURE_DNE;
}



bool ChipsRegion::add_command (ChipsCommand *inCmd, int &outStatus)
{
   bool success = true;
   if ((outStatus = cmd_manager->add_cmd(inCmd)) != ERR_STATUS_GOOD)
   {
      if (outStatus == ERR_CMD_LOCKED)
         outStatus = ERR_STATUS_GOOD;
      success = false;
   }
   return success;
}

