// 
//  Copyright (C) 2006,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsStateStream.hh"
#include "ChipsErrs.hh"
#include "ChipsDefines.hh"

//define static members
int   ChipsStateStream::file_endian = 0;
int   ChipsStateStream::dbl_sz = 8;
int   ChipsStateStream::flt_sz = 4;
int   ChipsStateStream::lng_sz = 4;
int   ChipsStateStream::int_sz = 4;
int   ChipsStateStream::sht_sz = 2;
int   ChipsStateStream::bln_sz = 1;
int   ChipsStateStream::chr_sz = 1;
int   ChipsStateStream::ckpt_sz = 4;
int   ChipsStateStream::my_endian = 0;
int   ChipsStateStream::my_dbl_sz = sizeof(double);
int   ChipsStateStream::my_flt_sz = sizeof(float);
int   ChipsStateStream::my_lng_sz = sizeof(long);
int   ChipsStateStream::my_int_sz = sizeof(int);
int   ChipsStateStream::my_sht_sz = sizeof(short);
int   ChipsStateStream::my_bln_sz = sizeof(bool);
int   ChipsStateStream::my_chr_sz = sizeof(char);
int   ChipsStateStream::my_ckpt_sz = sizeof(int);


ChipsStateStream::ChipsStateStream(const char *s, ios_base::openmode 
   mode, int opts):fstream(s, mode)
{
   error = ERR_STATUS_GOOD; 
   tempfnum = tempnum = 1; 

   if (is_open())
   {
      char cc; 
      int csize = sizeof(cc); 
      long endian_check= 1;
      my_endian   = !(*((char *)(&endian_check)));

      if ((mode & ios::binary) != ios::binary)
      {
         precision(16);  // set default precision to handle doubles  
         mymode = 1; 
	 // ascii state mode -no nothing 
      }
      else if ((mode & ios::out) == ios::out)
      {
         mymode = 0; 
	 // write out header info- endian flag and type sizes
         char hdr[] = "chips2-state";
         write(hdr, 13);

         cc =  my_endian;
	 write((char*)&cc, csize);
	 cc = my_dbl_sz;
	 write((char*)&cc, csize);
	 cc = my_flt_sz;
	 write((char*)&cc, csize);
	 cc = my_int_sz;
	 write((char*)&cc, csize);
	 cc = my_bln_sz;
	 write((char*)&cc, csize);

         int sz = V_CHECKPOINT_VAL;
         write_int(&sz);
         sz = vmajor = V_STATE_MAJ_VER; 
         write_int(&sz);
         sz = vminor = V_STATE_MIN_VER; 
         write_int(&sz);
  
         // write out additional types
         cc = my_lng_sz;
         write((char*)&cc, csize);
         cc = my_sht_sz;
         write((char*)&cc, csize);
         cc = my_chr_sz;
         write((char*)&cc, csize);
      }
      else if ((mode & ios::in) == ios::in)
      {
         // get header info- endian and type sizes
         char hdr[13];

         read((char*)&hdr, 13);
         if (strcmp(hdr, "chips2-state"))
         {
            error = ERR_NOT_STATEFILE;
            close();
         }
         else
         {
            read((char*)&cc, csize);
            file_endian = (int)cc; 
            read((char*)&cc, csize);
            dbl_sz = (int)cc; 
            read((char*)&cc, csize);
            flt_sz = (int)cc; 
            read((char*)&cc, csize);
            int_sz = (int)cc; 
            read((char*)&cc, csize);
            bln_sz = (int)cc; 

            int ckpnt;
            read_int(&ckpnt, 1);
            if (ckpnt != V_CHECKPOINT_VAL) 
            {
               close();
               error = ERR_NOT_STATEFILE;
            }
            else
            {
               read_int(&vmajor, 1);
               read_int(&vminor, 1);
               if (!version_compatible(V_STATE_MAJ_VER, V_STATE_MIN_VER))
                  error = ERR_STATEFILE_MAJ; 
            }

            // read in additional types
            read((char*)&cc, csize);
            lng_sz = (int)cc;
            read((char*)&cc, csize);
            sht_sz = (int)cc;
            read((char*)&cc, csize);
            chr_sz = (int)cc;
         }
      }
      if (s)  fname = s;
   }
   else   error = ERR_FILE_OPEN;
}


ChipsStateStream::~ChipsStateStream()
{
}


int ChipsStateStream::read_double(double* dd, int cnt)
{
   int rr, len, rval = ERR_STATUS_GOOD;
   double *td = new double[cnt]; 
   len = cnt * dbl_sz; 
   read((char*)&td[0], len);

   if (my_endian != file_endian)
   {
      union
      {
         double d;
         unsigned char b[8];
      } dat1, dat2;

      // endian change- swap byte orderings
      for (rr = cnt; rr--; )
      {
         dat1.d = td[rr];
         dat2.b[0] = dat1.b[7];
         dat2.b[1] = dat1.b[6];
         dat2.b[2] = dat1.b[5];
         dat2.b[3] = dat1.b[4];
         dat2.b[4] = dat1.b[3];
         dat2.b[5] = dat1.b[2];
         dat2.b[6] = dat1.b[1];
         dat2.b[7] = dat1.b[0];
         dd[rr] = dat2.d;
      }
   } 
   else
   {
      // same endian type- copy over for now
      for (rr = cnt; rr--; )   dd[rr] = td[rr]; 
   }
   delete[] td;
   return rval; 
}


int ChipsStateStream::read_float(float* ff, int cnt)
{
   int rr, len, rval = ERR_STATUS_GOOD;
   float *tf = new float[cnt]; 
   len = cnt * flt_sz; 
   read((char*)&tf[0], len);

   if (my_endian != file_endian)
   {
      union
      {
         float f;
         unsigned char b[4];
      } dat1, dat2;

      // endian change- swap byte orderings
      for (rr = cnt; rr--; )
      {
         dat1.f = tf[rr];
         dat2.b[0] = dat1.b[3];
         dat2.b[1] = dat1.b[2];
         dat2.b[2] = dat1.b[1];
         dat2.b[3] = dat1.b[0];
         ff[rr] = dat2.f;
      }
   } 
   else
   {
      // same endian type- copy over for now
      for (rr = cnt; rr--; )   ff[rr] = tf[rr]; 
   }
   delete[] tf;
   return rval; 
}


int ChipsStateStream::read_long(long* ii, int cnt)
{
   int rr, len, msk, rval = ERR_STATUS_GOOD;
   long *ti = new long[cnt];
   len = cnt * lng_sz;
   read((char*)&ti[0], len);

   // QUESTION: What to do if we're going from a 64 bit long type
   // to a 32 bit long type. What happens if the endians are different.

   if (my_endian != file_endian)
   {
      // endian change- swap byte orderings
      for (rr = cnt; rr--; )
      {
         ii[rr] = 0;
         for (int jj=lng_sz, pos=0; --jj; pos++)
         {
            msk = 0xff << (jj*8);
            if (jj >= lng_sz / 2)
               ii[rr] |= ((ti[rr] & msk) >> ((jj-pos)*8));
            else
               ii[rr] |= ((ti[rr] & msk) << ((pos-jj)*8));
         }
      }
   }
   else
   {
      // same endian type- copy over for now
      for (rr = cnt; rr--; )   ii[rr] = (long)ti[rr];
   }
   delete[] ti;
   return rval;
}


int ChipsStateStream::read_int(int* ii, int cnt)
{
   unsigned int *ti = new unsigned int[cnt]; 
   int rr, len = cnt * int_sz; 
   read((char*)&ti[0], len);

   if (my_endian != file_endian)
   {
      // endian change- swap byte orderings
      for (rr = cnt; rr--; )
      {
         ii[rr] = ((((ti[rr]) & 0xff000000) >> 24) | \
                   (((ti[rr]) & 0x00ff0000) >> 8)  | \
                   (((ti[rr]) & 0x0000ff00) << 8)  | \
                   (((ti[rr]) & 0x000000ff) << 24));
      }
   } 
   else
   {
      // same endian type- copy over for now
      for (rr = cnt; rr--; )   ii[rr] = (int)ti[rr]; 
   }
   delete[] ti;
   return ERR_STATUS_GOOD; 
}


int ChipsStateStream::read_short(short* ii, int cnt)
{
   unsigned short *ti = new unsigned short[cnt];
   int rr, len = cnt * sht_sz;
   read((char*)&ti[0], len);

   if (my_endian != file_endian)
   {
      // endian change- swap byte orderings
      for (rr = cnt; rr--; )
         ii[rr] = (((ti[rr] & 0xff00) >> 8) | ((ti[rr] & 0x00ff) << 8));
   }
   else
   {
      // same endian type- copy over for now
      for (rr = cnt; rr--; )   ii[rr] = (short)ti[rr];
   }
   delete[] ti;
   return ERR_STATUS_GOOD;
}


int ChipsStateStream::read_bool(bool* bb, int cnt)
{
   int len = cnt * bln_sz; 
   unsigned char *tb = new unsigned char[len]; 
   read((char*)&tb[0], len);
   bool ans; 
   for (int rr = cnt; rr--; )
   {
      ans = false;
      for (int cc = bln_sz; cc--; )   ans |= (tb[--len]);
      bb[rr] = ans; 
   }
   delete[] tb;
   return ERR_STATUS_GOOD; 
}


int ChipsStateStream::read_char(char* bb, int cnt)
{
   int len = cnt * chr_sz;
   read((char*)&bb[0], len);
   return ERR_STATUS_GOOD;
}


int ChipsStateStream::read_string(string* ss)
{
   int rval = ERR_STATUS_GOOD;
   if (ss)
   {
      int ln;
      read_int(&ln);
      if (ln)
      {
         char* sval = new char[ln];
         read(sval, ln);
         *ss = sval;
         delete[] sval; 
      }
      else 
         *ss = "\0";
   }
   else  rval = ERR_STATUS_BAD;
   return rval;
}


int ChipsStateStream::read_checkpoint()
{
   int ck, ckswap, rval = ERR_STATUS_GOOD;
   read((char*)&ck, ckpt_sz);
   
   // swap byte ordering if endian change
   if (my_endian != file_endian)
      ckswap = (((ck & 0xff000000) >> 24) | \
                ((ck & 0x00ff0000) >> 8)  | \
                ((ck & 0x0000ff00) << 8)  | \
                ((ck & 0x000000ff) << 24));
   else
      ckswap = ck;

   if (ckswap != V_CHECKPOINT_VAL)   rval = ERR_STATEFILE_COR;

   return rval; 
}


int ChipsStateStream::write_double(double* dd, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)dd, cnt*my_dbl_sz);
   return rval; 
}


int ChipsStateStream::write_float(float* ff, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)ff, cnt*my_flt_sz);
   return rval; 
}


int ChipsStateStream::write_long(long* ii, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)ii, cnt*my_lng_sz);
   return rval;
}


int ChipsStateStream::write_int(int* ii, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)ii, cnt*my_int_sz);
   return rval; 
}


int ChipsStateStream::write_short(short* ii, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)ii, cnt*my_sht_sz);
   return rval;
}


int ChipsStateStream::write_bool(bool* bb, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write((char*)bb, cnt*my_bln_sz);
   return rval; 
}


int ChipsStateStream::write_char(char* ii, int cnt)
{
   int rval = ERR_STATUS_GOOD;
   write(ii, cnt*my_chr_sz);
   return rval;
}


int ChipsStateStream::write_string(string* ss)
{
   int rval = ERR_STATUS_GOOD;
   if (ss)
   {
      int ln = ss->length() +1; 
      write_int(&ln);
      write((char*)ss->c_str(), ln); 
   }
   else  rval = ERR_STATUS_BAD;
   return rval;
}


int ChipsStateStream::write_checkpoint()
{
   int rval = ERR_STATUS_GOOD;
   int ck = V_CHECKPOINT_VAL;
   write((char*)&ck, sizeof(ck));
   return rval;
}

void ChipsStateStream::append_to_template(string fileinfo)
{
   if (fileinfo.length())
   {
      tempdata.append(fileinfo);
      tempdata.append("\n");
   }
}

string ChipsStateStream::use_tempname()
{
   char buf[80];
   sprintf(buf, "MyData_%1d", tempnum++);
   tempname = buf;
   return tempname;
}

