// 
//  Copyright (C) 2004-2009,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsSystem.hh"
#include "ChipsEngine.hh"
#include "ChipsMessages.hh"
#include "ChipsMatrix.hh"
#include "VtkEngine.hh"
#include <stdio.h>
#include <sstream>

#ifndef HEADLESS
#include "ChipsEventManager.hh"
#include "ChipsEventObserver.hh"
#include "glib.h"
ChipsEventManager *event_manager=NULL;

class ChipsSystemEventObserver : ChipsEventObserver
{
public:
   ChipsSystemEventObserver ()
   {
      if (event_manager)
      {
        this->register_event (ChipsEventManager::PickEvent);
        this->register_event (ChipsEventManager::RangeZoomEvent);
        this->register_event (ChipsEventManager::RangeZoomXEvent);
        this->register_event (ChipsEventManager::RangeZoomYEvent);
      }

      this->pick_set = false;
      this->range_set = false;
   }
   
   virtual void execute_event (int event, void *data)
   {
      switch (event)
      {
         case ChipsEventManager::PickEvent:
         {
            CHIPS_PICK_EVT_P pick_event = (CHIPS_PICK_EVT_P)data;
            this->pick_evt.x = pick_event->x;
            this->pick_evt.y = pick_event->y;
            this->pick_evt.x_str = pick_event->x_str;
            this->pick_evt.y_str = pick_event->y_str;
            this->pick_evt.event = pick_event->event;
            this->pick_evt.key = pick_event->key;
            this->pick_set = true;
         }
         break;
         case ChipsEventManager::RangeZoomEvent:
         case ChipsEventManager::RangeZoomXEvent:
         case ChipsEventManager::RangeZoomYEvent:
         {
            this->range_set = true;
         }
      }
   }

   void reset_range_set () { this->range_set = false; }
   bool is_range_set () { return this->range_set; }

   void reset_pick_set () { this->pick_set = false; }

   bool get_pick_event (CHIPS_PICK_EVT &out_pick_evt)
   {
      if (!this->pick_set)
         return false;

      out_pick_evt.x = pick_evt.x;
      out_pick_evt.y = pick_evt.y;
      out_pick_evt.x_str = pick_evt.x_str;
      out_pick_evt.y_str = pick_evt.y_str;
      out_pick_evt.event = pick_evt.event;
      out_pick_evt.key = pick_evt.key;
      return true;
   }


protected:
   CHIPS_PICK_EVT pick_evt;
   bool pick_set;
   bool range_set;
};


#endif

ChipsSystem::ChipsSystem(bool load_prefs)
{
   eng = new VtkEngine(load_prefs);
#ifndef HEADLESS
   event_manager = new ChipsEventManager (eng);
#endif
}

  
ChipsSystem::~ChipsSystem()
{
   if (eng)
   {
      delete eng; 
      eng = NULL;
   }
#ifndef HEADLESS
   if (event_manager)
   {
      delete event_manager;
      event_manager = NULL;
   }
#endif
}


int ChipsSystem::load_preferences(string in_file)
{
   string filepath = in_file;
   // resolve the path to the file, if not provided
   if (((in_file.rfind('/') == string::npos) || (in_file.find("./") == 0) ||
       (in_file.find("../") == 0)) && in_file.length())
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(in_file);
      }
   }
   return eng->load_preferences(filepath);
}


int ChipsSystem::save_preferences(string in_file)
{
   string filepath = in_file;
   // resolve the path to the file, if not provided
   if (filepath.length())
   {
      if ((filepath.rfind('/') == string::npos) || (filepath.find("./") == 0) ||
          (filepath.find("../") == 0))
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
            filepath.append("/");
            filepath.append(in_file);
         }
      }
   }
   else
   {
      // if no path was provided, we should attempt to write to the
      // user's .chips.rc file provided that clobber is enabled.
      filepath = getenv("HOME");
      if (filepath.length()) filepath.append("/");
      else   return ERR_RESFLE_PATH; 
      filepath.append(".chips.rc");
   }
   return eng->save_preferences(filepath);
}

int ChipsSystem::set_preference(string pref, string value)
{
   return eng->set_preference(pref, value);
}


char* ChipsSystem::get_preference(string pref)
{
   return eng->get_preference(pref);
}


int ChipsSystem::set_line_stipple(V_LINE_TYPE line, int val)
{
   return eng->set_line_stipple(line, val);
}


int ChipsSystem::get_line_stipple(V_LINE_TYPE line, int &out_val)
{
   return eng->get_line_stipple(line, out_val);
}


int ChipsSystem::load_state(string fname)
{
   string filepath = fname;
   // resolve the path to the file, if not provided
   if (((fname.rfind('/') == string::npos) || (fname.find("./") == 0) ||
       (fname.find("../") == 0)) && fname.length())
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(fname);
      }
   }
   return eng->load_state(filepath);
}


int ChipsSystem::save_state(string fname, int mode)
{
   string filepath = fname;
   // resolve the path to the file, if not provided
   if (((fname.rfind('/') == string::npos) || (fname.find("./") == 0) ||
       (fname.find("../") == 0)) && fname.length())
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(fname);
      }
   }
   return eng->save_state(filepath, false, false, mode);
}


int ChipsSystem::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE start,
   string property,  string value)  
{
   set_id_spec(cid);
   return eng->set_cascading_property(id_spec, start, property, value);
}


int ChipsSystem::move(ChipsId* cid, V_OBJECT_TYPE obj, double xx, double yy, 
   int mode)
{
   set_id_spec(cid);
   return eng->move(id_spec, obj, xx, yy, mode);
}


int ChipsSystem::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj, int& out_val)
{
   set_id_spec(cid);
   return eng->get_object_count(id_spec, obj, out_val);
}


int ChipsSystem::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   set_id_spec(cid);
   return eng->get_object_coordinfo(id_spec, obj, out_val, out_id);
}


int ChipsSystem::add_window(ChipsId* cid, string attlist)
{
   int rval;
   set_id_spec(cid);
   if ((rval = eng->add_window(id_spec, attlist)) == ERR_STATUS_GOOD)
      cid->reset_frame_id(""); 
   return rval; 
}


int ChipsSystem::add_window(ChipsId* cid, float ww, float hh, 
   string format, string attlist)
{
   int rval;
   set_id_spec(cid);
   if ((rval = eng->add_window(id_spec, ww, hh, format, attlist)) == 
        ERR_STATUS_GOOD)
      cid->reset_frame_id(""); 
   return rval;
}


int ChipsSystem::delete_window(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_window(id_spec);
}


int ChipsSystem::load_font(string fontname, string fontpath, 
   V_FONTSTYLE_TYPE style)
{
   string filepath = fontpath;
   // resolve the path to the file, if not provided
   if (fontpath.length() && ((fontpath.rfind('/') == string::npos) ||
       (fontpath.find("./") == 0) || (fontpath.find("../") == 0)))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(fontpath);
      }
   }
   return eng->load_font(fontname, fontpath, style);
}

char* ChipsSystem::info(ChipsId* cid, V_OBJECT_TYPE obj, bool descend)
{
   set_id_spec(cid, 1);
   return eng->info(id_spec, obj, descend);
}

char* ChipsSystem::info_coordinate(ChipsId* cid)
{
   set_id_spec(cid, 1);
   return eng->info_coordinate(id_spec);
}

char* ChipsSystem::info_depth(ChipsId* cid, int depth)
{
   set_id_spec(cid, 1);
   return eng->info_depth(id_spec, depth);
}


char* ChipsSystem::info_bound_axes(ChipsId* cid)
{
   set_id_spec(cid, 1);
   return eng->info_bound_axes(id_spec);
}


char* ChipsSystem::info_current()
{
   return eng->list_current();
}


char* ChipsSystem::get_current_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->get_current_frame(id_spec);
}


int ChipsSystem::set_current_window(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_window);
}


int ChipsSystem::set_current_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_frame);
}


int ChipsSystem::set_current_image(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_image);
}

int ChipsSystem::set_current_colorbar(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_colorbar);
}

int ChipsSystem::set_current_plot(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_plot);
}


int ChipsSystem::set_current_curve(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_curve);
}

int ChipsSystem::set_current_contour(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_contour);
}


int ChipsSystem::set_current_histogram(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_histogram);
}


int ChipsSystem::set_current_axis(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_axis);
}


int ChipsSystem::set_current_line(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_line);
}


int ChipsSystem::set_current_label(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_label);
}


int ChipsSystem::set_current_point(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_point);
}


int ChipsSystem::set_current_legend(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_legend);
}

int ChipsSystem::set_current_region(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->set_current(id_spec, chips_region);
}



int ChipsSystem::add_frame(ChipsId* cid, string attlist)
{
   set_id_spec(cid);
   return eng->add_frame(id_spec, attlist);
}


int ChipsSystem::add_frame(ChipsId* cid, float x1, float y1, 
                          float x2, float y2, string attlist)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD) 
   {
      set_id_spec(cid);
      rval = eng->add_frame(id_spec, x1, y1, x2, y2, attlist);
   }
   return rval;
}


int ChipsSystem::delete_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_frame(id_spec);
}


int ChipsSystem::set_framemode(ChipsId* cid, int mode)
{
   set_id_spec(cid);
   return eng->set_framemode(id_spec, mode);
}


int ChipsSystem::tile(ChipsId* cid, unsigned int cols, unsigned int rows)
{
   set_id_spec(cid);
   return eng->tile(id_spec, cols, rows);
}


int ChipsSystem::row_tile(ChipsId* cid, unsigned int rows)
{
   set_id_spec(cid);
   return eng->row_tile(id_spec, rows);
}


int ChipsSystem::col_tile(ChipsId* cid, unsigned int cols)
{
   set_id_spec(cid);
   return eng->col_tile(id_spec, cols);
}


int ChipsSystem::prev_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->prev_frame(id_spec);
}


int ChipsSystem::next_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->next_frame(id_spec);
}


int ChipsSystem::reposition_frame(ChipsId* cid, float x1,
                                  float y1, float x2, float y2)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD)  
   {
      set_id_spec(cid);
      rval = eng->reposition_frame(id_spec, x1, y1, x2, y2);
   }
   return rval;
}


int ChipsSystem::display_depth(ChipsId* cid, int depth)
{
   set_id_spec(cid);
   return eng->display_depth(id_spec, depth);
}


int ChipsSystem::hide_depth(ChipsId* cid, int depth)
{
   set_id_spec(cid);
   return eng->hide_depth(id_spec, depth);
}

int ChipsSystem::collapse_depths(ChipsId* cid, int scnt, int* srcs, int dest)
{
   set_id_spec(cid);
   return eng->collapse_depths(id_spec, srcs, scnt, dest);
}


int ChipsSystem::set_default_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_DEFAULT_DEPTH);
}


int ChipsSystem::get_default_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_DEFAULT_DEPTH);
}


int ChipsSystem::shuffle(ChipsId* cid, V_OBJECT_TYPE obj, V_SHUFFLE_TYPE way)
{
   set_id_spec(cid);
   return eng->rearrange_depth(id_spec, obj, (int)way);
}

int ChipsSystem::shuffle_forward(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   return eng->rearrange_depth(id_spec, obj, (int)chips_forward);
}

int ChipsSystem::shuffle_backward(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   return eng->rearrange_depth(id_spec, obj, (int)chips_backward);
}

int ChipsSystem::shuffle_front(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   return eng->rearrange_depth(id_spec, obj, (int)chips_front);
}

int ChipsSystem::shuffle_back(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   return eng->rearrange_depth(id_spec, obj, (int)chips_back);
}


int ChipsSystem::reposition_plot(ChipsId* cid, float x1,
                                 float y1, float x2, float y2)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->reposition_plot(id_spec, x1, y1, x2, y2);
   }
   return rval;
}


int ChipsSystem::adjust_grid_gaps(ChipsId* cid, float xgap, float ygap)
{
   set_id_spec(cid);
   return eng->adjust_grid_gaps(id_spec, xgap, ygap);
}


int ChipsSystem::adjust_grid_xgap(ChipsId* cid, float xgap)
{
   set_id_spec(cid);
   return eng->adjust_grid_xgap(id_spec, xgap);
}

int ChipsSystem::adjust_grid_ygap(ChipsId* cid, float ygap)
{
   set_id_spec(cid);
   return eng->adjust_grid_ygap(id_spec, ygap);
}


int ChipsSystem::adjust_grid_xrelsize(ChipsId* cid, int col, float ratio)
{
   set_id_spec(cid);
   return eng->adjust_grid_xrelsize(id_spec, col, ratio);
}


int ChipsSystem::adjust_grid_yrelsize(ChipsId* cid, int row, float ratio)
{
   set_id_spec(cid);
   return eng->adjust_grid_yrelsize(id_spec, row, ratio);
}

int ChipsSystem::adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* sizes)
{
   set_id_spec(cid);
   return eng->adjust_grid_xrelsizes(id_spec, cnt, sizes);
}


int ChipsSystem::adjust_grid_yrelsizes(ChipsId* cid, int cnt, float* sizes)
{
   set_id_spec(cid);
   return eng->adjust_grid_yrelsizes(id_spec, cnt, sizes);
}


int ChipsSystem::grid_objects(ChipsId* cid, int cols, int rows, float xgap,
   float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   return eng->grid_objects(id_spec, cols, rows, xgap, ygap, usage,
                            lcnt, lids);
}


int ChipsSystem::row_grid_objects(ChipsId* cid, int cnt, int* vals,
    float xgap, float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   return eng->grid_objects(id_spec, 0, cnt, vals, xgap, ygap, 
                            usage, lcnt, lids);
}

int ChipsSystem::col_grid_objects(ChipsId* cid, int cnt, int* vals,
    float xgap, float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   return eng->grid_objects(id_spec, 1, cnt, vals, xgap, ygap,
                            usage, lcnt, lids);
}


int ChipsSystem::swap_object_positions(ChipsId* cid, string lid2)
{
   set_id_spec(cid);
   return eng->swap_object_positions(id_spec, lid2);
}


int ChipsSystem::view_single_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->view_single_frame(id_spec);
}


int ChipsSystem::view_placed_frame(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->view_placed_frame(id_spec);
}


int ChipsSystem::set_frame_bgcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), SET_FRAME_COLOR);
}


int ChipsSystem::get_frame_bgcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_FRAME_COLOR);
}


int ChipsSystem::set_frame_border_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_FRAME_BORDER_COLOR);
}


int ChipsSystem::get_frame_border_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_FRAME_BORDER_COLOR);
}


int ChipsSystem::set_frame_transparency(ChipsId* cid, bool seethrough)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&seethrough),
                            SET_FRAME_TRANSPARENT);
}


int ChipsSystem::get_frame_transparency(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_FRAME_TRANSPARENT);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::display_frame_border(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), 
                            SET_FRAME_BORDER_VIS);
}


int ChipsSystem::hide_frame_border(ChipsId* cid)
{
   int iv = 0; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),
                            SET_FRAME_BORDER_VIS);
}


int ChipsSystem::get_frame_border_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_FRAME_BORDER_VIS);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::display_frame(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),
                            SET_FRAME_VISIBLE);
}


int ChipsSystem::hide_frame(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),
                            SET_FRAME_VISIBLE);
}


int ChipsSystem::get_frame_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_FRAME_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_frame_scale(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val), SET_FRAME_SCALE);
}


int ChipsSystem::get_frame_scale(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_FRAME_SCALE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::blink_frames(ChipsId* cid, string fid2, float duration)
{
   set_id_spec(cid);
   return eng->blink_frames(id_spec, fid2, duration);
}


int ChipsSystem::stop_blink(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->stop_blink(id_spec);
}


int ChipsSystem::arrange_frames(ChipsId* cid, int cnt, string* fids)
{
   set_id_spec(cid);
   return eng->arrange_frames(id_spec, cnt, fids);
}


int ChipsSystem::add_plot(ChipsId* cid, string attlist)
{
   set_id_spec(cid);
   return eng->add_plot(id_spec, attlist);
}


int ChipsSystem::add_plot(ChipsId* cid,
   float px1, float py1, float px2, float py2, string attlist)
{
   int rval;
   if ((rval = normalized_area_check(px1, py1, px2, py2)) == ERR_STATUS_GOOD)  
   {
      set_id_spec(cid);
      rval = eng->add_plot(id_spec, px1, py1, px2, py2, attlist);
   }
   return rval;
}


int ChipsSystem::delete_plot(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_plot(id_spec);
}


int ChipsSystem::clear_plot(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->clear_plot(id_spec);
}


int ChipsSystem::get_plot_title_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_DEPTH);
}

int ChipsSystem::set_plot_title_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_PLOT_DEPTH);
}



int ChipsSystem::add_curve(ChipsId* cid, long cnt, double* x, double* y, 
   Transform *transform, string attlist)
{
   set_id_spec(cid);
   return eng->add_curve(id_spec, cnt, x, y, NULL, NULL, NULL, NULL, 
                         NULL, NULL, transform, attlist);
}


int ChipsSystem::add_curve(ChipsId* cid, long cnt, double* x, double*y, 
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform, string attlist)
{
   set_id_spec(cid);
   return eng->add_curve(id_spec, cnt, x, y, ydown, yup, xdown, xup, ylims,
                         xlims, transform, attlist);
}


int ChipsSystem::delete_curve(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_curve(id_spec);
}

int ChipsSystem::display_curve(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CURVE_VISIBLE);
}


int ChipsSystem::hide_curve(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CURVE_VISIBLE);
}


int ChipsSystem::get_curve_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val), 
                            SET_CURVE_SYMBOL_STYLE);
}


int ChipsSystem::get_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_SYMBOL_STYLE);
}


int ChipsSystem::set_curve_symbolangle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CURVE_SYMBOL_ANGLE);
}


int ChipsSystem::get_curve_symbolangle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_SYMBOL_ANGLE);
}

int ChipsSystem::set_curve_linestyle(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CURVE_LINE_STYLE);
}


int ChipsSystem::get_curve_linestyle(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_LINE_STYLE);
}


int ChipsSystem::set_curve_symbolfill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_fill),
                               SET_CURVE_SYMBOL_FILL);
}


int ChipsSystem::get_curve_symbolfill(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_SYMBOL_FILL);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_symbolsize(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_CURVE_SYMBOL_SIZE);
}


int ChipsSystem::get_curve_symbolsize(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_SYMBOL_SIZE);
}


int ChipsSystem::set_curve_symbolcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), 
                            SET_CURVE_SYMBOL_COLOR);
}


int ChipsSystem::get_curve_symbolcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CURVE_SYMBOL_COLOR);
}


int ChipsSystem::set_curve_linethickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_CURVE_LINE_THICKNESS);
}


int ChipsSystem::get_curve_linethickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CURVE_LINE_THICKNESS);
}


int ChipsSystem::set_curve_linecolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), 
                            SET_CURVE_LINE_COLOR);
}


int ChipsSystem::get_curve_linecolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_LINE_COLOR);
}


int ChipsSystem::set_curve_errthickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness), 
                            SET_CURVE_ERR_THICKNESS);
}


int ChipsSystem::get_curve_errthickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CURVE_ERR_THICKNESS);
}


int ChipsSystem::set_curve_errcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_CURVE_ERR_COLOR);
}


int ChipsSystem::get_curve_errcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_ERR_COLOR);
}

int ChipsSystem::set_curve_errcaplength(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_CURVE_ERR_CAPLEN);
}

int ChipsSystem::get_curve_errcaplength(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_ERR_CAPLEN);
}

int ChipsSystem::set_curve_limitoverride(ChipsId* cid, bool in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                               SET_CURVE_LIMOVERRIDE);
}


int ChipsSystem::get_curve_limitoverride(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_LIMOVERRIDE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_limitlength(ChipsId* cid, double in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_CURVE_LIMLENGTH);
}


int ChipsSystem::get_curve_limitlength(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_LIMLENGTH);
}




int ChipsSystem::get_curve_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4]; 
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_CURVE_RANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      outx1 = out_val[0]; 
      outx2 = out_val[1]; 
      outy1 = out_val[2]; 
      outy2 = out_val[3]; 
   }
   return rval;
}


int ChipsSystem::get_curve_xrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2]; 
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_CURVE_XRANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}


int ChipsSystem::get_curve_yrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_CURVE_YRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}


int ChipsSystem::set_curve_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_CURVE_DEPTH);
}


int ChipsSystem::get_curve_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_DEPTH);
}


int ChipsSystem::set_curve_errup(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),SET_CURVE_ERR_UP);
}


int ChipsSystem::get_curve_errup(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_ERR_UP);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_errdown(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CURVE_ERR_DOWN);
}


int ChipsSystem::get_curve_errdown(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_ERR_DOWN);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_errleft(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CURVE_ERR_LEFT);
}


int ChipsSystem::get_curve_errleft(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_ERR_LEFT);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_errright(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CURVE_ERR_RIGHT);
}


int ChipsSystem::get_curve_errright(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CURVE_ERR_RIGHT);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CURVE_BADDATA);
}


int ChipsSystem::get_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CURVE_BADDATA);
}


int ChipsSystem::set_curve_errstyle(ChipsId* cid, string in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_style.c_str(),
                            SET_CURVE_ERR_STYLE);
}


int ChipsSystem::get_curve_errstyle(ChipsId* cid, string& out_style)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_style, SET_CURVE_ERR_STYLE);
}

int ChipsSystem::limits(ChipsId* cid, double xmin, double xmax, double ymin, double ymax)
{
   set_id_spec(cid);
   return eng->limits(id_spec, XY_AXIS, xmin, xmax, ymin, ymax);
}


int ChipsSystem::limits(ChipsId* cid, int dim, double min, double max)
{
   set_id_spec(cid);
   return eng->limits(id_spec, dim, min, max);
}


int ChipsSystem::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE amin,
   V_AUTOLIMIT_TYPE amax)
{
   set_id_spec(cid);
   return eng->limits(id_spec, dim, amin, amax);
}


int ChipsSystem::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE amax)
{
   set_id_spec(cid);
   return eng->limits(id_spec, dim, min, amax);
}


int ChipsSystem::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE amin,
   double max)
{
   set_id_spec(cid);
   return eng->limits(id_spec, dim, amin, max);
}

int ChipsSystem::limits(ChipsId* pcid, int dim, V_OBJECT_TYPE obj, 
   ChipsId* ccid)
{
   set_id_spec(pcid);
   return eng->limits(id_spec, dim, obj, ccid);
}

int ChipsSystem::pick_limits(ChipsId* cid, int dim)
{
#ifndef HEADLESS
   ChipsSystemEventObserver event_observer;
   int old_state = event_manager->get_state ();
   int state = ChipsEventManager::RangeZoomState;
   double event_pos[] = {0,0};

   event_manager->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&state);
   event_manager->invoke_event (event_pos, ChipsEventManager::StartRangeZoomEvent, (void *)&dim);

   while (!event_observer.is_range_set ())
      while (g_main_context_iteration(NULL, FALSE));

   event_manager->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&old_state);

   return ERR_STATUS_GOOD;
#else
   return ERR_FEATURE_DNE;
#endif
}



int ChipsSystem::limits(ChipsId* pcid, int dim, V_OBJECT_TYPE obj,
   string oid)
{
   set_id_spec(pcid);
   ChipsId tmpB;
   tmpB.reset_window_id("");
   if (obj == chips_curve)
      tmpB.set_curve_id(oid);
   else if (obj == chips_contour)
      tmpB.set_contour_id(oid);
   else if (obj == chips_histogram)
      tmpB.set_histogram_id(oid);
   else if (obj == chips_image)
      tmpB.set_image_id(oid);
   else 
      return ERR_LIMIT_OBJ; 
   return eng->limits(id_spec, dim, obj, &tmpB);
}

int ChipsSystem::zoom(ChipsId* pcid, float scale)
{
   set_id_spec(pcid);
   return eng->zoom(id_spec, scale);
}

int ChipsSystem::zoom(ChipsId* pcid)
{
   set_id_spec(pcid);
   return eng->zoom(id_spec);
}

int ChipsSystem::panto(ChipsId* pcid, double x, double y)
{
   set_id_spec(pcid);
   return eng->panto(id_spec, x, y);
}

int ChipsSystem::pan(ChipsId* pcid)
{
   set_id_spec(pcid);
   return eng->pan(id_spec);
}


int ChipsSystem::set_plot_title(ChipsId* cid, string in_title)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_title.c_str(), SET_PLOT_TITLE);
}


int ChipsSystem::get_plot_title(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE);
}


int ChipsSystem::set_plot_title_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), 
                            SET_PLOT_TITLE_COLOR);
}


int ChipsSystem::get_plot_title_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_COLOR);
}


int ChipsSystem::set_plot_title_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_PLOT_TITLE_ANGLE);
}


int ChipsSystem::get_plot_title_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_ANGLE);
}


int ChipsSystem::set_plot_title_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_PLOT_TITLE_HALIGN);
}


int ChipsSystem::get_plot_title_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_HALIGN);
}

int ChipsSystem::set_plot_title_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_PLOT_TITLE_VALIGN);
}


int ChipsSystem::get_plot_title_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_VALIGN);
}


int ChipsSystem::set_plot_title_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)font.c_str(), SET_PLOT_TITLE_FONT);
}


int ChipsSystem::get_plot_title_font(ChipsId* cid, string& out_font)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_font, SET_PLOT_TITLE_FONT);
}


int ChipsSystem::set_plot_title_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(),
                            SET_PLOT_TITLE_FONTSTYLE);
}


int ChipsSystem::get_plot_title_fontstyle(ChipsId* cid, string& out_style)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_style,SET_PLOT_TITLE_FONTSTYLE);
}


int ChipsSystem::set_plot_title_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_PLOT_TITLE_SIZE);
}


int ChipsSystem::get_plot_title_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_SIZE);
}


int ChipsSystem::set_plot_title_xpos(ChipsId* cid, float xpos_norm)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&xpos_norm),
                            SET_PLOT_TITLE_X);
}


int ChipsSystem::get_plot_title_xpos(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_X);
}


int ChipsSystem::set_plot_title_ypos(ChipsId* cid, float ypos_norm)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&ypos_norm),
                            SET_PLOT_TITLE_Y);
}


int ChipsSystem::get_plot_title_ypos(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TITLE_Y);
}


int ChipsSystem::set_plot_axisstyle(ChipsId* cid, V_AXISSTYLE_TYPE style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&style),
                            SET_PLOT_AXIS_STYLE);
}


int ChipsSystem::set_plot_axisstyle(ChipsId* cid, string style)
{
   int rval = ERR_STATUS_BAD;
   V_AXISSTYLE_TYPE astyle;
   if ((rval = string_to_v_axisstyle(style.c_str(),&astyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&astyle),
                               SET_PLOT_AXIS_STYLE);
   }
   return rval;
}

int ChipsSystem::get_plot_axisstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_AXIS_STYLE);
}


int ChipsSystem::set_plot_cornerstyle(ChipsId* cid, string style)
{
   int rval = ERR_STATUS_BAD;
   V_CORNERSTYLE_TYPE cstyle;
   if ((rval = string_to_v_cornerstyle(style.c_str(),&cstyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&cstyle),
                               SET_PLOT_CORNER);
   }
   return rval;
}


int ChipsSystem::set_plot_cornerstyle(ChipsId* cid, V_CORNERSTYLE_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val), SET_PLOT_CORNER);
}


int ChipsSystem::get_plot_cornerstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_CORNER);
}


int ChipsSystem::get_plot_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_PLOT_RANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      outx1 = out_val[0];
      outx2 = out_val[1];
      outy1 = out_val[2];
      outy2 = out_val[3];
   }
   return rval;
}


int ChipsSystem::get_plot_xrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_PLOT_XRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}


int ChipsSystem::get_plot_yrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2];
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&out_val, GET_PLOT_YRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}


int ChipsSystem::set_plot_xlabel(ChipsId* cid, string in_label)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_label.c_str(),SET_PLOT_XLABEL);
}


int ChipsSystem::set_plot_ylabel(ChipsId* cid, string in_label)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_label.c_str(),SET_PLOT_YLABEL);
}


int ChipsSystem::display_plot(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_PLOT_VISIBLE);
}


int ChipsSystem::hide_plot(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec,static_cast<void*>(&iv),SET_PLOT_VISIBLE);
}


int ChipsSystem::get_plot_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_PLOT_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::get_plot_aspect_ratio(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_ASPECT_RATIO);
}

int ChipsSystem::get_plot_aspect_width(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_ASPECT_WIDTH);
}

int ChipsSystem::get_plot_aspect_height(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_ASPECT_HEIGHT);
}


int ChipsSystem::set_plot_aspect_ratio(ChipsId* cid, string ratio)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)ratio.c_str(),
                            SET_PLOT_ASPECT_RATIO);
}


int ChipsSystem::set_data_aspect_ratio(ChipsId* cid, string ratio)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)ratio.c_str(),
                            SET_DATA_ASPECT_RATIO);
}

int ChipsSystem::get_data_aspect_ratio(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_DATA_ASPECT_RATIO);
}

int ChipsSystem::display_axis(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_AXIS_VISIBLE);
}


int ChipsSystem::hide_axis(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_AXIS_VISIBLE);
}


int ChipsSystem::get_axis_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_plot_leftmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&margin_sz),
                            SET_PLOT_LMARGIN);
}


int ChipsSystem::get_plot_leftmargin(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_LMARGIN);
}


int ChipsSystem::set_plot_rightmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&margin_sz),
                            SET_PLOT_RMARGIN);
}


int ChipsSystem::get_plot_rightmargin(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_RMARGIN);
}


int ChipsSystem::set_plot_topmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&margin_sz),
                            SET_PLOT_TMARGIN);
}


int ChipsSystem::get_plot_topmargin(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_TMARGIN);
}


int ChipsSystem::set_plot_bottommargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&margin_sz),
                            SET_PLOT_BMARGIN);
}


int ChipsSystem::get_plot_bottommargin(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_PLOT_BMARGIN);
}


int ChipsSystem::reverse_axes(ChipsId* cid, int dim)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&dim),SET_AXIS_REVERSE);
}


int ChipsSystem::set_axis_scalable(ChipsId* cid, bool scalable)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&scalable),
                            SET_AXIS_SCALABLE);
}


int ChipsSystem::display_major_ticks(ChipsId* cid)
{
   int ival = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&ival),
                            SET_AXIS_MAJ_VIS);
}


int ChipsSystem::hide_major_ticks(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_AXIS_MAJ_VIS);
}


int ChipsSystem::get_major_ticks_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_MAJ_VIS);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   set_id_spec(cid);
   return eng->set_arbitrary_tick_positions(id_spec, arraysize, pos, vals);
}


int ChipsSystem::set_axis_major_mode(ChipsId* cid, string in_mode)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_mode.c_str(), 
                            SET_AXIS_MAJ_MODE);
}


int ChipsSystem::get_axis_major_mode(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJ_MODE);
}


int ChipsSystem::set_axis_major_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness), 
                            SET_AXIS_MAJ_THICKNESS);
}


int ChipsSystem::get_axis_major_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJ_THICKNESS);
}


int ChipsSystem::set_axis_major_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_AXIS_MAJ_STYLE);
}


int ChipsSystem::set_axis_major_style(ChipsId* cid, string style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle; 
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD) 
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_AXIS_MAJ_STYLE);
   }
   return rval;
}


int ChipsSystem::get_axis_major_style(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJ_STYLE);
}


int ChipsSystem::set_axis_ticklabel_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_AXIS_TICKLABEL_STYLE);
}


int ChipsSystem::set_axis_ticklabel_style(ChipsId* cid, string style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_AXIS_TICKLABEL_STYLE);
   }
   return rval;
}


int ChipsSystem::get_axis_ticklabel_style(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TICKLABEL_STYLE);
}


int ChipsSystem::display_minor_ticks(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_AXIS_MIN_VIS);
}


int ChipsSystem::hide_minor_ticks(ChipsId* cid)
{
   int iv = 0; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_AXIS_MIN_VIS);
}


int ChipsSystem::get_minor_ticks_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_MIN_VIS);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_axis_minor_mode(ChipsId* cid, string in_mode)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_mode.c_str(), SET_AXIS_MIN_MODE);
}


int ChipsSystem::get_axis_minor_mode(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MIN_MODE);
}


int ChipsSystem::set_axis_minor_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_AXIS_MIN_THICKNESS);
}


int ChipsSystem::get_axis_minor_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MIN_THICKNESS);
}


int ChipsSystem::set_axis_minor_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_AXIS_MIN_STYLE);
}



int ChipsSystem::set_axis_minor_style(ChipsId* cid, string style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval = string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_AXIS_MIN_STYLE);
   }
   return rval;
}


int ChipsSystem::get_axis_minor_style(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MIN_STYLE);
}


int ChipsSystem::set_axis_parallel_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_OFFSET_PAR);
}


int ChipsSystem::get_axis_parallel_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_OFFSET_PAR);
}


int ChipsSystem::set_axis_perpendicular_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_OFFSET_PERP);
}


int ChipsSystem::get_axis_perpendicular_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_OFFSET_PERP);
}


int ChipsSystem::set_axis_automin(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),SET_AXIS_AUTOMIN);
}


int ChipsSystem::get_axis_automin(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_AUTOMIN);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_axis_automax(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),SET_AXIS_AUTOMAX);
}


int ChipsSystem::get_axis_automax(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_AUTOMAX);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_axis_pad(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val), SET_AXIS_PAD);
}


int ChipsSystem::get_axis_pad(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_PAD);
}


int ChipsSystem::set_axis_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_AXIS_DEPTH);
}


int ChipsSystem::get_axis_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_DEPTH);
}


int ChipsSystem::get_axis_scale(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, GET_AXIS_SCALE);
}


int ChipsSystem::set_axis_ticklabel_color(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_val.c_str(),
                           SET_AXIS_TICKLABEL_COLOR);
}


int ChipsSystem::get_axis_ticklabel_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TICKLABEL_COLOR);
}


int ChipsSystem::set_axis_ticklabel_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_AXIS_TICKLABEL_SIZE);
}


int ChipsSystem::get_axis_ticklabel_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TICKLABEL_SIZE);
}


int ChipsSystem::set_axis_ticklabel_offset(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_AXIS_TICKLABEL_OFFSET);
}


int ChipsSystem::get_axis_ticklabel_offset(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_AXIS_TICKLABEL_OFFSET);
}


int ChipsSystem::set_axis_ticklabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_TICKLABEL_ANGLE);
}


int ChipsSystem::get_axis_ticklabel_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_AXIS_TICKLABEL_ANGLE);
}


int ChipsSystem::set_axis_ticklabel_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_TICKLABEL_HALIGN);
}


int ChipsSystem::get_axis_ticklabel_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_AXIS_TICKLABEL_HALIGN);
}

int ChipsSystem::set_axis_ticklabel_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_TICKLABEL_VALIGN);
}


int ChipsSystem::get_axis_ticklabel_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_AXIS_TICKLABEL_VALIGN);
}


int ChipsSystem::set_axis_ticklabel_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)font.c_str(),
                            SET_AXIS_TICKLABEL_FONT);
}


int ChipsSystem::get_axis_ticklabel_font(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_AXIS_TICKLABEL_FONT);
}


int ChipsSystem::set_axis_ticklabel_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(),
                            SET_AXIS_TICKLABEL_FNTSTL);
}


int ChipsSystem::get_axis_ticklabel_fontstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_AXIS_TICKLABEL_FNTSTL);
}


int ChipsSystem::set_axis_label_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), 
                            SET_AXIS_LABEL_COLOR);
}


int ChipsSystem::get_axis_label_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_COLOR);
}


int ChipsSystem::set_axis_label_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis),
                            SET_AXIS_LABEL_VIS);
}


int ChipsSystem::get_axis_label_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_LABEL_VIS);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_axis_label_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)font.c_str(), SET_AXIS_LABEL_FONT);
}


int ChipsSystem::get_axis_label_font(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_FONT);
}


int ChipsSystem::set_axis_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(),
                            SET_AXIS_LABEL_FONTSTYLE);
}


int ChipsSystem::get_axis_label_fontstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_AXIS_LABEL_FONTSTYLE);
}


int ChipsSystem::set_axis_label_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_AXIS_LABEL_SIZE);
}


int ChipsSystem::get_axis_label_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_SIZE);
}


int ChipsSystem::set_axis_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_LABEL_ANGLE);
}


int ChipsSystem::get_axis_label_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_ANGLE);
}


int ChipsSystem::set_axis_xlabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_XLAB_ANG);
}


int ChipsSystem::get_axis_xlabel_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_XLAB_ANG);
}


int ChipsSystem::set_axis_ylabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_YLAB_ANG);
}


int ChipsSystem::get_axis_ylabel_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_YLAB_ANG);
}


int ChipsSystem::set_axis_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_LABEL_HALIGN);
}


int ChipsSystem::get_axis_label_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_HALIGN);
}


int ChipsSystem::set_axis_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_AXIS_LABEL_VALIGN);
}


int ChipsSystem::get_axis_label_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_LABEL_VALIGN);
}


int ChipsSystem::linear_scale(ChipsId* cid, int dim)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&dim),
                            SET_PLOT_LINEARSCALE);
}


int ChipsSystem::log_scale(ChipsId* cid, int dim)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&dim),
                            SET_PLOT_LOGSCALE);
}


int ChipsSystem::set_label_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_color.c_str(), SET_LABEL_COLOR);
}


int ChipsSystem::get_label_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_COLOR);
}



int ChipsSystem::set_label_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_text.c_str(), SET_LABEL_TEXT);
}


int ChipsSystem::get_label_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_TEXT);
}


int ChipsSystem::set_label_size(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                           SET_LABEL_SIZE);
}


int ChipsSystem::get_label_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_DEPTH);
}

int ChipsSystem::set_label_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_LABEL_DEPTH);
}


int ChipsSystem::get_label_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_SIZE);
}



int ChipsSystem::display_label(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),SET_LABEL_VISIBLE);
}


int ChipsSystem::hide_label(ChipsId* cid)
{
   int iv = 0; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),SET_LABEL_VISIBLE);
}


int ChipsSystem::get_label_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_LABEL_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val), 
                               SET_LABEL_ANGLE);
}


int ChipsSystem::get_label_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_ANGLE);
}


int ChipsSystem::set_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val), 
                            SET_LABEL_HALIGN);
}


int ChipsSystem::get_label_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_HALIGN);
}

int ChipsSystem::set_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val), 
                            SET_LABEL_VALIGN);
}


int ChipsSystem::get_label_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LABEL_VALIGN);
}


int ChipsSystem::set_label_font(ChipsId* cid, string in_font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_font.c_str(), SET_LABEL_FONT);
}


int ChipsSystem::get_label_font(ChipsId* cid, string& out_font)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_font, SET_LABEL_FONT);
}


int ChipsSystem::set_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(), SET_LABEL_STYLE);
}


int ChipsSystem::get_label_fontstyle(ChipsId* cid, string& out_style)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_style, SET_LABEL_STYLE);
}


int ChipsSystem::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans, string attlist)
{
   set_id_spec(cid);
   return eng->add_contour(id_spec, dat, xdim, ydim, numlev, levels, trans, attlist);
}


int ChipsSystem::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int numlev, double* levels, Transform *trans, string attlist)
{
   set_id_spec(cid);
   return eng->add_contour(id_spec, cnt, x, y, z, numlev, levels, trans, attlist);
}


int ChipsSystem::delete_contour(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_contour(id_spec);
}

int ChipsSystem::display_contour(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CONTOUR_VISIBLE);
}


int ChipsSystem::hide_contour(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CONTOUR_VISIBLE);
}


int ChipsSystem::get_contour_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CONTOUR_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_contour_mode(ChipsId* cid, string md)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)md.c_str(), SET_CONTOUR_MODE);
}


int ChipsSystem::get_contour_mode(ChipsId* cid, string& out_md)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_md, SET_CONTOUR_MODE);
}


int ChipsSystem::set_contour_style(ChipsId* cid, string styl)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)styl.c_str(), SET_CONTOUR_STYLE);
}


int ChipsSystem::get_contour_style(ChipsId* cid, string& out_styl)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_styl, SET_CONTOUR_STYLE);
}


int ChipsSystem::set_contour_color(ChipsId* cid, string inc)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)inc.c_str(), SET_CONTOUR_COLOR);
}


int ChipsSystem::get_contour_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_COLOR);
}


int ChipsSystem::set_contour_thickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_CONTOUR_THICKNESS);
}


int ChipsSystem::get_contour_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_THICKNESS);
}

int ChipsSystem::get_contour_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_CONTOUR_RANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      outx1 = out_val[0];
      outx2 = out_val[1];
      outy1 = out_val[2];
      outy2 = out_val[3];
   }
   return rval;
}


int ChipsSystem::get_contour_xrange(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2]; 
   set_id_spec(cid);
   int rval = eng->get_property(id_spec,(void*)&out_val,GET_CONTOUR_XRANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}

int ChipsSystem::get_contour_yrange(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2]; 
   set_id_spec(cid);
   int rval = eng->get_property(id_spec,(void*)&out_val,GET_CONTOUR_YRANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}


int ChipsSystem::get_contour_zrange(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2];
   set_id_spec(cid);
   int  rval = eng->get_property(id_spec,(void*)&out_val,GET_CONTOUR_ZRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}

int ChipsSystem::set_contour_levels(ChipsId* cid, unsigned int arraysize,
   double* pos)
{
   set_id_spec(cid);
   return eng->set_contour_levels(id_spec, arraysize, pos);
}

int ChipsSystem::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   set_id_spec(cid);
   return eng->get_contour_levels(id_spec, pos);
}



int ChipsSystem::set_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_CONTOUR_ALGORITHM);
}


int ChipsSystem::get_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_ALGORITHM);
}


int ChipsSystem::set_contour_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_interval),
                            SET_CONTOUR_INTERVAL);
}


int ChipsSystem::get_contour_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_INTERVAL);
}



int ChipsSystem::set_contour_numlevels(ChipsId* cid, int in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CONTOUR_NUMLEVELS);
}


int ChipsSystem::get_contour_numlevels(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_NUMLEVELS);
}

int ChipsSystem::set_contour_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_CONTOUR_DEPTH);
}


int ChipsSystem::get_contour_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CONTOUR_DEPTH);
}


int ChipsSystem::add_histogram(ChipsId* cid, long cnt, double* bin_low,
   double* bin_high, double* counts, double* yd, double* yu, string attlist)
{
   set_id_spec(cid);
   return eng->add_histogram(id_spec, cnt, bin_low, bin_high, counts,
                             yd, yu, attlist);
}


int ChipsSystem::delete_histogram(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_histogram(id_spec);
}


int ChipsSystem::display_histogram(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_HISTO_VISIBLE);
}


int ChipsSystem::hide_histogram(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_HISTO_VISIBLE);
}


int ChipsSystem::get_histogram_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_HISTO_VISIBLE);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_HISTO_SYMBOL_STYLE);
}


int ChipsSystem::get_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_SYMBOL_STYLE);
}


int ChipsSystem::set_histogram_symbolangle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_HISTO_SYMBOL_ANGLE);
}


int ChipsSystem::get_histogram_symbolangle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_SYMBOL_ANGLE);
}


int ChipsSystem::set_histogram_linestyle(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_HISTO_LINE_STYLE);
}


int ChipsSystem::get_histogram_linestyle(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid); 
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_LINE_STYLE);
}


int ChipsSystem::set_histogram_symbolfill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_fill),
                            SET_HISTO_SYMBOL_FILL);
}


int ChipsSystem::get_histogram_symbolfill(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_HISTO_SYMBOL_FILL);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_histogram_symbolsize(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_HISTO_SYMBOL_SIZE);
}


int ChipsSystem::get_histogram_symbolsize(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_SYMBOL_SIZE);
}


int ChipsSystem::set_histogram_symbolcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_HISTO_SYMBOL_COLOR);
}


int ChipsSystem::get_histogram_symbolcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_HISTO_SYMBOL_COLOR);
}


int ChipsSystem::set_histogram_linethickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_HISTO_LINE_THICKNESS);
}


int ChipsSystem::get_histogram_linethickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_HISTO_LINE_THICKNESS);
}


int ChipsSystem::set_histogram_linecolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_HISTO_LINE_COLOR);
}


int ChipsSystem::get_histogram_linecolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_LINE_COLOR);
}


int ChipsSystem::set_histogram_errthickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_HISTO_ERR_THICKNESS);
}

int ChipsSystem::get_histogram_errthickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_HISTO_ERR_THICKNESS);
}

int ChipsSystem::set_histogram_errcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_HISTO_ERR_COLOR);
}

int ChipsSystem::get_histogram_errcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_ERR_COLOR);
}

int ChipsSystem::set_histogram_errcaplength(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_HISTO_ERR_CAPLEN);
}

int ChipsSystem::get_histogram_errcaplength(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_ERR_CAPLEN);
}

int ChipsSystem::set_histogram_errup(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),SET_HISTO_ERR_UP);
}


int ChipsSystem::get_histogram_errup(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_HISTO_ERR_UP);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_histogram_errdown(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_HISTO_ERR_DOWN);
}


int ChipsSystem::get_histogram_errdown(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_HISTO_ERR_DOWN);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_histogram_dropline(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_HISTO_DROPLINE);
}


int ChipsSystem::get_histogram_dropline(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_HISTO_DROPLINE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_histogram_errstyle(ChipsId* cid, string in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_style.c_str(),
                            SET_HISTO_ERR_STYLE);
}


int ChipsSystem::get_histogram_errstyle(ChipsId* cid, string& out_style)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_style, SET_HISTO_ERR_STYLE);
}

int ChipsSystem::set_histogram_fillcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_HISTO_FILL_COLOR);
}


int ChipsSystem::get_histogram_fillcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_FILL_COLOR);
}


int ChipsSystem::set_histogram_fillopacity(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)&in_val, SET_HISTO_FILL_OPACITY);
}


int ChipsSystem::get_histogram_fillopacity(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_FILL_OPACITY);
}

int ChipsSystem::set_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_HISTO_FILL_STYLE);
}


int ChipsSystem::get_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_FILL_STYLE);
}


int ChipsSystem::get_histogram_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_HISTO_RANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      outx1 = out_val[0];
      outx2 = out_val[1];
      outy1 = out_val[2];
      outy2 = out_val[3];
   }
   return rval;
}


int ChipsSystem::get_histogram_xrange(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2]; 
   set_id_spec(cid);
   int rval = eng->get_property(id_spec,(void*)&out_val,GET_HISTO_XRANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}


int ChipsSystem::get_histogram_yrange(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2]; 
   set_id_spec(cid);
   int rval = eng->get_property(id_spec,(void*)&out_val,GET_HISTO_YRANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}


int ChipsSystem::set_histogram_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_HISTO_DEPTH);
}


int ChipsSystem::get_histogram_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_HISTO_DEPTH);
}


int ChipsSystem::add_axis(ChipsId* cid, int dim, float pos,
   float minx, float maxx, float miny, float maxy, Transform *transform, string aid)
{
   set_id_spec(cid);
   if ((chips_strcmp(aid.c_str(), "bx1") == 0) || 
       (chips_strcmp(aid.c_str(), "bx2") == 0) ||
       (chips_strcmp(aid.c_str(), "by1") == 0) ||
       (chips_strcmp(aid.c_str(), "by2") == 0))
      return ERR_BOX_AXIS_ID; 
   else
      return eng->add_axis(id_spec, dim, pos, minx, maxx, miny, maxy, transform, aid);
}

int ChipsSystem::delete_axis(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_axis(id_spec);
}


int ChipsSystem::bind_axes(ChipsId* cid, string lid, string aid)
{
   set_id_spec(cid);
   return eng->bind_axes(id_spec, lid, aid);
}

int ChipsSystem::unbind_axes(ChipsId* cid, string lid, string aid)
{
   set_id_spec(cid);
   return eng->unbind_axes(id_spec, lid, aid);
}


int ChipsSystem::set_axis_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_AXIS_COLOR);
}


int ChipsSystem::get_axis_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_COLOR);
}


int ChipsSystem::set_axis_major_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_AXIS_MAJOR_COLOR);
}


int ChipsSystem::get_axis_major_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJOR_COLOR);
}


int ChipsSystem::set_axis_minor_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_AXIS_MINOR_COLOR);
}


int ChipsSystem::get_axis_minor_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINOR_COLOR);
}


int ChipsSystem::set_axis_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_text.c_str(), SET_AXIS_TEXT);
}


int ChipsSystem::get_axis_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TEXT);
}


int ChipsSystem::set_axis_xlabel_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_text.c_str(), SET_AXIS_XLAB_TXT);
}


int ChipsSystem::get_axis_xlabel_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_XLAB_TXT);
}


int ChipsSystem::set_axis_ylabel_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_text.c_str(), SET_AXIS_YLAB_TXT);
}


int ChipsSystem::get_axis_ylabel_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_YLAB_TXT);
}


int ChipsSystem::set_axis_tickformat(ChipsId* cid, string in_format)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_format.c_str(),
                            SET_AXIS_TICKFORMAT);
}


int ChipsSystem::get_axis_tickformat(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TICKFORMAT);
}


int ChipsSystem::set_axis_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness), 
                            SET_AXIS_THICKNESS);
}


int ChipsSystem::get_axis_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_THICKNESS);
}


int ChipsSystem::set_axis_tickstyle(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style), 
                            SET_AXIS_TICKSTYLE);
}


int ChipsSystem::set_axis_tickstyle(ChipsId* cid, string in_style)
{
   int rval = ERR_STATUS_BAD;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval=string_to_v_tickstyle(in_style.c_str(),&tstyle))==ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_AXIS_TICKSTYLE);
   }
   return rval;
}


int ChipsSystem::get_axis_tickstyle(ChipsId* cid, string& out_val)
{
   // tickstyle is actually majortick style- use that method
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_TICKSTYLE);
}


int ChipsSystem::set_axis_major_num(ChipsId* cid, int in_num)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_num),
                            SET_AXIS_MAJOR_NUM);
}


int ChipsSystem::get_axis_major_num(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJOR_NUM);
}


int ChipsSystem::set_axis_minor_num(ChipsId* cid, int in_num)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_num),
                            SET_AXIS_MINOR_NUM);
}


int ChipsSystem::get_axis_minor_num(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINOR_NUM);
}


int ChipsSystem::set_axis_major_length(ChipsId* cid, int in_len)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_len),
                               SET_AXIS_MAJORLENGTH);
}


int ChipsSystem::get_axis_major_length(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJORLENGTH);
}


int ChipsSystem::set_axis_minor_length(ChipsId* cid, int in_len)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_len),
                            SET_AXIS_MINORLENGTH);
}


int ChipsSystem::get_axis_minor_length(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINORLENGTH);
}


int ChipsSystem::set_axis_major_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_interval),
                            SET_AXIS_MAJORINT);
}


int ChipsSystem::get_axis_major_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJORINT);
}


int ChipsSystem::set_axis_minor_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_interval),
                            SET_AXIS_MINORINT);
}


int ChipsSystem::get_axis_minor_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINORINT);
}


int ChipsSystem::set_axis_ticklabel_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis),
                            SET_AXIS_TICKLABEL_VIS);
}


int ChipsSystem::get_axis_ticklabel_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_TICKLABEL_VIS);
   out_val = (val != 0);
   return rval;
}



int ChipsSystem::set_axis_majorgrid_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis),
                            SET_AXIS_MAJGRD_VIS);
}


int ChipsSystem::get_axis_majorgrid_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_MAJGRD_VIS);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_axis_majorgrid_thickness(ChipsId* cid, float in_thick)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thick),
                            SET_AXIS_MAJGRD_THICK);
}


int ChipsSystem::get_axis_majorgrid_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJGRD_THICK);
}


int ChipsSystem::set_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&style),
                            SET_AXIS_MAJGRD_STYLE);
}


int ChipsSystem::get_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJGRD_STYLE);
}


int ChipsSystem::set_axis_majorgrid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_AXIS_MAJGRD_COLOR);
}


int ChipsSystem::get_axis_majorgrid_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MAJGRD_COLOR);
}


int ChipsSystem::set_axis_minorgrid_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis), 
                            SET_AXIS_MINGRD_VIS);
}


int ChipsSystem::get_axis_minorgrid_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_AXIS_MINGRD_VIS);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_axis_minorgrid_thickness(ChipsId* cid, float in_thick)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thick),
                            SET_AXIS_MINGRD_THICK);
}


int ChipsSystem::get_axis_minorgrid_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINGRD_THICK);
}


int ChipsSystem::set_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&style),
                            SET_AXIS_MINGRD_STYLE);
}


int ChipsSystem::get_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINGRD_STYLE);
}


int ChipsSystem::set_axis_minorgrid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_AXIS_MINGRD_COLOR);
}


int ChipsSystem::get_axis_minorgrid_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_AXIS_MINGRD_COLOR);
}


int ChipsSystem::get_axis_range(ChipsId* cid, double& out1, double& out2)
{
   double out_val[2]; 
   set_id_spec(cid);
   int rval = eng->get_property(id_spec,(void*)&out_val,GET_AXIS_RANGE);
   if (rval == ERR_STATUS_GOOD)  
   {
      out1 = out_val[0]; 
      out2 = out_val[1]; 
   }
   return rval;
}


int ChipsSystem::add_image(ChipsId* cid, string filename, Transform *transform, string attlist)
{
   set_id_spec(cid);
   return eng->add_image(id_spec, filename, transform, attlist);
}


int ChipsSystem::add_image(ChipsId* cid,
   string filename, float ix1, float iy1, float ix2, float iy2, 
   Transform *transform, string attlist)
{
   set_id_spec(cid);
   return eng->add_image(id_spec, filename, ix1, iy1, ix2, iy2, transform, attlist);
}


int ChipsSystem::add_image(ChipsId* cid, double* rvals, double *gvals, 
   double *bvals, double *avals, int xaxis, int yaxis, Transform *transform, string attlist)
{
   set_id_spec(cid);
   return eng->add_image(id_spec, rvals, gvals, bvals, avals, xaxis, yaxis, transform, attlist);
}



int ChipsSystem::delete_image(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_image(id_spec);
}

int ChipsSystem::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{
   set_id_spec(cid);
   return eng->set_image_channel(id_spec, vals, xaxis, yaxis, channel);
}

int ChipsSystem::remove_image_channel(ChipsId* cid, int channel)
{
   set_id_spec(cid);
   return eng->remove_image_channel(id_spec, channel);
}

int ChipsSystem::print_image (ChipsId* cid, const string &outfile, const string &prefs)
{
   set_id_spec(cid);
   string filepath = outfile;
   // resolve the path to the output file, if not provided
   if (outfile.length() && ((outfile.rfind('/') == string::npos) ||
       (outfile.find("./") == 0) || (outfile.find("../") == 0)))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(outfile);
      }
   }
   return eng->print_image(id_spec, filepath, prefs);

}

int ChipsSystem::set_image_colormap(ChipsId* cid, V_COLORMAP_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_IMAGE_CMAP);
}

int ChipsSystem::get_image_colormap(ChipsId* cid, V_COLORMAP_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_IMAGE_CMAP);
}

int ChipsSystem::set_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_IMAGE_INTERPOLATION);
}

int ChipsSystem::get_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_IMAGE_INTERPOLATION);
}

int ChipsSystem::set_image_invert_colormap(ChipsId* cid, bool inv)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&inv),
                               SET_IMAGE_INV_CMAP);
}

int ChipsSystem::get_image_invert_colormap(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_IMAGE_INV_CMAP);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_image_colormap_interpolate(ChipsId* cid, bool interp)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&interp),
                               SET_IMAGE_CMAP_INTERPOLATE);
}

int ChipsSystem::get_image_colormap_interpolate(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_IMAGE_CMAP_INTERPOLATE);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_image_scale_channels(ChipsId* cid, bool scale)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&scale),
                               SET_IMAGE_SCALE_CHANNELS);
}

int ChipsSystem::get_image_scale_channels(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_IMAGE_SCALE_CHANNELS);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_image_colormap_size(ChipsId* cid, int size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&size),
                               SET_IMAGE_CMAP_SIZE);
}

int ChipsSystem::get_image_colormap_size(ChipsId* cid, int& out_size)
{
   int rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&out_size, SET_IMAGE_CMAP_SIZE);
   return rval;
}

int ChipsSystem::set_image_alpha(ChipsId* cid, double min, double max)
{
   double range[2] = { min, max };
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)&range, SET_IMAGE_ALPHA);
}

int ChipsSystem::get_image_alpha(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2];
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&out_val, SET_IMAGE_ALPHA);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}

int ChipsSystem::set_image_threshold(ChipsId* cid, double min, double max)
{
   double range[2] = { min, max };
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)&range, SET_IMAGE_THRESHOLD);
}

int ChipsSystem::get_image_threshold(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[2];
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&out_val, SET_IMAGE_THRESHOLD);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}

int ChipsSystem::set_image_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_IMAGE_DEPTH);
}

int ChipsSystem::get_image_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_IMAGE_DEPTH);
}

int ChipsSystem::display_image(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_IMAGE_VISIBLE);
}

int ChipsSystem::hide_image(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_IMAGE_VISIBLE);
}

int ChipsSystem::get_image_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_IMAGE_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::get_image_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_IMAGE_RANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      outx1 = out_val[0];
      outx2 = out_val[1];
      outy1 = out_val[2];
      outy2 = out_val[3];
   }
   return rval;
}

int ChipsSystem::get_image_xrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_IMAGE_XRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}

int ChipsSystem::get_image_yrange(ChipsId* cid, double& out1, double& out2)
{
   int rval = ERR_STATUS_BAD;
   double out_val[4];
   set_id_spec(cid);
   rval = eng->get_property(id_spec,(void*)&out_val,GET_IMAGE_YRANGE);
   if (rval == ERR_STATUS_GOOD)
   {
      out1 = out_val[0];
      out2 = out_val[1];
   }
   return rval;
}


int ChipsSystem::add_colorbar(ChipsId* cid, double xpos, double ypos,
   string obj_id, V_OBJECT_TYPE obj, string attlist)
{
   set_id_spec(cid);
   return eng->add_colorbar(id_spec, xpos, ypos, obj_id, obj, attlist);
}

int ChipsSystem::delete_colorbar(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_colorbar(id_spec);
}

int ChipsSystem::set_colorbar_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_CBAR_DEPTH);
}

int ChipsSystem::get_colorbar_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_DEPTH);
}

int ChipsSystem::set_colorbar_width(ChipsId* cid, float in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_CBAR_WIDTH);
}

int ChipsSystem::get_colorbar_width(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_WIDTH);
}

int ChipsSystem::set_colorbar_length(ChipsId* cid, float in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_CBAR_LENGTH);
}

int ChipsSystem::get_colorbar_length(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LENGTH);
}

int ChipsSystem::set_colorbar_orientation(ChipsId* cid, int inval)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&inval),
                           SET_CBAR_ORIENTATION);
}

int ChipsSystem::get_colorbar_orientation(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_ORIENTATION);
}

int ChipsSystem::display_colorbar(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CBAR_VISIBLE);
}

int ChipsSystem::hide_colorbar(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_CBAR_VISIBLE);
}

int ChipsSystem::get_colorbar_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CBAR_VISIBLE);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_colorbar_border_visible(ChipsId* cid, bool val)
{
   int iv = (int)val; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),
                            SET_CBAR_BORDERVIS);
}

int ChipsSystem::get_colorbar_border_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CBAR_BORDERVIS);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_colorbar_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_HALIGN);
}

int ChipsSystem::get_colorbar_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_HALIGN);
}

int ChipsSystem::set_colorbar_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_VALIGN);
}

int ChipsSystem::get_colorbar_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_VALIGN);
}

int ChipsSystem::set_colorbar_label_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), 
                            SET_CBAR_LABEL_COLOR);
}

int ChipsSystem::get_colorbar_label_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_COLOR);
}

int ChipsSystem::set_colorbar_label_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis),
                            SET_CBAR_LABEL_VIS);
}

int ChipsSystem::get_colorbar_label_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CBAR_LABEL_VIS);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_colorbar_label_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)font.c_str(), SET_CBAR_LABEL_FONT);
}

int ChipsSystem::get_colorbar_label_font(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_FONT);
}

int ChipsSystem::set_colorbar_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(),
                            SET_CBAR_LABEL_FONTSTYLE);
}

int ChipsSystem::get_colorbar_label_fontstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CBAR_LABEL_FONTSTYLE);
}

int ChipsSystem::set_colorbar_label_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_CBAR_LABEL_SIZE);
}

int ChipsSystem::get_colorbar_label_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_SIZE);
}

int ChipsSystem::set_colorbar_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_LABEL_ANGLE);
}

int ChipsSystem::get_colorbar_label_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_ANGLE);
}

int ChipsSystem::set_colorbar_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_LABEL_HALIGN);
}

int ChipsSystem::get_colorbar_label_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_HALIGN);
}

int ChipsSystem::set_colorbar_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_LABEL_VALIGN);
}

int ChipsSystem::get_colorbar_label_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_VALIGN);
}

int ChipsSystem::set_colorbar_label_location(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_CBAR_LABEL_LOCATION);
}

int ChipsSystem::set_colorbar_label_location(ChipsId* cid, string style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_CBAR_LABEL_LOCATION);
   }
   return rval;
}

int ChipsSystem::get_colorbar_label_location(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_LABEL_LOCATION);
}

int ChipsSystem::set_colorbar_label_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)(in_text.c_str ()),
                            SET_CBAR_TEXT);
}

int ChipsSystem::get_colorbar_label_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TEXT);
}


int ChipsSystem::set_colorbar_ticklabel_color(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_val.c_str(),
                           SET_CBAR_TICKLABEL_COLOR);
}

int ChipsSystem::get_colorbar_ticklabel_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICKLABEL_COLOR);
}

int ChipsSystem::set_colorbar_ticklabel_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_CBAR_TICKLABEL_SIZE);
}

int ChipsSystem::get_colorbar_ticklabel_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICKLABEL_SIZE);
}

int ChipsSystem::set_colorbar_ticklabel_offset(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&sz),
                            SET_CBAR_TICKLABEL_OFFSET);
}

int ChipsSystem::get_colorbar_ticklabel_offset(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_CBAR_TICKLABEL_OFFSET);
}

int ChipsSystem::set_colorbar_ticklabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_TICKLABEL_ANGLE);
}

int ChipsSystem::get_colorbar_ticklabel_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CBAR_TICKLABEL_ANGLE);
}

int ChipsSystem::set_colorbar_ticklabel_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_TICKLABEL_HALIGN);
}

int ChipsSystem::get_colorbar_ticklabel_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_CBAR_TICKLABEL_HALIGN);
}

int ChipsSystem::set_colorbar_ticklabel_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_TICKLABEL_VALIGN);
}

int ChipsSystem::get_colorbar_ticklabel_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val,SET_CBAR_TICKLABEL_VALIGN);
}

int ChipsSystem::set_colorbar_ticklabel_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)font.c_str(),
                            SET_CBAR_TICKLABEL_FONT);
}

int ChipsSystem::get_colorbar_ticklabel_font(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CBAR_TICKLABEL_FONT);
}

int ChipsSystem::set_colorbar_ticklabel_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)style.c_str(),
                            SET_CBAR_TICKLABEL_FNTSTL);
}

int ChipsSystem::get_colorbar_ticklabel_fontstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_CBAR_TICKLABEL_FNTSTL);
}

int ChipsSystem::set_colorbar_ticklabel_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&vis),
                            SET_CBAR_TICKLABEL_VIS);
}

int ChipsSystem::get_colorbar_ticklabel_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_CBAR_TICKLABEL_VIS);
   out_val = (val != 0);
   return rval;
}

int ChipsSystem::set_colorbar_tick_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_CBAR_TICK_THICKNESS);
}

int ChipsSystem::get_colorbar_tick_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid); 
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_THICKNESS);
}

int ChipsSystem::set_colorbar_tickstyle(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_CBAR_TICK_STYLE);
}

int ChipsSystem::set_colorbar_tickstyle(ChipsId* cid, string style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle; 
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD) 
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_CBAR_TICK_STYLE);
   }
   return rval;
}

int ChipsSystem::get_colorbar_tickstyle(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_STYLE);
}

int ChipsSystem::set_colorbar_mode(ChipsId* cid, string in_mode)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_mode.c_str(), 
                            SET_CBAR_TICK_MODE);
}

int ChipsSystem::get_colorbar_mode(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_MODE);
}

int ChipsSystem::set_colorbar_parallel_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_OFFSET_PAR);
}


int ChipsSystem::get_colorbar_parallel_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_OFFSET_PAR);
}


int ChipsSystem::set_colorbar_perpendicular_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_CBAR_OFFSET_PERP);
}


int ChipsSystem::get_colorbar_perpendicular_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_OFFSET_PERP);
}

int ChipsSystem::set_colorbar_tick_location(ChipsId* cid, string in_style)
{
   int rval;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval=string_to_v_tickstyle(in_style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = eng->set_property(id_spec, static_cast<void*>(&tstyle),
                               SET_CBAR_TICK_LOCATION);
   }
   return rval;
}

int ChipsSystem::set_colorbar_tick_location(ChipsId* cid, int in_style)
{
   int rval;
   set_id_spec(cid);
   rval = eng->set_property(id_spec, static_cast<void*>(&in_style),
                               SET_CBAR_TICK_LOCATION);
   return rval;
}

int ChipsSystem::get_colorbar_tick_location(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_LOCATION);
}

int ChipsSystem::set_colorbar_tick_visible(ChipsId* cid, bool vis)
{
   int rval;
   set_id_spec(cid);
   rval = eng->set_property(id_spec, static_cast<void*>(&vis),
                               SET_CBAR_TICK_VIS); 
   return rval;
}

int ChipsSystem::get_colorbar_tick_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_VIS);
}

int ChipsSystem::set_colorbar_tick_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)in_color.c_str(),
                           SET_CBAR_TICK_COLOR);
}

int ChipsSystem::get_colorbar_tick_color(ChipsId* cid, string& out_color)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_color, SET_CBAR_TICK_COLOR);
}

int ChipsSystem::set_colorbar_tick_length(ChipsId* cid, int in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,(void*)&in_thickness,
                           SET_CBAR_TICK_LENGTH);
}

int ChipsSystem::get_colorbar_tick_length(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_CBAR_TICK_LENGTH);
}



int ChipsSystem::add_label(ChipsId* cid, double xpos, 
   double ypos, string text, string attlist)
{
   set_id_spec(cid);
   return eng->add_label(id_spec, xpos, ypos, text, attlist);
}

int ChipsSystem::delete_label(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_label(id_spec);
}


int ChipsSystem::add_line(ChipsId* cid,
   double startx, double starty, double endx, double endy, string attlist)
{
   set_id_spec(cid);
   return eng->add_line(id_spec, startx, starty, endx, endy, attlist);
}


int ChipsSystem::add_line(ChipsId* cid,
   double* xx, double* yy, unsigned int count, string attlist)
{
   set_id_spec(cid);
   return eng->add_line(id_spec, xx, yy, count, attlist);
}


int ChipsSystem::delete_line(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_line(id_spec);
}


int ChipsSystem::set_line_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), SET_LINE_COLOR);
}


int ChipsSystem::get_line_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LINE_COLOR);
}


int ChipsSystem::get_line_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LINE_DEPTH);
}

int ChipsSystem::set_line_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_LINE_DEPTH);
}


int ChipsSystem::set_line_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_LINE_THICKNESS);
}


int ChipsSystem::get_line_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid); 
   return eng->get_property(id_spec, (void*)&out_val, SET_LINE_THICKNESS);
}


int ChipsSystem::set_line_style(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_LINE_STYLE);
}


int ChipsSystem::get_line_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LINE_STYLE);
}


int ChipsSystem::set_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_LINE_EXTEND);
}


int ChipsSystem::get_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_LINE_EXTEND);
}


int ChipsSystem::display_line(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_LINE_VISIBLE);
}


int ChipsSystem::hide_line(ChipsId* cid)
{
   int iv = 0; 
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_LINE_VISIBLE);
}


int ChipsSystem::get_line_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_LINE_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::add_point(ChipsId* cid, double xx, double yy, string attlist)
{
   set_id_spec(cid);
   return eng->add_point(id_spec, xx, yy, attlist);
}


int ChipsSystem::add_point(ChipsId* cid,
   double xx, double  yy, double  zz, string attlist)
{
   set_id_spec(cid);
   return eng->add_point(id_spec, xx, yy, zz, attlist);
}


int ChipsSystem::delete_point(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_point(id_spec);
}


int ChipsSystem::set_point_size(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_size),
                            SET_POINT_SYMBOL_SIZE);
}


int ChipsSystem::get_point_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_POINT_SYMBOL_SIZE);
}


int ChipsSystem::get_point_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_POINT_DEPTH);
}

int ChipsSystem::set_point_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_POINT_DEPTH);
}


int ChipsSystem::set_point_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                            SET_POINT_SYMBOL_ANGLE);
}


int ChipsSystem::get_point_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_POINT_SYMBOL_ANGLE);
}


int ChipsSystem::set_point_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_POINT_SYMBOL_COLOR);
}


int ChipsSystem::get_point_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec,(void*)&out_val,SET_POINT_SYMBOL_COLOR);
}


int ChipsSystem::set_point_fill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_fill),
                            SET_POINT_SYMBOL_FILL);
}


int ChipsSystem::get_point_fill(ChipsId* cid, bool& out_fill)
{
   int val, rval; 
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_POINT_SYMBOL_FILL);
   out_fill = (val != 0); 
   return rval;
}


int ChipsSystem::set_point_symbol(ChipsId* cid, V_SYMBOL_TYPE in_symbol)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_symbol),
                            SET_POINT_SYMBOL_STYLE);
}


int ChipsSystem::get_point_symbol(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_POINT_SYMBOL_STYLE);
}


int ChipsSystem::display_point(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),SET_POINT_VISIBLE);
}


int ChipsSystem::hide_point(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv),SET_POINT_VISIBLE);
}


int ChipsSystem::get_point_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_POINT_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_window_title(ChipsId* cid, string title)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)title.c_str(), SET_WINDOW_TITLE);
}


int ChipsSystem::get_window_title(ChipsId* cid, string& out_title)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_title, SET_WINDOW_TITLE);
}


int ChipsSystem::set_redraw(bool val)
{
   return eng->set_redraw(val);
}


int ChipsSystem::set_window_redraw(ChipsId* cid,  bool draw)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&draw),
                            SET_WINDOW_REDRAW);
}


int ChipsSystem::get_window_redraw(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_WINDOW_REDRAW);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::set_window_bgcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), SET_WINDOW_COLOR);
}


int ChipsSystem::get_window_bgcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_WINDOW_COLOR);
}


int ChipsSystem::iconify_window(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->iconify_window(id_spec);
}


int ChipsSystem::deiconify_window(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->deiconify_window(id_spec);
}


int ChipsSystem::display_grid(ChipsId* cid, int scale)
{
   set_id_spec(cid);
   return eng->display_grid(id_spec, scale);
}


int ChipsSystem::hide_grid(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->hide_grid(id_spec);
}


int ChipsSystem::set_grid_linestyle(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_GRID_LINESTYLE);
}


int ChipsSystem::set_grid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), SET_GRID_COLOR);
}


int ChipsSystem::set_grid_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                           SET_GRID_THICKNESS);
}


int ChipsSystem::set_grid_scale(ChipsId* cid, int in_scale)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_scale), 
                            SET_GRID_SCALE);
}


int ChipsSystem::set_layout_linestyle(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_style),
                            SET_LAYOUT_LINESTYLE);
}


int ChipsSystem::set_layout_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(), SET_LAYOUT_COLOR);
}


int ChipsSystem::set_layout_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_LAYOUT_THICKNESS);
}


int ChipsSystem::display_frame_layout(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->display_layout(id_spec);
}


int ChipsSystem::hide_frame_layout(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->hide_layout(id_spec);
}


int ChipsSystem::add_legend(ChipsId* cid, const char** curve_ids,
   const char** curve_labels, string lgid)
{
   set_id_spec(cid);
   if (chips_strcmp(lgid.c_str(), "all") == 0)
      return ERR_USED_ALL_AS_ID;
   else
      return eng->add_legend(id_spec, curve_ids, curve_labels, lgid);
}

int ChipsSystem::delete_legend(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_legend(id_spec);
}

int ChipsSystem::hide_legend(ChipsId* cid)
{
   int iv = 2; 
   set_id_spec(cid);
   return eng->set_property(id_spec,static_cast<void*>(&iv),SET_LEGEND_VISIBLE);
}

int ChipsSystem::display_legend(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   return eng->set_property(id_spec,static_cast<void*>(&iv),SET_LEGEND_VISIBLE);
}

int ChipsSystem::set_legend_color(ChipsId* cid, string in_clr)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_clr.c_str(), SET_LEGEND_COLOR);
}

int ChipsSystem::set_legend_borders(ChipsId* cid, bool showborders)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&showborders),
                            SET_LEGEND_BORDERS);
}

int ChipsSystem::set_legend_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness), 
                            SET_LEGEND_THICKNESS);
}

int ChipsSystem::set_legend_transparency(ChipsId* cid, bool seethrough)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&seethrough),
                            SET_LEGEND_TRANSPARENT);
}


int ChipsSystem::add_region(ChipsId* cid,
   double* xx, double* yy, unsigned int count, string attlist)
{
   set_id_spec(cid);
   return eng->add_region(id_spec, xx, yy, count, attlist);
}


int ChipsSystem::add_region(ChipsId* cid,
   int sides, double xx, double yy, double radius, string attlist)
{
   set_id_spec(cid);
   return eng->add_region(id_spec, sides, xx, yy, radius, attlist);
}


int ChipsSystem::delete_region(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_region(id_spec);
}


int ChipsSystem::display_region(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&iv), SET_REGN_VISIBLE);
}


int ChipsSystem::hide_region(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   return eng->set_property(id_spec,static_cast<void*>(&iv),SET_REGN_VISIBLE);
}

int ChipsSystem::get_region_visible(ChipsId* cid, bool& out_val)
{
   int val, rval;
   set_id_spec(cid);
   rval = eng->get_property(id_spec, (void*)&val, SET_REGN_VISIBLE);
   out_val = (val != 0);
   return rval;
}


int ChipsSystem::get_region_depth(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_DEPTH);
}

int ChipsSystem::set_region_depth(ChipsId* cid, int in_depth)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_depth),
                           SET_REGN_DEPTH);
}


int ChipsSystem::set_region_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec,static_cast<void*>(&in_val),SET_REGN_ANGLE);
}


int ChipsSystem::get_region_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_ANGLE);
}


int ChipsSystem::set_region_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_thickness),
                            SET_REGN_EDGE_THICKNESS);
}


int ChipsSystem::get_region_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_EDGE_THICKNESS);
}


int ChipsSystem::set_region_style(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_REGN_EDGE_STYLE);
}


int ChipsSystem::get_region_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_EDGE_STYLE);
}

int ChipsSystem::set_region_fillstyle(ChipsId* cid, V_FILL_TYPE val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&val),
                            SET_REGN_FILL_STYLE);
}


int ChipsSystem::get_region_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_FILL_STYLE);
}


int ChipsSystem::set_region_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                            SET_REGN_EDGE_COLOR);
}


int ChipsSystem::get_region_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_EDGE_COLOR);
}


int ChipsSystem::set_region_fillcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, (void*)in_color.c_str(),
                           SET_REGN_FILL_COLOR);
}


int ChipsSystem::get_region_fillcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_FILL_COLOR);
}

int ChipsSystem::set_region_opacity(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   return eng->set_property(id_spec, static_cast<void*>(&in_val),
                               SET_REGN_OPACITY);
}


int ChipsSystem::get_region_opacity(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   return eng->get_property(id_spec, (void*)&out_val, SET_REGN_OPACITY);
}


int ChipsSystem::pick(ChipsId* cid, int count, unsigned int mask)
{
   return ERR_FEATURE_DNE;
}


int ChipsSystem::get_pick(ChipsId* cid,
   vector<CHIPS_PICK_EVT> &picklist, int count, unsigned int mask)
{
#ifndef HEADLESS
   ChipsSystemEventObserver event_observer;
   int rval = ERR_STATUS_GOOD;
   
   int old_state = event_manager->get_state ();
   int state = ChipsEventManager::PickState;
   CHIPS_PICK_EVT pick_event;
   double event_pos[] = {0,0};
   COORD_SYS old_coordsys;
   string old_coordid;
   ChipsId tmp_cid;

   rval = eng->setup_pick (cid, count, mask, tmp_cid);

   if (rval == ERR_STATUS_GOOD)
   {
      event_manager->get_display_coordinates (old_coordsys, old_coordid);
      event_manager->set_display_coordinates (tmp_cid.get_coord_sys (),
                                           tmp_cid.get_coord_id ());

      event_manager->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&state);
      event_manager->invoke_event (event_pos, ChipsEventManager::StartPickEvent, (void *)&mask);

      picklist.clear ();
      picklist.reserve(500);

      while (picklist.size () != (unsigned int)count)
      {
         while (g_main_context_iteration(NULL, FALSE));
         if (event_observer.get_pick_event (pick_event))
         {
            if ((int)pick_event.key == 27)
               break;
            else
            {
               picklist.push_back (pick_event);
               event_observer.reset_pick_set ();
            }
         }
      }

      event_manager->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&old_state);
      event_manager->set_display_coordinates (old_coordsys, old_coordid);
   }

   
#else
   int rval = ERR_FEATURE_DNE;
#endif
   return rval;
}


int ChipsSystem::print_frame(ChipsId* cid, string outfile)
{
   set_id_spec(cid);
   string filepath = outfile;
   // resolve the path to the output file, if not provided
   if (outfile.length() && ((outfile.rfind('/') == string::npos) ||
       (outfile.find("./") == 0) || (outfile.find("../") == 0)))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(outfile);
      }
   }
   return eng->print_frame(id_spec, filepath);
}


int ChipsSystem::print_window(ChipsId* cid, string outfile, string prefs)
{
   set_id_spec(cid);
   string filepath = outfile;
   // resolve the path to the output file, if not provided
   if (outfile.length() && ((outfile.rfind('/') == string::npos) ||
       (outfile.find("./") == 0) || (outfile.find("../") == 0)))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(outfile);
      }
   }
   return eng->print_window(id_spec, filepath, prefs);
}


int ChipsSystem::delete_all_windows(ChipsId* cid)
{
   set_id_spec(cid, 1);
   return eng->delete_all_windows(id_spec);
}


int ChipsSystem::delete_all_frames(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_all_frames(id_spec);
}


int ChipsSystem::delete_all_plots(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_all_plots(id_spec);
}


int ChipsSystem::delete_all_images(ChipsId* cid)
{
   set_id_spec(cid);
   return eng->delete_all_images(id_spec);
}


int ChipsSystem::normalized_area_check(float x1, float y1, float x2, float y2)
{
   int rval = ERR_STATUS_BAD;
   if (x1 < 0)  
      rval = ERR_NORM_DATA_RANGE;
   else if (x1 > 1)
      rval = ERR_NORM_DATA_RANGE;
   else if (x2 <= x1)
      rval = ERR_NORM_XMAX_LE_XMIN;
   else if ((x2 > 1) || (y1 < 0) || (y1 > 1))
      rval = ERR_NORM_DATA_RANGE;
   else if (y2 <= y1)
      rval = ERR_NORM_YMAX_LE_YMIN;
   else if (y2 > 1)
      rval = ERR_NORM_DATA_RANGE; 
   else 
      rval = ERR_STATUS_GOOD; 
   return rval; 
}


int ChipsSystem::undo(int depth)
{
   set_id_spec();
   return eng->undo(depth);
}


int ChipsSystem::undo(string tag)
{
   set_id_spec();
   return eng->undo(tag);
}


int ChipsSystem::redo(int depth)
{
   set_id_spec();
   return eng->redo(depth);
}


int ChipsSystem::redo(string tag)
{
   set_id_spec();
   return eng->redo(tag);
}


int ChipsSystem::set_undo_tag(string tag)
{
   set_id_spec();
   return eng->set_undo_tag(tag);
}


int ChipsSystem::clear_undo_tag(string tag)
{
   set_id_spec();
   return eng->clear_undo_tag(tag);
}

int ChipsSystem::open_undo_block()   
{ 
   set_id_spec();
   return eng->open_undo_block(); 
}


int ChipsSystem::close_undo_block()
{ 
   set_id_spec();
   return eng->close_undo_block(); 
}


int ChipsSystem::open_undo_buffer()   
{ 
   set_id_spec();
   return eng->open_undo_buffer(); 
}


int ChipsSystem::close_undo_buffer()
{ 
   set_id_spec();
   return eng->close_undo_buffer(); 
}


int ChipsSystem::discard_undo_buffer()
{ 
   set_id_spec();
   return eng->discard_undo_buffer(); 
}


void ChipsSystem::set_id_spec(ChipsId* supplied, bool use_all)
{
   if (!supplied)
   {
      id_spec = &default_ids; 
      if (use_all)
         id_spec->set_all();
      else 
         id_spec->reset_window_id(""); 
   }
   else 
      id_spec = supplied;
}

int ChipsSystem::convert_coordinate (double *in_val, double *out_val, int count,
                               ChipsId *cid1, ChipsId *cid2, int axis)
{
   int rval;
   ChipsVector in_vec (count+1);
   ChipsVector out_vec (count+1);

   if (!in_val || !out_val)
      return ERR_STATUS_BAD;

   for (int ii=0; ii<count; ii++)
      in_vec[ii] = in_val[ii];

   set_id_spec ();
   rval = eng->convert_coordinate(in_vec, out_vec, cid1, cid2, axis);

   for (int ii=0; ii<count; ii++)
      out_val[ii] = out_vec[ii];

   return rval;
}

int ChipsSystem::convert_coordinate (double in_val, double &out_val,
                               ChipsId *cid1, ChipsId *cid2, int axis)
{
   // Create a vector of size 2 to to account for a homogenous factor
   ChipsVector in_vec (2);  
   ChipsVector out_vec (2);
   int rval;

   in_vec[0] = in_val;

   set_id_spec ();
   rval = eng->convert_coordinate (in_vec, out_vec, cid1, cid2, axis);

   out_val = out_vec[0];

   return rval;
}


char* ChipsSystem::get_errors(int mode)
{
   return eng->get_errors(mode);
}

int ChipsSystem::add_client_error(int errcode, string errtext)
{
   return eng->add_client_error(errcode, errtext);
}

int ChipsSystem::set_error_verbosity(int val)
{
   return eng->set_error_verbosity(val);
}

int ChipsSystem::clear_errors()
{
   return eng->clear_errors();
}

int ChipsSystem::get_error_verbosity(int &out_val)
{
   return eng->get_error_verbosity(out_val);
}

int ChipsSystem::get_error_count(int mode, int &out_val)
{
   return eng->get_error_count(mode, out_val);
}

int ChipsSystem::get_error_types(int &out_val)
{
   return eng->get_error_types(out_val);
}

int ChipsSystem::load_fill(string in_file, int slot)
{
   string filepath = in_file;
   // resolve the path to the file, if not provided
   if (in_file.length() && ((in_file.rfind('/') == string::npos) || 
       (in_file.find("./") == 0) || (in_file.find("../") == 0)))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(in_file);
      }
   }
   return eng->load_fill(filepath, slot);
}

int ChipsSystem::load_colormap(long len, double* r, double* g, 
	double* b, double* alpha, int slot)
{
  return eng->load_colormap(len, r, g, b, alpha, slot);
}

int ChipsSystem::load_colormap(string in_file, int slot)
{
   string filepath = in_file;
   // resolve the path to the file, if not provided
   if (((in_file.rfind('/') == string::npos) || (in_file.find("./") == 0) ||
       (in_file.find("../") == 0)) && in_file.length())
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(in_file);
      }
   }
   return eng->load_colormap(filepath, slot);
}

int ChipsSystem::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   return eng->highlight_object(id_spec, obj);
}

