// 
//  Copyright (C) 2004,2007-2009,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSSYSTEM_HH
#define CHIPSSYSTEM_HH

#include <string>
#include <list>
#include <vector>
#include <iostream>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ChipsDefinesP.hh"
#include "ChipsId.hh"

#include "Transform.hh"

class ChipsEngine;
class ChipsObject;

using namespace std;


class ChipsSystem
{
   public:

   ChipsSystem(bool load_prefs=true);
  
   virtual ~ChipsSystem();

   // preferences
   int     load_preferences(string in_file=""); 
   int     save_preferences(string in_file="");
   int     set_preference(string pref, string value);  
   char*   get_preference(string pref);  
   int     set_line_stipple(V_LINE_TYPE line, int val);
   int     get_line_stipple(V_LINE_TYPE line, int &out_val);
   int     load_fill(string fname, int slot=chips_userfill1); 
   
   int     load_colormap(long len, double* r, double* g, 
			 double* b, double* alpha, int slot=chips_usercmap1);
   int     load_colormap(string fname, int slot=chips_usercmap1);

   int     load_state(string fname); 
   int     save_state(string fname, int mode=0); 
   int     undo(string tag);
   int     undo(int depth=1);
   int     redo(string tag);
   int     redo(int depth=1);
   int     set_undo_tag(string tag);
   int     clear_undo_tag(string tag);
   int     open_undo_block();
   int     close_undo_block();
   int     open_undo_buffer();
   int     close_undo_buffer();
   int     discard_undo_buffer();

   int     set_cascading_property(ChipsId* cid, V_OBJECT_TYPE start,
                                  string property,  string value);  
   int     move(ChipsId* cid, V_OBJECT_TYPE obj, double xx, double yy, 
                int mode=V_MOVE_ABSOLUTE); 
   int     get_object_count(ChipsId* cid, V_OBJECT_TYPE obj, int& out_val); 
   int     get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj, 
                                COORD_SYS& out_val, string& out_id); 
   int     highlight_object(ChipsId* cid, V_OBJECT_TYPE obj); 

   // window operations 
   int     add_window(ChipsId* cid, string attlist = "");
   int     add_window(ChipsId* cid, float ww, float hh, string format="", 
                      string attlist="");
   int     delete_window(ChipsId* cid);
   int     set_current_window(ChipsId* cid);
   int     set_window_title(ChipsId* cid, string title); 
   int     get_window_title(ChipsId* cid, string& out_val); 
   int     set_window_bgcolor(ChipsId* cid, string in_color); 
   int     get_window_bgcolor(ChipsId* cid, string& out_val); 
   int     set_window_redraw(ChipsId* cid, bool draw); 
   int     get_window_redraw(ChipsId* cid, bool& out_val);
   int     iconify_window(ChipsId* cid);
   int     deiconify_window(ChipsId* cid);
   int     display_grid(ChipsId* cid, int scale=0);
   int     hide_grid(ChipsId* cid);
   int     set_redraw(bool val);

   int     display_frame_layout(ChipsId* cid);
   int     hide_frame_layout(ChipsId* cid);
   int     set_frame_bgcolor(ChipsId* cid, string in_color); 
   int     get_frame_bgcolor(ChipsId* cid, string& out_val); 
   int     set_frame_border_color(ChipsId* cid, string in_color); 
   int     get_frame_border_color(ChipsId* cid, string& out_val); 
   int     get_frame_border_visible(ChipsId* cid, bool& out_val);
   int     display_frame_border(ChipsId* cid);
   int     hide_frame_border(ChipsId* cid);
   int     set_layout_color(ChipsId* cid, string in_color); 
   int     set_layout_thickness(ChipsId* cid, float in_thickness);
   int     set_layout_linestyle(ChipsId* cid, V_LINE_TYPE in_style);
   int     set_grid_color(ChipsId* cid, string in_color); 
   int     set_grid_thickness(ChipsId* cid, float in_thickness);
   int     set_grid_linestyle(ChipsId* cid, V_LINE_TYPE in_style);
   int     set_grid_scale(ChipsId* cid, int in_scale);
   int     pick(ChipsId* cid, int count = 1, unsigned int mask = 0);
   int     get_pick(ChipsId* cid, vector<CHIPS_PICK_EVT> &picklist,
                int count = 1, unsigned int mask = 0);

   // frame operations
   int     add_frame(ChipsId* cid, string attlist = "");
   int     add_frame(ChipsId* cid, float x1, float y1, float x2, float y2, 
                     string attlist = "");
   int     delete_frame(ChipsId* cid);
   int     set_framemode(ChipsId* cid, int mode);
   int     tile(ChipsId* cid, unsigned int cols, unsigned int rows);
   int     row_tile(ChipsId* cid, unsigned int rows);
   int     col_tile(ChipsId* cid, unsigned int cols);
   int     prev_frame(ChipsId* cid);
   int     next_frame(ChipsId* cid);
   int     reposition_frame(ChipsId* cid, float x1, 
                            float y1, float x2, float y2);
   int     view_single_frame(ChipsId* cid);
   int     view_placed_frame(ChipsId* cid);
   int     arrange_frames(ChipsId* cid, int cnt, string* fids);
   int     set_frame_transparency(ChipsId* cid, bool seethrough); 
   int     get_frame_transparency(ChipsId* cid, bool& out_val);
   int     set_frame_scale(ChipsId* cid, bool val); 
   int     get_frame_scale(ChipsId* cid, bool& out_val);
   int     blink_frames(ChipsId* cid, string fid2, float duration=3.0);
   int     stop_blink(ChipsId* cid); 
   int     hide_frame(ChipsId* cid);
   int     display_frame(ChipsId* cid);
   int     get_frame_visible(ChipsId* cid, bool& out_val);
   int     set_current_frame(ChipsId* cid);
   int     display_depth(ChipsId* cid, int depth);
   int     hide_depth(ChipsId* cid, int depth);
   int     collapse_depths(ChipsId* cid, int cnt, int* srcs, int dest);
   int     set_default_depth(ChipsId* cid, int in_val);
   int     get_default_depth(ChipsId* cid, int& out_val);
   int     shuffle(ChipsId* cid, V_OBJECT_TYPE obj, V_SHUFFLE_TYPE way);
   int     shuffle_forward(ChipsId* cid, V_OBJECT_TYPE obj);
   int     shuffle_backward(ChipsId* cid, V_OBJECT_TYPE obj);
   int     shuffle_front(ChipsId* cid, V_OBJECT_TYPE obj);
   int     shuffle_back(ChipsId* cid, V_OBJECT_TYPE obj);
   int     adjust_grid_xgap(ChipsId* cid, float xgap);
   int     adjust_grid_ygap(ChipsId* cid, float ygap);
   int     adjust_grid_gaps(ChipsId* cid, float xgap, float ygap);
   int     adjust_grid_xrelsize(ChipsId* cid, int col, float ratio);
   int     adjust_grid_yrelsize(ChipsId* cid, int row, float ratio);
   int     adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* sizes);
   int     adjust_grid_yrelsizes(ChipsId* cid, int cnt, float* sizes);
   int     grid_objects(ChipsId* cid, int cols, int rows, float xgap, 
                        float ygap, int usage = 0, int lcnt=0, 
                        string* lids = NULL);
   int     row_grid_objects(ChipsId* cid, int cnt, int* vals,
                        float xg, float yg, int usage = 0, int lcnt = 0,
                        string* lids = NULL);
   int     col_grid_objects(ChipsId* cid, int cnt, int* vals,
                        float xg, float yg, int usage = 0, int lcnt = 0,
                        string* lids = NULL);
   int     swap_object_positions(ChipsId* cid, string lid2);

   // plot operations
   int     add_plot(ChipsId* cid, string attlist = "");
   int     add_plot(ChipsId* cid, float px1, float py1, float px2, float py2, 
                    string attlist = "");
   int     delete_plot(ChipsId* cid);
   int     clear_plot(ChipsId* cid);
   int     set_plot_title_depth(ChipsId* cid, int in_val);
   int     get_plot_title_depth(ChipsId* cid, int& out_val);
   int     reposition_plot(ChipsId* cid, float x1, 
                           float y1, float x2, float y2);
   int     limits(ChipsId* cid, double xmin, double xmax, double ymin, double ymax);
   int     limits(ChipsId* cid, int dim, double min, double max);
   int     limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin, 
                  V_AUTOLIMIT_TYPE automax);
   int     limits(ChipsId* cid, int dim, double min, V_AUTOLIMIT_TYPE automax);
   int     limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin, double max); 
   int     limits(ChipsId* pcid, int dim, V_OBJECT_TYPE obj, ChipsId* ccid); 
   int     limits(ChipsId* pcid, int dim, V_OBJECT_TYPE obj, string oid); 
   int     pick_limits(ChipsId* cid, int dim=XY_AXIS); 
   int     zoom (ChipsId* cid, float scale);
   int     zoom (ChipsId* cid);
   int     panto (ChipsId* cid, double x, double y);
   int     pan (ChipsId* cid);

   int     log_scale(ChipsId* cid, int dim=XY_AXIS);
   int     linear_scale(ChipsId* cid, int dim=XY_AXIS);
   int     set_plot_title(ChipsId* cid, string in_title);
   int     get_plot_title(ChipsId* cid, string& out_val);
   int     set_plot_title_color(ChipsId* cid, string in_color);
   int     get_plot_title_color(ChipsId* cid, string& out_val);
   int     set_plot_title_angle(ChipsId* cid, float in_val);
   int     get_plot_title_angle(ChipsId* cid, float& out_val);
   int     set_plot_title_halign(ChipsId* cid, float in_val);
   int     get_plot_title_halign(ChipsId* cid, float& out_val);
   int     set_plot_title_valign(ChipsId* cid, float in_val);
   int     get_plot_title_valign(ChipsId* cid, float& out_val);
   int     set_plot_title_size(ChipsId* cid, int sz);
   int     get_plot_title_size(ChipsId* cid, int& out_val);
   int     set_plot_title_font(ChipsId* cid, string font);
   int     get_plot_title_font(ChipsId* cid, string& out_val);
   int     set_plot_title_fontstyle(ChipsId* cid, string style);
   int     get_plot_title_fontstyle(ChipsId* cid, string& out_val);
   int     set_plot_title_xpos(ChipsId* cid, float in_norm);
   int     get_plot_title_xpos(ChipsId* cid, float& out_val);
   int     set_plot_title_ypos(ChipsId* cid, float in_norm);
   int     get_plot_title_ypos(ChipsId* cid, float& out_val);
   int     get_plot_range(ChipsId* cid, double& outx1, double& outx2,
                          double& outy1, double& outy2);
   int     get_plot_xrange(ChipsId* cid, double& out1, double& out2);
   int     get_plot_yrange(ChipsId* cid, double& out1, double& out2);

   int     set_plot_xlabel(ChipsId* cid, string in_lab);
   int     set_plot_ylabel(ChipsId* cid, string in_lab);
   int     set_plot_axisstyle(ChipsId* cid, string style); 
   int     set_plot_axisstyle(ChipsId* cid, V_AXISSTYLE_TYPE style); 
   int     get_plot_axisstyle(ChipsId* cid, string& out_val);
   int     set_plot_cornerstyle(ChipsId* cid, V_CORNERSTYLE_TYPE val);
   int     set_plot_cornerstyle(ChipsId* cid, string style);
   int     get_plot_cornerstyle(ChipsId* cid, string& out_val);
   int     hide_plot(ChipsId* cid);
   int     display_plot(ChipsId* cid);
   int     get_plot_visible(ChipsId* cid, bool& out_val);
   int     display_axis(ChipsId* cid);
   int     hide_axis(ChipsId* cid);
   int     get_axis_visible(ChipsId* cid, bool& out_val);
   int     set_data_aspect_ratio(ChipsId* cid, string style);
   int     get_data_aspect_ratio(ChipsId* cid, string& out_val);
   int     set_plot_aspect_ratio(ChipsId* cid, string style);
   int     get_plot_aspect_ratio(ChipsId* cid, string& out_val);
   int     get_plot_aspect_width(ChipsId* cid, float& out_val);
   int     get_plot_aspect_height(ChipsId* cid, float& out_val);

   int     set_plot_leftmargin(ChipsId* cid, float margin_sz); 
   int     get_plot_leftmargin(ChipsId* cid, float& out_val);
   int     set_plot_rightmargin(ChipsId* cid, float margin_sz); 
   int     get_plot_rightmargin(ChipsId* cid, float& out_val);
   int     set_plot_topmargin(ChipsId* cid, float margin_sz); 
   int     get_plot_topmargin(ChipsId* cid, float& out_val);
   int     set_plot_bottommargin(ChipsId* cid, float margin_sz); 
   int     get_plot_bottommargin(ChipsId* cid, float& out_val);
   int     reverse_axes(ChipsId* cid, int dim=XY_AXIS);
   int     set_axis_scalable(ChipsId* cid, bool scalable); 
   int     display_major_ticks(ChipsId* cid);
   int     hide_major_ticks(ChipsId* cid);
   int     get_major_ticks_visible(ChipsId* cid, bool& out_val);
   int     display_minor_ticks(ChipsId* cid);
   int     hide_minor_ticks(ChipsId* cid);
   int     get_minor_ticks_visible(ChipsId* cid, bool& out_val);
   int     set_current_plot(ChipsId* cid);
   int     set_arbitrary_tick_positions(ChipsId* cid, unsigned int arraysize, 
                               double* pos, string* vals=NULL);

   int     set_axis_major_mode(ChipsId* cid, string in_mode); 
   int     get_axis_major_mode(ChipsId* cid, string& out_val); 
   int     set_axis_major_thickness(ChipsId* cid, float in_thickness);
   int     get_axis_major_thickness(ChipsId* cid, float& out_val);
   int     set_axis_major_style(ChipsId* cid, string in_style);
   int     set_axis_major_style(ChipsId* cid, int in_style);
   int     get_axis_major_style(ChipsId* cid, string& out_val);
   int     set_axis_minor_mode(ChipsId* cid, string in_mode); 
   int     get_axis_minor_mode(ChipsId* cid, string& out_val); 
   int     set_axis_minor_thickness(ChipsId* cid, float in_thickness);
   int     get_axis_minor_thickness(ChipsId* cid, float& out_val);
   int     set_axis_minor_style(ChipsId* cid, string in_style);
   int     set_axis_minor_style(ChipsId* cid, int in_style);
   int     get_axis_minor_style(ChipsId* cid, string& out_val);
   int     set_axis_parallel_offset(ChipsId* cid, float in_val);
   int     get_axis_parallel_offset(ChipsId* cid, float& out_val);
   int     set_axis_perpendicular_offset(ChipsId* cid, float in_val);
   int     get_axis_perpendicular_offset(ChipsId* cid, float& out_val);
   int     set_axis_automin(ChipsId* cid, bool on); 
   int     get_axis_automin(ChipsId* cid, bool& out_val);
   int     set_axis_automax(ChipsId* cid, bool on); 
   int     get_axis_automax(ChipsId* cid, bool& out_val);
   int     set_axis_pad(ChipsId* cid, double in_val);
   int     get_axis_pad(ChipsId* cid, double& out_val);
   int     set_axis_depth(ChipsId* cid, int in_val);
   int     get_axis_depth(ChipsId* cid, int& out_val);
   int     get_axis_scale(ChipsId* cid, int& out_val);
   int     set_axis_label_color(ChipsId* cid, string in_color);
   int     get_axis_label_color(ChipsId* cid, string& out_val);
   int     set_axis_label_size(ChipsId* cid, int sz);
   int     get_axis_label_size(ChipsId* cid, int& out_val);
   int     set_axis_label_angle(ChipsId* cid, float sz);
   int     get_axis_label_angle(ChipsId* cid, float& out_val);
   int     set_axis_xlabel_angle(ChipsId* cid, float sz);
   int     get_axis_xlabel_angle(ChipsId* cid, float& out_val);
   int     set_axis_ylabel_angle(ChipsId* cid, float sz);
   int     get_axis_ylabel_angle(ChipsId* cid, float& out_val);
   int     set_axis_label_halign(ChipsId* cid, float sz);
   int     get_axis_label_halign(ChipsId* cid, float& out_val);
   int     set_axis_label_valign(ChipsId* cid, float sz);
   int     get_axis_label_valign(ChipsId* cid, float& out_val);
   int     set_axis_label_font(ChipsId* cid, string font);
   int     get_axis_label_font(ChipsId* cid, string& out_val);
   int     set_axis_label_fontstyle(ChipsId* cid, string style);
   int     get_axis_label_fontstyle(ChipsId* cid, string& out_val);
   int     set_axis_label_visible(ChipsId* cid, bool vis);
   int     get_axis_label_visible(ChipsId* cid, bool& out_val);
   int     set_axis_ticklabel_style(ChipsId* cid, string style);
   int     set_axis_ticklabel_style(ChipsId* cid, int in_style);
   int     get_axis_ticklabel_style(ChipsId* cid, string& out_val);
   int     set_axis_ticklabel_color(ChipsId* cid, string in_color);
   int     get_axis_ticklabel_color(ChipsId* cid, string& out_val);
   int     set_axis_ticklabel_size(ChipsId* cid, int in_val);
   int     get_axis_ticklabel_size(ChipsId* cid, int& out_val);
   int     set_axis_ticklabel_angle(ChipsId* cid, float in_val);
   int     get_axis_ticklabel_angle(ChipsId* cid, float& out_val);
   int     set_axis_ticklabel_halign(ChipsId* cid, float in_val);
   int     get_axis_ticklabel_halign(ChipsId* cid, float& out_val);
   int     set_axis_ticklabel_valign(ChipsId* cid, float in_val);
   int     get_axis_ticklabel_valign(ChipsId* cid, float& out_val);
   int     set_axis_ticklabel_font(ChipsId* cid, string in_font);
   int     get_axis_ticklabel_font(ChipsId* cid, string& out_val);
   int     set_axis_ticklabel_fontstyle(ChipsId* cid, string style);
   int     get_axis_ticklabel_fontstyle(ChipsId* cid, string& out_val);
   int     set_axis_ticklabel_offset(ChipsId* cid, int in_val);
   int     get_axis_ticklabel_offset(ChipsId* cid, int& out_val);
   int     set_axis_ticklabel_visible(ChipsId* cid, bool vis);
   int     get_axis_ticklabel_visible(ChipsId* cid, bool& out_val);
   int     set_axis_majorgrid_visible(ChipsId* cid, bool vis);
   int     get_axis_majorgrid_visible(ChipsId* cid, bool& out_val);
   int     set_axis_majorgrid_thickness(ChipsId* cid, float in_thickness);
   int     get_axis_majorgrid_thickness(ChipsId* cid, float& out_val);
   int     set_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE in_style);
   int     get_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_axis_majorgrid_color(ChipsId* cid, string in_color);
   int     get_axis_majorgrid_color(ChipsId* cid, string& out_val);
   int     set_axis_minorgrid_visible(ChipsId* cid, bool vis);
   int     get_axis_minorgrid_visible(ChipsId* cid, bool& out_val);
   int     set_axis_minorgrid_thickness(ChipsId* cid, float in_thickness);
   int     get_axis_minorgrid_thickness(ChipsId* cid, float& out_val);
   int     set_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE in_style);
   int     get_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_axis_minorgrid_color(ChipsId* cid, string in_color);
   int     get_axis_minorgrid_color(ChipsId* cid, string& out_val);

   // curve operations 
   int     add_curve(ChipsId* cid, long cnt, double* x, double* y, 
                     Transform *transform, string attlist = "");
   int     add_curve(ChipsId* cid, long cnt, double* x, double* y, 
                     double* ydown, double* yup, 
                     double* xdown, double* xup, 
                     double* ylims, double* xlims, 
                     Transform *transform, string attlist = "");
   int     delete_curve(ChipsId* cid);
   int     display_curve(ChipsId* cid);
   int     hide_curve(ChipsId* cid);
   int     get_curve_visible(ChipsId* cid, bool& out_val);
   int     set_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE val);
   int     get_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val);
   int     set_curve_symbolfill(ChipsId* cid, bool in_fill);
   int     get_curve_symbolfill(ChipsId* cid, bool& out_val);
   int     set_curve_symbolsize(ChipsId* cid, int in_size);
   int     get_curve_symbolsize(ChipsId* cid, int& out_val);
   int     set_curve_symbolcolor(ChipsId* cid, string in_color);
   int     get_curve_symbolcolor(ChipsId* cid, string& out_val);
   int     set_curve_symbolangle(ChipsId* cid, float in_val);
   int     get_curve_symbolangle(ChipsId* cid, float& out_val);
   int     set_curve_linestyle(ChipsId* cid, V_LINE_TYPE val);
   int     get_curve_linestyle(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_curve_linethickness(ChipsId* cid, float in_size);
   int     get_curve_linethickness(ChipsId* cid, float& out_val);
   int     set_curve_linecolor(ChipsId* cid, string in_color);
   int     get_curve_linecolor(ChipsId* cid, string& out_val);
   int     set_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE val);
   int     get_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE& out_val);
   int     set_curve_errstyle(ChipsId* cid, string style);
   int     get_curve_errstyle(ChipsId* cid, string& out_val);
   int     set_curve_errup(ChipsId* cid, bool val);
   int     get_curve_errup(ChipsId* cid, bool& out_val);
   int     set_curve_errdown(ChipsId* cid, bool val);
   int     get_curve_errdown(ChipsId* cid, bool& out_val);
   int     set_curve_errleft(ChipsId* cid, bool val);
   int     get_curve_errleft(ChipsId* cid, bool& out_val);
   int     set_curve_errright(ChipsId* cid, bool val);
   int     get_curve_errright(ChipsId* cid, bool& out_val);
   int     set_curve_errthickness(ChipsId* cid, float in_thickness);
   int     get_curve_errthickness(ChipsId* cid, float& out_val);
   int     set_curve_errcolor(ChipsId* cid, string in_color);
   int     get_curve_errcolor(ChipsId* cid, string& out_val);
   int     set_curve_errcaplength(ChipsId* cid, int in_size);
   int     get_curve_errcaplength(ChipsId* cid, int& out_val);
   int     set_curve_limitlength(ChipsId* cid, double in_val);
   int     get_curve_limitlength(ChipsId* cid, double& in_val);
   int     set_curve_limitoverride(ChipsId* cid, bool val);
   int     get_curve_limitoverride(ChipsId* cid, bool& out_val);
   int     set_current_curve(ChipsId* cid);
   int     get_curve_range(ChipsId* cid, double& outx1, double& outx2,
                           double& outy1, double& outy2);
   int     get_curve_xrange(ChipsId* cid, double& out1, double& out2);
   int     get_curve_yrange(ChipsId* cid, double& out1, double& out2);
   int     set_curve_depth(ChipsId* cid, int in_val);
   int     get_curve_depth(ChipsId* cid, int& out_val);

   // contour operations 
   int     add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
                       unsigned int numlev = 0, double* levels = NULL,
		       Transform *trans=NULL, string attlist = ""); 
   int     add_contour(ChipsId* cid, long cnt, double* x, double* y,
                      double* z, unsigned int numlev = 0, double* levels = NULL,
                      Transform *trans=NULL, string attlist = ""); 
   int     delete_contour(ChipsId* cid);
   int     display_contour(ChipsId* cid);
   int     hide_contour(ChipsId* cid);
   int     get_contour_visible(ChipsId* cid, bool& out_val);
   int     set_contour_mode(ChipsId*, string mode); 
   int     get_contour_mode(ChipsId*, string& out_val); 
   int     set_contour_style(ChipsId*, string style); 
   int     get_contour_style(ChipsId*, string& out_val); 
   int     set_contour_color(ChipsId*, string in_color); 
   int     get_contour_color(ChipsId*, string& out_val); 
   int     set_contour_levels(ChipsId* cid, unsigned int arraysize, 
                              double* pos);
   int     get_contour_levels(ChipsId* cid, vector<double> &pos);
   int     set_contour_thickness(ChipsId* cid, float in_size);
   int     get_contour_thickness(ChipsId* cid, float& out_val);
   int     set_current_contour(ChipsId* cid);
   int     get_contour_range(ChipsId* cid, double& outx1, double& outx2,
                             double& outy1, double& outy2);
   int     get_contour_xrange(ChipsId* cid, double& out1, double& out2);
   int     get_contour_yrange(ChipsId* cid, double& out1, double& out2);
   int     get_contour_zrange(ChipsId* cid, double& out1, double& out2);
   int     set_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE val);
   int     get_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE& out_val);
   int     set_contour_numlevels(ChipsId* cid, int in_val);
   int     get_contour_numlevels(ChipsId* cid, int& out_val);
   int     set_contour_interval(ChipsId* cid, double in_val);
   int     get_contour_interval(ChipsId* cid, double& out_val);
   int     set_contour_depth(ChipsId* cid, int in_val);
   int     get_contour_depth(ChipsId* cid, int& out_val);

   // histogram operations
   int     add_histogram(ChipsId* cid, long cnt, double* bin_low,  
                         double* bin_high, double* counts, 
                         double* yd = NULL, double* yu = NULL,  
                         string attlist = ""); 
   int     delete_histogram(ChipsId* cid);
   int     display_histogram(ChipsId* cid);
   int     hide_histogram(ChipsId* cid);
   int     get_histogram_visible(ChipsId* cid, bool& out_val);
   int     set_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE val);
   int     get_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val);
   int     set_histogram_symbolfill(ChipsId* cid, bool in_fill);
   int     get_histogram_symbolfill(ChipsId* cid, bool& out_val);
   int     set_histogram_symbolsize(ChipsId* cid, int in_size);
   int     get_histogram_symbolsize(ChipsId* cid, int& out_val);
   int     set_histogram_symbolcolor(ChipsId* cid, string in_color);
   int     get_histogram_symbolcolor(ChipsId* cid, string& out_val);
   int     set_histogram_symbolangle(ChipsId* cid, float in_val);
   int     get_histogram_symbolangle(ChipsId* cid, float& out_val);
   int     set_histogram_linestyle(ChipsId* cid, V_LINE_TYPE val);
   int     get_histogram_linestyle(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_histogram_linethickness(ChipsId* cid, float in_size);
   int     get_histogram_linethickness(ChipsId* cid, float& out_val);
   int     set_histogram_linecolor(ChipsId* cid, string in_color);
   int     get_histogram_linecolor(ChipsId* cid, string& out_val);
   int     set_histogram_errstyle(ChipsId* cid, string style);
   int     get_histogram_errstyle(ChipsId* cid, string& out_val);
   int     set_histogram_errup(ChipsId* cid, bool val);
   int     get_histogram_errup(ChipsId* cid, bool& out_val);
   int     set_histogram_errdown(ChipsId* cid, bool val);
   int     get_histogram_errdown(ChipsId* cid, bool& out_val);
   int     set_histogram_dropline(ChipsId* cid, bool val);
   int     get_histogram_dropline(ChipsId* cid, bool& out_val);
   int     set_histogram_errthickness(ChipsId* cid, float in_thickness);
   int     get_histogram_errthickness(ChipsId* cid, float& out_val);
   int     set_histogram_errcolor(ChipsId* cid, string in_color);
   int     get_histogram_errcolor(ChipsId* cid, string& out_val);
   int     set_histogram_errcaplength(ChipsId* cid, int in_size);
   int     get_histogram_errcaplength(ChipsId* cid, int& out_val);
   int     set_histogram_fillcolor(ChipsId* cid, string in_color);
   int     get_histogram_fillcolor(ChipsId* cid, string& out_val);
   int     set_histogram_fillopacity(ChipsId* cid, double in_val);
   int     get_histogram_fillopacity(ChipsId* cid, double& out_val);
   int     set_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE in_style);
   int     get_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val);
   int     set_current_histogram(ChipsId* cid);
   int     get_histogram_range(ChipsId* cid, double& outx1, double& outx2,
                               double& outy1, double& outy2);
   int     get_histogram_xrange(ChipsId* cid, double& out1, double& out2);
   int     get_histogram_yrange(ChipsId* cid, double& out1, double& out2);
   int     set_histogram_depth(ChipsId* cid, int in_val);
   int     get_histogram_depth(ChipsId* cid, int& out_val);

   // axis  operations 
   int     add_axis(ChipsId* cid, int dim, float pos, float minx,
                    float maxx, float miny, float maxy, Transform *transform,
                    string aid="");
   int     delete_axis(ChipsId* cid);
   int     bind_axes(ChipsId* cid, string lid, string aid);
   int     unbind_axes(ChipsId* cid, string lid, string aid);
   int     set_axis_color(ChipsId* cid, string in_color);
   int     get_axis_color(ChipsId* cid, string& out_val);
   int     set_axis_major_color(ChipsId* cid, string in_color);
   int     get_axis_major_color(ChipsId* cid, string& out_val);
   int     set_axis_minor_color(ChipsId* cid, string in_color);
   int     get_axis_minor_color(ChipsId* cid, string& out_val);
   int     set_axis_text(ChipsId* cid, string in_text);
   int     get_axis_text(ChipsId* cid, string& out_val);
   int     set_axis_xlabel_text(ChipsId* cid, string in_text);
   int     get_axis_xlabel_text(ChipsId* cid, string& out_val);
   int     set_axis_ylabel_text(ChipsId* cid, string in_text);
   int     get_axis_ylabel_text(ChipsId* cid, string& out_val);
   int     set_axis_tickformat(ChipsId* cid, string in_format);
   int     get_axis_tickformat(ChipsId* cid, string& out_val);
   int     set_axis_thickness(ChipsId* cid, float in_thickness);
   int     get_axis_thickness(ChipsId* cid, float& out_val);
   int     set_axis_tickstyle(ChipsId* cid, int in_style);
   int     set_axis_tickstyle(ChipsId* cid, string in_style);
   int     get_axis_tickstyle(ChipsId* cid, string& out_val);
   int     set_axis_major_num(ChipsId* cid, int in_num);
   int     get_axis_major_num(ChipsId* cid, int& out_val);
   int     set_axis_minor_num(ChipsId* cid, int in_num);
   int     get_axis_minor_num(ChipsId* cid, int& out_val);
   int     set_axis_major_length(ChipsId* cid, int in_len);
   int     get_axis_major_length(ChipsId* cid, int& out_val);
   int     set_axis_minor_length(ChipsId* cid, int in_len);
   int     get_axis_minor_length(ChipsId* cid, int& out_val);
   int     set_axis_major_interval(ChipsId* cid, double in_len);
   int     get_axis_major_interval(ChipsId* cid, double& out_val);
   int     set_axis_minor_interval(ChipsId* cid, double in_len);
   int     get_axis_minor_interval(ChipsId* cid, double& out_val);
   int     set_current_axis(ChipsId* cid);
   int     get_axis_range(ChipsId* cid, double& out1, double& out2);

   // image operations
   int     add_image(ChipsId* cid, string filename, 
                     Transform *transform, string attlist="");
   int     add_image(ChipsId* cid, string filename, float ix1, float iy1,
                     float ix2, float iy2, Transform *transform,  string attlist = "");
   int     add_image(ChipsId* cid, double *rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, 
                     Transform *transform, string attlist = "");
   int     delete_image(ChipsId* cid);
   int     set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel);
   int     remove_image_channel(ChipsId* cid, int channel);
   int     print_image (ChipsId *cid, const string &outfile, const string &prefs);
   int     set_current_image(ChipsId* cid);
   int     set_image_colormap(ChipsId* cid, V_COLORMAP_TYPE val);
   int     get_image_colormap(ChipsId* cid, V_COLORMAP_TYPE& out_val);
   int     set_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE val);
   int     get_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE& out_val);
   int     set_image_invert_colormap(ChipsId* cid, bool inv);
   int     get_image_invert_colormap(ChipsId* cid, bool& out_val);
   int     set_image_colormap_interpolate(ChipsId* cid, bool interp);
   int     get_image_colormap_interpolate(ChipsId* cid, bool& out_val);
   int     set_image_colormap_size(ChipsId* cid, int size);
   int     get_image_colormap_size(ChipsId* cid, int& out_size);
   int     set_image_alpha(ChipsId* cid, double min, double max);
   int     get_image_alpha(ChipsId* cid, double& out1, double& out2);
   int     set_image_threshold(ChipsId* cid, double min, double max);
   int     get_image_threshold(ChipsId* cid, double& out1, double& out2);
   int     set_image_depth(ChipsId* cid, int in_val);
   int     get_image_depth(ChipsId* cid, int& out_val);
   int     display_image(ChipsId* cid);
   int     set_image_scale_channels(ChipsId* cid, bool scale);
   int     get_image_scale_channels(ChipsId* cid, bool& out_val);
   int     hide_image(ChipsId* cid);
   int     get_image_visible(ChipsId* cid, bool& out_val);
   int     get_image_range(ChipsId* cid, double& outx1, double& outx2,
                           double& outy1, double& outy2);
   int     get_image_xrange(ChipsId* cid, double& out1, double& out2);
   int     get_image_yrange(ChipsId* cid, double& out1, double& out2);



   // colorbar operations
   int     add_colorbar(ChipsId* cid, double xpos, double ypos,
                          string obj_id="", V_OBJECT_TYPE obj=chips_not_object, 
                          string attlist = "");
   int     delete_colorbar(ChipsId* cid);
   int     set_current_colorbar(ChipsId* cid);
   int     set_colorbar_depth(ChipsId* cid, int in_val);
   int     get_colorbar_depth(ChipsId* cid, int& out_val);
   int     set_colorbar_width(ChipsId* cid, float in_val);
   int     get_colorbar_width(ChipsId* cid, float& out_val);
   int     set_colorbar_length(ChipsId* cid, float in_val);
   int     get_colorbar_length(ChipsId* cid, float& out_val);
   int     display_colorbar(ChipsId* cid);
   int     hide_colorbar(ChipsId* cid);
   int     get_colorbar_visible(ChipsId* cid, bool& out_val);
   int     set_colorbar_border_visible(ChipsId* cid, bool val);
   int     get_colorbar_border_visible(ChipsId* cid, bool& val);
   int     set_colorbar_halign(ChipsId* cid, float sz);
   int     get_colorbar_halign(ChipsId* cid, float& out_val);
   int     set_colorbar_valign(ChipsId* cid, float sz);
   int     get_colorbar_valign(ChipsId* cid, float& out_val);
   int     set_colorbar_orientation(ChipsId* cid, int inval);
   int     get_colorbar_orientation(ChipsId* cid, int& out_val);
   int     set_colorbar_label_color(ChipsId* cid, string in_color);
   int     get_colorbar_label_color(ChipsId* cid, string& out_val);
   int     set_colorbar_label_visible(ChipsId* cid, bool vis);
   int     get_colorbar_label_visible(ChipsId* cid, bool& out_val);
   int     set_colorbar_label_font(ChipsId* cid, string font);
   int     get_colorbar_label_font(ChipsId* cid, string& out_val);
   int     set_colorbar_label_fontstyle(ChipsId* cid, string style);
   int     get_colorbar_label_fontstyle(ChipsId* cid, string& out_val);
   int     set_colorbar_label_size(ChipsId* cid, int sz);
   int     get_colorbar_label_size(ChipsId* cid, int& out_val);
   int     set_colorbar_label_angle(ChipsId* cid, float sz);
   int     get_colorbar_label_angle(ChipsId* cid, float& out_val);
   int     set_colorbar_label_halign(ChipsId* cid, float sz);
   int     get_colorbar_label_halign(ChipsId* cid, float& out_val);
   int     set_colorbar_label_valign(ChipsId* cid, float sz);
   int     get_colorbar_label_valign(ChipsId* cid, float& out_val);
   int     set_colorbar_label_location(ChipsId* cid, string style);
   int     set_colorbar_label_location(ChipsId* cid, int in_style);
   int     get_colorbar_label_location(ChipsId* cid, string& out_val);
   int     set_colorbar_label_text(ChipsId* cid, string text);
   int     get_colorbar_label_text(ChipsId* cid, string& out_val);
   int     set_colorbar_ticklabel_color(ChipsId* cid, string in_color);
   int     get_colorbar_ticklabel_color(ChipsId* cid, string& out_val);
   int     set_colorbar_ticklabel_size(ChipsId* cid, int in_val);
   int     get_colorbar_ticklabel_size(ChipsId* cid, int& out_val);
   int     set_colorbar_ticklabel_offset(ChipsId* cid, int in_val);
   int     get_colorbar_ticklabel_offset(ChipsId* cid, int& out_val);
   int     set_colorbar_ticklabel_angle(ChipsId* cid, float in_val);
   int     get_colorbar_ticklabel_angle(ChipsId* cid, float& out_val);
   int     set_colorbar_ticklabel_halign(ChipsId* cid, float in_val);
   int     get_colorbar_ticklabel_halign(ChipsId* cid, float& out_val);
   int     set_colorbar_ticklabel_valign(ChipsId* cid, float in_val);
   int     get_colorbar_ticklabel_valign(ChipsId* cid, float& out_val);
   int     set_colorbar_ticklabel_font(ChipsId* cid, string in_font);
   int     get_colorbar_ticklabel_font(ChipsId* cid, string& out_val);
   int     set_colorbar_ticklabel_fontstyle(ChipsId* cid, string style);
   int     get_colorbar_ticklabel_fontstyle(ChipsId* cid, string& out_val);
   int     set_colorbar_ticklabel_visible(ChipsId* cid, bool vis);
   int     get_colorbar_ticklabel_visible(ChipsId* cid, bool& out_val);
   int     set_colorbar_tick_thickness(ChipsId* cid, float in_thickness);
   int     get_colorbar_tick_thickness(ChipsId* cid, float& out_val);
   int     set_colorbar_tickstyle(ChipsId* cid, string in_style);
   int     set_colorbar_tickstyle(ChipsId* cid, int in_style);
   int     get_colorbar_tickstyle(ChipsId* cid, string& out_val);
   int     set_colorbar_mode(ChipsId* cid, string in_mode);
   int     get_colorbar_mode(ChipsId* cid, string& out_val);
   int     set_colorbar_parallel_offset(ChipsId* cid, float in_val);
   int     get_colorbar_parallel_offset(ChipsId* cid, float& out_val);
   int     set_colorbar_perpendicular_offset(ChipsId* cid, float in_val);
   int     get_colorbar_perpendicular_offset(ChipsId* cid, float& out_val);
   int     set_colorbar_tick_location(ChipsId* cid, string in_style);
   int     set_colorbar_tick_location(ChipsId* cid, int in_style);
   int     get_colorbar_tick_location(ChipsId* cid, string& out_val);
   int     set_colorbar_tick_visible(ChipsId* cid, bool vis);
   int     get_colorbar_tick_visible(ChipsId* cid, bool& out_val);
   int     set_colorbar_tick_color(ChipsId* cid, string in_color);
   int     get_colorbar_tick_color(ChipsId* cid, string& out_color);
   int     set_colorbar_tick_length(ChipsId* cid, int in_thickness);
   int     get_colorbar_tick_length(ChipsId* cid, int& out_val);





   // label operations
   int     add_label(ChipsId* cid, double xpos, double ypos, 
                     string text, string attlist = "");
   int     delete_label(ChipsId* cid);
   int     set_label_color(ChipsId* cid, string in_color);
   int     get_label_color(ChipsId* cid, string& out_val);
   int     set_label_text(ChipsId* cid, string in_text);
   int     get_label_text(ChipsId* cid, string& out_val);
   int     set_label_size(ChipsId* cid, int in_val);
   int     get_label_size(ChipsId* cid, int& out_val);
   int     set_label_depth(ChipsId* cid, int in_val);
   int     get_label_depth(ChipsId* cid, int& out_val);
   int     set_label_angle(ChipsId* cid, float in_val);
   int     get_label_angle(ChipsId* cid, float& out_val);
   int     set_label_halign(ChipsId* cid, float in_val);
   int     get_label_halign(ChipsId* cid, float& out_val);
   int     set_label_valign(ChipsId* cid, float in_val);
   int     get_label_valign(ChipsId* cid, float& out_val);
   int     set_label_font(ChipsId* cid, string in_font);
   int     get_label_font(ChipsId* cid, string& out_val);
   int     set_label_fontstyle(ChipsId* cid, string in_style);
   int     get_label_fontstyle(ChipsId* cid, string& out_val);
   int     hide_label(ChipsId* cid);
   int     display_label(ChipsId* cid);
   int     get_label_visible(ChipsId* cid, bool& out_val);
   int     set_current_label(ChipsId* cid);

   // line operations 
   int     add_line(ChipsId* cid, double startx, double starty,
                    double endx, double endy, string attlist = "");
   int     add_line(ChipsId* cid, double* xx, double* yy, 
                    unsigned int count, string attlist = "");
   int     delete_line(ChipsId* cid);
   int     set_line_color(ChipsId* cid, string in_color);
   int     get_line_color(ChipsId* cid, string& out_val);
   int     set_line_thickness(ChipsId* cid, float in_thickness);
   int     get_line_thickness(ChipsId* cid, float& out_val);
   int     set_line_style(ChipsId* cid, V_LINE_TYPE in_style);
   int     get_line_style(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE in_val);
   int     get_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE& out_val);
   int     set_line_depth(ChipsId* cid, int in_val);
   int     get_line_depth(ChipsId* cid, int& out_val);
   int     hide_line(ChipsId* cid);
   int     display_line(ChipsId* cid);
   int     get_line_visible(ChipsId* cid, bool& out_val);
   int     set_current_line(ChipsId* cid);

   // point operations
   int     add_point(ChipsId* cid,  double xx, double yy,
                     string attlist = "");
   int     add_point(ChipsId* cid, double xx, double yy, double zz,
                     string attlist = "");
   int     delete_point(ChipsId* cid);
   int     set_point_color(ChipsId* cid, string in_color);
   int     get_point_color(ChipsId* cid, string& out_val);
   int     set_point_size(ChipsId* cid, int in_size);
   int     get_point_size(ChipsId* cid, int& out_val);
   int     set_point_angle(ChipsId* cid, float in_val);
   int     get_point_angle(ChipsId* cid, float& out_val);
   int     set_point_fill(ChipsId* cid, bool in_fill);
   int     get_point_fill(ChipsId* cid, bool& out_val);
   int     set_point_depth(ChipsId* cid, int in_val);
   int     get_point_depth(ChipsId* cid, int& out_val);
   int     set_point_symbol(ChipsId* cid, V_SYMBOL_TYPE in_symbol);
   int     get_point_symbol(ChipsId* cid, V_SYMBOL_TYPE& out_val);
   int     hide_point(ChipsId* cid);
   int     display_point(ChipsId* cid);
   int     get_point_visible(ChipsId* cid, bool& out_val);
   int     set_current_point(ChipsId* cid);

   // legend operations
   int     add_legend(ChipsId* cid, const char** curve_ids, 
                      const char** curve_labels, string lgid = "");
   int     delete_legend(ChipsId* cid);
   int     hide_legend(ChipsId* cid);
   int     display_legend(ChipsId* cid);
   int     set_legend_color(ChipsId* cid, string in_color);
   int     set_legend_borders(ChipsId* cid, bool showborders);
   int     set_legend_thickness(ChipsId* cid, float in_thickness);
   int     set_legend_transparency(ChipsId* cid, bool seethrough); 
   int     set_current_legend(ChipsId* cid);

   // region operations 
   int     add_region(ChipsId* cid, double* xx, double* yy, 
                    unsigned int count, string attlist = "");
   int     add_region(ChipsId* cid, int sides, double xx, double yy, 
                      double radius, string attlist = "");
   int     delete_region(ChipsId* cid);
   int     hide_region(ChipsId* cid);
   int     display_region(ChipsId* cid);
   int     get_region_visible(ChipsId* cid, bool& out_val);
   int     set_region_depth(ChipsId* cid, int in_val);
   int     get_region_depth(ChipsId* cid, int& out_val);
   int     set_region_angle(ChipsId* cid, float in_val);
   int     get_region_angle(ChipsId* cid, float& out_val);
   int     set_region_thickness(ChipsId* cid, float in_thickness);
   int     get_region_thickness(ChipsId* cid, float& out_val);
   int     set_region_style(ChipsId* cid, V_LINE_TYPE in_style);
   int     get_region_style(ChipsId* cid, V_LINE_TYPE& out_val);
   int     set_region_color(ChipsId* cid, string in_color);
   int     get_region_color(ChipsId* cid, string& out_val);
   int     set_region_fillcolor(ChipsId* cid, string in_color);
   int     get_region_fillcolor(ChipsId* cid, string& out_val);
   int     set_region_fillstyle(ChipsId* cid, V_FILL_TYPE in_style);
   int     get_region_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val);
   int     set_region_opacity(ChipsId* cid, double in_val);
   int     get_region_opacity(ChipsId* cid, double& out_val);
   int     set_current_region(ChipsId* cid);



   // print operations
   int     print_window(ChipsId* cid, string outfile="", string prefs="");
   int     print_frame(ChipsId* cid, string outfile="");
   int     set_umask(mode_t mode)    { return ((int)umask(mode)); }

   // font operations
   int     load_font(string fontname, string fontpath,
                     V_FONTSTYLE_TYPE style = chips_normal); 

   char*   info(ChipsId* cid, V_OBJECT_TYPE obj=chips_window, bool descend=true);
   char*   info_coordinate(ChipsId* cid);
   char*   info_depth(ChipsId* cid, int depth=0);
   char*   info_bound_axes(ChipsId* cid); 
   char*   info_current();
   char*   get_current_frame(ChipsId* cid);

   // 'delete all' operations 
   int     delete_all_windows(ChipsId* cid);
   int     delete_all_frames(ChipsId* cid);
   int     delete_all_plots(ChipsId* cid);
   int     delete_all_images(ChipsId* cid);

   int     add_client_error(int errcode, string errtext); 
   char*   get_errors(int mode=0);
   int     clear_errors();
   int     set_error_verbosity(int val);
   int     get_error_verbosity(int &out_val);
   int     get_error_count(int mode, int &out_val);
   int     get_error_types(int &out_val);

   // coordinate operations
   int     add_data_transform (int row, int column, double **transform, string id="");
   int     remove_data_transform (string id="");
   int     apply_data_transform (ChipsId *cid, int row, int column, double **transform, string id="");
   int     apply_data_transform (ChipsId *cid, string id=""); 
   int     invert_data_transform (ChipsId *cid);
   int     convert_coordinate (double *in_val, double *out_val, int count, 
                               ChipsId *cid1, ChipsId *cid2, int axis = XY_AXIS);
   int     convert_coordinate (double in_val, double &out_val, 
                               ChipsId *cid1, ChipsId *cid2, int axis);

   ChipsEngine*  get_engine()  { return eng; }

  protected:
   int normalized_area_check(float x1, float y1, float x2, float y2); 
   void set_id_spec(ChipsId* supplied=NULL, bool use_all=0); 

  private:
   ChipsEngine* eng;   
   ChipsId      default_ids; 
   ChipsId*     id_spec; 
  
};

#endif 
