// 
//  Copyright (C) 2006,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsTextSupport.hh"
#include "ChipsFontList.hh"

#include <stdio.h>
#include <string.h>

typedef struct state_t
{
   double         angle;     // angle of text
   int            type;      // sub, norm, super
   V_FONTSTYLE_TYPE style;   // normal, bold, italic
   string         fontfile;  // fontfile to use
   string         font;      // font of label
   ChipsColor     color;     // color info for this element
   ELEM_P_T       parent;    // parent node in format tree
} STATE_T, *STATE_P_T;


extern ChipsFontList* theFonts;

#define CMODE_STANDARD        0
#define CMODE_SYMBOL          1
#define CMODE_SUBSCRIPT       2
#define CMODE_SUPERSCRIPT     3
#define CMODE_ITALIC          4
#define CMODE_ROMAN           5 
#define CMODE_UPRIGHT         6 
#define CMODE_BOLD            7 
#define CMODE_MEDIUM          8
#define CMODE_SANSERIF        9 
#define CMODE_TYPEWRITER     10 

static ELEM_P_T     fmt_tree;  
static list<STATE_P_T> state;
static STATE_P_T   curr_sp;
static int          popscript;   
int                 size;

// local function prototypes 
int       add_next(string text, string fnt, ELEM_PP_T pos);
int       add_subscript(string text, string fnt, ELEM_PP_T pos);
int       add_superscript(string text, string fnt, ELEM_PP_T pos);
int       get_token(char* in, char** start, char** stop, int& mode);
char*     parse_token_subscript(char* val, ELEM_PP_T e_p);
char*     parse_token_superscript(char* val, ELEM_PP_T e_p);
char*     parse_token_symbol(char* val, ELEM_PP_T e_p);
char*     parse_token_text(char* val, ELEM_PP_T e_p);
char*     parse_token_typestyle(char* val, ELEM_PP_T e_p, int comp);


// convert label string into tree structure to support sub/superscripts
int format_text(const string &text, const string &def_font, 
                V_FONTSTYLE_TYPE def_style, const string &def_color, 
                int def_size, ELEM_PP_T result)
{
   int sts = 0;
   int len;
   ELEM_P_T curr_p = NULL;
   char *buf;

   len = text.length();
   fmt_tree = NULL; 

   // set beginning text font and fontstyle
   size = def_size; 
   STATE_P_T sp = new STATE_T;
   sp->style = def_style; 
   sp->font = def_font;
   sp->type = 99; 
   sp->color.set_foreground_color(def_color);
   sp->parent = NULL; 
   state.push_back(sp);
   curr_sp = sp; 

   if ((buf = new char[len+1]) != NULL)
   {
      int tok, escaped = 0;
      char *pos, *tok_start = NULL, *tok_stop = NULL;

      // copy the string into temp buffer to tokenize 
      strcpy(buf, text.c_str());

      // replace any newlines ('\n') with spaces (' ') for now since
      // vtk can't support them cleanly with multiple font sizes yet
      for (tok = len; tok--;) if (buf[tok] == '\n') buf[tok] = ' ';
      tok = 0;

      pos = buf; 
      while (get_token(pos, &tok_start, &tok_stop, escaped) == 0)
      {
         switch(escaped)
         {
            case CMODE_SUPERSCRIPT:
               pos = parse_token_superscript(tok_start, &curr_p);
            break;
            case CMODE_SUBSCRIPT:
               pos = parse_token_subscript(tok_start, &curr_p);
            break;
            case CMODE_SYMBOL:
               pos = parse_token_symbol(tok_start, &curr_p);
            break;
            case CMODE_ITALIC:       // fallthrough intended
            case CMODE_BOLD:         // fallthrough intended
            case CMODE_UPRIGHT:      // fallthrough intended
            case CMODE_MEDIUM:       // fallthrough intended
            case CMODE_ROMAN:        // fallthrough intended
            case CMODE_SANSERIF:     // fallthrough intended
            case CMODE_TYPEWRITER:
               pos = parse_token_typestyle(tok_start, &curr_p, escaped);
            break; 
            default: 
               pos = parse_token_text(tok_start, &curr_p);
            break; 
         }
      }
      delete [] buf;
   }

   // remove state list entries
   STATE_P_T sob;
   for(list<STATE_P_T>::iterator ii=state.begin(); ii != state.end(); ++ii)
   {
      sob = (*ii);
      delete sob;
   }
   state.clear();

   *result = fmt_tree; 
   return sts; 
}


int add_next(string txt, string fnt, ELEM_PP_T pos)  
{
   if (!pos)  return -3;   // NULL pointer

   if (*pos == NULL)
   {
      // emtpy list
      if ((*pos = new ELEM_T) != NULL)
      {
         (*pos)->text = txt;
         (*pos)->super = NULL;
         (*pos)->next = NULL;
         (*pos)->sub = NULL;
         (*pos)->parent = NULL;
         (*pos)->size = size; 
         (*pos)->color.set_foreground_color(curr_sp->color.get_color()); 
         if (theFonts) 
         {
            (*pos)->fontfile = theFonts->get_path(fnt, curr_sp->style); 
            if ((*pos)->fontfile.length() == 0)
            {
               string tmpfont = "helvetica"; 
               (*pos)->fontfile = theFonts->get_path(tmpfont, curr_sp->style); 
            }
         }
         else  (*pos)->fontfile = "DEFAULT_HELVETICA"; 

         // set head of format tree
         fmt_tree = *pos;
      }
      else return -1; // alloc failed
   }
   else
   {
      // move to last segment
      while((*pos)->next != NULL) (*pos) = (*pos)->next;  

      if ((*pos)->next == NULL)
      {
	 string newff;
         if (theFonts) 
         {
            newff = theFonts->get_path(fnt, curr_sp->style); 
            if (newff.length() == 0)
            {
               string tmpfont = "helvetica";
               newff = theFonts->get_path(tmpfont, curr_sp->style);
            }
         }
         else  newff = "DEFAULT_HELVETICA"; 

        if (((*pos)->sub == NULL) && ((*pos)->super == NULL) &&
             (newff == (*pos)->fontfile) && 
	     ((*pos)->color.get_name() == curr_sp->color.get_name()) )
	 {
	    // append text to previous entry 
	    (*pos)->text.append(txt); 
	 }
	 else 
	 {
            ELEM_P_T tmp;
            if ((tmp = new ELEM_T) != NULL)
            {
               tmp->text = txt;
               tmp->super = NULL;
               tmp->next = NULL;
               tmp->sub = NULL;
               tmp->parent = *pos;
               (*pos)->next = tmp;
	       tmp->fontfile = newff;
               tmp->size = size; 
               tmp->color.set_foreground_color(curr_sp->color.get_color()); 
            }
            else return -1; // alloc failed
         }
      }
      else return -2;   // next contains entry
   }
   return 0;
}

int add_superscript(string txt, string fnt, ELEM_PP_T pos) 
{
   if (!pos)  return -3;   // NULL pointer

   if (*pos == NULL)
   {
      add_next("\0", fnt, pos);
   }
   if (*pos != NULL)
   {
      if ((*pos)->super == NULL)
      {
         ELEM_P_T tmp;
         if ((tmp = new ELEM_T) != NULL)
         {
            tmp->text = txt;
            tmp->super = NULL;
            tmp->next = NULL;
            tmp->sub = NULL;
            tmp->parent = *pos;
            tmp->size = size; 
            tmp->color.set_foreground_color(curr_sp->color.get_color());
            (*pos)->super = tmp;
            if (theFonts)
            {
               tmp->fontfile = theFonts->get_path(fnt, curr_sp->style);
               if (tmp->fontfile.length() == 0)
               {
                  string tmpfont = "helvetica";
                  tmp->fontfile = theFonts->get_path(tmpfont, curr_sp->style);
               }
            }
            else  tmp->fontfile = "DEFAULT_HELVETICA";
         }
         else return -1; // alloc failed
      }
      else
      {
         add_next(txt, fnt, &((*pos)->super));
      }
   }
   else return -4; // failed to allocate head of list
   return 0;
}

int add_subscript(string txt, string fnt, ELEM_PP_T pos)
{
   if (!pos)  return -3;   // NULL pointer

   if (*pos == NULL)
   {
      add_next("\0", fnt, pos);
   }

   if (*pos != NULL)
   {
      if ((*pos)->sub == NULL)
      {
         ELEM_P_T tmp;
         if ((tmp = new ELEM_T) != NULL)
         {
            tmp->text = txt;
            tmp->super = NULL;
            tmp->next = NULL;
            tmp->sub = NULL;
            tmp->parent = *pos;
            tmp->size = size; 
            tmp->color.set_foreground_color(curr_sp->color.get_color());
            (*pos)->sub = tmp;
            if (theFonts)
            {
               tmp->fontfile = theFonts->get_path(fnt, curr_sp->style);
               if (tmp->fontfile.length() == 0)
               {
                  string tmpfont = "helvetica";
                  tmp->fontfile = theFonts->get_path(tmpfont, curr_sp->style);
               }
            }
            else  tmp->fontfile = "DEFAULT_HELVETICA";
         }
         else return -1; // alloc failed
      }
      else
      {
         add_next(txt, fnt, &((*pos)->sub));
      }
   }
   else return -4; // failed to allocate head of list
   return 0;
}

int delete_tree(ELEM_P_T iter)
{
   if (iter)
   {
      if (iter->super)
      {
         delete_tree(iter->super);
         iter->super = NULL;
      }
      if (iter->sub)
      {
         delete_tree(iter->sub);
         iter->sub = NULL;
      }
      if (iter->next)
      {
         delete_tree(iter->next);
         iter->next = NULL;
      }
      delete(iter);
   }

   return 0;
}


char* parse_token_symbol(char* val, ELEM_PP_T e_p)
{
   char* pos = val; 
   char letter[2] = {0, 0};

   if ((pos[0] == 'A') && (pos[1] == 'A') &&
            (!isalpha(pos[2])))
   {
      letter[0] = 197; 
      if((curr_sp->font == "helvetica") || (curr_sp->font == "times") ||
         (curr_sp->font == "courier"))
         add_next(letter, curr_sp->font, e_p);  
      else 
         add_next(letter, "times", e_p);    // ANGSTROM
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2; 
   }
   else if (((strstr(pos, "alpha") == &pos[0]) ||
            (strstr(pos, "Alpha") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // ALPHA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if (((strstr(pos, "beta") == &pos[0]) ||
            (strstr(pos, "Beta") == &pos[0])) &&
            (!isalpha(pos[4])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // BETA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if (((strstr(pos, "gamma") == &pos[0]) ||
            (strstr(pos, "Gamma") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // GAMMA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if (((strstr(pos, "delta") == &pos[0]) ||
            (strstr(pos, "Delta") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // DELTA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if (((strstr(pos, "epsilon") == &pos[0]) ||
            (strstr(pos, "Epsilon") == &pos[0])) &&
            (!isalpha(pos[7])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // EPSILON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if (((strstr(pos, "zeta") == &pos[0]) ||
            (strstr(pos, "Zeta") == &pos[0])) &&
            (!isalpha(pos[4])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // ZETA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if (((strstr(pos, "eta") == &pos[0]) ||
            (strstr(pos, "Eta") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      if (pos[0] == 'e')   letter[0] = 'h';
         else               letter[0] = 'H'; 
      add_next(letter, "greek", e_p);    // ETA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "Theta1") == &pos[0]) && !isalpha(pos[6]))
   {
      letter[0] = 74; 
      add_next(letter, "greek", e_p);    // THETA1 (variant)
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if (((strstr(pos, "theta") == &pos[0]) ||
            (strstr(pos, "Theta") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      if (pos[0] == 't')  letter[0] = 'q';
         else               letter[0] = 'Q';  
      add_next(letter, "greek", e_p);    // THETA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if (((strstr(pos, "iota") == &pos[0]) ||
            (strstr(pos, "Iota") == &pos[0])) &&
            (!isalpha(pos[4])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // IOTA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if (((strstr(pos, "kappa") == &pos[0]) ||
            (strstr(pos, "Kappa") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // KAPPA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if (((strstr(pos, "lambda") == &pos[0]) ||
            (strstr(pos, "Lambda") == &pos[0])) &&
            (!isalpha(pos[6])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // LAMBDA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if (((strstr(pos, "mu") == &pos[0]) ||
            (strstr(pos, "Mu") == &pos[0])) &&
            (!isalpha(pos[2])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // MU
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if (((strstr(pos, "nu") == &pos[0]) ||
            (strstr(pos, "Nu") == &pos[0])) &&
            (!isalpha(pos[2])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // NU
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if (((pos[0] == 'x') || (pos[0] == 'X')) && 
            (pos[1] == 'i') && (!isalpha(pos[2])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // XI
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if (((strstr(pos, "omicron") == &pos[0]) ||
            (strstr(pos, "Omicron") == &pos[0])) &&
            (!isalpha(pos[7])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // OMICRON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if (((pos[0] == 'p') || (pos[0] == 'P')) && 
            (pos[1] == 'i') && (!isalpha(pos[2])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // PI
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2; 
   }
   else if (((strstr(pos, "rho") == &pos[0]) ||
            (strstr(pos, "Rho") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);   // RHO
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if (((strstr(pos, "sigma") == &pos[0]) ||
            (strstr(pos, "Sigma") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // SIGMA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "Stigma") == &pos[0]) && !isalpha(pos[6])) 
   {
      letter[0] = 86;
      add_next(letter, "greek", e_p);    // STIGMA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if (((strstr(pos, "tau") == &pos[0]) ||
            (strstr(pos, "Tau") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      letter[0] = pos[0];
      add_next(letter, "greek", e_p);    // TAU
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if (((strstr(pos, "upsilon") == &pos[0]) ||
            (strstr(pos, "Upsilon") == &pos[0])) &&
            (!isalpha(pos[7])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // UPSILON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "phi1") == &pos[0]) && !isalpha(pos[4]))
   {
      letter[0] = 106; 
      add_next(letter, "greek", e_p);    // PHI1 (variant)
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if (((strstr(pos, "phi") == &pos[0]) ||
            (strstr(pos, "Phi") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      if (pos[0] == 'p')  letter[0] = 'f';
         else               letter[0] = 'F'; 
      add_next(letter, "greek", e_p);    // PHI
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if (((strstr(pos,"chi") == &pos[0])||(strstr(pos,"Chi") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      letter[0] = pos[0]; 
      add_next(letter, "greek", e_p);    // CHI
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if (((strstr(pos, "psi") == &pos[0]) ||
            (strstr(pos, "Psi") == &pos[0])) &&
            (!isalpha(pos[3])))
   {
      if (pos[0] == 'p')  letter[0] = 'y';
         else               letter[0] = 'Y'; 
      add_next(letter, "greek", e_p);    // PSI
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if (((strstr(pos, "omega") == &pos[0]) ||
            (strstr(pos, "Omega") == &pos[0])) &&
            (!isalpha(pos[5])))
   {
      if (pos[0] == 'o')  letter[0] = 'w';
         else               letter[0] = 'W'; 
      add_next(letter, "greek", e_p);    // OMEGA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "omegapi") == &pos[0]) && !isalpha(pos[7]))
   {
      letter[0] = 118; 
      add_next(letter, "greek", e_p);    // OMEGAPI (variant)
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "Star") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 189;
      add_next(letter, "greek", e_p);    // FIVE SIDED STAR
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if ((strstr(pos, "cuberoot") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 191;
      add_next(letter, "greek", e_p);    // CUBE ROOT
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "FiveStar") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 189;
      add_next(letter, "greek", e_p);    // FIVE SIDED STAR
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "SixStar") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 92;
      add_next(letter, "greek", e_p);    // SIX SIDED STAR
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "Asterisk") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 93;
      add_next(letter, "greek", e_p);    // Asterisk
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "hbar") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 94;
      add_next(letter, "greek", e_p); 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if ((strstr(pos, "aries") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 161;
      add_next(letter, "greek", e_p);    // ARIES
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "eighthnote") == &pos[0]) && (!isalpha(pos[10]))) 
   {
      letter[0] = 162;
      add_next(letter, "greek", e_p);    // 8TH NOTE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=10;
   }
   else if ((strstr(pos, "maltese") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 211;
      add_next(letter, "greek", e_p);    // MALTESE CROSS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "pounds") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 163;
      add_next(letter, "times", e_p);    // POUNDS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "pm") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 177;
      add_next(letter, "greek", e_p);    // PLUS/MINUS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "rightturn") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 178;
      add_next(letter, "greek", e_p);    // CLOCKWISE ARROW
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "div") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 184;
      add_next(letter, "greek", e_p);    // DIVISION
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "times") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 180;
      add_next(letter, "greek", e_p);   // MULTIPLICATION
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "copyright") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 169;
      add_next(letter, "greek", e_p);    // COPYRIGHT
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "aa") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 229;
      if((curr_sp->font == "helvetica") || (curr_sp->font == "times") ||
         (curr_sp->font == "courier"))
         add_next(letter, curr_sp->font, e_p);
      else
         add_next(letter, "times", e_p);    // a w/ circle 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "ae") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 230;
      if((curr_sp->font == "helvetica") || (curr_sp->font == "times") ||
         (curr_sp->font == "courier"))
         add_next(letter, curr_sp->font, e_p);
      else
         add_next(letter, "times", e_p);    // ae
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "AE") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 198;
      if((curr_sp->font == "helvetica") || (curr_sp->font == "times") ||
         (curr_sp->font == "courier"))
         add_next(letter, curr_sp->font, e_p);
      else
         add_next(letter, "times", e_p);    // AE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "clubsuit") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 167;
      add_next(letter, "greek", e_p);    // CLUB
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "heartsuit") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 166;
      add_next(letter, "greek", e_p);    // HEART
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "spadesuit") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 165;
      add_next(letter, "greek", e_p);    // SPADE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "diamondsuit") == &pos[0]) && (!isalpha(pos[11])))
   {
      letter[0] = 168;
      add_next(letter, "greek", e_p);    // DIAMOND
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=11;
   }
   else if ((strstr(pos, "forall") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 34;
      add_next(letter, "greek", e_p);    // FOR ALL
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "hexagon") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 35;
      add_next(letter, "greek", e_p);    // HEXAGON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "exists") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 36;
      add_next(letter, "greek", e_p);    // EXISTS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "earth") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 38;
      add_next(letter, "greek", e_p);    // EARTH
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "pentagon") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 37;
      add_next(letter, "greek", e_p);    // PENTAGON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "ScissorRight") == &pos[0]) && (!isalpha(pos[12])))
   {
      letter[0] = 39;
      add_next(letter, "greek", e_p);    // SCISSORS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=12;
   }
   else if ((strstr(pos, "angle") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 208;
      add_next(letter, "greek", e_p);    // ANGLE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "neq") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 185;
      add_next(letter, "greek", e_p);    // NOT EQUAL  
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "equiv") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 186;
      add_next(letter, "greek", e_p);    // EQUIV
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "approx") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 187;
      add_next(letter, "greek", e_p);    // APPROX
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "bullet") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 183;
      add_next(letter, "greek", e_p);    // BULLET
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "infty") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 170;
      add_next(letter, "greek", e_p);    // INFINITY
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "otimes") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 196;
      add_next(letter, "greek", e_p);    // O TIMES 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "oplus") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 197;
      add_next(letter, "greek", e_p);    // O PLUS 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "odot") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 212;
      add_next(letter, "greek", e_p);    // SUN 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if ((strstr(pos, "sun") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 212;
      add_next(letter, "greek", e_p);    // SUN 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "astrosun") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 212;
      add_next(letter, "greek", e_p);    // SUN 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "smiley") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 203;
      add_next(letter, "greek", e_p);    // SMILEY FACE 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "frowny") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 126;
      add_next(letter, "greek", e_p);    // FROWNY FACE 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "watch") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 123;
      add_next(letter, "greek", e_p);    // WRIST WATCH
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "pluto") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 215;
      add_next(letter, "greek", e_p);    // PLUTO 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "mercury") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 210;
      add_next(letter, "greek", e_p);    // MERCURY
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "jupiter") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 217;
      add_next(letter, "greek", e_p);    // JUPITER
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "saturn") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 218;
      add_next(letter, "greek", e_p);    // SATURN
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "neptune") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 213;
      add_next(letter, "greek", e_p);    // NEPTUNE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "venus") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 226;
      add_next(letter, "greek", e_p);    // VENUS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "mars") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 227;
      add_next(letter, "greek", e_p);    // MARS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if ((strstr(pos, "uranus") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 228;
      add_next(letter, "greek", e_p);    // URANUS
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "rightmoon") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 229;
      add_next(letter, "greek", e_p);    // LAST QUARTER MOON
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "leftrightarrow") == &pos[0]) && (!isalpha(pos[14])))
   {
      letter[0] = 171;
      add_next(letter, "greek", e_p);    // LEFT-RIGHT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=14;
   }
   else if ((strstr(pos, "leftarrow") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 198;
      add_next(letter, "greek", e_p);    // LEFT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "uparrow") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 173;
      add_next(letter, "greek", e_p);    // UP ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "rightarrow") == &pos[0]) && (!isalpha(pos[10])))
   {
      letter[0] = 174;
      add_next(letter, "greek", e_p);    // RIGHT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=10;
   }
   else if ((strstr(pos, "downarrow") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 175;
      add_next(letter, "greek", e_p);    // DOWN ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "Leftrightarrow") == &pos[0]) && (!isalpha(pos[14])))
   {
      letter[0] = 219;
      add_next(letter, "greek", e_p);    // WIDE LEFT-RIGHT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=14;
   }
   else if ((strstr(pos, "Leftarrow") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 220;
      add_next(letter, "greek", e_p);    // WIDE LEFT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "Uparrow") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 221;
      add_next(letter, "greek", e_p);    // WIDE UP ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "Rightarrow") == &pos[0]) && (!isalpha(pos[10])))
   {
      letter[0] = 222;
      add_next(letter, "greek", e_p);    // WIDE RIGHT ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=10;
   }
   else if ((strstr(pos, "Downarrow") == &pos[0]) && (!isalpha(pos[9])))
   {
      letter[0] = 223;
      add_next(letter, "greek", e_p);    // WIDE DOWN ARROW 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=9;
   }
   else if ((strstr(pos, "aleph") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 192;
      add_next(letter, "greek", e_p);    // ALEPH 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "in") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 206;
      add_next(letter, "greek", e_p);    // IN 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "notin") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 207;
      add_next(letter, "greek", e_p);    // NOT IN 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "cap") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 199;
      add_next(letter, "greek", e_p);    // CAP
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "cup") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 200;
      add_next(letter, "greek", e_p);    // CUP
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "poi") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 96;
      add_next(letter, "greek", e_p);    // Place of Interest 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "supset") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 201;
      add_next(letter, "greek", e_p);    // SUPERSET
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "supseteq") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 202;
      add_next(letter, "greek", e_p);    // SUPERSET EQ
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "subset") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 204;
      add_next(letter, "greek", e_p);    // SUBSET
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "subseteq") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 205;
      add_next(letter, "greek", e_p);    // SUBSET EQ
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "leq") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 163;
      add_next(letter, "greek", e_p);    // LESS THAN OR EQUAL
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "geq") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 179;
      add_next(letter, "greek", e_p);    // GREATER THAN OR EQUAL
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "propto") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 181;
      add_next(letter, "greek", e_p);    // PROPTO
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "partial") == &pos[0]) && (!isalpha(pos[7])))
   {
      letter[0] = 182;
      add_next(letter, "greek", e_p);    // PARTIAL
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "Im") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 193;
      add_next(letter, "greek", e_p);    // IM
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "Re") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 194;
      add_next(letter, "greek", e_p);    // RE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "wp") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 195;
      add_next(letter, "greek", e_p);    // WP
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2;
   }
   else if ((strstr(pos, "ldots") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 188;
      add_next(letter, "greek", e_p);    // ELLIPSIS 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "leftmoon") == &pos[0]) && (!isalpha(pos[8])))
   {
      letter[0] = 190;
      add_next(letter, "greek", e_p);    // FIRST QUARTER MOON 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=8;
   }
   else if ((strstr(pos, "cong") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 64;
      add_next(letter, "greek", e_p);    // CONGRUENT
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if (((strstr(pos, "diamond") == &pos[0]) ||
             (strstr(pos, "lozenge") == &pos[0]))   && (!isalpha(pos[7])))
   {
      letter[0] = 224;
      add_next(letter, "greek", e_p);    // DIAMOND/LOZENGE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=7;
   }
   else if ((strstr(pos, "langle") == &pos[0]) && (!isalpha(pos[6])))
   {
      letter[0] = 225;
      add_next(letter, "greek", e_p);    // LANGLE
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=6;
   }
   else if ((strstr(pos, "lceil") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 124;
      add_next(letter, "greek", e_p);    // LCEIL
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "nabla") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 209;
      add_next(letter, "greek", e_p);    // NABLA
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "surd") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 214;
      add_next(letter, "greek", e_p);    // SURD
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=4;
   }
   else if ((strstr(pos, "neg") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 172;
      add_next(letter, "greek", e_p);    // NEG
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "int") == &pos[0]) && (!isalpha(pos[3])))
   {
      letter[0] = 125;
      add_next(letter, "greek", e_p);    // INT
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=3;
   }
   else if ((strstr(pos, "textdegree") == &pos[0]) && (!isalpha(pos[10])))
   {
      letter[0] = 176;
      if((curr_sp->font == "helvetica") || (curr_sp->font == "times") ||
         (curr_sp->font == "courier"))
         add_next(letter, curr_sp->font, e_p);
      else
         add_next(letter, "times", e_p);    // degree symbol 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=10;
   }
   else if ((strstr(pos, "circ") == &pos[0]) && (!isalpha(pos[4])))
   {
      letter[0] = 176;
      add_next(letter, "greek", e_p);   // CIRC
      pos+=4;
   }

   else if ((strstr(pos, "_") == &pos[0]) && (!isalpha(pos[1])))
   {
      letter[0] = '_';
      add_next(letter, curr_sp->font, e_p);   // underscore 
      pos++;
   }
   else if ((strstr(pos, "^") == &pos[0]) && (!isalpha(pos[1])))
   {
      letter[0] = '^';
      add_next(letter, curr_sp->font, e_p);   // carat
      pos++;
   }
   else if (strstr(pos, "{") == &pos[0])
   {
      letter[0] = '{';
      add_next(letter, curr_sp->font, e_p);
      pos++;
   }
   else if (strstr(pos, "}") == &pos[0])
   {
      letter[0] = '}';
      add_next(letter, curr_sp->font, e_p);
      pos++;
   }
   else if ((strstr(pos, "sf") == &pos[0]) && (!isalpha(pos[2])))
   {
      curr_sp->font = "helvetica"; 
      pos+=2;
   }
   else if ((strstr(pos, "tt") == &pos[0]) && (!isalpha(pos[2])))
   {
      curr_sp->font = "courier"; 
      pos+=2;
   }
   else if ((strstr(pos, "rm") == &pos[0]) && (!isalpha(pos[2])))
   {
      curr_sp->font = "times";
      pos+=2;
   }
   else if ((strstr(pos, "bf") == &pos[0]) && (!isalpha(pos[2])))
   {
      curr_sp->style = chips_bold;
      pos+=2;
   }
   else if ((strstr(pos, "it") == &pos[0]) && (!isalpha(pos[2])))
   {
      curr_sp->style = chips_italic;
      pos+=2;
   }
   else if ((strstr(pos, "color{") == &pos[0]))
   {
      pos +=6;
      char* ff = strstr(pos, "}");
      if (ff)
      {
         *ff = '\0';
         curr_sp->color.set_foreground_color(pos);
         pos = ff; 
         pos++; 
      } 
   }
   else if ((strstr(pos, "`{A}") == &pos[0]) || (strstr(pos, "`A") == &pos[0]) ||
            (strstr(pos, "grave{A}") == &pos[0]))
   {
      letter[0] = 192;
      add_next(letter, curr_sp->font, e_p);  // GRAVE ACCENT A
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{E}") == &pos[0]) || (strstr(pos, "`E") == &pos[0]) ||
            (strstr(pos, "grave{E}") == &pos[0]))
   {
      letter[0] = 200;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT E
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{I}") == &pos[0]) || (strstr(pos, "`I") == &pos[0]) ||
            (strstr(pos, "grave{I}") == &pos[0]))
   {
      letter[0] = 204;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT I
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{O}") == &pos[0]) || (strstr(pos, "`O") == &pos[0]) ||
            (strstr(pos, "grave{O}") == &pos[0]))
   {
      letter[0] = 210;
      add_next(letter, curr_sp->font, e_p);  // GRAVE ACCENT O 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{U}") == &pos[0]) || (strstr(pos, "`U") == &pos[0]) ||
            (strstr(pos, "grave{U}") == &pos[0]))
   {
      letter[0] = 217;
      add_next(letter, curr_sp->font, e_p);  // GRAVE ACCENT U 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{a}") == &pos[0]) || (strstr(pos, "`a") == &pos[0]) ||
            (strstr(pos, "grave{a}") == &pos[0]))
   {
      letter[0] = 224;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT a
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{e}") == &pos[0]) || (strstr(pos, "`e") == &pos[0]) ||
            (strstr(pos, "grave{e}") == &pos[0]))
   {
      letter[0] = 232;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT e
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{i}") == &pos[0]) || (strstr(pos, "`i") == &pos[0]) ||
            (strstr(pos, "grave{i}") == &pos[0]))
   {
      letter[0] = 236;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT i
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{o}") == &pos[0]) || (strstr(pos, "`o") == &pos[0]) ||
            (strstr(pos, "grave{o}") == &pos[0]))
   {
      letter[0] = 242;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // GRAVE ACCENT o
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "`{u}") == &pos[0]) || (strstr(pos, "`u") == &pos[0]) ||
            (strstr(pos, "grave{u}") == &pos[0]))
   {
      letter[0] = 249;
      add_next(letter, curr_sp->font, e_p);  // GRAVE ACCENT u 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[0] == 'g' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{A}") == &pos[0]) || (strstr(pos, "'A") == &pos[0]) ||
            (strstr(pos, "acute{A}") == &pos[0]))
   {
      letter[0] = 193;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT A
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{E}") == &pos[0]) || (strstr(pos, "'E") == &pos[0]) ||
            (strstr(pos, "acute{E}") == &pos[0]))
   {
      letter[0] = 201;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT E
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{I}") == &pos[0]) || (strstr(pos, "'I") == &pos[0]) ||
            (strstr(pos, "acute{I}") == &pos[0]))
   {
      letter[0] = 205;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT I
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{O}") == &pos[0]) || (strstr(pos, "'O") == &pos[0]) ||
            (strstr(pos, "acute{O}") == &pos[0]))
   {
      letter[0] = 211;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT O 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{U}") == &pos[0]) || (strstr(pos, "'U") == &pos[0]) ||
            (strstr(pos, "acute{U}") == &pos[0]))
   {
      letter[0] = 218;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT U 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{a}") == &pos[0]) || (strstr(pos, "'a") == &pos[0]) ||
            (strstr(pos, "acute{a}") == &pos[0]))
   {
      letter[0] = 225;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT a
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{e}") == &pos[0]) || (strstr(pos, "'e") == &pos[0]) ||
            (strstr(pos, "acute{e}") == &pos[0]))
   {
      letter[0] = 233;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT e
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{i}") == &pos[0]) || (strstr(pos, "'i") == &pos[0]) ||
            (strstr(pos, "acute{i}") == &pos[0]))
   {
      letter[0] = 237;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT i
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{o}") == &pos[0]) || (strstr(pos, "'o") == &pos[0]) ||
            (strstr(pos, "acute{o}") == &pos[0]))
   {
      letter[0] = 243;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // ACUTE ACCENT o
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{u}") == &pos[0]) || (strstr(pos, "'u") == &pos[0]) ||
            (strstr(pos, "acute{u}") == &pos[0]))
   {
      letter[0] = 250;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT u 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{Y}") == &pos[0]) || (strstr(pos, "'Y") == &pos[0]) ||
            (strstr(pos, "acute{Y}") == &pos[0]))
   {
      letter[0] = 221;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT Y 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "'{y}") == &pos[0]) || (strstr(pos, "'y") == &pos[0]) ||
            (strstr(pos, "acute{y}") == &pos[0]))
   {
      letter[0] = 253;
      add_next(letter, curr_sp->font, e_p);  // ACUTE ACCENT y 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'c' && pos[5] == '{') pos +=8;
            else  pos+=2; 
   }
   else if (strstr(pos, "hat{A}") == &pos[0])
   {
      letter[0] = 194;
      add_next(letter, curr_sp->font, e_p);  // CIRCUMFLEX A
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos += 6; 
   }
   else if (strstr(pos, "hat{E}") == &pos[0])
   {
      letter[0] = 202;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX E
      pos += 6; 
   }
   else if (strstr(pos, "hat{I}") == &pos[0])
   {
      letter[0] = 206;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX I
      pos += 6; 
   }
   else if (strstr(pos, "hat{O}") == &pos[0])
   {
      letter[0] = 212;
      add_next(letter, curr_sp->font, e_p);  // CIRCUMFLEX O 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos += 6; 
   }
   else if (strstr(pos, "hat{U}") == &pos[0])
   {
      letter[0] = 219;
      add_next(letter, curr_sp->font, e_p);  // CIRCUMFLEX U 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos += 6; 
   }
   else if (strstr(pos, "hat{a}") == &pos[0])
   {
      letter[0] = 226;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX a
      pos += 6; 
   }
   else if (strstr(pos, "hat{e}") == &pos[0]) 
   {
      letter[0] = 234;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX e
      pos += 6; 
   }
   else if (strstr(pos, "hat{i}") == &pos[0])
   {
      letter[0] = 238;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX i
      pos += 6; 
   }
   else if (strstr(pos, "hat{o}") == &pos[0])
   {
      letter[0] = 244;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // CIRCUMFLEX o
      pos += 6;
   }
   else if (strstr(pos, "hat{u}") == &pos[0])
   {
      letter[0] = 251;
      add_next(letter, curr_sp->font, e_p);  // CIRCUMFLEX u 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos += 6;
   }
   else if ((strstr(pos, "\"{A}") == &pos[0]) || (strstr(pos, "\"A") == &pos[0]) ||
            (strstr(pos, "ddot{A}") == &pos[0]))
   {
      letter[0] = 196;
      add_next(letter, curr_sp->font, e_p);  // UMLAUT A
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{E}") == &pos[0]) || (strstr(pos, "\"E") == &pos[0]) ||
            (strstr(pos, "ddot{E}") == &pos[0]))
   {
      letter[0] = 203;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT E
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{I}") == &pos[0]) || (strstr(pos, "\"I") == &pos[0]) ||
            (strstr(pos, "ddot{I}") == &pos[0]))
   {
      letter[0] = 207;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT I
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{O}") == &pos[0]) || (strstr(pos, "\"O") == &pos[0]) ||
            (strstr(pos, "ddot{O}") == &pos[0]))
   {
      letter[0] = 214;
      add_next(letter, curr_sp->font, e_p);  // UMLAUT O 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{U}") == &pos[0]) || (strstr(pos, "\"U") == &pos[0]) ||
            (strstr(pos, "ddot{U}") == &pos[0]))
   {
      letter[0] = 220;
      add_next(letter, curr_sp->font, e_p);  // UMLAUT U 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{a}") == &pos[0]) || (strstr(pos, "\"a") == &pos[0]) ||
            (strstr(pos, "ddot{a}") == &pos[0]))
   {
      letter[0] = 228;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT a 
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{e}") == &pos[0]) || (strstr(pos, "\"e") == &pos[0]) ||
            (strstr(pos, "ddot{e}") == &pos[0]))
   {
      letter[0] = 235;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT e
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{i}") == &pos[0]) || (strstr(pos, "\"i") == &pos[0]) ||
            (strstr(pos, "ddot{i}") == &pos[0]))
   {
      letter[0] = 239;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT i
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{o}") == &pos[0]) || (strstr(pos, "\"o") == &pos[0]) ||
            (strstr(pos, "ddot{o}") == &pos[0]))
   {
      letter[0] = 246;
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;  // UMLAUT o
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{u}") == &pos[0]) || (strstr(pos, "\"u") == &pos[0]) ||
            (strstr(pos, "ddot{u}") == &pos[0]))
   {
      letter[0] = 252;
      add_next(letter, curr_sp->font, e_p);  // UMLAUT u 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "\"{y}") == &pos[0]) || (strstr(pos, "\"y") == &pos[0]) ||
            (strstr(pos, "ddot{y}") == &pos[0]))
   {
      letter[0] = 255;
      add_next(letter, curr_sp->font, e_p);  // UMLAUT y 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[4] == '{' && pos[1] == 'd')  pos+=7;   
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{A}") == &pos[0]) || (strstr(pos, "~A") == &pos[0]) ||
            (strstr(pos, "tilde{A}") == &pos[0]))
   {
      letter[0] = 195;
      add_next(letter, curr_sp->font, e_p);  // TILDE A 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{N}") == &pos[0]) || (strstr(pos, "~N") == &pos[0]) ||
            (strstr(pos, "tilde{N}") == &pos[0]))
   {
      letter[0] = 209;
      add_next(letter, curr_sp->font, e_p);  // TILDE N 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{O}") == &pos[0]) || (strstr(pos, "~O") == &pos[0]) ||
            (strstr(pos, "tilde{O}") == &pos[0]))
   {
      letter[0] = 213;
      add_next(letter, curr_sp->font, e_p);  // TILDE O 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{a}") == &pos[0]) || (strstr(pos, "~a") == &pos[0]) ||
            (strstr(pos, "tilde{a}") == &pos[0]))
   {
      letter[0] = 227;
      add_next(letter, curr_sp->font, e_p);  // TILDE a 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{n}") == &pos[0]) || (strstr(pos, "~n") == &pos[0]) ||
            (strstr(pos, "tilde{n}") == &pos[0]))
   {
      letter[0] = 241;
      add_next(letter, curr_sp->font, e_p);  // TILDE n 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "~{o}") == &pos[0]) || (strstr(pos, "~o") == &pos[0]) ||
            (strstr(pos, "tilde{o}") == &pos[0]))
   {
      letter[0] = 245;
      add_next(letter, curr_sp->font, e_p);  // TILDE o 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4; 
         else if (pos[1] == 'i' and pos[5] == '{') pos+=8;
            else  pos+=2; 
   }
   else if ((strstr(pos, "c{C}") == &pos[0]) || (strstr(pos, "cC") == &pos[0]))
   {
      letter[0] = 199;
      add_next(letter, curr_sp->font, e_p);  // CEDILLA C 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4;   else  pos+=2; 
   }
   else if ((strstr(pos, "c{c}") == &pos[0]) || (strstr(pos, "cc") == &pos[0]))
   {
      letter[0] = 231;
      add_next(letter, curr_sp->font, e_p);  // CEDILLA c 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      if (pos[1] == '{') pos+=4;   else  pos+=2; 
   }
   else if ((strstr(pos, "DH") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 208;
      add_next(letter, curr_sp->font, e_p);  // DH 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2; 
   }
   else if ((strstr(pos, "dh") == &pos[0]) && (!isalpha(pos[2])))
   {
      letter[0] = 240;
      add_next(letter, curr_sp->font, e_p);  // dh 
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=2; 
   }
   else if ((strstr(pos, "fraka") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 230;
      add_next(letter, "greek", e_p);    // fraktur a
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakb") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 231;
      add_next(letter, "greek", e_p);    // fraktur b
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakc") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 232;
      add_next(letter, "greek", e_p);    // fraktur c
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakd") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 233;
      add_next(letter, "greek", e_p);    // fraktur d
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frake") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 234;
      add_next(letter, "greek", e_p);    // fraktur e
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakf") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 235;
      add_next(letter, "greek", e_p);    // fraktur f
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakg") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 236;
      add_next(letter, "greek", e_p);    // fraktur g
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakh") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 237;
      add_next(letter, "greek", e_p);    // fraktur h
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "fraki") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 238;
      add_next(letter, "greek", e_p);    // fraktur i
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakj") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 239;
      add_next(letter, "greek", e_p);    // fraktur j
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakk") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 240;
      add_next(letter, "greek", e_p);    // fraktur k
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakl") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 241;
      add_next(letter, "greek", e_p);    // fraktur l
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakm") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 242;
      add_next(letter, "greek", e_p);    // fraktur m
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakn") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 243;
      add_next(letter, "greek", e_p);    // fraktur n
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frako") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 244;
      add_next(letter, "greek", e_p);    // fraktur o
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakp") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 245;
      add_next(letter, "greek", e_p);    // fraktur p
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakq") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 246;
      add_next(letter, "greek", e_p);    // fraktur q
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakr") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 247;
      add_next(letter, "greek", e_p);    // fraktur r
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "fraks") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 248;
      add_next(letter, "greek", e_p);    // fraktur s
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakt") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 249;
      add_next(letter, "greek", e_p);    // fraktur t
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "fraku") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 250;
      add_next(letter, "greek", e_p);    // fraktur u
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakv") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 251;
      add_next(letter, "greek", e_p);    // fraktur v
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakw") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 252;
      add_next(letter, "greek", e_p);    // fraktur w
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakx") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 253;
      add_next(letter, "greek", e_p);    // fraktur x
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "fraky") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 254;
      add_next(letter, "greek", e_p);    // fraktur y
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((strstr(pos, "frakz") == &pos[0]) && (!isalpha(pos[5])))
   {
      letter[0] = 255;
      add_next(letter, "greek", e_p);    // fraktur z
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=5;
   }
   else if ((pos[0] == '\\') && (!isalpha(pos[1])))
   {
      // NOTE: This deviates from latex which treats \\ as a newline
      letter[0] = pos[0];  
      add_next(letter,  curr_sp->font, e_p);    // BACKSLASH
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos+=1;
   }
   else              // OTHER ESCAPED SEQUENCE (SYMBOL/FONT)
   {
      add_next(pos, curr_sp->font, e_p);
      pos += strlen(pos);
      if ((*e_p)->next) *e_p = (*e_p)->next;
   }
   if (popscript) 
   {
      popscript = 0; 
      while((*e_p)->parent && (*e_p)->parent->next)
         *e_p = (*e_p)->parent;
      if ((*e_p)->parent) *e_p = (*e_p)->parent;
   }
   return pos; 
}

char* parse_token_subscript(char* val, ELEM_PP_T e_p)
{
   char* pos = val; 

   if (pos[0] == '\\' || pos[0] == '\0') 
   {
      char letter[2] = {0,0}; 
      add_subscript(letter, curr_sp->font, e_p);
      if ((*e_p)->sub) *e_p = (*e_p)->sub;
      popscript = 1; 
   }
   else if (pos[0] != '{')
   {
      char tmp = pos[1];
      pos[1] = '\0';
      add_subscript(pos, curr_sp->font, e_p);
      pos[1] = tmp;
      pos++;
   }
   else
   {
      int ll, ff, bcnt = 0, done = 0;
      pos++; 
      STATE_P_T tsp = new STATE_T;
      tsp->angle = curr_sp->angle; 
      tsp->type = CMODE_SUBSCRIPT; 
      tsp->style = curr_sp->style; 
      tsp->font = curr_sp->font; 
      tsp->color.set_foreground_color(curr_sp->color.get_name()); 
      tsp->parent = *e_p;
      state.push_back(tsp);
      curr_sp = tsp;
      ll = strlen(pos);
      for (ff = 0; (!done && (ff < ll)); ff++)
      {
         if (pos[ff] == '}') 
	 {
            if (curr_sp && (curr_sp->type == CMODE_SUBSCRIPT))
	    {
	       pos[ff] = '\0';
               add_subscript(pos, curr_sp->font, e_p);
	       pos = &pos[ff];
	       pos++;
               done = 1; 
	    }
	    if (!state.empty())
	    {
               if (curr_sp->type != 99) 
               {
                  STATE_P_T sob = state.back(); 
                  delete sob; 
                  state.pop_back();
               }
               curr_sp = state.back();
            }
         }
         else if (pos[ff] == '{')
	 {
            pos[ff] = '\0';
            add_subscript(pos, curr_sp->font, e_p);
            if ((*e_p)->sub) *e_p = (*e_p)->sub;
            pos = &pos[ff];
            pos++;
            done = 1;

            tsp = new STATE_T;
            tsp->angle = curr_sp->angle;
            tsp->type = 0;
            tsp->style = curr_sp->style;
            tsp->font = curr_sp->font;
            tsp->color.set_foreground_color(curr_sp->color.get_name());
	    tsp->parent = *e_p; 
            state.push_back(tsp);
            curr_sp = tsp; 
	    bcnt++;
	 }
      }
      if (!done)
      {
         char letter[2] = {0,0};
	 add_subscript(letter, curr_sp->font, e_p);
         if ((*e_p)->sub) *e_p = (*e_p)->sub;
         for (ff = 0; ff < bcnt; ff++)
	 {
            if (curr_sp->type != 99)
            {
               STATE_P_T sob = state.back();
               delete sob;
               state.pop_back();
            }
            curr_sp = state.back();
         }
      }
   }
   return pos; 
}


char* parse_token_superscript(char* val, ELEM_PP_T e_p)
{
   char* pos = val; 

   if (pos[0] == '\\' || pos[0] == '\0') 
   {
      char letter[2] = {0,0}; 
      add_superscript(letter, curr_sp->font, e_p);
      if ((*e_p)->super) *e_p = (*e_p)->super;
      popscript = 1; 
   }
   else if (pos[0] != '{')
   {
      char tmp = pos[1];
      pos[1] = '\0';
      add_superscript(pos, curr_sp->font, e_p);
      pos[1] = tmp;
      pos++;
   }
   else
   {
      int ff, ll, bcnt = 0, done = 0;
      pos++; 
      STATE_P_T tsp = new STATE_T;
      tsp->angle = curr_sp->angle;
      tsp->type = CMODE_SUPERSCRIPT;
      tsp->style = curr_sp->style;
      tsp->font = curr_sp->font;
      tsp->color.set_foreground_color(curr_sp->color.get_name());
      tsp->parent = *e_p; 
      state.push_back(tsp);
      curr_sp = tsp;
      ll = strlen(pos);
      for (ff = 0; (!done && (ff < ll)); ff++)
      {
         if (pos[ff] == '}') 
	 {
            if (curr_sp && (curr_sp->type == CMODE_SUPERSCRIPT))
	    {
	       pos[ff] = '\0';
	       add_superscript(pos, curr_sp->font, e_p);
	       pos = &pos[ff];
	       pos++;
               done =1; 
	    }
            if (!state.empty()) 
	    {
               if (curr_sp->type != 99)
               {
                  STATE_P_T sob = state.back();
                  delete sob;
                  state.pop_back();
               }
               curr_sp = state.back();
            }
         }
         else if (pos[ff] == '{')
	 {
            pos[ff] = '\0';
	    add_superscript(pos, curr_sp->font, e_p);
            if ((*e_p)->super) *e_p = (*e_p)->super;
            pos = &pos[ff];
            pos++;
            done = 1;
            tsp = new STATE_T;
            tsp->angle = curr_sp->angle;
            tsp->type = 0;
            tsp->style = curr_sp->style;
            tsp->font = curr_sp->font;
            tsp->color.set_foreground_color(curr_sp->color.get_name());
	    tsp->parent = *e_p; 
            state.push_back(tsp);
            curr_sp = tsp;
	    bcnt++;
	 }
      }
      if (!done)
      {
         char letter[2] = {0,0};
	 add_superscript(letter, curr_sp->font, e_p);
         if ((*e_p)->super) *e_p = (*e_p)->super;
         for (ff = 0; ff < bcnt; ff++)
	 {
            if (curr_sp->type != 99)
            {
               STATE_P_T sob = state.back();
               delete sob;
               state.pop_back();
            }
            curr_sp = state.back();
         }
      }
   }
   return pos; 
}

char* parse_token_text(char* val, ELEM_PP_T e_p)
{
   char* pos = val; 
   int done = 0;
   int ff, ll = strlen(pos);
   for (ff = 0; (!done && (ff < ll)); ff++)
   {
      if (pos[ff] == '}') 
      {
         STATE_P_T tsp = NULL;
         if (!state.empty())   tsp = state.back();
         if (tsp)
         {
            curr_sp = tsp; 
            pos[ff] = '\0';
            add_next(pos, curr_sp->font, e_p);
            pos = &pos[ff];
            pos++;
            if (tsp->type == 0)
            {
               if ((*e_p) && (*e_p)->next)   *e_p = (*e_p)->next;
            }
            else if ((tsp->type == CMODE_SUPERSCRIPT) ||
                     (tsp->type == CMODE_SUBSCRIPT))
            {
               if (tsp->parent)
                  *e_p = tsp->parent;
               else
               {
                  while((*e_p)->parent && (*e_p)->parent->next)
                     *e_p = (*e_p)->parent;
                  if ((*e_p)->parent) *e_p = (*e_p)->parent;
               }
            }
            if (!state.empty())
            {
               if (curr_sp->type != 99)
               {
                  delete curr_sp;
                  state.pop_back();
                  curr_sp = state.back();
               }
            }
         }
         else if (pos[0] == '}')
         {
            pos = &pos[1];
         }
         done = 1;
      }
      else if (pos[ff] == '{')
      {
	 pos[ff] = '\0';
	 if (ff>0)   add_next(pos, curr_sp->font, e_p);
         if ((*e_p) && (*e_p)->next) *e_p = (*e_p)->next;
         pos = &pos[ff];
         pos++;
         done = 1; 
         STATE_P_T tsp = new STATE_T;
         tsp->angle = curr_sp->angle;
         tsp->type = 0;
         tsp->style = curr_sp->style;
         tsp->font = curr_sp->font;
         tsp->color.set_foreground_color(curr_sp->color.get_name());
	 tsp->parent = *e_p; 
         state.push_back(tsp);
         curr_sp = tsp; 
      }
   }
   if (!done)
   {
      add_next(val, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;
      pos = val + strlen(val);
   }
   return pos; 
}

int get_token(char* in, char** start, char** stop, int& mode)
{
   static char rep = '\0';
   static char* prev = NULL;

   int rval = 0; 
   if (in && start && stop)
   {
      char* last;
      // reset end character of last token if necessary
      if (prev && (in == prev) && (rep != '\0')) in[0] = rep;

      if (in[0] != '\0') 
      {
         if (in[0] == '_')
         {
            mode = CMODE_SUBSCRIPT; 
            *start = &in[1]; 
         }
         else if (in[0] == '^')
         {
            mode = CMODE_SUPERSCRIPT; 
            *start = &in[1]; 
         }
         else if (in[0] == '\\')
         {
            if ((strstr(in, "textit") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_ITALIC;
               *start = &in[7];
            }
            else if ((strstr(in, "textup") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_UPRIGHT;
               *start = &in[7];
            }
            else if ((strstr(in, "textbf") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_BOLD;
               *start = &in[7];
            }
            else if ((strstr(in, "textmd") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_MEDIUM;
               *start = &in[7];
            }
            else if ((strstr(in, "textrm") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_ROMAN;
               *start = &in[7];
            }
            else if ((strstr(in, "textsf") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_SANSERIF;
               *start = &in[7];
            }
            else if ((strstr(in, "texttt") == &in[1]) && (!isalpha(in[7])))
            {
               mode = CMODE_TYPEWRITER;
               *start = &in[7];
            }
            else if (strlen(in) > 1 && (in[1] == '_' || in[1] == '^')) 
            {
               // handle underscore or carat  
               mode = CMODE_STANDARD; 
               *start = &in[1]; 
            }
            else
            {
               mode = CMODE_SYMBOL; 
               *start = &in[1]; 
            }
         }
         else 
         {
            mode = CMODE_STANDARD; 
            *start = &in[0]; 
         }

	 // new token or inside previous one?
         if (!prev || (in == prev)) 
         {
            last = *start;
	    last++;
            while ((*last != '\\') && (*last != '_') && (*last != '^') &&
                   (*last != '\0'))  last++; 
            rep = *last;
            if (*last != '\0')
            {
               *last = '\0';  
            *stop = last; 
            prev = last; 
            }
	    else 
	    {
            *stop = last; 
            prev = NULL; 
	    }
        }
      }
      else  rval = 1; // done 
   }
   else  rval = -1;  // no string to parse

   return rval; 
}


char* parse_token_typestyle(char* val, ELEM_PP_T e_p, int comp)
{
   char* pos = val;
   V_FONTSTYLE_TYPE tmpstyle = curr_sp->style;

   if (pos[0] == '\\')
   {
      char letter[2] = {0,0};
      add_next(letter, curr_sp->font, e_p);
      if ((*e_p)->next) *e_p = (*e_p)->next;
      popscript = 1;
   }
   else if (pos[0] != '{')
   {
      int ndx = 0;
      // ignore white space
      while ( isspace(pos[ndx])) ndx++; 
      char tmp = pos[++ndx];
      pos[ndx] = '\0';
      switch(comp)
      {
         case CMODE_ROMAN:
            add_next(pos, "times", e_p);
         break;
         case CMODE_SANSERIF:
            add_next(pos, "helvetica", e_p);
         break;
         case CMODE_TYPEWRITER:
            add_next(pos, "courier", e_p);
         break;
         case CMODE_ITALIC:
            if (curr_sp->style == chips_bold)
               curr_sp->style = chips_bolditalic;
            else
               curr_sp->style = chips_italic;
            add_next(pos, curr_sp->font, e_p);
         break;
         case CMODE_UPRIGHT:
            if (curr_sp->style == chips_bolditalic) 
               curr_sp->style = chips_bold;
            else if (curr_sp->style != chips_bold) 
               curr_sp->style = chips_normal;
            add_next(pos, curr_sp->font, e_p);
         break;
         case CMODE_BOLD:
            if (curr_sp->style == chips_italic)
               curr_sp->style = chips_bolditalic;
            else
               curr_sp->style = chips_bold;
            add_next(pos, curr_sp->font, e_p);
         break;
         case CMODE_MEDIUM:
            if (curr_sp->style == chips_bolditalic)
               curr_sp->style = chips_italic;
            else if (curr_sp->style != chips_italic)
               curr_sp->style = chips_normal;
            add_next(pos, curr_sp->font, e_p);
         break;
         default:
         break;  
      }
      pos[ndx] = tmp;
      pos = &pos[ndx];
      curr_sp->style = tmpstyle;
   }
   else
   {
      pos++;
      STATE_P_T tsp = new STATE_T;
      tsp->angle = curr_sp->angle;
      tsp->type = comp;
      tsp->style = curr_sp->style;
      tsp->font = curr_sp->font;
      tsp->color.set_foreground_color(curr_sp->color.get_name());
      tsp->parent = *e_p; 
      state.push_back(tsp);

      switch(comp)
      {
         case CMODE_ROMAN:
            tsp->font = "times";
         break;
         case CMODE_SANSERIF:
            tsp->font = "helvetica";
         break;
         case CMODE_TYPEWRITER:
            tsp->font = "courier";
         break;
         case CMODE_ITALIC:
            if (tsp->style == chips_bold)
               tsp->style = chips_bolditalic;
            else
               tsp->style = chips_italic;
         break;
         case CMODE_UPRIGHT:
            if (tsp->style == chips_bolditalic) 
               tsp->style = chips_bold;
            else if (tsp->style != chips_bold) 
               tsp->style = chips_normal;
         break;
         case CMODE_BOLD:
            if (tsp->style == chips_italic)
               tsp->style = chips_bolditalic;
            else
               tsp->style = chips_bold;
         break;
         case CMODE_MEDIUM:
            if (tsp->style == chips_bolditalic)
               tsp->style = chips_italic;
            else if (tsp->style != chips_italic)
               tsp->style = chips_normal;
         break;
         default: 
         break; 
      }
      curr_sp = tsp;

      int bcnt = 0, done = 0;
      int ff, ll = strlen(pos);
      for (ff = 0; (!done && (ff < ll)); ff++)
      {
         if (pos[ff] == '}')
         {
            if (curr_sp && (curr_sp->type == comp))
            {
               pos[ff] = '\0';
               add_next(pos, curr_sp->font, e_p);
               if ((*e_p)->next) *e_p = (*e_p)->next;
               pos = &pos[ff];
               pos++;
               done =1;
            }
            if (curr_sp->type != 99)
            {
               STATE_P_T sob = state.back();
               delete sob;
               state.pop_back();
            }
            curr_sp = state.back();
         }
         else if (pos[ff] == '{')
         {
            pos[ff] = '\0';
            add_next(pos, curr_sp->font, e_p);
            if ((*e_p)->next) *e_p = (*e_p)->next;
            pos = &pos[ff];
            pos++;
            done = 1;

            tsp = new STATE_T;
            tsp->angle = curr_sp->angle;
            tsp->type = 0;
            tsp->style = curr_sp->style;
            tsp->font = curr_sp->font;
            tsp->color.set_foreground_color(curr_sp->color.get_name());
	    tsp->parent = *e_p; 
            state.push_back(tsp);
            curr_sp = tsp;
            bcnt++;
         }
      }
      if (!done)
      {
         for (ff = 0; ff < bcnt; ff++)
         {
            if (curr_sp->type != 99)
            {
               STATE_P_T sob = state.back();
               delete sob;
               state.pop_back();
            }
            curr_sp = state.back();
         }
      }
   }
   return pos;
}


