// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsTransformCommand.hh"
#include "ChipsStateStream.hh"
#include "ChipsTransformFactory.hh"
#include "ChipsContour.hh"


ChipsTransformCommand::ChipsTransformCommand(ChipsObject *co,
   int ctype, Transform *otrans, Transform *ntrans)
{
   ChipsTransformFactory tf;

   this->co = co;
   this->old_transform = ChipsTransformFactory::copy_transform (otrans);
   this->new_transform = ChipsTransformFactory::copy_transform (ntrans);
   this->cmd_type = ctype;  
   this->weight=1;
}

ChipsTransformCommand::~ChipsTransformCommand()
{
   if (old_transform)
      delete old_transform;
   if (new_transform)
      delete new_transform;
}

ChipsTransformCommand::ChipsTransformCommand(ChipsObject* co,
   ChipsStateStream& in)
{
   this->co = co;
   deserialize(in);
   weight=1;
}

ChipsTransformCommand::ChipsTransformCommand(ChipsStateStream& in)
{
   this->co = NULL;
   deserialize(in);
   weight=1;
}

int ChipsTransformCommand::execute ()
{
   int rval = ERR_STATUS_GOOD;
   ChipsContour *ctr;

   // current only supported chips object is contour

   if (!co)
      return ERR_STATUS_BAD;

   // cmd_type of 1 is set_transform
   // cmd_type of 2 is apply_transform
   // cmd_type of 3 is invert_transform

   switch (cmd_type)
   {
      case 1:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->set_transform (new_transform, false);
      break;
      case 2:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->apply_transform (false);
      break;
      case 3:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->apply_transform_inverse (false);
      break;
      default:
         rval = ERR_STATUS_BAD;
      break;
   }

   return rval;
}

int ChipsTransformCommand::unexecute ()
{
   int rval = ERR_STATUS_BAD;
   ChipsContour *ctr;

   // current only supported chips object is contour

   if (!co)
      return ERR_STATUS_BAD;

   // cmd_type of 1 is set_transform
   // cmd_type of 2 is apply_transform
   // cmd_type of 3 is invert_transform

   switch (cmd_type)
   {
      case 1:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->set_transform (old_transform, false);
      break;
      case 2:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->apply_transform_inverse (false);
      break;
      case 3:
         if ((ctr = dynamic_cast <ChipsContour *>(co)))
            rval = ctr->apply_transform (false);
      break;
      default:
         rval = ERR_STATUS_BAD;
      break;
   }

   return rval;
}


void ChipsTransformCommand::serialize(ChipsStateStream& out)
{
   int val = 18;
   bool trans_exist = true;
   out.write_int(&val);
   serialize_object(out);
   out.write_int(&ref);
   out.write_int(&block_ref);

   out.write_int(&cmd_type);

   if (old_transform)
   {
      trans_exist = true;
      ChipsTransformFactory tf;
      out.write_bool (&trans_exist);
      tf.serialize (old_transform, out);
   }
   else
   {
      trans_exist = false;
      out.write_bool (&trans_exist);
   }

   if (new_transform)
   {
      trans_exist = true;
      ChipsTransformFactory tf;
      out.write_bool (&trans_exist);
      tf.serialize (new_transform, out);
   }
   else
   {
      trans_exist = false;
      out.write_bool (&trans_exist);
   }

}


void ChipsTransformCommand::deserialize(ChipsStateStream& in)
{
   bool trans_exist;

   in.read_int(&ref);
   in.read_int(&block_ref);

   in.read_int(&cmd_type);

   in.read_bool (&trans_exist);
   if (trans_exist)
   {
      ChipsTransformFactory tf;
      old_transform = tf.deserialize (in);
   }
   else
      old_transform = static_cast <Transform *> (0);

   in.read_bool (&trans_exist);
   if (trans_exist)
   {
      ChipsTransformFactory tf;
      new_transform = tf.deserialize (in);
   }
   else
      new_transform = static_cast <Transform *> (0);

}



void ChipsTransformCommand::print ()
{
   cerr << "   [ "<<ref<<", "<< block_ref <<"] " <<"Transform " << endl;
}


