// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsTransformFactory.hh"
#include "ChipsStateStream.hh"

#include <list>
#include <string.h>

#ifndef NO_CMD_MANAGER
#include "ChipsCmdManager.hh"
extern ChipsCmdManager* cmd_manager;
#endif

ChipsTransformFactory::ChipsTransformFactory ()
{
}

ChipsTransformFactory::~ChipsTransformFactory ()
{
}

Transform *ChipsTransformFactory::checkValidTransform (Transform *trans)
{
   if (dynamic_cast <WCSTransform *>(trans) != 0)
   {
      string *tctype = (string *)(trans->get_parameter_value ("CTYPE"));
      if (tctype[0] != "RA---TAN" && tctype[1] != "DEC--TAN")
      {
         //trans = NULL;// - Old method. Now we replace with a 2D transform that's the identity matrix
         //This is neccesary so it won't auto-use the physical transform a plot has.
         
         //Initializes to identity matrix automatically
         trans = ChipsTransformFactory::new_transform (ChipsTransformFactory::LINEAR2D);
         
         #ifndef NO_CMD_MANAGER
         cmd_manager->add_err(ERR_TRANS_NOT_SUPP, "1", 1, tctype[0].c_str(), tctype[1].c_str());
         //cmd_manager->add_err ("Unsupported transform. Using base coordiantes", rval, "1", 3);
         #endif
      }
   }
   
   return trans;
}

unsigned char *ChipsTransformFactory::serialize (Transform *trans, int &length)
{
   list<TransParam*> trans_params;
   list<TransParam*>::iterator iter;
   dmDataType trval = DT_UNK;
   int ival = 0;
   long lval = 0;
   int size = 0;

   unsigned char *out = NULL;
   int pos = 0;

   length = 0;
   if (!trans) 
      return out;

   length = get_transform_size (trans);

   // this extra int accounts for the off the size of the parameter list being
   // stored off.  Note this happens right before the parameter list loop
   length += sizeof (int);

   // mark which type of transform this is 
   if ((dynamic_cast <WCSTransform *>(trans) != 0) || (dynamic_cast <WCSTANTransform *>(trans) != 0))
   {
      length += 4;
      out = new unsigned char [length];
      memset (out, 0, length);
      memcpy (out, "WCS\0", 4);
      pos = 4;
   }
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
   {
      length += 9;
      out = new unsigned char [length];
      memset (out, 0, length);
      memcpy (out, "LINEAR2D\0", 9);
      pos = 9;
   }
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
   {
      length += 6;
      out = new unsigned char [length];
      memset (out, 0, length);
      memcpy (out, "SPLIT\0", 6);
      pos = 6;
   }
   else
   {
      length += 5;
      out = new unsigned char [length];
      memset (out, 0, length);
      memcpy (out, "base\0", 5);
      pos = 5;
   }

   // copy in the name
   memcpy (&out[pos],trans->get_name ().c_str (), trans->get_name ().length ()+1);
   pos += trans->get_name ().length ()+1;


   // go through parameter list and copy out all parameter values
   trans_params = trans->get_parameter_list ();
   ival = trans_params.size ();
   memcpy (&out[pos],(void *)(&ival), sizeof (ival));
   pos += sizeof (ival);
   for (iter = trans_params.begin (); iter != trans_params.end (); iter++)
   {
      memcpy (&out[pos], (*iter)->get_name ().c_str (),  (*iter)->get_name ().length () + 1);
      pos += (*iter)->get_name ().length () + 1;
      memcpy (&out[pos], (*iter)->get_desc ().c_str (),  (*iter)->get_desc ().length () + 1);
      pos += (*iter)->get_desc ().length () + 1;
      memcpy (&out[pos], (*iter)->get_parent ().c_str (),  (*iter)->get_parent ().length () + 1);
      pos += (*iter)->get_parent ().length () + 1;
      lval = (*iter)->get_nvals ();
      memcpy (&out[pos], (void *)(&lval), sizeof (lval));
      pos += sizeof (lval);
      trval = (*iter)->get_datatype ();
      memcpy (&out[pos], (void *)(&trval), sizeof (trval));
      pos += sizeof (trval);

      if ((*iter)->get_datatype () == DT_CHAR)
      {
         char **text_val = static_cast <char **>((*iter)->get_value());
         for (long ii=0; ii<(*iter)->get_nvals (); ii++)
         {
            size = sizeof_dmDataType (DT_CHAR, text_val[ii]);
            memcpy (&out[pos], text_val[ii], size);
            pos += (size +1); // add one for null terminator 
         }
      }
      else
      {
         size = sizeof_dmDataType ((*iter)->get_datatype (), NULL) * (*iter)->get_nvals ();
         memcpy (&out[pos], (*iter)->get_value(), size);
         pos += size;
      }

   }

   return out;
}

void ChipsTransformFactory::serialize_ascii (Transform *trans, fstream &out)
{
   list<TransParam*> trans_params;
   list<TransParam*>::iterator iter;
   int datatype = 0;
   int nvals = 0;

   out.precision (15);

   if (!trans)
      return;

   if (out.is_open () == false)
      return;

   out << "type: ";
   // mark which type of transform this is
   if ((dynamic_cast <WCSTransform *>(trans) != 0) || (dynamic_cast <WCSTANTransform *>(trans) != 0))
      out << "WCS";
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
      out << "LINEAR2D";
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
      out << "SPLIT";
   else
      out << "base";

   out << endl;

   // copy in the name
   out << "name: " << trans->get_name () << endl;

   out << "\nparameter list begin:\n";
   // go through parameter list and copy out all parameter values
   trans_params = trans->get_parameter_list ();
   for (iter = trans_params.begin (); iter != trans_params.end (); iter++)
   {
      out << "name: " << (*iter)->get_name ().c_str () << endl;
      out << "desc: " << (*iter)->get_desc ().c_str () << endl;
      out << "parent: " << (*iter)->get_parent ().c_str () << endl;
      out << "nvals: " << (*iter)->get_nvals () << endl;
      out << "datatype: " << (*iter)->get_datatype () << endl;
      out << "value: ";

      datatype = static_cast <int>((*iter)->get_datatype ());
      nvals = (*iter)->get_nvals ();

      if ((*iter)->get_value())
      {
         string spacer = "";
         switch (datatype)
         {
            case DT_UINT2:       // fallthrough intentional
            case DT_INT2:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <short *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_UINT4:        // fallthrough intentional
            case DT_INT4:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <long *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_REAL4:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <float *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_REAL8:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <double *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_CHAR:
            {
//                Question how to handle spacers
            }
            break;
            case DT_BIT:          // fallthrough intentional
            case DT_UINT1:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <char *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_LOGICAL:
               for (int ii=0; ii<nvals; ii++)
               {
                  out << spacer << (static_cast <bool *>((*iter)->get_value()))[ii];
                  if (spacer.length () == 0)
                     spacer = ",";
               }
            break;
            case DT_UNK:  // fallthrough intentional
            default:
            break;
         }
      }
      else 
         out << "NULL\n";

      out << endl << endl;
   }
}


void ChipsTransformFactory::serialize (Transform *trans, ChipsStateStream &out)
{
   list<TransParam*> trans_params;
   list<TransParam*>::iterator iter;
   int ival = 0;
   long num_vals = 0;
   string sval;

   if (!trans)
      return;

   // mark which type of transform this is
   if ((dynamic_cast <WCSTransform *>(trans) != 0) || (dynamic_cast <WCSTANTransform *>(trans) != 0))
      sval = "WCS";
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
      sval = "LINEAR2D";
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
      sval = "SPLIT";
   else
      sval = "base";

   out.write_string (&sval);


   // copy in the name
   sval = trans->get_name ();
   out.write_string (&sval);

   // go through parameter list and copy out all parameter values
   trans_params = trans->get_parameter_list ();
   ival = trans_params.size ();
   out.write_int (&ival);
   for (iter = trans_params.begin (); iter != trans_params.end (); iter++)
   {
      sval = (*iter)->get_name ();
      out.write_string (&sval);

      sval = (*iter)->get_desc ();
      out.write_string (&sval);

      sval = (*iter)->get_parent ();
      out.write_string (&sval);

      num_vals = static_cast <int>((*iter)->get_nvals ());
      out.write_long (&num_vals);

      ival = static_cast <int>((*iter)->get_datatype ());
      out.write_int (&ival);

      switch (ival)
      { 
         case DT_UINT2:       // fallthrough intentional
         case DT_INT2:
            out.write_short (static_cast <short *>((*iter)->get_value()), num_vals);
         break;
         case DT_UINT4:        // fallthrough intentional
         case DT_INT4:
            out.write_long (static_cast <long *>((*iter)->get_value()), num_vals);
         break;
         case DT_REAL4:
            out.write_float (static_cast <float *>((*iter)->get_value()), num_vals);
         break;
         case DT_REAL8:
            out.write_double (static_cast <double *>((*iter)->get_value()), num_vals);
         break;
         case DT_CHAR:
         {
            char **text_val = static_cast <char **>((*iter)->get_value());
            for (long ii=0; ii<(*iter)->get_nvals (); ii++)
            {
               sval = text_val[ii];
               out.write_string (&sval);
            }
         }
         break;
         case DT_BIT:          // fallthrough intentional
         case DT_UINT1:
            out.write_char (static_cast <char *>((*iter)->get_value()), num_vals);
         break;
         case DT_LOGICAL:
            out.write_bool (static_cast <bool *>((*iter)->get_value()), num_vals);
         break;
         case DT_UNK:  // fallthrough intentional
         default:
         break;
      }
   }
}

                                                       


Transform *ChipsTransformFactory::deserialize (unsigned char *in, int length)
{
   Transform *trans = static_cast <Transform *>(0);
   TransParam *param;
   string trans_type;
   string trans_name;
   int num_params;
   
   dmDataType param_datatype = DT_UNK;
   string param_name;
   string param_desc;
   string param_parent;
   long param_nvals;
   int size=0;

   int pos = 0;

   if (!in || (length == 0))
      return trans;

   // find out which type of transform this is
   trans_type = (char *)(&in[pos]);
   pos += trans_type.length () + 1;

   if ((trans_type == "WCSTAN") || (trans_type == "WCS"))
      trans = new WCSTransform;
   else if (trans_type == "LINEAR2D")
      trans = new LINEAR2DTransform;
   else if (trans_type == "SPLIT")
      trans = new SPLITTransform;
   else
      trans = new Transform;

   // copy in the name
   trans_name = (char *)(&in[pos]);
   trans->set_name (trans_name);
   pos += trans_name.length () + 1;

   // go through parameter list and copy out all parameter values
   memcpy ((unsigned char *)(&num_params), &in[pos], sizeof (num_params));
   pos += sizeof (num_params);
   for (int ii=0; ii<num_params; ii++)
   {
      // Gather basic information
      param_name =reinterpret_cast <char *>(&in[pos]);
      pos += param_name.length () + 1;

      param_desc = reinterpret_cast <char *>(&in[pos]);
      pos += param_desc.length () + 1;

      param_parent = reinterpret_cast <char *>(&in[pos]);
      pos += param_parent.length () + 1;

      memcpy ((unsigned char *)(&param_nvals), &in[pos], sizeof (param_nvals));
      pos += sizeof (param_nvals);

      memcpy ((unsigned char *)(&param_datatype), &in[pos], sizeof (param_datatype));
      pos += sizeof (param_datatype);

      if (param_datatype != DT_CHAR)
         size = sizeof_dmDataType (param_datatype, NULL) * param_nvals;


      // Create the parameter without values
      param = new TransParam (param_name, static_cast <void *>(0), 
		      param_datatype, param_desc, param_parent, param_nvals);
         

      // Copy the parameter values from the byte stream
      switch (param_datatype)
      {
         case DT_UINT2:
         {
            unsigned short *param_value = new unsigned short [param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_INT2:
         {
            short *param_value = new short [param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_UINT4:
         {
            unsigned long *param_value = new unsigned long[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_INT4:
         {
            long *param_value = new long[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_REAL4:
         {
            float *param_value = new float[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_REAL8:
         {
            double *param_value = new double[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_CHAR:
         {

            string *param_value = new string[param_nvals];
            for (long ii=0; ii<param_nvals; ii++)
            { 
               char *tmp_str = reinterpret_cast <char *>(&in[pos]);
               param_value[ii] = tmp_str;
               pos += (param_value[ii].length ()+1);
            }
            param->set_value (param_value, param_nvals);
         }
         break;
         case DT_BIT:          // fallthrough intentional
         case DT_UINT1:
         {
	    unsigned char *param_value = new unsigned char[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case DT_LOGICAL:
         {
            bool *param_value = new bool[param_nvals];
            memcpy ((void *)param_value, &in[pos], size);
            param->set_value (param_value, param_nvals);
            delete []param_value;
            pos += size;
         }
         break;
         case (DT_UNK):  // fallthrough intentional
         default:
         break;
      }
      trans->set_parameter (param);
      delete param;

   }

   return trans;

}



Transform *ChipsTransformFactory::deserialize (ChipsStateStream &in)
{
   Transform *trans = static_cast <Transform *>(0);
   TransParam *param;
   string trans_type;
   string trans_name;
   string trans_flavor;
   int num_params;

   dmDataType param_datatype = DT_UNK;
   string param_name;
   string param_desc;
   string param_parent;
   long param_nvals;
   int ival;

   // find out which type of transform this is
   in.read_string (&trans_type);

   if ((trans_type == "WCSTAN") || (trans_type == "WCS"))
      trans = new WCSTransform;
   else if (trans_type == "LINEAR2D")
      trans = new LINEAR2DTransform;
   else if (trans_type == "SPLIT")
      trans = new SPLITTransform;
   else
      trans = new Transform;

   // copy in the name
   in.read_string (&trans_name);
   trans->set_name (trans_name);

   if (in.get_minor_version () < 12)
   {
       // copy in the flavor
       in.read_string (&trans_flavor);
   }

   // go through parameter list and copy out all parameter values
   in.read_int (&num_params);
   for (int ii=0; ii<num_params; ii++)
   {
      // Gather basic information about the parameter
      in.read_string (&param_name);
      in.read_string (&param_desc);
      in.read_string (&param_parent);
      in.read_long (&param_nvals);

      in.read_int (&ival);
      param_datatype = static_cast <dmDataType> (ival);

      // Create the parameter without values
      param = new TransParam (param_name, static_cast <void *>(0),
                      param_datatype, param_desc, param_parent, param_nvals);

      // Copy the parameter values from the state stream
      switch (param_datatype)
      {
         case DT_UINT2:
         {
            unsigned short *param_value = new unsigned short[param_nvals];
            in.read_short (reinterpret_cast <short *>(param_value), param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         }
         break;
         case DT_INT2:
         {
            short *param_value = new short[param_nvals];
            in.read_short (param_value, param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         } 
         break;
         case DT_UINT4:
         {
            unsigned long *param_value = new unsigned long[param_nvals];
            in.read_long (reinterpret_cast <long *>(param_value), param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         }
         break;
         case DT_INT4:
         {
            long *param_value = new long[param_nvals];
            in.read_long (param_value, param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         }
         break;
         case DT_REAL4:
         {
            float *param_value = new float[param_nvals];
            in.read_float (param_value, param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         }
         break;
         case DT_REAL8:
         {
            double *param_value = new double[param_nvals];
            in.read_double (param_value, param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;
         }
         break;
         case DT_CHAR:
         {
            string *param_value = new string[param_nvals];
            for (long ii=0; ii<param_nvals; ii++)
               in.read_string (&param_value[ii]);
            param->set_value (param_value, param_nvals);
         }
         break;
         case DT_BIT:          // fallthrough intentional
         case DT_UINT1:
         {
            char *param_value = new char[param_nvals];
            in.read_char (param_value, param_nvals);
            param->set_value (reinterpret_cast <unsigned char *>(param_value), param_nvals);
            delete []param_value;
         }
         break;
         case DT_LOGICAL:
         {
            bool *param_value = new bool[param_nvals];
            in.read_bool (param_value, param_nvals);
            param->set_value (param_value, param_nvals);
            delete []param_value;

         }
         break;
         case DT_UNK:  // fallthrough intentional
         default:
         break;
      }
      trans->set_parameter (param);
      delete param;

   }

   return trans;
}

Transform *ChipsTransformFactory::copy_transform (Transform *trans)
{

   Transform *out_trans = static_cast <Transform *> (0);


   if (!trans)
      return out_trans;

   out_trans = trans->copy ();

   return out_trans;
}


int ChipsTransformFactory::get_transform_size (Transform *trans)
{
   list<TransParam*> trans_params;
   list<TransParam*>::iterator iter;
   int size = 0;

   if (!trans)
      return 0;

   size += trans->get_name ().length () * sizeof (char) + 1;

   trans_params = trans->get_parameter_list ();
   for (iter = trans_params.begin (); iter != trans_params.end (); iter++)
   {
      // include +1 on the strings for null terminators
      size += (*iter)->get_name ().length () * sizeof (char) + 1;
      size += (*iter)->get_desc ().length () * sizeof (char) + 1;
      size += (*iter)->get_parent ().length () * sizeof (char) + 1;
      size += sizeof ((*iter)->get_nvals ());
      size += sizeof ((*iter)->get_datatype ());

      // if data type is a string then go through each string
      // and calculate the size
      if ((*iter)->get_datatype () == DT_CHAR)
      {
         char **text_val = static_cast <char **> ((*iter)->get_value());
         for (long ii=0; ii<(*iter)->get_nvals (); ii++)
            size += sizeof_dmDataType (DT_CHAR, text_val[ii]) + 1;
      }
      else
         size += (sizeof_dmDataType ((*iter)->get_datatype (), NULL) * (*iter)->get_nvals ());
      
   }

   return size; 
}

size_t ChipsTransformFactory::sizeof_dmDataType (dmDataType dt, void *value)
{
   size_t size = 0;
   switch (dt)
   {
      case (DT_UINT2):       // fallthrough intentional
      case (DT_INT2):
         size = sizeof (short);
      break;
      case (DT_UINT4):        // fallthrough intentional
      case (DT_INT4):
         size = sizeof (long);
      break;
      case (DT_REAL4):
         size = sizeof (float);
      break;
      case (DT_REAL8):
         size = sizeof (double);
      break;
      case (DT_CHAR):
         // this assumes that value is null terminated. This
         // assumption may need to be changed
         if (value != NULL)
            size = strlen (static_cast <char *>(value));
         else
            size = 0;
      break;
      case (DT_BIT):          // fallthrough intentional
      case (DT_UINT1):
         size = sizeof (char);
      break;
      case (DT_LOGICAL):
         size = sizeof (bool);
      break;
      case (DT_UNK):  // fallthrough intentional
      default:
         size = 0;
      break;
   }
  
   return size; 
}


ChipsTransformFactory::TransformType ChipsTransformFactory::get_transform_type (Transform *trans)
{
   ChipsTransformFactory::TransformType type;
   if (dynamic_cast <WCSTransform *>(trans) != 0)
      type = ChipsTransformFactory::WCSTAN;
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
      type = ChipsTransformFactory::LINEAR2D;
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
      type = ChipsTransformFactory::SPLIT;
   else
      type = ChipsTransformFactory::UNKNOWN;

   return type;
}

Transform * ChipsTransformFactory::new_transform (TransformType type)
{
   Transform *transform = NULL;
   switch (type)
   {
      case ChipsTransformFactory::WCSTAN:
         transform = new WCSTransform;
      break;
      case ChipsTransformFactory::SPLIT:
         transform = new SPLITTransform;
      break;
      case ChipsTransformFactory::LINEAR2D:
      {
         double **matrix;
         transform = new LINEAR2DTransform;

         // set to the indentity matrix
         matrix = new double*[3];
         for (int ii=3; ii--;)
         {
            matrix[ii] = new double[3];
            for (int jj=3; jj--;)
               matrix[ii][jj] = 0;
            matrix[ii][ii] = 1;
         }
         transform->set_transform_matrix (&matrix[0]);
         for (int ii=3; ii--;)
            delete []matrix[ii];
         delete []matrix;

      }
      case ChipsTransformFactory::UNKNOWN:
      break;
   }

   return transform;
}     

bool ChipsTransformFactory::compare (Transform *trans1, Transform *trans2)
{
   list<TransParam*> trans_params1;
   list<TransParam*>::iterator iter1;
   list<TransParam*> trans_params2;
   list<TransParam*>::iterator iter2;
   int datatype;
   int nvals;

   if (get_transform_type (trans1) != get_transform_type (trans2))
      return false;

   // dont compare the name 

   // go through parameter list and compare all parameter values
   trans_params1 = trans1->get_parameter_list ();
   trans_params2 = trans2->get_parameter_list ();

   if (trans_params1.size () != trans_params2.size ())
      return false;


   /**** this assumes the parameter lists are sorted in the same order. 
    **** this could be problematic but for now assume that it's not
    ****/

   iter1 = trans_params1.begin ();
   iter2 = trans_params2.begin ();
   for (; iter1 != trans_params1.end (); iter1++, iter2++)
   {
      if ((*iter1)->get_name () != (*iter2)->get_name ())
         return false;

      // ignore the parameter description

      datatype = (*iter1)->get_datatype ();
      if ((*iter1)->get_datatype () != (*iter2)->get_datatype ())
         return false;

      nvals = (*iter1)->get_nvals ();
      if ((*iter1)->get_nvals () != (*iter2)->get_nvals ())
         return false;


      switch (datatype)
      {
         case DT_LOGICAL:
         {
            bool *value1 = (bool *)((*iter1)->get_value ());
            bool *value2 = (bool *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_INT2:
         {
            short *value1 = (short *)((*iter1)->get_value ());
            short *value2 = (short *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_INT4:
         {
            long *value1 = (long *)((*iter1)->get_value ());
            long *value2 = (long *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_REAL4:
         {
            float *value1 = (float *)((*iter1)->get_value ());
            float *value2 = (float *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_REAL8:
         {
            double *value1 = (double *)((*iter1)->get_value ());
            double *value2 = (double *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_UINT2:
         {
            unsigned short *value1 = (unsigned short *)((*iter1)->get_value ());
            unsigned short *value2 = (unsigned short *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_UINT4:
         {
            unsigned long *value1 = (unsigned long *)((*iter1)->get_value ());
            unsigned long *value2 = (unsigned long *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_UINT1:
         {
            char *value1 = (char *)((*iter1)->get_value ());
            char *value2 = (char *)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (value1[ii] != value2[ii])
                  return false;
         }
         break;
         case DT_CHAR:
         {
            char **value1 = (char **)((*iter1)->get_value ());
            char **value2 = (char **)((*iter2)->get_value ());
            for (int ii=nvals; ii--;)
               if (strcmp (value1[ii], value2[ii]) != 0)
                  return false;
         }
         default:
         break;
      }
   }

   return true;
}

void *ChipsTransformFactory::get_parameter_copy (Transform *trans, const string &parameter)
{
   void *parameter_value = NULL;
   string local_parameter = "";
   TransParam *param = NULL;
   if (dynamic_cast <WCSTransform *>(trans) != 0)
   {
      if (parameter == "rotation")
      {
         // calculate the rotation here as it needs to be converted 
         // to radians 

         parameter_value = (double *)(trans->get_parameter_value ("CROTA"));
         if (parameter_value)
         {
            double *angle = new double[1];
            *angle = *(double *)parameter_value;
            *angle = *angle * V_PI/180.;
            parameter_value = (void *)angle;
         }

         // unset the local parameter so that parameter is not retrieved below
         local_parameter = "";

      }
      else if (parameter == "scale")
         local_parameter = "CDELT";
      else
         local_parameter = parameter;
   }
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
   {
      if (parameter == "rotation")
         local_parameter = "ROTATION";
      else if (parameter == "scale")
         local_parameter = "SCALE";
      else
         local_parameter = parameter;
   }
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
   {
      local_parameter = parameter;
   }

   param = trans->get_parameter (local_parameter.c_str ());
   if (param)
   {
      void *tmp_param_value = param->get_value ();
      switch (param->get_datatype ())
            {
         case DT_UINT2:
         {
            parameter_value = (void *)(new unsigned short[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((unsigned short *)parameter_value)[ii] = ((unsigned short *)tmp_param_value)[ii];
         }   
         break;
         case DT_INT2:
         {
            parameter_value = (void *)(new short[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((short *)parameter_value)[ii] = ((short *)tmp_param_value)[ii];

         }
         break;
         case DT_UINT4:
         {
            parameter_value = (void *)(new unsigned long[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((unsigned long *)parameter_value)[ii] = ((unsigned long *)tmp_param_value)[ii];

         }
         break;
         case DT_INT4:
         {
            parameter_value = (void *)(new long[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((long *)parameter_value)[ii] = ((long *)tmp_param_value)[ii];

         }
         break;
         case DT_REAL4:
         {
            parameter_value = (void *)(new float[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((float *)parameter_value)[ii] = ((float *)tmp_param_value)[ii];

         }
         break;
         case DT_REAL8:
         {
            parameter_value = (void *)(new double[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((double *)parameter_value)[ii] = ((double *)tmp_param_value)[ii];

         }
         break;
         case DT_CHAR:
         {
            parameter_value = (void *)(new string[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((string *)parameter_value)[ii] = ((char **)tmp_param_value)[ii];

         }
         break;
         case DT_BIT:          // fallthrough intentional
         case DT_UINT1:
         {
            parameter_value = (void *)(new char[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((char *)parameter_value)[ii] = ((char *)tmp_param_value)[ii];

         }
         break;
         case DT_LOGICAL:
         {
            parameter_value = (void *)(new bool[param->get_nvals ()]);
            for (int ii=param->get_nvals (); ii--;)
               ((bool *)parameter_value)[ii] = ((bool *)tmp_param_value)[ii];
         }
         break;
         case DT_UNK:  // fallthrough intentional
         default:
         break;
      }

   }
   

   return parameter_value;
}

