// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSTRANSFORMFACTORY_HH
#define CHIPSTRANSFORMFACTORY_HH

#include "Transform.hh"
#include "TransParam.hh"
#include "ChipsDefinesP.hh"

#include <iostream>
#include <string>
#include <fstream>

using namespace std;

class ChipsStateStream;

class ChipsTransformFactory
{
   public:
      enum TransformType {UNKNOWN, LINEAR2D, WCSTAN, SPLIT};

      ChipsTransformFactory ();
      virtual ~ChipsTransformFactory ();

      static Transform *checkValidTransform(Transform *trans);
      static void serialize (Transform *trans, ChipsStateStream &out);
      static unsigned char *serialize (Transform *trans, int &length);
      static void serialize_ascii (Transform *trans, fstream &out);
      static Transform *deserialize (unsigned char *in, int length);
      static Transform *deserialize (ChipsStateStream &in);

      static TransformType get_transform_type (Transform *trans);

      static Transform *new_transform (TransformType type);
      static bool compare (Transform *trans1, Transform *trans2);
      static Transform *copy_transform (Transform *trans);

      static void *get_parameter_copy (Transform *trans, const string &parameter);

      template <class T>
      static void set_parameter (Transform *trans, const string &parameter, T *value, int nvals);

   protected:
      static int get_transform_size (Transform *trans);
      static size_t sizeof_dmDataType (dmDataType dt, void *value);
};

template <class T>
void ChipsTransformFactory::set_parameter (Transform *trans, const string &parameter, T *value, int nvals)
{
   string local_parameter = "";
   TransParam *param;

   if (dynamic_cast <WCSTransform *>(trans) != 0)
   {
      if (parameter == "rotation")
      {
         // set the rotation here as it needs to be converted
         // to degrees 

         T angle;
         if (value)
            angle = *value * V_PI/180.;
         param = trans->get_parameter ("CROTA");
         param->set_value (&angle, 1);


         // unset the local parameter so that parameter is not retrieved below
         local_parameter = "";

         
      }
      else if (parameter == "scale")
         local_parameter = "CDELT";
      else
         local_parameter = parameter;
   }
   else if (dynamic_cast <LINEAR2DTransform *>(trans) != 0)
   {
      if (parameter == "rotation")
         local_parameter = "ROTATION";
      else if (parameter == "scale")
         local_parameter = "SCALE";
      else
         local_parameter = parameter;
   }
   else if (dynamic_cast <SPLITTransform *>(trans) != 0)
   {
      local_parameter = parameter;
   }

   if (local_parameter.size () > 0)
   {
      param = trans->get_parameter (local_parameter.c_str ());
      param->set_value (value, nvals);
   }

}

#endif
