// 
//  Copyright (C) 2010  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsWCSTANAxisAlgorithm.hh"
#include "ChipsDataSystem.hh"
#include "Transform.hh"
#include "ChipsAxis.hh"
#include "ChipsDefinesP.hh"
#include "ChipsMatrix.hh"
#include "ChipsCmdManager.hh"

#include <cmath>

extern ChipsCmdManager* cmd_manager;

using namespace std;

ChipsWCSTANAxisAlgorithm::ChipsWCSTANAxisAlgorithm (int orientation, ChipsAxis *axis) : ChipsAxisAlgorithm (orientation, axis)
{
   this->padded_min = -1;
   this->padded_max = 1;
   this->absolute_min = 0;
   this->absolute_max = 1;

   this->transformed_padded_min = -1;
   this->transformed_padded_max = 1;
   this->transformed_absolute_min = -1;
   this->transformed_absolute_max = 1;

   this->last_pad = 0;
   this->last_automin = false;
   this->last_automax = false;

   this->grid_slices = 50;
   this->tick_spacing = 0;
   this->minortick_spacing = 0;

   this->transformed_tickpos = NULL;

   this->scalable = false;

   this->max_set = 0;
   this->min_set = 0;

   this->last_axis_position = 0;
   this->grid_modified = true;
   this->sexagesimal = false;

   this->grid_extent[0] = 0;
   this->grid_extent[1] = 0;

   this->major_mode = chips_limits;

}

ChipsWCSTANAxisAlgorithm::ChipsWCSTANAxisAlgorithm (ChipsWCSTANAxisAlgorithm &copy) : ChipsAxisAlgorithm (copy)
{
   this->max_set = copy.max_set;
   this->min_set = copy.max_set;
   this->padded_min = copy.padded_min;
   this->padded_max = copy.padded_max;
   this->absolute_min = copy.absolute_min;
   this->absolute_max = copy.absolute_max;

   this->transformed_padded_min = copy.transformed_padded_min;
   this->transformed_padded_max = copy.transformed_padded_max;
   this->transformed_absolute_min = copy.transformed_absolute_min;
   this->transformed_absolute_max = copy.transformed_absolute_max;

   this->last_pad = copy.last_pad;
   this->last_automin = copy.last_automin;
   this->last_automax = copy.last_automax;
   this->last_axis_position = copy.last_axis_position;

   this->grid_slices = copy.grid_slices;
   this->tick_spacing = copy.tick_spacing;
   this->minortick_spacing = copy.minortick_spacing;
   this->grid_modified = true;

   this->sexagesimal = copy.sexagesimal;

   this->transformed_tickpos = NULL;
   if (this->tickcnt > 0)
   {
      this->transformed_tickpos = new double[this->tickcnt];
      for (int ii=this->tickcnt; ii--;)
         this->transformed_tickpos[ii] = copy.transformed_tickpos[ii];
   }

   this->grid_extent[0] = copy.grid_extent[0];
   this->grid_extent[1] = copy.grid_extent[1];

   // *** interval, count and arbitrary modes aren't supported in wcs. 
   // *** change to limits so that the plot can still be created. this
   // *** should be changed as these modes are supported
   this->set_major_mode (this->major_mode);
   this->set_minor_mode (this->minor_mode);
}

ChipsWCSTANAxisAlgorithm::~ChipsWCSTANAxisAlgorithm ()
{
   if (this->transformed_tickpos)
      delete []this->transformed_tickpos;
}

ChipsWCSTANAxisAlgorithm * ChipsWCSTANAxisAlgorithm::clone ()
{
   return new ChipsWCSTANAxisAlgorithm (*this);
}

void ChipsWCSTANAxisAlgorithm::set_scalable(bool val)
{
}

void ChipsWCSTANAxisAlgorithm::set_scaletype (int val)
{
}

void ChipsWCSTANAxisAlgorithm::set_tick_positions (double *ticks, int count)
{
/*
   this->modified = true;

   if (this->tickpos)
      delete []this->tickpos;

   this->tickpos = NULL;
   this->tickcnt = 0;
   this->arbitrary_set = false;

   if (!ticks)
      return;

   this->tickpos = new double[count];
   this->tickcnt = count;
   for (int ii=count; ii--;)
      this->tickpos[ii] = ticks[ii];

   this->arbitrary_set = false;
*/
}

void ChipsWCSTANAxisAlgorithm::set_arbitrary_positions (double *positions, string *text, int count)
{
   this->modified = true;

   if (this->arbpos)
   {
      delete []this->arbpos;
      this->arbpos = NULL;
   }

   if (this->arbtext)
   {
      delete []this->arbtext;
      this->arbtext = NULL;
   }

   this->arbcnt = 0;

   if (!positions)
      return;

   if (text)
   {
      this->arbtext = new string[count];
      for (int ii = count; ii--;)
         this->arbtext[ii] = text[ii];
   }
   this->arbcnt = count;
   this->arbpos = new double[count];
   for (int ii = count; ii--;)
      this->arbpos[ii] = positions[ii];
}

int ChipsWCSTANAxisAlgorithm::set_absolute_min (double val, int scaletype, bool transformed)
{
   if (scaletype == V_SCALE_LOG)
      return ERR_STATUS_BAD;

   if (transformed)
   {
      if (this->transformed_absolute_min != val)
         this->modified = true;
      this->transformed_absolute_min = val;
      this->min_set = 1;
   }
   else
   {
      if (this->absolute_min != val)
         this->modified = true;
      this->absolute_min = val;
      this->min_set = 2;
   }

   return ERR_STATUS_GOOD;
}


double ChipsWCSTANAxisAlgorithm::get_min (bool transformed)
{
   if (transformed)
      return this->transformed_padded_min;
   return this->padded_min;
}

double ChipsWCSTANAxisAlgorithm::get_absolute_min (int scaletype, bool transformed)
{
   if (transformed)
      return this->transformed_absolute_min;
   return this->absolute_min;
}

int ChipsWCSTANAxisAlgorithm::set_absolute_max (double val, int scaletype, bool transformed)
{
   if (scaletype == V_SCALE_LOG)
      return ERR_STATUS_GOOD;

   if (transformed)
   {
      if (this->transformed_absolute_max != val)
         this->modified = true;
      this->transformed_absolute_max = val;
      this->max_set = 1;
   }
   else
   {
      if (this->absolute_max != val)
         this->modified = true;
      this->absolute_max = val;
      this->max_set = 2;
   }


   return ERR_STATUS_GOOD;
}

double ChipsWCSTANAxisAlgorithm::get_max (bool transformed)
{
   if (transformed)
      return this->transformed_padded_max;
   return this->padded_max;
}

double ChipsWCSTANAxisAlgorithm::get_absolute_max (int scaletype, bool transformed)
{
   if (transformed)
      return this->transformed_absolute_max;
   return this->absolute_max;
}


int ChipsWCSTANAxisAlgorithm::set_absolute_minmax (double min, double max, int scaletype, bool transformed)
{
   this->set_absolute_min (min, scaletype, transformed);
   this->set_absolute_max (max, scaletype, transformed);

   return ERR_STATUS_GOOD;
}


int ChipsWCSTANAxisAlgorithm::update_tick_positions(bool &out_updated)

{
   int rval = ERR_STATUS_GOOD;

   out_updated = false;

   if (!this->modified)
      return rval;

   this->reset_adjusted_parameters ();

   switch (this->major_mode)
   {
      case chips_nice:
      case chips_limits:
         rval = this->limit_tick_positions ();
      break;
      case chips_interval:
         rval = this->interval_tick_positions ();
      break;
      case chips_count:
         rval = this->count_tick_positions ();
      break;
      case chips_arbitrary:
         rval = this->arbitrary_tick_positions ();
      break;
      default:
         rval = ERR_STATUS_BAD;
      break;
   }

   // the algorithms were rerun so assume they were updated
   out_updated = true;

   // all parameters were used to calculated; reset modified flag
   this->modified = false;
   this->grid_modified = true;
   return rval;
}

int ChipsWCSTANAxisAlgorithm::calculate_tick_limits(bool automin, bool automax, double pad)

{
   int rval = ERR_STATUS_GOOD;

   if (!this->axis)
      return ERR_STATUS_BAD;

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   if (dsCoord == NULL)
      return ERR_STATUS_BAD;

   if (!this->modified && (this->last_pad == pad) &&
        (this->last_automax == automax) &&
        (this->last_automin == automin) &&
	(this->last_axis_position == this->axis->get_norm_pos ()) &&
        (!dsCoord->get_modified ()))
      return rval;

   this->last_pad = pad;
   this->last_automax = automax;
   this->last_automin = automin;
   this->last_axis_position = this->axis->get_norm_pos ();

   this->reset_adjusted_parameters ();

   switch (this->major_mode)
   {
      case chips_nice:
         rval = this->nice_tick_limits ();
      break;
      case chips_limits:
         rval = this->limit_tick_limits (automin, automax, pad);
      break;
      case chips_interval:
         rval = this->interval_tick_limits (automin, automax, pad);
      break;
      case chips_count:
         rval = this->count_tick_limits (automin, automax, pad);
      break;
      case chips_arbitrary:
         rval = this->arbitrary_tick_limits (automin, automax, pad);
      break;
      default:
         rval = ERR_STATUS_BAD;
      break;
   }

   if (!rval)
      this->modified = true;

   return rval;
}


int ChipsWCSTANAxisAlgorithm::count_tick_limits(bool automin, bool automax, double pad)
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::count_tick_positions()
{
/*
   double xpmin, xpmax, dist;

   if (this->pollCrossed)
   {
      if (this->min < 0)
         this->min += (-90 - this->min)
      else
         this->max += (90 - this->max)
   }

   scale2(this->min, this->max, major_num-1, &xpmin, &xpmax, &this->tick_spacing);
   this->min = xpmin;
   this->max = xpmin +  (major_num-1) * dist;

   // handle wrap around
   if ((this->max >= 360) && (this->orientation == RA))
      this->max -= 360
   else if (this->orientation == DEC)
   {
      if (this->max >= 90)
      {
         this->max -= (90-this->max);
         this->pollCrossed = true;
      }
      else if (this->min <= -90)
      {
         this->min += (90-this->min);
         this->pollCrossed = true;
      }
      else
         this->pollCrossed = false;

   }

   // clearup any old tick positions
   if (this->tickpos)
   {
      delete []this->tickpos;
      this->tickpos = NULL;
   }

   // if the count is at least 2 and less than a defined number of
   // ticks (nominally 100) then use count mode
   if ((major_num > 1) && (major_num <  V_SIZE_MAX_VAL))
   {
      this->tickcnt = major_num;
      this->tickpos = new double[this->tickcnt];
      if (this->tickpos)
      {
         unsigned int rr;
         if ((this->tick_spacing > DBL_EPSILON) && (xpmin < DBL_EPSILON) &&
             (xpmin > -1*DBL_EPSILON))
            this->tickpos[0] = 0;
         else
            this->tickpos[0] = xpmin;
         for (rr = 1; rr < this->tickcnt; rr++)
         {
            this->tickpos[rr] = this->tickpos[rr-1] + dist;
            if ((this->tick_spacing > DBL_EPSILON) && (this->tickpos[rr] < DBL_EPSILON) &&
                (this->tickpos[rr] > -1*DBL_EPSILON))
               this->tickpos[rr] = 0;
         }
      }
   }
   if (this->tickpos)
   {
      // use recomputed tick locations
//      reset_string_labels();
//      return change_range();
   }
   else
   {
      // too high a count or allocation failed- try limit mode
      return  limit_tick_positions(0);
   }
*/

   return ERR_STATUS_GOOD;
}


int ChipsWCSTANAxisAlgorithm::limit_tick_limits(bool automin, bool automax, double pad)
{
   double padmin, padmax;
   int rval = ERR_STATUS_GOOD;

   if ((this->min_set==0) || (this->max_set==0))
      return ERR_STATUS_BAD;

   ChipsVector min_coord;
   ChipsVector max_coord;
   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   if (dsCoord == NULL)
      return ERR_STATUS_BAD;

   string tmp_id = dsCoord->get_id_no_wcs ();

   // Assume the coordinate system is up to date with the absolute positions
   // Convert the transformed data to image space to calculate padding
   if ((this->min_set == 1) || ((this->min_set & 1) && dsCoord->get_modified ()))
   {
      min_coord = dsCoord->get_lower_bounds ("");
      min_coord[index] = this->transformed_absolute_min;
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, "", DATA, tmp_id);
      this->absolute_min = min_coord[index];
   }
   else if ((this->min_set == 2) || ((this->min_set & 2) && dsCoord->get_modified ()))
   {
      min_coord = dsCoord->get_lower_bounds ();
      ChipsVector tmp_coord(min_coord.get_size ());
      min_coord[index] = this->absolute_min;
      dsCoord->convert_cs_to_cs (&min_coord, &tmp_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_absolute_min = tmp_coord[index];
   }
   else
   {
      min_coord = dsCoord->get_lower_bounds ();
      min_coord[index] = this->absolute_min;
   }

   if ((this->max_set == 1) || ((this->max_set & 1) && dsCoord->get_modified ()))

   {
      max_coord = dsCoord->get_upper_bounds ("");

      ChipsVector tmp_coord (max_coord.get_size ());
      max_coord[index] = this->transformed_absolute_max;
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, "", DATA, tmp_id);
      this->absolute_max = max_coord[index];
   }
   if ((this->max_set == 2) || ((this->max_set & 2) && dsCoord->get_modified ()))
   {
      max_coord = dsCoord->get_upper_bounds ();

      ChipsVector tmp_coord (max_coord.get_size ());
      max_coord[index] = this->absolute_max;
      dsCoord->convert_cs_to_cs (&max_coord, &tmp_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_absolute_max = tmp_coord[index];
   }
   else
   {
      max_coord = dsCoord->get_upper_bounds ();
      max_coord[index] = this->absolute_max;
   }


   // adjust by pad factor
   double tmp_min = min_coord[index];
   double tmp_max = max_coord[index];
   double tmp = (tmp_max - tmp_min)*pad;
   if (automin)
      padmin = tmp_min - tmp;
   else
      padmin = tmp_min;
   if (automax)
      padmax = tmp_max + tmp;
   else
      padmax = tmp_max;


   this->padded_min = padmin;
   this->padded_max = padmax;

   min_coord[index] = padmin;
   max_coord[index] = padmax;

//   dsCoord->modify_limits (min_coord, max_coord, tmp_id);

   // convert back to world space
   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               DATA, tmp_id, DATA, "");
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               DATA, tmp_id, DATA, "");

   transform_valid = false;

/*   this->transformed_padded_min = min_coord[index];
   this->transformed_padded_max = max_coord[index];
*/
   this->max_set |= 4;
   this->min_set |= 4;

   return rval;
}


int ChipsWCSTANAxisAlgorithm::limit_tick_positions()
{
   double xpmin, xpmax, base;
   int rval = ERR_STATUS_GOOD;

   ChipsVector min_coord;
   ChipsVector max_coord;

   Transform  *transform;

   string tmp_id;

   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   int opp_index = (index == 0) ? 1 : 0;
   int poll_exists = 0;
   int spacing_sign = 1;

   bool reversed = this->axis->get_reverse ();
   double img_hit[2];
   double tmp_hit[2];

   double *tmp_tickpos;
   double *tmp_transformed_tickpos;

   int num_hits = 0;
   int tmp_count = 0;
   bool valid = false;

   double min_coord_trns = 0;
   double max_coord_trns = 0;

   double alt_limit = 0;

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   if (dsCoord == NULL)
      return ERR_STATUS_BAD;

   if (!this->min_set || !this->max_set)
      return ERR_STATUS_BAD;


   transform = dsCoord->get_transform (dsCoord->get_default_transform_id ());

   if (transform == NULL)
      return ERR_STATUS_BAD;

   this->get_axis_position (true, min_coord, max_coord);

   min_coord_trns = min_coord[index];
   max_coord_trns = max_coord[index];

   // convert the bounds to image coordinates
   tmp_id = dsCoord->get_id_no_wcs ();
   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               DATA, "", DATA, tmp_id);
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               DATA, "", DATA, tmp_id);


   poll_exists = check_poll ();

   // if the max coord is less than the min coord assume we crossed zero
   // for RA -- ***** NOTE: Need to account for mirror flips in the transform
   if (this->orientation == RA_AXIS)
   {
      // handle crossing zero
      if (fabs(min_coord_trns-max_coord_trns) > 180)
      {
         if  (reversed)
            max_coord_trns+=360;
         else
            min_coord_trns+=360;
      }

      // handle crossing a poll
      if (((min_coord_trns < max_coord_trns) && !reversed) ||
          ((min_coord_trns > max_coord_trns) && reversed))
         reversed = !reversed;

      if (this->sexagesimal)
      {
         if (!reversed)
            this->scale1_sexg_ra(max_coord_trns, min_coord_trns, 4, &xpmax, &xpmin, &
(this->tick_spacing), &base);
         else
            this->scale1_sexg_ra(min_coord_trns, max_coord_trns, 4, &xpmin, &xpmax, &
(this->tick_spacing), &base);
      }
      else
      {
         if (!reversed)
            this->scale1(max_coord_trns, min_coord_trns, 5, &xpmax, &xpmin, &(this->tick_spacing), &base);
         else
            this->scale1(min_coord_trns, max_coord_trns, 5, &xpmin, &xpmax, &(this->tick_spacing), &base);
      }

      spacing_sign = -1;
   }
   else if (this->orientation == DEC_AXIS)
   {
      // when the poll exists use the 90 as an alternative upper 
      // or lower limit
      alt_limit = 90*poll_exists;
      
      // check if the extents hit the axis twice if so then
      // the tick label cross a poll
      if (poll_exists == 0)
      {
         valid = false;
         for (int ii=0; ii<2; ii++)
         {
            if (ii == 0)
               num_hits = get_dec_intersection (min_coord_trns, min_coord[opp_index], DEC_AXIS, transform, &tmp_hit[0]);
            else
               num_hits = get_dec_intersection (max_coord_trns, min_coord[opp_index], DEC_AXIS, transform, &tmp_hit[0]);

            if (num_hits == 2)
            {
               img_hit[opp_index] = min_coord[opp_index];
               for (int jj=0; jj<2; jj++)
               {
                  valid = false;
                  if ((tmp_hit[jj] >= (min_coord[index]-FLT_EPSILON)) && (tmp_hit[jj] <= (max_coord[index]+FLT_EPSILON)))
                  {
                     img_hit[index] = tmp_hit[jj];
                     if (ii == 0)
                        valid = this->check_point (&img_hit[0], min_coord_trns, index, transform);
                     else
                        valid = this->check_point (&img_hit[0], max_coord_trns, index, transform);
                  }
                  if (!valid)
                     break;
               }
            }

            if (valid)
               break;
         }

         // if there are valid point on either side of the poll
         // use either the opposite side of the plot as a new
         // limit or use the poll -- this keeps the spacing between
         // ticks and grids more readable
         if (valid)
         {
            ChipsVector data_max = dsCoord->get_upper_bounds ("");

            if (((max_coord_trns > 0) && (max_coord_trns < data_max[index])) ||
                ((max_coord_trns < 0) && (max_coord_trns > data_max[index])))
               alt_limit = data_max[index];
            else if (max_coord_trns > 0)
               alt_limit = 90;
            else
               alt_limit = -90;

         }
      }

      // adjust the tick limit if necessary
      if (alt_limit != 0)
      {
         if (fabs(min_coord_trns - alt_limit) < fabs (max_coord_trns - alt_limit))
            min_coord_trns = alt_limit;
         else
         {
            max_coord_trns = min_coord_trns;
            min_coord_trns = alt_limit;
         }
      }
      
      // check if the plot went over the top of a poll such that grid
      // is essentially flipped
      if (((min_coord_trns > max_coord_trns) && (min_coord_trns > 0) && !reversed) ||
          ((max_coord_trns > min_coord_trns) && (max_coord_trns > 0) && reversed))
      {
         max_coord_trns = 180 - max_coord_trns;
         min_coord_trns = 180 - min_coord_trns;
      }
      else if (((min_coord_trns > max_coord_trns) && (min_coord_trns < 0) && !reversed) ||
               ((max_coord_trns > min_coord_trns) && (max_coord_trns < 0) && reversed))
      {
      
         min_coord_trns = -180 - min_coord_trns;
         max_coord_trns = -180 - max_coord_trns;
      }

      // figure out the tick scaling
      if (this->sexagesimal)
      {
         if (!reversed)
            this->scale1_sexg_dec(min_coord_trns, max_coord_trns, 4, &xpmin, &xpmax, &(this->tick_spacing), &base);
         else
            this->scale1_sexg_dec(max_coord_trns, min_coord_trns, 4, &xpmax, &xpmin, &(this->tick_spacing), &base);
      }
      else
      {
         if (!reversed)
            this->scale1(min_coord_trns, max_coord_trns, 5, &xpmin, &xpmax, &(this->tick_spacing), &base);
         else
            this->scale1(max_coord_trns, min_coord_trns, 5, &xpmax, &xpmin, &(this->tick_spacing), &base);
      }
   }
         
   if (reversed)
      spacing_sign *= -1;

   this->tick_spacing *= spacing_sign;

   // the spacing sign is used later for when the dec cross
   // a poll
   spacing_sign = 1;

   this->tickcnt = (unsigned int)(fabs(xpmax - xpmin) / fabs(this->tick_spacing) + 1.5);

   // clearup any old tick positions
   if (this->tickpos)
   {
      delete []this->tickpos;
      this->tickpos = NULL;
   }

   if (this->transformed_tickpos)
   {
      delete []this->transformed_tickpos;
      this->transformed_tickpos = NULL;
   }

   // make sure we have a reasonable number of ticks
   // if not create 2 tick marks

   if ((this->tickcnt <= 0) || (this->tickcnt > 500))
   {
      this->tickcnt = 2;
      this->tick_spacing = xpmax-xpmin;
   }

   // setup new tick locations
   this->tickpos = new double[this->tickcnt*2]; // dec can have twice as many ticks
   this->transformed_tickpos = new double[this->tickcnt*2];

   tmp_tickpos = new double[this->tickcnt];
   tmp_transformed_tickpos = new double[this->tickcnt];

   if (this->tickpos)
   {
      unsigned int rr; 
      double alpha_delta = xpmin;
      double tol = 10e-12;

      for (rr = 0; rr < this->tickcnt; rr++)
      { 
         if (!approx_equal (this->tick_spacing, 0, tol, tol) &&
             approx_equal (alpha_delta, 0, tol, tol))
            alpha_delta = 0;
//         if ((fabs (this->tick_spacing) > DBL_EPSILON) && (fabs(alpha_delta) < FLT_EPSILON))
//            alpha_delta = 0;

         if (this->orientation == RA_AXIS)
         {

            // handle crossing zero
            if (alpha_delta < 0)
               alpha_delta += 360;
            if (alpha_delta >= 360)
               alpha_delta -= 360;
            this->tickpos[rr] = get_ra_intersection (alpha_delta, min_coord[opp_index], RA_AXIS, transform);

            img_hit[index] = this->tickpos[rr];
            img_hit[opp_index] = min_coord[opp_index];

            // verify the value retrieved is the expected value
            // if it's not push the tickmark out of the plot
            if (!this->check_point (&img_hit[0], alpha_delta, index, transform))
               this->tickpos[rr] = max_coord[index] + 10;
         }
         else 
         { 
            bool tickpos_set = false;

            if (alpha_delta < -90)
            {
               alpha_delta = -180 - alpha_delta;
               spacing_sign *= -1;
            }
            else if (alpha_delta > 90)
            {
               alpha_delta = 180 - alpha_delta;
               spacing_sign *= -1;
            }
            num_hits = get_dec_intersection (alpha_delta, min_coord[opp_index], DEC_AXIS, transform, &this->tickpos[rr]);

            // if there was more than one hit copy the second hit into 
            // a temporary array to be added back at the end of tickpos
            img_hit[opp_index] = min_coord[opp_index];
            for (int ii=0; ii<num_hits; ii++)
            {

               img_hit[index] = this->tickpos[rr+ii];
               // verify the value retrieved is the expected value
               // plusses/minusses are lost in interesection eqation so
               // this may not always be the case
               if (this->check_point (&img_hit[0], alpha_delta, index, transform))
               {
                  
                  if (!tickpos_set)
                  {
                     this->tickpos[rr] = this->tickpos[rr+ii];
                     tickpos_set = true;
                  }
                  else
                  {
                     tmp_tickpos[tmp_count] = this->tickpos[rr+ii];
                     tmp_transformed_tickpos[tmp_count] = alpha_delta;
                     tmp_count++;
                  }
               }
            }

            // if there was no hit set the location to be some clipped point
            // its necessary for determining minor tick marks.
            if ((num_hits == 0) || !tickpos_set)
               this->tickpos[rr] = max_coord[index] + 10;
         }

         this->transformed_tickpos[rr] = alpha_delta;

         alpha_delta += (this->tick_spacing * spacing_sign);
      }

      if (tmp_count > 0)
      {
         for (int rr=0; rr < tmp_count; rr++)
         {
            this->tickpos[this->tickcnt+rr] = tmp_tickpos[(tmp_count-1)-rr];
            this->transformed_tickpos[this->tickcnt+rr] = tmp_transformed_tickpos[(tmp_count-1)-rr];
         }
      }

      // copy over the min and max tick positions found
      this->grid_extent[0] = this->transformed_tickpos[0];
      this->grid_extent[1] = this->transformed_tickpos[this->tickcnt-1];
   
      if (this->orientation == DEC_AXIS)
      {
         // count the number of useful ticks keep one tick just outside
         // the each boundary
         int count = this->tickcnt;
         this->tickcnt = 0;
         bool exit=false;

         double min_extent = min_coord[index];
         double max_extent = max_coord[index];

         if (reversed)
         {
            min_extent = max_coord[index];
            max_extent = min_coord[index];
         }
         for (int rr=0; rr<count+tmp_count; rr++)
         {
            if ((this->tickpos[rr] > min_extent) && (this->tickpos[rr] < max_extent))
            {
               this->tickpos[this->tickcnt] = this->tickpos[rr];
               this->transformed_tickpos[this->tickcnt] = this->transformed_tickpos[rr];
               this->tickcnt++;
               exit = true;
            }
            else if (exit)
            {
               // single hit ticks can cause tick positions to appear 
               // out of order -- ex: dec crosses zero
               if ((this->tickpos[rr] <= this->tickpos[0]) && (this->tickpos[rr] >= this->tickpos[1]))
               {
                  this->tickpos[0] = this->tickpos[rr];
                  this->transformed_tickpos[0] = this->transformed_tickpos[rr];
               }
               else
               {
                  
                  this->tickpos[this->tickcnt] = this->tickpos[rr];
                  this->transformed_tickpos[this->tickcnt] = this->transformed_tickpos[rr];
                  this->tickcnt++;
               
               }
            }
            else
            {
               this->tickpos[0] = this->tickpos[rr];
               this->transformed_tickpos[0] = this->transformed_tickpos[rr];
               this->tickcnt = 1;
            }
         }
      }

      delete []tmp_tickpos;
      delete []tmp_transformed_tickpos;
   }

   if (this->minor_mode == chips_nice)
   {
      if (base == 2.0)     this->minor_num_adjusted = 3;
         else              this->minor_num_adjusted = 4;

      this->count_minortick_positions ();
      this->minor_mode_adjusted = chips_arbitrary;
   }


   return rval;
}

int ChipsWCSTANAxisAlgorithm::nice_tick_limits()
{
   double xpmin, xpmax, base;
   int rval = ERR_STATUS_GOOD;

   if ((this->min_set==0) || (this->max_set==0))
      return ERR_STATUS_BAD;

   ChipsVector min_coord;
   ChipsVector max_coord;
   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   bool reversed = this->axis->get_reverse ();

   if (dsCoord == NULL)
      return ERR_STATUS_BAD;

   string tmp_id = dsCoord->get_id_no_wcs ();

   // Assume the coordinate system is up to date with the absolute positions
   // Convert the transformed data to image space to calculate padding
   if ((this->min_set == 1) || ((this->min_set & 1) && dsCoord->get_modified ()))

   {
      min_coord = dsCoord->get_lower_bounds ("");
      min_coord[index] = this->transformed_absolute_min;
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, "", DATA, tmp_id);
      this->absolute_min = min_coord[index];
   }
   else if ((this->min_set == 2) || ((this->min_set & 2) && dsCoord->get_modified ()))
   {
      min_coord = dsCoord->get_lower_bounds ();
      min_coord[index] = this->absolute_min;
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_absolute_min = min_coord[index];
   }

   if ((this->max_set == 1) || ((this->max_set & 1) && dsCoord->get_modified ()))
   {
      max_coord = dsCoord->get_upper_bounds ("");
      max_coord[index] = this->transformed_absolute_max;
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, "", DATA, tmp_id);
      this->absolute_max = max_coord[index];
   }
   else if ((this->max_set == 2) || ((this->max_set & 2) && dsCoord->get_modified ()))
   {
      max_coord = dsCoord->get_upper_bounds ();
      max_coord[index] = this->absolute_max;
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_absolute_max = max_coord[index];
   }


   if (this->orientation == RA_AXIS)
   {
      double tmp_min = this->transformed_absolute_min;
      double tmp_max = this->transformed_absolute_max;
      if ((tmp_min < tmp_max) && !reversed)
         tmp_min += 360;
      else if ((tmp_min > tmp_max) && reversed)
         tmp_max += 360;

      if (!reversed)
         this->scale1(tmp_max, tmp_min, 5, &xpmax, &xpmin, &(this->tick_spacing), &base);
      else
         this->scale1(tmp_min, tmp_max, 5, &xpmin, &xpmax, &(this->tick_spacing), &base);
      if (xpmin >= 360)
         xpmin -= 360;
   }
   else if (this->orientation == DEC_AXIS)
   {
      int poll_exists = check_poll ();
      double tmp_min = this->transformed_absolute_min;
      double tmp_max = this->transformed_absolute_max;

      if (poll_exists > 0)
         tmp_max = 180 - tmp_max;
      else if (poll_exists < 0)
         tmp_min = -180 - tmp_min;
      if (!reversed)
         this->scale1(tmp_min, tmp_max, 5, &xpmin, &xpmax, &(this->tick_spacing), &base);
      else
         this->scale1(tmp_max, tmp_min, 5, &xpmax, &xpmin, &(this->tick_spacing), &base);
   }

   this->transformed_padded_min = xpmin;
   this->transformed_padded_max = xpmax;

   
   min_coord = dsCoord->get_lower_bounds ("");
   min_coord[index] = xpmin;
   max_coord = dsCoord->get_upper_bounds ("");
   max_coord[index] = xpmax;


   // convert back to world space
   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               DATA, "", DATA, tmp_id);
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               DATA, "", DATA, tmp_id);

   this->max_set = 3;
   this->min_set = 3;

   transform_valid = true;

   return rval;
}

int ChipsWCSTANAxisAlgorithm::nice_tick_positions()
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::arbitrary_tick_limits(bool automin, bool automax,
                                             double pad)
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::arbitrary_tick_positions()
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::interval_tick_limits(bool automin, bool automax, double pad)
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::interval_tick_positions()
{
   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::count_minortick_positions ()
{
   double tmp_min;
   double tmp_max;
   double tmp_int; 
   double tmp_val;
   double tmp_img[2];
   int num_hits;
   int poll_exists;

   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   int opp_index = (index == 0) ? 1 : 0;

   bool reversed = this->axis->get_reverse ();

   if (this->tickcnt <= 0)
      return ERR_STATUS_BAD;

   if (this->minortickpos)
   {
      delete []this->minortickpos;
      this->minortickpos = NULL;
      this->minortickcnt = 0;
   }

   if (this->minor_num_adjusted == 0)
      return ERR_STATUS_GOOD;

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   ChipsVector min_pos_img(0,0);
   ChipsVector max_pos_img(1,1);
   ChipsVector min_pos(0,0);
   ChipsVector max_pos(1,1);

   if (dsCoord == NULL)
      return ERR_STATUS_BAD;

   if (!this->min_set || !this->max_set)
      return ERR_STATUS_BAD;

   Transform *transform = dsCoord->get_transform (dsCoord->get_default_transform_id ());

   if (transform == NULL)
      return ERR_STATUS_BAD;

   if (this->orientation == DEC_AXIS)
      poll_exists = check_poll ();

   this->get_axis_position (false, min_pos, max_pos);
    
   // copy some tickmarks before the first tick
   this->minortickcnt = (this->tickcnt-1)*this->minor_num_adjusted;
   this->minortickpos = new double[this->minortickcnt];

   // set the minor tick spacing
   this->minortick_spacing = this->tick_spacing / (this->minor_num_adjusted+1);

   for(unsigned int ii=0; ii<this->tickcnt-1; ii++)
   {

      /**** DO ALL MAJOR TICKS FALL EXTEND BEYOND PLOT BOUNDS???? ****/

      if (!reversed)
      {
         tmp_min = this->transformed_tickpos[ii];
         tmp_max = this->transformed_tickpos[ii+1];
      }
      else
      {
         tmp_max = this->transformed_tickpos[ii];
         tmp_min = this->transformed_tickpos[ii+1];
      }

      if ((this->orientation == RA_AXIS) && tmp_min<tmp_max)
         tmp_int = (tmp_max-(tmp_min+360))/(this->minor_num_adjusted+1);
      else
         tmp_int = (tmp_max-tmp_min)/(this->minor_num_adjusted+1);

      for(int jj=0; jj<this->minor_num_adjusted; jj++)
      {
         tmp_val = tmp_min+(jj+1)*tmp_int;

         if (this->orientation == RA_AXIS)
         {
            if (tmp_val < 0)
               tmp_val += 360;
            else if (tmp_val >= 360)
               tmp_val -= 360;
            this->minortickpos[ii*this->minor_num_adjusted+jj] = get_ra_intersection (tmp_val, min_pos[opp_index], RA_AXIS, transform);
         }
         else
         {
            num_hits = get_dec_intersection (tmp_val, min_pos[opp_index], DEC_AXIS, transform, tmp_img);
	    if (num_hits > 0)
            {
               double tmp_min_img = this->tickpos[ii];
               double tmp_max_img = this->tickpos[ii+1];
               if (tmp_min_img > tmp_max_img)
               {
                  double tmp = tmp_min_img;
                  tmp_min_img = tmp_max_img;
                  tmp_max_img = tmp;
               }
               for (int kk=0; kk<num_hits; kk++)
               {
                  if ((tmp_img[kk] >= tmp_min_img) && (tmp_img[kk] <= tmp_max_img))
                     this->minortickpos[ii*this->minor_num_adjusted+jj] = tmp_img[kk];
               }
            }
            else
               this->minortickcnt--;
         }
      }
   }

   return ERR_STATUS_GOOD;
}


double *** ChipsWCSTANAxisAlgorithm::calculate_grid (int &out_line_count, int &out_slices, int tick_type)
{
   double ***grid;

   if (!this->grid_modified)
      return NULL;

   // tick_type is 0 for major ticks and 1 for minor ticks
   double local_tickspacing = this->tick_spacing;

   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   int opp_index = (index == 0) ? 1 : 0;
   int poll_exists = 0;
   double hit_point = 0;
   double exclude_point = 0;
   int misses=0;
   int direction = 1;
   bool ticks_tested = false;

   int no_intersection = 0;

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   ChipsVector min_coord_img(0,0);
   ChipsVector max_coord_img(1,1);
   ChipsVector min_coord(0,0);
   ChipsVector max_coord(1,1);

   int slice=0;
   int max_line_count;
   out_line_count=0;


   if (dsCoord == NULL)
      return NULL;

   if (!this->min_set || !this->max_set)
      return NULL;


   Transform *transform = dsCoord->get_transform (dsCoord->get_default_transform_id ());

   if (transform == NULL)
      return NULL;

   if (this->axis->get_reverse ())
      local_tickspacing *= -1;

   if (tick_type == 1)
      local_tickspacing = this->minortick_spacing;

   if (local_tickspacing == 0)
      return NULL;

   poll_exists = check_poll ();

   out_slices = this->grid_slices;
   if (this->orientation == RA_AXIS)
      out_slices = 2;

   min_coord_img = dsCoord->get_lower_bounds ();
   max_coord_img = dsCoord->get_upper_bounds ();
   min_coord = dsCoord->get_lower_bounds ("");
   max_coord = dsCoord->get_upper_bounds ("");


   // estimate the number of grid lines that will be used
   max_line_count = fabs ((grid_extent[0] - grid_extent[1]) / local_tickspacing);
   // use 360 when poll exists
   if ((this->orientation == RA_AXIS) && poll_exists)
      max_line_count = ceil(360.0 / fabs(local_tickspacing));

   if (max_line_count == 0)
      return NULL;

   // create memory for the grid
   grid = new double**[max_line_count];
   for (int ii=0; ii<max_line_count; ii++)
   {
      grid[ii] = new double*[out_slices];
      for (int jj=0; jj<out_slices; jj++)
         grid[ii][jj] = new double[2];
   }

   min_coord_img = dsCoord->get_lower_bounds ();
   max_coord_img = dsCoord->get_upper_bounds ();

   if (min_coord_img[0] > max_coord_img[0])
   {
      double tmp = min_coord_img[0];
      min_coord_img[0] = max_coord_img[0];
      max_coord_img[0] = tmp;
   }
   if (min_coord_img[1] > max_coord_img[1])
   {
      double tmp = min_coord_img[1];
      min_coord_img[1] = max_coord_img[1];
      max_coord_img[1] = tmp;
   }

   // figure out the direction of that the tick spacing is going
   hit_point = this->grid_extent[0];
   exclude_point = this->grid_extent[0];
   if ((this->orientation == RA_AXIS) && (fabs (this->grid_extent[1] - this->grid_extent[0]) > 180))
   {
      if (this->grid_extent[1] > this->grid_extent[0])
         direction = copysignf (direction,(this->grid_extent[1] - (this->grid_extent[0]+360))/local_tickspacing);
      else
         direction = copysignf (direction,(this->grid_extent[1]+360 - this->grid_extent[0])/local_tickspacing);
   }


   // calculate the grid for RA axis when no poll exists
   if ((this->orientation == RA_AXIS) && !poll_exists)
   {
      double hit1[] = {0,0};
      double hit2[] = {0,0};

      // if this is a minor tick ignore the grid line if
      // it falls on a major tick line
      if (tick_type == 1)
      {
         hit_point += (direction*local_tickspacing);
         exclude_point += (direction*this->tick_spacing);
      }

      while (no_intersection < 2)
      {
         misses = 0;
      
         // check intersection with each plot edge -- each grid line
         // consists of two points. find which plot edge the line
         // intersects 

         hit1[1] = min_coord_img[opp_index];
         hit2[1] = max_coord_img[opp_index];
         hit1[0] = get_ra_intersection (hit_point, hit1[1], RA_AXIS, transform);
         hit2[0] = get_ra_intersection (hit_point, hit2[1], RA_AXIS, transform);

         if (hit1[0] < min_coord_img[index])
         {
            hit1[0] = min_coord_img[index];
            hit1[1] = get_ra_intersection (hit_point, hit1[0], DEC_AXIS, transform);
            if ((hit1[1] < min_coord_img[opp_index]) || (hit1[1] > max_coord_img[opp_index]))
               misses++;
         }
         else if (hit1[0] > max_coord_img[index])
         {
            hit1[0] = max_coord_img[index];
            hit1[1] = get_ra_intersection (hit_point, hit1[0], DEC_AXIS, transform);
            if ((hit1[1] < min_coord_img[opp_index]) || (hit1[1] > max_coord_img[opp_index]))
               misses++;
         }

         if (hit2[0] < min_coord_img[index])
         {
            hit2[0] = min_coord_img[index];
            hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);
            if ((hit2[1] < min_coord_img[opp_index]) || (hit2[1] > max_coord_img[opp_index]))
               misses++;
         }
         else if (hit2[0] > max_coord_img[index])
         {
            hit2[0] = max_coord_img[index];
            hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);
            if ((hit2[1] < min_coord_img[opp_index]) || (hit2[1] > max_coord_img[opp_index]))
               misses++;
         }

         // if neither edge hit then their are no more grid lines 
         // in this direction
         if (misses >= 2)
         {
            // if there hasn't been a hit yet keep
            // trying in this direction
            if (out_line_count > 0)
            {
               direction = -direction;
               hit_point = this->grid_extent[0];
               exclude_point = this->grid_extent[0] + (direction*this->tick_spacing);
               no_intersection++;
            }
         }
         else
         {

            // RA lines are always straight only need the two slices
            grid[out_line_count][slice][0] = hit1[0];
            grid[out_line_count][slice++][1] = hit1[1];

            grid[out_line_count][slice][0] = hit2[0];
            grid[out_line_count][slice++][1] = hit2[1];
            out_line_count++;
         }

         // test all ticks and then test beyond the tick marks
         hit_point += (direction*local_tickspacing);

         // if this is a minor tick ignore the grid line if
         // it falls on a major tick line
         if ((tick_type == 1) && flt_equal(hit_point, exclude_point))
         {
            hit_point += (direction*local_tickspacing);
            exclude_point += (direction*this->tick_spacing);
         }

         slice=0;

         // break out if there are too many lines 
         // most likely a bug has been hit
         if (out_line_count > 5000)
            break;
         // create more memory
         resize_grid (&grid, out_line_count, out_slices, max_line_count);
      }
   }
   // calculate the grid for DEC axis when no poll exists
   else if ((this->orientation == DEC_AXIS) && !poll_exists)
   {
      int num_hits = 0;
      int total_num_hits = 0;
      int max_no_intersection = 2;

      double ra_hits[8];
      double hit[2];
      double *tmp_hit;

      double min_ra, max_ra;

      double hit1[] = {-1,-1};
      double tol = 10e-8;

      // if this is a minor tick ignore the grid line if
      // it falls on a major tick line
      if (tick_type == 1)
      {
         hit_point += (direction*local_tickspacing);
         exclude_point += (direction*this->tick_spacing);
      }

      // loop through until all possible intersections have been explored
      while (no_intersection < max_no_intersection)
      {
         misses = 0;
         total_num_hits = 0;

         // check intersection with the plot edges -- there should
         // be at least one intersection with each plot edge. There
         // a possibility of two.
         num_hits = get_dec_intersection (hit_point, min_coord_img[opp_index], DEC_AXIS, transform, hit);

         hit1[opp_index] = min_coord_img[opp_index];
         for (int ii=0; ii<num_hits; ii++)
         {
            if ((hit[ii] <= max_coord_img[index]) && (hit[ii] >= min_coord_img[index]))
            {
               hit1[index] = hit[ii];
               tmp_hit = (double *)transform->apply ((void *)(&hit1[0]),dmDOUBLE,1);
               // verify the value retrieved is the expected value
               // plusses/minusses are lost in interesection eqation so
               // this may not always be the case
/*               if (flt_equal (tmp_hit[1],hit_point))
*/
               if (approx_equal (tmp_hit[1],hit_point,tol,tol))
                  ra_hits[total_num_hits++] = tmp_hit[0];
               delete []tmp_hit;
            }
         }

         num_hits = get_dec_intersection (hit_point, max_coord_img[opp_index], DEC_AXIS, transform, hit);
         hit1[opp_index] = max_coord_img[opp_index];
         for (int ii=0; ii<num_hits; ii++)
         {
            if ((hit[ii] <= max_coord_img[index]) && (hit[ii] >= min_coord_img[index]))
            {
               hit1[index] = hit[ii];
               tmp_hit = (double *)transform->apply ((void *)(&hit1[0]),dmDOUBLE,1);
               // verify the value retrieved is the expected value
               // plusses/minusses are lost in interesection eqation so
               // this may not always be the case
/*               if (flt_equal (tmp_hit[1],hit_point))
*/
               if (approx_equal (tmp_hit[1],hit_point,tol,tol))
                  ra_hits[total_num_hits++] = tmp_hit[0];
               delete []tmp_hit;
            }

         }

         num_hits = get_dec_intersection (hit_point, min_coord_img[index], RA_AXIS, transform, hit);
         hit1[index] = min_coord_img[index];
         for (int ii=0; ii<num_hits; ii++)
         {
            if ((hit[ii] <= max_coord_img[opp_index]) && (hit[ii] >= min_coord_img[opp_index]))
            {
               hit1[opp_index] = hit[ii];
               tmp_hit = (double *)transform->apply ((void *)(&hit1[0]),dmDOUBLE,1);

               // verify the value retrieved is the expected value
               // plusses/minusses are lost in interesection eqation so
               // this may not always be the case
/*               if (flt_equal (tmp_hit[1],hit_point))
*/
               if (approx_equal (tmp_hit[1],hit_point,tol,tol))
                  ra_hits[total_num_hits++] = tmp_hit[0];
               delete []tmp_hit;
            }
         }

         num_hits = get_dec_intersection (hit_point, max_coord_img[index], RA_AXIS, transform, hit);
         hit1[index] = max_coord_img[index];
         for (int ii=0; ii<num_hits; ii++)
         {

            if ((hit[ii] <= max_coord_img[opp_index]) && (hit[ii] >= min_coord_img[opp_index]))
            {
               hit1[opp_index] = hit[ii];
               tmp_hit = (double *)transform->apply ((void *)(&hit1[0]),dmDOUBLE,1);

               // verify the value retrieved is the expected value
               // plusses/minusses are lost in interesection eqation so
               // this may not always be the case
/*               if (fabs (tmp_hit[1]-hit_point) < FLT_EPSILON)
*/
               if (approx_equal (tmp_hit[1],hit_point,tol,tol))
                  ra_hits[total_num_hits++] = tmp_hit[0];
               delete []tmp_hit;
            }
         }


         // if none of the edges had an intersection then there
         // are no more grid lines in this direction
         if (total_num_hits == 0)
         {
            // if there have been no hits in this direction
            // keep trying until one is found
            if (ticks_tested)
            {
               direction = -direction;
               hit_point = this->grid_extent[0];
               exclude_point = this->grid_extent[0] + (direction*this->tick_spacing);
               no_intersection++;
            }
         }
         else
         {
            // go through each of the ra/decs to find the smallest and largest ra
            max_ra = min_ra = ra_hits[0];

            for (int ii=1; ii<total_num_hits; ii++)
            {
               if (max_ra < ra_hits[ii])
                  max_ra = ra_hits[ii];
               else if (min_ra > ra_hits[ii])
                  min_ra = ra_hits[ii];
            }

            // check if it crossed zero -- there cant be more than 180
            // degrees on a single line 
            if (max_ra - min_ra > 180)
            {
               for (int ii=0; ii<total_num_hits; ii++)
               {
                  if (ra_hits[ii] < 180)
                  {
                     ra_hits[ii] += 360;
                     if (min_ra < ra_hits[ii])
                        min_ra = ra_hits[ii];
                  }
               }
            }


            // calculate some number of slices for each grid line
            hit1[0] = min_ra;
            hit1[1] = hit_point;
	    double spacing = (max_ra - min_ra)/(out_slices-1);
            for (int ii=0; ii<out_slices; ii++)
            {
               tmp_hit = (double *)transform->invert ((void *)(&hit1[0]),dmDOUBLE,1);
               grid[out_line_count][ii][0] = tmp_hit[0];
               grid[out_line_count][ii][1] = tmp_hit[1];
               hit1[0] += spacing;
	       delete []tmp_hit;
            }
            out_line_count++;
         }

         // test all ticks and then test beyond the tick marks
         hit_point += (direction*local_tickspacing);

         // if this is a minor tick ignore the grid line if
         // it falls on a major tick line
         if ((tick_type == 1) && flt_equal(hit_point, exclude_point))
         {
            hit_point += (direction*local_tickspacing);
            exclude_point += (direction*this->tick_spacing);
         }

         if (((hit_point < this->grid_extent[0]) || (hit_point > this->grid_extent[1])) &&
             ((hit_point > this->grid_extent[0]) || (hit_point < this->grid_extent[1])))
            ticks_tested = true;

         // kick out if there are too many lines.
         // most likely a corner case bug
         if (out_line_count > 5000)
            break;

         // create more memory
         resize_grid (&grid, out_line_count, out_slices, max_line_count);
      }
   }
   // calculate the grid for RA axis when poll exists
   else if ((this->orientation == RA_AXIS) && poll_exists)
   {
      ChipsVector hit1 (0,0);
      ChipsVector hit2 (0,0); 
      double degrees_covered = 0;

      // if this is a minor tick ignore the grid line if
      // it falls on a major tick line
      if (tick_type == 1)
      {
         hit_point += (fabs (local_tickspacing));
         exclude_point += (fabs(this->tick_spacing));
      }


      hit1[1] = 90 * poll_exists;

      // convert the poll to image coordinates
      dsCoord->convert_cs_to_cs (&hit1, &hit1, 1, DATA, "", DATA, dsCoord->get_id_no_wcs ());

      // if the poll exists then find all grid lines which start at the
      // poll and emmenate out. this should cover 360 degrees
      while ((degrees_covered < 360) && (out_line_count != max_line_count))
      {

         // find which plot edge the grid line intersects
         hit2[1] = min_coord_img[opp_index];
         hit2[0] = get_ra_intersection (hit_point, hit2[1], RA_AXIS, transform);
         if (hit2[0] < min_coord_img[index])
         {
            hit2[0] = min_coord_img[index];
            hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);

            // verify the intersection found is a valid if not try the
            // other side of the plot
            if (!this->check_point (&hit2[0], hit_point, 0, transform))
            {
               hit2[0] = max_coord_img[index];
               hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);
            }

         }
         else if (hit2[0] > max_coord_img[index])
         {
            hit2[0] = max_coord_img[index];
            hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);

            // verify the intersection found is a valid if not try the
            // other side of the plot
            if (!this->check_point (&hit2[0], hit_point, 0, transform))
            {
               hit2[0] = min_coord_img[index];
               hit2[1] = get_ra_intersection (hit_point, hit2[0], DEC_AXIS, transform);
            }
         }
         else
         {

            // verify the intersection found is a valid if not try the
            // other side of the plot
            if (!this->check_point (&hit2[0], hit_point, 0, transform))
            {
               hit2[1] = max_coord_img[opp_index];
               hit2[0] = get_ra_intersection (hit_point, hit2[1], RA_AXIS, transform);
            }
         }


         // RA lines are always straight only need the two slices
         grid[out_line_count][0][0] = hit1[0];
         grid[out_line_count][0][1] = hit1[1];

         grid[out_line_count][1][0] = hit2[0];
         grid[out_line_count][1][1] = hit2[1];
         out_line_count++;

         hit_point += fabs(local_tickspacing);

         if (hit_point >= 360)
            hit_point -= 360;

         // if this is a minor tick ignore the grid line if
         // it falls on a major tick line
         if ((tick_type == 1) && flt_equal(hit_point, exclude_point))
         {
            hit_point += (fabs (local_tickspacing));
            exclude_point += (fabs(this->tick_spacing));
            if (exclude_point >= 360)
               exclude_point -= 360;
         }

         degrees_covered += fabs(local_tickspacing);
      }
   }
   // calculate the grid for DEC axis when poll exists
   else if ((this->orientation == DEC_AXIS) && poll_exists)
   {
      double *tmp_hit;
      double hit[] = {0,0};
      double poll_radius = 0;
      double *hit_poll;
      double corner_dist = 0;
      double tmp_dist = 0;
      double spacing = 0;
      double start_ra = 0;

      exclude_point = poll_exists * 90;

      local_tickspacing = fabs (local_tickspacing);
      direction = -1 * poll_exists;

      hit[1] = poll_exists * 90;
      hit_poll = (double *)transform->invert ((void *)(&hit[0]),dmDOUBLE,1);

      // find the maximum distance between the poll and one of the corner
      corner_dist = sqrt (pow(hit_poll[0] - min_coord_img[0], 2) + pow(hit_poll[1] - min_coord_img[1], 2));
      tmp_dist = sqrt (pow(hit_poll[0] - max_coord_img[0], 2) + pow(hit_poll[1] - max_coord_img[1], 2));

      // figure out the ra of the corner
      hit[0] = min_coord_img[0];
      hit[1] = min_coord_img[1];
      tmp_hit = (double *)transform->apply ((void *)(&hit[0]),dmDOUBLE,1);
      hit[0] = tmp_hit[0];
      delete[] tmp_hit;

      if (corner_dist < tmp_dist)
      {
         corner_dist = tmp_dist;
         hit[0] = max_coord_img[0];
         hit[1] = max_coord_img[1];
         tmp_hit = (double *)transform->apply ((void *)(&hit[0]),dmDOUBLE,1);
         hit[0] = tmp_hit[0];
         delete[] tmp_hit;
      }
      corner_dist = sqrt (pow(hit_poll[0] - min_coord_img[0], 2) + pow(hit_poll[1] - max_coord_img[1], 2));
      if (corner_dist < tmp_dist)
      {
         corner_dist = tmp_dist;
         hit[0] = min_coord_img[0];
         hit[1] = max_coord_img[1];
         tmp_hit = (double *)transform->apply ((void *)(&hit[0]),dmDOUBLE,1);
         hit[0] = tmp_hit[0];
         delete[] tmp_hit;
      }
      tmp_dist = sqrt (pow(hit_poll[0] - max_coord_img[0], 2) + pow(hit_poll[1] - min_coord_img[1], 2));
      if (corner_dist < tmp_dist)
      {
         corner_dist = tmp_dist;
         hit[0] = max_coord_img[0];
         hit[1] = min_coord_img[1];
         tmp_hit = (double *)transform->apply ((void *)(&hit[0]),dmDOUBLE,1);
         hit[0] = tmp_hit[0];
         delete[] tmp_hit;
      }


      hit[1] = poll_exists * 90;
      hit[1] += (direction*local_tickspacing);
      start_ra = hit[0];
      // if this is a minor tick ignore the grid line if
      // it falls on a major tick line
      if (tick_type == 1)
      {
         hit[1] += (direction*local_tickspacing);
         exclude_point += (direction*fabs(this->tick_spacing));
      }

      tmp_hit = (double *)transform->invert ((void *)(&hit[0]),dmDOUBLE,1);

      poll_radius = sqrt (pow(hit_poll[0]-tmp_hit[0], 2) + pow(hit_poll[1]-tmp_hit[1], 2));
      delete []tmp_hit;


      // create concentric circles around the poll. continue the circles
      // until radius of the grid line exceeds the edge of the plot
      while (poll_radius < corner_dist)
      {

         // create circle
         spacing = 360.0/(out_slices-1);
         for (int ii=0; ii<out_slices; ii++)
         {
            tmp_hit = (double *)transform->invert ((void *)(&hit[0]),dmDOUBLE,1);

            if (check_point (tmp_hit, hit[1], 1, transform))
            {
               grid[out_line_count][ii][0] = tmp_hit[0];
               grid[out_line_count][ii][1] = tmp_hit[1];
            }
            else if (ii > 0)
            {
               grid[out_line_count][ii][0] = grid[out_line_count][ii-1][0];
               grid[out_line_count][ii][1] = grid[out_line_count][ii-1][1];
            }
            else
            {
               grid[out_line_count][ii][0] = 0;
               grid[out_line_count][ii][1] = 0;
            }

            hit[0] += spacing;
            delete []tmp_hit;
         }
         hit[0] = start_ra;


         // calculate the radius
         hit[1] += (direction*local_tickspacing);
         // if this is a minor tick ignore the grid line if
         // it falls on a major tick line
         if ((tick_type == 1) && flt_equal(hit[1], exclude_point))
         {
            hit[1] += (direction*local_tickspacing);
            exclude_point += (direction*fabs(this->tick_spacing));
         }

         tmp_hit = (double *)transform->invert ((void *)(&hit[0]),dmDOUBLE,1);
         if (check_point (tmp_hit, hit[1], 1, transform))
            poll_radius = sqrt (pow(hit_poll[0]-tmp_hit[0], 2) + pow(hit_poll[1]-tmp_hit[1], 2));
         else
            poll_radius = corner_dist;
         delete []tmp_hit;

         out_line_count++;
         resize_grid (&grid, out_line_count, out_slices, max_line_count);
      }
   }


   // squeeze the grid array
   for (int ii=out_line_count; ii < max_line_count; ii++)
   {
      for (int jj=0; jj<out_slices; jj++)
         delete []grid[ii][jj];
      delete []grid[ii];
   }

   return grid;
}


// SCALE2 algorithm (C.R. Lewart, CACM, Oct 73, v16 n10)
// finds a new range to nicely divide data into exactly n linear intervals
// ** modified for degrees **
void ChipsWCSTANAxisAlgorithm::scale2(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist)
{
   double interval, scaled, fm, del = 0.000001;
   double lin_int[] = { 1.0, 2.0, 5.0, 10.0, 20.0 };
   double deg_int[] = {15.0, 30.0, 45.0, 60.0, 90.0};
   double *vint = &lin_int[0];
   int nal, m1, m2, nx, np, done = 0, ii = 1;
   int max_interval = 5;

   // account for wrap around 360
   if (inmin > inmax)
   {
      double tmp_max = inmax;
      inmax = inmin + 360;
      inmin = tmp_max;
   }

   // scale the min/max to be with reasonable limits
   while (inmin >= 360)
   {
      inmin -= 360;
      inmax -= 360;
   }

   while (inmax-inmin >= 360)
      inmax -= 360;

   // calculate the interval scale
   interval = (inmax-inmin)/(double)slices;
   nal = (int) log10(interval);
   if (interval < 1)   --nal;
   scaled = interval/pow((double)10.0, nal);

   // if the interval is greater than the smallest degrees use
   // degree interval calculations -- that is intervals greater
   // than 15 should use nice degree intervals
   if (interval >= deg_int[0])
   {
      vint = &deg_int[0];
      max_interval = 4;
      scaled = interval;
      nal = 0;
   }


   // find an appropriate intervael
   while (!done && (ii <= max_interval))
   {
      if ((scaled < vint[ii - 1] + del) || (ii == max_interval))
      {
         *dist = vint[ii - 1] * pow((double)10.0, nal);
         m1 = (int)(fm = inmin/(*dist));
         if (fabs((double)m1 + 1.0 - fm) < del) ++m1;
         *outmin = *dist * m1;
         fm = inmax/(*dist);
         m2 = (int)(fm + 1);
         if (fabs(fm + 1.0 -(double)m2) < del)  --m2;
         *outmax = (*dist) * m2;
         if ((np = m2 - m1) <= slices) done = 1;
      }
      ++ii;
   }

   nx = (slices - np) / 2;
   *outmin -= (double) nx * (*dist);
   *outmax = *outmin + (double)slices * (*dist);
   if (*outmin > inmin)   *outmin = inmin;
   if (*outmax < inmax)   *outmax = inmax;
}

// SCALE1 algorithm (C.R. Lewart, CACM, Oct 73, v16 n10)
// finds a new range to nicely divide data into approx n linear intervals
// ** modified for degrees
void ChipsWCSTANAxisAlgorithm::scale1(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist, double *base)
{
   double interval, scaled, fm, baseval;
   int nal, m;

   // scale the min/max to be with reasonable limits
   while (inmin >= 360)
   {
      inmin -= 360;
      inmax -= 360;
   }

   while (inmax-inmin > 360)
      inmax -= 360;


   if (inmax == inmin)  ++inmax;
   if (!slices) slices = 5;     // avoid div by 0- use 5 as desired num ticks

   // calculate the interval scale
   interval = (inmax-inmin)/(double)slices;
   nal = (int) log10(interval);
   if (interval < 1) --nal;

   // scale interval between 1-10 and find value base on the geometric mean
   // if the interval is greater than the smallest degrees use
   // degree interval calculations -- that is intervals greater
   // than 15 should use nice degree intervals
   if (interval <= 15)
   {
      if ((scaled = interval/pow((double)10.0, nal)) < 1.41421356237) baseval = 1.0;
         else if (scaled < 3.16227766017)                     baseval = 2.0;
            else  if (scaled < 7.07106781187)                 baseval = 5.0;
               else                                           baseval = 10;
   }
   else
   {
      if ((scaled = interval/pow((double)10.0, nal)) < sqrt(1.5*3)) baseval = 1.5;
         else if (scaled < sqrt(1.5*3.0*4.5))               baseval = 3.0;
            else  if (scaled < sqrt(1.5*3.0*4.5*6))         baseval = 4.5;
               else if (scaled < sqrt(1.5*3.0*4.5*6*9))     baseval = 6.0;
                  else                                      baseval = 9.0;
   }

     
   if (base)  *base = baseval;

   if ((*dist = copysign(baseval * pow((double)10.0, nal), (inmax - inmin))) != 0)
   {
      // compute interval size
      fm = inmin/(*dist);
      if (fm < 0.0)   m = (int)fm - 1;
         else   m = (int)fm;
      if (fabs((double)m+1.0-fm) < 0.0000000000000001)  ++m;

      // find new minimum and maximum
      *outmin = (*dist) * (double)m;
      fm = inmax/(*dist);
      if (fm < -1.0)  m = (int)fm;
         else m = (int)fm + 1;
      if (fabs(fm+1.0-(double)m) < 0.0000000000000001)  --m;
      *outmax = (*dist) * (double)m;

      // adjust for roundoff
      if (*outmin > inmin)   *outmin = inmin;
      if (*outmax < inmax)   *outmax = inmax;
   }
   else
   {
      *outmin = inmin;
      *outmax = inmax;
   }
}

void ChipsWCSTANAxisAlgorithm::scale1_sexg_dec(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist, double *base)
{

   double min_degree, min_minute, min_second;
   double max_degree, max_minute, max_second;
   // convert the input to dec
   deg_to_dec (inmin, min_degree, min_minute, min_second);
   deg_to_dec (inmax, max_degree, max_minute, max_second);

   // adjust the max to span just that unit
   double degrees[] = {min_degree,max_degree};
   double minutes[] = {min_minute, max_minute+(max_degree-min_degree)*60};
   double seconds[] = {min_second, max_second+(minutes[1]-min_minute)*60};

   double tmp;
   if (degrees[0] > degrees[1])
   {
      tmp = degrees[1];
      degrees[1] = degrees[0];
      degrees[0] = tmp;
   }
   degrees[1]++; // pad out the degrees

   if (minutes[0] > minutes[1])
   {
      tmp = minutes[1];
      minutes[1] = minutes[0];
      minutes[0] = tmp;
   }
   minutes[1]++; // pad out the minutes

   if (seconds[0] > seconds[1])
   {
      tmp = seconds[1];
      seconds[1] = seconds[0];
      seconds[0] = tmp;
   }


   // return everything in degrees
   if ((degrees[1] - degrees[0]) >= slices-1)
      scale1(degrees[0], degrees[1], slices, outmin, outmax, dist, base);
   else if ((minutes[1] - minutes[0]) >= slices-1)
   {
      scale1(minutes[0], minutes[1], slices, outmin, outmax, dist, base);
      // check for a swap and convert back to degrees note that the seconds are
      // based off of the mininum value
      if (minutes[0] == min_minute)
      {
         *outmin = (*outmin)/60. + min_degree;
         *outmax = (*outmax)/60. + min_degree;
      }
      else
      {
         *outmin = (*outmin)/60. + max_degree;
         *outmax = (*outmax)/60. + max_degree;
      }

      (*dist) /= 60.;
   }
   else
   {
      scale1(seconds[0], seconds[1], slices, outmin, outmax, dist, base);
      // check for a swap and convert back to degrees note that the seconds are
      // based off of the mininum value
      if (seconds[0] == min_second)
      {
         *outmin = (((*outmin)/60.) + min_minute)/60. + min_degree;
         *outmax = (((*outmax)/60.) + min_minute)/60. + min_degree;
      }
      else
      {
         *outmin = (((*outmin)/60.) + max_minute)/60. + max_degree;
         *outmax = (((*outmax)/60.) + max_minute)/60. + max_degree;
      }
      (*dist) /= 3600.;
   }
}


void ChipsWCSTANAxisAlgorithm::scale1_sexg_ra(double inmin, double inmax, int slices,
   double* outmin, double* outmax, double* dist, double *base)
{

   double min_hour, min_minute, min_second;
   double max_hour, max_minute, max_second;

   // convert the input to ra
   deg_to_ra (inmin, min_hour, min_minute, min_second, false);
   deg_to_ra (inmax, max_hour, max_minute, max_second, false);

   // adjust the max to span just that unit
   double hours[] = {min_hour,max_hour};
   double minutes[] = {min_minute, max_minute+(max_hour-min_hour)*60};
   double seconds[] = {min_second, max_second+(minutes[1]-min_minute)*60};

   double tmp;
   if (hours[0] > hours[1])
   {
      tmp = hours[1];
      hours[1] = hours[0];
      hours[0] = tmp;
   }
   hours[1]++; // pad out the hours

   if (minutes[0] > minutes[1])
   {
      tmp = minutes[1];
      minutes[1] = minutes[0];
      minutes[0] = tmp;
   }
   minutes[1]++; // pad out the minutes

   if (seconds[0] > seconds[1])
   {
      tmp = seconds[1];
      seconds[1] = seconds[0];
      seconds[0] = tmp;
   }


   // return everything in degrees
   if ((hours[1] - hours[0]) >= slices-1)
   {
      scale1(hours[0], hours[1], slices, outmin, outmax, dist, base);
      (*outmin) *= 15;
      (*outmax) *= 15;
      (*dist) *= 15;
   }
   else if ((minutes[1] - minutes[0]) >= slices-1)
   {
      scale1(minutes[0], minutes[1], slices, outmin, outmax, dist, base);
      // check for a swap and convert back to degrees note that the seconds are
      // based off of the mininum value
      if (minutes[0] == min_minute)
      {
         *outmin = (*outmin)*.25 + min_hour*15;
         *outmax = (*outmax)*.25 + min_hour*15;
      }
      else
      {
         *outmin = (*outmin)*.25 + max_hour*15;
         *outmax = (*outmax)*.25 + max_hour*15;
      }

      (*dist) *= .25;
   }
   else
   {
      scale1(seconds[0], seconds[1], slices, outmin, outmax, dist, base);
      // check for a swap and convert back to degrees note that the seconds are
      // based off of the mininum value
      if (seconds[0] == min_second)
      {
         *outmin = (((*outmin)/60.) + min_minute)*.25 + min_hour*15;
         *outmax = (((*outmax)/60.) + min_minute)*.25 + min_hour*15;
      }
      else
      {
         *outmin = (((*outmin)/60.) + max_minute)*.25 + max_hour*15;
         *outmax = (((*outmax)/60.) + max_minute)*.25 + max_hour*15;
      }
      (*dist) *= (15/3600.);
   }
}

void ChipsWCSTANAxisAlgorithm::deg_to_ra (double in_degree, double &out_hour, 
                                            double &out_minute, double &out_second, 
                                            bool wrap_hour)
{
   out_second = in_degree/15.0;
   out_second = modf(out_second, &out_hour);
   if (wrap_hour)
      out_hour = fmod(out_hour, 24);       // wrap at 24 hours
   if (fabs (out_second) < FLT_EPSILON)
      out_second = 0;
   out_second = modf(out_second*60.0, &out_minute);
   if (fabs (out_second) < FLT_EPSILON)
      out_second = 0;
   out_second *= 60; /* ??? + 0.00137; */
}


void ChipsWCSTANAxisAlgorithm::deg_to_dec (double in_degree, double &out_degree,
                                            double &out_minute, double &out_second)
{
   out_second = modf(in_degree, &out_degree);
   if (out_second >= 0)
      out_second = modf(out_second*60.0 + FLT_EPSILON, &out_minute);
   else
      out_second = modf(out_second*60.0 - FLT_EPSILON, &out_minute);
   out_second *= 60; /* ??? + 0.00137; */

}


void ChipsWCSTANAxisAlgorithm::reset_string_labels(const string& tickformat)
{
   unsigned int rr;

   if ((major_mode == chips_arbitrary) && this->arbtext)
   {
      if (this->sexagesimal)
      {
         this->modified = true;
         this->sexagesimal = false;
      }
      if (this->tick_text != NULL)
      {
         // tick text has changed- recompute tick_text
         delete[] this->tick_text;
         this->ttext_cnt = this->arbcnt;
         this->tick_text = new string[this->arbcnt];
      }
   }
   else if (strchr(tickformat.c_str(), 'z') || strchr(tickformat.c_str(), 'Z'))
   {
      // format using 3.2 x10^2 instead of 3.2e+02 (%z=%e, %Z=%g)

      if (this->sexagesimal)
      {
         bool updated;
         this->modified = true;
         this->sexagesimal = false;
         this->update_tick_positions (updated);
      }

      char tmpfmt[80];
      char tmpval[80];
      bool neg;
      strcpy(tmpfmt, tickformat.c_str());
      char* cpos = strchr(tmpfmt, 'z');
      if (!cpos)
      {
         cpos = strchr(tmpfmt, 'Z');
         cpos = strchr(tmpfmt, 'Z');
         if (cpos) *cpos = 'g';
      }
      else   *cpos = 'e';

      if (this->tick_text) delete []this->tick_text;
      this->tick_text = new string[this->tickcnt];
      this->ttext_cnt = tickcnt;

      unsigned int set_cnt = 0;
      if (tickformat == "%0.0z")
      {
         for (rr = 0; rr < this->tickcnt; rr++)
         {
            sprintf(tmpval, "%.3e", this->transformed_tickpos[rr]);
            if (strstr(tmpval, "1.000e") == &tmpval[0])  set_cnt++;
               else if (strstr(tmpval, "-1.000e") == &tmpval[0])  set_cnt++;
                  else if (strstr(tmpval, "0.000e") == &tmpval[0])  set_cnt++;
                     else if (this->transformed_tickpos[rr]< get_min() || this->transformed_tickpos[rr] > this->get_max())
                        set_cnt++;
         }
         if (set_cnt == this->tickcnt) strcpy(tmpfmt, "%.3e");
      }
      for (rr = 0; rr < this->tickcnt; rr++)
      {
         sprintf(tmpval, tmpfmt, this->transformed_tickpos[rr]);
         cpos = strchr(tmpval, 'e');
         if (cpos)
         {
            *cpos = '\0';
            // handle %0.0z as only 10^x
            if (set_cnt != this->tickcnt || this->transformed_tickpos[rr] != 0)
            {
               if (set_cnt != this->tickcnt)
               {
                  this->tick_text[rr] = tmpval;
                  this->tick_text[rr].append(" \\times 10^{");
               }
               else if (this->transformed_tickpos[rr] > 0)
                  this->tick_text[rr] = ("10^{");
               else
                  this->tick_text[rr] = ("-10^{");
               cpos++;
               neg =  (*cpos == '-');
               cpos++;
               if (neg) this->tick_text[rr].append("-");
               if (*cpos == '\0')  this->tick_text[rr].append("0");
               else               this->tick_text[rr].append(cpos);
               this->tick_text[rr].append("}");
            }
            else   this->tick_text[rr] = "0";
         }
         else   this->tick_text[rr] = tmpval;
      }
   }
   else if (strstr(tickformat.c_str(),"ra") || strstr(tickformat.c_str(),"RA"))
   {
      // using sexagesimal (HMS) format for right ascension (ra)
      if (!this->sexagesimal)
      {
         this->modified = true;
         this->sexagesimal = true;
      }
      sexagesimal_ra_ticks(tickformat);
   }
   else if (strstr(tickformat.c_str(),"dec") || strstr(tickformat.c_str(),"DEC"))
   {
      // using sexagesimal (DMS) format for declination (dec)
      if (!this->sexagesimal)
      {
         this->modified = true;
         this->sexagesimal = true;
      }
      sexagesimal_dec_ticks(tickformat);
   }
   else
   {
      char tmpval[80];

      if (this->sexagesimal)
      {
         bool updated;
         this->modified = true;
         this->sexagesimal = false;
         this->update_tick_positions (updated);
      }

      if (this->tick_text) delete []this->tick_text;
      this->tick_text = new string[this->tickcnt];
      this->ttext_cnt = tickcnt;

      for (rr = 0; rr < this->tickcnt; rr++)
      {
         sprintf(tmpval, tickformat.c_str (), this->transformed_tickpos[rr]);
         this->tick_text[rr] = tmpval;
      }
   }
}

void ChipsWCSTANAxisAlgorithm::serialize (ChipsStateStream &out)
{
   bool bval;
   int ival;

   out.write_int ((int *)&this->ttext_cnt);
   out.write_int ((int *)&this->tickcnt);
   out.write_int ((int *)&this->arbcnt);

   if (this->tick_text)
   {
      for (unsigned int ii=this->ttext_cnt; ii--;)
         out.write_string (&this->tick_text[ii]);
   }
   if (this->tickpos)
      out.write_double (this->tickpos, (int)this->tickcnt);
   if (this->transformed_tickpos)
      out.write_double (this->transformed_tickpos, (int)this->tickcnt);
   if (this->arbpos)
      out.write_double (this->arbpos, (int)this->arbcnt);

   bval = (this->arbtext != NULL);
   out.write_bool (&bval);

   if (this->arbtext)
   {
      for (unsigned int ii=this->ttext_cnt; ii--;)
         out.write_string (this->arbtext);
   }

   ival = this->major_mode;
   out.write_int (&ival);
   ival = this->major_mode_adjusted;
   out.write_int (&ival);
   ival = this->minor_mode;
   out.write_int (&ival);
   ival = this->minor_mode_adjusted;
   out.write_int (&ival);
   out.write_int (&this->minor_num);
   out.write_int (&this->minor_num_adjusted);
   out.write_int (&this->major_num);
   out.write_double (&this->major_int);
   out.write_double (&this->minor_int);

   out.write_double (&this->padded_min);
   out.write_double (&this->padded_max);
   out.write_double (&this->transformed_padded_min);
   out.write_double (&this->transformed_padded_max);
   out.write_double (&this->absolute_min);
   out.write_double (&this->absolute_max);
   out.write_double (&this->transformed_absolute_min);
   out.write_double (&this->transformed_absolute_max);

   out.write_bool (&this->arbitrary_set);

   out.write_double (&this->last_pad);
   out.write_bool (&this->last_automin);
   out.write_bool (&this->last_automax);
   out.write_double (&this->last_axis_position);

   out.write_int (&this->orientation);

   out.write_bool (&this->sexagesimal);

   out.write_int (&this->grid_slices);
   out.write_double (&this->tick_spacing);

   out.write_int (&this->min_set);
   out.write_int (&this->max_set);

   out.write_double (&this->minortick_spacing);
   out.write_double (this->grid_extent, 2);

}

void ChipsWCSTANAxisAlgorithm::deserialize (ChipsStateStream &in)
{
   bool bval;
   int ival;

   in.read_int ((int *)&this->ttext_cnt);
   in.read_int ((int *)&this->tickcnt);
   in.read_int ((int *)&this->arbcnt);

   if (this->tick_text)
   {
      delete []this->tick_text;
      this->tick_text = NULL;
   }

   if (this->ttext_cnt)
   {
      this->tick_text = new string[this->ttext_cnt];
      for (unsigned int ii=this->ttext_cnt; ii--;)
         in.read_string (&this->tick_text[ii]);
   }
   if (this->tickpos)
   {
      delete []this->tickpos;
      this->tickpos = NULL;
   }

   if (this->transformed_tickpos)
   {
      delete []this->transformed_tickpos;
      this->transformed_tickpos = NULL;
   }


   if (this->tickcnt)
   {
      this->tickpos = new double[this->tickcnt];
      in.read_double (this->tickpos, (int)this->tickcnt);
   }

   if (this->tickcnt)
   {
      this->transformed_tickpos = new double[this->tickcnt];
      in.read_double (this->transformed_tickpos, (int)this->tickcnt);
   }

   if (this->arbpos)
   {
      delete []this->arbpos;
      this->arbpos = NULL;
   }

   if (this->arbcnt)
   {
      this->arbpos = new double[this->arbcnt];
      in.read_double (this->arbpos, (int)this->arbcnt);
   }

   in.read_bool (&bval);

   if (this->arbtext)
   {
      delete []this->arbtext;
      this->arbtext = NULL;
   }

   if (bval)
   {
      this->arbtext = new string[this->arbcnt];
      for (int ii=this->arbcnt; ii--;)
         in.read_string (&this->arbtext[ii]);
   }

   in.read_int (&ival);
   this->major_mode = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&ival);
   this->major_mode_adjusted = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&ival);
   this->minor_mode = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&ival);
   this->minor_mode_adjusted = static_cast <V_TICKMODE_TYPE>(ival);
   in.read_int (&this->minor_num);
   in.read_int (&this->minor_num_adjusted);
   in.read_int (&this->major_num);
   in.read_double (&this->major_int);
   in.read_double (&this->minor_int);

   in.read_double (&this->padded_min);
   in.read_double (&this->padded_max);
   in.read_double (&this->transformed_padded_min);
   in.read_double (&this->transformed_padded_max);

   in.read_double (&this->absolute_min);
   in.read_double (&this->absolute_max);

   in.read_double (&this->transformed_absolute_min);
   in.read_double (&this->transformed_absolute_max);

   in.read_bool (&this->arbitrary_set);

   in.read_double (&this->last_pad);
   in.read_bool (&this->last_automin);
   in.read_bool (&this->last_automax);
   in.read_double (&this->last_axis_position);

   in.read_int (&this->orientation);

   in.read_bool (&this->sexagesimal);

   in.read_int (&this->grid_slices);
   in.read_double (&this->tick_spacing);

   in.read_int (&this->min_set);
   in.read_int (&this->max_set);

   if (in.get_minor_version() >= 10)
   {
      in.read_double (&this->minortick_spacing);
      in.read_double (this->grid_extent, 2);
   }
}

void ChipsWCSTANAxisAlgorithm::reset_adjusted_parameters ()
{
   this->minor_mode_adjusted = this->minor_mode;
   this->minor_num_adjusted = this->minor_num;
   this->major_mode_adjusted = this->major_mode;
}

int ChipsWCSTANAxisAlgorithm::sexagesimal_ra_ticks(const string &tickformat)
{
   // converts tick positions to sexagesimal format for ra
   if (this->tickcnt == 0)
   {
//      cmd_manager->add_err("Cannot set tick format to sexagesimal if there are no tick marks.", id, 1);
      return ERR_STATUS_BAD;
   }

   // set 'nice' tick locations for non-user specified tick locations
   // for now- simply plot log scale values at their decimal degree locations
/*   if ((this->major_mode != chips_arbitrary) && (this->scaletype != V_SCALE_LOG))
      this->compute_sexagesimal_locations(0);
*/
   // create labels
   string hmfmt;
   double tpos, hd, md, ss;
   unsigned int rr;
   char tmpfmt[80], tmpval[80], tmpspec[80];
   char *fmt = tmpspec;

   strcpy(tmpspec, tickformat.c_str());
   char* cpos = strstr(tmpspec, "ra");
   if (!cpos)   cpos = strstr(tmpspec, "RA");
   *cpos = '\0';
   if (tmpspec[0] == '%') fmt++;
   if (!strlen(fmt))
   {
      strcpy(fmt, "02.0");
      if (this->tickcnt > 0)
      {
         deg_to_ra (this->tick_spacing, hd, md, ss);
         ss = fabs(ss);
         // if the tick spacing is less than 1
         if (((hd > -FLT_EPSILON) && (hd < FLT_EPSILON)) &&
             ((md > -FLT_EPSILON) && (md < FLT_EPSILON)) &&
             ((ss < 1) && (ss > 0)))
         {
            ss = -log10 (ss);
            sprintf (fmt, "02.%d",(int)ceil(ss));
         }
      }
   }     
   if (strstr(tickformat.c_str(), "ra1"))
   {
      hmfmt = "%02d:%02d:";
      sprintf(tmpfmt, "%%%sf", fmt);
   }
   else if (strstr(tickformat.c_str(), "ra2"))
   {
      hmfmt="%02d %02d ";
      sprintf(tmpfmt, "%%%sf", fmt);
   }
   else if (strstr(tickformat.c_str(), "RA3"))
   {
      hmfmt="%02dH %02dM ";
      sprintf(tmpfmt, "%%%sfS", fmt);
   }
   else if (strstr(tickformat.c_str(), "ra3"))
   {
      hmfmt = "%02dh %02dm ";
      sprintf(tmpfmt, "%%%sfs", fmt);
   }
   else if (strstr(tickformat.c_str(), "RA4"))
   {
      hmfmt = "%02d^H %02d^M ";
      sprintf(tmpfmt, "%%%sf^S", fmt);
   }
   else
   {
      hmfmt = "%02d^h %02d^m ";
      sprintf(tmpfmt, "%%%sf^s", fmt);
   }

   if (this->tick_text) delete []this->tick_text;
   this->tick_text = new string[this->tickcnt];
   this->ttext_cnt = this->tickcnt;

   char* sixty_p;
   for (rr = 0; rr < this->tickcnt; rr++)
   {
      if ((tpos = fmod(this->transformed_tickpos[rr], 360.0)) < 0) tpos+=360;
      deg_to_ra (tpos, hd, md, ss);

      // account for roundup to 60 seconds
      sprintf(tmpspec, tmpfmt, ss);
      if ((sixty_p = strstr(tmpspec,"60")) != NULL)
      {
         *sixty_p = '0';
         md++;
      }
      // account for roundup to 60 minutes;
      sprintf(tmpval,hmfmt.c_str(), (int)hd, (int)md);
      if ((sixty_p = strstr(tmpval,"60")) != NULL)
      {
         // account for 23h to 00h
         if (hd == 23)  sprintf(tmpval,hmfmt.c_str(), 0, 0);
            else        sprintf(tmpval,hmfmt.c_str(), (int)hd+1, 0);
      }
      this->tick_text[rr] = tmpval;
      this->tick_text[rr].append(tmpspec);
   }
   return ERR_STATUS_GOOD;
}


int ChipsWCSTANAxisAlgorithm::sexagesimal_dec_ticks(const string &tickformat)
{
   // converts tick positions to sexagesimal format for dec
   if (this->tickcnt == 0)
   {
//      cmd_manager->add_err("Cannot set tick format to sexagesimal if there are no tick marks.", id, 1);
      return ERR_STATUS_BAD;
   }

   // create labels
   double hd, md, ss;
   unsigned int swapsign, rr;
   char tmpfmt[80], tmpval[80], tmpspec[80];
   char *fmt = tmpspec;

   strcpy(tmpspec, tickformat.c_str());
   char* cpos = strstr(tmpspec, "dec");
   if (!cpos)   cpos = strstr(tmpspec, "DEC");
   *cpos = '\0';
   if (tmpspec[0] == '%') fmt++;
   if (!strlen(fmt))
   {
      strcpy(fmt, "02.0");
      if (this->tickcnt > 0)
      {
         deg_to_dec (this->tick_spacing, hd, md, ss);
         ss = fabs(ss);
         // if the tick spacing is less than 1
         if (((hd > -FLT_EPSILON) && (hd < FLT_EPSILON)) &&
             ((md > -FLT_EPSILON) && (md < FLT_EPSILON)) &&
             ((ss < 1) && (ss > 0)))
         {
            ss = -log10 (ss);
            sprintf (fmt, "02.%d",(int)ceil(ss));
         }
      }
   }
   if (strstr(tickformat.c_str(), "dec1"))
      sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);
   else if (strstr(tickformat.c_str(), "dec2"))
      sprintf(tmpfmt, "%%+02d %%02d %%%sf", fmt);
   else if (strstr(tickformat.c_str(), "DEC3"))
      sprintf(tmpfmt, "%%+02dD %%02dM %%%sfS", fmt);
   else if (strstr(tickformat.c_str(), "dec3"))
      sprintf(tmpfmt, "%%+02dd %%02dm %%%sfs", fmt);
   else
      sprintf(tmpfmt, "%%+02d^o %%02d' %%%sf\"", fmt);

   if (tick_text)
      delete []this->tick_text;
   this->tick_text = new string[tickcnt];
   this->ttext_cnt = this->tickcnt;

   for (rr = 0; rr < this->tickcnt; rr++)
   {
      deg_to_dec (this->transformed_tickpos[rr], hd, md, ss);
      // handle degrees between -1:0 with a - prefix infront of degrees
      if ((swapsign = ((hd <= 0) && ((md < 0) || (ss < 0)))))
      {
         if (md < 0) md *= -1;
         if (ss < 0) ss *= -1;
      }
      // handle minutes field set to +/-60 due to precision issues
      if ((md == 60) || (md == -60))
      {
         md = 0;
         hd++;
      }

      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss);
      this->tick_text[rr] = tmpval;
      if (swapsign)  this->tick_text[rr][0] = '-';
   }
   return ERR_STATUS_GOOD;
}


bool ChipsWCSTANAxisAlgorithm::fix_minmax (double &minval, double &maxval, bool transformed)
{
   ChipsVector min_coord (0,0);
   ChipsVector max_coord (1,1);
   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   bool reversed = false;
   bool changed = false;

   if (dsCoord == NULL)
      return false;

   string tmp_id = dsCoord->get_id_no_wcs ();

   this->get_axis_position (transformed, min_coord, max_coord);

   min_coord[index] = minval;
   max_coord[index] = maxval;

   if (transformed)
   {
      // convert the coordinate to image space
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               DATA, "", DATA, tmp_id);
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               DATA, "", DATA, tmp_id);
   }

   double tmp_minval = min_coord[index];
   double tmp_maxval = max_coord[index];

   if (tmp_minval > tmp_maxval)
   {
      // adjust to ensure min is less than max
      double tmp = tmp_maxval;
      tmp_maxval = tmp_minval;
      tmp_minval = tmp;

      reversed = true;
      changed = true;
   }
   else
   {
      reversed = false;
      if (tmp_minval == tmp_maxval && tmp_minval == this->padded_min && tmp_maxval == this->padded_max)
      {
         // adjust to ensure range is not 1 point
         if (tmp_minval != 0)
         {
            tmp_minval *= 0.95;
            tmp_maxval *= 1.05;
         }
         else
         {
            tmp_minval = -1.0;
            tmp_maxval = 1.0;
         }
         changed = true;
      }
   }

   if (changed)
   {
      min_coord[index] = tmp_minval;
      max_coord[index] = tmp_maxval;

      if (transformed)
      {
         // convert back to world space
         dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               DATA, tmp_id, DATA, "");
         dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               DATA, tmp_id, DATA, "");
      }

      minval = min_coord[index];
      maxval = max_coord[index];
   }

   return reversed;
}


void ChipsWCSTANAxisAlgorithm::get_axis_position (bool use_transform, ChipsVector &min_coord, ChipsVector &max_coord)
{
   int index = (this->orientation == RA_AXIS) ? 0 : 1;
   int opp_index = (index == 0) ? 1 : 0;

   min_coord = ChipsVector (0,0);
   max_coord = ChipsVector (1,1);

   ChipsDataSystem *dsCoord = this->axis->get_data_system ();

   if (dsCoord == NULL)
      return;

   string tmp_id = "";

   if (!use_transform)
      tmp_id = dsCoord->get_id_no_wcs ();


   // convert from frame to plot to get the correct position of the plot
   min_coord[opp_index] = this->axis->get_norm_pos ();
   max_coord[opp_index] = this->axis->get_norm_pos ();

   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               FRAME_NORM, "", PLOT_NORM, "");
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               FRAME_NORM, "", PLOT_NORM, "");

   // get the data coordinates from the plot
   min_coord[index] = 0;
   max_coord[index] = 1;
   dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1,
                               PLOT_NORM, "", DATA, tmp_id);
   dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1,
                               PLOT_NORM, "", DATA, tmp_id);
}


int ChipsWCSTANAxisAlgorithm::check_poll ()
{
   int rval = 0;
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   ChipsVector pos_poll(0,90);
   ChipsVector neg_poll(0,-90);
   ChipsVector img_poll(0,0);
   ChipsVector lower_bound = dsCoord->get_lower_bounds ();
   ChipsVector upper_bound = dsCoord->get_upper_bounds ();


   if (this->validate_coordinates (pos_poll, dsCoord->get_id ()))
   {
      dsCoord->convert_cs_to_cs (&pos_poll, &img_poll, 1, DATA, "", DATA, dsCoord->get_id_no_wcs ());


      // check if the image coordinate is in the plot area
      if ((lower_bound[0] <= img_poll[0]) && (img_poll[0] <= upper_bound[0]) &&
          (lower_bound[1] <= img_poll[1]) && (img_poll[1] <= upper_bound[1]))
         rval = 1;
   }
   else if (this->validate_coordinates (neg_poll, dsCoord->get_id ()))
   {

      // check the other poll
      dsCoord->convert_cs_to_cs (&neg_poll, &img_poll, 1, DATA, "", DATA, dsCoord->get_id_no_wcs ());

      // check if the image coordinate is in the plot area
      if ((lower_bound[0] <= img_poll[0]) && (img_poll[0] <= upper_bound[0]) &&
          (lower_bound[1] <= img_poll[1]) && (img_poll[1] <= upper_bound[1]))
            rval = -1;
   }

   return rval;  
}

double ChipsWCSTANAxisAlgorithm::get_ra_intersection (double alpha, double image_point, int orientation, Transform *transform)
{
   double out_image_point;

   double *crpix = static_cast <double *>(transform->get_parameter_value ("CRPIX"));
   double *crval = static_cast <double *>(transform->get_parameter_value ("CRVAL"));
   double *cdelt = static_cast <double *>(transform->get_parameter_value ("CDELT"));
   double *crota = static_cast <double *>(transform->get_parameter_value ("CROTA"));
   double rho = 0;

   if (crota)
      rho = (*crota) * V_PI/180.;

   ChipsMatrix m(3,3);
   ChipsVector pts0(3);
   ChipsVector pts1(3);
   ChipsVector proj_pts(3);

   // create a CD matrix as specified by fits standards
   m.identity ();
   m[0][0] = cdelt[0]*cos(rho);
   m[0][1] = copysignf (cdelt[1],cdelt[0])*sin(rho);
   m[1][0] = -copysignf (cdelt[0],cdelt[1])*sin(rho);
   m[1][1] = cdelt[1]*cos(rho);


   if (orientation != RA_AXIS)
   {
      pts1[0] = pts0[0] = image_point-crpix[0];
      pts0[1] = -crpix[1];
      pts1[1] = 10-crpix[1];
   }
   else
   {
      pts0[0] = -crpix[0];
      pts1[1] = pts0[1] = image_point-crpix[1];
      pts0[0] = 10-crpix[0];
   }
   
   double xp0, xp1, yp0, yp1;
   double t = 0;

   double d_alphas = (alpha - crval[0])*V_PI/180.;
   double delta_p = crval[1]*V_PI/180.;
   double sin_delta_p = sin (delta_p);
   double cos_delta_p = cos (delta_p);

   double cos_phi_p;

   proj_pts = m*pts0;
   xp0 = proj_pts[0];
   yp0 = proj_pts[1];

   proj_pts = m*pts1;
   xp1 = proj_pts[0];
   yp1 = proj_pts[1];


   if (crval[0] == 90)
      cos_phi_p = 1;
   else
      cos_phi_p = -1;

   // Because the project point might be rotated first find the
   // intersection of the ra/dec coordinate with the parametric line
   // at time t. From t figure out the actual projected point
   if (orientation == RA_AXIS)
   {
//      xp = (((1 - (sin_delta_p*sin_delta_p)) * 180 / (cos_delta_p*V_PI)) + yp * (sin_delta_p * cos_phi_p + (1./tan(d_alphas))*sin_phi_p)) / (sin_delta_p*sin_phi_p - (1./tan(d_alphas))*cos_phi_p);

      // Take the above equation and divide into sub parts so that its in
      // the form of
      //     xp(t) = A + Byp(t)
      //       where:  xp(t) = xp(0) + (xp(1) - xp(0))
      //               yp(t) = yp(0) + (yp(1) - yp(0))
      double div = ((-1./tan(d_alphas))*cos_phi_p);
      double A = (1 - (sin_delta_p*sin_delta_p)) * 180 / (cos_delta_p*V_PI)/div;
      double B = (sin_delta_p * cos_phi_p)/div;

      t = (xp0 - A - B*yp0) / (B*(yp1-yp0) - xp1 + xp0);
   }
   else
   {
      // yp = ((xp * (sin_delta_p*sin_phi_p - (1./tan(d_alphas))*cos_phi_p)) - ((1 - sin_delta_p*sin_delta_p) * 180/ (cos_delta_p*V_PI))) / (sin_delta_p * cos_phi_p + (1./tan(d_alphas))*sin_phi_p)

      // Take the above equation and divide into sub parts so that its in
      // the form of
      //     yp(t) = Bxp(t) - A
      //       where:  xp(t) = xp(0) + (xp(1) - xp(0))
      //               yp(t) = yp(0) + (yp(1) - yp(0))


      double div = sin_delta_p * cos_phi_p;
      double A = ((1 - sin_delta_p*sin_delta_p) * 180/ (cos_delta_p*V_PI)) / div;
      double B = (-(1./tan(d_alphas))*cos_phi_p) / div;

      t = (yp0 + A - B*xp0) / (B*(xp1-xp0) - yp1 + yp0);
   }

   // solve xp and yp at t
   proj_pts = ChipsVector (xp0 + ((xp1-xp0) * t), yp0 + ((yp1-yp0) * t));

   // unproject the point
   m = m.inverse ();
   pts0 = m*proj_pts;


   out_image_point = 0;
   if (orientation == RA_AXIS)
      out_image_point = pts0[0] + crpix[0];
   else
      out_image_point = pts0[1] + crpix[1];

   return out_image_point;

}


int ChipsWCSTANAxisAlgorithm::get_dec_intersection (double delta, double image_point, int orientation, Transform *transform, double out_image_values[])
{

   double *crpix = static_cast <double *>(transform->get_parameter_value ("CRPIX"));
   double *crval = static_cast <double *>(transform->get_parameter_value ("CRVAL"));
   double *cdelt = static_cast <double *>(transform->get_parameter_value ("CDELT"));
   double *crota = static_cast <double *>(transform->get_parameter_value ("CROTA"));
   double rho = 0;

   double delta_p = crval[1]*V_PI/180.;
   double sin_delta_p = sin (delta_p);
   double cos_delta_p = cos (delta_p);

   double cos_phi_p;
//   double sin_phi_p = 0;

   double A1,A2,B,C2;
   double a,b,c;
   double d,e1,f1,f2,f3;

   double xp0, xp1, yp0, yp1;

   const double DEGREES = 180./V_PI;
   const double RADIANS = V_PI/180.;


   if (crota)
      rho = (*crota) * RADIANS;

   ChipsMatrix m(3,3);
   ChipsVector pts(3);
   ChipsVector pts0(3);
   ChipsVector pts1(3);
   ChipsVector proj_pts(3);

   // create a CD matrix as specified by fits standards
   m.identity ();
   m[0][0] = cdelt[0]*cos(rho);
   m[0][1] = copysignf (cdelt[1],cdelt[0])*sin(rho);
   m[1][0] = -copysignf (cdelt[0],cdelt[1])*sin(rho);
   m[1][1] = cdelt[1]*cos(rho);


   if (orientation != RA_AXIS)
   {
      pts1[0] = pts0[0] = image_point-crpix[0];
      pts0[1] = -crpix[1];
      pts1[1] = 10-crpix[1];
   }
   else
   {
      pts0[0] = -crpix[0];
      pts1[1] = pts0[1] = image_point-crpix[1];
      pts0[0] = 10-crpix[0];
   }

   proj_pts = m*pts0;
   xp0 = proj_pts[0];
   yp0 = proj_pts[1];

   proj_pts = m*pts1;
   xp1 = proj_pts[0];
   yp1 = proj_pts[1];

   if (crval[0] == 90)
      cos_phi_p = 1;
   else
      cos_phi_p = -1;

   // Because the project point might be rotated first find the
   // intersection of the ra/dec coordinate with the parametric line
   // at time t. From t figure out the actual projected point

   if (orientation != RA_AXIS)
   {
/*
      A = (1 - cos_delta_p*cos_delta_p)*180/V_PI; // + xp*cos_delta_p*sin_delta_p*sin_phi_p;
      B = cos_delta_p*sin_delta_p*cos_phi_p;
      C = sin(delta*V_PI/180.)*sin_delta_p;

      a = (B*B - C*C);
      b = (-2*B*A);
      c = (A*A - ((C*C)*((180/V_PI)*(180/V_PI) + xp*xp)));
*/

      // Take the above equations and add in xp(t) in the form of x0 + (x1-x0)*t
      // and yp(t) in the same form. Then solve for t.

      A1 = (1 - cos_delta_p*cos_delta_p)*DEGREES;
//      A2 = cos_delta_p*sin_delta_p*sin_phi_p;      <--- sin_phi_p = 0
      B = cos_delta_p*sin_delta_p*cos_phi_p;
      C2 = sin(delta*RADIANS)*sin_delta_p;
      C2 *= C2; // C^2

      // here d,e,f replace a,b,c above ... this is to keep the descrimate
      // variables seperate. 
      d = B*B - C2;
      // e = e1 + e2*t2 -- but e2 is zero
      e1 = -2*B*A1;
      // f = f1 + f2*t f3*t^2
      f1 = (A1*A1) - C2*DEGREES*DEGREES - C2*xp0*xp0;
      f2 = -2*C2*(xp1-xp0)*xp0;
      f3 = -C2*((xp1-xp0)*(xp1-xp0));

      // 0 = a*t^2 + b*t + c
      a = d*((yp1-yp0)*(yp1-yp0)) + f3;
      b = (2*d*yp0+e1)*(yp1-yp0) + f2;
      c = d*yp0*yp0 + e1*yp0 + f1;
   }
   else
   {
/*
      A = (1 - cos_delta_p*cos_delta_p)*180/V_PI - (yp*cos_delta_p*sin_delta_p*cos_phi_p);
      B = 0; //-cos_delta_p*sin_delta_p*sin_phi_p
      C = sin(delta*V_PI/180.)*sin_delta_p;

      a = -(C*C); //(B**2 - C**2)
      b = 0; //(-2*B*A)
      c = (A*A - ((C*C)*((180/V_PI)*(180/V_PI) + yp*yp)));
*/

      // Take the above equations and add in xp(t) in the form of x0 + (x1-x0)*t
      // and yp(t) in the same form. Then solve for t.

      A1 = (1 - cos_delta_p*cos_delta_p)*DEGREES - yp0*cos_delta_p*sin_delta_p*cos_phi_p;
      A2 = -(yp1-yp0)*cos_delta_p*sin_delta_p*cos_phi_p;
      // A = A1 + A2*t
//      B = -cos_delta_p*sin_delta_p*sin_phi_p         <--- sin_phi_p = 0
      C2 = sin(delta*RADIANS)*sin_delta_p;
      C2 *= C2;  // C^2

      d = -C2;

      // e = 0 because B is zero

      // f = (A1+A2*t)**2 - (C**2)*((180/math.pi)**2 + (yp0 + (yp1-yp0)t)**2)
      //   = A1**2 + 2*A1*A2*t + (A2**2)*t**2 - (C**2)*(180/math.pi)**2 - 
      //     (C**2)*(yp0**2 + 2*yp0*(yp1- yp0)*t + ((yp1-yp0)**2)*t**2
         
      f1 = A1*A1 - C2*DEGREES*DEGREES - C2*yp0*yp0;
      f2 = 2*A1*A2 - C2*2*yp0*(yp1-yp0);
      f3 = A2*A2 - C2*(yp1-yp0)*(yp1-yp0);
      // f = f1 + f2t +f3t**2

      // descriminate values
      // 0 = dx^2 + ex + f
      // d * xp0**2 + 2*d*xp0*(xp1-xp0))*t + d*((xp1-xp0)**2)*t**2 + e1*xp0 + e1*(xp1-xp0)*t + e2*xp0*t + e2*(xp1-xp0)*t**2 + f1 + f2t +f3t**2

      a = d*(xp1-xp0)*(xp1-xp0) + f3;
      b = 2*d*xp0*(xp1-xp0) + f2;
      c = d * xp0*xp0 + f1;
   }

   double discrim = b*b - 4*a*c;
   double pt[2];
   int num_points = 0;

   if (dbl_equal (discrim, 0))
   {
      pt[0] = -b/(2*a);
      num_points = 1;
   }
   else if (dbl_equal (a, 0))
   {
      pt[0] = -c/b;
      num_points = 1;
   }
   else if (discrim > 0)
   {
      pt[0] = (-b + sqrt (discrim))/(2*a);
      pt[1] = (-b - sqrt (discrim))/(2*a);
      num_points = 2;
   }

   m = m.inverse ();
   for (int ii=0; ii<num_points; ii++)
   {
      proj_pts = ChipsVector (xp0+(xp1-xp0)*pt[ii], yp0+(yp1-yp0)*pt[ii]);
      pts = m*proj_pts;

      if (orientation != RA_AXIS)
         out_image_values[ii] = pts[1] + crpix[1];
      else
         out_image_values[ii] = pts[0] + crpix[0];
   }
   return num_points;

}

void ChipsWCSTANAxisAlgorithm::transform_tick_limits ()
{
   ChipsDataSystem *dsCoord = this->axis->get_data_system ();
   int index = (this->orientation == RA_AXIS) ? 0 : 1;

   if (dsCoord == NULL)
      return;

   string tmp_id = dsCoord->get_id_no_wcs ();
   ChipsVector min_coord;
   ChipsVector max_coord;

   if (transform_valid && (max_set & 4))
   {
      max_coord = dsCoord->get_upper_bounds ("");
      max_coord[index] = this->transformed_padded_max;
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, "", DATA, tmp_id);
      this->padded_max = max_coord[index];
   }
   else if (this->max_set & 4)
   {
      max_coord = dsCoord->get_upper_bounds ();
      max_coord[index] = this->padded_max;
      dsCoord->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_padded_max = max_coord[index];
   }

   if (transform_valid && (min_set & 4))
   {
      min_coord = dsCoord->get_lower_bounds ("");
      min_coord[index] = this->transformed_padded_min;
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, "", DATA, tmp_id);
      this->padded_min = min_coord[index];
   }
   else if (this->min_set & 4)
   {
      min_coord = dsCoord->get_lower_bounds ();
      min_coord[index] = this->padded_min;
      dsCoord->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, tmp_id, DATA, "");
      this->transformed_padded_min = min_coord[index];
   }

}

bool ChipsWCSTANAxisAlgorithm::validate_coordinates (const ChipsVector &coord, const string &ds_id)
{
   ChipsDataSystem *ds = this->axis->get_data_system (ds_id);
   ChipsVector tmp_coord (coord);
   ChipsVector comp_coord (coord);
   string img_id = ds->get_id_no_wcs ();

   bool valid = false;

   ds->convert_cs_to_cs (&coord, &tmp_coord, 1, DATA, ds_id, DATA, img_id);
   ds->convert_cs_to_cs (&tmp_coord, &comp_coord, 1, DATA, img_id, DATA, ds_id);

   // test if the y coordinate is the same

   // the original and comparison coordinate are polls than the
   // only validation of the xaxis is that it's between 0 and 360
   if (flt_equal (comp_coord[1], coord[1]))
   {

      if ((flt_equal (coord[1], 90)) || (flt_equal (coord[1], 90)))
      {
         // the original and comparison coordinate are polls than the
         // only validation of the xaxis is that it's between 0 and 360
         if ((comp_coord[0] >= 0) && (comp_coord[0] < 360))
            valid = true;
      }
      else
      {
         // handle wrap around on the xaxis
         if ((flt_equal (coord[0],0) || flt_equal (coord[0],360)) &&
             (flt_equal (comp_coord[0],0) || flt_equal (comp_coord[0],360)))
            valid = true;
         else if (flt_equal (comp_coord[0], coord[0]))
            valid = true;
      }

   }

   return valid;
}

double* ChipsWCSTANAxisAlgorithm::get_tick_positions (bool transformed)
{
   if (transformed)
      return this->transformed_tickpos;

   return this->tickpos;
}

void ChipsWCSTANAxisAlgorithm::format_point (const string &tickformat, double pnt, string &formatted_pnt)
{
   char tmp_pnt[80];
/*   // for arbitrary points return the interpolated enumerated value
   if (major_mode == chips_arbitrary)
   {
      sprintf (tmp_pnt, "%f", pnt);
      formatted_pnt = tmp_pnt;
   }
*/

   // format z/Z for using 3.2 x10^2 instead of 3.2e+02 (%z=%e, %Z=%g)
   // don't support z or Z -- use alternative
   if (strchr(tickformat.c_str(), 'z'))
   {
      string tmp_format = tickformat;
      tmp_format[tickformat.length ()-1] = 'e';
      sprintf (tmp_pnt, tmp_format.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
   else if (strchr(tickformat.c_str(), 'Z'))
   {
      string tmp_format = tickformat;
      tmp_format[tickformat.length ()-1] = 'g';
      sprintf (tmp_pnt, tmp_format.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
   else if (strstr(tickformat.c_str(),"ra") || strstr(tickformat.c_str(),"RA"))
   {

      string hmfmt;
      double tpos, hd, md, ss;
      char tmpfmt[80], tmpval[80], tmpspec[80];
      char *fmt = tmpspec;
      char* sixty_p;

      strcpy(tmpspec, tickformat.c_str());
      char* cpos = strstr(tmpspec, "ra");
      if (!cpos)   cpos = strstr(tmpspec, "RA");
      *cpos = '\0';
      if (tmpspec[0] == '%') fmt++;
      if (!strlen(fmt))  strcpy(fmt, "02.0");

      // always use : deliminated format
      hmfmt = "%02d:%02d:";
      sprintf(tmpfmt, "%%%sf", fmt);

      if ((tpos = fmod(pnt, 360.0)) < 0) tpos+=360;
      ss = tpos/15.0;
      ss = modf(ss, &hd);
      hd = fmod(hd, 24);       // wrap at 24 hours
      ss = modf(ss*60.0 + FLT_EPSILON, &md);
      ss = ss * 60 + 0.00137;

      // account for roundup to 60 seconds
      sprintf(tmpspec, tmpfmt, ss);
      if ((sixty_p = strstr(tmpspec,"60")) != NULL)
      {
         *sixty_p = '0';
         md++;
      }
      // account for roundup to 60 minutes;
      sprintf(tmpval,hmfmt.c_str(), (int)hd, (int)md);
      if ((sixty_p = strstr(tmpval,"60")) != NULL)
      {
         // account for 23h to 00h
         if (hd == 23)  sprintf(tmpval,hmfmt.c_str(), 0, 0);
            else        sprintf(tmpval,hmfmt.c_str(), (int)hd+1, 0);
      }
      formatted_pnt = tmpval;
      formatted_pnt.append(tmpspec);
   }
   else if (strstr(tickformat.c_str(),"dec") || strstr(tickformat.c_str(),"DEC"))
   {
      // using sexagesimal (DMS) format for declination (dec)
      double hd, md, ss;
      unsigned int swapsign;
      char tmpfmt[80], tmpval[80], tmpspec[80];
      char *fmt = tmpspec;

      strcpy(tmpspec, tickformat.c_str());
      char* cpos = strstr(tmpspec, "dec");
      if (!cpos)   cpos = strstr(tmpspec, "DEC");
      *cpos = '\0';
      if (tmpspec[0] == '%') fmt++;
      if (!strlen(fmt))  strcpy(fmt, "02.0");
/*      if (strstr(tickformat.c_str(), "dec1"))
         sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);
      else if (strstr(tickformat.c_str(), "dec2"))
         sprintf(tmpfmt, "%%+02d %%02d %%%sf", fmt);
      else if (strstr(tickformat.c_str(), "DEC3"))
         sprintf(tmpfmt, "%%+02dD %%02dM %%%sfS", fmt);
      else if (strstr(tickformat.c_str(), "dec3"))
         sprintf(tmpfmt, "%%+02dd %%02dm %%%sfs", fmt);
      else
         sprintf(tmpfmt, "%%+02d^o %%02d' %%%sf\"", fmt);
*/
      // always use : deliminated format
      sprintf(tmpfmt, "%%+02d:%%02d:%%%sf", fmt);


      ss = modf(pnt, &hd);
      if (ss >= 0)
         ss = modf(ss*60.0 + FLT_EPSILON, &md);
      else
         ss = modf(ss*60.0 - FLT_EPSILON, &md);
      ss *= 60;
      // handle degrees between -1:0 with a - prefix infront of degrees
      if ((swapsign = ((hd <= 0) && ((md < 0) || (ss < 0)))))
      {
         if (md < 0) md *= -1;
         if (ss < 0) ss *= -1;
      }
      // handle minutes field set to +/-60 due to precision issues
      if ((md == 60) || (md == -60))
      {
         md = 0;
         hd++;
      }

      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss);
      formatted_pnt = tmpval;
      if (swapsign)  formatted_pnt[0] = '-';
   }
   else
   {
      sprintf(tmp_pnt, tickformat.c_str (), pnt);
      formatted_pnt = tmp_pnt;
   }
}


void ChipsWCSTANAxisAlgorithm::resize_grid (double ****out_grid, int line_count, int slices, int &max_line_count)
{
   double ***grid = *out_grid;
   if (line_count >= max_line_count)
   {
      double ***resized_grid = new double**[max_line_count];
      for (int ii=0; ii<max_line_count; ii++)
      {
         resized_grid[ii] = new double*[slices];
         for (int jj=0; jj<slices; jj++)
         {
            resized_grid[ii][jj] = new double[2];
            resized_grid[ii][jj][0] = grid[ii][jj][0];
            resized_grid[ii][jj][1] = grid[ii][jj][1];
            delete []grid[ii][jj];
         }
         delete []grid[ii];
      }
      delete []grid;

      *out_grid = grid = new double**[max_line_count*2];
      for (int ii=0; ii<max_line_count*2; ii++)
      {
         grid[ii] = new double*[slices];
         for (int jj=0; jj<slices; jj++)
         {
            grid[ii][jj] = new double[2];
            if (ii < max_line_count)
            {
               grid[ii][jj][0] = resized_grid[ii][jj][0];
               grid[ii][jj][1] = resized_grid[ii][jj][1];
               delete []resized_grid[ii][jj];
            }
         }
         if (ii < max_line_count)
            delete []resized_grid[ii];
      }
      delete []resized_grid;
      max_line_count *= 2;
   }

}


bool ChipsWCSTANAxisAlgorithm::check_point (double *img_point, double comp_point, int index, Transform *transform)
{

   bool valid;
   double *tmp_point;
   float tol = 10e-6;

   tmp_point = (double *)transform->apply ((void *)(img_point),dmDOUBLE,1);
   if (flt_equal (comp_point, 0) || flt_equal (comp_point,360))
      valid = (approx_equal (tmp_point[index], 0, tol, tol) || approx_equal (tmp_point[index],360,tol,tol));
   else
      valid = approx_equal (tmp_point[index],comp_point,tol, tol);
   delete []tmp_point;

   return valid;
}

int ChipsWCSTANAxisAlgorithm::set_major_mode (V_TICKMODE_TYPE mode)
{
   if ((mode != chips_nice) && (mode != chips_limits))
   {
      string error = "WCS axes do not support the axis mode ";
      error += v_tickmode_to_string(mode);
      error += ". They currently only support limits and nice modes.";
      cmd_manager->add_err(error, this->axis->get_id (), 0);
      return ERR_STATUS_BAD;
   }

   if (this->major_mode != mode)
      this->modified=true;
   this->major_mode=mode;

   return ERR_STATUS_GOOD;
}

int ChipsWCSTANAxisAlgorithm::set_minor_mode (V_TICKMODE_TYPE mode)
{
   if ((mode != chips_nice) && (mode != chips_count))
   {
      string error = "WCS axes do not support the axis mode ";
      error += v_tickmode_to_string(mode);
      error += ". They currently only support count and nice modes.";
      cmd_manager->add_err(error, this->axis->get_id (), 0);
      return ERR_STATUS_BAD;
   }

   if (this->minor_mode != mode)
      this->modified=true;
   this->minor_mode=mode;

   return ERR_STATUS_GOOD;
}

