// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSWCSTANAXISALGORITHM_HH
#define CHIPSWCSTANAXISALGORITHM_HH

#include <string>
#include <stdlib.h>
#include <stdio.h>
#include "ChipsAxisAlgorithm.hh"

class ChipsAxis;
class Transform;

using namespace std;

class ChipsWCSTANAxisAlgorithm : public ChipsAxisAlgorithm
{
  public:

   ChipsWCSTANAxisAlgorithm(int orientation, ChipsAxis *axis);
   ChipsWCSTANAxisAlgorithm(ChipsWCSTANAxisAlgorithm &copy);
   virtual ~ChipsWCSTANAxisAlgorithm();

   virtual ChipsWCSTANAxisAlgorithm * clone ();


   virtual void       set_scalable(bool val);
   virtual void       set_scaletype (int val);

   virtual double     get_min (bool transformed=true);
   virtual double     get_max (bool transformed=true);

   virtual double     get_absolute_min (int scaletype=-1, bool transformed=true);
   virtual int        set_absolute_min (double val, int scaletype=-1, bool transformed=true);
   virtual double     get_absolute_max (int scaletype=-1, bool transformed=true);
   virtual int        set_absolute_max (double val, int scaletype=-1, bool transformed=true);
   virtual int        set_absolute_minmax (double min, double max, int scaletype=-1, bool transformed=true);

   virtual double*    get_tick_positions (bool transformed);
   virtual void       set_tick_positions (double *ticks, int count);
   virtual void       set_arbitrary_positions (double *positions, string *text, int count);

   virtual int        set_major_mode (V_TICKMODE_TYPE mode);
   virtual int        set_minor_mode (V_TICKMODE_TYPE mode);

   virtual int        calculate_tick_limits(bool automin, bool automax,
                                  double pad);
   virtual int        update_tick_positions(bool &out_updates);
   virtual void       reset_string_labels(const string& tickformat);

   virtual void       transform_tick_limits ();
   virtual bool       fix_minmax (double &minval, double &maxval, bool transformed);
   virtual bool       validate_coordinates (const ChipsVector &coord, const string &ds_id);

   virtual void       serialize (ChipsStateStream &out);
   virtual void       deserialize (ChipsStateStream &in);

   virtual void format_point (const string &tickformat, double pnt, string &formatted_pnt);

  protected:

   void       get_axis_position (bool transform, ChipsVector &min_coord, ChipsVector &max_coord);

   void       reset_adjusted_parameters ();

   int        sexagesimal_dec_ticks(const string& tickformat);
   int        sexagesimal_ra_ticks(const string& tickformat);

   int        count_tick_positions();
   int        limit_tick_positions();
   int        nice_tick_positions();
   int        arbitrary_tick_positions();
   int        interval_tick_positions();

   int        nice_tick_limits();
   int        count_tick_limits(bool automin, bool automax,
                                             double pad);
   int        limit_tick_limits(bool automin, bool automax, double pad);
   int        arbitrary_tick_limits(bool automin, bool automax,
                                             double pad);
   int        interval_tick_limits(bool automin, bool automax,
                                             double pad);

   int        count_minortick_positions();
   int        interval_minortick_positions();


   double *** calculate_grid (int &out_line_count, int &out_slices, int tick_type);

   static void scale1_sexg_dec(double inmin, double inmax, int slices,
                        double* outmin, double* outmax, double* dist, 
                        double *base);

   static void scale1_sexg_ra(double inmin, double inmax, int slices,
                       double* outmin, double* outmax, double* dist, 
                       double *base);

   static void deg_to_ra (double in_degree, double &out_hour,
                          double &out_minute, double &out_second, bool wrap_hour=true);

   static void deg_to_dec (double in_degree, double &out_degree,
                           double &out_minute, double &out_second);



              // nice tick methods (C.R. Lewart, CACM, Oct 73,v16 n10)
              // modified for degrees
   static void scale1(double min, double max, int n, double* outmin,
                     double* outmax, double* dist, double* base=NULL);
   static void scale2(double min, double max, int n, double* outmin,
                     double* outmax, double* dist);
   int       check_poll ();
   static double get_ra_intersection (double alpha, double img_point, int orientation, Transform *transform);
   static int get_dec_intersection (double delta, double img_point, int orientation, Transform *transform, double out_image_values[]);

  private:
   static void resize_grid (double ****grid, int line_count, int slices, int &max_line_count);
   static bool check_point (double *img_point, double comp_point, int index, Transform *transform);

  protected:


   double     absolute_min;  // current minimum range value 
   double     absolute_max;  // current maximum range value 
   double     padded_min;    // current minimum range value with padding
   double     padded_max;    // current maximum range value with padding

   double     transformed_padded_min;
   double     transformed_padded_max;
   double     transformed_absolute_min;
   double     transformed_absolute_max;

   bool       sexagesimal;   // are we displaying sexagesimal (hh:mm:ss)

   int        grid_slices;   // the number of slices for each grid line
   double     tick_spacing;  // the spacing between ticks in degrees
   double     grid_extent[2]; // the potential span of transformed tick position

   double     last_pad;
   bool       last_automin;
   bool       last_automax;
   double     last_axis_position;

   double    *transformed_tickpos; // transformed tick positions

   int        max_set;
   int        min_set;

   bool       grid_modified; // recalculate grid coordinates

   double     minortick_spacing; // spacing between minor ticks

  private:

     


};


#endif 
