// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSWINDOW_HH
#define CHIPSWINDOW_HH

#include <string>
#include <map>
#include <vector>

#include "ChipsObject.hh" 
#include "ChipsId.hh" 
#include "ChipsFrame.hh" 
#include "ChipsColor.hh" 
#include "ChipsCoordinate.hh"
#include "ChipsStateStream.hh"

#ifndef HEADLESS
#include "gtk/gtk.h"
#endif

#include "Transform.hh"

class ChipsEngine;
class ChipsPlot;
class ChipsGtkWindow;

using namespace std;

class ChipsWindow: public ChipsObject 
{
  public:

   ChipsWindow(ChipsEngine* eng, string id);
   ChipsWindow(ChipsEngine* eng, ChipsStateStream& in);
   ChipsWindow(ChipsEngine* eng, float width, float height, string format="", 
              string id=""); 

   virtual    void object_delete ();

   void       initialize(bool create_coord=true);  // set defaults 

   // methods for window titlebar interactions 
   string     get_title()  { return title; } 
   int        set_title(string val);
   int        set_title_suffix (string val);

   // methods for window dimensions 
   int        get_width()  { return width; } 
   int        get_height()  { return height; } 

   // methods for frame interactions 
   int        add_frame(ChipsId* cid);
   int        add_frame(ChipsId* cid, float xmin, float ymin, float xmax,
                        float ymax);
   int        delete_frame(ChipsId* cid, bool force_del=false);
   int        reposition_frame(const ChipsId* cid, float x1, float y1, 
                 float x2, float y2);
   int        blink_frames(ChipsId* cid, string fid2, float interval=3.0);
   int        stop_blink(ChipsId* cid);

   int        display_depth(ChipsId* cid, int depth);
   int        hide_depth(ChipsId* cid, int depth);
   int        collapse_depths(ChipsId* cid, int* sources, int srccnt, int dest);
   int        rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner);

   string     get_current()  { return current; }
   int        set_current(const string &val); 
   int        set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec=true);
   int        prev_frame(ChipsId* cid, bool rec=true);
   int        next_frame(ChipsId* cid, bool rec=true);

   // layout/grid methods
   void       show_layout();
   void       hide_layout();
   int        get_framemode()  { return framemode; }
   int        set_framemode(int mode, bool rec=true); 
   int        set_layout_color(string in_color);
   string     get_layout_color()   { return layout_color.get_name(); }
   int        set_layout_thickness(float val);
   float      get_layout_thickness()   { return layout_thickness; }
   int        set_layout_linestyle(V_LINE_TYPE val);
   V_LINE_TYPE get_layout_linestyle() { return layout_linestyle; }

   int        display_grid(int val); 
   int        hide_grid(); 
   bool       grid_visible()   { return vis_grid; } 
   int        set_grid_color(string in_color);
   string     get_grid_color()     { return grid_color.get_name(); }
   int        set_grid_thickness(float val);
   float      get_grid_thickness()  { return grid_thickness; }
   int        set_grid_linestyle(V_LINE_TYPE val);
   V_LINE_TYPE get_grid_linestyle() { return grid_linestyle; }
   int        set_grid_scale(int in_scale);
   int        get_grid_scale()      { return gscale; }


   void       clear(); 

   int        highlight_object(ChipsId* cid, V_OBJECT_TYPE obj, bool vis);
   int        move_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj,
                          double xx, double yy, int mode);
   int        resize_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *bounds);
   int        rotate_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle);
   int        select_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected);
   int        get_annotation_position (const ChipsId* cid, V_OBJECT_TYPE obj, ChipsVector &pos);
   int        change_annotation_coordinate (const ChipsId* cid, V_OBJECT_TYPE obj, COORD_SYS coordsys, const string& coordid);


   int        get_object_count(ChipsId* cid, V_OBJECT_TYPE obj, int &out_val);
   int        get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj, 
                                   COORD_SYS& out_val, string& out_id);
   int        get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num);
   int        set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj);
   int        set_property(ChipsId* cid, void* in_val, int in_property, 
                           bool rec=true);
   int        get_property(ChipsId* cid, void* out_val, int out_property);

   int        adjust_grid_xgap(ChipsId* cid, float xg); 
   int        adjust_grid_ygap(ChipsId* cid, float yg); 
   int        adjust_grid_xrelsize(ChipsId* cid, int col, float ratio); 
   int        adjust_grid_yrelsize(ChipsId* cid, int row, float ratio); 
   int        adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* ratios); 
   int        adjust_grid_yrelsizes(ChipsId* cid, int cnt, float* ratios); 
   int        adjust_grid_gaps(ChipsId* cid, float xg, float yg); 
   int        grid_objects(ChipsId* cid,int cols,int rows, float xg, float yg,
                           int usage = 0, int lcnt=0, string* lids = NULL); 
   int        grid_objects(ChipsId* cid,int orient, int cnt, int* vals,
                           float xg, float yg,int usage=0, int lcnt=0,
                           string* lids=NULL); 
   int        swap_object_positions(ChipsId* cid, string id2);

   // plot operations 
   int        add_plot(ChipsId* cid);
   int        add_plot(ChipsId* cid, float x1, float y1, float x2, float y2);
   int        delete_plot(ChipsId* cid);
   int        clear_plot(ChipsId* cid);
   int        reposition_plot(ChipsId* cid, float x1, float y1, 
                 float x2, float y2);
   int        limits(ChipsId* cid, int dim, double xmin, double xmax, 
                     double ymin, double ymax);
   int        limits(ChipsId* cid, int dim, double min, double max);
   int        limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                     V_AUTOLIMIT_TYPE automax);
   int        limits(ChipsId* cid, int dim, double min,
                     V_AUTOLIMIT_TYPE automax);
   int        limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
                     double max);
   int        zoom (ChipsId* cid, float scale);
   int        panto (ChipsId* cid, double x, double y);
   int        pan  (ChipsId* cid, double x1, double y1, double x2, double y2);

   // curve operations
   int        add_curve(ChipsId* cid, long cnt, double* x, double* y,
                        double* ydown, double* yup, double* xdown, double* xup,
                        double* ylims, double* xlims, Transform *transform);
   int        delete_curve(ChipsId* cid);

   // contour operations
   int        add_contour(ChipsId* cid, long cnt, double* x, double* y,
                        double* z, unsigned int numlev, double* levels, 
                        Transform *trans);
   int        add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
                         unsigned int numlev, double* levels, Transform *trans); 
   int        delete_contour(ChipsId* cid);
   int        set_contour_levels(ChipsId* cid, unsigned int arraysize,
                                 double* pos);
   int        get_contour_levels(ChipsId* cid, vector<double> &pos);

   // histogram operations
   int        add_histogram(ChipsId* cid, long cnt, 
                            double* bin_low, double* bin_high, double* counts,
                            double* yd = NULL, double* yu = NULL);
   int        delete_histogram(ChipsId* cid);

   // axis  operations
   int        add_axis(ChipsId* cid, int dim, float pos, float minx, float maxx, float miny, float maxy, Transform *transform);
   int        delete_axis(ChipsId* cid);
   int        set_arbitrary_tick_positions(ChipsId* cid, unsigned int arraysize,
                               double* pos, string* vals=NULL);
   int        bind_axes(ChipsId*cid, const string lid, const string aid);
   int        unbind_axes(ChipsId*cid, const string lid, const string aid);
   int        format_axis_point (const ChipsId* cid, double pnt, string &formatted_pnt);


    // image operations
   int        add_image(ChipsId* cid, string filename, Transform *transform);
   int        add_image(ChipsId* cid, string filename,
                        float ix1, float iy1, float ix2, float iy2,
                        Transform *transform);
   int        add_image(ChipsId* cid, double* rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, Transform *transform);
   int        delete_image(ChipsId* cid);
   int        set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel);
   int        remove_image_channel(ChipsId* cid, int channel);
   int        print_image (ChipsId *cid, const string &outfile);

   int        add_colorbar(ChipsId* cid, double xpos, double ypos,
                            string obj_id="", V_OBJECT_TYPE obj=chips_not_object);
   int        delete_colorbar(ChipsId* cid);

   // label object operations
   int        add_label(ChipsId* cid, double xpos, double ypos, string text);
   int        delete_label(ChipsId* cid);
   int        add_line(ChipsId* cid, double startx, double starty, double endx,
			double endy);
   int        add_line(ChipsId* cid, double* xx, double* yy,unsigned int count);
   int        delete_line(ChipsId* cid);

   int        add_point(ChipsId* cid, double xx, double yy);
   int        add_point(ChipsId* cid, double xx, double yy, double zz);
   int        delete_point(ChipsId* cid);

   // legend operations
   int        add_legend(ChipsId* cid, const char** curve_ids,
                         const char** curve_labels, string lgid = "");
   int        delete_legend(ChipsId* cid);

   // region operations
   int        add_region(ChipsId* cid, double* xx, double* yy,
                         unsigned int count);
   int        add_region(ChipsId* cid, int sides, double xx, double yy,
                         double radius);
   int        delete_region(ChipsId* cid);

   // coordinate operations
   int     convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, ChipsId *cid2, int axis);
   int    get_current_coord (ChipsId *cid);
   int    get_coord_object_count(ChipsId *cid, int objtype = chips_not_object);
   bool   check_coordinate_exists (ChipsId *cid);

   // methods for window state 
   void       iconify();
   void       deiconify();
   bool       iconic(); 

   char*      info(ChipsId* cid, V_OBJECT_TYPE obj=chips_frame,
                   bool descend=true, bool current_obj=false, bool full=true);
   char*      info_depth(ChipsId* cid, int depth=0);
   char*      info_coordinate (ChipsId* cid);
   char*      info_bound_axes(ChipsId* cid);
   char*      list_current();
   bool       is_current ();

   void       activate(); 

   void       update_display(); 
 
   int        tile(unsigned int cols, unsigned int rows, bool rec=true); 
   int        row_tile(unsigned int rows, bool rec=true); 
   int        col_tile(unsigned int cols, bool rec=true); 
   void       set_col_row_vals(unsigned int rows, unsigned int cols)
              { tile_rows = rows; tile_cols =cols; }

   int        single_frame(string fid, bool rec=true);
   int        arrange_frames(int cnt, string* fids); 
   int        get_order(list<string> &out_order); 

   string     get_color()    { return color.get_name(); }
   int        set_color(string in_color);
   int        adjust_default_colors(const string newbg, const string newfg, 
                                    string& oldbg, string& oldfg);
   int        set_default_colors_ps(string newbg = "", string newfg = "");
   int        set_default_colors_display(string newbg = "", string newfg = "");

   int        update_print_scale (double scale); 

   void       toggle_blink_frame();

   void       set_redraw_ready();
   void       set_redraw(bool val);
   bool       get_redraw()           { return redraw_on; }

   int        set_export_orient(int val);
   int        get_export_orient()    { return export_orient; } 
   int        set_export_font(bool val);
   int        get_export_font()      { return export_font; } 
   int        set_export_clobber(bool val);
   int        get_export_clobber()      { return export_clobber; } 

   // page setup
   int         set_page_lmargin(float margin_sz);
   float       get_page_lmargin()    {  return  leftmargin; }
   int         set_page_rmargin(float margin_sz);
   float       get_page_rmargin()    {  return  rightmargin; }
   int         set_page_tmargin(float margin_sz);
   float       get_page_tmargin()    {  return  topmargin; }
   int         set_page_bmargin(float margin_sz);
   float       get_page_bmargin()    {  return  bottommargin; }
   int         set_page_width(float page_sz);
   float       get_page_width()      {  return  page_width; }
   int         set_page_height(float page_sz);
   float       get_page_height()     {  return  page_height; }
   int         set_page_size(string page_sz);
   string      get_page_size(); 
   int         set_page_scale_width(float scale_sz);
   float       get_page_scale_width() {  return  page_scale_width; }
   int         set_page_scale_height(float scale_sz);
   float       get_page_scale_height() {  return  page_scale_height; }
   int	       set_page_units (string val);
   string      get_page_units ();
   int         set_fit_to_page(bool flag);
   bool        get_fit_to_page()     {  return  fit_to_page; }
   int         set_printmethod(const string &method);
   string      get_printmethod()     {  return  printmethod; }
   int         set_page_keep_aspect(bool flag);
   bool        get_page_keep_aspect() {  return  keep_aspect; }
   int         set_export_dpi (int dpi);
   int         get_export_dpi () { return export_dpi; }

   ChipsEngine*  get_engine()        { return eng; }
   ChipsObject*  get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                            bool include_deref=false);


   int        print(string outfile, string format, V_PRINTFILE_TYPE scheme, string* wrotename = NULL);
   bool	      get_print_mode () { return print_mode; }
   bool       get_drawready () {return drawready; }

   // virtual functions 
   virtual int do_print(string outfile, int fmt=0, int colorSpace=6, bool add_ext=true); 
   virtual int redraw(); 
   virtual int render_layout(); 
   virtual int change_layout_color(); 
   virtual int change_layout_thickness(); 
   virtual int change_layout_linestyle(); 
   virtual int render_grids(); 
   virtual int change_grid_visible();
   virtual int change_grid_color(); 
   virtual int change_grid_thickness(); 
   virtual int change_grid_linestyle(); 
   virtual int change_order();
   virtual int change_color();
   virtual int change_size ();
   virtual int change_redraw ();
   virtual int show ();
   virtual int hide ();
   virtual int re_order_frames ();
   virtual void remap_hide ();
   virtual void remap_show ();

 
   virtual int reference_object (ChipsObject *co, bool change_cnt, const string &curr="");
   virtual int dereference_object (ChipsObject *co, bool change_cnt, const string &curr="");
   virtual int clear_cmd (ChipsObject *co);


   virtual ChipsFrame* new_frame(string fid="");
   virtual ChipsFrame* new_frame(ChipsStateStream& in);
   virtual ChipsFrame* new_frame(float xmin, float ymin, float xmax,
      float ymax, string fid);

   virtual void serialize(ChipsStateStream& out);
   virtual void serialize_ascii(ChipsStateStream& out, int keep=1);
   void deserialize(ChipsStateStream& in);
   int  deserialize_frames(ChipsStateStream& in);
   virtual int serialize_object(ChipsObject *co, ChipsStateStream& out);
   virtual ChipsObject *deserialize_object (V_OBJECT_TYPE co_type, 
                                                   ChipsStateStream& in);

   int update_all ();

   bool       redraw_blocked();

   void       set_window_num (int val) {window_num = val;};
   int        get_window_num () {return window_num;};

   virtual void disable_all_interaction ();
   virtual void enable_all_interaction ();


  /************************************************
   *   Temporary Variables for testing !
   ************************************************/

   char* create_frame_id();

#ifndef HEADLESS
   static int  blink(gpointer data);
#endif


  // set the gui class to be a friend
  friend class ChipsGtkWindow;

  protected:

   virtual ~ChipsWindow();

#ifndef HEADLESS
   virtual ChipsGtkWindow *new_gui_window (string geom, bool focus);
#endif

   int get_frames (const ChipsId *cid, list <ChipsFrame *>&out_frames, bool allow_all);

   void    disable_frame_interaction (const string &frame_id);
   void    enable_frame_interaction (const string &frame_id);

   void    insert_frame (ChipsFrame *cf);
   void    erase_frame (const string &fid);

   void    update_grid_properties(); 


   ChipsEngine* eng; 
   ChipsFrame* get_frame(string id="");

   void       initialize_gui(string geom="");
   int        store_event(double xval, double yval, const string &xstr,
                           const string &ystr, int evt, char key=0);

   virtual void       set_window_size();
   virtual void       set_window_size(int inPixelWidth,int inPixelHeight);
   virtual float      conversion_factor(V_UNITS_TYPE &val);

   bool       activated;   // The window has been activated
   float      raw_width;  // window width 
   float      raw_height; // window height
   int        width;      // window width in pixels 
   int        height;     // window height in pixels
   float      scale_width;  // scales the window height to fit the monitor resolution
   float      scale_height; // scales the window width to fit the monitor resolution
   map<string, ChipsFrame*>frame; // list of frames associated with the window
   map<unsigned int, ChipsFrame*>dframe; // list of dereferenced frames
   list<ChipsFrame*>frame_order; // list of frame orders
   string     prefix;     // window title prefix 
   string     title;      // title for window pane
   V_UNITS_TYPE units;    // units for window dimensions 
   string     current;    // current frame(s) - id or 'all'

   int        framemode;  // placed = 0, single = 1, tiled = 2
   int        tile_rows;  // number of rows for tiling 
   int        tile_cols;  // number of cols for tiling 

   bool       vis_layout; // layout is visible?
   ChipsColor layout_color; // color of frame border (layout) lines 
   V_LINE_TYPE layout_linestyle;   // type of line
   float      layout_thickness; // thickness of frame border (layout) lines

   // grids 
   int        gscale;     // scale type for grid  
   bool       vis_grid;   // grid is visible?
   bool       focus;      // should new chips window grap focus
   ChipsColor grid_color; // color of grid lines 
   V_LINE_TYPE grid_linestyle;    // type of line
   float      grid_thickness; // thickness of grid 
   bool       snaptogrid;

   ChipsColor color;      // background color of window 

   // blink support
   int        duration;   // blink time interval in milliseconds
   int        prev_mode;  // previous frame mode
   int        timer;      // id of gtk timer 
   int        blinkframe; // frame to blink: frame 1 or 2 of blink
   string     lastframe;  // 'current' frame before the blink operation 
   string     frame_1;    // first frame to blink  
   string     frame_2;    // second frame to blink  
   int        frame_cnt;  // reference count for frame id creation
   int        frame_ord;  // frame position reference
   bool       redraw_on;  // is redraw enabled/disabled 
   bool       drawready;  // window ready for redraw 
   bool       smoothing;  // window smoothing enabled
   bool       print_mode; // window is now setting up for printing
   bool       export_font; // export font to output ps file 
   bool       export_clobber;  // overwrite existing file?
   int        export_orient; // portrait == 0, landscape == 1
   V_PRINTFILE_TYPE export_scheme; // export color scheme
   string     export_format; 
   int        export_dpi;  // export dot per inch
   // page info
   float      leftmargin;
   float      rightmargin;
   float      topmargin;
   float      bottommargin;
   float      page_width;
   float      page_height;
   V_UNITS_TYPE page_units;
   float      page_scale_width;
   float      page_scale_height;
   bool       fit_to_page;
   bool       keep_aspect;
   string     printmethod;

   // default color schemes
   string     fg_display; // foreground display color 
   string     bg_display; // background display color 

   int window_num;        // this window's order 0th, 1st, 2nd, etc.

#ifndef HEADLESS
   ChipsGtkWindow *gui_window; 
#endif
};

#endif 
