// 
//  Copyright (C) 2006,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsWindowSystem.hh"
#include <algorithm>
#include <sstream>

unsigned long ChipsWindowSystem::global_ref_id = 0;

ChipsWindowSystem::ChipsWindowSystem (ChipsEngine *en):ChipsCoordinate (NULL)
{
   this->eng = en;
   pix_width  = 1;
   pix_height = 1;
   scale_width = 1;
   scale_height = 1;
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
}

ChipsWindowSystem::ChipsWindowSystem (ChipsEngine *en, int wdth, int hght, double scale_wdth, double scale_hght):ChipsCoordinate (NULL)
{
   this->eng = en;
   pix_width  = wdth;
   pix_height = hght;
   scale_width = scale_wdth;
   scale_height = scale_hght;
   this->ref_id = this->global_ref_id;
   this->global_ref_id++;
   this->create_id (this->id);
}

ChipsWindowSystem::ChipsWindowSystem (ChipsStateStream &in, ChipsEngine *en)
{
   this->eng = en;
   this->deserialize (in);
}


ChipsWindowSystem::~ChipsWindowSystem ()
{
   if (!frame_systems.empty ())
   {
      map<string, ChipsFrameSystem *>::iterator ii;
      for (ii = frame_systems.begin (); ii!=frame_systems.end (); ii++)
         ii->second->remove_parent ();
      frame_systems.clear ();
   }
}

ChipsWindowSystem::ChipsWindowSystem (const ChipsWindowSystem &ws):ChipsCoordinate (ws)
{
   this->eng = ws.eng;
   this->pix_width = ws.pix_width;
   this->pix_height = ws.pix_height;
   this->scale_width = ws.scale_width;
   this->scale_height = ws.scale_height;
   this->frame_systems = ws.frame_systems;
   this->ref_id = ws.ref_id;
}

void ChipsWindowSystem::operator = (const ChipsWindowSystem &ws)
{
   if (this == &ws)
      return;

   this->pix_width = ws.pix_width;
   this->pix_height = ws.pix_height;
   this->scale_width = ws.scale_width;
   this->scale_height = ws.scale_height;
   this->frame_systems = ws.frame_systems;
}

int ChipsWindowSystem::convert_cs_to_cs (const ChipsVector *in, ChipsVector *out, int size,
                                 COORD_SYS from_cs, const string &from_id,
                                 COORD_SYS to_cs, const string &to_id)
{
   int rval = ERR_STATUS_BAD;
   map <string, ChipsFrameSystem *>::iterator from_fs;
   map <string, ChipsFrameSystem *>::iterator to_fs;

   if (!in || !out)
      return ERR_STATUS_BAD;

   if (size <= 0)
      return ERR_STATUS_BAD;

   if ((from_cs == WINDOW_NORM) || (from_cs == PIXEL) || (from_cs == PIXEL_REAL)){
      if ((from_id.length () > 0) && (id != from_id)) return ERR_STATUS_BAD;
   }

   if ((to_cs == WINDOW_NORM) || (to_cs == PIXEL) || (to_cs == PIXEL_REAL)){
      if ((to_id.length () > 0) && (id != to_id)) return ERR_STATUS_BAD;
   }

   if (to_cs == FRAME_NORM){
      to_fs = frame_systems.find (to_id);
      if (to_fs == frame_systems.end ())
         return ERR_COORD_DNE;
   }
   
   switch (from_cs){
      case WINDOW_NORM:
	 if (to_cs == WINDOW_NORM){
            for (int ii=size; ii--;)
               out[ii] = in[ii];
            rval =  ERR_STATUS_GOOD;
         }
         else if ((to_cs == PIXEL_REAL) || (to_cs == PIXEL) || 	
		  (to_cs == FRAME_NORM)){
	    double scl_h = 1;
	    double scl_w = 1;
	    if (to_cs == PIXEL_REAL){
	       scl_h = scale_height;
	       scl_w = scale_width;
	    }
	    for (int ii=size; ii--;){
	       out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
	       if (out[ii].get_size () == 1)
		  return ERR_STATUS_BAD;
	       else{
		  out[ii][0] = in[ii][0]*pix_width*scl_w;
                  out[ii][1] = in[ii][1]*pix_height*scl_h;
	       }
 	    }
            rval =  ERR_STATUS_GOOD;
         }
	 if (to_cs == FRAME_NORM){
	    rval = to_fs->second->convert_cs_to_cs (out, out, size, PIXEL, id, 
						to_cs, to_id);
	 }
      break;
      case PIXEL:
         if (to_cs == WINDOW_NORM){
            for (int ii=size; ii--;){
	       out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
               if (out[ii].get_size () == 1)
	          return ERR_STATUS_BAD;
	       else{
                  out[ii][0] = in[ii][0]/pix_width;
                  out[ii][1] = in[ii][1]/pix_height;
	       }
	    }
            rval =  ERR_STATUS_GOOD;
         }
	 else if (to_cs == PIXEL_REAL){
            for (int ii=size; ii--;){
               out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
               if (out[ii].get_size () == 1)
	          return ERR_STATUS_BAD;
	       else {
                  out[ii][0] = in[ii][0]*scale_width;
                  out[ii][1] = in[ii][1]*scale_height;
               }

 	    }
            rval =  ERR_STATUS_GOOD;
         }
	 else if (to_cs == PIXEL){
            for (int ii=size; ii--;)
               out[ii] = in[ii];
            rval =  ERR_STATUS_GOOD;
         }
	 else if (to_cs == FRAME_NORM){
	    rval = to_fs->second->convert_cs_to_cs (in, out, size, PIXEL, id,
                                                to_cs, to_id);
         }
      break;
      case PIXEL_REAL:
         if (to_cs == WINDOW_NORM){
            for (int ii=size; ii--;){
               out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
               if (out[ii].get_size () < 2)
		  return ERR_STATUS_BAD;
               else{
                  out[ii][0] = in[ii][0]/(pix_width*scale_width);
                  out[ii][1] = in[ii][1]/(pix_height*scale_height);
	       }
            }
            rval =  ERR_STATUS_GOOD;
         }
         else if (to_cs == PIXEL){
            for (int ii=size; ii--;){
               out[ii] = in[ii];
	       // if the size is one use the dimension to determine
               // which axis to convert
               if (out[ii].get_size () < 2)
	          return ERR_STATUS_BAD;
               else{
                  out[ii][0] = in[ii][0]/scale_width;
                  out[ii][1] = in[ii][1]/scale_height;
	       }
            }
            rval =  ERR_STATUS_GOOD;
         }
         else if (to_cs == PIXEL_REAL){
            for (int ii=size; ii--;)
               out[ii] = in[ii];
            rval =  ERR_STATUS_GOOD;
         }
         else if (to_cs == FRAME_NORM){
            rval = to_fs->second->convert_cs_to_cs (in, out, size, PIXEL, id,
                                                to_cs, to_id);
         }
      break;
      case FRAME_NORM:
         from_fs = frame_systems.find (from_id);
         if (from_fs == frame_systems.end ())
            return ERR_STATUS_BAD;
	 if ((to_cs == WINDOW_NORM) || (to_cs == PIXEL))
	    rval = from_fs->second->convert_cs_to_cs (in, out, size, from_cs, 
						from_id, to_cs, to_id);
	 else if (to_cs == FRAME_NORM){
            rval = from_fs->second->convert_cs_to_cs (in, out, size, from_cs,
                                                from_id, PIXEL, id);
	    if (rval == ERR_STATUS_GOOD)
	       rval = to_fs->second->convert_cs_to_cs (out, out, size, PIXEL, 
						id, to_cs, to_id);
         }
	 else 
	    rval = ERR_STATUS_BAD;
      break;
      default: rval = ERR_STATUS_BAD;
   }
   return rval;
}

int ChipsWindowSystem::add_coord_sys (ChipsCoordinate *cs)
{
   if (cs->get_id().length () == 0)
      return ERR_STATUS_BAD;

   if (frame_systems.find (cs->get_id()) != frame_systems.end ())
      return ERR_STATUS_BAD;

   frame_systems[cs->get_id ()] = (ChipsFrameSystem *)cs;
   cs->set_parent (this);
   return ERR_STATUS_GOOD;
}


int ChipsWindowSystem::remove_coord_sys (const string &id)
{
   map <string, ChipsFrameSystem *>::iterator fs;

   if ((fs = frame_systems.find (id)) == frame_systems.end ())
      return ERR_STATUS_BAD;
//   delete fs->second;
   fs->second->remove_parent ();
   frame_systems.erase(fs);
   return ERR_STATUS_GOOD;
}


ChipsCoordinate *ChipsWindowSystem::get_coord (COORD_SYS coord_sys, const string &id)
{
   ChipsCoordinate *coord = NULL;
   if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
       (coord_sys == PIXEL_REAL) || (coord_sys == COORD_NONE)){
      if ((id.length() == 0) || (id == this->id))
         coord = this;
   }
   else
   {
      map <string, ChipsFrameSystem *>::iterator fs;

      int pos = id.find_first_of (".", 0);
      pos = id.find_first_of (".", pos+1);

      string tmp_id = id.substr (0, pos);

      if ((fs = frame_systems.find (tmp_id)) != frame_systems.end ())
         coord = fs->second;
      
      if ((coord_sys != FRAME_NORM) && coord)
         coord->get_coord (coord_sys, id);
      else if ((coord_sys == FRAME_NORM) && !coord)
      {
         // check the frame old ids for backwards compatibility
         for (fs = frame_systems.begin (); fs != frame_systems.end (); fs++)
            if (id == fs->second->get_old_id ())
               coord = fs->second;
      }
   }

   return coord;

}

int ChipsWindowSystem::insert (ChipsObject *co, bool use_co_info)
{
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   int rval = ERR_STATUS_BAD;
   if (co == NULL)
      return ERR_STATUS_BAD;

   // Store object local forces the object to remain
   // in this coordinate systems list. It overrides
   // the coordinate objects coord system
   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case COORD_NONE:
      {
         if ((coord_id.length () == 0) || (coord_id == id)){
            chips_objects.push_back (co);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval = ERR_COORD_DNE;
      }
      break;
      case FRAME_NORM:
      {
         map <string, ChipsFrameSystem *>::iterator fs;

         if ((fs = frame_systems.find (coord_id)) != frame_systems.end ())
            rval = fs->second->insert (co, use_co_info);
         else
            rval = ERR_COORD_DNE;
      }
      break;
   }


   return rval;
}


int ChipsWindowSystem::remove (ChipsObject *co, bool use_co_info)
{
   int rval = ERR_STATUS_BAD;
   COORD_SYS coord_sys = COORD_NONE;
   string coord_id = "";
   if (co == NULL)
      return ERR_STATUS_BAD;

   if (use_co_info && !store_object_local)
      co->get_coord_system (coord_sys, coord_id);

   switch (coord_sys){
      case WINDOW_NORM:
      case PIXEL:
      case PIXEL_REAL:
      case COORD_NONE:
      {
         list <ChipsObject *>::iterator ii = find (chips_objects.begin (),
                                                chips_objects.end (), co);
         if (ii != chips_objects.end ()){
            chips_objects.erase (ii);
            rval = ERR_STATUS_GOOD;
         }
      }
      break;
      case FRAME_NORM:
      {
         map <string, ChipsFrameSystem *>::iterator fs;

         if ((fs = frame_systems.find (coord_id)) != frame_systems.end ())
            rval = fs->second->remove (co, use_co_info);
         else
            rval = ERR_COORD_DS_DNE;
      }
   }

   return rval;
}


bool ChipsWindowSystem::coord_exists (COORD_SYS coord_sys, const string &id)
{
   bool rval = false;
   if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
       (coord_sys == PIXEL_REAL) || (coord_sys == COORD_NONE)){
      if ((id.length() == 0) || (id == this->id))
         rval = true;
   }
   else
   {
      ChipsCoordinate *tmp_coord = this->get_coord (coord_sys, id);
      rval = (tmp_coord != NULL);
   }

   return rval;
}

void ChipsWindowSystem::get_coord_list (list <string> &outList)
{
   map <string, ChipsFrameSystem *>::iterator ii;

   for (ii = frame_systems.begin (); ii != frame_systems.end (); ++ii)
      outList.push_back (ii->first);
}




int ChipsWindowSystem::update_obj_positions (bool update_all)
{
   int rval = ERR_STATUS_GOOD;

   list <ChipsObject *>::iterator ii;
   for (ii = chips_objects.begin (); ii != chips_objects.end (); ++ii){
       rval = (*ii)->reset_position ();
       if (rval != ERR_STATUS_GOOD)
          return rval;
   }

   if (update_all){
      map <string, ChipsFrameSystem *>::iterator iter;
      for (iter = frame_systems.begin (); iter != frame_systems.end (); iter++){
         if ((rval=iter->second->update_obj_positions(update_all)) != ERR_STATUS_GOOD)
            break;
      }
   }

   return rval;
};

void ChipsWindowSystem::set_dim (int win_width, int win_height)
{
   pix_width = win_width;
   pix_height = win_height;
   update_obj_positions ();
};

void ChipsWindowSystem::get_dim (int &win_width, int &win_height)
{
   win_width = pix_width;
   win_height = pix_height;
};

void ChipsWindowSystem::set_scale (double width, double height)
{
   scale_width = width;
   scale_height = height;
};

void ChipsWindowSystem::get_scale (double &width, double &height)
{
   width = scale_width;
   height = scale_height;
};


int ChipsWindowSystem::deserialize (ChipsStateStream &in)
{
   int sz = 0;
   int rval = ERR_STATUS_GOOD;
   ChipsFrameSystem *frame_sys;
 
   if (in.get_minor_version() > 6)
   {
      in.read_long ((long *)&(this->global_ref_id));
      in.read_long ((long *)&(this->ref_id));
      in.read_string (&id);
   }
   else
   {
      this->ref_id = this->global_ref_id;
      this->global_ref_id++;
      in.read_string (&id);
      this->create_id (this->id);
   }
   in.read_bool (&visible);
   in.read_int (&pix_width);
   in.read_int (&pix_height);
   in.read_double (&scale_width);
   in.read_double (&scale_height);


   // check for corruption
   if (in.read_checkpoint () != ERR_STATUS_GOOD)
      rval = ERR_STATEFILE_INC;
   else{

      in.read_int (&sz);

      for (int ii = 0; !rval && ii<sz; ii++){
         frame_sys = new ChipsFrameSystem (in, this);
         if (frame_sys)
            frame_systems[frame_sys->get_id ()] = frame_sys;
         else
            rval = ERR_STATEFILE_COR;
      }
   }

   return rval;

};

void ChipsWindowSystem::serialize (ChipsStateStream &out)
{
   int sz = 0;
   
   map <string, ChipsFrameSystem *>::iterator ii;
   if (out.get_minor_version() > 6)
   {
      out.write_long ((long *)&(this->global_ref_id));
      out.write_long ((long *)&(this->ref_id));
   }
     
   out.write_string (&id);
   out.write_bool (&visible);
   out.write_int (&pix_width);
   out.write_int (&pix_height);
   out.write_double (&scale_width);
   out.write_double (&scale_height);

   out.write_checkpoint ();

   sz = frame_systems.size ();
   out.write_int (&sz);

   for (ii = frame_systems.begin (); ii != frame_systems.end (); ii++)
      ii->second->serialize (out);
};

void ChipsWindowSystem::create_id (string &curr_id)
{
   stringstream tmp_id;

   tmp_id << "ds" << this->ref_id;
   curr_id = tmp_id.str ();
}
