// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//



#include "ChipsEventManager.hh"
#include "ChipsEventObserver.hh"
#include "ChipsDefines.hh"
#include "ChipsEngine.hh"
#include "ChipsObject.hh"
#include "ChipsCmdManager.hh"
#include "ChipsMatrix.hh"
#include "ChipsGuiMain.hh"

#include "ChipsPlot.hh"

#include <sstream>

extern ChipsCmdManager* cmd_manager;
extern ChipsGuiMain*    chipsGui;

ChipsEventObjectInfo::ChipsEventObjectInfo () 
{
   this->object_type = chips_not_object;
   this->subobject_type = chips_not_subobject;
}

ChipsEventObjectInfo::ChipsEventObjectInfo (const ChipsId &cid, 
                             V_OBJECT_TYPE type, V_SUBOBJECT_TYPE subtype)
{ 
   this->cid = cid;
   this->object_type = type;
   this->subobject_type = subtype ;
}

ChipsEventObjectInfo::ChipsEventObjectInfo (const ChipsEventObjectInfo &info)
{
   this->cid = info.cid;
   this->object_type = info.object_type;
   this->subobject_type = info.subobject_type;
}


void ChipsEventObjectInfo::operator = (const ChipsEventObjectInfo &info)
{
   this->cid = info.cid;
   this->object_type = info.object_type;
   this->subobject_type = info.subobject_type;
}



ChipsEventManager::ChipsEventManager (ChipsEngine *eng)
{
   this->engine = eng;

   this->start_event_pos[0] = 0.0;
   this->start_event_pos[1] = 0.0;
   this->event_started = false;


   this->selected_info.object_type = chips_not_object;
   this->state = ChipsEventManager::NoState;

   this->display_coord_sys = COORD_NONE;
   this->display_coord_id = "";
   this->display_coord_set = false;

   highlight = false;

   this->select_start = true;
   this->subselect_finished = false;

   this->add_substate = ChipsEventManager::NoAdd;

   this->block_subselect = false;
   this->block_select = false;


   this->annotation_points_xx = NULL;
   this->annotation_points_yy = NULL;
   this->num_annotation_points = 0;
 
}

ChipsEventManager::~ChipsEventManager ()
{
   if (this->num_annotation_points > 0)
   {
      delete []this->annotation_points_xx;
      delete []this->annotation_points_yy;
      this->num_annotation_points = 0;
   }
}

void ChipsEventManager::set_state (int state)
{
   this->state = state;
}

/*
void ChipsEventManager::set_add_substate (int state)
{
   this->add_substate = state;
}
*/


bool ChipsEventManager::is_valid_state (int state)
{
   bool rval = true;
   switch (state)
   {
      case ChipsEventManager::PickState:
      case ChipsEventManager::PanState:
      case ChipsEventManager::DragZoomState:
      case ChipsEventManager::RangeZoomState:
      {
         if (this->selected_info.object_type != chips_not_object)
         {
            ChipsId cid (this->selected_info.cid);
            cid.set_coord_sys (DATA);
            cid.set_coord_id ("");
            rval = this->engine->check_coordinate_exists (&cid);
         }
         else if (this->subselected_info.object_type != chips_not_object)
         {
            ChipsId cid (this->subselected_info.cid);
            cid.set_coord_sys (DATA);
            cid.set_coord_id ("");
            rval = this->engine->check_coordinate_exists (&cid);
         }
         else
            rval = false;
      }
      break;
      case ChipsEventManager::AddState:
      {
         if (this->selected_info.object_type != chips_not_object)
         {
            ChipsId cid (this->selected_info.cid);
            cid.set_coord_sys (PLOT_NORM);
            cid.set_coord_id ("");
            rval = this->engine->check_coordinate_exists (&cid);
         }
         else if (this->subselected_info.object_type != chips_not_object)
         {
            ChipsId cid (this->subselected_info.cid);
            cid.set_coord_sys (PLOT_NORM);
            cid.set_coord_id ("");
            rval = this->engine->check_coordinate_exists (&cid);
         }
         else
            rval = false;
      }
      break;

      default:
         rval = true;
      break;
   }

   return rval;
}

void ChipsEventManager::display_event_position (double event_pos[2])
{
   ChipsId cid_pix;
   ChipsId cid_data;
   ChipsVector pos (event_pos[0], event_pos[1]);
   stringstream text;
   string x_str;
   string y_str;

   cid_pix.set_coord_sys (PIXEL_REAL);

   if ((this->display_coord_sys == COORD_NONE) || !this->display_coord_set)
      this->engine->get_current_coord (&cid_data);
   else
   {
      cid_data.set_coord_sys (this->display_coord_sys);
      if (this->display_coord_sys == DATA)
         cid_data.set_coord_id (this->display_coord_id);
   }

   engine->convert_coordinate (pos, pos, &cid_pix,
                           &cid_data, XY_AXIS);

   if (cid_data.get_coord_sys () == DATA)
   {
      cid_data.set_axis_id (cid_data.get_xaxis_id ());
      cid_data.set_axis_dim (X_AXIS);
      engine->format_axis_point (&cid_data, pos[0], x_str);
      cid_data.set_axis_id (cid_data.get_yaxis_id ());
      cid_data.set_axis_dim (Y_AXIS);
      engine->format_axis_point (&cid_data, pos[1], y_str);
   }
   else
   {
      stringstream str;
      str << pos[0];
      x_str = str.str ();
      str.str ("");
      str << pos[1];
      y_str = str.str ();
   }



   text << "(" << x_str << "," << y_str << ")" << ends;
   this->engine->append_title_text (&cid_data, text.str ());
}

void ChipsEventManager::set_display_coordinates (COORD_SYS coord_sys, const string &coord_id)
{
   this->display_coord_sys = coord_sys;
   this->display_coord_id = coord_id;
   this->display_coord_set = (this->display_coord_sys != COORD_NONE);
}

void ChipsEventManager::get_display_coordinates (COORD_SYS &coord_sys, string &coord_id)
{
   coord_sys = this->display_coord_sys;
   coord_id = this->display_coord_id;
}



void ChipsEventManager::start_event (double event_pos[2])
{
   // this prevents nested events
   if (this->event_started)
      return;

   this->start_event_pos[0] = event_pos[0];
   this->start_event_pos[1] = event_pos[1];
   this->event_started = true;

   cmd_manager->open_block();
}

void ChipsEventManager::end_event ()
{
   // this prevents nested events
   if (!this->event_started)
      return;

   this->event_started = false;
   cmd_manager->close_block();
}


void ChipsEventManager::invoke_event (double event_pos[2], int event, void *data, ChipsEventObserver *caller)
{

   void *event_data = data;
   int state_change[2];
   CHIPS_PICK_EVT pick_evt;
   ChipsEventObjectInfo object_info;

   switch (event)
   {
      case ChipsEventManager::SelectEvent:
      {
         // *** THIS SHOULD BE REMOVED WHEN WE MOVE TO MULTIPLE SELECTIONS ***
         this->invoke_event (event_pos, ChipsEventManager::UnselectAllEvent, NULL);

         ChipsEventObjectInfo *info = (ChipsEventObjectInfo *)data;
         if (info != NULL)
         {
/*            // when in select mode only allow selection of container classes
            if (state != ChipsEventManager::SelectState)
            {
               if (!this->check_container_type (info->object_type))
                 info->object_type = chips_plot;
            }
*/              
            this->select (*info);
          }
      }
      break;
      case ChipsEventManager::UnselectEvent:
      {
         ChipsEventObjectInfo *info = (ChipsEventObjectInfo *)data;
         if (info != NULL)
            this->unselect (*info);
      }
      break;
      case ChipsEventManager::UnselectAllEvent:
      {
         this->unselect_all ();
      }
      break;
      case ChipsEventManager::RightClickEvent:
      {
         event_data = engine->get_object (&selected_info.cid, selected_info.object_type);
      }
      break;
      case ChipsEventManager::PanEvent:
      {
         double *start_pos = (double *)data;
         this->pan (start_pos, event_pos);
      }
      break;
      case ChipsEventManager::RangeZoomEvent:
      {
         double *range = (double *)data;
         this->pick_limits(range);
      }
      break;
      case ChipsEventManager::RangeZoomXEvent:
      {
         double *range = (double *)data;
         this->pick_limits(range, X_AXIS);
      }
      break;
      case ChipsEventManager::RangeZoomYEvent:
      {
         double *range = (double *)data;
         this->pick_limits(range, Y_AXIS);
      }
      break;
      case ChipsEventManager::ZoomEvent:
      {
         float zoomAmt = *((float *)data);
         this->zoom (zoomAmt);
      }
      break;
      case ChipsEventManager::MoveEvent:
      {
         double *start_pos = ((double *)data);
         this->move (start_pos, event_pos);
      }
      break;
      case ChipsEventManager::ResizeEvent:
      {
         double *bounds = ((double *)data);
         this->resize (bounds);
      }
      break;
      case ChipsEventManager::RotateEvent:
      {
         float angle = *((float *)data);
         this->rotate (angle);
      }
      break;
      case ChipsEventManager::DeleteEvent:
      {
         this->delete_object ();
      }
      break;
      case ChipsEventManager::ChangeStateEvent:
      {
         // end any add event currently running
         if (add_substate != ChipsEventManager::NoAdd)
            this->invoke_event (event_pos, ChipsEventManager::EndAddEvent, NULL);

         this->invoke_event (event_pos, ChipsEventManager::UnselectAllEvent, NULL);
         state_change[0] = this->state;
         state_change[1] = *((int *)data);
         this->state = state_change[1];
         event_data = (void *)(&state_change[0]);
      }
      break;
      case ChipsEventManager::ChangeAddStateEvent:
      {
         int val = ChipsEventManager::AddState;

         // go to select mode to edit the properties
         this->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&val);
         this->invoke_event (event_pos, ChipsEventManager::UnselectAllEvent, NULL);
         this->add_substate = *((int*)data);
      }
      break;
      case ChipsEventManager::PickEvent:
      {
         int *pickinfo = (int *)data;
         this->set_pickevent (event_pos, pickinfo, pick_evt);

         event_data = (void *)(&pick_evt);
      }
      break;
      case ChipsEventManager::AddEvent:
      {
         vector<double> *points = (vector<double> *)data;

         this->block_subselect = true;
         this->block_select = true;

         // if this is the first time open a block to capture 
         // multiple events
         if (num_annotation_points == 0)
            cmd_manager->open_block();

         if (add_substate == ChipsEventManager::PointAdd)
            add_point (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::PolyLineAdd)
            add_polyline (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::HLineAdd)
            add_hline (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::VLineAdd)
            add_vline (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::PolygonAdd)
            add_polygon (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::EllipseAdd)
            add_ellipse (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::BoxAdd)
            add_box (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::LabelAdd)
            add_label (tmp_selected_info, *points);
         else if (add_substate == ChipsEventManager::ColorBarAdd)
            add_colorbar (tmp_selected_info, *points);

      }
      break;
      case ChipsEventManager::EndAddEvent:
      {
         int val = ChipsEventManager::SelectState;

         if (add_substate == ChipsEventManager::PolygonAdd)
         {
            ChipsId cid;
            if (num_annotation_points >= 2)
               engine->undo ();
            if (num_annotation_points > 2)
            {
               int rval = engine->add_region (&cid, this->annotation_points_xx,this->annotation_points_yy, this->num_annotation_points);
               if (rval && chipsGui)
                  chipsGui->log_error ();
            }
         }
         else if ((add_substate == ChipsEventManager::EllipseAdd) ||
                 (add_substate == ChipsEventManager::BoxAdd))
         {
            ChipsId cid;
            if (num_annotation_points > 2)
            {
               int rval = engine->add_region (&cid, this->annotation_points_xx,this->annotation_points_yy, this->num_annotation_points);
               if (rval && chipsGui)
                  chipsGui->log_error ();
            }
         }


         if (this->num_annotation_points > 0)
         {
            delete []this->annotation_points_xx;
            delete []this->annotation_points_yy;
            this->annotation_points_xx = NULL;
            this->annotation_points_yy = NULL;
            this->num_annotation_points = 0;
         }



         engine->end_override_preferences ();

         this->block_subselect = false;
         this->block_select = false;

         // select the object
         if (add_substate != ChipsEventManager::NoAdd)
         {
            add_substate = ChipsEventManager::NoAdd;
            this->invoke_event (event_pos, ChipsEventManager::ChangeStateEvent, (void *)&val);
            this->invoke_event (event_pos, ChipsEventManager::SelectEvent, &tmp_selected_info);
         }

         cmd_manager->close_block();

      }
      break;
      default:
      break;
   }

   if (this->observers.find (event) != this->observers.end())
   {
      list<ChipsEventObserver *>::iterator iter = this->observers[event].begin ();

      for (; iter != this->observers[event].end (); iter++)
      {
         if (*iter != caller)
            (*iter)->execute_event (event, event_data);
      }
   }
}

bool ChipsEventManager::request_subselect (const ChipsEventObjectInfo &info)
{
   bool rval = false;
   if (this->state == ChipsEventManager::NoState)
      return false;

   if (this->block_subselect)
      return false;

   if (this->subselect_finished)
      return false;

   if (this->select_start || 
      (this->tmp_selected_info.object_type == this->subselect_type) ||
      (this->tmp_selected_info.object_type != info.object_type))
   {
      rval = this->request_select (info);

      if (rval)
      {
         ChipsObject *obj = engine->get_object (&info.cid, info.object_type);
         ChipsObject *selected_object = engine->get_object (&this->selected_info.cid, info.object_type);
         if (obj == selected_object)
           this->subselect_finished = true;
         this->subselect_type = info.object_type;
      }
   }
   return rval;

}

bool ChipsEventManager::request_select (const ChipsEventObjectInfo &info)
{
   if (this->state == ChipsEventManager::NoState)
      return false;

   if (this->block_select)
      return false;

   bool rval = false;

   if (this->select_start)
   {
      this->select_start = false;
      rval = true;

   }
   // if this is a new frame check to see if the new frame falls on top
   // of the old one
   else if (strcmp(this->tmp_selected_info.cid.get_frame_id (), info.cid.get_frame_id ()) != 0)
   {
      string selected_frame = this->tmp_selected_info.cid.get_frame_id ();
      string object_frame = info.cid.get_frame_id ();

      if (this->compare_frame_order (object_frame, selected_frame))
         rval = true;
   }
   else if (info.object_type == chips_frame)
      ; // if the object type is a chips_frame punt. the case for switching
        // frames handled above.
   else if (info.object_type == chips_plot)
   {
      if (this->tmp_selected_info.object_type == chips_frame)
         rval = true;
      else if (this->tmp_selected_info.object_type == chips_plot)
      {
         // Give precidence to the plot title over the plot
         if ((this->tmp_selected_info.subobject_type == chips_not_subobject) ||
             (info.subobject_type == chips_plot_title))
            rval = true;
      }
   }
   else
   {
      // if the current select is a frame or a plot don't bother 
      // comparing depths
      if ((this->tmp_selected_info.object_type == chips_frame) || 
          (this->tmp_selected_info.object_type == chips_plot))
         rval = true;
      else
      {
         int obj_depth, obj_shuffle;
         int selected_depth, selected_shuffle;

     
         rval = this->engine->get_object_depth_info (&info.cid, info.object_type, obj_depth, obj_shuffle);
         rval = this->engine->get_object_depth_info (&this->tmp_selected_info.cid, this->tmp_selected_info.object_type, selected_depth, selected_shuffle);


         if ((selected_depth < obj_depth) || 
          ((selected_depth == obj_depth) && (selected_shuffle < obj_shuffle)))
            rval = true;
      }
   }

   if (rval == true)
   {
      this->tmp_selected_info = info;

      // something actual was selected reset teh subselect
      this->subselect_type = chips_not_object;
   }


   return rval;

}

void ChipsEventManager::finalize_select (bool invoke_select)
{
   // if no request has been done return
   if (this->select_start)
      return;

   double pos[] = {0,0};

   this->select_start = true;
   this->subselect_finished = false;
   this->subselected_info.cid.clear ();
   this->subselected_info.object_type = chips_not_object;

   // Check if we're trying to select the plot area. If so
   // don't select the plot unless it's not the container which
   // is currently selected
   if ((this->tmp_selected_info.object_type == this->subselect_type)  && (this->tmp_selected_info.subobject_type == chips_not_subobject))
   {
      // if the object is a container and is a new container select 
      // that container
      if (this->check_container_type (this->tmp_selected_info.object_type))
      {
         bool cid_same = true;
         switch (this->tmp_selected_info.object_type)
         {
            case chips_plot:
            {
               const char *new_id = this->tmp_selected_info.cid.get_plot_id ();
               const char *old_id = this->selected_info.cid.get_plot_id ();

               cid_same =  (cid_same && (strcmp (new_id, old_id) == 0));
            }
            case chips_frame: // fallthrough intended
            {
               const char *new_id = this->tmp_selected_info.cid.get_frame_id ();
               const char *old_id = this->selected_info.cid.get_frame_id ();

               cid_same = (cid_same && (strcmp (new_id, old_id) == 0));
            }
            case chips_window: // fallthrough intended
            {
               const char *new_id = this->tmp_selected_info.cid.get_window_id ();
               const char *old_id = this->selected_info.cid.get_window_id ();

               cid_same = (cid_same && (strcmp (new_id, old_id) == 0));
            }
            break;
            default:
            break;
         }

           
         // if select is not invoked during a subselect then 
         // objects may not be added correctly in particular interactively
         // adding polygons
         if (!cid_same) // && invoke_select)
            this->invoke_event (pos, ChipsEventManager::SelectEvent, &this->tmp_selected_info);
      }
      if (this->selected_info.object_type != chips_not_object)
         this->invoke_event (pos, ChipsEventManager::UnselectEvent, &this->selected_info);

      
      this->subselected_info = tmp_selected_info;
   }
   else if (!this->is_selected (this->tmp_selected_info))
   {
      if  (invoke_select) 
      {
         if (this->selected_info.object_type != chips_not_object)
            this->invoke_event (pos, ChipsEventManager::UnselectEvent, &this->selected_info);

         this->invoke_event (pos, ChipsEventManager::SelectEvent, &this->tmp_selected_info);
      }
      
      this->selected_info = tmp_selected_info;
   }
 
   this->subselect_type = chips_not_object;

}

void ChipsEventManager::select (const ChipsEventObjectInfo &info)
{

   if (info.object_type == chips_not_object)
      return;

   this->selected_info = info;

   if (!this->display_coord_set)
   {
      ChipsId cid(this->selected_info.cid);
      this->engine->get_current_coord (&cid);
      this->display_coord_sys = cid.get_coord_sys ();
      this->display_coord_id = cid.get_coord_id ();
   }

   this->engine->set_current (&this->selected_info.cid, this->selected_info.object_type, false);
   if (this->state == ChipsEventManager::SelectState)
   {

      this->select_start = true;
      this->engine->select(&this->selected_info.cid, this->selected_info.object_type, this->selected_info.subobject_type, true);
      if (highlight)
         this->engine->highlight_object(&this->selected_info.cid, this->selected_info.object_type); 
   }
}

void ChipsEventManager::unselect (const ChipsEventObjectInfo &info)
{
   bool valid_unselect = (((this->state == ChipsEventManager::SelectState) ||
                          (this->state == ChipsEventManager::AddState)) &&
                          (info.object_type != chips_not_object));

   if (this->is_selected (info))
   {
      // select must happen before selected info change as changing
      // selected_info could change info if they are the same
      if (valid_unselect)
          this->engine->select(&info.cid, info.object_type, info.subobject_type, false);
      this->selected_info.object_type = chips_not_object;
   }
   else if (valid_unselect)
      this->engine->select(&info.cid, info.object_type, info.subobject_type, false);
}


void ChipsEventManager::unselect_all ()
{
   if ((this->state != ChipsEventManager::SelectState) &&
       (this->state != ChipsEventManager::AddState))
      return;

   if (this->selected_info.object_type != chips_not_object)
   {
      this->engine->select(&this->selected_info.cid, this->selected_info.object_type, this->selected_info.subobject_type, false);
   }

   this->selected_info.object_type = chips_not_object;

}

void ChipsEventManager::delete_object ()
{
   if (this->selected_info.object_type == chips_not_object)
      return;

   ChipsId cid (this->selected_info.cid);
   int rval;

   switch (this->selected_info.object_type)
   {
      case chips_window:
         rval = this->engine->delete_window (&cid);
      break;
      case chips_frame:
         rval = this->engine->delete_frame (&cid);
      break;
      case chips_plot:
         rval = this->engine->delete_plot (&cid);
      break;
      case chips_curve:
         rval = this->engine->delete_curve (&cid);
      break;
      case chips_histogram:
         rval = this->engine->delete_histogram (&cid);
      break;
      case chips_contour:
         rval = this->engine->delete_contour (&cid);
      break;
      case chips_image:
         rval = this->engine->delete_image (&cid);
      break;
      case chips_label:
         rval = this->engine->delete_label (&cid);
      break;
      case chips_line:
         rval = this->engine->delete_line (&cid);
      break;
      case chips_point:
         rval = this->engine->delete_point (&cid);
      break;
      case chips_axis:
         rval = this->engine->delete_axis (&cid);
      break;
      case chips_colorbar:
         rval = this->engine->delete_colorbar (&cid);
      break;
      case chips_region:
         rval = this->engine->delete_region (&cid);
      break;
      default:
      break;
   }

   if (!rval)
      this->selected_info.object_type = chips_not_object;
}


void ChipsEventManager::move (double start_pos[], double event_pos[])
{
   if (this->selected_info.object_type == chips_not_object)
      return;

   ChipsId cid(this->selected_info.cid);
   cid.set_coord_sys (PIXEL_REAL);
   cid.set_coord_id ("");

   this->engine->move (&cid,
                     this->selected_info.object_type, 
                     this->selected_info.subobject_type,
                     event_pos[0] - start_pos[0],
                     event_pos[1] - start_pos[1],
                     V_MOVE_DELTA);

}

void ChipsEventManager::resize (double bounds[])
{
   if (this->selected_info.object_type == chips_not_object)
      return;

   ChipsId cid(this->selected_info.cid);
   cid.set_coord_sys (PIXEL_REAL);
   cid.set_coord_id ("");

   this->engine->resize (&cid,
                     this->selected_info.object_type,
                     this->selected_info.subobject_type,
                     bounds);

}

void ChipsEventManager::rotate (float angle)
{
   if (this->selected_info.object_type == chips_not_object)
      return;

   this->engine->rotate (&this->selected_info.cid,
                     this->selected_info.object_type,
                     this->selected_info.subobject_type,
                     angle);

}


void ChipsEventManager::pan (double start_pos[], double event_pos[])
{

   ChipsId cid1;
   if (this->selected_info.object_type != chips_not_object)
      cid1 = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid1 = this->subselected_info.cid;
   else
      return;

   ChipsVector start (start_pos[0], start_pos[1]);
   ChipsVector end (event_pos[0], event_pos[1]);
   ChipsId cid2(cid1);

   engine->get_current_coord (&cid1);

   cid2.set_coord_sys (PIXEL_REAL);
   cid2.set_coord_id ("");
   engine->convert_coordinate (start, start, &cid2,
                           &cid1, XY_AXIS);
   engine->convert_coordinate (end, end, &cid2,
                           &cid1, XY_AXIS);

   engine->pan (&cid1, start[0], start[1], end[0], end[1]);

}

void ChipsEventManager::zoom (float scale)
{
   ChipsId cid;
   if (this->selected_info.object_type != chips_not_object)
      cid = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid = this->subselected_info.cid;
   else
      return;


   this->engine->zoom (&cid, scale);
   
}

void ChipsEventManager::pick_limits (double range[], int dim)
{
   ChipsId cid1;
   if (this->selected_info.object_type != chips_not_object)
      cid1 = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid1 = this->subselected_info.cid;
   else
      return;


   ChipsVector lower_bounds (range[0],range[1]);
   ChipsVector upper_bounds (range[2],range[3]);
   ChipsId cid2 (cid1);

//   cid1.set_coord_sys(DATA);
   engine->get_current_coord (&cid1);

// Note: This may not be the cleanest approach, but the alternatives didn't
// seem much better. Tag on a 'none' to the end of the id to get the bounds
// in image coordinates. This is faster and cleaner than getting the bounds
// in wcs coordinates.

   if (strlen (cid1.get_coord_id ()) > 0)
   {
      string coord_id = cid1.get_coord_id ();
      coord_id += "_none";
      cid1.set_coord_id (coord_id);
   }

   cid2.set_coord_sys (PIXEL_REAL);
   engine->convert_coordinate (lower_bounds, lower_bounds, &cid2,
                           &cid1, XY_AXIS);
   engine->convert_coordinate (upper_bounds, upper_bounds, &cid2,
                           &cid1, XY_AXIS);

   this->engine->limits (&cid1, dim, lower_bounds[0], upper_bounds[0], lower_bounds[1], upper_bounds[1]);


}

void ChipsEventManager::set_pickevent (double pos[2], int pickinfo[2],
                       CHIPS_PICK_EVT &outpick)
{
   ChipsId cid1;
   if (this->selected_info.object_type != chips_not_object)
      cid1 = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid1 = this->subselected_info.cid;


   ChipsVector event_pos (pos[0],pos[1]);
   ChipsId cid2 (cid1);
   string x_str = "";
   string y_str = "";

//   cid1.set_coord_sys(DATA);
   if ((this->display_coord_sys == COORD_NONE) || !this->display_coord_set)
      this->engine->get_current_coord (&cid1);
   else
   {
      cid1.set_coord_sys (this->display_coord_sys);
      if (this->display_coord_sys == DATA)
         cid1.set_coord_id (this->display_coord_id);
   }

   cid2.set_coord_sys (PIXEL_REAL);
   engine->convert_coordinate (event_pos, event_pos, &cid2,
                           &cid1, XY_AXIS);
   if (cid1.get_coord_sys () == DATA)
   {
      cid1.set_axis_id (cid1.get_xaxis_id ());
      cid1.set_axis_dim (X_AXIS);
      engine->format_axis_point (&cid1, event_pos[0], x_str);
      cid1.set_axis_id (cid1.get_yaxis_id ());
      cid1.set_axis_dim (Y_AXIS);
      engine->format_axis_point (&cid1, event_pos[1], y_str);
   }
   else
   {
      stringstream str;
      str << event_pos[0] << ends;
      x_str = str.str ();
      str.str ("");
      str << event_pos[1] << ends;
      y_str = str.str ();
   }



   outpick.x = event_pos[0];
   outpick.y = event_pos[1];
   outpick.x_str = x_str;
   outpick.y_str = y_str;
   outpick.event = pickinfo[0];
   outpick.key = (char)pickinfo[1];


}

void ChipsEventManager::add_hline (ChipsEventObjectInfo &object_info, const vector<double> &points)
{
   ChipsId cid1;
   ChipsVector pos1(0.1,0.1);
   ChipsVector pos2(.9,.9);
   vector<double> tmp_points(4);

   if (points.size () < 2)
      return;


   if (this->selected_info.object_type != chips_not_object)
      cid1 = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid1 = this->subselected_info.cid;
   else
      cid1 = object_info.cid;

   ChipsId cid2 (cid1);

   cid1.set_coord_sys (PLOT_NORM);
   cid2.set_coord_sys (PIXEL_REAL);
   engine->convert_coordinate (pos1, pos1, &cid1, &cid2, X_AXIS);
   engine->convert_coordinate (pos2, pos2, &cid1, &cid2, X_AXIS);

   tmp_points[0] = pos1[0];
   tmp_points[1] = points[1];
   tmp_points[2] = pos2[0];
   tmp_points[3] = points[1];

   engine->override_preferences("extend=both", "line");

   this->add_polyline (object_info, tmp_points);

}


void ChipsEventManager::add_vline (ChipsEventObjectInfo &object_info, const vector<double> &points)
{
   ChipsId cid1;
   ChipsVector pos1(0.1,0.1);
   ChipsVector pos2(0.9,0.9);
   vector<double> tmp_points(4);

   if (points.size () < 2)
      return;


   if (this->selected_info.object_type != chips_not_object)
      cid1 = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      cid1 = this->subselected_info.cid;
   else
      cid1 = object_info.cid;

   ChipsId cid2 (cid1);

   cid1.set_coord_sys (PLOT_NORM);
   cid2.set_coord_sys (PIXEL_REAL);
   engine->convert_coordinate (pos1, pos1, &cid1, &cid2, Y_AXIS);
   engine->convert_coordinate (pos2, pos2, &cid1, &cid2, Y_AXIS);

   tmp_points[0] = points[0];
   tmp_points[1] = pos1[0];
   tmp_points[2] = points[0];
   tmp_points[3] = pos2[0];

   engine->override_preferences("extend=both", "line");

   this->add_polyline (object_info, tmp_points);
}


void ChipsEventManager::add_polyline (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   int rval;
   ChipsId *cid;

   this->setup_add_object (object_info, points);

   cid = &object_info.cid;

   if (this->num_annotation_points < 2)
      return;

   if (this->num_annotation_points > 2)
      engine->undo ();

   rval = engine->add_line (cid, this->annotation_points_xx,
                             this->annotation_points_yy, 
                             this->num_annotation_points);
   if (rval && chipsGui)
      chipsGui->log_error ();

   object_info.object_type = chips_line;
   object_info.subobject_type = chips_not_subobject;
}

void ChipsEventManager::add_point (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   int rval;
   ChipsId *cid;

   this->setup_add_object (object_info, points);
   cid = &object_info.cid;

   rval = engine->add_point (cid, this->annotation_points_xx[0], 
                                  this->annotation_points_yy[0]);
   if (rval && chipsGui)
      chipsGui->log_error ();

   object_info.object_type = chips_point;
   object_info.subobject_type = chips_not_subobject;

}

void ChipsEventManager::add_colorbar (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   int rval;
   ChipsId *cid;
   bool coordset = this->display_coord_set;
   COORD_SYS coord_sys = this->display_coord_sys;
   string coord_id = this->display_coord_id;

   // set the coord sys to plot norm by default
   this->display_coord_sys = PLOT_NORM;
   this->display_coord_id = "";
   this->display_coord_set = true;

   this->setup_add_object (object_info, points);
   cid = &object_info.cid;

   // center the colorbar to the plot based on default orientation
   char *pref = engine->get_preference("colorbar.orientation");
   if (pref && !strcmp(pref,"horizontal"))
      this->annotation_points_xx[0] = 0.5;
   else 
      this->annotation_points_yy[0] = 0.5;

   rval = engine->add_colorbar (cid, this->annotation_points_xx[0],
                                  this->annotation_points_yy[0]);
   if (rval && chipsGui)
      chipsGui->log_error ();

   // reset the display coord
   this->display_coord_sys = coord_sys;
   this->display_coord_id = coord_id;
   this->display_coord_set = coordset;


   object_info.object_type = chips_colorbar;
   object_info.subobject_type = chips_not_subobject;

}


void ChipsEventManager::add_polygon (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   this->add_polyline (object_info, points);

   object_info.object_type = chips_region;
   object_info.subobject_type = chips_not_subobject;

}

void ChipsEventManager::add_ellipse (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   this->setup_add_object (object_info, points);

   object_info.object_type = chips_region;
   object_info.subobject_type = chips_not_subobject;

}

void ChipsEventManager::add_box (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   this->setup_add_object (object_info, points);

   object_info.object_type = chips_region;
   object_info.subobject_type = chips_not_subobject;

}

void ChipsEventManager::add_label (ChipsEventObjectInfo &object_info,
                       const vector<double> &points)
{
   int rval;
   ChipsId *cid;

   this->setup_add_object (object_info, points);
   cid = &object_info.cid;

   rval = engine->add_label (cid, this->annotation_points_xx[0],
                                  this->annotation_points_yy[0], "");
   if (rval && chipsGui)
      chipsGui->log_error ();

   int ival = 1; 
   engine->set_property(cid, static_cast<void*>(&ival), SET_LABEL_BBOX);

   object_info.object_type = chips_label;
   object_info.subobject_type = chips_not_subobject;


}

void ChipsEventManager::setup_add_object (ChipsEventObjectInfo &object_info, const vector<double> &points)
{
   ChipsId *cid = &object_info.cid;

   if (this->selected_info.object_type != chips_not_object)
      *cid = this->selected_info.cid;
   else if (this->subselected_info.object_type != chips_not_object)
      *cid = this->subselected_info.cid;

   if (points.size () % 2 != 0)
      return;


   if (this->num_annotation_points > 0)
   {
      delete []this->annotation_points_xx;
      delete []this->annotation_points_yy;
      this->num_annotation_points = 0;
   }

   num_annotation_points = points.size () / 2;
   ChipsId cid2 (*cid);
   string x_str = "";
   string y_str = "";

   double *xx = new double[num_annotation_points];
   double *yy = new double[num_annotation_points];

   this->annotation_points_xx = xx;
   this->annotation_points_yy = yy;


   if ((this->display_coord_sys == COORD_NONE) || !this->display_coord_set)
      this->engine->get_current_coord (cid);
   else
   {
      cid->set_coord_sys (this->display_coord_sys);
      if (this->display_coord_sys == DATA)
         cid->set_coord_id (this->display_coord_id);
   }

   cid2.set_coord_sys (PIXEL_REAL);

   ChipsVector *pos1 = new ChipsVector[num_annotation_points];
   for (int ii=num_annotation_points; ii--;)
      pos1[ii] = ChipsVector (points[ii*2], points[ii*2+1]);

   engine->convert_coordinate (pos1, pos1, num_annotation_points, &cid2, cid, XY_AXIS);

   for (int ii=num_annotation_points; ii--;)
   {
      xx[ii] = pos1[ii][0];
      yy[ii] = pos1[ii][1];
   }

}

void ChipsEventManager::register_observer (ChipsEventObserver *observer, int event)
{
   this->observers[event].push_back (observer);
}

void ChipsEventManager::unregister_observer (ChipsEventObserver *observer, int event)
{
   if (this->observers.find (event) == this->observers.end())
      return;

   list<ChipsEventObserver *>::iterator iter = this->observers[event].begin ();

   for (; iter != this->observers[event].end (); iter++)
      if (*iter == observer)
         break;

   if (iter != this->observers[event].end ())
      this->observers[event].erase (iter);

}

bool ChipsEventManager::is_selected (const ChipsEventObjectInfo &info)
{
   bool rval = false;
   if ((this->selected_info.object_type == info.object_type) && 
       (this->selected_info.subobject_type == info.subobject_type))
   {
      ChipsObject *obj = engine->get_object (&info.cid, info.object_type);
      ChipsObject *selected_object = engine->get_object (&this->selected_info.cid, this->selected_info.object_type);

      rval = (obj == selected_object);
   }

   return rval;

}

bool ChipsEventManager::is_subselected (const ChipsEventObjectInfo &info)
{
   bool rval = false;
 
   if ((this->subselected_info.object_type != info.object_type) ||
       (this->subselected_info.subobject_type != info.subobject_type))
      return false;

   ChipsObject *obj = engine->get_object (&info.cid, info.object_type);
   ChipsObject *selected_object = engine->get_object (&this->subselected_info.cid, info.object_type);

   rval = (obj == selected_object);


   return rval;
}


bool ChipsEventManager::compare_frame_order (const string &frame1,
                                             const string &frame2) const
{
   ChipsId cid;
   bool rval = false;
   list<string> frame_order;
   list<string>::iterator iter;

   this->engine->get_frame_order (&cid, frame_order);
   for (iter=frame_order.begin (); iter!=frame_order.end(); iter++)
   {
      // if we find frame1 then frame2 must be on top
      if (*iter == frame1)
      {
         rval = false;
         break;
      }
      // otherwise the frame1 must be on top
      else if (*iter == frame2)
      {
         rval = true;
         break;
      }
   }
   return rval;
}

bool ChipsEventManager::check_container_type (V_OBJECT_TYPE object_type)
{
   return ((object_type == chips_window) || 
           (object_type == chips_frame) ||
           (object_type == chips_plot));
}

