// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef CHIPSEVENTMANAGER_HH
#define CHIPSEVENTMANAGER_HH

#include "ChipsId.hh"
#include "ChipsDefines.hh" // for CHIPS_PICK_EVT

#include <map>
#include <list>
#include <string>
#include <vector>


class ChipsEngine;
class ChipsEventObserver;

class ChipsEventObjectInfo
{
   public:
      ChipsId cid;
      V_OBJECT_TYPE object_type;
      V_SUBOBJECT_TYPE subobject_type;

      ChipsEventObjectInfo ();
      ChipsEventObjectInfo (const ChipsEventObjectInfo& info);
      ChipsEventObjectInfo (const ChipsId &cid, V_OBJECT_TYPE type, V_SUBOBJECT_TYPE subtype=chips_not_subobject);

      void operator = (const ChipsEventObjectInfo &info);
};


class ChipsEventManager
{
   public:

      enum _InteractionState_
      {
         NoState,
         PanState,
         RangeZoomState,  // pick limits
         SelectState,
         DragZoomState,  
         AddState,
         DeleteState,
         PickState
      };

      enum _AddSubState_
      {
         NoAdd,
         PointAdd,
         PolyLineAdd,
         HLineAdd,
         VLineAdd,
         PolygonAdd,
         EllipseAdd,
         BoxAdd,
         LabelAdd,
         ColorBarAdd
      };

      enum _Events_
      {
         SelectEvent,
         UnselectEvent,
         UnselectAllEvent,
         RightClickEvent,
         PanEvent,
	 StartRangeZoomEvent, // setup pick limits
         RangeZoomEvent,  // pick limits
         RangeZoomXEvent, // pick limits on X-Axis
         RangeZoomYEvent, // pick limits on Y-Axis
         ZoomEvent,
         MoveEvent,
         ResizeEvent,
         RotateEvent,
         DeleteEvent,
         InfoEvent,        // report errors and info
         StartPickEvent,
         PickEvent,
         ChangeStateEvent,
         ChangeAddStateEvent,  // change the state to Add State
         AddEvent,             // add an object
         EndAddEvent           // finished adding the object
      };


      ChipsEventManager (ChipsEngine *eng);
      ~ChipsEventManager ();

      // event specific functions
      void set_state (int state);
      int get_state () {return this->state;};
      bool is_valid_state () {return this->is_valid_state (this->state);};
      bool is_valid_state (int state);
      int get_add_substate () { return this->add_substate;};
//      void set_add_substate (int state);

      // Display the event position
      void display_event_position (double event_pos[2]);
      void set_display_coordinates (COORD_SYS coord_sys, const string &coord_id);
      void get_display_coordinates (COORD_SYS &coord_sys, string &coord_id);

    
      void start_event (double event_pos[2]);
      void invoke_event (double event_pos[2], int event, void *data, ChipsEventObserver *caller=NULL);
      void end_event ();


      // Select an object
  
      // select a single object. this is designed to be used with multiple
      // selects defining the actual selected object by depth.  
      // finalize_select should be called to end the request_select cycle
      bool request_select (const ChipsEventObjectInfo &info);
      bool request_subselect (const ChipsEventObjectInfo &info);

      // ends the request_select sequence and invokes the select event
      void finalize_select (bool invoke_select);

      // Get the engine
      ChipsEngine *get_engine () {return this->engine;};

      // Get the selected object
      void get_selected (ChipsEventObjectInfo &info)
         {info = this->selected_info;}

      void register_observer (ChipsEventObserver* observer, int event);
      void unregister_observer (ChipsEventObserver* observer, int event);

      double *get_start_event_pos () {return &this->start_event_pos[0];};

      bool is_selected (const ChipsEventObjectInfo &info);
      bool is_subselected (const ChipsEventObjectInfo &info);

      void set_highlight_on_select(bool val) { highlight = val; } 
      bool get_highlight_on_select()  { return highlight; }

   protected:


      // Operations that work on selected objects
      void move (double start_pos[], double event_pos[]);
      void resize (double bounds[]);
      void rotate (float angle);
      void delete_object ();

      // Pan and zoom
      void pan (double start_pos[], double event_pos[]);
      void zoom (float scale);
      void pick_limits (double range[], int dim=XY_AXIS);

      // general selection functions
      void select (const ChipsEventObjectInfo &info);
      void unselect (const ChipsEventObjectInfo &info);
      void unselect_all ();

      // pick functions
      void set_pickevent (double pos[2], int pickinfo[2], 
                       CHIPS_PICK_EVT &outpickevent);

      // add functions
      void setup_add_object (ChipsEventObjectInfo &object_info, 
                       const vector<double> &points);
      
      void add_hline (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_vline (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_polyline (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_point (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_polygon (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_ellipse (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_box (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_label (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);
      void add_colorbar (ChipsEventObjectInfo &object_info,
                       const vector<double> &points);


      


      // compare frame order of two frames -- true if frame1 is
      // on top of frame2
      bool compare_frame_order (const string &frame1, const string &frame2) const;

      // check object type category
      bool check_container_type (V_OBJECT_TYPE object_type);

      ChipsEventObjectInfo selected_info;
      ChipsEventObjectInfo subselected_info;
      ChipsEventObjectInfo tmp_selected_info;
      bool select_start;
      bool subselect_finished;
      bool subselect_requested;
      V_OBJECT_TYPE subselect_type;

      int state; 
      int add_substate;

      COORD_SYS display_coord_sys;
      string display_coord_id;
      bool display_coord_set;

      double start_event_pos[2];
      bool event_started;
      bool highlight;

      bool block_subselect;
      bool block_select;

      ChipsEngine *engine;

      double *annotation_points_xx;
      double *annotation_points_yy;
      int     num_annotation_points;

      map<int, list<ChipsEventObserver*> > observers;
};

#endif


