// 
//  Copyright (C) 2008-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGtkWindow.hh"
#include "ChipsWindow.hh"
#include "ChipsEngine.hh"
#include "ChipsId.hh"
#include "ChipsEventManager.hh"

#include <iostream>
#include <stdlib.h>

using namespace std;

GdkGLConfig *ChipsGtkWindow::glconfig = NULL;

extern ChipsEventManager *event_manager;

ChipsGtkWindow::ChipsGtkWindow (ChipsWindow *parent, string geom, bool focus)
{

   this->eventSafeFlag = false;
   this->parent = parent;
   this->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   this->drawing_area =  gtk_drawing_area_new ();
   this->has_focus = true;

   // setup opengl -- note this is here because vtk needs it
   // it's not clear to me that this is the correct place to do this
   // however, trying to do it in vtk causes issues
   if (glconfig == NULL)
   {
      glconfig = gdk_gl_config_new_by_mode ((GdkGLConfigMode)(GDK_GL_MODE_RGB  |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE));

      if (glconfig == NULL)
      {
         g_print ("*** Cannot find the double-buffered visual.\n");
         g_print ("*** Trying single-buffered visual.\n");

         /* Try single-buffered visual */
         glconfig = gdk_gl_config_new_by_mode ((GdkGLConfigMode)(GDK_GL_MODE_RGB
   |
                                            GDK_GL_MODE_DEPTH));
         if (glconfig == NULL)
         {
            g_print ("*** No appropriate OpenGL-capable visual found.\n");
            exit (1);
         }
      }
   }

   // Set OpenGL-capability to the widget.
   gtk_widget_set_gl_capability (this->drawing_area,
                                this->glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

   // if a geometry hint such as '+0-100' was specified try to use it
   if (geom.length())
      if (!gtk_window_parse_geometry (GTK_WINDOW (this->window), geom.c_str()))
         g_print ("*** Invalid geometry hint request ignored.\n");

   gtk_window_set_title (GTK_WINDOW (this->window), "");
   gtk_window_set_default_size (GTK_WINDOW(this->window), 300, 300);

   gtk_window_set_focus_on_map(GTK_WINDOW(this->window), focus);

   gtk_container_set_reallocate_redraws (GTK_CONTAINER (this->window), TRUE);

   // setup events
   gtk_widget_add_events (this->drawing_area,
                             GDK_EXPOSURE_MASK       |
                             GDK_ENTER_NOTIFY_MASK   |
                             GDK_LEAVE_NOTIFY_MASK   );

   gtk_widget_add_events (this->window, GDK_DELETE   |
					GDK_BUTTON_PRESS_MASK |
                                        GDK_CONFIGURE);

   // connect events
   g_signal_connect_after (G_OBJECT (this->drawing_area), "realize",
                          G_CALLBACK (this->on_realize),
                          static_cast<void *> (this));

   g_signal_connect (G_OBJECT (this->window), "delete_event",
                        G_CALLBACK(on_delete),
                        static_cast<void*> (this));

   g_signal_connect (G_OBJECT (this->drawing_area), "expose_event", 
                        G_CALLBACK(on_expose),
                        static_cast<void*> (this));

   g_signal_connect (G_OBJECT (this->window), "button-press-event",
                        G_CALLBACK(on_focus),
                        static_cast<void*> (this));

   g_signal_connect (G_OBJECT (this->drawing_area), "enter-notify-event",
                        G_CALLBACK(on_enter),
                        static_cast<void*> (this));

   g_signal_connect (G_OBJECT (this->drawing_area), "leave-notify-event",
                        G_CALLBACK(on_leave),
                        static_cast<void*> (this));

   g_signal_connect (G_OBJECT (this->window), "configure_event",
                        G_CALLBACK(on_configure),
                        static_cast<void*> (this));




   GTK_WIDGET_SET_FLAGS(this->drawing_area, GTK_CAN_FOCUS);

   gtk_widget_set_double_buffered(this->drawing_area, false);

   gtk_container_add (GTK_CONTAINER (this->window), this->drawing_area);
   gtk_widget_realize (this->drawing_area);
}

ChipsGtkWindow::~ChipsGtkWindow ()
{
   gtk_widget_destroy(GTK_WIDGET (this->window));

   this->window = NULL;
   this->drawing_area = NULL;
   this->parent = NULL;
}

void ChipsGtkWindow::show_window ()
{
   gtk_widget_realize (this->drawing_area);
   gtk_widget_show_all (this->window);
}

void ChipsGtkWindow::hide_window ()
{
   if (!GTK_WIDGET_MAPPED (this->window))
     return;
   gtk_widget_hide(this->window);
   gtk_widget_unrealize (this->drawing_area);
}

void *ChipsGtkWindow::get_xwindow ()
{
   return ((void *)(GDK_WINDOW_XWINDOW(this->drawing_area->window)));
}


void ChipsGtkWindow::set_title (const string &title)
{
   if (this->has_focus)
      gtk_window_set_title (GTK_WINDOW (this->window), title.c_str());
}

void ChipsGtkWindow::get_title (string &title)
{
   title = GTK_WINDOW (this->window)->title;
}


void ChipsGtkWindow::set_default_size (int width, int height)
{
   gtk_window_set_default_size (GTK_WINDOW(this->window), width, height);
}

void ChipsGtkWindow::get_default_size (int &width, int &height)
{
   gint *w;
   gint *h;

   gtk_window_get_default_size (GTK_WINDOW(this->window), w, h);

   width = *w;
   height = *h;
}


void ChipsGtkWindow::iconify ()
{
   gtk_window_iconify(GTK_WINDOW(this->window));
}

void ChipsGtkWindow::deiconify ()
{
   gtk_window_deiconify(GTK_WINDOW(this->window));
}


void ChipsGtkWindow::draw_rectangle (bool filled, int x, int y, int width, int height)
{
   gdk_draw_rectangle(this->drawing_area->window, this->drawing_area->style->fg_gc[0], filled, x, y, width, height);
}


void ChipsGtkWindow::set_fg_color (float red, float green, float blue)
{
   GdkColor clr;

   clr.red = static_cast<guint16> (65535 * red);
   clr.green = static_cast<guint16> (65535 * green);
   clr.blue = static_cast<guint16> (5535 * blue);
        
   gdk_gc_set_rgb_fg_color(this->drawing_area->style->fg_gc[0], &clr);
}



void ChipsGtkWindow::on_realize (GtkWidget *widget, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);

   gtk_window->parent->remap_show ();
}

gboolean ChipsGtkWindow::on_delete (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);

   if (!gtk_window->eventSafeFlag)
      return false;

   ChipsId cid;
   cid.set_window_id(gtk_window->parent->get_id());
   ((ChipsEngine*)gtk_window->parent->get_engine())->delete_window(&cid);

   return true;

}

void ChipsGtkWindow::on_focus (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);

   ChipsId cid (gtk_window->parent->get_id ().c_str ());
   double pos[2] = {0,0};
   ChipsEventObjectInfo unselect_info;
   ChipsEventObjectInfo select_info (cid, chips_window);

   // don't allow users to change the currency if in a pick
   if ((!gtk_window->parent->is_current ()) &&
       (event_manager->get_state () != ChipsEventManager::PickState))
   {
      event_manager->invoke_event (pos, ChipsEventManager::UnselectAllEvent, &select_info);
      event_manager->invoke_event (pos, ChipsEventManager::SelectEvent, &select_info);
   }
}

void ChipsGtkWindow::on_enter (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);
   gtk_window->has_focus = true;

}

void ChipsGtkWindow::on_leave (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);


   gtk_window->parent->set_title (gtk_window->parent->get_title ());
   gtk_window->has_focus = false;

}


void ChipsGtkWindow::on_expose (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);
   
   GdkEventExpose* evt = (GdkEventExpose*)event;
   gdk_window_begin_paint_region(gtk_window->window->window, evt->region);
   
   gtk_window->parent->set_redraw_ready();
   gtk_window->parent->redraw ();
 
   gdk_window_end_paint(gtk_window->window->window);

}

void ChipsGtkWindow::on_configure (GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGtkWindow *gtk_window = static_cast <ChipsGtkWindow *>(data);

   GdkEventConfigure *evt = (GdkEventConfigure*)event;
   guint width = evt->width;
   guint height = evt->height;

   if (!gtk_window->eventSafeFlag)
      return;

   gtk_window->parent->set_window_size(width,height);
   gtk_window->parent->update_display();
   gtk_window->parent->redraw ();
}

