// 
//  Copyright (C) 2011,2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiAxisFrame.hh"
#include "ChipsGuiColorWidget.hh"
#include "ChipsGuiLabelWidget.hh"
#include "ChipsGuiLevelsWidget.hh"
#include "ChipsGuiPivotPointWidget.hh"
#include "ChipsGuiMain.hh"
#include "ChipsGuiPropPane.hh"
#include "ChipsEngine.hh"
#include "ChipsId.hh"
#include "ChipsAxis.hh"
#include "ChipsDataSystem.hh"
#include "ChipsCmdManager.hh"
#include "config.h"

#include <sstream>

#if HAVE_IEEEFP_H 
#include <ieeefp.h>
#endif 

using namespace std;

extern ChipsGuiMain* chipsGui;
extern ChipsCmdManager* cmd_manager;

class ChipsSetEngineAxisPropCallback : public ChipsGuiPropFrameCallback
{
   public:
      ChipsSetEngineAxisPropCallback (int property, ChipsId *cid) 
      { 
         this->property = property; 
         this->axis_cid = cid;
      };
      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         if ((data != NULL) && eng && this->axis_cid)
            eng->set_property(this->axis_cid, data, this->property);
      }

   protected:
      int property;
      ChipsId *axis_cid;
};


class ChipsAxisMoveCallback : public ChipsGuiPropFrameCallback
{
   public:
      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         double *pos = static_cast<double *>(data);
         if ((pos != NULL) && eng)
            eng->move(&cid, chips_label, pos[0], pos[1], V_MOVE_ABSOLUTE);
      }
};

class ChipsAxisSetArbitraryCallback : public ChipsGuiPropFrameCallback
{
   public:
      class ChipsAxisSetArbitraryCallbackData
      {
         public:
            double *ticks;
            string *text;
            int num_ticks;
      };

      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         ChipsAxisSetArbitraryCallbackData *cb_data = static_cast<ChipsAxisSetArbitraryCallbackData *>(data);
         if (cb_data && eng)
            eng->set_arbitrary_tick_positions(&cid, cb_data->num_ticks, 
                                             cb_data->ticks, cb_data->text);
      }
};




// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiAxisFrame::ChipsGuiAxisFrame():ChipsGuiPropFrame(),ChipsGuiTooltipObject()
{

   _ticklabel_color = NULL;
   _minorgrid_color = NULL;
   _majorgrid_color = NULL;
   _minortick_color = NULL;
   _majortick_ticks = NULL;
   _majortick_ticksB = NULL;
   _majortick_color = NULL;
   _label = NULL;
   _color = NULL;

   _object_states = STATE_CHNG_PROP_AXS;  

}

// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiAxisFrame::~ChipsGuiAxisFrame()
{
//  gtk_widget_destroy(_frame);

   if (_ticklabel_color)
      delete _ticklabel_color;
   if (_minorgrid_color)
      delete _minorgrid_color;
   if (_majorgrid_color)
      delete _majorgrid_color;
   if (_minortick_color)
      delete _minortick_color;
   if (_majortick_ticks)
      delete _majortick_ticks;
   if (_majortick_ticksB)
      delete _majortick_ticksB;
   if (_majortick_color)
      delete _majortick_color;
   if (_ticklabel_alignment)
      delete _ticklabel_alignment;
   if (_label)
      delete _label;
   if (_color)
      delete _color;
}

void ChipsGuiAxisFrame::build_form()
{

   GtkWidget*   label;

   _frame = gtk_frame_new("Axis Properties");
   _nbook = gtk_notebook_new ();
   gtk_notebook_set_tab_pos (GTK_NOTEBOOK(_nbook), GTK_POS_TOP);
   gtk_notebook_popup_enable(GTK_NOTEBOOK(_nbook));
   gtk_notebook_set_scrollable(GTK_NOTEBOOK(_nbook), GTK_POLICY_AUTOMATIC);
   gtk_container_add(GTK_CONTAINER(_frame), _nbook);

   label = gtk_label_new("Axis");
   

   // setup axis notebook
   {
      GtkWidget*   frame1;
      GtkWidget*   table1;
      GtkWidget*   frame2;
      GtkWidget*   frame3;

      GtkWidget*   vbox;
      GtkWidget*   app_hbox1;
      GtkWidget*   app_hbox2;
      GtkWidget*   app_hbox3;
      GtkWidget*   app_hbox4;
      GtkWidget*   app_hbox5;
      GtkWidget*   app_frame;

      GtkWidget*   mode_hbox1;
      GtkWidget*   mode_hbox2;
      GtkWidget*   mode_hbox3;
      GtkWidget*   mode_vbox1;
      GtkWidget*   mode_vbox2;
      GtkWidget*   mode_hbox;
      GtkCellRenderer* cell1;


      vbox = gtk_vbox_new (FALSE, 3);
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), vbox, label);

      // setup appearance

      frame1 = gtk_frame_new("Appearance");
      gtk_frame_set_label_align(GTK_FRAME(frame1), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), frame1, FALSE, FALSE,5);

      table1 = gtk_table_new (3,4,FALSE);
      gtk_container_add(GTK_CONTAINER(frame1), table1);


      _color = new ChipsGuiColorWidget();
      _color->build_form ();

      _thick = gtk_spin_button_new_with_range(0.5, 10, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_thick), 1);

      _tickstyle = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_tickstyle), "inside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_tickstyle), "outside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_tickstyle), "centered");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_tickstyle),1);

      _majorgridvisible = gtk_toggle_button_new_with_label ("Show Major Grid");
      _minorgridvisible = gtk_toggle_button_new_with_label ("Show Minor Grid");

      
      _apply_current = gtk_radio_button_new_with_label (NULL, "X Axis");
      _apply_xy = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (_apply_current), "XY Axes");

      app_hbox5 = gtk_hbox_new (FALSE, 2);
      app_frame = gtk_frame_new ("");
      gtk_frame_set_shadow_type (GTK_FRAME(app_frame), GTK_SHADOW_ETCHED_OUT);
      label = gtk_label_new("Changes Applied To:");
      gtk_box_pack_start (GTK_BOX (app_hbox5), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox5), _apply_current, FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (app_hbox5), _apply_xy, FALSE, FALSE, 0);

      gtk_container_add(GTK_CONTAINER(app_frame), app_hbox5);
      gtk_table_attach(GTK_TABLE(table1), app_frame, 0, 4, 2, 3,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);


      app_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Color");
      gtk_box_pack_start (GTK_BOX (app_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox1), _color->base_widget (), FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(table1), app_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Tick Style");
      gtk_box_pack_start (GTK_BOX (app_hbox2), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox2), _tickstyle, TRUE, TRUE, 5);

      gtk_table_attach(GTK_TABLE(table1), app_hbox2, 2, 3, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Thickness");
      gtk_box_pack_start (GTK_BOX (app_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox3), _thick, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(table1), app_hbox3, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);



      app_hbox4 = gtk_hbutton_box_new ();
      gtk_button_box_set_layout (GTK_BUTTON_BOX (app_hbox4), GTK_BUTTONBOX_START);
      gtk_box_pack_start (GTK_BOX (app_hbox4), _majorgridvisible, FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (app_hbox4), _minorgridvisible, FALSE, FALSE, 0);

      gtk_table_attach(GTK_TABLE(table1), app_hbox4, 0, 2, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      // setup label
      
      frame2 = gtk_frame_new("Label");
      gtk_frame_set_label_align(GTK_FRAME(frame2), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), frame2, FALSE, FALSE,5);

      _label_text = gtk_entry_new ();
      gtk_container_add (GTK_CONTAINER(frame2), _label_text);


      // setup limits

      frame3 = gtk_frame_new("Tick Setup");
      gtk_frame_set_label_align(GTK_FRAME(frame3), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), frame3, TRUE, TRUE,5);

      mode_hbox = gtk_hbox_new (FALSE, 2);
      gtk_container_add(GTK_CONTAINER(frame3), mode_hbox);

      mode_vbox2 = gtk_vbox_new (FALSE, 3);
      gtk_box_pack_start(GTK_BOX(mode_hbox), mode_vbox2, FALSE, FALSE,5);

      GType types[] = {G_TYPE_DOUBLE, G_TYPE_STRING};
      _majortick_ticks = new ChipsGuiLevelsWidget(2,types);
      _majortick_ticks->build_form();
      _majortick_ticks->set_column_title ("Tick", 0);
      _majortick_ticks->set_column_title ("Value", 1);
      _majortick_ticks->set_min_level (1);


      _majortick_mode_list = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
      build_majortick_mode_list (_majortick_mode_list, ChipsAxisAlgorithmFactory::UNKNOWN);
      _majortick_mode = gtk_combo_box_new_with_model(GTK_TREE_MODEL (_majortick_mode_list));
      cell1 = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _majortick_mode ), cell1, FALSE );
      gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(_majortick_mode),cell1,"text",0,NULL);

      _majortick_interval = gtk_spin_button_new_with_range(0, G_MAXDOUBLE, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_interval), 5);

      _majortick_count = gtk_spin_button_new_with_range(2, 500, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_count), 0);


      mode_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Mode");
      gtk_box_pack_start (GTK_BOX (mode_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox1), _majortick_mode, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox1, FALSE, FALSE,5);

      mode_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Interval");
      gtk_box_pack_start (GTK_BOX (mode_hbox2), label, FALSE, FALSE, 3);
      gtk_box_pack_start (GTK_BOX (mode_hbox2), _majortick_interval, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox2, FALSE, FALSE,5);


      mode_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Num Ticks");
      gtk_box_pack_start (GTK_BOX (mode_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox3), _majortick_count, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox3, FALSE, FALSE,5);

      mode_vbox1 = gtk_vbox_new (FALSE, 2);
      label = gtk_label_new ("Manual Ticks");
      gtk_box_pack_start (GTK_BOX (mode_vbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_vbox1), _majortick_ticks->base_widget (), TRUE, TRUE, 5);
      gtk_box_pack_start(GTK_BOX(mode_hbox), mode_vbox1, TRUE, TRUE,5);

   }

   // setup the label page
   {
      _label = new ChipsGuiLabelWidget();
      _label->build_form ();
      _label->set_position_type (ChipsGuiLabelWidget::FLOAT,
                              ChipsGuiLabelWidget::INT);

      label = gtk_label_new("Label");
      _label_page_num = gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), _label->base_widget (), label);

   }

   // setup the tick label page
   {
      GtkWidget*   table;
      GtkWidget*   fontframe;
      GtkWidget*   positionframe;
      GtkWidget*   formatframe;
      GtkWidget*   font_hbox1;
      GtkWidget*   font_hbox2;
      GtkCellRenderer* cell1;
      GtkCellRenderer* cell2;
      GtkCellRenderer* cell3;

      GtkWidget*   pos_table;
      GtkWidget*   pos_hbox1;
      GtkWidget*   pos_hbox2;
      GtkWidget*   pos_hbox3;

      GtkWidget*   format_table;
      GtkWidget*   format_text_hbox;
     
      GtkWidget*   vis_hbox;
    

      label = gtk_label_new ("Tick Label");
      table = gtk_table_new (4,2,FALSE);
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), table, label);

      fontframe = gtk_frame_new("Font");
      gtk_frame_set_label_align(GTK_FRAME(fontframe), .0, .5);
      gtk_table_attach(GTK_TABLE(table), fontframe, 0, 2, 0, 1,
                       (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), GTK_SHRINK,5,5);

      font_hbox1 = gtk_hbox_new (FALSE, 4);
      gtk_container_add(GTK_CONTAINER(fontframe), font_hbox1);

      _ticklabel_font = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL( ChipsGuiPropPane::font_info.get_font_list ()), 0);
      gtk_box_pack_start (GTK_BOX (font_hbox1), _ticklabel_font, FALSE, FALSE, 2);

      cell1 = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _ticklabel_font ), cell1, FALSE );

      _ticklabel_size = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL( ChipsGuiPropPane::font_info.get_size_list ()), 0);
      gtk_widget_set_size_request (_ticklabel_size, 65, -1);
      gtk_box_pack_start (GTK_BOX (font_hbox1), _ticklabel_size, FALSE, FALSE, 2);
      cell2 = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _ticklabel_size ), cell2, FALSE );

      font_hbox2 = gtk_hbutton_box_new ();
      gtk_button_box_set_layout (GTK_BUTTON_BOX(font_hbox2), GTK_BUTTONBOX_CENTER);
      gtk_box_pack_start (GTK_BOX (font_hbox1), font_hbox2, FALSE, FALSE, 2);

      _ticklabel_bold = gtk_toggle_button_new_with_label (GTK_STOCK_BOLD);
      gtk_button_set_use_stock (GTK_BUTTON(_ticklabel_bold), TRUE);
      gtk_box_pack_start (GTK_BOX (font_hbox1), _ticklabel_bold, FALSE, FALSE, 0);


      _ticklabel_italic = gtk_toggle_button_new_with_label (GTK_STOCK_ITALIC);
      gtk_button_set_use_stock (GTK_BUTTON(_ticklabel_italic), TRUE);
      gtk_box_pack_start (GTK_BOX (font_hbox1), _ticklabel_italic, FALSE, FALSE, 0);

      _ticklabel_color = new ChipsGuiColorWidget();
      _ticklabel_color->build_form ();
      gtk_box_pack_start (GTK_BOX (font_hbox1), _ticklabel_color->base_widget (), FALSE, FALSE, 5);

      // add position frame
      positionframe = gtk_frame_new("Position");
      gtk_frame_set_label_align(GTK_FRAME(positionframe), .0, .5);
      gtk_table_attach(GTK_TABLE(table), positionframe, 0, 1, 1, 2, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),5, 5);

      pos_table = gtk_table_new (4, 2, FALSE);
      gtk_container_add(GTK_CONTAINER(positionframe), pos_table);

      pos_hbox1 = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(pos_table), pos_hbox1, 0,1,0,1, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), GTK_SHRINK,5, 5);

      label = gtk_label_new("Offset");
      gtk_box_pack_start(GTK_BOX(pos_hbox1), label, FALSE, FALSE,5);

      _ticklabel_offset = gtk_spin_button_new_with_range (-9999, 9999,1.0);
      gtk_box_pack_start(GTK_BOX(pos_hbox1), _ticklabel_offset, FALSE, FALSE,5);
      gtk_spin_button_set_digits (GTK_SPIN_BUTTON(_ticklabel_offset), 0);

      pos_hbox2 = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(pos_table), pos_hbox2, 0,1,1,2, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), GTK_SHRINK,5, 5);

      label = gtk_label_new("Angle");
      gtk_box_pack_start(GTK_BOX(pos_hbox2), label, FALSE, FALSE,5);

      _ticklabel_angle = gtk_spin_button_new_with_range (0, 359.99999, 1);
      gtk_box_pack_start(GTK_BOX(pos_hbox2), _ticklabel_angle, FALSE, FALSE,5);
      gtk_spin_button_set_wrap (GTK_SPIN_BUTTON(_ticklabel_angle), TRUE);

      pos_hbox3 = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(pos_table), pos_hbox3, 0,1,2,3, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), GTK_SHRINK,5, 5);
      label = gtk_label_new("Style");
      gtk_box_pack_start(GTK_BOX(pos_hbox3), label, FALSE, FALSE,5);

      _ticklabel_style = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_ticklabel_style), "inside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_ticklabel_style), "outside");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_ticklabel_style),1);
      gtk_box_pack_start(GTK_BOX(pos_hbox3), _ticklabel_style, FALSE, FALSE,5);

      // setup alignment
      _ticklabel_alignment = new ChipsGuiPivotPointWidget();
      _ticklabel_alignment->build_form (true);
      gtk_table_attach(GTK_TABLE(pos_table), _ticklabel_alignment->base_widget (), 1,2,0,4, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),5, 5);


      // setup format
      formatframe = gtk_frame_new("Format");
      gtk_frame_set_label_align(GTK_FRAME(formatframe), .0, .5);
      gtk_table_attach(GTK_TABLE(table), formatframe, 0, 1, 2, 3, (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),5, 5);


      format_table =  gtk_table_new (2, 1, FALSE);
      gtk_container_add(GTK_CONTAINER(formatframe), format_table);
      
      format_text_hbox = gtk_hbox_new (FALSE, 1);
      label = gtk_label_new("Custom");
      _ticklabel_format_custom= gtk_entry_new ();
      gtk_widget_set_size_request (_ticklabel_format_custom, 80, -1);
      gtk_box_pack_start(GTK_BOX(format_text_hbox), label, FALSE, FALSE,5);
      gtk_box_pack_start(GTK_BOX(format_text_hbox), _ticklabel_format_custom, FALSE, FALSE,5);

      _format_list = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
      build_format_list (_format_list);
      _ticklabel_format_box = gtk_combo_box_new_with_model(GTK_TREE_MODEL (_format_list));
      cell3 = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _ticklabel_format_box ), cell3, FALSE );
      gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(_ticklabel_format_box),cell3,"markup",0,NULL); 

      gtk_combo_box_set_active(GTK_COMBO_BOX(_ticklabel_format_box),1);
      gtk_table_attach(GTK_TABLE(format_table), _ticklabel_format_box, 0,1,0,1, GTK_SHRINK, GTK_SHRINK,5, 5);
      gtk_table_attach(GTK_TABLE(format_table), format_text_hbox, 0,1,1,2, GTK_SHRINK, GTK_SHRINK,5, 5);

      vis_hbox = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(table), vis_hbox, 0, 1, 3, 4, GTK_FILL, GTK_SHRINK,5, 5);

      _ticklabel_visible = gtk_check_button_new_with_label ("Visible");
      gtk_box_pack_start (GTK_BOX (vis_hbox), _ticklabel_visible, FALSE, FALSE, 0);

   }

   // setup majortick page
   {
      GtkWidget*   app_frame;
      GtkWidget*   vbox;
      GtkWidget*   app_table;
      GtkWidget*   app_hbox1;
      GtkWidget*   app_hbox2;
      GtkWidget*   app_hbox3;
      GtkWidget*   app_hbox5;
      GtkWidget*   vis_hbox;

      GtkWidget*   mode_hbox;
      GtkWidget*   mode_hbox1;
      GtkWidget*   mode_hbox2;
      GtkWidget*   mode_hbox3;
      GtkWidget*   mode_vbox1;
      GtkWidget*   mode_vbox2;


      label = gtk_label_new ("Major Tick");
      vbox = gtk_vbox_new (FALSE, 3);
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), vbox, label);

      app_frame = gtk_frame_new("Appearance");
      gtk_frame_set_label_align(GTK_FRAME(app_frame), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), app_frame, FALSE, FALSE,5);

      app_table = gtk_table_new (2,2,FALSE);
      gtk_container_add(GTK_CONTAINER(app_frame), app_table);

      _majortick_color = new ChipsGuiColorWidget();
      _majortick_color->build_form ();

      _majortick_thick = gtk_spin_button_new_with_range(0.5, 10, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_thick), 1);

      _majortick_length = gtk_spin_button_new_with_range(0, 100000, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_length), 1);

      _majortick_visible = gtk_check_button_new_with_label ("Visible");

      _majortick_style = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majortick_style), "inside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majortick_style), "outside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majortick_style), "centered");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_majortick_style),1);

      app_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Color");
      gtk_box_pack_start (GTK_BOX (app_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox1), _majortick_color->base_widget (), FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Thickness");
      gtk_box_pack_start (GTK_BOX (app_hbox2), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox2), _majortick_thick, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox2, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Length");
      gtk_box_pack_start (GTK_BOX (app_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox3), _majortick_length, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox3, 0, 1, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox5 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Style");
      gtk_box_pack_start (GTK_BOX (app_hbox5), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox5), _majortick_style, TRUE, TRUE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox5, 1, 2, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      // add majortick mode setup clone from first tab
      GtkWidget* frame5 = gtk_frame_new("Tick Setup");
      gtk_frame_set_label_align(GTK_FRAME(frame5), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), frame5, TRUE, TRUE,5);

      mode_hbox = gtk_hbox_new (FALSE, 2);
      gtk_container_add(GTK_CONTAINER(frame5), mode_hbox);

      mode_vbox2 = gtk_vbox_new (FALSE, 3);
      gtk_box_pack_start(GTK_BOX(mode_hbox), mode_vbox2, FALSE, FALSE,5);

      GType types[] = {G_TYPE_DOUBLE, G_TYPE_STRING};
      _majortick_ticksB = new ChipsGuiLevelsWidget(2,types);
      _majortick_ticksB->build_form();
      _majortick_ticksB->set_column_title ("Tick", 0);
      _majortick_ticksB->set_column_title ("Value", 1);
      _majortick_ticksB->set_min_level (1);

      _majortick_mode_listB = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
      build_majortick_mode_list (_majortick_mode_listB, ChipsAxisAlgorithmFactory::UNKNOWN);
      _majortick_modeB = gtk_combo_box_new_with_model(GTK_TREE_MODEL (_majortick_mode_listB));
      GtkCellRenderer* mycell = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _majortick_modeB ), mycell, FALSE );
      gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(_majortick_modeB),mycell,"text",0,NULL);

      _majortick_intervalB = gtk_spin_button_new_with_range(0, G_MAXDOUBLE, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_intervalB), 5);

      _majortick_countB = gtk_spin_button_new_with_range(2, 500, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majortick_countB), 0);

      mode_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Mode");
      gtk_box_pack_start (GTK_BOX (mode_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox1), _majortick_modeB, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox1, FALSE, FALSE,5);

      mode_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Interval");
      gtk_box_pack_start (GTK_BOX (mode_hbox2), label, FALSE, FALSE, 3);
      gtk_box_pack_start (GTK_BOX (mode_hbox2), _majortick_intervalB, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox2, FALSE, FALSE,5);

      mode_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Num Ticks");
      gtk_box_pack_start (GTK_BOX (mode_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox3), _majortick_countB, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(mode_vbox2), mode_hbox3, FALSE, FALSE,5);

      mode_vbox1 = gtk_vbox_new (FALSE, 2);
      label = gtk_label_new ("Manual Ticks");
      gtk_box_pack_start (GTK_BOX (mode_vbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_vbox1), _majortick_ticksB->base_widget (), TRUE, TRUE, 5);
      gtk_box_pack_start(GTK_BOX(mode_hbox), mode_vbox1, TRUE, TRUE,5);

      // visible
      vis_hbox = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox), vis_hbox, FALSE, FALSE,5);
      gtk_box_pack_start (GTK_BOX (vis_hbox), _majortick_visible, FALSE, FALSE, 0);
   }

   // setup minortick page
   {
      GtkWidget*   app_frame;
      GtkWidget*   mode_frame;
      GtkWidget*   vbox;
      GtkWidget*   app_table;
      GtkWidget*   mode_table;

      GtkWidget*   app_hbox1;
      GtkWidget*   app_hbox2;
      GtkWidget*   app_hbox3;
      GtkWidget*   app_hbox5;

      GtkWidget*   mode_hbox1;
      GtkWidget*   mode_hbox2;
      GtkWidget*   mode_hbox3;
      GtkWidget*   vis_hbox;

      GtkCellRenderer* cell1;

      label = gtk_label_new ("Minor Tick");
      vbox = gtk_vbox_new (FALSE, 3);
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), vbox, label);

      app_frame = gtk_frame_new("Appearance");
      gtk_frame_set_label_align(GTK_FRAME(app_frame), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), app_frame, FALSE, FALSE,5);

      app_table = gtk_table_new (3,2,FALSE);
      gtk_container_add(GTK_CONTAINER(app_frame), app_table);

      _minortick_color = new ChipsGuiColorWidget();
      _minortick_color->build_form ();

      _minortick_thick = gtk_spin_button_new_with_range(0.5, 10, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_minortick_thick), 1);

      _minortick_length = gtk_spin_button_new_with_range(0, 100000, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_minortick_length), 1);

      _minortick_visible = gtk_check_button_new_with_label ("Visible");

      _minortick_style = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minortick_style), "inside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minortick_style), "outside");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minortick_style), "centered");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_minortick_style),1);

      app_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Color");
      gtk_box_pack_start (GTK_BOX (app_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox1), _minortick_color->base_widget (), FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Thickness");
      gtk_box_pack_start (GTK_BOX (app_hbox2), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox2), _minortick_thick, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox2, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      app_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Length");
      gtk_box_pack_start (GTK_BOX (app_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox3), _minortick_length, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox3, 0, 1, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);


      app_hbox5 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Style");
      gtk_box_pack_start (GTK_BOX (app_hbox5), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (app_hbox5), _minortick_style, TRUE, TRUE, 5);
      gtk_table_attach(GTK_TABLE(app_table), app_hbox5, 1, 2, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);


      // modes

      mode_frame = gtk_frame_new("Mode Setup");
      gtk_frame_set_label_align(GTK_FRAME(mode_frame), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), mode_frame, FALSE, FALSE,5);

      mode_table = gtk_table_new (2,2,FALSE);
      gtk_container_add(GTK_CONTAINER(mode_frame), mode_table);

      _minortick_mode_list = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
      build_minortick_mode_list (_minortick_mode_list, ChipsAxisAlgorithmFactory::UNKNOWN);
      _minortick_mode = gtk_combo_box_new_with_model(GTK_TREE_MODEL (_minortick_mode_list));
      cell1 = gtk_cell_renderer_text_new();
      gtk_cell_layout_pack_start( GTK_CELL_LAYOUT( _minortick_mode ), cell1, FALSE );
      gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(_minortick_mode),cell1,"text",0,NULL);
      gtk_combo_box_set_button_sensitivity (GTK_COMBO_BOX(_minortick_mode), GTK_SENSITIVITY_AUTO);

      _minortick_interval = gtk_spin_button_new_with_range(0, G_MAXDOUBLE, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_minortick_interval), 5);

      _minortick_count = gtk_spin_button_new_with_range(1, 500, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_minortick_count), 0);


      mode_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Mode");
      gtk_box_pack_start (GTK_BOX (mode_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox1), _minortick_mode, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(mode_table), mode_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|
GTK_FILL), 5, 5);

      mode_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Interval");
      gtk_box_pack_start (GTK_BOX (mode_hbox2), label, FALSE, FALSE, 3);
      gtk_box_pack_start (GTK_BOX (mode_hbox2), _minortick_interval, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(mode_table), mode_hbox2, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);


      mode_hbox3 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Num Ticks");
      gtk_box_pack_start (GTK_BOX (mode_hbox3), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (mode_hbox3), _minortick_count, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(mode_table), mode_hbox3, 1, 2, 1, 2,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      // visible
      vis_hbox = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox), vis_hbox, FALSE, FALSE,5);
      gtk_box_pack_start (GTK_BOX (vis_hbox), _minortick_visible, FALSE, FALSE, 0);

   }

   // setup grid page
   {
      GtkWidget*   major_frame;
      GtkWidget*   minor_frame;
      GtkWidget*   vbox;
      GtkWidget*   major_table;
      GtkWidget*   minor_table;

      GtkWidget*   major_hbox1;
      GtkWidget*   major_hbox2;
      GtkWidget*   major_hbox5;

      GtkWidget*   minor_hbox1;
      GtkWidget*   minor_hbox2;
      GtkWidget*   minor_hbox5;
      GtkWidget*   vis_hbox1;
      GtkWidget*   vis_hbox2;


      label = gtk_label_new ("Grids");
      vbox = gtk_vbox_new (FALSE, 2);
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), vbox, label);

      major_frame = gtk_frame_new("Major Grids");
      gtk_frame_set_label_align(GTK_FRAME(major_frame), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), major_frame, FALSE, FALSE,5);

      major_table = gtk_table_new (2,3,FALSE);
      gtk_container_add(GTK_CONTAINER(major_frame), major_table);


      _majorgrid_color = new ChipsGuiColorWidget();
      _majorgrid_color->build_form ();

      _majorgrid_thick = gtk_spin_button_new_with_range(0.5, 10, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_majorgrid_thick), 1);

      _majorgrid_visible = gtk_check_button_new_with_label ("Visible");

      _majorgrid_style = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "no line");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "solid");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "dot");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "dot longdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "dot shortdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "longdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "shortdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_majorgrid_style), "shortdash longdash");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_majorgrid_style),1);

      major_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Color");
      gtk_box_pack_start (GTK_BOX (major_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (major_hbox1), _majorgrid_color->base_widget (), FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(major_table), major_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      major_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Thickness");
      gtk_box_pack_start (GTK_BOX (major_hbox2), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (major_hbox2), _majorgrid_thick, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(major_table), major_hbox2, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      vis_hbox1 = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(major_table), vis_hbox1, 0, 1, 1, 2,
         GTK_FILL, GTK_SHRINK, 5, 5);
      gtk_box_pack_start (GTK_BOX (vis_hbox1), _majorgrid_visible, FALSE, FALSE, 0);


      major_hbox5 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Style");
      gtk_box_pack_start (GTK_BOX (major_hbox5), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (major_hbox5), _majorgrid_style, TRUE, TRUE, 5);
      gtk_table_attach(GTK_TABLE(major_table), major_hbox5, 2, 3, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      // minor grids

      minor_frame = gtk_frame_new("Minor Grids");
      gtk_frame_set_label_align(GTK_FRAME(minor_frame), .0, .5);
      gtk_box_pack_start(GTK_BOX(vbox), minor_frame, FALSE, FALSE,5);

      minor_table = gtk_table_new (2,3,FALSE);
      gtk_container_add(GTK_CONTAINER(minor_frame), minor_table);


      _minorgrid_color = new ChipsGuiColorWidget();
      _minorgrid_color->build_form ();

      _minorgrid_thick = gtk_spin_button_new_with_range(0.5, 10, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_minorgrid_thick), 1);

      _minorgrid_visible = gtk_check_button_new_with_label ("Visible");

      _minorgrid_style = gtk_combo_box_new_text();
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "no line");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "solid");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "dot");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "dot longdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "dot shortdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "longdash");
      gtk_combo_box_append_text(GTK_COMBO_BOX(_minorgrid_style), "shortdash longdash");
      gtk_combo_box_set_active(GTK_COMBO_BOX(_minorgrid_style),1);

      minor_hbox1 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Color");
      gtk_box_pack_start (GTK_BOX (minor_hbox1), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (minor_hbox1), _minorgrid_color->base_widget (), FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(minor_table), minor_hbox1, 0, 1, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      minor_hbox2 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Thickness");
      gtk_box_pack_start (GTK_BOX (minor_hbox2), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (minor_hbox2), _minorgrid_thick, FALSE, FALSE, 5);
      gtk_table_attach(GTK_TABLE(minor_table), minor_hbox2, 1, 2, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

      vis_hbox2 = gtk_hbox_new (FALSE, 2);
      gtk_table_attach(GTK_TABLE(minor_table), vis_hbox2, 0, 1, 1, 2,
         GTK_FILL, GTK_SHRINK, 5, 5);
      gtk_box_pack_start (GTK_BOX (vis_hbox2), _minorgrid_visible, FALSE, FALSE, 0);



      minor_hbox5 = gtk_hbox_new (FALSE, 2);
      label = gtk_label_new("Style");
      gtk_box_pack_start (GTK_BOX (minor_hbox5), label, FALSE, FALSE, 5);
      gtk_box_pack_start (GTK_BOX (minor_hbox5), _minorgrid_style, TRUE, TRUE, 5);
      gtk_table_attach(GTK_TABLE(minor_table), minor_hbox5, 2, 3, 0, 1,
         (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), (GtkAttachOptions)(GTK_EXPAND|GTK_FILL), 5, 5);

   }

   // Setup General Page
   {
      GtkWidget*   hbox1;
      GtkWidget*   hbox2;
      GtkWidget*   hbox3;
      GtkWidget*   hbox4;
      GtkWidget*   vbox1;
      GtkWidget*   padding_hbox;


      
      vbox1 = gtk_vbox_new (FALSE, 2);
      label = gtk_label_new("General");
      gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), vbox1, label);

      hbox1 = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE,10);
      hbox2 = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox1), hbox2, FALSE, FALSE,10);
      hbox3 = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox1), hbox3, FALSE, FALSE,10);
      hbox4 = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox1), hbox4, FALSE, FALSE,10);



      _depth = gtk_spin_button_new_with_range(1, G_MAXINT, 1);
      gtk_spin_button_set_digits(GTK_SPIN_BUTTON(_depth), 0);

      label = gtk_label_new("Depth");
      gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE,10);
      gtk_box_pack_start(GTK_BOX(hbox1), _depth, FALSE, FALSE,10);

      _visible = gtk_check_button_new();

      label = gtk_label_new("Visible");
      gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE,10);
      gtk_box_pack_start(GTK_BOX(hbox2), _visible, FALSE, FALSE,10);

      _automin = gtk_check_button_new ();
      label = gtk_label_new("Enable Auto Min");
      gtk_box_pack_start(GTK_BOX(hbox3), label, FALSE, FALSE,10);
      gtk_box_pack_start(GTK_BOX(hbox3), _automin, FALSE, FALSE,10);

      _automax = gtk_check_button_new ();
      label = gtk_label_new("Enable Auto Max");
      gtk_box_pack_start(GTK_BOX(hbox4), label, FALSE, FALSE,10);
      gtk_box_pack_start(GTK_BOX(hbox4), _automax, FALSE, FALSE,10);

      _padding = gtk_spin_button_new_with_range(0, 100, .01);

      padding_hbox = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start(GTK_BOX(vbox1), padding_hbox, FALSE, FALSE,10);

      label = gtk_label_new("Limits Padding");
      gtk_box_pack_start(GTK_BOX(padding_hbox), label, FALSE, FALSE, 5);
      gtk_box_pack_start(GTK_BOX(padding_hbox), _padding, FALSE, FALSE, 5);
      

   }

   g_signal_connect(GTK_OBJECT(_nbook), "switch-page",  G_CALLBACK(change_tab_callback),  (void*)this);
   _color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (color_callback, this));
   _thick_id = g_signal_connect(G_OBJECT(_thick), "value-changed", G_CALLBACK(thick_callback), (void *)this);
   _tickstyle_id = g_signal_connect(G_OBJECT(_tickstyle), "changed", G_CALLBACK(tickstyle_callback), (void*)this);
   _automin_id = g_signal_connect(G_OBJECT(_automin), "toggled", G_CALLBACK(automin_callback), (void*)this);
   _automax_id = g_signal_connect(G_OBJECT(_automax), "toggled", G_CALLBACK(automax_callback), (void*)this);
   _padding_id = g_signal_connect(G_OBJECT(_padding), "value-changed", G_CALLBACK(padding_callback), (void*)this);
   _label_text_id = g_signal_connect(G_OBJECT(_label_text), "changed", G_CALLBACK(label_text_callback), (void*)this);
   _majorgridvisible_id = g_signal_connect(G_OBJECT(_majorgridvisible), "toggled", G_CALLBACK(majorgrid_visible_callback), (void*)this);
   _minorgridvisible_id = g_signal_connect(G_OBJECT(_minorgridvisible), "toggled", G_CALLBACK(minorgrid_visible_callback), (void*)this);
   _apply_current_id = g_signal_connect(G_OBJECT(_apply_current), "toggled", G_CALLBACK(apply_axis_callback), (void*)this);
   _apply_xy_id = g_signal_connect(G_OBJECT(_apply_xy), "toggled", G_CALLBACK(apply_axis_callback), (void*)this);

   _label->add_callback ("change_color", new ChipsGuiLabelWidgetCallback (label_color_change_callback, this));
   _label->add_callback ("change_angle", new ChipsGuiLabelWidgetCallback (label_angle_change_callback, this));
   _label->add_callback ("change_text", new ChipsGuiLabelWidgetCallback (label_text_change_callback, this));
   _label->add_callback ("change_fontstyle", new ChipsGuiLabelWidgetCallback (label_fontstyle_change_callback, this));
   _label->add_callback ("change_font", new ChipsGuiLabelWidgetCallback (label_font_change_callback, this));
   _label->add_callback ("change_align", new ChipsGuiLabelWidgetCallback (label_align_change_callback, this));
   _label->add_callback ("change_size", new ChipsGuiLabelWidgetCallback (label_size_change_callback, this));
   _label->add_callback ("change_position", new ChipsGuiLabelWidgetCallback (label_position_change_callback, this));
   _label->add_callback ("change_visible", new ChipsGuiLabelWidgetCallback (label_visible_change_callback, this));

   _majortick_color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (majortick_color_callback, this));
   _majortick_ticks->add_callback ("change_levels", new ChipsGuiLevelsWidgetCallback (majortick_ticks_callback, this));
   _majortick_ticksB->add_callback ("change_levels", new ChipsGuiLevelsWidgetCallback (majortick_ticks_callback, this));
   _majortick_ticks->add_callback ("delete_levels", new ChipsGuiLevelsWidgetCallback (majortick_ticks_callback, this));
   _majortick_ticksB->add_callback ("delete_levels", new ChipsGuiLevelsWidgetCallback (majortick_ticks_callback, this));
   _majortick_thick_id = g_signal_connect(G_OBJECT(_majortick_thick), "value-changed", G_CALLBACK(majortick_thick_callback), (void*)this);
   _majortick_style_id = g_signal_connect(G_OBJECT(_majortick_style), "changed", G_CALLBACK(majortick_style_callback), (void*)this);
   _majortick_length_id = g_signal_connect(G_OBJECT(_majortick_length), "value-changed", G_CALLBACK(majortick_length_callback), (void*)this);
   _majortick_visible_id = g_signal_connect(G_OBJECT(_majortick_visible), "toggled", G_CALLBACK(majortick_visible_callback), (void*)this);
   _majortick_mode_id = g_signal_connect(G_OBJECT(_majortick_mode), "changed", G_CALLBACK(majortick_mode_callback), (void*)this);
   _majortick_modeB_id = g_signal_connect(G_OBJECT(_majortick_modeB), "changed", G_CALLBACK(majortick_mode_callback), (void*)this);
   _majortick_interval_id = g_signal_connect(G_OBJECT(_majortick_interval), "value-changed", G_CALLBACK(majortick_interval_callback), (void*)this);
   _majortick_intervalB_id = g_signal_connect(G_OBJECT(_majortick_intervalB), "value-changed", G_CALLBACK(majortick_interval_callback), (void*)this);
   _majortick_count_id = g_signal_connect(G_OBJECT(_majortick_count), "value-changed", G_CALLBACK(majortick_count_callback), (void*)this);
   _majortick_countB_id = g_signal_connect(G_OBJECT(_majortick_countB), "value-changed", G_CALLBACK(majortick_count_callback), (void*)this);

   _minortick_color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (minortick_color_callback, this));
   _minortick_thick_id = g_signal_connect(G_OBJECT(_minortick_thick), "value-changed", G_CALLBACK(minortick_thick_callback), (void*)this);
   _minortick_style_id = g_signal_connect(G_OBJECT(_minortick_style), "changed", G_CALLBACK(minortick_style_callback), (void*)this);
   _minortick_length_id = g_signal_connect(G_OBJECT(_minortick_length), "value-changed", G_CALLBACK(minortick_length_callback), (void*)this);
   _minortick_visible_id = g_signal_connect(G_OBJECT(_minortick_visible), "toggled", G_CALLBACK(minortick_visible_callback), (void*)this);
   _minortick_mode_id = g_signal_connect(G_OBJECT(_minortick_mode), "changed", G_CALLBACK(minortick_mode_callback), (void*)this);
   _minortick_interval_id = g_signal_connect(G_OBJECT(_minortick_interval), "value-changed", G_CALLBACK(minortick_interval_callback), (void*)this);
   _minortick_count_id = g_signal_connect(G_OBJECT(_minortick_count), "value-changed", G_CALLBACK(minortick_count_callback), (void*)this);

   _majorgrid_color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (majorgrid_color_callback, this));
   _majorgrid_thick_id = g_signal_connect(G_OBJECT(_majorgrid_thick), "value-changed", G_CALLBACK(majorgrid_thick_callback), (void*)this);
   _majorgrid_style_id = g_signal_connect(G_OBJECT(_majorgrid_style), "changed", G_CALLBACK(majorgrid_style_callback), (void*)this);
   _majorgrid_visible_id = g_signal_connect(G_OBJECT(_majorgrid_visible), "toggled", G_CALLBACK(majorgrid_visible_callback), (void*)this);

   _minorgrid_color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (minorgrid_color_callback, this));
   _minorgrid_thick_id = g_signal_connect(G_OBJECT(_minorgrid_thick), "value-changed", G_CALLBACK(minorgrid_thick_callback), (void*)this);
   _minorgrid_style_id = g_signal_connect(G_OBJECT(_minorgrid_style), "changed", G_CALLBACK(minorgrid_style_callback), (void*)this);
   _minorgrid_visible_id = g_signal_connect(G_OBJECT(_minorgrid_visible), "toggled", G_CALLBACK(minorgrid_visible_callback), (void*)this);

   _ticklabel_color->add_callback ("change_color", new ChipsGuiColorWidgetCallback (ticklabel_color_callback, this));
   _ticklabel_font_id = g_signal_connect(G_OBJECT(_ticklabel_font), "changed", G_CALLBACK(ticklabel_font_callback), (void*)this);
   _ticklabel_bold_id = g_signal_connect(G_OBJECT(_ticklabel_bold), "toggled", G_CALLBACK(ticklabel_fontstyle_callback), (void*)this);
   _ticklabel_italic_id = g_signal_connect(G_OBJECT(_ticklabel_italic), "toggled", G_CALLBACK(ticklabel_fontstyle_callback), (void*)this);
   _ticklabel_size_id = g_signal_connect(G_OBJECT(_ticklabel_size), "changed", G_CALLBACK(ticklabel_size_callback), (void*)this);
   _ticklabel_angle_change_id = g_signal_connect(G_OBJECT(_ticklabel_angle), "value-changed", G_CALLBACK(ticklabel_angle_callback), (void*)this);
   _ticklabel_offset_id = g_signal_connect(G_OBJECT(_ticklabel_offset), "value-changed", G_CALLBACK(ticklabel_offset_callback), (void*)this);
   _ticklabel_style_id = g_signal_connect(G_OBJECT(_ticklabel_style), "changed", G_CALLBACK(ticklabel_style_callback), (void*)this);

   _ticklabel_alignment->add_callback ("change_pivot", new ChipsGuiPivotPointWidgetCallback (ticklabel_align_callback, this));
   _ticklabel_alignment->add_callback ("enable_xauto", new ChipsGuiPivotPointWidgetCallback (ticklabel_halign_auto_callback, this));
   _ticklabel_alignment->add_callback ("enable_yauto", new ChipsGuiPivotPointWidgetCallback (ticklabel_valign_auto_callback, this));

   _ticklabel_visible_change_id = g_signal_connect(G_OBJECT(_ticklabel_visible), "toggled", G_CALLBACK(ticklabel_visible_callback), (void*)this);

   _ticklabel_format_custom_id = g_signal_connect(G_OBJECT(_ticklabel_format_custom), "activate", G_CALLBACK(ticklabel_format_custom_callback), (void*)this);
   _ticklabel_format_custom_focusout_id = g_signal_connect(G_OBJECT(_ticklabel_format_custom), "focus-out-event", G_CALLBACK(ticklabel_format_custom_focusout_callback), (void*)this);
   _ticklabel_format_box_id = g_signal_connect(G_OBJECT(_ticklabel_format_box), "changed", G_CALLBACK(ticklabel_format_callback), (void*)this);

   _depth_id = g_signal_connect(G_OBJECT(_depth), "value-changed", G_CALLBACK(depth_callback), (void*)this);
   _visible_id = g_signal_connect(G_OBJECT(_visible), "toggled", G_CALLBACK(visible_callback), (void*)this);

   setup_tooltip(_color->base_widget(),"Sets all color properties to the specified value");
   setup_tooltip(_thick,"Sets all line thickness properties to the specified value");
   setup_tooltip(_tickstyle,"Controls the positioning of the major and minor tick marks");
   setup_tooltip(_majorgridvisible,"Displays grid lines at the major tick marks");
   setup_tooltip(_minorgridvisible,"Display grid lines at the minor tick marks"); 
   setup_tooltip(_apply_current,"Apply changes to the currently selected axis");
   setup_tooltip(_apply_xy,"apply the changes to the selected axis and the other axis which defines the current coordinate system");


   setup_tooltip(_ticklabel_color->base_widget(),"Sets the color to use for the tick labels");
   setup_tooltip(_ticklabel_font,"Sets the font to use for the tick labels");
   setup_tooltip(_ticklabel_size,"Sets the point size of the tick labels");
   setup_tooltip(_ticklabel_bold,"Specifies whether tick labels should use a bold fontstyle");
   setup_tooltip(_ticklabel_italic,"Specifies whether the tick labels should be italicized");
   setup_tooltip(_ticklabel_style,"Specifies whether the tick labels should appear inside/outside of the plot");
   setup_tooltip(_ticklabel_format_custom,"Print format specification for tick labels");
   setup_tooltip(_ticklabel_angle,"Specifies the rotation angle for tick labels");
   setup_tooltip(_ticklabel_offset,"The offset in pixels from the axis base to the tick labels");
   setup_tooltip(_ticklabel_format_box,"Select pre-canned tick label format specifications"); 
   setup_tooltip(_ticklabel_visible,"Should the tick label be displayed or hidden");
   setup_tooltip(_majortick_color->base_widget(),"Sets the color to use for the major ticks");
   setup_tooltip(_majortick_thick,"Sets the thickness of the major ticks");
   setup_tooltip(_majortick_visible,"Should the major ticks be displayed or hidden");
   setup_tooltip(_majortick_length,"Sets the length of the major tick marks");
   setup_tooltip(_majortick_style,"Controls the positioning of major tick marks");
   setup_tooltip(_minortick_color->base_widget(),"Sets the color to use for the minor ticks");
   setup_tooltip(_minortick_thick,"Sets the thickness of the minor ticks");
   setup_tooltip(_minortick_visible,"Should the minor ticks be displayed or hidden");
   setup_tooltip(_minortick_length,"Sets the length of the minor tick marks");
   setup_tooltip(_minortick_style,"Controls the positioning of minor tick marks");
   setup_tooltip(_majortick_mode,"Specifies the mode used to determine tick mark spacing");
   setup_tooltip(_majortick_modeB,"Specifies the mode used to determine tick mark spacing");
   setup_tooltip(_majortick_interval,"Specifies the distance between ticks in Interval mode");
   setup_tooltip(_majortick_intervalB,"Specifies the distance between ticks in Interval mode");
   setup_tooltip(_majortick_count,"Suggests the number of major ticks in Count mode");
   setup_tooltip(_majortick_countB,"Suggests the number of major ticks in Count mode");
   setup_tooltip(_majortick_ticks->base_widget(),"Sets the majortick locations and tick labels for Manual (Arbitrary) mode");
   setup_tooltip(_majortick_ticksB->base_widget(),"Sets the majortick locations and tick labels for Manual (Arbitrary) mode");

   setup_tooltip(_minortick_mode,"Specifies the mode used to determine minor tick mark spacing");
   setup_tooltip(_minortick_interval,"Specifies the distance between minor ticks in Interval mode");
   setup_tooltip(_minortick_count,"Suggests the number of minor ticks in Count mode");

   setup_tooltip(_majorgrid_color->base_widget(),"Sets the color to use for the major grid lines");
   setup_tooltip(_majorgridvisible,"Should the axis major grids be displayed or hidden");
   setup_tooltip(_majorgrid_visible,"Should the axis major grids be displayed or hidden");
   setup_tooltip(_majorgrid_thick,"Set the thickness of the major grid lines");
   setup_tooltip(_majorgrid_style,"Sets the pattern used to display major grid lines");
   setup_tooltip(_minorgrid_color->base_widget(),"Sets the color to use for the minor grid lines");
   setup_tooltip(_minorgridvisible,"Should the axis minor grids be displayed or hidden");
   setup_tooltip(_minorgrid_visible,"Should the axis minor grids be displayed or hidden");
   setup_tooltip(_minorgrid_thick,"Set the thickness of the minor grid lines");
   setup_tooltip(_minorgrid_style,"Sets the pattern used to display minor grid lines");
   setup_tooltip(_label_text,"text (including limited latex support) for the axis label");
   setup_tooltip(_depth,"Specifies the rendering layer (depth) where the axis is drawn");
   setup_tooltip(_visible,"Should the axis be displayed or hidden");
   setup_tooltip(_automin, "Automatically set minimum value when a curve, histogram, contour, or image is added.");
   setup_tooltip(_automax, "Automatically set maximum value when a curve, histogram, contour, or image is added.");
   setup_tooltip(_padding, "Percentage of padding to add for limits, interval, or arbitray modes");

}

void ChipsGuiAxisFrame::color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)
   {
      cmd_manager->open_block ();
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_MAJOR_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_MINOR_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_LABEL_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_TICKLABEL_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_MAJGRD_COLOR);
      axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())),
                                 SET_AXIS_MINGRD_COLOR);

      axisFrame->_majortick_color->set_chips_color (color);
      axisFrame->_minortick_color->set_chips_color (color);
      axisFrame->_label->set_color (color);
      axisFrame->_ticklabel_color->set_chips_color (color);
      axisFrame->_majorgrid_color->set_chips_color (color);
      axisFrame->_minorgrid_color->set_chips_color (color);
      cmd_manager->close_block ();
   }


}

void ChipsGuiAxisFrame::thick_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   float fval = val;
   if (val == 0) fval = 0.5;

   if (axisFrame)
   {
      cmd_manager->open_block ();
      axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_THICKNESS);
      axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MAJ_THICKNESS);
      axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MIN_THICKNESS);
      axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MAJGRD_THICK);
      axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MINGRD_THICK);

      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_thick), axisFrame->_majortick_thick_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_minortick_thick), axisFrame->_minortick_thick_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majorgrid_thick), axisFrame->_majorgrid_thick_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_minorgrid_thick), axisFrame->_minorgrid_thick_id);

      gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majortick_thick), fval);
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_minortick_thick), fval);
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majorgrid_thick), fval);
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_minorgrid_thick), fval);

      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_thick), axisFrame->_majortick_thick_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minortick_thick), axisFrame->_minortick_thick_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majorgrid_thick), axisFrame->_majorgrid_thick_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minorgrid_thick), axisFrame->_minorgrid_thick_id);


      cmd_manager->close_block ();
   }
}

void ChipsGuiAxisFrame::tickstyle_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_axis_property(static_cast<void*>(&val), SET_AXIS_TICKSTYLE);

}

void ChipsGuiAxisFrame::automin_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   bool val = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_AUTOMIN);
}

void ChipsGuiAxisFrame::automax_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   bool val = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_AUTOMAX);
}

void ChipsGuiAxisFrame::padding_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_PAD);

}

void ChipsGuiAxisFrame::label_text_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string text = gtk_entry_get_text(GTK_ENTRY(widget));
   if (axisFrame)
   {
      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(text.c_str())), SET_AXIS_TEXT);
      axisFrame->_label->set_text (text);
   }
}

void ChipsGuiAxisFrame::majorgrid_visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   if (axisFrame)
   {
      bool val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
      if (widget == axisFrame->_majorgrid_visible)
      {
         g_signal_handler_block(GTK_OBJECT(axisFrame->_majorgridvisible), axisFrame->_majorgridvisible_id);
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(axisFrame->_majorgridvisible), val);
         g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majorgridvisible), axisFrame->_majorgridvisible_id);
      }
      else
      {
         g_signal_handler_block(GTK_OBJECT(axisFrame->_majorgrid_visible), axisFrame->_majorgrid_visible_id);
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(axisFrame->_majorgrid_visible), val);
         g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majorgrid_visible), axisFrame->_majorgrid_visible_id);

      }
      axisFrame->set_engine_axis_property(static_cast<void*>(&val), SET_AXIS_MAJGRD_VIS);
   }

}

void ChipsGuiAxisFrame::minorgrid_visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   if (axisFrame)
   {
      bool val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
      if (widget == axisFrame->_minorgrid_visible)
      {
         g_signal_handler_block(GTK_OBJECT(axisFrame->_minorgridvisible), axisFrame->_minorgridvisible_id);
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(axisFrame->_minorgridvisible), val);
         g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minorgridvisible), axisFrame->_minorgridvisible_id);
      }
      else
      {
         g_signal_handler_block(GTK_OBJECT(axisFrame->_minorgrid_visible), axisFrame->_minorgrid_visible_id);
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(axisFrame->_minorgrid_visible), val);
         g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minorgrid_visible), axisFrame->_minorgrid_visible_id);

      }
      axisFrame->set_engine_axis_property(static_cast<void*>(&val), SET_AXIS_MINGRD_VIS);
   }


}


void ChipsGuiAxisFrame::label_color_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_color ();
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_LABEL_COLOR);

}

void ChipsGuiAxisFrame::label_angle_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float angle = widget->get_angle ();
   if (axisFrame) axisFrame->set_engine_property(static_cast<void*>(&angle),
                                 SET_AXIS_LABEL_ANGLE);

}

void ChipsGuiAxisFrame::label_text_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string text = widget->get_text ();

   if (axisFrame)
   {
      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(text.c_str())), SET_AXIS_TEXT);

      g_signal_handler_block(GTK_OBJECT(axisFrame->_label_text), axisFrame->_label_text_id);
      gtk_entry_set_text(GTK_ENTRY(axisFrame->_label_text), text.c_str ());
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_label_text), axisFrame->_label_text_id);
   }

}

void ChipsGuiAxisFrame::label_font_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string font = widget->get_font ();
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(font.c_str())), SET_AXIS_LABEL_FONT);

}

void ChipsGuiAxisFrame::label_fontstyle_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = widget->get_fontstyle ();
   const char *fontstyle = v_fontstyle_to_string((V_FONTSTYLE_TYPE)val);
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(fontstyle)), SET_AXIS_LABEL_FONTSTYLE);

}

void ChipsGuiAxisFrame::label_align_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float valign, halign;

   widget->get_align (halign, valign);
   if (axisFrame)
   {
      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&halign),
                                     SET_AXIS_LABEL_HALIGN);
      axisFrame->set_engine_property(static_cast<void*>(&valign),
                                     SET_AXIS_LABEL_VALIGN);
      cmd_manager->close_block ();
   }
}

void ChipsGuiAxisFrame::label_size_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int size = widget->get_size ();
   if (axisFrame) axisFrame->set_engine_property(static_cast<void*>(&size),
                                 SET_AXIS_LABEL_SIZE);
}

void ChipsGuiAxisFrame::label_position_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float posx, posy;
   widget->get_position (posx, posy);

   if (axisFrame)
   {
      ChipsAxis *ax = dynamic_cast<ChipsAxis *>(axisFrame->_eng->get_object (&(axisFrame->_cid), chips_axis));
      cmd_manager->open_block ();
      if (ax)
      {
         if (ax->get_orientation () == X_AXIS)
         {
            axisFrame->set_engine_property(static_cast<void*>(&posx),
                                    SET_AXIS_OFFSET_PAR);
            axisFrame->set_engine_property(static_cast<void*>(&posy),
                                    SET_AXIS_OFFSET_PERP);
         }
         else
         {
            axisFrame->set_engine_property(static_cast<void*>(&posy),
                                    SET_AXIS_OFFSET_PAR);
            axisFrame->set_engine_property(static_cast<void*>(&posx),
                                    SET_AXIS_OFFSET_PERP);
         }
      }
      cmd_manager->close_block ();
   }
}


void ChipsGuiAxisFrame::label_visible_change_callback(ChipsGuiLabelWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   bool visible = widget->get_visible ();
   if (axisFrame) axisFrame->set_engine_property(static_cast<void*>(&visible),
                                 SET_AXIS_LABEL_VIS);

}


void ChipsGuiAxisFrame::majortick_color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_MAJOR_COLOR);
   
}

void ChipsGuiAxisFrame::majortick_ticks_callback(ChipsGuiLevelsWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   ChipsAxisSetArbitraryCallback set_arbitrary_callback;
   ChipsAxisSetArbitraryCallback::ChipsAxisSetArbitraryCallbackData set_data;
   double *ticks;
   string *text;
   int num_ticks = 0;

   double *curr_ticks;
   string *curr_text;
   unsigned int curr_num_ticks = 0;
   bool changed = false;
   ChipsAxis *axis = NULL;

   axis = dynamic_cast<ChipsAxis *>(axisFrame->_eng->get_object (&axisFrame->_cid, chips_axis));

   widget->get_column (&ticks, num_ticks, 0);
   widget->get_column (&text, num_ticks, 1);

   set_data.num_ticks = 0;
   for (int ii=0; ii<num_ticks; ii++)
   {
      if (finite (ticks[ii]))
      {
         ticks[set_data.num_ticks] = ticks[ii];
         text[set_data.num_ticks] = text[ii];
         set_data.num_ticks++;
      }
   }

   if (axis)
      axis->get_arbitrary_tick_positions (curr_num_ticks, &curr_ticks, &curr_text);

   // check to see if the ticks or text has changed
   if (curr_num_ticks == (unsigned int)num_ticks)
   {
      for (int ii=curr_num_ticks; ii-- && !changed;)
         if (curr_ticks[ii] != ticks[ii])
            changed = true;

      for (int ii=curr_num_ticks; ii-- && !changed;)
         if (curr_text[ii] != text[ii])
            changed = true;
   }
   else 
      changed = true;

   delete []curr_ticks;
   delete []curr_text;

   set_data.text = text;
   set_data.ticks = ticks;
   if (changed && axisFrame)
   {
      GtkTreeIter iter;
      const char *str_val = v_tickmode_to_string (chips_arbitrary);
      cmd_manager->open_block ();
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);

      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_list, iter, (int)chips_arbitrary))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_mode), &iter);
      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_listB, iter, (int)chips_arbitrary))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_modeB), &iter);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);

      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MAJ_MODE);

      axisFrame->update_property (&set_arbitrary_callback, &set_data);
      cmd_manager->close_block ();
   }

   delete []set_data.ticks;
   delete []set_data.text;

}


void ChipsGuiAxisFrame::majortick_thick_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   float fval = val;
   if (val == 0) fval = 0.5;

   if (axisFrame) axisFrame->set_engine_property(static_cast<void*>(&fval), SET_AXIS_MAJ_THICKNESS);

}

void ChipsGuiAxisFrame::majortick_style_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MAJ_STYLE);

}

void ChipsGuiAxisFrame::majortick_length_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MAJORLENGTH);

}

void ChipsGuiAxisFrame::majortick_visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = (int)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MAJ_VIS);

}

void ChipsGuiAxisFrame::majortick_mode_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   GValue value = {0,};
   GtkTreeIter iter;
   int val;
   const char *str_val;

   GtkWidget* wtmp = axisFrame->get_axistab_majortick_widget();
   gint tval = gtk_combo_box_get_active(GTK_COMBO_BOX (widget));
   g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
   g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);
   if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter))
   {
      if (widget == wtmp)  // axis tab 
      {
         gtk_tree_model_get_value (GTK_TREE_MODEL(axisFrame->_majortick_mode_list), &iter, 1, &value);
         val = g_value_get_int (&value);
         str_val = v_tickmode_to_string ((V_TICKMODE_TYPE)val);
         if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MAJ_MODE);
        wtmp = axisFrame->get_ticktab_majortick_widget();
      }
      else 
      {
         gtk_tree_model_get_value (GTK_TREE_MODEL(axisFrame->_majortick_mode_listB), &iter, 1, &value);
         val = g_value_get_int (&value);
         str_val = v_tickmode_to_string ((V_TICKMODE_TYPE)val);
         if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MAJ_MODE);
      }
      if (wtmp)  gtk_combo_box_set_active(GTK_COMBO_BOX(wtmp), tval);
   }
   g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
   g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);
}

void ChipsGuiAxisFrame::majortick_interval_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)
   {
      GtkTreeIter iter;
      const char *str_val = v_tickmode_to_string (chips_interval);

      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MAJORINT);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_minortick_interval), axisFrame->_minortick_interval_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_interval), axisFrame->_majortick_interval_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_intervalB), axisFrame->_majortick_intervalB_id);

      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_list, iter, (int)chips_interval))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_mode), &iter);
      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_listB, iter, (int)chips_interval))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_modeB), &iter);
      gtk_spin_button_set_range(GTK_SPIN_BUTTON(axisFrame->_minortick_interval), 0, val);
      if (widget == axisFrame->_majortick_interval)
         gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majortick_intervalB), val);
      else 
         gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majortick_interval), val);

      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minortick_interval), axisFrame->_minortick_interval_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_interval), axisFrame->_majortick_interval_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_intervalB), axisFrame->_majortick_intervalB_id);

      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MAJ_MODE);

      cmd_manager->close_block ();

   }

}

void ChipsGuiAxisFrame::majortick_count_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)
   {
      GtkTreeIter iter;
      const char *str_val = v_tickmode_to_string (chips_count);
      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MAJOR_NUM);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);
      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_list, iter, (int)chips_count))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_mode), &iter);
      if (axisFrame->get_tickmode_iter (axisFrame->_majortick_mode_listB, iter, (int)chips_count))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_majortick_modeB), &iter);
      if (widget == axisFrame->_majortick_count)
         gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majortick_countB), val);
      else 
         gtk_spin_button_set_value(GTK_SPIN_BUTTON(axisFrame->_majortick_count), val);

      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_mode), axisFrame->_majortick_mode_id);
      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_majortick_modeB), axisFrame->_majortick_modeB_id);

      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MAJ_MODE);
      cmd_manager->close_block ();
   }
}

void ChipsGuiAxisFrame::minortick_color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_MINOR_COLOR);

}

void ChipsGuiAxisFrame::minortick_thick_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   float fval = val;
   if (val == 0) fval = 0.5;

   if (axisFrame) axisFrame->set_engine_property(static_cast<void*>(&fval), SET_AXIS_MIN_THICKNESS);

}

void ChipsGuiAxisFrame::minortick_style_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MIN_STYLE);

}

void ChipsGuiAxisFrame::minortick_length_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MINORLENGTH);
}

void ChipsGuiAxisFrame::minortick_visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = (int)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MIN_VIS);

}

void ChipsGuiAxisFrame::minortick_mode_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   GValue value = {0,};
   GtkTreeIter iter;
   int val;
   const char *str_val;

   if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter))
   {
      gtk_tree_model_get_value (GTK_TREE_MODEL(axisFrame->_minortick_mode_list), &iter, 1, &value);
      val = g_value_get_int (&value);
      str_val = v_tickmode_to_string ((V_TICKMODE_TYPE)val);

      if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MIN_MODE);
   }

}

void ChipsGuiAxisFrame::minortick_interval_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)
   {
      GtkTreeIter iter;
      const char *str_val = v_tickmode_to_string (chips_interval);
      
      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MINORINT);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_minortick_mode), axisFrame->_minortick_mode_id);

      if (axisFrame->get_tickmode_iter (axisFrame->_minortick_mode_list, iter, (int)chips_interval))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_minortick_mode), &iter);

      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minortick_mode), axisFrame->_minortick_mode_id);

      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MIN_MODE);
      cmd_manager->close_block ();
   }
}

void ChipsGuiAxisFrame::minortick_count_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)
   {
      GtkTreeIter iter;
      const char *str_val = v_tickmode_to_string (chips_count);
      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_MINOR_NUM);
      g_signal_handler_block(GTK_OBJECT(axisFrame->_minortick_mode), axisFrame->_minortick_mode_id);
      if (axisFrame->get_tickmode_iter (axisFrame->_minortick_mode_list, iter, (int)chips_count))
         gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_minortick_mode), &iter);

      g_signal_handler_unblock(GTK_OBJECT(axisFrame->_minortick_mode), axisFrame->_minortick_mode_id);

      axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_val)), SET_AXIS_MIN_MODE);
      cmd_manager->close_block ();
   }

}

void ChipsGuiAxisFrame::majorgrid_color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)  axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_MAJGRD_COLOR);

}

void ChipsGuiAxisFrame::majorgrid_thick_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   float fval = val;
   if (val == 0) fval = 0.5;

   if (axisFrame) axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MAJGRD_THICK);

}

void ChipsGuiAxisFrame::majorgrid_style_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_axis_property(static_cast<void*>(&val), SET_AXIS_MAJGRD_STYLE);
}

void ChipsGuiAxisFrame::minorgrid_color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)  axisFrame->set_engine_axis_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_MINGRD_COLOR);

}

void ChipsGuiAxisFrame::minorgrid_thick_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   float fval = val;
   if (val == 0) fval = 0.5;

   if (axisFrame) axisFrame->set_engine_axis_property(static_cast<void*>(&fval), SET_AXIS_MINGRD_THICK);
}

void ChipsGuiAxisFrame::minorgrid_style_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_axis_property(static_cast<void*>(&val), SET_AXIS_MINGRD_STYLE);
}


void ChipsGuiAxisFrame::ticklabel_color_callback(ChipsGuiColorWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   string color = widget->get_chips_color ();
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(color.c_str())), SET_AXIS_TICKLABEL_COLOR);

}

void ChipsGuiAxisFrame::ticklabel_font_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   char *str_font = gtk_combo_box_get_active_text(GTK_COMBO_BOX (widget));

   if (str_font)
   {
      if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(str_font), SET_AXIS_TICKLABEL_FONT);
      g_free (str_font);
   }

}

void ChipsGuiAxisFrame::ticklabel_fontstyle_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int set_style = 0;
   V_FONTSTYLE_TYPE style;

   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (axisFrame->_ticklabel_bold)))
      set_style += 1;
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (axisFrame->_ticklabel_italic)))
      set_style += 2;

   if (set_style == 1)
      style = chips_bold;
   else if (set_style == 2)
      style = chips_italic;
   else if (set_style == 3)
      style = chips_bolditalic;
   else
      style = chips_normal;

   const char *str_style = v_fontstyle_to_string((V_FONTSTYLE_TYPE)style);
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(str_style)), SET_AXIS_TICKLABEL_FNTSTL);

}

void ChipsGuiAxisFrame::ticklabel_size_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);

   char *str_size = gtk_combo_box_get_active_text(GTK_COMBO_BOX (widget));
   int int_size;

   if (str_size)
   {
      int_size = atoi (str_size);
      if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&int_size), SET_AXIS_TICKLABEL_SIZE);

      g_free (str_size);

   }

}

void ChipsGuiAxisFrame::ticklabel_angle_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));

   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_TICKLABEL_ANGLE);
}

void ChipsGuiAxisFrame::ticklabel_offset_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));

   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_TICKLABEL_OFFSET);

}

void ChipsGuiAxisFrame::ticklabel_style_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_TICKLABEL_STYLE);

}

void ChipsGuiAxisFrame::ticklabel_align_callback(ChipsGuiPivotPointWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float halign, valign;

   widget->get_pivot_position (halign, valign);

   if (axisFrame)
   {
      cmd_manager->open_block ();
      axisFrame->set_engine_property(static_cast<void*>(&halign), SET_AXIS_TICKLABEL_HALIGN);
      axisFrame->set_engine_property(static_cast<void*>(&valign), SET_AXIS_TICKLABEL_VALIGN);
      cmd_manager->close_block ();
   }
}

void ChipsGuiAxisFrame::ticklabel_halign_auto_callback(ChipsGuiPivotPointWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float hval = AUTO_ALIGN;
   float vval;
   ChipsAxis *axis;

   if (!axisFrame)
      return;

   axis = dynamic_cast<ChipsAxis *>(axisFrame->_eng->get_object (&(axisFrame->_cid), chips_axis));

   if (!axis)
      return;

   axisFrame->set_engine_property(static_cast<void*>(&hval), SET_AXIS_TICKLABEL_HALIGN);
   hval = axis->get_ticklabel_halign (true);
   vval = axis->get_ticklabel_valign (true);
   axisFrame->_ticklabel_alignment->set_pivot_position (hval, vval);


}

void ChipsGuiAxisFrame::ticklabel_valign_auto_callback(ChipsGuiPivotPointWidget *widget, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   float hval;
   float vval = AUTO_ALIGN;
   ChipsAxis *axis;

   if (!axisFrame)
      return;
   axis = dynamic_cast<ChipsAxis *>(axisFrame->_eng->get_object (&(axisFrame->_cid), chips_axis));

   if (!axis)
      return;

   axisFrame->set_engine_property(static_cast<void*>(&vval), SET_AXIS_TICKLABEL_VALIGN);

   hval = axis->get_ticklabel_halign (true);
   vval = axis->get_ticklabel_valign (true);
   axisFrame->_ticklabel_alignment->set_pivot_position (hval, vval);

}

void ChipsGuiAxisFrame::ticklabel_format_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   GtkTreeIter iter;
   string format = "";
   GValue value={0,};

   if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter))
   {
      gtk_tree_model_get_value (GTK_TREE_MODEL(axisFrame->_format_list), &iter,
                                    1, &value);
      format = g_value_get_string (&value);
   }

   g_signal_handler_block(GTK_OBJECT(axisFrame->_ticklabel_format_custom), axisFrame->_ticklabel_format_custom_id);
   if (format.length () > 0)
      gtk_entry_set_text (GTK_ENTRY (axisFrame->_ticklabel_format_custom), format.c_str ());
   else
      format = gtk_entry_get_text (GTK_ENTRY (axisFrame->_ticklabel_format_custom));
   g_signal_handler_unblock(GTK_OBJECT(axisFrame->_ticklabel_format_custom), axisFrame->_ticklabel_format_custom_id);
   axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(format.c_str ())), SET_AXIS_TICKFORMAT);
}

bool ChipsGuiAxisFrame::ticklabel_format_custom_focusout_callback(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   axisFrame->ticklabel_format_custom_callback (widget, data);
   return false;
}


void ChipsGuiAxisFrame::ticklabel_format_custom_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   GtkTreeIter iter;
   GtkTreeIter last_iter;
   string format = "";
   string tmp_format = "";
   GValue value={0,};

   format = gtk_entry_get_text (GTK_ENTRY (axisFrame->_ticklabel_format_custom));

   gtk_tree_model_get_iter_root (GTK_TREE_MODEL(axisFrame->_format_list), &iter);
   do
   {
      last_iter = iter;
      gtk_tree_model_get_value (GTK_TREE_MODEL(axisFrame->_format_list), &iter,
                                    1, &value);
      tmp_format = g_value_get_string (&value);
      // find matching combo option (ra is same as ra4)
      if (tmp_format == format || ((format == "%ra" && tmp_format == "%ra4") ||
          (format == "%RA" && tmp_format == "%RA4")))
         break;
      g_value_unset (&value);
   } while (gtk_tree_model_iter_next (GTK_TREE_MODEL(axisFrame->_format_list), &iter));

   g_signal_handler_block(GTK_OBJECT(axisFrame->_ticklabel_format_box), axisFrame->_ticklabel_format_box_id);
   gtk_combo_box_set_active_iter(GTK_COMBO_BOX(axisFrame->_ticklabel_format_box),&last_iter);
   g_signal_handler_unblock(GTK_OBJECT(axisFrame->_ticklabel_format_box), axisFrame->_ticklabel_format_box_id);
   axisFrame->set_engine_property(static_cast<void*>(const_cast<char *>(format.c_str ())), SET_AXIS_TICKFORMAT);
}




void ChipsGuiAxisFrame::ticklabel_visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   bool val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_TICKLABEL_VIS);

}

void ChipsGuiAxisFrame::depth_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_DEPTH);

}

void ChipsGuiAxisFrame::visible_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   int val = (int)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
   if (axisFrame)  axisFrame->set_engine_property(static_cast<void*>(&val), SET_AXIS_VISIBLE);

}

void ChipsGuiAxisFrame::apply_axis_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);
   bool active = (bool)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

   if (axisFrame && active)
      axisFrame->set_apply_xy (widget == axisFrame->_apply_xy);
}


void ChipsGuiAxisFrame::initialize(ChipsId* id)
{
   V_FONTSTYLE_TYPE tmp_fontstyle;
   V_TICKMODE_TYPE tmp_tickmode_type;
   ChipsAxisAlgorithmFactory::TransformType axis_type;
   ChipsAxis *axis = NULL;
   string tmp_tickstyle;
   GtkTreeIter iter;
   stringstream sstream;

   if (!id)  return; 

   if (chipsGui && !_eng) 
      _eng = chipsGui->get_engine(); 

   axis = dynamic_cast<ChipsAxis *>(_eng->get_object (id, chips_axis));

   if (!axis)
      return;

   axis_type = axis->get_algorithm_type ();

   if (&_cid != id)
   {
      _cid.reset_window_id(id->get_window_id()); 
      _cid.set_frame_id(id->get_frame_id()); 
      _cid.set_plot_id(id->get_plot_id()); 
      _cid.set_axis_id(id->get_axis_id()); 
   }

   _set_cid = _cid;

   _eng->get_property(&_cid, (void*)&_colorval,SET_AXIS_COLOR);
   _eng->get_property(&_cid, (void*)&_paddingval,SET_AXIS_PAD);
   _eng->get_property(&_cid, (void*)&_depthval,SET_AXIS_DEPTH);
   _eng->get_property(&_cid, (void*)&_majortick_colorval,SET_AXIS_MAJOR_COLOR);
   _eng->get_property(&_cid, (void*)&_minortick_colorval,SET_AXIS_MINOR_COLOR);
   _eng->get_property(&_cid, (void*)&_thicknessval,SET_AXIS_THICKNESS);
   _eng->get_property(&_cid, (void*)&_majortick_thicknessval,SET_AXIS_MAJ_THICKNESS);
   _eng->get_property(&_cid, (void*)&_minortick_thicknessval,SET_AXIS_MIN_THICKNESS);
   _eng->get_property(&_cid, (void*)&_labelval,SET_AXIS_TEXT);
   _eng->get_property(&_cid, (void*)&tmp_tickstyle,SET_AXIS_TICKSTYLE);
   string_to_v_tickstyle (tmp_tickstyle.c_str (), &_tickstyleval);
   _eng->get_property(&_cid, (void*)&tmp_tickstyle,SET_AXIS_MAJ_STYLE);
   string_to_v_tickstyle (tmp_tickstyle.c_str (), &_majortick_styleval);
   _eng->get_property(&_cid, (void*)&tmp_tickstyle,SET_AXIS_MIN_STYLE);
   string_to_v_tickstyle (tmp_tickstyle.c_str (), &_minortick_styleval);
   _eng->get_property(&_cid, (void*)&tmp_tickstyle,SET_AXIS_TICKLABEL_STYLE);
   string_to_v_tickstyle (tmp_tickstyle.c_str (), &_ticklabel_styleval);
   _eng->get_property(&_cid, (void*)&_ticklabel_formatval,SET_AXIS_TICKFORMAT);
   _eng->get_property(&_cid, (void*)&_majortick_numval,SET_AXIS_MAJOR_NUM);
   _eng->get_property(&_cid, (void*)&_minortick_numval,SET_AXIS_MINOR_NUM);
   _eng->get_property(&_cid, (void*)&_majortick_intval,SET_AXIS_MAJORINT);
   _eng->get_property(&_cid, (void*)&_minortick_intval,SET_AXIS_MINORINT);
   _eng->get_property(&_cid, (void*)&_majortick_modeval,SET_AXIS_MAJ_MODE);
   _eng->get_property(&_cid, (void*)&_minortick_modeval,SET_AXIS_MIN_MODE);
   _eng->get_property(&_cid, (void*)&_majortick_visibleval,SET_AXIS_MAJ_VIS);
   _eng->get_property(&_cid, (void*)&_minortick_visibleval,SET_AXIS_MIN_VIS);
   _eng->get_property(&_cid, (void*)&_visibleval,SET_AXIS_VISIBLE);
   _eng->get_property(&_cid, (void*)&_majortick_lengthval,SET_AXIS_MAJORLENGTH);
   _eng->get_property(&_cid, (void*)&_label_offsetparval,SET_AXIS_OFFSET_PAR);
   _eng->get_property(&_cid, (void*)&_label_offsetperpval,SET_AXIS_OFFSET_PERP);
   _eng->get_property(&_cid, (void*)&_ticklabel_offsetval,SET_AXIS_TICKLABEL_OFFSET);
   _eng->get_property(&_cid, (void*)&_ticklabel_fontval,SET_AXIS_TICKLABEL_FONT);
   _eng->get_property(&_cid, (void*)&_ticklabel_fontstyleval,SET_AXIS_TICKLABEL_FNTSTL);
   _eng->get_property(&_cid, (void*)&_ticklabel_sizeval,SET_AXIS_TICKLABEL_SIZE);
   _eng->get_property(&_cid, (void*)&_ticklabel_colorval,SET_AXIS_TICKLABEL_COLOR);
   _eng->get_property(&_cid, (void*)&_ticklabel_angleval,SET_AXIS_TICKLABEL_ANGLE);
   _eng->get_property(&_cid, (void*)&_ticklabel_halignval,SET_AXIS_TICKLABEL_HALIGN);
   _eng->get_property(&_cid, (void*)&_ticklabel_valignval,SET_AXIS_TICKLABEL_VALIGN);
   _eng->get_property(&_cid, (void*)&_ticklabel_visibleval,SET_AXIS_TICKLABEL_VIS);
   _eng->get_property(&_cid, (void*)&_label_visibleval,SET_AXIS_LABEL_VIS);
   _eng->get_property(&_cid, (void*)&_label_fontval,SET_AXIS_LABEL_FONT);
   _eng->get_property(&_cid, (void*)&_label_fontstyleval,SET_AXIS_LABEL_FONTSTYLE);
   _eng->get_property(&_cid, (void*)&_label_angleval,SET_AXIS_LABEL_ANGLE);
   _eng->get_property(&_cid, (void*)&_label_halignval,SET_AXIS_LABEL_HALIGN);
   _eng->get_property(&_cid, (void*)&_label_valignval,SET_AXIS_LABEL_VALIGN);
   _eng->get_property(&_cid, (void*)&_label_sizeval,SET_AXIS_LABEL_SIZE);
   _eng->get_property(&_cid, (void*)&_label_colorval,SET_AXIS_LABEL_COLOR);
   _eng->get_property(&_cid, (void*)&_autominval,SET_AXIS_AUTOMIN);
   _eng->get_property(&_cid, (void*)&_automaxval,SET_AXIS_AUTOMAX);
   _eng->get_property(&_cid, (void*)&_minortick_lengthval,SET_AXIS_MINORLENGTH);
   _eng->get_property(&_cid, (void*)&_majorgrid_colorval,SET_AXIS_MAJGRD_COLOR);
   _eng->get_property(&_cid, (void*)&_majorgrid_styleval,SET_AXIS_MAJGRD_STYLE);
   _eng->get_property(&_cid, (void*)&_majorgrid_visibleval,SET_AXIS_MAJGRD_VIS);
   _eng->get_property(&_cid, (void*)&_majorgrid_thicknessval,SET_AXIS_MAJGRD_THICK);
   _eng->get_property(&_cid, (void*)&_minorgrid_colorval,SET_AXIS_MINGRD_COLOR);
   _eng->get_property(&_cid, (void*)&_minorgrid_styleval,SET_AXIS_MINGRD_STYLE);
   _eng->get_property(&_cid, (void*)&_minorgrid_visibleval,SET_AXIS_MINGRD_VIS);
   _eng->get_property(&_cid, (void*)&_minorgrid_thicknessval,SET_AXIS_MINGRD_THICK);



//   labelobj->get_position (_posxval, _posyval);

   g_signal_handler_block(GTK_OBJECT(_thick), _thick_id);
   g_signal_handler_block(GTK_OBJECT(_tickstyle), _tickstyle_id);
   g_signal_handler_block(GTK_OBJECT(_automin), _automin_id);
   g_signal_handler_block(GTK_OBJECT(_automax), _automax_id);
   g_signal_handler_block(GTK_OBJECT(_padding), _padding_id);
   g_signal_handler_block(GTK_OBJECT(_majorgridvisible), _majorgridvisible_id);
   g_signal_handler_block(GTK_OBJECT(_minorgridvisible), _minorgridvisible_id);
   g_signal_handler_block(GTK_OBJECT(_label_text), _label_text_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_thick), _majortick_thick_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_style), _majortick_style_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_length), _majortick_length_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_visible), _majortick_visible_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_mode), _majortick_mode_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_interval), _majortick_interval_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_intervalB), _majortick_intervalB_id);
   g_signal_handler_block(GTK_OBJECT(_majortick_count), _majortick_count_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_thick), _minortick_thick_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_style), _minortick_style_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_length), _minortick_length_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_visible), _minortick_visible_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_mode), _minortick_mode_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_interval), _minortick_interval_id);
   g_signal_handler_block(GTK_OBJECT(_minortick_count), _minortick_count_id);
   g_signal_handler_block(GTK_OBJECT(_majorgrid_thick), _majorgrid_thick_id);
   g_signal_handler_block(GTK_OBJECT(_majorgrid_style), _majorgrid_style_id);
   g_signal_handler_block(GTK_OBJECT(_majorgrid_visible), _majorgrid_visible_id);
   g_signal_handler_block(GTK_OBJECT(_minorgrid_thick), _minorgrid_thick_id);
   g_signal_handler_block(GTK_OBJECT(_minorgrid_style), _minorgrid_style_id);
   g_signal_handler_block(GTK_OBJECT(_minorgrid_visible), _minorgrid_visible_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_font), _ticklabel_font_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_bold), _ticklabel_bold_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_italic), _ticklabel_italic_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_size), _ticklabel_size_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_angle), _ticklabel_angle_change_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_offset), _ticklabel_offset_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_style), _ticklabel_style_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_visible), _ticklabel_visible_change_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_format_custom), _ticklabel_format_custom_id);
   g_signal_handler_block(GTK_OBJECT(_ticklabel_format_box), _ticklabel_format_box_id);


   g_signal_handler_block(GTK_OBJECT(_depth), _depth_id);
   g_signal_handler_block(GTK_OBJECT(_visible), _visible_id);

   g_signal_handler_block(GTK_OBJECT(_apply_xy), _apply_xy_id);
   g_signal_handler_block(GTK_OBJECT(_apply_current), _apply_current_id);

   // initialize general properties
   _color->set_chips_color (_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_thick), _thicknessval);
   gtk_combo_box_set_active(GTK_COMBO_BOX(_tickstyle),(int)_tickstyleval);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_automin), _autominval);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_automax), _automaxval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_padding), _paddingval);
   gtk_entry_set_text(GTK_ENTRY(_label_text),(const gchar *)_labelval.c_str());
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_majorgridvisible), _majorgrid_visibleval);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_minorgridvisible), _minorgrid_visibleval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_depth), _depthval);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_visible), _visibleval);


   // initialize label
   _label->set_color(_label_colorval);
   _label->set_size (_label_sizeval);
   _label->set_font (_label_fontval);
   _label->set_angle (_label_angleval);
   _label->set_align (_label_halignval, _label_valignval);
   _label->set_visible (_label_visibleval);
   _label->set_text (_labelval);
   string_to_v_fontstyle(_label_fontstyleval.c_str (), &tmp_fontstyle);
   _label->set_fontstyle (tmp_fontstyle);


   if (_apply_xy_map.count (axis->get_ref_cnt ()) == 0) 
      _apply_xy_map[axis->get_ref_cnt ()] = true;

   if (_apply_xy_map [axis->get_ref_cnt ()])
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(_apply_xy), 1);
   else
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(_apply_current), 1);
   set_apply_xy (_apply_xy_map[axis->get_ref_cnt ()]);
   

   _cid.set_axis_dim (axis->get_orientation ());
   if (axis->get_orientation () == X_AXIS)
   {
      double label_xrange[] = {-.5, .5};
      double label_yrange[] = {-9999, 9999}; 
      _label->set_position_type (ChipsGuiLabelWidget::FLOAT,
                              ChipsGuiLabelWidget::INT);
      _label->set_position_range (label_xrange, .01, label_yrange, 1);
      _label->set_position (_label_offsetparval, _label_offsetperpval);

      gtk_button_set_label (GTK_BUTTON(_apply_current), "X Axis");

   }
   else
   {
      double label_yrange[] = {-.5, .5};
      double label_xrange[] = {-9999, 9999};

      _label->set_position_type (ChipsGuiLabelWidget::INT,
                              ChipsGuiLabelWidget::FLOAT);
      _label->set_position_range (label_xrange, 1, label_yrange, .01);
      _label->set_position (_label_offsetperpval, _label_offsetparval);

      gtk_button_set_label (GTK_BUTTON(_apply_current), "Y Axis");
   }


   // initialize tick label
   sstream << _ticklabel_sizeval << ends;
   gtk_entry_set_text (GTK_ENTRY (GTK_BIN (_ticklabel_size)->child), sstream.str ().c_str ());

   {
      int index = 0;
      const gchar *tmp_font;
      GValue value={0,};
      bool font_set = false;
      GtkListStore *font_list;

      font_list = ChipsGuiPropPane::font_info.get_font_list ();

      // this doesn't work if the default value is not in the model tree
      // it's not clear from the documentation how to make this work
      gtk_tree_model_get_iter_root (GTK_TREE_MODEL(font_list), &iter);
      do
      {
         gtk_tree_model_get_value (GTK_TREE_MODEL(font_list), &iter,
                                    0, &value);
         tmp_font = g_value_get_string (&value);
         if (chips_strcmp (tmp_font,_ticklabel_fontval.c_str ()) == 0)
         {
            font_set = true;
            break;
         }
         index++;
         g_value_unset (&value);
      } while (gtk_tree_model_iter_next (GTK_TREE_MODEL(font_list), &iter));

      if (font_set)
         gtk_combo_box_set_active(GTK_COMBO_BOX(_ticklabel_font),index);
      else
         gtk_combo_box_set_active(GTK_COMBO_BOX(_ticklabel_font), -1);
   }


   
   string_to_v_fontstyle(_ticklabel_fontstyleval.c_str (), &tmp_fontstyle);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_ticklabel_visible), _ticklabel_visibleval);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_ticklabel_bold), FALSE);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_ticklabel_italic), FALSE);
   if ((tmp_fontstyle == chips_bold) || (tmp_fontstyle == chips_bolditalic))
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_ticklabel_bold), TRUE);
   if ((tmp_fontstyle == chips_italic) || (tmp_fontstyle == chips_bolditalic))
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_ticklabel_italic), TRUE);

   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_ticklabel_angle), _ticklabel_angleval);
   _ticklabel_color->set_chips_color (_ticklabel_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_ticklabel_offset), _ticklabel_offsetval);

   gtk_combo_box_set_active (GTK_COMBO_BOX (_ticklabel_style), (int)_ticklabel_styleval);

   _ticklabel_alignment->set_auto_x (_ticklabel_halignval == AUTO_ALIGN);
   _ticklabel_alignment->set_auto_y (_ticklabel_valignval == AUTO_ALIGN);
   
   // re-assign the halign/valign val with actual value
   _ticklabel_halignval = axis->get_ticklabel_halign (true);
   _ticklabel_valignval = axis->get_ticklabel_valign (true);
   _ticklabel_alignment->set_pivot_position (_ticklabel_halignval, _ticklabel_valignval);

   // compute tick format
   if (_ticklabel_formatval.length () > 0)
   {
      string tmp_format;
      GValue value={0,};
      GtkTreeIter last_iter;

      gtk_tree_model_get_iter_root (GTK_TREE_MODEL(_format_list), &iter);
      do
      {
         last_iter = iter;
         gtk_tree_model_get_value (GTK_TREE_MODEL(_format_list), &iter,
                                    1, &value);
         tmp_format = g_value_get_string (&value);
         if (tmp_format == _ticklabel_formatval)
            break;
         else if (tmp_format == "%ra4" && (_ticklabel_formatval == "%ra" ||
            _ticklabel_formatval == "%RA"))
            break;  // handle special case of %ra or %RA mapping to %ra4 
         else if (tmp_format == "%dec" && _ticklabel_formatval == "%DEC")
            break;  // handle special case of %DEC mapping to %dec 
         g_value_unset (&value);
      } while (gtk_tree_model_iter_next (GTK_TREE_MODEL(_format_list), &iter));

      gtk_combo_box_set_active_iter(GTK_COMBO_BOX(_ticklabel_format_box),&last_iter);
      gtk_entry_set_text (GTK_ENTRY (_ticklabel_format_custom), _ticklabel_formatval.c_str ());

   }
   else
   {
      gtk_combo_box_set_active(GTK_COMBO_BOX(_ticklabel_format_box), -1);
      gtk_entry_set_text (GTK_ENTRY (_ticklabel_format_custom), "");
   }


   // initialize major ticks
   build_majortick_mode_list (_majortick_mode_list, axis_type);
   build_majortick_mode_list (_majortick_mode_listB, axis_type);
   build_minortick_mode_list (_minortick_mode_list, axis_type);


   _majortick_color->set_chips_color (_majortick_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majortick_thick), _majortick_thicknessval);
   gtk_combo_box_set_active(GTK_COMBO_BOX(_majortick_style),(int)_majortick_styleval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majortick_length), _majortick_lengthval);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_majortick_visible),_majortick_visibleval);

   string_to_v_tickmode (_majortick_modeval.c_str (), &tmp_tickmode_type);

   if (get_tickmode_iter (_majortick_mode_list, iter, (int)tmp_tickmode_type))
      gtk_combo_box_set_active_iter(GTK_COMBO_BOX(_majortick_mode), &iter);
   if (get_tickmode_iter (_majortick_mode_listB, iter, (int)tmp_tickmode_type))
      gtk_combo_box_set_active_iter(GTK_COMBO_BOX(_majortick_modeB), &iter);

   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majortick_interval), _majortick_intval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majortick_intervalB), _majortick_intval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majortick_count), _majortick_numval);

   if (axis_type == ChipsAxisAlgorithmFactory::WCSTAN)
   {
     _majortick_ticks->disable_widget ();
     _majortick_ticksB->disable_widget ();
     gtk_widget_set_sensitive (_majortick_interval, FALSE);
     gtk_widget_set_sensitive (_majortick_intervalB, FALSE);
     gtk_widget_set_sensitive (_majortick_count, FALSE);
     gtk_widget_set_sensitive (_minortick_interval, FALSE);
     gtk_widget_set_sensitive (_minortick_count, FALSE);
   }
   else
   {
     _majortick_ticks->enable_widget ();
     _majortick_ticksB->enable_widget ();
     reset_arbitrary_tick_list ();
     gtk_widget_set_sensitive (_majortick_interval, TRUE);
     gtk_widget_set_sensitive (_majortick_intervalB, TRUE);
     gtk_widget_set_sensitive (_majortick_count, TRUE);
     gtk_widget_set_sensitive (_minortick_interval, TRUE);
     gtk_widget_set_sensitive (_minortick_count, TRUE);
   }

   // initialize minor ticks

   _minortick_color->set_chips_color (_minortick_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_minortick_thick), _minortick_thicknessval);
   gtk_combo_box_set_active(GTK_COMBO_BOX(_minortick_style),(int)_minortick_styleval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_minortick_length), _minortick_lengthval);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_minortick_visible),_minortick_visibleval);

   string_to_v_tickmode (_minortick_modeval.c_str (), &tmp_tickmode_type);

   if (get_tickmode_iter (_minortick_mode_list, iter, (int)tmp_tickmode_type))
      gtk_combo_box_set_active_iter(GTK_COMBO_BOX(_minortick_mode), &iter);

   if (_minortick_intval > _majortick_intval)
      _minortick_intval = _majortick_intval;
   gtk_spin_button_set_range(GTK_SPIN_BUTTON(_minortick_interval), 0, _majortick_intval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_minortick_interval), _minortick_intval);
   if (_minortick_intval > _majortick_intval)
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_minortick_count), _minortick_numval);

   // initialize major grids
   _majorgrid_color->set_chips_color (_majorgrid_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_majorgrid_thick), _majorgrid_thicknessval);
   gtk_combo_box_set_active(GTK_COMBO_BOX(_majorgrid_style),(int)_majorgrid_styleval);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_majorgrid_visible),_majorgrid_visibleval);


   // initialize minor grids
   _minorgrid_color->set_chips_color (_minorgrid_colorval);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(_minorgrid_thick), _minorgrid_thicknessval);
   gtk_combo_box_set_active(GTK_COMBO_BOX(_minorgrid_style),(int)_minorgrid_styleval);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (_minorgrid_visible),_minorgrid_visibleval);

 
   g_signal_handler_unblock(GTK_OBJECT(_thick), _thick_id);
   g_signal_handler_unblock(GTK_OBJECT(_tickstyle), _tickstyle_id);
   g_signal_handler_unblock(GTK_OBJECT(_automin), _automin_id);
   g_signal_handler_unblock(GTK_OBJECT(_automax), _automax_id);
   g_signal_handler_unblock(GTK_OBJECT(_padding), _padding_id);
   g_signal_handler_unblock(GTK_OBJECT(_majorgridvisible), _majorgridvisible_id);
   g_signal_handler_unblock(GTK_OBJECT(_minorgridvisible), _minorgridvisible_id);
   g_signal_handler_unblock(GTK_OBJECT(_label_text), _label_text_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_thick), _majortick_thick_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_style), _majortick_style_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_length), _majortick_length_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_visible), _majortick_visible_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_mode), _majortick_mode_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_interval), _majortick_interval_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_intervalB), _majortick_intervalB_id);
   g_signal_handler_unblock(GTK_OBJECT(_majortick_count), _majortick_count_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_thick), _minortick_thick_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_style), _minortick_style_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_length), _minortick_length_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_visible), _minortick_visible_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_mode), _minortick_mode_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_interval), _minortick_interval_id);
   g_signal_handler_unblock(GTK_OBJECT(_minortick_count), _minortick_count_id);      g_signal_handler_unblock(GTK_OBJECT(_majorgrid_thick), _majorgrid_thick_id);
   g_signal_handler_unblock(GTK_OBJECT(_majorgrid_style), _majorgrid_style_id);
   g_signal_handler_unblock(GTK_OBJECT(_majorgrid_visible), _majorgrid_visible_id);
   g_signal_handler_unblock(GTK_OBJECT(_minorgrid_thick), _minorgrid_thick_id);
   g_signal_handler_unblock(GTK_OBJECT(_minorgrid_style), _minorgrid_style_id);
   g_signal_handler_unblock(GTK_OBJECT(_minorgrid_visible), _minorgrid_visible_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_font), _ticklabel_font_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_bold), _ticklabel_bold_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_italic), _ticklabel_italic_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_size), _ticklabel_size_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_angle), _ticklabel_angle_change_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_offset), _ticklabel_offset_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_style), _ticklabel_style_id);   
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_visible), _ticklabel_visible_change_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_format_custom), _ticklabel_format_custom_id);
   g_signal_handler_unblock(GTK_OBJECT(_ticklabel_format_box), _ticklabel_format_box_id);

   g_signal_handler_unblock(GTK_OBJECT(_depth), _depth_id);
   g_signal_handler_unblock(GTK_OBJECT(_visible), _visible_id);

   g_signal_handler_unblock(GTK_OBJECT(_apply_xy), _apply_xy_id);
   g_signal_handler_unblock(GTK_OBJECT(_apply_current), _apply_current_id);



}

void ChipsGuiAxisFrame::build_format_list (GtkListStore *format_list)
{
   GtkTreeIter iter;

   if (!format_list)
      return;

   gtk_list_store_clear (format_list);

   gtk_list_store_append( format_list, &iter );
   gtk_list_store_set( format_list, &iter, 0, "Integer", 1, "\%d", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Float", 1, "\%f", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Compact", 1, "\%g", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Power10", 1, "\%z", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Power10 Compact", 1, "\%0.0z", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Power10 Optional ", 1, "\%Z", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00:00:00)", 1, "\%ra1", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00 00 00)", 1, "\%ra2", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00h 00m 00s)", 1, "\%ra3", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00H 00M 00S)", 1, "\%RA3", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00<sup>h</sup>00<sup>m</sup>00<sup>s</sup>)", 1, "\%ra4", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "RA (00<sup>H</sup>00<sup>M</sup>00<sup>S</sup>)", 1, "\%RA4", -1);
   gtk_list_store_append( format_list, &iter);

   gtk_list_store_set( format_list, &iter, 0, "Dec (00<sup>o</sup>00'00\")", 1, "\%dec", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Dec (00:00:00)", 1, "\%dec1", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Dec (00 00 00)", 1, "\%dec2", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Dec (00d 00m 00s)", 1, "\%dec3", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Dec (00D 00M 00S)", 1, "\%DEC3", -1);
   gtk_list_store_append( format_list, &iter);
   gtk_list_store_set( format_list, &iter, 0, "Custom", 1, "", -1);


}

void ChipsGuiAxisFrame::build_majortick_mode_list (GtkListStore *mode_list, ChipsAxisAlgorithmFactory::TransformType axis_type)
{
   GtkTreeIter iter;

   if (!mode_list)
      return;

   gtk_list_store_clear (mode_list);

   if (axis_type == ChipsAxisAlgorithmFactory::WCSTAN)
   {
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "nice", 1, (int)chips_nice, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "limits", 1, (int)chips_limits, -1);
   }
   else
   {
      gtk_list_store_append( mode_list, &iter );
      gtk_list_store_set( mode_list, &iter, 0, "count", 1, (int)chips_count, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "interval", 1, (int)chips_interval, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "nice", 1, (int)chips_nice, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "limits", 1, (int)chips_limits, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "manual", 1, (int)chips_arbitrary, -1);
   }
}

void ChipsGuiAxisFrame::build_minortick_mode_list (GtkListStore *mode_list, ChipsAxisAlgorithmFactory::TransformType axis_type)
{
   GtkTreeIter iter;

   if (!mode_list)
      return;

   gtk_list_store_clear (mode_list);

   if (axis_type == ChipsAxisAlgorithmFactory::WCSTAN)
   {
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "count", 1, (int)chips_nice, -1);
   }
   else
   {
      gtk_list_store_append( mode_list, &iter );
      gtk_list_store_set( mode_list, &iter, 0, "count", 1, (int)chips_count, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "interval", 1, (int)chips_interval, -1);
      gtk_list_store_append( mode_list, &iter);
      gtk_list_store_set( mode_list, &iter, 0, "nice", 1, (int)chips_nice, -1);
   }
}





void ChipsGuiAxisFrame::reset_arbitrary_tick_list ()
{
   ChipsAxis *axis = NULL;
   double *tick_positions;
   string *tick_text;
   unsigned int num_ticks;

   axis = dynamic_cast<ChipsAxis *>(_eng->get_object (&_cid, chips_axis));

   if (!axis)
      return;

   axis->get_arbitrary_tick_positions (num_ticks, &tick_positions, &tick_text);

   _majortick_ticks->initialize_column (tick_positions, num_ticks, 0);
   _majortick_ticks->initialize_column (tick_text, num_ticks, 1);

   _majortick_ticksB->initialize_column (tick_positions, num_ticks, 0);
   _majortick_ticksB->initialize_column (tick_text, num_ticks, 1);

   delete []tick_positions;
   delete []tick_text;

}

bool ChipsGuiAxisFrame::get_tickmode_iter (GtkListStore *mode_list, GtkTreeIter &iter, int value)
{
   GValue gvalue = {0,};
   bool succeed = false;

   gtk_tree_model_get_iter_root (GTK_TREE_MODEL(mode_list), &iter);
   do
   {
      gtk_tree_model_get_value (GTK_TREE_MODEL(mode_list), &iter,
                                    1, &gvalue);
      if (value ==  g_value_get_int (&gvalue))
      {
         succeed = true;
         break;
      }
      g_value_unset (&gvalue);
   } while (gtk_tree_model_iter_next (GTK_TREE_MODEL(mode_list), &iter));

   return succeed;

}

void ChipsGuiAxisFrame::display_subobject (V_SUBOBJECT_TYPE subtype)
{
   if (subtype == chips_axis_label)
      gtk_notebook_set_current_page (GTK_NOTEBOOK(_nbook), _label_page_num);
}

void ChipsGuiAxisFrame::set_apply_xy (bool enable)
{
   ChipsAxis *ax = dynamic_cast<ChipsAxis *>(_eng->get_object (&_cid, chips_axis));

   int orientation = ax->get_orientation ();

   _set_cid.set_axis_dim (X_AXIS);
   _set_cid.set_axis_id ("");

   if (orientation == X_AXIS)
   {
      _set_cid.set_xaxis_id (ax->get_id ());
      _set_cid.set_yaxis_id ("");
   }
   else
   {
      _set_cid.set_yaxis_id (ax->get_id ());
      _set_cid.set_xaxis_id ("");
   }


   if (enable)
   {
      ChipsDataSystem *ds;
      _apply_xy_map[ax->get_ref_cnt ()] = true;
      _eng->get_current_coord (&_set_cid);

      if (_set_cid.get_coord_sys () == DATA)
      {
         ds = ax->get_data_system (_set_cid.get_coord_id ());
         if (ds)
         {
            _set_cid.set_xaxis_id (ds->get_xaxis_id ());
            _set_cid.set_yaxis_id (ds->get_yaxis_id ());
         }
      } 
   }
   else 
      _apply_xy_map[ax->get_ref_cnt ()] = false;

   
}

void ChipsGuiAxisFrame::set_engine_axis_property (void *val, int property, bool ignore)
{
   ChipsSetEngineAxisPropCallback callback (property, &_set_cid);
   this->update_property (&callback, val, ignore);
}

void ChipsGuiAxisFrame::change_tab_callback (GtkNotebook *notebook,
   GtkNotebookPage *notebook_page, int page, void *data)
{
   ChipsGuiAxisFrame *axisFrame = static_cast <ChipsGuiAxisFrame*>(data);

   if (!axisFrame)  return;
   if (!(page == 0 || page == 3)) return;
   axisFrame->reload_arbitrary_tick_list(page); 
}

void ChipsGuiAxisFrame::reload_arbitrary_tick_list(int page)
{
   // changing to 'Axis' or 'Majortick' tab- update Manual ticks table
   if (!(page == 0 || page == 3)) return; 

   if (!_eng) return; 

   ChipsAxis *axis = dynamic_cast<ChipsAxis *>(_eng->get_object (&_cid, chips_axis));

   if (!axis)  return; 

   ChipsGuiLevelsWidget* widget = NULL; 
   double *curr_ticks;
   string *curr_text;
   unsigned int curr_num_ticks = 0;
   double *ticks;
   int num_ticks = 0;
   string *text;
   bool changed = false;

   axis->get_arbitrary_tick_positions (curr_num_ticks, &curr_ticks, &curr_text);

   if (page == 0)  widget = _majortick_ticks; 
      else         widget = _majortick_ticksB; 
   if (!widget) return; 
   widget->get_column (&ticks, num_ticks, 0);
   widget->get_column (&text, num_ticks, 1);

   // check to see if the ticks or text has changed
   if (curr_num_ticks == (unsigned int)num_ticks)
   {
      for (int ii=curr_num_ticks; ii-- && !changed;)
         if ((curr_ticks[ii] != ticks[ii]) || (curr_text[ii] != text[ii]))
            changed = true;
   }
   else 
      changed = true;

   if (changed)
   {
      // copy data
      widget->initialize_column(curr_ticks, curr_num_ticks, 0);
      widget->initialize_column(curr_text, curr_num_ticks, 1);
   }
}
