// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiColorWidget.hh"

#include "ChipsGuiMain.hh"
#include "ChipsEngine.hh"
#include "ChipsId.hh"
#include "ChipsColor.hh"

#include <iostream>

extern ChipsGuiMain* chipsGui;

class ChipsGuiColorImage
{
   public:
      ChipsGuiColorImage (int width, int height)
      {
         this->width = width;
         this->height = height;
         this->pixbuf = NULL;
         this->data = NULL;
         this->image = NULL;
         this->image_built = false;
         this->color_rep = "";
      }

      ~ChipsGuiColorImage ()
      {
         if (this->data)
            delete []this->data;
 
         if (this->image)
            gtk_widget_destroy (this->image);
      }


      void build_image (float red, float green, float blue)
      {
         if (this->image_built)
            return;

         this->data = new guchar[3*width*height];

         this->pixbuf = gdk_pixbuf_new_from_data (this->data,
                                                    GDK_COLORSPACE_RGB,
                                                    FALSE,
                                                    8,
                                                    this->width,
                                                    this->height,
                                                    this->width*3,
                                                    NULL, NULL);

         this->image = gtk_image_new_from_pixbuf (this->pixbuf);
         this->set_color (red, green, blue);

         this->image_built = true;


      }

      void destroy_image ()
      {
         gtk_widget_destroy (this->image);
         delete []this->data;

         this->data = NULL;
         this->image = NULL;

         this->image_built = false;
      }

      void set_color (float red, float green, float blue)
      {

         this->red = red;
         this->green = green;
         this->blue = blue;
         for (int rr=0; rr<this->height; rr++)
            for (int cc=0; cc<this->width; cc++)
            {
               this->data[3*(cc+(rr*width))] = (gchar)(red*255);
               this->data[3*(cc+(rr*width))+1] = (gchar)(green*255);
               this->data[3*(cc+(rr*width))+2] = (gchar)(blue*255);
            }

         gtk_image_set_from_pixbuf (GTK_IMAGE (this->image), this->pixbuf);
      }

      void get_color (float &red, float &green, float &blue)
      {
         red = this->red;
         green = this->green;
         blue = this->blue;
      }

      void set_color_rep (const string &rep, string defstr="")
      {
         this->color_rep = rep;
         this->def_str = defstr; 
      }

      string get_color_rep ()
      {
         return this->color_rep;
      }

      string get_default_str()
      {
         return this->def_str;
      }

      GtkWidget *get_image () { return this->image; };


   protected:

      int width;
      int height;
      guchar *data;
      GtkWidget *image;
      GdkPixbuf *pixbuf;
      bool image_built;
      float red;
      float green;
      float blue;
      string color_rep;
      string def_str; 
};

ChipsGuiColorWidgetCallback::ChipsGuiColorWidgetCallback (void (*callback)(ChipsGuiColorWidget *, void *), void *data)
{
   _callback = callback;
   _callbackData = data;
}

void ChipsGuiColorWidgetCallback::execute (ChipsGuiWidget *caller, void *)
{
   ChipsGuiColorWidget *colorWidget = dynamic_cast<ChipsGuiColorWidget *>(caller);
   _callback (colorWidget, _callbackData);
}



ChipsGuiColorWidget::ChipsGuiColorWidget ()
{
   _currentColor = NULL;
   _toolBar = NULL;
   _colorButton = NULL;

   _standardColors.clear ();

   _valid_callback_ids.push_back ("change_color");

   _foregroundColor = true;
   _prevsel = NULL; 
   _prevclr = ""; 


}

ChipsGuiColorWidget::~ChipsGuiColorWidget ()
{
   map<string, ChipsGuiColorImage *>::iterator ii;

   if (_currentColor) 
      delete _currentColor;

   if (_menuItem)  
      delete []_menuItem; 

   for (ii = _standardColors.begin (); ii != _standardColors.end (); ii++)
      delete ii->second;
   _standardColors.clear ();
}

void ChipsGuiColorWidget::build_form ()
{
   _base_widget = _toolBar = gtk_toolbar_new ();
   AtkObject *accessible;

   // add accessibility info
   accessible = gtk_widget_get_accessible (_toolBar);
   atk_object_set_name (accessible, "colorToolbar");
   atk_object_set_description (accessible, "Color button toolbar\n");

   // setup the color button 
   {
      _currentColor = new ChipsGuiColorImage (40, 20);
      _currentColor->build_image (0, 0, 0);
      _colorButton = gtk_menu_tool_button_new (_currentColor->get_image (), "TEST");
      g_signal_connect (_colorButton, "clicked", G_CALLBACK (custom_color_clicked_callback), this);

      gtk_toolbar_insert (GTK_TOOLBAR(_toolBar), _colorButton, -1);
      gtk_tool_item_set_expand(GTK_TOOL_ITEM(_colorButton), FALSE);
      gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(_colorButton), FALSE);
      gtk_toolbar_set_show_arrow (GTK_TOOLBAR(_toolBar), FALSE);
      gtk_toolbar_set_icon_size (GTK_TOOLBAR(_toolBar), GTK_ICON_SIZE_BUTTON);
      gtk_toolbar_set_style (GTK_TOOLBAR(_toolBar), GTK_TOOLBAR_ICONS);

      accessible = gtk_widget_get_accessible (GTK_WIDGET (_colorButton));
      atk_object_set_name (accessible, "colorButton");
      atk_object_set_description (accessible, "Color Button\n");
      atk_object_set_role (accessible, ATK_ROLE_MENU_BAR);

   }

   // setup menu
   {
      load_standard_colors ();

      _menu = gtk_menu_new ();

      _defaultItem = gtk_menu_item_new_with_label ("Default");
      GtkWidget *customItem = gtk_menu_item_new_with_label ("Custom");
      update_default_color_string(); 
      list<string>::iterator ii;
      int rr = 0;
      int cc = 0;
      int max_row = (int) _colorNames.size () / 3;
      int zz = 0;

      _menuItem = new GtkWidget*[_colorNames.size() + 2]; 
      for (ii=_colorNames.begin (); ii!=_colorNames.end (); ii++, rr++)
      {
         ChipsGuiColorImage *color_image = _standardColors[*ii];
         _menuItem[zz] = gtk_image_menu_item_new ();
         gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (_menuItem[zz]), color_image->get_image ());
         gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM (_menuItem[zz]), TRUE);
         gtk_menu_item_set_label (GTK_MENU_ITEM (_menuItem[zz]),  ii->c_str ());
         gtk_menu_attach (GTK_MENU(_menu), _menuItem[zz], cc, cc+1, rr, rr+1);
         gtk_image_menu_item_get_always_show_image (GTK_IMAGE_MENU_ITEM(_menuItem[zz]));
         g_signal_connect (_menuItem[zz], "activate", G_CALLBACK (standard_color_clicked_callback), this);
         zz++; 

         if (rr == max_row)
         {
            rr = -1;
            cc++;
         }
      }

      GtkWidget *seperator = gtk_separator_menu_item_new ();
      gtk_menu_append (GTK_MENU (_menu), seperator);

      gtk_menu_append (GTK_MENU (_menu), _defaultItem);
      g_signal_connect (_defaultItem, "activate", G_CALLBACK (default_color_clicked_callback), this);

      gtk_menu_append (GTK_MENU (_menu), customItem);
      g_signal_connect (customItem, "activate", G_CALLBACK (custom_color_clicked_callback), this);

      gtk_widget_show_all (_menu);

      _menuItem[zz++] = _defaultItem; 
      _menuItem[zz] = customItem; 

      gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON(_colorButton), _menu);
      g_signal_connect (GTK_MENU_TOOL_BUTTON(_colorButton), "show-menu", G_CALLBACK (menu_clicked_callback), this);

      accessible = gtk_widget_get_accessible (_menu);
      atk_object_set_name (accessible, "colorButtonMenu");
      atk_object_set_description (accessible, "Menu for preset colors \n");
      atk_object_set_role (accessible, ATK_ROLE_MENU);

   }
   
}

void ChipsGuiColorWidget::show ()
{
   gtk_widget_show_all (_toolBar);
}

void ChipsGuiColorWidget::hide ()
{
   gtk_widget_hide (_toolBar);
}


void ChipsGuiColorWidget::custom_color_clicked_callback(GtkWidget *widget, gpointer *data)
{
   ChipsGuiColorWidget *colorWidget = (ChipsGuiColorWidget *)data;
   GtkWidget *colorDialog = gtk_color_selection_dialog_new ("Custom Color");
   GtkWidget *colorSelection = gtk_color_selection_dialog_get_color_selection (GTK_COLOR_SELECTION_DIALOG (colorDialog));
   GdkColor gcolor = {0,0,0,0};
   float r,g,b;
   int result;
   

   colorWidget->_currentColor->get_color (r,g,b);
   gcolor.red = r * 65535;
   gcolor.green = g * 65535;
   gcolor.blue = b * 65535;

   gtk_color_selection_set_current_color (GTK_COLOR_SELECTION (colorSelection), &gcolor);

   result = gtk_dialog_run(GTK_DIALOG(colorDialog));

   if (result == GTK_RESPONSE_OK)
   {
      gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(colorSelection),
                   &gcolor);

      colorWidget->_currentColor->set_color (gcolor.red/65535., gcolor.green/65535., gcolor.blue/65535.);
      colorWidget->_currentColor->set_color_rep ("");
      colorWidget->highlight_color();
      if (colorWidget->_callbacks.count ("change_color") > 0)
         colorWidget->_callbacks["change_color"]->execute (colorWidget, NULL);
   }

   gtk_widget_destroy(colorDialog);

}

void ChipsGuiColorWidget::default_color_clicked_callback(GtkWidget *widget, gpointer *data)
{
   ChipsGuiColorWidget *colorWidget = (ChipsGuiColorWidget *)data;
   colorWidget->set_chips_color ("default");
   if (colorWidget->_callbacks.count ("change_color") > 0)
      colorWidget->_callbacks["change_color"]->execute (colorWidget, NULL);
}

void ChipsGuiColorWidget::standard_color_clicked_callback(GtkWidget *widget, gpointer *data)
{
   ChipsGuiColorWidget *colorWidget = (ChipsGuiColorWidget *)data;
   string label = gtk_menu_item_get_label (GTK_MENU_ITEM (widget));
   colorWidget->set_chips_color (label);  

   if (colorWidget->_callbacks.count ("change_color") > 0)
      colorWidget->_callbacks["change_color"]->execute (colorWidget, NULL);
}

void ChipsGuiColorWidget::menu_clicked_callback(GtkWidget *widget, gpointer *data)
{
   ChipsGuiColorWidget *colorWidget = (ChipsGuiColorWidget *)data;
   if (colorWidget)  colorWidget->update_default_color_string();
}


void ChipsGuiColorWidget::set_chips_color (const string &colorval)
{
   ChipsColor clr;
   if (colorval == "default" && chipsGui)
   {
      ChipsEngine* eng = chipsGui->get_engine();
      char* tmp;
      if (_foregroundColor)
         tmp = eng->get_preference("foreground.display");
      else
         tmp = eng->get_preference("background.display");
      if (tmp)   clr.set_color(tmp);
         else    clr.set_color(colorval);
      _currentColor->set_color (clr.red (), clr.green (), clr.blue ());
      if (tmp)  _currentColor->set_color_rep (colorval, tmp);
         else   _currentColor->set_color_rep (colorval);

      free (tmp);
   }
   else
   {
      clr.set_color(colorval);
      _currentColor->set_color (clr.red (), clr.green (), clr.blue ());
      _currentColor->set_color_rep (colorval);
   }
   highlight_color(); 
}

string ChipsGuiColorWidget::get_chips_color ()
{
   string str_color;
   char clrs[32];
   float r,g,b;

   str_color = _currentColor->get_color_rep ();
   
   if (str_color.length () == 0)
   {
      _currentColor->get_color (r,g,b);
      sprintf(clrs,"%02x%02x%02x", (unsigned int)(r*255), (unsigned int)(g*255), (unsigned int)(b*255) );
      str_color = clrs;
   }

   return str_color;
}

void ChipsGuiColorWidget::update_default_color_string()
{
   if (!chipsGui)  return; 
   ChipsEngine* eng = chipsGui->get_engine();
   if (!eng) return;
   char* tmp;
   if (_foregroundColor)
      tmp = eng->get_preference("foreground.display");
   else
      tmp = eng->get_preference("background.display");
   
   if (tmp) 
   {
      string myclr = get_chips_color();
      string deflbl;
      if (myclr == "default")
      {
         // need to highlight and update default color 
         char *markup = g_markup_printf_escaped ("<span style=\"italic\" bgcolor=\"yellow\">Default   (%s)</span>", tmp);
         GtkWidget* gw3 = gtk_bin_get_child(GTK_BIN(_defaultItem));
         gtk_label_set_markup (GTK_LABEL (gw3), markup);
         ChipsColor clr;

         float rr,gg,bb;
         _currentColor->get_color (rr,gg,bb);
         clr.set_color(tmp);
         if (rr != clr.red() || gg != clr.green() || bb != clr.blue())
         {
            _currentColor->set_color (clr.red (), clr.green (), clr.blue ());
         }
      }
      else
      {
         deflbl = "Default   (";
         deflbl.append(tmp);
         deflbl.append(")");
         gtk_menu_item_set_label(GTK_MENU_ITEM(_defaultItem), deflbl.c_str());
      }
      free (tmp);
   } 
}


void ChipsGuiColorWidget::highlight_color()
{
   string myclr = get_chips_color();
   if (myclr != _prevclr && 
       !(myclr == "default" && _prevclr == "Default")) 
   {
      if (_prevsel)
      {
         // unhighlight previous selection 
         gtk_menu_item_set_label(GTK_MENU_ITEM(_prevsel), _prevclr.c_str()); 
      } 

      GtkWidget* gw2 = gtk_menu_get_active(GTK_MENU(_menu));
      const gchar *clr2 = gtk_menu_item_get_label(GTK_MENU_ITEM(gw2));

      if (myclr == clr2) 
      {
         GtkWidget* gw3 = gtk_bin_get_child(GTK_BIN(gw2));
         char *markup = g_markup_printf_escaped ("<span style=\"italic\" bgcolor=\"yellow\">%s</span>", clr2);
         gtk_label_set_markup (GTK_LABEL (gw3), markup);
         _prevsel = gw2;
         _prevclr = clr2;  
      } 
      else 
      {
         int done=  0, rr, sz = _colorNames.size(); 

         if (myclr == "default")
         {
            string dstr =  _currentColor->get_default_str();

               char *markup = g_markup_printf_escaped ("<span style=\"italic\" bgcolor=\"yellow\">Default   (%s)</span>", dstr.c_str());
            GtkWidget* gw3 = gtk_bin_get_child(GTK_BIN(_menuItem[sz]));
            gtk_label_set_markup (GTK_LABEL (gw3), markup);
            _prevsel = _menuItem[sz];
            _prevclr = "Default";
            done = 1;
         }

         for (rr = 0; rr < sz && !done; rr++)
         {
             const gchar *clr2 = gtk_menu_item_get_label(GTK_MENU_ITEM(_menuItem[rr]));
            if (clr2 && strcmp(clr2,myclr.c_str()) == 0)
            {
               char *markup = g_markup_printf_escaped ("<span style=\"italic\" bgcolor=\"yellow\">%s</span>", clr2);
               GtkWidget* gw3 = gtk_bin_get_child(GTK_BIN(_menuItem[rr]));
               gtk_label_set_markup (GTK_LABEL (gw3), markup);
                _prevsel = _menuItem[rr];
                _prevclr = clr2;
               done = 1; 
            }
         }  

         if (!done)  // check for custom color  
         {
            ChipsColor cc;
            if (cc.set_color(myclr) == ERR_STATUS_GOOD) 
            {
               char *markup = g_markup_printf_escaped ("<span style=\"italic\" bgcolor=\"yellow\">Custom   (%s)</span>", myclr.c_str());
               GtkWidget* gw3 = gtk_bin_get_child(GTK_BIN(_menuItem[sz+1]));
               gtk_label_set_markup (GTK_LABEL (gw3), markup);
               _prevsel = _menuItem[sz+1];
               _prevclr = "Custom";
               done = 1;
            }  
         } 
      }
   }

}

void ChipsGuiColorWidget::load_standard_colors ()
{
   ChipsColor chips_color;
   list <string>::iterator ii;

   _colorNames.push_back ("white");
   _colorNames.push_back ("black");
   _colorNames.push_back ("red");
   _colorNames.push_back ("orange");
   _colorNames.push_back ("brown");
   _colorNames.push_back ("yellow");
   _colorNames.push_back ("green");
   _colorNames.push_back ("cyan");
   _colorNames.push_back ("blue");
   _colorNames.push_back ("purple");
   _colorNames.push_back ("slateblue");
   _colorNames.push_back ("magenta");
   _colorNames.push_back ("navy");
   _colorNames.push_back ("mediumblue");
   _colorNames.push_back ("cornflower");
   _colorNames.push_back ("steelblue");
   _colorNames.push_back ("skyblue");
   _colorNames.push_back ("seagreen");
   _colorNames.push_back ("forest");
   _colorNames.push_back ("olive");
   _colorNames.push_back ("lime");
   _colorNames.push_back ("chartreuse");
   _colorNames.push_back ("springgreen");
   _colorNames.push_back ("aquamarine");
   _colorNames.push_back ("turquoise");
   _colorNames.push_back ("sienna");
   _colorNames.push_back ("peru");
   _colorNames.push_back ("chocolate");
   _colorNames.push_back ("goldenrod");
   _colorNames.push_back ("tan");
   _colorNames.push_back ("gold");
   _colorNames.push_back ("khaki");
   _colorNames.push_back ("coral");
   _colorNames.push_back ("salmon");
   _colorNames.push_back ("tomato");
   _colorNames.push_back ("peach");
   _colorNames.push_back ("pink");
   _colorNames.push_back ("mistyrose");
   _colorNames.push_back ("maroon");
   _colorNames.push_back ("firebrick");
   _colorNames.push_back ("darkred");
   _colorNames.push_back ("orchid");
   _colorNames.push_back ("plum");
   _colorNames.push_back ("violet");
   _colorNames.push_back ("thistle");
   _colorNames.push_back ("slategray");
   _colorNames.push_back ("gray");
   _colorNames.push_back ("beige");
   _colorNames.push_back ("wheat");
   _colorNames.push_back ("bisque");
   _colorNames.push_back ("linen");
   _colorNames.push_back ("seashell");
   _colorNames.push_back ("snow");
   _colorNames.push_back ("lavender");
   _colorNames.push_back ("azure");
   _colorNames.push_back ("honeydew");
   _colorNames.push_back ("ivory");

   _standardColors.clear ();
   for (ii = _colorNames.begin (); ii != _colorNames.end (); ii++)
   {
      ChipsGuiColorImage *color = new ChipsGuiColorImage (5,5);
      _standardColors[*ii] = color;
      chips_color.set_color (*ii);
      color->build_image (chips_color.red (), 
                          chips_color.green (), 
                          chips_color.blue ());
      color->set_color_rep (*ii);
   }
      
}

void ChipsGuiColorWidget::enable ()
{
   gtk_widget_set_sensitive (GTK_WIDGET(_colorButton), TRUE);
}

void ChipsGuiColorWidget::disable ()
{
   gtk_widget_set_sensitive (GTK_WIDGET(_colorButton), FALSE);
}

