// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiLabelFrame.hh"
#include "ChipsGuiLabelPropWidget.hh"
#include "ChipsGuiMain.hh"
#include "ChipsEngine.hh"
#include "ChipsId.hh"
#include "ChipsGuiPropPane.hh"
#include "ChipsGuiFontInfo.hh"
#include "ChipsCmdManager.hh"
#include "ChipsMatrix.hh"

#include <algorithm>

using namespace std;

extern ChipsGuiMain* chipsGui;

class ChipsLabelMoveCallback : public ChipsGuiPropFrameCallback
{
   public:
      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         double *pos = static_cast<double *>(data);
         if ((pos != NULL) && eng)
            eng->move(&cid, chips_label, pos[0], pos[1], V_MOVE_ABSOLUTE);
      }
};



// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiLabelFrame::ChipsGuiLabelFrame():ChipsGuiPropFrame(),ChipsGuiTooltipObject()
{
   _labelprop = NULL;
   _object_states = STATE_CHNG_PROP_LBL;
   _object_states |= STATE_CHNG_AXS_FMT;
}

// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiLabelFrame::~ChipsGuiLabelFrame()
{
   if (_labelprop)  delete _labelprop; 
}

void ChipsGuiLabelFrame::build_form()
{

   _frame = gtk_frame_new("Label Properties");

   _labelprop = new ChipsGuiLabelPropWidget ();
   _labelprop->build_form ();

   gtk_container_add(GTK_CONTAINER(_frame), _labelprop->base_widget ());


   _labelprop->add_callback ("change_color", new ChipsGuiLabelPropWidgetCallback (color_change_callback, this));
   _labelprop->add_callback ("change_text", new ChipsGuiLabelPropWidgetCallback (text_change_callback, this));
   _labelprop->add_callback ("change_font", new ChipsGuiLabelPropWidgetCallback (font_change_callback, this));
   _labelprop->add_callback ("change_fontstyle", new ChipsGuiLabelPropWidgetCallback (fontstyle_change_callback, this));
   _labelprop->add_callback ("change_size", new ChipsGuiLabelPropWidgetCallback (size_change_callback, this));
   _labelprop->add_callback ("change_angle", new ChipsGuiLabelPropWidgetCallback (angle_change_callback, this));
   _labelprop->add_callback ("change_position", new ChipsGuiLabelPropWidgetCallback (position_change_callback, this));
   _labelprop->add_callback ("change_coordinate", new ChipsGuiLabelPropWidgetCallback (coordinate_callback, this));
   _labelprop->add_callback ("change_halign", new ChipsGuiLabelPropWidgetCallback (halign_change_callback, this));
   _labelprop->add_callback ("change_valign", new ChipsGuiLabelPropWidgetCallback (valign_change_callback, this));
   _labelprop->add_callback ("change_depth", new ChipsGuiLabelPropWidgetCallback (depth_callback, this));
   _labelprop->add_callback ("change_visibility", new ChipsGuiLabelPropWidgetCallback (vis_callback, this));

}


void ChipsGuiLabelFrame::set_color(const string &color)
{
   set_engine_property((void*)color.c_str(), SET_LABEL_COLOR); 
}

void ChipsGuiLabelFrame::set_angle(float val)
{
   set_engine_property(static_cast<void*>(&val), SET_LABEL_ANGLE); 
   
}

void ChipsGuiLabelFrame::set_font(const string &val)
{
   set_engine_property(static_cast<void*>(const_cast<char*>(val.c_str())),
      SET_LABEL_FONT);
}

void ChipsGuiLabelFrame::set_depth(int val)
{
   set_engine_property(static_cast<void*>(&val), SET_LABEL_DEPTH);
}

void ChipsGuiLabelFrame::set_fontstyle(int val)
{
   const char *style = v_fontstyle_to_string((V_FONTSTYLE_TYPE)val);
   set_engine_property(static_cast<void*>(const_cast<char *>(style)),
      SET_LABEL_STYLE);
}

void ChipsGuiLabelFrame::set_halign(float val)
{
   set_engine_property(static_cast<void*>(&val), SET_LABEL_HALIGN);
}

void ChipsGuiLabelFrame::set_valign(float val)
{
   set_engine_property(static_cast<void*>(&val), SET_LABEL_VALIGN);
}

void ChipsGuiLabelFrame::set_size(int val)
{
   set_engine_property(static_cast<void*>(&val), SET_LABEL_SIZE);
}

void ChipsGuiLabelFrame::set_text(const string &val)
{
   set_engine_property(static_cast<void*>(const_cast<char*>(val.c_str ())),
      SET_LABEL_TEXT);
}

void ChipsGuiLabelFrame::set_visibility(bool val)
{
   int ival = val;
   set_engine_property(static_cast<void*>(&ival), SET_LABEL_VISIBLE);
}


void ChipsGuiLabelFrame::set_position(double xpos, double ypos)
{
   ChipsLabelMoveCallback move_callback;
   double pos[2] = {xpos, ypos};
   this->update_property (&move_callback, &pos[0]);
}

void ChipsGuiLabelFrame::set_coordinate(int coordinate)
{
   ChipsId cid;
   COORD_SYS coordsys = (COORD_SYS)coordinate;
   string coordid = "";

   _eng->get_current_coord (&cid);

   if ((coordsys == DATA) && (cid.get_coord_sys () != DATA))
      chipsGui->log_error ("Data coordinate system does not exists.");
   else
      coordid = cid.get_coord_id ();

   coordid = ChipsCoordinate::extract_id (coordsys, coordid);

   _eng->change_annotation_coordinate (&cid, chips_label, coordsys, coordid);
}




void ChipsGuiLabelFrame::color_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame)  return; //ensure we have needed pointer 
   string clr = widget->get_color ();
     
   labelFrame->set_color(clr);
}

void ChipsGuiLabelFrame::text_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   string text = widget->get_text ();
   labelFrame->set_text(text); 
}

void ChipsGuiLabelFrame::depth_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   int depth = widget->get_depth ();
   labelFrame->set_depth(depth);   

}

void ChipsGuiLabelFrame::font_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   string font = widget->get_font ();
   labelFrame->set_font (font);
}

void ChipsGuiLabelFrame::fontstyle_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   int fontstyle = widget->get_fontstyle ();
   labelFrame->set_fontstyle (fontstyle);

}

void ChipsGuiLabelFrame::size_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   int size = widget->get_size ();
   labelFrame->set_size (size);

}

void ChipsGuiLabelFrame::angle_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame) return; //ensure we have needed pointer
   float angle = widget->get_angle ();
   labelFrame->set_angle (angle);

}

void ChipsGuiLabelFrame::position_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);
   double x,y;

   if (!labelFrame)  return; //ensure we have needed pointer
   widget->get_position (x,y);
   labelFrame->set_position(x,y);
}

void ChipsGuiLabelFrame::coordinate_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame)  return; //ensure we have needed pointer
   int coordinate = widget->get_coordinate ();
   labelFrame->set_coordinate (coordinate);
}

void ChipsGuiLabelFrame::halign_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame)  return; //ensure we have needed pointer
   float halign = widget->get_halign ();
   labelFrame->set_halign (halign);
}

void ChipsGuiLabelFrame::valign_change_callback(ChipsGuiLabelPropWidget *widget, void *data)
{

   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);
   
   if (!labelFrame)  return; //ensure we have needed pointer
   float valign = widget->get_valign ();
   labelFrame->set_valign (valign);

}

void ChipsGuiLabelFrame::vis_callback(ChipsGuiLabelPropWidget *widget, void *data)
{
   ChipsGuiLabelFrame *labelFrame = static_cast <ChipsGuiLabelFrame*>(data);

   if (!labelFrame)  return; //ensure we have needed pointer
   bool val = widget->get_visibility ();
   labelFrame->set_visibility (val);
}


void ChipsGuiLabelFrame::initialize(ChipsId* id)
{
   int depthval;
   string fontval;
   V_FONTSTYLE_TYPE fontstyleval;
   float halignval;
   float valignval;
   int sizeval;
   string textval;
   int visval;
   COORD_SYS coordsys;
   string coordid;
   ChipsVector pos(3);
   string colorval;
   float angleval;
   string tmp_fontstyleval;


   if (!id)  return; 

   if (chipsGui && !_eng) 
      _eng = chipsGui->get_engine(); 

   if (id != &_cid)
   {
      _cid.reset_window_id(id->get_window_id()); 
      _cid.set_frame_id(id->get_frame_id()); 
      _cid.set_plot_id(id->get_plot_id()); 
      _cid.set_label_id(id->get_label_id()); 
   }

   _eng->get_property(&_cid, (void*)&colorval, SET_LABEL_COLOR);
   _eng->get_property(&_cid, (void*)&angleval, SET_LABEL_ANGLE);
   _eng->get_property(&_cid, (void*)&fontval, SET_LABEL_FONT);
   _eng->get_property(&_cid, (void*)(&tmp_fontstyleval), SET_LABEL_STYLE);
   _eng->get_property(&_cid, (void*)&depthval, SET_LABEL_DEPTH);
   _eng->get_property(&_cid, (void*)&sizeval, SET_LABEL_SIZE);
   _eng->get_property(&_cid, (void*)&halignval, SET_LABEL_HALIGN);
   _eng->get_property(&_cid, (void*)&valignval, SET_LABEL_VALIGN);
   _eng->get_property(&_cid, (void*)&textval, SET_LABEL_TEXT);
   _eng->get_property(&_cid, (void*)&visval, SET_LABEL_VISIBLE);

   _eng->get_object_coordinfo (&_cid, chips_label, coordsys, coordid);
   _eng->get_annotation_position (&_cid, chips_label, pos);

   string_to_v_fontstyle (tmp_fontstyleval.c_str (), &fontstyleval);

   _labelprop->set_color (colorval);
   _labelprop->set_coordinate ((int)coordsys);
   _labelprop->set_angle (angleval);
   _labelprop->set_position (pos[0],pos[1]);
   _labelprop->set_font (fontval);
   _labelprop->set_fontstyle (fontstyleval);
   _labelprop->set_size (sizeval);
   _labelprop->set_halign (halignval);
   _labelprop->set_valign (valignval);
   _labelprop->set_text (textval);
   _labelprop->set_depth (depthval);
   _labelprop->set_visibility (visval);

   // get tick formats
   int xmode = 0, ymode = 0;
   string xformat, yformat;
   ChipsId tmpid (_cid);
   tmpid.set_axis_id(_cid.get_xaxis_id());
   tmpid.set_axis_dim (X_AXIS);
   _eng->get_property(&tmpid, (void*)&xformat, SET_AXIS_TICKFORMAT);
   tmpid.set_axis_id(_cid.get_yaxis_id());
   tmpid.set_axis_dim (Y_AXIS);
   _eng->get_property(&tmpid, (void*)&yformat, SET_AXIS_TICKFORMAT);

   transform(xformat.begin(), xformat.end(), xformat.begin(), ::tolower);
   transform(yformat.begin(), yformat.end(), yformat.begin(), ::tolower);

   if (xformat.find("ra") != string::npos ||
       xformat.find("dec") != string::npos)
      xmode = 1;
   if (yformat.find("ra") != string::npos ||
       yformat.find("dec") != string::npos)
      ymode = 1;

   _labelprop->set_position_format (xmode, ymode);

   // enable/disable the data coord pulldown entry based on existing coords
   ChipsId mcid;
   mcid.set_coord_sys(DATA);
   if (_eng->check_coordinate_exists(&mcid))  
      _labelprop->enable_data_coordinate();
   else 
      _labelprop->disable_data_coordinate();

}


