// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiLevelsWidget.hh"
#include <cstdlib>
#include <cstring>
#include <sstream>
#include <gdk/gdkkeysyms.h>

#include "ChipsGuiMain.hh"

using namespace std;

extern ChipsGuiMain* chipsGui;

ChipsGuiLevelsWidgetCallback::ChipsGuiLevelsWidgetCallback (void (*callback)(ChipsGuiLevelsWidget *, void *), void *data)
{
   _callback = callback;
   _callbackData = data;
}

void ChipsGuiLevelsWidgetCallback::execute (ChipsGuiWidget *caller, void *)
{
   ChipsGuiLevelsWidget *levelWidget = dynamic_cast<ChipsGuiLevelsWidget *>(caller);
   _callback (levelWidget, _callbackData);
}


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiLevelsWidget::ChipsGuiLevelsWidget(int num_columns, GType type[])
{
   treeview = NULL;
   min_level = 0;
   last_iter = NULL; 

   _valid_callback_ids.push_back ("change_levels");
   _valid_callback_ids.push_back ("delete_levels");
   _valid_callback_ids.push_back ("add_levels");

   _num_columns = num_columns;
   _column_types = new GType[num_columns];

   

   _column_titles = new string [num_columns];
   for (int ii=num_columns; ii--;)
   {
      _column_types[ii] = type[ii];
      _column_titles[ii] = "";
   }
}

// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiLevelsWidget::~ChipsGuiLevelsWidget()
{
   delete []_column_types;
   delete []_column_titles;
}

void ChipsGuiLevelsWidget::add_item()
{
   GtkTreeIter iter;
   gtk_list_store_append (GTK_LIST_STORE (items_model), &iter);
   // add item as NaN so user knows to change it 
   GValue value = {0, };

   for (unsigned int ii=0; ii<_num_columns; ii++)
   {
      init_gvalue (value, _column_types[ii]);  
      gtk_list_store_set_value (GTK_LIST_STORE(items_model), &iter, ii, &value);
      g_value_unset (&value);
   }

   // move to new cell and set focus 
   GtkTreeModel *model = gtk_tree_view_get_model (treeview);
   GtkTreePath *path = gtk_tree_model_get_path(model, &iter);
   GtkTreeViewColumn* cl = gtk_tree_view_get_column(GTK_TREE_VIEW(treeview), 0);
   gtk_tree_view_set_cursor(GTK_TREE_VIEW(treeview),path, cl, TRUE);

   if (gtk_tree_model_iter_n_children (GTK_TREE_MODEL (items_model), NULL) == min_level+1)
      gtk_widget_set_sensitive (_delBtn, TRUE);
}

void ChipsGuiLevelsWidget::delete_item()
{
   GtkTreeIter iter;
   GtkTreeModel *model = gtk_tree_view_get_model (treeview);
   GtkTreeSelection *selection = gtk_tree_view_get_selection (treeview);

   if (gtk_tree_selection_get_selected (selection, NULL, &iter))
   {
      gint i;
      GtkTreePath *path;
      int num_levels = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (model), NULL);

      if (num_levels > min_level)
      {
         path = gtk_tree_model_get_path (model, &iter);
         i = gtk_tree_path_get_indices (path)[0];
         gtk_list_store_remove (GTK_LIST_STORE (model), &iter);

         gtk_tree_path_free (path);

         if (num_levels - 1 == min_level)
            gtk_widget_set_sensitive (_delBtn, FALSE);
      }
   }
}

void ChipsGuiLevelsWidget::edit_item(GtkCellRendererText *cell, 
   const gchar *path_string, const gchar *new_text)
{
   GtkTreeModel *model = gtk_tree_view_get_model (treeview);
   GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
   GtkTreeIter iter;
   int column = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (cell), "column"));

   gtk_tree_model_get_iter (model, &iter, path);

   GValue oval = {0,}; 
   GValue oval2 = {0,}; 
   GValue value = {0,};
   GValue value2 = {0,};
   gtk_tree_model_get_value (model, &iter, column, &oval);

   g_value_init(&value, _column_types[column]);

   if (!set_string_to_gvalue (value, new_text,  _column_types[column]))
      g_value_copy (&oval, &value);  // invalid value- reset to previous value  
   
   gtk_list_store_set_value(GTK_LIST_STORE(model), &iter, column, &value);
 
   if (column == 0 && _num_columns == 2  && _column_types[1] == G_TYPE_STRING)  
   {
      // when editing first column (positions) and more than one column exists
      // move to the values column afterwards 
      if (last_iter)   gtk_tree_iter_free(last_iter);
      last_iter = gtk_tree_iter_copy(&iter); 
      GtkTreeViewColumn* cl = gtk_tree_view_get_column(GTK_TREE_VIEW(treeview), 1);
      gtk_tree_model_get_value (model, &iter, 1, &oval2);
      const gchar *gc = g_value_get_string(&oval2);
      if (gc && !strlen(gc))
      {
         // if the second column is blank, seed it with the position value
         g_value_init(&value2, _column_types[1]);
         set_string_to_gvalue (value2, new_text,  _column_types[1]);
         gtk_list_store_set_value(GTK_LIST_STORE(model), &iter, 1, &value2);
      }
      g_value_unset(&oval2);

      // since gtk treeviews update themselves on idle, we add a single run
      // idle callback that repositions the cursor to the value column for
      // the entry we want (tracked by last_iter) 
      g_idle_add( (GSourceFunc)set_cursor, this );
   }

   g_value_unset(&oval);

   gtk_tree_path_free (path);
}

void ChipsGuiLevelsWidget::del_item_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiLevelsWidget* me = (ChipsGuiLevelsWidget*)data;
   if (me)
   {
      me->delete_item(); 
      if (me->_callbacks.count ("delete_levels") > 0)
         me->_callbacks["delete_levels"]->execute (me, NULL);
   }
}

gboolean ChipsGuiLevelsWidget::del_item_keypress_callback(GtkWidget *widget,
                                             GdkEvent *event,
                                             gpointer data)
{
   ChipsGuiLevelsWidget* me = (ChipsGuiLevelsWidget*)data;

   if ((event->key.keyval != GDK_Delete) && 
       (event->key.keyval != GDK_KP_Delete) && 
       (event->key.keyval != GDK_BackSpace))
      return FALSE;

   if (me)
   {
      me->delete_item();
      if (me->_callbacks.count ("delete_levels") > 0)
         me->_callbacks["delete_levels"]->execute (me, NULL);
      return TRUE;
   }
   return FALSE;
}


void ChipsGuiLevelsWidget::add_item_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiLevelsWidget* me = (ChipsGuiLevelsWidget*)data;
   if (me)
   {
      me->add_item(); 
      if (me->_callbacks.count ("add_levels") > 0)
         me->_callbacks["add_levels"]->execute (me, NULL);
   }
}

void ChipsGuiLevelsWidget::edit_item_callback(GtkCellRendererText *cell, const gchar *path_string, const gchar *new_text, gpointer data)
{
   ChipsGuiLevelsWidget* me = (ChipsGuiLevelsWidget*)data;
   if (me)
   {
      me->edit_item(cell, path_string, new_text);
      if (me->_callbacks.count ("change_levels") > 0)
         me->_callbacks["change_levels"]->execute (me, NULL);
   }

}

void ChipsGuiLevelsWidget::build_form()
{
   AtkObject *accessible;

   _base_widget = vbox = gtk_vbox_new(FALSE, 5);

   sw = gtk_scrolled_window_new(NULL, NULL);
   gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
                                        GTK_SHADOW_ETCHED_IN);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
   gtk_box_pack_start(GTK_BOX(vbox), sw, TRUE, TRUE, 0);

   /* create model */
   items_model = GTK_TREE_MODEL(gtk_list_store_newv(_num_columns, &_column_types[0]));
//   gtk_list_store_clear(GTK_LIST_STORE(items_model));

   /* create tree view */
   treeview = GTK_TREE_VIEW(gtk_tree_view_new_with_model(items_model));
   gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
   gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)),
                                   GTK_SELECTION_SINGLE);

   g_signal_connect(GTK_WIDGET(treeview), "key-press-event",
                                G_CALLBACK(del_item_keypress_callback), this);


   // add column
   for (unsigned int ii=0; ii<_num_columns; ii++)
   {
      GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
      g_object_set(renderer, "editable", TRUE, NULL); 
      g_signal_connect(renderer, "edited", G_CALLBACK (edit_item_callback), this);
      g_object_set_data(G_OBJECT(renderer), "column", GINT_TO_POINTER(ii));
      
      gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW(treeview),
                                               -1, _column_titles[ii].c_str (), renderer,
                                               "text", ii, NULL);
   }
   GtkTreeViewColumn* tvcol = gtk_tree_view_get_column(GTK_TREE_VIEW(treeview), 0); 
   gtk_tree_view_column_set_sort_column_id(tvcol, 0); 

   g_object_unref(items_model);
   gtk_container_add(GTK_CONTAINER(sw), GTK_WIDGET(treeview));

   /* some buttons */
   hbox = gtk_hbox_new(FALSE,4);
   gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

   _addBtn = gtk_button_new_with_label("Add");
   g_signal_connect(_addBtn, "clicked",
                     G_CALLBACK(add_item_callback), this);
   gtk_box_pack_start(GTK_BOX(hbox), _addBtn, TRUE, TRUE, 0);

   _delBtn = gtk_button_new_with_label("Remove");
   g_signal_connect(_delBtn, "clicked",
                     G_CALLBACK(del_item_callback), this);
   gtk_box_pack_start(GTK_BOX(hbox), _delBtn, TRUE, TRUE, 0);

   // add accessibility info
   accessible = gtk_widget_get_accessible (_base_widget);
   atk_object_set_name (accessible, "levelsWidget");
   atk_object_set_description (accessible, "Level widget\n");
   accessible = gtk_widget_get_accessible (_addBtn);
   atk_object_set_name (accessible, "addButton");
   atk_object_set_description (accessible, "Level widget add button\n");
   accessible = gtk_widget_get_accessible (_delBtn);
   atk_object_set_name (accessible, "deleteButton");
   atk_object_set_description (accessible, "Level widget delete button\n");
   

}

void ChipsGuiLevelsWidget::initialize_column (double *vals, int cnt, int column)
{
   initialize_column<double> (vals, cnt, column);
}

void ChipsGuiLevelsWidget::initialize_column (float *vals, int cnt, int column)
{
   initialize_column<float> (vals, cnt, column);
}

void ChipsGuiLevelsWidget::initialize_column (string *vals, int cnt, int column)
{
   initialize_column<string> (vals, cnt, column);
}

void ChipsGuiLevelsWidget::initialize_column (int *vals, int cnt, int column)
{
   initialize_column<int> (vals, cnt, column);
}


void ChipsGuiLevelsWidget::get_column (double **elems, int &out_cnt, int column)
{
   *elems = get_column<double> (out_cnt, column);
}

void ChipsGuiLevelsWidget::get_column (float **elems, int &out_cnt, int column)
{
   *elems = get_column<float> (out_cnt, column);
}

void ChipsGuiLevelsWidget::get_column (string **elems, int &out_cnt, int column)
{
   *elems = get_column<string> (out_cnt, column);
}

void ChipsGuiLevelsWidget::get_column (int **elems, int &out_cnt, int column)
{
   *elems = get_column<int> (out_cnt, column);
}



void ChipsGuiLevelsWidget::set_gvalue (GValue &value, void *val, GType type)
{
   switch (type)
   {
      case G_TYPE_INT:
         g_value_set_int (&value, *(int *)val);
         break;
      case G_TYPE_FLOAT:
         g_value_set_float (&value, *(float *)val);
         break;
      case G_TYPE_DOUBLE:
         g_value_set_double (&value, *(double *)val);
         break;
      case G_TYPE_STRING:
      {
         const char *c_str = ((string *)val)->c_str ();
         g_value_set_string (&value, c_str);
         break;
      }
   }
}

bool ChipsGuiLevelsWidget::set_string_to_gvalue (GValue &value, const string &val, GType type)
{
   bool succeed = true;
   stringstream ss;
   ss.str (val);
   switch (type)
   {
      case G_TYPE_INT:
      {
         int ival;
         ss >> ival;
         g_value_set_int (&value, ival);
         succeed = !ss.fail ();
         break;
      }
      case G_TYPE_FLOAT:
      {
         float fval;
         ss >> fval;
         g_value_set_float (&value, fval);
         succeed = !ss.fail ();
         break; 
      }
      case G_TYPE_DOUBLE:
      {
         double dval;
         ss >> dval;
         g_value_set_double (&value, dval);
         succeed = !ss.fail ();
         break;
      }
      case G_TYPE_STRING:
      {
         const char *c_str = val.c_str ();
         g_value_set_string (&value, c_str);
         succeed = true;
         break;
      }
      default:
         succeed = false;
      break;
   }
   return succeed;
}

void ChipsGuiLevelsWidget::init_gvalue (GValue &value, GType type)
{
   g_value_init(&value, type);
   switch (type)
   {
      case G_TYPE_INT:
         g_value_set_int (&value, 0);
         break;
      case G_TYPE_FLOAT:
      {
         char* ptr = NULL;
         float fval = strtod("nan", &ptr);
         g_value_set_float (&value, fval);
         break;
      }
      case G_TYPE_DOUBLE:
      {
         char* ptr = NULL;
         double dval = strtod("nan", &ptr);
         g_value_set_double (&value, dval);
         break;
      }
      case G_TYPE_STRING:
      {
         const char c_str[] = {"\0"};
         g_value_set_string (&value, c_str);
         break;
      }
   }
}

void ChipsGuiLevelsWidget::get_gvalue_data (GValue &value, int &data)
{
   data = g_value_get_int (&value);
}

void ChipsGuiLevelsWidget::get_gvalue_data (GValue &value, float &data)
{
   data = g_value_get_float (&value);
}

void ChipsGuiLevelsWidget::get_gvalue_data (GValue &value, double &data)
{
   data = g_value_get_double (&value);
}

void ChipsGuiLevelsWidget::get_gvalue_data (GValue &value, string &data)
{
   data = g_value_get_string (&value);
}

      
void ChipsGuiLevelsWidget::show ()
{
   gtk_widget_show_all (_base_widget);
}

void ChipsGuiLevelsWidget::hide ()
{
   gtk_widget_hide (_base_widget);
}

void ChipsGuiLevelsWidget::set_column_title (const string &title, unsigned int column)
{
   if (column >= _num_columns)
      return;

   _column_titles[column] = title;

   if (treeview)
   {
      GtkTreeViewColumn *gtk_column = gtk_tree_view_get_column (GTK_TREE_VIEW (treeview), column);
      gtk_tree_view_column_set_title (gtk_column, title.c_str ());
   }
      
      
}

void ChipsGuiLevelsWidget::disable_widget ()
{
   gtk_widget_set_sensitive (_addBtn, FALSE);
   gtk_widget_set_sensitive (_delBtn, FALSE);

   gtk_widget_set_sensitive (GTK_WIDGET(treeview), FALSE);

}

void ChipsGuiLevelsWidget::enable_widget ()
{
   gtk_widget_set_sensitive (_addBtn, TRUE);
   gtk_widget_set_sensitive (_delBtn, TRUE);

   gtk_widget_set_sensitive (GTK_WIDGET(treeview), TRUE);
}


gboolean ChipsGuiLevelsWidget::set_cursor(ChipsGuiLevelsWidget *data)
{
   // set cell location in the treeview based on the position info
   // stored in last_iter
   ChipsGuiLevelsWidget* me = (ChipsGuiLevelsWidget*)data;
   GtkTreeView* tv = me->treeview;     
   GtkTreeViewColumn* cl = gtk_tree_view_get_column(GTK_TREE_VIEW(tv), 1);
   GtkTreeModel *model = gtk_tree_view_get_model(tv);
   GtkTreePath *tmppath = gtk_tree_model_get_path(model, me->last_iter);
   gtk_tree_view_set_cursor(GTK_TREE_VIEW(tv), tmppath, cl, TRUE);
   gtk_tree_path_free(tmppath);
   return( FALSE );  // so set_cursor is only called once per edit
}

