// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiLineFrame.hh"
#include "ChipsGuiMain.hh"
#include "ChipsEngine.hh"
#include "ChipsLine.hh"
#include "ChipsGuiLinePropWidget.hh"
#include "ChipsCmdManager.hh"

#include <algorithm>

using namespace std;

extern ChipsGuiMain* chipsGui;

class ChipsLineMoveCallback : public ChipsGuiPropFrameCallback
{
   public:
      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         double *pos = static_cast<double *>(data);
         if ((pos != NULL) && eng)
            eng->move(&cid, chips_line, pos[0], pos[1], V_MOVE_ABSOLUTE);
      }
};


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiLineFrame::ChipsGuiLineFrame():ChipsGuiPropFrame(),ChipsGuiTooltipObject()
{
   _lineprop = NULL;
   _object_states = STATE_CHNG_PROP_LINE;
   _object_states |= STATE_CHNG_AXS_FMT;
}

// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiLineFrame::~ChipsGuiLineFrame()
{
   if (_lineprop)
      delete _lineprop;
}

void ChipsGuiLineFrame::build_form()
{
   _frame = gtk_frame_new("Line Properties");

   _lineprop = new ChipsGuiLinePropWidget ();
   _lineprop->build_form ();

   gtk_container_add(GTK_CONTAINER(_frame), _lineprop->base_widget ());

   _lineprop->add_callback ("change_color", new ChipsGuiLinePropWidgetCallback (color_callback, this));
   _lineprop->add_callback ("change_position", new ChipsGuiLinePropWidgetCallback (position_callback, this));
   _lineprop->add_callback ("change_coordinate", new ChipsGuiLinePropWidgetCallback (coordinate_callback, this));
   _lineprop->add_callback ("change_style", new ChipsGuiLinePropWidgetCallback (style_callback, this));
   _lineprop->add_callback ("change_depth", new ChipsGuiLinePropWidgetCallback (depth_callback, this));
   _lineprop->add_callback ("change_thickness", new ChipsGuiLinePropWidgetCallback (thick_callback, this));
   _lineprop->add_callback ("change_extent", new ChipsGuiLinePropWidgetCallback (extent_callback, this));
   _lineprop->add_callback ("change_visibility", new ChipsGuiLinePropWidgetCallback (vis_callback, this));

}

void ChipsGuiLineFrame::set_color(string color)
{
   set_engine_property ((void*)color.c_str(), SET_LINE_COLOR); 
}

void ChipsGuiLineFrame::set_position(double xpos, double ypos)
{
   ChipsLineMoveCallback move_callback;
   double pos[2] = {xpos, ypos};
   this->update_property (&move_callback, &pos[0]);
}

void ChipsGuiLineFrame::set_coordinate(int coordinate)
{
   ChipsId cid;
   COORD_SYS coordsys = (COORD_SYS)coordinate;
   string coordid = "";

   _eng->get_current_coord (&cid);

   if ((coordsys == DATA) && (cid.get_coord_sys () != DATA))
      chipsGui->log_error ("Data coordinate system does not exists.");
   else
      coordid = cid.get_coord_id ();

   coordid = ChipsCoordinate::extract_id (coordsys, coordid);

   _eng->change_annotation_coordinate (&cid, chips_line, coordsys, coordid);
}

void ChipsGuiLineFrame::set_thick(float val)
{
   set_engine_property (static_cast<void*>(&val), 
      SET_LINE_THICKNESS); 
}

void ChipsGuiLineFrame::set_style(int val)
{
   set_engine_property (static_cast<void*>(&val),
      SET_LINE_STYLE);
}

void ChipsGuiLineFrame::set_depth(int val)
{
   set_engine_property (static_cast<void*>(&val),
      SET_LINE_DEPTH);
}

void ChipsGuiLineFrame::set_extent(int val)
{
   set_engine_property (static_cast<void*>(&val),
      SET_LINE_EXTEND);
}

void ChipsGuiLineFrame::set_visibility(bool val)
{
   int ival = val;
   set_engine_property(static_cast<void*>(&ival), SET_LINE_VISIBLE);
}


void ChipsGuiLineFrame::color_callback(ChipsGuiLinePropWidget *widget, void *data)
{
   ChipsGuiLineFrame *linFrame = static_cast <ChipsGuiLineFrame*>(data);

   if (!linFrame)  return; //ensure we have needed pointer 
     
   string clr = widget->get_color ();

   linFrame->set_color(clr);
}

void ChipsGuiLineFrame::position_callback(ChipsGuiLinePropWidget *widget, void *data)
{
   ChipsGuiLineFrame *pntFrame = static_cast <ChipsGuiLineFrame*>(data);
   double x,y;

   if (!pntFrame)  return; //ensure we have needed pointer
   widget->get_position (x,y);
   pntFrame->set_position(x,y);
}

void ChipsGuiLineFrame::coordinate_callback(ChipsGuiLinePropWidget *widget, void *data)
{
   ChipsGuiLineFrame *pntFrame = static_cast <ChipsGuiLineFrame*>(data);

   if (!pntFrame)  return; //ensure we have needed pointer
   int coordinate = widget->get_coordinate ();
   pntFrame->set_coordinate (coordinate);
}


void ChipsGuiLineFrame::style_callback(ChipsGuiLinePropWidget *widget, gpointer data)
{
   ChipsGuiLineFrame *linFrame = static_cast <ChipsGuiLineFrame*>(data);
   int val = widget->get_style ();
   if (linFrame)  linFrame->set_style(val); 
}

void ChipsGuiLineFrame::thick_callback(ChipsGuiLinePropWidget *widget, gpointer data)
{
   ChipsGuiLineFrame *linFrame = static_cast <ChipsGuiLineFrame*>(data);
   float fval = widget->get_thick ();
   if (linFrame)  linFrame->set_thick(fval);
}

void ChipsGuiLineFrame::depth_callback(ChipsGuiLinePropWidget *widget, gpointer data)
{
   ChipsGuiLineFrame *linFrame = static_cast <ChipsGuiLineFrame*>(data);
   int val = widget->get_depth ();
   if (linFrame)  linFrame->set_depth(val); 
}

void ChipsGuiLineFrame::extent_callback(ChipsGuiLinePropWidget *widget, gpointer data)
{
   ChipsGuiLineFrame *linFrame = static_cast <ChipsGuiLineFrame*>(data);
   int val = widget->get_extent ();
   if (linFrame)  linFrame->set_extent(val); 
}

void ChipsGuiLineFrame::vis_callback(ChipsGuiLinePropWidget *widget, gpointer data)
{
   ChipsGuiLineFrame *myFrame = static_cast <ChipsGuiLineFrame*>(data);
   bool val = widget->get_visibility ();
   if (myFrame)  myFrame->set_visibility(val);
}

void ChipsGuiLineFrame::initialize(ChipsId* id)
{
   COORD_SYS coordsys;
   string coordid;
   ChipsVector pos(3);
   string colorval;
   float thickval;
   int depthval;
   int extentval;
   bool visval;
   V_LINE_TYPE styleval;

   if (!id)  return; 

   if (chipsGui && !_eng) 
      _eng = chipsGui->get_engine(); 

   if (id != &_cid)
   {
      _cid.reset_window_id(id->get_window_id()); 
      _cid.set_frame_id(id->get_frame_id()); 
      _cid.set_plot_id(id->get_plot_id()); 
      _cid.set_line_id(id->get_line_id()); 
   }

   _eng->get_property(&_cid, (void*)&colorval, SET_LINE_COLOR);
   _eng->get_property(&_cid, (void*)&extentval, SET_LINE_EXTEND);
   _eng->get_property(&_cid, (void*)&thickval, SET_LINE_THICKNESS);
   _eng->get_property(&_cid, static_cast<void*>(&styleval), 
      SET_LINE_STYLE);
   _eng->get_property(&_cid, (void*)&depthval, SET_LINE_DEPTH);
   _eng->get_property(&_cid, (void*)&visval, SET_LABEL_VISIBLE);
   _eng->get_object_coordinfo (&_cid, chips_line, coordsys, coordid);
   _eng->get_annotation_position (&_cid, chips_line, pos);

   _lineprop->set_color (colorval);
   _lineprop->set_coordinate ((int)coordsys);
   _lineprop->set_position (pos[0],pos[1]);
   _lineprop->set_extent (extentval);
   _lineprop->set_thick (thickval);
   _lineprop->set_style (styleval);
   _lineprop->set_depth (depthval);
   _lineprop->set_visibility (visval);

   // get tick formats
   int xmode = 0, ymode = 0;
   string xformat, yformat;
   ChipsId tmpid (_cid);
   tmpid.set_axis_id(_cid.get_xaxis_id());
   tmpid.set_axis_dim (X_AXIS);
   _eng->get_property(&tmpid, (void*)&xformat, SET_AXIS_TICKFORMAT);
   tmpid.set_axis_id(_cid.get_yaxis_id());
   tmpid.set_axis_dim (Y_AXIS);
   _eng->get_property(&tmpid, (void*)&yformat, SET_AXIS_TICKFORMAT);

   transform(xformat.begin(), xformat.end(), xformat.begin(), ::tolower);
   transform(yformat.begin(), yformat.end(), yformat.begin(), ::tolower);

   if (xformat.find("ra") != string::npos ||
       xformat.find("dec") != string::npos)
      xmode = 1;
   if (yformat.find("ra") != string::npos ||
       yformat.find("dec") != string::npos)
      ymode = 1;

   _lineprop->set_position_format (xmode, ymode);

   // enable/disable the data coord pulldown entry based on existing coords
   ChipsId mcid;
   mcid.set_coord_sys(DATA);
   if (_eng->check_coordinate_exists(&mcid))  
      _lineprop->enable_data_coordinate();
   else 
      _lineprop->disable_data_coordinate();

}

