// 
//  Copyright (C) 2011-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiPointFrame.hh"
#include "ChipsGuiPointPropWidget.hh"
#include "ChipsGuiMain.hh"
#include "ChipsEngine.hh"
#include "ChipsCoordinate.hh"
#include "ChipsCmdManager.hh"

#include <algorithm>

using namespace std;

extern ChipsGuiMain* chipsGui;

class ChipsPointMoveCallback : public ChipsGuiPropFrameCallback
{
   public:
      virtual void execute (ChipsEngine *eng, ChipsId &cid, void *data)
      {
         double *pos = static_cast<double *>(data);
         if ((pos != NULL) && eng)
            eng->move(&cid, chips_point, pos[0], pos[1], V_MOVE_ABSOLUTE);
      }
};


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiPointFrame::ChipsGuiPointFrame():ChipsGuiPropFrame(),ChipsGuiTooltipObject()
{
   _pointprop = NULL;
   _object_states = STATE_CHNG_PROP_PNT;
   _object_states |= STATE_CHNG_AXS_FMT;
}

// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiPointFrame::~ChipsGuiPointFrame()
{
   if (_pointprop)
      delete _pointprop;
}

void ChipsGuiPointFrame::build_form()
{
   _frame = gtk_frame_new("Point Properties");

   _pointprop = new ChipsGuiPointPropWidget ();
   _pointprop->build_form ();

   gtk_container_add(GTK_CONTAINER(_frame), _pointprop->base_widget ());

   _pointprop->add_callback ("change_color", new ChipsGuiPointPropWidgetCallback (color_callback, this));
   _pointprop->add_callback ("change_position", new ChipsGuiPointPropWidgetCallback (position_callback, this));
   _pointprop->add_callback ("change_angle", new ChipsGuiPointPropWidgetCallback (angle_callback, this));
   _pointprop->add_callback ("change_coordinate", new ChipsGuiPointPropWidgetCallback (coordinate_callback, this));
   _pointprop->add_callback ("change_style", new ChipsGuiPointPropWidgetCallback (style_callback, this));
   _pointprop->add_callback ("change_depth", new ChipsGuiPointPropWidgetCallback (depth_callback, this));
   _pointprop->add_callback ("change_size", new ChipsGuiPointPropWidgetCallback (size_callback, this));
   _pointprop->add_callback ("change_fill", new ChipsGuiPointPropWidgetCallback (fill_callback, this));
   _pointprop->add_callback ("change_visibility", new ChipsGuiPointPropWidgetCallback (vis_callback, this));



   
}

void ChipsGuiPointFrame::set_color(string color)
{
   set_engine_property((void*)color.c_str(), SET_POINT_SYMBOL_COLOR); 
}

void ChipsGuiPointFrame::set_position(double xpos, double ypos)
{
   ChipsPointMoveCallback move_callback;
   double pos[2] = {xpos, ypos};
   this->update_property (&move_callback, &pos[0]);
}

void ChipsGuiPointFrame::set_coordinate(int coordinate)
{
   ChipsId cid;
   COORD_SYS coordsys = (COORD_SYS)coordinate;
   string coordid = "";

   _eng->get_current_coord (&cid);

   if ((coordsys == DATA) && (cid.get_coord_sys () != DATA))
      chipsGui->log_error ("Data coordinate system does not exists.");
   else
      coordid = cid.get_coord_id ();

   coordid = ChipsCoordinate::extract_id (coordsys, coordid);

   _eng->change_annotation_coordinate (&cid, chips_point, coordsys, coordid);
}

void ChipsGuiPointFrame::set_size(int val)
{
   set_engine_property(static_cast<void*>(&val), 
      SET_POINT_SYMBOL_SIZE); 
}

void ChipsGuiPointFrame::set_fill(bool fill)
{
   set_engine_property(static_cast<void*>(&fill),
      SET_POINT_SYMBOL_FILL);
}

void ChipsGuiPointFrame::set_style(int val)
{
   set_engine_property(static_cast<void*>(&val),
      SET_POINT_SYMBOL_STYLE);
}

void ChipsGuiPointFrame::set_depth(int val)
{
   set_engine_property(static_cast<void*>(&val),
      SET_POINT_DEPTH);
}

void ChipsGuiPointFrame::set_angle(float val)
{
   set_engine_property(static_cast<void*>(&val),
      SET_POINT_SYMBOL_ANGLE);
}

void ChipsGuiPointFrame::set_visibility(bool val)
{
   int ival = val;
   set_engine_property(static_cast<void*>(&ival), SET_POINT_VISIBLE);
}

void ChipsGuiPointFrame::color_callback(ChipsGuiPointPropWidget *widget, void *data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);

   if (!pntFrame)  return; //ensure we have needed pointer 
   string clr = widget->get_color (); 
   pntFrame->set_color(clr);
}

void ChipsGuiPointFrame::position_callback(ChipsGuiPointPropWidget *widget, void *data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);
   double x,y;

   if (!pntFrame)  return; //ensure we have needed pointer
   widget->get_position (x,y);
   pntFrame->set_position(x,y);
}

void ChipsGuiPointFrame::coordinate_callback(ChipsGuiPointPropWidget *widget, void *data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);

   if (!pntFrame)  return; //ensure we have needed pointer
   int coordinate = widget->get_coordinate ();
   pntFrame->set_coordinate (coordinate);
}

void ChipsGuiPointFrame::angle_callback(ChipsGuiPointPropWidget *widget, void *data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);

   if (!pntFrame)  return; //ensure we have needed pointer
   float angle = widget->get_angle ();
   pntFrame->set_angle (angle);
}



void ChipsGuiPointFrame::style_callback(ChipsGuiPointPropWidget *widget, gpointer data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);
   int val = widget->get_style ();
   if (pntFrame)  pntFrame->set_style(val); 
}

void ChipsGuiPointFrame::size_callback(ChipsGuiPointPropWidget *widget, gpointer data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);
   int vv = widget->get_size ();
   if (pntFrame)  pntFrame->set_size(vv);
}

void ChipsGuiPointFrame::depth_callback(ChipsGuiPointPropWidget *widget, gpointer data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);
   int val = widget->get_depth ();
   if (pntFrame)  pntFrame->set_depth(val); 
}

void ChipsGuiPointFrame::fill_callback(ChipsGuiPointPropWidget *widget, gpointer data)
{
   ChipsGuiPointFrame *pntFrame = static_cast <ChipsGuiPointFrame*>(data);
   bool fill = widget->get_fill ();
   if (pntFrame)  pntFrame->set_fill(fill);
}

void ChipsGuiPointFrame::vis_callback(ChipsGuiPointPropWidget *widget, gpointer data)
{
   ChipsGuiPointFrame *myFrame = static_cast <ChipsGuiPointFrame*>(data);
   bool val = widget->get_visibility ();
   if (myFrame)  myFrame->set_visibility(val);
}

void ChipsGuiPointFrame::initialize(ChipsId* id)
{
   COORD_SYS coordsys;
   string coordid;
   ChipsVector pos(3);
   string colorval;
   int sizeval;
   int depthval;
   bool fillval;
   bool visval;
   float angleval;
   V_SYMBOL_TYPE styleval;

   if (chipsGui && !_eng) 
      _eng = chipsGui->get_engine(); 

   if (id != &_cid)
   {
      _cid.reset_window_id(id->get_window_id()); 
      _cid.set_frame_id(id->get_frame_id()); 
      _cid.set_plot_id(id->get_plot_id()); 
      _cid.set_point_id(id->get_point_id()); 
   }


   _eng->get_property(&_cid, (void*)&colorval, SET_POINT_SYMBOL_COLOR);
   _eng->get_property(&_cid, (void*)&fillval, SET_POINT_SYMBOL_FILL);
   _eng->get_property(&_cid, (void*)&angleval, SET_POINT_SYMBOL_ANGLE);
   _eng->get_property(&_cid, (void*)&sizeval, SET_POINT_SYMBOL_SIZE);
   _eng->get_property(&_cid, static_cast<void*>(&styleval), 
         SET_POINT_SYMBOL_STYLE);
   _eng->get_property(&_cid, (void*)&depthval, SET_POINT_DEPTH);
   _eng->get_object_coordinfo (&_cid, chips_point, coordsys, coordid);
   _eng->get_annotation_position (&_cid, chips_point, pos);


   _pointprop->set_color (colorval);
   _pointprop->set_coordinate ((int)coordsys);
   _pointprop->set_angle (angleval);
   _pointprop->set_position (pos[0],pos[1]);
   _pointprop->set_fill (fillval);
   _pointprop->set_size (sizeval);
   _pointprop->set_style (styleval);
   _pointprop->set_depth (depthval);
   _pointprop->set_visibility (visval);

   // get tick formats
   int xmode = 0, ymode = 0;
   string xformat, yformat;
   ChipsId tmpid (_cid);
   tmpid.set_axis_id(_cid.get_xaxis_id());
   tmpid.set_axis_dim (X_AXIS);
   _eng->get_property(&tmpid, (void*)&xformat, SET_AXIS_TICKFORMAT);
   tmpid.set_axis_id(_cid.get_yaxis_id());
   tmpid.set_axis_dim (Y_AXIS);
   _eng->get_property(&tmpid, (void*)&yformat, SET_AXIS_TICKFORMAT);

   transform(xformat.begin(), xformat.end(), xformat.begin(), ::tolower);
   transform(yformat.begin(), yformat.end(), yformat.begin(), ::tolower);

   if (xformat.find("ra") != string::npos ||
       xformat.find("dec") != string::npos)
      xmode = 1;
   if (yformat.find("ra") != string::npos ||
       yformat.find("dec") != string::npos)
      ymode = 1;

   _pointprop->set_position_format (xmode, ymode);

   // enable/disable the data coord pulldown entry based on existing coords
   ChipsId mcid;
   mcid.set_coord_sys(DATA);
   if (_eng->check_coordinate_exists(&mcid))  
      _pointprop->enable_data_coordinate();
   else 
      _pointprop->disable_data_coordinate();
}


