// 
//  Copyright (C) 2011  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiSettings.hh"
#include "ChipsDefaults.hh"
#include "ChipsDefines.hh"

#include "vtkBoundingBoxRepresentation.hh"
#include "vtkAddAnnotationRepresentation.hh"
#include "vtkChipsRepresentation.hh"
#include "vtkProperty2D.h"
#include "ChipsColor.hh"
#include "ChipsEngine.hh"
#include "ChipsGuiMain.hh"
#include "ChipsGuiTooltipObject.hh"

extern ChipsDefaults* theDefaults;
extern ChipsGuiMain* chipsGui;

using namespace std;

ChipsGuiSettings::ChipsGuiSettings()
{
   initialize();   // set our default values  
 
   load();         // load in- user saved preference values  
}

ChipsGuiSettings::~ChipsGuiSettings()
{
}

void ChipsGuiSettings::initialize()
{
   tbar_style = 2;  // both icon & label 
   pane_tbar = pane_nav = pane_info = pane_prop = pane_shell = true; // show all
   set_tooltips(true);   // show tooltips 
   zoombutton = true;  // show zoom button in panel 
   wingrids = false;     // display window grids
#ifdef CHIPS_GUI_ADDDEL
   delbutton = true;   // show delete button in panel 
#endif
   defbutton = "select";  // default mode- overridden by chipsrc value 
   selcolor = "cyan";  // default selection bounding box anchor color
   rangecolor = "magenta"; 

   this->set_selcolor (selcolor);
   this->set_rangecolor (rangecolor);

}

// loads values from chips preferences file 
int ChipsGuiSettings::load()
{
   if (!theDefaults)   return 1;   // error- no defaults file
   
   pane_tbar = theDefaults->get_toolbarvis();
   pane_nav = theDefaults->get_navpanevis(); 
   pane_info = theDefaults->get_infpanevis();
   pane_prop = theDefaults->get_proppanevis();
   pane_shell = theDefaults->get_shellpanevis(); 
   zoombutton = theDefaults->get_tbarzoombtn();
#ifdef CHIPS_GUI_ADDDEL
   delbutton  = theDefaults->get_tbardelbtn();
#endif
   defbutton  = theDefaults->get_tbarmode();
   wingrids = theDefaults->get_showgrid();
   set_tooltips(theDefaults->get_guitips());
   set_tbar_style(theDefaults->get_tbarstyle()); 

    
   return 0; 
}

// write values out to the chips preferences file 
int ChipsGuiSettings::save()
{
   if (!theDefaults)   return 1;   // error- no defaults file

   // update all pref values 
   theDefaults->set_preference("gui.view.toolbar", (pane_tbar ? "true":"false")); 
   theDefaults->set_preference("gui.view.propertyframe", (pane_prop ? "true":"false")); 
   theDefaults->set_preference("gui.view.navframe", (pane_nav ? "true":"false")); 
   theDefaults->set_preference("gui.view.infoframe", (pane_info ? "true":"false")); 
   theDefaults->set_preference("gui.view.shellframe", (pane_shell ? "true":"false")); 

   theDefaults->set_preference("gui.tooltips", (tooltips ? "true":"false"));
   theDefaults->set_preference("gui.toolbar.zoom", (zoombutton ? "true":"false"));
#ifdef CHIPS_GUI_ADDDEL
   theDefaults->set_preference("gui.toolbar.delete", (delbutton ? "true":"false"));
#endif
   theDefaults->set_preference("gui.toolbar.style", tbar_style_as_string());
   theDefaults->set_preference("gui.toolbar.mode", defbutton);

   // save changes 
   theDefaults->save_file(); 

   return 0; 
}

int ChipsGuiSettings::set_tbar_style(int val)
{
   int rv = 0; 
   switch(val) 
   {
      case 0:   // icon only
      case 1:   // label only 
      case 2:   // both   
         tbar_style = val; 
      break;
      default:
         rv = 1;  // invalid value 
      break; 
   }
   return rv; 
}

int ChipsGuiSettings::set_tbar_style(string val)
{
   if (val == "icons")            tbar_style = 0;
      else if (val == "labels")   tbar_style = 1;
         else if (val == "both")  tbar_style = 2;
            else return 1; // invalid value
   return 0; 
}


int ChipsGuiSettings::set_defbutton(const string &val)
{
   if (chips_strcmp(val.c_str(), "pan") == 0) 
      defbutton = "pan"; 
   else if (chips_strcmp(val.c_str(), "select") == 0)
      defbutton = "select";
   else if (chips_strcmp(val.c_str(), "range") == 0)
      defbutton = "range";
   else if (chips_strcmp(val.c_str(), "zoom") == 0)
      defbutton = "zoom";
#ifdef CHIPS_GUI_ADDDEL
   else if (chips_strcmp(val.c_str(), "add") == 0)
      defbutton = "add";
   else if (chips_strcmp(val.c_str(), "delete") == 0)
      defbutton = "delete";
#endif
   else
      return 1;   // invalid value   
   return 0;      // set def to one of the mode buttons
}

string ChipsGuiSettings::tbar_style_as_string()
{
   static string sval; 
  
   if (tbar_style == 0)          sval = "icons";
      else if (tbar_style == 1)  sval = "labels";
         else                    sval = "both";
   return sval;  
}

int ChipsGuiSettings::set_selcolor(const string &val)
{
   if (valid_color(val))
   {
      selcolor = val; 
      vtkProperty2D *prop = vtkChipsRepresentation::GetGlobalBorderProperty ();
      if (prop)
      {
         ChipsColor chips_color (val);
         prop->SetColor (chips_color.red (), chips_color.green (), chips_color.blue ());
         if (chipsGui && chipsGui->get_engine())
            chipsGui->get_engine()->redraw_window ("all");

      }

      return 0; 
   } 
   return 1; 
}

int ChipsGuiSettings::set_rangecolor(const string &val)
{
   if (valid_color(val))
   {
      rangecolor = val;
      vtkProperty2D *prop = vtkBoundingBoxRepresentation::GetGlobalBorderProperty ();
      if (prop)
      {
         ChipsColor chips_color (val);
         prop->SetColor (chips_color.red (), chips_color.green (), chips_color.blue ());
      }
      prop = vtkAddAnnotationRepresentation::GetGlobalAnnotationProperty ();
      if (prop)
      {
         ChipsColor chips_color (val);
         prop->SetColor (chips_color.red (), chips_color.green (), chips_color.blue ());
      }

      return 0;
   }
   return 1;
}

void ChipsGuiSettings::set_wingrids(bool val)
{
   theDefaults->set_preference("window.grid.visible", (val ? "true":"false"));
}

void ChipsGuiSettings::set_tooltips(bool val)
{ 
   tooltips = val; 
   ChipsGuiTooltipObject::show_tooltips(tooltips);
}



