// 
//  Copyright (C) 2011  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsGuiShellPane.hh"
#include "ChipsGuiErrLog.hh"
#include <gdk/gdkkeysyms.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include <vte.h>
#include <fstream>
#include <iostream>

#define FGINDEX  1
#define BGINDEX  3 


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiShellPane::ChipsGuiShellPane(string serverid)
{
   //_frame = gtk_frame_new("Shell");
   _serverid = serverid; 
   _initfile = ""; 
   _frame = gtk_frame_new(NULL);
   _table = gtk_table_new (2,2, FALSE);
   gtk_container_add (GTK_CONTAINER(_frame), _table);

   _nbook = gtk_notebook_new();
   gtk_notebook_set_tab_pos (GTK_NOTEBOOK(_nbook), GTK_POS_TOP);
   gtk_notebook_set_scrollable(GTK_NOTEBOOK(_nbook), GTK_POLICY_AUTOMATIC);

   setup_chips_shell();
    
   // The notebook doesn't handle events when the user reselects the current
   // tab- use an event box to capture a click on the current tab 
   _errlbl = gtk_label_new("Errors");
   _errbox = gtk_event_box_new (); 
   gtk_container_add (GTK_CONTAINER(_errbox), _errlbl);
   gtk_widget_show(_errlbl); 
    
   _errlog = new ChipsGuiErrLog(); 
   gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), _errlog->base_widget(), _errbox);

   gtk_table_attach(GTK_TABLE(_table),_nbook,0,2,0,2,
        (GtkAttachOptions) (GTK_EXPAND |GTK_FILL),(GtkAttachOptions) (GTK_EXPAND|GTK_FILL),0,0);

   // handle notebook error tab click 
   gtk_widget_set_events(_errbox, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(G_OBJECT(_errbox),"button-press-event",
      G_CALLBACK(event_callback), (void*)this);

   // handle notebook page changes  
   g_signal_connect(G_OBJECT(_nbook),"switch-page",
      G_CALLBACK(switchpage_callback), (void*)this);

   // handle key presses in shell widget 
   gtk_widget_set_events(_shell, GDK_KEY_PRESS_MASK);
   g_signal_connect(G_OBJECT(_shell),"key-press-event",
      G_CALLBACK(event_callback), (void*)this);
}


// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiShellPane::~ChipsGuiShellPane()
{
  gtk_widget_destroy(_shell);
  gtk_widget_destroy(_nbook);
  gtk_widget_destroy(_frame);
  delete _errlog; 
  _errlog = NULL; 
  if (_initfile.length() && !access(_initfile.c_str(), F_OK))
     remove(_initfile.c_str());
}

void ChipsGuiShellPane::setup_chips_shell()
{
   _shell = vte_terminal_new ();
   char initbuf[256]; 
   string dname, fname;
   char *params[] = {"chips", initbuf, NULL}; 
   GPid child; 

   // handle connecting vte shell to the server and overriding quit/disconnect 
   int pd = (int)getpid();
   char* lgname = getlogin();
   if (lgname)  sprintf(initbuf,"/tmp/.chips.%s.%1d.gui_init", lgname, pd);
      else      sprintf(initbuf,"/tmp/.chips.%1d.gui_init", pd);
   ofstream out(initbuf);
   out << "try:"<<endl<<"   import os"<<endl<<"   connect(\"" << _serverid << "\",\"guicmdline\")" << endl;
   out << "   wd = get_server_cwd()"<<endl<< "   if len(wd):   os.chdir(wd)"<< endl;
   out << "   def quit():"<<endl<<"     print 'Please close the Chips Gui to exit the shell'"<<endl<<endl;
   out << "   def exit():"<<endl<<"     print 'Please close the Chips Gui to exit the shell'"<<endl<<endl;
   out << "   def disconnect():"<<endl<<"     print 'Please close the Chips Gui to disconnect from the server'"<<endl<<endl;
   out << "finally:"<<endl;
   out << "   os.remove(\"" << initbuf << "\")" << endl;
   out.close();
   _initfile = initbuf; 

   child = vte_terminal_fork_command (VTE_TERMINAL(_shell), "chips", params, NULL, "~/", FALSE, FALSE, FALSE);

   gtk_widget_set_size_request(GTK_WIDGET(_shell), 10, 10);
   vte_terminal_set_size(VTE_TERMINAL(_shell), 80, 10);

   GtkWidget *nblabel = gtk_label_new("Shell");
   gtk_notebook_append_page(GTK_NOTEBOOK(_nbook), _shell, nblabel);
}

void ChipsGuiShellPane::show()
{
   gtk_widget_show(GTK_WIDGET(_frame));

   // set terminal window color- appears that it needs to be done after show 
   GtkStyle *style = gtk_widget_get_style(GTK_WIDGET(_errlog->base_widget()));
   vte_terminal_set_background_image(VTE_TERMINAL(_shell),NULL);
   vte_terminal_set_color_background(VTE_TERMINAL(_shell), &style->base[0]);
   vte_terminal_set_color_foreground(VTE_TERMINAL(_shell), &style->text[0]);
   vte_terminal_set_color_highlight(VTE_TERMINAL(_shell), &style->base[1]);
}

void ChipsGuiShellPane::hide()
{
   gtk_widget_hide(GTK_WIDGET(_frame));
}

int ChipsGuiShellPane::get_height()
{
   int ht = 0;

   if (_frame)   ht = _frame->allocation.height;
   return ht;
}

void ChipsGuiShellPane::connect(string serv)
{
   string cmd = "connect('";
   // append specified serverid if provided 
   if (serv.length()) cmd.append(serv);
      else          cmd.append(_serverid);
   cmd.append("')\n");
   vte_terminal_feed_child(VTE_TERMINAL(_shell), cmd.c_str(), cmd.length());
}

void ChipsGuiShellPane::disconnect()
{
   string cmd = "disconnect()\n";
   vte_terminal_feed_child(VTE_TERMINAL(_shell), cmd.c_str(), cmd.length());
}

void ChipsGuiShellPane::log_error(string err)
{
   if (_errlog)
   { 
      _errlog->add_message(err);
      // when an error gets logged, the error tab label is displayed in red
      gtk_label_set_markup(GTK_LABEL(_errlbl), "<span foreground=\"red\">Errors</span>");
   }
}

void ChipsGuiShellPane::reset_error_label()
{
   // clear out the Error label color indication 
   gtk_label_set_markup(GTK_LABEL(_errlbl), "Errors");
}

gboolean ChipsGuiShellPane::event_callback(GtkWidget* widget,
    GdkEvent  *event, gpointer data)
{
   ChipsGuiShellPane *shellPane = static_cast <ChipsGuiShellPane*>(data);

   if (event && shellPane)
   {
      if (event->type == GDK_KEY_PRESS)
      {
         GdkEventKey *gek = (GdkEventKey*)event;
         if ((gek->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK && 
             (gek->keyval == GDK_d || gek->keyval == GDK_D))
         {
            shellPane->log_error("<Control><d> in the Gui Shell Frame ignored- Please close the Chips Gui to exit the shell\n"); 
            return 1; // ignore <ctrl><d> or <ctrl><D>   don't pass on to shell 
         }  
      }
      else if (event->type == GDK_BUTTON_PRESS)
      {
         shellPane->reset_error_label();
      } 
   }
   return 0; // pass on to shell 
}


void ChipsGuiShellPane::switchpage_callback(GtkNotebook *notebook,
   GtkWidget   *page,  guint page_num, gpointer data)
{
   ChipsGuiShellPane *shellPane = static_cast <ChipsGuiShellPane*>(data);

   if (page_num == 1 && shellPane)
      shellPane->reset_error_label();
}




