// 
//  Copyright (C) 2012-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//



#include "ChipsGuiSpinWidget.hh"
#include "ChipsDefinesP.hh"

#include <sstream>
#include <cstdlib>

#include <ctype.h>
#include <math.h>
#include <gdk/gdkkeysyms.h>

#define cgswSEPCHARS "dmsh: "

using namespace std;

ChipsGuiSpinWidgetCallback::ChipsGuiSpinWidgetCallback (void (*callback)(ChipsGuiSpinWidget *, void *), void *data)
{
   _callback = callback;
   _callbackData = data;
}

void ChipsGuiSpinWidgetCallback::execute (ChipsGuiWidget *caller, void *)
{
   ChipsGuiSpinWidget *coordWidget = dynamic_cast<ChipsGuiSpinWidget *>(caller);
   _callback (coordWidget, _callbackData);
}


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
ChipsGuiSpinWidget::ChipsGuiSpinWidget(int type):ChipsGuiWidget(),ChipsGuiTooltipObject()
{
   _dval = 0;
   _sval = "0:00:00"; 
   _type = type;
   _precision = 5; 
   _stepsize = 1; 
   _scrollcnt = 0;
   _scrolltimestamp = 0; 
   _enable_popup = true;
   build_form();

   // setup the optional callbacks 
   _valid_callback_ids.push_back ("change_value");
   _valid_callback_ids.push_back ("change_step");
}


// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
ChipsGuiSpinWidget::~ChipsGuiSpinWidget()
{
}

void ChipsGuiSpinWidget::build_form()
{

   _adj = gtk_adjustment_new (0, -G_MAXDOUBLE, G_MAXDOUBLE, 1, 10, 0);
   _base_w = gtk_entry_new();

   _mode = 0;  
 
   _btn_press_id = g_signal_connect(G_OBJECT(GTK_ENTRY(_base_w)), 
                    "button-press-event", G_CALLBACK(entry_button_press_cb),
                    (void*)this);

   _key_press_id = g_signal_connect(G_OBJECT(GTK_ENTRY(_base_w)), 
                    "key-press-event", G_CALLBACK(entry_key_press_cb), 
                    (void*)this);

   _key_release_id = g_signal_connect(G_OBJECT(GTK_ENTRY(_base_w)), 
                    "key-release-event", G_CALLBACK(entry_key_release_cb), 
                    (void*)this);

   _focus_out_id = g_signal_connect (G_OBJECT(GTK_ENTRY(_base_w)),
                    "focus-out-event", G_CALLBACK(focus_out_cb),
                    (void*)this);

   _scroll_id = g_signal_connect(G_OBJECT(GTK_ENTRY(_base_w)), "scroll-event",
                    G_CALLBACK(entry_scroll_cb), (void*)this);

}

void ChipsGuiSpinWidget::set_value(double val)
{
   _dval = val; 
}

void ChipsGuiSpinWidget::set_value(string val)
{
   _sval = val;
}

string ChipsGuiSpinWidget::get_string_value()
{
   return gtk_entry_get_text (GTK_ENTRY(_base_w));
}

void ChipsGuiSpinWidget::show ()
{
   gtk_widget_show (_base_w);
}

void ChipsGuiSpinWidget::hide ()
{
   gtk_widget_hide (_base_w);
}

bool ChipsGuiSpinWidget::entry_button_press_cb(GtkWidget *wid,
   GdkEventButton *event, gpointer data)
{
   if (event->type == GDK_BUTTON_PRESS &&  event->button == 2)
   {
      ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
      me->popup_menu();
      return true; // press handled via popup- don't forward event
   }
   return false; // pass along button press  
}

bool ChipsGuiSpinWidget::entry_key_release_cb(GtkWidget *wid,
   GdkEvent *event, gpointer data)
{
   GdkEventKey *evt = (GdkEventKey*)event;
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);

   // keep the scroll step size constant while the up/down arrow key
   // is pressed, but on a key release, recalculate the step size
   if (me && (evt->keyval == GDK_Up || evt->keyval == GDK_Down))
   {
      me->reset_scrollcnt();
   }
   return false; 
}


bool ChipsGuiSpinWidget::entry_key_press_cb(GtkWidget *wid,
   GdkEvent *event, gpointer data)
{
   GdkEventKey *evt = (GdkEventKey*)event;
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);

   if (!me)  return false; 
   if (evt->keyval == GDK_Up)
   {
      me->process_scroll(0); 
      return true; // handle ourselves
   }
   else if (evt->keyval == GDK_Down)
   {
      me->process_scroll(1); 
      return true; // handle ourselves
   }
   else if (evt->keyval == GDK_Return || evt->keyval == GDK_Tab)
   {
      // check for mode change
      int sts = me->validate_text(); 
      if (sts != -1)
         me->set_mode(sts); 
      else 
         return true;   // invalid text- don't process

      // treat text as entered and run change_value cb
      me->run_change_value_cb();
      return true; 
   }
   else if ((evt->keyval <= 0x7f) && ispunct(evt->keyval) &&
            ((evt->keyval != 'e') && (evt->keyval != 'E') &&
            (evt->keyval != '.') && (evt->keyval != ':') &&
            (evt->keyval != ',') && (evt->keyval != '+') && (evt->keyval != '-'))) 
   {
      return true; // ignore most letters and punctuation 
   }
   return false; // pass along key press to default handler

}

void ChipsGuiSpinWidget::run_change_value_cb()
{
   if (_callbacks.count ("change_value") > 0)
         _callbacks["change_value"]->execute(this, NULL);
}

bool ChipsGuiSpinWidget::entry_scroll_cb(GtkWidget *wid, GdkEvent *event, 
   gpointer data)
{
   GdkEventScroll *evt = (GdkEventScroll*)event;
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
   if (!me) return true;
  
   // check the timestamp and reset the scroll step size if necessary
   me->reset_scrollcnt(evt->time); 

   if (evt->direction == GDK_SCROLL_UP)
      me->process_scroll(0); 
   else if (evt->direction == GDK_SCROLL_DOWN)
      me->process_scroll(1);
   return true;
}

void ChipsGuiSpinWidget::popup_menu()
{
   GtkWidget *menu, *mitem1, *mitem2;

   menu = gtk_menu_new();
   mitem1 =  gtk_radio_menu_item_new_with_label(NULL, "Numeric");
   if (_type == 0) 
      mitem2 =  gtk_radio_menu_item_new_with_label_from_widget(GTK_RADIO_MENU_ITEM(mitem1), "RA (HH:MM:SS)");
   else 
      mitem2 =  gtk_radio_menu_item_new_with_label_from_widget(GTK_RADIO_MENU_ITEM(mitem1), "Dec (DD:MM:SS)");
   if (_mode == 0)
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem1),TRUE);
   else
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem2),TRUE);

   g_signal_connect(mitem1, "activate", (GCallback)set_numeric_cb,  this);
   g_signal_connect(mitem2, "activate", (GCallback)set_sexagesimal_cb,  this);

   gtk_menu_shell_append(GTK_MENU_SHELL(menu), mitem1);
   gtk_menu_shell_append(GTK_MENU_SHELL(menu), mitem2);
   gtk_widget_show_all(menu);

   gtk_widget_set_sensitive (mitem1, _enable_popup);
   gtk_widget_set_sensitive (mitem2, _enable_popup);

   gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 0,
                  gdk_event_get_time(NULL));
}

void ChipsGuiSpinWidget::set_numeric_cb(GtkWidget *widget, gpointer data)
{
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
   if (me)  me->set_mode(0);
}

void ChipsGuiSpinWidget::set_sexagesimal_cb(GtkWidget *widget, gpointer data)
{
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
   if (me)  me->set_mode(1);
}

int ChipsGuiSpinWidget::ra_to_deg(string val, double &dval)
{
   string part[3];
   double dd, mm, ss;
   int cnt = 0;
   char buf[80];
   char *where, *err = 0;
   strcpy(buf, val.c_str());
   where = strtok(buf, cgswSEPCHARS);
   while(where && cnt < 3)
   {
      if (strlen(where))   part[cnt++] = where;
      where = strtok(NULL, cgswSEPCHARS);
   }
   if (cnt == 1)  // single numeric value
   {
      dd = strtod(part[0].c_str(), &err);
      if (err[0] != part[0][part[0].length()])  return -1;  // invalid text
      if (val.find(':') != string::npos) 
      {
         // numeric value followed by colon- treat as ra hours
         if (dd >= 0)   dval = 15*(dd + ((ss/60.0)+mm)/60.0);
            else       dval = 15*(dd - ((ss/60.0)+mm)/60.0);

      } 
      else 
         dval = dd;
   }
   else  if (cnt == 3)
   {
      // hms format
      dd  = strtod(part[0].c_str(), &err);
      if (err[0] != part[0][part[0].length()])  return -2;  // invalid hours
      mm  = strtod(part[1].c_str(), &err);
      if (err[0] != part[1][part[1].length()])  return -3;  // invalid min
      ss  = strtod(part[2].c_str(), &err);
      if (err[0] != part[2][part[2].length()])  return -4;  // invalid sec
      if (dd >= 0)   dval = 15*(dd + ((ss/60.0)+mm)/60.0);
         else       dval = 15*(dd - ((ss/60.0)+mm)/60.0);
   }
   else
      return -1;   // invalid input string
   // wrap around 
   if (dval > 360)
   {
      dval -= 360.0; 
   } 
   else if (dval < 0)
      dval += 360.0; 
   return 0;
}

int ChipsGuiSpinWidget::dec_to_deg(string val, double &dval)
{
   string part[3];
   double dd, mm, ss;
   int cnt = 0;
   char buf[80];
   char *where, *err = 0;
   strcpy(buf, val.c_str());
   where = strtok(buf, cgswSEPCHARS);
   while(where && cnt < 3)
   {
      if (strlen(where))   part[cnt++] = where;
      where = strtok(NULL, cgswSEPCHARS);
   }
   if (cnt == 1)  // single numeric value
   {
      dd = strtod(part[0].c_str(), &err);
      if (err[0] != part[0][part[0].length()])  return -1;  // invalid text
      if (val.find(':') != string::npos)
      {
         if (dd > 0)   dval = dd + ((ss/60.0)+mm)/60.0;
            else       dval = dd - ((ss/60.0)+mm)/60.0;
      }
      else   dval = dd;
   }
   else  if (cnt == 3)
   {
      // dms format
      dd  = strtod(part[0].c_str(), &err);
      if (err[0] != part[0][part[0].length()])  return -2;  // invalid degrees
      mm  = strtod(part[1].c_str(), &err);
      if (err[0] != part[1][part[1].length()])  return -3;  // invalid min
      ss  = strtod(part[2].c_str(), &err);
      if (err[0] != part[2][part[2].length()])  return -4;  // invalid sec
      if (dd > 0)   dval = dd + ((ss/60.0)+mm)/60.0;
         else       dval = dd - ((ss/60.0)+mm)/60.0;
   }
   else
      return -1;   // invalid input string
   if (dval > 90 || dval < -90)
      return -6;   // invalid value
   return 0;
}

int ChipsGuiSpinWidget::deg_to_ra(double val, string &sval)
{
   string hmfmt =  "%02d:%02d:", sfmt="%04.1f";
   double tpos, hd, md, ss;
   char tmpval[80], tmpspec[80];
   char *sixty_p;
   if ((tpos = fmod(val, 360.0)) < 0) tpos+=360;
   ss = tpos/15.0;
   ss = modf(ss, &hd);
   hd = fmod(hd, 24);       // wrap at 24 hours

/**
   ss = modf(ss*60.0 + FLT_EPSILON, &md);
   ss = ss * 60 + 0.00137;
**/
   ss = modf(ss*60.0, &md);
   ss = ss * 60.0;


//WLM- try increased precision on sexagesimal seconds 
sfmt="%07.4f";

   // account for roundup to 60 seconds
   sprintf(tmpspec, sfmt.c_str(), ss);
   if ((sixty_p = strstr(tmpspec,"60")) != NULL)
   {
      *sixty_p = '0';
      md++;
   }
   // account for roundup to 60 minutes;
   sprintf(tmpval,hmfmt.c_str(), (int)hd, (int)md);
   if ((sixty_p = strstr(tmpval,"60")) != NULL)
   {
      // account for 23h to 00h
      if (hd == 23)  sprintf(tmpval, hmfmt.c_str(), 0, 0);
         else        sprintf(tmpval, hmfmt.c_str(), (int)hd+1, 0);
   }
   sval = tmpval; 
   sval.append(tmpspec);
   return 0; 
}

int ChipsGuiSpinWidget::deg_to_dec(double dval, string &sval)
{
   double hd, md, ss, val;
   unsigned int swapsign;
   char tmpfmt[80], tmpval[80];

   // WLM031612 - handle flip for > 90 
   int tm = dval / 360;
   val = dval - (tm * 360);
   if (val > 180)  val = 180 - val;
      else if (val < -180)  val = -180 -val;
   if (val > 90) val = 180 - val;  
      else if (val < -90) val = -180 - val;  
   strcpy(tmpfmt, "%+02d:%02d:%04.1f");
   ss = modf(val, &hd);
   if (ss >= 0)
      ss = modf(ss*60.0 + FLT_EPSILON, &md);
   else
      ss = modf(ss*60.0 - FLT_EPSILON, &md);
   ss *= 60;
   // handle degrees between -1:0 with a - prefix infront of degrees
   if ((swapsign = ((hd <= 0) && ((md < 0) || (ss < 0)))))
   {
      if (md < 0) md *= -1;
      if (ss < 0) ss *= -1;
   }
   // handle minutes field set to +/-60 due to precision issues
   if ((md == 60) || (md == -60))
   {
      md = 0;
      hd++;
   }
   sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss);

   if (strstr(tmpval,":60:") == NULL && strstr(tmpval,":60"))
   {
      // handle invalid seconds conversion- dd:mm:60 
      ss = 0;
      md+=1.0;  
      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss); 
   }
   if (strstr(tmpval,":60:") != NULL)
   {
      // handle invalid minutes conversion- dd:60:ss 
      if (hd >= 0)  hd+= 1.0;
         else       hd-= 1.0;
      md = 0; 

      sprintf(tmpval, tmpfmt, (int)hd, (int)md, ss); 
   }
   sval = tmpval;
   if (swapsign)  sval[0] = '-';
 
   return 0; 
}

void ChipsGuiSpinWidget::set_mode(int mode)
{
   if (mode != _mode)   
   {
      string sval;
      double dval; 

      if (_mode == 1)
      {
         // use sexagesimal 
         sval = (char*)(gtk_entry_get_text (GTK_ENTRY(_base_w)));
         string tmpval;
         if (_type == 0) 
         {
            if ( deg_to_ra(_dval, tmpval) != 0)
               dval = gtk_adjustment_get_value(GTK_ADJUSTMENT(_adj)); 
            if (tmpval == sval) 
               dval = _dval; 
            else 
            {
               char * pEnd;
               double td = strtod(sval.c_str(), &pEnd);
               if (pEnd && *pEnd == '\0')   dval = td;
            }
         }
         else 
         {
            if ( deg_to_dec(_dval, tmpval) != 0)
               dval = gtk_adjustment_get_value(GTK_ADJUSTMENT(_adj)); 
            if (tmpval == sval) 
               dval = _dval; 
            else
            {
               char * pEnd;
               double td = strtod(sval.c_str(), &pEnd);
               if (pEnd && *pEnd == '\0')   dval = td;
            }
         }
         gchar* text = g_strdup_printf("%f",dval);
//WLM         g_signal_handler_block(GTK_ENTRY(_base_w), _process_output_id);
         gtk_entry_set_text (GTK_ENTRY(_base_w), text);
         _sval = text;
         _dval = dval; 
//WLM         g_signal_handler_unblock(GTK_ENTRY(_base_w), _process_output_id);
         g_free (text);
      }
      else 
      {
         // use degrees 
         dval = get_numeric_value();
         // use sexagesimal
         if (_type == 0)
         {
            deg_to_ra(dval, sval);
         }
         else
         {
            deg_to_dec(dval, sval);
         }
         gchar* text = g_strdup_printf("%s",sval.c_str());
//WLM         g_signal_handler_block(GTK_ENTRY(_base_w), _process_output_id);
         gtk_entry_set_text (GTK_ENTRY(_base_w), text);
         _sval = text;
         _dval = dval; 
//WLM         g_signal_handler_unblock(GTK_ENTRY(_base_w), _process_output_id);
         g_free (text);
      }
   }
   _mode = mode; 
}


int ChipsGuiSpinWidget::process_output_cb(GtkEntry *wid,
   gpointer data)
{
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
   if (!me) return FALSE; 

   double dvalue;
   gchar *text;
   char* sval = (char*)(gtk_entry_get_text (GTK_ENTRY(wid)));

   if (sval && me->get_mode() == 1)
   {
      text = g_strdup_printf("%s",sval);
      if (me->get_type() == 0)   me->ra_to_deg(text, dvalue);
         else   me->dec_to_deg(text, dvalue); 
      me->set_value(dvalue); 
      gtk_adjustment_set_value(GTK_ADJUSTMENT(me->_adj) ,dvalue);
   }
   else
   {
      dvalue = me->get_numeric_value(); 
      text = g_strdup_printf ("%g", dvalue);
      me->set_value(dvalue);
   }
   me->set_value(text);
   gtk_entry_set_text (GTK_ENTRY(wid), text);
   g_free (text);

   return TRUE;  // block default output handling
}

double ChipsGuiSpinWidget::get_numeric_value()
{
   const char* txt = gtk_entry_get_text(GTK_ENTRY(_base_w));
   char * pEnd;
   double dval = _dval; 

   if (txt)   dval = strtod (txt, &pEnd);
   if (pEnd == &txt[strlen(txt)])  
   {
     _dval = dval;
   }
   else if (txt)
   {
      // not numeric- try converting from sexagesimal 
      string tmpval;
      if (_type == 0)
      {
         deg_to_ra(_dval, tmpval);
         if (tmpval != txt)  
            ra_to_deg(txt, dval);
         else 
            dval = _dval;
      }
      else
      {
         deg_to_dec(_dval, tmpval);
         if (tmpval != txt)
            dec_to_deg(txt, dval);
         else 
            dval = _dval;
      }
   }
   return dval; 
}

void ChipsGuiSpinWidget::process_scroll(int dir)
{
   if (_scrollcnt++ == 0)
   {
      // first scroll event- adjust spin step 
      reset_scrollsize(); 
   }  
   string val = gtk_entry_get_text (GTK_ENTRY(_base_w));
   if (_mode == 0)  // numeric 
   {
      // for now, increment by 1 degree
      if (dir == 0)
      {  
         _dval += _stepsize; 
      }
      else 
      {
         _dval -= _stepsize; 
      }
      int prec = val.length(); 
      size_t sz = val.find('.'); 
      if (sz != string::npos)   prec = prec - sz -1; 
      char fmt[30];
      if (prec > 8)  sprintf(fmt, "%%.%de", prec);
         else  sprintf(fmt, "%%.%df", prec);
      gchar *text = g_strdup_printf (fmt, _dval);

      if (val == text)
      {
         // string value is same- adjust precision by 1
         sprintf(fmt, "%%.%df", prec+1);
         g_free (text);
         text = g_strdup_printf (fmt, _dval);
      }
      if ((val + ".0" == text) || (val + "0" == text))
      {
         // string value is still same- adjust by step's precision
         char sfmt[30];
         char *spos;
         sprintf(sfmt,"%e",_stepsize);
         if ((spos = strstr(sfmt, "e-")) != NULL) 
         {
            sprintf(sfmt, "%%.%sf", spos+2); 
            g_free (text);
            text = g_strdup_printf(sfmt, _dval);
         }
      } 
      set_value(text);
      gtk_entry_set_text (GTK_ENTRY(_base_w), text);
      g_free (text);
   }
   else             // sexagesimal  
   {
      string tmpval;
      double tmpd; 
      if (_type == 0)  // ra
      {
         deg_to_ra(_dval, tmpval);
         if (tmpval != val)

            ra_to_deg(val, tmpd);
         else
            tmpd = _dval;
         if (dir == 0)
         {
            tmpd += _stepsize;
         }
         else
         {
            tmpd = tmpd - _stepsize;
         }
         deg_to_ra(tmpd, val);
      }
      else 
      {
         deg_to_dec(_dval, tmpval);
         if (tmpval != val)
            dec_to_deg(val, tmpd);
         else
            tmpd = _dval;
         if (dir == 0)
         {
            tmpd += _stepsize; 
//            tmpd += .000278; 
         }
         else 
         {
            tmpd -= _stepsize; 
 //           tmpd -= .000278; 
         }
         deg_to_dec(tmpd, val);
      }
      gtk_entry_set_text(GTK_ENTRY(_base_w), val.c_str());
      _dval = tmpd; 
   }
   validate_text(); 
   run_change_value_cb();
}

void ChipsGuiSpinWidget::value_changed_callback(GtkWidget *widget, gpointer data)
{
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);

   double dval = me->get_numeric_value(); 

   if (me->_callbacks.count ("change_value") > 0)
      me->_callbacks["change_value"]->execute (me, NULL);
}

bool ChipsGuiSpinWidget::focus_out_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   ChipsGuiSpinWidget *me = static_cast <ChipsGuiSpinWidget*>(data);
      // check for mode change
      int sts = me->validate_text();
      if (sts != -1)
         me->set_mode(sts);
      else
         return true;   // invalid text- don't process

      // treat text as entered and run change_value cb
      me->run_change_value_cb();
      return true;

   return false;
}

void ChipsGuiSpinWidget::set(double val, int mode)
{
   _dval = val; 
   if (mode == 0 || mode == 1)  _mode = mode; 
   if (_mode == 0)
   {
      std::ostringstream strs;
      strs.precision(12);
      strs << _dval;
      _sval = strs.str();
   }
   else if (_mode == 1) 
   {
      if (_type == 0)   deg_to_ra(_dval, _sval);
         else           deg_to_dec(_dval, _sval);
   }
   gtk_entry_set_text(GTK_ENTRY(_base_w), _sval.c_str());
}

int ChipsGuiSpinWidget::validate_text()
{
   const char* val = gtk_entry_get_text (GTK_ENTRY(_base_w));
   double tmpd; 
   int rv; 

   // check for numeric value 
   char * pEnd;
   tmpd = strtod(val, &pEnd);
   if (pEnd == &val[strlen(val)])
   {
      flag_invalid_text(0);   // valid numeric value 

      // if the user typed in a numeric value- set precison accordingly
      const char* dpnt = strchr(val,'.'); 
      if (dpnt)  set_precision(strlen(dpnt)-1); 
      return 0; 
   }

   // check ra/dec conversion or double 
   if (_type == 0)  // ra
      rv = ra_to_deg(val, tmpd);
   else
      rv = dec_to_deg(val, tmpd);
   // valid ra/dec string  
   if (rv == 0)    
   {
      flag_invalid_text(0); // valid ra/dec value 
      return 1; 
   }

   // highlight invalid string
   flag_invalid_text(1);   
   return -1; 
}


// change text to red if invalid (set is 1)  
void ChipsGuiSpinWidget::flag_invalid_text(bool set)
{
   const GdkColor RED_COLOR = { 0, 65535, 0, 0 };
   const GdkColor *color;
   if (set)
   {
      color = &RED_COLOR;
   }
   else    color = NULL;

   gtk_widget_modify_text(GTK_WIDGET(_base_w), GTK_STATE_NORMAL, color);
   gtk_widget_modify_text(GTK_WIDGET(_base_w), GTK_STATE_ACTIVE, color);
   gtk_widget_modify_text(GTK_WIDGET(_base_w), GTK_STATE_PRELIGHT, color);
   gtk_widget_modify_text(GTK_WIDGET(_base_w), GTK_STATE_SELECTED, color);
}

void ChipsGuiSpinWidget::set_precision(int digits)
{
   _precision = digits; 
}


void ChipsGuiSpinWidget::reset_scrollsize()
{
   // if a change_step callback has been defined, run it 
   if (_callbacks.count ("change_step") > 0)
      _callbacks["change_step"]->execute(this, NULL);
}

void ChipsGuiSpinWidget::reset_scrollcnt(unsigned int timestamp)
{
   // if we stop scrolling for more that 2.5 seconds, recalculate the
   // scrolling step size 
   if (timestamp - _scrolltimestamp > 2500)   reset_scrollcnt();  
   _scrolltimestamp = timestamp;  
}

void ChipsGuiSpinWidget::disable ()
{
   gtk_widget_set_sensitive (_base_w, FALSE);
}

void ChipsGuiSpinWidget::enable ()
{
   gtk_widget_set_sensitive (_base_w, TRUE);
}

void ChipsGuiSpinWidget::disable_menu ()
{
   _enable_popup = false;
}

void ChipsGuiSpinWidget::enable_menu ()
{
   _enable_popup = true;
}

