# 
#  Copyright (C) 2007  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


_object_types = dict (ChipsWindow='window',
                      ChipsBorder='border',
                      ChipsFrame='frame',
                      ChipsTitle='title',
                      ChipsPlotCorner='corner',
                      ChipsPlotAspect='aspect',
                      ChipsPlot='plot',
                      ChipsMajorGrid='majorgrid',
                      ChipsMinorGrid='minorgrid',
                      ChipsMajorTick='majortick',
                      ChipsMinorTick='minortick',
                      ChipsAxisXBase='x',
                      ChipsAxisYBase='y',
                      ChipsAxisLabBase='label',
                      ChipsAxisLabel='label',
                      ChipsAxisOffset='offset',
                      ChipsAxisTicklabel='ticklabel',
                      ChipsAxis='axis',
                      ChipsSymbol='symbol',
                      ChipsErr='err',
                      ChipsHistogramErr='err',
                      ChipsHistogramFill='fill',
                      ChipsHistogram='histogram',
                      ChipsColorBarTicklabel='ticklabel',
                      ChipsCurveLine='line',
                      ChipsCurveLimit='limit',
                      ChipsCurve='curve',
                      ChipsContour='contour',
                      ChipsLabel='label',
                      ChipsLine='line',
                      ChipsPoint='point',
                      ChipsRegionFill='fill',
                      ChipsRegionEdge='edge',
                      ChipsRegion='region',
                      ChipsExport='export',
                      ChipsForeground='foreground',
                      ChipsBackground='background',
                      ChipsPick='pick',
                      ChipsInfo='info',
		      ChipsDefaultPrefs='default',
                      ChipsPreferences='preferences',
                      ChipsImage='image',
                      ChipsColorBar="colorbar",
                      ChipsColorBarLabel="label",
                      ChipsColorBarBorder="border",
                      ChipsColorBarTick="tick",
                      ChipsWildCard="all")
      


# Abstract attribute class
class ChipsAttributeObject (object):
   """
   Parent class for all attribute class
   """
   def __init__ (self):
      self.all = ChipsWildCard(self)

   def __print__ (self, depth=0, prefix=''):
      'Prints out a structure as member=val list with 1 entry per line'
      if depth != 0:
         prefix = prefix + _object_types[self.__class__.__name__]
         if (prefix != ''):
            prefix = prefix + '.'
      t = ''
      for x in dir(self):
         if not x.startswith ('_'):
            attr = getattr(self, x)
            # if attribute is a class instance- recursive call to print
            if isinstance (attr, ChipsAttributeObject):
               t = t + attr.__print__(depth+1, prefix)
            elif x != 'all':
               t = t + prefix + x + ' = ' + '%s'%getattr(self,x) + '\n'
      return t

   def __set_all__(self, prop, value):
      'sets all fields ending with prop to value'
      cnt = 0
      for x in dir(self):
         if not x.startswith ('_'):
            attr = getattr(self, x)
            # if attribute is a class instance- recursive call to process
            if isinstance (attr, ChipsAttributeObject):
               cnt += attr.__set_all__(prop, value)
               # handle compound prop such as fill.color 
               if prop.find('.') != -1:
                  pt = prop.partition('.')
                  if len(pt) == 3 and pt[0] == x:
                     cnt += attr.__set_all__(pt[2], value)
            elif prop in x:
               if hasattr(self,prop):
                  setattr(self,prop,value) 
                  cnt += 1
      return cnt 


class ChipsWindow (ChipsAttributeObject):
   """
   Chips attribute object class for windows. Contains all attributes which
   can be set and retrieved for/from a chips window.
   """
   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsWindow, self).__init__ ()

      self.height = None
      self.width = None
      self.units = None
      self.scaleheight = None
      self.scalewidth = None
      self.smoothing = None
      self.display = None
      self.stem = None
      self.bgcolor = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsBorder (ChipsAttributeObject):
   """
   Chips attribute object class for a frame border. Contains all attributes  
   which can be set and retrieved for/from a chips frame border.
   """
   def __init__(self):
      self.color = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsFrame (ChipsAttributeObject):
   """
   Chips attribute object class for frames. Contains all attributes which
   can be set and retrieved for/from a chips frame border.
   """
   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsFrame, self).__init__ ()

      self.bgcolor = None
      self.transparency = None
      self.border = ChipsBorder()
      self.mode = None
      self.aspect = None
      self.scale = None
      self.stem = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsTitle (ChipsAttributeObject):
   """
   Chips attribute object class for plot titles. Contains all attributes which
   can be set and retrieved for/from a chips plot title.
   """

   def __init__(self):
      self.xpos = None
      self.ypos = None
      self.angle = None
      self.depth = None
      self.halign = None
      self.valign = None
      self.color = None
      self.size = None
      self.font = None
      self.fontstyle = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsPlotCorner (ChipsAttributeObject):
   """
   Chips attribute object class for plot corners. Contains all attributes which
   can be set and retrieved for/from a chips plot corner.
   """
   def __init__(self):
      self.style = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsPlotAspect (ChipsAttributeObject):
   """
   Chips attribute object class for plot aspect related parameters. Contains all 
   attributes which can be set and retrieved for/from chips plot aspect.
   """
   def __init__(self):
        self.ratio = None
   def __str__(self):
        return self.__print__()
   def __repr__(self):
        return self.__print__()


class ChipsPlot (ChipsAttributeObject):
   """
   Chips attribute object class for plots. Contains all attributes which
   can be set and retrieved for/from a chips plot.
   """
   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsPlot, self).__init__ ()

      self.title = ChipsTitle()
      self.style = None
      self.corner = ChipsPlotCorner()
      self.leftmargin = None
      self.rightmargin = None
      self.topmargin = None
      self.bottommargin = None
      self.aspect = ChipsPlotAspect()
      self.stem = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsMajorGrid (ChipsAttributeObject):
   """
   Chips attribute object class for an axis' major grid. Contains all attributes
   which can be set and retrieved for/from a chips axis major grid.
   """

   def __init__(self):
      self.visible = None
      self.style = None
      self.color = None
      self.thickness = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsMinorGrid (ChipsAttributeObject):
   """
   Chips attribute object class for an axis' minor grid. Contains all attributes
   which can be set and retrieved for/from a chips axis minor grid.
   """

   def __init__(self):
      self.visible = None
      self.style = None
      self.color = None
      self.thickness = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsMajorTick (ChipsAttributeObject):
   """
   Chips attribute object class for axis major ticks. Contains all attributes 
   which can be set and retrieved for/from a chips axis major tick.
   """

   def __init__(self):
      self.mode = None
      self.interval = None
      self.count = None
      self.thickness = None
      self.color = None
      self.style = None
      self.length = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsMinorTick (ChipsAttributeObject):
   """
   Chips attribute object class for axis minor ticks. Contains all attributes 
   which can be set and retrieved for/from a chips axis minor tick.
   """
   def __init__(self):
      self.mode = None
      self.interval = None
      self.count = None
      self.thickness = None
      self.color = None
      self.style = None
      self.length = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisLabBase  (ChipsAttributeObject):
   """
   Chips attribute object class for a an axis label base. Contains attributes 
   which can be set and retrieved for/from a chips axis label base.
   """
   def __init__(self):
      self.text = None
      self.angle = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisXBase (ChipsAttributeObject):
   """
   Chips attribute object class for an x axis base. Contains all attributes
   which can be set and retrieved for/from a chips x axis base.
   """

   def __init__(self):
      self.stem = None
      self.label = ChipsAxisLabBase()
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisYBase (ChipsAttributeObject):
   """
   Chips attribute object class for a y axis base. Contains all attributes 
   which can be set and retrieved for/from a chips y axis base.
   """

   def __init__(self):
      self.stem = None
      self.label = ChipsAxisLabBase()
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisLabel (ChipsAttributeObject):
   """
   Chips attribute object class for an axis label. Contains all attributes 
   which can be set and retrieved for/from a chips axis label.
   """

   def __init__(self):
      self.color = None
      self.size = None
      self.font = None
      self.fontstyle = None
      self.halign = None
      self.valign = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisOffset (ChipsAttributeObject):
   """
   Chips attribute object class for an axis offset. Contains all attributes 
   which can be set and retrieved for/from a chips axis offset.
   """

   def __init__(self):
      self.parallel = None
      self.perpendicular = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxisTicklabel (ChipsAttributeObject):
   """
   Chips attribute object class for axis tick labels. Contains all attributes 
   which can be set and retrieved for/from a chips axis tick label.
   """

   def __init__(self):
      self.style = None
      self.color = None
      self.size = None
      self.angle = None
      self.valign = None
      self.halign = None
      self.font = None
      self.fontstyle = None
      self.offset = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsAxis (ChipsAttributeObject):
   """
   Chips attribute object class for axes. Contains all attributes which
   can be set and retrieved for/from a chips axis.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsAxis, self).__init__ ()

      self.x = ChipsAxisXBase()
      self.y = ChipsAxisYBase()
      self.depth = None
      self.color = None
      self.thickness = None
      self.label = ChipsAxisLabel()
      self.majorgrid = ChipsMajorGrid()
      self.minorgrid = ChipsMinorGrid()
      self.automin = None
      self.automax = None
      self.pad = None
      self.offset = ChipsAxisOffset()
      self.tickformat = None
      self.ticklabel = ChipsAxisTicklabel()
      self.tickstyle = None
      self.majortick = ChipsMajorTick()
      self.minortick = ChipsMinorTick()
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsSymbol (ChipsAttributeObject):
   """
   Chips attribute object class for symbols. Contains all attributes which
   can be set and retrieved for/from a chips symbol.
   """

   def __init__(self):
      self.color = None
      self.style = None
      self.size = None
      self.angle = None
      self.fill = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsErr (ChipsAttributeObject):
   """
   Chips attribute object class for error bars. Contains all attributes which
   can be set and retrieved for/from a chips error bars.
   """

   def __init__(self):
      self.style = None
      self.up = None
      self.down = None
      self.left = None
      self.right = None
      self.color = None
      self.caplength = None
      self.thickness = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsHistogramErr (ChipsAttributeObject):
   """
   Chips attribute object class for histogram error bars. Contains all 
   attributes which can be set and retrieved for/from a chips histogram error
   bar.
   """

   def __init__(self):
      self.style = None
      self.up = None
      self.down = None
      self.color = None
      self.caplength = None
      self.thickness = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()


class ChipsCurveLine (ChipsAttributeObject):
   """
   Chips attribute object class for curve lines. Contains all attributes which
   can be set and retrieved for/from a chips curve line.
   """

   def __init__(self):
      self.color = None
      self.thickness = None
      self.style = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsCurveLimit (ChipsAttributeObject):
   """
   Chips attribute object class for curve limits. Contains all attributes which
   can be set and retrieved for/from a chips curve limit.
   """

   def __init__(self):
      self.length = None
      self.override = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsWildCard():
   """
   Chips attribute object to handle wildcard 
   """
   def __init__ (self, parent):
      self.__local_parent__ = parent

   def __setattr__ (self, name, value):
      if name != "__local_parent__":
         self.__local_parent__.__set_all__ (name, value)
      else:
         self.__dict__[name] = value

   def __str__(self):
      return ""
         
   def __repr__(self):
      return ""

class ChipsCurve (ChipsAttributeObject):
   """
   Chips attribute object class for curve. Contains all attributes which
   can be set and retrieved for/from a chips curve.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsCurve, self).__init__ ()

      self.stem = None
      self.depth = None
      self.baddata = None
      self.line = ChipsCurveLine()
      self.symbol = ChipsSymbol()
      self.err = ChipsErr()
      self.limit = ChipsCurveLimit()
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsHistogramFill (ChipsAttributeObject):
   """
   Chips attribute object class for a histogram's fill. Contains all attributes 
   which can be set and retrieved for/from a chips histogram's fill.
   """

   def __init__(self):
      self.color = None
      self.opacity = None
      self.style = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsHistogram (ChipsAttributeObject):
   """
   Chips attribute object class for histograms. Contains all attributes which
   can be set and retrieved for/from a chips histogram.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsHistogram, self).__init__ ()

      self.stem = None
      self.depth = None
      self.line = ChipsCurveLine()
      self.symbol = ChipsSymbol()
      self.err = ChipsHistogramErr()
      self.dropline = None
      self.fill = ChipsHistogramFill ()
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsContour (ChipsAttributeObject):
   """
   Chips attribute object class for contours. Contains all attributes which
   can be set and retrieved for/from a chips contour.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsContour, self).__init__ ()

      self.stem = None
      self.depth = None
      self.color = None
      self.thickness = None
      self.style = None
      self.levels = None
      self.numlevels = None
      self.algorithm = None
      self.mode = None
      self.interval = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsImage (ChipsAttributeObject):
   """
   Chips attribute object class for images. Contains all attributes which
   can be set and retrieved for/from a chips image
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsImage, self).__init__ ()

      self.stem = None
      self.id = None
      self.colormap = None
      self.invert_colormap = None
      self.colormap_interpolate = None
      self.colormap_size = None
      self.interpolation = None
      self.scale_channels = None
      self.alpha = None
      self.threshold = None
      self.depth = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsColorBarBorder (ChipsAttributeObject):
   """
   Chips attribute object class for a color bar border. Contains all attributes
   which can be set and retrieved for/from a chips color bar border.
   """
   def __init__(self):
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()


class ChipsColorBarLabel (ChipsAxisLabel):
   """
   Chips attribute object class for color bar labels. Contains all attributes 
   which can be set and retrieved for/from a chips color bar label.
   """

   def __init__(self):
      ChipsAxisLabel.__init__ (self)
      self.text = None
      self.angle = None
      self.location = None

class ChipsColorBarTick (ChipsAttributeObject):
   """
   Chips attribute object class for color bar ticks. Contains all attributes
   which can be set and retrieved for/from a chips color bar tick.
   """
   def __init__(self):
      self.mode = None
      self.thickness = None
      self.color = None
      self.style = None
      self.length = None
      self.visible = None
      self.location = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsColorBarTicklabel (ChipsAttributeObject):
   """
   Chips attribute object class for colorbar tick labels. Contains all 
   attributes which can be used with a chips colorbar tick label.
   """

   def __init__(self):
      self.color = None
      self.size = None
      self.angle = None
      self.valign = None
      self.halign = None
      self.font = None
      self.fontstyle = None
      self.offset = None
      self.visible = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()


class ChipsColorBar (ChipsAttributeObject):
   """
   Chips attribute object class for color bars. Contains all attributes 
   which can be set and retrieved for/from a chips color bar.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsColorBar, self).__init__ ()

      self.stem = None
      self.id = None
      self.depth = None
      self.length = None
      self.width = None
      self.border = ChipsColorBarBorder ()
      self.halign = None
      self.valign = None
      self.orientation = None
      self.label = ChipsColorBarLabel ()
      self.offset = ChipsAxisOffset ()
      self.tick = ChipsColorBarTick ()
      self.ticklabel = ChipsColorBarTicklabel ()
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()


class ChipsLabel (ChipsAttributeObject):
   """
   Chips attribute object class for labels. Contains all attributes which
   can be set and retrieved for/from a chips label.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsLabel, self).__init__ ()

      self.stem = None
      self.depth = None
      self.color = None
      self.size = None
      self.angle = None
      self.valign = None
      self.halign = None
      self.font = None
      self.fontstyle = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsLine (ChipsAttributeObject):
   """
   Chips attribute object class for lines. Contains all attributes which
   can be set and retrieved for/from a chips line.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsLine, self).__init__ ()

      self.stem = None
      self.depth = None
      self.color = None
      self.thickness = None
      self.style = None
      self.extend = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsPoint (ChipsAttributeObject):
   """
   Chips attribute object class for points. Contains all attributes which
   can be set and retrieved for/from a chips point.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsPoint, self).__init__ ()

      self.stem = None
      self.depth = None
      self.color = None
      self.size = None
      self.style = None
      self.fill = None
      self.angle = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsRegionFill (ChipsAttributeObject):
   """
   Chips attribute object class for a region's fill. Contains all attributes 
   which can be set and retrieved for/from a chips region's fill.
   """

   def __init__(self):
      self.color = None
      self.style = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsRegionEdge (ChipsAttributeObject):
   """
   Chips attribute object class for a region's edge. Contains all attributes 
   which can be set and retrieved for/from a chips region's edge.
   """

   def __init__(self):
      self.thickness = None
      self.style = None
      self.color = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsRegion (ChipsAttributeObject):
   """
   Chips attribute object class for regions. Contains all attributes which
   can be set and retrieved for/from a chips region.
   """

   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsRegion, self).__init__ ()

      self.stem = None
      self.depth = None
      self.fill = ChipsRegionFill()
      self.edge = ChipsRegionEdge()
      self.angle = None
      self.opacity = None
      self.id = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsExport (ChipsAttributeObject):
   """
   Chips attribute object class for export. Contains all attributes which
   can be set and retrieved when printing a window.
   """

   def __init__(self):
      self.fonts = None
      self.format = None
      self.orientation = None
      self.colorscheme = None
      self.leftmargin = None
      self.rightmargin = None
      self.topmargin = None
      self.bottommargin = None
      self.pagewidth = None
      self.pageheight = None
      self.pagesize = None
      self.pageunits = None
      self.scaleheight = None
      self.scalewidth = None
      self.keepaspect = None
      self.fittopage = None
      self.clobber = None
      self.printdialog = None
      self.dpi = None
      self.printmethod = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsForeground (ChipsAttributeObject):
   """
   Chips attribute object class for foreground. Contains all attributes which
   can be set and retrieved for/from the foreground preference property.
   """

   def __init__(self):
      self.file = None
      self.display = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsBackground (ChipsAttributeObject):
   """
   Chips attribute object class for foreground. Contains all attributes which
   can be set and retrieved for/from the background preference property.
   """
   def __init__(self):
      self.file = None
      self.display = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsPick (ChipsAttributeObject):
   """
   Chips attribute object class for pick. Contains all attributes which
   can be set and retrieved when using the pick/get_pick commands.
   """

   def __init__(self):
      self.mask = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()

class ChipsInfo (ChipsAttributeObject):
   """
   Chips attribute object class for info. Contains all attributes which
   can be set and retrieved when using the info command.
   """

   def __init__ (self):
      self.current = None
      self.depth = None
      self.coord = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()
   def __set_all__(self, prop, value):
      'no op- we do not support wild card settings for this object'
      return 0



class ChipsDefaultPrefs (ChipsAttributeObject):
   """
   Chips attribute object class for info. Contains all attributes which
   are have a 'default' object prefix.
   """

   def __init__ (self):
      self.depth = None
      self.verbosity = None
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()
   def __set_all__(self, prop, value):
      'no op- we do not support wild card settings for this object'
      return 0



class ChipsPreferences (ChipsAttributeObject):
   """
   Chips attribute object class for preferences. Contains all attributes which
   can be set and retrieved when setting preferences.
   """
   
   def __init__(self):
      # initialize ChipsAttribute first
      super(ChipsPreferences, self).__init__ ()

      self.window = ChipsWindow()
      self.frame = ChipsFrame()
      self.plot = ChipsPlot()
      self.axis = ChipsAxis()
      self.curve = ChipsCurve()
      self.histogram = ChipsHistogram()
      self.contour = ChipsContour()
      self.image = ChipsImage()
      self.colorbar = ChipsColorBar()
      self.label = ChipsLabel()
      self.line = ChipsLine()
      self.point = ChipsPoint()
      self.region = ChipsRegion()
      self.foreground= ChipsForeground()
      self.background= ChipsBackground()
      self.export = ChipsExport()
      self.pick = ChipsPick()
      self.info = ChipsInfo()
      self.default = ChipsDefaultPrefs()
   def __str__(self):
      return self.__print__()
   def __repr__(self):
      return self.__print__()


