# 
#  Copyright (C) 2009,2013  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# Defined are high level user interface routines which interact
# with crates

import struct
import os
import array
import numpy
import pychips as chips
from pychips import advanced
from pychips import utils as __utils__

from pytransform import *
from pycrates import *

__all__ = ["ChipsContour", "get_contour", "set_contour", "add_contour", "add_curve", "add_histogram", "add_image", "get_image", "delete_image", "set_image_channel", "make_figure", "make_script", "binary_to_fits"]

# overloaded chips classes
class ChipsContour (chips.ChipsContour):
   def __init__(self):
     chips.ChipsContour.__init__(self)
     self.wcs = None

# check if the object is a crate
def _is_crate_ (obj):
  """
    _is_crate_ (obj)

    return if the object is either an instance of TABLECrate or an IMAGECrate
  """
  return  (isinstance (obj, TABLECrate) or isinstance (obj, IMAGECrate))

def __get_axis_label__(cdat):
   """
      __get_axis_label__(cdat)

      takes in a CrateDate object and sets up the string for axis label
   """
   my_lab = None
   if isinstance (cdat, CrateData):
      my_lab = cdat.name
      uns = cdat.unit
      if uns != None and (len(uns) != 0):
         my_lab = my_lab + ' (' + uns + ')'
      my_lab = my_lab.replace('_','\\_')
   return my_lab


# overloaded contour functions
def get_contour (id=None, property=None):
   if type(id) == str:
      oid = chips.pychips_base.ChipsId()
      oid.set_contour_id(id)
   elif isinstance(id, chips.pychips_base.ChipsId):
      oid = id
   elif id == None:
      oid = chips.pychips_base.ChipsId()
   else:
      raise AttributeError("chips ERROR: Unknown or unexpected parameter type")

   result = None
   if property and property != 'all':
      if (property.find('contour') == 0):    # remove preceeding 'contour.'
         property = property.replace('contour.','')
      if property == 'wcs':
         transform = advanced.get_contour_transform ()
         if transform:
            result = transform.get_name ()
      else:
         result = chips.get_contour (id, property)
   else:

      result = chips.get_contour (id, property)
      # deprecated 
#      result = ChipsContour();
#
#      tmp_result = chips.get_contour (id, property)
#      for memb in dir (tmp_result):
#         if not memb.startswith ('_'):
#            attr = getattr (tmp_result, memb)
#            setattr (result, memb, attr)
#      transform = advanced.get_contour_transform ()
#      if transform:
#         val = transform.get_name ()
#      else:
#         val = None
#      result.wcs = val


   return result

get_contour.__doc__ = chips.get_contour.__doc__


def set_contour (id=None, values=None, crate=None):

   if crate == None and values == None and id == None:
      raise AttributeError("chips ERROR: No attribute values specified")

   # ensure values and id are in the correct order
   if _is_crate_ (crate):
      tmp_crate = crate
   elif _is_crate_ (values):
      tmp_crate = values
   elif _is_crate_ (id):
      raise AttributeError("chips ERROR: First element of set_contour cannot be a crate")
   else:
      tmp_crate = None

   if values == None or _is_crate_ (values):   # handle values without ids
      mid = None
      if not isinstance(id, chips.pychips_base.ChipsId):
         vals = id
      else:
         vals = None
   else:
      # both provided- need to verify order
      if type(id) == str or isinstance(id, chips.pychips_base.ChipsId) or id == None:
         mid = id;
         vals = values;
      else:
         mid = values
         vals = id

   if vals == None:
      raise AttributeError("chips ERROR: No attribute values specified.")

   oid = __utils__.set_id (mid, chips.pychips_base.chips_contour)
   chips.pychips_base.lock()

   # if input is a string attribute list convert to attribute value pairs
   if type(vals) == str:
      vals = __utils__.string_to_attribute_list(vals)

   # convert data structure to ChipsContour object
   ob = ChipsContour ()

   try:
      ob = __utils__.data_to_attribute_object (ob, vals)
   except:
      chips.pychips_base.unlock()
      raise

   object_type = __utils__.get_object_type (ob)
   if object_type == None:
      chips.pychips_base.unlock()
      raise AttributeError("chips ERROR: You cannot call set_contour with %s "%type(ob))

   try:
      chips.pychips_base.open_undo_block ()
      chips.pychips_base.open_undo_buffer()

      if hasattr(ob, 'wcs') and ob.wcs != None and tmp_crate != None:
         try:
            transform = get_axis_transform (tmp_crate, ob.wcs)
         except:
            try:
               transform = get_transform (tmp_crate, ob.wcs)
            except:
               chips.pychips_base.discard_undo_buffer()
               chips.pychips_base.close_undo_block ()
               chips.pychips_base.unlock()
               raise

         #concatinate the physical and world coordinate systems
         if (ob.wcs.upper () == 'WORLD') or (ob.wcs == 'EQPOS'):
            phys_trans = get_axis_transform (tmp_crate, 'PHYSICAL')
            if (phys_trans):
               transform = _concat_wcs_transform_ (phys_trans, transform)

         advanced.set_contour_transform (oid, transform)
   except:
      chips.pychips_base.discard_undo_buffer()
      chips.pychips_base.close_undo_block ()
      chips.pychips_base.unlock()
      raise
   else:
      chips.pychips_base.close_undo_buffer()

   try:
      chips.set_contour (oid, ob)
   finally:
      chips.pychips_base.close_undo_block()
      chips.pychips_base.unlock()

set_contour.__doc__ = chips.set_contour.__doc__


#TEMPORARY CODE !!!! Needed because right now transforms
#always expect to go from physical to world. 
def _concat_wcs_transform_ (phys_trans, world_trans):
   if not phys_trans or not world_trans:
      return None

   mtrx1 = numpy.asmatrix (phys_trans.get_transform_matrix ())
   mtrx2 = numpy.asmatrix (world_trans.get_transform_matrix ())

   mtrx = mtrx1 * mtrx2

   world_param_list = world_trans.get_parameter_list ()
   concat_trans = WCSTANTransform(world_trans.get_name (), world_param_list)
   concat_trans.set_transform_matrix (mtrx)
   return concat_trans

def add_image (*args):
   """
   add_image([ChipsId,] [red_image] [,green_image] [,blue_image] [,alpha_image], x-dim, y-dim, [,attribute_list])
   add_image([ChipsId,] [IMAGECrate|file name] [,IMAGECrate|file name] [,IMAGECrate|file name] [,IMAGECrate|file name], [,attribute_list)
   """
   __add_image__(*args)


def __add_image__ (*args):
   """
   add_image([ChipsId,] [red_image] [,green_image] [,blue_image] [,alpha_image], x-dim, y-dim, [,attribute_list])
   add_image([ChipsId,] [IMAGECrate|file name] [,IMAGECrate|file name] [,IMAGECrate|file name] [,IMAGECrate|file name], [,attribute_list)

   Returns the wcs transform component names used or None
   """
   curr_arg_indx = 0
   new_args = []
   img = [None, None, None, None]
   first_image = None
   image_dim = [None, None]
   xname = None
   yname = None

   # format the arg list so that attribute structures are converted to
   # strings if they exist
   args = __utils__.format_add_argslist (0,False,__utils__.get_object_type(chips.ChipsImage()),*args)

   if (len (args) > 0):
      # Check if the first arg is a chips id
      if isinstance (args[curr_arg_indx], chips.pychips_base.ChipsId):
         new_args.append (args[curr_arg_indx])
         curr_arg_indx = curr_arg_indx + 1


   for ii in range (4):
      if (len (args) > curr_arg_indx):
         # Check if the current arg is a file name
         if isinstance (args[curr_arg_indx], str):
            try:
               img[ii] = IMAGECrate (args[curr_arg_indx], 'r')
            except Exception as e:
               # if this is the first attribute assume then it must
               # be a file name or if it's not the the last attribute
               # if this is the last attribute assume that the string
               # is an attribute string and continue processing
               if ii == 0 or curr_arg_indx != len(args)-1:
                  raise e
               else:
                  img[ii] = None
                  break
            if first_image == None:
               first_image = img[ii]
            curr_arg_indx = curr_arg_indx + 1
         # Check if the current arg is an image crate
         elif isinstance (args[curr_arg_indx], IMAGECrate) :
            img[ii] = args[curr_arg_indx];
            curr_arg_indx = curr_arg_indx + 1
            if first_image == None:
               first_image = img[ii]
         elif args[curr_arg_indx] == None or \
              __utils__.check_seq (args[curr_arg_indx]): 
            img[ii] = args[curr_arg_indx];
            curr_arg_indx = curr_arg_indx + 1

   # there was at least one crate
   if first_image != None:

      transform = None
      trans_name = None
      # add each image to the argument list
      for ii in range (4):

         if isinstance (img[ii], IMAGECrate):

            # add crate information and original entries into new arg list
            new_args.append (img[ii].get_image ().values)

            shape = img[ii].get_shape ()
            xdim_index = len(shape)-1
            ydim_index = len(shape)-2
            # check that the dims are the same
            if image_dim[0] == None:
               image_dim[0] = shape[xdim_index]
            elif image_dim[0] != shape[xdim_index]:
               raise AttributeError("chips ERROR: The dimensions of the each image must be equal. The dimensions of image %d differs from the previous image."%(ii))

            if image_dim[1] == None:

               # if y index is less than zero than it's a 1D array
               if ydim_index >= 0:
                  image_dim[1] = shape[ydim_index]
               else:
                  image_dim[1] = 1
            elif image_dim[1] != shape[ydim_index]:
               raise AttributeError("chips ERROR: The dimensions of the each image must be equal. The dimensions of image %d differs from the previous image."%(ii))

         # just add the sequence
         else:
            new_args.append (img[ii])


      # add the dimensions to the list
      new_args.append (image_dim[0])
      new_args.append (image_dim[1])
      

      # pull the last element from the arglist. Ingore if there's only 
      # one element
      if len(args) > curr_arg_indx:
         arg_end = args[len(args)-1]
      else:
         arg_end = None


      for i in range (curr_arg_indx,len(args)-1):
         new_args.append (args[i])
         curr_arg_indx = curr_arg_indx + 1


      # if the last argument is a string parse it for wcs and remove the 
      # wcsentry from the list
      if (type (arg_end) == str):
         (trans_name, arg_end) = __utils__.pop_attrlist ('wcs', arg_end)

         # if the name exists try to pull it from the crate
         if trans_name:
            # if the transform name was passed into the function use it
            try:
               transform = get_axis_transform (first_image, trans_name)
            except:
               transform = get_transform (first_image, trans_name)
      elif arg_end != None:
         # arg end wasn't an attribute list put it back on the argument
         # list before the transforms get added
         new_args.append (arg_end)
         arg_end = None


      # if the transform has not been set try each of the default
      # transform categrories
      if not trans_name:
         try:
            transform = get_axis_transform (first_image, 'WORLD')
         except:
            try:
               transform = get_axis_transform (first_image, 'PHYSICAL')
            except:
               try:
                  transform = get_axis_transform (first_image, 'LOGICAL')
               except:
                  pass
         else:
            trans_name = 'WORLD'
            
      #If we're using a world translation, check if it's a supported type
      #Warn and fall back to physical if it's not
      #Also change trans_name so we don't accidentally concat later
      if trans_name and trans_name.upper() == 'WORLD':
         if not _legal_world_transform(transform):
            ctype_values = _get_ctype_values(transform)
            try:
               transform = get_axis_transform (first_image, 'PHYSICAL')
               trans_name = 'PHYSICAL'
               print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using physical transform.".format(ctype_values)
               #raise AttributeError("chips ERROR: '"+table.get_filename () + "' contains < 2 columns")
            except LookupError:
               try:
                  transform = get_axis_transform (first_image, 'LOGICAL')
                  trans_name = 'LOGICAL'
                  print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using logical transform.".format(ctype_values)
                  if not transform:  #If there's no actual transform (and there won't be), Create an identity-matrix 2D transform so this won't have the plot's default transform applied to it. 
                     transform = LINEAR2DTransform()
                     for parameter in transform.get_parameter_list(): #Initialize parameters
                        if parameter.get_name() == "SCALE":
                           parameter.set_value(numpy.array([1, 1]))
                        if parameter.get_name() == "ROTATION":
                           parameter.set_value(numpy.float64(0))
                        if parameter.get_name() == "OFFSET":
                           parameter.set_value(numpy.array([0, 0]))
                  
               except LookupError:
                  #We current'y can't reach this code - Logical in get_axis_transform returns a None.  
                  #Makes sense since we don't actually need to change the object at all to do a logical transform.  
                  #Keeping this here for futureproofing.  
                  transform = None
                  trans_name = None
                  print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using base coordinates.".format(ctype_values)

      #concatinate the physical and world coordinate systems
      if trans_name and ((trans_name.upper() == 'WORLD') or (trans_name.upper() == 'EQPOS')):
         phys_trans = get_axis_transform (first_image, 'PHYSICAL')
         if (phys_trans):
            transform = _concat_wcs_transform_ (phys_trans, transform)

      if transform:
         new_args.append (transform)

         # get wcs component names for axes
         try:
            tmpax= first_image.get_axis(transform.get_name())
            if tmpax != None:
               axis_names = tmpax.get_cptslist ()
               try:
                  xname = axis_names[0]
                  yname = axis_names[1]
               except:
                  pass
         except KeyError:
            pass #We should only hit this if we made up a logical transform


      if (arg_end):
         new_args.append (arg_end)

      # handle meta data
      if first_image != None:
         tmstr = first_image.get_filename()
         if isinstance (new_args[len(new_args)-1], str):
            if tmstr.find('"') != -1:
               new_args[len(new_args)-1] = new_args[len(new_args)-1] + " meta=" + repr(tmstr)
            else:
               new_args[len(new_args)-1] += " meta=\""
               new_args[len(new_args)-1] += tmstr
               new_args[len(new_args)-1] += "\""
         else:
            if tmstr.find('"') != -1:
               new_args.append(" meta="+repr(tmstr))
            else:
               new_args.append(" meta=\""+tmstr+"\"")

      new_args = tuple (new_args)
      chips.add_image (*new_args)
   else:
      chips.add_image (*args)

   return xname, yname


def get_image(*args): 
   """
   get_image([id [,property]]) 

   Returns a ChipsImage object with attribute values of requested image. If an
   id string or ChipsId is provied, a specific property value may also be specified'
   """
   result = chips.get_image(*args)
   return result

def delete_image (id=None):
   if type(id) == str:
      oid = chips.pychips_base.ChipsId()
      oid.set_image_id(id)
   elif isinstance(id, chips.pychips_base.ChipsId):
      oid = id
   elif id == None:
      oid = chips.pychips_base.ChipsId()
   else:
      raise AttributeError("chips ERROR: Unknown or unexpected parameter type")

   result = chips.delete_image(id);
   return result

delete_image.__doc__ = chips.delete_image.__doc__

def set_image_channel (*args):
   """
   set_image_channel ([id,] image_array, x-dim, y-dim, channel)
   set_image_channel ([id,] image_array|IMAGECrate|file name, channel)

   Sets the specified image channel (0=red|1=green|2=blue|3=alpha) with
   the provided image values. If the optional ChipsId or image id
   string is not supplied, the routine will operate on the current image.
   """

   curr_arg_indx = 0
   new_args = []

   # pull off the chips id if it exists
   if isinstance (args[curr_arg_indx], chips.pychips_base.ChipsId):
      new_args.append (args[curr_arg_indx])
      curr_arg_indx += 1

   # if there are three arguments then the first argument must be 
   # a string id
   elif len (args[curr_arg_indx:]) == 3:
      if isinstance (args[curr_arg_indx], str):
         new_args.append (args[curr_arg_indx])
         curr_arg_indx += 1
      else:
         raise AttributeError("chips ERROR: Unexpect parameter type for first argument");

   # if there are only two arguments assume that it is file name
   # and channel
   if len (args[curr_arg_indx:]) == 2 and __utils__.check_seq(args[curr_arg_indx]) == False:
      image_crate = None
      if isinstance (args[curr_arg_indx], str):
         image_crate = IMAGECrate(args[curr_arg_indx], 'r')
      elif isinstance (args[curr_arg_indx], IMAGECrate) :
         image_crate = args[curr_arg_indx]
      else:
         raise AttributeError("chips ERROR: Unexpect parameter type for argument %d. Expected file name or image crate"%(curr_arg_indx+1))

      shape = image_crate.get_shape ()
      new_args.append (image_crate.get_image ().values)
      new_args.append (shape[len(shape)-1])
      new_args.append (shape[len(shape)-2])
      curr_arg_indx += 1

   # pull off the rest of the list
   for ii in args[curr_arg_indx:]:
      new_args.append (ii)

   new_args = tuple (new_args)

   chips.set_image_channel (*new_args)


# overloaded contour functions
def add_contour (*args):
   """
   add_contour([ChipsId,] image_array, x-dim, y-dim [,levels] [,transform] [,attribute_list])
   add_contour([ChipsId,] x-indices, y-indices, image_array [,attribute_list])
   add_contour([ChipsId,] IMAGECrate|file name [,attribute_list)

   Adds a contour from the image provided. Only regularly gridded images are
   supported. When an image crate or file name is provided, the default wcs
   transform is applied automatically. Control over which transform to apply
   can be achieved by adjusting the wcs field in the attribute list
   """
   __add_contour__(*args); 

def __add_contour__ (*args):
   """
   add_contour([ChipsId,] image_array, x-dim, y-dim [,levels] [,transform] [,attribute_list])
   add_contour([ChipsId,] x-indices, y-indices, image_array [,attribute_list])
   add_contour([ChipsId,] IMAGECrate|file name [,attribute_list)

   Adds a contour from the image provided. Only regularly gridded images are
   supported. When an image crate or file name is provided, the default wcs
   transform is applied automatically. Control over which transform to apply
   can be achieved by adjusting the wcs field in the attribute list.
 
   Returns the wcs transform component names used or None
   """
   curr_arg_indx = 0
   new_args = []
   img = None
   xname = None
   yname = None

   if (len (args) > 0):
      # Check if the first arg is a chips id
      if isinstance (args[curr_arg_indx], chips.pychips_base.ChipsId):
         new_args.append (args[curr_arg_indx])
         curr_arg_indx = curr_arg_indx + 1

   if (len (args) > curr_arg_indx):
      # Check if the current arg is a file name 
      if isinstance (args[curr_arg_indx], str):
         img = IMAGECrate (args[curr_arg_indx], 'r')
         curr_arg_indx = curr_arg_indx + 1
      # Check if the current arg is an image crate
      elif (isinstance (args[curr_arg_indx], IMAGECrate)):
         img = args[curr_arg_indx];
         curr_arg_indx = curr_arg_indx + 1

   # format the arg list so that attribute structures are converted to
   # strings if they exist
   args = __utils__.format_add_argslist (0,False,__utils__.get_object_type(ChipsContour()),*args)
   
   if (img):

      transform = None
      trans_name = None

      # add crate information and original entries into new arg list
      new_args.append (img.get_image ().values)

      shape = img.get_shape ()
      new_args.append (shape[len(shape)-1])
      new_args.append (shape[len(shape)-2])

      # pull the last element from the arglist. Ingore if there's only one
      # element
      if len(args) > curr_arg_indx:
         arg_end = args[len(args)-1]
      else:
         arg_end = None


      for i in range (curr_arg_indx,len(args)-1):
         new_args.append (args[i])
         curr_arg_indx = curr_arg_indx + 1


      # if the last argument is a string parse it for wcs and remove the wcs
      # entry from the list
      if (type (arg_end) == str):
         (trans_name, arg_end) = __utils__.pop_attrlist ('wcs', arg_end)

         # if the name exists try to pull it from the crate
         if trans_name:
            # if the transform name was passed into the function use it
            try:
               transform = get_axis_transform (img, trans_name)
            except:
               transform = get_transform (img, trans_name)
      elif arg_end != None:
         # arg end wasn't an attribute list put it back on the argument
         # list before the transforms get added
         new_args.append (arg_end)
         arg_end = None


      # if the transform has not been set try each of the default 
      # transform categrories
      if not trans_name:
         try:
            transform = get_axis_transform (img, 'WORLD')
         except:
            try:
               transform = get_axis_transform (img, 'PHYSICAL')
            except:
               try:
                  transform = get_axis_transform (img, 'LOGICAL')
               except:
                  pass
         else:
            trans_name = 'WORLD'

      #If we're using a world translation, check if it's a supported type
      #Warn and fall back to physical if it's not
      #Also change trans_name so we don't accidentally concat later
      if trans_name and trans_name.upper() == 'WORLD':
         if not _legal_world_transform(transform):
            ctype_values = _get_ctype_values(transform)
            try:
               transform = get_axis_transform (img, 'PHYSICAL')
               trans_name = 'PHYSICAL'
               print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using physical transform.".format(ctype_values)
               #raise AttributeError("chips ERROR: '"+table.get_filename () + "' contains < 2 columns")
            except LookupError:
               try:
                  transform = get_axis_transform (img, 'LOGICAL')
                  trans_name = 'LOGICAL'
                  print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using logical transform.".format(ctype_values)
                  if not transform:  #If there's no actual transform (and there won't be), Create an identity-matrix 2D transform so this won't have the plot's default transform applied to it. 
                     transform = LINEAR2DTransform()
                     for parameter in transform.get_parameter_list(): #Initialize parameters
                        if parameter.get_name() == "SCALE":
                           parameter.set_value(numpy.array([1, 1]))
                        if parameter.get_name() == "ROTATION":
                           parameter.set_value(numpy.float64(0))
                        if parameter.get_name() == "OFFSET":
                           parameter.set_value(numpy.array([0, 0]))
               except LookupError:
                  #We currently can't reach this code - Logical in get_axis_transform returns a None.
                  #Makes sense since we don't actually need to change the object at all to do a logical transform.
                  #Keeping this here for futureproofing.
                  transform = None
                  trans_name = None
                  print "chips WARNING: Unable to use world transform - {0} is an unsupported type. Using base coordinates.".format(ctype_values)

      #concatinate the physical and world coordinate systems
      if trans_name and ((trans_name.upper() == 'WORLD') or (trans_name.upper() == 'EQPOS')):
         phys_trans = get_axis_transform (img, 'PHYSICAL')
         if (phys_trans):
            transform = _concat_wcs_transform_ (phys_trans, transform)

      if transform:
         new_args.append (transform)

         # get wcs component names for axes 
         try:
            tmpax=img.get_axis(transform.get_name())
            if tmpax != None:
               axis_names = tmpax.get_cptslist ()
               try:
                  xname = axis_names[0]
                  yname = axis_names[1]
               except:
                  pass
         except KeyError:
            pass #We should only hit this if we made up a logical transform 

      if (arg_end):
         new_args.append (arg_end)

      # handle meta data for contour of images 
      if img: 
         tmstr = img.get_filename()
         if isinstance (new_args[len(new_args)-1], str):
            if tmstr.find('"') != -1:
               new_args[len(new_args)-1] = new_args[len(new_args)-1] + " meta=" + repr(tmstr)
            else:
               new_args[len(new_args)-1] += " meta=\""
               new_args[len(new_args)-1] += tmstr
               new_args[len(new_args)-1] += "\""
         else:
            if tmstr.find('"') != -1:
               new_args.append(" meta="+repr(tmstr))
            else:
               new_args.append(" meta=\""+tmstr+"\"")

      new_args = tuple (new_args)
      chips.add_contour (*new_args)
   else:
      chips.add_contour (*args)

   return xname, yname

def _get_ctype_values(transform):
   """Returns the values in CTYPE concatenated into a single string, seperated by commas. Returns empty string if CTYPE is not found"""
   str_list = []
   for parameter in transform.get_parameter_list():
      if parameter.get_name() == "CTYPE":
         for value in parameter.get_value():
            str_list.append(value)
   return ", ".join(str_list)         


def _legal_world_transform (world_transform):
   ctype_values = _get_ctype_values(world_transform)
   if ctype_values == "RA---TAN, DEC--TAN":
      return True
   else:
      return False

def _get_vector_components (cratedata, return_list=None):

   if return_list == None:
      return_list = []

   # search through each component of a vector column
   cptslist = cratedata.get_cptslist ()

   # if the cratedata has no component it is the component
   # add it to the return list. otherwise recursively search
   # for all component vectors
   if cptslist is None:
     return_list.append (cratedata)
   else:
      for name in cptslist:
         col = cratedata.get_vector_component (name)
         _get_vector_components (col, return_list)
   return return_list     
       

def _copy_cratedata_to_array (cratedata, tnull_list=None):
   crate_vals = None

   # copy the data crate array as a float so that nan values can
   # be set if necessary. Only copy those arrays which have tnulls
   # which will be processed further
   if cratedata.get_nullval () != None:
      crate_vals = cratedata.values.astype (float)
      if tnull_list != None:
         tnull_list.append ((cratedata,crate_vals))
   else:
      crate_vals = cratedata.values
   return crate_vals


def add_curve (*args):

   """
   add_curve([ChipsId,] x, y [,error_list] [,limits_list] [,attributes])'
   add_curve([ChipsId,] TABLECrate|file name, [, attributes])'

   adds a curve with optional error bars and/or upper/lower limits. Error
   list may be None, a single array without list brackets, or from one to
   four arrays enclosed in list brackets. Limits_list may be None a single
   array without list brackets, or one or two arrays enclosed in brackets.
   CrateData columns may be used interchangably with arrays.

   Alternatively a TABLECrate or file name may be passed as an argument.
   In these cases the first 2 columns of the file will be used as x and
   y. The 3rd and 4th columns (if they exist) will be used as y errors
   """
   __add_curve__(*args)

def __add_curve__(*args):

   """
   __add_curve__([ChipsId,] x, y [,error_list] [,limits_list] [,attributes])'
   __add_curve__([ChipsId,] TABLECrate|file name, [, attributes])'

   adds a curve with optional error bars and/or upper/lower limits. Error
   list may be None, a single array without list brackets, or from one to
   four arrays enclosed in list brackets. Limits_list may be None a single
   array without list brackets, or one or two arrays enclosed in brackets.
   CrateData columns may be used interchangably with arrays.

   Alternatively a TABLECrate or file name may be passed as an argument.
   In these cases the first 2 columns of the file will be used as x and
   y. The 3rd and 4th columns (if they exist) will be used as y errors 
   """

   max_cols = 4
   curr_arg_indx = 0
   new_args = []
   table = None
   xname = None
   yname = None

   attribute_list = None

   if (len (args) > 0):
      # Check if the first arg is a chips id
      if isinstance (args[curr_arg_indx], chips.pychips_base.ChipsId):
         new_args.append (args[curr_arg_indx])
         curr_arg_indx = curr_arg_indx + 1

   if (len (args) > curr_arg_indx):
      table = None
      # Check if the current arg is a file name
      if isinstance (args[curr_arg_indx], str):
         table = TABLECrate (args[curr_arg_indx], 'r')
      elif isinstance (args[curr_arg_indx], TABLECrate):
         table = args[curr_arg_indx]
      if table:

         # format the argument list so that if an attribute list exists it
         # will be a string this is necessary for adding meta data -- this
         # could be done at the start of the function; however, if the user
         # forgot a file name and only had an attribute list then the
         # attribute list will be treated as a file name, which will cause 
         # an error but an incorrect error
         args = __utils__.format_add_argslist (0,False,__utils__.get_object_type(chips.ChipsCurve()),*args)

         curr_arg_indx = curr_arg_indx + 1


         # get column names. retrieve only component names and no virtual column
         colnames=get_col_names(table, False, True)
         ncols=len(colnames)

         # less than 2 cols in the file
         if ( ncols < 2 ):
            raise AttributeError("chips ERROR: '"+table.get_filename () + "' contains < 2 columns")

         if ( ncols > max_cols ) :
            ncols = max_cols;

         # put the first 2 table columns into the new arg list
         for indx in range(2) :
            new_args.append(table.get_column(colnames[indx]))
         xname = __get_axis_label__(table.get_column(colnames[0])) 
         yname = __get_axis_label__(table.get_column(colnames[1])) 

         # put the other two columns into a list to ensure that they are
         # treated as errors
         err_list = []
         for indx in range (max_cols-2,ncols):
            err_list.append(table.get_column(colnames[indx]))

         if len(err_list) > 0:
            new_args.append (err_list)

         #put the rest of the original args list into the new arg list
         for indx in range (curr_arg_indx,len(args)):
            new_args.append (args[indx])

         # handle meta data
         if table != None:
            tmstr = table.get_filename(); 
            if isinstance (new_args[len(new_args)-1], str):
               if tmstr.find('"') != -1:
                  new_args[len(new_args)-1] = new_args[len(new_args)-1] + " meta=" + repr(tmstr)
               else:
                  new_args[len(new_args)-1] += " meta=\""
                  new_args[len(new_args)-1] += tmstr
                  new_args[len(new_args)-1] += "\""
            else:
               if tmstr.find('"') != -1:
                  new_args.append(" meta="+repr(tmstr))
               else:
                  new_args.append(" meta=\""+tmstr+"\"")

         # recursively parse the rest of this list
         tmpxname = None
         tmpyname = None
         (tmpxname, tmpyname) = __add_curve__(*new_args)
         if tmpxname != None:
            xname = tmpxname
         if tmpyname != None:
            yname = tmpyname

         return  xname, yname

      # go through the list and convert CrateData and lists of CrateData
      # to an array.
      tmpcolcnt = 0
      cols = []
      for indx in range (curr_arg_indx,len(args)):

         if isinstance (args[indx], CrateData):

            # this will return a list of component columns even if the
            # cratedata is not a vector colum 
            cratedata_list = _get_vector_components (args[indx]);

            for cratedata in cratedata_list:
               new_args.append (_copy_cratedata_to_array (cratedata, cols))
               if tmpcolcnt == 0:
                  xname = __get_axis_label__(cratedata) 
                  tmpcolcnt += 1
               elif tmpcolcnt == 1:
                  yname = __get_axis_label__(cratedata) 
                  tmpcolcnt += 1
         elif __utils__.check_seq (args[indx]):
            crate_list = True
            tmp_args = []

            for arg in args[indx]:
               # Only look at lists of crate data and lists of list
               # lists of any thing else don't deal with for now. This
               # keeps us from having to go through an entire data column
               # each of which is actually a list
               if not isinstance (arg, CrateData) and not __utils__.check_seq (arg) and arg != None: 
                  crate_list = False
                  break;

            if crate_list:
               for arg in args[indx]:
                  if isinstance (arg, CrateData):
                      # this will return a list of component columns even if the
                      # cratedata is not a vector column
                      cratedata_list = _get_vector_components (arg);

                      for cratedata in cratedata_list:
                         tmp_args.append (_copy_cratedata_to_array (cratedata, cols))
                         if tmpcolcnt == 0:
                            xname = __get_axis_label__(cratedata) 
                            tmpcolcnt += 1
                         elif tmpcolcnt == 1:
                            yname = __get_axis_label__(cratedata) 
                            tmpcolcnt += 1
                  else:
                     tmp_args.append (arg)
            else:
               tmp_args = args[indx]

            new_args.append (tmp_args)

         else:
            new_args.append (args[indx])

      for col in cols:
         crate_data = col[0]
         col_values = col[1]
         tnull = crate_data.get_nullval ()
         if tnull is not None:
            elemIndx = numpy.where (crate_data.values == tnull)
            col_values[elemIndx] = numpy.NaN 

#   # handle meta data
#   if table != None:
#      if isinstance (new_args[len(new_args)-1], str):
#         new_args[len(new_args)-1] += " meta=\""
#         new_args[len(new_args)-1] += table.get_filename()
#         new_args[len(new_args)-1] += "\""
#      else:
#         new_args.append("meta=\""+table.get_filename()+"\"")

   new_args = tuple (new_args)
   chips.add_curve (*new_args)

   return xname, yname



def add_histogram (*args):

   """
   add_histogram([ChipsId,] counts [,attribute_list])
   add_histogram([ChipsId,] bins, counts [,attribute_list])
   add_histogram([ChipsId,] bin_lo, bin_hi, counts [,yerrs] [,attribute_list])
   add_histogram([ChipsId,] bin_lo, bin_hi, ydown, yup, counts [,attribute_list])
   add_histogram ([ChipsId,] TABLECrate|file name [,attribute_list])
   add_histogram ([ChipsId,] IMAGECrate|file name [,attribute_list])

   Adds a histogram using the provided data, creating window, frame, or plot
   as needed. Crate Data columns may be used interchangeably with any array
   argument. When passing in a TABLECrate or file name the first 2 columns are
   used as bins and counts. If the 1st column is an array it will be used 
   as bin_lo and bin_hi. When passing an IMAGECrate, the image must be 1D or
   only have 1 dimension with size > 1.
   """

   # initialize converted argument list
   new_args=[]
   argc = len(args)
   pos = 0
   ncols=0

   # determine what arguments have been passed

   if (argc > 0):

      # check for ChipsId
      if (isinstance(args[pos], chips.ChipsId)):
         new_args.append(args[pos])
         pos+=1

      # check if arg is a file spec
      table = None
      image = None
      if ( isinstance(args[pos], str)):
         try:
            table=TABLECrate(args[pos], 'r')
         except TypeError as e:
            # try as an image
            try:
               image=IMAGECrate(args[pos])
            except:
               table = None
               image = None
      elif isinstance (args[pos], TABLECrate):
         table = args[pos]
      elif isinstance (args[pos], IMAGECrate):
         image = args[pos]
         
      if (image):
         shape = image.get_shape()
         dimcnt = len (filter (lambda x: x > 1, shape))
         if dimcnt != 1:
            raise AttributeError("chips ERROR: Supplied image must be a 1D image or slice- there were %d dimensions with length greater than 1." % mycnt)
            return

      # format the argument list so that if an attribute list exists it
      # will be a string this is necessary for adding meta data -- this
      # could be done at the start of the function; however, if the user
      # forgot a file name and only had an attribute list then the
      # attribute list will be treated as a file name, which will cause
      # an error but an incorrect error
      args = __utils__.format_add_argslist (0,False,__utils__.get_object_type(chips.ChipsHistogram()),*args)


      if table:

         pos+=1

         col_names = get_col_names (table, False, True)
         ncols=len (col_names)

         # less than 2 cols in the file
         if ( ncols < 2 ):
            raise AttributeError("chips ERROR: %s must contain two columns" % args[pos-1])
         else :
            mybins=table.get_column(col_names[0])
            if mybins.get_size() == 2:
               binlo,binhi=numpy.column_stack(mybins.values)
               new_args.append(binlo)
               new_args.append(binhi)
            else:
               new_args.append(mybins.values)

            cnts = table.get_column(1).values
            new_args.append(cnts)

      elif image:

         pos+=1
        
         cnts = image.get_image ().values
         new_args.append(cnts)

      elif ((argc-pos)>0):
         tmp_args = []
         for rr in range(pos, argc):
            if (isinstance(args[rr], CrateData)):
               _get_vector_components (args[rr], tmp_args);
            else:
               tmp_args.append (args[rr])

         if (len (tmp_args) > 6):
            raise AttributeError("chips ERROR: expected at most 5 data objects in the argument list.")
         else:
            for arg in tmp_args:
               if (isinstance(arg, CrateData)):
                  new_args.append(arg.values)
                  pos+=1;
               else:
                  new_args.append(arg)
                  pos+=1

#      # copy remaining parameters into new argument list
      for i in range(pos, argc) :
            new_args.append(args[i])

      # handle meta data
      if table != None:
         tmstr = table.get_filename()
         if isinstance (new_args[len(new_args)-1], str):
            if tmstr.find('"') != -1:
               new_args[len(new_args)-1] = new_args[len(new_args)-1] + " meta=" + repr(tmstr)
            else:
               new_args[len(new_args)-1] += " meta=\""
               new_args[len(new_args)-1] += tmstr
               new_args[len(new_args)-1] += "\""
         else:
            if tmstr.find('"') != -1:
               new_args.append(" meta="+repr(tmstr))
            else:
               new_args.append(" meta=\""+tmstr+"\"")
      elif image != None:
         tmstr = image.get_filename()
         if isinstance (new_args[len(new_args)-1], str):
            if tmstr.find('"') != -1:
               new_args[len(new_args)-1] = new_args[len(new_args)-1] + " meta=" + repr(tmstr)
            else:
               new_args[len(new_args)-1] += " meta=\""
               new_args[len(new_args)-1] += tmstr
               new_args[len(new_args)-1] += "\""
         else:
            if tmstr.find('"') != -1:
               new_args.append(" meta="+repr(tmstr))
            else:
               new_args.append(" meta=\""+tmstr+"\"")

      chips.add_histogram (*new_args)


def make_figure(arg1=None, arg2=None, arg3=None, arg4=None):
   """
   make_figure ([id] Crate|file name [, figure type] [, attribute list])

   creates a new curve, contour, histogram or image from the data provided. It 
   clears and uses an existing plot area when possible and will create
   one if none exist. Axis labels and plot title will be filled in from
   the data if possible. By default a curve or contour will be created
   depending on the data. This can be overriden by setting the figure type
   to curve, contour histogram or image.
   
   """
   chips.pychips_base.lock()
   cur_inf = chips.info_current()
   if cur_inf == None:
      clearp = False
   elif (cur_inf.count('   Plot [') == 0):
      clearp = False
   elif (cur_inf.count('   Plot [') == 1):
      # need to clear out the plot
      clearp = True
   else:
      chips.pychips_base.unlock()
      raise AttributeError("chips ERROR: can not use make_figure when more than a single plot is current.")

   # track number of windows
   win_count = chips.pychips_base.get_object_count(chips.pychips_base.chips_window)
  
   # id       = ChipsId object to override currency
   # fname    = user specified file name
   # mycrate  = crate containing the appropriate data
   # mytype   = type of figure - chips_curve|contour|histogram
   # atts     = user specified attribute settings

   mycrate = None

   mytype = None
   atts = None
   if (isinstance(arg1, chips.ChipsId)):
      id = arg1
      if isinstance(arg2, str):
         fname = arg2
         if __utils__.isinstance_int(arg3):
	    if arg3 == chips.pychips_base.chips_curve or arg3 == chips.pychips_base.chips_contour or arg3 == chips.pychips_base.chips_histogram or arg3 == chips.pychips_base.chips_image:
	       mytype = arg3
	    else:
	       chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: To specify figure type, please use either a string or enumeration.")
	    atts = arg4
         elif isinstance(arg3, str):
	    tmp = arg3.lower()
	    if tmp == "curve":
	       mytype = chips.pychips_base.chips_curve
	    elif tmp == "contour":
	       mytype = chips.pychips_base.chips_contour
	    elif tmp == "histogram":
	       mytype = chips.pychips_base.chips_histogram
            elif tmp == "image":
               mytype = chips.pychips_base.chips_image
            elif (tmp.count('=') != 0):
	       # treat as attribute list
	       atts = arg3
	       mytype = None
	    else:
	       chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: expected third argument to either specify figure type (curve|contour|histogram|image) or be an attribute list")
         else:
	    # treat as attribute list
	    atts = arg3

	 # load crate if necessary
	 if (len(fname) != 0):
            if mytype == chips.pychips_base.chips_contour or mytype == chips.pychips_base.chips_image:
	       mycrate = IMAGECrate(fname,'r')
            elif mytype == chips.pychips_base.chips_curve:
	       mycrate = TABLECrate(fname,'r')
            elif mytype == chips.pychips_base.chips_histogram:
               try:
                  mycrate = TABLECrate(fname, 'r')
               except TypeError as e:
                  mycrate=IMAGECrate(fname, 'r')
            else:
               try:
                  mycrate = TABLECrate(fname,'r')
                  mytype = chips.pychips_base.chips_curve
               except TypeError as e:
                  mycrate = IMAGECrate(fname,'r')
               
                  shape = mycrate.get_shape()
                  dimcnt = len (filter (lambda x: x > 1, shape))
                  if len(shape) ==1 or dimcnt == 1:
                     mytype = chips.pychips_base.chips_histogram
                  else:
                     mytype = chips.pychips_base.chips_contour

      elif _is_crate_(arg2):
         mycrate = arg2
         fname = mycrate.get_filename()

         if __utils__.isinstance_int(arg3):
            if arg3 == chips.pychips_base.chips_curve or \
                arg3 == chips.pychips_base.chips_contour or \
                arg3 == chips.pychips_base.chips_histogram or \
                arg3 == chips.pychips_base.chips_image:
                mytype = arg3
                atts = arg4
            else:
                chips.pychips_base.unlock()
                raise AttributeError("chips ERROR: To specify figure type, please use either a string or enumeration.")
         elif isinstance(arg3, str):
            tmp = arg3.lower()
            if tmp == "curve":
               mytype = chips.pychips_base.chips_curve
               atts = arg4
            elif tmp == "contour":
               mytype = chips.pychips_base.chips_contour
               atts = arg4
            elif tmp == "histogram":
               mytype = chips.pychips_base.chips_histogram
               atts = arg4
            elif tmp == "image":
               mytype = chips.pychips_base.chips_image
               atts = arg4
            elif (tmp.count('=') != 0):
               # treat as attribute list
               atts = arg3
               mytype = None
            else:
               chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: expected third argument to either specify figure type (curve|contour|histogram|image) or be an attribute list")
         else:
            # treat as attribute list
            atts = arg3

      else:
	 chips.pychips_base.unlock()
         raise AttributeError("chips ERROR: expected a filename or Crate object after the ChipsId parameter.")
   else:
      id = chips.ChipsId()
      id.clear()
      if isinstance(arg1, str):
         fname = arg1
         if __utils__.isinstance_int(arg2):
	    if arg2 == chips.pychips_base.chips_curve or \
                arg2 == chips.pychips_base.chips_contour or \
                arg2 == chips.pychips_base.chips_histogram or \
                arg2 == chips.pychips_base.chips_image:
	       mytype = arg2
	    else:
	       chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: To specify figure type, please use either a string or enumeration.")
	    atts = arg3
         elif isinstance(arg2, str):
	    tmp = arg2.lower()
	    if tmp == "curve":
	       mytype = chips.pychips_base.chips_curve
	       atts = arg3
	    elif tmp == "contour":
	       mytype = chips.pychips_base.chips_contour
	       atts = arg3
	    elif tmp == "histogram":
	       mytype = chips.pychips_base.chips_histogram
	       atts = arg3
            elif tmp == "image":
               mytype = chips.pychips_base.chips_image
               atts = arg3
            elif (tmp.count('=') != 0):
	       # treat as attribute list
	       atts = arg2
	       mytype = None
	    else:
	       chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: expected third argument to either specify figure type (curve|contour|histogram) or be an attribute list")
         else:
	    # treat as attribute list
            atts = arg2 

            
	 # load crate if necessary
	 if (len(fname) != 0):
            if mytype == chips.pychips_base.chips_contour or mytype == chips.pychips_base.chips_image:
	       mycrate = IMAGECrate(fname,'r')
            elif mytype == chips.pychips_base.chips_curve:
	       mycrate = TABLECrate(fname,'r')
            elif mytype == chips.pychips_base.chips_histogram:
               try:
	          mycrate = TABLECrate(fname,'r')
               except TypeError as e:
                  # try as an image
                  mycrate=IMAGECrate(fname,'r')
            else:
               try:
                  mycrate = TABLECrate(fname,'r')
                  mytype = chips.pychips_base.chips_curve
               except TypeError as e:
                  mycrate = IMAGECrate(fname,'r')
                  
                  shape = mycrate.get_shape()
                  dimcnt = len (filter (lambda x: x > 1, shape))
                  if len(shape) ==1 or dimcnt == 1:
                     mytype = chips.pychips_base.chips_histogram
                  else:
                     mytype = chips.pychips_base.chips_contour
      elif _is_crate_(arg1):
         mycrate = arg1
	 fname = mycrate.get_filename ()
         if __utils__.isinstance_int(arg2):
            if arg2 == chips.pychips_base.chips_curve or \
                arg2 == chips.pychips_base.chips_contour or \
                arg2 == chips.pychips_base.chips_histogram or \
                arg2 == chips.pychips_base.chips_image:
               mytype = arg2
	       atts = arg3
            else:
	       chips.pychips_base.unlock()
	       raise AttributeError("chips ERROR: To specify figure type, please use either a string or enumeration.")
	 elif isinstance(arg2, str):
	    tmp = arg2.lower()
            if tmp == "curve":
               mytype = chips.pychips_base.chips_curve
               atts = arg3
            elif tmp == "contour":
               mytype = chips.pychips_base.chips_contour
               atts = arg3
            elif tmp == "histogram":
               mytype = chips.pychips_base.chips_histogram
               atts = arg3
            elif tmp == "image":
               mytype = chips.pychips_base.chips_image
               atts = arg3
            elif (tmp.count('=') != 0):
               # treat as attribute list
               atts = arg2
               mytype = chips.pychips_base.chips_curve
            else:
               chips.pychips_base.unlock()
               raise AttributeError("chips ERROR: expected third argument to either specify figuretype (curve|contour|histogram|image) or be an attribute list")
         else:
	    # treat as attribute list
            atts = arg2
      else:
	 chips.pychips_base.unlock()
         raise AttributeError("chips ERROR: expected a filename or Crate object after the ChipsId parameter.")

   # verify that we have the correct type of data to use
   if (mytype == chips.pychips_base.chips_contour or mytype == chips.pychips_base.chips_image) and mycrate.is_image() != 1:
      chips.pychips_base.unlock()
      raise AttributeError("chips ERROR: Contouring and imaging require image data.")
   if mytype == chips.pychips_base.chips_curve and mycrate.is_image() == 1: 
      chips.pychips_base.unlock()
      raise AttributeError("chips ERROR: Plotting curves requires table data.")
   if not mytype:
      if mycrate.is_image() != True:
         mytype = chips.pychips_base.chips_curve
      else:
         shape = mycrate.get_shape()
         dimcnt = len (filter (lambda x: x > 1, shape))
         if len(shape) ==1 or dimcnt == 1:
            mytype = chips.pychips_base.chips_histogram
         else:
            mytype = chips.pychips_base.chips_contour
   try:
      chips.pychips_base.open_undo_buffer()

      if clearp:
         chips.clear_plot()

      my_xlab = None
      my_ylab = None

      # set up plot title
      if mycrate.key_exists("OBJECT"):
         titlestr = mycrate.get_key_value("OBJECT")
      elif mycrate.key_exists("TITLE"):
         titlestr = mycrate.get_key_value("TITLE")
      else:
         titlestr = fname;
      # remove path and any dm filter spec
      startpos = titlestr.rfind('/') + 1
      endpos = titlestr.find('[')
      if endpos == -1:
         endpos = len(titlestr)
      titlestr = titlestr[startpos:endpos]
      # want underscores to be underscores and not latex subscripts
      titlestr = titlestr.replace('_','\\_')

      if atts == None:
         atts = ''

      if mytype == chips.pychips_base.chips_contour:
         (my_xlab, my_ylab) = __add_contour__(id, mycrate, atts)
      elif mytype == chips.pychips_base.chips_image:
         (my_xlab, my_ylab) = __add_image__(id, mycrate, atts)
      elif mytype == chips.pychips_base.chips_curve:
         (my_xlab, my_ylab) = __add_curve__(id, mycrate, atts)
      else:
         add_histogram(id, mycrate, atts)
         # add axis labels based on column names/units
         if isinstance (mycrate, TABLECrate):
            mycol0 = mycrate.get_column(0)
            mycol1 = mycrate.get_column(1)
            labstr = mycol0.name
            uns = mycol0.unit
            if uns != None and (len(uns) != 0):
               labstr = labstr + ' (' + uns + ')'
            my_xlab = labstr.replace('_','\\_')
            labstr = mycol1.name
            uns = mycol1.unit
            if uns != None and len(uns) != 0:
               labstr = labstr + ' (' + uns + ')'
            my_ylab = labstr.replace('_','\\_')
         else:
            mycol0 = ""
            mycol1 = ""
	    
      # add plot labels if know
      if isinstance(my_xlab, str):
         chips.set_plot_xlabel(my_xlab)
      if isinstance(my_ylab, str):
         chips.set_plot_ylabel(my_ylab)

      # add plot title if value set
      if len(titlestr) != 0:
         chips.set_plot_title(titlestr)
      # if new window, need to set the redraw 
      if win_count == 0:
         chips.pychips_base.set_window_redraw(__utils__.bool_cast(chips.pychips_base.get_preference('window.display')))

   except:
      chips.pychips_base.discard_undo_buffer()
      chips.pychips_base.unlock()
      raise
   else:
      chips.pychips_base.close_undo_buffer()
      chips.pychips_base.unlock()



def make_script(filename, overwrite=False):
   """
   make_script(filename, overwrite)

   Generates a python script that replicates the current state of the chips
   server. This command will create a script that uses crates so that the
   template may be executed without needing to set the template data variables.
   This routine generates a script with the given file name as well as stores
   the data for curves, contours, histograms, and images in a separate files
   of the same name but with .My(Data|Curve)_# extenstions. Preference settings
   are stored in a file with a .prefs extension.
   """
   if type(filename) != str:
      raise AttributeError("chips ERROR: expected a string filename")

   chips.make_script(filename, overwrite)

   tmpname = filename + "_TmP_CrAtE.Py"
   infile=open(filename)
   outfile=open(tmpname, 'w')
   outfile.write("from pycrates import *     # import crates i/o routines\n")
   done = False
   cname = []
   while not done:
      line = infile.readline()
      if not line: break
      if ((line.find('MyData_') == 0 or line.find('CurveData_') == 0)  and line.find('= DataObj()  # data-') != -1):
         # parse out data obj name and data file
         dobj = line[0:line.find(' =')]
         dfile = line[line.find('data-')+6:]
         dfile = dfile.replace("\n","")
         binary_to_fits(dfile,overwrite)
         cname = [dobj.replace('Data','Crate')]
         outfile.write(cname[0] + " = TABLECrate(\""+dfile.rstrip('.bin')+".fits\")\n")
         line=line.rstrip('.bin\n')+".fits\n"
         outfile.write(line)
      elif ((line.find('MyData_') == 0 or line.find('CurveData_') == 0) and line.find("insert data array here") != -1):
         colname = line[line.find('.')+1:line.find(' ')]
         repstr = cname[0] + ".get_column(\"" + colname + "\").values"
         line = line.replace("#  insert data array here", repstr)
         outfile.write(line)
      elif ((line.find('MyData_') == 0 or line.find('ImageData_') == 0)  and line.find('= DataObj()  # image data-') != -1):
         # parse out data obj name and data file
         dobj = line[0:line.find(' =')]
         dfile = line[line.find('data-')+6:]
         dfile = dfile.replace("\n","")
         dfile_list = dfile.split (",")

         # images can have multiple files associated with them make a crate for
         # each file
         cname = []
         ii = 0
         for dfile in dfile_list:
            binary_to_fits(dfile,overwrite)
            tmp_cname = dobj.replace('Data','Crate')
            if len (dfile_list) > 1:
               tmp_cname += ("_"+str(ii))
            cname.append (tmp_cname)
            outfile.write(tmp_cname + " = IMAGECrate(\""+dfile.rstrip('.bin')+".fits\")\n")
            ii += 1
         line=line.rstrip('.bin\n')+".fits\n"
         outfile.write(line)
         
      elif ((line.find('MyData_') == 0 or line.find('ImageData_') == 0) and line.find("insert image array here") != -1):
         colname = line[line.find('.')+1:line.find(' ')]
         if len(cname) > 0:
            repstr = cname[0] + ".get_image().values"
            del (cname[0])
         else:
            raise RuntimeError ("chips ERROR: Unexpected number of channels in image template.")
         line = line.replace("#  insert image array here", repstr)
         outfile.write(line)

      else:
         outfile.write(line)
   infile.close()
   outfile.close()
   os.remove(filename)
   os.rename(tmpname, filename)

def binary_to_fits(infile, overwrite=False):
   """
     binary_2_fits(infile)
   """

   f = open(infile,"rb")
   outfile = infile.rstrip(".bin")+".fits"

   hdr = array.array('l')
   hdr.read(f, 4)
   # image header                          table header  
   # [     1     ][     2     ]        [     0     ][   ncols   ] 
   # [  ax1 len  ][  ax2 len  ]        [   nrows   ][colname sz ] 

   if hdr[0] == 0:   # Table Data 
      cnames = array.array('c')
      cnames.read(f, hdr[3] * struct.calcsize('d'))
      cols = cnames.tostring()
      cols=cols.split('\0')

      cr = TABLECrate(mode='rw')
      for ll in range(0, hdr[1]):
         data = array.array('d')
         data.read(f, hdr[2])
         darr = numpy.array(data)
         cd = CrateData()
         cd.name = cols[ll]
         cd.values = darr
         cr.add_column(cd)
      write_file(cr, outfile, overwrite)

      # workaround for Crate bug #12940
      cr.__del__ ()

   elif hdr[0] == 1:   # Image Data
      sz = hdr[2] * hdr[3] 
      data = array.array('d')
      data.read(f, sz)
      darr = numpy.array(data)
      f.close()
      cd = CrateData()
      img = darr.reshape(1, hdr[2], hdr[3])
      cd.values = img
      cr = IMAGECrate(mode='rw')
      add_piximg(cr, cd)
      write_file(cr, outfile, overwrite)
   if os.access(infile,os.W_OK) and os.access(outfile,os.R_OK) and infile.endswith(".bin"):
      os.remove(infile)

      # workaround for Crate bug #12940
      cr.__del__ ()

