# 
#  Copyright (C) 2007,2009,2013  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# Utility functions used to help create higher level user interface functions

from pychips import attribute_objects as attr_objs
import pychips_base as __pychips_base__
import shlex

import numpy
import itertools



def get_object_type (obj):
   try:
      name = obj.__class__.__name__
   except:
      return None
   try:
      return attr_objs._object_types[name]
   except:
      return name

def convert_orientation(val):
   'Converts a horizontal or vertical offset string to enumeration 0 and 1'
   if isinstance(val, str):
      low = val.lower()
      if low == 'horizontal' or low == '0' or low == '4':
         val = 0
      elif low == 'vertical' or low == '1' or low == '5':
         val = 1
      else:
         val = None
   elif val == 4: 
      val = 0
   elif val == 5:
      val = 1
   elif val != 0 and val != 1:
      val = None

   return val

def convert_offset(val):
   'Converts a horizontal or vertical offset string to a float'
   if isinstance_numeric(val):
      return val
   elif isinstance(val, str):
      low = val.lower()
      if low == 'left':
         val = 0.0
      elif low == 'center' or low == 'base':
         val = 0.5
      elif low == 'right' or low == 'top':
         val = 1.0
      elif low == 'auto':
         val = __pychips_base__.AUTO_ALIGN
      else:
         val = float(low)
      return val
   else:
      return None

def convert_depth(val):
   'Converts a string depth value to an integer'
   if isinstance_int(val):
      return val
   elif isinstance(val, str):
      low = val.lower()
      if low == 'default':
         val = __pychips_base__.V_USE_DEFAULT_DEPTH
      else:
         val = int(low)
      return val
   elif isinstance_float(val):
      val = int(val)
      return val
   else:
      return None

def bool_cast(val):
   'Converts a string (true|False|on|OFF|etc...)  to a boolean value'
   if isinstance(val, str):
      # since built in bool() only returns false for empty strings
      vlo = val.lower()
      if vlo == 'false' or vlo == 'off' or vlo == 'no' or vlo == '0' or \
         vlo == 'f' or vlo == 'n':
         return(False)
      elif vlo == 'true' or vlo == 'on' or vlo == 'yes' or vlo == '1' or \
         vlo == 't' or vlo == 'y':
         return(True)
      else:
         return(None)
   else:
      # use built in bool cast
      return bool(val)

def check_seq  (s):
   try: 
      iter(s) 
   except: ret = False
   else:
      ret = (isinstance(s, str) == 0)
   return ret;

def isinstance_numeric (n):
   return isinstance_int (n) or isinstance_float (n)

def isinstance_int (n):
   'This replaces isinstance (int) to be more generic'

   if isinstance (n, int):
      return True
   try:
      if n + 0 != int (n):
         return False
   except:
      return False

   return True 

def isinstance_float (n):
   'This replaces isinstance (float) to be more generic'

   if isinstance (n, float):
      return True
   try:
      # if the number is an int then it's not a float
      if isinstance_int (n):
         return False
      if n + 0 != float (n):
         return False
   except:
      return False

   return True

def array_length (a):
   a_length = 0

   # try to cast a the length of a 2d array to a 1d array
   # this assumes that user is using either 2D or 1D lists
   # anything else wont be caught by this check
   if isinstance (a, numpy.ndarray):
      a_length = 1
      for ii in a.shape:
         a_length *= ii
   else:
      try:
         a_length = len (list ( itertools.chain (*a)))
      except:
         a_length = len(a)
   return a_length


def is_object_list(inp):
   'Returns whether a list is composed of base chips objects'
   if check_seq (inp) and not isinstance (inp, dict):
      for elem in inp:
         if isinstance(elem, attr_objs.ChipsAttributeObject) == False:
            return False
      return True
   else:
      return False

def convert2attlist(inp, prefix=''):
   'Converts a structure or list to the string "a=v1 b=v2" format used by chips'
   t = ''

   # return if the input is already a string
   if isinstance(inp, str):
      return inp

   if type(inp) == dict:
      for attr,val in inp.iteritems ():
         if isinstance(attr,str):
            xtyp = type(val)
            if isinstance(val, str):
               xtyp = str
            if isinstance_int(val):
               xtyp = int
            if isinstance_float(val):
               xtyp = float
            if check_seq (val):
               xtyp = list
            split_attr = attr.split ('.')
            if xtyp == int and (('color' in split_attr) \
                or  ('bgcolor' in split_attr)) :
               # color stored as an int- output 6 digit hex string (rrggbb)
               t = t + prefix + attr + '=' + '%.6x'%val + ' '
            elif isinstance_numeric(val):
               t = t + prefix + attr + '=' + '%s'%val + ' '
            elif xtyp == str:
               if val.find(' ') != -1:
                  if val.find("'") == 0 and val.rfind("'") + 1 == len(val):
                     t = t + prefix + attr + '=%s'%val + ' '
                  elif val.find("\"") == 0 and val.rfind("\"") + 1 == len(val):
                     t = t + prefix + attr + '=%s'%val + ' '
                  else:
                     t = t + prefix + attr + '=' + repr(val) + ' '
               else:
                  t = t + prefix + attr + '=%s'%val + ' '
            elif xtyp == bool:
               if attr == True:
                  t = t + prefix + attr + '=true '
               else:
                  t = t + prefix + attr + '=false '
            elif xtyp == list:
               list_attr = '['
               delim = ''
               for subelem in val:
                  list_attr = list_attr + delim + "%s"%subelem
                  delim = ','
               list_attr = list_attr + '] '
               t = t + prefix + attr + '=' + list_attr
         else:
            raise AttributeError("chips ERROR: Expected string attribute name in list.")
            return
      return t

   if check_seq (inp):
      # convert list to string

      # handle list of objects
      if is_object_list(inp):
         for elem in range(0, len(inp)):
            t = t + convert2attlist(inp[elem])
         return t
      elif (len(inp) % 2) == 1:
         raise AttributeError("chips ERROR: Expected even number of entries (attr-value) pairs.")
         return None
      else:
         for elem in range(0, len(inp), 2):
            if isinstance(inp[elem], str):
               xtyp = type(inp[elem+1])
               if isinstance(inp[elem+1], str):
                  xtyp = str
               if isinstance_int(inp[elem+1]):
                  xtyp = int
               if isinstance_float(inp[elem+1]):
                  xtyp = float
               if check_seq (inp[elem+1]):
                  xtyp = list
               split_attr = inp[elem].split ('.')
               if xtyp == int and (('color' in split_attr) or ('bgcolor' in split_attr)):
                  # color stored as an int- output 6 digit hex string (rrggbb)
                  t = t + prefix + inp[elem] + '=' + '%.6x'%inp[elem+1] + ' '
               elif xtyp == int or xtyp == float:
                  t = t + prefix + inp[elem] + '=' + '%s'%inp[elem+1] + ' '
               elif xtyp == str:
                  if inp[elem+1].find(' ') != -1:
                     if inp[elem+1].find("'") == 0 and inp[elem+1].rfind("'") + 1 == len(inp[elem+1]):
                        t = t + prefix + inp[elem] + '=%s'%inp[elem+1] + ' '
                     elif inp[elem+1].find("\"") == 0 and inp[elem+1].rfind("\"") + 1 == len(inp[elem+1]):
                        t = t + prefix + inp[elem] + '=%s'%inp[elem+1] + ' '
                     else:
                        t = t + prefix + inp[elem] + '=' + repr(inp[elem+1]) + ' '
                  else:
                     t = t + prefix + inp[elem] + '=%s'%inp[elem+1] + ' '
               elif xtyp == bool:
                  if inp[elem+1] == True:
                     t = t + prefix + inp[elem] + '=true '
                  else:
                     t = t + prefix + inp[elem] + '=false '
               elif xtyp == list:
                  list_attr = '['
                  delim = ''
                  for subelem in inp[elem+1]:
                     list_attr = list_attr + delim + "%s"%subelem
                     delim = ','
                  list_attr = list_attr + '] '
                  t = t + prefix + inp[elem] + '=' + list_attr
            else:
               raise AttributeError("chips ERROR: Expected string attribute name in list.")
               return
         return t
   else:
      # convert structure to string
      object_type = get_object_type (inp);
      if object_type != None:
         prefix = prefix + object_type
         if (prefix != ''):
            prefix = prefix + '.'

   t = ''
   for x in dir(inp):
      if getattr(inp,x) != None and x[0] != '_':
         xval = getattr(inp, x)
         xtyp = type(getattr(inp, x))
         ## majortick workaround- since resource isn't axis.majortick
         if prefix == 'axis.majortick.':  prefix = 'majortick.'
         if prefix == 'axis.minortick.':  prefix = 'minortick.'

         if isinstance_int (xval) and (x == 'color' or x == 'bgcolor'):
            # color stored as an int- output as a 6 digit hex string (rrggbb)
            t = t + prefix + x + '=' + '%.6x'%getattr(inp,x) + ' '
         elif isinstance(xval, str) or isinstance_numeric (xval):
            t = t + prefix + x + '=' + '%s'%getattr(inp,x) + ' '
         elif xtyp == bool:
            if getattr(inp,x) == True:
               t = t + prefix + x + '=true '
            else:
               t = t + prefix + x + '=false '
         elif isinstance (xval, list):
            t = t + prefix + x + '='

            list_attr = '['
            delim = ''
            for subelem in xval:
               list_attr = list_attr + delim + "%s"%subelem
               delim = ','
            list_attr = list_attr + '] '
            t = t + list_attr
         else:
            t = t + convert2attlist(getattr(inp, x), prefix)
   return t

def pop_attrlist (search_key, attributes):
   'takes a string version of an attribute list removes specified key from it. returns the value of the key if found, and the new attribute list'

   if (type (attributes) != str):
      raise AttributeError ("chips ERROR: attribute list must be in string format")

   attr_list = string_to_attribute_list (attributes)

   is_key = True
   key_found = False
   new_attr_list = ""
   key_value = None
   for index in attr_list:
      # odd values are keywords
      if is_key:
         if index == search_key:
#         if key.find (search_key) > -1:
            key_found = True
         else:
            if new_attr_list == "":
               new_attr_list = index
            else:
               new_attr_list += " " + index
               
      elif key_found:
         key_value = index
         key_found = False
      else:
         new_attr_list = new_attr_list + "=" + index
      is_key = not is_key

   return (key_value, new_attr_list)


def format_add_argslist (min_num_strings, check_coords, att_obj_name, *args):
   'format the arg list so that attribute structures are converted to strings if they exist'
   arg_begin = None
   arg_end = None
   if (len(args) != 0):
      arg_begin = args[0]
      arg_end = args[len(args)-1]

   start_arg = int(arg_begin != None)

   # if the last argument is a list but its empty return
   if check_seq (arg_end) and len(arg_end) == 0:
      return args



   # count the number of strings to ensure we have aleast the
   # minimum required
   num_strings = 0
   for ii in args:
      if isinstance (ii, str):
         num_strings = num_strings + 1

   # check to see if the last arg is a list or chips attribute object
   # if it's a list check to see if its first argument is a string
   # this is to differentiate data from the attributes. This wont
   # work when we go to plotting non-numerics
   if type(arg_end) == dict or (check_seq(arg_end) and isinstance(arg_end[0],str)) or is_object_list(arg_end) or (isinstance (arg_end, attr_objs.ChipsAttributeObject)) or (isinstance(arg_end,str) and num_strings > min_num_strings):
      new_args = []


      # TEMPORARY FIX
      # extract the coord sys value and add it to the chips id
      if check_coords:
         att_list = convert2attlist (arg_end)
         if is_object_list(arg_end) or isinstance (arg_end, attr_objs.ChipsAttributeObject):
            (coord_value, arg_end) = pop_attrlist (att_obj_name+'.coordsys',att_list)
         else:
            (coord_value, arg_end) = pop_attrlist ('coordsys', convert2attlist (arg_end))

         if coord_value != None:
            if isinstance(arg_begin, __pychips_base__.ChipsId):
               arg_begin.set_coord_sys (__pychips_base__.string_to_v_coord_sys(coord_value))
            else:
               id = __pychips_base__.ChipsId ()
               coord_sys = __pychips_base__.string_to_v_coord_sys(coord_value)
               if not __pychips_base__.valid_coord_sys_type (coord_sys):
                  raise AttributeError ('chips ERROR: The (%s) coordinate system does not exist'%coord_value)
               id.set_coord_sys (coord_sys)
               new_args.append (id)
      else:
         arg_end = convert2attlist (arg_end)

      # if args only have one element then end args will cover
      # the whatever arguments needed to be added to the list
      if len(args) != 1:
         if arg_begin != None:
            new_args.append (arg_begin)

         for x in range(start_arg,len(args)-1):
            new_args.append(args[x])
         
      if arg_end != None:
         new_args.append (arg_end)

      new_args = tuple (new_args)
   else:
      new_args = args
   return new_args

def set_id(id=None, typ=None):
   'Returns a ChipsId structure to pass into system calls'
   if isinstance(id, str):
      myid = __pychips_base__.ChipsId()
      if typ == None:
         raise AttributeError("chips ERROR: id specified without specifying the object type")
         return None
      else:
         myid.set_object_id(typ, id)
   elif isinstance(id, __pychips_base__.ChipsId):
      myid = id
   elif id == None:
      myid = __pychips_base__.ChipsId()
   else:
      raise AttributeError("chips ERROR: Unknown or unexpected parameter type")
      return None
   return myid;

def string_to_list (val, cast_type=None):
   """
   converts a string of format [a,b,c] to an actual list. an optional
   argument cast_type can be past to convert each element to a particular
   type
   """

   # if the input is already a list just return
   if check_seq (val) :
      return val
 
   if val == None:
      return None

   try:
      val = val.lstrip ()
      val = val.rstrip ()
   except:
      raise AttributeError ("chips ERROR: The attribute element is expected to be a string") 
   if not val.startswith('[') or not val.endswith (']'):
      raise AttributeError ("chips ERROR: The attribute element should be a list which starts '[' and ends with ']'")

   val = val[1:len(val)-1]
   val = val.lstrip ()

   # if the list is empty return
   if val == '':
      return None

   val = val.split (',')


   if cast_type != None:
      try:
         # convert each element of the list to the desired type
         for ii in range (len(val)):
            val[ii] = cast_type (val[ii])
      except:
         cast_type_str = str(cast_type)
         cast_type_str = cast_type_str[cast_type_str.find ("'"):cast_type_str.rfind ("'")+1]
         
         raise AttributeError ("chips ERROR: All elements of the attribute list are expected to be of type %s. The following is not valid, %s"%(cast_type_str,val[ii]))

   return val


def string_to_attribute_list(val):
   """
   converts a string of the format 'a.b=c d.e="f g"' to a list of entries
   ['a.b','c','d.e',"f g"']
   """
   z=shlex.split(val)
   l = []

   # rejoin string lists 'x=[1,', '2,', '3]' => 'x=[1 2 3]'
   tmp_list = []
   tmp_elem = ''
   open_list = False
   for elem in z:
      if not open_list:
         tmp_elem = elem

      if elem.find ('[') != -1:
         tmp_elem = ''
         open_list = True
      if elem.find (']') != -1:
         open_list = False
         tmp_elem = tmp_elem + elem

      if open_list:
         tmp_elem = tmp_elem + elem + ' '
      else:
         tmp_list.append (tmp_elem)
  
   z = tmp_list

   for elem in z:
      if elem.find(' ') != -1:
         a=elem.replace('=','="',1);
         a+='"'
      else:
         a=elem
      b=a.split('=',1)
      l.append(b[0])
      if (len(b) ==2):
         l.append(b[1])
   return l

def data_to_attribute_object (obj, data):
   if type(data) == dict:
      for attr,val in data.iteritems ():
         try:
            lvls = attr.split('.')
         except:
            raise AttributeError("chips ERROR: Expected string attribute name in list.")
         else:
            if len(lvls) == 1:   # handle single segment (bgcolor) attribute
               if hasattr(obj, attr):
                  setattr(obj, attr, val)
               else:
                  raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+attr+"' in list.")
            elif len(lvls) == 2: # handle two segment (border.color) attribute
               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     setattr(lev1, lvls[1], val)
                  elif lvls[0] == 'err' and lvls[1] == '*':
                     # curve/histogram- err.* special case 
                     if isinstance(obj, attr_objs.ChipsCurve):
                        setattr(lev1, 'up', val)
                        setattr(lev1, 'down', val)
                        setattr(lev1, 'left', val)
                        setattr(lev1, 'right', val)
                     elif isinstance(obj, attr_objs.ChipsHistogram):
                        setattr(lev1, 'up', val)
                        setattr(lev1, 'down', val)
                  else:
                     raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+attr+"' in list.")
               elif lvls[0] == '*':
                  cnt = obj.__set_all__(lvls[1], val);
                  if cnt == 0:
                     raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1])
               else:
                  raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1])
            elif len(lvls) == 3: # handle three segment (a.b.c) attribute
               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     lev2 = getattr(lev1, lvls[1])
                     if hasattr(lev2, lvls[2]):
                        setattr(lev2, lvls[2], val);
                     elif lvls[2] == '*' and lvls[1] == 'err':
                        # special case handling for (curve|histogram).err.*
                        if lvls[0] == 'curve':
                           setattr(lev2, 'up', val)
                           setattr(lev2, 'down', val)
                           setattr(lev2, 'left', val)
                           setattr(lev2, 'right', val)
                        elif lvls[0] == 'histogram':
                           setattr(lev2, 'up', val)
                           setattr(lev2, 'down', val)
                        else:
                           raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1]+"."+lvls[2])
                     else:
                        cnt = obj.__set_all__(lvls[2], val);
                        if cnt == 0:
                           raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[2]+"' in list.")
                  else:
                     cnt = obj.__set_all__(lvls[2], val);
                     if cnt == 0:
                        raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[1]+"' in list.")
               else:
                  raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+attr+"' in list.")
            elif len(lvls) == 4: # handle four segment (a.b.c.d) attribute
               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     lev2 = getattr(lev1, lvls[1])
                     if hasattr(lev2, lvls[2]):
                        lev3 = getattr(lev2, lvls[2])
                        if hasattr(lev3, lvls[3]):
                           setattr(lev3, lvls[3], val);
                        else:
                           cnt = obj.__set_all__(lvls[3], val);
                           if cnt == 0:
                              raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[3]+"' in list.")
                     else:
                        cnt = obj.__set_all__(lvls[2], val);
                        if cnt == 0:
                           raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[2]+"' in list.")
                  else:
                     cnt = obj.__set_all__(lvls[2], val);
                     if cnt == 0:
                        raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[1]+"' in list.")
               else:
                  raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+attr+"' in list.")

   # handle a list or attribute object
   elif check_seq (data):
      if (len(data) % 2) == 1:
         __pychips_base__.unlock()
         raise AttributeError("chips ERROR: Expected attribute settings as paired values such as \"color=blue\" or [\"color\",\"blue\"].")

      for elem in range(0, len(data), 2):
         if isinstance(data[elem], str):
            lvls = data[elem].split('.')
            if len(lvls) == 1:   # handle single segment (bgcolor) attribute
               if hasattr(obj, data[elem]):
		  field = getattr(obj, data[elem])
		  if isinstance(field, attr_objs.ChipsAttributeObject):
                     raise AttributeError("chips ERROR: "+obj.__class__.__name__+" member '"+data[elem]+"' is a class- not an attribute.")
		  else:
                     setattr(obj, data[elem], data[elem+1])
               else:
                  raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+data[elem]+"' in list.")
            elif len(lvls) == 2: # handle two segment (border.color) attribute

               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     setattr(lev1, lvls[1], data[elem+1])
                  elif lvls[0] == 'err' and lvls[1] == '*':
                     # curve/histogram err.* special case
                     if isinstance(obj, attr_objs.ChipsCurve):
                        setattr(lev1, 'up', data[elem+1])
                        setattr(lev1, 'down', data[elem+1])
                        setattr(lev1, 'left', data[elem+1])
                        setattr(lev1, 'right', data[elem+1])
                     elif isinstance(obj, attr_objs.ChipsHistogram):
                        setattr(lev1, 'up', data[elem+1])
                        setattr(lev1, 'down', data[elem+1])
                  else:
                     raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1])
               elif lvls[0] == '*':
                  cnt = obj.__set_all__(lvls[1], data[elem+1]);
                  if cnt == 0:
                     raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1])
               else:
                  raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1])

            elif len(lvls) == 3: # handle three segment (a.b.c) attribute
               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     lev2 = getattr(lev1, lvls[1])
                     if hasattr(lev2, lvls[2]):
                        setattr(lev2, lvls[2], data[elem+1])
                     elif lvls[2] == '*' and lvls[1] == 'err':
                        # special case handling for (curve|histogram).err.*
                        if lvls[0] == 'curve':
                           setattr(lev2, 'up', data[elem+1])
                           setattr(lev2, 'down', data[elem+1])
                           setattr(lev2, 'left', data[elem+1])
                           setattr(lev2, 'right', data[elem+1])
                        elif lvls[0] == 'histogram':
                           setattr(lev2, 'up', data[elem+1])
                           setattr(lev2, 'down', data[elem+1])
                        else:
                           raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1]+"."+lvls[2])
                     else:
                        raise AttributeError("chips ERROR: Unknown or unsupported <obj>.<attribute> entry: "+lvls[0]+"."+lvls[1]+"."+lvls[2])
                  elif lvls[1] == '*':
                     cnt = lev1.__set_all__(lvls[2], data[elem+1]);
                     if cnt == 0:
                        raise AttributeError("chips ERROR: Invalid "+lev1.__class__.__name__+" attribute '"+data[elem]+"' in list.")
                  else:
                     raise AttributeError("chips ERROR: Invalid "+lev1.__class__.__name__+" attribute '"+data[elem]+"' in list.")
               elif lvls[0] == '*':
                  if lvls[1]  == '*' and lvls[2] in ['color','bgcolor','thickness','size']:
                     cnt = obj.__set_all__(lvls[2], data[elem+1]);
                  elif lvls[1] in [ 'fill','line','symbol','err','label','ticklabel']:
                     pp = lvls[1] + "." + lvls[2]
                     cnt = obj.__set_all__(pp, data[elem+1]);
                     if cnt == 0:
                        raise AttributeError("chips ERROR: Invalid "+lvls[1] + "." + lvls[2]+" attribute '"+data[elem]+"' in list.")
                  
               else:
                   raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+data[elem]+"' in list.")
            elif len(lvls) == 4: # handle four segment (a.b.c.d) attribute
               if hasattr(obj, lvls[0]):
                  lev1 = getattr(obj, lvls[0])
                  if hasattr(lev1, lvls[1]):
                     lev2 = getattr(lev1, lvls[1])
                     if hasattr(lev2, lvls[2]):
                        lev3 = getattr(lev2, lvls[2])
                        if hasattr(lev3, lvls[3]):
                           setattr(lev3, lvls[3], data[elem+1]);
                        else:
                           cnt = obj.__set_all__(lvls[3], data[elem+1]);
                           if cnt == 0:
                              raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[3]+"' in list.")
                     else:
                        cnt = obj.__set_all__(lvls[2], data[elem+1]);
                        if cnt == 0:
                           raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[2]+"' in list.")
                  else:
                     cnt = obj.__set_all__(lvls[2], data[elem+1]);
                     if cnt == 0:
                        raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[1]+"' in list.")
               else:
                  raise AttributeError("chips ERROR: Invalid "+obj.__class__.__name__+" attribute '"+lvls[0]+"' in list.")
         else:
            raise AttributeError("chips ERROR: Expected string attribute name in list.")
   else:
      obj = data

   return obj


def get_window_redraw_list():
   'returns a list containing window ids, and corresponding redraw values'
   i = __pychips_base__.info_current()
   lst = []
   if i != None:
      cnt = i.count("Window [")
      if cnt > 0:
	 endpos = 0
         for elem in range(0,cnt):
            startpos = i.find('Window [', endpos)
	    if startpos != -1:
               endpos = i.find(']\n', startpos)
	       if endpos != -1:
                  lst.append(i[startpos + 8:endpos]) 
                  lst.append(__pychips_base__.get_window_redraw(i[startpos + 8:endpos]))
   return lst


def reset_window_draw_values(lst):
   'given a list containing window ids and redraw values reset the window redraw values'
   if check_seq(lst) and len(lst) % 2 == 0:
      for i in range (0, len(lst), 2):
         __pychips_base__.set_window_redraw(lst[i], lst[i+1])
	 

def get_current_window_id():
   'returns the id of the current window'
   i = __pychips_base__.info_current()
   if i == None:
      wid = None
   else:
      cnt = i.count("Window [")
      if cnt > 1:
         wid = 'all'
      else:
         wid = i[i.find('Window [') + 8: i.find(']\n')]
   return wid


def get_color_as_string (val):
   'get a color value as a string whether its in a hex form or string'
   color = val
   if not isinstance (val, str):
      try:
         color = '%.6x' % val
      except:
         pass
   return color

class ChipsPrintableString (str):
   'wraps strings that strings are formatted nicely without using print'
   def __init__ (self, string):
      self.str = string
   def __str__ (self):
      return self.str
   def __repr__ (self):
      return self.str


	 
	 
