#include "VtkColorMap.hh"

#include <vtkLookupTable.h>
#include <vtkUnsignedCharArray.h>
#include <vtkMath.h>

VtkColorMap::VtkColorMap () : ChipsColorMap ()
{
   this->colorTable = vtkLookupTable::New ();

   // default the number of table values
   this->colorTable->SetNumberOfTableValues (256);
   this->colorTable->SetRampToLinear ();

   this->tableType = VtkColorMap::HSV;
   this->tableTypeChanged = false;

   this->dataTable = NULL;

   this->colorScale = 0;
   this->colorRamp = 0;

   this->alphaOverride = false;
   this->thresholdSet = false;

   this->scalarRange[0] = -1;
   this->scalarRange[1] = -1;

   this->clear_positions ();

   this->built = false;
}


VtkColorMap::~VtkColorMap ()
{
   this->colorTable->Delete ();
   if (this->dataTable)
   {
      for (int ii=0; ii<this->tableSize; ii++)
         delete [](this->dataTable[ii]);
      delete []this->dataTable;
   }
}

void VtkColorMap::set_table_size (int size)
{
   this->colorTable->SetNumberOfTableValues (size);
}

int  VtkColorMap::get_table_size ()
{
   return this->colorTable->GetNumberOfTableValues ();
}

void VtkColorMap::set_alpha_range (double alpha_range[])
{
   this->alphaOverride = true;
   this->set_alpha_range (alpha_range[0], alpha_range[1]);
}
void VtkColorMap::get_alpha_range (double alpha_range[])
{
   // only get the range if the alpha was set
   if (this->alphaOverride)
   {
      alpha_range[0] = this->rgbaMap[3].front ().second;
      alpha_range[1] = this->rgbaMap[3].back ().second;
   }
   else
   {
      // if the alpha range is not set default to 1,1
      alpha_range[0] = 1;
      alpha_range[1] = 1;
   }
}

void VtkColorMap::set_color_scale (int scale)
{
   switch (scale)
   {
      case 0:
         this->colorTable->SetScaleToLinear ();
      break;
      case 1:
         this->colorTable->SetScaleToLog10 ();
      break;
      default:
      break;
   }
}

int  VtkColorMap::get_color_scale ()
{
   return this->colorScale;
}

void VtkColorMap::set_color_ramp (int ramp)
{
   switch (ramp)
   {
      case 0:
         this->colorTable->SetRampToLinear ();
      break;
      case 1:
         this->colorTable->SetRampToSCurve ();
      break;
      case 2:
         this->colorTable->SetRampToSQRT ();
      break;
      default:
      break;
   }

   this->colorRamp = ramp;
}

int  VtkColorMap::get_color_ramp ()
{
   return this->colorRamp;
}

void VtkColorMap::set_scalar_threshold (double threshold[])
{
   this->thresholdSet = true;
   this->colorTable->SetTableRange (threshold);
}

void VtkColorMap::get_scalar_threshold (double threshold[])
{
   if (this->thresholdSet)
      this->colorTable->GetTableRange (threshold);
   else
   {
      threshold[0] = -1;
      threshold[1] = -1;
   }
}


void VtkColorMap::set_scalar_range (double range[])
{
   this->scalarRange[0] = range[0];
   this->scalarRange[1] = range[1];

   if (!this->thresholdSet)
      this->set_scalar_threshold (range);
}

void VtkColorMap::get_scalar_range (double range[])
{
   range[0] = this->scalarRange[0];
   range[1] = this->scalarRange[1];
}


void VtkColorMap::set_blue_position (double position, double value)
{
   this->set_table_type (VtkColorMap::RGB);
   pair <double, double> map_value (position, value);
   this->rgbaMap[2].push_back (map_value);
   this->built = false;
}

void VtkColorMap::set_green_position (double position, double value)
{
   this->set_table_type (VtkColorMap::RGB);
   pair <double, double> map_value (position, value);
   this->rgbaMap[1].push_back (map_value);
   this->built = false;
}

void VtkColorMap::set_red_position (double position, double value)
{
   this->set_table_type (VtkColorMap::RGB);
   pair <double, double> map_value (position, value);
   this->rgbaMap[0].push_back (map_value);
   this->built = false;
}

void VtkColorMap::set_alpha_position (double position, double value)
{
   pair <double, double> map_value (position, value);
   this->rgbaMap[3].push_back (map_value);
   this->built = false;
}

void VtkColorMap::clear_positions (int position)
{

   if (position & 1)
      this->rgbaMap[0].clear ();
   if (position & 2)
      this->rgbaMap[1].clear ();
   if (position & 4)
      this->rgbaMap[2].clear ();
   if (position & 8)
   {
      // when reseting the alpha position also clear the
      // alpha range flag
      this->alphaOverride = false;
      this->rgbaMap[3].clear ();
   }

   this->built = false;
}


void VtkColorMap::set_blue_range (double min, double max)
{
   this->clear_positions (4);
   this->set_blue_position (0, min);
   this->set_blue_position (1, max);
}

void VtkColorMap::set_green_range (double min, double max)
{
   // clear the green position
   this->clear_positions (2);

   this->set_green_position (0, min);
   this->set_green_position (1, max);

}

void VtkColorMap::set_red_range (double min, double max)
{
   // clear the red position
   this->clear_positions (1);

   this->set_red_position (0, min);
   this->set_red_position (1, max);
}

void VtkColorMap::set_hue_position (double position, double value)
{
   this->set_table_type (VtkColorMap::HSV);
   pair <double, double> map_value (position, value);
   this->rgbaMap[0].push_back (map_value);
   this->built = false;
}

void VtkColorMap::set_saturation_position (double position, double value)
{
   this->set_table_type (VtkColorMap::HSV);
   pair <double, double> map_value (position, value);
   this->rgbaMap[1].push_back (map_value);
   this->built = false;
}

void VtkColorMap::set_value_position (double position, double value)
{
   this->set_table_type (VtkColorMap::HSV);
   pair <double, double> map_value (position, value);
   this->rgbaMap[2].push_back (map_value);
   this->built = false;
}


void VtkColorMap::set_hue_range (double min, double max)
{
   this->set_table_type (VtkColorMap::HSV);

   // clear the hue position
   this->clear_positions (1);
   this->set_hue_position (0, min);
   this->set_hue_position (1, max);

   this->built = false;
}


void VtkColorMap::set_value_range (double min, double max)
{
   this->set_table_type (VtkColorMap::HSV);

   // clear the value position
   this->clear_positions (4);
   this->set_value_position (0, min);
   this->set_value_position (1, max);

   this->built = false;
}

void VtkColorMap::set_saturation_range (double min, double max)
{
   this->set_table_type (VtkColorMap::HSV);

   // clear the saturation position
   this->clear_positions (2);

   this->set_saturation_position (0, min);
   this->set_saturation_position (1, max);

   this->built = false;
}

void VtkColorMap::set_alpha_range (double min, double max)
{
   
   // clear the alpha position
   this->clear_positions (8);

   // indicate that alpha range was explicity set. this is nessary
   // for tables to know whether alpha was overriden by the user
   this->alphaOverride = true; 
   this->set_alpha_position (0, min);
   this->set_alpha_position (1, max);

   this->built = false;
}

void VtkColorMap::set_table (double **table, int table_size)
{
   this->set_table_type (VtkColorMap::TABLE);

   if (this->dataTable)
   {
      for (int ii=0; ii<this->tableSize; ii++)
         delete [](this->dataTable[ii]);
      delete []this->dataTable;
   }
   this->tableSize = table_size;
   this->dataTable = new double* [table_size];

   for (int ii=0; ii<this->tableSize; ii++)
   {
      this->dataTable[ii] = new double [4];
      for (int jj=0; jj<4; jj++)
         this->dataTable[ii][jj] = table[ii][jj];
   }
 
   this->built = false;
}

void VtkColorMap::set_data_range (double min, double max)
{
   this->colorTable->SetTableRange (min, max);
}

void VtkColorMap::invert_table ()
{
   // make sure the table is built before trying to access it
   if (((this->tableType == VtkColorMap::TABLE) || 
       (this->tableType == VtkColorMap::RGB)) && !this->built)
      this->update ();

   this->inverted = !this->inverted;

   this->build_inverted_table ();
}

void VtkColorMap::build_inverted_table ()
{
   switch (this->tableType)
   {
      case VtkColorMap::HSV: // fallthrough intended
/*      {
         double range[2];
         double invertRange[2];
      
         this->colorTable->GetHueRange (range);
         invertRange[0] = range[1];
         invertRange[1] = range[0];
         this->colorTable->SetHueRange (invertRange);
         this->colorTable->GetValueRange (range);
         invertRange[0] = range[1];
         invertRange[1] = range[0];
         this->colorTable->SetValueRange (invertRange);
         this->colorTable->GetSaturationRange (range);
         invertRange[0] = range[1];
         invertRange[1] = range[0];
         this->colorTable->SetSaturationRange (invertRange);
      }
      break;
*/
      case VtkColorMap::TABLE:  // fallthrough intended
      case VtkColorMap::RGB:
      {

         double tmpRgba1[4];
         double tmpRgba2[4];
         double tmpValue;
         int numColors = this->get_table_size ();
 
         // invert the table
         for (int ii=0; ii<numColors/2; ii++)
         {
            this->colorTable->GetTableValue (ii, tmpRgba1);
            this->colorTable->GetTableValue (numColors-ii-1, tmpRgba2);
            for (int jj=0; jj<4; jj++)
            {
               tmpValue = tmpRgba1[jj];
               tmpRgba1[jj] = tmpRgba2[jj];
               tmpRgba2[jj] = tmpValue;
            }
            this->colorTable->SetTableValue (ii, tmpRgba1);
            this->colorTable->SetTableValue (numColors-ii-1, tmpRgba2);
         }
      }
      break;
   }
}

void VtkColorMap::update ()
{

   switch (this->tableType)
   {
      case VtkColorMap::HSV:   // fallthrough intentional
/*         // automatically interpolated
      break;
*/
      case VtkColorMap::RGB:
      {
         int numColors = this->get_table_size ();
         int prevRgbaPos[] = {0, 0, 0, 0};
         int nextRgbaPos[] = {1, 1, 1, 1};

         double prevRgbaValue[] = {0, 0, 0, 1};
         double nextRgbaValue[] = {0, 0, 0, 1};

         double rgbaValue[4];

         list <pair <double, double> >:: iterator rgba_iter[4];

         for (int ii=0; ii<4; ii++)
            rgba_iter[ii] = rgbaMap[ii].begin ();

         // initialize previous and next iterators to the first
         // two indicies in the rgba map
         for (int ii=0; ii<4; ii++)
         {
            // only update if the rgba map exists
            if (rgba_iter[ii] != rgbaMap[ii].end ())
            {
               prevRgbaPos[ii] = (int)(rgba_iter[ii]->first * numColors);
               prevRgbaValue[ii] = rgba_iter[ii]->second;

               rgba_iter[ii]++;

               nextRgbaPos[ii] = prevRgbaPos[ii];
               nextRgbaValue[ii] = prevRgbaValue[ii];

               // if a second entry exists set the next iterator to it
               if (rgba_iter[ii] != rgbaMap[ii].end ())
               {
                  nextRgbaPos[ii] = (int)(rgba_iter[ii]->first * numColors);
                  nextRgbaValue[ii] = rgba_iter[ii]->second;
               }
               
            }
            else 
               nextRgbaPos[ii] *= numColors;
         }

         // populate the table

         for (int ii=0; ii<numColors; ii++)
         {
            for (int jj=0; jj<4; jj++)
            {
               if ((prevRgbaPos[jj] < ii) && 
                   (nextRgbaPos[jj] >= ii) && 
                   !this->interpolate)
                  rgbaValue[jj] = prevRgbaValue[jj];
               else
                  rgbaValue[jj] = this->linear_interpolate (
                        prevRgbaPos[jj], nextRgbaPos[jj],
                        prevRgbaValue[jj], nextRgbaValue[jj], ii);
            }

            if (this->tableType == VtkColorMap::HSV)
               vtkMath::HSVToRGB (rgbaValue, rgbaValue);

            this->colorTable->SetTableValue (ii, rgbaValue);

            // update the iterators if they have passed the 
            // current index
            for (int jj=0; jj<4; jj++)
            {
               if (nextRgbaPos[jj] <= ii)
               {
                  prevRgbaPos[jj] = nextRgbaPos[jj];
                  prevRgbaValue[jj] = nextRgbaValue[jj];

                  if (rgba_iter[jj] != rgbaMap[jj].end ())
                     rgba_iter[jj]++;

                  if (rgba_iter[jj] != rgbaMap[jj].end ())
                  {
                     nextRgbaPos[jj] = (int)(rgba_iter[jj]->first * numColors);
                     nextRgbaValue[jj] = rgba_iter[jj]->second;
                  }
               }
            }
         }
      }
      break;
      case VtkColorMap::TABLE:
      {
         int numColors = 0;
         int tableIndex = 0;
         double step = 0;

         int rgbCount = 4;
         int alphaCount = 0;

         double alphaRange[] = {1,1};

         double rgbaValue[4];

         numColors = this->get_table_size ();

         if (this->interpolate)
            step = (double)(this->tableSize-1) / (double)(numColors-1);
         else
            step = (double)(this->tableSize) / (double)(numColors);

         // if the alpha range is set use them instead of the table values
         if (this->alphaOverride)
         {
            // get the alpha range 
            this->get_alpha_range (alphaRange);
            rgbCount = 3;
            alphaCount = 1;
         }

         for (int ii=0; ii<numColors; ii++)
         {

            if (this->interpolate)
            {

               for (int jj=0; jj<rgbCount; jj++)
               {
                  if (tableIndex+1 >= this->tableSize)
                     rgbaValue[jj] = this->dataTable[tableIndex][jj];
                  else
                     rgbaValue[jj] = this->linear_interpolate (
                                          tableIndex, 
                                          tableIndex + 1,
                                          this->dataTable[tableIndex][jj],
                                          this->dataTable[tableIndex+1][jj],
                                          ii*step);
               }

               if (alphaCount > 0)
               {
                  rgbaValue[3] = this->linear_interpolate (
                                          0, this->tableSize,
                                          alphaRange[0],
                                          alphaRange[1],
                                          ii*step);
               }
            }
            else
            {
               for (int jj=0; jj<rgbCount; jj++)
                  rgbaValue[jj] = this->dataTable[tableIndex][jj];

               // always interpolate the alpha range
               if (alphaCount > 0)
                  rgbaValue[3] = this->linear_interpolate (0, this->tableSize,
                                          alphaRange[0],
                                          alphaRange[1],
                                          tableIndex);
            }

            this->colorTable->SetTableValue (ii, rgbaValue);

            tableIndex = (int)((ii+1)*step);

         }
      }
      break;
      default:
      break;
   }

/*   if ((this->tableType == VtkColorMap::HSV) &&
       this->tableTypeChanged)
   {
      // vtk wont recognize the update unless the actual data has been
      // changed. since the last command was not a hsv it cant be
      // guaranteed that the current hsv is differnt from the last
      // hsv submitted. the following forces vtk to recognize the change
      double range[2];
      double invertRange[2];
      this->colorTable->GetHueRange (range);
      this->colorTable->SetHueRange (invertRange[0], invertRange[1]);
      this->colorTable->SetHueRange (range);

      this->colorTable->ForceBuild ();
   }
   else
*/
   this->colorTable->Build ();

   this->tableTypeChanged = false;

   this->built = true;

   // rebuild the table with new values inverted
   if (this->inverted)
      this->build_inverted_table ();
}


void VtkColorMap::serialize (ChipsStateStream& out)
{
   short type = static_cast <short> (this->tableType);
   int ival;
   double alphaRange[] = {1,1};

   out.write_bool (&inverted);
   out.write_bool (&interpolate);
   out.write_int (&colormap_type);

   out.write_short (&type);
   
   switch (this->tableType)
   {
      case VtkColorMap::HSV:  // fallthrough intentional
/*      {
         double range[2];
         this->colorTable->GetHueRange (range);
         out.write_double (&range[0], 2);
         this->colorTable->GetValueRange (range);
         out.write_double (&range[0], 2);
         this->colorTable->GetSaturationRange (range);
         out.write_double (&range[0], 2);
      }
      break;
*/
      case VtkColorMap::RGB:
      {
         int ival;
         double dval;
         list <pair <double, double> >::iterator iter;
         for (int ii=0; ii<4; ii++)
         {
            ival = this->rgbaMap[ii].size ();
            out.write_int (&ival);
            iter = this->rgbaMap[ii].begin ();  
            for (; iter != rgbaMap[ii].end (); iter++)
            {
               dval = iter->first;
               out.write_double (&dval);
               dval = iter->second;
               out.write_double (&dval);
            }
         }
      }
      break;
      case VtkColorMap::TABLE:
      {
         out.write_int (&this->tableSize);
         for (int ii=0; ii<this->tableSize; ii++)
            out.write_double (this->dataTable[ii], 4);
      }
      break;
   }

 
   out.write_int (&colorRamp);
   out.write_int (&colorScale);
   out.write_double (this->colorTable->GetTableRange (), 2);

   ival = this->get_table_size ();
   out.write_int (&ival);

   out.write_double (&this->scalarRange[0], 2);

   out.write_bool (&this->alphaOverride);
   if (this->alphaOverride)
   {
      this->get_alpha_range (alphaRange);
      out.write_double (&alphaRange[0], 2);
   }

   out.write_bool (&this->thresholdSet);


}

void VtkColorMap::deserialize (ChipsStateStream& in)
{
   int ival;
   short sval;

   in.read_bool (&inverted);
   in.read_bool (&interpolate);
   in.read_int (&colormap_type);

   in.read_short (&sval);

   this->tableType = static_cast <TABLETYPE> (sval);

   switch (this->tableType)
   {
      case VtkColorMap::HSV:  // fall through intentional
/*      {
         double range[2];
         in.read_double (&range[0], 2);
         this->colorTable->SetHueRange (range);
         in.read_double (&range[0], 2);
         this->colorTable->SetValueRange (range);
         in.read_double (&range[0], 2);
         this->colorTable->SetSaturationRange (range);
      }
      break;
*/
      case VtkColorMap::RGB:
      {
         int ival;
         double dval1;
         double dval2;
         pair <double, double> map_value;

         for (int ii=0; ii<4; ii++)
         {
            in.read_int (&ival);

            for (int jj=0; jj<ival; jj++)
            {
               in.read_double (&dval1);
               in.read_double (&dval2);
               map_value.first = dval1;
               map_value.second = dval2;
               this->rgbaMap[ii].push_back (map_value);

            }
         }
      }
      break;
      case VtkColorMap::TABLE:
      {
         in.read_int (&this->tableSize);
         this->dataTable = new double* [this->tableSize];
         for (int ii=0; ii<this->tableSize; ii++)
         {
            this->dataTable[ii] = new double[4];
            in.read_double (this->dataTable[ii], 4);
         }
      }
      break;
   }


   double thresh[2];
   in.read_int (&colorRamp);
   in.read_int (&colorScale);

   in.read_double (&thresh[0], 2);
   this->colorTable->SetTableRange (thresh);
   in.read_int (&ival);
   this->set_table_size (ival);

   if (((in.get_major_version() == 3) && (in.get_minor_version () >= 4)) ||
      (in.get_major_version() > 3))
   {
      double alphaRange[] = {1,1};
      in.read_double (&this->scalarRange[0], 2);
      in.read_bool (&this->alphaOverride);
      if (this->alphaOverride)
      {
         in.read_double (&alphaRange[0], 2);
         this->set_alpha_range (alphaRange);
      }

      in.read_bool (&this->thresholdSet);
   }


   this->update ();
}



double VtkColorMap::linear_interpolate (double position1, double position2, 
                                 double value1, double value2, double index)
{
   double interpValue = 0; 
   // if the previous position is greater than the current
   // table index then dont interpolate anything
   if (position1 >= index)
      interpValue = value1;
   // if the next position is less than the current
   // table index then dont interpolate anything
   else if (position2 <= index)
      interpValue = value2;
   else
   {
      // linearly interpolate between the previous
      // position and next position
      interpValue =
             ((value2 * (index-position1)) +
              (value1 * (position2-index))) /
             (position2 - position1);
   }

   return interpValue;
}

void VtkColorMap::set_table_type (TABLETYPE type)
{
   if (this->tableType != type)
   {
      this->tableTypeChanged = true;

      // due to table switch clear all positions except alpha
      this->clear_positions (7);
   }

   this->tableType = type;
}

unsigned char *VtkColorMap::get_computed_table (int &out_size)
{
   vtkUnsignedCharArray *vtkTable = this->colorTable->GetTable ();
   int size = this->colorTable->GetNumberOfTableValues () * 4;
   unsigned char *table = new unsigned char [size];

   for (int ii=size; ii--;)
      table[ii] = vtkTable->GetValue (ii);

   out_size = size;
   return table;
}
