// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <sstream>
#include "VtkCurve.hh"
#include "vtkDoubleArray.h"
#include "vtkPoints.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkGlyphSource2D.h"
#include "vtkProperty2D.h"
#include "vtkDoubleArray.h"
#include "vtkAxisActor2DExtra.hh"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "VtkUtils.hh"

#ifndef HEADLESS
#include "vtkChipsPickableWidget.hh"
#include "vtkPlotDataRepresentation.hh"
#endif



VtkCurve::VtkCurve(vtkRenderer *inRenderer,
		   ChipsObject* par, 
		   unsigned long cnt, double* x, double* y,
                   const string& transform_id,
	           const string& cid, bool compute):ChipsCurve(par,cnt,x,y,transform_id,cid,compute)
{
  long index = 0;

  renderer = inRenderer;
  plotData = vtkPlotDataListActor2DExtra::New ();
  if ( inRenderer )
    {

      inRenderer->AddActor (plotData);

#ifndef HEADLESS
      // setup gui widget
      vtkPlotDataRepresentation *plotDataRep = vtkPlotDataRepresentation::New ();
      ChipsId chipsId;
      plotDataRep->SetPlotDataActor (plotData);

      this->populate_chips_id (chipsId);
      plotDataWidget = vtkChipsPickableWidget::New ();
      plotDataWidget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
      plotDataWidget->SetRepresentation (plotDataRep);
      plotDataWidget->SetIdInfo (chipsId, this->get_object_type());
      plotDataWidget->ResizableOff ();
      plotDataWidget->RotateableOff ();
      plotDataWidget->MoveableOff ();
      plotDataWidget->SetDefaultRenderer (renderer);
      plotDataWidget->EnabledOn ();
      plotDataRep->Delete ();
#endif

      for (list<CHIPS_RANGE_P>::const_iterator tmpRangePtr=indices.begin(); tmpRangePtr!=indices.end();tmpRangePtr++,index++)
        {
          if ( create_curvedata(index,0,0,0,0,0,0,*tmpRangePtr) != ERR_STATUS_GOOD)
            {
              cerr<<"Error: VtkCurve::VtkCurve - AddData Failed"<<endl;
            };
        };


      change_linethickness();
      change_linestyle();
      change_linecolor(); 
      change_symbolsize(); 
      change_symbolstyle(); 
      change_symbolcolor();
      change_errcolor();
      change_errstyle();
      change_errup();
      change_errdown();
      change_errleft();
      change_errright();
      change_errthickness();
      change_visible();
      change_limitslength();
      change_limitsoverride();
      change_scale (0,0);
    };

};

VtkCurve::VtkCurve(vtkRenderer *inRenderer, ChipsObject* par,
                   ChipsStateStream& in): ChipsCurve(par, in)
{
  renderer = inRenderer;
  plotData = vtkPlotDataListActor2DExtra::New ();
  if ( inRenderer)
    {
      vtkDoubleArray * upErr=vtkDoubleArray::New();
      vtkDoubleArray * downErr=vtkDoubleArray::New();
      vtkDoubleArray * leftErr=vtkDoubleArray::New();
      vtkDoubleArray * rightErr=vtkDoubleArray::New();
      vtkDoubleArray * yLim=vtkDoubleArray::New();
      vtkDoubleArray * xLim=vtkDoubleArray::New();
      long index = 0;

      inRenderer->AddActor (plotData);

#ifndef HEADLESS
      // setup gui widget
      vtkPlotDataRepresentation *plotDataRep = vtkPlotDataRepresentation::New ();
      ChipsId chipsId; 
      plotDataRep->SetPlotDataActor (plotData);

      this->populate_chips_id (chipsId);
      plotDataWidget = vtkChipsPickableWidget::New ();
      plotDataWidget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor
 ());
      plotDataWidget->SetRepresentation (plotDataRep);
      plotDataWidget->SetIdInfo (chipsId, this->get_object_type());
      plotDataWidget->ResizableOff ();
      plotDataWidget->RotateableOff ();
      plotDataWidget->MoveableOff ();
      plotDataWidget->SetDefaultRenderer (renderer);
      plotDataWidget->EnabledOn ();
      plotDataRep->Delete ();
#endif

       for (list<CHIPS_RANGE_P>::const_iterator tmpRangePtr=indices.begin(); tmpRangePtr!=indices.end();tmpRangePtr++, index++)
         {
           if ( create_curvedata(index, upErr,downErr,leftErr,rightErr,yLim,xLim,*tmpRangePtr)!= ERR_STATUS_GOOD)
             {
               cerr<<"Error: VtkCurve::VtkCurve - AddData Failed"<<endl;
             };
         };



       change_linethickness();
       change_linestyle();
       change_linecolor(); 
       change_symbolsize(); 
       change_symbolstyle(); 
       change_symbolcolor();
       change_errcolor(); 
       change_errstyle(); 
       change_errup(); 
       change_errdown(); 
       change_errleft(); 
       change_errright(); 
       change_errthickness(); 
       change_limitslength();
       change_limitsoverride();
       change_visible();
       change_scale (0,0);

       upErr->Delete();
       downErr->Delete();
       leftErr->Delete();
       rightErr->Delete();
       yLim->Delete();
       xLim->Delete();
    };
};




VtkCurve::VtkCurve(vtkRenderer *inRenderer,
		   ChipsObject* par, 
		   unsigned long cnt, double* x, double* y,
		   double* ydown, double* yup, double* xdown, double* xup,
                   double* ylims, double* xlims, 
                   const string& transform_id,
                   const string& cid, bool compute):ChipsCurve(par,cnt,x,y,ydown,yup,xdown,xup,ylims,xlims,transform_id,cid,compute)
{   
  renderer = inRenderer;
  plotData = vtkPlotDataListActor2DExtra::New ();

  if ( inRenderer)
    {
      vtkDoubleArray * upErr=vtkDoubleArray::New();
      vtkDoubleArray * downErr=vtkDoubleArray::New();
      vtkDoubleArray * leftErr=vtkDoubleArray::New();
      vtkDoubleArray * rightErr=vtkDoubleArray::New();
      vtkDoubleArray * yLim=vtkDoubleArray::New();
      vtkDoubleArray * xLim=vtkDoubleArray::New();
      long index = 0;

      inRenderer->AddActor (plotData);
     
#ifndef HEADLESS 
      // setup gui widget
      vtkPlotDataRepresentation *plotDataRep = vtkPlotDataRepresentation::New ();
      ChipsId chipsId;
      plotDataRep->SetPlotDataActor (plotData);

      this->populate_chips_id (chipsId);
      plotDataWidget = vtkChipsPickableWidget::New ();
      plotDataWidget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor
 ());
      plotDataWidget->SetRepresentation (plotDataRep);
      plotDataWidget->SetIdInfo (chipsId, this->get_object_type());
      plotDataWidget->ResizableOff ();
      plotDataWidget->RotateableOff ();
      plotDataWidget->MoveableOff ();
      plotDataWidget->SetDefaultRenderer (renderer);
      plotDataWidget->EnabledOn ();
      plotDataRep->Delete ();
#endif

      for (list<CHIPS_RANGE_P>::const_iterator tmpRangePtr=indices.begin(); tmpRangePtr!=indices.end();tmpRangePtr++, index++)
        {

          if ( create_curvedata(index,upErr,downErr,leftErr,rightErr,yLim,xLim,*tmpRangePtr)!= ERR_STATUS_GOOD)
            {
              cerr<<"Error: VtkCurve::VtkCurve - AddData Failed"<<endl;
            };

        };



      change_linethickness();
      change_linestyle();
      change_linecolor(); 
      change_symbolsize(); 
      change_symbolstyle(); 
      change_symbolcolor();
      change_errcolor(); 
      change_errstyle(); 
      change_errup(); 
      change_errdown(); 
      change_errleft(); 
      change_errright(); 
      change_errthickness(); 
      change_limitslength();
      change_limitsoverride();
      change_visible();
      change_scale (0, 0);
      
      upErr->Delete();
      downErr->Delete();
      leftErr->Delete();
      rightErr->Delete();
      yLim->Delete();
      xLim->Delete();
    };
};

VtkCurve::~VtkCurve()
{

   if (renderer)
     renderer->RemoveActor (plotData);
   plotData->Delete ();
#ifndef HEADLESS
   plotDataWidget->Delete ();
#endif
};

int VtkCurve::render_err()
{
  return ERR_FEATURE_DNE;
};

int VtkCurve::change_linethickness()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
   {
     dataPtr->GetLineActor()->GetProperty()->SetLineWidth(linethickness);
   };
  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_linestyle()
{

  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      switch(linestyle)
	{
	case chips_noline:
	  dataPtr->SetLinesVisibility(0);
	  break;
	case chips_dotlongdash: 
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_dotlongdash_pattern);
	  break;
	case chips_dotshortdash: 
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_dotshortdash_pattern);
	  break;
	case chips_longdash: 
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_longdash_pattern);
	  break;
	case chips_shortdash:
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_shortdash_pattern);
	  break;
	case chips_shortdashlongdash:
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_shortdashlongdash_pattern);
	  break;
	case chips_dot:
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_dot_pattern);
  	  break;
	case chips_solid:
	default:
	  dataPtr->SetLinesVisibility(1);
	  dataPtr->GetLineActor()->GetProperty()->SetLineStipplePattern(chips_solid_pattern);
        };
   };       
   return ERR_STATUS_GOOD;
}; 
 
int VtkCurve::change_linecolor()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->GetLineActor()->GetProperty()->SetColor(linecolor.red(),linecolor.green(),linecolor.blue());
    };
  
  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_symbolsize()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->SetGlyphScale(symbolsize);
    }
#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif
  
  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_symbolstyle()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->SetGlyphAngle(symbolangle);
      switch (symbolstyle)
	{
	case chips_none:
	  dataPtr->SetPointsVisibility(0);
	  break;
	case chips_circle:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(7);
	  break;
	case chips_cross:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(4);
	  dataPtr->SetGlyphAngle(symbolangle + 45);
	  break;
	case chips_diamond:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(8);
	  break;
	case chips_downtriangle:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(5);
	  dataPtr->SetGlyphAngle(symbolangle + 180);
	  break;
	case chips_plus:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(4);
	  break;
	case chips_square:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(6);
	  break;
	case chips_uptriangle:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(5);
	  break;
	case chips_arrow:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(10);
	  break;
	case chips_point_type:
	default:
	  dataPtr->SetPointsVisibility(1);
	  dataPtr->SetGlyphType(1);
	};
	  
	  
      dataPtr->GetPointActor()->GetProperty()->SetColor(symbolcolor.red(),symbolcolor.green(),symbolcolor.blue());
	  
      if ( symbolfilled)
	{
	  dataPtr->SetGlyphFilled(1);
	}
      else
	{
	  dataPtr->SetGlyphFilled(0) ;
	};
	  
      dataPtr->SetGlyphScale(symbolsize);
#ifndef HEADLESS
      plotDataWidget->UpdateRepresentation ();
#endif
    };
  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_symbolcolor()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->GetPointActor()->GetProperty()->SetColor(symbolcolor.red(),symbolcolor.green(),symbolcolor.blue());
    };

  return ERR_STATUS_GOOD;
};



int VtkCurve::change_symbolfilled()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      if ( symbolfilled)
	{
	  dataPtr->SetGlyphFilled(1);
	}
      else
	{
	  dataPtr->SetGlyphFilled(0) ; 
	};
    };

  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_errcolor()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->GetErrBarActor()->GetProperty()->SetColor(errcolor.red(),errcolor.green(),errcolor.blue());
      dataPtr->GetDataLimitsActor()->GetProperty()->SetColor(errcolor.red(),errcolor.green(),errcolor.blue());
    };
  
  return ERR_STATUS_GOOD;

};
 
int VtkCurve::change_errstyle()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    { 
      switch( errstyle)
	{
	case chips_bar:
	  dataPtr->SetErrBarCapVisibility(0);
	  dataPtr->SetDataLimitsCapVisibility(0);
	  break;
	case chips_cap:
	default:
          dataPtr->SetErrBarCapLength(errcaplen); 
	  dataPtr->SetErrBarCapVisibility(1);
	  dataPtr->SetDataLimitsCapVisibility(1);
	};
    }
#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_errup()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());) 
    {

      if ( errup ){
	dataPtr->SetUpErrBarVisibility(1);
	dataPtr->SetYLimitsVisibility(1);
      }
      else{
	dataPtr->SetUpErrBarVisibility(0);
	dataPtr->SetYLimitsVisibility(0);
      }
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_errdown()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {

      if ( errdown ){
	dataPtr->SetDownErrBarVisibility(1);
	dataPtr->SetYLimitsVisibility(1);
      }
      else{
	dataPtr->SetDownErrBarVisibility(0);
	dataPtr->SetYLimitsVisibility(0);
      };
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
}; 

int VtkCurve::change_errleft()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {

      if ( errleft ){
	dataPtr->SetLeftErrBarVisibility(1);
	dataPtr->SetXLimitsVisibility(1);
      }
      else{
	dataPtr->SetLeftErrBarVisibility(0);
	dataPtr->SetXLimitsVisibility(0);
      }
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
  
}; 

int VtkCurve::change_errright()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      if ( errright ){
	dataPtr->SetRightErrBarVisibility(1);
	dataPtr->SetXLimitsVisibility(1);
      }
      else{
	dataPtr->SetRightErrBarVisibility(0);
	dataPtr->SetXLimitsVisibility(0);
      }
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;

}; 

int VtkCurve::change_errthickness()
{
  if ( errthickness < 0 )
    return ERR_STATUS_BAD;
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->GetErrBarActor()->GetProperty()->SetLineWidth(errthickness);
      dataPtr->GetDataLimitsActor()->GetProperty()->SetLineWidth(errthickness);
    };
  
  return ERR_STATUS_GOOD;
}; 


int VtkCurve::change_limitslength()
{
  if ( limlength < 0 )
    return ERR_STATUS_BAD;

  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->SetDataLimitsBarLength (limlength);
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
};



int VtkCurve::change_limitsoverride()
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      dataPtr->SetDataLimitsOverride (limoverride);
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
};



int VtkCurve::create_curvedata(long inPlotDataId,
				vtkDoubleArray* inUpErr,
				vtkDoubleArray* inDownErr,
				vtkDoubleArray* inLeftErr,
				vtkDoubleArray* inRightErr,
				vtkDoubleArray* inYLim,
				vtkDoubleArray* inXLim,
				CHIPS_RANGE_P inRangeInfo)
{

  if ( inRangeInfo )
    {
      if ((inRangeInfo->lo > inRangeInfo->hi) || (inRangeInfo->lo<0) || (inRangeInfo->hi<0))
	{
	  return ERR_STATUS_BAD;
	}
    };


  vtkPolyData* curveData = vtkPolyData::New ();
  curveData->Initialize();

  if ( inUpErr)
    {
      inUpErr->Initialize();
    };

  if ( inDownErr)
    {
      inDownErr->Initialize();
    };

  if ( inLeftErr)
    {
      inLeftErr->Initialize();
    };

  if ( inRightErr)
    {
      inRightErr->Initialize();
    };
  
  if ( inYLim)
    {
      inYLim->Initialize();
    };
  
  if ( inXLim)
    {
      inXLim->Initialize();
    };
  

  vtkPoints* points = vtkPoints::New();
  points->SetDataTypeToDouble();
  
  int increment;
  int low,high;
  int tmpIndx=0;
  
  if ( inRangeInfo)
    {
      low=inRangeInfo->lo;
      high=inRangeInfo->hi;
      if (inRangeInfo->inclusive )
	{
	  increment=1;
	}
      else
	{
	  if ( inRangeInfo->hi==inRangeInfo->lo)
	    increment=1;
	  else
	    increment=inRangeInfo->hi-inRangeInfo->lo;
	};
    }
  else
    {
      increment=1;
      low=0;
      high=num_pts-1;
    };

  for (int i=low;i <= high; i +=increment) 
    {
      points->InsertPoint(tmpIndx++,xx[i],yy[i],0);


      if ( yup && inUpErr )
	{
	  inUpErr->InsertNextValue(yup[i]);
	}
      
      if ( ydown && inDownErr)
	{
	  inDownErr->InsertNextValue(ydown[i]);
	}
  
      if ( xup && inRightErr )
	{
	  inRightErr->InsertNextValue(xup[i]);
	}
      
      if ( xdown && inLeftErr )
	{
	  inLeftErr->InsertNextValue(xdown[i]);
	};

      if ( ylim )
	{
	  inYLim->InsertNextValue(ylim[i]);
	};

      if ( xlim )
	{
	  inXLim->InsertNextValue(xlim[i]);
	};
    };
  
  
  
  curveData->SetPoints(points);
  //  curveData->GetPointData()->SetScalars(scalars);
  
  

  points->Delete();
  //  scalars->Delete();

  // Add data to vtk actor
  plotData->AddData (inPlotDataId, curveData, inUpErr, inDownErr, inLeftErr, inRightErr, inXLim, inYLim);


#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif
  curveData->Delete ();

  return ERR_STATUS_GOOD;

};

int VtkCurve::change_depth()
{
  plotData->SetLayerNumber (depth*10+1);
  return ERR_STATUS_GOOD;
}

int VtkCurve::change_visible ()
{
  int tmp_vis = (!get_parent()->is_visible ()) ? 0 : is_visible ();

  list<CHIPS_RANGE_P>::const_iterator tmpRangePtr=indices.begin();
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      if ( tmp_vis)
	{
	  if (((*tmpRangePtr)->mask & display_mask) == (*tmpRangePtr)->mask)  
	    {
	      dataPtr->SetVisibility(tmp_vis);
	    }
	  else
	    {
	      dataPtr->SetVisibility(!tmp_vis); 
	    };
  	}
      else
	{
	    dataPtr->SetVisibility(tmp_vis);
	};
	tmpRangePtr++;
     }

#ifndef HEADLESS
   plotDataWidget->GetRepresentation ()->SetVisibility (tmp_vis);
#endif

   return ERR_STATUS_GOOD;

}

int VtkCurve::update_data ()
{

  plotData->RemoveData ();

  vtkDoubleArray * upErr=vtkDoubleArray::New();
  vtkDoubleArray * downErr=vtkDoubleArray::New();
  vtkDoubleArray * leftErr=vtkDoubleArray::New();
  vtkDoubleArray * rightErr=vtkDoubleArray::New();
  vtkDoubleArray * yLim=vtkDoubleArray::New();
  vtkDoubleArray * xLim=vtkDoubleArray::New();
  long index = 0;


  for (list<CHIPS_RANGE_P>::const_iterator tmpRangePtr=indices.begin(); tmpRangePtr!=indices.end();tmpRangePtr++, index++)
    {
      if ( create_curvedata(index,upErr,downErr,leftErr,rightErr,yLim,xLim,*tmpRangePtr)!= ERR_STATUS_GOOD)
        {
          cerr<<"Error: VtkCurve::VtkCurve - AddData Failed"<<endl;
        };
    }


    change_linethickness();
    change_linestyle();
    change_linecolor();
    change_symbolsize();
    change_symbolstyle();
    change_symbolcolor();
    change_errcolor();
    change_errstyle();
    change_errup();
    change_errdown();
    change_errleft();
    change_errright();
    change_errthickness();
    change_visible();
    change_limitslength();
    change_limitsoverride();
    
    upErr->Delete();
    downErr->Delete();
    leftErr->Delete();
    rightErr->Delete();
    yLim->Delete();
    xLim->Delete();

    return ERR_STATUS_GOOD;
}

int VtkCurve::set_plot_location (double x1, double y1, double x2, double y2)
{
   double pos1[] = {x1,y1};
   double pos2[] = {x2,y2};
   plotData->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
   plotData->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();
   
   VtkUtils::FrameNorm2ViewportNorm (coord, renderer, pos1);
   VtkUtils::FrameNorm2ViewportNorm (coord, renderer, pos2);

   plotData->GetPositionCoordinate()->SetValue(pos1[0],pos1[1]);
   plotData->GetPosition2Coordinate()->SetValue(pos2[0],pos2[1]);

   plotData->UpdatePosition ();

#ifndef HEADLESS
   plotDataWidget->UpdateRepresentation ();
#endif

   return ERR_STATUS_GOOD;
}

int VtkCurve::change_range (const ChipsVector &low_range, const ChipsVector &high_range) 
{
  double tmpRangeX[2];
  double tmpRangeY[2];
  vtkPlotDataActor2DExtra *dataPtr = NULL;

  tmpRangeX[0] = low_range[0];
  tmpRangeY[0] = low_range[1];
  tmpRangeX[1] = high_range[0];
  tmpRangeY[1] = high_range[1];


  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      // Vtk does checking to see if the ranges are the same before updating
      // the actors. Actors wont get rerendered if ranges do not change. 
      dataPtr->SetXRange(tmpRangeX);
      dataPtr->SetYRange(tmpRangeY);

    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();
#endif

  return ERR_STATUS_GOOD;
}

int VtkCurve::change_scale (int inXScale, int inYScale)
{
  vtkPlotDataActor2DExtra *dataPtr = NULL;
  for (plotData->InitTraversal (); (dataPtr = plotData->GetNextPlotDataActor ());)
    {
      // Vtk does checking to see if the ranges are the same before updating
      // the actors. Actors wont get rerendered if ranges do not change.
      dataPtr->SetXScale(inXScale);
      dataPtr->SetYScale(inYScale);
    };

#ifndef HEADLESS
  plotDataWidget->UpdateRepresentation ();

#endif
  return ERR_STATUS_GOOD;
}

void VtkCurve::select ()
{
#ifndef HEADLESS
  if (plotDataWidget)
    plotDataWidget->Select (true);
#endif
}

void VtkCurve::unselect ()
{
#ifndef HEADLESS
  if (plotDataWidget)
    plotDataWidget->Select (false);
#endif
}

void VtkCurve::disable_interaction ()
{
#ifndef HEADLESS
  vtkPlotDataRepresentation *rep = reinterpret_cast<vtkPlotDataRepresentation *> (this->plotDataWidget->GetRepresentation ());
  rep->DisableBorder ();
  this->plotDataWidget->EnabledOff ();
#endif
}

void VtkCurve::enable_interaction ()
{
#ifndef HEADLESS
  if (this->is_interaction_blocked ())
    return;
  vtkPlotDataRepresentation *rep = reinterpret_cast<vtkPlotDataRepresentation *> (this->plotDataWidget->GetRepresentation ());

  this->plotDataWidget->EnabledOn ();
  rep->EnableBorder ();
#endif
}


