// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkDepthControl.hh"

VtkDepthControl::VtkDepthControl (vtkRenderer *inRenderer, int val): 
   ChipsDepthControl(val)
{
   renderer = inRenderer;
}

VtkDepthControl::VtkDepthControl (vtkRenderer *inRenderer, 
  ChipsFrame* frm, ChipsStateStream& in): ChipsDepthControl(frm, in)
{
   renderer = inRenderer;
   if (frm)   deserialize_obj_table(frm, in);
}



VtkDepthControl::~VtkDepthControl ()
{
  
}


// Sorts a list starting at a particular layer. Anything below that
// layer is assumed to be in order
int VtkDepthControl::sort_depth (int start_layer)
{
   vtkActor2DCollection *actors_list = renderer->GetActors2D ();
// There's something wrong with the insertion routine for actor2d
// where it seems to drop actors when inserting in the middle of the
// list. Use stl instead.
//   vtkActor2DCollection *new_actors_list = vtkActor2DCollection::New ();
   list <vtkActor2D *> new_actors_list;
   vtkActor2D *actor;

   if (actors_list->GetNumberOfItems () == 0)
      return ERR_STATUS_BAD;

   if (!renderer)
      return ERR_STATUS_BAD;

   // All vtk layers are base 10, while chips is base 1. Multiply the
   // the start layer by 10 to get it in the correct scale.
   start_layer *= 10;

   // Get to the correct layer to start sorting
   for (actors_list->InitTraversal(); actor = actors_list->GetNextActor2D();){
      if (actor->GetLayerNumber () >= start_layer)
         new_actors_list.push_back (actor);
   }

   new_actors_list.sort (&list_comp);

//   new_actors_list->Sort ();
   int layer = 0;
   list <vtkActor2D *>::iterator actor_list_iter = new_actors_list.begin ();
   for (; actor_list_iter != new_actors_list.end (); actor_list_iter++){
      actor = (*actor_list_iter);
      renderer->RemoveActor2D (actor);
      renderer->AddActor2D (actor);
      // Actors being sorted should ALWAYS be the depth x 10 + 1
      // After the actor gets sorted it should now be base 10.
      if ((layer = actor->GetLayerNumber ()) % 10 != 0)
         actor->SetLayerNumber (layer - 1);

   }

   return ERR_STATUS_GOOD;

}

// Finds the first unsorted item and sorts it into the list
// This assumes that the rest of the list is well sorted
int VtkDepthControl::sort_single_item ()
{
   vtkActor2DCollection *actors_list = renderer->GetActors2D ();
   vtkActor2D *unsorted_actor = NULL;
   vtkActor2D *prev_actor = NULL;
   vtkActor2D *curr_actor = NULL;
   bool init_prev = true;
   bool found_unsorted = false;
   int rval = ERR_STATUS_GOOD;

   if (actors_list->GetNumberOfItems () == 0)
      return ERR_STATUS_BAD;

   if (!renderer)
      return ERR_STATUS_BAD;
   // Get to the correct layer to start sorting
   for (actors_list->InitTraversal(); !found_unsorted && (curr_actor = actors_list->GetNextActor2D());){
      if (init_prev){
         prev_actor = curr_actor;
         init_prev = false;
      }

      // By a definition chips enforces. Out of order items will not be of base
      // 10. Find the first non-base 10 layer and sort it
      if (curr_actor->GetLayerNumber () % 10 != 0)
         unsorted_actor = curr_actor;

      // if previous actor is greater than current actor then the actors
      // are out of order
      if (prev_actor->GetLayerNumber () > curr_actor->GetLayerNumber ())
         found_unsorted = true;
      prev_actor = curr_actor;
   }

   if (found_unsorted){
      if (unsorted_actor){
         // Push the item onto end of the renderer
         renderer->RemoveActor (unsorted_actor);
         renderer->AddActor (unsorted_actor);

         // Put the last item back into the correct place
         rval = this->sort_last_item ();
      }
      else rval = ERR_STATUS_BAD;
   }
   else if (unsorted_actor){
      // The actors depth changed but it remains sorted
      // Reset its layer number
      unsorted_actor->SetLayerNumber (unsorted_actor->GetLayerNumber ()-1);
   }

   return rval;
}


// Pulls the last item from the list and sorts into the list
int VtkDepthControl::sort_last_item ()
{
   vtkActor2DCollection *actors_list = renderer->GetActors2D ();
   vtkActor2D *last_actor;
   vtkActor2D *actor;
   int layer;


   if (actors_list->GetNumberOfItems () == 0)
      return ERR_STATUS_BAD;

   if (!renderer)
      return ERR_STATUS_BAD;

   last_actor = actors_list->GetLastActor2D ();
   layer = last_actor->GetLayerNumber ();

   // Actors being sorted should ALWAYS be the depth x 10 + 1
   // After the actor gets sorted it should now be base 10.
   if (layer % 10 != 0)
      last_actor->SetLayerNumber (layer - 1);

   // Remove all actor that are depths greater than the last actor.
   // Then add them back on at the end of the list
   for (actors_list->InitTraversal(); actor = actors_list->GetNextActor2D();){
      if (actor->GetLayerNumber () > layer){
         renderer->RemoveActor (actor);
         renderer->AddActor2D (actor);
      }
   }

   return ERR_STATUS_GOOD;
}


