// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkFrame.hh" 
#include "vtkPoints.h"
#include "vtkCellArray.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkProperty2D.h"
#include "vtkRendererCollection.h"
#include "ChipsWindow.hh"
#include "VtkPlot.hh"
#include "VtkDepthControl.hh"

#include "vtkActor2DCollection.h"
#include "vtkActor2D.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkCoordinate.h"

#include "VtkUtils.hh"

#ifndef HEADLESS
#include "vtkChipsWidget.hh"
#include "vtkChipsRepresentation.hh"
#endif


VtkFrame::VtkFrame(vtkRenderWindow* inRenWin, ChipsObject* par, int pos, 
                               string val) : ChipsFrame(par, val)
{
   initialize (inRenWin, pos);
   depth_table = new VtkDepthControl(frameRenderer, def_depth);
}

VtkFrame::VtkFrame(vtkRenderWindow* inRenWin, ChipsObject* par, int pos, 
                               ChipsStateStream& in) : ChipsFrame(par, in)
{
   initialize (inRenWin, pos);
   depth_table = NULL; 
   change_visible ();
}

VtkFrame::VtkFrame(vtkRenderWindow* inRenWin, ChipsObject* par, int pos, float x1, float x2, 
			       float y1, float y2, 
			       string val) : ChipsFrame(par, x1, x2, y1, y2, val)
{
   initialize (inRenWin, pos);
   depth_table = new VtkDepthControl(frameRenderer, def_depth);
}
  
VtkFrame::~VtkFrame()
{
   if (plane_data)
      plane_data->Delete ();
   if (box_data)
      box_data->Delete ();
   frameRenderer->RemoveActor (planeActor);
   frameRenderer->RemoveActor (boundingBoxActor);

   remove_renderer ();
   
   planeActor->Delete ();
   boundingBoxActor->Delete ();
   frameRenderer->Delete ();
}

void VtkFrame::initialize (vtkRenderWindow* inRenWin, int pos)
{

   ChipsId cid;
   this->populate_chips_id (cid);

   plane_height = 0;
   plane_width = 0;

   plane_data = NULL;
   box_data = NULL;

   frameRenderer = vtkRenderer::New();
   frameRenderer->ResetCameraClippingRange();
   frameRenderer->SetBackground(0, 0, 0);

   planeActor = vtkActor2D::New ();
   if (planeActor)  planeActor->SetVisibility (opaque);
   boundingBoxActor = vtkActor2D::New ();

   create_plane ();
   create_bounding_box ();

   frameRenderer->AddActor (planeActor);
   frameRenderer->AddActor (boundingBoxActor);

#ifndef HEADLESS
   frame_rep = vtkChipsRepresentation::New ();

   frame_widget = vtkChipsWidget::New ();
   frame_widget->SetInteractor (inRenWin->GetInteractor ());
   frame_widget->SetRepresentation (frame_rep);
   frame_widget->SetIdInfo (cid, this->get_object_type());
   frame_widget->RotateableOff ();
   frame_widget->SetDefaultRenderer (frameRenderer);
   frame_widget->EnabledOn ();
#endif


   insert_renderer (frameRenderer, inRenWin, pos);
}


int VtkFrame::change_position()
{
   int retVal = 1;
   double size[2] = {bbx2-bbx1, bby2-bby1};
   int origin[] = {1,1};
   vtkPoints *pts;

   frameRenderer->SetViewport (bb_norm_x1, bb_norm_y1, bb_norm_x2, bb_norm_y2);
/*
   // This check is necessary because the display chops
   // off the line when its at the viewport bounds. 
   if (((ChipsWindow *)get_parent())->get_print_mode ())
   {
      origin[0] = 0;
      origin[1] = 0;
   }

   pts = box_data->GetPoints ();
   pts->SetPoint (0, origin[0], origin[1], 0);
   pts->SetPoint (1, origin[0], size[1], 0);
   pts->SetPoint (2, size[0], size[1], 0);
   pts->SetPoint (3, size[0], origin[1], 0);

   pts = plane_data->GetPoints ();
   pts->SetPoint (0, 0, 0, 0);
   pts->SetPoint (1, 0, size[1], 0);
   pts->SetPoint (2, size[0], size[1], 0);
   pts->SetPoint (3, size[0], 0, 0);

*/

   double min_pos[] = {0, 0};
   double max_pos[] = {1, 1};

   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, min_pos);
   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, max_pos);

   pts = box_data->GetPoints ();
   pts->SetPoint (0, min_pos[0], min_pos[1], 0);
   pts->SetPoint (1, min_pos[0], max_pos[1], 0);
   pts->SetPoint (2, max_pos[0], max_pos[1], 0);
   pts->SetPoint (3, max_pos[0], min_pos[1], 0);

   pts = plane_data->GetPoints ();
   pts->SetPoint (0, min_pos[0], min_pos[1], 0);
   pts->SetPoint (1, min_pos[0], max_pos[1], 0);
   pts->SetPoint (2, max_pos[0], max_pos[1], 0);
   pts->SetPoint (3, max_pos[0], min_pos[1], 0);



   

#ifndef HEADLESS
   vtkCoordinate* pos1 = this->frame_rep->GetPositionCoordinate ();
   vtkCoordinate* pos2 = this->frame_rep->GetPosition2Coordinate ();
   double low[2] = {0,0};
   double up[2] = {1,1};
/*   double low[2] = {bbx1,bby1};
   double up[2] = {bbx2, bby2};

   this->frameRenderer->DisplayToNormalizedDisplay(low[0], low[1]);
   this->frameRenderer->NormalizedDisplayToViewport(low[0], low[1]);
   this->frameRenderer->ViewportToNormalizedViewport(low[0], low[1]);

   this->frameRenderer->DisplayToNormalizedDisplay(up[0], up[1]);
   this->frameRenderer->NormalizedDisplayToViewport(up[0], up[1]);
   this->frameRenderer->ViewportToNormalizedViewport(up[0], up[1]);
*/

   pos1->SetCoordinateSystemToNormalizedViewport ();
   pos2->SetCoordinateSystemToNormalizedViewport ();

   pos1->SetValue (low[0], low[1]);
   pos2->SetValue (up[0], up[1]);
#endif

   return retVal;
}

void VtkFrame::change_mode (int mode)
{
   switch (mode)
   {
      case V_FRAMEMODE_PLACED:
#ifndef HEADLESS
         this->frame_widget->MoveableOn ();
         this->frame_widget->ResizableOn ();
#endif
      break;
      case V_FRAMEMODE_SINGLE:
      case V_FRAMEMODE_TILED:
#ifndef HEADLESS
         this->frame_widget->MoveableOff ();
         this->frame_widget->ResizableOff ();
#endif
      break;
   }
}

int VtkFrame::redraw ()
{
   int retVal;
   change_position ();
   return retVal;
}

ChipsPlot* VtkFrame::new_plot(string plid)
{
   string tmpid;

   if (plid.length())
      tmpid = plid;
   else
      tmpid = create_plot_id();
   return new VtkPlot(frameRenderer, this, tmpid);
}

ChipsPlot* VtkFrame::new_plot(ChipsStateStream& in)
{
   return new VtkPlot(frameRenderer, this, in);
}

ChipsPlot* VtkFrame::new_plot(float posx1, float posy1, float posx2,
   float posy2, string plid)
{
   string tmpid;

   if (plid.length())
      tmpid = plid;
   else
      tmpid = create_plot_id();

   return new VtkPlot(frameRenderer, this, posx1, posy1, posx2, posy2, tmpid);
}


int VtkFrame::change_visible ()
{
   int tmp_vis = (!visible) ? 0 : display_border;
   boundingBoxActor->SetVisibility (tmp_vis);
   tmp_vis = (!visible) ? 0 : opaque;
   planeActor->SetVisibility (tmp_vis);

   if (tmp_vis)
   {
#ifndef HEADLESS
      this->frame_rep->SetVisibility (tmp_vis);
#endif
   }



   return ERR_STATUS_GOOD;
}


int VtkFrame::change_color()
{
   if (visible == false)
      return ERR_STATUS_GOOD;
   else if (planeActor)
      planeActor->GetProperty()->SetColor(color.red(), color.green(), 
                                          color.blue());
   else  return ERR_STATUS_BAD; 
   return ERR_STATUS_GOOD; 
}

int VtkFrame::change_border_color()
{
   if (visible == false)
      return ERR_STATUS_GOOD;
   else if (boundingBoxActor)
      boundingBoxActor->GetProperty ()->SetColor(border_color.red(),
         border_color.green(), border_color.blue());
   else  return ERR_STATUS_BAD; 
   return ERR_STATUS_GOOD; 
}


int VtkFrame::change_transparency()
{
   if (planeActor)
      planeActor->SetVisibility (opaque);
   return 0;
}


int VtkFrame::change_border_visibility()
{
   if (visible == false)
      return ERR_STATUS_GOOD;
   else if (boundingBoxActor)
      boundingBoxActor->SetVisibility (display_border);
   else  return ERR_STATUS_BAD;
   return ERR_STATUS_GOOD;
}


void VtkFrame::create_plane ()
{
   double min_pos[] = {0, 0};
   double max_pos[] = {1, 1};
   vtkPoints *pts = vtkPoints::New ();

   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, min_pos);
   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, max_pos);

   pts->SetNumberOfPoints (4);
   pts->InsertPoint (0, min_pos[0], min_pos[1], 0);
   pts->InsertPoint (1, min_pos[0], max_pos[1], 0);
   pts->InsertPoint (2, max_pos[0], max_pos[1], 0);
   pts->InsertPoint (3, max_pos[0], min_pos[1], 0);


   vtkCellArray *arr = vtkCellArray::New ();
   arr->InsertNextCell (4);
   arr->InsertCellPoint (0);
   arr->InsertCellPoint (1);
   arr->InsertCellPoint (2);
   arr->InsertCellPoint (3);

   if (plane_data == NULL)
      plane_data = vtkPolyData::New ();

   plane_data->SetPoints (pts);
   plane_data->SetPolys (arr);

   vtkPolyDataMapper2D *mapper = vtkPolyDataMapper2D::New ();
   mapper->SetInput (plane_data);

   vtkCoordinate *coordinateSystem = vtkCoordinate::New ();
   coordinateSystem->SetCoordinateSystemToNormalizedViewport ();
   mapper->SetTransformCoordinate (coordinateSystem);
   coordinateSystem->Delete ();

   planeActor->SetMapper (mapper);

   planeActor->GetProperty()->SetColor(color.red(), color.green(), color.blue());
   planeActor->SetPosition (0.0, 0.0);

   pts->Delete ();
   arr->Delete ();
   mapper->Delete ();
}

void VtkFrame::create_bounding_box ()
{
   int *size = this->frameRenderer->GetSize ();
   vtkPoints *pts = vtkPoints::New ();

   double min_pos[] = {0, 0};
   double max_pos[] = {1, 1};

   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, min_pos);
   VtkUtils::FrameNorm2ViewportNorm (coord, this->frameRenderer, max_pos);

   pts->SetNumberOfPoints (4);
   pts->InsertPoint (0, min_pos[0], min_pos[1], 0);
   pts->InsertPoint (1, min_pos[0], max_pos[1], 0);
   pts->InsertPoint (2, max_pos[0], max_pos[1], 0);
   pts->InsertPoint (3, max_pos[0], min_pos[1], 0);

   vtkCellArray *arr = vtkCellArray::New ();
   arr->InsertNextCell (5);
   arr->InsertCellPoint (0);
   arr->InsertCellPoint (1);
   arr->InsertCellPoint (2);
   arr->InsertCellPoint (3);
   arr->InsertCellPoint (0);

   if (box_data == NULL)
      box_data = vtkPolyData::New ();

   box_data->SetPoints (pts);
   box_data->SetLines (arr);  

   vtkPolyDataMapper2D *mapper = vtkPolyDataMapper2D::New ();
   mapper->SetInput (box_data);

   vtkCoordinate *coordinateSystem = vtkCoordinate::New ();
//   coordinateSystem->SetCoordinateSystemToViewport ();
   coordinateSystem->SetCoordinateSystemToNormalizedViewport ();
   mapper->SetTransformCoordinate (coordinateSystem);
   coordinateSystem->Delete ();

   if (boundingBoxActor)
   {
      boundingBoxActor->SetMapper (mapper);
      boundingBoxActor->GetProperty ()->SetColor(border_color.red(), 
         border_color.green(), border_color.blue());
      boundingBoxActor->SetPosition (0.0, 0.0);
      boundingBoxActor->SetVisibility (display_border);
   }
   pts->Delete ();
   arr->Delete ();
   mapper->Delete ();
}

void VtkFrame::insert_renderer (vtkRenderer *new_ren, vtkRenderWindow *renWin, int pos)
{
   vtkRendererCollection *rc;
   vtkRenderer *ren;

   if (renWin == NULL)
      return;
   renWin->SetNumberOfLayers (renWin->GetNumberOfLayers() + 1);

   // Go through the list of the renderers and increase every renderer
   // that comes after this one by one. The zeroth renderer is renderered
   // on top
   rc = renWin->GetRenderers ();
   for (rc->InitTraversal(); (ren = rc->GetNextItem()); ){
      if (ren->GetLayer() >= pos)
         ren->SetLayer(ren->GetLayer()+1);
   }


   new_ren->SetLayer (pos);
   renWin->AddRenderer (new_ren);
}

void VtkFrame::remove_renderer ()
{
   vtkRendererCollection *rc;
   vtkRenderer *ren;
   vtkRenderWindow *renWin = frameRenderer->GetRenderWindow ();
   int max_layer = renWin->GetNumberOfLayers();

   renWin->RemoveRenderer (frameRenderer);

   // if this frame renderer was the top layer reset the render windows
   // number of layer to the next highest layer.
   if (frameRenderer->GetLayer () == (max_layer-1))
   {
      max_layer = 0;

      // Go through the list of the renderers find the new highest layer
      rc = renWin->GetRenderers ();
      for (rc->InitTraversal(); (ren = rc->GetNextItem()); ){
         if (ren->GetLayer() >= max_layer)
            max_layer = ren->GetLayer();
      }
      renWin->SetNumberOfLayers(max_layer+1);
   }

}



int VtkFrame::set_order (int order_num)
{
   frameRenderer->SetLayer (order_num);
   return ERR_STATUS_GOOD;
}


int  VtkFrame::deserialize_depth_table(ChipsStateStream& in)
{
   depth_table = new VtkDepthControl(frameRenderer, this, in);
   return ERR_STATUS_GOOD;
}

void VtkFrame::select ()
{
#ifndef HEADLESS
  if (frame_widget)
    frame_widget->Select (true);
#endif
}

void VtkFrame::unselect ()
{
#ifndef HEADLESS
  if (frame_widget)
    frame_widget->Select (false);
#endif
}

void VtkFrame::disable_interaction ()
{
#ifndef HEADLESS
  this->frame_widget->SetEnabled (false);
  this->frame_rep->DisableBorder ();

  this->disable_plot_interaction ("all");
#endif
}

void VtkFrame::enable_interaction ()
{
#ifndef HEADLESS
  if (this->is_interaction_blocked ())
      return;
  this->frame_widget->SetEnabled (true);
  this->frame_rep->EnableBorder ();

  this->enable_plot_interaction ("all");
#endif
}


void VtkFrame::add_grid (vtkActor2D *grid)
{
  if (grid)
    this->frameRenderer->AddActor (grid);
}
