// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkLabel.hh"
#include "VtkFrame.hh"
#include "vtkCoordinate.h"
#include "vtkCamera.h"
#include "vtkRenderWindow.h"

#include "vtkTextActorExtra.hh"
#include "ChipsCoordinate.hh"

#ifndef HEADLESS
#include "vtkChipsTextWidget.hh"
#include "vtkChipsTextRepresentation.hh"
#endif

#include "ChipsId.hh"

#include "VtkUtils.hh"


VtkLabel::VtkLabel(vtkRenderer* inRenderer,ChipsObject* par, const string& lbid,
                  ChipsCoordinate *inCoord, COORD_SYS coord_sys, double inXPos,
                  double inYPos,const string& val):ChipsLabel(par,lbid,inCoord,coord_sys,inXPos,inYPos,val)
{
  renderer = inRenderer;

  vtkText=vtkTextActorExtra::New();
  AddFormattedText(format,vtkText, get_scale_factor ());
  vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
  vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();

#ifndef HEADLESS
  ChipsId cid;
  this->populate_chips_id (cid);

  vtkChipsTextRepresentation *label_rep = vtkChipsTextRepresentation::New ();
  label_rep->SetTextActor (vtkText);
  label_rep->SetMaxFontSize (V_SIZE_MAX_VAL);


  this->label_widget = vtkChipsTextWidget::New ();
  this->label_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->label_widget->SetRepresentation (label_rep);
  this->label_widget->SetIdInfo (cid, this->get_object_type());
  this->label_widget->SetDefaultRenderer (renderer);
  this->label_widget->EnabledOn ();
  label_rep->Delete ();
#endif

  change_position();
  change_color();
  change_rotation();
  change_alignment();
  renderer->AddActor(vtkText);


};

VtkLabel::VtkLabel(vtkRenderer* inRenderer,ChipsObject* par, ChipsCoordinate *inCoord, COORD_SYS coord_sys, ChipsStateStream& in):
   ChipsLabel(par, inCoord, coord_sys, in)
{
  renderer = inRenderer;

  vtkText=vtkTextActorExtra::New();
  AddFormattedText(format,vtkText, get_scale_factor ());
  vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
  vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();
  renderer->AddActor(vtkText);

#ifndef HEADLESS
  vtkChipsTextRepresentation *label_rep = vtkChipsTextRepresentation::New ();
  label_rep->SetTextActor (vtkText);
  label_rep->SetMaxFontSize (V_SIZE_MAX_VAL);

  ChipsId cid;
  this->populate_chips_id (cid);

  this->label_widget = vtkChipsTextWidget::New ();
  this->label_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor
 ());
  this->label_widget->SetRepresentation (label_rep);
  this->label_widget->SetIdInfo (cid, this->get_object_type());
  this->label_widget->SetDefaultRenderer (renderer);
  this->label_widget->EnabledOn ();
  label_rep->Delete ();
#endif

  change_position();
  change_color();
  change_rotation();
  change_alignment();
  change_visible ();


};


VtkLabel::~VtkLabel()
{
  if ( renderer)
    {
      renderer->RemoveActor(vtkText);
      vtkText->Delete();
    }

#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Delete ();
#endif
};

int VtkLabel::redraw()
{

  change_position();
  change_color();
  change_rotation();
  change_alignment();
  return ERR_STATUS_GOOD;
};

int VtkLabel::change_text()
{  

  if (vtkText)
    {
      vtkText->Clear();
      AddFormattedText(format,vtkText,get_scale_factor ());
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif
       double rr,gg,bb; 
       get_rgb_color(rr,gg,bb);
       vtkText->setBBoxColor(rr,gg,bb);
    }
  else
    return ERR_STATUS_BAD ;


  return ERR_STATUS_GOOD;
};

int VtkLabel::change_font()
{
  return change_text();

};

int VtkLabel::change_size()
{ 
  return change_text();
  
};

int VtkLabel::change_bbox()
{
  if ( vtkText )
    {
      vtkText->SetdrawBBox(bbox);
    }
  return ERR_STATUS_GOOD;
};

int VtkLabel::change_depth ()
{
  // Axes will handle depth themselves
  if ( vtkText )
    {
      vtkText->SetLayerNumber (depth*10+1);
    }
  return ERR_STATUS_GOOD;
};

int VtkLabel::change_scale_factor()
{ 
  return change_text();
  
};

int VtkLabel::change_position()
{
  int rval = ERR_STATUS_GOOD;
  if ( renderer)
    {
      ChipsVector tmp_orig (0,0);
      vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
      vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();

      if (coord)
         rval = coord->convert_cs_to_cs (&orig, &tmp_orig, 1, coord_system, data_system_id, FRAME_NORM, "");
      else
      {
         tmp_orig[0] = orig[0];
         tmp_orig[1] = orig[1];
      }

      if (rval != ERR_STATUS_GOOD)
      {
         tmp_orig[0] = -1;
         tmp_orig[1] = -1;
      }


      VtkUtils::FrameNorm2ViewportNorm (coord, renderer, &tmp_orig[0]);

      vtkText->GetPositionCoordinate()->SetValue(tmp_orig[0], tmp_orig[1]);

      if ((coord_system == DATA) && !rval)
       {
         vtkCoordinate *Location[2];
         ChipsCoordinate *tmp_coord= NULL;
	 double tmpActorLow[] = {0.0, 0.0};
         double tmpActorHigh[] = {1.0, 1.0};

         if (coord)
            tmp_coord = coord->get_coord (PLOT_NORM, "");

         if (tmp_coord)
         {
            tmpActorLow[0]= tmp_coord->get_lower_bounds ()[0];
            tmpActorLow[1]= tmp_coord->get_lower_bounds ()[1];
            tmpActorHigh[0]= tmp_coord->get_upper_bounds ()[0];
            tmpActorHigh[1]= tmp_coord->get_upper_bounds ()[1];
         }

         Location[0] = vtkCoordinate::New ();
         Location[1] = vtkCoordinate::New ();
         Location[0]->SetCoordinateSystemToNormalizedViewport();
         Location[1]->SetCoordinateSystemToNormalizedViewport();

         VtkUtils::FrameNorm2ViewportNorm (coord, renderer, tmpActorLow);
         VtkUtils::FrameNorm2ViewportNorm (coord, renderer, tmpActorHigh);

         Location[0]->SetValue(tmpActorLow[0],tmpActorLow[1]);
         Location[1]->SetValue(tmpActorHigh[0],tmpActorHigh[1]);

         // Check for clipped data
         if ((tmp_orig[0] < tmpActorLow[0]) || (tmp_orig[0] > tmpActorHigh[0]) ||
             (tmp_orig[1] < tmpActorLow[1]) || (tmp_orig[1] > tmpActorHigh[1]))
            vtkText->GetPositionCoordinate()->SetValue(-1, -1);

         Location[0]->Delete ();
         Location[1]->Delete ();
       }

      // This error is fine. It just indicates that the point needs
      // to be clipped
      if (rval == ERR_NEG_LOG_VAL)
        rval = ERR_STATUS_GOOD;


      vtkText->Modified();

#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif
    }

  return rval;
};

int VtkLabel::change_color()
{
  return change_text();
};

int VtkLabel::change_visible ()
{
   int tmp_vis = (!get_parent()->is_visible ()) ? 0 : is_visible();

   if (vtkText)
     vtkText->SetVisibility (tmp_vis);
#ifndef HEADLESS
   this->label_widget->GetRepresentation ()->SetVisibility (tmp_vis);
   if (this->label_widget && tmp_vis)
   {
       this->label_widget->GetRepresentation ()->SetVisibility (tmp_vis);
   }
#endif

   return ERR_STATUS_GOOD;

}



int VtkLabel::change_rotation()
{
  if (vtkText)
    {
      vtkText->GetTextProperty()->SetOrientation(angle);
      vtkText->Modified();
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif

    }

  return ERR_STATUS_GOOD;
};


int VtkLabel::change_alignment()
{
  if ( vtkText)
    {
      vtkText->SetVertAlignmentOffset(vert_alignment);
      vtkText->SetHorizAlignmentOffset(horiz_alignment);
      vtkText->Modified();
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif

    }

  return ERR_STATUS_GOOD;
};


void VtkLabel::select ()
{
#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Select (true);
#endif
}

void VtkLabel::unselect ()
{
#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Select (false);
#endif
}


int VtkLabel::compute_resize( double bounds[])
{

  // bounds should be in display coordinates 

  double outPosition[2];

  int font = ComputeFontSizeAndPosFromBBox (this->vtkText, renderer, bounds, outPosition);
  int rval = this->set_property(static_cast<void*>(&font), SET_LABEL_SIZE);
  
  COORD_SYS coord_sys = PIXEL_REAL;
  string coord_id = "";

  if (!rval)
     this->move (coord_sys, coord_id, outPosition[0], outPosition[1], V_MOVE_ABSOLUTE);

  return rval;
}

void VtkLabel::disable_interaction ()
{
#ifndef HEADLESS
  vtkChipsTextRepresentation *rep = reinterpret_cast<vtkChipsTextRepresentation *> (this->label_widget->GetRepresentation ());
  rep->DisableBorder ();
  this->label_widget->EnabledOff ();
#endif
}

void VtkLabel::enable_interaction ()
{
#ifndef HEADLESS
  if (this->is_interaction_blocked ())
     return;

  vtkChipsTextRepresentation *rep = reinterpret_cast<vtkChipsTextRepresentation *> (this->label_widget->GetRepresentation ());

  this->label_widget->EnabledOn ();
  rep->EnableBorder ();
#endif
}

