// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkPlot.hh"

#include "vtkTextProperty.h"
#include "vtkAxisActor2DExtra.hh"
#include "VtkHistogram.hh"
#include "VtkCurve.hh"
#include "VtkAxis.hh"
#include "VtkContour.hh"
#include "VtkImage.hh"
#include "VtkPlotLabel.hh"
#include "VtkLabel.hh"
#include "VtkLine.hh"
#include "VtkPoint.hh"
#include "VtkPolygon.hh"
#include "VtkColorBar.hh"
#include "ChipsCoordinate.hh"
#include "ChipsCmdManager.hh"

#ifndef HEADLESS
#include "vtkRenderWindow.h"
#include "vtkBoundingBoxWidget.hh"
#include "vtkBoundingBoxRepresentation.hh"
#include "vtkAddAnnotationWidget.hh"
#include "vtkAddAnnotationRepresentation.hh"
#include "vtkChipsPlotWidget.hh"
#include "vtkChipsRepresentation.hh"
#endif

#include <sstream>


extern ChipsCmdManager* cmd_manager;

VtkPlot::VtkPlot(vtkRenderer* inRenderer,ChipsObject* par, const string& val):
 ChipsPlot(par,val)
{
  renderer=inRenderer;

  title = new VtkPlotLabel(renderer, this,"title",title_xpos_norm,title_ypos_norm, "");
  cmd_manager->remove_last_err (); 

#ifndef HEADLESS
  this->bounding_box = vtkBoundingBoxWidget::New ();
  this->bounding_box->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->bounding_box->SetDefaultRenderer (renderer);
  this->bounding_box->EnabledOn ();

  this->annotation_widget = vtkAddAnnotationWidget::New ();
  this->annotation_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->annotation_widget->SetDefaultRenderer (renderer);
  this->annotation_widget->EnabledOn ();


  ChipsId cid;
  this->populate_chips_id (cid);

  this->plot_rep = vtkChipsRepresentation::New ();

  this->plot_widget = vtkChipsPlotWidget::New ();
  this->plot_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->plot_widget->SetRepresentation (plot_rep);
  this->plot_widget->SetIdInfo (cid, this->get_object_type());
  this->plot_widget->RotateableOff ();
  this->plot_widget->SetDefaultRenderer (renderer);
  this->plot_widget->EnabledOn ();
#endif

  this->recalculate_position ();
};


VtkPlot::VtkPlot(vtkRenderer* inRenderer, ChipsObject* par, ChipsStateStream& in):
 ChipsPlot(par, in)
{
  renderer=inRenderer;

  if (deserialize_title)
  {
     title = new VtkPlotLabel(renderer,this, in); 
     cmd_manager->remove_last_err (); 
  }
  else
  {
     title = new VtkPlotLabel(renderer,this,"title",title_xpos_norm,title_ypos_norm, ""); 
     cmd_manager->remove_last_err (); 
  }

     if ((bottom = new VtkAxis (renderer, this, coord, PLOT_NORM, in)) != NULL)
        handle_new_xaxis(bottom, false);
     if ((top = new VtkAxis (renderer, this,  coord, PLOT_NORM,in)) != NULL)
        handle_new_xaxis(top, false);
     if ((left = new VtkAxis (renderer, this, coord, PLOT_NORM, in)) != NULL)
        handle_new_yaxis(left, false);
     if ((right = new VtkAxis (renderer, this, coord, PLOT_NORM, in)) != NULL)
        handle_new_yaxis(right, false);

     // create coordinate systems for the axes -- these coordinate
     // systems are for internal use only
     add_coordinate_system (top, right, NULL, false);
     add_coordinate_system (bottom, left, NULL, false);

     if (bindings)
     {
        bindings->add_binding(bottom, top);
        bindings->add_binding(left, right);
     }
     cur_x = cur_y = NULL;  // don't leave border's as current axes

#ifndef HEADLESS
  this->bounding_box = vtkBoundingBoxWidget::New ();
  this->bounding_box->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->bounding_box->EnabledOn ();

  this->annotation_widget = vtkAddAnnotationWidget::New ();
  this->annotation_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->annotation_widget->EnabledOn ();



  ChipsId cid;
  this->populate_chips_id (cid);

  this->plot_rep = vtkChipsRepresentation::New ();

  this->plot_widget = vtkChipsPlotWidget::New ();
  this->plot_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->plot_widget->SetRepresentation (plot_rep);
  this->plot_widget->SetIdInfo (cid, this->get_object_type());
  this->plot_widget->RotateableOff ();
  this->plot_widget->SetDefaultRenderer (renderer);
  this->plot_widget->SetPriority (.95);
  this->plot_widget->EnabledOn ();
#endif


};

VtkPlot::VtkPlot(vtkRenderer* inRenderer, ChipsObject* par, 
		     float x1, float y1, float x2, float y2, 
		     const string& val): ChipsPlot(par,x1,y1,x2,y2,val)
{
  renderer=inRenderer;
  title = new VtkPlotLabel(renderer,this,"title",title_xpos_norm,title_ypos_norm, "");

  cmd_manager->remove_last_err (); 

#ifndef HEADLESS
  this->bounding_box = vtkBoundingBoxWidget::New ();
  this->bounding_box->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->bounding_box->EnabledOn ();

  this->annotation_widget = vtkAddAnnotationWidget::New ();
  this->annotation_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->annotation_widget->EnabledOn ();


  ChipsId cid;
  this->populate_chips_id (cid);

  this->plot_rep = vtkChipsRepresentation::New ();

  this->plot_widget = vtkChipsPlotWidget::New ();
  this->plot_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->plot_widget->SetRepresentation (plot_rep);
  this->plot_widget->SetIdInfo (cid, this->get_object_type());
  this->plot_widget->RotateableOff ();
  this->plot_widget->SetDefaultRenderer (renderer);
  this->plot_widget->SetPriority (.95);
  this->plot_widget->EnabledOn ();
#endif

  this->recalculate_position ();

};
 

VtkPlot::~VtkPlot()
{
  if (title)
     title->object_delete ();

#ifndef HEADLESS
  if (this->bounding_box)
     this->bounding_box->Delete ();

  if (this->annotation_widget)
     this->annotation_widget->Delete ();

  if (this->plot_widget)
     this->plot_widget->Delete ();

  if (this->plot_rep)
     this->plot_rep->Delete ();
#endif
};


int VtkPlot::change_visible ()
{
  int tmp_vis = (!get_parent()->is_visible ()) ? 0 : is_visible();
  int curr_vis = visible;
  visible = tmp_vis && !offscreen;
  if (coord)
      coord->set_visible (visible);
  title->change_visible ();
  this->change_objects_visibility ();

#ifndef HEADLESS
   //this->enable_interaction ();
  this->bounding_box->GetRepresentation ()->SetVisibility (visible);
  this->annotation_widget->GetRepresentation ()->SetVisibility (visible);
  this->plot_rep->SetVisibility (visible);
#endif


  visible = curr_vis;  

  return ERR_STATUS_GOOD;
}

void VtkPlot::enable_picklimits (bool enabled)
{
#ifndef HEADLESS
  if (enabled)
  {
    this->bounding_box->Unblock ();
  }
  else
  {
    this->bounding_box->Block ();
  }
#endif
}

void VtkPlot::enable_addannotation (bool enabled)
{
#ifndef HEADLESS
  if (enabled)
  {
    this->annotation_widget->Unblock ();
  }
  else
  {
    this->annotation_widget->Block ();
  }
#endif
}


int VtkPlot::recalculate_position ()
{
  int rval = ChipsPlot::recalculate_position ();

#ifndef HEADLESS
  if (!rval)
  {
    ChipsVector min_bounds = this->coord->get_lower_bounds ();
    ChipsVector max_bounds = this->coord->get_upper_bounds ();
    vtkBoundingBoxRepresentation *bounding_box_rep = reinterpret_cast<vtkBoundingBoxRepresentation *>(this->bounding_box->GetRepresentation ());

    this->coord->convert_cs_to_cs(&min_bounds, &min_bounds, 1, FRAME_NORM, "", PIXEL_REAL, "");
    this->coord->convert_cs_to_cs(&max_bounds, &max_bounds, 1, FRAME_NORM, "", PIXEL_REAL, "");

    bounding_box_rep->SetMinBoundPos (&min_bounds[0]);
    bounding_box_rep->SetMaxBoundPos (&max_bounds[0]);

    // Frame Normalized is NOT a match for vtk viewport normalized
    // Make sure we have things matching up correctly

    this->renderer->DisplayToNormalizedDisplay(min_bounds[0], min_bounds[1]);
    this->renderer->NormalizedDisplayToViewport(min_bounds[0], min_bounds[1]);
    this->renderer->ViewportToNormalizedViewport(min_bounds[0], min_bounds[1]);

    this->renderer->DisplayToNormalizedDisplay(max_bounds[0], max_bounds[1]);
    this->renderer->NormalizedDisplayToViewport(max_bounds[0], max_bounds[1]);
    this->renderer->ViewportToNormalizedViewport(max_bounds[0], max_bounds[1]);

    vtkCoordinate* pos1 = this->plot_rep->GetPositionCoordinate ();
    vtkCoordinate* pos2 = this->plot_rep->GetPosition2Coordinate ();

    pos1->SetCoordinateSystemToNormalizedViewport ();
    pos2->SetCoordinateSystemToNormalizedViewport ();

    pos1->SetValue (min_bounds[0], min_bounds[1]);
    pos2->SetValue (max_bounds[0]-min_bounds[0], max_bounds[1]-min_bounds[1]);

    this->plot_rep->Modified();
  }
#endif



  return rval;
}

void VtkPlot::set_data_aspect_ratio (const ChipsId *cid) 
{
  if (!coord || !cid)
    return;

  string dsid = cid->get_coord_id();
  ChipsDataSystem *ds = NULL;

  string ax1 = cid->get_xaxis_id();
  string ax2 = cid->get_yaxis_id();
  ChipsAxis *ax = get_xaxis(ax1),
            *ay = get_yaxis(ax2);

  // first attempt to use the DATA coord sys id from the ChipsId 
  if (dsid.length() && cid->get_coord_sys() == DATA)
    ds=dynamic_cast<ChipsDataSystem *>(coord->get_coord(DATA,dsid));

  // didn't get the data sys yet, try using specified/current axes
  if (!ds)
    ds = get_datasystem (ax, ay);

  if (ds)
  {
#ifndef HEADLESS
    vtkBoundingBoxRepresentation *bounding_box_rep = reinterpret_cast<vtkBoundingBoxRepresentation *>(this->bounding_box->GetRepresentation ());
 
    bounding_box_rep->SetUseRatio (ds->use_ratio ());
#endif

    if (ds->use_ratio ())
    {
      string ratio = ds->get_aspect_ratio ();
      float aspect[2];
      int rval = string_to_ratio_values(ratio.c_str(), aspect[0], aspect[1]);
#ifndef HEADLESS
      if (!rval)
        bounding_box_rep->SetRatio (aspect);
#endif
    }
  }
}



ChipsCurve* VtkPlot::new_curve(long cnt, double* x, double* y, double* ydown,
                               double* yup, double* xdown, double* xup,
                               double* ylims, double* xlims,
                               const string & transform_id, const string & cvid)
{
   VtkCurve *curveData;
   string tmpid;

   if (!cvid.empty())
      tmpid = cvid;
   else
      tmpid = create_curve_id();

    curveData = new VtkCurve(renderer,this, cnt, x, y, 
                        ydown, yup, xdown, xup, ylims, xlims, transform_id, tmpid); 


   return curveData;
};


ChipsHistogram* VtkPlot::new_histogram(long cnt, double* lo, double* hi,
                                       double* counts, double* yd, double* yu, 
                                       const string& hid)
{
   string tmpid;

   if (!hid.empty())
      tmpid = hid;
   else
      tmpid = create_histogram_id();

   return new VtkHistogram(renderer, this, cnt, lo, hi, counts, yd, yu, tmpid);

  return NULL;
};


ChipsAxis* VtkPlot::new_axis(ChipsCoordinate *inCoord, COORD_SYS sys,
                              double inMinX, double inMinY, double inMaxX,
                              double inMaxY, int inOrientation,
                              ChipsAxisAlgorithmFactory::TransformType type,
                              const string& aid)
{
  string tmpid;

  if (!aid.empty())
    tmpid = aid;
  else
    tmpid = create_axis_id(inOrientation);

  VtkAxis *ax = new VtkAxis (renderer,inCoord,sys,inMinX,inMinY,inMaxX,inMaxY,inOrientation,type,this, tmpid);
  return ax;

};


ChipsAxis* VtkPlot::new_axis(ChipsCoordinate *inCoord, COORD_SYS sys,
			      double inMinX, double inMinY, double inMaxX,
                              double inMaxY, int inOrientation,
			      double inMin,double inMax,
                              ChipsAxisAlgorithmFactory::TransformType type,
                              const string& aid)
{
  string tmpid;

  if (!aid.empty())
    tmpid = aid;
  else
    tmpid = create_axis_id(inOrientation);

  VtkAxis *ax = new VtkAxis (renderer,inCoord,sys,inMinX,inMinY,inMaxX,inMaxY,inOrientation,inMin,inMax,type,this, tmpid);
  return ax;

};


ChipsAxis* VtkPlot::new_axis(ChipsStateStream& in)
{
  VtkAxis *tmp_axis = new VtkAxis (renderer, this, coord, coord_system, in);

  return tmp_axis;

};


ChipsImage* VtkPlot::new_image(ChipsStateStream& in)
{
  VtkImage* tmp_img =  new VtkImage(renderer, this, in);
  return tmp_img;
};

ChipsCurve* VtkPlot::new_curve(ChipsStateStream& in)
{
  VtkCurve* tmp_crv =  new VtkCurve(renderer, this, in);
  return tmp_crv;
};


ChipsContour* VtkPlot::new_contour(ChipsStateStream& in)
{
  VtkContour* tmp_cntr =  new VtkContour(renderer, this, in);
  return tmp_cntr;
};


ChipsHistogram* VtkPlot::new_histogram(ChipsStateStream& in)
{
  VtkHistogram* tmp_hst =  new VtkHistogram(renderer, this, in);
  return tmp_hst;
};


ChipsContour* VtkPlot::new_contour(long cnt, double* x, double* y, double* z,
				   unsigned int count, double *levels,
				   string transform_id, const string& ctid)
{
   string tmpid;

   if (ctid.length())
      tmpid = ctid;
   else
     tmpid = create_contour_id();
   return new VtkContour(renderer,this, cnt, x, y,z, count, levels, transform_id, tmpid);
   
};


ChipsLabel* VtkPlot::new_label(ChipsCoordinate *inCoord, COORD_SYS coord_sys,
      double xpos, double ypos, string text, string tid)
{
   string tmpid;

   if (tid.length())
      tmpid = tid;
   else
      tmpid = create_label_id();
   return new VtkLabel(renderer, this, tmpid, inCoord, coord_sys, xpos, ypos, text);
}

ChipsLabel* VtkPlot::new_label(ChipsStateStream& in)
{
   return new VtkLabel(renderer, this, coord, coord_system, in);
}


ChipsLine* VtkPlot::new_line(ChipsCoordinate *inCoord, COORD_SYS coord_sys,
   double startx, double starty, double endx, double endy, string lnid)
{
   string tmpid;

   if (lnid.length())
      tmpid = lnid;
   else
      tmpid = create_line_id();

   return new VtkLine(renderer, this, inCoord, coord_sys, startx,
		      starty,endx, endy, tmpid);
}


ChipsLine* VtkPlot::new_line(ChipsCoordinate *inCoord, COORD_SYS coord_sys,
   double* xx, double* yy, unsigned int count, string lnid)
{
   string tmpid;

   if (lnid.length())
      tmpid = lnid;
   else
      tmpid = create_line_id();
   return new VtkLine(renderer, this, inCoord, coord_sys, xx, yy, count, tmpid);
}

ChipsLine* VtkPlot::new_line(ChipsStateStream& in)
{
   return new VtkLine(renderer, this, coord, coord_system, in);
}


ChipsPoint* VtkPlot::new_point(ChipsCoordinate *inCoord, COORD_SYS coord_sys, double xx, double yy, string ptid)
{
   string tmpid;
   if (ptid.length())
      tmpid = ptid;
   else
      tmpid = create_point_id();

   return new VtkPoint(renderer, this, inCoord, coord_sys, xx, yy, tmpid);
}


ChipsImage* VtkPlot::new_image(string fname, string transform_id, string iid)
{
   string tmpid;

   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();

   return new VtkImage(renderer, this, fname, transform_id, tmpid);
}


ChipsImage* VtkPlot::new_image(string fname, float posx1, float posy1,
   float posx2, float posy2, string transform_id, string iid)
{
   string tmpid;

   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();

   return new VtkImage(renderer, this, fname, posx1, posy1, posx2,
                       posy2, transform_id, tmpid);
}


ChipsImage* VtkPlot::new_image(double* rvals, double *gvals, double *bvals, 
                               double *avals, int xaxis, int yaxis, 
                               string transform_id, string iid)
{
   string tmpid;

   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();

   return new VtkImage(renderer, this, xaxis, yaxis, rvals, gvals, bvals, 
                       avals, transform_id, tmpid);
}


ChipsPoint* VtkPlot::new_point(ChipsCoordinate *inCoord, COORD_SYS coord_sys, double xx, double yy, double zz, string ptid)
{
   string tmpid;

   if (ptid.length())
      tmpid = ptid;
   else      tmpid = create_point_id();

   return new VtkPoint(renderer, this, inCoord, coord_sys, xx, yy, zz, tmpid);
}

ChipsPoint* VtkPlot::new_point(ChipsStateStream& in)
{
   return new VtkPoint(renderer, this,  coord, coord_system, in);
}


ChipsRegion*  VtkPlot::new_polygon(ChipsCoordinate *inCoord, COORD_SYS sys,
                                    double* xx, double* yy, unsigned int count,
                                    string rid)
{
   string tmpid;
   if (rid.length())
      tmpid = rid;
   else
      tmpid = create_region_id();
   return new VtkPolygon(renderer, this, inCoord, sys, xx, yy, count, tmpid);

};

ChipsRegion*  VtkPlot::new_polygon(ChipsCoordinate *inCoord, COORD_SYS sys,
                                    int sides, double xx, double yy,
                                    double radius, string rid)
{
   string tmpid;

   if (rid.length())
      tmpid = rid;
   else
      tmpid = create_region_id();
   return new VtkPolygon(renderer, this, inCoord, sys, sides, xx, yy, radius, tmpid);
};

ChipsRegion*  VtkPlot::new_polygon(ChipsStateStream& in)
{
   return new VtkPolygon(renderer, this,  coord, coord_system, in);
}


ChipsColorBar* VtkPlot::new_colorbar(ChipsCoordinate *inCoord,
   COORD_SYS sys,  double xpos, double ypos, string tid)
{
   string tmpid;
   if (tid.length())
      tmpid = tid;
   else
       tmpid = create_cbar_id();
   return new VtkColorBar(renderer, this, inCoord, sys, xpos, ypos, tmpid);
}

ChipsColorBar* VtkPlot::new_colorbar(ChipsCoordinate *inCoord,
   COORD_SYS sys,  double xpos, double ypos, unsigned short cmap_id, string tid)
{
   string tmpid;
   if (tid.length())
      tmpid = tid;
   else
       tmpid = create_cbar_id();
   return new VtkColorBar(renderer, this, inCoord, sys, xpos, ypos, cmap_id, tmpid);
}

ChipsColorBar* VtkPlot::new_colorbar(ChipsStateStream& in)
{
   return new VtkColorBar(renderer, this, coord, coord_system, in);
}

void VtkPlot::select ()
{
#ifndef HEADLESS
  if (plot_widget)
    plot_widget->Select (true);
#endif
}

void VtkPlot::unselect ()
{
#ifndef HEADLESS
  if (plot_widget)
    plot_widget->Select (false);
#endif
}

void VtkPlot::disable_interaction ()
{
#ifndef HEADLESS
   this->plot_widget->EnabledOff ();
   this->bounding_box->EnabledOff ();
   this->annotation_widget->EnabledOff ();
   this->plot_rep->DisableBorder ();

   this->disable_object_interaction ("all", chips_label);
   this->disable_object_interaction ("all", chips_line);
   this->disable_object_interaction ("all", chips_point);
   this->disable_object_interaction ("all", chips_axis);
   this->disable_object_interaction ("all", chips_region);
   this->disable_object_interaction ("all", chips_colorbar);
   this->disable_object_interaction ("all", chips_curve);
   this->disable_object_interaction ("all", chips_histogram);
   this->disable_object_interaction ("all", chips_contour);
   this->disable_object_interaction ("all", chips_image);
#endif
}

void VtkPlot::enable_interaction ()
{
#ifndef HEADLESS
   if (this->is_interaction_blocked ())
      return;
   this->plot_widget->EnabledOn ();
   this->bounding_box->EnabledOn ();
   this->annotation_widget->EnabledOn ();

   this->plot_rep->EnableBorder ();

   this->enable_object_interaction ("all", chips_label);
   this->enable_object_interaction ("all", chips_line);
   this->enable_object_interaction ("all", chips_point);
   this->enable_object_interaction ("all", chips_axis);
   this->enable_object_interaction ("all", chips_region);
   this->enable_object_interaction ("all", chips_colorbar);
   this->enable_object_interaction ("all", chips_curve);
   this->enable_object_interaction ("all", chips_histogram);
   this->enable_object_interaction ("all", chips_contour);
   this->enable_object_interaction ("all", chips_image);
#endif

}

