// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkPlotLabel.hh"
#include "VtkFrame.hh"
#include "vtkCoordinate.h"
#include "vtkCamera.h"
#include "vtkRenderWindow.h"

#include "vtkTextActorExtra.hh"
#include "ChipsCoordinate.hh"

#ifndef HEADLESS
#include "vtkChipsTextWidget.hh"
#include "vtkChipsTextRepresentation.hh"
#endif

#include "ChipsId.hh"
#include "ChipsPlot.hh"

#include "VtkUtils.hh"


VtkPlotLabel::VtkPlotLabel(vtkRenderer* inRenderer,ChipsObject* par, 
                  const string& lbid, double inXPos, double inYPos,
                  const string& val):ChipsLabel(par,lbid,NULL,COORD_NONE,
                  inXPos,inYPos,val)
{
  renderer = inRenderer;

  vtkText=vtkTextActorExtra::New();
  AddFormattedText(format,vtkText, get_scale_factor ());
  vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
  vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();

#ifndef HEADLESS
  ChipsId cid;
  this->populate_chips_id (cid);

  vtkChipsTextRepresentation *label_rep = vtkChipsTextRepresentation::New ();
  label_rep->SetTextActor (vtkText);
  label_rep->SetMaxFontSize (V_SIZE_MAX_VAL);


  this->label_widget = vtkChipsTextWidget::New ();
  this->label_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor ());
  this->label_widget->SetRepresentation (label_rep);
  this->label_widget->SetIdInfo (cid, chips_plot, chips_plot_title);
  this->label_widget->SetDefaultRenderer (renderer);
  this->label_widget->EnabledOn ();
  label_rep->Delete ();
#endif

  change_color();
  change_rotation();
  change_alignment();
  renderer->AddActor(vtkText);


};

VtkPlotLabel::VtkPlotLabel(vtkRenderer* inRenderer,ChipsObject* par, ChipsStateStream& in):
   ChipsLabel(par, NULL, COORD_NONE, in)
{
  renderer = inRenderer;

  vtkText=vtkTextActorExtra::New();
  AddFormattedText(format,vtkText, get_scale_factor ());
  vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
  vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();
  renderer->AddActor(vtkText);

#ifndef HEADLESS
  vtkChipsTextRepresentation *label_rep = vtkChipsTextRepresentation::New ();
  label_rep->SetTextActor (vtkText);
  label_rep->SetMaxFontSize (V_SIZE_MAX_VAL);

  ChipsId cid;
  this->populate_chips_id (cid);

  this->label_widget = vtkChipsTextWidget::New ();
  this->label_widget->SetInteractor (renderer->GetRenderWindow ()->GetInteractor
 ());
  this->label_widget->SetRepresentation (label_rep);
  this->label_widget->SetIdInfo (cid, this->get_object_type());
  this->label_widget->SetDefaultRenderer (renderer);
  this->label_widget->EnabledOn ();
  label_rep->Delete ();
#endif

  change_color();
  change_rotation();
  change_alignment();
  change_visible ();


};


VtkPlotLabel::~VtkPlotLabel()
{
  if ( renderer)
    {
      renderer->RemoveActor(vtkText);
      vtkText->Delete();
    }

#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Delete ();
#endif
};

int VtkPlotLabel::redraw()
{

  change_color();
  change_rotation();
  change_alignment();
  return ERR_STATUS_GOOD;
};

int VtkPlotLabel::change_text()
{  

  if (vtkText)
    {
      vtkText->Clear();
      AddFormattedText(format,vtkText,get_scale_factor ());
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif
    }
  else
    return ERR_STATUS_BAD ;


  return ERR_STATUS_GOOD;
};

int VtkPlotLabel::change_font()
{
  return change_text();

};

int VtkPlotLabel::change_size()
{ 
  return change_text();
  
};


int VtkPlotLabel::change_depth ()
{
  // Axes will handle depth themselves
  if ( vtkText )
    {
      vtkText->SetLayerNumber (depth*10+1);
    }
  return ERR_STATUS_GOOD;
};

int VtkPlotLabel::change_scale_factor()
{ 
  return change_text();
  
};

int VtkPlotLabel::change_position()
{
  int rval = ERR_STATUS_GOOD;
  if ( renderer)
    {
      ChipsVector tmp_orig (orig);
      vtkText->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();
      vtkText->GetPosition2Coordinate()->SetCoordinateSystemToNormalizedViewport();

      VtkUtils::FrameNorm2ViewportNorm (parent->get_chips_coord (), renderer, &tmp_orig[0]);

      vtkText->GetPositionCoordinate()->SetValue(tmp_orig[0], tmp_orig[1]);

      vtkText->Modified();

#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif
    }

  return rval;
};


int VtkPlotLabel::change_color()
{
  return change_text();
};

int VtkPlotLabel::change_visible ()
{
   int tmp_vis = (!get_parent()->is_visible ()) ? 0 : is_visible();

   if (vtkText)
     vtkText->SetVisibility (tmp_vis);

#ifndef HEADLESS
   if (this->label_widget && tmp_vis)
   {
       //this->label_widget->EnabledOn ();
       this->label_widget->GetRepresentation ()->SetVisibility (tmp_vis);
   }
#endif

   return ERR_STATUS_GOOD;

}



int VtkPlotLabel::change_rotation()
{
  if (vtkText)
    {
      vtkText->GetTextProperty()->SetOrientation(angle);
      vtkText->Modified();
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif

    }

  return ERR_STATUS_GOOD;
};


int VtkPlotLabel::change_alignment()
{
  if ( vtkText)
    {
      vtkText->SetVertAlignmentOffset(vert_alignment);
      vtkText->SetHorizAlignmentOffset(horiz_alignment);
      vtkText->Modified();
#ifndef HEADLESS
      if (this->label_widget)
        this->label_widget->UpdateRepresentation ();
#endif

    }

  return ERR_STATUS_GOOD;
};


void VtkPlotLabel::select ()
{
#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Select (true);
#endif
}

void VtkPlotLabel::unselect ()
{
#ifndef HEADLESS
  if (this->label_widget)
    this->label_widget->Select (false);
#endif
}


int VtkPlotLabel::compute_resize( double bounds[])
{
  // bounds should be in display coordinates

  double outPosition[2];
  ChipsId cid;

  ChipsPlot *plot_parent = dynamic_cast <ChipsPlot *>(this->parent);

  if (!plot_parent)
    return ERR_STATUS_BAD;

  this->populate_chips_id (cid);
  cid.set_coord_sys (PIXEL_REAL);
  cid.set_coord_id ("");

  int font = ComputeFontSizeAndPosFromBBox (this->vtkText, renderer, bounds, outPosition);
  int rval = plot_parent->set_property(&cid, static_cast<void*>(&font), SET_PLOT_TITLE_SIZE);

  if (!rval)
    plot_parent->move_object (&cid, chips_plot, chips_plot_title, outPosition[0], outPosition[1], V_MOVE_ABSOLUTE);

  return rval;

}

void VtkPlotLabel::disable_interaction ()
{
#ifndef HEADLESS
  vtkChipsTextRepresentation *rep = reinterpret_cast<vtkChipsTextRepresentation *> (this->label_widget->GetRepresentation ());
  rep->DisableBorder ();
  this->label_widget->EnabledOff ();
#endif
}

void VtkPlotLabel::enable_interaction ()
{
#ifndef HEADLESS
  if (this->is_interaction_blocked ())
	return;

  vtkChipsTextRepresentation *rep = reinterpret_cast<vtkChipsTextRepresentation *> (this->label_widget->GetRepresentation ());

  this->label_widget->EnabledOn ();
  rep->EnableBorder ();
#endif
}

