// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "VtkWindow.hh"

#include <stdio.h>
#include <unistd.h>
#include <sstream>

#include "ChipsDefaults.hh"
#include "ChipsWindowSystem.hh"
#include "VtkEngine.hh"

#ifndef HEADLESS
#include "VtkGtkWindow.hh"
#endif

#include "vtkRenderer.h"
#include "vtkRendererCollection.h"
#include "vtkMapper.h"

#include "vtkWindowToImageFilter.h"      // rasterized support (png, jpg)
#include "vtkPNGWriter.h"                // png support 
#include "vtkJPEGWriter.h"               // jpg support 
#include "vtkGL2PSExporter.h"            // ps, eps, pdf support 
#include "vtkImageData.h"
#include "vtkImageMapper.h"
#include "vtkCamera.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkPoints.h"
#include "vtkCellArray.h"
#include "vtkProperty2D.h"
#include "vtkIdList.h"
#include "vtkCoordinate.h"

#include "vtkXOpenGLOffScreenRenderWindow.h"

extern ChipsDefaults* theDefaults; 

VtkWindow::VtkWindow(ChipsEngine* en, string val):ChipsWindow(en, val)
{
   initialize ();
#ifndef HEADLESS
   this->gui_window->setEventSafe(true);
#endif

}

VtkWindow::VtkWindow(ChipsEngine* en, ChipsStateStream& in):ChipsWindow(en, in)
{
   initialize ();
#ifndef HEADLESS
   this->gui_window->setEventSafe(true);
#endif
}

VtkWindow::VtkWindow(ChipsEngine* en, float ww, float hh, string format, string val):ChipsWindow(en, ww, hh, format, val)
{
   initialize ();
#ifndef HEADLESS
   this->gui_window->setEventSafe(true);
#endif

}

VtkWindow::~VtkWindow()
{
   base_renderer->Delete();

   if (redraw_renderer)
      redraw_renderer->Delete ();

   if (backdrop_actor)
      backdrop_actor->Delete ();

   if (renWin)
      renWin->Delete();

};

void VtkWindow::initialize ()
{
   remapping = false;

   redraw_renderer = NULL;
   backdrop_actor = NULL;
   
   renWin = NULL;

   base_renderer = vtkRenderer::New();
   change_color();    // set default window background color 
   base_renderer->SetViewport (0, 0, width, height);

   renWin = vtkRenderWindow::New();

#ifndef HEADLESS
   string tmp_geom = "";
   if (gui_window)
      delete gui_window;
   if (theDefaults)
      tmp_geom = theDefaults->get_windowgeom();

   initialize_gui (tmp_geom);
   gui_window->set_default_size (width, height);

   renWin->SetWindowId(gui_window->get_xwindow ());
#endif

   renWin->SetSize(width,height);

   renWin->SetNumberOfLayers (1);
   renWin->AddRenderer (base_renderer);
   if (smoothing)   renWin->LineSmoothingOn();

   // setup grid
   this->gridData = vtkPolyData::New ();


   vtkCoordinate *coordinate = vtkCoordinate::New ();
   coordinate->SetCoordinateSystemToDisplay ();

   this->gridMapper = vtkPolyDataMapper2D::New();
   this->gridMapper->SetInput(this->gridData);
   this->gridMapper->SetTransformCoordinate (coordinate);

   this->gridActor = vtkActor2D::New();
   this->gridActor->SetMapper(this->gridMapper);


   this->render_grids ();

   this->base_renderer->AddActor (this->gridActor);


   if (!activated)
   {
      activate();
      activated = true;
   }

   change_grid_color ();
   change_grid_thickness ();
   change_grid_linestyle ();

}

void VtkWindow::set_window_size()
{
  ChipsWindow::set_window_size();
  renWin->SetSize(width,height);

  if (redraw_renderer)
     this->draw_backdrop ();

};

void VtkWindow::set_window_size(int inPixelWidth,int inPixelHeight)
{
  ChipsWindow::set_window_size(inPixelWidth, inPixelHeight);
   renWin->SetSize(width,height);

   if (redraw_renderer)
     this->draw_backdrop ();

};

int VtkWindow::do_print(string outfile, int fmt, int colorSpace, bool add_ext)
{
   int rval = ERR_STATUS_GOOD; 
   bool printDev = false;
   bool margins_set = false;
   bool pagesize_set = false;
// Cant use offscreen renderer because of a bug in apple x-client
   vtkRenderWindow* print_renWin = NULL;
   vtkMapper *vm = NULL;
   int leftmargin_p;
   int rightmargin_p;
   int topmargin_p;
   int bottommargin_p;
   int page_width_p;
   int page_height_p;
   int margin_width;
   int margin_height;
   float conv = conversion_factor (page_units);
   int orig_width = width;
   int orig_height = height;
   string local_printmethod = this->printmethod;

   bool display_changed = false;
   bool cleanup_print_renwin = false;
   string tmp_outfile;           // output file name for pdf

   // Create file for chips to write out when printing to printer
   // and when writing to pdf
   if ((outfile.length() == 0))
   {
      char buf[128];
      printDev = true;

      sprintf(buf, "/tmp/.chips.%d", getpid());
      outfile = buf;

   }

   set_default_colors_ps();

   if (scale_width && scale_height)
   {
      // set size back to true ps size (remove scale)
      width=(int)(width/scale_width);
      height=(int)(height/scale_height);
      ((ChipsWindowSystem *)coord)->set_scale (1,1);

      display_changed = true;
   }


   // Only set margines and pagesize if using PS or PDF formats
   if ((fmt == 0) || (fmt == 2))
   {
      if (page_width && page_height){

         // if page units were not set. use the current window units
         if (page_units != chips_nounit)
            conv = conversion_factor (page_units);
         else
            conv = conversion_factor (units);

         page_width_p = (int)(page_width * conv);
         page_height_p = (int)(page_height * conv);
      }
      else{
         if (export_orient){
            page_width_p = height;
            page_height_p = width;
         }
         else{
            page_width_p = width;
            page_height_p = height;
         }
         conv = conversion_factor (units);
      }
      pagesize_set = true;

	

      if (!fit_to_page){
         if (page_scale_width)
            width = (int)(page_scale_width*width);
         if (page_scale_height)
            height = (int)(page_scale_height*height);
      }

      if (leftmargin || rightmargin || topmargin || bottommargin){
         margins_set = true;
         leftmargin_p = (int)(leftmargin * conv);
	 bottommargin_p = (int)(bottommargin * conv);
	 rightmargin_p = (int)(rightmargin * conv);
	 topmargin_p = (int)(topmargin * conv);

         if (export_orient){
	    margin_width = (pagesize_set) ? page_height_p : height+leftmargin_p;
            margin_height = (pagesize_set) ? page_width_p : width+bottommargin_p;

         }
         else{
            margin_width = (pagesize_set) ? page_width_p : width+leftmargin_p;
	    margin_height = (pagesize_set) ? page_height_p : height+bottommargin_p;
	 }
	 margin_width -= (rightmargin_p+leftmargin_p);
         margin_height -= (topmargin_p+bottommargin_p);

         if ((margin_width < 0) || (margin_height < 0)){
	    width = orig_width;
            height = orig_height;
	    redraw ();
	    return ERR_MARGIN_VS_PAGE;
	 }
      }
   }


   if ((margins_set || pagesize_set) && fit_to_page){
      // fit the page to be the size within the margins if they 
      // exist or to be the page size
      int tmp_width, tmp_height;
      if (margins_set){
	 tmp_width = margin_width;
	 tmp_height = margin_height;
      }
      else{
	 tmp_width = (export_orient) ? page_height_p : page_width_p;
	 tmp_height = (export_orient) ? page_width_p : page_height_p;
      }

      // Keep the aspect ratio of the window
      if (keep_aspect)
      {
	 if (tmp_width/(float)width > tmp_height/(float)height){
	    width =  (int)(tmp_height* width/(float)height);
	    height = tmp_height;
	 }
	 else{ 
	    height = (int)(tmp_width * height/(float)width);
            width = tmp_width;
         }
      }
      else {
	 width = tmp_width;
	 height = tmp_height;
      }


      display_changed = true;
   }

   if ((fmt == 0) || (fmt == 2)){
      // If the pagesize was not set use the width and height
      // of the window after all modifications have been made.
      if (!pagesize_set){
         if (export_orient){
            page_width_p = height;
            page_height_p = width;
         }
         else{
            page_width_p = width;
            page_height_p = height;
         }
         pagesize_set = true;
      }
   }


   // Since the offscreen renderer cant be used currently only write to
   // the back buffer to minimalize some of the display changes.
//   renWin->SetSwapBuffers(0);

   // turn off grids if they were on
   this->gridActor->SetVisibility (0);

   // disable widgets while printing
   this->disable_all_interaction ();

   // Remove the renderer before printing. This renderer contains an image
   // of that last things rendered before redraw was turned off
   if (redraw_renderer)
      renWin->RemoveRenderer (redraw_renderer);


   if (display_changed){
      update_display();

      ((ChipsWindowSystem *)coord)->set_dim (width, height);

      // Change the size of the render window to fit the new width
      // This is NOT necessary if we do offscreen rendering
/*      renWin->SetSize(width,height);
*/

   }

   if ((export_dpi != 72) && (fmt != 3) && (fmt != 4) && (fmt != 5))
      update_print_scale (export_dpi/72.0);


   // if the window hasn't been drawn then it has to be copied over
   // to an offscreen buffer for printing
   if (!this->drawready)
      local_printmethod = "xoffscreen";

   if (local_printmethod == "framebuffer")
   {
      print_renWin = vtkRenderWindow::New ();
      print_renWin->OffScreenRenderingOn ();
      copy_render_window (renWin, print_renWin);
      cleanup_print_renwin = true;
   }
   else if (local_printmethod == "xoffscreen")
   {
      print_renWin = vtkXOpenGLOffScreenRenderWindow::New ();
      print_renWin->OffScreenRenderingOn ();
      copy_render_window (renWin, print_renWin);
      cleanup_print_renwin = true;
   }
   else
   {
      print_renWin = renWin;
      cleanup_print_renwin = false;
      print_renWin->SetSize(width,height);
   }

   vm = (vtkMapper*)vtkMapper::New();
   vm->SetGlobalImmediateModeRendering (1);


   vtkGL2PSExporter *psExporter = NULL;
   switch(fmt)
   {
      case 3:   // PNG
      {
         if (add_ext)
            outfile.append(".png");
         vtkWindowToImageFilter *w2if = vtkWindowToImageFilter::New();
         w2if->SetInput(print_renWin);
         vtkPNGWriter *pngw = vtkPNGWriter::New();
         pngw->SetInput(w2if->GetOutput());
         pngw->SetFileName(outfile.c_str());
         pngw->Write();
         pngw->Delete();
	 w2if->Delete();
      }
      break; 
      case 4:   // JPG  -- fallthrough intended
      case 5:   // JPEG
      {
         if (add_ext && (fmt==4))
            outfile.append(".jpg");
         if (add_ext && (fmt==5))
            outfile.append(".jpeg");
         vtkWindowToImageFilter *w2if = vtkWindowToImageFilter::New();
         w2if->SetInput(print_renWin);
         vtkJPEGWriter *jpgw = vtkJPEGWriter::New();
         jpgw->SetInput(w2if->GetOutput());
         jpgw->SetFileName(outfile.c_str());
         jpgw->Write();
         jpgw->Delete();
         w2if->Delete();
      }
      break; 
      case 1:   // EPS
         if (!add_ext)
         {
            int ext_pos = outfile.find_last_of (".", outfile.length ());
            outfile = outfile.substr(0,ext_pos);
         }
         psExporter = vtkGL2PSExporter::New();
         psExporter->SetRenderWindow(print_renWin);
         psExporter->SetFileFormatToEPS();
         psExporter->SetSilent(1);
         psExporter->SetCompress(0);
         psExporter->SetSort(0);
         psExporter->SetFilePrefix(outfile.c_str());
//         psExporter->SetLandscape(export_orient);
	 if ( export_font)
	   psExporter->SetFontDataImport(1);
	 else
	   psExporter->SetFontDataImport(0);
	 psExporter->SetColorSpace(colorSpace);
         psExporter->SetImageDPI (export_dpi);
         psExporter->Write();
         psExporter->Delete();
      break; 
      case 2:   // PDF
         if (!add_ext)
         {
            int ext_pos = outfile.find_last_of (".", outfile.length ());
            outfile = outfile.substr(0,ext_pos);
         }
         psExporter = vtkGL2PSExporter::New();
         psExporter->SetRenderWindow(print_renWin);
         psExporter->SetFileFormatToPDF();
         psExporter->SetSilent(1);
         psExporter->SetCompress(0);
         psExporter->SetSort(0);
         psExporter->SetFilePrefix(outfile.c_str());
         psExporter->SetLandscape(export_orient);
	 if ( export_font)
	   psExporter->SetFontDataImport(1);
	 else
	   psExporter->SetFontDataImport(0);
	 psExporter->SetColorSpace(colorSpace);
	 if (margins_set){
	   psExporter->SetPageMargins (leftmargin_p, bottommargin_p,
						rightmargin_p, topmargin_p);
	   psExporter->SetStartPosition (leftmargin_p, bottommargin_p);
	 }
	 if (pagesize_set)
	   psExporter->SetPageSize (page_width_p, page_height_p);
         psExporter->SetImageDPI (export_dpi);
         psExporter->Write();
         psExporter->Delete();
      break; 
      case 0:   // PS - fallthrough intended 
      default:
         if (!add_ext)
         {
            int ext_pos = outfile.find_last_of (".", outfile.length ());
            outfile = outfile.substr(0,ext_pos);
         }

         psExporter = vtkGL2PSExporter::New();
         psExporter->SetRenderWindow(print_renWin);

         psExporter->SetFileFormatToPS();
         psExporter->SetSilent(1);
         psExporter->SetCompress(0);
         psExporter->SetSort(0);
         psExporter->SetFilePrefix(outfile.c_str());
         psExporter->SetLandscape(export_orient);

	 if ( export_font)
	   psExporter->SetFontDataImport(1);
	 else
	   psExporter->SetFontDataImport(0);
	 psExporter->SetColorSpace(colorSpace);
	 if (margins_set){
           psExporter->SetPageMargins (leftmargin_p, bottommargin_p,
                                                rightmargin_p, topmargin_p);
           psExporter->SetStartPosition (leftmargin_p, bottommargin_p);
         }
         if (pagesize_set){
           psExporter->SetPageSize (page_width_p, page_height_p);
	 }

         psExporter->SetImageDPI (export_dpi);

         psExporter->Write();
         psExporter->Delete();
      break; 
   }

   if (printDev) {

      // make sure we can access the temp file to print
      // png and jpg already have extensions in name
      if (fmt == 0)    outfile.append (".ps");
         else if (fmt == 1)  outfile.append (".eps");
            else if (fmt == 2)  outfile.append (".pdf");

      if (access(outfile.c_str(), R_OK) == 0) 
      {
         string cmd;
         if (printDev){
	    cmd = " lpr " + outfile;
	    system(cmd.c_str());
	 }
         // now remove the file 
         remove(outfile.c_str());
      } 
      else   
         rval = ERR_STATUS_BAD; 
   }

   print_mode = false;

   vm->SetGlobalImmediateModeRendering (0);
   vm->Delete ();

   if (display_changed)
   {
      // reset size back to scaled window size
      width=orig_width;
      height=orig_height;

      ((ChipsWindowSystem *)coord)->set_scale (scale_width,scale_height);
      ((ChipsWindowSystem *)coord)->set_dim ((int)(width/scale_width), (int)(height/scale_height));

      update_display();

      // if the display had changed make sure the sub components are
      // redrawn with the correct values (ie if the print_window size 
      // does not match the render window size then objects need to be
      // redrawn to match
      renWin->Modified ();

      // Change the size of the render window to fit the new width
      // This is NOT necessary if we do offscreen rendering
      /*renWin->SetSize(width,height);
       */
   }


   // if the render window is new object delete it
   if (cleanup_print_renwin)
      print_renWin->Delete ();
   else
      print_renWin->SetSize(width,height);

   if ((export_dpi != 72) && (fmt != 3) && (fmt != 4) && (fmt != 5))
      update_print_scale (1);

   // Add the redraw renderer back to the window
   if (redraw_renderer)
      renWin->AddRenderer (redraw_renderer);

   // turn grids back on if they were on
   change_grid_visible ();

   // disable widgets while printing
   this->enable_all_interaction ();

   renWin->SetSwapBuffers(1);
   set_default_colors_display();
   redraw();
 

   return rval; 
};

int VtkWindow::redraw()
{
   if (drawready && !redraw_blocked())
      renWin->Render();
   
   return ERR_STATUS_GOOD;
};


int VtkWindow::render_layout()
{

  return ERR_STATUS_BAD;
}; 

int VtkWindow::change_layout_color()
{

  return ERR_STATUS_BAD;
}; 

int VtkWindow::change_layout_thickness()
{

  return ERR_STATUS_BAD;
}; 

int VtkWindow::render_grids()
{

   float rowSpacing = 0;
   float colSpacing = 0;
   int numRows = 0;
   int numCols = 0;

   update_grid_properties();

   switch(gscale)
   {
      case V_GRID_SCALE_HALF_INCH:
      {
         V_UNITS_TYPE units = chips_inch;
         colSpacing = (conversion_factor (units)/2.)*scale_width;
         rowSpacing = (conversion_factor (units)/2.)*scale_height;
      }
      break;
      case V_GRID_SCALE_DEFAULT:   // fallthrough intended
      case V_GRID_SCALE_QTR_INCH:
      {
         V_UNITS_TYPE units = chips_inch;
         colSpacing = (conversion_factor (units)/4.)*scale_width;
         rowSpacing = (conversion_factor (units)/4.)*scale_height;

      }
      break;
      case V_GRID_SCALE_CENTIMETER:
      {
         V_UNITS_TYPE units = chips_cm;
         colSpacing = (conversion_factor (units))*scale_width;
         rowSpacing = (conversion_factor (units))*scale_height;

      }
      break;
      case V_GRID_SCALE_NORMALIZED:
         colSpacing = width / 10.;
         rowSpacing = height / 10.;
      break;
   }


   if ((colSpacing != 0) && (rowSpacing != 0))
   {
      vtkPoints *points = vtkPoints::New ();
      vtkCellArray *cells = vtkCellArray::New ();
      vtkIdList *pointIds = vtkIdList::New ();
      int pointId = 0;

      numRows = ceil (height / rowSpacing);
      numCols = ceil (width / colSpacing);

      points->Allocate ((numRows+1)*(numCols+1));

      for (int row=0; row <= numRows; row++)
         for (int col=0; col <= numCols; col++)
         {
            double y = row*rowSpacing;
            double x = col*colSpacing;
            if (y >= height)
               y = height - .05;
            else if (y <= 0)
               y = .05;

            if (col*colSpacing >= width)
               x = width - .05;
            else if (x <= 0)
               x = .05;
            
            points->InsertNextPoint (x,y,0);
         }

            

      cells->Allocate (1, numRows*numCols);
      pointIds->SetNumberOfIds (3);

      // compute interior
      for (int row=1; row < numRows; row++)
         for (int col=0; col < numCols-1; col++)
         {
            pointId = col + (row * (numCols+1));
            pointIds->SetId (0,pointId);
            pointIds->SetId (1,pointId+1);
            pointIds->SetId (2,pointId+numCols+2);

            cells->InsertNextCell(pointIds);
         }

      // compute right edge 
      pointIds->SetNumberOfIds (2);
      for (int row=1; row < numRows; row++)
      {
            pointId = numCols - 1 + (row * (numCols+1)) ;
            pointIds->SetId (0,pointId);
            pointIds->SetId (1,pointId+1);

            cells->InsertNextCell(pointIds);
      }

      // compute buttom edge
      for (int col=1; col < numCols; col++)
      {
            pointId = col;
            pointIds->SetId (0,pointId);
            pointIds->SetId (1,pointId+numCols+1);

            cells->InsertNextCell(pointIds);
      }


      this->gridData->SetPoints (points);
      this->gridData->SetLines (cells);

      points->Delete ();
      cells->Delete ();
      pointIds->Delete ();
   }
   return ERR_STATUS_GOOD;
}; 

int VtkWindow::change_grid_color()
{
  this->gridActor->GetProperty ()->SetColor (grid_color.red (), grid_color.green (), grid_color.blue ());
  return ERR_STATUS_GOOD;
}; 

int VtkWindow::change_grid_thickness()
{
  this->gridActor->GetProperty()->SetLineWidth(grid_thickness);
  return ERR_STATUS_GOOD;
}; 

int VtkWindow::change_grid_visible ()
{
  this->gridActor->SetVisibility (this->vis_grid);
  return ERR_STATUS_GOOD;
}

int VtkWindow::change_grid_linestyle ()
{
   switch (grid_linestyle){
      case chips_noline:
         this->gridActor->SetVisibility(0);
         break;
      case chips_dotlongdash:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_dotlongdash_pattern);
         break;
      case chips_dotshortdash:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_dotshortdash_pattern);
         break;
      case chips_longdash:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_longdash_pattern);
         break;
      case chips_shortdash:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_shortdash_pattern);
         break;
      case chips_shortdashlongdash:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_shortdashlongdash_pattern);
         break;
      case chips_dot:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_dot_pattern);
         break;
      case chips_solid:
      default:
         this->gridActor->SetVisibility(this->vis_grid);
         this->gridActor->GetProperty()->SetLineStipplePattern(chips_solid_pattern);
         break;

   };

   return ERR_STATUS_GOOD;
}




int VtkWindow::change_order()
{
  return ERR_STATUS_BAD;

};


int VtkWindow::change_color()
{
   if (base_renderer)
      base_renderer->SetBackground(color.red(), color.green(), color.blue());

   return 0;
}

int VtkWindow::change_size ()
{
   /* NOT TESTED YET 
   renWin->SetSize(width,height);
   if (base_renderer)
      base_renderer->SetViewport (0, 0, width, height);
   */
   return ERR_STATUS_GOOD;
}

int VtkWindow::change_redraw ()
{

#ifndef HEADLESS
   if (this->drawready && !this->remapping)
   {

      // Remove the redraw renderer. Either it we no longer
      // have redraw off or we'll add a new one.
      if (redraw_renderer)
      {
         renWin->RemoveRenderer (redraw_renderer);
	 renWin->SetNumberOfLayers (renWin->GetNumberOfLayers()-1);
	 redraw_renderer->Delete ();
	 backdrop_actor->Delete ();
         backdrop_actor = NULL;
	 redraw_renderer = NULL;
      }


      if (!redraw_on)
      {
	 // The redraw renderer is necessary because we need
         // a viewport that covers the entire window so when
         // the raster position is set it's set to one that
         // is in window coordinates. This is not implicitly done by 
         // vtk. The raster position is necessary for SetRGBAPixelData
	 redraw_renderer = vtkRenderer::New();
         redraw_renderer->SetViewport (0, 0, width, height);
	 redraw_renderer->SetLayer (renWin->GetNumberOfLayers());

         renWin->SetSwapBuffers(0);
         if (!redraw_blocked())
	    renWin->Render ();

         // Add backdrop to renderer
	 backdrop_actor = vtkActor2D::New ();
         this->draw_backdrop ();
         backdrop_actor->GetProperty()->SetColor (color.red(), color.green(), color.blue());
         redraw_renderer->AddActor (backdrop_actor);

         // Add Image to renderer
	 vtkWindowToImageFilter *img_filter = vtkWindowToImageFilter::New();

         // Convert vtk actor to images
         img_filter->SetReadFrontBuffer (0);
         img_filter->SetInputBufferTypeToRGB ();
         img_filter->SetInput (renWin);

         img_filter->Update ();
	 
         vtkImageMapper *img_mapper = vtkImageMapper::New ();
         img_mapper->SetInput (img_filter->GetOutput ());
	 img_mapper->SetColorWindow( 255);
  	 img_mapper->SetColorLevel( 127.5);

         vtkActor2D *img_actor = vtkActor2D::New ();
         img_actor->SetMapper (img_mapper);
	 img_actor->GetPositionCoordinate()->SetCoordinateSystemToNormalizedViewport();

         img_actor->SetPosition(0.0,0.0);
         img_actor->SetPosition2(1.0,1.0);
         img_actor->SetLayerNumber (1);

         redraw_renderer->AddActor (img_actor);

	 renWin->SetNumberOfLayers (renWin->GetNumberOfLayers()+1);
         renWin->AddRenderer (redraw_renderer);


	 img_actor->Delete ();
	 img_mapper->Delete ();
	 img_filter->Delete ();
      }

      renWin->SetSwapBuffers(1);
   }
#endif 
   return ERR_STATUS_GOOD; 
}


ChipsFrame* VtkWindow::new_frame(string fid)
{
   string tmpid;
   int renderer_pos;
   if (fid.length())
      tmpid = fid;
   else
      tmpid = create_frame_id();

   // If the redraw is off the redraw renderer will be active
   // This renderer should always be on top
   if (redraw_renderer)
      renderer_pos = renWin->GetNumberOfLayers() - 1;
   else
      renderer_pos = renWin->GetNumberOfLayers();

   VtkFrame *frame = new VtkFrame (renWin, this, renderer_pos, tmpid);

   frame->add_grid (this->gridActor);
   return frame;

}

ChipsFrame* VtkWindow::new_frame(ChipsStateStream& in)
{
    int renderer_pos;

    // If the redraw is off the redraw renderer will be active
    // This renderer should always be on top
    if (redraw_renderer)
      renderer_pos = renWin->GetNumberOfLayers() - 1;
    else
      renderer_pos = renWin->GetNumberOfLayers();

    VtkFrame *frame = new VtkFrame (renWin, this, renderer_pos, in);

    frame->add_grid (this->gridActor);
    return frame;


}

ChipsFrame* VtkWindow::new_frame(float xmin, float ymin, float xmax,
      float ymax, string fid)
{
   string tmpid;
   int renderer_pos;

   if (fid.length())
      tmpid = fid;
   else
      tmpid = create_frame_id();

   // If the redraw is off the redraw renderer will be active
   // This renderer should always be on top
   if (redraw_renderer)
      renderer_pos = renWin->GetNumberOfLayers() - 1;
   else
      renderer_pos = renWin->GetNumberOfLayers();


   VtkFrame *frame = new VtkFrame(renWin, this, renderer_pos, xmin, ymin, xmax, ymax, tmpid);

   frame->add_grid (this->gridActor);
   return frame;
}

void VtkWindow::copy_render_window (vtkRenderWindow* renWin1, vtkRenderWindow* renWin2)
{
   vtkRenderer *renderer1, *renderer2;
   vtkProp *prop;
   vtkPropCollection *prop_collection;
   vtkRendererCollection *renderer_collection = renWin1->GetRenderers();
   bool init = true;

   vtkMapper *mapper = (vtkMapper*)vtkMapper::New();
   renWin2->SetSize(width,height);

   mapper->SetGlobalImmediateModeRendering (1);

   // Create new renderers for each of the renderers in
   // this render window. Then copy over each actor.
   renWin2->SetNumberOfLayers(renWin1->GetNumberOfLayers ());
   for (renderer_collection->InitTraversal(); (renderer1 = renderer_collection->GetNextItem());)
   {
  
      renderer2 = vtkRenderer::New ();
      if (init)
      {
          renderer2->SetBackground(color.red(), color.green(), color.blue());
          init = false;
      }
      renderer2->SetViewport(renderer1->GetViewport ());

      prop_collection = renderer1->GetViewProps();
      for (prop_collection->InitTraversal(); (prop = prop_collection->GetNextProp()); )
         renderer2->AddViewProp (prop);

      renderer2->SetLayer (renderer1->GetLayer ());
      renWin2->AddRenderer(renderer2);

      renderer2->Delete ();

   }

   mapper->SetGlobalImmediateModeRendering (0);
   mapper->Delete ();
}


void VtkWindow::move_render_window (vtkRenderWindow* renWin1, vtkRenderWindow* renWin2)
{
   vtkRenderer *renderer1;
   vtkRendererCollection *renderer_collection = renWin1->GetRenderers();

   renWin2->SetSize(width,height);

   // Create new renderers for each of the renderers in
   // this render window. Then copy over each actor.
   renWin2->SetNumberOfLayers(renWin1->GetNumberOfLayers ());
   for (renderer_collection->InitTraversal(); (renderer1 = renderer_collection->GetNextItem());)
   {
      renWin2->AddRenderer(renderer1);
      renWin1->RemoveRenderer(renderer1);
   }

}


int VtkWindow::re_order_frames ()
{

   // start at one to account for the base renderer in VtkWindow
   int offset = 1;

   map<string, ChipsFrame*>::iterator ii;
   ChipsFrame *cf;
   for(ii=frame.begin(); ii != frame.end(); ++ii)
   {
      cf = (*ii).second;
      cf->set_order (offset + cf->get_frame_num ());
   }

   return ERR_STATUS_GOOD;
}

void VtkWindow::remap_hide ()
{
#ifndef HEADLESS
   // no need to remap if the window has already been hidden
   if (this->remapping || !this->drawready)
      return; 

   vtkRenderWindow *hiddenRenWin = vtkRenderWindow::New();
//   vtkRenderWindow *hiddenRenWin = vtkRenderWindow::New();
   hiddenRenWin->OffScreenRenderingOn ();

   // if the redraw renderer exists remove it now as it has ties to the
   // current render window
   if (this->redraw_renderer)
   {
      this->renWin->RemoveRenderer (redraw_renderer);
      this->renWin->SetNumberOfLayers (renWin->GetNumberOfLayers()-1);
      this->redraw_renderer->Delete ();
      this->backdrop_actor->Delete ();
      this->backdrop_actor = NULL;
      this->redraw_renderer = NULL;
   }

   // make sure all interactions are disabled
   this->disable_all_interaction ();
   this->block_interaction ();

   move_render_window (this->renWin, hiddenRenWin);

   ((VtkGtkWindow *)this->gui_window)->set_render_window (NULL);
   this->renWin->Delete ();
   this->renWin = hiddenRenWin;
   this->remapping = true;

#endif
}

void VtkWindow::remap_show ()
{

#ifndef HEADLESS
   if (!this->remapping)
      return;

   vtkRenderWindow *hiddenRenWin = this->renWin;

   this->renWin = vtkRenderWindow::New();
   if (smoothing)   renWin->LineSmoothingOn();
   this->move_render_window (hiddenRenWin, renWin);

   this->renWin->SetWindowId(this->gui_window->get_xwindow ());
   ((VtkGtkWindow *)this->gui_window)->set_render_window (renWin);
   
   //cerr << "VtkWindow enable_all_interaction" << endl;
   this->unblock_interaction ();
   this->enable_all_interaction ();

   hiddenRenWin->Delete ();
   hiddenRenWin = NULL;

   this->remapping = false;

   // this checks to see if the redraw status has been changed
   // while remapping. changing redraw status during a remap
   // causes x to crash
   this->change_redraw ();

#endif
}



void VtkWindow::draw_backdrop ()
{
   // make sure the backdrop actor exists before drawing the backdrop
   if (!this->backdrop_actor)
      return;

   vtkPoints *pts = vtkPoints::New ();
   vtkCellArray *arr = vtkCellArray::New ();
   vtkPolyData *backdrop = vtkPolyData::New ();
   vtkPolyDataMapper2D *bd_mapper = vtkPolyDataMapper2D::New ();

   pts->SetNumberOfPoints (4);
   pts->InsertPoint (0, 0, 0, 0);
   pts->InsertPoint (1, 0, height, 0);
   pts->InsertPoint (2, width, height, 0);
   pts->InsertPoint (3, width, 0, 0);

   arr->InsertNextCell (5);
   arr->InsertCellPoint (0);
   arr->InsertCellPoint (1);
   arr->InsertCellPoint (2);
   arr->InsertCellPoint (3);
   arr->InsertCellPoint (0);

   backdrop->SetPoints (pts);
   backdrop->SetPolys (arr);

   
   bd_mapper->SetInput (backdrop);

   backdrop_actor->SetMapper (bd_mapper);

   bd_mapper->Delete ();
   backdrop->Delete ();
   arr->Delete ();
   pts->Delete ();
}


#ifndef HEADLESS
ChipsGtkWindow *VtkWindow::new_gui_window (string geom, bool focus)
{
   return new VtkGtkWindow(renWin, this, geom, focus);
}
#endif

