// 
//  Copyright (C) 2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "vtkAddAnnotationWidget.hh"
#include "vtkAddAnnotationRepresentation.hh"
#include "vtkCommand.h"
#include "vtkChipsCommand.hh"
#include "vtkChipsWidgetEvent.hh"
#include "vtkKeySym.hh"
#include "gui/ChipsEventManager.hh"
#include "gui/ChipsEventObserver.hh"

#include "vtkCallbackCommand.h"
#include "vtkObjectFactory.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkWidgetEventTranslator.h"
#include "vtkWidgetCallbackMapper.h" 
#include "vtkAbstractPropPicker.h"
#include "vtkEvent.h"

#include "ChipsEngine.hh"

extern ChipsEventManager *event_manager;


vtkStandardNewMacro(vtkAddAnnotationWidget);

class vtkAddAnnotationEventObserver : ChipsEventObserver
{
public:
  vtkAddAnnotationEventObserver (vtkAddAnnotationWidget *parent)
  {
    this->register_event (ChipsEventManager::EndAddEvent);
    this->parent = parent;
  }

  void execute_event (int event, void *data)
  {
     this->parent->EndInteraction ();
  }

protected:
  vtkAddAnnotationWidget *parent;
};

    


//-------------------------------------------------------------------------
vtkAddAnnotationWidget::vtkAddAnnotationWidget()
{
  this->WidgetState = vtkAddAnnotationWidget::Start;
  this->WidgetRep = NULL;
  this->eventObserver = new vtkAddAnnotationEventObserver (this);

  this->CallbackMapper->SetCallbackMethod(vtkCommand::MouseMoveEvent,
                                          vtkChipsWidgetEvent::Move,
                                          this, vtkAddAnnotationWidget::OnMouseMove);
  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonReleaseEvent,
                                          vtkChipsWidgetEvent::EndAdd,
                                          this, vtkAddAnnotationWidget::OnLeftButtonRelease);
  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonPressEvent,
                                          vtkChipsWidgetEvent::Add,
                                          this, vtkAddAnnotationWidget::OnLeftButtonPress);
  this->CallbackMapper->SetCallbackMethod(vtkCommand::KeyPressEvent,
                                          vtkChipsWidgetEvent::KeyPressed,
                                          this, vtkAddAnnotationWidget::OnKeyPress);


  this->AnnotationPoints.reserve (100);

  this->SetPriority (.5);
}

//-------------------------------------------------------------------------
vtkAddAnnotationWidget::~vtkAddAnnotationWidget()
{
  if (this->eventObserver)
    delete this->eventObserver;
}

void vtkAddAnnotationWidget::SetBlock (bool flag)
{
  if (flag)
    {
    if (!this->AnnotationPoints.empty ())
      this->EndInteraction();
    this->WidgetState = vtkAddAnnotationWidget::Blocked;
    }
    // modified to not reset state while in the middle of adding a label
    else if (this->WidgetState !=  vtkAddAnnotationWidget::DrawText)  
       this->WidgetState = vtkAddAnnotationWidget::Start;
}

void vtkAddAnnotationWidget::EndInteraction ()
{
  double pos[2] = {0,0};

  if (this->WidgetState == vtkAddAnnotationWidget::Blocked)
    return;

  if (this->WidgetState == DrawText)
  {
     // remove bounding box around label 
     ChipsEngine *eng = event_manager->get_engine();
     if (eng)
     {
        int ival = 0;
        ChipsId cid;
        eng->set_property(&cid, static_cast<void*>(&ival), SET_LABEL_BBOX);
     }
  }
  this->AnnotationPoints.clear ();
  this->WidgetState = vtkAddAnnotationWidget::Start; 
  this->WidgetRep->EndWidgetInteraction (pos);
  this->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);

  this->Render();
}


void vtkAddAnnotationWidget::OnLeftButtonPress (vtkAbstractWidget* w)
{
  vtkAddAnnotationWidget* self = reinterpret_cast<vtkAddAnnotationWidget *>(w);

  if (self->SubOnLeftButtonRelease (w))
     return;

  if (self->WidgetState == vtkAddAnnotationWidget::Blocked)
    return;

  if (event_manager->get_state () != ChipsEventManager::AddState)
    return;

  if (!event_manager->is_valid_state ())
    return;

  if (event_manager->get_add_substate() == ChipsEventManager::LabelAdd &&
      self->WidgetState == vtkAddAnnotationWidget::DrawText)
  {
    // already started adding label- end it 
    double eventPos[2];
    eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
    eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);
    event_manager->invoke_event(eventPos, ChipsEventManager::EndAddEvent, NULL);
    ChipsEngine *eng = event_manager->get_engine();
    int ival = 0;
    ChipsId cid; 
    if (eng) eng->set_property(&cid, static_cast<void*>(&ival), SET_LABEL_BBOX);
    return;
  } 

  if ((event_manager->get_add_substate () != ChipsEventManager::EllipseAdd) &&
      (event_manager->get_add_substate () != ChipsEventManager::BoxAdd))
    return;

  double eventPos[2];
  vtkAddAnnotationRepresentation *addAnnoRep = reinterpret_cast<vtkAddAnnotationRepresentation*>(self->WidgetRep);
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);

  self->StartInteraction();
  self->InvokeEvent(vtkCommand::StartInteractionEvent,NULL);


  switch (event_manager->get_add_substate ())
    {
    case ChipsEventManager::EllipseAdd:
      addAnnoRep->SetDrawMode (vtkAddAnnotationRepresentation::Ellipse);
      self->WidgetState = vtkAddAnnotationWidget::DrawEllipse;
      self->WidgetRep->StartWidgetInteraction (eventPos);
    break;
    case ChipsEventManager::BoxAdd:
      addAnnoRep->SetDrawMode (vtkAddAnnotationRepresentation::Box);
      self->WidgetState = vtkAddAnnotationWidget::DrawBox;
      self->WidgetRep->StartWidgetInteraction (eventPos);
    break;
    default:
    break;
    }

}

  


void vtkAddAnnotationWidget::OnLeftButtonRelease (vtkAbstractWidget* w)
{
  vtkAddAnnotationWidget* self = reinterpret_cast<vtkAddAnnotationWidget *>(w);

  if (self->SubOnLeftButtonRelease (w))
     return;

  if (self->WidgetState == vtkAddAnnotationWidget::Blocked)
    return;

  if (event_manager->get_state () != ChipsEventManager::AddState)
    return;

  if (!event_manager->is_valid_state ())
    return;

  double eventPos[2];
  bool good = true;
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);

  if ((event_manager->get_add_substate () == ChipsEventManager::PolygonAdd) &&
     (self->AnnotationPoints.size () >= 4))
    {
    int num_points = self->AnnotationPoints.size () / 2;


    // check to see if the last segment added intersects
    // with any of the previous segments. 
    double x1 = eventPos[0];
    double y1 = eventPos[1];
    double x2 = self->AnnotationPoints[self->AnnotationPoints.size ()-2];
    double y2 = self->AnnotationPoints[self->AnnotationPoints.size ()-1];

    double n1, n2, d;
    for (int ii=num_points-2; ii--;)
      {
      double x3 = self->AnnotationPoints[ii*2];
      double y3 = self->AnnotationPoints[ii*2+1];
      double x4 = self->AnnotationPoints[(ii+1)*2];
      double y4 = self->AnnotationPoints[(ii+1)*2+1];

      n1 = (x4 - x3)*(y1 - y3) - (y4 - y3)*(x1 - x3);
      n2 = (x2 - x1)*(y1 - y3) - (y2 - y1)*(x1 - x3);
      d = (y4 - y3)*(x2 - x1) - (x4 - x3)*(y2 - y1);

      if (d != 0)
        {
        if ((((n1/d) >= 0) && ((n1/d) <= 1)) &&
            (((n2/d) >= 0) && ((n2/d) <= 1)))
          {
          good = false;
          break;
          }
        }
      }
    }

  if (good)
    {
    vtkAddAnnotationRepresentation *addAnnoRep = reinterpret_cast<vtkAddAnnotationRepresentation*>(self->WidgetRep);

    switch (event_manager->get_add_substate ())
      {
      case ChipsEventManager::PolyLineAdd:
      case ChipsEventManager::PolygonAdd:
        self->AnnotationPoints.push_back (eventPos[0]);
        self->AnnotationPoints.push_back (eventPos[1]);

        event_manager->invoke_event (eventPos, ChipsEventManager::AddEvent, (void *)&(self->AnnotationPoints));
        addAnnoRep->SetDrawMode (vtkAddAnnotationRepresentation::Line);
        self->WidgetRep->StartWidgetInteraction (eventPos);
        self->WidgetState = vtkAddAnnotationWidget::Continue;

        break;
      case ChipsEventManager::EllipseAdd:
        {
        addAnnoRep->GetPoints (self->AnnotationPoints);
        event_manager->invoke_event (eventPos, ChipsEventManager::AddEvent, (void *)&(self->AnnotationPoints));
        event_manager->invoke_event (eventPos, ChipsEventManager::EndAddEvent, NULL);
        self->WidgetState = vtkAddAnnotationWidget::Start;
        }
        break;
      case ChipsEventManager::BoxAdd:
        {
        addAnnoRep->GetPoints (self->AnnotationPoints);
        event_manager->invoke_event (eventPos, ChipsEventManager::AddEvent, (void *)&(self->AnnotationPoints));
        event_manager->invoke_event (eventPos, ChipsEventManager::EndAddEvent, NULL);
        self->WidgetState = vtkAddAnnotationWidget::Start;
        }
        break;
      case ChipsEventManager::LabelAdd:
        {
        self->AnnotationPoints.push_back (eventPos[0]);
        self->AnnotationPoints.push_back (eventPos[1]);
        event_manager->invoke_event (eventPos, ChipsEventManager::AddEvent, (void *)&(self->AnnotationPoints));
        self->WidgetState = vtkAddAnnotationWidget::DrawText;
        }
        break;
      default:
        self->AnnotationPoints.push_back (eventPos[0]);
        self->AnnotationPoints.push_back (eventPos[1]);
        event_manager->invoke_event (eventPos, ChipsEventManager::AddEvent, (void *)&(self->AnnotationPoints));
        event_manager->invoke_event (eventPos, ChipsEventManager::EndAddEvent, NULL);
        self->WidgetState = vtkAddAnnotationWidget::Start;
        break;
      }

    self->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);

    self->Render();
    }

}


void vtkAddAnnotationWidget::OnMouseMove (vtkAbstractWidget* w)
{
  vtkAddAnnotationWidget* self = reinterpret_cast<vtkAddAnnotationWidget *>(w);

  if (self->SubOnMouseMove (w))
    return;

  if ((self->WidgetState == vtkAddAnnotationWidget::Start) ||
      (self->WidgetState == vtkAddAnnotationWidget::Blocked))
    return;

  if (event_manager->get_state () != ChipsEventManager::AddState)
    return;

  double eventPos[2];
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);

  self->WidgetRep->WidgetInteraction (eventPos);

  self->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);
  self->Render();
}

void vtkAddAnnotationWidget::OnKeyPress (vtkAbstractWidget* w)
{
  vtkAddAnnotationWidget* self = reinterpret_cast<vtkAddAnnotationWidget *>(w);

  if ( self->SubOnKeyPress(w) )
    return;

  if (self->WidgetState == vtkAddAnnotationWidget::Blocked)
    return;

  if (event_manager->get_state () != ChipsEventManager::AddState)
    return;

  if (event_manager->get_add_substate() == ChipsEventManager::LabelAdd)
  {
     unsigned char keyCode = self->Interactor->GetKeyCode ();
     char* keysym = self->Interactor->GetKeySym ();

     ChipsEngine *eng = event_manager->get_engine();
     if (!eng)
        return;

     ChipsId cid;
     string curval; 
     int ival = 0;
     eng->get_property(&cid, (void*)&curval, SET_LABEL_TEXT);
     self->WidgetState = vtkAddAnnotationWidget::DrawText;
     if (keyCode == vtkKeySym::Backspace || keyCode == vtkKeySym::Delete)
     {
        // erase last character 
        if (curval.length())  curval.erase(curval.end()-1); 
     } 
     else if (keyCode == vtkKeySym::Esc || (keysym && !strcmp(keysym,"\r")))
     {
        // [Escape] or [Return] hit- exit add label mode
        double eventPos[2];
        eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
        eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);
        event_manager->invoke_event(eventPos, ChipsEventManager::EndAddEvent, NULL);
        eng->set_property(&cid, static_cast<void*>(&ival), SET_LABEL_BBOX);
        return;
     }
     else if (keysym) 
     {
        curval += keysym;  // add text 
     }
     eng->set_property(&cid, static_cast<void*>(const_cast<char*>(curval.c_str ())), SET_LABEL_TEXT);
     // hide bounding box, except if no text
     ival = 1;
     eng->set_property(&cid, static_cast<void*>(&ival), SET_LABEL_BBOX);
  }
}

//----------------------------------------------------------------------
void vtkAddAnnotationWidget::CreateDefaultRepresentation()
{
  if ( ! this->WidgetRep )
    {
    this->WidgetRep = vtkAddAnnotationRepresentation::New ();
    }
}

//-------------------------------------------------------------------------
void vtkAddAnnotationWidget::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  this->CallbackMapper->PrintSelf (os, indent);
}

