#include "vtkBoundingBoxRepresentation.hh"
#include "vtkRenderer.h"
#include "vtkWindow.h"
#include "vtkObjectFactory.h"

#include "vtkPoints.h"
#include "vtkCellArray.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkActor2D.h"
#include "vtkProperty2D.h"
#include "vtkTransform.h"
#include "vtkTransformPolyDataFilter.h"

vtkStandardNewMacro(vtkBoundingBoxRepresentation);

vtkProperty2D *vtkBoundingBoxRepresentation::GlobalBorderProperty = vtkProperty2D::New ();

//-------------------------------------------------------------------------
vtkBoundingBoxRepresentation::vtkBoundingBoxRepresentation()
{

  // Initial positioning information
  this->PositionCoordinate = vtkCoordinate::New();
  this->PositionCoordinate->SetCoordinateSystemToDisplay();
  this->PositionCoordinate->SetValue(0.05, 0.05);
  this->Position2Coordinate = vtkCoordinate::New();
  this->Position2Coordinate->SetCoordinateSystemToDisplay();
  this->Position2Coordinate->SetReferenceCoordinate(this->PositionCoordinate);

  vtkPoints *borderPoints = vtkPoints::New();
  borderPoints->SetDataTypeToDouble();
  borderPoints->SetNumberOfPoints(4);
  borderPoints->SetPoint(0, 0.0, 0.0, 0.0); //may be updated by the subclass
  borderPoints->SetPoint(1, 1.0, 0.0, 0.0);
  borderPoints->SetPoint(2, 1.0, 1.0, 0.0);
  borderPoints->SetPoint(3, 0.0, 1.0, 0.0);
  vtkCellArray *borderLine = vtkCellArray::New();
  borderLine->InsertNextCell(5);
  for (int ii=0; ii<4; ii++)
    borderLine->InsertCellPoint(ii);
  borderLine->InsertCellPoint(0);

  vtkPolyData *borderPolyData = vtkPolyData::New();
  borderPolyData->SetPoints(borderPoints);
  borderPolyData->SetLines(borderLine);

  borderPoints->Delete ();
  borderLine->Delete();
  this->BorderTransform = vtkTransform::New();
  this->BorderTransformFilter = vtkTransformPolyDataFilter::New();
  this->BorderTransformFilter->SetTransform(this->BorderTransform);
  this->BorderTransformFilter->SetInput(borderPolyData);

  borderPolyData->Delete ();

  this->BorderMapper = vtkPolyDataMapper2D::New();
  this->BorderMapper->SetInputConnection(this->BorderTransformFilter->GetOutputPort());
  this->BorderActor = vtkActor2D::New();
  this->BorderActor->SetMapper(this->BorderMapper);

  this->BorderProperty = vtkProperty2D::New();
//  this->BorderActor->SetProperty(this->BorderProperty);
  this->BorderActor->SetProperty (GlobalBorderProperty);

  this->BorderProperty->SetColor (1, 0, 1);
  this->BorderActor->SetVisibility (0);

  this->DragOrigin[0] = true;
  this->DragOrigin[1] = true;

  this->MinBoundPos[0] = -1;
  this->MinBoundPos[1] = -1;

  this->MaxBoundPos[0] = -1;
  this->MaxBoundPos[1] = -1;

  this->UseRatio = false;
  this->Ratio[0] = 1;
  this->Ratio[1] = 1;

  this->UseGlobalProperty = true;

}

//----------------------------------------------------------------------------
vtkBoundingBoxRepresentation::~vtkBoundingBoxRepresentation()
{
  this->BorderTransform->Delete ();
  this->BorderTransformFilter->Delete ();
  this->BorderMapper->Delete ();
  this->BorderActor->Delete ();
  this->BorderProperty->Delete ();

  this->PositionCoordinate->Delete ();
  this->Position2Coordinate->Delete ();

}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::StartWidgetInteraction(double eventPos[2])
{
  this->StartEventPosition[0] = eventPos[0];
  this->StartEventPosition[1] = eventPos[1];

  this->PositionCoordinate->SetValue(eventPos[0],eventPos[1]);
  this->Position2Coordinate->SetValue(1,1);

  this->PositionCoordinate->GetComputedDisplayValue(this->Renderer);
  this->Position2Coordinate->GetComputedDisplayValue(this->Renderer);

  // make sure this is the top of the render list
  this->Renderer->RemoveViewProp (this);
  this->Renderer->AddViewProp (this);

  this->BorderActor->SetVisibility (1);

  this->Modified();


}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::WidgetInteraction(double eventPos[2])
{
/*
  // convert to normalized viewport coordinates
  this->Renderer->DisplayToNormalizedDisplay(XF,YF);
  this->Renderer->NormalizedDisplayToViewport(XF,YF);
  this->Renderer->ViewportToNormalizedViewport(XF,YF);
*/

  // there are four parameters that can be adjusted
  double *fpos1 = this->PositionCoordinate->GetValue();
  double *fpos2 = this->Position2Coordinate->GetValue();

  double par1[2];
  double par2[2];
  par1[0] = fpos1[0];
  par1[1] = fpos1[1];
  par2[0] = fpos1[0] + fpos2[0];
  par2[1] = fpos1[1] + fpos2[1];

  double delX = eventPos[0] - this->StartEventPosition[0];
  double delY = eventPos[1] - this->StartEventPosition[1];

  if (this->DragOrigin[0])
  {
    if ((par1[0] + delX >= par2[0]) || (eventPos[0] > par2[0]))
      this->DragOrigin[0] = false;
  }
  else
  {
    if ((par2[0] + delX <= par1[0]) || (eventPos[0] < par1[0]))
      this->DragOrigin[0] = true;
  }


  if (this->DragOrigin[1])
  {
    if ((par1[1] + delY >= par2[1]) || (eventPos[1] > par2[1]))
      this->DragOrigin[1] = !this->DragOrigin[1];
  }
  else
  {
    if ((par2[1] + delY <= par1[1]) || (eventPos[1] < par1[1]))
      this->DragOrigin[1] = !this->DragOrigin[1];
  }


  for (int ii=0; ii<2; ii++)
    {
    if (this->DragOrigin[ii])
      par1[ii] = eventPos[ii];
    else
      par2[ii] = eventPos[ii];

    if ((this->MinBoundPos[ii] > -1) && (par1[ii] < this->MinBoundPos[ii] ))
      par1[ii] = this->MinBoundPos[ii];

    if ((this->MaxBoundPos[ii] > -1) && (par2[ii] > this->MaxBoundPos[ii] ))
      par2[ii] = this->MaxBoundPos[ii];
    }


  if (this->UseRatio)
    {
    double aspect = this->Ratio[0]/this->Ratio[1];  // width/height
    double boxAspect = (par2[0] - par1[0]) / (par2[1] - par1[1]);
    double scale = 0;
    double exceededBounds = false;
    int index1 = 0;
    int index2 = 1;

    // adjust the width
    if (boxAspect < aspect)
      {
      scale = aspect/boxAspect;
      index1 = 0;
      index2 = 1;
      }
    else
      {
      scale = boxAspect/aspect;
      index1 = 1;
      index2 = 0;
      }
      
    // adjust the position to fit the scale
    if (this->DragOrigin[index1])
      {
      par1[index1] = par2[index1] - (scale * (par2[index1] - par1[index1]));

      // make sure the new bounds haven't exceed the max bounds
      if ((this->MinBoundPos[index1] > -1) && (par1[index1] < this->MinBoundPos[index1] ))
        {
        par1[index1] = this->MinBoundPos[index1];
        exceededBounds = true;
        }
      }
    else
      {
      par2[index1] = (scale * (par2[index1] - par1[index1])) + par1[index1];

      // make sure the new bounds haven't exceed the max bounds
      if ((this->MaxBoundPos[index1] > -1) && (par2[index1] > this->MaxBoundPos[index1] ))
        {
        par2[index1] = this->MaxBoundPos[index1];
        exceededBounds = true;
        }
      }

    // if the bounds had been exceded recalculate height bounds
    if (exceededBounds)
      {
      boxAspect = (par2[0] - par1[0]) / (par2[1] - par1[1]);

      scale = (index1 == 0) ? boxAspect/aspect : aspect/boxAspect;

      // adjust the position to fit the scale
      if (this->DragOrigin[index2])
        par1[index2] = par2[index2] - (scale * (par2[index2] - par1[index2]));
      else
        par2[index2] = (scale * (par2[index2] - par1[index2])) + par1[index2];
      }
    }

  // Modify the representation
  if (par2[0] > par1[0] && par2[1] > par1[1])
    {
    this->PositionCoordinate->SetValue(par1[0],par1[1]);
    this->Position2Coordinate->SetValue(par2[0] - par1[0], par2[1] - par1[1]);

    this->StartEventPosition[0] = eventPos[0];
    this->StartEventPosition[1] = eventPos[1];
    }

  this->Modified();
}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::EndWidgetInteraction(double eventPos[2])
{
  this->BorderActor->SetVisibility (0);
}

void vtkBoundingBoxRepresentation::GetBoxRange(double range[])
{
  double *fpos1 = this->PositionCoordinate->GetValue();
  double *fpos2 = this->Position2Coordinate->GetValue();

  range[0] = fpos1[0];
  range[1] = fpos1[1];
  range[2] = fpos1[0] + fpos2[0];
  range[3] = fpos1[1] + fpos2[1];
}



//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::BuildRepresentation()
{
  if ( this->GetMTime() > this->BuildTime ||
       (this->Renderer && this->Renderer->GetVTKWindow() &&
        this->Renderer->GetVTKWindow()->GetMTime() > this->BuildTime) )
    {

    // Set things up
    double *pos1 = this->PositionCoordinate->
      GetComputedDoubleViewportValue(this->Renderer);
    double *pos2 = this->Position2Coordinate->
      GetComputedDoubleViewportValue(this->Renderer);

    // Now transform the canonical widget into display coordinates
    double tx = pos1[0];
    double ty = pos1[1];
    double sx = (pos2[0]-pos1[0]);
    double sy = (pos2[1]-pos1[1]);

    this->BorderTransform->Identity();
    this->BorderTransform->Translate(tx, ty, 0.0);
    this->BorderTransform->Scale(sx,sy,1);

    this->BuildTime.Modified();
    }
}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::ReleaseGraphicsResources(vtkWindow *w)
{
  this->BorderActor->ReleaseGraphicsResources(w);
/*  this->BorderOutlineActor->ReleaseGraphicsResources(w);
*/
}

//-------------------------------------------------------------------------
int vtkBoundingBoxRepresentation::RenderOverlay(vtkViewport *w)
{
  int count = 0;
  if ( ! this->BorderActor->GetVisibility() )
    {
    return 0;
    }
  this->BuildRepresentation();
//  count += this->BorderOutlineActor->RenderOverlay(w);
  count += this->BorderActor->RenderOverlay(w);
  return count;
}

//-------------------------------------------------------------------------
int vtkBoundingBoxRepresentation::RenderOpaqueGeometry(vtkViewport *w)
{
  int count = 0;

  if ( ! this->BorderActor->GetVisibility() )
    {
    return 0;
    }
  this->BuildRepresentation();
//  count += this->BorderOutlineActor->RenderOpaqueGeometry(w);
  count += this->BorderActor->RenderOpaqueGeometry(w);
  return count;
}

//-----------------------------------------------------------------------------
int vtkBoundingBoxRepresentation::RenderTranslucentPolygonalGeometry(vtkViewport *w)
{
  int count = 0;
  if ( ! this->BorderActor->GetVisibility() )
    {
    return 0;
    }
//  count += this->BorderOutlineActor->RenderTranslucentPolygonalGeometry(w);
  this->BuildRepresentation();
  count += this->BorderActor->RenderTranslucentPolygonalGeometry(w);
  return count;
}

//-----------------------------------------------------------------------------
// Description:
// Does this prop have some translucent polygonal geometry?
int vtkBoundingBoxRepresentation::HasTranslucentPolygonalGeometry()
{
  int count = 0;
  if ( ! this->BorderActor->GetVisibility() )
    {
    return 0;
    }
//  count += this->BorderOutlineActor->HasTranslucentPolygonalGeometry();
  this->BuildRepresentation();
  count += this->BorderActor->HasTranslucentPolygonalGeometry();
  return count;
}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::SetUseGlobalProperty (bool enable)
{
  this->UseGlobalProperty = enable;

  if (enable)
    this->BorderActor->SetProperty (GlobalBorderProperty);
  else
    this->BorderActor->SetProperty (this->BorderProperty);


}

//-------------------------------------------------------------------------
void vtkBoundingBoxRepresentation::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
}




