#include "vtkBoundingBoxWidget.hh"
#include "vtkBoundingBoxRepresentation.hh"
#include "vtkCommand.h"
#include "vtkChipsCommand.hh"
#include "vtkChipsWidgetEvent.hh"
#include "vtkKeySym.hh"
#include "gui/ChipsEventManager.hh"
#include "gui/ChipsEventObserver.hh"


#include "vtkCallbackCommand.h"
#include "vtkObjectFactory.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkWidgetEventTranslator.h"
#include "vtkWidgetCallbackMapper.h" 
#include "vtkAbstractPropPicker.h"
#include "vtkEvent.h"

extern ChipsEventManager *event_manager;


vtkStandardNewMacro(vtkBoundingBoxWidget);

class vtkBoundingBoxEventObserver : ChipsEventObserver
{
public:
  vtkBoundingBoxEventObserver (vtkBoundingBoxWidget *parent)
  {
    this->register_event (ChipsEventManager::StartRangeZoomEvent);
    this->parent = parent;
  }

  void execute_event (int event, void *data)
  {
     this->parent->dim = *(int*) data;
  }

protected:
  vtkBoundingBoxWidget *parent;
};

    


//-------------------------------------------------------------------------
vtkBoundingBoxWidget::vtkBoundingBoxWidget()
{
  this->WidgetState = vtkBoundingBoxWidget::Start;
  this->WidgetRep = NULL;
  this->eventObserver = new vtkBoundingBoxEventObserver (this);

  this->CallbackMapper->SetCallbackMethod(vtkCommand::MouseMoveEvent,
                                          vtkChipsWidgetEvent::Move,
                                          this, vtkBoundingBoxWidget::OnMouseMove);
  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonPressEvent,
                                          vtkChipsWidgetEvent::Resize,
                                          this, vtkBoundingBoxWidget::OnLeftButtonPress);
  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonReleaseEvent,
                                          vtkChipsWidgetEvent::EndSelect,
                                          this, vtkBoundingBoxWidget::OnLeftButtonRelease);

  this->dim = 3;

  this->SetPriority (.5);
}

//-------------------------------------------------------------------------
vtkBoundingBoxWidget::~vtkBoundingBoxWidget()
{
  if (this->eventObserver)
    delete this->eventObserver;
}

void vtkBoundingBoxWidget::SetBlock (bool flag)
{
  if (flag)
    {
    this->WidgetState = vtkBoundingBoxWidget::Blocked;
    if (this->WidgetState != vtkBoundingBoxWidget::Start)
      {
      this->EndInteraction();
      }
    }
  else
    this->WidgetState = vtkBoundingBoxWidget::Start;
}
    

void vtkBoundingBoxWidget::OnLeftButtonPress (vtkAbstractWidget* w)
{

  vtkBoundingBoxWidget* self = reinterpret_cast<vtkBoundingBoxWidget *>(w);

  if (self->SubOnLeftButtonPress (w))
     return;

  if (self->WidgetState != vtkBoundingBoxWidget::Start)
    return;

  if (event_manager->get_state () != ChipsEventManager::RangeZoomState)
    return;

  if (!event_manager->is_valid_state ())
    return;

  double eventPos[2];
  vtkBoundingBoxRepresentation *boxRep = reinterpret_cast<vtkBoundingBoxRepresentation*>(self->WidgetRep);
  double *minBound = boxRep->GetMinBoundPos ();
  double *maxBound = boxRep->GetMaxBoundPos ();
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);

  // make sure the event position is within the min/max bounds
  if (((minBound[0] >= 0) && (eventPos[0] < minBound[0])) ||
      ((minBound[1] >= 0) && (eventPos[1] < minBound[1])) ||
      ((maxBound[0] >= 0) && (eventPos[0] > maxBound[0])) ||
      ((maxBound[1] >= 0) && (eventPos[1] > maxBound[1])))
    return;

  self->WidgetState = vtkBoundingBoxWidget::Select;

  self->WidgetRep->StartWidgetInteraction (eventPos);

  self->StartInteraction();
  self->InvokeEvent(vtkCommand::StartInteractionEvent,NULL);

}


void vtkBoundingBoxWidget::OnLeftButtonRelease (vtkAbstractWidget* w)
{
  vtkBoundingBoxWidget* self = reinterpret_cast<vtkBoundingBoxWidget *>(w);

  if (self->SubOnLeftButtonRelease (w))
     return;

  if ((self->WidgetState == vtkBoundingBoxWidget::Start) ||
      (self->WidgetState == vtkBoundingBoxWidget::Blocked))
    return;

  double eventPos[2];
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);

  self->WidgetRep->EndWidgetInteraction (eventPos);
  if (self->WidgetState == vtkBoundingBoxWidget::Expand)
    {
    self->WidgetState = vtkBoundingBoxWidget::Start;
    if (event_manager->get_state () == ChipsEventManager::RangeZoomState)
      {
      double select_range[4];
      vtkBoundingBoxRepresentation *boxRep = reinterpret_cast<vtkBoundingBoxRepresentation*>(self->WidgetRep);
      boxRep->GetBoxRange (select_range);

      if (self->dim == 1)
        event_manager->invoke_event (eventPos, ChipsEventManager::RangeZoomXEvent, select_range);
      else if (self->dim == 2)
        event_manager->invoke_event (eventPos, ChipsEventManager::RangeZoomYEvent, select_range);
      else if (self->dim == 3)
        event_manager->invoke_event (eventPos, ChipsEventManager::RangeZoomEvent, select_range);

      }
    }
    

  self->EndInteraction();
  self->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);
  self->Render();
}


void vtkBoundingBoxWidget::OnMouseMove (vtkAbstractWidget* w)
{
  vtkBoundingBoxWidget* self = reinterpret_cast<vtkBoundingBoxWidget *>(w);

  if (self->SubOnMouseMove (w))
     return;

  if ((self->WidgetState == vtkBoundingBoxWidget::Start) ||
      (self->WidgetState == vtkBoundingBoxWidget::Blocked))
    return;

  if (event_manager->get_state () != ChipsEventManager::RangeZoomState)
    return;

  double eventPos[2];
  eventPos[0] = static_cast<double>(self->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(self->Interactor->GetEventPosition()[1]);


  self->WidgetState = vtkBoundingBoxWidget::Expand;

  self->WidgetRep->WidgetInteraction (eventPos);

  self->EndInteraction();
  self->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);
  self->Render();
}

//----------------------------------------------------------------------
void vtkBoundingBoxWidget::CreateDefaultRepresentation()
{
  if ( ! this->WidgetRep )
    {
    this->WidgetRep = vtkBoundingBoxRepresentation::New ();
    }
}

//-------------------------------------------------------------------------
void vtkBoundingBoxWidget::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  this->CallbackMapper->PrintSelf (os, indent);
}

