/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkChipsRepresentation.h

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkChipsRepresentation - represent a vtkBorderWidget
// .SECTION Description
// This class is used to represent and render a vtBorderWidget. To
// use this class, you need to specify the two corners of a rectangular
// region.
//
// The class is typically subclassed so that specialized representations can
// be created.  The class defines an API and a default implementation that
// the vtkChipsRepresentation interacts with to render itself in the scene.

// .SECTION Caveats
// The separation of the widget event handling (e.g., vtkBorderWidget) from
// the representation (vtkChipsRepresentation) enables users and developers
// to create new appearances for the widget. It also facilitates parallel
// processing, where the client application handles events, and remote
// representations of the widget are slaves to the client (and do not handle
// events).

// .SECTION See Also
// vtkBorderWidget vtkTextWidget


#ifndef __vtkChipsRepresentation_h
#define __vtkChipsRepresentation_h

#include "vtkWidgetRepresentation.h"
#include "vtkCoordinate.h" //Because of the viewport coordinate macro

class vtkPoints;
class vtkPolyData;
class vtkTransform;
class vtkTransformPolyDataFilter;
class vtkPolyDataMapper2D;
class vtkActor2D;
class vtkProperty2D;
class vtkGlyphSource2D;
class vtkGlyph2D;


class VTK_WIDGETS_EXPORT vtkChipsRepresentation : public vtkWidgetRepresentation
{
public:
  // Description:
  // Instantiate this class.
  static vtkChipsRepresentation *New();

  // Description:
  // Define standard methods.
  vtkTypeMacro(vtkChipsRepresentation,vtkWidgetRepresentation);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Specify opposite corners of the box defining the boundary of the
  // widget. By default, these coordinates are in the normalized viewport
  // coordinate system, with Position the lower left of the outline, and
  // Position2 relative to Position. Note that using these methods are
  // affected by the ProportionalResize flag. That is, if the aspect ratio of
  // the representation is to be preserved (e.g., ProportionalResize is on),
  // then the rectangle (Position,Position2) is a bounding rectangle. Also,
  vtkViewportCoordinateMacro(Position);
  vtkViewportCoordinateMacro(Position2);

//ETX
  // Description:
  // Specify when and if the border should appear. If ShowBorder is "on",
  // then the border will appear if the border is enabled. If the border
  // is off the border will never appear.  
  void SetShowBorder (bool flag);
  vtkGetMacro(ShowBorder,bool);
  vtkBooleanMacro(ShowBorder,bool);

  // Description:
  // When the border is enabled, then the reqests of ShowBorder are
  // recognized. If teh border is disabled, it will not appear.
  void SetBorderEnabled (bool flag);
  vtkGetMacro(BorderEnabled,bool);
  vtkBooleanMacro(BorderEnabled,bool);
  void EnableBorder() {this->SetBorderEnabled(true);}
  void DisableBorder() {this->SetBorderEnabled(false);}

  // Description:
  // When the rotation is enabled, then the reqests of ShowBorder are
  // recognized. If teh rotation is disabled, it will not appear.
  void SetRotationEnabled (bool flag);
  vtkGetMacro(RotationEnabled,bool);
  vtkBooleanMacro(RotationEnabled,bool);
  void EnableRotation() {this->SetRotationEnabled(true);}
  void DisableRotation() {this->SetRotationEnabled(false);}


  // Description:
  // Set/Get whether the widget should use global properties or local properties
  virtual void SetUseGlobalProperty (bool enable);
  vtkGetMacro (UseGlobalProperty, bool);

  // Description:
  // Get the global properties 
  static vtkProperty2D *GetGlobalBorderProperty () {return GlobalBorderProperty;};

  
  // Description:
  // Specify the properties of the border.
  vtkGetObjectMacro(BorderProperty,vtkProperty2D);

  // Description:
  // Specify the properties of the border.
  vtkGetObjectMacro(RotationProperty,vtkProperty2D);


  vtkGetVectorMacro(StartEventPosition,double,2);
  vtkGetVectorMacro(CurrentEventPosition,double,2);

  // Description:
  // Indicate whether resizing operations should keep the x-y directions
  // proportional to one another. Also, if ProportionalResize is on, then
  // the rectangle (Position,Position2) is a bounding rectangle, and the
  // representation will be placed in the rectangle in such a way as to 
  // preserve the aspect ratio of the representation.
  vtkSetMacro(ProportionalResize,int);
  vtkGetMacro(ProportionalResize,int);
  vtkBooleanMacro(ProportionalResize,int);

  // Description:
  // Specify a minimum and/or maximum size (in pixels) that this representation
  // can take. These methods require two values: size values in the x and y
  // directions, respectively.
  vtkSetVector2Macro(MinimumSize,int);
  vtkGetVector2Macro(MinimumSize,int);
  vtkSetVector2Macro(MaximumSize,int);
  vtkGetVector2Macro(MaximumSize,int);

  // Description:
  // The tolerance representing the distance to the widget (in pixels)
  // in which the cursor is considered to be on the widget, or on a
  // widget feature (e.g., a corner point or edge).
  vtkSetClampMacro(Tolerance,int,1,10);
  vtkGetMacro(Tolerance,int);

  // Description:
  // This is a modifier of the interaction state. When set, widget interaction
  // allows the border (and stuff inside of it) to be translated with mouse
  // motion.
  virtual void SetMoving (bool flag) {this->Moving = flag;};
  vtkGetMacro(Moving,bool);
  vtkBooleanMacro(Moving,bool);

  // Description:
  // This is a modifier of the interaction state. When set, widget interaction
  // allows the border (and stuff inside of it) to be translated with mouse
  // motion.
  virtual void SetResizing (bool flag) {this->Resizing = flag;};
  vtkGetMacro(Resizing,bool);
  vtkBooleanMacro(Resizing,bool);

  // Description:
  // This is a modifier of the interaction state. When set, widget interaction
  // allows the border (and stuff inside of it) to be translated with mouse
  // motion.
  virtual void SetRotating (bool flag) {this->Rotating = flag;};
  vtkGetMacro(Rotating,bool);
  vtkBooleanMacro(Rotating,bool);

  // Description:
  // Get the current angle
  vtkGetMacro (Angle, float);



//BTX
  // Description:
  // Define the various states that the representation can be in.
  enum _InteractionState
  {
    Outside=0,
    Inside,
    On,
    RightEdge,
    LeftEdge,
    TopEdge,
    BottomEdge,
    RightHandle,
    LeftHandle,
    TopHandle,
    BottomHandle,
    TopRightHandle,
    TopLeftHandle,
    BottomLeftHandle,
    BottomRightHandle,
    RotateHandle
  };
//ETX

  // Description:
  // Subclasses should implement these methods. See the superclasses'
  // documentation for more information.
  virtual void BuildRepresentation();
  virtual void StartWidgetInteraction(double eventPos[2]);
  virtual void WidgetInteraction(double eventPos[2]);
  virtual void EndWidgetInteraction (double eventPos[2]);
  virtual void GetSize(double size[2])
    {size[0]=1.0; size[1]=1.0;}
  virtual int ComputeBorderInteractionState(int X, int Y, int modify=0);
  virtual int ComputeInteractionState(int X, int Y, int modify=0);

  virtual bool IsInteractionStateOnObject () {return this->OnObject;};

  // Description:
  // Update the representative with any changes to the system.
  virtual void Update();

  
  // Description:
  // These methods are necessary to make this representation behave as
  // a vtkProp.
  virtual void GetActors2D(vtkPropCollection*);
  virtual void ReleaseGraphicsResources(vtkWindow*);
  virtual int RenderOverlay(vtkViewport*);
  virtual int RenderOpaqueGeometry(vtkViewport*);
  virtual int RenderTranslucentPolygonalGeometry(vtkViewport*);
  virtual int HasTranslucentPolygonalGeometry();

  
protected:
  vtkChipsRepresentation();
  ~vtkChipsRepresentation();

  // Ivars
  bool          ShowBorder;
  bool          RotationEnabled;
  bool          BorderEnabled;
  vtkProperty2D *BorderProperty;
  vtkProperty2D *RotationProperty;
  int           ProportionalResize;
  int           Tolerance;

  //interaction states
  bool          Moving;
  bool          Resizing;
  bool          Rotating;
  
  // Layout (position of lower left and upper right corners of border)
  vtkCoordinate *PositionCoordinate;
  vtkCoordinate *Position2Coordinate;

  // Sometimes subclasses must negotiate with their superclasses
  // to achieve the correct layout.
  bool Negotiated;
  virtual void NegotiateLayout();
  virtual bool CheckNegotiateLayout() {return !this->Negotiated;}

  // Border representation. Subclasses may use the BWTransform class
  // to transform their geometry into the region surrounded by the border.
  vtkGlyphSource2D           *BorderGlyphSource;
  vtkGlyph2D                 *BorderGlyphs;
  vtkPolyData                *BorderPolyData;
  vtkTransform               *BorderTransform;
  vtkTransformPolyDataFilter *BorderTransformFilter;
  vtkPolyDataMapper2D        *BorderMapper;
  vtkActor2D                 *BorderActor;

  vtkGlyphSource2D           *RotationGlyphSource;
  vtkGlyph2D                 *RotationGlyphs;
  vtkPolyData                *RotationPolyData;
  vtkTransform               *RotationTransform;
  vtkTransformPolyDataFilter *RotationTransformFilter;
  vtkPolyDataMapper2D        *RotationMapper;
  vtkActor2D                 *RotationActor;

  float                      RotationPointOffset;

  

  double CurrentEventPosition[3];

  // Constraints on size
  int MinimumSize[2];
  int MaximumSize[2];

  bool OnObject;

  float Angle;

  static vtkProperty2D       *GlobalBorderProperty;
  bool   UseGlobalProperty;

private:
  vtkChipsRepresentation(const vtkChipsRepresentation&);  //Not implemented
  void operator=(const vtkChipsRepresentation&);  //Not implemented
};

#endif
