// 
//  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "vtkChipsWidget.hh"
#include "vtkChipsRepresentation.hh"
#include "vtkCommand.h"
#include "vtkChipsCommand.hh"
#include "vtkChipsWidgetEvent.hh"
#include "vtkKeySym.hh"

#include "ChipsEventManager.hh"

#include "vtkCallbackCommand.h"
#include "vtkObjectFactory.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkWidgetEventTranslator.h"
#include "vtkWidgetCallbackMapper.h" 
#include "vtkEvent.h"
#include "vtkCoordinate.h"

#include "vtkGtkRenderWindowInteractor.hh"

extern ChipsEventManager *event_manager;


vtkStandardNewMacro(vtkChipsWidget);

class vtkFinalizeSelectCallback : public vtkCommand
{
public:
  vtkFinalizeSelectCallback () {this->parent = NULL;};
  static vtkFinalizeSelectCallback *New ()
    { return new vtkFinalizeSelectCallback; }
  
  virtual void Execute(vtkObject*, unsigned long eventId, void*)
    {
      if (!event_manager)
        return;
      switch (eventId)
        {
        case vtkChipsCommand::LeftButtonDoubleClickEvent:
        case vtkCommand::LeftButtonPressEvent:
          event_manager->finalize_select (event_manager->get_state () == ChipsEventManager::SelectState);
          if (this->parent)
            {
            this->parent->FinalizeSelect ();
            }
          break;

        }
    }
  void SetParent (vtkChipsWidget *parent) { this->parent = parent; };
protected:
  vtkChipsWidget *parent;
};

//-------------------------------------------------------------------------
vtkChipsWidget::vtkChipsWidget()
{
  this->WidgetState = vtkChipsWidget::Start;
  this->Selectable = 1;
  this->Resizable = 1;
  this->Moveable = 1;
  this->Rotateable = 1;
  
  this->objectInfo = new ChipsEventObjectInfo ();

  this->finalizeSelect = vtkFinalizeSelectCallback::New ();
  this->finalizeSelect->SetParent (this);

 
  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonPressEvent,
                                          vtkChipsWidgetEvent::Select,
                                          this, vtkChipsWidget::SelectAction);

  this->CallbackMapper->SetCallbackMethod(vtkCommand::LeftButtonReleaseEvent,
                                          vtkChipsWidgetEvent::EndSelect,
                                          this, vtkChipsWidget::EndSelectAction);
  this->CallbackMapper->SetCallbackMethod(vtkChipsCommand::LeftButtonDoubleClickEvent,
                                          vtkChipsWidgetEvent::DoubleClickSelect,
                                          this, vtkChipsWidget::DoubleClickSelectAction);

  this->CallbackMapper->SetCallbackMethod(vtkCommand::MouseMoveEvent,
                                          vtkChipsWidgetEvent::Move,
                                          this, vtkChipsWidget::MoveAction);



  this->CallbackMapper->SetCallbackMethod(vtkCommand::KeyPressEvent,
                                          vtkChipsWidgetEvent::KeyPressed,
                                          this, vtkChipsWidget::KeyPressAction);

  this->CallbackMapper->SetCallbackMethod(vtkCommand::KeyReleaseEvent,
                                          vtkChipsWidgetEvent::KeyReleased,
                                          this, vtkChipsWidget::KeyReleaseAction);

   this->SetPriority (.9);
}

//-------------------------------------------------------------------------
vtkChipsWidget::~vtkChipsWidget()
{
  this->SetEnabled (0);
  if (this->objectInfo)
    delete this->objectInfo;

  if (this->finalizeSelect)
    this->finalizeSelect->Delete ();


}

void vtkChipsWidget::SetEnabled(int enabling)
{
  if (enabling == this->Enabled)
    return;

  if (this->Interactor == NULL)
    return;

  if (this->Interactor->GetRenderWindow () == NULL)
    return;

  int manageCursor = this->ManagesCursor;

  // don't initialize the cursor as the representations may not be setup.
  this->ManagesCursor = 0;
  this->Superclass::SetEnabled (enabling);
  this->ManagesCursor = manageCursor;

  if (this->Enabled)
    {
    this->Interactor->AddObserver (vtkCommand::LeftButtonPressEvent, this->finalizeSelect, this->Priority-0.1);
    this->Interactor->AddObserver (vtkChipsCommand::LeftButtonDoubleClickEvent, this->finalizeSelect, this->Priority-0.1);

    }
  else
    {
    this->Interactor->RemoveObserver(this->finalizeSelect);
    }
    

}

//-------------------------------------------------------------------------
void vtkChipsWidget::SetCursor(int cState)
{
  vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);

  if (!this->Resizable)
    {
    switch (cState)
      {
      case vtkChipsRepresentation::BottomLeftHandle:
      case vtkChipsRepresentation::BottomRightHandle:
      case vtkChipsRepresentation::TopLeftHandle:
      case vtkChipsRepresentation::TopRightHandle:
      case vtkChipsRepresentation::TopHandle:
      case vtkChipsRepresentation::BottomHandle:
      case vtkChipsRepresentation::LeftHandle:
      case vtkChipsRepresentation::RightHandle:
        return;
        break;
      default:
        break;
      }
    }


  if (!this->Moveable)
    {
    switch (cState)
      {
      case vtkChipsRepresentation::TopEdge:
      case vtkChipsRepresentation::BottomEdge:
      case vtkChipsRepresentation::LeftEdge:
      case vtkChipsRepresentation::RightEdge:
      case vtkChipsRepresentation::On:
        return;
        break;
      default:
        break;
      }

    }

  if (!this->Rotateable)
    {
    switch (cState)
      {
      case vtkChipsRepresentation::RotateHandle:
        return;
        break;
      default:
        break;
      }

    }

  switch (cState)
    {
    // calling "SetCurrentCursor should not be done here. However, to 
    // would mean editting vtkObserverMediator which is less desirable
    // Subclassing the class doesn't work because the function which needs
    // modifying isn't virtual.
    case vtkChipsRepresentation::BottomLeftHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZESW);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZESW);
      break;
    case vtkChipsRepresentation::BottomRightHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZESE);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZESE);
      break;
    case vtkChipsRepresentation::TopLeftHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZENW);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZENW);
      break;
    case vtkChipsRepresentation::TopRightHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZENE);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZENE);
      break;
    case vtkChipsRepresentation::TopHandle:
    case vtkChipsRepresentation::BottomHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZENS);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZENS);
      break;
    case vtkChipsRepresentation::LeftHandle:
    case vtkChipsRepresentation::RightHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_SIZEWE);
      else
        this->RequestCursorShape(VTK_CURSOR_SIZEWE);
      break;
    case vtkChipsRepresentation::TopEdge:
    case vtkChipsRepresentation::BottomEdge:
    case vtkChipsRepresentation::LeftEdge:
    case vtkChipsRepresentation::RightEdge:
    case vtkChipsRepresentation::On:
      if ( guiInteractor )
        guiInteractor->SetCurrentCursor(VTK_CURSOR_FLEUR);
      else
        this->RequestCursorShape(VTK_CURSOR_HAND);
      break;        
    case vtkChipsRepresentation::RotateHandle:
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_ROTATE);
      else
        this->RequestCursorShape(VTK_CURSOR_DEFAULT);
      break;

    default:
      if ( guiInteractor )
        guiInteractor->SetCurrentCursor(VTK_CURSOR_DEFAULT);
      else
        this->RequestCursorShape(VTK_CURSOR_DEFAULT);
    }
}


//-------------------------------------------------------------------------
void vtkChipsWidget::DoubleClickSelectAction(vtkAbstractWidget *w)
{
  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);

  if ( self->SubclassDoubleClickSelectAction() )
    return;

  if (event_manager->get_state () != ChipsEventManager::SelectState)
  {
     int state = ChipsEventManager::SelectState;
     double eventPos[2] = {0,0};
     event_manager->invoke_event (eventPos, ChipsEventManager::ChangeStateEvent, &state);
  }

  self->SelectAction (w);

}

//-------------------------------------------------------------------------
void vtkChipsWidget::SelectAction(vtkAbstractWidget *w)
{

  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);
  vtkChipsRepresentation *rep = reinterpret_cast<vtkChipsRepresentation*>(self->WidgetRep);

  if ( self->SubclassSelectAction() )
    return;

  if (( self->WidgetState != vtkChipsWidget::Selected ) && 
      ( !rep->IsInteractionStateOnObject () ))
    return;

  // Picked something inside the widget
  int X = self->Interactor->GetEventPosition()[0];
  int Y = self->Interactor->GetEventPosition()[1];

  double eventPos[2];
  eventPos[0] = static_cast<double>(X);
  eventPos[1] = static_cast<double>(Y);

  if (( self->WidgetRep->GetInteractionState() == vtkChipsRepresentation::Outside ) ||
      ( self->WidgetRep->GetInteractionState() == vtkChipsRepresentation::Inside ))
    {
    self->WidgetState = vtkChipsWidget::Start;
    }

  else
    {
 
    bool selected = false;
  
    // We are definitely selected
//    self->GrabFocus(self->EventCallbackCommand);

    self->WidgetRep->StartWidgetInteraction(eventPos);

    if (self->Selectable)
      selected = event_manager->request_select (*self->objectInfo);

    if (selected &&
        (event_manager->get_state () == ChipsEventManager::SelectState))
      {


      if(self->Resizable &&
         ((rep->GetInteractionState() == vtkChipsRepresentation::RightHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::LeftHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::TopHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::BottomHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::TopRightHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::TopLeftHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::BottomRightHandle) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::BottomLeftHandle)))
        {
        rep->ResizingOn ();
        }

      else if (self->Moveable &&
         ((rep->GetInteractionState() == vtkChipsRepresentation::RightEdge) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::LeftEdge) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::TopEdge) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::BottomEdge) ||
          (rep->GetInteractionState() == vtkChipsRepresentation::On)))

        {
        rep->MovingOn ();
        }
      else if (self->Rotateable &&
         (rep->GetInteractionState() == vtkChipsRepresentation::RotateHandle))
        {
        rep->RotatingOn ();
        }

      self->WidgetState = vtkChipsWidget::Modify;
      }
    }
                       
//  self->EventCallbackCommand->SetAbortFlag(1);
  self->StartInteraction();
  self->InvokeEvent(vtkCommand::StartInteractionEvent,NULL);
  self->Render ();
}

//-------------------------------------------------------------------------
void vtkChipsWidget::MoveAction(vtkAbstractWidget *w)
{
  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(self->WidgetRep);

  if ( self->SubclassMoveAction() )
    {
    return;
    }

  if ( self->WidgetState == vtkChipsWidget::Selected ) 
     self->SetCursor (self->WidgetRep->GetInteractionState());

  // compute some info we need for all cases
  int X = self->Interactor->GetEventPosition()[0];
  int Y = self->Interactor->GetEventPosition()[1];

  // Update the rep state
  if ( self->WidgetState != vtkChipsWidget::Modify)
    {
    int modify = (self->Interactor->GetControlKey ()) ? vtkKeySym::RightControl : 0;
    chipsRep->ComputeInteractionState(X, Y, modify);
    }
  else
    {

    double newEventPosition[2];
    newEventPosition[0] = static_cast<double>(X);
    newEventPosition[1] = static_cast<double>(Y);

    // handle resizing
    if(chipsRep->GetResizing ())
      {
      // Okay, adjust the representation (the widget is currently selected)

      chipsRep->WidgetInteraction(newEventPosition);

      vtkCoordinate *pos1 = chipsRep->GetPositionCoordinate ();
      vtkCoordinate *pos2 = chipsRep->GetPosition2Coordinate ();
      double *fpos1 = pos1->GetComputedDoubleDisplayValue (self->CurrentRenderer);
      double *fpos2 = pos2->GetComputedDoubleDisplayValue (self->CurrentRenderer);
      double boundingBox[4] = {fpos1[0], fpos1[1], fpos2[0], fpos2[1]};



      event_manager->invoke_event (newEventPosition, ChipsEventManager::ResizeEvent, &boundingBox[0]);

      }
    else if (chipsRep->GetMoving ())
      {

      // Okay, adjust the representation (the widget is currently selected)
      double startPosition[2];
      chipsRep->GetCurrentEventPosition (startPosition);
      chipsRep->WidgetInteraction(newEventPosition);
      event_manager->invoke_event (newEventPosition, ChipsEventManager::MoveEvent, startPosition);

      }

    else if (chipsRep->GetRotating ())
      {

      // Okay, adjust the representation (the widget is currently selected)
      double startPosition[2];
      float angle;
      chipsRep->GetCurrentEventPosition (startPosition);
      chipsRep->WidgetInteraction(newEventPosition);
      angle = chipsRep->GetAngle ();
      event_manager->invoke_event (newEventPosition, ChipsEventManager::RotateEvent, &angle);

      }



    // start a drag
    self->InvokeEvent(vtkCommand::InteractionEvent,NULL);
    self->Render();
    }

}

//-------------------------------------------------------------------------
void vtkChipsWidget::EndSelectAction(vtkAbstractWidget *w)
{
  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(self->WidgetRep);

  if ( self->SubclassEndSelectAction() )
    return;

  if ( self->WidgetState != vtkChipsWidget::Modify )
    return;


  if (chipsRep->GetResizing ())
    chipsRep->ResizingOff ();
  if (chipsRep->GetMoving ())
    chipsRep->MovingOff ();
  if (chipsRep->GetRotating ())
    chipsRep->RotatingOff ();

  double newEventPosition[2] = {self->Interactor->GetEventPosition()[0],
                                  self->Interactor->GetEventPosition()[1]};
  // Return state to not modifying
//  self->ReleaseFocus();
  self->WidgetState = vtkChipsWidget::Selected;
  chipsRep->EndWidgetInteraction (newEventPosition);
  

  // stop adjusting
//  self->EventCallbackCommand->SetAbortFlag(1);
  self->EndInteraction();
  self->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);
}

//-------------------------------------------------------------------------
void vtkChipsWidget::KeyPressAction(vtkAbstractWidget *w)
{

  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);

  if ( self->SubclassKeyPressAction() )
    return;

  unsigned char keyCode = self->Interactor->GetKeyCode ();

  switch (keyCode)
    {
    case vtkKeySym::Up:
    case vtkKeySym::Down:
    case vtkKeySym::Left:
    case vtkKeySym::Right:
      self->KeyPressMove (keyCode);
    break;
    default:
    break;
    }

}


//-------------------------------------------------------------------------
void vtkChipsWidget::KeyReleaseAction(vtkAbstractWidget *w)
{

  vtkChipsWidget *self = reinterpret_cast<vtkChipsWidget*>(w);

  if ( self->SubclassKeyReleaseAction() )
    return;
}


void vtkChipsWidget::KeyPressMove (unsigned char key)
{
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(this->WidgetRep);

  if ((this->WidgetState != vtkChipsWidget::Selected) ||
      !this->Moveable)
    return;

  chipsRep->MovingOn ();

  // convert to normalized viewport coordinates
  double eventPos[2];
  double startPosition[2];
  eventPos[0] = static_cast<double>(this->Interactor->GetEventPosition()[0]);
  eventPos[1] = static_cast<double>(this->Interactor->GetEventPosition()[1]);
  chipsRep->StartWidgetInteraction(eventPos);

  switch (key)
    {
    case vtkKeySym::Up:
      eventPos[1] += 1;
      break;
    case vtkKeySym::Down:
      eventPos[1] -= 1;
      break;
    case vtkKeySym::Left:
      eventPos[0] -= 1;
      break;
    case vtkKeySym::Right:
      eventPos[0] += 1;
      break;
    }

    chipsRep->GetCurrentEventPosition (startPosition);
    chipsRep->WidgetInteraction(eventPos);
    event_manager->invoke_event (eventPos, ChipsEventManager::MoveEvent, startPosition);


  // start a drag
  this->EventCallbackCommand->SetAbortFlag(1);
  this->InvokeEvent(vtkCommand::InteractionEvent,NULL);
  this->Render();


}

//----------------------------------------------------------------------
void vtkChipsWidget::Select (bool flag)
{
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(this->WidgetRep);

  chipsRep->SetShowBorder (flag);

  if (flag)
    {
    // Move the widget to the top of the render order
    if (this->CurrentRenderer)
      {
      this->CurrentRenderer->RemoveViewProp (this->WidgetRep);
      this->CurrentRenderer->AddViewProp (this->WidgetRep);
      }
    if (this->WidgetState == vtkChipsWidget::Start)
       this->WidgetState = vtkChipsWidget::Selected;
    }
  else
    {
    this->WidgetState = vtkChipsWidget::Start;
    chipsRep->MovingOff ();
    chipsRep->ResizingOff ();
    }
  this->Render ();

}

void vtkChipsWidget::SetRotateable (int rotateable)
{
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(this->WidgetRep);

  chipsRep->SetRotationEnabled (rotateable);
  this->Rotateable = rotateable;

}

//----------------------------------------------------------------------
void vtkChipsWidget::CreateDefaultRepresentation()
{
  if ( ! this->WidgetRep )
    {
    this->WidgetRep = vtkChipsRepresentation::New();
    }
}

//-------------------------------------------------------------------------
void vtkChipsWidget::UpdateRepresentation()
{
  vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(this->WidgetRep);

  if ( chipsRep )
    chipsRep->Update ();
}

//-------------------------------------------------------------------------
void vtkChipsWidget::FinalizeSelect ()
{
  if (!event_manager->is_selected (*this->objectInfo))
    {
    vtkChipsRepresentation *chipsRep = reinterpret_cast<vtkChipsRepresentation*>(this->WidgetRep);
    this->WidgetState = vtkChipsWidget::Start;
    chipsRep->MovingOff ();
    chipsRep->ResizingOff ();
    }
  else
    this->SetCursor(this->WidgetRep->GetInteractionState());
}

void vtkChipsWidget::SetIdInfo (const ChipsId &cid, V_OBJECT_TYPE objectType)
{
  this->SetIdInfo (cid, objectType, chips_not_subobject);
}

void vtkChipsWidget::SetIdInfo (const ChipsId &cid, V_OBJECT_TYPE objectType, V_SUBOBJECT_TYPE subObjectType)
{ 
  this->objectInfo->cid = cid; 
  this->objectInfo->object_type = objectType; 
  this->objectInfo->subobject_type = subObjectType;
};



//-------------------------------------------------------------------------
void vtkChipsWidget::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
  
  os << indent << "Selectable: " << (this->Selectable ? "On\n" : "Off\n");
  os << indent << "Resizable: " << (this->Resizable ? "On\n" : "Off\n");

  this->CallbackMapper->PrintSelf (os, indent);
}
