// 
//  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef __vtkChipsWidget_h
#define __vtkChipsWidget_h

#include "vtkAbstractWidget.h"
#include "ChipsId.hh"

class vtkChipsRepresentation;
class vtkFinalizeSelectCallback;
class ChipsEventObjectInfo;


class VTK_WIDGETS_EXPORT vtkChipsWidget : public vtkAbstractWidget
{
public:
  // Description:
  // Method to instantiate class.
  static vtkChipsWidget *New();

  // Description;
  // Standard methods for class.
  vtkTypeMacro(vtkChipsWidget,vtkAbstractWidget);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Indicate whether the interior region of the widget can be selected or
  // not. If not, then events (such as left mouse down) allow the user to
  // "move" the widget, and no selection is possible. Otherwise the
  // SelectRegion() method is invoked.
  vtkSetMacro(Selectable,int);
  vtkGetMacro(Selectable,int);
  vtkBooleanMacro(Selectable,int);


  // Description:
  // Indicate whether the boundary of the widget can be resized.
  // If not, the cursor will not change to "resize" type when mouse
  // over the boundary.
  vtkSetMacro(Resizable,int);
  vtkGetMacro(Resizable,int);
  vtkBooleanMacro(Resizable,int);

  // Description:
  // Indicate whether the boundary of the widget can be rotated.
  // If not, the cursor will not change to "rotate" type when mouse
  // over the boundary.
  void SetRotateable (int rotateable);
  vtkGetMacro(Rotateable,int);
  vtkBooleanMacro(Rotateable,int);

  // Description:
  // Indicate whether the boundary of the widget can be rotated.
  // If not, the cursor will not change to "rotate" type when mouse
  // over the boundary.
  vtkSetMacro(Moveable,int);
  vtkGetMacro(Moveable,int);
  vtkBooleanMacro(Moveable,int);


  // Description:
  // Specify an instance of vtkWidgetRepresentation used to represent this
  // widget in the scene. Note that the representation is a subclass of vtkProp
  // so it can be added to the renderer independent of the widget.
  void SetRepresentation(vtkChipsRepresentation *r)
    {this->Superclass::SetWidgetRepresentation(reinterpret_cast<vtkWidgetRepresentation*>(r));}

  // Description:
  // Specify the chips id and object type of the referring chips object
  void SetIdInfo (const ChipsId &cid, V_OBJECT_TYPE objectType, V_SUBOBJECT_TYPE subObjectType);
  void SetIdInfo (const ChipsId &cid, V_OBJECT_TYPE objectType);
  
  // Description:
  // Create the default widget representation if one is not set. 
  virtual void CreateDefaultRepresentation();

  // Description:
  // Update the representative with any changes to the system.
  virtual void UpdateRepresentation();

  // Description:
  // Request the widget to execute a select action
  void Select (bool flag);

  // Description:
  // Methods that satisfy the superclass' API.
  virtual void SetEnabled(int);

  friend class vtkFinalizeSelectCallback;


protected:
  vtkChipsWidget();
  ~vtkChipsWidget();

  //enable the selection of the region interior to the widget
  int Selectable;
  int Resizable;
  int Moveable;
  int Rotateable;

  //processes the registered events
  static void SelectAction(vtkAbstractWidget*);
  static void EndSelectAction(vtkAbstractWidget*);
  static void DoubleClickSelectAction(vtkAbstractWidget*);
  static void MoveAction(vtkAbstractWidget*);
  static void KeyPressAction(vtkAbstractWidget*);
  static void KeyReleaseAction(vtkAbstractWidget*);



  // Special internal methods to support subclasses handling events.
  // If a non-zero value is returned, the subclass is handling the event.
  virtual int SubclassSelectAction() {return 0;}
  virtual int SubclassEndSelectAction() {return 0;}
  virtual int SubclassDoubleClickSelectAction() {return 0;}
  virtual int SubclassMoveAction() {return 0;}
  virtual int SubclassKeyPressAction() {return 0;};
  virtual int SubclassKeyReleaseAction() {return 0;};

  
  // helper methods for cursoe management
  virtual void SetCursor(int State);

  // helper method for moving via keypress
  virtual void KeyPressMove (unsigned char key);

  // helper to finalize selection
  virtual void FinalizeSelect ();

//BTX
  //widget state
  int WidgetState;
  enum _WidgetState{Start=0,Modify,Selected};
//ETX

  ChipsEventObjectInfo *objectInfo;

   vtkFinalizeSelectCallback *finalizeSelect;
/*  static int finalizeSelectCount;
*/

private:
  vtkChipsWidget(const vtkChipsWidget&);  //Not implemented
  void operator=(const vtkChipsWidget&);  //Not implemented
};

#endif
