/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkInteractorStyleChips.cc,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkInteractorStyleChips.hh"

#include "vtkCallbackCommand.h"
#include "vtkCommand.h"
#include "vtkObjectFactory.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkAssemblyPath.h"
#include "vtkAbstractPropPicker.h"
#include "vtkAreaPicker.h"
#include "vtkRenderer.h"

#include "vtkPoints.h"
#include "vtkCellArray.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkActor2D.h"
#include "vtkProperty2D.h"
#include "vtkTransform.h"
#include "vtkTransformPolyDataFilter.h"
#include "vtkActor2DCollection.h"

#include "vtkRenderWindow.h"
#include "vtkGtkRenderWindowInteractor.hh"

#include "vtkKeySym.hh"

#include "vtkChipsWidget.hh"
#include "ChipsEventManager.hh"
#include "ChipsEventObserver.hh"
#include "ChipsDefines.hh"   // needed for pick masks


extern ChipsEventManager* event_manager;


vtkCxxRevisionMacro(vtkInteractorStyleChips, "$Revision: 1.25 $");
vtkStandardNewMacro(vtkInteractorStyleChips);

class vtkScrollCountCallback : public vtkCommand
{
public:
  static vtkScrollCountCallback *New ()
    {
    vtkScrollCountCallback *cb = new vtkScrollCountCallback;
    cb->TimerId = 0;
    cb->Parent = NULL;
    cb->RunCount = 0;
    return cb;
    }

  virtual void Execute (vtkObject *caller, unsigned long eventId, void *callData)
    {
    if (!Parent)
      return;

    if (vtkCommand::TimerEvent == eventId)
      {
      int tid = *static_cast<int *>(callData);

      if (tid == this->TimerId)
        {
        this->Parent->resetScrollCallback = true;
        this->TimerId = 0;
        this->RunCount++;
        }
      }
    }

  void SetTimerId (int tid) {this->TimerId = tid;};
  void SetParent (vtkInteractorStyleChips *parent) {this->Parent = parent;};
  int GetRunCount () {return this->RunCount;};
  void ResetRunCount () {this->RunCount=0;};

protected:
  int TimerId; 
  int RunCount;
  vtkInteractorStyleChips *Parent;
   
   
};

class vtkScrollCountResetCallback : public vtkCommand
{
public:
  static vtkScrollCountResetCallback *New ()
    {
    vtkScrollCountResetCallback *cb = new vtkScrollCountResetCallback;
    cb->Parent = NULL;
    return cb;
    }

  virtual void Execute (vtkObject *caller, unsigned long eventId, void *callData
)
    {
    if (!Parent)
      return;

    if (vtkCommand::TimerEvent == eventId)
      {
      int tid = *static_cast<int *>(callData);

      if (tid== this->TimerId)
        this->Parent->scrollCallback->ResetRunCount ();
      }
    }

  void SetTimerId (int tid) {this->TimerId = tid;};
  int GetTimerId () {return this->TimerId;};
  void SetParent (vtkInteractorStyleChips *parent) {this->Parent = parent;};

protected:
  int TimerId;
  vtkInteractorStyleChips *Parent;
};



class vtkInteractorEventObserver : ChipsEventObserver
{
public:
  vtkInteractorEventObserver (vtkInteractorStyleChips *parent)
  {
    this->register_event (ChipsEventManager::ChangeStateEvent);
    this->register_event (ChipsEventManager::StartPickEvent);
    this->parent = parent;

    this->pickmask = 0;
  }

  void execute_state_event ()
  {
    int state = event_manager->get_state ();
    vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (parent->Interactor);

    switch (state)
      {
      case ChipsEventManager::PanState:
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_HAND);
        else
          parent->RequestCursorShape(VTK_CURSOR_HAND);
      break;
      case ChipsEventManager::DragZoomState:
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_DOUBLEARROW);
        else
          parent->RequestCursorShape(VTK_CURSOR_DEFAULT);
      break;
      case ChipsEventManager::AddState:
      case ChipsEventManager::RangeZoomState:
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_CROSSHAIR);
        else
          parent->RequestCursorShape(VTK_CURSOR_CROSSHAIR);
      break;
      default:
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_DEFAULT);
        else
          parent->RequestCursorShape(VTK_CURSOR_DEFAULT);
      break;

      }
  }

  void execute_event (int event, void *data)
  {
    switch (event)
      {
      case ChipsEventManager::ChangeStateEvent:
        this->execute_state_event ();
        break;
      case ChipsEventManager::StartPickEvent:
        {
        this->pickmask = *(int*) data;
        }
      }
  }

  int get_pickmask () {return this->pickmask;};

protected:
  vtkInteractorStyleChips *parent;
  int pickmask;
};



//----------------------------------------------------------------------------
vtkInteractorStyleChips::vtkInteractorStyleChips() 
{
  this->Tolerance = 2;
  this->EventPosition[0] = 0;
  this->EventPosition[1] = 0;

  this->eventObserver = new vtkInteractorEventObserver (this);
  this->scrollCallback = vtkScrollCountCallback::New ();
  this->scrollCallback->SetParent (this);
  this->scrollResetCallback = vtkScrollCountResetCallback::New ();
  this->scrollResetCallback->SetParent (this);

  this->scrollCallbackAdded = false;
  this->ScrollCount = 0;
  this->MaxScrollCount = 20;
  this->resetScrollCallback = true;
/*
  vtkPoints *annotationPoints = vtkPoints::New();
  annotationPoints->SetDataTypeToDouble();
  annotationPoints->SetNumberOfPoints(2);
  annotationPoints->SetPoint(0, 0.0, 0.0, 0.0);
  annotationPoints->SetPoint(1, 1.0, 0.0, 0.0);
  vtkCellArray *annotationLine = vtkCellArray::New();
  annotationLine->InsertNextCell(5);
  for (int ii=0; ii<2; ii++)
    annotationLine->InsertCellPoint(ii);
  annotationLine->InsertCellPoint(0);

  AnnotationPolyData = vtkPolyData::New ();
  AnnotationPolyData->SetPoints(annotationPoints);
  AnnotationPolyData->SetLines(annotationLine);

  annotationPoints->Delete ();
  annotationLine->Delete();

  this->AnnotationMapper = vtkPolyDataMapper2D::New();
  this->AnnotationMapper->SetInput(this->AnnotationPolyData);
  this->AnnotationActor = vtkActor2D::New();
  this->AnnotationActor->SetMapper(this->AnnotationMapper);

  this->AnnotationProperty = vtkProperty2D::New();
  this->AnnotationActor->SetProperty(this->AnnotationProperty);

  this->AnnotationProperty->SetColor (1, 1, 1);
  this->AnnotationActor->SetVisibility (0);

  AnnotationPoints.reserve (100);
*/
  
}

//----------------------------------------------------------------------------
vtkInteractorStyleChips::~vtkInteractorStyleChips() 
{
  delete this->eventObserver;
/*
  this->AnnotationPolyData->Delete ();
  this->AnnotationMapper->Delete ();
  this->AnnotationActor->Delete ();
  this->AnnotationProperty->Delete ();
*/
}

void vtkInteractorStyleChips::OnKeyPress ()
{ 
  vtkRenderWindowInteractor *rwi = this->Interactor;
  if (event_manager->get_state () == ChipsEventManager::PickState)
    {
    if (((this->eventObserver->get_pickmask () & KEY_PRESS) == KEY_PRESS) &&
        (rwi->GetKeyCode() != vtkKeySym::Esc))
      {
      int pickinfo[] = {KEY_PRESS,(int)(rwi->GetKeySym ()[0])};
      event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
      }
    }
}

void vtkInteractorStyleChips::OnKeyRelease ()
{
  vtkRenderWindowInteractor *rwi = this->Interactor;
  if (event_manager->get_state () == ChipsEventManager::PickState)
    {
    if (((this->eventObserver->get_pickmask () & KEY_RELEASE) == KEY_RELEASE) &&
        (rwi->GetKeyCode() != vtkKeySym::Esc))
      {
      int pickinfo[] = {KEY_PRESS,(int)(rwi->GetKeySym ()[0])};
      event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
      }
    }
}


void vtkInteractorStyleChips::OnChar()
{
  if (!event_manager) return;

  vtkRenderWindowInteractor *rwi = this->Interactor;
  int state = event_manager->get_state ();

  int *eventPos = rwi->GetEventPosition();
  this->EventPosition[0] = eventPos[0];
  this->EventPosition[1] = eventPos[1];

  if (state == ChipsEventManager::AddState && event_manager->get_add_substate()
      == ChipsEventManager::LabelAdd) 
  {
    return; // pass through - keyboard input handled as label input
  }

  switch (rwi->GetKeyCode())
    {
    case vtkKeySym::r:
      {
      if (state != ChipsEventManager::PickState)
        {
        state = ChipsEventManager::RangeZoomState;
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::ChangeStateEvent, (void *)&state);
        }
      }
    break;
    case vtkKeySym::p:
      {
      if (state != ChipsEventManager::PickState)
        {
        state = ChipsEventManager::PanState;
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::ChangeStateEvent, (void *)&state);
        }
      }
    break;
    case vtkKeySym::s:
      {
      if (state != ChipsEventManager::PickState)
        {
        state = ChipsEventManager::SelectState;
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::ChangeStateEvent, (void *)&state);
        }
      }
    break;
    case vtkKeySym::Backspace:  // this is for the mac
    case vtkKeySym::Delete:
      {
      event_manager->invoke_event (this->EventPosition, ChipsEventManager::DeleteEvent, NULL);
      }
    break;
    case vtkKeySym::Plus:
      {
      float zoomAmt = 1.1;
      event_manager->invoke_event (this->EventPosition, ChipsEventManager::ZoomEvent, (void *)&zoomAmt);
      }
    break;
    case vtkKeySym::Minus:
      {
      float zoomAmt = 0.9;
      event_manager->invoke_event (this->EventPosition, ChipsEventManager::ZoomEvent, (void *)&zoomAmt);
      }
    break;
    case vtkKeySym::Esc:
      {
      if (state == ChipsEventManager::PickState)
        {
        int pickinfo[] = {KEY_PRESS,27};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      else if (state == ChipsEventManager::AddState)
        {
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::EndAddEvent, NULL);
        }
      else if (state == ChipsEventManager::SelectState)
        {
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::UnselectAllEvent, NULL);
        }
      }
    break;
    default:
    break;
    }
}


void vtkInteractorStyleChips::OnLeftButtonDown()
{

  if (!event_manager) return;

  vtkRenderWindowInteractor *rwi = this->Interactor;
  int state = event_manager->get_state ();
  this->LeftMousePress = true;

/*  event_manager->finalize_select ();
*/

  int *eventPos = rwi->GetEventPosition();
  this->EventPosition[0] = eventPos[0];
  this->EventPosition[1] = eventPos[1];


  switch (state)
    {

    case ChipsEventManager::RangeZoomState:
    case ChipsEventManager::SelectState:
      {
      event_manager->start_event (this->EventPosition);
      }
    break;
    case ChipsEventManager::PanState:
      {
      if (!event_manager->is_valid_state ())
        this->LeftMousePress = false;
      else
        {
        vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_DRAG);
        else
          this->RequestCursorShape(VTK_CURSOR_DEFAULT);

        event_manager->start_event (this->EventPosition);
        }
      }
    break;
    case ChipsEventManager::AddState:
      {
      vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_CROSSHAIR);
      else
        this->RequestCursorShape(VTK_CURSOR_CROSSHAIR);
      }
    break;
    case ChipsEventManager::DragZoomState:
      {
      if (!event_manager->is_valid_state ())
        this->LeftMousePress = false;
      else
        {
        event_manager->start_event (this->EventPosition);
        }
      }
    break;
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN1_DOWN) == BTN1_DOWN)
        {
        int pickinfo[] = {BTN1_DOWN, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    case ChipsEventManager::NoState:
    default:
    break;
    }
}


void vtkInteractorStyleChips::OnMouseMove()
{
  if (!event_manager) return;

  int X = this->Interactor->GetEventPosition()[0];
  int Y = this->Interactor->GetEventPosition()[1];

  vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);

  int state = event_manager->get_state ();

  double lastPosition[2] = {this->EventPosition[0], this->EventPosition[1]};
  this->EventPosition[0] = X;
  this->EventPosition[1] = Y;

  event_manager->display_event_position (this->EventPosition);

  if ((!this->LeftMousePress) || (state == ChipsEventManager::SelectState))
    {
    this->FindPokedRenderer(X, Y);
    if (this->CurrentRenderer)
      {
      this->Interactor->StartPickCallback();
      vtkAbstractPropPicker *picker =
            vtkAbstractPropPicker::SafeDownCast(this->Interactor->GetPicker());
      vtkAreaPicker *areaPicker = vtkAreaPicker::SafeDownCast (picker);
      if (areaPicker)
        {
        areaPicker->AreaPick(X-this->Tolerance, Y-this->Tolerance,
                   X+this->Tolerance, Y+this->Tolerance,
                   this->CurrentRenderer);
        }
      else if (picker)
        {
        picker->Pick(X, Y, 0, this->CurrentRenderer);
        }
      this->Interactor->EndPickCallback();
      }
    }


  switch (state)
    {

    case ChipsEventManager::SelectState:
      break;
    case ChipsEventManager::PanState:
      {
      if (this->LeftMousePress)
        {
        if (event_manager->is_valid_state ())
          {
          event_manager->invoke_event (this->EventPosition, ChipsEventManager::PanEvent, (void *)&lastPosition[0]);
          }
        }
      else
        {
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_HAND);
        else
          this->RequestCursorShape(VTK_CURSOR_HAND);
        }
      }
    break;
    case ChipsEventManager::DragZoomState:
      {
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_DOUBLEARROW);
      else
        this->RequestCursorShape(VTK_CURSOR_DEFAULT);

      if (this->LeftMousePress)
        {
        if (event_manager->is_valid_state ())
          {
          float zoomAmt = 0;
          float dy = this->EventPosition[1] - lastPosition[1];
          if (dy > 0)
             zoomAmt = 1.1;
          else if (dy < 0)
             zoomAmt = 0.9;
          
          if (zoomAmt != 0)
             event_manager->invoke_event (this->EventPosition, ChipsEventManager::ZoomEvent, (void *)&zoomAmt);

          }
        }
      }
    break;
    case ChipsEventManager::AddState:
      if (event_manager->get_add_substate() != ChipsEventManager::LabelAdd)
      {
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_CROSSHAIR);
        else
          this->RequestCursorShape(VTK_CURSOR_CROSSHAIR);
      }
    break;
    case ChipsEventManager::RangeZoomState:
      {
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_CROSSHAIR);
      else
        this->RequestCursorShape(VTK_CURSOR_CROSSHAIR);
      }
    break;
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & MOTION_NOTIFY) == MOTION_NOTIFY)
        {
        int pickinfo[] = {MOTION_NOTIFY, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    case ChipsEventManager::NoState:
    default:
    break;
    }

}

void vtkInteractorStyleChips::OnMouseWheelForward()
{
   if (!event_manager) return;


   if (this->resetScrollCallback)
     {
     if (!this->scrollCallbackAdded)
       {
       this->Interactor->AddObserver (vtkCommand::TimerEvent, this->scrollCallback);
       this->Interactor->AddObserver (vtkCommand::TimerEvent, this->scrollResetCallback);
       this->scrollCallbackAdded = true;
       }
     if (this->scrollCallback->GetRunCount () == 1)
       {
       this->ScrollCount = this->MaxScrollCount;
       int tid = this->Interactor->CreateOneShotTimer (20);
       this->scrollCallback->SetTimerId (tid);

       // if we've recieved a lot of scroll events remove the
       // reset callback
       this->Interactor->DestroyTimer (this->scrollResetCallback->GetTimerId ());
       }
     else
       {
       this->ScrollCount = 0;
       this->scrollCallback->ResetRunCount ();
       int tid = this->Interactor->CreateOneShotTimer (150);
       this->scrollCallback->SetTimerId (tid);

       // this ensures that if the scroll are coming in slowly that
       // they are reset periodically
       tid = this->Interactor->CreateOneShotTimer (200);
       this->scrollResetCallback->SetTimerId (tid);

       }


     this->resetScrollCallback = false;
     }

   if (this->ScrollCount == this->MaxScrollCount)
     return;

   this->ScrollCount++;

   this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
   this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

   float zoomAmt = 1.1;
   

   event_manager->invoke_event (this->EventPosition, ChipsEventManager::ZoomEvent, (void *)&zoomAmt);
}

void vtkInteractorStyleChips::OnMouseWheelBackward()
{
   if (!event_manager) return;

   if (this->resetScrollCallback)
     {
     if (!this->scrollCallbackAdded)
       {
       this->Interactor->AddObserver (vtkCommand::TimerEvent, this->scrollCallback);
       this->Interactor->AddObserver (vtkCommand::TimerEvent, this->scrollResetCallback);
       this->scrollCallbackAdded = true;
       }

     if (this->scrollCallback->GetRunCount () == 1)
       {
       this->ScrollCount = this->MaxScrollCount;
       int tid = this->Interactor->CreateOneShotTimer (20);
       this->scrollCallback->SetTimerId (tid);

       // if we've recieved a lot of scroll events remove the 
       // reset callback 
       this->Interactor->DestroyTimer (this->scrollResetCallback->GetTimerId ());
       }
     else
       {
       this->ScrollCount = 0;
       this->scrollCallback->ResetRunCount ();
       int tid = this->Interactor->CreateOneShotTimer (150);
       this->scrollCallback->SetTimerId (tid);

       // this ensures that if the scroll are coming in slowly that 
       // they are reset periodically 
       tid = this->Interactor->CreateOneShotTimer (200);
       this->scrollResetCallback->SetTimerId (tid);
       }

     this->resetScrollCallback = false;
     }

   if (this->ScrollCount == this->MaxScrollCount)
     return;

   this->ScrollCount++;


   this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
   this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

   float zoomAmt = 0.9;


   event_manager->invoke_event (this->EventPosition, ChipsEventManager::ZoomEvent, (void *)&zoomAmt);

}


void vtkInteractorStyleChips::OnLeftButtonUp()
{

  if (!event_manager) return;

  int state = event_manager->get_state ();


  switch (state)
    {
    case ChipsEventManager::RangeZoomState:
    case ChipsEventManager::SelectState:
      {
      event_manager->end_event ();
      }
    break;
    case ChipsEventManager::PanState:
      {
      if (this->LeftMousePress)
        event_manager->end_event ();

      vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);
      if (guiInteractor)
        guiInteractor->SetCurrentCursor(VTK_CURSOR_HAND);
      else
        this->RequestCursorShape(VTK_CURSOR_HAND);

      }
    break;
    case ChipsEventManager::DragZoomState:
      {
      if (this->LeftMousePress)
        event_manager->end_event ();
      }
    break;
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN1_UP) == BTN1_UP)
        {
        int pickinfo[] = {BTN1_UP, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    case ChipsEventManager::AddState:
      // change the cursor while entering text via add_label interaction
      if (event_manager->get_add_substate() == ChipsEventManager::LabelAdd)
      {
        vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);
        if (guiInteractor)
          guiInteractor->SetCurrentCursor(VTK_CURSOR_PENCIL);
        else
          this->RequestCursorShape(VTK_CURSOR_PENCIL);
      }
    break;
    case ChipsEventManager::NoState:
    default:
    break;
    }

    this->LeftMousePress = false;

}

void vtkInteractorStyleChips::OnMiddleButtonDown()
{
  if (!event_manager) return;
  int state = event_manager->get_state ();

  this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
  this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

  switch (state)
    {
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN2_DOWN) == BTN2_DOWN)
        {
        int pickinfo[] = {BTN2_DOWN, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    default:
    break;
    }
}

void vtkInteractorStyleChips::OnMiddleButtonUp()
{
  if (!event_manager) return;
  int state = event_manager->get_state ();

  this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
  this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

  switch (state)
    {
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN2_UP) == BTN2_UP)
        {
        int pickinfo[] = {BTN2_UP, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    default:
    break;
    }
}

void vtkInteractorStyleChips::OnRightButtonDown()
{
  if (!event_manager) return;

  this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
  this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

  int state = event_manager->get_state ();

  switch (state)
    {
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN3_DOWN) == BTN3_DOWN)
        {
        int pickinfo[] = {BTN3_DOWN, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    default:
      event_manager->invoke_event (EventPosition, ChipsEventManager::RightClickEvent, NULL);
    break;
    }
}

void vtkInteractorStyleChips::OnRightButtonUp()
{
  if (!event_manager) return;
  int state = event_manager->get_state ();

  this->EventPosition[0] = this->Interactor->GetEventPosition()[0];
  this->EventPosition[1] = this->Interactor->GetEventPosition()[1];

  switch (state)
    {
    case ChipsEventManager::PickState:
      {
      if ((this->eventObserver->get_pickmask () & BTN3_UP) == BTN3_UP)
        {
        int pickinfo[] = {BTN3_UP, -1};
        event_manager->invoke_event (this->EventPosition, ChipsEventManager::PickEvent, (void *)&pickinfo);
        }
      }
    break;
    default:
    break;
    }
}



void vtkInteractorStyleChips::OnUser()
{
//   cout << "User event" << endl;
}

void vtkInteractorStyleChips::OnEnter()
{
/*
  // reset the cursor style when you enter the window
  vtkGtkRenderWindowInteractor *guiInteractor = vtkGtkRenderWindowInteractor::SafeDownCast (this->Interactor);
  if (guiInteractor)
    guiInteractor->SetCurrentCursor(VTK_CURSOR_DEFAULT);
*/
}


void vtkInteractorStyleChips::OnLeave()
{
//   cout << "Elvis has left the building" << endl;
}

void vtkInteractorStyleChips::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);
}
