/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkOpenGLPolyDataMapper2D.cxx

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkOpenGLPolyDataMapper2D.h"

#include "vtkActor2D.h"
#include "vtkCellArray.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
#include "vtkPlane.h"
#include "vtkPlaneCollection.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"
#include "vtkProperty2D.h"
#include "vtkScalarsToColors.h"
#include "vtkUnsignedCharArray.h"
#include "vtkViewport.h"
#include "vtkWindow.h"
#include "vtkgluPickMatrix.h"
#include "vtkToolkits.h"  // for VTK_USE_GL2PS

#ifdef VTK_USE_GL2PS
#ifdef CXC_USE_CHIPS
#include <list>  
using namespace std;
#endif // CXC_USE_CHIPS
#include "vtk_gl2ps.h"
#include "vtkGL2PSExporter.h"
#endif // VTK_USE_GL2PS

#include <math.h>

#ifndef VTK_IMPLEMENT_MESA_CXX
vtkStandardNewMacro(vtkOpenGLPolyDataMapper2D);
#endif

void vtkOpenGLPolyDataMapper2D::RenderOverlay(vtkViewport* viewport,
                                              vtkActor2D* actor)
{
  int            numPts;
  vtkPolyData    *input=static_cast<vtkPolyData *>(this->GetInput());
  int            j;
  vtkPoints      *p, *displayPts;
  vtkCellArray   *aPrim;
  vtkUnsignedCharArray *c=NULL;
  unsigned char  *rgba;
  unsigned char  color[4];
  vtkIdType      *pts = 0;
  vtkIdType      npts = 0;
  int            cellScalars = 0;
  int            cellNum = 0;
  vtkPlaneCollection *clipPlanes;
  vtkPlane           *plane;
  int                 i,numClipPlanes;
  double              planeEquation[4];
  vtkDataArray*  t = 0;

  vtkDebugMacro (<< "vtkOpenGLPolyDataMapper2D::Render");

  if ( input == NULL ) 
    {
    vtkErrorMacro(<< "No input!");
    return;
    }
  else
    {
    input->Update();
    numPts = input->GetNumberOfPoints();
    } 

  if (numPts == 0)
    {
    vtkDebugMacro(<< "No points!");
    return;
    }
  
  if ( this->LookupTable == NULL )
    {
    this->CreateDefaultLookupTable();
    }

  // Texture and color by texture
  t = input->GetPointData()->GetTCoords();
  if ( t ) 
    {
    int tDim = t->GetNumberOfComponents();
    if (tDim != 2)
      {
      vtkDebugMacro(<< "Currently only 2d textures are supported.\n");
      t = 0;
      }
    }

  // if something has changed regenrate colors and display lists
  // if required
  //
  if ( this->GetMTime() > this->BuildTime || 
       input->GetMTime() > this->BuildTime || 
       this->LookupTable->GetMTime() > this->BuildTime ||
       actor->GetProperty()->GetMTime() > this->BuildTime)
    {
    // sets this->Colors as side effect
    this->MapScalars(actor->GetProperty()->GetOpacity());
    this->BuildTime.Modified();
    }

  // Get the position of the actor
  int size[2];
  size[0] = viewport->GetSize()[0];
  size[1] = viewport->GetSize()[1];
  double *vport = viewport->GetViewport();
  int* actorPos = 
    actor->GetPositionCoordinate()->GetComputedViewportValue(viewport);

  // get window info
  double *tileViewPort = viewport->GetVTKWindow()->GetTileViewport();
  double visVP[4];
  visVP[0] = (vport[0] >= tileViewPort[0]) ? vport[0] : tileViewPort[0];
  visVP[1] = (vport[1] >= tileViewPort[1]) ? vport[1] : tileViewPort[1];
  visVP[2] = (vport[2] <= tileViewPort[2]) ? vport[2] : tileViewPort[2];
  visVP[3] = (vport[3] <= tileViewPort[3]) ? vport[3] : tileViewPort[3];
  if (visVP[0] >= visVP[2])
    {
    return;
    }
  if (visVP[1] >= visVP[3])
    {
    return;
    }
  size[0] = 
    vtkMath::Round(size[0]*(visVP[2] - visVP[0])/(vport[2] - vport[0]));
  size[1] = 
    vtkMath::Round(size[1]*(visVP[3] - visVP[1])/(vport[3] - vport[1]));
  
  // Set up the font color from the text actor
  double*  actorColor = actor->GetProperty()->GetColor();
  color[0] = static_cast<unsigned char>(actorColor[0] * 255.0);
  color[1] = static_cast<unsigned char>(actorColor[1] * 255.0);
  color[2] = static_cast<unsigned char>(actorColor[2] * 255.0);
  color[3] = static_cast<unsigned char>(
    255.0*actor->GetProperty()->GetOpacity());

  // Transform the points, if necessary
  p = input->GetPoints();
  if ( this->TransformCoordinate )
    {
#ifdef CXC_USE_CHIPS
    double *dtmp;
#else
    int *itmp;
#endif
    numPts = p->GetNumberOfPoints();
    displayPts = vtkPoints::New();
    displayPts->SetNumberOfPoints(numPts);
    for ( j=0; j < numPts; j++ )
      {
      this->TransformCoordinate->SetValue(p->GetPoint(j));
#ifdef CXC_USE_CHIPS
      dtmp = this->TransformCoordinate->GetComputedDoubleViewportValue(viewport);
      displayPts->SetPoint(j,dtmp[0], dtmp[1], 0.0);
#else
      itmp = this->TransformCoordinate->GetComputedViewportValue(viewport);
      displayPts->SetPoint(j,itmp[0], itmp[1], 0.0);
#endif
      }
    p = displayPts;
    }

  // Set up the coloring
  if ( this->Colors )
    {
    c = this->Colors;
    if (!input->GetPointData()->GetScalars())
      {
      cellScalars = 1;
      }
    }
  vtkDebugMacro(<< c);
  vtkDebugMacro(<< cellScalars);

  // set the colors for the foreground
  glColor4ubv(color);

  // push a 2D matrix on the stack
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  if(viewport->GetIsPicking())
    {
    vtkgluPickMatrix(viewport->GetPickX(), viewport->GetPickY(),
                     viewport->GetPickWidth(),
                     viewport->GetPickHeight(),
                     viewport->GetOrigin(), viewport->GetSize());
    }
  
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();

  if ( ! t)
    {
    glDisable( GL_TEXTURE_2D );
    }
  glDisable( GL_LIGHTING );
   
    
  // Assume we want to do Zbuffering for now.
  // we may turn this off later
  glDepthMask(GL_TRUE);

  int *winSize = viewport->GetVTKWindow()->GetSize();
  
  int xoff = static_cast<int>(actorPos[0] - (visVP[0] - vport[0])*
                              winSize[0]);
  int yoff = static_cast<int>(actorPos[1] - (visVP[1] - vport[1])*
                              winSize[1]);

  if ( actor->GetProperty()->GetDisplayLocation() == 
       VTK_FOREGROUND_LOCATION )
    {
    glOrtho(-xoff,-xoff + size[0],
            -yoff, -yoff +size[1], 0, 1);
    }  
  else
    {
    glOrtho(-xoff,-xoff + size[0],
            -yoff, -yoff + size[1], -1, 0);
    }
    
  // Clipping plane stuff
  clipPlanes = this->ClippingPlanes;

  if (clipPlanes == NULL)
    {
    numClipPlanes = 0;
    }
  else
    {
    numClipPlanes = clipPlanes->GetNumberOfItems();
    if (numClipPlanes > 4)
      {
      vtkErrorMacro(<< "Only 4 clipping planes are used with 2D mappers");
      }
    }

  for (i = 0; i < numClipPlanes; i++)
    {
    glEnable(static_cast<GLenum>(GL_CLIP_PLANE0+i));
    }

  for (i = 0; i < numClipPlanes; i++)
    {
    plane = static_cast<vtkPlane *>(clipPlanes->GetItemAsObject(i));

    planeEquation[0] = plane->GetNormal()[0];
    planeEquation[1] = plane->GetNormal()[1];
    planeEquation[2] = plane->GetNormal()[2];
    planeEquation[3] = -(planeEquation[0]*plane->GetOrigin()[0]+
                         planeEquation[1]*plane->GetOrigin()[1]+
                         planeEquation[2]*plane->GetOrigin()[2]);
    glClipPlane(static_cast<GLenum>(GL_CLIP_PLANE0+i),planeEquation);
    }

#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
  if ( numClipPlanes > 0 )
    {
    double ll[] = {-xoff,-yoff};
    double ur[] = {-xoff+size[0], -yoff +size[1]};
    double tmpViewport[] = {visVP[0] * winSize[0], visVP[1] * winSize[1],
                            visVP[2] * winSize[0], visVP[3] * winSize[1]};
    this->ClipViewport  (ll, ur, tmpViewport);
    }

#endif // VTK_USE_GL2PS
#endif // CXC_USE_CHIPS

  // Set the PointSize
  glPointSize(actor->GetProperty()->GetPointSize());

  double *dptr;
  aPrim = input->GetVerts();
  glBegin(GL_POINTS);
  for (aPrim->InitTraversal(); aPrim->GetNextCell(npts,pts); cellNum++)
    { 
    for (j = 0; j < npts; j++) 
      {
      if (c) 
        {
        if (cellScalars)
          {
          rgba = c->GetPointer(4*cellNum);
          }
        else
          {
          rgba = c->GetPointer(4*pts[j]);
          }
        glColor4ubv(rgba);
        }
      if (t)
        {
        glTexCoord2dv(t->GetTuple(pts[j]));
        }  
      // this is done to work around an OpenGL bug, otherwise we could just
      // call glVertex2dv
      dptr = p->GetPoint(pts[j]);
      glVertex3d(dptr[0],dptr[1],0);
      }
    }
  glEnd();

  // Set pointsize and linewidth for GL2PS output.
#ifdef VTK_USE_GL2PS
  gl2psPointSize(actor->GetProperty()->GetPointSize()*
                 vtkGL2PSExporter::GetGlobalPointSizeFactor());
  gl2psLineWidth(actor->GetProperty()->GetLineWidth()*
                 vtkGL2PSExporter::GetGlobalLineWidthFactor());
#endif // VTK_USE_GL2PS

  // Set the LineWidth
  glLineWidth(actor->GetProperty()->GetLineWidth());

  // Set the LineStipple
  if (actor->GetProperty()->GetLineStipplePattern() != 0xFFFF)
    {
    glEnable (GL_LINE_STIPPLE);
#ifndef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
    gl2psEnable(GL2PS_LINE_STIPPLE);
#endif // VTK_USE_GL2PS
#endif // CXC_USE_CHIPS
    glLineStipple (actor->GetProperty()->GetLineStippleRepeatFactor(), 
                   actor->GetProperty()->GetLineStipplePattern());
#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
    gl2psEnable(GL2PS_LINE_STIPPLE);
#endif // VTK_USE_GL2PS
#endif // CXC_USE_CHIPS
    }
  else
    {
    glDisable (GL_LINE_STIPPLE);
#ifdef VTK_USE_GL2PS
    gl2psDisable(GL2PS_LINE_STIPPLE);
#endif // VTK_USE_GL2PS
    }

  aPrim = input->GetLines();
  for (aPrim->InitTraversal(); aPrim->GetNextCell(npts,pts); cellNum++)
    {
#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
      gl2psEnable(GL2PS_POLYLINE);
#endif // VTK_USE_GL2PS
#endif // CXC_USE_CHIPS

    glBegin(GL_LINE_STRIP);
    for (j = 0; j < npts; j++)
      {
      if (c)
        {
        if (cellScalars)
          {
          rgba = c->GetPointer(4*cellNum);
          }
        else
          {
          rgba = c->GetPointer(4*pts[j]);
          }
        glColor4ubv(rgba);
        }
      if (t)
        {
        glTexCoord2dv(t->GetTuple(pts[j]));
        }  
      // this is done to work around an OpenGL bug, otherwise we could just
      // call glVertex2dv
      dptr = p->GetPoint(pts[j]);
      glVertex3d(dptr[0],dptr[1],0);
      }
    glEnd();

#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
    gl2psDisable(GL2PS_POLYLINE);
#endif // VTK_USE_GL2PS
#endif // CXC_USE_CHIPS

    }

#ifdef CXC_USE_CHIPS
  // Set the Polygon Stipple
  if (actor->GetProperty()->GetUsePolygonStipple())
    {
    glEnable (GL_POLYGON_STIPPLE);

    glPixelStorei (GL_UNPACK_LSB_FIRST, 1);

    glPolygonStipple (actor->GetProperty()->GetPolygonStipplePattern());

#ifdef VTK_USE_GL2PS
    gl2psEnable (GL2PS_POLYGON_STIPPLE);
#endif

    }
  else
    {
    glDisable (GL_POLYGON_STIPPLE);
#ifdef VTK_USE_GL2PS
    gl2psDisable (GL2PS_POLYGON_STIPPLE);
#endif

    }
#endif // CXC_USE_CHIPS

  aPrim = input->GetStrips();
  for (aPrim->InitTraversal(); aPrim->GetNextCell(npts,pts); cellNum++)
    {
    glBegin(GL_TRIANGLE_STRIP);
    for (j = 0; j < npts; j++)
      {
      if (c)
        {
        if (cellScalars)
          {
          rgba = c->GetPointer(4*cellNum);
          }
        else
          {
          rgba = c->GetPointer(4*pts[j]);
          }
        glColor4ubv(rgba);
        }
      if (t)
        {
        glTexCoord2dv(t->GetTuple(pts[j]));
        }  
      // this is done to work around an OpenGL bug, otherwise we could just
      // call glVertex2dv
      dptr = p->GetPoint(pts[j]);
      glVertex3d(dptr[0],dptr[1],0);
      }
    glEnd();
    }

  aPrim = input->GetPolys();
  for (aPrim->InitTraversal(); aPrim->GetNextCell(npts,pts); cellNum++)
    {
    glBegin(GL_POLYGON);
    for (j = 0; j < npts; j++)
      {
      if (c)
        {
        if (cellScalars)
          {
          rgba = c->GetPointer(4*cellNum);
          }
        else
          {
          rgba = c->GetPointer(4*pts[j]);
          }
        glColor4ubv(rgba);
        }
      if (t)
        {
        glTexCoord2dv(t->GetTuple(pts[j]));
        }  
      // this is done to work around an OpenGL bug, otherwise we could just
      // call glVertex2dv
      dptr = p->GetPoint(pts[j]);
      glVertex3d(dptr[0],dptr[1],0);
      }
    glEnd();
    }

  if ( this->TransformCoordinate )
    {
    p->Delete();
    }

  for (i = 0; i < numClipPlanes; i++)
    {
    glDisable(static_cast<GLenum>(GL_CLIP_PLANE0+i));
    }

#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
  if ( numClipPlanes > 0 )
    {
    gl2psEndClipPlane ();
    }
#endif
#endif // CXC_USE_CHIPS

  // push a 2D matrix on the stack
  glMatrixMode( GL_PROJECTION);
  glPopMatrix();
  glMatrixMode( GL_MODELVIEW);
  glPopMatrix();
  glEnable( GL_LIGHTING);
  
  // Turn it back on in case we've turned it off
  glDepthMask( GL_TRUE );
  glDisable( GL_TEXTURE_2D );
}

#ifdef CXC_USE_CHIPS
#ifdef VTK_USE_GL2PS
void vtkOpenGLPolyDataMapper2D::ClipViewport  (double *lowerLeft, double *upperRight, double *viewportPositions)
{

  if (!lowerLeft || !upperRight || !viewportPositions)
     return;

  if (this->ClippingPlanes == NULL)
     return;

  double *point1 = new double[3];
  double *point2 = new double[3];

  list <double *>clipPoints;
  list <double *>cornerPoints;

  list<double *>::iterator currPoint1;
  list<double *>::iterator currPoint2;
  list<double *>::iterator tmpPoint1;
  list<double *>::iterator tmpPoint2;

  bool *ignorePlanes = static_cast<bool *>(0);

  int numEdges = 4;
  int index = 0;
  int intersect = 1;
  double t=0;
  vtkPlane           *plane;
  int numClipPlanes = 0;


  // initialize the corner points
  double *tmp = new double[3];
  tmp[0] = lowerLeft[0];
  tmp[1] = lowerLeft[1];
  tmp[2] = 0;

  cornerPoints.push_back (tmp);

  tmp = new double[3];
  tmp[0] = lowerLeft[0];
  tmp[1] = upperRight[1];
  tmp[2] = 0;

  cornerPoints.push_back (tmp);

  tmp = new double[3];
  tmp[0] = upperRight[0];
  tmp[1] = upperRight[1];
  tmp[2] = 0;

  cornerPoints.push_back (tmp);

  tmp = new double[3];
  tmp[0] = upperRight[0];
  tmp[1] = lowerLeft[1];
  tmp[2] = 0;

  cornerPoints.push_back (tmp);

  currPoint1 = cornerPoints.begin ();
  currPoint2 = currPoint1;
  currPoint2++;

  numClipPlanes = this->ClippingPlanes->GetNumberOfItems();

  ignorePlanes = new bool[numClipPlanes];

  for (int ii=0; ii<numEdges; ii++)
    {
    // copy the corners into a temporary points
    for (int jj=0; jj<3; jj++)
      {
      point1[jj] = (*currPoint1)[jj];
      point2[jj] = (*currPoint2)[jj];
      }

    // dont ignore any planes to start clipping
    for (int jj=0; jj<numClipPlanes; jj++)
      ignorePlanes[jj] = false;

    // if the entire line was clipped then continue looping until
    // either the entire line isn't clipped or the entire viewport is clipped
    while ( (index = this->ClipLine (point1, point2, ignorePlanes)) < numClipPlanes)
      {
      // find where the plane intersects the bottom

      plane = (vtkPlane *)(this->ClippingPlanes)->GetItemAsObject(index);

      // since the following code creates a line which is the plane this plan
      // can be ignored the next time the line is clipping is checked. thi s
      // is necessary because round off error may cause the new line to get
      // clipped again by the same plane
      ignorePlanes[index] = true;

      tmpPoint1 = (currPoint1 == cornerPoints.begin())  ? cornerPoints.end () : currPoint1;
      tmpPoint2 = (currPoint2 == cornerPoints.begin())  ? cornerPoints.end () : currPoint2;
      tmpPoint1--;
      tmpPoint2--;
      intersect = vtkPlane::IntersectWithLine (*tmpPoint1,
                                               *tmpPoint2,
                                               plane->GetNormal(),
                                               plane->GetOrigin(), t, point1);

      // try intersecting with the other edge
      if (intersect == 0)
        {
        // this corner is completely clipped. remove it
        delete [](*tmpPoint2);
        tmpPoint2 = cornerPoints.erase (tmpPoint2);
        if (tmpPoint2 == cornerPoints.end ())
           tmpPoint2 = cornerPoints.begin();

        tmpPoint1 = (tmpPoint1 == cornerPoints.begin())  ? cornerPoints.end () : tmpPoint1;
        tmpPoint2 = (tmpPoint2 == cornerPoints.begin())  ? cornerPoints.end () : tmpPoint2;
        tmpPoint1--;
        tmpPoint2--;

        intersect = vtkPlane::IntersectWithLine (*tmpPoint1,
                                               *tmpPoint2,
                                               plane->GetNormal(),
                                               plane->GetOrigin(), t, point1);

        // reset current point1 because it would have been deleted
        currPoint1 = tmpPoint2;
        numEdges--;
        }

      // the plane clips the entire image
      if (intersect == 0)
        break;

      // find where the plane intersects the top
//      plane = (vtkPlane *)(this->ClippingPlanes)->GetItemAsObject(index);
      tmpPoint1 = currPoint1;
      tmpPoint2 = currPoint2;
      tmpPoint1++;
      tmpPoint2++;

      if (tmpPoint1 == cornerPoints.end())
        tmpPoint1 = cornerPoints.begin ();
      if (tmpPoint2 == cornerPoints.end())
        tmpPoint2 = cornerPoints.begin ();

      intersect = vtkPlane::IntersectWithLine (*tmpPoint1,
                                               *tmpPoint2,
                                               plane->GetNormal(),
                                               plane->GetOrigin(), t, point2);

      // try intersecting with the other edge
      if (intersect == 0)
        {
        // this corner is completely clipped. remove it
        delete [](*tmpPoint1);
        tmpPoint1 = cornerPoints.erase (tmpPoint1);
        if (tmpPoint1 == cornerPoints.end ())
           tmpPoint1 = cornerPoints.begin();

        //tmpPoint1++;
        tmpPoint2++;

        if (tmpPoint1 == cornerPoints.end())
          tmpPoint1 = cornerPoints.begin ();
        if (tmpPoint2 == cornerPoints.end())
          tmpPoint2 = cornerPoints.begin ();

        intersect = vtkPlane::IntersectWithLine (*tmpPoint1,
                                               *tmpPoint2,
                                               plane->GetNormal(),
                                               plane->GetOrigin(), t, point2);

        // reset current point2 because it would have been deleted
        currPoint2 = tmpPoint1;
        numEdges--;
        }

      // the plane clips the entire image
      if (intersect == 0)
        break;

      for (int jj=0; jj<3; jj++)
        {
        (*currPoint1)[jj] = point1[jj];
        (*currPoint2)[jj] = point2[jj];
        }

      }

    if (intersect == 0)
      {
      // remove any points that may have been added to the clipPoint list
      for (tmpPoint1 = clipPoints.begin (); tmpPoint1 != clipPoints.end (); tmpPoint1++)
        delete [](*tmpPoint1);
      clipPoints.clear ();

      break;
      }
    else
      {

      // store these points in a list
      clipPoints.push_back (point1);
      clipPoints.push_back (point2);

      // select the next set of corner points
      currPoint1 = currPoint2;
      currPoint2++;
      if (currPoint2 == cornerPoints.end ())
        currPoint2 = cornerPoints.begin ();

      // create memory for new points
      point1 = new double[3];
      point2 = new double[3];
      }

    }
  GLfloat *path = static_cast<GLfloat *>(0);
  int jj=0;
  double normPointX = 0;
  double normPointY = 0;

  if (clipPoints.size() > 0)
     path = new GLfloat[clipPoints.size()*2];

  for (tmpPoint1 = clipPoints.begin (); tmpPoint1 != clipPoints.end (); tmpPoint1++)
    {
    // check for repeating points
    if ((jj < 1) || ((path[jj-2] != static_cast<GLfloat>((*tmpPoint1)[0])) ||
                     (path[jj-1] != static_cast<GLfloat>((*tmpPoint1)[1]))))
      {

      normPointX = (static_cast<GLfloat>((*tmpPoint1)[0]) - lowerLeft[0])/static_cast<float>(upperRight[0] - lowerLeft[0]);
      normPointY = (static_cast<GLfloat>((*tmpPoint1)[1]) - lowerLeft[1])/static_cast<float>(upperRight[1] - lowerLeft[1]);
      path[jj++] = normPointX * (viewportPositions[2] - viewportPositions[0]) + viewportPositions[0];
      path[jj++] = normPointY * (viewportPositions[3] - viewportPositions[1]) + viewportPositions[1];
      }

    }

  gl2psReplaceClipPlane (numClipPlanes, path, jj);
  delete []path;
  delete []ignorePlanes;

  for (tmpPoint1 = clipPoints.begin (); tmpPoint1 != clipPoints.end (); tmpPoint1++)
    delete [](*tmpPoint1);
  clipPoints.clear ();

  for (tmpPoint1 = cornerPoints.begin (); tmpPoint1 != cornerPoints.end (); tmpPoint1++)
    delete [](*tmpPoint1);
  cornerPoints.clear ();

  if (point1) delete []point1;
  if (point2) delete []point2;
}

// based off of bline clipping
int vtkOpenGLPolyDataMapper2D::ClipLine  (double *point1, double *point2, bool *ignorePlanes)
{

  vtkPlaneCollection *clipPlanes;
  vtkPlane           *plane;
  int                 numClipPlanes;
  char                p1OutBits = 0;
  char                p2OutBits = 0;
  double             *p1BoundCoords;
  double             *p2BoundCoords;
  double              tmpPoint[3];
  double              tmpPoint1[3];
  double              tmpPoint2[3];
  double              tIn = 0;
  double              tOut = 1;
  double              t;


  clipPlanes = this->ClippingPlanes;

  if (clipPlanes == NULL)
    {
       return 0;
    }
  else
    {
    numClipPlanes = clipPlanes->GetNumberOfItems();
    }

  p1BoundCoords = new double[numClipPlanes];
  p2BoundCoords = new double[numClipPlanes];

  for (int ii = 0; ii < numClipPlanes; ii++)
    {
    if (ignorePlanes && !ignorePlanes[ii])
      {
      plane = (vtkPlane *)clipPlanes->GetItemAsObject(ii);

      // calculate bline's boundary coordinates
      p1BoundCoords[ii] = plane->FunctionValue(point1);
      p2BoundCoords[ii] = plane->FunctionValue(point2);

      // determine which coordinates lay outside the plane
      if (p1BoundCoords[ii] < 0)
        p1OutBits |= (1 << ii);
      if (p2BoundCoords[ii] < 0)
        p2OutBits |= (1 << ii);
      }
    else
      {
         p1BoundCoords[ii] = 0;
         p2BoundCoords[ii] = 0;
      }
    }

  // check to see if the entire line is outside the clipping plane
  if ((p1OutBits & p2OutBits) != 0)
    {
    char combineOut = p1OutBits & p2OutBits;

    // find one plane which clips the entire line and return it
    for (int ii = 0; ii < numClipPlanes; ii++)
      {
      if (((combineOut & (1 << ii)) >> ii) == 1)
         {
         delete []p1BoundCoords; 
         delete []p2BoundCoords; 
         return ii;
         }
      }
    }
  // return if none of the line is clipped
  if ((p1OutBits | p2OutBits) == 0)
    {
    delete []p1BoundCoords; 
    delete []p2BoundCoords; 
    return numClipPlanes;
    }

  // initialize tmppoint1 and tmppoint2
  for (int ii = 0; ii < 3; ii++)
    {
    tmpPoint1[ii] = point1[ii];
    tmpPoint2[ii] = point2[ii];
    }

  for (int ii = 0; ii < numClipPlanes; ii++)
    {
    plane = (vtkPlane *)clipPlanes->GetItemAsObject(ii);

    // if the plane clips the line record where
    if (p2BoundCoords[ii] < 0)
      {
      vtkPlane::IntersectWithLine (point1, point2,
                                       plane->GetNormal(),
                                       plane->GetOrigin(), t, tmpPoint);

      if (t < tOut)
        {
        tOut = t;
        for (int jj=0; jj<3; jj++)
          tmpPoint2[jj] = tmpPoint[jj];
        }
      }

    if (p1BoundCoords[ii] < 0)
      {
      vtkPlane::IntersectWithLine (point1, point2,
                                       plane->GetNormal(),
                                       plane->GetOrigin(), t, tmpPoint);

      if (t > tIn)
        {
        tIn = t;
        for (int jj=0; jj<3; jj++)
          tmpPoint1[jj] = tmpPoint[jj];
        }
      }

    // if the points resolve to a single point treat this as if
    // the entire line was clipped
    bool singlePoint = true;
    for (int jj=0; jj<3; jj++)
      if (tmpPoint1[jj] != tmpPoint2[jj])
        singlePoint = false;

    if (singlePoint)
      {
      delete []p1BoundCoords;
      delete []p2BoundCoords;
      return ii;
      }
    }

    for (int ii = 0; ii < 3; ii++)
      {
      point1[ii] = tmpPoint1[ii];
      point2[ii] = tmpPoint2[ii];
      }

  delete []p1BoundCoords;
  delete []p2BoundCoords;

  return numClipPlanes;
}
#endif
#endif // CXC_USE_CHIPS

  
//----------------------------------------------------------------------------
void vtkOpenGLPolyDataMapper2D::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
}
