// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkAxisActor2DExtra.h,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.
 
     This file has been derived from version of vtkXYPlotActor.hh 
     Distribution and modifications to this file are governed by the above
     copyright.
 =========================================================================*/
// .NAME vtkAxisActor2DExtra - Create an axis with tick marks and labels
// .SECTION Description
// vtkAxisActor2DExtra creates an axis with tick marks, labels, and/or a title,
// depending on the particular instance variable settings. vtkAxisActor2DExtra is
// a 2D actor; that is, it is drawn on the overlay plane and is not
// occluded by 3D geometry. To use this class, you typically specify two
// points defining the start and end points of the line (x-y definition using
// vtkCoordinate class), the number of labels, and the data range
// (min,max). You can also control what parts of the axis are visible
// including the line, the tick marks, the labels, and the title.  You can 
// also specify the label format (a printf style format).
//
// This class decides what font size to use and how to locate the labels. It
// also decides how to create reasonable tick marks and labels. The number
// of labels and the range of values may not match the number specified, but
// should be close.
//
// Labels are drawn on the "right" side of the axis. The "right" side is
// the side of the axis on the right as you move from Position to Position2. 
// The way the labels and title line up with the axis and tick marks depends on
// whether the line is considered horizontal or vertical.
//
// The vtkActor2D instance variables Position and Position2 are instances of 
// vtkCoordinate. Note that the Position2 is an absolute position in that 
// class (it was by default relative to Position in vtkActor2D).
//
// What this means is that you can specify the axis in a variety of coordinate
// systems. Also, the axis does not have to be either horizontal or vertical.
// The tick marks are created so that they are perpendicular to the axis.
//
// Set the text property/attributes of the title and the labels through the 
// vtkTextProperty objects associated to this actor.
//
// .SECTION See Also
// vtkAxisActor3D can be used to create axes in world coordinate space.
// vtkActor2D vtkTextMapper vtkPolyDataMapper2D vtkScalarBarActor
// vtkCoordinate vtkTextProperty

#ifndef __vtkAxisActor2DExtra_h
#define __vtkAxisActor2DExtra_h

#include "vtkActor2D.h"
#include "vtkPlane.h"

#include <vector>
#include <list>
#include <string>

using namespace std;

class vtkPolyDataMapper2D;
class vtkPolyData;
class vtkTextMapper;
class vtkTextProperty;
class vtkTextActorExtra;
class vtkTextActorFormat;

class vtkAxisActor2DExtra : public vtkActor2D
{
public:
  vtkTypeRevisionMacro(vtkAxisActor2DExtra,vtkActor2D);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Instantiate object.
  static vtkAxisActor2DExtra *New();

  // Description:
  // Specify the position of the first point defining the axis.
  // Note: backward compatibility only, use vtkActor2D's Position instead.
  virtual vtkCoordinate *GetPoint1Coordinate() 
    { return this->GetPositionCoordinate(); };
  virtual void SetPoint1(double x[2]) { this->SetPosition(x); };
  virtual void SetPoint1(double x, double y) { this->SetPosition(x,y); };
  virtual double *GetPoint1() { return this->GetPosition(); };

  // Description:
  // Specify the position of the second point defining the axis. Note that
  // the order from Point1 to Point2 controls which side the tick marks
  // are drawn on (ticks are drawn on the right, if visible).
  // Note: backward compatibility only, use vtkActor2D's Position2 instead.
  virtual vtkCoordinate *GetPoint2Coordinate() 
    { return this->GetPosition2Coordinate(); };
  virtual void SetPoint2(double x[2]) { this->SetPosition2(x); };
  virtual void SetPoint2(double x, double y) { this->SetPosition2(x,y); };
  virtual double *GetPoint2() { return this->GetPosition2(); };

  // Description:
  // Specify the (min,max) axis range. This will be used in the generation
  // of labels, if labels are visible.
  int SetRange(double* inRange, int index);
  int SetRange(double inMinX,double inMaxX, double inMinY, double inMaxY, int index);
  vtkGetVectorMacro(Range,double,2);
  vtkGetMacro (RangeIndex, int);

  vtkGetVectorMacro(AdjustedRange,double,2);
  void GetAdjustedRange(double& outMin, double& outMax){ outMin=this->AdjustedRange[0];outMax=this->AdjustedRange[1];};

  vtkSetMacro(ShowLimitsTicks, int);
  vtkGetMacro(ShowLimitsTicks, int);


  vtkSetMacro(MajTicksMode, int);
  vtkGetMacro(MajTicksMode, int);


  vtkSetMacro(MinTicksMode, int);
  vtkGetMacro(MinTicksMode, int);



  int  SetMajTicksLocations(const vector<double>& inMajLocations,const vector<bool>& inLabelLocations, const vector<vtkTextActorFormat>& inLabelFormats);
  int  SetMinTicksLocations(const vector<double>& inMinLocations);

  void SetClipLocations (double inMinX, double inMinY,double inMaxX,double inMaxY);



  // Description:
  // Set/Get the number of annotation labels to show.
  vtkSetMacro(NumberOfMajTicks, int);
  vtkGetMacro(NumberOfMajTicks, int);


  //Tick Displaying mode setup functions


  vtkSetMacro(NumberOfMinTicks, int);
  vtkGetMacro(NumberOfMinTicks, int);
  
  vtkSetMacro(MajTicksInterval, double);
  vtkGetMacro(MajTicksInterval, double);

  
  vtkSetMacro(MinTicksInterval, double);
  vtkGetMacro(MinTicksInterval, double);
  



  //  vtkSetMacro(LabelFontSize, int);
  //  vtkGetMacro(LabelFontSize, int);
  
  
  
  // Description:
  // Set/Get the format with which to print the labels on the scalar
  // bar.
  vtkSetStringMacro(LabelFormat);
  vtkGetStringMacro(LabelFormat);

  vtkTextActorExtra* GetTitleActor(){return this->TitleActor;};

  // Description:
  // Get maj ticks actor;
  vtkActor2D* GetMajTicks(){ return this->MajTicksActor;};
  
  // Description:
  // Get min ticks actor;
  vtkActor2D* GetMinTicks(){ return this->MinTicksActor;};



  // Description:
  // Get maj grid actor;
  vtkActor2D* GetMajGrid(){ return this->MajGridActor;};

  
  // Description:
  // Get min grid actor;
  vtkActor2D* GetMinGrid(){ return this->MinGridActor;};




  // Description:
  // Set/Get the labels text property.
  virtual void SetLabelTextProperty(vtkTextProperty *p);
  vtkGetObjectMacro(LabelTextProperty,vtkTextProperty);
      
  // Description:
  // Set/Get the length of the maj tick marks (expressed in pixels or display
  // coordinates). 
  vtkSetClampMacro(MajTicksLength, int, 0, 100);
  vtkGetMacro(MajTicksLength, int);
  

 // Description:
  // Set/Get the length of the min tick marks (expressed in pixels or display
  // coordinates). 
  vtkSetClampMacro(MinTicksLength, int, 0, 100);
  vtkGetMacro(MinTicksLength, int);
  

  // Description:
  // Set/Get the min value when converting to log

  void SetMinLogLimit( double inMinLogLimit)
  { if ( inMinLogLimit> 0)
    { this->MinLogLimit=inMinLogLimit;this->AdjustRange();this->Modified();}
  ;};
  double  GetMinLogLimit(){return this->MinLogLimit;};
  


  // Description:
  // Set/Get the offset of the ticks (expressed in pixels or display
  // coordinates). The offset is the distance of ticks from axis
  // or other objects.
  vtkSetMacro(MajTickOffset, int);
  vtkGetMacro(MajTickOffset, int);
  

  vtkSetMacro(MinTickOffset, int);
  vtkGetMacro(MinTickOffset, int);


  // Description:
  // Set/Get the offset of the labels (expressed in pixels or display
  // coordinates). The offset is the distance of labels from tick marks
  // or other objects.
  vtkSetMacro(LabelOffset, int);
  vtkGetMacro(LabelOffset, int);
  

  // Description:
  // Set/Get the offset of the title The offset is 
  // the distance of title from center of
  //  axis in percentage of axis' length(valid values 0.0 to 1.0)
  vtkSetMacro(TitleParallelOffset, double);
  vtkGetMacro(TitleParallelOffset, double);


  // Description:
  // Set/Get the offset of the title (expressed in pixels or display
  // coordinates). The offset is the distance of title from axis
  vtkSetMacro(TitlePerpendicularOffset, float);
  vtkGetMacro(TitlePerpendicularOffset, float);
  


  // Description:
  // Set/Get visibility of the axis line.
  vtkSetMacro(AxisVisibility, int);
  vtkGetMacro(AxisVisibility, int);
  vtkBooleanMacro(AxisVisibility, int);

  // Description:
  // Set/Get visibility of the axis tick marks.
  vtkSetMacro(MajTicksVisibility, int);
  vtkGetMacro(MajTicksVisibility, int);
  vtkBooleanMacro(MajTicksVisibility, int);

  vtkSetMacro(MinTicksVisibility, int);
  vtkGetMacro(MinTicksVisibility, int);
  vtkBooleanMacro(MinTicksVisibility, int);



  // Description:
  // Set/Get visibility of the axis labels.
  vtkSetMacro(LabelVisibility, int);
  vtkGetMacro(LabelVisibility, int);
  vtkBooleanMacro(LabelVisibility, int);

  // Description:
  // Set/Get visibility of the axis title.
  vtkSetMacro(TitleVisibility, int);
  vtkGetMacro(TitleVisibility, int);
  vtkBooleanMacro(TitleVisibility, int);



  // Description:
  // Set/Get visibility of maj grids;
  vtkSetMacro(MajGridVisibility, int);
  vtkGetMacro(MajGridVisibility, int);
  vtkBooleanMacro(MajGridVisibility, int);


  // Description:
  // Set/Get visibility of min grids;
  vtkSetMacro(MinGridVisibility, int);
  vtkGetMacro(MinGridVisibility, int);
  vtkBooleanMacro(MinGridVisibility, int);



  // Description:
  // Set/Get scale of axis(linear,log,...)
  void SetScale(int inScaleX, int inScaleY);
  void GetScale (int &inScaleX, int &inScaleY);


  // Description:
  // Set/Get Flipping of labels around axis
  vtkSetMacro(TitleFlip, int);
  vtkGetMacro(TitleFlip, int);
  vtkBooleanMacro(TitleFlip, int);


  // Description:
  // Set/Get Flipping of major ticks
  vtkSetMacro(MajorTicksFlip, int);
  vtkGetMacro(MajorTicksFlip, int);
  vtkBooleanMacro(MajorTicksFlip, int);


  // Description:
  // Set/Get Flipping of minor ticks
  vtkSetMacro(MinorTicksFlip, int);
  vtkGetMacro(MinorTicksFlip, int);
  vtkBooleanMacro(MinorTicksFlip, int);


  // Description:
  // Set/Get Flipping of labels
  vtkSetMacro(LabelsFlip, int);
  vtkGetMacro(LabelsFlip, int);
  vtkBooleanMacro(LabelsFlip, int);


  vtkSetMacro(FrontFillerLength, double);
  vtkGetMacro(FrontFillerLength, double);

  vtkSetMacro(BackFillerLength, double);
  vtkGetMacro(BackFillerLength, double);
  

  // Description:
  // Set/Get filler(miter style) - 0 none, 1 square,
  //                               2 down triangle
  //                               3 up triangle
  //                               4 mid triangle

  vtkSetClampMacro(FrontFillerStyle, int, 0, 4);
  vtkGetMacro(FrontFillerStyle, int);

  vtkSetClampMacro(BackFillerStyle, int, 0, 4);
  vtkGetMacro(BackFillerStyle, int);

  // Description:  Filler alignment - 0 outside
  //                                  1 center
  //                                  2 inside

  vtkSetClampMacro(FrontFillerAlignment, int, 0, 2);
  vtkGetMacro(FrontFillerAlignment, int)

    vtkSetClampMacro(BackFillerAlignment, int, 0, 2);
  vtkGetMacro(BackFillerAlignment, int)


  // Axis alignment 0 bottom, 1 center, 2 top  
  vtkSetClampMacro(Alignment, int, 0, 2);
  vtkGetMacro(Alignment, int)



  void SetMajTickLabelVertAlignmentOffset(double inAlignment);
  void SetMajTickLabelHorizAlignmentOffset(double inAlignment);
  double GetMajTickLabelVertAlignmentOffset(){  return this->MajTickLabelVertAlignmentOffset;};
  double GetMajTickLabelHorizAlignmentOffset(){  return this->MajTickLabelHorizAlignmentOffset;};


  void SetGridPositions (double inMinX, double inMinY, double inMaxX, double inMaxY);
  void GetGridPositions (double &outMinX, double &outMinY, double &outMaxX, double &outMaxY);

  void SetMajGridValues (double ***inGrid, int inSlices, int inLines, bool copy=true);
  double*** GetMajGridValues (int &outSlices, int &outLines);

  void SetMinGridValues (double ***inGrid, int inSlices, int inLines, bool copy=true);
  double*** GetMinGridValues (int &outSlices, int &outLines);


  // Clip data
  void SetClipData(int flag);
  int GetClipData() {return ClipData;};


  // Description:
  // Draw the axis. 
  int RenderOverlay(vtkViewport* viewport);
  int RenderOpaqueGeometry(vtkViewport* viewport);
  virtual int RenderTranslucentPolygonalGeometry(vtkViewport *) {return 0;}

  // Description:
  // Does this prop have some translucent polygonal geometry?
  virtual int HasTranslucentPolygonalGeometry();

  // Description:
  // Release any graphics resources that are being consumed by this actor.
  // The parameter window could be used to determine which graphic
  // resources to release.
  void ReleaseGraphicsResources(vtkWindow *);

  // Description:
  // This method computes the range of the axis given an input range. 
  // It also computes the number of tick marks given a suggested number.
  // (The number of tick marks includes end ticks as well.)
  // The number of tick marks computed (in conjunction with the output
  // range) will yield "nice" tick values. For example, if the input range
  // is (0.25,96.7) and the number of ticks requested is 10, the output range
  // will be (0,100) with the number of computed ticks to 11 to yield tick
  // values of (0,10,20,...,100).
  static void ComputeRange(double inRange[2], 
                           double outRange[2],
                           int inNumTicks, 
                           int &outNumTicks, 
                           double &interval,
			   int inScale,
			   double inMinLogLimit);

  // Description:
  // General method to computes font size from a representative size on the 
  // viewport (given by size[2]). The method returns the font size (in points)
  // and the string height/width (in pixels). It also sets the font size of the
  // instance of vtkTextMapper provided. The factor is used when you're trying
  // to create text of different size-factor (it is usually = 1 but you can
  // adjust the font size by making factor larger or smaller).
  static int SetFontSize(vtkViewport *viewport, 
                         vtkTextMapper *textMapper, 
                         int *targetSize, 
                         double factor, 
                         int *stringSize);
  static int SetMultipleFontSize(vtkViewport *viewport, 
                                 vtkTextMapper **textMappers, 
                                 int nbOfMappers, 
                                 int *targetSize,
                                 double factor, 
                                 int *stringSize);

  // Description:
  // Shallow copy of an axis actor. Overloads the virtual vtkProp method.
  void ShallowCopy(vtkProp *prop);



protected:
  vtkAxisActor2DExtra();
  ~vtkAxisActor2DExtra();



  vtkTextProperty *LabelTextProperty;



  double Range[4];
  double AdjustedRange[4];
  int    RangeIndex;


  int   ShowLimitsTicks;
  int   MajTicksMode;
  int   MinTicksMode;
  int   NumberOfMajTicks;
  int   NumberOfMinTicks;
  double MajTicksInterval;
  double MinTicksInterval;


  char  *LabelFormat;
  int   MajTicksLength;
  int   MinTicksLength;
  int   MajTickOffset;
  int   MinTickOffset;
  int   LabelOffset;
  double   TitleParallelOffset;
  float   TitlePerpendicularOffset;

  int   Scale[2];

  double MinLogLimit;

  int   TitleFlip;
  int   LabelsFlip;
  int   MajorTicksFlip;
  int   MinorTicksFlip;


  int   AxisVisibility;
  int   MajTicksVisibility;
  int   MinTicksVisibility;
  int   LabelVisibility;
  int   TitleVisibility;
  int   MajGridVisibility;
  int   MinGridVisibility;

  int   Alignment;


  double FrontFillerLength;
  int   FrontFillerAlignment;

  double BackFillerLength;
  int   BackFillerAlignment;

  int   FrontFillerStyle;
  int   BackFillerStyle;

  double  LastPosition[2];
  double   LastPosition2[2];
  
  virtual void BuildLineAxis(vtkViewport *viewport);
  virtual void BuildPolygonAxis(vtkViewport *viewport);

  static void SetOffsetPosition(double xTick[3], double theta, 
                                double inXOffset, double inYOffset,
				int inAnnotationsFlip,
				vtkActor2D *actor);

  static int MapScale(double* inDataRange, 
		      double inPixelRangeMin, 
		      double inPixelRangeMax,
		      int inScale,
		      double inMinLogLimit,
	              double inValue,bool clip,double& outValue);

  
  void CalculateMajTicksLocationsByLocations(double inMinRange, double inMaxRange,const vector<double>& inTickLocations,const vector<bool>& inLabelLocations,const vector<vtkTextActorFormat>& inLabelFormats,vector<double>& outTickLocations,vector<bool>& outLabelLocations,vector<vtkTextActorFormat>& outLabelFormats);
  
  void CalculateTicksLocationsByLocations(double inMinRange, double inMaxRange,const vector<double>& inTickLocations,vector<double>& outTickLocations);

  
  void CalculateMajTicksLocationsByInterval(double inMinRange, double inMaxRange,
					    double inInterval,vector<double>& outValues,
					    double& outMajReferenceTick);

  void CalculateMajTicksLocationsByCount(double inMinRange, double inMaxRange,
					 int inCount,vector<double>& outValues);

  void CalculateMinTicksLocationsByInterval(double inMinRange, double inMaxRange,
					    double inInterval,double inMajReferenceTick,
					    const vector<double>& inMajValues,
					    vector<double>& outValues);

  void CalculateMinTicksLocationsByInterval(double inMinRange, double inMaxRange,
					     double inInterval,
					     const vector<double>& inMajValues,
					     vector<double>& outValues);

  void CalculateMinTicksLocationsByCount(double inMinRange, double inMaxRange,
					 double inCount,double inMajReferenceTick,
					    const vector<double>& inMajValues,
					    vector<double>& outValues);


  void CalculateMinTicksLocationsByCount(double inMinRange, double inMaxRange,
					 int inCount,
					 const vector<double>& inMajValues,
					 vector<double>& outValues);

  
  void BuildLabels(list<vtkTextActorExtra*>& inLabelActors, unsigned int inCount);
  
  void BuildLabels(list<vtkTextActorExtra*>& inLabelActors, const vector<bool>& inLabelFlags);
  
  void AdjustRange();

  bool IsIntFormat(char * inFormatString);


  vtkTextActorExtra*   TitleActor;

  list<vtkTextActorExtra*> LabelActors;

  vtkPolyData         *Axis;
  vtkPolyDataMapper2D *AxisMapper;
  vtkActor2D          *AxisActor;


  vtkPolyData         *MajTicks;
  vtkPolyDataMapper2D *MajTicksMapper;
  vtkActor2D          *MajTicksActor;

  vtkPolyData         *MinTicks;
  vtkPolyDataMapper2D *MinTicksMapper;
  vtkActor2D          *MinTicksActor;

  vector<double>      MajTicksLocations;
  vector<double>      MinTicksLocations;
  vector<bool>        MajTicksLabelLocations;
  vector<vtkTextActorFormat>      MajTicksLabelFormats;



  vtkPolyData         *MajGrid;
  vtkPolyDataMapper2D *MajGridMapper;
  vtkActor2D          *MajGridActor;

  vtkPolyData         *MinGrid;
  vtkPolyDataMapper2D *MinGridMapper;
  vtkActor2D          *MinGridActor;



  double MajTickLabelVertAlignmentOffset;
  double MajTickLabelHorizAlignmentOffset;

  vtkCoordinate           *GridStart;
  vtkCoordinate           *GridStop;

  vtkPlane                *ClipPlanes[4];
  vtkCoordinate           *ClipLocation[2];

  bool  ClipData;

  vtkTimeStamp  BuildTime;

  double                  ***MajGridValues;
  int                     NumOfMajGridLines;
  int                     NumOfMajGridSlices;

  double                  ***MinGridValues;
  int                     NumOfMinGridLines;
  int                     NumOfMinGridSlices;


private:
  vtkAxisActor2DExtra(const vtkAxisActor2DExtra&);  // Not implemented.
  void operator=(const vtkAxisActor2DExtra&);  // Not implemented.
};


#endif
