// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkChipsTransform.cxx,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkChipsTransform.hh"
#include "vtkObjectFactory.h"
#include "vtkMath.h"
#include <math.h>
#include <stdlib.h>

#include "ChipsTransformFactory.hh"

vtkCxxRevisionMacro(vtkChipsTransform, "$Revision: 1.4 $");
vtkStandardNewMacro(vtkChipsTransform);

//----------------------------------------------------------------------------
vtkChipsTransform::vtkChipsTransform()
{
   chipsTransform = static_cast <Transform *> (0);
}

vtkChipsTransform::~vtkChipsTransform()
{
   if (chipsTransform)
      delete chipsTransform;
}

void vtkChipsTransform::PrintSelf(ostream& os, vtkIndent indent)
{
  vtkWarpTransform::PrintSelf(os,indent);
}

void vtkChipsTransform::InternalDeepCopy(vtkAbstractTransform *transform)
{
  Transform *tmpChipsTrans;
  vtkChipsTransform *tmpTransform = 
    dynamic_cast <vtkChipsTransform *> (transform);

  // copy these even though they aren't used
  this->SetInverseTolerance(tmpTransform->InverseTolerance);
  this->SetInverseIterations(tmpTransform->InverseIterations);

  // copy the inverse flag, which is used
  if (this->InverseFlag != tmpTransform->InverseFlag)
    {
    this->InverseFlag = tmpTransform->InverseFlag;
    this->Modified();
    }

  // copy the chips transform
  if (tmpChipsTrans = tmpTransform->GetChipsTransform ())
      this->SetChipsTransform (tmpChipsTrans);
  
}

vtkAbstractTransform *vtkChipsTransform::MakeTransform()
{
  return vtkChipsTransform::New();
}

template<class T>
void vtkChipsTransform::vtkApplyTransform(const T inPoint[3], T outPoint[3],
                                        T derivative[3][3])
{
  double *chipsMatrix = static_cast <double *>(0);
  outPoint[0] = inPoint[0];
  outPoint[1] = inPoint[1];
/* Currently hard code this to be 2D
  outPoint[2] = inPoint[2];
*/
  outPoint[2] = 1.0;

  if (chipsTransform)
    {
      T *tmp_out = static_cast <T*> (chipsTransform->apply (outPoint, 
                                this->GetTransformDataType (outPoint), 1));
      if (tmp_out)
      {
         outPoint[0] = tmp_out[0];
         outPoint[1] = tmp_out[1];
         delete [] tmp_out;

      }
      /* something which was discussed for the future. Will need to tackle
         for vectors 
      chipsMatrix = chipsTransform->get_matrix ();
      */
    }

  if (derivative && chipsMatrix)
    {
       derivative[0][0] = (T)chipsMatrix[0];
       derivative[0][1] = (T)chipsMatrix[1];
       derivative[0][2] = (T)chipsMatrix[2];

       derivative[1][0] = (T)chipsMatrix[3];
       derivative[1][1] = (T)chipsMatrix[4];
       derivative[1][2] = (T)chipsMatrix[5];

       derivative[2][0] = (T)chipsMatrix[6];
       derivative[2][1] = (T)chipsMatrix[7];
       derivative[2][2] = (T)chipsMatrix[8];
    }
  else if (derivative)
    {
       vtkMath::Identity3x3 (derivative);
    }
}

template<class T>
void vtkChipsTransform::vtkApplyInverse(const T inPoint[3], T outPoint[3],
                                        T derivative[3][3])
{
  double *chipsMatrix = static_cast <double *>(0);
  outPoint[0] = inPoint[0];
  outPoint[1] = inPoint[1];

/* Currently hard code this to be 2D 
  outPoint[2] = inPoint[2];
*/
  outPoint[2] = 1.0;

  if (chipsTransform)
    {

      T *tmp_out = static_cast <T*> (chipsTransform->invert (outPoint,
                                this->GetTransformDataType (outPoint), 1));
      if (tmp_out)
      {
         outPoint[0] = tmp_out[0];
         outPoint[1] = tmp_out[1];
         delete []tmp_out;
      }

      /* something which was discussed for the future. Will need to tackle
         for vectors
      chipsMatrix = chipsTransform->get_matrix ();
      */
    }

  if (derivative && chipsMatrix)
    {
       derivative[0][0] = (T)chipsMatrix[0];
       derivative[0][1] = (T)chipsMatrix[1];
       derivative[0][2] = (T)chipsMatrix[2];

       derivative[1][0] = (T)chipsMatrix[3];
       derivative[1][1] = (T)chipsMatrix[4];
       derivative[1][2] = (T)chipsMatrix[5];

       derivative[2][0] = (T)chipsMatrix[6];
       derivative[2][1] = (T)chipsMatrix[7];
       derivative[2][2] = (T)chipsMatrix[8];
    }
  else if (derivative)
    {
       vtkMath::Identity3x3 (derivative);
    }

}

void vtkChipsTransform::ForwardTransformPoint(const float inPoint[3],
                                                    float outPoint[3])
{
  vtkApplyTransform<float>(inPoint, outPoint, (float (*)[3])0);
}

void vtkChipsTransform::ForwardTransformPoint(const double inPoint[3],
                                                    double outPoint[3])
{
  vtkApplyTransform<double>(inPoint, outPoint, (double (*)[3])0);
}

void vtkChipsTransform::ForwardTransformDerivative(
                                                       const float inPoint[3],
                                                       float outPoint[3],
                                                       float derivative[3][3])
{
  vtkApplyTransform<float>(inPoint, outPoint, derivative);
}

void vtkChipsTransform::ForwardTransformDerivative(
                                                       const double inPoint[3],
                                                       double outPoint[3],
                                                       double derivative[3][3])
{
  vtkApplyTransform<double>(inPoint, outPoint, derivative);
}

void vtkChipsTransform::InverseTransformPoint(const float inPoint[3],
                                                    float outPoint[3])
{
  vtkApplyInverse<float>(inPoint, outPoint, (float (*)[3])0);
}

void vtkChipsTransform::InverseTransformPoint(const double inPoint[3],
                                                    double outPoint[3])
{
  vtkApplyInverse<double>(inPoint, outPoint, (double (*)[3])0);
}

void vtkChipsTransform::InverseTransformDerivative(
                                                       const float inPoint[3],
                                                       float outPoint[3],
                                                       float derivative[3][3])
{
  vtkApplyInverse<float>(inPoint, outPoint, derivative);
}

void vtkChipsTransform::InverseTransformDerivative(
                                                       const double inPoint[3],
                                                       double outPoint[3],
                                                       double derivative[3][3])
{
  vtkApplyInverse<double>(inPoint, outPoint, derivative);
}

void vtkChipsTransform::SetChipsTransform (Transform *trans)
{
  if (chipsTransform)
     delete chipsTransform;

  chipsTransform = trans->copy ();

}

