// 
//  Copyright (C) 2004-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkScalarBarActorExtra.h,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.
 
     This file has been derived from version of vtkXYPlotActor.hh 
     Distribution and modifications to this file are governed by the above
     copyright.
 =========================================================================*/
// .NAME vtkScalarBarActorExtra - Create an axis with tick marks and labels
// .SECTION Description
// vtkScalarBarActorExtra creates an axis with tick marks, labels, and/or a title,
// depending on the particular instance variable settings. vtkScalarBarActorExtra is
// a 2D actor; that is, it is drawn on the overlay plane and is not
// occluded by 3D geometry. To use this class, you typically specify two
// points defining the start and end points of the line (x-y definition using
// vtkCoordinate class), the number of labels, and the data range
// (min,max). You can also control what parts of the axis are visible
// including the line, the tick marks, the labels, and the title.  You can 
// also specify the label format (a printf style format).
//
// This class decides what font size to use and how to locate the labels. It
// also decides how to create reasonable tick marks and labels. The number
// of labels and the range of values may not match the number specified, but
// should be close.
//
// Labels are drawn on the "right" side of the axis. The "right" side is
// the side of the axis on the right as you move from Position to Position2. 
// The way the labels and title line up with the axis and tick marks depends on
// whether the line is considered horizontal or vertical.
//
// The vtkActor2D instance variables Position and Position2 are instances of 
// vtkCoordinate. Note that the Position2 is an absolute position in that 
// class (it was by default relative to Position in vtkActor2D).
//
// What this means is that you can specify the axis in a variety of coordinate
// systems. Also, the axis does not have to be either horizontal or vertical.
// The tick marks are created so that they are perpendicular to the axis.
//
// Set the text property/attributes of the title and the labels through the 
// vtkTextProperty objects associated to this actor.
//
// .SECTION See Also
// vtkAxisActor3D can be used to create axes in world coordinate space.
// vtkActor2D vtkTextMapper vtkPolyDataMapper2D vtkScalarBarActor
// vtkCoordinate vtkTextProperty

#ifndef __vtkScalarBarActorExtra_h
#define __vtkScalarBarActorExtra_h

#include "vtkActor2D.h"

#include <vtkstd/vector>
#include <vtkstd/list>

using namespace std;

class vtkPolyDataMapper2D;
class vtkPolyData;
class vtkTextMapper;
class vtkTextProperty;
class vtkTextActorExtra;
class vtkTextActorFormat;
class vktPlane;
class vtkScalarsToColors;
class vtkPlane;

#ifndef VTK_ORIENT_HORIZONTAL
  #define VTK_ORIENT_HORIZONTAL 0
#endif

#ifndef VTK_ORIENT_VERTICAL
  #define VTK_ORIENT_VERTICAL 1
#endif


class vtkScalarBarActorExtra : public vtkActor2D
{
public:
  vtkTypeRevisionMacro(vtkScalarBarActorExtra,vtkActor2D);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Instantiate object.
  static vtkScalarBarActorExtra *New();

  // Description:
  // Specify the position of the first point defining the axis.
  // Note: backward compatibility only, use vtkActor2D's Position instead.
  virtual vtkCoordinate *GetPoint1Coordinate() 
    { return this->GetPositionCoordinate(); };
  virtual void SetPoint1(double x[2]) { this->SetPosition(x); };
  virtual void SetPoint1(double x, double y) { this->SetPosition(x,y); };
  virtual double *GetPoint1() { return this->GetPosition(); };

  // Description:
  // Control the orientation of the scalar bar.
  void SetOrientation (int orientation);
  vtkGetMacro(Orientation, int);
  void SetOrientationToHorizontal()
       {this->SetOrientation(VTK_ORIENT_HORIZONTAL);};
  void SetOrientationToVertical() {this->SetOrientation(VTK_ORIENT_VERTICAL);};


  // Description:
  // Specify the position of the second point defining the axis. Note that
  // the order from Point1 to Point2 controls which side the tick marks
  // are drawn on (ticks are drawn on the right, if visible).
  // Note: backward compatibility only, use vtkActor2D's Position2 instead.
  virtual vtkCoordinate *GetPoint2Coordinate() 
    { return this->GetPosition2Coordinate(); };
  virtual void SetPoint2(double x[2]) { this->SetPosition2(x); };
  virtual void SetPoint2(double x, double y) { this->SetPosition2(x,y); };
  virtual double *GetPoint2() { return this->GetPosition2(); };

  int  SetMajorTicksLocations(const vtkstd::vector<double>& inMajorValues);

  void SetClipLocations (double inMinX, double inMinY,double inMaxX,double inMaxY);
  vtkGetVector2Macro (ClipLocation, vtkCoordinate*);
  vtkGetVector2Macro (Location, vtkCoordinate*);


  // Description:
  // Set/Get the format with which to print the labels on the scalar
  // bar.
  vtkSetStringMacro(LabelFormat);
  vtkGetStringMacro(LabelFormat);

  vtkTextActorExtra* GetTitleActor(){return this->TitleActor;};

  // Description:
  // Get maj ticks actor;
  vtkActor2D* GetMajorTicks(){ return this->MajorTicksActor;};
  

  // Description:
  // Get maj ticks actor;
  vtkActor2D* GetBoundingBox(){ return this->BoundingBoxActor;};

  // Description:
  // Set/Get the labels text property.
  virtual void SetLabelTextProperty(vtkTextProperty *p);
  vtkGetObjectMacro(LabelTextProperty,vtkTextProperty);

      
  // Description:
  // Set/Get the length of the maj tick marks (expressed in pixels or display
  // coordinates). 
  vtkSetClampMacro(MajorTicksLength, int, 0, 100);
  vtkGetMacro(MajorTicksLength, int);
  


  // Description:
  // Set/Get the offset of the ticks (expressed in pixels or display
  // coordinates). The offset is the distance of ticks from axis
  // or other objects.
  vtkSetMacro(MajorTickOffset, int);
  vtkGetMacro(MajorTickOffset, int);
  

  // Description:
  // Set/Get the offset of the labels (expressed in pixels or display
  // coordinates). The offset is the distance of labels from tick marks
  // or other objects.
  vtkSetMacro(LabelOffset, int);
  vtkGetMacro(LabelOffset, int);
  

  // Description:
  // Set/Get the offset of the title The offset is 
  // the distance of title from center of
  //  axis in percentage of axis' length(valid values 0.0 to 1.0)
  vtkSetMacro(TitleParallelOffset, double);
  vtkGetMacro(TitleParallelOffset, double);


  // Description:
  // Set/Get the offset of the title (expressed in pixels or display
  // coordinates). The offset is the distance of title from axis
  vtkSetMacro(TitlePerpendicularOffset, float);
  vtkGetMacro(TitlePerpendicularOffset, float);
  


  // Description:
  // Set/Get visibility of the axis line.
  vtkSetMacro(ScalarBarVisibility, int);
  vtkGetMacro(ScalarBarVisibility, int);
  vtkBooleanMacro(ScalarBarVisibility, int);

  // Description:
  // Set/Get visibility of the axis line.
  vtkSetMacro(BoundingBoxVisibility, int);
  vtkGetMacro(BoundingBoxVisibility, int);
  vtkBooleanMacro(BoundingBoxVisibility, int);

  // Description:
  // Set/Get visibility of the axis tick marks.
  vtkSetMacro(MajorTicksVisibility, int);
  vtkGetMacro(MajorTicksVisibility, int);
  vtkBooleanMacro(MajorTicksVisibility, int);


  // Description:
  // Set/Get visibility of the axis labels.
  vtkSetMacro(LabelVisibility, int);
  vtkGetMacro(LabelVisibility, int);
  vtkBooleanMacro(LabelVisibility, int);

  // Description:
  // Set/Get visibility of the axis title.
  vtkSetMacro(TitleVisibility, int);
  vtkGetMacro(TitleVisibility, int);
  vtkBooleanMacro(TitleVisibility, int);


  // Description:
  // Set/Get Flipping of labels around axis
  vtkSetMacro(TitleFlip, int);
  vtkGetMacro(TitleFlip, int);
  vtkBooleanMacro(TitleFlip, int);


  // Description:
  // Set/Get Flipping of major ticks
  vtkSetMacro(MajorTicksFlip, int);
  vtkGetMacro(MajorTicksFlip, int);
  vtkBooleanMacro(MajorTicksFlip, int);


  // Description:
  // Set/Get Flipping of labels
  vtkSetMacro(LabelsFlip, int);
  vtkGetMacro(LabelsFlip, int);
  vtkBooleanMacro(LabelsFlip, int);


  // Description:
  // Set/Get Flipping of ticks and tick labels
  vtkSetMacro(BarFlip, int);
  vtkGetMacro(BarFlip, int);
  vtkBooleanMacro(BarFlip, int);

  // Description:
  // Set/Get weither the scalar bar should use scalar ranges
  // from the lookup table or those set in the color bar
  vtkSetMacro(UseLookupTableRange, int);
  vtkGetMacro(UseLookupTableRange, int);
  vtkBooleanMacro(UseLookupTableRange, int);

  // Description:
  // Set/Get the scalar range the scalar bar should represent.
  // These values are only used if UseLookupTableRange is false
  vtkSetVector2Macro(ScalarRange,double);
  vtkGetVectorMacro(ScalarRange,double,2);


  // Description:
  // Set/Get the vtkLookupTable to use. The lookup table specifies the number
  // of colors to use in the table (if not overridden), as well as the scalar
  // range.
  virtual void SetLookupTable(vtkScalarsToColors*);
  vtkGetObjectMacro(LookupTable,vtkScalarsToColors);


  void SetMajorTickLabelVertAlignmentOffset(double inAlignment);
  void SetMajorTickLabelHorizAlignmentOffset(double inAlignment);
  double GetMajorTickLabelVertAlignmentOffset(){  return this->MajorTickLabelVertAlignmentOffset;};
  double GetMajorTickLabelHorizAlignmentOffset(){  return this->MajorTickLabelHorizAlignmentOffset;};


  void SetOrigin (double origin[]);
  double *GetOrigin () {return &(this->ScalarBarOrigin)[0];};
  void SetLength (float length);
  float GetLength () {return this->ScalarBarLength;};
  void SetWidth (float width);
  float GetWidth () {return this->ScalarBarWidth;};
  void SetVertAlignmentOffset (double offset);
  double GetVertAlignmentOffset () {return this->ScalarBarVertAlignmentOffset;};
  void SetHorizAlignmentOffset (double offset);
  double GetHorizAlignmentOffset () {return this->ScalarBarHorizAlignmentOffset;};
  



  // Clip data
  void SetClipData(int flag);
  int GetClipData() {return ClipData;};


  // Description:
  // Draw the axis. 
  int RenderOverlay(vtkViewport* viewport);
  int RenderOpaqueGeometry(vtkViewport* viewport);
  virtual int RenderTranslucentPolygonalGeometry(vtkViewport *) {return 0;}

  // Description:
  // Does this prop have some translucent polygonal geometry?
  virtual int HasTranslucentPolygonalGeometry();

  // Description:
  // Release any graphics resources that are being consumed by this actor.
  // The parameter window could be used to determine which graphic
  // resources to release.
  void ReleaseGraphicsResources(vtkWindow *);


  // Description:
  // General method to computes font size from a representative size on the 
  // viewport (given by size[2]). The method returns the font size (in points)
  // and the string height/width (in pixels). It also sets the font size of the
  // instance of vtkTextMapper provided. The factor is used when you're trying
  // to create text of different size-factor (it is usually = 1 but you can
  // adjust the font size by making factor larger or smaller).
  static int SetFontSize(vtkViewport *viewport, 
                         vtkTextMapper *textMapper, 
                         int *targetSize, 
                         double factor, 
                         int *stringSize);
  static int SetMultipleFontSize(vtkViewport *viewport, 
                                 vtkTextMapper **textMappers, 
                                 int nbOfMappers, 
                                 int *targetSize,
                                 double factor, 
                                 int *stringSize);

  // Description:
  // Shallow copy of an axis actor. Overloads the virtual vtkProp method.
  void ShallowCopy(vtkProp *prop);

  void UpdatePositions (vtkViewport *viewport);



protected:
  vtkScalarBarActorExtra();
  ~vtkScalarBarActorExtra();



  vtkTextProperty *LabelTextProperty;



  char  *LabelFormat;
  int   MajorTicksLength;
  int   MajorTickOffset;
  int   LabelOffset;
  double   TitleParallelOffset;
  float TitlePerpendicularOffset;

  int   TitleFlip;
  int   LabelsFlip;
  int   MajorTicksFlip;
  int   BarFlip;

  int   ScalarBarVisibility;
  int   BoundingBoxVisibility;
  int   MajorTicksVisibility;
  int   LabelVisibility;
  int   TitleVisibility;

  double  LastPosition[2];
  double   LastPosition2[2];
  
  virtual void BuildScalarBar (vtkViewport *viewport);

  static void SetOffsetPosition(double xTick[3], double theta, 
                                double inXOffset, double inYOffset,
				int inAnnotationsFlip,
				vtkActor2D *actor);

  static int MapScale(double* inDataRange, 
		      double inPixelRangeMin, 
		      double inPixelRangeMax,
		      int inScale,
		      double inMinLogLimit,double inValue,double& outValue);

  
  void BuildLabels(list<vtkTextActorExtra*>& inLabelActors, unsigned int inCount);
  
  bool IsIntFormat(char * inFormatString);
  void GenerateScalarBar (double offset[2], double width, double length);
  bool GetLogScale ();



  vtkTextActorExtra*   TitleActor;

  list<vtkTextActorExtra*> LabelActors;

  vtkPolyData         *ScalarBar;
  vtkPolyDataMapper2D *ScalarBarMapper;
  vtkActor2D          *ScalarBarActor;

  vtkPolyData         *BoundingBox;
  vtkPolyDataMapper2D *BoundingBoxMapper;
  vtkActor2D          *BoundingBoxActor;


  vtkPolyData         *MajorTicks;
  vtkPolyDataMapper2D *MajorTicksMapper;
  vtkActor2D          *MajorTicksActor;

  vtkstd::vector<double>      MajorTicksLocations;
  vtkstd::vector<vtkTextActorFormat>      MajorTicksLabelFormats;

  double MajorTickLabelVertAlignmentOffset;
  double MajorTickLabelHorizAlignmentOffset;

  vtkPlane                *ClipPlanes[4];
  vtkCoordinate           *ClipLocation[2];

  vtkCoordinate           *Location[2];

  bool  ClipData;

  vtkScalarsToColors *LookupTable;
  vtkTimeStamp  BuildTime;

  double ScalarBarOrigin[2];
  float ScalarBarWidth;
  float ScalarBarLength;
  double ScalarBarVertAlignmentOffset;
  double ScalarBarHorizAlignmentOffset;

  int Orientation;

  bool UseLookupTableRange;
  double ScalarRange[2];

private:

  bool updatePositions;

  vtkScalarBarActorExtra(const vtkScalarBarActorExtra&);  // Not implemented.
  void operator=(const vtkScalarBarActorExtra&);  // Not implemented.
};


#endif
