// 
//  Copyright (C) 2004-2012  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sstream>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include "ChipsEngine.hh"
#include "ChipsDefaults.hh"
#include "ChipsFillPattern.hh"
#include "ChipsFontList.hh"
#include "ChipsCmdManager.hh"
#include "ChipsStateStream.hh"
#include "ChipsDeleteCommand.hh"
#include "ChipsPrintDialog.hh"
#include "ChipsColorMapTable.hh"
#include "ChipsTransformFactory.hh"



ChipsDefaults* theDefaults = NULL; 
ChipsFontList* theFonts = NULL; 
ChipsFillPattern* fill_patterns = NULL; 
ChipsCmdManager* cmd_manager = NULL;
ChipsColorMapTable *colormap_table = NULL;

// Engine constructor 
ChipsEngine::ChipsEngine(bool load_prefs)
{
   theDefaults = new ChipsDefaults(load_prefs);
   theFonts = new ChipsFontList(); 
   fill_patterns = new ChipsFillPattern(); 

   colormap_table = new ChipsColorMapTable ();
   window_cnt = 0;
   buffer_depth = 0; 
   cmd_manager = new ChipsCmdManager (this);
   cmd_manager->set_err_verbosity(theDefaults->get_defaultverbosity());
   cmd_manager->set_cmd_lock(false);
   redraw_block = false;
}


// engine destructor 
ChipsEngine::~ChipsEngine()
{
   ChipsId cid; 
   // bug 13031: for now- don't force delete the windows (they get
   // cleaned up from the dwindow list)- this improves deletion speed w/
   // interactors 
   delete_all_windows(&cid);
   map<unsigned int, ChipsWindow *>::iterator ii;
   for (ii = dwindow.begin (); ii != dwindow.end (); ii++)
      ii->second->object_delete ();
   dwindow.clear ();
   window_cnt = 0; 
   if (theDefaults)
   {
      delete theDefaults;
      theDefaults = NULL; 
   }
   if (theFonts)
   {
      delete theFonts;
      theFonts = NULL; 
   }
   if (cmd_manager)
   {
      delete cmd_manager;
      cmd_manager = NULL;
   }
   if (colormap_table)
   {
      delete colormap_table;
      colormap_table = NULL;
   }
}

// load preferences file
int ChipsEngine::load_preferences(string in_file)
{
   int  rval;
   cmd_manager->clear_errs();
   if (theDefaults)
      rval = theDefaults->load_file(in_file); 
   else   
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"preferences object"); 
   cmd_manager->set_cmd_lock(false);
   return rval; 
}

// save preferences file
int ChipsEngine::save_preferences(string in_file)
{
   int  rval;
   cmd_manager->clear_errs();
   if (theDefaults)
      rval = theDefaults->save_file(in_file); 
   else   
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"preferences object"); 
   return rval; 
}


// set preference value
int ChipsEngine::set_preference(string pref, string value)
{
   int  rval, update = 0;
   cmd_manager->clear_errs();
   if (theDefaults)
   {
      cmd_manager->open_block();

      //order issue during undo/redo
      if ((pref == "foreground.display" || pref == "background.display") && 
          window.size())
      {
         list<ChipsWindow*>::iterator ii = window.begin();
         ChipsDefDispColorCommand *cmd = new ChipsDefDispColorCommand((*ii));
         if (cmd_manager->add_cmd(cmd) == ERR_STATUS_BAD)
            delete cmd;
         update = 1; 
      }
      rval = theDefaults->set_preference(pref, value);

      // handle default display.fg/bg changes
      if ((pref == "foreground.display" || pref == "background.display") && 
          window.size())
      {
         update_default_display_colors();
         list<ChipsWindow*>::iterator ii = window.begin();
         ChipsDefDispColorCommand *cmd = new ChipsDefDispColorCommand((*ii));
         if (cmd_manager->add_cmd(cmd) == ERR_STATUS_BAD)
            delete cmd;
      }

      if (update)   update_default_display_colors();
      cmd_manager->close_block();
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"preferences object"); 
   return rval;
}

// set preference value
int ChipsEngine::override_preferences(string prefs, string object)
{
   int  rval;
   cmd_manager->clear_errs();
   if (theDefaults)
      rval = theDefaults->set_overrides(prefs, object);
   else
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"preferences object");
   return rval;
}

int ChipsEngine::end_override_preferences()
{
   theDefaults->reset_defaults();
   return ERR_STATUS_GOOD;
}

// get preference value
char* ChipsEngine::get_preference(string pref)
{
   char* buf = NULL;
   cmd_manager->clear_errs();
   if (theDefaults)
      buf = theDefaults->get_preference(pref);
   return buf;
}

int ChipsEngine::load_state(string fname, unsigned int mask)
{
   int rval = ERR_STATUS_GOOD;
   char backup[80];
   static int my_pid = getpid();
   // The following should be replaced by a common function from ChipsDefine
   // Since the same exact task is done by the chips server

   string filename = "/tmp/.chips";
   struct passwd *pw = getpwuid(getuid());
   if (pw && pw->pw_name[0] != '\0')
   {
      filename.append(".");
      filename.append(pw->pw_name);
   }

   if (access(filename.c_str(), W_OK) != 0)
   {
      mkdir(filename.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
      if (access(filename.c_str(), W_OK) != 0)  filename = "/tmp";
   }
   filename.append("/");

   // This is the end  of what needs to be replaced

   cmd_manager->clear_errs();

   sprintf(backup, "%schips_backup.%1d",filename.c_str (), my_pid);
   if (strstr(backup, "chips_backup") == NULL)  
      sprintf(backup, "%schips_backup",filename.c_str ());
   
   if (fname.length() == 0)
   {
      cmd_manager->add_err(ERR_FILE_NOT_SUPPLIED); 
      return ERR_FILE_NOT_SUPPLIED;
   }

   ChipsStateStream in(fname.c_str(), ios::in|ios::binary);
   if ((rval = in.get_error()) == ERR_STATUS_GOOD)
   {
      cmd_manager->set_cmd_lock(true);
      // Code should save the current state here so that if 
      // the load fails we can go back to our pre-load state 
      // ie.   save_state("/tmp/my_previous"); 
      if (fname != backup)   save_state(backup, true);
      if ((mask&V_CLEAR_ALL) == V_CLEAR_ALL)    delete_all_windows(NULL, true);

      if ((mask&V_LOAD_STATE) == V_LOAD_STATE)
      {
         ChipsWindow* cw=NULL;
         int wlen;

         if (theDefaults)
         {
            // serialize the preferences
            theDefaults->deserialize(in);
            if (in.read_checkpoint() != ERR_STATUS_GOOD)  return ERR_STATEFILE_COR;
         }
         if (in.get_minor_version() > 0 && fill_patterns)
         {
            fill_patterns->deserialize(in);
	    if (in.read_checkpoint() != ERR_STATUS_GOOD)  return ERR_STATEFILE_COR;
         }

         if (in.get_minor_version () > 0 && colormap_table)
         {
            colormap_table->deserialize(in);
            if (in.read_checkpoint() != ERR_STATUS_GOOD)  return ERR_STATEFILE_COR;
         }

         in.read_int(&wlen);
	 string win_id;

         // if all windows weren't cleared but there are no windows in
         // the state file then clear all window regardless. According
         // to the state file there are no windows and that trumps.
         if ((wlen == 0) && ((mask&V_CLEAR_ALL) != V_CLEAR_ALL))
            delete_all_windows(NULL, true);

         for (int rr =0; ((rr < wlen) && (rval == ERR_STATUS_GOOD)); rr++)
         {
	    cw = NULL;

            // add window 
	    if ((mask&V_CLEAR_ALL) != V_CLEAR_ALL)
            {
	       int length = in.tellg ();
	       in.read_string(&win_id);
	       cw = get_window (win_id);
	       in.seekg (length, ios_base::beg);
	    }
	    if (cw)
            {
	       cw->deserialize (in);

               // since the window is just being deserialized, its possible
               // that the window was visible before the load state but
               // not visible after. If its not visible make sure its hidden.
               if (!cw->is_visible ())
                  cw->hide ();
	       current = cw->get_id();
	       cw->update_all ();
	       ChipsId cid (win_id.c_str (), "all");
	       cw->delete_frame (&cid,true);
	       rval = cw->deserialize_frames(in);
	    }
            else if ((cw = new_window(in)) != NULL)
            { 
               rval = cw->deserialize_frames(in);
               window.push_back(cw);
               current = cw->get_id();
            }

	    /* NOTE: If CLEAR_ALL isn't performed and load state may
	       have extra windows still lingering that were created
	       after the state that was loaded. Still needs investigation */
         }
         in.read_string(&current);

         // Load in the deleted windows
         int sz;
         in.read_int (&sz);
         for(int jj=0; jj < sz; ++jj)
         {
            // Disabling redraw keeps the windows hidden. Use this
            // instead of a force_hide on the window, as a force_hide
            // would have to come after window was created (creating
            // a flicker on the screen
            this->disable_redraw ();
            if ((cw = new_window(in)) != NULL)
            {
               rval = cw->deserialize_frames(in);
               dwindow[cw->get_ref_cnt()] = cw;
            }
            this->enable_redraw ();
         }
         in.read_int(&window_cnt);
         int rc;
         in.read_int(&rc);
         ChipsObject::set_cur_ref_cnt((unsigned int)rc);
      }
      else if ((mask&V_LOAD_COMMANDS) == V_LOAD_COMMANDS)
      {
         int ss = sizeof(int);
         int offset;
         in.seekg(-ss, ios::end);
         in.read_int(&offset);
         in.seekg(-offset, ios::end);
      }
      // If clearall delete command manager so that all remnence of the
      // old command manager are erased
      if (((mask&V_CLEAR_ALL) == V_CLEAR_ALL) && (rval == ERR_STATUS_GOOD))
      {
         delete cmd_manager;
         cmd_manager = new ChipsCmdManager (this);
      }
      if (((mask&V_LOAD_COMMANDS) == V_LOAD_COMMANDS) &&
          (rval == ERR_STATUS_GOOD))
         rval = cmd_manager->deserialize(this, in);

      if (rval != ERR_STATUS_GOOD)
      {
	 // load failed- restore back up 
         in.close(); 
         load_state(backup, mask); 
         if (strstr(backup, "chips_backup"))  remove(backup);
         return rval; 
      }
      else  if (strstr(backup, "chips_backup"))  remove(backup);

      // write out undo state
      cmd_manager->set_cmd_lock(false);
   }
   else cmd_manager->add_err(rval); 
   return rval; 
}

int ChipsEngine::save_state(string fname, bool override, bool override_block,
   int mode)
{
   int rval = ERR_STATUS_GOOD;

   if (!override)  cmd_manager->clear_errs();

   // cannot save state if undo block is open. Internal save states
   // can override this and write accordingly while in a block
   if (cmd_manager->in_block () && !override_block)
   {
      cmd_manager->add_err(ERR_UNDO_BLOCK_OPN);
      return ERR_UNDO_BLOCK_OPN;
   }

   // determine access status
   if (fname.length() == 0)
   {
      cmd_manager->add_err(ERR_FILE_NOT_SUPPLIED);
      return ERR_FILE_NOT_SUPPLIED;
   }
   else if (!access(fname.c_str(), F_OK))
   {
      // internal save states should clobber the backup file 
      // regardless of clobber setttings- external save states
      // should honor export.clobber value.
      if (!override)
      {
         if (!theDefaults->get_exportclobber())
         {
            cmd_manager->add_err(ERR_FILE_CLOBBER_OFF,"",0,fname.c_str());
            return ERR_FILE_CLOBBER_OFF;
         }
         else if (access(fname.c_str(), W_OK) != 0)
         {
            cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,fname.c_str()); 
            return ERR_ACCESS_PERMISSION;
         }
      }
   }

   // set state file to binary if mode = 0, template = 1
   ios_base::openmode opmode = ios::out;
   if (mode == 0)   opmode |= ios::binary;
   ChipsStateStream out(fname.c_str(), opmode, mode);

   if ((rval = out.get_error()) == ERR_STATUS_GOOD)
   {
      list<ChipsWindow*>::iterator ii;
      if (out.is_binary())  // standard binary state file
      {
         int sz;

         if (theDefaults)
         {
            // serialize the preferences 
            theDefaults->serialize(out); 
            out.write_checkpoint(); 
         }
         if (out.get_minor_version() > 0 && fill_patterns)
         {
            fill_patterns->serialize(out);
            out.write_checkpoint();
         }
         if (out.get_minor_version () > 0 && colormap_table)
         {
            colormap_table->serialize(out);
            out.write_checkpoint();
         }

         sz = window.size(); 
         out.write_int(&sz);
         for(ii=window.begin(); ii != window.end(); ++ii)
            (*ii)->serialize(out);
         out.write_string(&current); 

         sz = dwindow.size ();
         out.write_int(&sz);
         map<unsigned int, ChipsWindow*>::iterator jj;
         for(jj=dwindow.begin(); jj != dwindow.end(); ++jj)
            jj->second->serialize(out);

         out.write_int(&window_cnt);
         int rc = (int)ChipsObject::get_cur_ref_cnt();
         out.write_int(&rc);

         out.seekp(0, ios::end);
         streampos sp = out.tellp();
         cmd_manager->serialize(out);

         // store offset to the Undo/Redo Commands
         out.seekp(0, ios::end);
         streampos sp2 = out.tellp();
         long offset = sp2 - sp;
         int ioff = offset + sizeof(int); 
         out.write_int(&ioff); 
      }
      else 
      {
         // ascii template 
         int simplify = 0; // allow consolidation of ascii state
         string prefs = fname;  // save current default settings
         prefs.append(".prefs");
         if ((access(prefs.c_str(), F_OK) != 0 ||
              access(prefs.c_str(), W_OK) == 0) && theDefaults)
            theDefaults->save_file(prefs, override);

         out << "\nopen_undo_block()          # treat script as single command"<<endl;
         out << "clear()                    # remove all existing chips objects\n"<<endl;
         out << "\n# This script assumes that the user's default preferences were used"<<endl;
         out <<  "# to generate the output and were not changed- so the following " << endl;
         out <<  "# load_preferences() command is unnecessary. To guarantee that this"<< endl;
         out <<  "# script uses the correct settings to replicate the saved state, " << endl;
         out <<  "# uncomment out the following line." << endl;
         out << "# load_preferences(\""<<prefs<<"\")\n"<<endl;

         // write out any user loaded colormaps
         if (colormap_table)   colormap_table->serialize_ascii(out);

         for(ii=window.begin(); ii != window.end(); ++ii)
            (*ii)->serialize_ascii(out, simplify++);

         // window currency
         ii = window.end();
         ii--;
         string lastid = (*ii)->get_id();
         if (window.size() > 1 && current != lastid)
            out << "current_window(\""<<current<<"\")"<<endl;

         out << "\nclose_undo_block()          # end treat script as single command"<<endl;
         out.close();

         // if state contains templates (ie, curve, histogram, contour, or 
         // image data) we need to generate a template header and append our
         // ascii state file to it.  
         string val = out.get_data_template();
         if (val.length())
         {
            string fname2 = fname;
            fname2.append(".template_tmp");
            ChipsStateStream out2(fname2.c_str(), opmode, mode);
            out2 << "### Data Templates" << endl;
            out2 << "# The template variables below are used to specify the data needed for"<<endl;
            out2 << "# objects such as curves. The values may be substituted with other file"<<endl;
            out2 << "# names or data arrays as appropriate. It is the user's responsibility to"<<endl;
            out2 << "# ensure that data arrays are all the correct length, that provided column"<<endl;
            out2 << "# names exist, and that if files are provided, they contain the appropriate"<<endl;
            out2 << "# columns in the correct order. Note that the variables are provided in two"<<endl;
            out2 << "# flavors- as filenames and as data structures: the former makes use of the"<<endl;
            out2 << "# simpler file interface while the latter offers more flexibility."<<endl;
            out2 << val << endl;
           
            // append the ascii state commands stored in the state file 
            ifstream in2(fname.c_str());
            out2 << in2.rdbuf();
            in2.close();
            out2.close();
            // replace the ascii state file with the updated version
            remove(fname.c_str());
            rename(fname2.c_str(), fname.c_str());
         }
      }
   }
   return rval; 
}


int ChipsEngine::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE start,
   string property,  string value)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();

   if (id.length() == 0)  id = current;
   if (theDefaults)
   {
      cmd_manager->open_block();
      string tval, prefix = property.substr(0,4);
      if (chips_strcmp(prefix.c_str(),"all.") != 0) 
      {
         tval = "all.";
	 tval.append(property);
      }
      else
         tval = property;
      if ((rval = theDefaults->set_preference(tval, value)) == 
          ERR_UNKNOWN_PROPERTY)
      {
         cmd_manager->remove_last_err();
	 tval = "Unknown or unsupported <obj>.<attribute> cascade entry: ";
	 tval.append(property);
         cmd_manager->add_err(tval, rval);
      }
      if (rval == ERR_STATUS_GOOD)
      {
         if (window.empty())
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
         {
            // get window
            if ((cw = get_window(id)) != NULL)
            {
               rval = cw->set_cascading_property(cid, start);
               cw->redraw();
            }
            else  
	       cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "window", id.c_str());
         }
         else 
         {
            list<ChipsWindow*>::iterator ii;
            int sts, pass = 0, failval = 0;
            for(ii=window.begin(); ii != window.end(); ++ii)
            {
               cw = *ii;
               if ((sts = cw->set_cascading_property(cid, start)) == ERR_STATUS_GOOD)
                  pass++;
               else if (!err_type_object_DNE(sts))
                  failval = sts;
               cw->redraw();
            }
            cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
               rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
         }
      }
      cmd_manager->close_block();
   }
   else
      cmd_manager->add_err(ERR_NO_PREFS_OBJ, id);
   return rval;
}

int ChipsEngine::move(ChipsId* cid, V_OBJECT_TYPE obj,
   double xx, double yy, int mode)
{
   return this->move (cid, obj, chips_not_subobject, xx, yy, mode);
}

int ChipsEngine::move(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj,
   double xx, double yy, int mode)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();
   if (id.length() == 0)  id = current;

   cmd_manager->clear_errs();

   if (rval == ERR_STATUS_GOOD)
   {
      cmd_manager->open_block();
      if (window.empty())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
      else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
      {
         // get window
         if ((cw = get_window(id)) != NULL)
         {
            rval = cw->move_object(cid, obj, subobj, xx, yy, mode);
            cw->redraw();
         }
         else  
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
      }
      else 
      {
         list<ChipsWindow*>::iterator ii;
         int rr, pass = 0, failval = 0;
   
         for(ii=window.begin(); ii != window.end(); ++ii)
         {
            cw = *ii;
            rr = cw->move_object(cid, obj, subobj, xx, yy, mode);
            if (rr == ERR_STATUS_GOOD)  
            {
               cw->redraw();
               pass++;
            }
            else if (!err_type_object_DNE(rr))
               failval = rr;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
         else  
            current = "all";
      }
      cmd_manager->close_block();
   }
   else 
      cmd_manager->add_err(rval, id);
   return rval;
}

int ChipsEngine::resize(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *resize_data)

{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();
   if (id.length() == 0)  id = current;

   cmd_manager->clear_errs();

   if (rval == ERR_STATUS_GOOD)
   {
      cmd_manager->open_block();
      if (window.empty())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
      else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
      {
         // get window
         if ((cw = get_window(id)) != NULL)
         {
            rval = cw->resize_object(cid, obj, subobj, resize_data);
            cw->redraw();
         }
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
      }
      else
      {
         list<ChipsWindow*>::iterator ii;
         int rr, pass = 0, failval = 0;

         for(ii=window.begin(); ii != window.end(); ++ii)
         {
            cw = *ii;
            rr = cw->resize_object(cid, obj, subobj, resize_data);
            if (rr == ERR_STATUS_GOOD)
            {
               cw->redraw();
               pass++;
            }
            else if (!err_type_object_DNE(rr))
               failval = rr;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
         else
            current = "all";
      }
      cmd_manager->close_block();
   }
   else
      cmd_manager->add_err(rval, id);
   return rval;
}

int ChipsEngine::select(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected)
{
   list<ChipsWindow*> cwindows;
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();

   if ((rval = get_windows (cid, cwindows, true)) == ERR_STATUS_GOOD)
   {
      list<ChipsWindow*>::iterator ii;
      int rr, pass = 0, failval = 0;
      string wid = cid->get_window_id ();

      cmd_manager->open_block();
      // run the command
      for(ii=cwindows.begin(); ii != cwindows.end(); ++ii)
      {
         cw = *ii;
         rr = cw->select_object(cid, obj, subobj, selected);
         if (rr == ERR_STATUS_GOOD)
         {
            cw->redraw();
            pass++;
         }
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }
      cmd_manager->close_block();

      // check all passed
      if (window.size () == cwindows.size())
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      }

      // update currency
      if (wid.length () > 0)
         current = wid;
   }

   return rval;

}



int ChipsEngine::rotate(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle)
{
   list<ChipsWindow*> cwindows;
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();

   if ((rval = get_windows (cid, cwindows, true)) == ERR_STATUS_GOOD)
   {
      list<ChipsWindow*>::iterator ii;
      int rr, pass = 0, failval = 0;
      string wid = cid->get_window_id ();

      cmd_manager->open_block();
      // run the command
      for(ii=cwindows.begin(); ii != cwindows.end(); ++ii)
      {
         cw = *ii;
         rr = cw->rotate_object(cid, obj, subobj, angle);
         if (rr == ERR_STATUS_GOOD)
         {
            cw->redraw();
            pass++;
         }
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }
      cmd_manager->close_block();

      // check all passed
      if (window.size () == cwindows.size())
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      }
      
      // update currency
      if (wid.length () > 0)
         current = wid;
   }

   return rval;
}

int ChipsEngine::get_annotation_position (const ChipsId* cid, 
                                          V_OBJECT_TYPE obj, ChipsVector &pos)
{
   list<ChipsWindow*> cwindows;
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();

   if ((rval = get_windows (cid, cwindows, false)) == ERR_STATUS_GOOD)
   {
      cmd_manager->open_block();
      // run the command
      cw = *(cwindows.begin());
      cw->get_annotation_position (cid, obj, pos);
      cmd_manager->close_block();
   }

   return rval;

}

int ChipsEngine::change_annotation_coordinate (const ChipsId* cid, 
                                          V_OBJECT_TYPE obj, 
                                          COORD_SYS coordsys, 
                                          const string& coordid)
{
   list<ChipsWindow*> cwindows;
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();

   if ((rval = get_windows (cid, cwindows, true)) == ERR_STATUS_GOOD)
   {
      list<ChipsWindow*>::iterator ii;
      int rr, pass = 0, failval = 0;
      string wid = cid->get_window_id ();

      cmd_manager->open_block();
      // run the command
      for(ii=cwindows.begin(); ii != cwindows.end(); ++ii)
      {
         cw = *ii;
         rr = cw->change_annotation_coordinate(cid, obj, coordsys, coordid);
         if (rr == ERR_STATUS_GOOD)
         {
            cw->redraw();
            pass++;
         }
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }
      cmd_manager->close_block();

      // check all passed
      if (window.size () == cwindows.size())
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      }

      // update currency
      if (wid.length () > 0)
         current = wid;
   }

   return rval;
}


int ChipsEngine::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj,
   int& out_val)
{
   int rval = ERR_STATUS_GOOD; 
   cmd_manager->clear_errs();
   out_val = 0; 
   if (!window.empty())
   {
      string id = cid->get_window_id();
      if (id.length() == 0)  id = current;
      if (chips_strcmp(id.c_str(), "all") != 0)
      {
         // get window
         ChipsWindow* cw = get_window(id);
         if (cw)
         {
            if (obj == chips_window)
               out_val = 1; 
            else 
            {
               // check out frames 
               rval = cw->get_object_count(cid, obj, out_val);
            }
         }
      }
      else
      {
         if (obj == chips_window)
            out_val = window.size();  
         else
         {
            list<ChipsWindow*>::iterator ii;
            int cnt; 
            for(ii=window.begin(); ii != window.end(); ++ii)
            {
               cnt = 0; 
               (*ii)->get_object_count(cid, obj, cnt);
               out_val += cnt; 
            }
         }
      }
   }
   return rval;
}


int ChipsEngine::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   if (!window.empty())
   {
      string id = cid->get_window_id();
      if (id.length() == 0)  id = current;

      // get window
      ChipsWindow* cw = get_window(id);
      if (cw)
      {
         if (obj == chips_window)
         {
            out_val = COORD_NONE;
            out_id = ""; 
         }
         else 
            rval = cw->get_object_coordinfo(cid, obj, out_val, out_id);
      }
      else if (chips_strcmp(id.c_str(), "all") != 0)
         cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else   cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
   return rval;
}

int ChipsEngine::get_coord_object_count(ChipsId *cid, int objtype)
{
   int cnt = 0;
   if (window.empty()) return cnt; 

   string id = cid->get_window_id();
   if (id.length() == 0)  id = current;
   if ((id != "all") || ((id == "all") && (window.size() == 1)))
   {
      ChipsWindow* cw = get_window(id);
      if (cw)  cnt = cw->get_coord_object_count(cid, objtype);
   }
   return cnt;
}

int ChipsEngine::set_line_stipple(V_LINE_TYPE line, int val)
{
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   switch(line)
   {
      case chips_solid:
         chips_solid_pattern = val; 
      break; 
      case chips_dot:
         chips_dot_pattern = val; 
      break; 
      case chips_dotlongdash:
         chips_dotlongdash_pattern = val; 
      break; 
      case chips_dotshortdash:
         chips_dotshortdash_pattern = val; 
      break; 
      case chips_longdash:
         chips_longdash_pattern = val; 
      break; 
      case chips_shortdash:
         chips_shortdash_pattern = val; 
      break; 
      case chips_shortdashlongdash:
         chips_shortdashlongdash_pattern = val; 
      break; 
      default: 
         cmd_manager->add_err(rval=ERR_INT_SUBSET, "", 0, "line type", line,
            "[chips_solid|chips_dot|chips_dotlongdash|chips_dotshortdash|chips_longdash|chips_shortdash|chips_shortdashlongdash|1-7]");
      break; 
   }
   return rval;
}


int ChipsEngine::get_line_stipple(V_LINE_TYPE line, int &out_val)
{
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   switch(line)
   {
      case chips_solid:
         out_val = chips_solid_pattern; 
      break; 
      case chips_dot:
         out_val = chips_dot_pattern; 
      break;
      case chips_dotlongdash:
         out_val = chips_dotlongdash_pattern; 
      break;
      case chips_dotshortdash:
         out_val = chips_dotshortdash_pattern; 
      break;
      case chips_longdash:
         out_val = chips_longdash_pattern; 
      break;
      case chips_shortdash:
         out_val = chips_shortdash_pattern; 
      break;
      case chips_shortdashlongdash:
         out_val = chips_shortdashlongdash_pattern; 
      break;
      default: 
         out_val = 0; 
         cmd_manager->add_err(rval=ERR_INT_SUBSET, "", 0, "line type", line,
            "[chips_solid|chips_dot|chips_dotlongdash|chips_dotshortdash|chips_longdash|chips_shortdash|chips_shortdashlongdash|1-7]");
      break;
   }
   return rval;
}


int ChipsEngine::setup_pick (const ChipsId *cid, int count, 
                             unsigned int &inout_mask, ChipsId &out_cid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   out_cid = *cid;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (theDefaults && (inout_mask == 0))
      inout_mask = theDefaults->get_pickmask ();

   if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
         if (count < 0)
            cmd_manager->add_err (rval=ERR_NO_INT_NEGATIVE, id, 0, count);
         else if (inout_mask == 0)
            cmd_manager->add_err (rval=ERR_NO_PICK_MASK);   // no mask set
         else if (!cw->get_redraw ())
            cmd_manager->add_err (rval=ERR_PICK_DISPLAY);
         else if (!this->check_coordinate_exists (&out_cid))
            cmd_manager->add_err (rval=ERR_COORD_DNE, id, 0, v_coord_sys_to_string (out_cid.get_coord_sys ()));
      }
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows");
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);

   if (rval == ERR_STATUS_GOOD)
   {
      if (out_cid.get_coord_sys () == COORD_NONE)
         get_current_coord (&out_cid);
      if (get_window(id) != NULL)
         current = id;
   }
   return rval;
}




// engine level method to set 'current' objects 
int ChipsEngine::set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec, bool setmask)
{
   int  rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   string tmpid = cid->get_window_id();

   // if we are setting something other than window and a layout spec
   // window was not provided, use current window val 
   if ((tmpid.length() == 0) && (what != chips_window))
      tmpid = current;
 
   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         if (what == chips_window)
         {
            if (!window.empty())
            {
               if (rec)
               {
                  list<ChipsWindow*>::iterator ii;
                  ii=window.begin(); 
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand((*ii),
                     current, tmpid, chips_window);
                  if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
                     delete cmd;
               }
               current = "all";
            }
            else  cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         }
         else
         {
            list<ChipsWindow*>::iterator ii;
            int sts, pass = 0, failval = 0;
            for(ii=window.begin(); ii != window.end(); ++ii)
               if ((sts = (*ii)->set_current(cid,what,rec)) == ERR_STATUS_GOOD)
                  pass++;
               else  if (!err_type_object_DNE(sts))
                  failval = sts;
            cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
               rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
            else 
               current = "all";
         }
      }
      else
      {
         ChipsWindow* tmp = get_window(tmpid);

         if (tmp)
            if (what == chips_window)
            {
               if (rec)
               {
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand(tmp, 
                     current, tmpid, chips_window);
                  if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
                     delete cmd;
               }
               current = tmpid;
            }
            else 
               rval = tmp->set_current(cid, what, rec);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "window", 
	       tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) current = tmpid; 
   }
   else  
      cmd_manager->add_err(rval=ERR_NO_OBJ_ID,"",0,"window");
   if (setmask && rval == ERR_STATUS_GOOD)
      cmd_manager->update_mask(STATE_CHNG_CURRENCY);
   return rval;
}


ChipsObject* ChipsEngine::get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                                      bool include_deref)
{
   ChipsObject* obj = NULL;
   unsigned int obj_ref;

   if (cid)
   {
      ChipsWindow* win;
      string wid = cid->get_window_id(); 
      obj_ref = cid->get_window_ref_id ();
      if ((wid != "all") || ((wid == "all") && (window.size() == 1)))
      {
         win = get_window(wid);
         if (win && (otype != chips_window)) 
            obj = win->get_object(cid, otype, include_deref);
         else if (win)
         {
            // if obj_ref is zero then accept otherwise confirm the reference count
            if ((obj_ref == 0) || ((obj_ref > 0) && (win->get_ref_cnt () == obj_ref)))
	       obj = win;
         }
      }
      if (include_deref && !obj)
      {
         ChipsWindow* win = NULL;
         if (dwindow.count (obj_ref) != 0)
            win = dwindow[obj_ref];
         if (win && (otype != chips_window))
            obj = win->get_object(cid, otype, include_deref);
         else
            obj = win;
      }
   }
   return obj;
}

int ChipsEngine::get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num)
{

   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   if (!window.empty())
   {
      string id = cid->get_window_id();
      if (id.length() == 0)  id = current;

      // get window
      ChipsWindow* cw = get_window(id);
      if (cw)
         rval = cw->get_object_depth_info(cid, otype, out_depth_num, out_shuffle_num);
      else if (chips_strcmp(id.c_str(), "all") != 0)
         cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else   cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
   return rval;


}

// add a new window to the engine's list of windows 
int ChipsEngine::add_window(ChipsId* cid, string id, bool reset_errs)
{
   int  rval = ERR_STATUS_GOOD;     
   ChipsWindow* cw = NULL;
   string wid; 
   if (reset_errs){
      cmd_manager->clear_errs();
   }
   if (id.length())
   {
      if ((rval = theDefaults->set_overrides(id, "window")) != ERR_STATUS_GOOD)
         return rval;
   }
   wid = theDefaults->get_windowid();

   if (wid.length() && get_window(wid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE, "", 0, wid.c_str(), "window");
   else 
      cw = new_window(wid);

   if (cw)
   {
      ChipsCreateCommand *cmd = new ChipsCreateCommand(cw, (wid.length()==0), current);
      if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
         delete cmd;

      cw->set_window_num (window_cnt);
      window.push_back(cw);
      current = cw->get_id();
      cid->set_window_id(current); 
   }

// This was placed here because of some X-sync error. I'm removing it
// because I cant recreate the error and believe that it has been fixed.
//   if (rval == ERR_STATUS_GOOD)  sleep(2); 

   if (id.length()) theDefaults->reset_defaults();

   return rval; 
}


// add a new window with the specified dimensions 
int ChipsEngine::add_window(ChipsId* cid, float ww, float hh, 
   string format, string id, bool reset_errs)
{
   int  rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw = NULL;
   string wid;
   V_UNITS_TYPE unit;
   
   if (reset_errs){
      cmd_manager->clear_errs();
   }
   if (id.length())
   {
      if ((rval = theDefaults->set_overrides(id, "window")) != ERR_STATUS_GOOD)
         return rval;
   }
   wid = theDefaults->get_windowid(); 

   if (hh <= 0) 
      cmd_manager->add_err(rval=ERR_DBL_MIN,"",0,"height", hh, 0.0); 
   else if (ww <= 0)
      cmd_manager->add_err(rval=ERR_DBL_MIN,"",0,"width", ww, 0.0); 
   else if ((format.length () > 0) &&
           ((rval = c2_string_to_units(format.c_str(), &unit, "window.units"))
                != ERR_STATUS_GOOD)); 
      // do nothing if an error is found. c2_string_to_units adds the
      // error to the command manager
   else if (id.length())
   {
      if (wid.length() && get_window(wid))
         cmd_manager->add_err(rval=ERR_ID_IN_USE, "", 0, wid.c_str(), "window");
      else   
         cw = new_window(ww, hh, format, wid);
   }
   else
      cw = new_window(ww, hh, format, id);


   if (cw)
   {
      bool chng = ((id.length() == 0) || (wid.length() == 0));
      ChipsCreateCommand *cmd = new ChipsCreateCommand(cw,chng,current);
      if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
         delete cmd;

      cw->set_window_num (window_cnt);
      window.push_back(cw);
      current = cw->get_id();
      cid->set_window_id(current); 
   }

// This was placed here because of some X-sync error. I'm removing it
// because I cant recreate the error and believe that it has been fixed.
//   if (rval == ERR_STATUS_GOOD)  sleep(2); 

   if (id.length()) theDefaults->reset_defaults(); 

   return rval; 
}


// remove a window from the engine 
int ChipsEngine::delete_window(ChipsId* cid, bool force_del)
{
   ChipsWindow* cw; 
   int rval = ERR_OBJ_VAL_DNE; 
   string wid = cid->get_window_id(); 

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (window.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
      return rval;
   }
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      bool opened_block = false;

      if (!force_del){
         cmd_manager->open_block ();
         opened_block = true;
      }

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = (*ii);
         if (force_del)
            cw->object_delete ();
	 else
	 {
            cw->force_hide (); // create offscreen renderer
            if (!cmd_manager->cmd_lock())
	    {
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cw, current);
               if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
                  delete cmd;
            }
            // insert the window into a list of derefernced windows
	    dwindow[cw->get_ref_cnt ()] = cw;
         }
      }
      window.clear(); 
      current=""; 
      cid->reset_window_id(""); 
      if (!force_del && opened_block) cmd_manager->close_block ();
      rval = ERR_STATUS_GOOD; 
   }
   else
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); rval && (ii != window.end()); ++ii)
      {
         if ((*ii)->get_id() == wid)
         {
            cw = (*ii);
            window.erase(ii);
            if (force_del)
               cw->object_delete ();
            else
	    {
               cw->force_hide (); // create offscreen renderer
               if (!cmd_manager->cmd_lock())
	       {
                  ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cw, cw->get_id ());
                  if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
                     delete cmd;
               }
               // insert the window into a list of derefernced windows
               dwindow[cw->get_ref_cnt ()] = cw;
            }
            cid->reset_window_id(""); 
            rval = ERR_STATUS_GOOD; 
	    break;
         }
      }
      // set current to last window created
      if (rval == ERR_STATUS_GOOD) 
      {
         if (!window.empty())
         {
	    cw = window.back();
            current = cw->get_id(); 
         }
         else   current = "";
      }
      else if (rval == ERR_OBJ_VAL_DNE)
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());

   }

   return rval;
}


int ChipsEngine::set_redraw(bool val)
{
   for(list<ChipsWindow*>::iterator ii=window.begin(); ii != window.end(); ++ii)
      (*ii)->set_redraw(val);
   return ERR_STATUS_GOOD; 
}

int ChipsEngine::append_title_text (const ChipsId *cid, string text)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   if (id.length() == 0)  id = current;

   if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->set_title_suffix (text);
      }
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows");
      else   // specified window does not exist
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);

   if ((rval == ERR_STATUS_GOOD) && (cw != NULL))  current = id;
   return rval;
}

void ChipsEngine::redraw_window(string wid)
{
   if (!redraw_block)
   {
      string val;
      // if an id provided, redraw that window, otherwise use currency
      if (wid.length())  val = wid;
      else               val = current;
      if (val == "all")
      {
         list<ChipsWindow*>::iterator ii;
         for(ii=window.begin(); ii != window.end(); ++ii)
         {
            // always try to reshow the window incase it was not
            // allowed to be shown during the block
            (*ii)->activate ();

            // redraw the window if necessary
            if ((*ii)->get_redraw())
               (*ii)->redraw();
         }
      }
      else
      {
         ChipsWindow* cw = get_window(val);
         if (cw)
         {
            // always try to reshow the window incase it was not
            // allowed to be shown during the block
            cw->activate ();

            // redraw the window if necessary
            if (cw->get_redraw())  
               cw->redraw();
         }
      }
   }
}


// add a frame  
int ChipsEngine::add_frame(ChipsId* cid, string fid, bool reset_errs)
{
   int  rval = ERR_STATUS_GOOD;   
   ChipsWindow *cw = NULL;
   string wid = cid->get_window_id();
   if(reset_errs){
      cmd_manager->clear_errs();
   }
   if (fid.length())
   {
      if ((rval = theDefaults->set_overrides(fid, "frame")) != ERR_STATUS_GOOD)
         return rval; 
   }

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            pass++;
            cw->redraw();
         }
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
            else  current = "all";
      }
   }
   else if (wid == "")
   {
      // need to create a new window 
      add_window(cid); 
      cmd_manager->set_cmd_lock(true);
      if ((cw = get_window(current.c_str())) != NULL)
         rval = cw->add_frame(cid);
      else 
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   else
   {
      if ((cw = get_window(wid.c_str())) != NULL)
      {
         if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = wid;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   }  

   if (fid.length())  theDefaults->reset_defaults();
   return rval;
}


// add a frame at the specified normalized coordinates 
int ChipsEngine::add_frame(ChipsId* cid, float x1, float y1, float x2, 
   float y2, string fid, bool reset_errs)
{
   int sts, pass = 0, failval = 0, rval = ERR_STATUS_GOOD;   
   ChipsWindow *cw = NULL;
   string wid = cid->get_window_id();
   if (reset_errs){
      cmd_manager->clear_errs();
   }
   if (fid.length())
   {
      if ((rval = theDefaults->set_overrides(fid, "frame")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (wid.length() == 0)  wid = current;
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->add_frame(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
         {
            pass++;
            cw->redraw();
         }
         else 
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
      {
         cid->set_window_id("all"); 
         current = "all"; 
      }
   }
   else if (wid == "")
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
         rval = cw->add_frame(cid, x1, y1, x2, y2);
         if (rval == ERR_STATUS_GOOD)  cid->set_window_id(current);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->add_frame(cid,x1,y1,x2,y2)) == ERR_STATUS_GOOD) 
      {
         cid->set_window_id(wid);
         cw->redraw();
         current = wid;
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "window", wid.c_str());
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (fid.length())  theDefaults->reset_defaults();
   return rval;
}


// remove a frame
int ChipsEngine::delete_frame(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id(); 

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
	 cid->set_frame_id(fid);
         if ((sts = cw->delete_frame(cid)) == ERR_STATUS_GOOD)
         {
            pass++; 
            cw->redraw();
         }
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->delete_frame(cid)) == ERR_STATUS_GOOD)
	 {
	    cw->update_display();   // in case we're in single frame mode
	    cw->redraw();
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);

/*   cmd_manager->set_cmd_lock(false);
*/
   if (rval == ERR_STATUS_GOOD){
      current = id; 
      cw->redraw ();
   }
   return rval; 
}



int ChipsEngine::set_property(ChipsId* cid, void* in_val, int in_property)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   if (id.length() == 0)  id = current;

   cmd_manager->clear_errs();
   cmd_manager->open_block();

   if (window.empty())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
   else if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int rr, pass = 0, failval = 0; 
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((rr=cw->set_property(cid, in_val, in_property)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            pass++; 
         }
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else 
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->set_property(cid,in_val,in_property); 
         cw->redraw();
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }

   cmd_manager->close_block();

   if ((rval == ERR_STATUS_GOOD) && (current != id) && cw)
   {
      // if window currency changed, store off command for undo/redo
      ChipsCurrentCommand *cmd = new ChipsCurrentCommand(cw, current,
                                                         id, chips_window);
      if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
         delete cmd;
      current = id; 
   }
   return rval;
}



int ChipsEngine::get_property(ChipsId* cid, void* out_val, int out_prop)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (window.empty())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
   else 
   {
      // get window
      if ((cw = get_window(id)) != NULL)
         rval = cw->get_property(cid, out_val, out_prop);
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows"); 
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   return rval;
}


// create a plot area
int ChipsEngine::add_plot(ChipsId* cid, string plid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();

   if (plid.length())
   {
      if ((rval = theDefaults->set_overrides(plid, "plot")) != ERR_STATUS_GOOD)
         return rval;
   }
   cmd_manager->open_block();

   if (id.length() == 0)  id = current;
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->add_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
         cid->set_frame_id(fid);
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_plot(cid)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id; 
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
         if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            if ((rval=cw->add_plot(cid)) == ERR_STATUS_GOOD)
               cid->set_window_id(current);
         }
      }
      else   cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   cmd_manager->close_block();
   if (plid.length())  theDefaults->reset_defaults();
   return rval;
}


// create a plot area
int ChipsEngine::add_plot(ChipsId* cid,
   float px1, float py1, float px2, float py2, string plid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (plid.length())
   {
      if ((rval = theDefaults->set_overrides(plid, "plot")) != ERR_STATUS_GOOD)
         return rval;
   }
   cmd_manager->open_block();

   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string fid = cid->get_frame_id();
      int sts, pass = 0, failval = 0;

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->add_plot(cid,px1,py1,px2,py2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
         cid->set_frame_id(fid);
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval=cw->add_plot(cid,px1,py1,px2,py2)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id; 
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
         if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD) 
         {
            if ((rval=cw->add_plot(cid,px1,py1,px2,py2)) == ERR_STATUS_GOOD)
               cid->set_window_id(current);
         }
      }
      else   cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   cmd_manager->close_block();
   if (plid.length())  theDefaults->reset_defaults();
   return rval;
}


// remove a plot 
int ChipsEngine::delete_plot(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();

   if (window.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
      return rval;
   }
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->delete_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else 
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->delete_plot(cid);
         if (rval == ERR_STATUS_GOOD)  cw->redraw(); 
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   if (rval == ERR_STATUS_GOOD)  current = id;
   cmd_manager->close_block();
   return rval;
}


// clear items from a plot
int ChipsEngine::clear_plot(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();

   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->clear_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->clear_plot(cid);
         if (rval == ERR_STATUS_GOOD)  cw->redraw();
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   cmd_manager->close_block();

   if ((rval == ERR_STATUS_GOOD) && (current != id) && cw)
   {
      // if window currency changed, store off command for undo/redo
      ChipsCurrentCommand *cmd = new ChipsCurrentCommand(cw, current,
                                                         id, chips_window);
      if (cmd_manager->add_cmd (cmd) == ERR_STATUS_BAD)
         delete cmd;
      current = id;
   }

   return rval;
}


// change location of a plot
int  ChipsEngine::reposition_plot(ChipsId* cid,
   float x1, float y1, float x2, float y2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   cmd_manager->open_block();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts=cw->reposition_plot(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
         {
            pass++;
            cw->redraw();
         }
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->reposition_plot(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());

   cmd_manager->close_block();

   return rval;
}


int ChipsEngine::limits(ChipsId* cid, int dim, double xmin, double xmax, double ymin, double ymax)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, xmin, xmax, ymin, ymax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->limits(cid, dim, xmin, xmax, ymin, ymax)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;

}

int ChipsEngine::limits(ChipsId* cid, int dim, double min, double max)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval; 
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, min, max)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else 
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->limits(cid, dim, min, max)) == ERR_STATUS_GOOD)
      { 
         cw->redraw(); 
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   V_AUTOLIMIT_TYPE automax)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, automin, automax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->limits(cid, dim, automin, automax)) == ERR_STATUS_GOOD)  
      {
         cw->redraw();
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str()); 
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE automax)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, min, automax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->limits(cid, dim, min, automax)) == ERR_STATUS_GOOD) 
      {
         cw->redraw();
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   double max)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, automin, max)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->limits(cid, dim, automin, max)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      };
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::limits(ChipsId* cid, int dim, V_OBJECT_TYPE obj, ChipsId* ccid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   // first get limits from appropriate obj
   double ov[4];
   cmd_manager->clear_errs();

   if (obj == chips_curve) 
      rval = get_property(ccid, (void*)&ov, GET_CURVE_RANGE);
   else if (obj == chips_contour) 
      rval = get_property(ccid, (void*)&ov, GET_CONTOUR_RANGE);
   else if (obj == chips_histogram) 
      rval = get_property(ccid, (void*)&ov, GET_HISTO_RANGE);
   else if (obj == chips_image) 
      rval = get_property(ccid, (void*)&ov, GET_IMAGE_RANGE);
   else 
   {
      cmd_manager->add_err(ERR_LIMIT_OBJ); 
      return ERR_LIMIT_OBJ;
   }

   // no need to proceed if we didn't get our limits
   if (rval != ERR_STATUS_GOOD)  return rval;

   if (id.length() == 0)  id = current;

   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->limits(cid, dim, ov[0], ov[1], ov[2], ov[3])) == ERR_STATUS_GOOD) 
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval=cw->limits(cid,dim,ov[0],ov[1],ov[2],ov[3])) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::zoom(ChipsId* cid)
{
   int rval = ERR_FEATURE_DNE;
   cmd_manager->clear_errs();
   cmd_manager->add_err ("The interactive zoom command has been deprecated. Interactive zoom is now available directly through the gui.", rval);

   return rval;
}

int ChipsEngine::zoom(ChipsId* cid, float scale)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->zoom(cid, scale)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->zoom(cid, scale)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      };
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::pan(ChipsId* pcid)
{
   int rval = ERR_FEATURE_DNE;
   cmd_manager->clear_errs();
   cmd_manager->add_err ("The interactive pan command has been deprecated. Interactive pan is now available directly through the gui.", rval);

   return rval;

}

int ChipsEngine::panto(ChipsId* cid, double x, double y)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->panto(cid, x, y)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->panto(cid, x, y)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      };
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::pan(ChipsId* cid, double x1, double y1, double x2, double y2)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval;
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->pan(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->pan(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      };
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


// create a curve
int ChipsEngine::add_curve(ChipsId* cid, long cnt, double* x, double* y,
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform, string cvid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->open_block();

   cmd_manager->clear_errs();
   
    //Check if transform is valid
    ChipsTransformFactory transform_factory;
    transform = transform_factory.checkValidTransform(transform);
      
   if (cvid.length())
   {
      if ((rval = theDefaults->set_overrides(cvid, "curve"))!= ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string fid = cid->get_frame_id();
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
         if ((sts = cw->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup, 
              ylims, xlims, transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {

      
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup,
                              ylims, xlims, transform)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id; 
         }
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);
      if ((cw = get_window(current.c_str())) != NULL)
      {
	 if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {   
            rval = cw->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup,
                                 ylims, xlims, transform);
            if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
         }
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   cmd_manager->close_block();
   if (cvid.length())  theDefaults->reset_defaults();

   return rval;
}


// remove a curve
int ChipsEngine::delete_curve(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_curve_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_curve_id(cvid);
         cw = *ii;
         if ((sts = cw->delete_curve(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else 
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->delete_curve(cid)) == ERR_STATUS_GOOD)
      {
         cw->redraw(); 
         current = id;
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int numlev, double* levels, Transform *trans,  string ctid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;
   bool oldredraw;

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   trans = transform_factory.checkValidTransform(trans);
   
   if (ctid.length())
   {
      if ((rval = theDefaults->set_overrides(ctid, "contour")) 
          != ERR_STATUS_GOOD)   return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();
   cmd_manager->set_autocommit(false);

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         tmpid.set_frame_id(fid); 
         if ((sts = (*ii)->add_contour(&tmpid, cnt, x, y, z, numlev, levels, trans)) ==
              ERR_STATUS_GOOD)
           pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         oldredraw = cw->get_redraw();
         cw->set_redraw(false);
         rval = cw->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
         if (rval == ERR_STATUS_GOOD)
         {
            current = id;
            cw->set_redraw(oldredraw);
            cw->redraw();
         }
         else   cw->set_redraw(oldredraw);
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
	 if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            rval = cw->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
            if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
         }
      }
      else   cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   cmd_manager->set_autocommit(true);

   if (ctid.length())  theDefaults->reset_defaults();
   if (rval != ERR_STATUS_GOOD)
      cmd_manager->clear_hold_buffer();
   else
      cmd_manager->commit_hold_buffer();
   cmd_manager->close_block();

   return rval;
}


int ChipsEngine::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans, string ctid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   trans = transform_factory.checkValidTransform(trans);
   
   if (ctid.length())
   {
      if ((rval = theDefaults->set_overrides(ctid, "contour")) 
          != ERR_STATUS_GOOD)   return rval;
   }

   if (id.length() == 0)  id = current;

   cmd_manager->open_block();
   cmd_manager->set_autocommit(false);

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string fid = cid->get_frame_id();
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         ChipsId tmpid = *cid;
         tmpid.set_frame_id(fid); 
         if ((sts = cw->add_contour(&tmpid, dat, xdim, ydim, numlev, levels, trans)) ==
              ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_contour(cid, dat, xdim, ydim, numlev, levels, trans);
         if (rval == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id;
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
	 if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            rval = cw->add_contour(cid, dat, xdim, ydim, numlev, levels, trans);
            if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
         }
      }
      else   cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   cmd_manager->set_autocommit(true);

   if (ctid.length())  theDefaults->reset_defaults();
   if (rval != ERR_STATUS_GOOD)
      cmd_manager->clear_hold_buffer();
   else
      cmd_manager->commit_hold_buffer();
   cmd_manager->close_block();

   return rval;
}


int ChipsEngine::delete_contour(ChipsId* cid)
{
  ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ctid = cid->get_contour_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_contour_id(ctid);
         if ((sts = cw->delete_contour(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->delete_contour(cid)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::set_contour_levels(ChipsId* cid,
   unsigned int arraysize, double* pos)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts=cw->set_contour_levels(cid,arraysize,pos)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval=cw->set_contour_levels(cid,arraysize,pos)) == ERR_STATUS_GOOD)
      {
         cw->redraw();
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (window.empty())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
   else
   {
      // get window
      if ((cw = get_window(id)) != NULL)
         rval = cw->get_contour_levels(cid, pos);
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows");
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }


   return rval;
}


int ChipsEngine::add_histogram(ChipsId* cid, long cnt, 
   double* bin_low, double* bin_high, double* counts, double* yd, double* yu, 
   string hid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (hid.length())
   {
      if ((rval = theDefaults->set_overrides(hid, "histogram")) != 
         ERR_STATUS_GOOD)     return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         if ((sts = (*ii)->add_histogram(&tmpid, cnt, bin_low, bin_high, 
             counts, yd, yu)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_histogram(cid, cnt, bin_low, bin_high, counts, yd, yu);
         if (rval == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(current.c_str())) != NULL)
      {
	 if ((rval = cw->add_frame(cid)) == ERR_STATUS_GOOD)
         {
            rval = cw->add_histogram(cid, cnt,  bin_low, bin_high, 
                                     counts, yd, yu);
            if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
         }
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (hid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::delete_histogram(ChipsId* cid)
{
  ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string hid = cid->get_histogram_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_histogram_id(hid);
         if ((sts = cw->delete_histogram(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->delete_histogram(cid))  == ERR_STATUS_GOOD) 
      {
         cw->redraw();
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


// add axis
int ChipsEngine::add_axis(ChipsId* cid, int dim, float pos,
   float minx, float maxx, float miny, float maxy, Transform *transform, string aid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   transform = transform_factory.checkValidTransform(transform);
   
   if (aid.length())
   {
      if ((rval = theDefaults->set_overrides(aid, "axis")) != ERR_STATUS_GOOD)
         return rval;
   }
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string fid = cid->get_frame_id();
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
	 ChipsId tmpid = *cid;

         if ((sts = cw->add_axis(&tmpid,dim,pos,minx,maxx,miny,maxy,transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
	 cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_axis(cid,dim,pos,minx,maxx,miny,maxy,transform)) == ERR_STATUS_GOOD)
	 {
	    cw->redraw();
	    current = id; 
	 }
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);
      if ((cw = get_window(current.c_str())) != NULL)
      {
         rval = cw->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);
         if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (aid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


// remove an axis
int ChipsEngine::delete_axis(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string aid = cid->get_axis_id();
      string axid = cid->get_xaxis_id();
      string ayid = cid->get_yaxis_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_axis_id(aid);
         cid->set_xaxis_id(axid);
         cid->set_yaxis_id(ayid);
	 if ((sts = (*ii)->delete_axis(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if ((rval = cw->delete_axis(cid)) == ERR_STATUS_GOOD) 
      {
         cw->redraw(); 
         current = id;
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   cmd_manager->open_block();
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         sts = cw->set_arbitrary_tick_positions(cid, arraysize, pos, vals);
         if (sts == ERR_STATUS_GOOD) 
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw(); 
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else 
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->set_arbitrary_tick_positions(cid, arraysize, pos,
              vals)) == ERR_STATUS_GOOD)  
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::bind_axes(ChipsId* cid, const string lid, const string aid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;
   if (id != "")
   {
      cmd_manager->open_block();
      // get window
      if ((cw = get_window(id)) != NULL)
         rval = cw->bind_axes(cid, lid, aid);
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows"); 
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
      cmd_manager->close_block();
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);

   if (rval == ERR_STATUS_GOOD)
      cw->redraw ();
   return rval;
}


int ChipsEngine::unbind_axes(ChipsId* cid, const string lid, const string aid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;
   if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
         rval = cw->unbind_axes(cid, lid, aid);
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows"); 
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
   return rval;
}

int ChipsEngine::format_axis_point(const ChipsId* cid, double pnt, string &formatted_pnt)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;
   if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
         rval = cw->format_axis_point(cid, pnt, formatted_pnt);
      else if (chips_strcmp(id.c_str(), "all") == 0)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"windows");
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
   return rval;
}



// add an image 
int ChipsEngine::add_image(ChipsId* cid, string filename, Transform *transform, string iid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   transform = transform_factory.checkValidTransform(transform);
   
   if (iid.length())
   {
      if ((rval = theDefaults->set_overrides(iid, "image")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
         if ((sts = cw->add_image(cid, filename, transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_image(cid, filename, transform)) == ERR_STATUS_GOOD)
         {
            cw->redraw(); 
            current = id; 
         }
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_image(cid, filename, transform);
         if (rval == ERR_STATUS_GOOD)  cid->set_window_id(current);
      }
      else cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   cmd_manager->close_block();

   if (iid.length())  theDefaults->reset_defaults();
   return rval;
}


// add an image 
int ChipsEngine::add_image(ChipsId* cid,  string filename, 
   float ix1, float iy1, float ix2, float iy2, Transform *transform, string iid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   transform = transform_factory.checkValidTransform(transform);
   
   if (iid.length())
   {
      if ((rval = theDefaults->set_overrides(iid, "image")) != ERR_STATUS_GOOD)
         return rval;
   }
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii = window.begin(); ii != window.end();++ii)
      {
         cw = *ii;
         cid->reset_frame_id(fid);
         if ((sts = cw->add_image(cid, filename, ix1, iy1, ix2, iy2, transform)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval=cw->add_image(cid,filename,ix1,iy1,ix2,iy2, transform)) == ERR_STATUS_GOOD)
         {
            cw->redraw(); 
            current = id; 
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_image(cid, filename, ix1, iy1, ix2, iy2, transform);
         if (rval == ERR_STATUS_GOOD)   cid->set_window_id(current);
      }
      else cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   } 
   cmd_manager->close_block();

   if (iid.length())  theDefaults->reset_defaults();
   return rval;
}

// add an image 
int ChipsEngine::add_image(ChipsId* cid, double* rvals, double* gvals, double *bvals, double *avals, int xaxis, int yaxis, 
   Transform *transform, string iid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   
   //Check if transform is valid
   ChipsTransformFactory transform_factory;
   transform = transform_factory.checkValidTransform(transform);
   
   if (iid.length())
   {
      if ((rval = theDefaults->set_overrides(iid, "image")) != ERR_STATUS_GOOD)
         return rval;
   }
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
         if ((sts = cw->add_image(cid,rvals,gvals,bvals,avals,xaxis,yaxis,transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform)) == ERR_STATUS_GOOD)
         {
            cw->redraw(); 
            current = id; 
         }
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid, "", false);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform);
         if (rval == ERR_STATUS_GOOD)  cid->set_window_id(current);
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   cmd_manager->close_block();

   if (iid.length())  theDefaults->reset_defaults();
   return rval;
}



// remove an image 
int ChipsEngine::delete_image(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string iid = cid->get_image_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_image_id(iid);
         if ((sts = (*ii)->delete_image(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else 
   {
      if ((cw = get_window(id)) != NULL)   // get window 
      {
         if ((rval = cw->delete_image(cid)) == ERR_STATUS_GOOD)  
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{

      ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string iid = cid->get_image_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_image_id(iid);
         if ((sts = (*ii)->set_image_channel(cid, vals, xaxis, yaxis, channel)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else
   {
      if ((cw = get_window(id)) != NULL)   // get window
      {
         if ((rval = cw->set_image_channel(cid, vals, xaxis, yaxis, channel)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::remove_image_channel(ChipsId* cid, int channel)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string iid = cid->get_image_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_image_id(iid);
         if ((sts = (*ii)->remove_image_channel(cid, channel)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else
   {
      if ((cw = get_window(id)) != NULL)   // get window
      {
         if ((rval = cw->remove_image_channel(cid, channel)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   cmd_manager->close_block();
   return rval;
}



int ChipsEngine::add_label(ChipsId* cid,
  double xpos, double ypos, string text, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "label")) != ERR_STATUS_GOOD)
         return rval;
   }
   cmd_manager->open_block();

   if (id.length() == 0)  id = current;
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid); 
         if ((sts = cw->add_label(cid, xpos, ypos, text)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_label(cid, xpos, ypos, text)) == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_label(cid, xpos, ypos, text)) == ERR_STATUS_GOOD)
            cid->set_window_id(current); 
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::delete_label(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string lid = cid->get_label_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_label_id(lid);  // label id gets cleared downstream
         if ((sts = (*ii)->delete_label(cid)) == ERR_STATUS_GOOD) 
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_label(cid)) == ERR_STATUS_GOOD)
   {
      cw->redraw();
      current = id; 
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_line(ChipsId* cid,
    double startx, double starty, double endx, double endy, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "line")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
	 if ((sts = cw->add_line(cid, startx, starty, endx, endy)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_line(cid, startx, starty, endx, endy);
         if (rval == ERR_STATUS_GOOD)
         {
	    current = id; 
	    cw->redraw();
	 }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_line(cid, startx, starty, endx, endy);
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_line(ChipsId* cid,
   double* xx, double* yy, unsigned int count, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "line")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
	 if ((sts = cw->add_line(cid, xx, yy, count)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_line(cid, xx, yy, count)) == ERR_STATUS_GOOD)
         {
	    current = id; 
	    cw->redraw();
	 }
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_line(cid, xx, yy, count);
      else 
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::delete_line(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;

      string lid = cid->get_line_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_line_id(lid);  // line id gets cleared downstream 
	 if ((sts = cw->delete_line(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_line(cid)) == ERR_STATUS_GOOD) 
   {
      cw->redraw();
      current = id;
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_point(ChipsId* cid, double xx, double yy, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "point")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_frame_id(fid);
         if ((sts = (*ii)->add_point(cid, xx, yy)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) == NULL)
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
      else if ((rval = cw->add_point(cid, xx, yy)) == ERR_STATUS_GOOD)
      {
         current = id;
         cw->redraw();
      }
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_point(cid, xx, yy);
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}

int ChipsEngine::add_colorbar(ChipsId* cid,
   double xpos, double ypos, string obj_id, V_OBJECT_TYPE obj, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "colorbar")) != ERR_STATUS_GOOD)
         return rval;
   }
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
	 if ((sts = cw->add_colorbar(cid, xpos, ypos, obj_id, obj)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
	 if ((rval = cw->add_colorbar(cid, xpos, ypos, obj_id, obj)) == ERR_STATUS_GOOD)
         {
            current = id;
            cw->redraw();
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_colorbar(cid, xpos, ypos, obj_id, obj);
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::delete_colorbar(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ptid = cid->get_colorbar_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_colorbar_id(ptid);
	 if ((sts = (*ii)->delete_colorbar(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_colorbar(cid)) == ERR_STATUS_GOOD) 
   {
      cw->redraw();
       current = id;
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_point(ChipsId* cid,
   double xx, double yy, double zz, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "point")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
	 if ((sts = cw->add_point(cid, xx, yy, zz)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
	 if ((rval = cw->add_point(cid, xx, yy, zz)) == ERR_STATUS_GOOD)
         {
            current = id;
            cw->redraw();
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_point(cid, xx, yy, zz);
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::delete_point(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ptid = cid->get_point_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_point_id(ptid);
	 if ((sts = (*ii)->delete_point(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_point(cid)) == ERR_STATUS_GOOD) 
   {
      cw->redraw();
       current = id;
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_legend(ChipsId* cid, const char** curve_ids,
   const char** curve_labels, string lgid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int  rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->add_legend(cid, curve_ids, curve_labels, lgid)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else 
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
         rval = cw->add_legend(cid,curve_ids,curve_labels,lgid); 
         if (rval == ERR_STATUS_GOOD)
         {
            cw->redraw();
            current = id; 
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   return rval;
}


int ChipsEngine::delete_legend(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string lid = cid->get_legend_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_legend_id(lid);
         if ((sts = (*ii)->delete_legend(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_legend(cid)) == ERR_STATUS_GOOD)
   {
      current = id; 
      cw->redraw();
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::add_region(ChipsId* cid, double* xx, double* yy,
   unsigned int count, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();
   bool oldredraw;

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval=theDefaults->set_overrides(tid, "region")) != ERR_STATUS_GOOD)
         return rval;
   }

   if (id.length() == 0)  id = current;

   cmd_manager->open_block();
   cmd_manager->set_autocommit(false); 

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string fid = cid->get_frame_id();

      if (window.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
         return rval; 
      }
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->add_region(cid, xx, yy, count)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         oldredraw = cw->get_redraw();
         cw->set_redraw(false);
         if ((rval = cw->add_region(cid, xx, yy, count)) == ERR_STATUS_GOOD)
         {
            current = id;
            cw->set_redraw(oldredraw);
            cw->redraw();
         }
         else 
            cw->set_redraw(oldredraw);
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_region(cid, xx, yy, count)) == ERR_STATUS_GOOD)
            cw->redraw();
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
   if (tid.length())  theDefaults->reset_defaults();

/*   cmd_manager->set_cmd_lock(false); 
*/
   cmd_manager->set_autocommit(true); 
   if (rval != ERR_STATUS_GOOD)
      cmd_manager->clear_hold_buffer(); 
   else
      cmd_manager->commit_hold_buffer(); 
   cmd_manager->close_block();

   return rval;
}

int ChipsEngine::add_region(ChipsId* cid, int sides, double xx, double yy,
   double radius, string tid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (tid.length())
   {
      if ((rval = theDefaults->set_overrides(tid, "region")) != ERR_STATUS_GOOD)
         return rval;
   }
   if (sides < 3)   // need at least 3 sides for a valid polygon
   {
      cmd_manager->add_err(rval=ERR_INT_MIN,id,0,"sides",sides,2);
      return rval;
   }
   else if (radius <= 0)
   {
      char buf[128];
      sprintf(buf, "The region radius value (%3.2f) must be greater than 0", 
              radius);  
      cmd_manager->add_err(buf, rval=ERR_REGION_CREATE_FAIL, id, 0);
      return rval;
   }

   if (id.length() == 0)  id = current;
   cmd_manager->open_block();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string fid = cid->get_frame_id();
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cid->set_frame_id(fid);
         if ((sts = cw->add_region(cid, sides, xx, yy, radius)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->add_region(cid, sides, xx, yy, radius)) == ERR_STATUS_GOOD)
         {
            current = id;
            cw->redraw();
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else
   {
      // need to create a new window
      add_window(cid);
      cmd_manager->set_cmd_lock(true);

      if ((cw = get_window(id)) != NULL)
         rval = cw->add_region(cid, sides, xx, yy, radius);
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,"",0,"window");
      cmd_manager->set_cmd_lock(false);
   }
/*   cmd_manager->set_cmd_lock(false);
*/
   if (tid.length())  theDefaults->reset_defaults();
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::delete_region(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   int  rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   cmd_manager->open_block();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string rid = cid->get_region_id();
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cid->set_region_id(rid);
         if ((sts = (*ii)->delete_region(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(id)) == NULL)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   else if ((rval = cw->delete_region(cid)) == ERR_STATUS_GOOD) 
   {
      cw->redraw();
      current = id;
   }
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::print_window(ChipsId* cid, string outfile, string prefs)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (prefs.length())
   {
      if ((rval=theDefaults->set_overrides(prefs,"export"))!=ERR_STATUS_GOOD)
         return rval;
   }

   if ((cw = get_window(wid)) != NULL)
   {
      ChipsPrintDialog* pd = new ChipsPrintDialog();
      if (pd)
      {
	 bool use_dialog = theDefaults->get_printdialog(); 
	 rval = pd->doit(cw, "Print Window", outfile, use_dialog); 
         current = wid;
         delete pd;
      }
   }
   else if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      // Currently cannot print all windows because each
      // window would write to the same file. Refer to RFE 10078
      cmd_manager->add_err(ERR_PRINT_ALL); 
      return ERR_PRINT_ALL;
   }
   else if (wid.length())
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   else
     cmd_manager->add_err("There is currently no window to print.",rval=ERR_OBJ_VAL_DNE,"",0);

   if (prefs.length()) theDefaults->reset_defaults();

   return rval;
}


int ChipsEngine::print_image(ChipsId* cid, string outfile, string prefs)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (prefs.length())
   {
      if ((rval=theDefaults->set_overrides(prefs,"export"))!=ERR_STATUS_GOOD)
         return rval;
   }

   if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->print_image (cid, outfile);
   }
   else if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      // Currently cannot print all windows because each
      // window would write to the same file. Refer to RFE 10078
      cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,"",0,"windows");
      return ERR_PRINT_ALL;
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());

   if (prefs.length()) theDefaults->reset_defaults();

   return rval;
}


int ChipsEngine::print_frame(ChipsId* cid, string outfile)
{
   cmd_manager->add_err(ERR_FEATURE_DNE);
   return ERR_FEATURE_DNE;
}


ChipsWindow* ChipsEngine::get_window(string val)
{
   ChipsWindow *cw = NULL;
   string id = val;

   if (id.length() == 0)  id = current;
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      if (window.size() == 1)
      {
         list<ChipsWindow*>::iterator ii = window.begin();
         cw = *ii;
      }
   }
   else if (id != "")
   {
      list<ChipsWindow*>::iterator ii;
      for(ii=window.begin(); ii != window.end() && !cw; ++ii)
         if ((*ii)->get_id() == id)  cw = *ii;
   }
   return cw; 
}


int ChipsEngine::load_font(string fontname, string fontpath, 
   V_FONTSTYLE_TYPE style)
{
   int rval = ERR_STATUS_BAD;

   cmd_manager->clear_errs();
   if (fontpath.length() == 0)
     cmd_manager->add_err("No font file specified.",rval=ERR_FONT_FILE_LOAD,"",0);
   else if (verify_fontfile_exists(fontpath))
      rval = theFonts->load_font(fontname, fontpath, style); 
   else 
      cmd_manager->add_err(rval=ERR_FONT_FILE_LOAD,"",0,fontpath.c_str());

   if (!rval)
      cmd_manager->update_mask (STATE_CHNG_FONTS);
   
   return rval; 
}


char* ChipsEngine::info(ChipsId* cid, V_OBJECT_TYPE obj, bool descend, bool full)
{
   ChipsWindow* cw = NULL; 
   string temp;
   char* ptr = NULL;
   char* fptr = NULL;
   string id = cid->get_window_id();

   if (id.length() == 0)  id = current;

   temp.clear(); 

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
	 int rev = 0;
         if (obj <= chips_window)
         {
            rev = (theDefaults->get_infocurrent() &&
                       ((current == cw->get_id()) || ( current == "all")));
            temp.append("Window [");
            if (rev)  temp.append("\x1b[7m");
            temp.append(cw->get_id());
            if (rev)  temp.append("\x1b[m");
            temp.append("]\n");
         }
         if (descend || (obj > chips_window))
         {
            if ((fptr = cw->info(cid, obj, descend, rev, full)) != NULL) 
            { 
               temp.append(fptr);
               free(fptr);
               fptr = NULL;
            }
         }
      }
   }
   else if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
	 int rev = 0;
         if (obj <= chips_window)
         {
            rev = (theDefaults->get_infocurrent() &&
                       ((current == cw->get_id()) || ( current == "all")));

            temp.append("Window [");
            if (rev)  temp.append("\x1b[7m");
            temp.append(cw->get_id()); 
            if (rev)  temp.append("\x1b[m");
            temp.append("]\n"); 
         }
         if (descend || (obj > chips_window))
         {
            if ((fptr = cw->info(cid, obj, descend, rev, full)) != NULL) 
            {
               temp.append(fptr);
               free(fptr);
               fptr = NULL;
            }
         }
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char)); 
      strcpy(ptr, temp.c_str()); 
   } 
   return ptr; 
}

char* ChipsEngine::info_coordinate (ChipsId* cid)
{
   ChipsWindow* cw = NULL;
   string temp;
   char* ptr = NULL;
   char* fptr = NULL;
   string id = cid->get_window_id();

   if (id.length() == 0)  id = current;

   temp.clear();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;

	 /* Check that a coordinate system exists */
         if ((fptr = cw->info_coordinate(cid)) != NULL)
         {
	    temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            free(fptr);
            fptr = NULL;
         }
      }
   }
   else if (id != "")
   {
      // get window
      cw = get_window(id);

      if (cw)
      {
	 /* Check that a coordinate system exists */
         if ((fptr = cw->info_coordinate(cid)) != NULL)
         {
	    temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            free(fptr);
            fptr = NULL;
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}

char* ChipsEngine::info_depth(ChipsId* cid, int depth)
{
   ChipsWindow* cw = NULL;
   char *ptr = NULL, *fptr = NULL;
   string temp, id = cid->get_window_id();

   if (id.length() == 0)  id = current;

   temp.clear();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      for(list<ChipsWindow*>::iterator ii=window.begin();ii!= window.end();++ii)
      {
         cw = *ii;
         if ((fptr = cw->info_depth(cid, depth)) != NULL)
         {
            temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            delete[] fptr;
         }
      }
   }
   else if (id != "")
   {
      // get window
      if ((cw = get_window(id)) != NULL)
      {
         if ((fptr = cw->info_depth(cid, depth)) != NULL)
         {
            temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            delete[] fptr;
         }
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsEngine::info_bound_axes(ChipsId* cid)
{
   ChipsWindow* cw = NULL;
   string temp;
   char* ptr = NULL;
   char* fptr = NULL;
   string id = cid->get_window_id();

   if (id.length() == 0)  id = current;

   temp.clear();

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;

         if ((fptr = cw->info_bound_axes(cid)) != NULL)
         {
	    temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            delete [] fptr;
         }
      }
   }
   else if (id != "")
   {
      // get window
      cw = get_window(id);

      if (cw)
      {
         if ((fptr = cw->info_bound_axes(cid)) != NULL)
         {
	    temp.append("Window [");
            temp.append(cw->get_id());
            temp.append("]\n");
            temp.append(fptr);
            delete [] fptr;
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsEngine::list_current()
{
   ChipsWindow* cw = NULL; 
   string temp;
   char* ptr = NULL;
   char* fptr = NULL;

   if ((current == "all") || (current == "ALL"))
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         temp.append("Window [");
         temp.append(cw->get_id());
         temp.append("]\n");

         if ((fptr = cw->list_current()) != NULL) 
         { 
            temp.append(fptr);
            free(fptr);
            fptr = NULL;
         }
      }
   }
   else if (current != "")
   {
      // get window
      cw = get_window(current); 

      if (cw)
      {
         temp.append("Window [");
         temp.append(cw->get_id()); 
         temp.append("]\n"); 

         if ((fptr = cw->list_current()) != NULL) 
         {
            temp.append(fptr);
            free(fptr);
            fptr = NULL;
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char)); 
      strcpy(ptr, temp.c_str()); 
   } 

   return ptr; 
}

int ChipsEngine::convert_coordinate (ChipsVector &in_val, ChipsVector &out_val,
                               ChipsId *cid1, ChipsId *cid2, int axis)
{
   return this->convert_coordinate (&in_val, &out_val, 1, cid1, cid2, axis);
}

int ChipsEngine::convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, ChipsId *cid2, 
                               int axis)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid;
   string wid2;
   cmd_manager->clear_errs();

   if (num_vectors == 0)
      return ERR_STATUS_GOOD;

   if (!cid1 || !cid2)   
   {
      cmd_manager->add_err(ERR_NO_ID_OBJ);
      return ERR_NO_ID_OBJ;   // ChipsId object not specified
   }
//   if (duration)  return ERR_BLINK_ACTIVE;  // can't perform op while blinking

   ChipsId id_in (*cid1);
   ChipsId id_out (*cid2);


   wid = cid1->get_window_id();
   wid2 = cid2->get_window_id ();

   if (wid.length() == 0)  wid = current;
   if (wid2.length () == 0) wid2 = current;

   if (wid != wid2)
   {
      cmd_manager->add_err(ERR_COORD_MULT_WIN,"",0,wid.c_str(),wid2.c_str());
      return ERR_COORD_MULT_WIN;
   }

   if ((cw = get_window(wid)) != NULL)
   {
      if (id_in.get_coord_sys () == COORD_NONE)
         rval = cw->get_current_coord (&id_in);
      
      if (!rval && (id_out.get_coord_sys () == COORD_NONE))
         rval = cw->get_current_coord (&id_out);

      if (!rval)
         rval = cw->convert_coordinate (in_val, out_val, num_vectors, &id_in, &id_out, axis);
   }
   else if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      cmd_manager->add_err(ERR_USED_ALL_AS_ID);
      return ERR_USED_ALL_AS_ID;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str()); 

   if (!rval)
      cw->redraw();

   return rval;
}

bool ChipsEngine::check_coordinate_exists (ChipsId *cid)
{
   bool rval = true;
   ChipsWindow* cw;
   string wid;
   cmd_manager->clear_errs();

   if (!cid)
      return false;   // ChipsId object not specified

   wid = cid->get_window_id ();

   if (wid.length() == 0)  wid = current;

   if ((cw = get_window(wid)) != NULL)
   {
      if (cid->get_coord_sys () == COORD_NONE)
         rval = true;
      else
         rval = cw->check_coordinate_exists (cid);
   }
   else if (chips_strcmp(wid.c_str(), "all") == 0)
      rval = false; // dont support multiple windows right now
   else
      rval = false;
   return rval;
}

int ChipsEngine::get_current_coord (ChipsId *cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid;
   cmd_manager->clear_errs();

   if (!cid)
   {
      cmd_manager->add_err(ERR_NO_ID_OBJ);
      return ERR_NO_ID_OBJ;   // ChipsId object not specified
   }

   wid = cid->get_window_id ();

   if (wid.length() == 0)  wid = current;


   if ((chips_strcmp(wid.c_str(), "all") != 0) || (window.size() == 1))
   {
      if ((cw = get_window(wid)) != NULL)
         rval = cw->get_current_coord (cid);
      else
         rval = ERR_STATUS_BAD;
   }
   else
   {
      cmd_manager->add_err(ERR_USED_ALL_AS_ID);
      return ERR_USED_ALL_AS_ID;
   }
   return rval;
}




int  ChipsEngine::iconify_window(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cw->iconify();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      cw->iconify();
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::deiconify_window(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cw->deiconify();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      cw->deiconify();
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::display_grid(ChipsId* cid, int scale)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   if (wid.length() == 0)  wid = current;
   cmd_manager->clear_errs();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         rval |= cw->display_grid(scale);
         cw->redraw();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if (!cw->grid_visible() || cw->get_grid_scale() != scale)
      {
         rval = cw->display_grid(scale);
         cw->redraw();
      }
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::hide_grid(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string id = cid->get_window_id();

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if (cw->grid_visible())
         {
            rval |= cw->hide_grid();
            cw->redraw();
         }
      }
      current = id;
   }
   else if ((cw = get_window(id)) != NULL)
   {
      if (cw->grid_visible())
      {
         rval = cw->hide_grid();
         cw->redraw();
      }
      current = id;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   return rval;
}


int  ChipsEngine::display_layout(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   string wid = cid->get_window_id();
   ChipsWindow* cw;

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cw->show_layout();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      cw->show_layout();
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::hide_layout(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         cw->hide_layout();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      cw->hide_layout();
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int ChipsEngine::adjust_grid_gaps(ChipsId* cid, float xgap, float ygap)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->adjust_grid_gaps(cid,xgap,ygap)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_gaps(cid, xgap, ygap);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_xgap(ChipsId* cid, float xgap)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->adjust_grid_xgap(cid,xgap)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_xgap(cid, xgap);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_ygap(ChipsId* cid, float ygap)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->adjust_grid_ygap(cid,ygap)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_ygap(cid, ygap);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_xrelsize(ChipsId* cid, int col, float ratio)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts=(*ii)->adjust_grid_xrelsize(cid,col,ratio))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_xrelsize(cid, col, ratio);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_yrelsize(ChipsId* cid, int row, float ratio)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts=(*ii)->adjust_grid_yrelsize(cid,row,ratio))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_yrelsize(cid, row, ratio);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* ratios)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts=(*ii)->adjust_grid_xrelsizes(cid,cnt,ratios))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_xrelsizes(cid, cnt, ratios);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::adjust_grid_yrelsizes(ChipsId* cid, int cnt, float* ratios)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts=(*ii)->adjust_grid_yrelsizes(cid,cnt,ratios))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->adjust_grid_yrelsizes(cid, cnt, ratios);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::grid_objects(ChipsId* cid, int cols, int rows, float xgap,
   float ygap, int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts=(*ii)->grid_objects(cid,cols,rows,xgap,ygap,usage,
                                      lcnt,lids)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->grid_objects(cid, cols, rows, xgap, ygap, usage, 
                              lcnt, lids);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
/*   cmd_manager->set_cmd_lock(false);
*/
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::grid_objects(ChipsId* cid, int orient, int cnt, int* vals,
   float xgap, float ygap, int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->grid_objects(cid,orient,cnt,vals,xgap,ygap,usage,
                                        lcnt,lids)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->grid_objects(cid, orient, cnt, vals, xgap, ygap, usage, 
                              lcnt, lids);
      if (rval == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
/*   cmd_manager->set_cmd_lock(false);
*/
   cmd_manager->close_block();
   return rval;
}


int ChipsEngine::swap_object_positions(ChipsId* cid, string lid2)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (id.length() == 0)  id = current;
   cmd_manager->open_block();
   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->swap_object_positions(cid, lid2))
             == ERR_STATUS_GOOD)  
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
      {
         if ((rval = cw->swap_object_positions(cid, lid2)) ==
	    ERR_STATUS_GOOD)  cw->redraw();
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "window", id.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_WINDOW);
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::tile(ChipsId* cid, unsigned int cols, unsigned int rows)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->tile(cols, rows)) == ERR_STATUS_GOOD)  
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->tile(cols, rows)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::row_tile(ChipsId* cid, unsigned int rows)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();

   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->row_tile(rows)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->row_tile(rows)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::col_tile(ChipsId* cid, unsigned int cols)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->col_tile(cols)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->col_tile(cols)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::prev_frame(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->prev_frame(cid)) == ERR_STATUS_GOOD) 
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->prev_frame(cid)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::next_frame(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->next_frame(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->next_frame(cid)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::set_framemode(ChipsId* cid, int mode)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->set_framemode(mode)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->set_framemode(mode)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


char* ChipsEngine::get_current_frame(ChipsId* cid)
{
   string temp; 
   ChipsWindow* cw;
   char* ptr = NULL;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         temp.append(cw->get_current());
         temp.append(" "); 
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      temp.append(cw->get_current());
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


int  ChipsEngine::reposition_frame(ChipsId* cid,
   float x1, float y1, float x2, float y2)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts=cw->reposition_frame(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->reposition_frame(cid,x1,y1,x2,y2)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::view_single_frame(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts=cw->single_frame(cid->get_frame_id())) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->single_frame(cid->get_frame_id())) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}


int  ChipsEngine::view_placed_frame(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;
   cmd_manager->open_block();
   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->set_framemode(0)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->set_framemode(0);
      if ((rval = cw->set_framemode(0)) == ERR_STATUS_GOOD) 
      {
         cw->redraw();
         current = wid;
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   cmd_manager->close_block();
   return rval;
}



int  ChipsEngine::arrange_frames(ChipsId* cid, int cnt, string* fids)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->arrange_frames(cnt, fids)) == ERR_STATUS_GOOD)  
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval=cw->arrange_frames(cnt,fids)) == ERR_STATUS_GOOD) cw->redraw();
      current = wid;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}

/*
char* ChipsEngine::get_frame_order(ChipsId* cid)
{
   char* cval = static_cast <char *>(0); 
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      string temp;
      char* tc;
      int ll; 

      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((tc = cw->get_order()) != NULL)
         {
            temp.append(tc);
            temp.append("\n"); 
            free(tc); 
         }
      }
      current = wid;
      if ((ll = temp.length()) > 0)
      {
         cval = (char*)malloc(ll*sizeof(char));
         strncpy(cval, temp.c_str(), ll-1);
         cval[ll-1] = '\0';
      }
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      cval = cw->get_order();
      current = wid;
   }
   return cval;
}
*/

int ChipsEngine::get_frame_order(const ChipsId* cid, list<string> &out_order)
{
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();
   int rval = ERR_STATUS_GOOD;

   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;

      for(ii=window.begin(); !rval && ii != window.end(); ++ii)
      {
         cw = *ii;
         rval = cw->get_order(out_order);
         if (rval == ERR_STATUS_GOOD)
            out_order.push_back("");
         else
            out_order.clear ();
      }
      current = wid;
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      rval = cw->get_order(out_order);
      current = wid;
   }
   return rval;
}




int ChipsEngine::blink_frames(ChipsId* cid, string fid2, float duration)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();

   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->blink_frames(cid, fid2, duration)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
         rval = cw->blink_frames(cid, fid2, duration);
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
   return rval;
}

int ChipsEngine::stop_blink(ChipsId* cid)
{
   ChipsWindow *cw = NULL;
   string id = cid->get_window_id();
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();

   if (id.length() == 0)  id = current;

   if (chips_strcmp(id.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         cw = *ii;
         if ((sts = cw->stop_blink(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         cw->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
   }
   else if (id != "")
   {
      if ((cw = get_window(id)) != NULL)
         rval = cw->stop_blink(cid);
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
   }
   else  cmd_manager->add_err(rval = ERR_NO_CUR_WINDOW);
   return rval;
}

int ChipsEngine::delete_all_windows(ChipsId* cid, bool force_del)
{
   int rval;
   cmd_manager->clear_errs();
   if (force_del)
      cmd_manager->clear_cmds (); 
   if (cid)
   {
      cid->set_window_id("all");
      rval = delete_window(cid, force_del);
   }
   else
   {
      ChipsId tmp;
      tmp.set_window_id("all");
      rval = delete_window(&tmp, force_del);
   }
/*   if (force_del)
      cmd_manager->clear_cmds ();
*/
   window_cnt = 0;
   return rval;
}


int ChipsEngine::delete_all_frames(ChipsId* cid)
{
   cmd_manager->clear_errs();
   cid->set_frame_id("all"); 
   return  delete_frame(cid);
}


int ChipsEngine::delete_all_plots(ChipsId* cid)
{
   cmd_manager->clear_errs();
   cid->set_plot_id("all"); 
   return  delete_plot(cid);
}


int ChipsEngine::delete_all_images(ChipsId* cid)
{
   cmd_manager->clear_errs();
   cid->set_image_id("all"); 
   return  delete_image(cid);
}



ChipsWindow* ChipsEngine::new_window(string window_id)
{
   ChipsWindow* cw = NULL;
   cmd_manager->clear_errs();

   if (window_id.length() == 0)
   {
      string tmpid = create_window_id();
      cw =   new ChipsWindow(this, tmpid);
   }
   else
      cw =   new ChipsWindow(this, window_id);

   return cw;
}


ChipsWindow* ChipsEngine::new_window(ChipsStateStream& in)
{
   cmd_manager->clear_errs();
   return new ChipsWindow(this, in);
}



ChipsWindow* ChipsEngine::new_window(float width, float height,
   string format,  string window_id)
{
   ChipsWindow* cw = NULL;

   cmd_manager->clear_errs();
   if (window_id.length() == 0)
   {
      string tmpid = create_window_id();
      cw =   new ChipsWindow(this, width, height, format, tmpid);
   }
   else 
      cw =   new ChipsWindow(this, width, height,format, window_id);
   return cw;
}


char* ChipsEngine::create_window_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_windowstem();
      else            stem = "win";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++window_cnt);

   // increment count til next unused window id
   while(get_window(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++window_cnt);
   }
   return temp_id;
}

bool ChipsEngine::verify_fontfile_exists(string fontpath)
{
   return false; 
}


int ChipsEngine::undo(int depth)
{
   // if undo commands are buffered, erase them before undoing commands on the stack
   cmd_manager->clear_errs();
   if (!cmd_manager->get_autocommit())
      discard_undo_buffer();
   return cmd_manager->unexecute_cmd(depth);
}

int ChipsEngine::undo(string tag)
{
   // if undo commands are buffered, erase them before undoing commands on the stack
   cmd_manager->clear_errs();
   if (!cmd_manager->get_autocommit())
      discard_undo_buffer();
   return cmd_manager->unexecute_cmd(tag);
}

int ChipsEngine::redo(int depth)
{
   cmd_manager->clear_errs();
   return cmd_manager->execute_cmd(depth);
}

int ChipsEngine::redo(string tag)
{
   cmd_manager->clear_errs();
   return cmd_manager->execute_cmd(tag);
}

int ChipsEngine::set_undo_tag(string tag)
{
   cmd_manager->clear_errs();
   // don't allow setting an undo tag w/in a buffer 
   if (!cmd_manager->get_autocommit())  
   {
      cmd_manager->add_err(ERR_UNDO_BUFFER_TAG);
      return ERR_UNDO_BUFFER_TAG;
   }
   return cmd_manager->add_tag(tag);
}

int ChipsEngine::clear_undo_tag(string tag)
{
   cmd_manager->clear_errs();
   return cmd_manager->remove_tag(tag);
}

int ChipsEngine::open_undo_block()
{
   cmd_manager->clear_errs();
   if (cmd_manager)
   {
      cmd_manager->open_block();
      return ERR_STATUS_GOOD; 
   }
   return ERR_STATUS_BAD;
}

int ChipsEngine::close_undo_block()
{
   cmd_manager->clear_errs();
   if (cmd_manager)
   {
      cmd_manager->close_block();
      return ERR_STATUS_GOOD; 
   }
   return ERR_STATUS_BAD; 
}

int ChipsEngine::open_undo_buffer()
{
   cmd_manager->clear_errs();
   buffer_depth++; // don't want to close prematurely- nesting workaround

   disable_redraw();

   // treat everything in undo buffer as a single block
   cmd_manager->open_block();

   // force undo commands to goto 'hold' buffer
   cmd_manager->set_autocommit(false);
   return ERR_STATUS_GOOD;
}


int ChipsEngine::close_undo_buffer()
{
   int rval;
   ChipsWindow* cw;
   cmd_manager->clear_errs();

   // if we had multiple open_undo_buffer calls, don't close unless we 
   // correspond to the final close_undo_buffer call
   if (buffer_depth > 1)
   {
      // close out the block
      cmd_manager->close_block();
      buffer_depth--;
      return ERR_STATUS_GOOD;
   }
   else
      buffer_depth = 0;
  
   // stop using the hold buffer
   cmd_manager->set_autocommit(true);

   enable_redraw(); 

   rval = cmd_manager->commit_hold_buffer();    // store off changes

   // close out the block
   cmd_manager->close_block();

   // refresh 'current' window so that the changes are visible
   if (current == "all" || ((cw = get_window(current)) != NULL))
   {
      redraw_window ();
   }
   return rval; 
}


int ChipsEngine::discard_undo_buffer()
{
   int rval; 
   cmd_manager->clear_errs();
   // if we had multiple open_undo_buffer calls, don't proceed unless we 
   // correspond to the final discard_undo_buffer call
   if (buffer_depth > 1)
   {
      buffer_depth--;
      cmd_manager->close_block();
      return ERR_STATUS_GOOD;
   }
   else
      buffer_depth = 0;

   // discard the undo/redo commands in the buffer 
   rval =  cmd_manager->clear_hold_buffer();

   enable_redraw();
   cmd_manager->set_autocommit(true);

   // close out the block
   cmd_manager->close_block();

   return rval;
}

int ChipsEngine::reference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int retVal = ERR_STATUS_BAD;
   if (co && co->get_object_type () == chips_window){
      ChipsWindow *cw =  static_cast <ChipsWindow *>(co);
      list <ChipsWindow *>::iterator win_ii = window.begin ();

      for (; (win_ii != window.end ()) && ((*win_ii)->get_window_num () < cw->get_window_num ()); win_ii++);
      window.insert (win_ii, cw);
     
      dwindow.erase (co->get_ref_cnt ()); 

      // If a current string is passed in use it. It is possible that
      // if a delete all occured then undone that all is current. Otherwise
      // use referenced frame
      if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_window(curr)))
         current = curr;
      else
         current = co->get_id ();
      if (change_cnt)
         window_cnt++;
      retVal = ERR_STATUS_GOOD;
   }
   return retVal;
}

int ChipsEngine::dereference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int retVal = ERR_STATUS_BAD;
   if (co && co->get_object_type () == chips_window){

      window.remove (static_cast <ChipsWindow *>(co));
      dwindow[co->get_ref_cnt ()] = static_cast <ChipsWindow *>(co);

      if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_window(curr)))
         current = curr; 
      else if (!window.empty())
         current =  window.back()->get_id ();
      else
         current = "";
      
      if (change_cnt)
         window_cnt--;
      retVal = ERR_STATUS_GOOD;
   }
   return retVal;
}

int ChipsEngine::clear_cmd (ChipsObject *co)
{
   cmd_manager->clear_errs();
   if (!co || co->get_object_type () != chips_window)
      return ERR_STATUS_BAD;

   map <unsigned int, ChipsWindow *>::iterator ii;
   ii = dwindow.find (co->get_ref_cnt ());
   if (ii != dwindow.end ())
   {
      ii->second->object_delete ();
      dwindow.erase (ii);
   }
   return ERR_STATUS_GOOD;
}



int  ChipsEngine::display_depth(ChipsId* cid, int depth)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   if (wid.length() == 0)  wid = current;
   cmd_manager->clear_errs();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->display_depth(cid, depth)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->display_depth(cid, depth)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}

int  ChipsEngine::hide_depth(ChipsId* cid, int depth)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   if (wid.length() == 0)  wid = current;
   cmd_manager->clear_errs();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts = (*ii)->hide_depth(cid, depth)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->hide_depth(cid, depth)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int  ChipsEngine::collapse_depths(ChipsId* cid, int* srcs, int scnt, int dest)
{
   int rval = ERR_STATUS_GOOD;   
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   if (wid.length() == 0)  wid = current;
   cmd_manager->clear_errs();

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts= (*ii)->collapse_depths(cid, srcs, scnt, dest)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->collapse_depths(cid, srcs, scnt, dest)) == 
          ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}


int ChipsEngine::rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner)
{
   int rval = ERR_STATUS_GOOD;
   ChipsWindow* cw;
   string wid = cid->get_window_id();
   cmd_manager->clear_errs();
   if (wid.length() == 0)  wid = current;

   if (chips_strcmp(wid.c_str(), "all") == 0)
   {
      list<ChipsWindow*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=window.begin(); ii != window.end(); ++ii)
      {
         if ((sts= (*ii)->rearrange_depth(cid, obj, manner)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         (*ii)->redraw();
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"windows");
      else
         current = "all";
   }
   else if ((cw = get_window(wid)) != NULL)
   {
      if ((rval = cw->rearrange_depth(cid, obj, manner)) == ERR_STATUS_GOOD)
      {
         current = wid;
         cw->redraw();
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",wid.c_str());
   return rval;
}

char* ChipsEngine::get_errors(int mode)
{
   string val;
   char* ptr = NULL;

   if (cmd_manager)
      val = cmd_manager->get_errs(mode);
   else
      val.clear();

   int ll = val.length() + 1;
   ptr = (char*)malloc(ll*sizeof(char));
   if (ll > 1)
      strcpy(ptr, val.c_str());
   else 
      ptr[0]='\0';
   return ptr;
}

int ChipsEngine::add_client_error(int errcode, string errtext)
{
   if (cmd_manager)
   {
      cmd_manager->clear_errs();
      cmd_manager->add_err(errtext, errcode,"",0);
   }
   return ERR_STATUS_GOOD;
}


int ChipsEngine::get_error_count(int type, int &out_val)
{
   int rval = ERR_STATUS_GOOD;
   if (!cmd_manager)
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"command manager"); 
   else
      out_val = cmd_manager->get_err_cnt(type);
   return rval;
}

int ChipsEngine::set_error_verbosity(int val)
{

   if (val < 0)
   {
      if (cmd_manager) cmd_manager->add_err(ERR_INT_MIN,"",0,"verbosity",val,0);
       return ERR_INT_MIN;
   }
   if (cmd_manager)   cmd_manager->set_err_verbosity(val);

   return ERR_STATUS_GOOD;
}

int ChipsEngine::clear_errors()
{
   if (cmd_manager)   cmd_manager->clear_errs();
   return ERR_STATUS_GOOD;
}

int ChipsEngine::get_error_verbosity(int &out_val)
{
   int rval = ERR_STATUS_GOOD;
   if (!cmd_manager)
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"command manager"); 
   else
      out_val = cmd_manager->get_err_verbosity();
   return rval;
}

int ChipsEngine::get_error_types(int &out_val)
{
   int rval = ERR_STATUS_GOOD;
   if (!cmd_manager)
      cmd_manager->add_err(rval=ERR_OBJ_DNE,"",0,"command manager");
   else
      out_val = cmd_manager->get_err_types();
   return rval;
}

int ChipsEngine::load_fill(string in_file, int slot)
{
   int rval = ERR_STATUS_GOOD; 
   cmd_manager->clear_errs();
   if (!fill_patterns)  
      cmd_manager->add_err(rval=ERR_FILL_TABLE_ACCESS);
   else if (!in_file.length()) 
      cmd_manager->add_err("No fill file specified.",rval=ERR_FILL_LOAD,"",0);
   if (rval != ERR_STATUS_GOOD)  return rval;
   return fill_patterns->load_file(in_file, slot);
}

int ChipsEngine::load_colormap(long len, double* r, double* g, 
	double* b, double* alpha, int slot)
{
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (!colormap_table)
      cmd_manager->add_err(rval=ERR_STATUS_BAD);
   if (rval != ERR_STATUS_GOOD) return rval;

   return colormap_table->read_array_list(len, static_cast<V_COLORMAP_TYPE>(slot), 
	r, g, b, alpha);
}

int ChipsEngine::load_colormap(string in_file, int slot)
{
   int rval = ERR_STATUS_GOOD;

   cmd_manager->clear_errs();
   if (!colormap_table)  
      cmd_manager->add_err(rval=ERR_STATUS_BAD);
   if (!in_file.length()) 
      cmd_manager->add_err("No colormap file specified.",rval=ERR_CMAP_LOAD,"",0);
   if (rval != ERR_STATUS_GOOD)  return rval;

   return colormap_table->load_file(in_file, static_cast<V_COLORMAP_TYPE>(slot));
}

int ChipsEngine::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj)
{
   ChipsWindow *cw = NULL;
   int  cnt, rval = ERR_STATUS_GOOD;
   string id = cid->get_window_id();
   list<ChipsWindow*>::iterator ii;

   if (id.length() == 0)  id = current;

   cmd_manager->clear_errs();

   // since highlighting takes several seconds, verify objects exist first
   get_object_count(cid, obj, cnt);
   if (cnt == 0)
   {
      string tval = "Unable to find the requested ";
      const char* obval = v_object_to_string(obj); 
      if (obval != NULL)  tval.append(obval);
         else             tval.append("object");
      tval.append(" to highlight");
      cmd_manager->add_err(tval, rval=ERR_OBJ_DNE);
      return rval; 
   }


   // get window
   if ((cw = get_window(id)) == NULL)
   {
      if (chips_strcmp(id.c_str(), "all") != 0)
      {
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
         return rval;
      }
   }

   ChipsId* tmp_cid = new ChipsId (*cid);
   if (obj == chips_window)
      tmp_cid->set_frame_id ("all");

   // cycle through highlights
   bool vis = false;
   for (cnt = 0; cnt < 6 && rval == ERR_STATUS_GOOD; cnt++)
   {
      // first hide targeted objects
      if (cw)
      {
         // single frame
         if (obj == chips_window)
            rval = cw->highlight_object(tmp_cid, chips_frame, (int)vis);
         else
            rval = cw->highlight_object(tmp_cid, obj, (int)vis);
      }
      else
      {
         for(ii=window.begin(); ii != window.end(); ++ii)
         {
            cw = *ii;
            if (obj == chips_window)
               rval = cw->highlight_object(tmp_cid, chips_frame, (int)vis);
            else
               rval = cw->highlight_object(tmp_cid, obj, (int)vis);
         }
         cw = NULL;
      }
      usleep(100000);   // wait
      vis = !vis;
   }
   delete tmp_cid; 

   return rval;
}

void ChipsEngine::set_ignore_state(bool value)
{
   if (cmd_manager)  cmd_manager->set_ignore_state(value);
}

bool ChipsEngine::get_ignore_state()
{
   if (cmd_manager)  return cmd_manager->get_ignore_state();
   else return false;
}

int ChipsEngine::get_windows (const ChipsId *cid, list <ChipsWindow *>&out_windows, bool allow_all)
{
   int  rval = ERR_STATUS_GOOD;
   ChipsWindow *cw;
   string id = cid->get_window_id();
   if (id.length() == 0)  id = current;

   if (rval == ERR_STATUS_GOOD)
   {
      if (window.empty())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"windows");
      else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
      {
         // get window
         if ((cw = get_window(id)) != NULL)
            out_windows.push_back (cw);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"window",id.c_str());
      }
      else if (allow_all)
      {
         list<ChipsWindow*>::iterator ii;

         for(ii=window.begin(); ii != window.end(); ++ii)
            out_windows.push_back (*ii);
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      }
      else
         cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
   }
   else
      cmd_manager->add_err(rval, id);
   return rval;
}


void ChipsEngine::update_default_display_colors()
{
   // iterate through existing windows are reload colors
   string fg = theDefaults->get_defaultfgdisp(),
          bg = theDefaults->get_defaultbgdisp();
   list<ChipsWindow*>::iterator ii;

   for(ii=window.begin(); ii != window.end(); ++ii)
   {
      (*ii)->set_default_colors_display(bg, fg);
      (*ii)->redraw();
   }
   cmd_manager->update_mask(STATE_CHNG_PROP_DEF); 
}
