// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <stdio.h>
#include <math.h>

#include "ChipsFrame.hh"
#include "ChipsDefaults.hh"
#include "ChipsWindow.hh"
#include "ChipsPlot.hh"
#include "ChipsImage.hh"
#include "ChipsLabel.hh"
#include "ChipsCmdManager.hh"
#include "ChipsStateStream.hh"
#include "ChipsRepoCommand.hh"
#include "ChipsAdjustRelsizeCommand.hh"
#include "ChipsDepthArrangeCommand.hh"
#include "ChipsDepthUpdateCommand.hh"
#include "ChipsDepthVisCommand.hh"
#include "ChipsDepthCollapseCommand.hh"
#include "ChipsDeleteCommand.hh"
#include "ChipsPlotSystem.hh"


extern ChipsDefaults* theDefaults;
extern ChipsCmdManager* cmd_manager;

// static members 
ChipsFrame::ChipsFrame(ChipsObject* par, string val):ChipsObject(par, val)
{
   id = val; 
   nx1 = ny1 = 0.0; 
   nx2 = ny2 = 1.0; 
   initialize(); 
}


ChipsFrame::ChipsFrame(ChipsObject* par, ChipsStateStream& in):ChipsObject(par, "")
{
   initialize(false);

   deserialize(in);

   if (deserialize_coord)
      coord = new ChipsFrameSystem (in, NULL);
   else 
   {
      if (in.get_minor_version() < 7)
         coord = parent->get_chips_coord()->get_coord (FRAME_NORM, id);
      else 
         coord = parent->get_chips_coord()->get_coord (FRAME_NORM, this->data_system_id);
   }

   if (coord)
      coord->store_local (true);
}


ChipsFrame::ChipsFrame(ChipsObject* par, float x1n, float y1n, float x2n, 
   float y2n, string val): ChipsObject(par, val)
{
   id = val;
   nx1 = x1n; 
   ny1 = y1n; 
   nx2 = x2n; 
   ny2 = y2n; 
   initialize(); 
}

void ChipsFrame::initialize( bool create_coord)
{
   curr_plot = "";
   plot_cnt = 0;
   spcnt = spcols = sprows = 0;
   if (theDefaults)
   {
      string temp = theDefaults->get_framecolor();
      set_color(temp); 
      temp = theDefaults->get_framebordercolor();
      set_border_color(temp); 
      display_border = theDefaults->get_framebordervisible();
      preserve_aspect = theDefaults->get_frameaspect();
      scale = theDefaults->get_framescale();
      opaque = !theDefaults->get_frametransparent();
      leftmargin = theDefaults->get_plotleftmargin();
      bottommargin = theDefaults->get_plotbottommargin();
      rightmargin = theDefaults->get_plotrightmargin();
      topmargin = theDefaults->get_plottopmargin();
      def_depth = theDefaults->get_defaultdepth(); 
   }
   else
   {
      set_color("default");
      set_border_color("default");
      display_border = false;
      preserve_aspect = true; 
      opaque = true; 
      scale = true; 
      leftmargin = rightmargin = topmargin = bottommargin = 0.1; 
      def_depth = V_DEFAULT_DEPTH;
   }
   object_type = chips_frame;
   if (create_coord)
   {
      coord = new ChipsFrameSystem (parent->get_chips_coord());
      data_system_id = coord->get_id ();
   }
   coord_system = WINDOW_NORM; 
   // This forces any object the is in either frame or window coordinate
   // systems to be store in frame coordinate system. This is set to true
   // here because currently there are no objects that aren't really attached
   // to the frame coordinate system. Even if they were created by the
   // user in PIXEL or WINDOW_NORM. Objects created in PLOT or DATA coordinates
   // will be store in those coordinate systems.
   if (create_coord)
      coord->store_local (true); 
   visible = true;
   bind_table = new ChipsAxisBind();
   scale_factor = 1.0; 
   init_width = ((ChipsWindow *)parent)->get_width();
   init_height = ((ChipsWindow *)parent)->get_height();
   spplot = NULL; 
   sparr = NULL;
   spxrel = spyrel = spleft = spright = sptop = spbottom = NULL; 
   yrelsum = xrelsum = 1; 
   deserialize_coord = false;
}

ChipsFrame::~ChipsFrame()
{
   if (spleft) delete []spleft;
   if (spbottom) delete []spbottom;
   if (spright) delete []spright;
   if (sptop) delete []sptop;
   if (spxrel) delete []spxrel;
   if (spyrel) delete []spyrel;
   if (spplot) delete []spplot;
   if (sparr) delete []sparr;

   if (coord)
      delete coord;
   if (bind_table)
      delete bind_table;
   if (depth_table)
      delete depth_table;
}

void ChipsFrame::object_delete ()
{
   ChipsId cid;
   cid.set_plot_id ("all");
   if (!(plot.empty() && dplot.empty())) 
      delete_plot (&cid, true);
   delete this;
}



int ChipsFrame::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int  sts, pass = 0, failval = 0, rval = ERR_STATUS_GOOD;

   if (depth_table)  depth_table->reset_addition_count(NULL);

   if ((obj == chips_window) || (obj == chips_frame))
   {
      // set frame properties
      int mask = theDefaults->get_allpropmask();
 
      if ((mask & PREF_ALL_BGCOLOR) == PREF_ALL_BGCOLOR)
      {
         // set background (frame) color 
         string clr = theDefaults->get_allbgcolor();
         sts = set_property(NULL, (void*)clr.c_str(), SET_FRAME_COLOR);
         if (sts == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      else if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         // set foreground (borders, etc...) colors  
         string clr = theDefaults->get_allcolor();
	 sts = set_property(NULL, (void*)clr.c_str(), SET_FRAME_BORDER_COLOR);
         if (sts == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      else if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         sts = set_property(NULL, static_cast<void*>(&vis), SET_FRAME_VISIBLE); 
         if (sts == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
         sts = set_property(NULL,static_cast<void*>(&vis),SET_FRAME_BORDER_VIS);
         if (sts == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      else if ((mask & PREF_ALL_ASPECT) == PREF_ALL_ASPECT)
      {
         keep_aspect(theDefaults->get_allaspect());
      }

      // set all objects in heirarchy
      map<string, ChipsPlot*>::iterator hh;
      for(hh=plot.begin(); hh != plot.end(); ++hh)
         if ((sts = (*hh).second->set_cascading_property(cid, obj)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
   }
   else 
   {
      string plid = cid->get_plot_id();
      if (plid.length() == 0)  plid = curr_plot;

      bind_table->reset_access_list();
      if (!chips_strcmp(plid.c_str(), "all"))
      {
         // object heirarchy starts lower
         map <string, ChipsPlot*>::iterator hh;
         for(hh=plot.begin(); hh != plot.end(); ++hh)
         {
            bind_table->set_access_plot((*hh).second);
            if ((sts = (*hh).second->set_cascading_property(cid, obj)) ==
                 ERR_STATUS_GOOD)
                pass++;
             else if (!err_type_object_DNE(sts))
                failval = sts;
         }
      }
      else if (plot.count(plid))
      {
         bind_table->set_access_plot(plot[plid]);
         if ((sts = (plot[plid])->set_cascading_property(cid, obj)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      else
         cmd_manager->add_err(failval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
      bind_table->reset_access_list();
   }
   if (depth_table)   depth_table->sort_check();
   cmd_manager->clear_errs(CLEAR_DNE_ERRS);
   if (failval)
      rval = failval;
   else if (!pass)
      cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frame");
   return rval;
}


int ChipsFrame::set_property(ChipsId* cid, void* in_val, int in_property,
   bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   bool set_frame_prop = true;

   if (depth_table)  depth_table->reset_addition_count(NULL);

   switch(in_property)
   {
      case SET_FRAME_COLOR:
      {
         string clr = (const char*)in_val;

         if (rec)  cmd = new ChipsSetCommand(this, in_property,
                    get_color(), clr, ((ChipsWindow *)parent)->get_current());
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
         }
         else 
            delete cmd; 
      }
      break; 
      case SET_FRAME_BORDER_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
              get_border_color(), clr, ((ChipsWindow *)parent)->get_current());
         if ((rval = set_border_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
         }
         else
            delete cmd;
      }
      break;
      case SET_FRAME_BORDER_VIS:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)display_border; 
	 if (rec)  cmd = new ChipsSetCommand(this, in_property,
              oval, *ival, ((ChipsWindow *)parent)->get_current());
         if ((rval = set_border_visibility((*ival == 1))) == ERR_STATUS_GOOD)
         {
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
         }
         else
            delete cmd;
      }
      break; 
      case SET_FRAME_VISIBLE:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = (int)visible;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
              oval, *ival, ((ChipsWindow *)parent)->get_current());
         if ((*ival) == 1)
         {
            show();
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
	 }
         else if ((*ival) == 0)
         {
            hide();
            if (rec)
            {
               if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
         else
         {
            cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"frame visible",
                                 *ival, "0:1");
            delete cmd;
         }
      }
      break;
      case SET_FRAME_TRANSPARENT:
      {
         bool* bval = static_cast<bool*>(in_val);

         if (rec)  cmd = new ChipsSetCommand(this, in_property,
              get_transparency(), *bval,((ChipsWindow *)parent)->get_current());
         if ((rval = set_transparency(*bval)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
         }
         else
            delete cmd;
      }
      break; 
      case SET_FRAME_SCALE:
      {
         bool* bval = static_cast<bool*>(in_val);
	 bool oval = scale; 
         scale = *bval;
	 if (rec)
	 {
	    cmd = new ChipsSetCommand(this, in_property,
              oval, *bval,((ChipsWindow *)parent)->get_current());
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
      }
      break;
      case SET_DEFAULT_DEPTH:
      {
         if (depth_table)
         {
            int* ival = static_cast<int*>(in_val);
            int oval = depth_table->get_default_depth();
            if (rec)  cmd = new ChipsSetCommand(this, in_property,
                 oval, *ival, ((ChipsWindow *)parent)->get_current());
            if ((rval = depth_table->set_default_depth(*ival)) == ERR_STATUS_GOOD)
            {
               if (rec)
               {
                  if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
               }
            }
            else
            {
               if (*ival == 0)  
                  cmd_manager->add_err(rval=ERR_DEF_DEPTH,id);
               else 
                  cmd_manager->add_err(rval=ERR_INT_MIN, id, 0, "default depth", *ival, 0);
               delete cmd;
            }
         }
         else  cmd_manager->add_err(rval=ERR_DEPTH_TABLE_DNE,id,0);
      }
      break; 
      default:        // all other objects are handled in plot
      {
         set_frame_prop = false;

         // record any changes in the axes
         bind_table->record_updates (true);
         bind_table->reset_access_list();

         string plid = cid->get_plot_id();
	 if (plid.length() == 0)  plid = curr_plot; 
         if (plot.empty())
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"plots");
	 else if (!chips_strcmp(plid.c_str(), "all"))
	 {
            bind_table->reset_access_list();
            map <string, ChipsPlot*>::iterator hh;
            int sts, pass = 0, failval = 0; 
            ChipsId* tcid;
            for(hh=plot.begin(); hh != plot.end(); ++hh)
            {
               bind_table->set_access_plot((*hh).second);
               tcid = new ChipsId(*cid);
               if ((sts=(*hh).second->set_property(tcid, in_val, in_property))==
                   ERR_STATUS_GOOD)
                  pass++;
               else if (!err_type_object_DNE(sts))
                        failval = sts;
               delete tcid;
            }
            cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
               rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else 
               this->set_current_plot (plid);
            bind_table->reset_access_list();
         }
         else if (plot.count(plid))
         {
            bind_table->reset_access_list(plot[plid]);
            if ((rval = (plot[plid])->set_property(cid, in_val, in_property)) ==
                ERR_STATUS_GOOD)
               set_current_plot (plid); 
            bind_table->reset_access_list();
         }
	 else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
         // update the limits across plots
         if (!rval)
            check_plot_updates ();
         bind_table->record_updates (false);
         bind_table->reset_access_list();
      }
      break; 
   }
   if (depth_table)   depth_table->sort_check();
   
   // need to add frame id for errors
   if (cmd_manager && rval)   cmd_manager->add_frame_id_to_errs(id); 

   if (cmd_manager && !rval && set_frame_prop)
      cmd_manager->update_mask (STATE_CHNG_PROP_FRM);

   return rval;
}


int ChipsFrame::get_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_FRAME_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_FRAME_BORDER_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_border_color();
      }
      break;
      case SET_FRAME_BORDER_VIS:
      {
         int *ival = (int*)out_val;
         *ival = get_border_visibility(); 
      }
      break; 
      case SET_FRAME_VISIBLE:
      {
         int *ival = (int*)out_val;
         *ival = is_visible();
      }
      break; 
      case SET_FRAME_TRANSPARENT:
      {
         int *ival = (int*)out_val;
         *ival = get_transparency(); 
      }
      break; 
      case SET_FRAME_SCALE:
      {
         int *ival = (int*)out_val;
         *ival = scale; 
      }
      break; 
      case SET_DEFAULT_DEPTH:
      {
         int *ival = (int*)out_val;
         if (depth_table)
            *ival = depth_table->get_default_depth();
         else
            *ival = V_USE_DEFAULT_DEPTH; 
      }
      break;
      default:     // all other objects handled in plot
      {
	 string plid = cid->get_plot_id();
	 if (plid.length() == 0)   plid = curr_plot; 
         if (plot.empty())
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"plots");
         else if (chips_strcmp(plid.c_str(), "all") == 0)
         {
            ChipsPlot* cp = get_plot(plid); 
            if (plot.size() == 1 && (cp != NULL))
               rval = plot[cp->get_id()]->get_property(cid, out_val, property);
            else
               cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
         }
         else if (plot.count(plid))
	    rval = (plot[plid])->get_property(cid, out_val, property);
	 else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
      }
      break;
   }
   return rval;
}


int ChipsFrame::add_plot(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   string plid = theDefaults->get_plotid();
   ChipsPlot* tmpplot = NULL; 
   if (plid.length() && plot.count(plid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,plid.c_str(), "plot");
   else
   {
      tmpplot = new_plot(plid);
      // if (in_split)   tmpplot = new_plot(def_x1, def_y1, def_x2, def_y2, plid);

      if (tmpplot)
      {
         bool cmd_lock = cmd_manager->cmd_lock ();
         if (!cmd_lock)
         {
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpplot, (plid.length()==0),curr_plot);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         cmd_manager->set_cmd_lock (true);

         tmpplot->load_depth_table(depth_table);
         tmpplot->post_initialize ();

         plot.insert(pair<string, ChipsPlot*>(tmpplot->get_id (), tmpplot));
         set_current_plot (tmpplot->get_id());
         cid->set_plot_id(curr_plot);
         string axid = tmpplot->get_current_xaxis_id();
         cid->set_xaxis_id(axid);
         axid = tmpplot->get_current_yaxis_id();
         cid->set_yaxis_id(axid);
         cid->set_coord_sys (PLOT_NORM);
         cid->set_coord_id ("");

         cmd_manager->set_cmd_lock (cmd_lock);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"plot");
   }
   return rval;
}


int ChipsFrame::add_plot(ChipsId* cid, float px1, float py1,
                        float px2, float py2)
{
   int rval = ERR_STATUS_GOOD;
   string plid = theDefaults->get_plotid();
   ChipsPlot* tmpplot = NULL; 

   if (plid.length() && (get_plot(plid) != NULL))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,plid.c_str(), "plot");
   else
   {
      if ((tmpplot = new_plot(px1, py1, px2, py2, plid)) != NULL)
      {
         bool cmd_lock = cmd_manager->cmd_lock ();
         if (!cmd_lock)
         {
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpplot, (plid.length()==0),curr_plot);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         cmd_manager->set_cmd_lock (true);

         tmpplot->load_depth_table(depth_table);
         tmpplot->post_initialize ();

         plot.insert(pair<string, ChipsPlot*>(tmpplot->get_id (), tmpplot));
         set_current_plot (tmpplot->get_id());
         cid->set_plot_id(curr_plot);
         string axid = tmpplot->get_current_xaxis_id();
         cid->set_xaxis_id(axid);
         axid = tmpplot->get_current_yaxis_id();
         cid->set_yaxis_id(axid);
         cid->set_coord_sys (PLOT_NORM);
         cid->set_coord_id ("");

         cmd_manager->set_cmd_lock (cmd_lock);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"plot");
   }
   return rval;
}


int ChipsFrame::limits(ChipsId* cid, int dim, double xmin, double xmax, 
   double ymin, double ymax)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map <string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts = (*ii).second->limits(cid, dim, xmin, xmax, ymin, ymax)) ==
	     ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      rval = (plot[plid])->limits(cid, dim, xmin, xmax, ymin, ymax);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots 
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();
   
   return rval;
}


int ChipsFrame::limits(ChipsId* cid, int dim, double min, double max)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map <string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second); 
         if ((sts = (*ii).second->limits(cid, dim, min, max)) == ERR_STATUS_GOOD) 
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]); 
      rval = (plot[plid])->limits(cid, dim, min, max);
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   V_AUTOLIMIT_TYPE automax)
{
   int rval = ERR_STATUS_GOOD; 
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts = (*ii).second->limits(cid, dim, automin, automax)) ==  
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]); 
      rval = (plot[plid])->limits(cid, dim, automin, automax);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE automax)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second); 
         if ((sts = (*ii).second->limits(cid, dim, min, automax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]); 
      rval = (plot[plid])->limits(cid, dim, min, automax);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   double max)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts=(*ii).second->limits(cid, dim, automin, max)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      rval = (plot[plid])->limits(cid, dim, automin, max);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::zoom(ChipsId* cid, float scale, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   bool cmd_lock = cmd_manager->cmd_lock ();
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   if (!rec)
      cmd_manager->set_cmd_lock (true);

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts=(*ii).second->zoom(cid, scale)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      rval = (plot[plid])->zoom(cid, scale);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   if (!rec)
      cmd_manager->set_cmd_lock (cmd_lock);

   return rval;
}

int ChipsFrame::panto(ChipsId* cid, double x, double y)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts=(*ii).second->panto(cid, x, y)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;

      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      rval = (plot[plid])->panto(cid, x, y);
   }
   else if (plid.length())
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::pan(ChipsId* cid, double x1, double y1, double x2, double y2, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   bool cmd_lock = cmd_manager->cmd_lock ();
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   if (!rec)
      cmd_manager->set_cmd_lock (true);

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts=(*ii).second->pan(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }

      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      rval = (plot[plid])->pan(cid, x1, y1, x2, y2);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   if (!rec)
      cmd_manager->set_cmd_lock (cmd_lock);

   return rval;
}




int ChipsFrame::delete_plot(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;

   ChipsPlot* cplot;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)   plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;

      if (plot.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
         return rval;
      }

      // delete all plots
      if (!force_del)
         cmd_manager->open_block ();

      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cplot = (*ii).second;
         cplot->unselect ();
         if (force_del)
            cplot->object_delete ();
         else{
            cplot->remove_plot_from_depth_table();
            this->dereference_plot (cplot);

            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cplot,curr_plot);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }
      if (!force_del) 
         cmd_manager->close_block ();
      else
      {
         // If force deleteing all plots also remove previously delete plots
         map <unsigned int, ChipsPlot*>::iterator   dplot_ii;
         for (dplot_ii = dplot.begin (); dplot_ii != dplot.end (); dplot_ii++)
            dplot_ii->second->object_delete ();
         dplot.clear ();
      }
      plot.clear();
      curr_plot = "";
      cid->set_plot_id("");
      rval = ERR_STATUS_GOOD;
      plot_cnt = 0;
   }
   else if (plot.count(plid))
   {
      map<string, ChipsPlot*>::iterator ii;
      ii = plot.find(plid);
      cplot = (*ii).second;
      cplot->unselect ();
      if (depth_table)
      {
         if (!force_del)
            cplot->remove_plot_from_depth_table();
         else 
            depth_table->remove_object(cplot);
      }
      plot.erase(ii);
      if (force_del)
         cplot->object_delete ();
      else
      {
         cmd_manager->open_block ();
         this->dereference_plot (cplot);

         if (!cmd_manager->cmd_lock ())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cplot,curr_plot);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         cmd_manager->close_block ();
      }
      cid->set_plot_id("");
      if (!plot.empty())
      {
         map<string, ChipsPlot* >::reverse_iterator ri = plot.rbegin();
         set_current_plot ((*ri).first);
      }
      else   set_current_plot ("");

      rval = ERR_STATUS_GOOD;
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}

int ChipsFrame::reposition_plot_area(ChipsPlot* plt, float px1, float py1,
   float px2, float py2, bool offscreen, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   float llx, lly, urx, ury;
   ChipsRepoCommand *cmd = NULL;

   if (!plt)
   {  
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot","");
      return ERR_OBJ_VAL_DNE;
   }

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();
   
   if (rec || offscreen)
   {
      llx = plt->get_plot_lmargin();
      lly = plt->get_plot_bmargin();
      urx = 1 - plt->get_plot_rmargin();
      ury = 1 - plt->get_plot_tmargin();
   }

   if (rec)
   {
      cmd = new ChipsRepoCommand (plt, llx, lly, urx, ury, 
                                  px1, py1, px2, py2, curr_plot);
   }
   if ((rval = plt->reposition(px1, py1, px2, py2)) == ERR_STATUS_GOOD)
   {
      if (rec)
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      if (offscreen)
         plt->move_offscreen(rec);
      else 
         plt->move_onscreen(rec);
   }
   else  delete cmd;

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::reposition_plot(const ChipsId* cid, float px1, float py1,
   float px2, float py2, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   string plid = "";
   if (cid)  plid = cid->get_plot_id();

   if (!plid.length())  plid = curr_plot;


   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map <string, ChipsPlot*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         if ((sts =reposition_plot_area((*ii).second,px1,py1,px2,py2,false,rec)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
   }
   else if (plot.count(plid))
      rval = reposition_plot_area(plot[plid], px1, py1, px2, py2, false, rec);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   return rval;
}


int ChipsFrame::add_curve(ChipsId* cid, long cnt, double* x, double* y,
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      ChipsId* tcid;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         tcid = new ChipsId(*cid);
         rval = (*ii).second->add_curve(tcid, cnt, x, y, ydown, yup, xdown, xup,
                                 ylims, xlims, transform);
         delete tcid;
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup, ylims,
                           xlims, transform);
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_curve(cid, cnt, x, y, ydown, yup, xdown, 
          xup, ylims, xlims, transform)) == ERR_STATUS_GOOD)
         set_current_plot (plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::delete_curve(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }

   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_curve_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_curve_id(cvid);
         if ((sts =(*ii).second->delete_curve(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ("all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_curve(cid);
      set_current_plot (plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}


int ChipsFrame::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int numlev, double* levels, Transform *trans)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      ChipsId* tcid;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         tcid = new ChipsId(*cid);
         rval = (*ii).second->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
         delete tcid;
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      if ((rval = (plot[plid])->add_contour(cid,cnt,x,y,z,numlev,levels,trans)) ==
           ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval)   cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      ChipsId* tcid;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         tcid = new ChipsId(*cid);
         rval = (*ii).second->add_contour(tcid, dat, xdim, ydim, numlev, levels, trans);
         delete tcid;
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      if (add_plot(cid) == ERR_STATUS_GOOD)
      {
         cmd_manager->set_cmd_lock (true);   
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_contour(cid, dat, xdim, ydim, numlev, levels, trans);
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      bind_table->set_access_plot(plot[plid]);
      if ((rval = plot[plid]->add_contour(cid,dat,xdim,ydim,numlev,levels,trans)) ==
          ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval)   cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::delete_contour(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map <string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)    plid = curr_plot;

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         rval = (*ii).second->delete_contour(cid);
         if (rval != ERR_STATUS_GOOD) break;
      }
      if (rval == ERR_STATUS_GOOD)  set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_contour(cid);
      set_current_plot (plid);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}



int ChipsFrame::set_contour_levels(ChipsId* cid, unsigned int arraysize, 
   double* pos)
{
   int rval; 
   string plid = cid->get_plot_id();
   if (plid.length() == 0)    plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      for(map<string,ChipsPlot*>::iterator ii=plot.begin();ii!=plot.end();++ii)
      {
         if ((sts = (*ii).second->set_contour_levels(cid,arraysize,pos)) ==
             ERR_STATUS_GOOD)
           pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->set_contour_levels(cid, arraysize, pos);
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);
   }
   if (rval)   cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   int rval;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)    plid = curr_plot;

   if (plot.empty())
     cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"plots");
   else if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      ChipsPlot* cp = get_plot(plid);
      if (cp) 
         rval = cp->get_contour_levels(cid, pos);
      else 
         cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   }
   else if (plot.count(plid))
      rval = (plot[plid])->get_contour_levels(cid, pos);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   return rval;
}



int ChipsFrame:: add_histogram(ChipsId* cid, long cnt, double* bin_low,
   double* bin_high, double* counts, double* yd, double* yu)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      ChipsId* tcid;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         tcid = new ChipsId(*cid);
         rval = (*ii).second->add_histogram(tcid,cnt,bin_low,bin_high,counts,yd,yu);
         delete tcid;
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      ChipsPlot* cp;
      if (add_plot(cid) == ERR_STATUS_GOOD)
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)
            rval = cp->add_histogram(cid, cnt,bin_low, bin_high,counts, yd, yu);
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->add_histogram(cid,cnt,bin_low,bin_high,counts,yd,yu);
      if (rval == ERR_STATUS_GOOD)  set_current_plot (plid);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval)   cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::delete_histogram(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)    plid = curr_plot;

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         if ((rval = (*ii).second->delete_histogram(cid)) != ERR_STATUS_GOOD)
             break;
      if (rval == ERR_STATUS_GOOD)  set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      rval = plot[plid]->delete_histogram(cid);
      set_current_plot (plid);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}


int ChipsFrame::add_axis(ChipsId* cid, int dim, float pos,
   float minx, float maxx, float miny, float maxy, Transform *transform)
{
   string plid = cid->get_plot_id();
   COORD_SYS csys = cid->get_coord_sys();
   string    csid = cid->get_coord_id();
   int rval = ERR_STATUS_GOOD;
   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      for(map<string, ChipsPlot*>::iterator ii=plot.begin();ii!=plot.end();++ii)
      {
	 // reset coords so the correct coord sys used for all additions
	 cid->set_coord_sys(csys);
	 cid->set_coord_id(csid);

         rval = (*ii).second->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      ChipsPlot* cp; 
      if (add_plot(cid) == ERR_STATUS_GOOD)
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)
            rval = cp->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      rval = plot[plid]->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);


      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);
   }
   if (rval)   cmd_manager->add_frame_id_to_errs(id);
   return rval; 
}


int ChipsFrame::delete_axis(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();
   string axid = cid->get_xaxis_id();
   string ayid = cid->get_yaxis_id();

   if (plid.length() == 0)    plid = curr_plot;

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         if (axid.length() && (ayid.length() == 0))
            rval = (*ii).second->delete_xaxis(cid);
         else if (ayid.length() && (axid.length() == 0))
            rval = (*ii).second->delete_yaxis(cid);
         else
            rval = (*ii).second->delete_axis(cid);
         if (rval != ERR_STATUS_GOOD) break;
      }
      if (rval == ERR_STATUS_GOOD)  set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      if (axid.length() && (ayid.length() == 0))
         rval = plot[plid]->delete_xaxis(cid);
      else if (ayid.length() && (axid.length() == 0))
         rval = plot[plid]->delete_yaxis(cid);
      else
         rval = plot[plid]->delete_axis(cid);
      set_current_plot (plid);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}


int ChipsFrame::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)    plid = curr_plot;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         if ((sts = (*ii).second->set_arbitrary_tick_positions(cid,arraysize,
             pos,vals)) == ERR_STATUS_GOOD)
           pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current_plot ("all");
   }
   else if (plot.count(plid))
   {
      rval = plot[plid]->set_arbitrary_tick_positions(cid, arraysize, pos,vals);
      set_current_plot (plid);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::bind_axes(ChipsId* cid, const string plid, const string aid,
   bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsPlot *pl1, *pl2;
   ChipsAxis *ax1 = NULL, *ax2 = NULL;
   string pid1 = cid->get_plot_id();
   string pid2 = plid;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (pid1.length() == 0)   pid1 = curr_plot;
   if (pid2.length() == 0)   pid2 = curr_plot;

   if (((pl1 = get_plot(pid1))!= NULL) && ((pl2 = get_plot(pid2))!= NULL))
   {
      string aid1 = cid->get_axis_id();

      // if the axis id is not set check the xaxis id and the axis id. Order
      // of precidence is x then y
      if (aid1.length () == 0)
      {
         aid1 = cid->get_xaxis_id ();
         if (aid1.length() == 0)
            aid1 = cid->get_yaxis_id ();
      }

      if ((ax1 = pl1->get_xaxis(aid1)) == NULL)  ax1 = pl1->get_yaxis(aid1);  
      if ((ax2 = pl2->get_xaxis(aid)) == NULL)  ax2 = pl2->get_yaxis(aid);  
      if ((rval = bind_table->check_valid_binding(ax1, ax2)) == ERR_STATUS_GOOD)
      {
         if (ax1 != ax2) 
         {
            ChipsBindCommand *cmd = NULL;
            double* dp = NULL;
            int*    ip = NULL;
            string* sp = NULL; 
            bind_table->compile_synch_data(ax2, &dp, &ip, &sp);
            if (rec)  cmd = new ChipsBindCommand(this, pid1, aid1, pid2, aid,
                                                 dp, ip, sp,true);
            if (dp) delete[] dp; 
            if (ip) delete[] ip; 
            if (sp) delete[] sp; 

            pl1->remove_box_axes(ax1);
            pl2->remove_box_axes(ax2);

            if ((rval = bind_table->add_binding(ax1, ax2)) == ERR_STATUS_GOOD)
            {
               // disable the rec option in synch_axes call since the bind
               // command above already handles it 
               bind_table->synch_axes(ax1, ax2, 0); 
               if (rec)
                  if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
            }
            else
            {
               if (rval == ERR_DUP_BINDING)
                  cmd_manager->add_err(rval, id, 0, pid1.c_str(), aid1.c_str(),
                                       pid2.c_str(), aid.c_str());
               else if (rval == ERR_CROSSDIM_BIND)
                  cmd_manager->add_err(rval, id, 0, aid1.c_str(), aid.c_str());
               else if (rval == ERR_CYC_DEP_BIND)
                  cmd_manager->add_err(rval, id, 0);
               delete cmd;
            }
         }
         else cmd_manager->add_err(rval=ERR_SELF_BIND,id,0, pid1.c_str(),
	                           aid1.c_str());
      }
      else if (!ax1)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis",aid1.c_str ());
      else if (!ax2)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis",aid.c_str ());
   }
   else if (pl1)
       cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",pid2.c_str());
   else if (pl2)
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",pid1.c_str());
   else if ((chips_strcmp(pid1.c_str(),"all")==0) ||
            (chips_strcmp(pid2.c_str(),"all"))==0)
      cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"plots");

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   if (pl2 && ax2 && (rval == ERR_STATUS_GOOD))
   {
      ChipsId mycid;
      mycid.set_plot_id(pl2->get_id());
      mycid.set_axis_id(ax2->get_id());
      if (ax2->get_scaletype () == V_SCALE_LOG)
         pl2->set_dimension_property(&mycid, ax2->get_orientation(), SET_PLOT_LOGSCALE, rec);
      else
         pl2->set_dimension_property(&mycid, ax2->get_orientation(), SET_PLOT_LINEARSCALE, rec);
   }

   return rval;
}

int ChipsFrame::unbind_axes(ChipsId* cid, const string plid, const string aid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsPlot *pl1, *pl2;
   string pid1 = cid->get_plot_id();
   string pid2 = plid;
   ChipsBindCommand *cmd = NULL;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (pid1.length() == 0)   pid1 = curr_plot;
   if (pid2.length() == 0)   pid2 = curr_plot;

   if (((pl1 = get_plot(pid1))!= NULL) && ((pl2 = get_plot(pid2))!= NULL))
   {
      ChipsAxis *ax1 = NULL, *ax2 = NULL;
      string aid1 = cid->get_axis_id();
      if ((ax1 = pl1->get_xaxis(aid1)) == NULL)  ax1 = pl1->get_yaxis(aid1);  
      if ((ax2 = pl2->get_xaxis(aid)) == NULL)  ax2 = pl2->get_yaxis(aid);  

      if (ax1 && ax2 && (ax1 != ax2)) 
      {
         if (rec)
         {
            double* dp = NULL;
            int*    ip = NULL;
            string* sp = NULL;
            bind_table->compile_synch_data(ax2, &dp, &ip, &sp);
            cmd = new ChipsBindCommand(this, pl1->get_id (), ax1->get_id (),
               pl2->get_id (), ax2->get_id (), dp, ip, sp, false);
            if (dp) delete[] dp;
            if (ip) delete[] ip;
            if (sp) delete[] sp;
         }

         // remove binding 
         bind_table->delete_binding(ax1, ax2);
         if (rec && (rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis");
   }
   else if (pl1)
       cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",pid2.c_str());
   else if (pl2) 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",pid1.c_str());
   else if ((chips_strcmp(pid1.c_str(),"all")==0) ||
            (chips_strcmp(pid2.c_str(),"all"))==0)
      cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"plots");

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::format_axis_point (const ChipsId* cid, double pnt, string &formatted_pnt)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;

   if (plid.length() == 0)   plid = curr_plot;
   if (!chips_strcmp(plid.c_str(), "all") && (plot.size() != 1))
   {
      cmd_manager->add_err(ERR_OBJ_INVALID_ALL,id,0,"plot");
      return ERR_OBJ_INVALID_ALL;
   }
   if (plid == "")
   {
       cmd_manager->add_err("No current frame to perform operation on",ERR_OBJ_DNE, id);
       return ERR_OBJ_DNE;
   }
   if (plot.count(plid) == 0)
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "plot", plid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   cp = get_plot(plid);
   if ((rval = cp->format_axis_point(cid, pnt, formatted_pnt)) != ERR_STATUS_GOOD)
      cmd_manager->add_frame_id_to_errs(id);
   return rval;
}


int ChipsFrame::add_image(ChipsId* cid, string filename, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->reset_plot_id(""); 
         rval = (*ii).second->add_image(cid, filename, transform);
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_image(&tcid, filename, transform);
         if (rval == ERR_STATUS_GOOD)
            cid->set_image_id (tcid.get_image_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_image(cid, filename, transform)
          ) == ERR_STATUS_GOOD)
         set_current_plot (plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::add_image(ChipsId* cid, string filename, 
   float ix1, float iy1, float ix2, float iy2, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->reset_plot_id("");
         rval = (*ii).second->add_image(cid, filename, ix1, iy1, ix2, iy2, transform);
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_image(&tcid, filename, ix1, iy1, ix2, iy2, transform);
         if (rval == ERR_STATUS_GOOD)
            cid->set_image_id (tcid.get_image_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_image(cid, filename, ix1, iy1, ix2, iy2, transform)
          ) == ERR_STATUS_GOOD)
         set_current_plot (plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::add_image(ChipsId* cid, double* rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         bind_table->set_access_plot((*ii).second);
         cid->reset_plot_id("");
         rval = (*ii).second->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform);
      }
      if (rval == ERR_STATUS_GOOD)   set_current_plot (plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_image(&tcid, rvals, gvals, bvals, avals, xaxis, yaxis, transform);
         if (rval == ERR_STATUS_GOOD)
            cid->set_image_id (tcid.get_image_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform)
          ) == ERR_STATUS_GOOD)
      {
         set_current_plot (plid);  
         bind_table->set_access_plot(plot[plid]);
      }
         
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::delete_image(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_image_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_image_id(cvid);
         if ((sts =(*ii).second->delete_image(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ("all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_image(cid);
      set_current_plot (plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}

int ChipsFrame::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_image_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_image_id(cvid);
         if ((sts =(*ii).second->set_image_channel(cid, vals, xaxis, yaxis, channel)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ("all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->set_image_channel(cid, vals, xaxis, yaxis, channel);
      set_current_plot (plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;

}

int ChipsFrame::remove_image_channel(ChipsId* cid, int channel)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_image_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_image_id(cvid);
         if ((sts =(*ii).second->remove_image_channel(cid, channel)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->remove_image_channel(cid, channel);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;

}

int ChipsFrame::print_image (ChipsId* cid, const string &filename, int fmt)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      ChipsPlot* cp = get_plot(plid);
      if (cp)
      {
         rval = cp->print_image(cid, filename, fmt);
         set_current_plot ( plid);
      }
      else 
         cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"plots");
   }   
   else if (plot.count(plid))
   {
      rval = (plot[plid])->print_image(cid, filename, fmt);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}

int ChipsFrame::add_label(ChipsId* cid, double xpos, double ypos, string text)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_label(cid, xpos, ypos, text);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_label(&tcid, xpos, ypos, text);
         if (rval == ERR_STATUS_GOOD)
            cid->set_label_id (tcid.get_label_id ());   
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_label(cid, xpos, ypos, text)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::delete_label(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_label_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_label_id(cvid);
         if ((sts =(*ii).second->delete_label(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_label(cid);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}


int ChipsFrame::add_line(ChipsId* cid, double startx, double starty,
   double endx, double endy)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_line(cid, startx, starty, endx, endy);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_line(&tcid, startx, starty, endx, endy);
         if (rval == ERR_STATUS_GOOD)
            cid->set_line_id (tcid.get_line_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_line(cid, startx, starty, endx, endy)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::add_line(ChipsId* cid, double* xx, double* yy,
                         unsigned int count)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_line(cid, xx, yy, count);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_line(&tcid, xx, yy, count);
         if (rval == ERR_STATUS_GOOD)
            cid->set_line_id (tcid.get_line_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_line(cid, xx, yy, count)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::delete_line(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_line_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_line_id(cvid);
         if ((sts =(*ii).second->delete_line(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_line(cid);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}

int ChipsFrame::add_colorbar(ChipsId* cid, double xpos, double ypos,
    string obj_id, V_OBJECT_TYPE mapobj)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_colorbar(cid, xpos, ypos, obj_id, mapobj);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_colorbar(&tcid, xpos, ypos, obj_id, mapobj);
         if (rval == ERR_STATUS_GOOD)
            cid->set_colorbar_id (tcid.get_colorbar_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_colorbar(cid, xpos, ypos, obj_id, mapobj)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::delete_colorbar(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_colorbar_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_colorbar_id(cvid);
         if ((sts =(*ii).second->delete_colorbar(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_colorbar(cid);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}

int ChipsFrame::add_point(ChipsId* cid, double xx, double yy)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_point(cid, xx, yy);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_point(&tcid, xx, yy);
         if (rval == ERR_STATUS_GOOD)
            cid->set_point_id (tcid.get_point_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_point(cid, xx, yy)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::add_point(ChipsId* cid, double xx, double yy, double zz)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_point(cid, xx, yy, zz);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_point(&tcid, xx, yy, zz);
         if (rval == ERR_STATUS_GOOD)
            cid->set_point_id (tcid.get_point_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_point(cid, xx, yy, zz)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::delete_point(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_point_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_point_id(cvid);
         if ((sts =(*ii).second->delete_point(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_point(cid);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}



int ChipsFrame::add_legend(ChipsId* cid,  const char** curve_ids,
   const char** curve_labels, string lgid)
{
   return ERR_FEATURE_DNE;
}


int ChipsFrame::delete_legend(ChipsId* cid)
{
   return ERR_FEATURE_DNE;
}


int ChipsFrame::add_region(ChipsId* cid, double* xx, double* yy,
                          unsigned int count)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_region(cid, xx, yy, count);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_region(&tcid, xx, yy, count);
         if (rval == ERR_STATUS_GOOD)
            cid->set_region_id (tcid.get_region_id ());
         cmd_manager->set_cmd_lock (cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
     if ((rval = (plot[plid])->add_region(cid, xx, yy, count))==ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::add_region(ChipsId* cid, int sides, double xx, double yy,
                           double radius)
{
   int rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   ChipsPlot* cp;    

   if (plid.length() == 0)   plid = curr_plot;
   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         rval = (*ii).second->add_region(cid, sides, xx, yy, radius);
      if (rval == ERR_STATUS_GOOD)   set_current_plot ( plid);  
   }
   else if (plid == "")  // plot doesn't exist
   {
      bool cmd_lock = cmd_manager->cmd_lock();

      // use a temporary id to maintain the coordinate system
      ChipsId tcid (*cid);

      if (add_plot(cid) == ERR_STATUS_GOOD)   
      {
         cmd_manager->set_cmd_lock (true);
         if ((cp = get_plot(curr_plot)) != NULL)   
            rval = cp->add_region(&tcid, sides, xx, yy, radius);
         if (rval == ERR_STATUS_GOOD)
            cid->set_region_id (tcid.get_region_id ());
         cmd_manager->set_cmd_lock(cmd_lock);
      }
   }
   else if (plot.count(plid))
   {
      if ((rval = (plot[plid])->add_region(cid, sides, xx, yy, radius)
          ) == ERR_STATUS_GOOD)
         set_current_plot ( plid);  
   }
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   if (rval == ERR_STATUS_GOOD && depth_table)   depth_table->sort_check();
   if (rval) cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::delete_region(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator ii;
   string plid = cid->get_plot_id();

   if (plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      return rval;
   }
   if (plid.length() == 0)    plid = curr_plot;

   if (chips_strcmp(plid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_region_id();
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         cid->set_region_id(cvid);
         if ((sts =(*ii).second->delete_region(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      if (failval)
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         rval = failval;
      }
      else
      {
         cmd_manager->clear_errs();
         if (!pass)cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
            else  set_current_plot ( "all");
      }
   }
   else if (plot.count(plid))
   {
      rval = (plot[plid])->delete_region(cid);
      set_current_plot ( plid);
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}



int ChipsFrame::convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, ChipsId *cid2, 
                               int axis)
{
   int rval = ERR_STATUS_GOOD;

   ChipsCoordinate *coord1=NULL;
   ChipsCoordinate *coord2=NULL;
   
   // This will have to be revisited when we go to 3D
   // The static size of 3 wont work.
   ChipsVector *tmp_in = in_val;
   ChipsVector *tmp_out = out_val;
   ChipsVector *single_axis_in = NULL;
   ChipsVector *single_axis_out = NULL;

   int pos = -1;

   // Handle single values by creating a vector and populating
   // the correct axis
   if ((axis == X_AXIS) || (axis == Y_AXIS))
   {
      single_axis_in = new ChipsVector [num_vectors];
      single_axis_out = new ChipsVector [num_vectors];
      pos = (axis == X_AXIS) ? 0 : 1;

      for (int ii=num_vectors; ii--;)
      {
         single_axis_in[ii] = ChipsVector (3);
         single_axis_in[ii][pos] = in_val[ii][0];
         single_axis_out[ii] = ChipsVector (3);
      }

      tmp_in = single_axis_in;
      tmp_out = single_axis_out;
   }
   else if (axis == XY_AXIS)
   {
      if (in_val[0].get_size () != 3)
         cmd_manager->add_err(rval=ERR_BAD_COORD_DIM_SZ,id,0);
   }
   else 
      cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",axis,
                           "X_AXIS|Y_AXIS|XY_AXIS");

   if (!cid1 || !cid2)  cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);

   // Find the correct coordinate system for each chips id
   if (!rval && ((cid1->get_coord_sys () == DATA) || 
       (cid1->get_coord_sys () == PLOT_NORM)))
   {
      ChipsId tmp_id (*cid1);
      coord1 = find_coordinate_system (tmp_id, rval);

      if (coord1)
         rval = coord1->convert_cs_to_cs (tmp_in, tmp_out, num_vectors, 
                                      tmp_id.get_coord_sys (),
                                      tmp_id.get_coord_id (), FRAME_NORM, "");
   }
   else if (!rval)
   {
      rval = coord->convert_cs_to_cs (tmp_in, tmp_out, num_vectors, 
                                 cid1->get_coord_sys (), "" , FRAME_NORM, "");
   }

   if (!rval && ((cid2->get_coord_sys () == DATA) ||
       (cid2->get_coord_sys () == PLOT_NORM)))
   {
      ChipsId tmp_id (*cid2);
      coord2 = find_coordinate_system (tmp_id, rval);
      if (coord2)
         rval = coord2->convert_cs_to_cs (tmp_out, tmp_out, num_vectors,
                                      FRAME_NORM, "", tmp_id.get_coord_sys (),
                                      tmp_id.get_coord_id ()); 
      else if (rval == ERR_OBJ_DNE)
      {
         // coord conversion failed since plot doesn't exist
         cmd_manager->add_err(rval=ERR_COORD_DNE,id,0,
            v_coord_sys_to_string(tmp_id.get_coord_sys()));
      }
   }
   else if (!rval && (cid2->get_coord_sys () != FRAME_NORM))
      rval = coord->convert_cs_to_cs (tmp_out, tmp_out, num_vectors,
                                 FRAME_NORM, "", cid2->get_coord_sys (), "");

   // Copy new value to output argument
   if (single_axis_out)
   {
      for (int ii=num_vectors; ii--;)
         out_val[ii][0] = single_axis_out[ii][pos];
      delete []single_axis_out;
      single_axis_in = NULL;
   }

   if (single_axis_in)
   {
      delete []single_axis_in;
      single_axis_in = NULL;
   }

   return rval;
}


int ChipsFrame::get_current_coord (ChipsId *cid)
{
   ChipsPlot *plt;
   if (!cid) 
   {
      cmd_manager->add_err(ERR_NO_ID_OBJ,id,0);
      return ERR_NO_ID_OBJ;
   }
   string oid = cid->get_plot_id ();
   if (oid.length () == 0) oid = curr_plot;

   // If the length is zero then the current coordinate is here
   if (oid.length () == 0){
      cid->set_coord_sys (FRAME_NORM);
      cid->set_coord_id ("");
   }
   // Find the current coordinate of the plot
   else{
      if (chips_strcmp(oid.c_str(), "all") != 0 || plot.size() == 1)
      {
         if ((plt = get_plot(oid)) != NULL)
            return plt->update_current_coord (cid);
      }
      else 
      {
         cmd_manager->add_err(ERR_USED_ALL_AS_ID,id,0);
         return ERR_USED_ALL_AS_ID;
      }
   }
   return ERR_STATUS_GOOD;
}


char* ChipsFrame::info(ChipsId* cid, V_OBJECT_TYPE obj, bool descend,
   bool current_obj, bool full)
{
   char* ptr = NULL;
   string temp, val;
   map <int, string>chron;
   char buffer[80];
   int  infcur   = theDefaults->get_infocurrent(); 

   if (!plot.empty() && ((obj == chips_plot) || (obj <= chips_frame)))
   {
      map<string, ChipsPlot*>::iterator ii;
      map <int, string>chron;
      ChipsPlot* cpt;

      // list all plot objects in frame
      for(ii = plot.begin(); ii != plot.end(); ++ii)
         chron.insert(pair<int,string>((*ii).second->get_ref_cnt(),(*ii).first));
      for(map <int, string>::iterator i2= chron.begin(); i2 != chron.end();i2++)
      {
         cpt = plot[(*i2).second];
         int rev = (infcur && current_obj && ((curr_plot == cpt->get_id()) || 
                    chips_strcmp(curr_plot.c_str(),"all") == 0));
         temp.append("    Plot [");
         if (rev)  temp.append("\x1b[7m");
         temp.append(cpt->get_id());
         if (rev)  temp.append("\x1b[m");
         if (!cpt->is_offscreen())
         {
            temp.append("]");
            if (full)
            {
	       sprintf(buffer, "   (%3.2f,%3.2f)  .. (%3.2f,%3.2f)", 
	           cpt->get_plot_lmargin(), cpt->get_plot_bmargin(),
	           1-cpt->get_plot_rmargin(), 1-cpt->get_plot_tmargin());
               temp.append(buffer);
            }
            temp.append("\n");
            if (descend || (obj > chips_plot))
            {
               // list out objects in plot
               char* plt_stuff = cpt->info(cid, obj, rev);
      
               if (plt_stuff)
               {
                  temp.append(plt_stuff);
                  free(plt_stuff);
               }
            }
         }
         else
         {
            temp.append("]");
            if (full)  temp.append("   (OFFSCREEN)");
            temp.append("\n");
         }
      }
   }

   // add in the annotations list- ordered chronologically
   map <int, string>::iterator i2; 
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      temp.append((*i2).second);
   chron.clear();

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsFrame::info_depth(ChipsId* cid, int depth)
{
   char *ptr = NULL, *deps;
   string temp = "";

   if (depth_table)
   {
      if ((deps = depth_table->list_depths(depth)) != NULL)
      {
         temp = deps;
         delete[] deps;
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = new char[ll];
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsFrame::info_coordinate(ChipsId* cid)
{
   char* ptr = NULL;
   string temp;
   map <int, string>chron;
   if (!plot.empty())
   {
      map<string, ChipsPlot*>::iterator ii;
      map <int, string>chron;
      ChipsPlot* cpt;

      // list all plot objects in frame
      for(ii = plot.begin(); ii != plot.end(); ++ii)
         chron.insert(pair<int,string>((*ii).second->get_ref_cnt(),(*ii).first));
      for(map <int, string>::iterator i2= chron.begin(); i2 != chron.end();i2++)
      {
         cpt = plot[(*i2).second];
         temp.append("    Plot [");
         temp.append(cpt->get_id());
         temp.append("]\n");

         if (cpt)
         {
            char* plt_stuff  = cpt->info_coordinate ();
            if (plt_stuff)
            {
               temp.append(plt_stuff);
               free(plt_stuff);
            }
         }
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsFrame::info_bound_axes(ChipsId* cid)
{
  ChipsPlot *pl = NULL;
   char* ptr = NULL;
   string temp;
   string lid = cid->get_plot_id();
   string aid = cid->get_axis_id();
   if (!aid.length())  aid = cid->get_xaxis_id();
   if (!aid.length())  aid = cid->get_yaxis_id();

   if (lid.length() == 0)    lid = curr_plot;

   temp.clear();

   if (!chips_strcmp(lid.c_str(), "all"))
   {
      if (aid.length() && chips_strcmp(aid.c_str(), "all"))
      {
         ChipsAxis* ax = pl->get_xaxis(aid);
         if (!ax)  ax = pl->get_yaxis(aid);
         if (ax)   temp.append(bind_table->get_bindings_list(ax));
      }
      else
         temp.append(bind_table->get_bindings_list(NULL));
   }
   else if (plot.count(lid))
   {
      pl = plot[lid];
      if (aid.length() && chips_strcmp(lid.c_str(), "all"))
      {
         ChipsAxis* ax = pl->get_xaxis(aid);
         if (!ax)  ax = pl->get_yaxis(aid);
         if (ax)   temp = bind_table->get_bindings_list(ax);
      }
      else
         temp = bind_table->get_bindings_list(NULL);
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = new char[ll];
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsFrame::list_current()
{
   char* ptr = NULL;
   string temp;
   ChipsId tmp_id;
   map <int, string>chron;

   int rval;

   if (curr_plot != "")
   {
      ChipsPlot* cp;
      char *pobjs = NULL;
      if (chips_strcmp(curr_plot.c_str(), "all") == 0)
      {
         map<string, ChipsPlot*>::iterator ii;
         for(ii = plot.begin(); ii != plot.end(); ++ii)
            chron.insert(pair<int,string>((*ii).second->get_ref_cnt(),
                         (*ii).first));
         for(map <int, string>::iterator i2=chron.begin();i2!=chron.end();i2++)
         {
            cp = plot[(*i2).second];
            temp.append("    Plot [");
            temp.append(cp->get_id());
            temp.append("]\n");
            if ((pobjs = cp->list_current()) != NULL)
            {
               temp.append(pobjs);
               free(pobjs);
            }
         }
         chron.clear();
      }
      else
      {
         if ((cp = get_plot(curr_plot)) != NULL)
         {
            temp.append("    Plot [");
            temp.append(curr_plot);
            temp.append("]\n");
            if ((pobjs = cp->list_current()) != NULL)
            {
               temp.append(pobjs);
               free(pobjs);
            }
         }
      }
   }
   rval = get_current_coord (&tmp_id);

   if (!rval)
   {
       switch (tmp_id.get_coord_sys ())
       {
          case FRAME_NORM:
             temp.append("    Coord Sys [Frame Normalized]\n");
          break;
          case PLOT_NORM:
             temp.append("    Coord Sys [Plot Normalized]\n");
          break;
          case DATA:
             temp.append("    Coord Sys [Data]\n");
             temp.append("    Coord Sys ID [");
             temp.append(tmp_id.get_coord_id ());
             temp.append("]\n");
          break;
          default: 
	     // do nothing 
	  break; 
      }
   }
   
   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


int ChipsFrame::set_display_location(int mode, int px1, int py1, int px2, 
   int py2)
{
   int  rval = ERR_STATUS_GOOD; 

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   switch(mode)
   {
      case 0:  // placed
      {
         bbx1 = (int)((px2 - px1) * nx1 + px1+.5); 
         bby1 = (int)((py2 - py1) * ny1 + py1+.5); 
         bbx2 = (int)((px2 - px1) * nx2 + px1+.5); 
         bby2 = (int)((py2 - py1) * ny2 + py1+.5); 
      }
      break;
      case 1:  // single      // fallthrough intended 
      case 2:  // tiled
      {
         if (keep_aspect())
         {
            float xs = nx2 - nx1;
            float ys = ny2 - ny1;
            float dx = px2 - px1;
            float dy = py2 - py1;
            float offset; 

            if ((dx  != 0) && (dy != 0) && (ys != 0))
            {
               if ((xs/dx) > (ys/dy)) 
               {
                  offset = (py2 - py1 - (dx * ys/xs))/2; 
                  bbx1 = px1; 
                  bbx2 = px2;
                  bby1 = (int)(py1 + offset);
                  bby2 = (int)(py2 - offset);
               } 
               else
               {
                  offset = (px2 - px1 - (dy * xs/ys))/2; 
                  bby1 = py1; 
                  bby2 = py2;
                  bbx1 = (int)(px1 + offset);
                  bbx2 = (int)(px2 - offset);
               }
            }
         }
         else   // fill whole display area
         {
            bbx1 = px1;
            bby1 = py1;
            bbx2 = px2;
            bby2 = py2; 
         } 
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_INT_MINMAX,id,0,"frame mode",mode,0,2);
      break;
   }

   // Calculate the actual normalized coordinates. These are effected
   //   by frame modes unlike nx1, ny1, etc.
   float ww = ((ChipsWindow *)parent)->get_width();
   float hh = ((ChipsWindow *)parent)->get_height();
   bb_norm_x1 = (bbx1)/(double)ww;
   bb_norm_y1 = (bby1)/(double)hh;
   bb_norm_x2 = (bbx2)/(double)ww;
   bb_norm_y2 = (bby2)/(double)hh;


   if (scale)
   {
      double delbbnx = bb_norm_x2 - bb_norm_x1; 
      double deltax= (nx2 - nx1);
      double delbbny = bb_norm_y2 - bb_norm_y1; 
      double deltay= (ny2 - ny1);
      deltax /= (ww/init_width);
      deltay /= (hh/init_height);
      if (delbbnx)  deltax/= delbbnx; 
         else deltax = 1; 
      if (delbbny)  deltay/= delbbny; 
         else deltay = 1; 
      if ((scale_factor = sqrt((deltax*deltax + deltay*deltay))) != 0)
         scale_factor = sqrt(2.0) / scale_factor; 
      else
         scale_factor = 1; 
   }
   else  scale_factor = 1; 
 
   ChipsVector min_pos (2);
   ChipsVector max_pos (2);

   if (bb_norm_x1 < bb_norm_x2){
      min_pos[0] = bb_norm_x1;
      max_pos[0] = bb_norm_x2;
   }
   else{
      min_pos[0] = bb_norm_x2;
      max_pos[0] = bb_norm_x1;
   }

   if (bb_norm_y1 < bb_norm_y2){
      min_pos[1] = bb_norm_y1;
      max_pos[1] = bb_norm_y2;
   }
   else{
      min_pos[1] = bb_norm_y2;
      max_pos[1] = bb_norm_y1;
   }

   ((ChipsFrameSystem *)coord)->set_pos (min_pos, max_pos);

   if (rval == ERR_STATUS_GOOD)   change_position();

   map <string, ChipsPlot*>::iterator   plot_ii;
   
   bind_table->reset_access_list();
   for (plot_ii = plot.begin (); plot_ii != plot.end (); plot_ii++)
   {
      bind_table->set_access_plot(plot_ii->second);
      plot_ii->second->recalculate_position();
   }

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval; 
}


int ChipsFrame::adjust_pos_by_delta(float dx, float dy)
{
   int rval = ERR_STATUS_GOOD;

   if ((dx < -1.0) || (dx > 1.0) || (dy < -1.0) || (dy > 1.0)) 
      cmd_manager->add_err(rval=ERR_NORM_DATA_RANGE,id,0);
   else if (((nx1 + dx) < 0.0) || ((nx1 + dx) > 1.0))
      cmd_manager->add_err(rval=ERR_EDGE_BOUNDARY,id,0,"frame");
   else if (((ny1 + dy) < 0.0) || ((ny1 + dy) > 1.0))
      cmd_manager->add_err(rval=ERR_EDGE_BOUNDARY,id,0,"frame");
   else if (((nx2 + dx) < 0.0) || ((nx2 + dx) > 1.0))
      cmd_manager->add_err(rval=ERR_EDGE_BOUNDARY,id,0,"frame");
   else if (((ny2 + dy) < 0.0) || ((ny2 + dy) > 1.0))
      cmd_manager->add_err(rval=ERR_EDGE_BOUNDARY,id,0,"frame");
   else
   {
      nx1 += dx; 
      ny1 += dy; 
      nx2 += dx; 
      ny2 += dy; 
      return rval; 
   } 
   return rval;
}


int ChipsFrame::reposition(float px1, float py1, float px2, float py2, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if ((px1 < 0.0) || (px1 > 1.0)) 
      cmd_manager->add_err(rval=ERR_NORM_DATA_RANGE,id,0);  
   else if ((px2 < 0.0) || (px2 > 1.0))
      cmd_manager->add_err(rval=ERR_NORM_DATA_RANGE,id,0);  
   else if ((py1 < 0.0) || (py1 > 1.0))
      cmd_manager->add_err(rval=ERR_NORM_DATA_RANGE,id,0);  
   else if ((py2 < 0.0) || (py2 > 1.0)) 
      cmd_manager->add_err(rval=ERR_NORM_DATA_RANGE,id,0);  
   else
   {
      if (rec)
      {
         ChipsRepoCommand *cmd = new ChipsRepoCommand(this, nx1, ny1, nx2, ny2,
            px1, py1, px2, py2, ((ChipsWindow *)parent)->get_current());
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      nx1 = (double)px1;
      ny1 = (double)py1;
      nx2 = (double)px2;
      ny2 = (double)py2; 
   }
   return rval; 
}


int ChipsFrame::move(COORD_SYS coord_sys, const string &coord_id, double xx,
                   double yy, int mode, bool rec)
{
   ChipsVector pos (xx, yy);
   int rval = ERR_STATUS_GOOD;
   ChipsMoveCommand* cmd = NULL;

   if (coord_sys == COORD_NONE)
      coord_sys = coord_system;
   else if (!coord->coord_exists (coord_sys, ""))
   {
      cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));
      return ERR_COORD_DNE;
   }

   if (rec)  cmd = new ChipsMoveCommand(this, nx1, ny1, xx, yy, coord_sys,
      coord_id, mode, ((ChipsWindow *)parent)->get_current());
  
   rval = coord->convert_cs_to_cs (&pos, &pos, 1, coord_sys, "", WINDOW_NORM, "");
   if (rval == ERR_STATUS_GOOD)
   {
      if (mode == V_MOVE_DELTA)
         rval = adjust_pos_by_delta(pos[0], pos[1]);
      else
         rval = adjust_pos_by_delta(pos[0]-nx1, pos[1]-ny1);
   }
   if ((rval == ERR_STATUS_GOOD) && rec)
   {
      if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }   
   }
   else 
      delete(cmd); 

   return rval;
}


int ChipsFrame::move_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj,
   double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD; 
   ChipsId tmp_id (*cid);
   COORD_SYS cd_sys = cid->get_coord_sys ();
   string oid, cd_id = "";
   ChipsCoordinate *tmp_coord;

   bool move_plot = ((obj == chips_plot) && (subobj == chips_not_subobject));

   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if (cd_sys != COORD_NONE)
   {
      tmp_coord = find_coordinate_system (tmp_id, rval);

      if (rval)
         return rval;  

      cd_sys = tmp_id.get_coord_sys ();
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((cd_sys == WINDOW_NORM) || (cd_sys == PIXEL) ||
          (cd_sys == PIXEL_REAL))
      {
         cd_id = "";
         tmp_id.set_coord_id ("");
      }
      else
         cd_id = tmp_coord->get_id ();
   }

   oid = cid->get_plot_id();
   if (oid.length() == 0)  oid = curr_plot;
   if (plot.empty()) 
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"plots");
      return ERR_OBJ_LIST_EMPTY;
   }

   if (move_plot)
   {
      if (cd_sys == PLOT_NORM)
         cmd_manager->add_err(rval=ERR_INV_MOVE,"",0,"plot",
                                 "plot normalized");
      else if (cd_sys == DATA)
         cmd_manager->add_err(rval=ERR_INV_MOVE,"",0,"plot", "data");
   }


   map <string, ChipsPlot*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts;
      rval = ERR_STATUS_GOOD;
      if (move_plot)
      {
         bind_table->reset_access_list();
         for(ii= plot.begin(); ii != plot.end(); ++ii)
         {
            bind_table->set_access_plot((*ii).second);
            if ((sts=(*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
               ERR_STATUS_GOOD)   rval = sts;
         }
      }
      else
      {
         for(ii= plot.begin(); ii != plot.end(); ++ii)
            if ((sts=(*ii).second->move_object(&tmp_id, obj, subobj, xx, yy, mode)) !=
               ERR_STATUS_GOOD)   rval = sts;
      }
   }
   else if (plot.count(oid))
   {
      if (move_plot)
      {
         bind_table->reset_access_list(plot[oid]);
         rval = plot[oid]->move(cd_sys, cd_id, xx, yy, mode); 
      }
      else
         rval = plot[oid]->move_object(&tmp_id, obj, subobj, xx, yy, mode);
   }
   else if (move_plot)
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   set_current_plot ( oid);

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}

int ChipsFrame::resize_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *bounds)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_plot_id();
   ChipsVector lower_left;
   ChipsVector upper_right;
   ChipsPlot *cp;

   bool resize_plot = ((obj == chips_plot) && (subobj == chips_not_subobject));


   if (!tmpid.length()) tmpid = curr_plot;

   if (tmpid.length())
   {

      if (resize_plot)
      {
         COORD_SYS cd_sys = cid->get_coord_sys ();
         string cd_id = cid->get_coord_id ();
         lower_left = ChipsVector (bounds[0], bounds[1]);
         upper_right = ChipsVector (bounds[2], bounds[3]);

         rval = coord->convert_cs_to_cs (&lower_left, &lower_left, 1,
                  cd_sys, cd_id, FRAME_NORM, "");
         rval = coord->convert_cs_to_cs (&upper_right, &upper_right, 1,
                  cd_sys, cd_id, FRAME_NORM, "");
         rval = this->reposition_plot (cid, lower_left[0], lower_left[1],
                                   upper_right[0], upper_right[1]);

      }
      else if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsPlot*>::iterator ii;
         for(ii=plot.begin(); ii != plot.end(); ++ii)
         {
            cp = (*ii).second;
            if ((sts = cp->resize_object(cid, obj, subobj, bounds)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");

         rval = ERR_FEATURE_DNE;
      }
      else
      {
         // move this frame
         cp = get_plot(tmpid);

         if (cp)
            rval = cp->resize_object(cid, obj, subobj, bounds);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "plot", tmpid.c_str());
         rval = ERR_FEATURE_DNE;
      }
      if (rval == ERR_STATUS_GOOD) set_current_plot (tmpid);
   }
   else  cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"plots");
   if (rval)  cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::rotate_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_plot_id();
   ChipsPlot *cp;


   if (!tmpid.length()) tmpid = curr_plot;

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsPlot*>::iterator ii;
         for(ii=plot.begin(); ii != plot.end(); ++ii)
         {
            cp = (*ii).second;
            if ((sts = cp->rotate_object(cid, obj, subobj, angle)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");

         rval = ERR_FEATURE_DNE;
      }
      else
      {

         // move this frame
         cp = get_plot(tmpid);

         if (cp)
            rval = cp->rotate_object(cid, obj, subobj, angle);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "plot", tmpid.c_str());
         rval = ERR_FEATURE_DNE;
      }
      if (rval == ERR_STATUS_GOOD) set_current_plot (tmpid);
   }
   else  cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"plots");
   if (rval)  cmd_manager->add_frame_id_to_errs(id);
   return rval;
}

int ChipsFrame::select_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_plot_id();
   ChipsPlot *cp;
   bool select_plot = ((obj == chips_plot) && (subobj == chips_not_subobject));


   if (!tmpid.length()) tmpid = curr_plot;

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsPlot*>::iterator ii;
         for(ii=plot.begin(); ii != plot.end(); ++ii)
         {
            cp = (*ii).second;
            if (select_plot)
            {
               pass++;
               if (selected) cp->select ();
                 else cp->unselect ();
            }
            else if ((sts = cp->select_object(cid, obj, subobj, selected)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");

         rval = ERR_FEATURE_DNE;
      }
      else
      {

         // move this frame
         cp = get_plot(tmpid);

         if (cp)
         {
            if (select_plot)
            {
               if (selected) cp->select ();
                 else cp->unselect ();
            }
            else
               rval = cp->select_object(cid, obj, subobj, selected);
         }
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "plot", tmpid.c_str());
         rval = ERR_FEATURE_DNE;
      }
      if (rval == ERR_STATUS_GOOD) set_current_plot (tmpid);
   }
   else  cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"plots");
   if (rval)  cmd_manager->add_frame_id_to_errs(id);
   return rval;
}


int ChipsFrame::get_annotation_position (const ChipsId* cid,
                                          V_OBJECT_TYPE obj, ChipsVector &pos)
{
   list<ChipsPlot*> cplots;
   ChipsPlot *cp = NULL;
   int  rval = ERR_STATUS_GOOD;

   if ((rval = get_plots (cid, cplots, false)) == ERR_STATUS_GOOD)
   {
      // run the command
      cp = *(cplots.begin());
      cp->get_annotation_position (cid, obj, pos);
   }

   return rval;

}

int ChipsFrame::change_annotation_coordinate (const ChipsId* cid,
                                          V_OBJECT_TYPE obj,
                                          COORD_SYS coordsys,
                                          const string& coordid)
{
   list<ChipsPlot*> cplots;
   ChipsPlot *cp = NULL;
   int  rval = ERR_STATUS_GOOD;

   if ((rval = get_plots (cid, cplots, true)) == ERR_STATUS_GOOD)
   {
      list<ChipsPlot*>::iterator ii;
      int rr, pass = 0, failval = 0;
      string pid = cid->get_plot_id ();

      // run the command
      for(ii=cplots.begin(); ii != cplots.end(); ++ii)
      {
         cp = *ii;
         rr = cp->change_annotation_coordinate(cid, obj, coordsys, coordid);
         if (rr == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }

      // check all passed
      if (plot.size () == cplots.size())
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      }

      // update currency
      if (pid.length () > 0)
         set_current_plot (pid);
   }

   return rval;
}



int ChipsFrame::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj,
   int& out_val)
{
   int rval = ERR_STATUS_GOOD;
        
   if (!cid)
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   else
   {
      out_val = 0; 
      string tid; 
      switch(obj)
      {
         case chips_label:        // fallthrough intended:
         case chips_line:         // fallthrough intended:
         case chips_point:        // fallthrough intended:
         case chips_region:       // fallthrough intended:
         case chips_contour:      // fallthrough intended:
	 case chips_curve:        // fallthrough intended:
	 case chips_image:        // fallthrough intended:
	 case chips_histogram:    // fallthrough intended:
         case chips_colorbar:     // fallthrough intended:
	 case chips_axis:        
	 {
	    tid = cid->get_plot_id();
            if (tid.length() == 0) tid = curr_plot;
            if (chips_strcmp(tid.c_str(), "all") == 0)
	    {
               map<string, ChipsPlot*>::iterator ii;
	       int tval = 0; 
	       for(ii=plot.begin(); ii != plot.end(); ++ii)
	       {
                  (*ii).second->get_object_count(cid, obj, tval);
	          out_val += tval; 
               }
	    }
            else if (plot.count(tid))
	       plot[tid]->get_object_count(cid, obj, out_val);
	 }
	 break;
	 case chips_plot:       
	 {
	    tid = cid->get_plot_id();
            if (tid.length() == 0) tid = curr_plot;
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = plot.size();
            else if (plot.count(tid))
               out_val = 1;
         }
         break;
         default:
         break;
      }
   }
   return rval;
}


int ChipsFrame::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   int rval = ERR_STATUS_GOOD;

   out_val = COORD_NONE;
   out_id = "";

   if (!cid)
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   else
   {
      string tid;
      if (obj == chips_plot)
      {
         tid = cid->get_plot_id();
         if (tid.length() == 0) tid = curr_plot;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            ChipsPlot* cp = get_plot();
            if (cp)
            {
               out_val = cp->get_coord_system();
               out_id = (out_val == DATA) ? cp->get_data_system_id() : "";
            }
            else  
               cmd_manager->add_err (rval=ERR_INVALID_GET_ALL,id,0);
         }
         else if (plot.count(tid))
         {
            out_val = plot[tid]->get_coord_system();
            out_id = (out_val == DATA) ? plot[tid]->get_data_system_id() : "";
         }
      }
      else
      {

         tid = cid->get_plot_id();
         if (tid.length() == 0) tid = curr_plot;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            ChipsPlot* cp = get_plot();
            if (cp)
               rval = cp->get_object_coordinfo(cid, obj, out_val, out_id);
            else 
               cmd_manager->add_err (rval=ERR_INVALID_GET_ALL,id,0);
         }
         else if (plot.count(tid))
            rval = plot[tid]->get_object_coordinfo(cid, obj, out_val, out_id);
      }
   }
   return rval;
}

int ChipsFrame::get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num)
{
   int rval = ERR_STATUS_BAD;

   out_depth_num = 0;
   out_shuffle_num = 0;


   if (!cid)
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   else if ((otype != chips_frame) && depth_table)
   {
      ChipsObject *obj = this->get_object (cid, otype);

      if (obj)
      {
         out_depth_num = obj->get_depth();
         out_shuffle_num  = depth_table->position_in_depth(obj->get_ref_cnt(), obj->get_depth());
         rval = ERR_STATUS_GOOD;
      }
   }

   return rval;
}




int ChipsFrame::set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid;
   switch(what)
   {
      case chips_line:      // fallthrough intended
      case chips_label:     // fallthrough intended 
      case chips_point:     // fallthrough intended 
      case chips_region:    // fallthrough intended 
      case chips_axis:      // fallthrough intended 
      case chips_curve:     // fallthrough intended 
      case chips_contour:   // fallthrough intended 
      case chips_image:     // fallthrough intended 
      case chips_colorbar:  // fallthrough intended 
      case chips_histogram: 
      {
	 string tmpid = cid->get_plot_id();
         if (tmpid.length() == 0)  tmpid = curr_plot;
	 if (tmpid.length())
	 {
            if (chips_strcmp(tmpid.c_str(), "all") == 0)
	    {
               map<string, ChipsPlot*>::iterator ii;
	       for(ii=plot.begin(); ii != plot.end(); ++ii)
	          if (what == chips_axis)
                     rval = (*ii).second->set_current_axis(cid, rec);
		  else if (what == chips_curve)
                     rval = (*ii).second->set_current_curve(cid, rec);
                  else if (what == chips_contour)
                     rval = (*ii).second->set_current_contour(cid, rec);
                  else if (what == chips_histogram)
                     rval = (*ii).second->set_current_histogram(cid, rec);
                  else if (what == chips_label)
                     rval = (*ii).second->set_current_label(cid, rec);
                  else if (what == chips_line)
                     rval = (*ii).second->set_current_line(cid, rec);
                  else if (what == chips_point)
                     rval = (*ii).second->set_current_point(cid, rec);
                  else if (what == chips_region)
                     rval = (*ii).second->set_current_region(cid, rec);
                  else if (what == chips_image)
                     rval = (*ii).second->set_current_image(cid, rec);
                  else if (what == chips_colorbar)
                     rval = (*ii).second->set_current_colorbar(cid, rec);
            }
	    else if (plot.count(tmpid))
	    {
	       if (what == chips_axis)
                  rval = plot[tmpid]->set_current_axis(cid, rec);
               else if (what == chips_curve)
                  rval = plot[tmpid]->set_current_curve(cid, rec);
               else if (what == chips_contour)
                  rval = plot[tmpid]->set_current_contour(cid, rec);
               else if (what == chips_histogram)
                  rval = plot[tmpid]->set_current_histogram(cid, rec);
               else if (what == chips_label)
                  rval = plot[tmpid]->set_current_label(cid, rec);
               else if (what == chips_line)
                  rval = plot[tmpid]->set_current_line(cid, rec);
               else if (what == chips_point)
                  rval = plot[tmpid]->set_current_point(cid, rec);
               else if (what == chips_region)
                  rval = plot[tmpid]->set_current_region(cid, rec);
               else if (what == chips_image)
                  rval = plot[tmpid]->set_current_image(cid, rec);
               else if (what == chips_colorbar)
                  rval = plot[tmpid]->set_current_colorbar(cid, rec);
	    }
	    if (rval == ERR_STATUS_GOOD)
	       set_current_plot ( tmpid);
            else
	       cmd_manager->add_frame_id_to_errs(id);
         }
	 else
	    cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"plots");
      }
      break; 
      case chips_plot:
      {
         tmpid = cid->get_plot_id();
         ChipsCurrentCommand *cmd;
         if (rec)
            cmd = new ChipsCurrentCommand(this, curr_plot, tmpid, what);
         if ((rval = set_current_plot(tmpid)) == ERR_STATUS_GOOD)
         {
            if (rec)
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
         }
         else if (rec) delete cmd;
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"object", what,
            "chips_window:chips_region");
      break;
   }
   return rval;
}

int ChipsFrame::set_current_plot(const string &val)
{
   int rval = ERR_STATUS_GOOD;
   string prev_curr_plot = curr_plot;
   map<string, ChipsPlot*>::iterator ii;


   if (plot.count(val))
   {
      curr_plot = val;
      plot[val]->enable_picklimits (true);
      plot[val]->enable_addannotation (true);
   }
   else if (val.length () == 0)
   {
      curr_plot = val;
      rval = ERR_STATUS_GOOD;
   }
   else if (chips_strcmp(val.c_str(), "all") || plot.empty())
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "plot", val.c_str());
   else
   {
      curr_plot = "all";
      
/*  *** Only enable for 1 plot ***
      for(ii=plot.begin(); ii != plot.end(); ++ii)
      {
         ii->second->enable_picklimits (true);
         ii->second->enable_addannotation (true);
      }
*/
      ii=plot.begin();
      ii->second->enable_picklimits (true);
      ii->second->enable_addannotation (true);

   }

   if (!rval && (prev_curr_plot != curr_plot))
   {
      if (chips_strcmp(prev_curr_plot.c_str(), "all") == 0)
      {
         for(ii=plot.begin(); ii != plot.end(); ++ii)
            if (ii->first != curr_plot)
            {
               ii->second->enable_picklimits (false);
               ii->second->enable_addannotation (false);
            }
      }
      else if ((plot.count (prev_curr_plot)) && (chips_strcmp(prev_curr_plot.c_str(), "all") != 0))
      {
         plot[prev_curr_plot]->enable_picklimits (false);
         plot[prev_curr_plot]->enable_addannotation (false);
      }
   }
 
   return rval; 
}


int ChipsFrame::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_background_color(in_color)) == ERR_STATUS_GOOD)
      rval = change_color();
   return rval;
}


int ChipsFrame::set_border_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = border_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      rval = change_border_color();
   return rval;
}


int ChipsFrame::adjust_default_colors()
{
   // go through all object lists and adjust the object colors 
   map<string, ChipsPlot* >::iterator pp;
   for (pp =plot.begin(); pp != plot.end(); ++pp)
      (*pp).second->adjust_default_colors();
   // adjust frame colors
   if (color.get_name() == "default")         change_color(); 
   if (border_color.get_name() == "default")  change_border_color();

   return ERR_STATUS_GOOD; 
}

int ChipsFrame::update_print_scale (double scale)
{
   // go through all object lists and adjust scale for printing
   map<string, ChipsPlot* >::iterator pp;
   for (pp =plot.begin(); pp != plot.end(); ++pp)
      (*pp).second->update_print_scale (scale);

   return ERR_STATUS_GOOD;
}



int ChipsFrame::set_transparency(bool transparency)
{
   int rval = ERR_STATUS_GOOD;

   if (opaque == transparency) 
   {
      opaque = !transparency;
      rval = change_transparency();
   }
   return rval; 
}

int ChipsFrame::set_border_visibility(bool vis)
{
   int rval = ERR_STATUS_GOOD;

   if (display_border != vis)
   {
      display_border = vis;
      rval = change_border_visibility();
   }
   return rval;
}


void ChipsFrame::get_bb_norm_pos (float &out_x1, float &out_y1, 
                                  float &out_x2, float &out_y2)
{
   out_x1 = static_cast <float> (bb_norm_x1);
   out_y1 = static_cast <float> (bb_norm_y1);
   out_x2 = static_cast <float> (bb_norm_x2);
   out_y2 = static_cast <float> (bb_norm_y2);
}


int ChipsFrame::adjust_grid_gaps(float xg, float yg)
{
   int rval = ERR_STATUS_GOOD;

   if (xg < 0.0)
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"horizontal gap",xg,0.0);
   else if (xg > 1.0)
      cmd_manager->add_err(rval=ERR_FLT_MAX,id,0,"horizontal gap",xg,1.0);
   else if (yg < 0.0) 
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"vertical gap",yg,0.0);
   else if (yg > 1.0)
      cmd_manager->add_err(rval=ERR_FLT_MAX,id,0,"vertital gap",yg,1.0);
   else if ((sprows != 0) && (yg*(sprows-1) > (1.0-topmargin-bottommargin)))
      cmd_manager->add_err(rval=ERR_YGAP_VS_ROWS,id,0,yg);  
   else if ((spcols != 0) && (xg*(spcols-1) > (1.0-leftmargin-rightmargin)))
      cmd_manager->add_err(rval=ERR_XGAP_VS_COLS,id,0,xg);  
   else if ((xg != xgap) || (yg != ygap))
   {
      int ff;
      xgap = xg;
      ygap = yg;
      ChipsPlot *tmpl;
      compute_object_pos();
      for (ff = 0; ff < spcnt; ff++)
      {
         if (spplot[ff].length())
         {
            if ((tmpl = get_plot(spplot[ff])) != NULL)
	    {
	       bool offscr = tmpl->is_offscreen();
               rval = reposition_plot_area(tmpl, spleft[ff], spbottom[ff],
                                           spright[ff], sptop[ff], offscr);
            }
         }
      }
   }

   return rval;
}


int ChipsFrame::swap_object_positions(string id1, string id2)
{
   int rval = ERR_STATUS_GOOD;

   if (id1 == id2)
      cmd_manager->add_err(rval=ERR_SWAP_SELF,id,0,id1.c_str());
   else if (id1.length() && id2.length())
   {
      ChipsPlot *tmpf1 = get_plot(id1);
      ChipsPlot *tmpf2 = get_plot(id2);

      if (tmpf1 && tmpf2)
      {
         float f1x0, f1y0, f1x1, f1y1, f2x0, f2y0, f2x1, f2y1;
	 bool p1off, p2off;
	 p1off = tmpf1->is_offscreen();
	 p2off = tmpf2->is_offscreen();
         f1x0 = tmpf1->get_plot_lmargin();
         f1y0 = tmpf1->get_plot_bmargin();
         f1x1 = 1 - tmpf1->get_plot_rmargin();
         f1y1 = 1 - tmpf1->get_plot_tmargin();
         f2x0 = tmpf2->get_plot_lmargin();
         f2y0 = tmpf2->get_plot_bmargin();
         f2x1 = 1 - tmpf2->get_plot_rmargin();
         f2y1 = 1 - tmpf2->get_plot_tmargin();

         reposition_plot_area(tmpf1, f2x0, f2y0, f2x1, f2y1, p2off);
         reposition_plot_area(tmpf2, f1x0, f1y0, f1x1, f1y1, p1off);

         for (int ff = 0; ff < spcnt; ff++)
         {
            if (spplot[ff] == id1)  spplot[ff] = id2;
               else if (spplot[ff] == id2)  spplot[ff] = id1;
         }
      }
      else if (tmpf1)
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",id2.c_str());
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",id1.c_str());
   }
   else if (id1.length())
      cmd_manager->add_err("The second plot id was not specified or was blank",
                           rval=ERR_NO_ID_OBJ, id);
   else
      cmd_manager->add_err("The first plot id was not specified or was blank",
                           rval=ERR_NO_ID_OBJ, id);

   return rval;
}


int ChipsFrame::grid_objects(int cols, int rows, float xg, float yg,
   int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if ((leftmargin + rightmargin) > 1)
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"left",leftmargin,
                           "right",rightmargin); 
   else if ((topmargin + bottommargin) > 1)
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"top",topmargin,
                           "bottom",bottommargin); 
   else if (rows < 1)
      cmd_manager->add_err(rval=ERR_INT_MIN,id,0,"rows",rows,0);
   else if (cols < 1)
      cmd_manager->add_err(rval=ERR_INT_MIN,id,0,"cols",cols,0);
   else if ((xg < 0.0) || (xg > 1.0))
      cmd_manager->add_err(rval=ERR_DBL_MINMAX,id,0,"xgap",xg,0,1.0);
   else if ((yg < 0.0) || (yg > 1.0)) 
      cmd_manager->add_err(rval=ERR_DBL_MINMAX,id,0,"ygap",yg,0,1.0);
   else if (((cols -1)*xg + leftmargin + rightmargin) > 1.0)
      cmd_manager->add_err(rval=ERR_XGAP_VS_COLS,id,0,xg);  
   else if (((rows -1)*yg + topmargin + bottommargin) > 1.0)
      cmd_manager->add_err(rval=ERR_YGAP_VS_ROWS,id,0,yg);  
   else if ((usage < 0) || (usage > 4))
      cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"usage", usage,
         "0:use only, 1:use fill, 2:fill, 3:exclude,4:exclude fill");

   else if (rval == ERR_STATUS_GOOD) 
   {
      int ori = 0, ocnt = 0, *oval = NULL;  
      if (((sprows == 0) && (spcols == 0)) || ((sprows != 0) && (spcols != 0)))
      {
         oval = NULL;  
         ori = ocnt = 0; 
      }
      else if (sprows == 0)
      {
         oval = sparr;
         ori = 1; 
         ocnt = spcols; 
      }
      else if (spcols == 0)
      {
         oval = sparr;
         ori = 1; 
         ocnt = sprows; 
      }
      ChipsSetSplitCommand* cmd = new ChipsSetSplitCommand(this, ori, ocnt,
         oval, spcols, sprows, xgap, ygap, spplot, curr_plot);
      if ((rval = reset_split_variables(0, 0, NULL, cols, rows, xg, yg))
          == ERR_STATUS_GOOD)
      {
         compute_object_pos();
         rval = arrange_objects(usage, lcnt, lids);
         if (rval == ERR_STATUS_GOOD)
         {
            cmd->set_new_values(0, 0, NULL, spcols, sprows, xgap, ygap, spplot,
                                curr_plot);  
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else 
            delete cmd; 
      }
   }

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::grid_objects(int orient, int cnt, int* vals, float xg,
   float yg, int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;
   int rr;

   // record any changes in the axes
   bind_table->record_updates (true);
   bind_table->reset_access_list();

   if ((lcnt  > 0) && lids)
   {
      for (rr = 0; !rval && (rr < lcnt); rr++)
         if (get_plot(lids[rr]) == NULL) 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",lids[rr].c_str());
   }

   if ((leftmargin + rightmargin) > 1)
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"left",leftmargin,
                           "right",rightmargin);
   else if ((topmargin + bottommargin) > 1)
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"top",topmargin,
                           "bottom",bottommargin);
   else if ((orient < 0) || (orient > 1))
      cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"orientation", orient,
                           "0 (row) | 1 (column)");
   else if (cnt < 1)
      cmd_manager->add_err(rval=ERR_INT_GTE,id,0,"cnt", cnt, 0);
   else if (vals == NULL)
      cmd_manager->add_err(rval= ERR_NO_VALS_SUPPLIED,id,0);
   else if ((xg < 0.0) || (xg > 1.0))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX,id,0,"xgap",xg,0,1.0);
   else if ((yg < 0.0) || (yg > 1.0))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX,id,0,"ygap",yg,0,1.0);
   else if ((usage < 0) || (usage > 4))
      cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"usage", usage,
         "0:use only, 1:use fill, 2:fill, 3:exclude,4:exclude fill");
   else
   {
      char errbuf[256];
      // verify array data
      int total = 0;
      for (rr = cnt;rr--;)
      {
         total += vals[rr];
         if (vals[rr] < 1)
         {
            if (orient == 0)
               sprintf(errbuf,"The rows array element %1d value (%1d) must be greater than 0.", rr, vals[rr]);
            else 
               sprintf(errbuf,"The columns array element %1d value (%1d) must be greater than 0.", rr, vals[rr]);
            cmd_manager->add_err(errbuf, rval=ERR_ARRAY_VALUES);
         }
         else if (vals[rr] > 50) 
         {
            if (orient == 0)
               sprintf(errbuf,"The rows array element %1d value (%1d) must be less than 50.", rr, vals[rr]);
            else 
               sprintf(errbuf,"The columns array element %1d value (%1d) must be less than 50.", rr, vals[rr]);
            cmd_manager->add_err(errbuf, rval=ERR_ARRAY_VALUES);
         }
      }
   }

   if (rval == ERR_STATUS_GOOD)
   {
      int ori = 0, ocnt = 0, *oval = NULL;
      if (((sprows == 0) && (spcols == 0)) || ((sprows != 0) && (spcols != 0)))
      {
         oval = NULL;
         ori = ocnt = 0;
      }
      else if (sprows == 0)
      {
         oval = sparr;
         ori = 1;
         ocnt = spcols;
      }
      else if (spcols == 0)
      {
         oval = sparr;
         ori = 1;
         ocnt = sprows;
      }
      ChipsSetSplitCommand* cmd = new ChipsSetSplitCommand(this, ori, ocnt,
         oval, spcols, sprows, xgap, ygap, spplot, curr_plot);

      if ((rval = reset_split_variables(orient, cnt, vals, 0, 0, xg, yg))
           == ERR_STATUS_GOOD)
      {
         compute_object_pos();
         rval = arrange_objects(usage, lcnt, lids);
         if (rval == ERR_STATUS_GOOD)
         {
            cmd->set_new_values(orient, cnt, vals, 0, 0, xgap, ygap, spplot,
                                curr_plot);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else
            delete cmd;
      }
   }

   // update the limits across plots
   if (!rval)
      check_plot_updates ();
   bind_table->record_updates (false);
   bind_table->reset_access_list();

   return rval;
}


int ChipsFrame::reset_split_variables(int orient, int cnt, int* vals, int cols,
    int rows, float xg, float yg, string* lids)
{
   int rr, rval = ERR_STATUS_GOOD;

   if (spleft) delete []spleft;
   if (spbottom) delete []spbottom;
   if (spright) delete []spright;
   if (sptop) delete []sptop;
   if (spxrel) delete []spxrel;
   if (spyrel) delete []spyrel;
   if (spplot) delete []spplot;
   if (sparr) delete []sparr;

   if (cnt)
   {
      sparr = new int[cnt];
      spcnt = 0;
      for (rr = cnt;rr--;) spcnt += (sparr[rr] = vals[rr]);
      if (orient == 1)
      {
         spcols = cnt;
         sprows = 0;
         xrelsum = spcols;
         yrelsum = spcnt;
      }
      else
      {
         spcols = 0;
         sprows = cnt;
         yrelsum = sprows;
         xrelsum = spcnt;
      }
   }
   else
   {
      spcnt = (int)((yrelsum = sprows = rows) * ( xrelsum = spcols = cols));
      sparr = NULL;
   }
   spleft = new float[spcnt];
   spbottom = new float[spcnt];
   spright = new float[spcnt];
   sptop = new float[spcnt];
   spxrel = new float[spcnt];
   spyrel = new float[spcnt];
   spplot = new string[spcnt];
   xgap = xg;
   ygap = yg;

   if (!spleft || !spright || !sptop || !spbottom || !spxrel || !spyrel ||
       !spplot)
      cmd_manager->add_err(rval=ERR_ALLOC_FAIL);
   else 
      for (rr = spcnt; rr--; )
      {
         spxrel[rr] = spyrel[rr] = 1.0;
         if (lids)  spplot[rr] = lids[rr];
      }
   return rval; 
}


ChipsObject* ChipsFrame::get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                                     bool include_deref)
{
   ChipsObject* obj = NULL;
   string my_id;
   int obj_ref = 0;
   if (cid)
   {
      ChipsPlot* pl = NULL;
      my_id = cid->get_plot_id();
      obj_ref = cid->get_plot_ref_id();
      if (chips_strcmp(my_id.c_str(), "all"))
         pl = get_plot(my_id);
      else if (plot.size() == 1)
         pl = get_plot("");

      switch(otype)
      {
         case chips_plot:
         {
            if (pl)
               if ((obj_ref == 0) || ((obj_ref > 0) && (pl->get_ref_cnt () == obj_ref)))
                  obj = pl;
            if (include_deref && !obj)
            {
               if (dplot.count (obj_ref) != 0)
                  obj = dplot[obj_ref];
            }
         }
         break;
         default:
         {
            // if obj_ref is not zero then confirm the reference count
            if (pl)
               if (((obj_ref > 0) && (pl->get_ref_cnt () != obj_ref)))
                  pl = NULL;
            if (include_deref && !pl)
            {
               if (dplot.count (obj_ref) != 0)
                  pl = dplot[obj_ref];
            }
            if (pl != NULL)
               obj = pl->get_object(cid,otype,include_deref);
         }
         break;
      }
   }
   return obj;
}


int ChipsFrame::delete_object(const string &id, V_OBJECT_TYPE otype)
{
   int rval;
   ChipsId cid;

   switch(otype)
   {
      case chips_plot:
      {
         cid.set_plot_id (id);
         rval = delete_plot (&cid);
      }
      break;
      case chips_label:       // fallthrough intended
      case chips_line:        // fallthrough intended
      case chips_point:       // fallthrough intended
      case chips_region:      // fallthrough intended
      case chips_curve:       // fallthrough intended
      case chips_contour:     // fallthrough intended
      case chips_histogram:   // fallthrough intended
      case chips_image:       // fallthrough intended
      case chips_colorbar:    // fallthrough intended
      case chips_axis:
         // ignore these cases for now
      default:
      break;
   }
   return rval;
}



int ChipsFrame::compute_object_pos()
{
   int rr, cc, lyr = 0, rval = ERR_STATUS_GOOD;

   if (xrelsum && yrelsum)
   {
      float hlen, vlen;
      if (sprows && spcols)
      {
         hlen = (1 - (leftmargin + rightmargin + (spcols-1)*xgap)) / xrelsum; 
         vlen = (1 - (topmargin + bottommargin + (sprows-1)*ygap)) / yrelsum; 

         for (rr = 0; rr < sprows; rr++)
         {
            if (rr == 0)
               sptop[lyr] = 1 - topmargin;
            else
               sptop[lyr] = spbottom[lyr -1] - ygap;
            spbottom[lyr] = sptop[lyr] - (spyrel[lyr] * vlen);
      
            spright[lyr] = leftmargin + (spxrel[lyr] * hlen);
            spleft[lyr++]   = leftmargin;

            for (cc = 1; cc < spcols; cc++)
            {
               spleft[lyr] = spright[lyr-1] + xgap;
               spright[lyr] = spleft[lyr] + (spxrel[lyr] * hlen);
               spbottom[lyr] = spbottom[lyr-1];
               sptop[lyr] = sptop[lyr-1];
               lyr++;
            }
         }
      }
      else if (sprows && sparr)
      {
         vlen = (1 - (topmargin + bottommargin + (sprows-1)*ygap)) / yrelsum; 

         for (rr = 0; rr < sprows; rr++)
         {
            if (rr == 0)
               sptop[lyr] = 1 - topmargin;
            else
               sptop[lyr] = spbottom[lyr -1] - ygap;
            spbottom[lyr] = sptop[lyr] - (spyrel[lyr] * vlen);

            hlen = (1-(leftmargin+rightmargin+(sparr[rr]-1)*xgap))/sparr[rr];
            spleft[lyr] = leftmargin;
            spright[lyr] = leftmargin + (spxrel[lyr] * hlen);
            lyr++;
   
            for (cc = 1; cc < sparr[rr]; cc++)
            {
               spleft[lyr] = spright[lyr-1] + xgap;
               spright[lyr] = spleft[lyr] + hlen;
               spbottom[lyr] = spbottom[lyr-1];
               sptop[lyr] = sptop[lyr-1];
               lyr++;
            }
         }
      }
      else if (spcols && sparr)
      {
         hlen = (1 - (leftmargin + rightmargin + (spcols-1)*xgap)) / xrelsum;

         for (cc = 0; cc < spcols; cc++)
         {
            if (cc == 0)
               spleft[lyr] = leftmargin;
            else
               spleft[lyr] = spright[lyr-1] + xgap;
            spright[lyr] = spleft[lyr] + (spxrel[lyr] * hlen);
   
            vlen = (1-(topmargin+bottommargin + (sparr[cc]-1)*ygap))/ sparr[cc];
            sptop[lyr] = 1 - topmargin;
            spbottom[lyr] = sptop[lyr] - vlen;
            lyr++;
            for (rr = 1; rr < sparr[cc]; rr++)
            {
               spleft[lyr] = spleft[lyr-1];
               spright[lyr] = spright[lyr-1];
               sptop[lyr] = spbottom[lyr -1] - ygap;
               spbottom[lyr] = sptop[lyr] - vlen;
               lyr++;
            }
         }
      }
      else  rval = ERR_STATUS_BAD;  
   }
   else  rval = ERR_STATUS_BAD;  // bad relsum  

   return rval; 
}


int ChipsFrame::arrange_objects(int usage, int lcnt, string* lids)
{
   ChipsPlot* cl;
   ChipsId blank;
   int rr, cc, lndx = 0, rval = ERR_STATUS_GOOD;

   // set all plots offscreen and reinsert them as necessary
   map<string, ChipsPlot*>::iterator ii;
   for (ii = plot.begin(); ii != plot.end(); ++ii)
      (*ii).second->move_offscreen(true);

   switch (usage)
   {
      case 0:  // USE ONLY  - fallthrough intended
      case 1:  // USE FILL
      {
         map<string, ChipsPlot*>::iterator ii;
         if (lids && lcnt)
         {
            if (lcnt > spcnt)  lcnt = spcnt;

            for (rr = 0; rr < lcnt; rr++)
            {
               if ((cl = get_plot(lids[rr])) != NULL)
               {
                  rval = reposition_plot_area(cl, spleft[rr], spbottom[rr],
                                              spright[rr],sptop[rr], false);
                  if (rval == ERR_STATUS_GOOD)  spplot[rr] = lids[rr];
               
                  set_current_plot ( cl->get_id());
               }
            }
            if (usage == 1)
            {
               for (lndx = lcnt; lndx < spcnt; lndx++)
               {
                  if ((rval = add_plot(&blank,spleft[lndx],spbottom[lndx],spright[lndx],sptop[lndx])) == ERR_STATUS_GOOD)
                  {
                     spplot[lndx] = curr_plot;
                  }
               }
               if (lcnt < spcnt)  set_current_plot ( spplot[lcnt]);
               else               set_current_plot ( spplot[spcnt-1]);
            }
         }
         else
         {
            lndx = 0;
            int lsize = plot.size();

	    // plots are mapped alphabetically-  "p1,p11,p12,p2,p3,..."
	    // reindex to arrange in the order users expect (chronological)
	    map <int, string>chron;
            ChipsPlot* cpt;

            for(ii = plot.begin(); ii != plot.end(); ++ii)
	       chron.insert(pair<int,string>((*ii).second->get_ref_cnt(),(*ii).first));
	    for(map <int, string>::iterator i2= chron.begin(); i2 != chron.end();i2++)
            {
	       cpt = plot[(*i2).second];
	       if (lndx < spcnt)
	       {
                  rval = reposition_plot_area(cpt,spleft[lndx],
                     spbottom[lndx],spright[lndx],sptop[lndx]);
                  if (rval==ERR_STATUS_GOOD)  spplot[lndx] = cpt->get_id();
	       }
	       else   cpt->move_offscreen(true);
               lndx++;
	    }

            if (usage == 1)
            {
               for (lndx = lsize; lndx < spcnt; lndx++)
               {
                  if ((rval = add_plot(&blank,spleft[lndx],spbottom[lndx],spright[lndx],sptop[lndx])) == ERR_STATUS_GOOD)
                  {
                     spplot[lndx] = curr_plot;
                  }
               }
               if (lsize < spcnt)  set_current_plot ( spplot[lsize]);
               else                set_current_plot ( spplot[spcnt-1]);
            }
            else if (lndx > 0) 
	    {
	       if (lndx > spcnt)
                  set_current_plot ( spplot[spcnt-1]);
               else
                  set_current_plot ( spplot[lndx-1]);
	    }
         }
      }
      break;
      case 2:  // FILL
      {
         for (lndx =0; lndx < spcnt; lndx++)
         {
            if ((rval = add_plot(&blank, spleft[lndx],spbottom[lndx], spright[lndx],sptop[lndx])) == ERR_STATUS_GOOD)
               spplot[lndx] = curr_plot;
         }
         set_current_plot ( spplot[0]);
      }
      break;
      case 3:  // EXCLUDE      - fallthrough intended
      case 4:  // EXCLUDE FILL
      {
         if (lids && lcnt)
         {
            lndx = rr = 0;
            int skip = 0, lsize = plot.size();
            map <string, ChipsPlot*>::iterator ii;
            string lyid;
            unsigned long *key_list  = new unsigned long [lcnt];

            map <int, string>chron;
            for(ii = plot.begin(); ii != plot.end(); ii++)
               chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), (*ii).first));
            map <int, string>::iterator i2 = chron.begin(); 

            while ((rr < lsize) && (lndx < spcnt))
            {
               skip = 0;
               lyid = (*i2).second;
               for (cc = lcnt; cc--; )   skip |= (lyid == lids[cc]);
               if (!skip)
               {
                  rval = reposition_plot_area(plot[(*i2).second], spleft[lndx],
                                   spbottom[lndx],spright[lndx],sptop[lndx],false);
                  if (rval == ERR_STATUS_GOOD)  spplot[lndx] = lyid;
                        lndx++;
               }
               i2++;
               rr++;
            }
            delete [] key_list;
            chron.clear();
            if (usage == 4)
            {
               for (rr = lndx; rr < spcnt; rr++)
                  if ((rval = add_plot(&blank, spleft[rr], spbottom[rr],
                      spright[rr],sptop[rr])) == ERR_STATUS_GOOD)
                     spplot[rr] = curr_plot;
            }
            else  set_current_plot ( lyid);
         }
         else if (usage == 4)
         {
            
            if ((rval = add_plot(&blank, spleft[0], spbottom[0], 
                spright[0], sptop[0])) == ERR_STATUS_GOOD)
               spplot[0] = curr_plot;

            for (lndx =1; lndx < spcnt; lndx++)
            {
               if ((rval = add_plot(&blank, spleft[lndx], spbottom[lndx],
                   spright[lndx],sptop[lndx])) == ERR_STATUS_GOOD)
                  spplot[lndx] = curr_plot;
            }
            set_current_plot ( spplot[0]);
         }
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"usage", usage,
            "0:use only, 1:use fill, 2:fill, 3:exclude,4:exclude fill");
      break;
   }  // end switch


   return rval;
}


int ChipsFrame::adjust_grid_xrelsize(int entry, float size, bool rec)
{
   int rval = ERR_STATUS_GOOD;
  if (spcols == 0)
      cmd_manager->add_err(rval=ERR_SPLITROW_X,id,0);
   else if ((entry < 1) || (entry > spcols))
      cmd_manager->add_err(rval=ERR_INT_MINMAX,id,0,"entry",entry,1,spcols);
   else if (size <= 0.0)
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"relsize ratio",size,0);
   else
   {
      ChipsId blank;
      int ff;

      ChipsAdjustRelsizeCommand* cmd = NULL;
      if (rec)  cmd = new ChipsAdjustRelsizeCommand(this,
         spcnt, X_AXIS, spxrel, xrelsum);

      if (sprows && spcols)
      {
         for (ff = spcnt; ff--; )
         {
            if ( (ff % spcols) == (entry -1))  spxrel[ff] = size;
            else                               spxrel[ff] = 1;
         }
      }
      else
      {
         int zz, gg = 0;
         for (ff = 0; ff < spcols; ff++)
         {
            if (ff == (entry -1))
               for (zz = 0; zz < sparr[ff]; zz++)   spxrel[gg++] = size;
            else
               for (zz = 0; zz < sparr[ff]; zz++)   spxrel[gg++] = 1;
         }
      }
      xrelsum = spcols - 1 + size;
      if (rec && cmd)
      {
         cmd->store_new_values(spcnt,spxrel,xrelsum);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      compute_object_pos();
      for (ff = 0; ff < spcnt; ff++)
      {
         if (spplot[ff].length() and plot.count(spplot[ff]))
         {
            rval = reposition_plot_area(plot[spplot[ff]], spleft[ff], spbottom[ff],
                                           spright[ff], sptop[ff], false);
            blank.set_plot_id("");
         }
      }
   }
   return rval;
}


int ChipsFrame::adjust_grid_yrelsize(int entry, float size, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (sprows == 0)
      cmd_manager->add_err(rval=ERR_SPLITCOL_Y,id,0);
   else if ((entry < 1) || (entry > sprows))
      cmd_manager->add_err(rval=ERR_INT_MINMAX,id,0,"entry",entry,1,spcols);
   else if (size <= 0.0)
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"relsize ratio",size,0);
   else
   {
      ChipsId blank;
      int ff;

      ChipsAdjustRelsizeCommand* cmd = NULL;
      if (rec)  cmd = new ChipsAdjustRelsizeCommand(this,
         spcnt, Y_AXIS, spyrel, yrelsum);

      if (sprows && spcols)
      {
         int zz, gg = 0;
         for (ff = 0; ff < sprows; ff++)
            if (ff == (entry -1))
               for (zz = 0; zz < spcols; zz++)   spyrel[gg++] = size;
            else
               for (zz = 0; zz < spcols; zz++)   spyrel[gg++] = 1;
      }
      else
      {
         int zz, gg = 0;
         for (ff = 0; ff < sprows; ff++)
            if (ff == (entry -1))
               for (zz = 0; zz < sparr[ff]; zz++)   spyrel[gg++] = size;
            else
               for (zz = 0; zz < sparr[ff]; zz++)   spyrel[gg++] = 1;
      }
      yrelsum = sprows - 1 + size;
      if (rec && cmd)
      {
         cmd->store_new_values(spcnt,spyrel,yrelsum);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      compute_object_pos();
      for (ff = 0; ff < spcnt; ff++)
      {
         if (spplot[ff].length() and plot.count(spplot[ff]))
         {
            rval = reposition_plot_area(plot[spplot[ff]], spleft[ff], spbottom[ff],
                                        spright[ff], sptop[ff], false);
            blank.set_plot_id("");
         }
      }
   }
   return rval;
}

int ChipsFrame::adjust_grid_xrelsizes(int count, float* sizes, bool rec)
{
   int ff;
   float totsize = 0.0;

   if (spcols == 0)
   {
      cmd_manager->add_err(ERR_SPLITROW_X,id,0);
      return ERR_SPLITROW_X; // can't adjust x of split rows
   }
   else if (count != spcols)
   {
      cmd_manager->add_err(ERR_EXPECTED_CNT,id,0,count,spcols);
      return ERR_EXPECTED_CNT;
   }
   for (ff = 0; ff < spcols; ff++)
      if (sizes[ff] > 0)  totsize += sizes[ff];
         else             return ERR_STATUS_BAD;

   ChipsId blank;
   int zz, gg = 0, rval = ERR_STATUS_GOOD;;

   ChipsAdjustRelsizeCommand* cmd = NULL;
      if (rec)  cmd = new ChipsAdjustRelsizeCommand(this,
         spcnt, X_AXIS, spxrel, xrelsum);

   if (sprows && spcols)
      for (ff = 0; ff < sprows; ff++)
         for (zz = 0; zz < spcols; zz++)  spxrel[gg++] = sizes[zz];
   else
      for (ff = 0; ff < spcols; ff++)
         for (zz = 0; zz < sparr[ff]; zz++)   spxrel[gg++] = sizes[ff];
   xrelsum = totsize;

   if (rec && cmd)
   {
      cmd->store_new_values(spcnt,spxrel,xrelsum);
      if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }
   }

   compute_object_pos();
   for (ff = 0; ff < spcnt; ff++)
      if (spplot[ff].length() and plot.count(spplot[ff]))
      {
         rval = reposition_plot_area(plot[spplot[ff]], spleft[ff], spbottom[ff],
                                     spright[ff], sptop[ff], false);
         blank.set_plot_id("");
      }
   return rval;
}


int ChipsFrame::adjust_grid_yrelsizes(int count, float* sizes, bool rec)
{
   int ff;
   float totsize = 0.0;

   if (sprows == 0)
   {
      cmd_manager->add_err(ERR_SPLITROW_X,id,0);
      return ERR_SPLITCOL_Y; // can't adjust y of split cols
   }
   else if (count != sprows)
   {
      cmd_manager->add_err(ERR_EXPECTED_CNT,id,0,count,spcols);
      return ERR_EXPECTED_CNT;
   }
   
   for (ff = 0; ff < sprows; ff++)
      if (sizes[ff] > 0)  totsize += sizes[ff];
         else             return ERR_STATUS_BAD;

   ChipsId blank;
   int zz, gg = 0, rval = ERR_STATUS_GOOD;;

   ChipsAdjustRelsizeCommand* cmd = NULL;
   if (rec)  cmd = new ChipsAdjustRelsizeCommand(this,
      spcnt, Y_AXIS, spyrel, yrelsum);

   if (sprows && spcols)
      for (ff = 0; ff < sprows; ff++)
         for (zz = 0; zz < spcols; zz++)  spyrel[gg++] = sizes[ff];
   else
      for (ff = 0; ff < sprows; ff++)
         for (zz = 0; zz < sparr[ff]; zz++)   spyrel[gg++] = sizes[ff];
   yrelsum = totsize;

   if (rec && cmd)
   {
      cmd->store_new_values(spcnt,spyrel,yrelsum);
      if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }
   }
   compute_object_pos();
   for (ff = 0; ff < spcnt; ff++)
      if (spplot[ff].length() and plot.count(spplot[ff]))
      {
         rval = reposition_plot_area(plot[spplot[ff]], spleft[ff], spbottom[ff],
                                     spright[ff], sptop[ff], false);
         blank.set_plot_id("");
      }
   return rval;
}


int ChipsFrame::reference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_GOOD;
   string coord_id;
   COORD_SYS coord_sys;
   ChipsCoordinate *tmp_coord=NULL;
   if (!co)
      return ERR_STATUS_BAD;

   co->get_coord_system (coord_sys, coord_id); 
   if (coord)
   {

      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         tmp_coord = coord->get_coord(FRAME_NORM, "");
      else
         tmp_coord = coord->get_coord (coord_sys, coord_id);
   }

   switch (co->get_object_type ()){
      case chips_plot:
      {
         ChipsPlot *cp = static_cast <ChipsPlot *>(co);
         plot[cp->get_id ()] = cp;
         dplot.erase (co->get_ref_cnt ());
         if (cp->get_chips_coord ())
            coord->add_coord_sys (cp->get_chips_coord ());

         // the coordinate was never initialized all object position need to be
         // recalculated
         if (!(cp->get_coord_initialized ()))
         {
            cp->set_coord_initialized (true);
            cp->update_object_positions ();
            cp->reset_object_positions ();
         }
         if (change_cnt)  plot_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_plot(curr)))
            set_current_plot ( curr);
         else
            set_current_plot ( co->get_id ());
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"object");
      break; 
   }
   if (rval == ERR_STATUS_GOOD)
   {
      co->change_visible ();
      redraw ();
   }

   return rval;
}

int ChipsFrame::dereference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_GOOD;
   bool cmd_locked = cmd_manager->cmd_lock ();
   ChipsId cid;
   if (!co)
      return ERR_STATUS_BAD;

   cmd_manager->set_cmd_lock (true);

   switch (co->get_object_type ()){
      case chips_plot:
      {
         cid.set_plot_id (co->get_id ());
         rval = this->delete_plot (&cid, false);

         // if current has a length overide what ever delete plot
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_plot(curr)))
            set_current_plot ( curr);

         if (change_cnt)
            plot_cnt--;
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"object");
      break; 
   }
   cmd_manager->set_cmd_lock (cmd_locked);
   return rval;
}


int ChipsFrame::dereference_plot (ChipsPlot *cplot)
{
   ChipsCoordinate *plot_coord;
   int rval = ERR_STATUS_GOOD;

   // this hides the plot without changing its visibility
   cplot->force_hide ();
   dplot [cplot->get_ref_cnt ()] = cplot;

   // delete all objects associated with this plot. use the coordinate
   // system to find which objects these are
   plot_coord = cplot->get_chips_coord ();
   if (plot_coord)
   {
      list <string> coord_list;
      list <string>::iterator coord_ii;
      list <ChipsObject *> obj_list;
      list <ChipsObject *>::iterator obj_ii;
      ChipsCoordinate *data_coord;

      plot_coord->get_coord_list (coord_list);
      rval = ERR_STATUS_GOOD;
      for (coord_ii = coord_list.begin (); !rval && coord_ii != coord_list.end (); coord_ii++)
      {
         data_coord = plot_coord->get_coord (DATA, *coord_ii);
         if (data_coord)
            data_coord->get_object_list (obj_list);
         for (obj_ii = obj_list.begin (); !rval && obj_ii != obj_list.end (); obj_ii++)
         {
            V_OBJECT_TYPE otype = (*obj_ii)->get_object_type ();
            if ((otype != chips_curve) &&
                (otype != chips_histogram) &&
                (otype != chips_image) &&
                (otype != chips_contour) &&
                (otype != chips_axis))
               rval = this->delete_object ((*obj_ii)->get_id (), otype);
         }
         obj_list.clear ();
      }
      if (!rval)
      {
         plot_coord->get_object_list (obj_list);
         for (obj_ii = obj_list.begin (); obj_ii != obj_list.end (); obj_ii++)
         {
            V_OBJECT_TYPE otype = (*obj_ii)->get_object_type ();
            if ((otype != chips_curve) &&
                (otype != chips_histogram) &&
                (otype != chips_image) &&
                (otype != chips_contour) &&
                (otype != chips_axis))
               // don't capture the rval. plot title will return an error when
               // trying to delete it
               this->delete_object ((*obj_ii)->get_id (), otype);
         }
      }

      // remove the coordinate system 
      if (!rval)
         rval = coord->remove_coord_sys (plot_coord->get_id ());
   }

   return rval;
}

int ChipsFrame::clear_cmd (ChipsObject *co)
{
   int rval = ERR_STATUS_GOOD;
   if (!co)
      return ERR_STATUS_BAD;

   switch (co->get_object_type ())
   {
      case chips_plot:
      {
         map <unsigned int, ChipsPlot *>::iterator ii;
         ii = dplot.find (co->get_ref_cnt ());
         if (ii != dplot.end ())
         {
            ii->second->object_delete ();
            dplot.erase (ii);
         }
      }
      break;
      default:
        rval = ERR_STATUS_BAD;
      break;
   }

   return rval;
}

void ChipsFrame::serialize(ChipsStateStream& out)
{
   return this->serialize(out, false);
}

void ChipsFrame::serialize(ChipsStateStream& out, bool write_coord)
{
   string col = get_color();
   int sz;
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   
   out.write_double(&nx1);
   out.write_double(&ny1);
   out.write_double(&nx2);
   out.write_double(&ny2);
   out.write_bool(&opaque); 
   out.write_bool(&display_border); 
   out.write_double(&scale_factor);

   if (out.get_minor_version() > 1) 
   {
      unsigned int mask = 0;
      if (spleft)   mask |= 0x01;
      if (spbottom) mask |= 0x02;
      if (spright)  mask |= 0x04;
      if (sptop)    mask |= 0x08;
      if (spxrel)   mask |= 0x10;
      if (spyrel)   mask |= 0x20; 
      if (spplot)   mask |= 0x40;
      out.write_int(&spcnt);
      sz = (int) mask; 
      out.write_int(&sz);
      if ((mask & 0x01) == 0x01)   out.write_float(spleft, spcnt);
      if ((mask & 0x02) == 0x02)   out.write_float(spbottom, spcnt);
      if ((mask & 0x04) == 0x04)   out.write_float(spright, spcnt);
      if ((mask & 0x08) == 0x08)   out.write_float(sptop, spcnt);
      if ((mask & 0x10) == 0x10)   out.write_float(spxrel, spcnt);
      if ((mask & 0x20) == 0x20)   out.write_float(spyrel, spcnt);

      if ((mask & 0x40) == 0x40)
         for (sz = 0; sz < spcnt; sz++) out.write_string(&spplot[sz]);

      if (sprows > 0 && sprows < 100)  
      {
         out.write_int(&sprows);
      }
      else 
      {
         sz = 0;
         out.write_int(&sz);
      }
      if (spcols > 0 && spcols < 100)
      {
         out.write_int(&spcols);
      }
      else
      {
         sz = 0;
         out.write_int(&sz);
      }
      if (sparr)
      {
         if (sprows > 0 && sprows < 100)  
         {
            out.write_int(&sprows);
            out.write_int(sparr, sprows);
         }
         else if (spcols > 0 && spcols < 100)  
         {
            out.write_int(&spcols);
            out.write_int(sparr, spcols);
         }
         else 
         {
            sz = 0;
            out.write_int(&sz);
         }
      }
      else 
      {
         sz = 0;
         out.write_int(&sz);
      }
      out.write_float(&xrelsum);
      out.write_float(&yrelsum);
   }
   out.write_string(&col);
   col = get_border_color();
   out.write_string(&col);
   out.write_bool(&visible);
   if (!coord) write_coord = false;
   out.write_bool (&write_coord);
   out.write_int(&plot_cnt);
   
   if (out.get_minor_version() > 6)
      out.write_string (&this->data_system_id);

   // This must always be the last thing serialized. Gets deserialized
   // in constructor.
   if (write_coord)  coord->serialize (out);

   // serialize objects- since the component portions of plots
   // (border axes) are created in the plot, but we want to account
   // for depths, we'll re-add all objects then resort them to the
   // appropriate order in the depth table
   sz = plot.size();
   out.write_int(&sz);
   out.write_int(&sz);
   map <int, string>chron;
   for(map<string, ChipsPlot*>::iterator pi=plot.begin();pi != plot.end(); ++pi)
      chron.insert(pair<int, string>((*pi).second->get_ref_cnt(), (*pi).first));
   for(map <int, string>::iterator i2 = chron.begin(); i2 != chron.end(); i2++)
      (plot[(*i2).second])->serialize(out);

   sz = dplot.size();
   out.write_int(&sz);
   map<unsigned int, ChipsPlot*>::iterator pj;
   for(pj=dplot.begin(); pj != dplot.end(); ++pj)
      pj->second->serialize(out, true);

   serialize_objects(out);   // now serialize other objects
}

void ChipsFrame::serialize_ascii(ChipsStateStream& out, int keep)
{
   map <int, string>chron;
   string color1 = get_color();
   string color2 = get_border_color();
   string tmpid = theDefaults->get_framestem();
   string att = "", pos = "";
   if (nx1 != 0 || ny1 != 0 || nx2 != 1.0 || ny2 != 1.0)
   {
      char buf[256];
      sprintf(buf,"%3.2f,%3.2f,%3.2f,%3.2f,", nx1,ny1,nx2,ny2);
      pos = buf;
   }
   if (theDefaults->get_framecolor() != color1)
   {
      att = "bgcolor=";
      att.append(color1);
   }
   if (theDefaults->get_framebordercolor() != color2)
   {
      if (att.length()) att.append(" ");
      att = "border.color=";
      att.append(color2);
   }
   if (theDefaults->get_framebordervisible() != display_border)
   {
      if (att.length()) att.append(" ");
      if (display_border)  att.append("border.visible=1");
      else                 att.append("border.visible=0");
   }
   if (theDefaults->get_frametransparent() == opaque)
   {
      if (att.length()) att.append(" ");
      if (opaque)  att.append("transparency=0");
      else         att.append("transparency=1");
   }
   if (att.length())   att.append(" ");
   if (att == "" && pos == "")
   {
      if (id.find(tmpid) != 0)        out << "add_frame(id=\""<< id <<"\")"<<endl;
         else if (plot.size() < 1 || keep)   out << "add_frame()" << endl;
   }
   else   out << "add_frame(" << pos << "\"" << att << "id=" << id <<"\")"<<endl;
   keep = 0;
   for(map<string,ChipsPlot*>::iterator pi=plot.begin();pi!=plot.end();++pi)
      chron.insert(pair<int, string>((*pi).second->get_ref_cnt(),(*pi).first));
   for(map <int, string>::iterator i2=chron.begin();i2!=chron.end();i2++)
   {
      (plot[(*i2).second])->serialize_ascii(out, keep++);
      asf_plot = (plot[(*i2).second])->get_id(); 
      // We need to handle OFFSCREEN plots- via split- incl. swap_obj_pos
      // and resizes- for now simply hide the plot
      if ((plot[(*i2).second])->is_offscreen())
      {
         if (sprows && spcols)
            out << "hide_plot(\""<<(plot[(*i2).second])->get_id()<<"\")   ## offscreen plot from split("<<sprows<<","<<spcols<<")" << endl;
         else
            out << "hide_plot(\""<<(plot[(*i2).second])->get_id()<<"\")   ## offscreen plot from split" << endl;
      }
   }
   serialize_objects(out);   // now serialize other objects
}


void ChipsFrame::serialize_objects(ChipsStateStream& out)
{
   if (out.is_binary())
   {
      if (depth_table)   depth_table->serialize(out);    
      
      out.write_checkpoint();
   
      // store off bind-table so axes are synched
      if (bind_table)  bind_table->serialize(out); 

      // store off object currency
      out.write_string(&curr_plot);
   }
   else
   {
      if (depth_table && !plot.empty()) 
         depth_table->serialize_ascii(out, asf_plot);
      if (bind_table)  bind_table->serialize_ascii(out);

      // now that we've output the depth table, we can set up our currency
      if (curr_plot.length()) 
      {
         for(map<string,ChipsPlot*>::iterator pi=plot.begin();pi!=plot.end();++pi)
         {
            // write 'current' (non-all) plot last to set currency
            if ((*pi).first != curr_plot)
            {
               (*pi).second->serialize_currency_ascii(out, plot.size());
            }
         }
         if (plot.count(curr_plot))
            plot[curr_plot]->serialize_currency_ascii(out, plot.size());
      }
   }

}

void ChipsFrame::deserialize(ChipsStateStream& in)
{
   string cname;
   int len;
   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;

   if (in.get_minor_version() < 13)
   {
      float fval;
      in.read_float(&fval);
      nx1 = fval;
      in.read_float(&fval);
      ny1 = fval;
      in.read_float(&fval);
      nx2 = fval;
      in.read_float(&fval);
      ny2 = fval;
   }
   else
   {
      in.read_double(&nx1);
      in.read_double(&ny1);
      in.read_double(&nx2);
      in.read_double(&ny2);
   }

   in.read_bool(&opaque);
   in.read_bool(&display_border);
   in.read_double(&scale_factor);

   if (in.get_minor_version() > 1)
   {
      unsigned int mask=0; 
      int rr; 
      in.read_int(&spcnt);
      in.read_int(&len);
      if (len > 0) mask = (unsigned int) len; 
      delete[] spleft;
      delete[] spbottom;
      delete[] spright;
      delete[] sptop;
      delete[] spxrel;
      delete[] spyrel;
      delete[] spplot;
      spleft = spbottom = spright = sptop = spxrel = spyrel = NULL; 
      spplot = NULL; 

      if (in.get_major_version() < 3 ||
          (in.get_major_version() == 3 && in.get_minor_version() < 16))
      {
         double *tmpd = new double[spcnt];
         if ((mask & 0x01) == 0x01)
         {
            spleft = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  spleft[rr] = tmpd[rr];
         }
         if ((mask & 0x02) == 0x02)
         {
            spbottom = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  spbottom[rr] = tmpd[rr];
         }
         if ((mask & 0x04) == 0x04)
         {
            spright = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  spright[rr] = tmpd[rr];
         }
         if ((mask & 0x08) == 0x08)
         {
            sptop = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  sptop[rr] = tmpd[rr];
         }
         if ((mask & 0x10) == 0x10)
         {
            spxrel = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  spxrel[rr] = tmpd[rr];
         }
         if ((mask & 0x20) == 0x20)
         {
            spyrel = new float[spcnt];
            in.read_double(tmpd, spcnt);
            for (rr = spcnt;rr--;)  spyrel[rr] = tmpd[rr];
         }
         delete[] tmpd;
      }
      else
      {
         if ((mask & 0x01) == 0x01)
         {
            spleft = new float[spcnt];
            in.read_float(spleft, spcnt);
         }  
         if ((mask & 0x02) == 0x02)
         {
            spbottom = new float[spcnt];
            in.read_float(spbottom, spcnt);
         }
         if ((mask & 0x04) == 0x04)
         {
            spright = new float[spcnt];
            in.read_float(spright, spcnt);
         }
         if ((mask & 0x08) == 0x08)
         {
            sptop = new float[spcnt];
            in.read_float(sptop, spcnt);
         }
         if ((mask & 0x10) == 0x10)
         {
            spxrel = new float[spcnt];
            in.read_float(spxrel, spcnt);
         }
         if ((mask & 0x20) == 0x20)
         {
            spyrel = new float[spcnt];
            in.read_float(spyrel, spcnt);
         }
      }
      if ((mask & 0x40) == 0x40)
      {
         spplot = new string[spcnt];
         for (len = 0; len < spcnt; len++) in.read_string(&spplot[len]);
      }

      in.read_int(&sprows);
      in.read_int(&spcols);
      in.read_int(&len);
      if (len > 0 && len < 100)
      {
         if (sparr)  delete[] sparr;
         sparr = new int[len];
         in.read_int(sparr, len);
      }

      if (in.get_major_version() < 3 ||
          (in.get_major_version() == 3 && in.get_minor_version() < 16))
      {
         double tmp_relsum;
         in.read_double(&tmp_relsum);
         xrelsum = tmp_relsum;
         in.read_double(&tmp_relsum);
         yrelsum = tmp_relsum;
      }
      else
      {
         in.read_float(&xrelsum); 
         in.read_float(&yrelsum); 
      }
   }

   in.read_string(&cname);
   set_color(cname); 
   in.read_string(&cname);
   set_border_color(cname); 
   in.read_bool(&visible);
   in.read_bool (&deserialize_coord);
   in.read_int(&plot_cnt);

   if (in.get_minor_version() > 6)
      in.read_string (&this->data_system_id);

   reposition(nx1, ny1, nx2, ny2, false);
}


int ChipsFrame::deserialize_bind_table(ChipsStateStream& in)
{
   int sz, rr, len;
   string plid1, aid1, plid2, aid2;
   ChipsPlot *p1, *p2;
   ChipsAxis *a1, *a2;

   in.read_int(&sz);
   in.read_int(&len);
   if (sz != len)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = sz; rr--; )
   {
      in.read_string(&plid1);
      in.read_string(&aid1);
      in.read_string(&plid2);
      in.read_string(&aid2);

      if (plot.count(plid1) && plot.count(plid2))
      {
         p1 = plot[plid1];
         p2 = plot[plid2];

         if ((a1=p1->get_xaxis(aid1)) == NULL) a1 = p1->get_yaxis(aid1);  
         if ((a2=p2->get_xaxis(aid2)) == NULL) a2 = p2->get_yaxis(aid2);  
         if (a1 && a2)  
         {
            p1->remove_box_axes(a1);
            if (p1 != p2)  p2->remove_box_axes(a2);
            bind_table->add_binding(a1, a2);
//	    bind_table->synch_axes(a1, a2);
            // this will hide 'covered' border axes 
            p1->set_axis_style(p1->get_axis_style()); 
            if (p1 != p2)   p2->set_axis_style(p1->get_axis_style()); 
         }
      }
   }
   if (in.get_minor_version () > 10)
      bind_table->deserialize (this,in);
   return ERR_STATUS_GOOD;
}

int ChipsFrame::deserialize_plot(ChipsStateStream& in)
{
   int ival, ival2, rval = ERR_STATUS_GOOD;

   in.read_int(&ival);     // get plot map size
   in.read_int(&ival2);

   if (ival == ival2)
   {
      ChipsPlot* plt;
      for (int rr=0; rr< ival2 && rval == ERR_STATUS_GOOD; rr++)
      {
         plt = new_plot(in);
         if (plt != NULL)
         {
            plot.insert(pair<string, ChipsPlot*>(plt->get_id (), plt));
            set_current_plot ( plt->get_id());

            rval = plt->deserialize_axis(in);
            if (in.get_minor_version() > 0)
               if (!rval)   rval = plt->deserialize_image(in);
            if (!rval)   rval = plt->deserialize_curve(in);
            if (!rval)   rval = plt->deserialize_contour(in);
            if (!rval)   rval = plt->deserialize_histogram(in);
	    if (!rval)   rval = plt->deserialize_annotations(in); 
         }
      }

      in.read_int(&ival);     // get plot map size
      for (int rr=0; rr< ival && rval == ERR_STATUS_GOOD; rr++)
      {
         plt = new_plot(in);
         if (plt != NULL)
         {
            dereference_plot (plt);
            if (!rval)   rval = plt->deserialize_axis(in);
            if (in.get_minor_version() > 0)
               if (!rval)   rval = plt->deserialize_image(in);
            if (!rval)   rval = plt->deserialize_curve(in);
            if (!rval)   rval = plt->deserialize_contour(in);
            if (!rval)   rval = plt->deserialize_histogram(in);
	    if (!rval)   rval = plt->deserialize_annotations(in); 
         }
      }
   }
   return rval; 
}


int ChipsFrame::deserialize_objects(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;

   if ((rval = deserialize_depth_table(in)) == ERR_STATUS_GOOD)
   {
      // need to pass depth table to each existing plot
      map<string, ChipsPlot*>::iterator hh;
      for(hh=plot.begin(); hh != plot.end(); ++hh)
         (*hh).second->load_depth_table(depth_table);

      // need to pass depth table to deleted plots
      map<unsigned int, ChipsPlot*>::iterator ii;
      for(ii=dplot.begin(); ii != dplot.end(); ++ii)
         (*ii).second->load_depth_table(depth_table);

      // get checkpoint and bind table
      if (in.read_checkpoint() != ERR_STATUS_GOOD)  return ERR_STATEFILE_INC;
         if ((rval=deserialize_bind_table(in)) != ERR_STATUS_GOOD) return rval; 

      in.read_string(&curr_plot);
   }
   return rval; 
}


ChipsPlot* ChipsFrame::new_plot(string plid)
{
   string tmpid;

   if (plid.length())
      tmpid = plid;
   else
      tmpid = create_plot_id();
   return new ChipsPlot(this, tmpid);
}


ChipsPlot* ChipsFrame::new_plot(ChipsStateStream& in)
{
   return new ChipsPlot(this, in);
}


ChipsPlot* ChipsFrame::new_plot(float posx1, float posy1, float posx2,
   float posy2, string plid)
{
   string tmpid;

   if (plid.length())
      tmpid = plid;
   else
      tmpid = create_plot_id();
   return new ChipsPlot(this, posx1, posy1, posx2, posy2, tmpid);
}


char* ChipsFrame::create_plot_id()
{
   static char temp_id[80];
   ++plot_cnt;
   if (theDefaults)
   {
      string stem = theDefaults->get_plotstem();
      sprintf(temp_id, "%s%1d", stem.c_str(), plot_cnt);
   }
   else
      sprintf(temp_id, "plt%1d", plot_cnt);
   return temp_id;
}


string ChipsFrame::get_default_plot_id()
{
   string tmpid;
   tmpid = theDefaults->get_plotid();
   return tmpid;
}


ChipsPlot* ChipsFrame::get_plot(const string& val)
{
   string plid;
   if (val.length())   plid = val;
      else    plid = curr_plot;
   if ((chips_strcmp(plid.c_str(), "all") == 0) && plot.size() == 1)
   {
      map<string, ChipsPlot*>::iterator ii = plot.begin();
      return ((*ii).second);
   }
   else if ((chips_strcmp(plid.c_str(), "all") != 0) && plot.count(plid))
      return plot[plid];
   else
      return NULL;
}

int ChipsFrame::display_depth(int depth, bool rec)
{
   int rval = ERR_STATUS_BAD;
   if (depth_table)
   {
      if (rec)
      {
         ChipsDepthVisCommand* cmd = new ChipsDepthVisCommand(this, depth, true);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      rval =  depth_table->display_depth(depth);
   }
   return rval;
}


int ChipsFrame::hide_depth(int depth, bool rec)
{
   int rval = ERR_STATUS_BAD;
   if (depth_table)
   {
      if (rec)
      {
         ChipsDepthVisCommand* cmd = new ChipsDepthVisCommand(this, depth, false);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      rval =  depth_table->hide_depth(depth);
   }
   return rval;
}


int ChipsFrame::add_object_at_depth(ChipsObject* obj)
{
   if (depth_table)   depth_table->add_object(obj);
   return ERR_STATUS_GOOD;
}


int ChipsFrame::add_object_to_depth_table(ChipsObject* obj)
{
   // add entry to depth table and set up undo/redo command
   if (depth_table)
   {
      ChipsDepthUpdateCommand* cmd =
         new ChipsDepthUpdateCommand(this, obj, 0);
      cmd_manager->add_cmd(cmd);
      depth_table->add_object(obj);
      return ERR_STATUS_GOOD;
   }
   return ERR_STATUS_BAD; 
}


int ChipsFrame::remove_object_from_depth_table(ChipsObject* obj)
{
   // remove entry from depth table and set up undo/redo command
   if (depth_table && obj)
   {
      int posn = depth_table->total_objects_in_depth(obj->get_depth()) - 
         depth_table->position_in_depth(obj->get_ref_cnt(), obj->get_depth());
      ChipsDepthUpdateCommand* cmd = new ChipsDepthUpdateCommand(this, obj,
         1, posn);
      cmd_manager->add_cmd(cmd);
      depth_table->remove_object(obj);
      return ERR_STATUS_GOOD;
   }
   return ERR_STATUS_BAD; 
}


int ChipsFrame::collapse_depths(int* sources, int src_cnt, int dest, bool rec)
{
   int rval;
   ChipsDepthCollapseCommand* cmd = NULL;
   if (depth_table)
   {
      if (rec)
         cmd = new ChipsDepthCollapseCommand(this, dest, sources, src_cnt);

      rval = depth_table->collapse_depths(sources, src_cnt, dest, cmd);
      if (rval == ERR_STATUS_GOOD)
      {
//         depth_table->sort_check();
         if (rec)
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
      }
      else
         delete cmd;

      return rval; 
   }
   else 
      return ERR_STATUS_BAD;
}

int ChipsFrame::uncollapse_depths(multimap<int, int> &dest_table, int src_depth)
{
   if (depth_table)
   {
      int rval = depth_table->uncollapse_depths(dest_table, src_depth);
//      if (rval == ERR_STATUS_GOOD)  depth_table->sort_check();
      return rval;
   }
   else
      return ERR_STATUS_BAD;
}


int ChipsFrame::get_depth()
{
   if (depth_table)
      return depth_table->get_default_depth();
   else if (theDefaults)
      return theDefaults->get_defaultdepth();
   else
      return V_DEFAULT_DEPTH; 
}


int ChipsFrame::rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   string oid; 
   ChipsDepthArrangeCommand* cmd = NULL;

   switch(obj)
   {
      case chips_label:        // fallthrough intended
      case chips_line:         // fallthrough intended
      case chips_point:        // fallthrough intended
      case chips_region:       // fallthrough intended
      case chips_axis:         // fallthrough intended
      case chips_curve:        // fallthrpugh intended
      case chips_contour:      //fallthrough intended 
      case chips_image:        //fallthrough intended 
      case chips_colorbar:     // fallthrough intended
      case chips_histogram:
      {
         oid = cid->get_plot_id();
         if (oid.length() == 0)  oid = curr_plot;
         if (!chips_strcmp(oid.c_str(), "all") && plot.size() != 1)  
         {
            cmd_manager->add_err(ERR_INVALID_ALL_USE,id,0);
            return ERR_INVALID_ALL_USE;
         }
         if (chips_strcmp(oid.c_str(), "all") != 0 && plot.count(oid) == 0) 
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         set_current_plot ( oid);
         ChipsPlot* cp = get_plot(oid); 
         return cp->rearrange_object(cid, obj, manner, rec);
      }
      break; 
      default:
         cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"object", obj,
            "chips_label|chips_line|chips_point|chips_region|chips_axis|chips_curve|chips_contour|chips_histogram|chips_image|chips_colorbar");
      break; 
   }

   if (rec && cmd)
   {
      if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }
   }

   return rval;
}

int ChipsFrame::update_depth_table(V_OBJECT_TYPE obj, string oid, int how,
   int pos)
{
   // this routine is used for updating the depth table during undo/redo of
   // object set depth calls 
   if (!depth_table)  return ERR_STATUS_BAD;   // no depth table to update
   if (obj != chips_plot)  return ERR_STATUS_BAD;
   return ERR_STATUS_GOOD;
}


ChipsCoordinate *ChipsFrame::find_coordinate_system (ChipsId &cid, int &err_status)
{
   COORD_SYS coord_sys = cid.get_coord_sys ();
   ChipsCoordinate *outCoord = static_cast <ChipsCoordinate *>(0);

   err_status = 0;
   if (coord_sys == COORD_NONE){
      err_status = get_current_coord (&cid);
      coord_sys = cid.get_coord_sys ();
   }

   if (((coord_sys == DATA) || (coord_sys == PLOT_NORM)) && !err_status)
   {
      ChipsPlot *plt = static_cast <ChipsPlot *>(0);

      string plt_id = cid.get_plot_id ();
      if (plt_id.length () != 0) 
         plt = get_plot (plt_id);
      else if (curr_plot.length () != 0)
         plt = get_plot (curr_plot);

      if (!plt){
         err_status = ERR_OBJ_DNE;
         return outCoord;
      }

      outCoord = plt->find_coordinate_system (cid, err_status);
   }
   else if (!err_status){
      outCoord = coord;
      if (outCoord && outCoord->coord_exists (coord_sys, ""))
         err_status = ERR_STATUS_GOOD;
      else
         cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));
   }
   return outCoord;
}

int  ChipsFrame::deserialize_depth_table(ChipsStateStream& in)
{
   return ERR_FEATURE_DNE; 
}

int ChipsFrame::show ()
{
   int rval = ERR_STATUS_GOOD;
   visible = true;

   if (coord)   coord->set_visible (true);

   map<string, ChipsPlot*>::iterator plot_ii;
   for (plot_ii =plot.begin(); !rval && plot_ii != plot.end(); ++plot_ii)
      rval = plot_ii->second->change_visible ();

   if (!rval)  rval = change_visible ();
   return rval;
}

int ChipsFrame::force_show ()
{
   // Only show the frame if frame is visible
   if (!visible)  return ERR_STATUS_GOOD;
   return show ();
}


int ChipsFrame::hide ()
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsPlot*>::iterator plot_ii;

   visible = false;
   if (coord)  coord->set_visible(false);

   for (plot_ii =plot.begin(); !rval && plot_ii != plot.end(); ++plot_ii)
      rval = plot_ii->second->change_visible ();

   if (!rval)  rval = change_visible ();
   return rval;
}


int ChipsFrame::clear_plot (ChipsId* cid)
{
   int  rval = ERR_STATUS_GOOD;
   string plid = cid->get_plot_id();
   if (plid.length() == 0)
      plid = curr_plot;
   else if (plid != curr_plot)
   {
      ChipsCurrentCommand *cmd = 
         new ChipsCurrentCommand(this, curr_plot, plid, chips_plot);
      if ((rval = set_current_plot(plid)) == ERR_STATUS_GOOD)
      {
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      else delete cmd;
   }

   if (!chips_strcmp(plid.c_str(), "all"))
   {
      map <string, ChipsPlot*>::iterator hh;
      int sts, pass = 0, failval = 0;
      for(hh=plot.begin(); hh != plot.end(); ++hh)
         if ((sts=(*hh).second->clear ()) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
   }
   else if (plot.count(plid))
   {
      rval = plot[plid]->clear ();
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"plot",plid.c_str());
   return rval;
}


int ChipsFrame::check_plot_updates ()
{
   int rval = ERR_STATUS_GOOD;

   // update the limits across plots
   list <string> plot_ids;
   list <string> axis_ids;
   list <string>::iterator ii;
   ChipsPlot *tmp_plot;

   int updates_recorded = bind_table->get_updates_recorded ();

   bind_table->get_recorded_plots (plot_ids);
   for (ii=plot_ids.begin (); ii!=plot_ids.end (); ii++)
   {
      tmp_plot = get_plot (*ii);

      if (!tmp_plot)
      {
         rval = ERR_STATUS_BAD;
         break;
      }
      bind_table->get_recorded_axes (*ii, axis_ids);

      if (updates_recorded & 1)
      {
         // limits were updated in the bind table
         tmp_plot->update_limits (axis_ids);
      }
      // a property was updated in bind table 
      if (updates_recorded & 2)
      {
         int property = bind_table->get_property_recorded ();
         tmp_plot->update_property (axis_ids, property);
      }
   }

   return rval;
}

int ChipsFrame::reset_relsizes(int count, int dimn, float *vals, float relsum)
{
   int rr;
   if (dimn == X_AXIS)
   {
      if (count != spcnt)
      {
         delete [] spxrel;
         spxrel = spxrel = new float[count];
      }
      for (rr = count; rr--; )   spxrel[rr] = vals[rr];
      xrelsum = relsum;
   }
   else
   {
      if (count != spcnt)
      {
         delete [] spyrel;
         spyrel = spyrel = new float[count];
      }
      for (rr = count; rr--; )   spyrel[rr] = vals[rr];
      yrelsum = relsum;
   }
   return ERR_STATUS_GOOD;
}

int ChipsFrame::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj, bool vis)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_plot_id();
   ChipsPlot* tmp = NULL;
   map<string, ChipsPlot*>::iterator ii;

   // make sure we have a plot to work on
   if (!tmpid.length()) tmpid = curr_plot;
   if (!tmpid.length() || plot.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"plots");
      return rval;
   }
   if ((tmp = get_plot(tmpid)) == NULL)
   {
      if (chips_strcmp(tmpid.c_str(), "all") != 0)
      {
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"plot",tmpid.c_str());
         return rval;
      }
   }
   if (tmp)
   {
      // single plot
      if (obj == chips_plot)
      {
         if (vis)  tmp->force_show();
            else   tmp->force_hide();
      }
      else
         rval = tmp->highlight_object(cid, obj, vis); 
   }
   else
   {
      // highlight all plots
      if (obj == chips_plot)
      {
         for(ii=plot.begin(); ii != plot.end(); ++ii)
         {
            if (vis)  ((*ii).second)->force_show();
               else   ((*ii).second)->force_hide();
         }
      }
      else 
      {
         for(ii=plot.begin(); ii != plot.end(); ++ii)
         {
            ((*ii).second)->highlight_object(cid, obj, vis);
         }
      }
   }
   return rval;
}

void ChipsFrame::enable_picklimits (bool flag)
{
   ChipsPlot* cp = NULL;
   map<string, ChipsPlot*>::iterator ii;

   // enable the current children
   if (chips_strcmp(curr_plot.c_str(), "all") == 0)
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         ((*ii).second)->enable_picklimits(flag);
   else
   {
      cp = get_plot(curr_plot);
      if (cp)
         cp->enable_picklimits (flag);
   }
}

void ChipsFrame::disable_all_interaction ()
{
   this->disable_interaction ();
   this->disable_plot_interaction ("all");
}

void ChipsFrame::enable_all_interaction ()
{
   this->enable_interaction ();
   this->enable_plot_interaction ("all");
}


void ChipsFrame::disable_plot_interaction (const string &plot_id)
{
   if (chips_strcmp(plot_id.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         ii->second->disable_all_interaction ();
   }
   else if (plot.count (plot_id))
      plot[plot_id]->disable_all_interaction ();
}

void ChipsFrame::enable_plot_interaction (const string &plot_id)
{
   if (chips_strcmp(plot_id.c_str(), "all") == 0)
   {
      map<string, ChipsPlot*>::iterator ii;
      for(ii=plot.begin(); ii != plot.end(); ++ii)
         ii->second->enable_all_interaction ();
   }
   else if (plot.count (plot_id))
      plot[plot_id]->enable_all_interaction ();
}

int ChipsFrame::get_plots (const ChipsId *cid, list <ChipsPlot *>&out_plots, bool allow_all)
{
   int  rval = ERR_STATUS_GOOD;
   ChipsPlot *cp;
   string id = cid->get_plot_id();
   if (id.length() == 0)  id = curr_plot;

   if (rval == ERR_STATUS_GOOD)
   {
      if (plot.empty())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"plots");
      else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
      {
         // get plot
         if ((cp = get_plot(id)) != NULL)
            out_plots.push_back (cp);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"plot",id.c_str());
      }
      else if (allow_all)
      {
         map<string, ChipsPlot*>::iterator ii;

         for(ii=plot.begin(); ii != plot.end(); ++ii)
            out_plots.push_back (ii->second);
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      }
      else
         cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
   }
   else
      cmd_manager->add_err(rval, id);
   return rval;
}



