// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsPlot.hh"
#include "ChipsPoint.hh"
#include "ChipsImage.hh"
#include "ChipsColorBar.hh"
#include "ChipsPolygon.hh"
#include "ChipsLine.hh"
#include "ChipsCmdManager.hh"
#include "ChipsPlotSystem.hh"
#include "ChipsDepthArrangeCommand.hh"
#include "ChipsCommand.hh"
#include "ChipsDepthUpdateCommand.hh"
#include "ChipsDataAspectCommand.hh"
#include "ChipsDeleteCommand.hh"
#include "ChipsTransformFactory.hh"
#include "ChipsAxisAlgorithmFactory.hh"

#include "Transform.hh"

#include <stdio.h>
#include <float.h>
#include <config.h>
#include <sstream>
#include <algorithm>

#ifdef HAVE_IEEEFP_H
#include <ieeefp.h>        // finite defined in ieeefp.h (solaris)
#else
#include <math.h>          // finite() defined here      (linux, OSX)
#endif

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;
typedef map <string, ChipsCurve* >curves_map;         
typedef map <string, ChipsContour* >contours_map;    
typedef map <string, ChipsHistogram* >histograms_map; 
typedef map <string, ChipsAxis* >axes_map;  
typedef map <string, ChipsImage* >images_map;  
typedef map <string, ChipsLabel* >labels_map;  
typedef map <string, ChipsLine* >lines_map;  
typedef map <string, ChipsRegion* >regions_map;  
typedef map <string, ChipsColorBar* >cbars_map;  
typedef map <string, ChipsPoint* >points_map;  

ChipsPlot::ChipsPlot(ChipsObject* par, const string& val):ChipsObject(par, val)
{
   initialized = false;
   id = val; 
   nx1 = ny1 = 0.0;
   nx2 = ny2 = 1.0;
   initialize(); 
   adjust_plot_anchors(); 
   update_title_position(false);
}


ChipsPlot::ChipsPlot(ChipsObject* par, ChipsStateStream& in):ChipsObject(par, "")
{
   initialized = false;
   nx1 = ny1 = 0.0;
   nx2 = ny2 = 1.0;
   initialize(false);
   deserialize(in);
   // All coordinates are created at the window level. Find the correct
   // coordinate system and use it
   if (deserialize_coord)
      coord = new ChipsPlotSystem (in, NULL);
   else if (parent && parent->get_chips_coord ())
   {
      if (in.get_minor_version() < 7)
      {
         coord = parent->get_chips_coord ()->get_coord (PLOT_NORM, id);
         this->data_system_id = coord->get_id ();
      }
      else
         coord = parent->get_chips_coord ()->get_coord (PLOT_NORM, this->data_system_id);
   }

  
   adjust_plot_anchors(); 
   update_title_position(false);
   if (!visible)
      hide (); 
}


ChipsPlot::ChipsPlot(ChipsObject* par, float x1n, float y1n, float x2n,
   float y2n, const string& val): ChipsObject(par, val)
{
   initialized = false;
   id = val; 

   // utilize the input coordinates as the corners of the plot
   // canvas where the actual curves are drawn 
   nx1 = 0;
   ny1 = 0;
   nx2 = 1;
   ny2 = 1;
   initialize();
   left_margin = x1n;
   bottom_margin = y1n;
   right_margin = 1 - x2n;
   top_margin = 1 - y2n;
   adjust_plot_anchors(); 
   update_title_position(false);
}


ChipsPlot::~ChipsPlot()
{
   if (legend)  delete legend; 
   map <string, ChipsAxis* >::iterator ax_iter;
   map <unsigned int, ChipsAxis* >::iterator dax_iter;
   map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;


   ChipsId cid;
   cid.set_curve_id("all");
   cid.set_contour_id("all");
   cid.set_histogram_id("all");
   cid.set_xaxis_id("all");
   cid.set_yaxis_id("all");
   cid.set_image_id("all"); 
   cid.set_label_id("all");
   cid.set_line_id("all");
   cid.set_point_id("all");
   cid.set_colorbar_id("all"); 
   cid.set_region_id("all");

   // if we have objects (or soft deleted objects), call appropriate deletes
   if (!(lines.empty() && dlines.empty()))       delete_line(&cid, true);
   if (!(images.empty() && dimages.empty()))     delete_image(&cid, true);
   if (!(labels.empty() && dlabels.empty()))     delete_label(&cid, true);
   if (!(points.empty() && dpoints.empty()))     delete_point(&cid, true);
   if (!(cbars.empty() && dcbars.empty()))       delete_colorbar(&cid, true); 
   if (!(regions.empty() && dregions.empty()))   delete_region(&cid, true);
   if (!(curves.empty() && dcurves.empty()))     delete_curve (&cid, true);
   if (!(contours.empty() && dcontours.empty())) delete_contour (&cid, true);
   if (!(histograms.empty() && dhistograms.empty())) delete_histogram (&cid, true);

   // have to delete manually because as border axes get delete
   // the delete_axis call has issues
   for(ax_iter=axes_x.begin();ax_iter!=axes_x.end(); ++ax_iter)
   {
      if (depth_table)   depth_table->remove_object(ax_iter->second);
      if (bindings)      bindings->remove_axis(ax_iter->second);
      ax_iter->second->object_delete ();
   };
   for(ax_iter=axes_y.begin();ax_iter!=axes_y.end(); ++ax_iter)
   {
      if (depth_table)   depth_table->remove_object(ax_iter->second);
      if (bindings)      bindings->remove_axis(ax_iter->second);
      ax_iter->second->object_delete ();
   };

   for(dax_iter=daxes_x.begin();dax_iter!=daxes_x.end(); ++dax_iter)
   {
      if (depth_table)   depth_table->remove_object(dax_iter->second);
      dax_iter->second->object_delete ();
   }
   for(dax_iter=daxes_y.begin();dax_iter!=daxes_y.end(); ++dax_iter)
   {
      if (depth_table)   depth_table->remove_object(dax_iter->second);
      dax_iter->second->object_delete ();
   }
   for(ds_iter=deleted_data_systems.begin();ds_iter!=deleted_data_systems.end(); ++ds_iter)
      delete ds_iter->second;
   deleted_data_systems.clear ();

   if (coord){
      ((ChipsPlotSystem *)coord)->delete_all_data_systems ();
      delete coord; 
   }

}

void ChipsPlot::object_delete ()
{
   // remove the title from the depth table
   if (depth_table && title)
      depth_table->remove_object(title);
   delete this;
}


void ChipsPlot::initialize(bool create_chipscoord)
{
   px0 = py0 = px1 = py1 = 0;
   title_halign = 0.5; 
   title_valign = 0.0; 
   bottom = top = left = right = NULL;
   corner_style = chips_miter;
   aspect_use = false;
   aspect_w = aspect_h = 0.0;
   aspect_ratio = ""; 

   if (theDefaults)
   {
      corner_style = theDefaults->get_cornerstyle();
      left_margin  = theDefaults->get_plotleftmargin(); 
      right_margin = theDefaults->get_plotrightmargin(); 
      top_margin   = theDefaults->get_plottopmargin(); 
      bottom_margin= theDefaults->get_plotbottommargin(); 
      axis_style   = theDefaults->get_axisstyle(); 
      title_depth  = theDefaults->get_plottitledepth();
      title_xpos   = theDefaults->get_plottitlexpos();
      title_ypos   = theDefaults->get_plottitleypos();
      title_angle  =  theDefaults->get_plottitleangle();
      title_halign =  theDefaults->get_plottitlehalign();
      title_valign =  theDefaults->get_plottitlevalign();
      title_size   = theDefaults->get_plottitlesize();
      title_color  = theDefaults->get_plottitlecolor();
      title_font   = theDefaults->get_plottitlefont();
      title_fontstyle = theDefaults->get_plottitlefontstyle();
      aspect_w     = theDefaults->get_plotaspect_width();     
      aspect_h     = theDefaults->get_plotaspect_height();
      aspect_ratio = theDefaults->get_plotratio(); 
   }
   else 
   {
      left_margin=0.2;
      right_margin=0.2;
      top_margin=0.2;
      bottom_margin=0.2;
      axis_style = chips_open; 
      title_depth = V_USE_DEFAULT_DEPTH;
      title_angle = 0.0;
      title_xpos = 0.45;
      title_ypos = 15;
      title_halign = title_halign = AUTO_ALIGN;
      title_size = 12;
      title_color = "white";
      title_font = "helvetica";
      title_fontstyle = chips_normal;
   }
   coord_system = FRAME_NORM; 
   object_type = chips_plot;
   title = NULL; 
   currline = currlabel = currcurv = currcont = currhist = currpoint = 
      currimage = currregion = currcbar = "";
   cur_x = cur_y = NULL;
   cur_xall = cur_yall = false; 
   legend = NULL; 
   curve_cnt = contour_cnt = histogram_cnt = 0;
   image_cnt = line_cnt = label_cnt = point_cnt = region_cnt = 0; 
   axis_x_cnt = axis_y_cnt = legend_cnt = cbar_cnt = 0;
   bindings = get_bind_table();
   depth_table = NULL;
   offscreen = false;

   set_aspect_ratio(aspect_ratio); 

   // Initialize plot coordinate system 
   // Use parent id because it will be unique to the system. Plot ids may
   //  not be unique
   if (create_chipscoord){
      coord = new ChipsPlotSystem (parent->get_chips_coord ());
      this->data_system_id = coord->get_id ();
      adjust_plot_anchors(); 
      coord_initialized = true;
   }
   else
   {
      coord = NULL;
      coord_initialized = false;
   }
   initialized = true;
}

void ChipsPlot::post_initialize ()
{
   initialize_title();
   if (depth_table && title)
      depth_table->add_object(title);
   add_border_axes();
}

int ChipsPlot::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj,
   int& out_val)
{
   int rval = ERR_STATUS_GOOD;

   if (!cid)
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   else
   {
      out_val = 0;
      string tid;
      switch(obj)
      {
         case chips_contour: 
         {
            tid = cid->get_contour_id();
            if (tid.length() == 0)  tid = currcont; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = contours.size ();
            else if (contours.count(tid))
               out_val = 1;
         }
         break; 
         case chips_curve: 
         {
            tid = cid->get_curve_id();
            if (tid.length() == 0)  tid = currcurv; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = curves.size ();
            else if (curves.count(tid))
               out_val = 1;
         }
         break; 
         case chips_histogram: 
         {
            tid = cid->get_histogram_id();
            if (tid.length() == 0)  tid = currhist; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = histograms.size ();
            else if (histograms.count(tid))
               out_val = 1;
         }
         break; 
	 case chips_image:
	 {
            tid = cid->get_image_id();
            if (tid.length() == 0)  tid = currimage; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = images.size ();
            else if (images.count(tid))
               out_val = 1;
	 }
	 break; 
	 case chips_label:
	 {
            tid = cid->get_label_id();
            if (tid.length() == 0)  tid = currlabel; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = labels.size ();
            else if (labels.count(tid))
               out_val = 1;
	 }
	 break; 
	 case chips_line:
	 {
            tid = cid->get_line_id();
            if (tid.length() == 0)  tid = currline; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = lines.size ();
            else if (lines.count(tid))
               out_val = 1;
	 }
	 break; 
	 case chips_point:
	 {
            tid = cid->get_point_id();
            if (tid.length() == 0)  tid = currpoint; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = points.size ();
            else if (points.count(tid))
               out_val = 1;
	 }
	 break; 
	 case chips_colorbar:
	 {
            tid = cid->get_colorbar_id();
            if (tid.length() == 0)  tid = currcbar; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = cbars.size ();
            else if (cbars.count(tid))
               out_val = 1;
	 }
	 break; 
	 case chips_region:
	 {
            tid = cid->get_region_id();
            if (tid.length() == 0)  tid = currregion; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
               out_val = regions.size ();
            else if (regions.count(tid))
               out_val = 1;
	 }
	 break; 
         case chips_axis:
         {
            tid = cid->get_axis_id();
            int axsz; 
            if (chips_strcmp(tid.c_str(), "all") == 0)
            {
               axsz = axes_x.size ();
               if (axsz > 2) axsz -=2;
               out_val = axsz;
               axsz = axes_y.size ();
               if (axsz > 2) axsz -=2;
               out_val += axsz;
            }
            else if (tid.length())
            {
               if (get_xaxis(tid) != NULL)
                  out_val = 1;
               else if (get_yaxis(tid) != NULL)
                  out_val = 1;
            }
            else
            {
               tid = cid->get_xaxis_id();
               string tid2 = cid->get_yaxis_id();
               if (tid.length() && (!tid2.length()))
               {
                  if (get_xaxis(tid) != NULL)
                     out_val = 1;
               }
               else if ((!tid.length()) && tid2.length())
               {
                  if (get_yaxis(tid2) != NULL)
                     out_val = 1;
               }
               else if (tid.length() && tid2.length())
               {
                  cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
               }
               else
               {
                  if (cur_x)
                     out_val += 1;
                  else if (cur_xall) 
                  {
                     axsz = axes_x.size ();
                     if (axsz > 2) axsz -=2; 
                     out_val += axsz; 
                  }
                  if (cur_y)
                     out_val += 1;
                  else if (cur_yall)
                  {
                     axsz = axes_y.size ();
                     if (axsz > 2) axsz -=2; 
                     out_val += axsz; 
                  }
               }
            }
         }
         break;
         default:   // do nothing
         break;
      }
   }
   return rval;
}

int ChipsPlot::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   int rval = ERR_STATUS_GOOD;

   if (!cid)
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   else
   {
      out_val = COORD_NONE;
      string tid;
      switch(obj)
      {
         case chips_contour:
         {
            ChipsContour* cc; 
            tid = cid->get_contour_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && contours.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cc = get_contour(tid)) != NULL)
            {
               out_val = DATA;
               out_id = cc->get_data_system_id ();
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",tid.c_str());
         }
         break;
         case chips_curve:
         {
            ChipsCurve* cc; 
            tid = cid->get_curve_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && curves.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cc = get_curve(tid)) != NULL)
            {
               out_val = DATA;
               out_id = cc->get_data_system_id ();
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",tid.c_str());
         }
         break;
         case chips_histogram:
         {
            ChipsHistogram* ch; 
            tid = cid->get_histogram_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && histograms.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((ch = get_histogram(tid)) != NULL)
            {
               out_val = DATA;
               out_id = ch->get_data_system_id ();
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",tid.c_str());
         }
         break;
         case chips_image:
         {
            ChipsImage* ci;
            tid = cid->get_image_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && images.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((ci = get_image(tid)) != NULL)
            {
               out_val = ci->get_coord_system (); 
               out_id = (out_val == DATA) ? ci->get_data_system_id() : "";
            }
            else
              cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",tid.c_str());
         }
         break;
         case chips_label:
         {
            ChipsLabel* cl;
            tid = cid->get_label_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && labels.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cl = get_label(tid)) != NULL)
            {
               out_val = cl->get_coord_system (); 
               out_id = (out_val == DATA) ? cl->get_data_system_id() : "";
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",tid.c_str());
         }
         break;
         case chips_line:
         {
            ChipsLine* cl;
            tid = cid->get_line_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && lines.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cl = get_line(tid)) != NULL)
            {
               out_val = cl->get_coord_system ();
               out_id = (out_val == DATA) ? cl->get_data_system_id() : "";
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",tid.c_str());
         }
         break;
         case chips_point:
         {
            ChipsPoint* cp;
            tid = cid->get_point_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && points.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cp = get_point(tid)) != NULL)
            {
               out_val = cp->get_coord_system ();
               out_id = (out_val == DATA) ? cp->get_data_system_id() : "";
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",tid.c_str());
         }
         break;
         case chips_colorbar:
         {
            ChipsColorBar* cp;
            tid = cid->get_colorbar_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && cbars.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cp = get_colorbar(tid)) != NULL)
            {
               out_val = cp->get_coord_system ();
               out_id = (out_val == DATA) ? cp->get_data_system_id() : "";
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",tid.c_str());
         }
         break;
         case chips_region:
         {
            ChipsRegion* cr;
            tid = cid->get_region_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && regions.size() != 1)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if ((cr = get_region(tid)) != NULL)
            {
               out_val = cr->get_coord_system ();
               out_id = (out_val == DATA) ? cr->get_data_system_id() : "";
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",tid.c_str());
         }
         break;
         case chips_axis:
         {
            ChipsAxis* ca; 
            tid = cid->get_axis_id();
            if (chips_strcmp(tid.c_str(), "all") == 0 && axes_x.size() !=3 &&
                axes_y.size() != 3)
               cmd_manager->add_err(rval=ERR_INVALID_ALL_USE,id,0);
            else if (tid.length())
            {
               if ((ca = get_xaxis(tid)) != NULL)
               {
                  out_val = ca->get_coord_system();
                  out_id = (out_val == DATA) ? ca->get_data_system_id() : "";
               }
               else if ((ca = get_yaxis(tid)) != NULL)
               {
                  out_val = ca->get_coord_system();
                  out_id = (out_val == DATA) ? ca->get_data_system_id() : "";
               }
               else 
                  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tid.c_str());
            }
            else
            {
               tid = cid->get_xaxis_id();
               string tid2 = cid->get_yaxis_id();
               if (tid.length() && (!tid2.length()))
               {
                  if ((ca = get_xaxis(tid)) != NULL)
                  {
                     out_val = ca->get_coord_system();
                     out_id = (out_val == DATA) ? ca->get_data_system_id() : "";
                  }
                  else 
                     cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tid.c_str());
               }
               else if ((!tid.length()) && tid2.length())
               {
                  if ((ca = get_yaxis(tid2)) != NULL)
                  {
                     out_val = ca->get_coord_system();
                     out_id = (out_val == DATA) ? ca->get_data_system_id() : "";
                  }
                  else 
                     cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tid2.c_str());
               }
               else
               {
                  string errstr = "Cannot retrieve a property from both X and Y axes at the same time.";
                  cmd_manager->add_err(errstr, rval=ERR_INVALID_GET_ALL,id,0); 
               }
            }
         }
         break;
         default:   // do nothing
         break;
      }
   }
   return rval;
}



int ChipsPlot::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   int oldvis, st, mask = theDefaults->get_allpropmask();

   if ((obj == chips_window) || (obj == chips_frame) || (obj == chips_plot))
   {
      if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         string clr = theDefaults->get_allcolor();
	 rval = set_property(NULL, (void*)clr.c_str(), SET_PLOT_TITLE_COLOR);
      }
      else if ((mask & PREF_ALL_FONT) == PREF_ALL_FONT)
      {
         string newfont = theDefaults->get_allfont();
         rval = set_property(NULL, (void*)newfont.c_str(), SET_PLOT_TITLE_FONT);
      }
      else if ((mask & PREF_ALL_FONTSTYLE) == PREF_ALL_FONTSTYLE)
      {
         string sval = v_fontstyle_to_string(theDefaults->get_allfontstyle());
         rval = set_property(NULL, (void*)sval.c_str(), SET_PLOT_TITLE_FONTSTYLE);
      }
      else if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
         oldvis = visible; 
         rval = set_property(NULL, static_cast<void*>(&vis), SET_PLOT_VISIBLE); 
      }
      else if ((mask & PREF_ALL_ANGLE) == PREF_ALL_ANGLE)
      {
         float ang = theDefaults->get_allangle();
         rval = set_property(NULL, static_cast<void*>(&ang), SET_PLOT_TITLE_ANGLE);
      }
      else if ((mask & PREF_ALL_DEPTH) == PREF_ALL_DEPTH)
      {
         int dep = theDefaults->get_alldepth();
         rval = set_property(NULL, static_cast<void*>(&dep), SET_PLOT_DEPTH);
      }

      // set axis properties
      for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
      {
         st = (*am).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }
      for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
      {
         st = (*am).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set curve properties
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
      {
	 st = (*cc).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set contour properties
      for(contours_map::iterator cc=contours.begin(); cc!=contours.end(); ++cc)
      {
	 st = (*cc).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set histogram properties
      for(histograms_map::iterator hh=histograms.begin(); hh!=histograms.end(); ++hh)
      {
	 st = (*hh).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set line properties
      for(lines_map::iterator ll=lines.begin(); ll!=lines.end(); ++ll)
      {
	 st = (*ll).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set label properties
      for(labels_map::iterator lb=labels.begin(); lb!=labels.end(); ++lb)
      {
	 st = (*lb).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set point properties
      for(points_map::iterator pp=points.begin(); pp!=points.end(); ++pp)
      {
	 st = (*pp).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set colorbar properties
      for(cbars_map::iterator ib=cbars.begin(); ib!=cbars.end(); ++ib)
      {
	 st = (*ib).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set region properties
      for(regions_map::iterator rr=regions.begin(); rr!=regions.end(); ++rr)
      {
	 st = (*rr).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // set image properties
      for(images_map::iterator ii=images.begin(); ii!= images.end(); ++ii)
      {
         st = (*ii).second->set_cascading_property(cid, obj);
         if (st != ERR_STATUS_GOOD)  rval = st;
      }

      // due to the way visibility of axes depends on its parent's visiblity
      // wrap the axis commands with plot commands so undo/redo work
      if ((mask & PREF_ALL_VISIBLE) == PREF_ALL_VISIBLE)
      {
         int vis = theDefaults->get_allvisible();
	 if (oldvis) show();  else hide(); 
         rval = set_property(NULL, static_cast<void*>(&vis), SET_PLOT_VISIBLE); 
      }
   }
   else if (obj == chips_axis)       // set axis properties 
   {
      string tmp_id = cid->get_axis_id(); 
      ChipsAxis* ax = NULL;
 
      if (!axes_x_exist() && !axes_y_exist())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"axes");
      else if (tmp_id.length())   // adjust specified single axis or 'all'
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
	    {
	       st = (*am).second->set_cascading_property(cid, obj);
	       if (st != ERR_STATUS_GOOD)  rval = st;
	    }
	    for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
	    {
	       st = (*am).second->set_cascading_property(cid, obj);
	       if (st != ERR_STATUS_GOOD)  rval = st;
	    }
         }
         else        // adjust single specified axis 
         {
            if ((ax = get_xaxis(tmp_id)) != NULL)
               rval = ax->set_cascading_property(cid, obj);
            if ((ax = get_yaxis(tmp_id)) != NULL)
               rval = ax->set_cascading_property(cid, obj);
         }
      }
      else 
      {
         string tmp_id2 = cid->get_yaxis_id();
         tmp_id = cid->get_xaxis_id();
   
         if (tmp_id.length())   // adjust x axis
         {
            if (chips_strcmp(tmp_id.c_str(), "all") == 0)
            {
	       for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
	       {
	          st = (*am).second->set_cascading_property(cid, obj);
	          if (st != ERR_STATUS_GOOD)  rval = st;
	       }
            }
            else if ((ax = get_xaxis(tmp_id)) != NULL)
                  rval = ax->set_cascading_property(cid, obj);
         }
         if (tmp_id2.length())     // adjust y axis
         {
            if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
            {
	       for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
	       {
	          st = (*am).second->set_cascading_property(cid, obj);
	          if (st != ERR_STATUS_GOOD)  rval = st;
	       }
            }
            else if ((ax = get_yaxis(tmp_id2)) != NULL)
                  rval = ax->set_cascading_property(cid, obj);
         }
         if ((tmp_id2.length() == 0) && (tmp_id2.length() == 0))
         {
            // adjust x and y axes
            if (cur_x) 
               rval = cur_x->set_cascading_property(cid, obj);
            else if (cur_xall)
	       for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
	          rval |= (*am).second->set_cascading_property(cid, obj);
            if (cur_y)
               rval = cur_y->set_cascading_property(cid, obj);
            else if (cur_yall)
	       for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
	          rval |= (*am).second->set_cascading_property(cid, obj);
         }
      }
   }
   else if (obj == chips_curve)      // set curve properties
   {
      string cvid = cid->get_curve_id();
      if (cvid.length() == 0)   cvid = currcurv;

      if (cvid.length() && (chips_strcmp(cvid.c_str(), "all") != 0))
      {
         ChipsCurve* cv; 
         if ((cv = get_curve(cvid)) != NULL)
            rval = cv->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",cvid.c_str());
      }
      else 
      {
	 for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
            if ((st = (*cc).second->set_cascading_property(cid, obj)) != 
                ERR_STATUS_GOOD) 
               rval = st;
      }
   }
   else if (obj == chips_contour)      // set contour properties
   {
      string ctid = cid->get_contour_id();
      if (ctid.length() == 0)   ctid = currcont;

      if (ctid.length() && (chips_strcmp(ctid.c_str(), "all") != 0))
      {
         ChipsContour* ct;
         if ((ct = get_contour(ctid)) != NULL)
            rval = ct->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());
      }
      else
      {
	 for(contours_map::iterator cc=contours.begin();cc!=contours.end();++cc)
            if ((st = (*cc).second->set_cascading_property(cid, obj)) != 
                ERR_STATUS_GOOD) 
               rval = st;
      }
   }
   else if (obj == chips_histogram)      // set histogram properties
   {
      string chid = cid->get_histogram_id();
      if (chid.length() == 0)   chid = currhist;

      if (chid.length() && (chips_strcmp(chid.c_str(), "all") != 0))
      {
         ChipsHistogram* ch;
         if ((ch = get_histogram(chid)) != NULL)
            rval = ch->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",chid.c_str());
      }
      else
      {
         for(histograms_map::iterator hh=histograms.begin();hh!=histograms.end();++hh)
            if ((st = (*hh).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_image)      // set image properties
   {
      string ciid = cid->get_image_id();
      if (ciid.length() == 0)   ciid = currimage;

      if (ciid.length() && (chips_strcmp(ciid.c_str(), "all") != 0))
      {
         ChipsImage* ci;
         if ((ci = get_image(ciid)) != NULL)
            rval = ci->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",ciid.c_str());
      }
      else
      {
         for(images_map::iterator ii=images.begin();ii!=images.end();++ii)
            if ((st = (*ii).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_colorbar)      // set colorbar properties
   {
      string cbid = cid->get_colorbar_id();
      if (cbid.length() == 0)   cbid = currcbar;

      if (cbid.length() && (chips_strcmp(cbid.c_str(), "all") != 0))
      {
         ChipsColorBar* cb;
         if ((cb = get_colorbar(cbid)) != NULL)
            rval = cb->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",cbid.c_str());
      }
      else
      {
         for(cbars_map::iterator cc=cbars.begin();cc!=cbars.end();++cc)
            if ((st = (*cc).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_line)      // set line properties
   {
      string lid = cid->get_line_id();
      if (lid.length() == 0)   lid = currline;
      if (lid.length() && (chips_strcmp(lid.c_str(), "all") != 0))
      {
         ChipsLine* cl;
         if ((cl = get_line(lid)) != NULL)
            rval = cl->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",lid.c_str());
      }
      else
      {
         for(lines_map::iterator ll=lines.begin();ll!=lines.end();++ll)
            if ((st = (*ll).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_label)      // set label properties
   {
      string lid = cid->get_label_id();
      if (lid.length() == 0)   lid = currlabel;
      if (lid.length() && (chips_strcmp(lid.c_str(), "all") != 0))
      {
         ChipsLabel* cl;
         if ((cl = get_label(lid)) != NULL)
            rval = cl->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",lid.c_str());
      }
      else
      {
         for(labels_map::iterator ll=labels.begin();ll!=labels.end();++ll)
            if ((st = (*ll).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_point)      // set point properties
   {
      string pid = cid->get_point_id();
      if (pid.length() == 0)   pid = currpoint;
      if (pid.length() && (chips_strcmp(pid.c_str(), "all") != 0))
      {
         ChipsPoint* cp;
         if ((cp = get_point(pid)) != NULL)
            rval = cp->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",pid.c_str());
      }
      else
      {
         for(points_map::iterator pp=points.begin();pp!=points.end();++pp)
            if ((st = (*pp).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   else if (obj == chips_region)      // set region properties
   {
      string rid = cid->get_region_id();
      if (rid.length() == 0)   rid = currregion;
      if (rid.length() && (chips_strcmp(rid.c_str(), "all") != 0))
      {
         ChipsRegion* cr;
         if ((cr = get_region(rid)) != NULL)
            rval = cr->set_cascading_property(cid, obj);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",rid.c_str());
      }
      else
      {
         for(regions_map::iterator rr=regions.begin();rr!=regions.end();++rr)
            if ((st = (*rr).second->set_cascading_property(cid, obj)) !=
                ERR_STATUS_GOOD)
               rval = st;
      }
   }
   return rval;
}


int ChipsPlot::set_property(ChipsId* cid, void* in_val, int in_property,
   bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsSetCommand *cmd = NULL;
   bool set_plot_prop = true;

   switch(in_property)
   {
      case SET_LEGEND_COLOR:           // fallthrough intended
      case SET_LEGEND_THICKNESS:       // fallthrough intended
      case SET_LEGEND_TRANSPARENT:     // fallthrough intended
      case SET_LEGEND_VISIBLE:         // fallthrough intended
      case SET_LEGEND_BORDERS:         // fallthrough intended
         if (legend)
            rval = legend->set_property(in_val, in_property);
         else
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"legends");
         set_plot_prop = false;
      break; 
      case SET_CURVE_SYMBOL_ANGLE:     // fallthrough intended
      case SET_CURVE_SYMBOL_STYLE:     // fallthrough intended
      case SET_CURVE_SYMBOL_COLOR:     // fallthrough intended
      case SET_CURVE_SYMBOL_SIZE:      // fallthrough intended
      case SET_CURVE_SYMBOL_FILL:      // fallthrough intended
      case SET_CURVE_LINE_STYLE:       // fallthrough intended
      case SET_CURVE_LINE_COLOR:       // fallthrough intended
      case SET_CURVE_LINE_THICKNESS:   // fallthrough intended
      case SET_CURVE_BADDATA:          // fallthrough intended
      case SET_CURVE_DEPTH:            // fallthrough intended
      case SET_CURVE_ERR_CAPLEN:       // fallthrough intended
      case SET_CURVE_ERR_UP:           // fallthrough intended
      case SET_CURVE_ERR_DOWN:         // fallthrough intended
      case SET_CURVE_ERR_LEFT:         // fallthrough intended
      case SET_CURVE_ERR_RIGHT:        // fallthrough intended
      case SET_CURVE_ERR_STYLE:        // fallthrough intended
      case SET_CURVE_ERR_COLOR:        // fallthrough intended
      case SET_CURVE_LIMLENGTH:        // fallthrough intended
      case SET_CURVE_LIMOVERRIDE:      // fallthrough intended
      case SET_CURVE_VISIBLE:          // fallthrough intended
      case SET_CURVE_ERR_THICKNESS:
         rval = set_curve_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_IMAGE_THRESHOLD:        // fallthrough intended
      case SET_IMAGE_ALPHA:            // fallthrough intended
      case SET_IMAGE_INTERPOLATION:    // fallthrough intended
      case SET_IMAGE_DEPTH:            // fallthrough intended
      case SET_IMAGE_VISIBLE:          // fallthrough intended
      case SET_IMAGE_CMAP:             // fallthrough intended
      case SET_IMAGE_CMAP_SIZE:        // fallthrough intended
      case SET_IMAGE_CMAP_INTERPOLATE: // fallthrough intended
      case SET_IMAGE_SCALE_CHANNELS:   // fallthrough intended
      case SET_IMAGE_INV_CMAP: 
         rval = set_image_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_CBAR_DEPTH:             // fallthrough intended
      case SET_CBAR_VISIBLE:           // fallthrough intended
      case SET_CBAR_BORDERVIS:         // fallthrough intended
      case SET_CBAR_HALIGN:            // fallthrough intended
      case SET_CBAR_VALIGN:            // fallthrough intended
      case SET_CBAR_LENGTH:            // fallthrough intended
      case SET_CBAR_ORIENTATION:       // fallthrough intended
      case SET_CBAR_TICK_MODE:         // fallthrough intended
      case SET_CBAR_LABEL_ANGLE:       // fallthrough intended
      case SET_CBAR_LABEL_COLOR:       // fallthrough intended
      case SET_CBAR_LABEL_FONT:        // fallthrough intended
      case SET_CBAR_LABEL_FONTSTYLE:   // fallthrough intended
      case SET_CBAR_LABEL_HALIGN:      // fallthrough intended
      case SET_CBAR_LABEL_VALIGN:      // fallthrough intended
      case SET_CBAR_LABEL_SIZE:        // fallthrough intended
      case SET_CBAR_LABEL_VIS:         // fallthrough intended
      case SET_CBAR_LABEL_LOCATION:    // fallthrough intended
      case SET_CBAR_TEXT:              // fallthrough intended
      case SET_CBAR_TICK_THICKNESS:    // fallthrough intended
      case SET_CBAR_TICK_LENGTH:       // fallthrough intended
      case SET_CBAR_TICK_COLOR:        // fallthrough intended
      case SET_CBAR_TICK_STYLE:        // fallthrough intended
      case SET_CBAR_TICK_VIS:          // fallthrough intended
      case SET_CBAR_TICK_LOCATION:     // fallthrough intended
      case SET_CBAR_TICKLABEL_ANGLE:   // fallthrough intended
      case SET_CBAR_TICKLABEL_COLOR:   // fallthrough intended
      case SET_CBAR_TICKLABEL_FONT:    // fallthrough intended
      case SET_CBAR_TICKLABEL_FNTSTL:  // fallthrough intended
      case SET_CBAR_TICKLABEL_OFFSET:  // fallthrough intended
      case SET_CBAR_TICKLABEL_SIZE:    // fallthrough intended
      case SET_CBAR_TICKLABEL_HALIGN:  // fallthrough intended
      case SET_CBAR_TICKLABEL_VALIGN:  // fallthrough intended
      case SET_CBAR_TICKLABEL_VIS:     // fallthrough intended
      case SET_CBAR_OFFSET_PAR:        // fallthrough intended
      case SET_CBAR_OFFSET_PERP:       // fallthrough intended
      case SET_CBAR_WIDTH: 
         rval = set_colorbar_property(cid, in_val, in_property);
         set_plot_prop = false;
      break; 
      case SET_HISTO_SYMBOL_ANGLE:     // fallthrough intended
      case SET_HISTO_SYMBOL_STYLE:     // fallthrough intended
      case SET_HISTO_SYMBOL_COLOR:     // fallthrough intended
      case SET_HISTO_SYMBOL_SIZE:      // fallthrough intended
      case SET_HISTO_SYMBOL_FILL:      // fallthrough intended
      case SET_HISTO_LINE_STYLE:       // fallthrough intended
      case SET_HISTO_LINE_COLOR:       // fallthrough intended
      case SET_HISTO_LINE_THICKNESS:   // fallthrough intended
      case SET_HISTO_DEPTH:            // fallthrough intended
      case SET_HISTO_ERR_UP:           // fallthrough intended
      case SET_HISTO_ERR_DOWN:         // fallthrough intended
      case SET_HISTO_DROPLINE:         // fallthrough intended
      case SET_HISTO_ERR_STYLE:        // fallthrough intended
      case SET_HISTO_ERR_COLOR:        // fallthrough intended
      case SET_HISTO_ERR_CAPLEN:       // fallthrough intended
      case SET_HISTO_VISIBLE:          // fallthrough intended
      case SET_HISTO_FILL_OPACITY:     // fallthrough intended
      case SET_HISTO_FILL_COLOR:       // fallthrough intended
      case SET_HISTO_FILL_STYLE:       // fallthrough intended
      case SET_HISTO_ERR_THICKNESS:   
         rval = set_histogram_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_CONTOUR_COLOR:          // fallthrough intended
      case SET_CONTOUR_STYLE:          // fallthrough intended
      case SET_CONTOUR_THICKNESS:      // fallthrough intended
      case SET_CONTOUR_MODE:           // fallthrough intended
      case SET_CONTOUR_DEPTH:          // fallthrough intended
      case SET_CONTOUR_TICK_VIS:       // fallthrough intended
      case SET_CONTOUR_TICK_COLOR:     // fallthrough intended
      case SET_CONTOUR_TICK_THICK:     // fallthrough intended
      case SET_CONTOUR_TICK_LENGTH:    // fallthrough intended
      case SET_CONTOUR_TICK_DIR:       // fallthrough intended
      case SET_CONTOUR_LAB_VIS:        // fallthrough intended
      case SET_CONTOUR_LAB_COLOR:      // fallthrough intended
      case SET_CONTOUR_LAB_SIZE:       // fallthrough intended
      case SET_CONTOUR_LAB_FONT:       // fallthrough intended
      case SET_CONTOUR_ALGORITHM:      // fallthrough intended
      case SET_CONTOUR_INTERVAL:       // fallthrough intended
      case SET_CONTOUR_NUMLEVELS:      // fallthrough intended
      case SET_CONTOUR_VISIBLE:        // fallthrough intended
      case SET_CONTOUR_LAB_FONTSTYLE:       
         rval = set_contour_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_LABEL_TEXT:             // fallthrough intended
      case SET_LABEL_FONT:             // fallthrough intended
      case SET_LABEL_STYLE:            // fallthrough intended
      case SET_LABEL_COLOR:            // fallthrough intended
      case SET_LABEL_ANGLE:            // fallthrough intended
      case SET_LABEL_HALIGN:           // fallthrough intended
      case SET_LABEL_VALIGN:           // fallthrough intended
      case SET_LABEL_VISIBLE:          // fallthrough intended
      case SET_LABEL_DEPTH:            // fallthrough intended
      case SET_LABEL_BBOX:             // fallthrough intended
      case SET_LABEL_SIZE:
         rval = set_label_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_LINE_STYLE:             // fallthrough intended
      case SET_LINE_COLOR:             // fallthrough intended
      case SET_LINE_THICKNESS:         // fallthrough intended
      case SET_LINE_DEPTH:             // fallthrough intended
      case SET_LINE_EXTEND:            // fallthrough intended
      case SET_LINE_VISIBLE:
         rval = set_line_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_POINT_SYMBOL_STYLE:     // fallthrough intended
      case SET_POINT_SYMBOL_COLOR:     // fallthrough intended
      case SET_POINT_SYMBOL_SIZE:      // fallthrough intended
      case SET_POINT_SYMBOL_FILL:      // fallthrough intended
      case SET_POINT_VISIBLE:          // fallthrough intended
      case SET_POINT_DEPTH:            // fallthrough intended
      case SET_POINT_SYMBOL_ANGLE:
         rval = set_point_property(cid, in_val, in_property);
         set_plot_prop = false;
      break; 
      case SET_REGN_EDGE_COLOR:        // fallthrough intended
      case SET_REGN_EDGE_STYLE:        // fallthrough intended
      case SET_REGN_EDGE_THICKNESS:    // fallthrough intended
      case SET_REGN_ANGLE:             // fallthrough intended
      case SET_REGN_FILL_STYLE:        // fallthrough intended
      case SET_REGN_VISIBLE:           // fallthrough intended
      case SET_REGN_OPACITY:           // fallthrough intended
      case SET_REGN_DEPTH:             // fallthrough intended
      case SET_REGN_FILL_COLOR:
         rval = set_region_property(cid, in_val, in_property);
         set_plot_prop = false;
      break;
      case SET_AXIS_COLOR:             // fallthrough intended
      case SET_AXIS_PAD:               // fallthrough intended
      case SET_AXIS_DEPTH:             // fallthrough intended
      case SET_AXIS_LOG_MIN:           // fallthrough intended
      case SET_AXIS_MAJOR_COLOR:       // fallthrough intended
      case SET_AXIS_MINOR_COLOR:       // fallthrough intended
      case SET_AXIS_THICKNESS:         // fallthrough intended
      case SET_AXIS_MAJ_THICKNESS:     // fallthrough intended
      case SET_AXIS_MIN_THICKNESS:     // fallthrough intended
      case SET_AXIS_TEXT:              // fallthrough intended
      case SET_AXIS_XLAB_TXT:          // fallthrough intended
      case SET_AXIS_YLAB_TXT:          // fallthrough intended
      case SET_AXIS_SCALABLE:          // fallthrough intended
      case SET_AXIS_TICKSTYLE:         // fallthrough intended
      case SET_AXIS_MAJ_STYLE:         // fallthrough intended
      case SET_AXIS_MIN_STYLE:         // fallthrough intended
      case SET_AXIS_TICKLABEL_STYLE:   // fallthrough intended
      case SET_AXIS_TICKFORMAT:        // fallthrough intended
      case SET_AXIS_MAJOR_NUM:         // fallthrough intended
      case SET_AXIS_MINOR_NUM:         // fallthrough intended
      case SET_AXIS_MAJORINT:          // fallthrough intended
      case SET_AXIS_MINORINT:          // fallthrough intended
      case SET_AXIS_MAJ_MODE:          // fallthrough intended
      case SET_AXIS_MIN_MODE:          // fallthrough intended
      case SET_AXIS_MAJ_VIS:           // fallthrough intended
      case SET_AXIS_MIN_VIS:           // fallthrough intended
      case SET_AXIS_VISIBLE:           // fallthrough intended
      case SET_AXIS_MAJORLENGTH:       // fallthrough intended
      case SET_AXIS_OFFSET_PAR:        // fallthrough intended
      case SET_AXIS_OFFSET_PERP:       // fallthrough intended
      case SET_AXIS_TICKLABEL_OFFSET:  // fallthrough intended
      case SET_AXIS_TICKLABEL_FONT:    // fallthrough intended
      case SET_AXIS_TICKLABEL_FNTSTL:  // fallthrough intended
      case SET_AXIS_TICKLABEL_SIZE:    // fallthrough intended
      case SET_AXIS_TICKLABEL_COLOR:   // fallthrough intended
      case SET_AXIS_TICKLABEL_ANGLE:   // fallthrough intended
      case SET_AXIS_TICKLABEL_HALIGN:  // fallthrough intended
      case SET_AXIS_TICKLABEL_VALIGN:  // fallthrough intended
      case SET_AXIS_TICKLABEL_VIS:     // fallthrough intended
      case SET_AXIS_LABEL_VIS:         // fallthrough intended
      case SET_AXIS_LABEL_FONT:        // fallthrough intended
      case SET_AXIS_LABEL_FONTSTYLE:   // fallthrough intended
      case SET_AXIS_LABEL_ANGLE:       // fallthrough intended
      case SET_AXIS_XLAB_ANG:          // fallthrough intended
      case SET_AXIS_YLAB_ANG:          // fallthrough intended
      case SET_AXIS_LABEL_HALIGN:      // fallthrough intended
      case SET_AXIS_LABEL_VALIGN:      // fallthrough intended
      case SET_AXIS_LABEL_SIZE:        // fallthrough intended
      case SET_AXIS_LABEL_COLOR:       // fallthrough intended
      case SET_AXIS_AUTOMIN:           // fallthrough intended
      case SET_AXIS_AUTOMAX:           // fallthrough intended
      case SET_AXIS_MINORLENGTH:       // fallthrough intended
      case SET_AXIS_MAJGRD_COLOR:      // fallthrough intended
      case SET_AXIS_MAJGRD_STYLE:      // fallthrough intended
      case SET_AXIS_MAJGRD_VIS:        // fallthrough intended
      case SET_AXIS_MAJGRD_THICK:      // fallthrough intended
      case SET_AXIS_MINGRD_COLOR:      // fallthrough intended
      case SET_AXIS_MINGRD_STYLE:      // fallthrough intended
      case SET_AXIS_MINGRD_VIS:        // fallthrough intended
      case SET_AXIS_MINGRD_THICK:   
         rval = set_axis_property(cid, in_val, in_property);
         set_plot_prop = false;
      break; 
      case SET_PLOT_VISIBLE:
      {
	 int* ival = static_cast<int*>(in_val);
         int oval = visible;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, ((ChipsFrame *)parent)->get_current_plot());
         if ((*ival) == 1)
         {
            show();
            if (rec && ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD))
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else if ((*ival) == 0)
         {
            hide();
            if (rec && ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD))
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else
         {
            cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"plot visible",
                                 *ival, "0:1");
            delete cmd;
         }
      }
      break; 
      case SET_PLOT_OFFSCREEN:
      {
         int* ival = static_cast<int*>(in_val);
         int oval = offscreen;
         offscreen = *ival;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, ((ChipsFrame *)parent)->get_current_plot());
         if (offscreen == 0 || offscreen == 1)
         {
            if (rec && ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD))
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else
         {
            cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"plot visible",
                                 *ival, "0:1");
	    offscreen = oval; 
            delete cmd;
         }
         if ((rval == ERR_STATUS_GOOD) && (offscreen != oval))
            this->change_visible(); 
      }
      break;
      case SET_PLOT_LOGSCALE:     // fallthrough intended
      case SET_PLOT_LINEARSCALE:  // fallthrough intended 
      case SET_AXIS_REVERSE:           
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_dimension_property(cid, *ival, in_property, rec); 
      }
      break; 
      case SET_PLOT_TITLE:
      {
         string val = (const char*)in_val;
	 string oval = get_title_text();
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_text(val)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      } 
      break; 
      case SET_PLOT_ASPECT_RATIO:
      {
	 string oval = get_aspect_ratio(); 
         string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_aspect_ratio(val, rec)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
            cmd_manager->update_mask(STATE_CHNG_LIMS);
         }
         else  delete cmd;
      }
      break; 
      case SET_DATA_ASPECT_RATIO:
      {
         if (coord)
         {
            string val = (const char*)in_val;
            string dsid = cid->get_coord_id();
            ChipsDataSystem *ds = NULL;

            string ax1 = cid->get_xaxis_id();
            string ax2 = cid->get_yaxis_id();
            ChipsAxis *ax = get_xaxis(ax1),
                      *ay = get_yaxis(ax2);

            // first attempt to use the DATA coord sys id from the ChipsId 
            if (dsid.length() && cid->get_coord_sys() == DATA)
               ds=dynamic_cast<ChipsDataSystem *>(coord->get_coord(DATA,dsid));

            // didn't get the data sys yet, try using specified/current axes
            if (!ds)
               ds = get_datasystem (ax, ay);

            // check for log scale
            if ((ax && ax->get_scaletype() == V_SCALE_LOG) ||   
                (ay && ay->get_scaletype() == V_SCALE_LOG))
            {
               cmd_manager->add_err(rval=ERR_DASP_LOG,id,0); 
            }
            else if (ds)  // if we found our data system set the ratio 
            {
               ChipsDataAspectCommand* dcmd = NULL; 
               string oval = ds->get_aspect_ratio(); 
               if (rec)  dcmd = new ChipsDataAspectCommand(this, ax1, ax2,
                  val, oval, ((ChipsFrame *)parent)->get_current_plot());

	       if ((rval = ds->set_aspect_ratio(val)) == ERR_STATUS_GOOD)
               {
                  if (rec && (rval=cmd_manager->add_cmd(dcmd))!=ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete dcmd;
                  }
                  if (val.length())
                  {
                     ChipsAxis* master = ax;

/*
                     ChipsAxis* slave = ay;
                     if (ds->get_aspect_width() < ds->get_aspect_height ())
                     {
                        master = ay;
                        slave = ax;
                     }
*/
                     
                     if ((rval = bindings->add_aspect_ratio(ax, ay,
                          ds->get_aspect_width(), ds->get_aspect_height())) == 
                          ERR_STATUS_GOOD)
                     { 
                        double min, max;
                        min = master->get_absolute_min (false);
                        max = master->get_absolute_max (false);
                        ChipsId tmpid;
                        rval = bindings->update_connected_nodes(master, min, max,
                           false, AUTOLIM_OFF, 1, rec);
                     }
                     else   ds->set_aspect_ratio(oval);
                  }
                  else
                  {
                     ax->set_use_pad (true);
                     ay->set_use_pad (true);
                     bindings->remove_aspect_ratio(ax, ay);
                  }

                  this->set_data_aspect_ratio (cid);
               }
               else
               {
                  string errstr = "The data aspect ratio value ("+ val + ") must be one of [\"<width>:<height>\"|\"\"]";
                  cmd_manager->add_err(errstr, rval=ERR_NO_PROP_VALUE,id,0);
                  delete dcmd;
               }
            }
            else  cmd_manager->add_err(rval=ERR_DASP_NO_DS,id,0); 
         }
      }
      break; 
      case SET_PLOT_DEPTH:
      {
         int* ival = static_cast<int*>(in_val);
         int newval, oval = get_title_depth();
         // since depth change shuffles objects only update if depth changed 
         if ((*ival) != oval)
         {
            if ((*ival) == V_USE_DEFAULT_DEPTH)
               newval = depth;   // plot's depth contain's default depth value
            else
               newval = *ival;
            if (rec)  cmd = new ChipsSetCommand(this, in_property,
               oval, newval, ((ChipsFrame *)parent)->get_current_plot());
            if (rec && title) remove_object_from_depth_table(title);
            if ((rval = set_title_depth(newval)) == ERR_STATUS_GOOD)
            {
               if (rec && ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD))
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
            else  delete cmd;
            if (rec && title) add_object_to_depth_table(title);
         }
      }
      break;
      case SET_PLOT_XLABEL:
      {
         rval = set_xlabel_text(cid, (char*)in_val);
      }
      break; 
      case SET_PLOT_YLABEL:
      {
         rval = set_ylabel_text(cid, (char*)in_val);
      }
      break; 
      case SET_PLOT_TITLE_X:
      {
	 float* fval = static_cast<float*>(in_val);
	 float oval = get_title_xpos(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_xpos(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_PLOT_TITLE_Y:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = get_title_ypos(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_ypos(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_PLOT_TITLE_COLOR:
      {
         string val = (const char*)in_val;
	 string oval = get_title_color();
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_color(val)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_PLOT_TITLE_SIZE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = get_title_size(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_size(*ival)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TITLE_FONT:
      {
         string val = (const char*)in_val;
	 string oval = get_title_font();
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_font(val)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TITLE_ANGLE:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = get_title_angle(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_angle(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TITLE_HALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = get_title_halign(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_halign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TITLE_VALIGN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = get_title_valign(); 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_valign(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TITLE_FONTSTYLE:
      {
         string val = (const char*)in_val;
	 string oval = get_title_fontstyle();
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, val, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_title_fontstyle(val)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_LMARGIN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = left_margin;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_plot_lmargin(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_PLOT_RMARGIN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = right_margin;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_plot_rmargin(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec &&(rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_TMARGIN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = top_margin;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_plot_tmargin(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_BMARGIN:
      {
         float* fval = static_cast<float*>(in_val);
	 float oval = bottom_margin;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *fval, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_plot_bmargin(*fval)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break;
      case SET_PLOT_AXIS_STYLE:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int) axis_style; 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_axis_style((V_AXISSTYLE_TYPE)*ival, rec)) == ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      case SET_PLOT_CORNER:
      {
         int* ival = static_cast<int*>(in_val);
	 int oval = (int) corner_style; 
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            oval, *ival, ((ChipsFrame *)parent)->get_current_plot());
         if ((rval = set_corner_style((V_CORNERSTYLE_TYPE)*ival)) == 
             ERR_STATUS_GOOD)
         {
            if (rec && (rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         else  delete cmd;
      }
      break; 
      default:
         cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,in_property);
      break;
   }

   if (cmd_manager && !rval && set_plot_prop)
      cmd_manager->update_mask (STATE_CHNG_PROP_PLT);

   return rval;
}

int ChipsPlot::set_axis_property(ChipsId* cid, void* in_val, int in_property)
{
   int sts, rval = ERR_STATUS_GOOD;
   string tmp_id = cid->get_axis_id(); 
   ChipsAxis* ax = NULL;
   bool found = false; 
   string curr_xid = cur_xall ? "all" : get_current_xaxis_id ();
   string curr_yid = cur_yall ? "all" : get_current_yaxis_id ();
   bool update_coord = false;

   if (!axes_x_exist() && !axes_y_exist())
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"axes");
   else if (tmp_id.length())   // adjust specified single axis or 'all'
   {
      if (chips_strcmp(tmp_id.c_str(), "all") == 0)
      {
         for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
         {
            if ((in_property != SET_AXIS_VISIBLE) || 
                ((in_property == SET_AXIS_VISIBLE) && 
                ((am->second != top) && (am->second != bottom))))
            {
               sts = this->set_axis_mode_property (am->second, in_val, in_property);
               if (sts)
                  sts = am->second->set_property(in_val, in_property);
               if (sts != ERR_STATUS_GOOD)  rval = sts;
            }
         }
         if (axes_x.size () > 3)
         {
            // only set current to 'all' if more than 1 x axis
            cur_xall = true;
            cur_x = NULL;
            cid->set_xaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_xid)
               update_coord = true;
         }
         for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
         {
            if ((in_property != SET_AXIS_VISIBLE) ||
                ((in_property == SET_AXIS_VISIBLE) &&
                (am->second != left) && (am->second != right)))
            {
               // for visiblity don't set the borders they are controlled by
               // axis style
               sts = this->set_axis_mode_property (am->second, in_val, in_property);
               if (sts)
                  sts = am->second->set_property(in_val, in_property);
               if (sts != ERR_STATUS_GOOD)  rval = sts;
            }
         }
         if (axes_y.size () > 3)
         {
            // only set current to 'all' if more than 1 y axis
            cur_yall = true;
            cur_y = NULL;
            cid->set_yaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_yid)
               update_coord = true;
         }
         if (in_property == SET_AXIS_THICKNESS)   adjust_corners(); 
      }
      else        // adjust single specified axis 
      {
         ax = get_xaxis(tmp_id.c_str());
         if ((in_property == SET_AXIS_VISIBLE)  &&
             ((ax == bottom) || (ax == top)))
            cmd_manager->add_err(rval=ERR_BOX_AXIS_VIS,id,0); 
         else if (ax && !rval)
         {
            rval = this->set_axis_mode_property (ax, in_val, in_property);
            if (rval)
               rval = ax->set_property(in_val, in_property);
            found = true;
	    if ((ax != top) && (ax != bottom))
	    {
               cur_x = ax;
               cur_xall = false;
	       // If the current axis has changed update coords
               if (tmp_id != curr_xid)
                  update_coord = true;	 
	    }
            if (in_property == SET_AXIS_THICKNESS)  adjust_corner_widths(ax); 
            cid->set_xaxis_id(tmp_id);
         }
         ax = get_yaxis(tmp_id.c_str());
         if ((in_property == SET_AXIS_VISIBLE)  &&
             ((ax == left) || (ax == right)))
            cmd_manager->add_err(rval=ERR_BOX_AXIS_VIS,id,0); 
         else if (ax && !rval)
         {
            rval = this->set_axis_mode_property (ax, in_val, in_property);
            if (rval)
               rval = ax->set_property(in_val, in_property);
            found = true;
	    if ((ax != left) && (ax != right))
	    {
               cur_y = ax;
               cur_yall = false; 
	       // If the current axis has changed update coords
               if (tmp_id != curr_yid)
                  update_coord = true;
	    }
            if (in_property == SET_AXIS_THICKNESS)  adjust_corner_widths(ax);
            cid->set_yaxis_id(tmp_id);
         }
         else if (!found && !rval) 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
      }
   }
   else 
   {
      string tmp_id2 = cid->get_yaxis_id();
      tmp_id = cid->get_xaxis_id();

      if (in_property == SET_AXIS_VISIBLE && (tmp_id == "bx1" ||  
          tmp_id == "bx2" || tmp_id2 == "by1" || tmp_id2 == "by2"))
      {
         cmd_manager->add_err(rval=ERR_BOX_AXIS_VIS,id,0);
      }
      else if ((tmp_id.length()) && (tmp_id2.length()))   // adjust both axes
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
            {
               if ((in_property != SET_AXIS_VISIBLE) ||
                   ((in_property == SET_AXIS_VISIBLE) &&
                   ((am->second != top) && (am->second != bottom))))
               {

                  sts = this->set_axis_mode_property (am->second, in_val, in_property);
                  if (sts)
                     sts = am->second->set_property(in_val, in_property);
                  if (sts != ERR_STATUS_GOOD)  rval = sts;
               }
            }
            found |= true;

            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
	       // If the current axis has changed update coords 
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_xaxis(tmp_id.c_str());
            if (ax && !rval)
            {
               rval = this->set_axis_mode_property (ax, in_val, in_property);
               if (rval)
                  rval = ax->set_property(in_val, in_property);
               found |= true;
	       if ((ax != bottom) && (ax != top))
	       {
                  cur_x = ax;
                  cur_xall = false;
		  // If the current axis has changed update coords
                  if (tmp_id != curr_xid)
                     update_coord = true;
	       }
               if (in_property == SET_AXIS_THICKNESS)  adjust_corner_widths(ax);
               cid->set_xaxis_id(tmp_id);
            }
         }
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
            {
               if ((in_property != SET_AXIS_VISIBLE) ||
                   ((in_property == SET_AXIS_VISIBLE) &&
                   ((am->second != left) && (am->second != right))))
               {
                  sts = this->set_axis_mode_property (am->second, in_val, in_property);
                  if (sts)
                     sts = am->second->set_property(in_val, in_property);
                  if (sts != ERR_STATUS_GOOD)  rval = sts;
               }
            }
            found |= true;
            if (axes_y.size () > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
	       // If the current axis has changed update coords 
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_yaxis(tmp_id2.c_str());
            if (ax && !rval)
            {
               rval = this->set_axis_mode_property (ax, in_val, in_property);
               if (rval)
                  rval = ax->set_property(in_val, in_property);
               found |= true;
	       if ((ax != left) && (ax != right))
	       {
                  cur_y = ax;
                  cur_yall = false;
		  // If the current axis has changed update coords
                  if (tmp_id2 != curr_yid)
                     update_coord = true;
	       }
               if (in_property == SET_AXIS_THICKNESS)  adjust_corner_widths(ax);
               cid->set_yaxis_id(tmp_id2);
            }
         }
         if (!found && !rval)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
      }
      else if (tmp_id.length())   // adjust x axis
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
            {
               if ((in_property != SET_AXIS_VISIBLE) ||
                   ((in_property == SET_AXIS_VISIBLE) &&
                   ((am->second != top) && (am->second != bottom))))
               {
                  sts = this->set_axis_mode_property (am->second, in_val, in_property);
                  if (sts)
                     sts = am->second->set_property(in_val, in_property);
                  if (sts != ERR_STATUS_GOOD)  rval = sts;
               }
            }
            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
	       // If the current axis has changed update coords 
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_xaxis(tmp_id.c_str());
            if (ax && !rval)
            {
               rval = this->set_axis_mode_property (ax, in_val, in_property);
               if (rval)
                  rval = ax->set_property(in_val, in_property);
	       if ((ax != bottom) && (ax != top))
	       {
                  cur_x = ax;
                  cur_xall = false;
		  // If the current axis has changed update coords 
                  if (tmp_id != curr_xid)
                     update_coord = true;
	       }
               cid->set_xaxis_id(tmp_id);
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
         }
      }
      else if (tmp_id2.length())     // adjust y axis
      {
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
            {
               if ((in_property != SET_AXIS_VISIBLE) ||
                   ((in_property == SET_AXIS_VISIBLE) &&
                   ((am->second != left) && (am->second != right))))
               {
                  sts = this->set_axis_mode_property (am->second, in_val, in_property);
                  if (sts)
                     sts = am->second->set_property(in_val, in_property);
                  if (sts != ERR_STATUS_GOOD)  rval = sts;
               }
            }
            if (axes_y.size () > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
	       // If the current axis has changed update coords
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_yaxis(tmp_id2.c_str());
            if (ax && !rval)
            {
               rval = this->set_axis_mode_property (ax, in_val, in_property);
               if (rval)
                  rval = ax->set_property(in_val, in_property);
	       if ((ax != left) && (ax != right))
	       {
                  cur_y = ax;
                  cur_yall = false;
		  // If the current axis has changed update coords
                  if (tmp_id2 != curr_yid)
                     update_coord = true;
	       }
               cid->set_yaxis_id(tmp_id2);
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id2.c_str());
         }
      }
      else
      {
	 int usedim = cid->get_axis_dim();
         int processed = 0; 

	 if ((usedim & X_AXIS) == X_AXIS)
	 {
            if ((in_property == SET_AXIS_VISIBLE)  &&
                ((cur_x == bottom) || (cur_x == top)))
               cmd_manager->add_err(rval=ERR_BOX_AXIS_VIS,id,0);
            // adjust x and y axes
            else if (cur_x && !rval)
            {
               rval = this->set_axis_mode_property (cur_x, in_val, in_property);
               if (rval)
                  rval = cur_x->set_property(in_val, in_property);
            }
            else if (cur_xall)
               for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
               {
                  if ((in_property != SET_AXIS_VISIBLE) ||
                      ((in_property == SET_AXIS_VISIBLE) &&
                      ((am->second != top) && (am->second != bottom))))
                  {
                     sts = this->set_axis_mode_property (am->second, in_val, in_property);
                     if (sts)
                        sts = am->second->set_property(in_val, in_property);
                     rval |= sts;
                  }
               }
	    else 
	       processed = X_AXIS;
	 }

	 if ((usedim & Y_AXIS) == Y_AXIS)
	 {
            if ((in_property == SET_AXIS_VISIBLE)  &&
                ((cur_y == left) || (cur_y == right)))
               cmd_manager->add_err(rval=ERR_BOX_AXIS_VIS,id,0);
            else if (cur_y && !rval)
            {
               rval = this->set_axis_mode_property (cur_y, in_val, in_property);
               if (rval)
                  rval = cur_y->set_property(in_val, in_property);
            }
            else if (cur_yall)
               for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
               {
                  if ((in_property != SET_AXIS_VISIBLE) ||
                      ((in_property == SET_AXIS_VISIBLE) &&
                      ((am->second != left) && (am->second != right))))
                  {
                     sts = this->set_axis_mode_property (am->second, in_val, in_property);
                     if (sts)
                        sts = am->second->set_property(in_val, in_property);
                     rval |= sts;
                  }
               }
	    else processed += (int)Y_AXIS;
         }
	 // attempting to set the property of border axes 
         if (processed == usedim)
	 {
	    if (in_property == SET_AXIS_TEXT || in_property == SET_AXIS_XLAB_TXT || in_property == SET_AXIS_YLAB_TXT)
	       cmd_manager->add_err(rval=ERR_BORDER_AXIS_LBL, id, 0);
	    else
	       cmd_manager->add_err(rval=ERR_BORDER_AXIS_PROP, id, 0);
	 }
      }
   }

   // if request was a successful axis visibility setting, reset axis style
   // to draw any boxed/closed axes which may have been obscured
   if (rval == ERR_STATUS_GOOD)
   { 
      bool check_axis_style = false;
      if (depth_table && in_property == SET_AXIS_DEPTH)
      {
         int val = *(static_cast<int *>(in_val));
         check_axis_style = depth_table->get_depth_visible (val);
      }
      else if (in_property == SET_AXIS_VISIBLE)
         check_axis_style = true;
 
      if (check_axis_style && (axis_style != chips_open))
         set_axis_style(axis_style);

      if (update_coord)
         rval = update_current_coord (cid);
   }

   return rval;
}


int ChipsPlot::get_axis_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;
   string axid;
   string tmp_id = cid->get_axis_id();
   ChipsAxis* ax = NULL;
   bool found = false;

   if (tmp_id.length())  
   {
      if (tmp_id == "all")
         cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
      else
      {
         // for now, we can only handle a single axis- so verify we
         // have the correct one...
         ax = get_xaxis(tmp_id.c_str());
         if (ax && (ax->get_id() == tmp_id))
         {
            found = true; 
         }
         else 
         {
            ax = get_yaxis(tmp_id.c_str());
            found = (ax && (ax->get_id() == tmp_id));
         }
         if (found)
            rval = ax->get_property(out_val, property);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
      }
   }
   else 
   {
      tmp_id = cid->get_xaxis_id();
      string tmp_id2 = cid->get_yaxis_id();

      if (tmp_id.length() && tmp_id2.length())
         cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
      else if (tmp_id.length())
      {
         if ((ax = get_xaxis(tmp_id)) != NULL)
            rval = ax->get_property(out_val, property);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
      }
      else if (tmp_id2.length())
      {
         if ((ax = get_yaxis(tmp_id2)) != NULL)
            rval = ax->get_property(out_val, property);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id2.c_str());
      }
      else
      {
	 int usedim = cid->get_axis_dim();
	 if (usedim == X_AXIS) 
         {
            if (cur_x)
               rval = cur_x->get_property(out_val, property);
            else if (cur_xall)
            {
               ChipsAxis* tmpx = get_xaxis();
               if (tmpx) 
                  rval = tmpx->get_property(out_val, property);
               else 
               {
                  string errstr = "Cannot retrieve a property from mulitple X axes at the same time.";
                  cmd_manager->add_err(errstr, rval=ERR_INVALID_GET_ALL,id,0);
               } 
            } 
            else 
            {
               string errstr = "Cannot retrieve current X axis property since there are no X axes.";
               cmd_manager->add_err(errstr, rval=ERR_NO_CURRENT_AXIS, id, 0);
            } 
         } 
         else if (usedim == Y_AXIS)
         {
            if (cur_y)
               rval = cur_y->get_property(out_val, property);
            else if (cur_yall)
            {
               ChipsAxis* tmpy = get_yaxis();
               if (tmpy)
                  rval = tmpy->get_property(out_val, property);
               else
               {
                  string errstr = "Cannot retrieve a property from mulitple Y axes at the same time.";
                  cmd_manager->add_err(errstr, rval=ERR_INVALID_GET_ALL,id,0);
               }
            }
            else
            {
               string errstr = "Cannot retrieve current Y axis property since there are no Y axes.";
               cmd_manager->add_err(errstr, rval=ERR_NO_CURRENT_AXIS, id, 0);
            } 
         }
         else if (cur_x && !(cur_y  || cur_yall))
            rval = cur_x->get_property(out_val, property);
         else if (cur_y && !(cur_x  || cur_xall))
            rval = cur_y->get_property(out_val, property);
         else
         {
            string errstr = "Cannot retrieve a property from both X and Y axes at the same time.";
            cmd_manager->add_err(errstr, rval=ERR_INVALID_GET_ALL,id,0);
         }  
      }
   }
   return rval;
}


int ChipsPlot::get_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;

   switch(property)
   {
      case SET_CONTOUR_COLOR:          // fallthrough intended
      case SET_CONTOUR_STYLE:          // fallthrough intended
      case SET_CONTOUR_THICKNESS:      // fallthrough intended
      case GET_CONTOUR_RANGE:          // fallthrough intended
      case GET_CONTOUR_XRANGE:         // fallthrough intended
      case GET_CONTOUR_YRANGE:         // fallthrough intended
      case GET_CONTOUR_ZRANGE:         // fallthrough intended
      case SET_CONTOUR_ALGORITHM:      // fallthrough intended
      case SET_CONTOUR_INTERVAL:       // fallthrough intended
      case SET_CONTOUR_NUMLEVELS:      // fallthrough intended
      case SET_CONTOUR_DEPTH:          // fallthrough intended
      case SET_CONTOUR_VISIBLE:        // fallthrough intended
      case SET_CONTOUR_MODE:      
         rval = get_contour_property(cid, out_val, property);
      break;
      case GET_CURVE_RANGE:            // fallthrough intended
      case GET_CURVE_XRANGE:           // fallthrough intended
      case GET_CURVE_YRANGE:           // fallthrough intended
      case SET_CURVE_SYMBOL_SIZE:      // fallthrough intended
      case SET_CURVE_SYMBOL_ANGLE:     // fallthrough intended
      case SET_CURVE_SYMBOL_COLOR:     // fallthrough intended
      case SET_CURVE_SYMBOL_STYLE:     // fallthrough intended
      case SET_CURVE_SYMBOL_FILL:      // fallthrough intended
      case SET_CURVE_LINE_THICKNESS:   // fallthrough intended
      case SET_CURVE_LINE_COLOR:       // fallthrough intended
      case SET_CURVE_DEPTH:            // fallthrough intended
      case SET_CURVE_BADDATA:          // fallthrough intended
      case SET_CURVE_LINE_STYLE:       // fallthrough intended
      case SET_CURVE_ERR_CAPLEN:       // fallthrough intended
      case SET_CURVE_ERR_COLOR:        // fallthrough intended
      case SET_CURVE_ERR_STYLE:        // fallthrough intended
      case SET_CURVE_ERR_UP:           // fallthrough intended
      case SET_CURVE_ERR_DOWN:         // fallthrough intended
      case SET_CURVE_ERR_LEFT:         // fallthrough intended
      case SET_CURVE_ERR_RIGHT:        // fallthrough intended
      case SET_CURVE_LIMLENGTH:        // fallthrough intended
      case SET_CURVE_LIMOVERRIDE:      // fallthrough intended
      case SET_CURVE_VISIBLE:          // fallthrough intended
      case SET_CURVE_ERR_THICKNESS:               
         rval = get_curve_property(cid, out_val, property);
      break;
      case SET_IMAGE_THRESHOLD:        // fallthrough intended
      case SET_IMAGE_ALPHA:            // fallthrough intended
      case SET_IMAGE_DEPTH:            // fallthrough intended
      case SET_IMAGE_INTERPOLATION:    // fallthrough intended
      case SET_IMAGE_VISIBLE:          // fallthrough intended
      case SET_IMAGE_CMAP:             // fallthrough intended
      case SET_IMAGE_CMAP_SIZE:        // fallthrough intended
      case SET_IMAGE_CMAP_INTERPOLATE: // fallthrough intended
      case SET_IMAGE_SCALE_CHANNELS:   // fallthrough intended
      case GET_IMAGE_RANGE:            // fallthrough intended
      case GET_IMAGE_XRANGE:           // fallthrough intended
      case GET_IMAGE_YRANGE:           // fallthrough intended
      case GET_IMAGE_CMAP_ID:          // fallthrough intended
      case SET_IMAGE_INV_CMAP:
         rval = get_image_property(cid, out_val, property);
      break;
      case SET_CBAR_DEPTH:             // fallthrough intended
      case SET_CBAR_VISIBLE:           // fallthrough intended
      case SET_CBAR_BORDERVIS:         // fallthrough intended
      case SET_CBAR_HALIGN:            // fallthrough intended
      case SET_CBAR_VALIGN:            // fallthrough intended
      case SET_CBAR_LENGTH:            // fallthrough intended
      case SET_CBAR_ORIENTATION:       // fallthrough intended
      case SET_CBAR_TICK_MODE:         // fallthrough intended
      case SET_CBAR_LABEL_ANGLE:       // fallthrough intended
      case SET_CBAR_LABEL_COLOR:       // fallthrough intended
      case SET_CBAR_LABEL_FONT:        // fallthrough intended
      case SET_CBAR_LABEL_FONTSTYLE:   // fallthrough intended
      case SET_CBAR_LABEL_HALIGN:      // fallthrough intended
      case SET_CBAR_LABEL_VALIGN:      // fallthrough intended
      case SET_CBAR_LABEL_SIZE:        // fallthrough intended
      case SET_CBAR_LABEL_VIS:         // fallthrough intended
      case SET_CBAR_LABEL_LOCATION:    // fallthrough intended
      case SET_CBAR_TEXT:              // fallthrough intended
      case SET_CBAR_TICK_THICKNESS:    // fallthrough intended
      case SET_CBAR_TICK_LENGTH:       // fallthrough intended
      case SET_CBAR_TICK_COLOR:        // fallthrough intended
      case SET_CBAR_TICK_STYLE:        // fallthrough intended
      case SET_CBAR_TICK_VIS:          // fallthrough intended
      case SET_CBAR_TICK_LOCATION:     // fallthrough intended
      case SET_CBAR_TICKLABEL_ANGLE:   // fallthrough intended
      case SET_CBAR_TICKLABEL_COLOR:   // fallthrough intended
      case SET_CBAR_TICKLABEL_FONT:    // fallthrough intended
      case SET_CBAR_TICKLABEL_FNTSTL:  // fallthrough intended
      case SET_CBAR_TICKLABEL_OFFSET:  // fallthrough intended
      case SET_CBAR_TICKLABEL_SIZE:    // fallthrough intended
      case SET_CBAR_TICKLABEL_HALIGN:  // fallthrough intended
      case SET_CBAR_TICKLABEL_VALIGN:  // fallthrough intended
      case SET_CBAR_TICKLABEL_VIS:     // fallthrough intended
      case SET_CBAR_OFFSET_PAR:        // fallthrough intended
      case SET_CBAR_OFFSET_PERP:       // fallthrough intended
      case SET_CBAR_WIDTH: 
         rval = get_colorbar_property(cid, out_val, property);
      break;
      case SET_HISTO_SYMBOL_STYLE:     // fallthrough intended
      case SET_HISTO_SYMBOL_COLOR:     // fallthrough intended
      case SET_HISTO_SYMBOL_SIZE:      // fallthrough intended
      case SET_HISTO_SYMBOL_ANGLE:     // fallthrough intended
      case SET_HISTO_SYMBOL_FILL:      // fallthrough intended
      case SET_HISTO_LINE_STYLE:       // fallthrough intended
      case SET_HISTO_LINE_COLOR:       // fallthrough intended
      case SET_HISTO_LINE_THICKNESS:   // fallthrough intended
      case SET_HISTO_DROPLINE:         // fallthrough intended
      case SET_HISTO_ERR_CAPLEN:       // fallthrough intended
      case SET_HISTO_ERR_STYLE:        // fallthrough intended
      case SET_HISTO_ERR_COLOR:        // fallthrough intended
      case SET_HISTO_ERR_UP:           // fallthrough intended
      case SET_HISTO_ERR_DOWN:         // fallthrough intended
      case SET_HISTO_DEPTH:            // fallthrough intended
      case GET_HISTO_RANGE:            // fallthrough intended
      case GET_HISTO_XRANGE:           // fallthrough intended
      case GET_HISTO_YRANGE:           // fallthrough intended
      case SET_HISTO_VISIBLE:          // fallthrough intended
      case SET_HISTO_FILL_OPACITY:     // fallthrough intended
      case SET_HISTO_FILL_COLOR:       // fallthrough intended
      case SET_HISTO_FILL_STYLE:       // fallthrough intended
      case SET_HISTO_ERR_THICKNESS:
         rval = get_histogram_property(cid, out_val, property);
      break;
      case SET_LABEL_TEXT:             // fallthrough intended
      case SET_LABEL_FONT:             // fallthrough intended
      case SET_LABEL_STYLE:            // fallthrough intended
      case SET_LABEL_COLOR:            // fallthrough intended
      case SET_LABEL_ANGLE:            // fallthrough intended
      case SET_LABEL_HALIGN:           // fallthrough intended
      case SET_LABEL_VALIGN:           // fallthrough intended
      case SET_LABEL_VISIBLE:          // fallthrough intended
      case SET_LABEL_DEPTH:            // fallthrough intended
      case SET_LABEL_SIZE:
         rval = get_label_property(cid, out_val, property);
      break;
      case SET_LINE_STYLE:             // fallthrough intended
      case SET_LINE_COLOR:             // fallthrough intended
      case SET_LINE_VISIBLE:           // fallthrough intended
      case SET_LINE_DEPTH:             // fallthrough intended
      case SET_LINE_EXTEND:            // fallthrough intended
      case SET_LINE_THICKNESS:
         rval = get_line_property(cid, out_val, property);
      break;
      case SET_POINT_SYMBOL_STYLE:     // fallthrough intended
      case SET_POINT_SYMBOL_COLOR:     // fallthrough intended
      case SET_POINT_SYMBOL_SIZE:      // fallthrough intended
      case SET_POINT_SYMBOL_FILL:      // fallthrough intended
      case SET_POINT_VISIBLE:          // fallthrough intended
      case SET_POINT_DEPTH:            // fallthrough intended
      case SET_POINT_SYMBOL_ANGLE:
         rval = get_point_property(cid, out_val, property);
      break;
      case SET_REGN_EDGE_COLOR:        // fallthrough intended
      case SET_REGN_EDGE_STYLE:        // fallthrough intended
      case SET_REGN_EDGE_THICKNESS:    // fallthrough intended
      case SET_REGN_FILL_STYLE:        // fallthrough intended
      case SET_REGN_FILL_COLOR:        // fallthrough intended
      case SET_REGN_ANGLE:             // fallthrough intended
      case SET_REGN_VISIBLE:           // fallthrough intended
      case SET_REGN_OPACITY:           // fallthrough intended
      case SET_REGN_DEPTH:
         rval = get_region_property(cid, out_val, property);
      break;
      case SET_AXIS_COLOR:             // fallthrough intended
      case SET_AXIS_PAD:               // fallthrough intended
      case SET_AXIS_DEPTH:             // fallthrough intended
      case SET_AXIS_MAJOR_COLOR:       // fallthrough intended
      case SET_AXIS_MINOR_COLOR:       // fallthrough intended
      case SET_AXIS_THICKNESS:         // fallthrough intended
      case SET_AXIS_MAJ_THICKNESS:     // fallthrough intended
      case SET_AXIS_MIN_THICKNESS:     // fallthrough intended
      case SET_AXIS_TEXT:              // fallthrough intended
      case SET_AXIS_XLAB_TXT:          // fallthrough intended
      case SET_AXIS_YLAB_TXT:          // fallthrough intended
      case SET_AXIS_TICKSTYLE:         // fallthrough intended
      case SET_AXIS_MAJ_STYLE:         // fallthrough intended
      case SET_AXIS_MIN_STYLE:         // fallthrough intended
      case SET_AXIS_TICKLABEL_STYLE:   // fallthrough intended
      case SET_AXIS_TICKFORMAT:        // fallthrough intended
      case SET_AXIS_MAJOR_NUM:         // fallthrough intended
      case SET_AXIS_MINOR_NUM:         // fallthrough intended
      case SET_AXIS_MAJORINT:          // fallthrough intended
      case SET_AXIS_MINORINT:          // fallthrough intended
      case SET_AXIS_MAJ_MODE:          // fallthrough intended
      case SET_AXIS_MIN_MODE:          // fallthrough intended
      case SET_AXIS_MAJORLENGTH:       // fallthrough intended
      case SET_AXIS_OFFSET_PAR:        // fallthrough intended
      case SET_AXIS_OFFSET_PERP:       // fallthrough intended
      case SET_AXIS_TICKLABEL_OFFSET:  // fallthrough intended
      case SET_AXIS_TICKLABEL_FONT:    // fallthrough intended
      case SET_AXIS_TICKLABEL_FNTSTL:  // fallthrough intended
      case SET_AXIS_TICKLABEL_SIZE:    // fallthrough intended
      case SET_AXIS_TICKLABEL_COLOR:   // fallthrough intended
      case SET_AXIS_TICKLABEL_ANGLE:   // fallthrough intended
      case SET_AXIS_TICKLABEL_HALIGN:  // fallthrough intended
      case SET_AXIS_TICKLABEL_VALIGN:  // fallthrough intended
      case SET_AXIS_TICKLABEL_VIS:     // fallthrough intended
      case SET_AXIS_LABEL_VIS:         // fallthrough intended
      case SET_AXIS_LABEL_FONT:        // fallthrough intended
      case SET_AXIS_LABEL_FONTSTYLE:   // fallthrough intended
      case SET_AXIS_LABEL_ANGLE:       // fallthrough intended
      case SET_AXIS_XLAB_ANG:          // fallthrough intended
      case SET_AXIS_YLAB_ANG:          // fallthrough intended
      case SET_AXIS_LABEL_HALIGN:      // fallthrough intended
      case SET_AXIS_LABEL_VALIGN:      // fallthrough intended
      case SET_AXIS_LABEL_SIZE:        // fallthrough intended
      case SET_AXIS_LABEL_COLOR:       // fallthrough intended
      case SET_AXIS_MAJGRD_COLOR:      // fallthrough intended
      case SET_AXIS_MAJGRD_STYLE:      // fallthrough intended
      case SET_AXIS_MAJGRD_VIS:        // fallthrough intended
      case SET_AXIS_MAJGRD_THICK:      // fallthrough intended
      case SET_AXIS_MINGRD_COLOR:      // fallthrough intended
      case SET_AXIS_MINGRD_STYLE:      // fallthrough intended
      case SET_AXIS_MINGRD_VIS:        // fallthrough intended
      case SET_AXIS_MINGRD_THICK:      // fallthrough intended
      case SET_AXIS_AUTOMIN:           // fallthrough intended
      case SET_AXIS_AUTOMAX:           // fallthrough intended
      case SET_AXIS_LOG_MIN:           // fallthrough intended
      case SET_AXIS_VISIBLE:           // fallthrough intended
      case SET_AXIS_MAJ_VIS:           // fallthrough intended
      case SET_AXIS_MIN_VIS:           // fallthrough intended
      case GET_AXIS_RANGE:             // fallthrough intended
      case SET_AXIS_REVERSE:           // fallthrough intended
      case GET_AXIS_SCALE:             // fallthrough intended
      case SET_AXIS_MINORLENGTH:
         rval = get_axis_property(cid, out_val, property);
      break; 
      case SET_PLOT_AXIS_STYLE:
      {
         V_AXISSTYLE_TYPE as = get_axis_style();
         string *sval =  (string*)out_val;
         *sval =v_axisstyle_to_string(as);
      }
      break;
      case SET_PLOT_TITLE:
      {
         string *sval =  (string*)out_val;
         *sval = get_title_text();
      }
      break;
      case SET_PLOT_CORNER:
      {
         string *sval =  (string*)out_val;
         *sval = get_corner_style();
      }
      break;
      case SET_PLOT_TITLE_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_title_color();
      }
      break;
      case SET_PLOT_ASPECT_RATIO:
      {
         string *sval =  (string*)out_val;
         *sval = get_aspect_ratio();
      }
      break;
      case SET_DATA_ASPECT_RATIO:
      {
	 string *sval =  (string*)out_val;
         string dsid = cid->get_coord_id();
         ChipsDataSystem *ds = NULL;

         string ax1 = cid->get_xaxis_id();
         string ax2 = cid->get_yaxis_id();
         ChipsAxis *ax = get_xaxis(ax1),
                   *ay = get_yaxis(ax2);

         *sval = "";

         // first attempt to use the DATA coord sys id from the ChipsId
         if (dsid.length())
         {
            COORD_SYS cds; 
            if ((cds = cid->get_coord_sys()) == DATA)
            {
               ds=dynamic_cast<ChipsDataSystem *>(coord->get_coord(DATA,dsid));
               if (!ds) 
               {
                  string errmsg = "The DATA coordinate system " + dsid + \
                     " does not exist in the current plot.";  

                  cmd_manager->add_err(errmsg.c_str(), rval=ERR_COORD_DS_DNE, 
                     id, 0);
               }
            }
            else 
            {
               char buf[256];
               const char* conv = v_coord_sys_to_string(cds);

               if (conv)
                  sprintf(buf, "Expected the coord sys value to be DATA instead of %s.\n", conv); 
               else 
                  sprintf(buf, "Expected the coord sys value to be DATA instead of %d.\n", (int) cds);
               cmd_manager->add_err(buf, rval=ERR_COORD_DS_DNE, id, 0);
            }
         }
         else if ((ds = get_datasystem (ax, ay)) == NULL)
         {
            if (ax1.length() && !ax)
            {
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0,
                                    "x axis", ax1.c_str());
            }
            else if (ax2.length() && !ay)
            {
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0,
                                    "y axis", ax2.c_str());
	    }
            else
            {
               cmd_manager->add_err(rval=ERR_COORD_DNE, id, 0, "DATA");
            }
         }
         if (ds)   *sval = ds->get_aspect_ratio();
      }
      break;
      case SET_PLOT_TITLE_ANGLE:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_title_angle();
      }
      break;
      case SET_PLOT_TITLE_HALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_title_halign();
      }
      break;
      case SET_PLOT_TITLE_VALIGN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_title_valign();
      }
      break;
      case SET_PLOT_TITLE_X:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_title_xpos();
      }
      break;
      case SET_PLOT_TITLE_Y:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_title_ypos();
      }
      break;
      case SET_PLOT_ASPECT_WIDTH:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_aspect_width();
      }
      break;
      case SET_PLOT_ASPECT_HEIGHT:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = get_aspect_height();
      }
      break;
      case SET_PLOT_VISIBLE:
      {
         int *iv = (int*)out_val;
         *iv = is_visible(); 
      }
      break;
      case SET_PLOT_TITLE_SIZE:
      {
         int *iv = (int*)out_val;
         *iv = get_title_size();
      }
      break;
      case SET_PLOT_DEPTH:
      {
         int *iv = (int*)out_val;
         *iv = get_title_depth();
      }
      break;
      case SET_PLOT_TITLE_FONT:
      {
         string *sval =  (string*)out_val;
         *sval = get_title_font();
      }
      break;
      case SET_PLOT_TITLE_FONTSTYLE:
      {
         string *sval =  (string*)out_val;
         *sval = get_title_fontstyle();
      }
      break; 
      case SET_PLOT_LMARGIN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = (float)get_plot_lmargin();
      }
      break;
      case SET_PLOT_RMARGIN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = (float)get_plot_rmargin();
      }
      break;
      case SET_PLOT_BMARGIN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = (float)get_plot_bmargin();
      }
      break;
      case SET_PLOT_TMARGIN:
      {
         float* fv = static_cast<float*>(out_val);
         *fv = (float)get_plot_tmargin();
      }
      break;
      case GET_PLOT_RANGE:
      {
         double axrange[2] = { 0.0, 0.0};
         double *range =  (double*)out_val;
         int init = 1;
         ChipsAxis *tmpAx = get_xaxis(cid->get_xaxis_id());
      
         if (tmpAx)
         {
            rval = tmpAx->get_property(axrange, GET_AXIS_RANGE);
            range[0] = axrange[0];
            range[1] = axrange[1];
         }
         else if (cur_xall)
         {
            for(axes_map::iterator am = axes_x.begin(); am!=axes_x.end(); ++am)
            {
	       rval |= (*am).second->get_property(axrange, GET_AXIS_RANGE);
               if (init)
               {
                  range[0] = axrange[0];
                  range[1] = axrange[1];
                  init = 0;
               }
               else
               {
                  if (axrange[0] < range[0])  range[0] = axrange[0];
                  if (axrange[1] > range[1])  range[1] = axrange[1];
               }
            }
         }
         else
            range[0] = range[1] = 0.0;     // initialize to 0 if no x axes
         axrange[0] = axrange[1] = 0.0; 
         tmpAx = get_yaxis(cid->get_yaxis_id());
         if (tmpAx)
         {
            rval = tmpAx->get_property(axrange, GET_AXIS_RANGE);
            range[2] = axrange[0];
            range[3] = axrange[1];
         }
         else if (cur_yall)
         {
            init = 1;
            for(axes_map::iterator am = axes_y.begin(); am!=axes_y.end(); ++am)
            {
	       rval |= (*am).second->get_property(axrange, GET_AXIS_RANGE);
               if (init)
               {
                  range[2] = axrange[0];
                  range[3] = axrange[1];
                  init = 0;
               }
               else
               {
                  if (axrange[0] < range[2])  range[2] = axrange[0];
                  if (axrange[1] > range[3])  range[3] = axrange[1];
               }
            }
         }
         else
            range[2] = range[3] = 0.0;     // initialize to 0 if no y axes
      }
      break;
      case GET_PLOT_DATA_RANGE:
      {
         ChipsAxis *tmpAx = get_xaxis(cid->get_xaxis_id());
         ChipsAxis *tmpAy = get_yaxis(cid->get_yaxis_id());
         double min_x, max_x, lin_x, min_y, max_y, lin_y;
         double tminx, tmaxx, tminy, tmaxy;
         int sts=0;
         if (tmpAx  && tmpAy)
         {
            sts = find_axis_min_max(tmpAx, tmpAx->get_orientation(), &min_x, &max_x, &lin_x, &tminx, &tmaxx, true, tmpAy);
            sts |= find_axis_min_max(tmpAy, tmpAy->get_orientation(), &min_y, &max_y, &lin_y, &tminy, &tmaxy, true, tmpAx);
         }

         double *range =  (double*)out_val;
         range[0] = tminx;
         range[1] = tmaxx;
         range[2] = tminy;
         range[3] = tmaxy;
         range[4] = lin_x; 
         range[5] = lin_y; 
      }
      break;
      case GET_PLOT_XRANGE:
      {
         double *range =  (double*)out_val;
         if (cur_x)
            rval = cur_x->get_property(out_val, GET_AXIS_RANGE);
         else if (cur_xall)
         {
            double axrange[2]; 
            int init = 1; 
            for(axes_map::iterator am = axes_x.begin(); am!=axes_x.end(); ++am)
            {
	       rval |= (*am).second->get_property(axrange, GET_AXIS_RANGE);
               if (init)
               {
                  range[0] = axrange[0];
                  range[1] = axrange[1];
                  init = 0;  
               }
               else
               {
                  if (axrange[0] < range[0])  range[0] = axrange[0]; 
                  if (axrange[1] > range[1])  range[1] = axrange[1]; 
               } 
            }
         }
         else
            range[0] = range[1] = 0.0;     // initialize to 0 if no x axes
      }
      break;
      case GET_PLOT_YRANGE:
      {
         double *range =  (double*)out_val;
         if (cur_y)
            rval = cur_y->get_property(out_val, GET_AXIS_RANGE);
         else if (cur_xall)
         {
            double axrange[2];
            int init = 1;
            axes_map::iterator ii;
            for(axes_map::iterator am = axes_y.begin(); am!=axes_y.end(); ++am)
            {
	       rval |= (*am).second->get_property(axrange, GET_AXIS_RANGE);
               if (init)
               {
                  range[0] = axrange[0];
                  range[1] = axrange[1];
                  init = 0;
               }
               else
               {
                  if (axrange[0] < range[0])  range[0] = axrange[0];
                  if (axrange[1] > range[1])  range[1] = axrange[1];
               }
            }
         }
         else
            range[0] = range[1] = 0.0;     // initialize to 0 if no y axes
      }
      break;
      default:
         cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,property);
      break;
   }
   return rval;
}



int ChipsPlot::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   int rval = ERR_STATUS_GOOD;
   string tmp_id = cid->get_axis_id(); 
   ChipsAxis* ax = NULL;
   bool found = false; 
   string curr_xid = cur_xall ? "all" : get_current_xaxis_id ();
   string curr_yid = cur_yall ? "all" : get_current_yaxis_id ();
   bool update_coord = false;

   if (tmp_id.length())   // adjust specified single axis or 'all'
   {
      if (chips_strcmp(tmp_id.c_str(), "all") == 0)
      {
	 for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
            (*am).second->set_tick_positions(arraysize, pos, vals);
	 for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
            (*am).second->set_tick_positions(arraysize, pos, vals);
         if (axes_x.size () > 3)
         {
            cur_xall = true;
            cur_x = NULL;
            cid->set_xaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_xid)
               update_coord = true;
         }
         if (axes_y.size () > 3)
         {
            cur_yall = true;
            cur_y = NULL;
            cid->set_yaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_yid)
	       update_coord = true;
         }
      }
      else        // adjust single specified axis 
      {
         ax = get_xaxis(tmp_id.c_str());
         if (ax)
         {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
            found = true;
            cur_x = ax;
            cur_xall = false; 
            cid->set_xaxis_id(tmp_id);
	    // If the current axis has changed update coords
            if (tmp_id != curr_xid)
               update_coord = true;
         }
         ax = get_yaxis(tmp_id.c_str());
         if (ax)
         {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
            found = true;
            cur_y = ax;
            cur_yall = false; 
            cid->set_yaxis_id(tmp_id);
	    // If the current axis has changed update coords
            if (tmp_id != curr_yid)
               update_coord = true;
         }
         if (!found)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
      }
   }
   else 
   {
      string tmp_id2 = cid->get_yaxis_id();
      tmp_id = cid->get_xaxis_id();

      if ((tmp_id.length()) && (tmp_id2.length()))   // adjust both axes
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
               (*am).second->set_tick_positions(arraysize, pos, vals);
            found |= true;
            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
	       // If the current axis has changed update coords
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_xaxis(tmp_id.c_str());
            if (ax)
            {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
               found |= true;
               cur_x = ax;
               cur_xall = false;
               cid->set_xaxis_id(tmp_id);
	       // If the current axis has changed update coords
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
         }
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am = axes_y.begin(); am!=axes_y.end(); ++am)
               (*am).second->set_tick_positions(arraysize, pos, vals);
            found |= true;
            if (axes_y.size () > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
	       // If the current axis has changed update coords
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_yaxis(tmp_id2.c_str());
            if (ax)
            {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
               found |= true;
               cur_y = ax;
               cur_yall = false;
               cid->set_yaxis_id(tmp_id2);
	       // If the current axis has changed update coords
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
         }
         if (!found)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id2.c_str());
      }
      else if (tmp_id.length())   // adjust x axis
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am = axes_x.begin(); am!=axes_x.end(); ++am)
               (*am).second->set_tick_positions(arraysize, pos, vals);
            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
	       // If the current axis has changed update coords
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_xaxis(tmp_id.c_str());
            if (ax)
            {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
               cur_x = ax;
               cur_xall = false;
               cid->set_xaxis_id(tmp_id);
	       // If the current axis has changed update coords
               if (tmp_id != curr_xid)
                  update_coord = true;
            }
            else 
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id.c_str());
         }
      }
      else if (tmp_id2.length())     // adjust y axis
      {
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am = axes_y.begin(); am!=axes_y.end(); ++am)
               (*am).second->set_tick_positions(arraysize, pos, vals);
            if (axes_y.size () > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
	       // If the current axis has changed update coords 
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
         }
         else
         {
            ax = get_yaxis(tmp_id2.c_str());
            if (ax)
            {
            list <ChipsAxis*> tl = bindings->get_axes(ax);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
               cur_y = ax;
               cur_yall = false;
               cid->set_yaxis_id(tmp_id2);
	       // If the current axis has changed update coords
               if (tmp_id2 != curr_yid)
                  update_coord = true;
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmp_id2.c_str());
         }
      }
      else
      {
         // adjust x and y axes 
         if (cur_x)
         {
            list <ChipsAxis*> tl = bindings->get_axes(cur_x);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
         }
         else if (cur_xall)
	    for(axes_map::iterator am = axes_x.begin(); am!=axes_x.end(); ++am)
               rval |= (*am).second->set_tick_positions(arraysize, pos, vals);
         if (cur_y)
         {
            list <ChipsAxis*> tl = bindings->get_axes(cur_y);
            list<ChipsAxis*>::iterator jj;
            for(jj=tl.begin(); jj != tl.end(); ++jj)
               rval = (*jj)->set_tick_positions(arraysize, pos, vals);
         }
         else if (cur_yall)
	    for(axes_map::iterator am = axes_y.begin(); am!=axes_y.end(); ++am)
               rval |= (*am).second->set_tick_positions(arraysize, pos, vals);
      }
   }
   if (update_coord && (rval == ERR_STATUS_GOOD))
      rval = update_current_coord (cid);
   return rval;
}


int ChipsPlot::set_curve_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string cvid;
   ChipsCurve* cv;

   if (cid)   cvid = cid->get_curve_id();
   if (cvid.length() == 0) cvid = currcurv;

   if (!curves_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"curves");
   else if (cvid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
         if ((st = (*cc).second->set_property(in_val, in_property)) != 
            ERR_STATUS_GOOD)
            rval = st;
   }
   else  if ((cv = get_curve(cvid)) != NULL)
      rval = cv->set_property(in_val, in_property); 
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",cvid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      currcurv = cvid;
      cid->set_curve_id(currcurv);
   }
   return rval;
}


int ChipsPlot::get_curve_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;
   string cvid;
   ChipsCurve* cv;

   if (cid)
   {
      cvid = cid->get_curve_id();
   }
   if (cvid.length() == 0) cvid = currcurv;

   if (!curves_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"curves");
   else if (cvid == "all" && curves.size() != 1)
      cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   else
   {
      if ((cv = get_curve(cvid)) != NULL)
      {
         switch(property)
         {
            case SET_CURVE_SYMBOL_SIZE:
            {
               int *iv = (int*)out_val;
               *iv = cv->get_symbolsize();
            }
            break;
            case SET_CURVE_SYMBOL_ANGLE:
            {
               float *fv = (float*)out_val;
               *fv = cv->get_symbolangle();
            }
            break;
            case SET_CURVE_SYMBOL_COLOR:
            {
               string *sval =  (string*)out_val;
               *sval = cv->get_symbolcolor();
            }
            break;
            case SET_CURVE_SYMBOL_STYLE:
            {
               int *iv = (int*)out_val;
               *iv = cv->get_symbolstyle();
            }
            break;
            case SET_CURVE_SYMBOL_FILL:       
            {
               int *iv = (int*)out_val;
               *iv = cv->get_symbolfill();
            }
            break; 
            case SET_CURVE_LINE_THICKNESS:   
            {
               float *fv = (float*)out_val;
               *fv = cv->get_linethickness();
            }
            break; 
            case SET_CURVE_LINE_COLOR:
            {
               string *sval =  (string*)out_val;
               *sval = cv->get_linecolor();
            }
            break;
            case SET_CURVE_LINE_STYLE:
            {
               int *iv = (int*)out_val;
               *iv = cv->get_linestyle();
            }
            break;
            case SET_CURVE_ERR_THICKNESS:
            {
               float *fv = (float*)out_val;
               *fv = cv->get_errthickness();
            }
            break;
            case SET_CURVE_ERR_COLOR:
            {
               string *sval =  (string*)out_val;
               *sval = cv->get_errcolor();
            }
            break;
            case SET_CURVE_ERR_STYLE:
            {
               string *sval =  (string*)out_val;
               *sval = v_errstyle_to_string(cv->get_errstyle());
            }
            break;
            case SET_CURVE_ERR_CAPLEN:       
            {
               int *iv = (int*)out_val;
               *iv = cv->get_errcaplength();
            }
            break; 
            case SET_CURVE_ERR_UP:       
            {
               int *iv = (int*)out_val;
               *iv = cv->get_errup();
            }
            break; 
            case SET_CURVE_ERR_DOWN:       
            {
               int *iv = (int*)out_val;
               *iv = cv->get_errdown();
            }
            break; 
            case SET_CURVE_ERR_LEFT:       
            {
               int *iv = (int*)out_val;
               *iv = cv->get_errleft();
            }
            break; 
            case SET_CURVE_ERR_RIGHT:      
            {
               int *iv = (int*)out_val;
               *iv = cv->get_errright();
            }
            break; 
            case SET_CURVE_LIMOVERRIDE:
            {
               int *iv = (int*)out_val;
               *iv = cv->get_limitsoverride();
            }
            break;
            case SET_CURVE_LIMLENGTH:
            {
               double *dv = (double*)out_val;
               *dv = cv->get_limitslength();
            }
            break;
            case GET_CURVE_RANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = cv->get_xmin(true);
               dv[1] = cv->get_xmax(true);
               dv[2] = cv->get_ymin(true);
               dv[3] = cv->get_ymax(true);
            }
            break;
            case GET_CURVE_XRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = cv->get_xmin(true);
               dv[1] = cv->get_xmax(true);
            }
            break;
            case GET_CURVE_YRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = cv->get_ymin(true);
               dv[1] = cv->get_ymax(true);
            }
            break;
            case SET_CURVE_BADDATA:
            {
               int *iv = (int*)out_val;
               *iv = cv->get_baddatamode();
            }
            break;
	    case SET_CURVE_DEPTH:
	    {
               int *iv = (int*)out_val;
               *iv = cv->get_depth();
	    }
            break; 
            case SET_CURVE_VISIBLE:
            {
               int *ival = (int*)out_val;
               *ival = cv->is_visible();
            }
            break;
            default:
               cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,property);
            break;
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",cvid.c_str());
   }
   return rval;
}


int ChipsPlot::set_image_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string imid;
   ChipsImage* im;

   if (cid)   imid = cid->get_image_id();
   if (imid.length() == 0) imid = currimage;

   if (!images_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"images");
   else if (imid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(images_map::iterator cc = images.begin(); cc != images.end(); ++cc)
         if ((st = (*cc).second->set_property(in_val, in_property)) != 
            ERR_STATUS_GOOD)
            rval = st;
   }
   else  if ((im = get_image(imid)) != NULL)
      rval = im->set_property(in_val, in_property); 
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",imid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      currimage = imid;
      cid->set_image_id(currimage);
   }
   return rval;
}


int ChipsPlot::get_image_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;
   string imid;
   ChipsImage* im;

   if (cid)
   {
      imid = cid->get_image_id();
   }
   if (imid.length() == 0) imid = currimage;

   if (!images_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"images");
   else if (imid == "all" && images.size() != 1)
      cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   else
   {
      if ((im = get_image(imid)) != NULL)
      {
         switch(property)
         {
	    case SET_IMAGE_THRESHOLD:
            {
               double *dv =  (double*)out_val;
               dv[0] = im->get_threshold_min();
               dv[1] = im->get_threshold_max();
            }
	    break; 
	    case SET_IMAGE_ALPHA:
            {
               double *dv =  (double*)out_val;
               dv[0] = im->get_alpha_min();
               dv[1] = im->get_alpha_max();
            }
	    break; 
            case SET_IMAGE_CMAP:
            {
               int *iv = (int*)out_val;
               *iv = im->get_colormap();
            }
            break;
	    case SET_IMAGE_INTERPOLATION:
	    {
               int *iv = (int*)out_val;
               *iv = im->get_interpolation();
	    }
	    break; 
            case SET_IMAGE_INV_CMAP:
            {
               int *ival = (int*)out_val;
               *ival = im->get_invert_colormap();
            }
            break;
	    case SET_IMAGE_CMAP_SIZE:
            {
               int *ival = (int*)out_val;
               *ival = im->get_colormap_size();
            }
            break;
            case SET_IMAGE_CMAP_INTERPOLATE:
            {
               int *ival = (int*)out_val;
               *ival = im->get_colormap_interpolate();
            }
            break;
            case SET_IMAGE_VISIBLE:
            {
               int *ival = (int*)out_val;
               *ival = im->is_visible();
            }
            break;
            case SET_IMAGE_SCALE_CHANNELS:
            {
               bool *bv = (bool*)out_val;
               *bv = im->get_normalize_channels();
            }
            break;
	    case SET_IMAGE_DEPTH:
	    {
               int *iv = (int*)out_val;
               *iv = im->get_depth();
	    }
            break;
            case GET_IMAGE_RANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = im->get_xmin(true);
               dv[1] = im->get_xmax(true);
               dv[2] = im->get_ymin(true);
               dv[3] = im->get_ymax(true);
            }
            break;
            case GET_IMAGE_XRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = im->get_xmin(true);
               dv[1] = im->get_xmax(true);
            }
            break;
            case GET_IMAGE_YRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = im->get_ymin(true);
               dv[1] = im->get_ymax(true);
            }
            break; 
            case GET_IMAGE_CMAP_ID:
            {
               unsigned int *ui = (unsigned int*)out_val;
               *ui = im->get_colormap_id(); 
            }
            break; 
            default:
               cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,property);
            break;
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",imid.c_str());
   }
   return rval;
}

int ChipsPlot::set_histogram_property(ChipsId* cid, void* in_val, 
   int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string chid;
   ChipsHistogram* ch;

   if (cid)   chid = cid->get_histogram_id();
   if (chid.length() == 0) chid = currhist;

   if (!histograms_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"histograms");
   else if (chid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(histograms_map::iterator hh = histograms.begin(); hh != 
         histograms.end(); ++hh)
         if ((st = (*hh).second->set_property(in_val, in_property)) !=
             ERR_STATUS_GOOD)
            rval = st;
   }
   else if ((ch = get_histogram(chid)) != NULL)
      rval = ch->set_property(in_val, in_property);
   else 
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",chid.c_str());

   if (rval == ERR_STATUS_GOOD)
   {
      currhist= chid;
      cid->set_histogram_id(currhist);
   }
   return rval;
}


int ChipsPlot::get_histogram_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;
   string chid;
   ChipsHistogram* ch;

   if (cid)
   {
      chid = cid->get_histogram_id();
   }
   if (chid.length() == 0) chid = currhist;

   if (!histograms_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"histograms");
   else if (chid == "all" && histograms.size() != 1)
      cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   else
   {
      if ((ch = get_histogram(chid)) != NULL)
      {
         switch(property)
         {
            case SET_HISTO_SYMBOL_COLOR:
            {
               string *sval =  (string*)out_val;
              *sval = ch->get_symbolcolor();
            }
            break;
            case SET_HISTO_SYMBOL_SIZE:
            {
               int *iv = (int*)out_val;
               *iv = ch->get_symbolsize();
            }
            break;
            case SET_HISTO_SYMBOL_ANGLE:
            {
               float *fv = (float*)out_val;
               *fv = ch->get_symbolangle();
            }
            break;
            case SET_HISTO_SYMBOL_STYLE:  
            {
               int *iv = (int*)out_val;
               *iv = ch->get_symbolstyle();
            }
            break; 
            case SET_HISTO_SYMBOL_FILL:       
            {
               int *iv = (int*)out_val;
               *iv = ch->get_symbolfill();
            }
            break; 
            case SET_HISTO_LINE_COLOR:
            {
               string *sval =  (string*)out_val;
              *sval = ch->get_linecolor();
            }
            break;
            case SET_HISTO_LINE_THICKNESS:
            {
               float *fv = (float*)out_val;
               *fv = ch->get_linethickness();
            }
            break;
            case SET_HISTO_LINE_STYLE:  
            {
               int *iv = (int*)out_val;
               *iv = ch->get_linestyle();
            }
            break; 
            case SET_HISTO_DROPLINE:       
            {
               int *iv = (int*)out_val;
               *iv = ch->get_dropline();
            }
            break; 
            case SET_HISTO_ERR_CAPLEN:
            {
               int *iv = (int*)out_val;
               *iv = ch->get_errcaplength();
            }
            break;
            case SET_HISTO_ERR_UP:       
            {
               int *iv = (int*)out_val;
               *iv = ch->get_errup();
            }
            break; 
            case SET_HISTO_ERR_DOWN:       
            {
               int *iv = (int*)out_val;
               *iv = ch->get_errdown();
            }
            break; 
            case SET_HISTO_ERR_STYLE:      
            {
               string *sval =  (string*)out_val;
               *sval = v_errstyle_to_string(ch->get_errstyle());
            }
            break; 
            case SET_HISTO_ERR_COLOR:      
            {
               string *sval =  (string*)out_val;
              *sval = ch->get_errcolor();
            }
            break; 
            case SET_HISTO_ERR_THICKNESS:
            {
               float *fv = (float*)out_val;
               *fv = ch->get_errthickness();
            }
            break; 
            case GET_HISTO_RANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ch->get_xmin(true);
               dv[1] = ch->get_xmax(true);
               dv[2] = ch->get_ymin(true);
               dv[3] = ch->get_ymax(true);
            }
            break;
            case GET_HISTO_XRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ch->get_xmin(true);
               dv[1] = ch->get_xmax(true);
            }
            break;
            case GET_HISTO_YRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ch->get_ymin(true);
               dv[1] = ch->get_ymax(true);
            }
            break;
            case SET_HISTO_DEPTH:
            {
               int *iv = (int*)out_val;
               *iv = ch->get_depth();
            }
            break;
            case SET_HISTO_VISIBLE:
            {
               int *ival = (int*)out_val;
               *ival = ch->is_visible();
            }
            break;
            case SET_HISTO_FILL_COLOR:
            {
               string *sval =  (string*)out_val;
              *sval = ch->get_fillcolor();
            }
            break;
            case SET_HISTO_FILL_OPACITY:
            {
               double *dval =  (double*)out_val;
              *dval = ch->get_fillopacity();
            }
            break;
            case SET_HISTO_FILL_STYLE:      
            {
               int *iv = (int*)out_val;
               *iv = ch->get_fillstyle();
            }
            break; 
            default:
               cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,property);
            break;
         }
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",chid.c_str());
   }
   return rval;
}


int ChipsPlot::set_contour_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string ctid;
   ChipsContour* ct;

   if (cid)   ctid = cid->get_contour_id();
   if (ctid.length() == 0) ctid = currcont;

   if (!contours_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"contours");
   else if (ctid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(contours_map::iterator cc=contours.begin(); cc!=contours.end(); ++cc)
         if ((st = (*cc).second->set_property(in_val, in_property)) !=
             ERR_STATUS_GOOD)
            rval = st;
   }
   else if ((ct = get_contour(ctid)) != NULL)
      rval = ct->set_property(in_val, in_property);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());

   if (rval == ERR_STATUS_GOOD)
   {
      currcont= ctid;
      cid->set_contour_id(currcont);
   }
   return rval;
}

int ChipsPlot::get_contour_property(ChipsId* cid, void* out_val, int property)
{
   int rval = ERR_STATUS_GOOD;
   string ctid;
   ChipsContour* ct;

   if (cid)
   {
      ctid = cid->get_contour_id();
   }
   if (ctid.length() == 0) ctid = currcont;

   if (!contours_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"contours");
   else if (ctid == "all" && contours.size() != 1)
      cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   else
   {
      if ((ct = get_contour(ctid)) != NULL)
      {
         switch(property)
         {
            case SET_CONTOUR_COLOR:
            {
               string *sval =  (string*)out_val;
               *sval = ct->get_color();
            }
            break;
            case SET_CONTOUR_STYLE:
            {
               string *sval =  (string*)out_val;
               *sval = v_line_to_string(ct->get_style());
            }
            break;
            case SET_CONTOUR_MODE:
            {
               string *sval =  (string*)out_val;
               *sval = v_tickmode_to_string(ct->get_mode());
            }
            break;
            case SET_CONTOUR_THICKNESS:
            {
               float *fv = (float*)out_val;
               *fv = ct->get_thickness();
            }
            break; 
            case GET_CONTOUR_RANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ct->get_xmin(true);
               dv[1] = ct->get_xmax(true);
               dv[2] = ct->get_ymin(true);
               dv[3] = ct->get_ymax(true);
            }
            break;
            case GET_CONTOUR_XRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ct->get_xmin(true);
               dv[1] = ct->get_xmax(true);
            }
            break;
            case GET_CONTOUR_YRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ct->get_ymin(true);
               dv[1] = ct->get_ymax(true);
            }
            break;
            case GET_CONTOUR_ZRANGE:
            {
               double *dv =  (double*)out_val;
               dv[0] = ct->get_zmin();
               dv[1] = ct->get_zmax();
            }
	    break;
	    case SET_CONTOUR_ALGORITHM:
	    {
               int *iv = (int*)out_val;
               *iv = (int)ct->get_algorithm();
	    }
            break;
	    case SET_CONTOUR_DEPTH:
	    {
               int *iv = (int*)out_val;
               *iv = (int)ct->get_depth();
	    }
            break;
            case SET_CONTOUR_VISIBLE:
            {
               int *ival = (int*)out_val;
               *ival = ct->is_visible();
            }
            break;
            case SET_CONTOUR_NUMLEVELS:
            {
               int *iv = (int*)out_val;
               *iv = (int)ct->get_numlevels();
            }
            break;
            case SET_CONTOUR_INTERVAL:
            {
               double *dv =  (double*)out_val;
               *dv = (double)ct->get_interval();
            }
            break;
            default:
               cmd_manager->add_err(rval=ERR_UNKNOWN_PROP_ID,id,0,property);
            break;
         }
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());
   }
   return rval;
}


int ChipsPlot::set_contour_levels(ChipsId* cid, unsigned int arraysize, 
   double* pos)
{
   int rval = ERR_STATUS_GOOD;
   string ctid;
   ChipsContour* ct;

   if (cid)   ctid = cid->get_contour_id();
   if (ctid.length() == 0) ctid = currcont;

   if (!contours_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"contours");
   else if (ctid == "all")
   {
      contours_map::iterator ii;
      list<ChipsContour*>::iterator jj;
      rval = ERR_STATUS_GOOD;

      for(ii=contours.begin(); ii != contours.end(); ++ii)
            rval = (*ii).second->set_levels(arraysize, pos);
   }
   else if ((ct = get_contour(ctid)) != NULL)
      rval = ct->set_levels(arraysize, pos);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      currcont= ctid;
      cid->set_contour_id(currcont);
   }
   return rval;
}

int ChipsPlot::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   int rval = ERR_STATUS_GOOD;
   string ctid;
   ChipsContour* ct;

   if (cid)   ctid = cid->get_contour_id();
   if (ctid.length() == 0) ctid = currcont;

   if (!contours_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"contours");
   else if (ctid == "all" && contours.size() != 1)
      cmd_manager->add_err(rval=ERR_INVALID_GET_ALL,id,0);
   else if ((ct = get_contour(ctid)) != NULL)
      rval = ct->get_levels(pos);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());

   return rval;
}

string  ChipsPlot::get_title_text()
{
   string rv = ""; 
   if (title)    rv = title->get_text();
   return rv; 
}

int ChipsPlot::set_title_text(const string& val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)    rval = title->set_text(val);
      else       cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title"); 
   return rval; 
}


int ChipsPlot::set_title_color(string val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
      rval = title->set_color(val);
   else 
      cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}


int ChipsPlot::set_aspect_ratio(string val, bool rec)
{
   int rval;
   if (!chips_strcmp(val.c_str(), "current"))
   {
      // fix aspect to current aspect ratio
      char value[80];
      if ((px1 - px0) == (py1 - py0)) 
         val = "1:1";
      else 
      {
         sprintf(value, "%2.1f:%2.1f", px1-px0, py1-py0);
         val = value; 
      }
   }
   else if (!chips_strcmp(val.c_str(), "fit"))
   {
      // fit aspect to current data object range 
      ChipsAxis *tmpax, *tmpay;
      char value[80];
      double min_x, max_x, lin_x, min_y, max_y, lin_y; 
      double tmp_tmin, tmp_tmax;
      int sts=0; 
      tmpax = get_xaxis(); 
      tmpay = get_yaxis ();
      if (tmpax && tmpay)
      {
         sts = find_axis_min_max(tmpax, tmpax->get_orientation (), &min_x, &max_x, &lin_x, &tmp_tmin, &tmp_tmax); 
         sts |= find_axis_min_max(tmpay, tmpay->get_orientation (), &min_y, &max_y, &lin_y, &tmp_tmin, &tmp_tmax); 
      }
      if (sts != ERR_STATUS_GOOD) return ERR_NO_DATA_FOR_AUTO; 
      if ((max_x - min_x) == (max_y - min_y))
         val = "1:1";
      else
      {
         sprintf(value, "%2.1f:%2.1f", max_x-min_x, max_y-min_y);
         val = value;
      }
   }
   if ((rval = string_to_ratio_values(val.c_str(), aspect_w, aspect_h)) ==
      ERR_STATUS_GOOD)
   {
      aspect_ratio = val;
      aspect_use = 1; 
      adjust_plot_anchors(); 
      recalculate_position(); 
   }
   else if (val == "")
   {
      aspect_ratio = "";
      aspect_use = 0; 
      aspect_w = aspect_h = 0.0;
      adjust_plot_anchors(); 
      recalculate_position(); 
      rval = ERR_STATUS_GOOD; 
   }
   else 
   {
      cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"plot aspect.ratio", val.c_str(), "\"current\"|\"fit\"|\"\"|\"<width>:<height>\"");
   }

   if (rval == ERR_STATUS_GOOD && bindings)
   {
      // if plot aspect has changed- recompute limits for data aspects
      ChipsAxis* tmpax = get_xaxis();
      if (tmpax)
      {
         double min, max;
         min = tmpax->get_absolute_min (false);
         max = tmpax->get_absolute_max (false);
         bindings->update_connected_nodes(tmpax, min, max, false, AUTOLIM_OFF, 1, rec);
      }
   }
   return rval;
}


string  ChipsPlot::get_title_color()
{
   string rv = "";
   if (title)   rv = title->get_color();
   return rv; 
}


int ChipsPlot::set_title_font(string val)
{
   if (title)  return title->set_font(val);
   cmd_manager->add_err(ERR_OBJ_DNE,id,0,"plot title");
   return ERR_OBJ_DNE;
}


string  ChipsPlot::get_title_font()
{
   string rv = "";
   if (title)   return title->get_font();
   return rv;
}

int ChipsPlot::set_title_size(int sz)
{
   if (title)  return title->set_size(sz);
   cmd_manager->add_err(ERR_OBJ_DNE,id,0,"plot title");
   return ERR_OBJ_DNE; 
}


int ChipsPlot::get_title_size()
{
   if (!title)  return 0;
   return title->get_size();
}

int ChipsPlot::set_title_depth(int dep)
{
   title_depth = dep; 
   if (!title)  return ERR_STATUS_GOOD;
   return title->set_depth(dep);
}

int ChipsPlot::get_title_depth()
{
   if (!title)   return title_depth;
   return (title_depth = title->get_depth());
}

int ChipsPlot::set_title_fontstyle(string val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
   {
      if ((rval = string_to_v_fontstyle(val.c_str(), &title_fontstyle))
          == ERR_STATUS_GOOD)
         title->set_fontstyle(title_fontstyle);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}

string  ChipsPlot::get_title_fontstyle()
{
   string sval = ""; 
   if (title)   sval = v_fontstyle_to_string(title->get_fontstyle());
   return sval; 
}

int ChipsPlot::set_axis_style(V_AXISSTYLE_TYPE val, bool rec)
{
   int ival, rval = ERR_STATUS_GOOD;

   switch(val)
   {
      case chips_open:
         axis_style = val;
         ival = 0; 
         bottom->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         top->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         left->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         right->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
      break; 
      case chips_box:
         axis_style = val;
         ival = 1; 
         bottom->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         bottom->hide_major();
         bottom->hide_minor();
         top->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         top->hide_major();
         top->hide_minor();
         left->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         left->hide_major();
         left->hide_minor();
         right->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         right->hide_major();
         right->hide_minor();
         cover_box_axes(XY_AXIS,rec); 
      break; 
      case chips_closed:
         axis_style = val;
         ival = 1; 
         bottom->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         bottom->show_major();
         bottom->show_minor();
         top->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         top->show_major();
         top->show_minor();
         left->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         left->show_major();
         left->show_minor();
         right->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec); 
         right->show_major();
         right->show_minor();
         cover_box_axes(XY_AXIS,rec); 
      break; 
      default:
         cmd_manager->add_err(rval=ERR_BAD_PLOTSTYLE,id,0);
      break;
   }
   return rval;
}


int ChipsPlot::set_corner_style(V_CORNERSTYLE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (valid_cornerstyle_type(val))
   { 
      corner_style = val; 
      rval = adjust_corner_style(); 
   }
   else
      cmd_manager->add_err(rval=ERR_BAD_CORNERSTYLE,id,0);
   return rval;
}


string  ChipsPlot::get_corner_style()
{
   return v_cornerstyle_to_string(corner_style);
}

int ChipsPlot::get_annotation_position (const ChipsId* cid,
                                          V_OBJECT_TYPE obj, ChipsVector &pos)
{

   int  rval = ERR_FEATURE_DNE;
   string oid;
   cid->get_object_id (obj, oid);

   switch (obj)
   {
      case chips_point:
      {
         if (oid.length() == 0)  oid = currpoint;
         if (points.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"points");
            return ERR_OBJ_LIST_EMPTY;
         }
         if (chips_strcmp(oid.c_str(), "all") == 0)
            cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
         else if (points.count(oid))
            points[oid]->get_position (pos);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",oid.c_str());
      }
      break;
      case chips_line:
      {
         if (oid.length() == 0)  oid = currline;
         if (lines.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"lines");
            return ERR_OBJ_LIST_EMPTY;
         }
         if (chips_strcmp(oid.c_str(), "all") == 0)
            cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
         else if (lines.count(oid))
            lines[oid]->get_position (pos);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",oid.c_str());
      }
      break;
      case chips_region:
      {
         if (oid.length() == 0)  oid = currregion;
         if (regions.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"regions");
            return ERR_OBJ_LIST_EMPTY;
         }
         if (chips_strcmp(oid.c_str(), "all") == 0)
            cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
         else if (regions.count(oid))
            regions[oid]->get_position (pos);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",oid.c_str());
      }
      break;
      case chips_label:
      {
         if (oid.length() == 0)  oid = currlabel;
         if (labels.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"labels");
            return ERR_OBJ_LIST_EMPTY;
         }
         if (chips_strcmp(oid.c_str(), "all") == 0)
            cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
         else if (labels.count(oid))
            labels[oid]->get_position (pos);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
      }
      break;
      case chips_colorbar:
      {
         if (oid.length() == 0)  oid = currcbar;
         if (cbars.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
            return ERR_OBJ_LIST_EMPTY;
         }
         if (chips_strcmp(oid.c_str(), "all") == 0)
            cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
         else if (cbars.count(oid))
            cbars[oid]->get_position (pos);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
      }
      break;
      default:
         rval = ERR_FEATURE_DNE;
      break;
   }

   return rval;

}


int ChipsPlot::change_annotation_coordinate (const ChipsId* cid,
                                          V_OBJECT_TYPE obj,
                                          COORD_SYS coordsys,
                                          const string& coordid)
{
   int  rval = ERR_FEATURE_DNE;
   string oid;
   cid->get_object_id (obj, oid);

   switch (obj)
   {
      case chips_point:
      {
         if (oid.length() == 0)  oid = currpoint;
         if (points.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"points");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsPoint*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=points.begin(); ii != points.end(); ++ii)
               ii->second->change_coordinate (coordsys, coordid);
         }
         else if (points.count(oid))
            points[oid]->change_coordinate (coordsys, coordid);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",oid.c_str());
      }
      break;
      case chips_line:
      {
         if (oid.length() == 0)  oid = currline;
         if (lines.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"line");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsLine*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=lines.begin(); ii != lines.end(); ++ii)
               ii->second->change_coordinate (coordsys, coordid);
         }
         else if (lines.count(oid))
            lines[oid]->change_coordinate (coordsys, coordid);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",oid.c_str());
      }
      break;
      case chips_region:
      {
         if (oid.length() == 0)  oid = currregion;
         if (regions.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"region");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsRegion*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=regions.begin(); ii != regions.end(); ++ii)
               ii->second->change_coordinate (coordsys, coordid);
         }
         else if (regions.count(oid))
            regions[oid]->change_coordinate (coordsys, coordid);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",oid.c_str());
      }
      break;
      case chips_label:
      {
         if (oid.length() == 0)  oid = currlabel;
         if (labels.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"label");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsLabel*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=labels.begin(); ii != labels.end(); ++ii)
               ii->second->change_coordinate (coordsys, coordid);
         }
         else if (labels.count(oid))
            labels[oid]->change_coordinate (coordsys, coordid);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
      }
      break;
      case chips_colorbar:
      {
         if (oid.length() == 0)  oid = currcbar;
         if (cbars.empty())
         {
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"colorbar");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsColorBar*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=cbars.begin(); ii != cbars.end(); ++ii)
               ii->second->change_coordinate (coordsys, coordid);
         }
         else if (cbars.count(oid))
            cbars[oid]->change_coordinate (coordsys, coordid);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
      }
      break;



      default:
         rval = ERR_FEATURE_DNE;
      break;
   }

   return rval;
}


int ChipsPlot::resize_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *bounds)
{
   int  rval = ERR_STATUS_GOOD;
   string oid;
   cid->get_object_id (obj, oid);
   COORD_SYS coord_sys = cid->get_coord_sys ();
   string cd_id = cid->get_coord_id ();



   switch (obj)
   {
      case chips_plot:
      {
         if (subobj == chips_plot_title)
            rval = this->title->resize(coord_sys, cd_id, bounds);
         else
            rval = ERR_FEATURE_DNE;
      }
      break;
      case chips_label:
      {
         if (oid.length() == 0)  oid = currlabel;
         if (labels.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"labels");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsLabel*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            int sts;
            for(ii=labels.begin(); ii != labels.end(); ++ii)
              if ((sts |= (*ii).second->resize (coord_sys, cd_id, bounds)) !=
                    ERR_STATUS_GOOD)   rval = sts;
         }
         else if (labels.count(oid))
            rval = labels[oid]->resize(coord_sys, cd_id, bounds);
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         currlabel = oid;
      }
      break;
      case chips_point:
      {
         if (oid.length() == 0)  oid = currpoint;
         if (points.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"points");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsPoint*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            int sts;
            for(ii=points.begin(); ii != points.end(); ++ii)
              if ((sts |= (*ii).second->resize (coord_sys, cd_id, bounds)) !=
                    ERR_STATUS_GOOD)   rval = sts;
         }
         else if (points.count(oid))
            rval = points[oid]->resize(coord_sys, cd_id, bounds);
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         currpoint = oid;
      }
      break;
      case chips_colorbar:
      {
         if (oid.length() == 0)  oid = currpoint;
         if (cbars.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbar");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsColorBar*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            int sts;
            for(ii=cbars.begin(); ii != cbars.end(); ++ii)
            {
               if (subobj == chips_not_subobject)
               
                  sts |= ii->second->resize (coord_sys, cd_id, bounds);
               else
                  sts |= ii->second->resize_object (subobj, coord_sys, cd_id, bounds);
              
               if (sts != ERR_STATUS_GOOD)
                  rval = sts;
                  
            }
         }
         else if (cbars.count(oid))
         {
            if (subobj == chips_not_subobject)
               rval = cbars[oid]->resize(coord_sys, cd_id, bounds);
            else
               rval = cbars[oid]->resize_object(subobj, coord_sys, cd_id, bounds);
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         currcbar = oid;
      }
      break;
      case chips_axis:
      {
         if (subobj != chips_not_subobject)
         {
            string xid = oid;
            string yid = oid;

            if (oid.length() == 0)
            {
               xid = (cur_xall) ? "all" : cur_x->get_id ();
               yid = (cur_yall) ? "all" : cur_y->get_id ();
            }
            if (axes_x.empty() && axes_y.empty ())
            {
               cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"axes");
               return ERR_OBJ_LIST_EMPTY;
            }
            map <string, ChipsAxis*>::iterator ii;
            if (chips_strcmp(xid.c_str(), "all") == 0)
            {
               for(ii=axes_x.begin(); ii != axes_x.end(); ++ii)
                  ii->second->resize_object (subobj, coord_sys, cd_id, bounds);
               rval = ERR_STATUS_GOOD;
            }
            else if (axes_x.count(xid))
            {
               axes_x[xid]->resize_object (subobj, coord_sys, cd_id, bounds);
               rval = ERR_STATUS_GOOD;
            }
            else
               rval = ERR_OBJ_VAL_DNE;

            if (chips_strcmp(yid.c_str(), "all") == 0)
            {
               for(ii=axes_y.begin(); ii != axes_y.end(); ++ii)
                  ii->second->resize_object (subobj, coord_sys, cd_id, bounds);
               rval = ERR_STATUS_GOOD;
            }
            else if (axes_y.count(yid))
            {
               axes_y[yid]->resize_object (subobj, coord_sys, cd_id, bounds);
               rval = ERR_STATUS_GOOD;
            }
            else
               rval=ERR_OBJ_VAL_DNE;

            if (rval == ERR_OBJ_VAL_DNE)
            {
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",oid.c_str());
               return ERR_OBJ_VAL_DNE;
            }
         }
      }
      break;
      default:
         rval = ERR_FEATURE_DNE;
      break;
   }
   return rval;

}

int ChipsPlot::select_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected)
{
   int  rval = ERR_STATUS_GOOD;
   string oid;
   cid->get_object_id (obj, oid);


   switch (obj)
   {
      case chips_plot:
      {
         if (subobj == chips_plot_title)
         {
           if (selected) this->title->select ();
             else this->title->unselect ();
         }
         else
           rval = ERR_FEATURE_DNE;
      }
      break;
      case chips_label:
      {
         if (oid.length() == 0)  oid = currlabel;
         if (labels.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"labels");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsLabel*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=labels.begin(); ii != labels.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (labels.count(oid))
         {
            if (selected) labels[oid]->select();
              else labels[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         currlabel = oid;
      }
      break;
      case chips_line:
      {
         if (oid.length() == 0)  oid = currline;
         if (lines.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"lines");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsLine*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=lines.begin(); ii != lines.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (lines.count(oid))
         {
            if (selected) lines[oid]->select();
              else lines[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_point:
      {
         if (oid.length() == 0)  oid = currpoint;
         if (points.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"points");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsPoint*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=points.begin(); ii != points.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (points.count(oid))
         {
            if (selected) points[oid]->select();
              else points[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }

      }
      break;
      case chips_region:
      {
         if (oid.length() == 0)  oid = currregion;
         if (regions.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"regions");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsRegion*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=regions.begin(); ii != regions.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (regions.count(oid))
         {
            if (selected) regions[oid]->select();
              else regions[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_axis:
      {
         string xid = oid;
         string yid = oid;

         if (oid.length() == 0)
         {
            xid = (cur_xall) ? "all" : cur_x->get_id ();
            yid = (cur_yall) ? "all" : cur_y->get_id ();
         }
         if (axes_x.empty() && axes_y.empty ())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"axes");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsAxis*>::iterator ii;
         if (chips_strcmp(xid.c_str(), "all") == 0)
         {
            for(ii=axes_x.begin(); ii != axes_x.end(); ++ii)
            {
               if (subobj == chips_not_subobject)
               {
                 if (selected) ii->second->select ();
                   else ii->second->unselect ();
               }
               else
                  ii->second->select_object (subobj, selected);
            }
            rval = ERR_STATUS_GOOD;
         }
         else if (axes_x.count(xid))
         {
            if (subobj == chips_not_subobject)
            {
               if (selected) axes_x[xid]->select();
                 else axes_x[xid]->unselect ();
            }
            else
               axes_x[xid]->select_object (subobj, selected);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval = ERR_OBJ_VAL_DNE;

         if (chips_strcmp(yid.c_str(), "all") == 0)
         {
            for(ii=axes_y.begin(); ii != axes_y.end(); ++ii)
            {
               if (subobj == chips_not_subobject)
               {
                 if (selected) ii->second->select ();
                   else ii->second->unselect ();
               }
               else
                  ii->second->select_object (subobj, selected);
            }
            rval = ERR_STATUS_GOOD;
         }
         else if (axes_y.count(yid))
         {
            if (subobj == chips_not_subobject)
            {
               if (selected) axes_y[yid]->select();
                 else axes_y[yid]->unselect ();
            }
            else
               axes_y[yid]->select_object (subobj, selected);
            rval = ERR_STATUS_GOOD;
         }
         else 
            rval=ERR_OBJ_VAL_DNE;

         if (rval == ERR_OBJ_VAL_DNE)
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_colorbar:
      {
         if (oid.length() == 0)  oid = currcbar;
         if (cbars.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsColorBar*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=cbars.begin(); ii != cbars.end(); ++ii)
            {
               if (subobj == chips_not_subobject)
               {
                  if (selected) ii->second->select ();
                    else ii->second->unselect ();
               }
               else
                  ii->second->select_object (subobj, selected);
            }
         }
         else if (cbars.count(oid))
         {
            if (subobj == chips_not_subobject)
            {
               if (selected) cbars[oid]->select();
                 else cbars[oid]->unselect ();
            }
            else
               cbars[oid]->select_object (subobj, selected);
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_curve:
      {
         if (oid.length() == 0)  oid = currcurv;
         if (curves.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"curves");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsCurve*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=curves.begin(); ii != curves.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (curves.count(oid))
         {
            if (selected) curves[oid]->select();
              else curves[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_histogram:
      {
         if (oid.length() == 0)  oid = currhist;
         if (histograms.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"histograms");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsHistogram*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=histograms.begin(); ii != histograms.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (histograms.count(oid))
         {
            if (selected) histograms[oid]->select();
              else histograms[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_contour:
      {
         if (oid.length() == 0)  oid = currcont;
         if (contours.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"contours");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsContour*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=contours.begin(); ii != contours.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (contours.count(oid))
         {
            if (selected) contours[oid]->select();
              else contours[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      case chips_image: 
      {
         if (oid.length() == 0)  oid = currimage;
         if (images.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"images");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsImage*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=images.begin(); ii != images.end(); ++ii)
            {
               if (selected) ii->second->select ();
                 else ii->second->unselect ();
            }
         }
         else if (images.count(oid))
         {
            if (selected) images[oid]->select();
              else images[oid]->unselect ();
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      default:
         rval = ERR_FEATURE_DNE;
      break;
   }
   return rval;

}

int ChipsPlot::rotate_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle)
{
   int  rval = ERR_STATUS_GOOD;


   switch (obj)
   {
      case chips_plot:
         if (subobj == chips_plot_title)
            this->set_property (cid, &angle, SET_PLOT_TITLE_ANGLE);
         else
            rval = ERR_FEATURE_DNE;
      break;
      case chips_label:
         this->set_property (cid, &angle, SET_LABEL_ANGLE);
      break;
      case chips_point:
         this->set_property (cid, &angle, SET_POINT_SYMBOL_ANGLE);
      break;
      case chips_region:
         this->set_property (cid, &angle, SET_REGN_ANGLE);
      break;
      case chips_colorbar:
      {
         string oid = cid->get_colorbar_id ();
         if (oid.length() == 0)  oid = currcbar;
         if (cbars.empty())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsColorBar*>::iterator ii;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(ii=cbars.begin(); ii != cbars.end(); ++ii)
               ii->second->rotate_object (subobj, angle);
         }
         else if (cbars.count(oid))
            cbars[oid]->rotate_object (subobj, angle);
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }

      }
      break;
      case chips_axis:
      {
         string xid = cid->get_axis_id ();
         string yid = cid->get_axis_id ();

         if (xid.length() == 0)
            xid = (cur_xall) ? "all" : cur_x->get_id ();
         if (yid.length() == 0)
            yid = (cur_yall) ? "all" : cur_y->get_id ();
            
         if (axes_x.empty() && axes_y.empty ())
         {
            cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"axes");
            return ERR_OBJ_LIST_EMPTY;
         }
         map <string, ChipsAxis*>::iterator ii;
         if (chips_strcmp(xid.c_str(), "all") == 0)
         {
            for(ii=axes_x.begin(); ii != axes_x.end(); ++ii)
               ii->second->rotate_object (subobj, angle);
            rval = ERR_STATUS_GOOD;
         }
         else if (axes_x.count(xid))
         {
            axes_x[xid]->rotate_object (subobj, angle);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval = ERR_OBJ_VAL_DNE;

         if (chips_strcmp(yid.c_str(), "all") == 0)
         {
            for(ii=axes_y.begin(); ii != axes_y.end(); ++ii)
               ii->second->rotate_object (subobj, angle);
            rval = ERR_STATUS_GOOD;
         }
         else if (axes_y.count(yid))
         {
            axes_y[yid]->rotate_object (subobj, angle);
            rval = ERR_STATUS_GOOD;
         }
         else
            rval=ERR_OBJ_VAL_DNE;

         if (rval == ERR_OBJ_VAL_DNE)
         {
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",xid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
      }
      break;
      default:
         rval = ERR_FEATURE_DNE;
      break;
   }
   return rval;

}

int ChipsPlot::move_object (ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj,
   double xx, double yy, int mode)
{

   int rval = ERR_STATUS_GOOD;
   ChipsId tmp_id (*cid);
   ChipsCoordinate *tmp_coord;
   COORD_SYS cd_sys = tmp_id.get_coord_sys ();

   if (cd_sys != COORD_NONE)
   {
      tmp_coord = find_coordinate_system (tmp_id, rval);
      if (rval)
          return rval;

      cd_sys = tmp_id.get_coord_sys ();
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((cd_sys == WINDOW_NORM) || (cd_sys == PIXEL) || (cd_sys == PIXEL_REAL)
)
         tmp_id.set_coord_id ("");
   }

   switch (obj)
   {
      case chips_plot:
      {
         if (subobj == chips_plot_title)
         {
            ChipsVector pos1 (this->title_xpos_norm, this->title_ypos_norm);
            ChipsVector pos2 (xx, yy);
            ChipsVector result (pos1);
            float tmp_xpos; 
            float tmp_ypos;

            string cd_id = tmp_id.get_coord_id ();

            tmp_coord->convert_cs_to_cs (&pos2, &pos2, 1, cd_sys, cd_id, FRAME_NORM, "");
            if (mode == V_MOVE_DELTA)
               pos1 += pos2;
            else
               pos1 = pos2;

            tmp_coord->convert_cs_to_cs (&pos1, &result, 1, FRAME_NORM, "", PLOT_NORM, "");
            tmp_xpos = result[0];

            // title y position if pixel offset from the top of the plot
            // calculate the top of the plot in pixel coordinates
            pos2[0] = 1;
            pos2[1] = 1;

            if (py1) pos2[1] = py1;
              else coord->convert_cs_to_cs(&pos2,&pos2,1,PLOT_NORM,"",PIXEL,"");
          
            tmp_coord->convert_cs_to_cs (&pos1, &result, 1, FRAME_NORM, "", PIXEL, "");
            tmp_ypos = result[1] - pos2[1];
            this->set_property (&tmp_id, &tmp_xpos, SET_PLOT_TITLE_X);
            this->set_property (&tmp_id, &tmp_ypos, SET_PLOT_TITLE_Y);
         }
         else
            rval = ERR_FEATURE_DNE;
      }
      break;   
      case chips_label:
         rval = this->move_label (&tmp_id, xx, yy, mode);
      break;
      case chips_line:
         rval = this->move_line (&tmp_id, xx, yy, mode);
      break;
      case chips_point:
         rval = this->move_point (&tmp_id, xx, yy, mode);
      break;
      case chips_region:
         rval = this->move_region (&tmp_id, xx, yy, mode);
      break;
      case chips_axis: 
         if (subobj == chips_not_subobject)
            rval = this->move_axis (&tmp_id, xx, yy, mode);
         else
         {
            string xid = tmp_id.get_axis_id ();
            string yid = tmp_id.get_axis_id ();
            string cd_id = tmp_id.get_coord_id ();

            if (xid.length() == 0)
               xid = (cur_xall) ? "all" : cur_x->get_id ();
            if (yid.length() == 0)
               yid = (cur_yall) ? "all" : cur_y->get_id ();

            if (axes_x.empty() && axes_y.empty ())
            {
               cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"axes");
               return ERR_OBJ_LIST_EMPTY;
            }
            map <string, ChipsAxis*>::iterator ii;
            if (chips_strcmp(xid.c_str(), "all") == 0)
            {
               int sts=ERR_STATUS_GOOD;
               for(ii=axes_x.begin(); ii != axes_x.end(); ++ii)
               {
                  if ((sts |= ii->second->move_object (subobj, cd_sys, cd_id, xx, yy, mode)) != ERR_STATUS_GOOD)
                     rval = sts;
               }

            }
            else if (axes_x.count(xid))
               rval = axes_x[xid]->move_object (subobj, cd_sys, cd_id, xx, yy, mode);
            else
               rval = ERR_OBJ_VAL_DNE;

            if (chips_strcmp(yid.c_str(), "all") == 0)
            {
               int sts = ERR_STATUS_GOOD;
               for(ii=axes_y.begin(); ii != axes_y.end(); ++ii)
               {
                  if ((sts |= ii->second->move_object (subobj, cd_sys, cd_id, xx, yy, mode)) != ERR_STATUS_GOOD)
                     rval = sts;
               }
            }
            else if (axes_y.count(yid))
               rval = axes_y[yid]->move_object (subobj, cd_sys, cd_id, xx, yy, mode);
            else
               rval=ERR_OBJ_VAL_DNE;

            if (rval == ERR_OBJ_VAL_DNE)
            {
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",cid->get_axis_id ());
               return ERR_OBJ_VAL_DNE;
            }
         }
      break;
      case chips_colorbar:
         if (subobj == chips_not_subobject)
            rval = this->move_colorbar (&tmp_id, xx, yy, mode);
         else
         {
            string oid = tmp_id.get_colorbar_id ();
            string cd_id = tmp_id.get_coord_id ();
            if (oid.length() == 0)  oid = currpoint;
            if (cbars.empty())
            {
               cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbar");
               return ERR_OBJ_LIST_EMPTY;
            }
            map <string, ChipsColorBar*>::iterator ii;
            if (chips_strcmp(oid.c_str(), "all") == 0)
            {
               int sts;
               for(ii=cbars.begin(); ii != cbars.end(); ++ii)
                  sts |= ii->second->move_object (subobj, cd_sys, cd_id, xx, yy, mode );
               if (sts != ERR_STATUS_GOOD)
                  rval = sts;
            }
            else if (cbars.count(oid))
            {
               rval = cbars[oid]->move_object(subobj, cd_sys, cd_id, xx, yy, mode);
            }
            else
            {
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",oid.c_str());
               return ERR_OBJ_VAL_DNE;
            }
            currcbar = oid;
         }
      break;
      default:
         rval = ERR_CANT_MOVE;
   }
   return rval;
}



int ChipsPlot::set_title_xpos(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
   {
     title_xpos = val;
     update_title_position();
   } 
   else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}


int ChipsPlot::set_title_ypos(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
   {     
     title_ypos = val;
     update_title_position();
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}


int ChipsPlot::set_title_angle(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)   
      title->set_angle(val);
   else
      cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}


float ChipsPlot::get_title_angle()
{
   float fval = 0;
   if (title)   fval = title->get_angle();
   return fval; 
}


int ChipsPlot::set_title_halign(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
   {
      if (val == AUTO_ALIGN)
         rval = title->set_horizalignment(0.5);
      else 
         rval = title->set_horizalignment(val);
   }
   else cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}

float ChipsPlot::get_title_halign()
{
   float fval = 0;
   if (title)  fval = title->get_horizalignment();
   return fval;
}

int ChipsPlot::set_title_valign(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (title)
   {
      if (val == AUTO_ALIGN)
         rval = title->set_vertalignment(0.5);
      else 
         rval = title->set_vertalignment(val);
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"plot title");
   return rval;
}

float ChipsPlot::get_title_valign()
{
   float fval = 0;
   if (title)   fval = title->get_vertalignment();
   return fval;
}

int ChipsPlot::set_xlabel_text(ChipsId* cid, const string& val)
{
   ChipsId tmp_cid(*cid);
   string axis_id = cid->get_axis_id();
   int rval = ERR_STATUS_GOOD;

   // check if the axis component is in the xaxis list
   // the axis component should supersede the xaxis component
   if (axis_id != "")
   {
      tmp_cid.set_axis_id ("");
      tmp_cid.set_xaxis_id (axis_id);
   } 
   else 
      tmp_cid.set_yaxis_id ("");

   tmp_cid.set_axis_dim (X_AXIS);

   // rerun set properties with the temporary id
   rval =  (this->set_property (&tmp_cid, (void*)val.c_str(), SET_AXIS_TEXT));

   if (rval == ERR_STATUS_GOOD)
   {
      // if the axis id was used up the real chips id with the
      // correct values.
      if (tmp_cid.get_xaxis_id () == axis_id)
         cid->set_xaxis_id (axis_id);
   }
   return rval;
}

int ChipsPlot::set_ylabel_text(ChipsId* cid, const string& val)
{
   ChipsId tmp_cid(*cid);
   string axis_id = cid->get_axis_id();
   int rval = ERR_STATUS_GOOD;

   // check if the axis component is in the yaxis list
   // the axis component should supersede the yaxis component
   if (axis_id != "")
   {
      tmp_cid.set_axis_id ("");
      tmp_cid.set_yaxis_id (axis_id);
   }
   else
      tmp_cid.set_xaxis_id ("");

   tmp_cid.set_axis_dim (Y_AXIS);

   // rerun set properties with the temporary id
   rval =  (this->set_property (&tmp_cid, (void*)val.c_str(), SET_AXIS_TEXT));
			        
   if (rval == ERR_STATUS_GOOD)
   {
      // if the axis id was used up the real chips id with the
      // correct values.
      if (tmp_cid.get_yaxis_id () == axis_id)
	 cid->set_yaxis_id (axis_id);
   }
   return rval;
}

int ChipsPlot::add_curve(ChipsId* cid, long cnt, double* x, double* y,
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   int created = 0;
   ChipsAxis *t1 = NULL, *t2 = NULL;
   string cvid = theDefaults->get_curveid();
   ChipsDataSystem *tmp_coord = NULL;
   string tmp_ds_id, transform_id = "";
   double co_min[2];
   double co_max[2];
   double co_lmin[2];

   // if valid axes ids are specified in ChipsId, use them 
   if (cid)
   {
      string xid = cid->get_xaxis_id(),  
             yid = cid->get_yaxis_id(); 
    
      if (xid.length()) 
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else  if ((t1 = get_xaxis(xid.c_str())) == NULL)
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"x axis",xid.c_str());
      }
      if (yid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t2 = get_yaxis(yid.c_str())) == NULL)
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"y axis",yid.c_str());
      }
   }
   if (cvid.length() && (get_curve(cvid) != NULL))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,cvid.c_str(), "curve");
   if (rval == ERR_STATUS_GOOD) 
   {
      bool update_coord = false;
      if (t1){
	 cur_x = t1;
	 update_coord = true;
      }
      if (t2){
	 cur_y = t2; 
	 update_coord = true;
      }

      // create the curve -- this is necessary to extract the base limits incase
      // this is the first object created
      rval = this->setup_current_axes (transform, created);

      if (rval != ERR_STATUS_GOOD)
         return rval;

      if (created != 0)
         tmp_coord = new ChipsDataSystem (NULL);
      else
         tmp_coord = get_datasystem(cur_x, cur_y);

      if (transform && tmp_coord)
         transform_id = tmp_coord->add_transform (transform);

      ChipsCurve* curv = new_curve(cnt, x, y, ydown, yup, xdown, xup, ylims,
                                      xlims, transform_id, cvid);

      cmd_manager->open_block ();
      if (curv)
      {
         if (tmp_coord){
            curv->set_chips_coord (tmp_coord);
            curv->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (curv);
            // Updates the range for the curve
            curv->reset_position ();
         }
         else
         {
            curv->object_delete ();
            stringstream error_message;
            if (cur_x && cur_y)
               error_message << "The current axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a data system. add_curve cannot continue.";
            else if (cur_x)
               error_message << "Could not create create a current y axis. add_curve cannot continue.";
            else
               error_message << "Could not create create a current x axis. add_curve cannot continue.";
            cmd_manager->add_err (error_message.str (), this->id, 0);
            cmd_manager->close_block ();
            return ERR_STATUS_BAD;
         }
         co_min[0] = curv->get_xmin (true);
         co_min[1] = curv->get_ymin (true);
         co_max[0] = curv->get_xmax (true);
         co_max[1] =  curv->get_ymax (true);
         co_lmin[0] = curv->get_xlmin ();
         co_lmin[1] = curv->get_ylmin ();
         add_object_to_depth_table(curv, false);
      }
      else
      {
         cmd_manager->close_block ();
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"curve");
         return rval;
      }

      rval = this->create_axes (transform, created, co_min, co_max);

      // add curve command after create axes
      if (!rval && (!cmd_manager->cmd_lock ()))
      {
         // if ctid is empty that means the curve_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(curv,(cvid.length()==0),currcurv);
         if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            delete cmd;

         ChipsDepthUpdateCommand* depth_cmd =
         new ChipsDepthUpdateCommand(this, curv, 0);
            cmd_manager->add_cmd(depth_cmd);
      }

      if (!rval && (tmp_coord != get_datasystem(cur_x, cur_y)))
      {
         delete tmp_coord;
         curv->set_chips_coord (NULL);
         tmp_coord = get_datasystem(cur_x, cur_y);
         if (!tmp_coord)
         {
            remove_object_from_depth_table(curv);
            curv->object_delete ();
            stringstream error_message;
            error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinate system.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {
            curv->set_chips_coord (tmp_coord);
            curv->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (curv);

            // get the coord sys and update with the transform
            if (transform)
            {
               transform_id = tmp_coord->add_transform (transform);
               curv->set_transform_id (transform_id);
            }
         }
      }

      if (cur_x && cur_y && !rval)
      {
         if (created != 0)  // added new axis so adjust corner joint widths
         {
            adjust_corner_widths(cur_x);
            adjust_corner_widths(cur_y);
            update_coord = true;
         }

         // The current axis change update the coord sys
         if (update_coord && (rval == ERR_STATUS_GOOD))
            rval = update_current_coord (cid);

         // updates the plot area of the curve
	 curv->set_plot_location(bx1, by1, bx2, by2);
	 curves.insert(curves_map::value_type(curv->get_id(), curv));
	 curv->set_curve_num (curve_cnt);
         currcurv = curv->get_id();
         cid->set_curve_id(currcurv);
         cid->set_xaxis_id(cur_x->get_id()); 
         cid->set_yaxis_id(cur_y->get_id()); 
         rearrange_object(cid, chips_curve, chips_front, true);           

         // if this is the first object added set the aspect ratio
         bool first_object = !images_exist() && !contours_exist()
                      && (curves.size () == 1) && !histograms_exist();

         ChipsAxis *v_axis[] = {cur_x, cur_y};
         bool axis_created[] = {(created & 1) ==1, (created & 2) ==2};


         rval = calculate_limits (v_axis, co_min, co_max, co_lmin,
                         axis_created, 2, first_object);

      }
      else if (cur_xall || cur_yall) 
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
      else if (cur_y && !cur_x)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"x axis","curve");
      else if (cur_x && !cur_y)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"y axis","curve");
      cmd_manager->close_block ();
   }
   return rval;
}

int ChipsPlot::delete_curve(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   string cvid;
   ChipsCurve* cv;

   if (cid)    cvid = cid->get_curve_id();
   if (cvid.length() == 0) cvid = currcurv; 

   if (cvid == "all")
   {
      if (curves.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"curves");
         return rval;
      }
      // delete all curves in plot
      if (!force_del) cmd_manager->open_block ();
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
      {
         cv = (*cc).second;
         cv->unselect ();
         if (depth_table)
         {
            if (!force_del)   remove_object_from_depth_table(cv);
            else              depth_table->remove_object(cv);
         }
         if (force_del)
            cv->object_delete ();
         else
	 {
            // hide the curve without changing the visibility
            cv->force_hide ();
            dcurves[cv->get_ref_cnt ()] = cv;
            if (cv->get_chips_coord ())
            {
               cv->get_chips_coord ()->remove (cv);
               cv->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cv,currcurv);
               if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
	       {
                  delete cmd;
                  return rval;
               }
            }
         }
      }
      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         map <unsigned int, ChipsCurve* >::iterator dcrv_iter;
         // If force deleteing all curves also remove previously delete curves 
         for(dcrv_iter=dcurves.begin();dcrv_iter!=dcurves.end(); ++dcrv_iter)
            dcrv_iter->second->object_delete ();
         dcurves.clear ();
      }
      currcurv = "";
      cid->set_curve_id("");
      curves.clear();
   }
   else
   {
      curves_map::iterator cc = curves.find(cvid);
      if (cc == curves.end())  
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"curve",cvid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      cv = (*cc).second;
     
      // found curve so remove it
      curves.erase(cc);
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(cv);
         else              depth_table->remove_object(cv);
      }
      if (force_del)
         cv->object_delete ();
      else
      {
         // hide the curve without changing the visibility
         cv->force_hide ();
         dcurves[cv->get_ref_cnt ()] = cv;
         if (cv->get_chips_coord ())
         {
            cv->get_chips_coord ()->remove (cv);
            cv->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock ()){
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cv, currcurv);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_curve_id("");
      cv->unselect ();
	    
      if (!curves.empty())
      {
         cc = curves.end();
         cc--;
         cv = (*cc).second;
         currcurv = cv->get_id();
      }
      else   currcurv = "";
   }
   cmd_manager->update_mask(STATE_CHNG_ADD_CRV);  
   return ERR_STATUS_GOOD;  
}

ChipsCurve* ChipsPlot::get_curve(const string& val)
{
   ChipsCurve* cv = NULL; 
   string cvid; 

   if (val.length())   cvid = val;
      else    cvid = currcurv; 
   if ((chips_strcmp(cvid.c_str(), "all") == 0) && curves.size() == 1)
   {
      map<string, ChipsCurve*>::iterator ii = curves.begin();
      return ((*ii).second);
   }
   else if (chips_strcmp(cvid.c_str(), "all") != 0)
      if (curves.count(cvid))  cv = curves[cvid];
   return cv; 
}

int ChipsPlot::get_current_curve(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;

   if (cid)
   {
      cid->set_curve_id(currcurv);

      // adjust axes ids to match curve id val 
      if (chips_strcmp(currcurv.c_str(), "all") != 0 || curves.size()==1)
      {
         ChipsCurve* tmp = get_curve(); 

         if (tmp)
         {
/***** NEEDS REVIEW * ARE CURRENT AXES DIFF FROM CURVE AXES ***/

            cid->set_xaxis_id(tmp->get_xaxis ()); 
            cid->set_yaxis_id(tmp->get_yaxis()); 
         } 
      } 
      else 
      {
         // since curve is set to all, clear the axes id vals
         cid->set_xaxis_id("");
         cid->set_yaxis_id("");
      }
   }
   else  
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   return rval;
}

int ChipsPlot::set_current_curve(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string cvid = cid->get_curve_id();

      if (chips_strcmp(cvid.c_str(), "all") != 0)
      {
	 if (!curves.count(cvid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "curve", cvid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currcurv, cvid, chips_curve);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcurv = cvid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currcurv,
                                                      cvid, chips_curve);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcurv = "all";
      }
   }
   return rval;
}

int ChipsPlot::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int count, double *levels, Transform *trans)
{
   int rval = ERR_STATUS_GOOD;
   int created = 0;
   ChipsAxis *t1 = NULL, *t2 = NULL;
   string ctid = theDefaults->get_contourid();

   ChipsDataSystem *tmp_coord = NULL;
   string transform_id = "";
   double co_min[2];
   double co_max[2];
   double co_lmin[2];


   // verify that the contour data is a reg grid
   if (((rval = prescreen_contour(cnt, x, y, z)) == ERR_STATUS_GOOD) && cid)
   {
      // if valid axes ids are specified in ChipsId, use them
      string xid = cid->get_xaxis_id(),
             yid = cid->get_yaxis_id();

      if (xid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t1 = get_xaxis(xid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",xid.c_str());
      }
      if (yid.length())
      {
         if (chips_strcmp(yid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t2 = get_yaxis(yid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",yid.c_str());
      }
   }
   if (ctid.length() && (get_contour(ctid) != NULL))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,ctid.c_str(), "contour");
   if (rval == ERR_STATUS_GOOD)
   {
      bool update_coord = false;
      if (t1){
	 cur_x = t1;
	 update_coord = true;
      }
      if (t2){
	 cur_y = t2;
	 update_coord = true;
      }

      // create the contour -- this is necessary to extract the base limits incase
      // this is the first object created
      rval = this->setup_current_axes (trans, created);
      if (rval != ERR_STATUS_GOOD)
         return rval;

      if (created != 0)
         tmp_coord = new ChipsDataSystem (NULL);
      else
         tmp_coord = get_datasystem(cur_x, cur_y);

      if (trans && tmp_coord)
         transform_id = tmp_coord->add_transform (trans);

      ChipsContour* cntr = new_contour(cnt, x, y, z, count, levels, transform_id, ctid);

      cmd_manager->open_block ();
      if (cntr)
      {
         if (tmp_coord){
            cntr->set_chips_coord (tmp_coord);
            cntr->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (cntr);
            // Updates the range for the curve
            cntr->reset_position ();
         }
         else
         {
            cntr->object_delete ();
            stringstream error_message;
            if (cur_x && cur_y)
               error_message << "The current axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a data system. add_contour cannot continue.";
            else if (cur_x)
               error_message << "Could not create create a current y axis. add_contour cannot continue.";
            else
               error_message << "Could not create create a current x axis. add_contour cannot continue.";
            
            cmd_manager->add_err (error_message.str (), this->id, 0);
            cmd_manager->close_block ();
            return ERR_STATUS_BAD;
         }
         co_min[0] = cntr->get_xmin (true);
         co_min[1] = cntr->get_ymin (true);
         co_max[0] = cntr->get_xmax (true);
         co_max[1] =  cntr->get_ymax (true);
         co_lmin[0] = cntr->get_xlmin ();
         co_lmin[1] = cntr->get_ylmin ();
         add_object_to_depth_table(cntr, false);
      }
      else
      {
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"contour");
         cmd_manager->close_block ();
         return rval;
      }



      rval = this->create_axes (trans, created);

      // add contour create command after create_axes
      if (!rval && (!cmd_manager->cmd_lock ()))
      {
         // if ctid is empty that means the curve_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(cntr,(ctid.length()==0),currcont);
         if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            delete cmd;

         ChipsDepthUpdateCommand* depth_cmd =
         new ChipsDepthUpdateCommand(this, cntr, 0);
            cmd_manager->add_cmd(depth_cmd);
      }

      if (!rval && (tmp_coord != get_datasystem(cur_x, cur_y)))
      {
         delete tmp_coord;
         cntr->set_chips_coord (NULL);
         tmp_coord = get_datasystem(cur_x, cur_y);
         if (!tmp_coord)
         {
            remove_object_from_depth_table(cntr);
            cntr->object_delete ();;
            stringstream error_message;
            error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinate system.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {
            cntr->set_chips_coord (tmp_coord);
            cntr->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (cntr);

            if (trans)
            {
               transform_id = tmp_coord->add_transform (trans);
               cntr->set_transform_id (transform_id);
            }
         }
      }
  
      if (cur_x && cur_y && !rval)
      {
         if (created != 0)  // added new axis so adjust corner joint widths
         {
            adjust_corner_widths(cur_x);
            adjust_corner_widths(cur_y);
            update_coord = true;
         }

	 // The current axis change update the coord sys
	 if (update_coord && (rval == ERR_STATUS_GOOD))
            rval = update_current_coord (cid);

         // updates the plot area of the contour
         cntr->set_plot_location(bx1, by1, bx2, by2);
         contours.insert(contours_map::value_type(cntr->get_id(), cntr));
	 cntr->set_contour_num (contour_cnt);
         currcont = cntr->get_id();
         cid->set_contour_id(currcont);
         cid->set_xaxis_id(cur_x->get_id());
         cid->set_yaxis_id(cur_y->get_id());
         rearrange_object(cid, chips_contour, chips_front, true);

         // if no other objects exist then set the data aspect ratio
         bool first_object = !images_exist() && !curves_exist()
                         && (contours.size () == 1) && !histograms_exist();

         ChipsAxis *v_axis[] = {cur_x, cur_y};
         bool axis_created[] = {(created & 1) ==1, (created & 2) ==2};


         rval = calculate_limits (v_axis, co_min, co_max, co_lmin,
                         axis_created, 2, first_object);
      }
      else if (cur_xall || cur_yall)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
      else if (cur_y && !cur_x)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"x axis","contour");
      else if (cur_x && !cur_y)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"y axis","contour");
      cmd_manager->close_block ();
   }
   return rval;
}

int ChipsPlot::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans)
{
   int cnt = 0, rval = ERR_STATUS_GOOD;
   int created = 0;
   ChipsAxis *t1 = NULL, *t2 = NULL;
   string ctid = theDefaults->get_contourid();
   ChipsDataSystem *tmp_coord = NULL;
   string transform_id = "";
   double *xx = new double[xdim*ydim];
   double *yy = new double[xdim*ydim];
   double co_min[2];
   double co_max[2];
   double co_lmin[2];

   for (int row = 0; row < ydim; row++)
      for (int col = 0; col < xdim; col++)
      {
         xx[cnt] = col + 1;
         yy[cnt] = row + 1;
         cnt++;
      }

   if (((rval=prescreen_contour(cnt, xx, yy, dat)) == ERR_STATUS_GOOD) && cid)
   {
      string xid = cid->get_xaxis_id(),
             yid = cid->get_yaxis_id();
      if (xid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t1 = get_xaxis(xid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",xid.c_str());
      }
      if (yid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t2 = get_yaxis(yid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",yid.c_str());
      }
   }
   if (ctid.length() && contours.count(ctid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,ctid.c_str(), "contour");
   if (rval == ERR_STATUS_GOOD)
   {
      bool update_coord = false;
      if (t1){
	 cur_x = t1;
	 update_coord = true;
      }
      if (t2){
	 cur_y = t2;
	 update_coord = true;
      }

      // create the contour -- this is necessary to extract the base limits incase
      // this is the first object created
      rval = this->setup_current_axes (trans, created);
      if (rval != ERR_STATUS_GOOD)
         return rval;
       
      if (created != 0)
         tmp_coord = new ChipsDataSystem (NULL);
      else
         tmp_coord = get_datasystem(cur_x, cur_y);

      if (trans && tmp_coord)
         transform_id = tmp_coord->add_transform (trans);

      ChipsContour* cntr = new_contour(cnt, xx, yy, dat, numlev,
                                                    levels, transform_id, ctid);

      cmd_manager->open_block ();

      if (cntr)
      {
         if (tmp_coord){
            cntr->set_chips_coord (tmp_coord);
            cntr->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (cntr);
            // Updates the range for the curve
            cntr->reset_position ();
         }
         else
         {
            cntr->object_delete ();
            stringstream error_message;
            if (cur_x && cur_y)
               error_message << "The current axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a data system. add_contour cannot continue.";
            else if (cur_x)
               error_message << "Could not create create a current y axis. add_contour cannot continue.";
            else
               error_message << "Could not create create a current x axis. add_contour cannot continue.";
            cmd_manager->add_err (error_message.str (), this->id, 0);
            cmd_manager->close_block ();
            return ERR_STATUS_BAD;
         }

         co_min[0] = cntr->get_xmin (true);
         co_min[1] = cntr->get_ymin (true);
         co_max[0] = cntr->get_xmax (true);
         co_max[1] =  cntr->get_ymax (true);
         co_lmin[0] = cntr->get_xlmin ();
         co_lmin[1] = cntr->get_ylmin ();
         add_object_to_depth_table(cntr, false);
      }
      else
      {
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"contour");
         cmd_manager->close_block ();
         return rval;
      }

      rval = this->create_axes (trans, created);

      // add contour create command after create_axes
      if (!rval && (!cmd_manager->cmd_lock ()))
      {
         // if ctid is empty that means the curve_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(cntr,(ctid.length()==0),currcont);
         if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            delete cmd;
         ChipsDepthUpdateCommand* depth_cmd =
         new ChipsDepthUpdateCommand(this, cntr, 0);
            cmd_manager->add_cmd(depth_cmd);
      }


      if (!rval && (tmp_coord != get_datasystem(cur_x, cur_y)))
      {
         delete tmp_coord;
         cntr->set_chips_coord (NULL);
         tmp_coord = get_datasystem(cur_x, cur_y);
         if (!tmp_coord)
         {
            remove_object_from_depth_table(cntr);
            cntr->object_delete ();
            stringstream error_message;
            error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinate system.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {
            // get the coord sys and update with the transform
            if (trans)
            {
               transform_id = tmp_coord->add_transform (trans);
               cntr->set_transform_id (transform_id);
            }

            cntr->set_chips_coord (tmp_coord);
            cntr->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (cntr);
         }
      }
  
      if (cur_x && cur_y && !rval)
      {

         if (created != 0)  // added new axis so adjust corner joint widths
         {
            adjust_corner_widths(cur_x);
            adjust_corner_widths(cur_y);
            update_coord = true;
         }

	 // The current axis change update the coord sys
	 if (update_coord && (rval == ERR_STATUS_GOOD))
            rval = update_current_coord (cid);

         // updates the plot area of the contour
         cntr->set_plot_location(bx1, by1, bx2, by2);
         contours.insert(contours_map::value_type(cntr->get_id(), cntr));
	 cntr->set_contour_num (contour_cnt);
         currcont = cntr->get_id();
         cid->set_contour_id(currcont);
         cid->set_xaxis_id(cur_x->get_id());
         cid->set_yaxis_id(cur_y->get_id());
         rearrange_object(cid, chips_contour, chips_front, true);

         // if this is the first object added set the aspect ratio
         bool first_object = !images_exist() && !curves_exist()
                      && (contours.size () == 1) && !histograms_exist();

         ChipsAxis *v_axis[] = {cur_x, cur_y};
         bool axis_created[] = {(created & 1) ==1, (created & 2) ==2};


         rval = calculate_limits (v_axis, co_min, co_max, co_lmin,
                         axis_created, 2, first_object);

         string tmp_ratio = tmp_coord->get_aspect_ratio();
         if (first_object && (tmp_ratio.length () == 0))
         {
            bindings->reset_access_list(this);
            bool cmd_lock = cmd_manager->cmd_lock ();
            cmd_manager->set_cmd_lock (false);
            string ratio = "1:1";
            this->set_property(cid, (void*)ratio.c_str(),
                            SET_DATA_ASPECT_RATIO);

            cmd_manager->set_cmd_lock (cmd_lock);
         }
      }
      else if (cur_xall || cur_yall)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
      else if (cur_y)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"x axis","contour");
      else
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"y axis","contour");
      cmd_manager->close_block ();
   }
   delete []xx;
   delete []yy;
   return rval;
}

int ChipsPlot::delete_contour(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   string ctid;
   ChipsContour* ct;

   if (contours.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"contours");
      return rval;
   }

   if (cid)  ctid = cid->get_contour_id();
   if (ctid.length() == 0)   ctid = currcont;

   if (ctid == "all")
   {
      if (!force_del) cmd_manager->open_block ();
      // delete all contours in plot
      for(contours_map::iterator cc=contours.begin();cc!=contours.end(); ++cc)
      {
         ct = (*cc).second;
         ct->unselect ();
         if (depth_table)
         {
            if (!force_del)   remove_object_from_depth_table(ct);
            else              depth_table->remove_object(ct);
         }
         if (force_del)
            ct->object_delete ();
         else
         {
            // hide the contour without changing the visibility
            ct->force_hide ();
            dcontours[ct->get_ref_cnt ()] = ct;
            if (ct->get_chips_coord ())
            {
               ct->get_chips_coord ()->remove (ct);
               ct->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ct,currcont);
               if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  delete cmd;
                  return rval;
               }
            }
         }
      }
      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         map <unsigned int, ChipsContour* >::iterator dctr_iter;
         // If force deleteing all contours also remove previously delete contours
         for(dctr_iter=dcontours.begin();dctr_iter!=dcontours.end(); ++dctr_iter)
            dctr_iter->second->object_delete ();
         dcontours.clear ();
      }
      currcont = "";
      cid->set_contour_id("");
      rval = ERR_STATUS_GOOD;
      contours.clear();
   }
   else
   {
      contours_map::iterator cc = contours.find(ctid);
      if (cc == contours.end())
      {
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",ctid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ct = (*cc).second;
      // found contour so remove it
      contours.erase(cc);
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(ct);
         else              depth_table->remove_object(ct);
      }
      if (force_del)
         ct->object_delete ();
      else
      {
         // hide the contour without changing the visibility
         ct->force_hide ();
         dcontours[ct->get_ref_cnt ()] = ct;
         if (ct->get_chips_coord ())
         {
            ct->get_chips_coord ()->remove (ct);
            ct->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock ()){
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ct,currcont);
            cmd_manager->add_cmd (cmd);
         }
      }
      cid->set_contour_id("");
      ct->unselect ();

      if (!contours.empty())
      {
         cc = contours.end();
         cc--;
         ct = (*cc).second;
         currcont = ct->get_id();
      }
      else   currcont = "";
   }
   cmd_manager->update_mask(STATE_CHNG_ADD_CNTR);  
   return ERR_STATUS_GOOD;
}

ChipsContour* ChipsPlot::get_contour(const string& val)
{
   ChipsContour* ct = NULL;
   string ctid;

   if (val.length())   ctid = val;
      else    ctid = currcont;
   if ((chips_strcmp(ctid.c_str(), "all") == 0) && contours.size() == 1)
   {
      map<string, ChipsContour*>::iterator ii = contours.begin();
      return ((*ii).second);
   }
   else if (chips_strcmp(ctid.c_str(), "all") != 0)
      if (contours.count(ctid))  ct = contours[ctid];
   return ct;
}

int ChipsPlot::get_current_contour(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;

   if (cid)
   {
      cid->set_contour_id(currcont);

      // adjust axes ids to match contour id val 
      if (chips_strcmp(currcont.c_str(), "all") != 0 || contours.size()==1)
      {
         ChipsContour* tmp = get_contour(); 

         if (tmp)
         {
/***** NEEDS REVIEW * ARE CURRENT AXES DIFF FROM CURVE AXES ***/
            cid->set_xaxis_id(tmp->get_xaxis ()); 
            cid->set_yaxis_id(tmp->get_yaxis()); 
         } 
      } 
      else 
      {
         // since contour is set to all, clear the axes id vals
         cid->set_xaxis_id("");
         cid->set_yaxis_id("");
      }
   }
   else
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   return rval;
}

int ChipsPlot::set_current_contour(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string ctid = cid->get_contour_id();

      if (chips_strcmp(ctid.c_str(), "all") != 0)
      {
         if (!contours.count(ctid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "contour", ctid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currcont, ctid, chips_contour);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcont = ctid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currcont,
               ctid, chips_contour);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcont = "all";
      }
   }
   return rval;
}

int ChipsPlot::add_histogram(ChipsId* cid,  long cnt, double* lo, double* hi,
   double* counts, double* yd, double* yu)
{
   int rval = ERR_STATUS_GOOD;
   int created = 0;
   ChipsAxis *t1 = NULL, *t2 = NULL;
   string hid = theDefaults->get_histogramid();
   ChipsDataSystem *tmp_coord = NULL;

   // if valid axes ids are specified in ChipsId, use them
   if (cid)
   {
      string xid = cid->get_xaxis_id();
      string yid = cid->get_yaxis_id();

      if (xid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t1 = get_xaxis(xid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",xid.c_str());
      }
      if (yid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
         else if ((t2 = get_yaxis(yid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",yid.c_str());
      }
   }
   if (hid.length() && (get_histogram(hid) != NULL))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,hid.c_str(), "histogram");
   if (rval == ERR_STATUS_GOOD)
   {
      bool update_coord = false;
      if (t1){
	 cur_x = t1;
	 update_coord = true;
      }
      if (t2){
  	 cur_y = t2;
	 update_coord = true;
      }

      cmd_manager->open_block ();
      rval = this->create_axes (NULL, created);

      if (!rval)
      {
         tmp_coord = get_datasystem(cur_x, cur_y);
         if (!tmp_coord)
         {
            stringstream error_message;
            if (cur_x && cur_y)
               error_message << "The current axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a data system. add_histogram cannot continue.";
            else if (cur_x)
               error_message << "Could not create create a current y axis. add_histogram cannot continue.";
            else
               error_message << "Could not create create a current x axis. add_histogram cannot continue.";
            cmd_manager->add_err (error_message.str (), this->id, 0);

            rval = ERR_STATUS_BAD;
         }
      }

      if (cur_x && cur_y && !rval)
      {
         if (created != 0)
         {
            adjust_corner_widths(cur_x);
            adjust_corner_widths(cur_y);
            update_coord = true;
         }

	 // The axis have changed update the coord sys
	 if (update_coord && (rval == ERR_STATUS_GOOD))
            rval = update_current_coord (cid);

         tmp_coord = get_datasystem(cur_x, cur_y);

         if (tmp_coord)
         {
            // transform data before it gets into histogram
            Transform *logical_transform = NULL;
            logical_transform = tmp_coord->get_transform (tmp_coord->get_default_transform_id ());
            if (logical_transform)
            {
               // this wont work exactly when in wcs. its here for completeness
               // and for linear transforms. since the yy values may differ
               // slightly from the low bin to the high bin the best bet is
               // an approximation

               // go through each of the data points and adjust
               // apply the transforms inverse

               double *invert_data;
               double *dataset  = new double[cnt*2];

               // populate the yy's first so that they are not repopulated
               // if both hi and lo are set
               for (int ii=0; ii<cnt; ii++)
                  dataset[ii*2+1] = counts[ii];

               // if low is supplied update the low bins
               if (lo)
               {
                  for (int ii=0; ii<cnt; ii++)
                     dataset[ii*2] = lo[ii];

                  invert_data = (double *)logical_transform->invert((void *)dataset, dmDOUBLE, cnt);

                  for (int ii=0; ii<cnt; ii++)
                  {
                     lo[ii] = invert_data[ii*2];
                     counts[ii] = invert_data[ii*2+1];
                  }
                  delete []invert_data;
               }

               // if hi is supplied update the hi bins
               if (hi)
               {
                  for (int ii=0; ii<cnt; ii++)
                     dataset[ii*2] = hi[ii];

                  invert_data = (double *)logical_transform->invert((void *)dataset, dmDOUBLE, cnt);

                  // this will overwrite the yy's if the low bins are also supplied
                  for (int ii=0; ii<cnt; ii++)
                  {
                     hi[ii] = invert_data[ii*2];
                     counts[ii] = invert_data[ii*2+1];
                  }
                  delete []invert_data;
               }
            }
         }
         ChipsHistogram* hst = new_histogram(cnt, lo, hi, counts, yd, yu, hid);
         if (hst)
         {
	    if (!cmd_manager->cmd_lock ()){
	       // if hid is empty that means the histogram_cnt changed
               ChipsCreateCommand *cmd = new ChipsCreateCommand(hst,(hid.length()==0),currhist);
               if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD){
	          delete cmd;
	          return rval;
	       }
            }

            if (tmp_coord){
               hst->set_chips_coord (tmp_coord);
               hst->set_coord_system (DATA, tmp_coord->get_id ());
               tmp_coord->insert (hst);
               // Updates the range for the curve
               hst->reset_position ();
            }

            // updates the plot area of the curve
            hst->set_plot_location(bx1, by1, bx2, by2);
            histograms.insert(histograms_map::value_type(hst->get_id(), hst));
            hst->set_histogram_num (histogram_cnt);
            currhist = hst->get_id();
            cid->set_histogram_id(currhist);
            cid->set_xaxis_id(cur_x->get_id());
            cid->set_yaxis_id(cur_y->get_id());
            add_object_to_depth_table(hst);

            // adjust the axes ranges to correspond to the data if
            // autolimits are on or this histogram created the axes 
            ChipsAxis *v_axis[] = {cur_x, cur_y};
            double co_min[] = {hst->get_xmin (true), hst->get_ymin (true)};
            double co_max[] = {hst->get_xmax (true), hst->get_ymax (true)};
            double co_lmin[] = {hst->get_xlmin (), hst->get_ylmin ()};
            bool axis_created[] = {(created & 1) ==1, (created & 2) ==2};

            rval = calculate_limits (v_axis, co_min, co_max, co_lmin,
                         axis_created, 2, (histograms.size () == 1) && 
                         !contours_exist() && !curves_exist() && 
                         !images_exist ());
         }
         else
         {
            cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"histogram"); 
 
            // if we created axes for this contour, remove them
            if ((created & 0x01) == 0x01)   delete_xaxis(NULL, true);
            if ((created & 0x02) == 0x02)   delete_yaxis(NULL, true);
         }
      }
      else if (cur_xall || cur_yall)
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes"); 
      else if (cur_y)
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"x axis","histogram");
      else
	 cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"y axis","histogram");
      cmd_manager->close_block ();
   }
   return rval;
}

int ChipsPlot::delete_histogram(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   string lid;
   ChipsHistogram* ch;

   if (histograms.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"histograms");
      return rval;
   }

   if (cid)   lid = cid->get_histogram_id();
   if (lid.length() == 0) lid = currhist;

   if (lid == "all")
   {
      // delete all curves in plot
      if (!force_del) cmd_manager->open_block ();
      for(histograms_map::iterator hh = histograms.begin(); hh != histograms.end(); ++hh)
      {
         ch = (*hh).second;
         ch->unselect ();
         if (depth_table)
         {
            if (!force_del)  remove_object_from_depth_table(ch);
            else             depth_table->remove_object(ch);
         }
         if (force_del)
            ch->object_delete ();
         else
         {
            // hide the histogram without changing the visibility
            ch->force_hide();
            dhistograms[ch->get_ref_cnt ()] = ch;
            if (ch->get_chips_coord ())
            {
               ch->get_chips_coord ()->remove (ch);
               ch->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ch,currhist);
               if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  delete cmd;
                  return rval;
               }
            }
         }
      }
      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         map <unsigned int, ChipsHistogram* >::iterator dhst_iter;
         // If force deleteing all histograms also remove previously delete histograms
         for(dhst_iter=dhistograms.begin();dhst_iter!=dhistograms.end(); ++dhst_iter)
            dhst_iter->second->object_delete ();
         dhistograms.clear ();
      }
      currhist = "";
      cid->set_histogram_id("");
      rval = ERR_STATUS_GOOD;
      histograms.clear();
   }
   else
   {
      histograms_map::iterator hh = histograms.find(lid);
      if (hh == histograms.end())
      {
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",lid.c_str());
         return ERR_OBJ_VAL_DNE; 
      }
      ch = (*hh).second;
      histograms.erase(hh);
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(ch);
         else              depth_table->remove_object(ch);
      }
      if (force_del)
         ch->object_delete ();
      else
      {
         // hide the histogram without changing the visibility
         ch->force_hide();
         dhistograms[ch->get_ref_cnt ()] = ch;
         if (ch->get_chips_coord ())
         {
            ch->get_chips_coord ()->remove (ch);
            ch->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock ()){
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ch,currhist);
            cmd_manager->add_cmd (cmd);
         }
      }
      cid->set_histogram_id("");
               rval = ERR_STATUS_GOOD;
      ch->unselect ();

      if (!histograms.empty())
      {
         hh = histograms.end();
         hh--;
         ch = (*hh).second;
         currhist = ch->get_id();
      }
      else   currhist = "";
   }
   cmd_manager->update_mask(STATE_CHNG_ADD_HSTG);  
   return ERR_STATUS_GOOD;
}

ChipsHistogram* ChipsPlot::get_histogram(const string& val)
{
   ChipsHistogram* ch = NULL;
   string chid;

   if (val.length())   chid = val;
      else    chid = currhist;
   if ((chips_strcmp(chid.c_str(), "all") == 0) && histograms.size() == 1)
   {
      map<string, ChipsHistogram*>::iterator ii = histograms.begin();
      return ((*ii).second);
   }
   else if (chips_strcmp(chid.c_str(), "all") != 0)
      if (histograms.count(chid))  ch = histograms[chid];
   return ch;
}

int ChipsPlot::set_current_histogram(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (cid)
   {
      string chid = cid->get_histogram_id();

      if (chips_strcmp(chid.c_str(), "all") != 0)
      {
         if (!histograms.count(chid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "histogram", chid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currhist, chid, chips_histogram);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }

         }
         currhist = chid;
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currhist,
                                                      chid, chips_histogram);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currhist = "all";
      }
   }
   return rval;
}


int ChipsPlot::get_current_histogram(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;

   if (cid)
   {
      cid->set_histogram_id(currhist);

      // adjust axes ids to match histogram id val
      if (chips_strcmp(currhist.c_str(), "all") != 0 || histograms.size() == 1)
      {
         ChipsHistogram* tmp = get_histogram();

         if (tmp)
         {
            cid->set_xaxis_id(tmp->get_xaxis());
            cid->set_yaxis_id(tmp->get_yaxis());
         }
      }
      else
      {
         // since histogram is set to all, clear the axes id vals
         cid->set_xaxis_id("");
         cid->set_yaxis_id("");
      }
   }
   else 
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   return rval;
}

int ChipsPlot::add_image(ChipsId* cid, string filename, Transform *transform)
{
   return ERR_FEATURE_DNE;
}

int ChipsPlot::add_image(ChipsId* cid, string filename,
   float ix1, float iy1, float ix2, float iy2, Transform *transform)
{
   return ERR_FEATURE_DNE;
}

int ChipsPlot::add_image(ChipsId* cid, double* rvals, double *gvals, 
                       double *bvals, double *avals, int xaxis, int yaxis,
                       Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   ChipsImage* tmpimage = NULL;
   ChipsAxis *t1 = NULL, *t2 = NULL;
   string imid = theDefaults->get_imageid();
   int created = 0;
   string tmp_ds_id, transform_id = "";
   ChipsDataSystem *tmp_coord = NULL;
   double co_min[2];
   double co_max[2];
   double co_lmin[2];

   if (imid.length() && images.count(imid))
   {
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,imid.c_str(), "image");
      return rval; 
   }

   if (cid)
   {
      string xid = cid->get_xaxis_id();
      string yid = cid->get_yaxis_id();
      if (xid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes");
         else if ((t1 = get_xaxis(xid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",xid.c_str());
      }
      if (yid.length())
      {
         if (chips_strcmp(xid.c_str(), "all") == 0)
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes");
         else if ((t2 = get_yaxis(yid.c_str())) == NULL)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",yid.c_str());
      }
   }

   if (rval == ERR_STATUS_GOOD)
   {
      bool update_coord = false;
      if (t1)
      {
         cur_x = t1;
         update_coord = true;
      }
      if (t2)
      {
         cur_y = t2;
         update_coord = true;
      }

      // create the image -- this is necessary to extract the base limits incase
      // this is the first object created
      rval = this->setup_current_axes (transform, created);
      if (rval != ERR_STATUS_GOOD)
         return rval;

      if (created != 0)
         tmp_coord = new ChipsDataSystem (NULL);
      else
         tmp_coord = get_datasystem(cur_x, cur_y);

      if (transform && tmp_coord)
         transform_id = tmp_coord->add_transform (transform);

      tmpimage = new_image(rvals, gvals, bvals, avals, xaxis, yaxis, transform_id, imid);
      if (tmpimage != NULL)
      {
         if (tmp_coord)
         {
            tmpimage->set_chips_coord (tmp_coord);
            tmpimage->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (tmpimage);
            // Updates the range for the image
            tmpimage->reset_position ();

         }
         else
         {
            stringstream error_message;
            if (cur_x && cur_y)
               error_message << "The current axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a data system. add_image cannot continue.";
            else if (cur_x)
               error_message << "Could not create create a current y axis. add_image cannot continue.";
            else
               error_message << "Could not create create a current x axis. add_image cannot continue.";
            cmd_manager->add_err (error_message.str (), this->id, 0);
            cmd_manager->close_block ();

            delete tmpimage;
            return ERR_STATUS_BAD;
         }
         co_min[0] = tmpimage->get_xmin (true);
         co_min[1] = tmpimage->get_ymin (true);
         co_max[0] = tmpimage->get_xmax (true);
         co_max[1] =  tmpimage->get_ymax (true);
         co_lmin[0] = tmpimage->get_xlmin ();
         co_lmin[1] = tmpimage->get_ylmin ();
         add_object_to_depth_table(tmpimage, false);
      }
      else
      {
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"image");
         cmd_manager->close_block ();
         return rval;
      }


      cmd_manager->open_block ();
      rval = this->create_axes (transform, created, co_min, co_max);

      // add image create command after create axes
      if (!rval && (!cmd_manager->cmd_lock ()))
      {
         // if imid is empty that means the contour_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpimage,(imid.length()==0));
         if ((rval = cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            delete cmd;
            return rval;
         }
         ChipsDepthUpdateCommand* depth_cmd =
         new ChipsDepthUpdateCommand(this, tmpimage, 0);
            cmd_manager->add_cmd(depth_cmd);
      }

      if (!rval && (tmp_coord != get_datasystem(cur_x, cur_y)))
      {
         // delete the temporary coordinate system and get the real one
         delete tmp_coord;
         tmpimage->set_chips_coord (NULL);
         tmp_coord = get_datasystem(cur_x, cur_y);
         if (!tmp_coord)
         {
            remove_object_from_depth_table(tmpimage);
            delete tmpimage;
            tmpimage = NULL;
            stringstream error_message;
            error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinate system.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {
            // get the coord sys and update with the transform
            if (transform)
            {
               transform_id = tmp_coord->add_transform (transform);
               tmpimage->set_transform_id (transform_id);
            }

            tmpimage->set_chips_coord (tmp_coord);
            tmpimage->set_coord_system (DATA, tmp_coord->get_id ());
            tmp_coord->insert (tmpimage);
         }

      }

      if (cur_x && cur_y && !rval)
      {
         if (created != 0)  // added new axis so adjust corner joint widths
         {
            adjust_corner_widths(cur_x);
            adjust_corner_widths(cur_y);
            update_coord = true;
         }

         // The current axis change update the coord sys
         if (update_coord && (rval == ERR_STATUS_GOOD))
            rval = update_current_coord (cid);

         // updates the plot area of the contour
         tmpimage->set_plot_location(bx1, by1, bx2, by2);

         tmpimage->set_image_num (image_cnt);
	 currimage = tmpimage->get_id();
         cid->set_image_id(currimage);
         cid->set_xaxis_id(cur_x->get_id());
         cid->set_yaxis_id(cur_y->get_id());
	 images.insert(pair<string, ChipsImage*>(currimage, tmpimage));

         // the image is created before the axes push it to the front
         rearrange_object (cid, chips_image, chips_front, true);

         // if no other objects exist then set the data aspect ratio
         bool first_object = !contours_exist() && !curves_exist()
                      && (images.size () == 1) && !histograms_exist();

         ChipsAxis *v_axis[] = {cur_x, cur_y};
         bool axis_created[] = {(created & 1) ==1, (created & 2) ==2};

         rval = calculate_limits (v_axis, co_min, co_max, co_lmin, 
                         axis_created, 2, first_object);

         string tmp_ratio = tmp_coord->get_aspect_ratio();
         if (first_object && (tmp_ratio.length () == 0))
         {
            bindings->reset_access_list(this);
            bool cmd_lock = cmd_manager->cmd_lock ();
            cmd_manager->set_cmd_lock (false);
            string ratio = "1:1";
            this->set_property(cid, (void*)ratio.c_str(),
                         SET_DATA_ASPECT_RATIO);
             cmd_manager->set_cmd_lock (cmd_lock);
         }
      }
      else if (cur_xall || cur_yall)
      {
         if (tmpimage) delete tmpimage;
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axes");
      }
      else if (cur_y)
      {
         if (tmpimage) delete tmpimage;
         cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"x axis","image");
      }
      else
      { 
         if (tmpimage) delete tmpimage;
         cmd_manager->add_err(rval=ERR_NO_CURRENT_AXIS,id,0,"y axis","image");
      }

      cmd_manager->close_block ();
   }

   return rval;
}

int ChipsPlot::delete_image(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsImage* cimg;
   string iid = cid->get_image_id();

   if (iid.length() == 0)   iid = currimage;

   if (chips_strcmp(iid.c_str(), "all") == 0)
   {
      map<string, ChipsImage* >::iterator ii;

      if (images.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"images");
         return rval;
      }

      // delete all images in frame
      if (!force_del)
         cmd_manager->open_block ();

      for(ii=images.begin(); ii != images.end(); ++ii)
      {
         cimg = (*ii).second;
         cimg->unselect ();
         if (depth_table)
         {
	    if (!force_del)   remove_object_from_depth_table(cimg);
            else              depth_table->remove_object(cimg);
         }
	 if (force_del)
            cimg->object_delete ();
	 else
         {
            // this hides the image without changing its visibility
            cimg->force_hide ();
            dimages [cimg->get_ref_cnt ()] = cimg;
	    if (cimg->get_chips_coord ())
            {
               cimg->get_chips_coord ()->remove(cimg);
               cimg->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cimg, currimage);
               if ((rval=cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }

      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         // If force deleteing all images also remove previously deleted images
         map <unsigned int, ChipsImage*>::iterator  dimage_ii;
         for (dimage_ii = dimages.begin();dimage_ii!=dimages.end();dimage_ii++)
            dimage_ii->second->object_delete();
         dimages.clear ();
      }

      images.clear();
      currimage = "";
      cid->set_image_id("");
      rval = ERR_STATUS_GOOD;
      image_cnt = 0;
   }
   else if (images.count(iid))
   {
      map<string, ChipsImage* >::iterator ii;
      ii = images.find(iid);
      cimg = (*ii).second;
      cimg->unselect ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(cimg);
         else              depth_table->remove_object(cimg);
      }
      images.erase(ii);
      if (force_del)
         cimg->object_delete();
      else
      {
         // this hides the image without changing its visibility
         cimg->force_hide ();
         dimages [cimg->get_ref_cnt ()] = cimg;
         if (cimg->get_chips_coord ())
         {
            cimg->get_chips_coord()->remove(cimg);
            cimg->set_chips_coord(NULL);
         }
         coord->remove(cimg);
         if (!cmd_manager->cmd_lock())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cimg, currimage);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_image_id(""); 
      if (!images.empty())
      {
         map<string, ChipsImage* >::reverse_iterator ri = images.rbegin();
         currimage = (*ri).first;
      }
      else   currimage = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "image", iid.c_str());
   if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_ADD_IMG);  
   return rval;
}

int ChipsPlot::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{
   int rval = ERR_STATUS_GOOD;
   string imid;
   ChipsImage* im;

   if (cid)   imid = cid->get_image_id();
   if (imid.length() == 0) imid = currimage;

   if (!images_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"images");
   else if (imid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(images_map::iterator cc = images.begin(); cc != images.end(); ++cc)
         if ((st = (*cc).second->add_channel(vals, xaxis, yaxis, channel)) != ERR_STATUS_GOOD)
            rval = st;
   }
   else  if ((im = get_image(imid)) != NULL)
      rval = im->add_channel(vals, xaxis, yaxis, channel);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",imid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      currimage = imid;
      cid->set_image_id(currimage);
   }
   return rval;
}

int ChipsPlot::remove_image_channel(ChipsId* cid, int channel)
{
   int rval = ERR_STATUS_GOOD;
   string imid;
   ChipsImage* im;

   if (cid)   imid = cid->get_image_id();
   if (imid.length() == 0) imid = currimage;

   if (!images_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"images");
   else if (imid == "all")
   {
      int st;
      rval = ERR_STATUS_GOOD;
      for(images_map::iterator cc = images.begin(); cc != images.end(); ++cc)
         if ((st = (*cc).second->remove_channel(channel)) != ERR_STATUS_GOOD)
            rval = st;
   }
   else  if ((im = get_image(imid)) != NULL)
      rval = im->remove_channel(channel);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",imid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      currimage = imid;
      cid->set_image_id(currimage);
   }
   return rval;
}

int ChipsPlot::print_image(ChipsId* cid, const string &filename, int fmt)
{
   int rval = ERR_STATUS_GOOD;
   string imid;
   ChipsImage* im;

   if (cid)   imid = cid->get_image_id();
   if (imid.length() == 0) imid = currimage;

   if (!images_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"images");
   else if (imid == "all")
      cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"images");
   else  if ((im = get_image(imid)) != NULL)
      rval = im->print(filename, fmt);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",imid.c_str());

   if (rval == ERR_STATUS_GOOD)
   {
      currimage = imid;
      cid->set_image_id(currimage);
   }
   return rval;
}

int ChipsPlot::add_axis (ChipsId* cid, int dim, double pos,double min,double max,
                         Transform *transform, bool setup_axis)
{
   int rval;
   switch (dim)
   {
      case XY_AXIS:
         rval = this->add_axis (cid, dim, pos, min, max, min, max, transform, setup_axis);
      break;
      case X_AXIS:
         rval = this->add_axis (cid, dim, pos, min, max, 0, 0, transform, setup_axis);
      break;
      case Y_AXIS:
         rval = this->add_axis (cid, dim, pos, 0, 0, min, max, transform, setup_axis);
      break;
   }

   return rval;
}

int ChipsPlot::add_axis (ChipsId* cid, int dim, double pos,double minx,double maxx,
                         double miny,double maxy, Transform *transform, 
                         bool setup_axis)
{
   ChipsAxis *ax = NULL;
   ChipsAxis *ay = NULL;
   int rval = ERR_STATUS_GOOD;

   string  xaid = theDefaults->get_axisxid();
   string  yaid = theDefaults->get_axisyid();

   ChipsAxisAlgorithmFactory::TransformType axis_type = ChipsAxisAlgorithmFactory::get_algorithm_type (transform);

   if (axis_type == ChipsAxisAlgorithmFactory::UNKNOWN)
   {
      cmd_manager->add_err("Unknown transform type. Transforms should either be LINEAR2D or WCSTAN", ERR_STATUS_BAD,id,0);
      return ERR_STATUS_BAD;
   }


   if (!((dim == X_AXIS) || (dim == Y_AXIS) || (dim == XY_AXIS) ||
       (dim == RA_AXIS) || (dim == DEC_AXIS) || (dim == RADEC_AXIS)))
   {
      cmd_manager->add_err(ERR_INT_SUBSET,id,0,"dimension",dim,
                           "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
      return ERR_INT_SUBSET;
   }

   ChipsCoordinate *tmp_coord = NULL;
   ChipsId tmpId (*cid);

   // coord id should only be used for data coordinates
   if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");
   if (tmpId.get_coord_sys () == COORD_NONE) tmpId.set_coord_sys (PLOT_NORM);

   // This updates the chips id for the appropriate coord sys
   tmp_coord = this->find_coordinate_system (tmpId, rval);

   // current axes only support plot norm and data coordinates
   if ((tmpId.get_coord_sys () != PLOT_NORM) && (tmpId.get_coord_sys () != DATA))
   {
      rval = ERR_STATUS_BAD;
      cmd_manager->add_err ("Axes can only be added in PLOT_NORM or DATA coordinates", ERR_STATUS_BAD, "", 0);
   }
   else if (!tmp_coord)
   {
      rval = ERR_COORD_DNE;
      cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(tmpId.get_coord_sys ()));
   }
   else if ((tmpId.get_coord_sys () == DATA) && (!cur_x || !cur_y))
   {
      rval = ERR_COORD_DNE;
      cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(tmpId.get_coord_sys ()));
   }
     

   if (rval != ERR_STATUS_GOOD)
      return rval;

   // Find the position of the axes
   ChipsVector lower_pos (0.0, 0.0);
   ChipsVector upper_pos (1.0, 1.0);

   // Convert length of the axis from plot normalized to whatever
   // coordinate system the axis is in
   if (coord)
   {
      coord->convert_cs_to_cs (&lower_pos, &lower_pos, 1, PLOT_NORM, "",
                                tmpId.get_coord_sys (), tmpId.get_coord_id ());
      coord->convert_cs_to_cs (&upper_pos, &upper_pos, 1, PLOT_NORM, "",
                                tmpId.get_coord_sys (), tmpId.get_coord_id ());
   }


   switch (dim)
   {
      case XY_AXIS:
      {
         if (xaid.length() && (get_xaxis(xaid)))
            cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,xaid.c_str(), "axis");
         else if ((yaid.length() && get_yaxis(yaid)))
            cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,yaid.c_str(), "axis");
         else
         {

            // Create the axes
            if (setup_axis)
            {
               ax = new_axis(tmp_coord,tmpId.get_coord_sys (),lower_pos[0],pos,
                                   upper_pos[1],pos,X_AXIS,minx,maxx,axis_type,xaid);
               ay = new_axis(tmp_coord,tmpId.get_coord_sys (),pos,lower_pos[1],pos,
                                   upper_pos[1],Y_AXIS,miny,maxy,axis_type,yaid);
            }
            else
            {
               ax = new_axis(tmp_coord,tmpId.get_coord_sys (),lower_pos[0],pos,
                                   upper_pos[1],pos,X_AXIS,axis_type,xaid);
               ay = new_axis(tmp_coord,tmpId.get_coord_sys (),pos,lower_pos[1],pos,
                                   upper_pos[1],Y_AXIS,axis_type,yaid);
            }
      

            // Create a coord system for each existing axis
            add_coordinate_system (ax, ay, transform);

            // for RA/DEC only allow on pair per coordinate system 
            // for XY_AXIS create a pairing for each existing axis
            if (dim == XY_AXIS)
            {
               for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
                  if ((am->second != left) && (am->second != right) && (am->second != ay))
                     add_coordinate_system (ax, am->second, transform);
               for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
                  if ((am->second != top) && (am->second != bottom) && (am->second != ax))
                     add_coordinate_system (am->second, ay, transform);
            }
         }
      }
      break;
      case X_AXIS:
      {
         if (xaid.length() && get_xaxis(xaid))
            cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,xaid.c_str(), "axis");

         else
         {

            int success = ERR_STATUS_BAD;

            // Create the axes
            if (setup_axis)
               ax = new_axis(tmp_coord,tmpId.get_coord_sys (),lower_pos[0],pos,
                                   upper_pos[1],pos,dim,minx,maxx,axis_type,xaid);
            else
               ax = new_axis(tmp_coord,tmpId.get_coord_sys (),lower_pos[0],pos,
                                   upper_pos[1],pos,dim,axis_type,xaid);

            for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
               if ((am->second != left) && (am->second != right))
                  // check to see if at least one coordinate system was made
                  success &= add_coordinate_system (ax, am->second, transform);

            // if no coordinate systems could be created, create one with
            // a single axis
            if (success != ERR_STATUS_GOOD)
               add_coordinate_system (ax, NULL, transform);
         }
      }
      break;
      case Y_AXIS:
      {
         if (yaid.length() && get_yaxis(yaid))
            cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,yaid.c_str(), "axis");

         else
         {

            int success = ERR_STATUS_BAD;

            // Create the axes 
            if (setup_axis)
               ay = new_axis(tmp_coord,tmpId.get_coord_sys (),pos,lower_pos[0],
                                   pos,upper_pos[1],dim,miny,maxy,axis_type,yaid);
            else
               ay = new_axis(tmp_coord,tmpId.get_coord_sys (),pos,lower_pos[0],
                                   pos,upper_pos[1],dim,axis_type,yaid);

            for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
               if ((am->second != top) && (am->second != bottom))
                  // check to see if at least one coordinate system was made
                  success &= add_coordinate_system (am->second, ay, transform);

            // if no coordinate systems could be created, create one with
            // a single axis
            if (success != ERR_STATUS_GOOD)
               add_coordinate_system (NULL, ay, transform);
         }
      }
   }

   // if val is empty that means the axis_y_cnt changed
   cmd_manager->open_block ();
   if (ax)
   {
      if (setup_axis)
         rval =  handle_new_xaxis(ax, (xaid.length()==0), cid);
      else
         cur_x = ax;
      add_object_to_depth_table(ax);
   }
   if (ay)
   {
      if (setup_axis)
         rval =  handle_new_yaxis(ay, (yaid.length()==0), cid);
      else
         cur_y = ay;
      add_object_to_depth_table(ay);
   }

   list<string> exclude_axes;
   update_data_coord_positions (ax, ay, exclude_axes);
   rval = update_current_coord(cid);
   cmd_manager->close_block ();

   return rval;
}

int ChipsPlot::handle_new_xaxis(ChipsAxis* ax, bool change_cnt, ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (ax)
   { 
      ax->set_grid_position(by1, by2);
      ax->set_plot_location(bx1, by1, bx2, by2);
      ax->update_tick_positions ();

      // Add a command if the command manager is not locked
      if (!cmd_manager->cmd_lock ()){
         ChipsCreateCommand *cmd;
         if (cur_xall)
            cmd = new ChipsCreateCommand(ax, change_cnt,"all");
         else if (cur_x)
            cmd = new ChipsCreateCommand(ax, change_cnt, cur_x->get_id ());
         else
            cmd = new ChipsCreateCommand(ax, change_cnt);
         rval = cmd_manager->add_cmd (cmd);
         if (rval != ERR_STATUS_GOOD){
            delete cmd;
            return rval;
         }
      }

      if (bindings)
      {
         if (axes_x.size() == 2)
         {
            if (!cmd_manager->cmd_lock())
            {
               ChipsBindCommand *bind_cmd;
               double* dp = NULL;
               int*    ip = NULL;
               string* sp = NULL;
   
               bindings->compile_synch_data(top,&dp, &ip, &sp);
               // This isn't the best approach another approach
               // should be looked into
               bind_cmd = new ChipsBindCommand(get_parent(), id, ax->get_id(),
                                     id, top->get_id(), dp, ip, sp, true);
               if ((rval = cmd_manager->add_cmd(bind_cmd)) != ERR_STATUS_GOOD)
                  delete bind_cmd;
               if (dp) delete []dp;
               if (ip) delete []ip;
               if (sp) delete []sp;
            }
            bindings->add_binding(ax, top);
            bindings->synch_axes(ax, top, false);
         }
         else bindings->add_axis(ax);   // update axis bind grpah  
      }
     
      if (top && bottom && (ax != top) && (ax != bottom))
         ax->set_label_text(theDefaults->get_axisxlabel());

      axes_x.insert (axes_map::value_type (ax->get_id(),ax));
      ax->set_axis_num (axes_x.size ());
      cur_x = ax;
      if (left && right)
         ax->set_corners(corner_style, left->get_thickness(), 
                         right->get_thickness());
      else
         ax->set_corners(corner_style, 0, 0);

      cur_xall = false; 
      if (cid)  cid->set_xaxis_id(ax->get_id()); 

      // the new axis may cover boxed/closed axes 
      if (axes_x.size () > 2)
         if (axis_style != chips_open)   cover_box_axes(ax);

   }
   else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"axis");
   return rval; 
}

int ChipsPlot::handle_new_yaxis(ChipsAxis* ay, bool change_cnt, ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (ay)
   {
      ay->set_grid_position(bx1, bx2);
      ay->set_plot_location(bx1, by1, bx2, by2);
      ay->update_tick_positions ();

      // add a command if the command manager is not locked
      if (!cmd_manager->cmd_lock ()){
         ChipsCreateCommand *cmd;
         if (cur_yall)
            cmd = new ChipsCreateCommand(ay, change_cnt, "all");
         else if (cur_y)
            cmd = new ChipsCreateCommand(ay, change_cnt, cur_y->get_id ());
         else
            cmd = new ChipsCreateCommand(ay, change_cnt);
         rval = cmd_manager->add_cmd (cmd);
         if (rval != ERR_STATUS_GOOD){
            delete cmd;
            return rval;
         }
      }

      if (bindings)
      {
         if (axes_y.size() == 2)
         {
            if (!cmd_manager->cmd_lock ())
            {
               ChipsBindCommand *bind_cmd;
               double* dp = NULL;
               int*    ip = NULL;
               string* sp = NULL;

               bindings->compile_synch_data(right, &dp, &ip, &sp);
               // This isn't the best approach another approach
               // should be looked into
   
               bind_cmd = new ChipsBindCommand(get_parent(), id, ay->get_id(), id,
                                              right->get_id(), dp, ip, sp, true);
               if ((rval = cmd_manager->add_cmd(bind_cmd)) != ERR_STATUS_GOOD)
                  delete bind_cmd;
               if (dp) delete []dp;
               if (ip) delete []ip;
               if (sp) delete []sp;
            }
            bindings->add_binding(ay, right);
            bindings->synch_axes(ay, right, false);
         }
         else bindings->add_axis(ay); 
      }
      if (left && right && (ay != left) && (ay != right))
          ay->set_label_text(theDefaults->get_axisylabel());

      axes_y.insert(axes_map::value_type(ay->get_id(), ay));
      ay->set_axis_num (axes_y.size ());
      cur_y = ay;
      if (top && bottom)
         ay->set_corners(corner_style, bottom->get_thickness(), 
                         top->get_thickness());
      else
         ay->set_corners(corner_style, 0, 0); 
      cur_yall = false; 
      if (cid)  cid->set_yaxis_id(ay->get_id()); 

      // the new axis may cover boxed/closed axes 
      if (axes_y.size () >2)
         if (axis_style != chips_open)   cover_box_axes(ay);

   }
   else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"axis"); 
   return rval; 
}

ChipsAxis* ChipsPlot::get_xaxis(const string& val)
{
   ChipsAxis* ax = NULL;
   if (val.length())   
   {
      if ((chips_strcmp(val.c_str(), "all") != 0) && (axes_x.count(val)))
	 ax = axes_x[val];
   }
   else if (cur_x)
      ax = cur_x;   
   else if (cur_xall && (axes_x.size () == 3))
      ax = axes_x.begin()->second;
   return ax;
}

ChipsAxis* ChipsPlot::get_yaxis(const string& val)
{
   ChipsAxis* ax = NULL;
   if (val.length())   
   {
      if ((chips_strcmp(val.c_str(), "all") != 0) && (axes_y.count(val)))
	 ax = axes_y[val];
   }
   else if (cur_y)
      ax = cur_y;   
   else if (cur_yall && (axes_y.size () == 3))
      ax = axes_y.begin()->second;
   return ax;
}

int ChipsPlot::set_current_xaxis(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD; 
   bool update_coord = false;

   if (cid)
   {
      // precedence-
      //   1) specified x_axis id
      //   2) specified axis id
      string aid = cid->get_xaxis_id();
      if (aid.length() == 0)  aid = cid->get_axis_id(); 

      if (chips_strcmp(aid.c_str(), "all") != 0)
      {
         if (axes_x.count(aid))
	 {
	    ChipsAxis *ax = axes_x[aid];
            if ((ax != top) && (ax != bottom))
            {
               if (rec)
               {
                  string tmpid;
                  if (cur_x)   tmpid = cur_x->get_id(); 
                     else if (cur_xall)  tmpid = "all"; 
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
                     tmpid, aid, chips_axis);
                  if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
               }
               cur_x = ax;
               cur_xall = false; 
               update_coord = true;
               rval = ERR_STATUS_GOOD; 
            }
            else 
               cmd_manager->add_err(rval=ERR_PLOT_BORDER_CURR, id, 0, aid.c_str());
         }
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "axis", aid.c_str());
      } 
      else if (axes_x.size () > 3)
      {
         if (rec)
         {
            string tmpid;
            if (cur_x)   tmpid = cur_x->get_id(); 
            else if (cur_xall)  tmpid = "all"; 
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, tmpid, aid,
                                                      chips_axis);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         cur_xall = true; 
         cur_x = NULL; 
	 update_coord = true;
      }
   }
   if (update_coord && (rval == ERR_STATUS_GOOD))
      rval = update_current_coord (cid);
   return rval; 
}

int ChipsPlot::set_current_yaxis(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD; 
   bool update_coord = false;

   if (cid)
   {
      // precedence-
      //   1) specified y_axis id
      //   2) specified axis id
      string aid = cid->get_yaxis_id();
      if (aid.length() == 0)  aid = cid->get_axis_id(); 

      if (chips_strcmp(aid.c_str(), "all") != 0)
      {
	 if (axes_y.count(aid))
         {
	    ChipsAxis *ax = axes_y[aid];
            if ((ax != left) && (ax != right))
            {
               if (rec)
               {
                  string tmpid;
                  if (cur_y)   tmpid = cur_y->get_id(); 
                     else if (cur_yall)  tmpid = "all"; 
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
                     tmpid, aid, chips_axis);
                  if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
               }
               cur_y = axes_y[aid];
               cur_yall = false; 
               update_coord = true;
               rval = ERR_STATUS_GOOD; 
            }
            else
               cmd_manager->add_err(rval=ERR_PLOT_BORDER_CURR, id, 0, aid.c_str());
         }
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "axis", aid.c_str());
      }
      else if (axes_y.size() > 3)
      {
         if (rec)
         {
            string tmpid;
            if (cur_y)   tmpid = cur_y->get_id(); 
            else if (cur_yall)  tmpid = "all"; 
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, tmpid, aid,
                                                      chips_axis);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         cur_yall = true; 
         cur_y = NULL; 
	 update_coord = true;
      }
   }
   if (update_coord && (rval == ERR_STATUS_GOOD))
      rval = update_current_coord (cid);
   return rval;
}

int ChipsPlot::set_current_axis(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (cid)
   {
      string aid = cid->get_axis_id();
      string xid = cid->get_xaxis_id();
      string yid = cid->get_yaxis_id();

      // precedence- 
      //   1) single specified x_axis id or y_axis id 
      //   2) a specified axis id  
      if ((aid.length()==0) && (xid.length() != 0) && (yid.length() == 0))
         rval = set_current_xaxis(cid, rec);
      else if ((aid.length()==0) && (xid.length() == 0) && (yid.length() != 0))
         rval = set_current_yaxis(cid, rec);
      else if (aid.length())
      {
         int r1, r2; 
         r1 = set_current_xaxis(cid, rec);

         // remove the error of the object not existing either it will exist
         // in the y axis or not. no point to two of the same errors
         if (r1 == ERR_OBJ_VAL_DNE)
	    cmd_manager->remove_last_err ();

         r2 = set_current_yaxis(cid, rec);
	 if ((r1 == ERR_STATUS_GOOD) || (r2 == ERR_STATUS_GOOD)) 
            rval = ERR_STATUS_GOOD; 
         else if (r1 != ERR_OBJ_VAL_DNE)
         {
            rval = r1;
            // remove the error of the object not existing. the axis
            // existed in the xaxis list
            if (r2 == ERR_OBJ_VAL_DNE)
               cmd_manager->remove_last_err ();
         }
         else
            rval = r2;
      } 
   }
   return rval;
}

int ChipsPlot::delete_axis(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;

   if (axes_x.empty() && axes_y.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"axes");
      return rval;
   }

   //precedence 
   //   1) single specified x_axis or y_axis id
   //   2) a specified axis id
   string axid = cid->get_xaxis_id();
   string ayid = cid->get_yaxis_id();
   string aid = cid->get_axis_id();

   if (axid.length() && (ayid.length() == 0))
   {
      rval = delete_xaxis(cid, force_del);
   }
   else if (ayid.length() && (axid.length() == 0))
   {
      rval = delete_yaxis(cid, force_del);
   }
   else if (aid.length() && (aid != "all"))
   {
      if (get_xaxis(aid) != NULL)
         rval = delete_xaxis(cid, force_del);
      else if (get_yaxis(aid) != NULL)
         rval = delete_yaxis(cid, force_del);
   }
   else
   {
      cmd_manager->open_block ();
      int sts1, sts2; 
      sts1 = delete_xaxis(cid, force_del);
      sts2 = delete_yaxis(cid, force_del);

      if ((sts1 == ERR_STATUS_GOOD) || (sts2 == ERR_STATUS_GOOD))
      {
         rval = ERR_STATUS_GOOD;
         cmd_manager->clear_errs();
      }
      else if (sts1 != ERR_STATUS_GOOD)
         rval = sts1;
      else
         rval = sts2;
      cmd_manager->close_block ();
   }
   return rval;
}

int ChipsPlot::delete_xaxis(ChipsId* cid, bool force_del)
{
   string aid = "";
   int rval = ERR_STATUS_GOOD;
   ChipsAxis* ax;
   bool tmp_cur_xall = cur_xall;
   ChipsAxis *tmp_cur_x = cur_x;

   // if an xaxis id is specified in cid, use it, otherwise use axis id from
   // cid. If neither is set, use the current x axis  
   if (cid) 
   {
      aid = cid->get_xaxis_id();
      if (aid.length() == 0)   aid = cid->get_axis_id();
   }
   if (aid.length()) 
   {
      if (!chips_strcmp(aid.c_str(), "all")) 
      {
         tmp_cur_xall = true;
	 tmp_cur_x = NULL;
      }
      else if (get_xaxis(aid))
      {
         tmp_cur_x = get_xaxis(aid);
	 tmp_cur_xall = false;
      }
      else 
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,aid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
   }

   if (!force_del) cmd_manager->open_block ();

   if (tmp_cur_xall)   // delete all axes 
   {
      if (curves_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"curve"); 
      else if (contours_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"contour"); 
      else if (histograms_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"histogram"); 
      else if (axes_x.size() == 2)  // empty except for borders 
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"x axis");
      else
      {
         bool cont = true;
         // Delete x and y then add false as  an argument to the function
         for(axes_map::iterator am=axes_x.begin(); cont && am!=axes_x.end() && !rval; ++am)
            if (check_for_data_sys_objs ((*am).second))
            {
               cont = false;
               cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"object"); 
            }
      }
      if (rval == ERR_STATUS_GOOD)
      {
         for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
         {
            ax = (*am).second;
            if ((ax != top) && (ax != bottom))
            {
               if (depth_table)
               {
                  if (!force_del)   remove_object_from_depth_table(ax);
                  else              depth_table->remove_object(ax);
               }
               if (force_del)
               {
                  // delete the coords from the xaxis first
                  remove_data_coord (ax, NULL);

                  if (bindings)  bindings->remove_axis(ax);
                  ax->object_delete ();
               }
               else
               {
                 rval = this->dereference_axis (ax);
               }
            }
         }
         axes_x.clear ();
         if (top)
            axes_x[top->get_id ()] = top;
         if (bottom)
            axes_x[bottom->get_id ()] = bottom;

         // If force deleteing all axes also remove previously delete axes
         if (force_del)
         {
            map <unsigned int, ChipsAxis* >::iterator dax_iter;
            for(dax_iter=daxes_x.begin();dax_iter!=daxes_x.end(); ++dax_iter)
               dax_iter->second->object_delete ();

             // delete all deleted data systems then too
             map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
             for(ds_iter=deleted_data_systems.begin();ds_iter!=deleted_data_systems.end(); ++ds_iter)
               delete ds_iter->second;
             deleted_data_systems.clear ();

         }

         cid->set_xaxis_id("");
         cur_x = NULL;
         cur_xall = false; 
         cid->set_coord_sys (PLOT_NORM);
         cid->set_coord_id ("");
         rval = ERR_STATUS_GOOD;
      }
   }
   else if (tmp_cur_x)
   {
      if (check_for_data_sys_objs(tmp_cur_x))
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"object"); 
      else if ((tmp_cur_x == top) || (tmp_cur_x == bottom)) 
         cmd_manager->add_err(rval=ERR_BOX_AXIS_DEL,id,0);
      else if (!tmp_cur_x)   
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis");
      else if (axes_x.count(tmp_cur_x->get_id()) == 0)
      {
         string tmpid = tmp_cur_x->get_id(); 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmpid.c_str());
      }
      else
      {
         ax = axes_x[tmp_cur_x->get_id()];
         ax->unselect ();
         if (depth_table)
         {
            if (!force_del)  remove_object_from_depth_table(ax);
            else             depth_table->remove_object(ax);
         }
         if (force_del)
         {
            if (bindings)  bindings->remove_axis(ax);
            // delete the coords from the xaxis first
            remove_data_coord (ax, NULL);

            ax->object_delete ();
         }
         else  rval = this->dereference_axis (ax);
         axes_map::iterator tam = axes_x.find(tmp_cur_x->get_id()); 
         axes_x.erase(tam);
         rval = ERR_STATUS_GOOD;
         if (axes_x.size () > 2)
         {
            cur_x = NULL;
            for (tam = axes_x.begin(); tam != axes_x.end(); tam++)
               if ((*tam).second && (*tam).second!=top && (*tam).second!=bottom)
                  cur_x = (*tam).second;
            rval = update_current_coord(cid);
         }
         else
         {
            cur_x = NULL;
            cid->set_coord_sys (PLOT_NORM);
            cid->set_coord_id ("");
         }
         cur_xall = false;
      }
   }
   else   cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"x axis");

   // if axis deletion was successful- reset axis style to draw in any
   // boxed/closed axes which may have been obscured
   if ((rval == ERR_STATUS_GOOD) && (axis_style != chips_open))
      set_axis_style(axis_style); 

   if (!force_del) cmd_manager->close_block ();

   return rval; 
}

int ChipsPlot::delete_yaxis(ChipsId* cid, bool force_del)
{
   string aid = "";
   int rval = ERR_STATUS_GOOD;
   ChipsAxis* ax;
   bool tmp_cur_yall = cur_yall;
   ChipsAxis *tmp_cur_y = cur_y;

   // if an yaxis id is specified in cid, use it, otherwise use axis id from
   // cid. If neither is set, use the current y axis  
   if (cid) 
   {
      aid = cid->get_yaxis_id();
      if (aid.length() == 0)   aid = cid->get_axis_id();
   }
   if (aid.length()) 
   {
      if (!chips_strcmp(aid.c_str(), "all")) 
      {
         tmp_cur_yall = true;
	 tmp_cur_y = NULL;
      }
      else if (get_yaxis(aid))
      {
         tmp_cur_y = get_yaxis(aid);
	 tmp_cur_yall = false;
      }
      else   cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,aid.c_str());
   }

   if (!force_del) cmd_manager->open_block ();

   if (tmp_cur_yall)   // delete all axes 
   {
      if (curves_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"curve");
      else if (contours_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"contour");
      else if (histograms_exist())
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"histogram");
      else if (axes_y.size() == 2)    // empty except for borders
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"y axis");
      else
      {
         bool cont = true;
         // Delete x and y then add false as  an argument to the function
         for(axes_map::iterator am=axes_y.begin(); cont && am!=axes_y.end() && !rval; ++am)
            if (check_for_data_sys_objs ((*am).second))
            {
               cont = false;
               cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"object"); 
            }
      }
      if (rval == ERR_STATUS_GOOD)
      {
         for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
         {
            ax = (*am).second;
            ax->unselect ();
            if ((ax != left) && (ax != right))
            {
               if (depth_table)
               {
                  if (!force_del)   remove_object_from_depth_table(ax);
                  else              depth_table->remove_object(ax);
               }
               if (force_del)
               {
                  // delete the coords from the xaxis first
                  remove_data_coord (NULL, ax);
                  if (bindings)  bindings->remove_axis(ax);
                  ax->object_delete ();
               }
               else
                  rval = this->dereference_axis (ax);
            }
         }

         axes_y.clear ();
         if (left)
            axes_y[left->get_id ()] = left;
         if (right)
            axes_y[right->get_id ()] = right;
         
         // If force deleteing all axes also remove previously delete axes
         if (force_del)
         {
            map <unsigned int, ChipsAxis* >::iterator dax_iter;
            for(dax_iter=daxes_y.begin();dax_iter!=daxes_y.end(); ++dax_iter)
               dax_iter->second->object_delete ();

             // delete all deleted data systems then too
             map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
             for(ds_iter=deleted_data_systems.begin();ds_iter!=deleted_data_systems.end(); ++ds_iter)
               delete ds_iter->second;
             deleted_data_systems.clear ();
         }

         cid->set_yaxis_id("");
         cur_y = NULL;
         cur_yall = false; 
         cid->set_coord_sys (PLOT_NORM);
         cid->set_coord_id ("");
         rval = ERR_STATUS_GOOD;
      }
   }
   else if (tmp_cur_y)
   {
      if (check_for_data_sys_objs (tmp_cur_y))
         cmd_manager->add_err(rval=ERR_AXIS_OBJ_EXISTS,id,0,"object"); 
      else if ((tmp_cur_y == left) || (tmp_cur_y == right)) 
         cmd_manager->add_err(rval=ERR_BOX_AXIS_DEL,id,0);
      else if (!tmp_cur_y)   
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis");
      else if (axes_y.count(tmp_cur_y->get_id()) == 0)
      {
         string tmpid = tmp_cur_y->get_id(); 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"axis",tmpid.c_str());
      }
      else
      {
         ax = axes_y[tmp_cur_y->get_id()];
         if (depth_table)
         {
            if (!force_del)   remove_object_from_depth_table(ax);
            else              depth_table->remove_object(ax);
         }
         if (force_del)
         {
            remove_data_coord (NULL, ax);
            if (bindings)  bindings->remove_axis(ax);
            ax->object_delete ();
         }
         else
            rval = this->dereference_axis (ax);

         axes_map::iterator tam = axes_y.find(tmp_cur_y->get_id()); 
         axes_y.erase (tam);
         rval = ERR_STATUS_GOOD;
         if (axes_y.size() > 2)
         {
            cur_y = NULL;
            for (tam = axes_y.begin(); tam != axes_y.end(); tam++)
               if ((*tam).second && (*tam).second!=left && (*tam).second!=right)
                  cur_y = (*tam).second;
            rval = update_current_coord(cid);
         }
         else
         {
            cur_y = NULL;
            cid->set_coord_sys (PLOT_NORM);
            cid->set_coord_id ("");
         }
         cur_yall = false;
      }
   }
   else    cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"y axis");

   // if axis deletion was successful- reset axis style to draw in any
   // boxed/closed axes which may have been obscured
   if ((rval == ERR_STATUS_GOOD) && (axis_style != chips_open))
      set_axis_style(axis_style); 

   if (!force_del) cmd_manager->close_block ();
   return rval; 
}

int ChipsPlot::move_axis(ChipsId* cid, double xx, double yy, int mode)
{
   int sts, rval = ERR_STATUS_GOOD;
   string tmp_id = cid->get_axis_id(); 
   ChipsAxis* ax = NULL;
   bool found = false; 
   string curr_xid = cur_xall ? "all" : get_current_xaxis_id ();
   string curr_yid = cur_yall ? "all" : get_current_yaxis_id ();
   bool update_coord = false;

   // Make sure the coordinate system exists
   COORD_SYS coord_sys = cid->get_coord_sys();
   string coord_id = cid->get_coord_id();
   
   if (!coord)
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(ERR_COORD_DS_DNE,id,0,v_coord_sys_to_string(coord_sys), coord_id.c_str ());
      else
         cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));
      return ERR_COORD_DNE;
   }
   else if (!coord->coord_exists (coord_sys, coord_id))
   {
      if (coord_sys == DATA)
         cmd_manager->add_err(ERR_COORD_DS_DNE,id,0,v_coord_sys_to_string(coord_sys), coord_id.c_str ());
      else
         cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));
      return ERR_COORD_DNE;
   }

   if (!axes_x_exist() && !axes_y_exist())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"axes");
   else if (tmp_id.length())   // adjust specified single axis or 'all'
   {
      if (chips_strcmp(tmp_id.c_str(), "all") == 0)
      {
         for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
	 {
            sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
	    if (sts != ERR_STATUS_GOOD)  rval = sts; 
         }
         if (axes_x.size () > 3)
         {
            // only set current to 'all' if more than 1 x axis
            cur_xall = true;
            cur_x = NULL;
            cid->set_xaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_xid)   update_coord = true;
         }
         for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
	 {
            sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
	    if (sts != ERR_STATUS_GOOD)  rval = sts; 
         }
         if (axes_y.size() > 3)
         {
            // only set current to 'all' if more than 1 y axis
            cur_yall = true;
            cur_y = NULL;
            cid->set_yaxis_id("all");
	    // If the current axis has changed update coords
            if (tmp_id != curr_yid)   update_coord = true;
         }
      }
      else        // adjust single specified axis 
      {
	 if (axes_x.count(tmp_id) == 1)
	 {
            ax = axes_x[tmp_id];
            rval = ax->move(coord_sys, coord_id, xx, yy, mode);
            found = true;
	    if ((ax != top) && (ax != bottom))
	    {
               cur_x = ax;
               cur_xall = false;
               if (tmp_id != curr_xid)   update_coord = true;	 
	    }
            cid->set_xaxis_id(tmp_id);
         }
	 if (axes_y.count(tmp_id) == 1)
	 {
            ax = axes_y[tmp_id];
            rval = ax->move(coord_sys, coord_id, xx, yy, mode);
            found = true;
	    if ((ax != left) && (ax != right))
	    {
               cur_y = ax;
               cur_yall = false; 
               if (tmp_id != curr_yid)   update_coord = true;
	    }
            cid->set_yaxis_id(tmp_id);
         }
         if (!found)
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,tmp_id.c_str());
      }
   }
   else 
   {
      string tmp_id2 = cid->get_yaxis_id();
      tmp_id = cid->get_xaxis_id();

      if ((tmp_id.length()) && (tmp_id2.length()))   // adjust both axes
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
	    for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
            {
               sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
               if (sts != ERR_STATUS_GOOD)  rval = sts; 
            }
            found |= true;

            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
               if (tmp_id != curr_xid)   update_coord = true;
            }
         }
         else
         {
            if ((ax = get_xaxis(tmp_id.c_str())) != NULL)
            {
               rval = ax->move(coord_sys, coord_id, xx, yy, mode);
               found |= true;
	       if ((ax != bottom) && (ax != top))
	       {
                  cur_x = ax;
                  cur_xall = false;
		  // If the current axis has changed update coords
                  if (tmp_id != curr_xid)  update_coord = true;
	       }
               cid->set_xaxis_id(tmp_id);
            }
         }
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
            {
               sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
	       if (sts != ERR_STATUS_GOOD)  rval = sts; 
            }
            found |= true;
            if (axes_y.size() > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
               if (tmp_id2 != curr_yid)   update_coord = true;
            }
         }
         else
         {
            if ((ax = get_yaxis(tmp_id2.c_str())) != NULL)
            {
               rval = ax->move(coord_sys, coord_id, xx, yy, mode);
               found |= true;
	       if ((ax != left) && (ax != right))
	       {
                  cur_y = ax;
                  cur_yall = false;
                  if (tmp_id2 != curr_yid)   update_coord = true;
	       }
               cid->set_yaxis_id(tmp_id2);
            }
         }
         if (!found) 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,tmp_id2.c_str());
      }
      else if (tmp_id.length())   // adjust x axis
      {
         if (chips_strcmp(tmp_id.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
            {
               sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
	       if (sts != ERR_STATUS_GOOD)  rval = sts; 
            }
            if (axes_x.size () > 3)
            {
               cur_x = NULL;
               cur_xall = true;
               cid->set_xaxis_id("all");
               if (tmp_id != curr_xid)   update_coord = true;
            }
         }
         else
         {
            ax = get_xaxis(tmp_id.c_str());
            if (ax)
            {
               rval = ax->move(coord_sys, coord_id, xx, yy, mode);
	       if ((ax != bottom) && (ax != top))
	       {
                  cur_x = ax;
                  cur_xall = false;
                  if (tmp_id != curr_xid)   update_coord = true;
	       }
               cid->set_xaxis_id(tmp_id);
            }
            else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,tmp_id.c_str());
         }
      }
      else if (tmp_id2.length())     // adjust y axis
      {
         if (chips_strcmp(tmp_id2.c_str(), "all") == 0)
         {
            for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
            {
               sts = (*am).second->move(coord_sys, coord_id, xx, yy, mode);
	       if (sts != ERR_STATUS_GOOD)  rval = sts; 
            }
            if (axes_y.size() > 3)
            {
               cur_y = NULL;
               cur_yall = true;
               cid->set_yaxis_id("all");
               if (tmp_id2 != curr_yid)  update_coord = true;
            }
         }
         else
         {
            if ((ax = get_yaxis(tmp_id2.c_str())) != NULL)
            {
               rval = ax->move(coord_sys, coord_id, xx, yy, mode);
	       if ((ax != left) && (ax != right))
	       {
                  cur_y = ax;
                  cur_yall = false;
                  if (tmp_id2 != curr_yid)   update_coord = true;
	       }
               cid->set_yaxis_id(tmp_id2);
            }
            else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,tmp_id2.c_str());
         }
      }
      else
      {
         // adjust x and y axes
         if (cur_x)
            rval = cur_x->move(coord_sys, coord_id, xx, yy, mode);
         else if (cur_xall)
         {
            for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
               rval |= (*am).second->move(coord_sys, coord_id, xx, yy, mode);
         }
         if (cur_y)
            rval = cur_y->move(coord_sys, coord_id, xx, yy, mode);
         else if (cur_yall)
         {
            for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
               rval |= (*am).second->move(coord_sys, coord_id, xx, yy, mode);
         }
      }
   }
   if (rval == ERR_STATUS_GOOD)
   { 
      // reset axis style to redraw any boxed/closed axes 
      if (axis_style != chips_open)   set_axis_style(axis_style);
   }
   return rval;
}

int ChipsPlot::get_current_xaxis(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      if (cur_xall)   cid->set_xaxis_id("all"); 
         else if (cur_x)   cid->set_xaxis_id(cur_x->get_id());
            else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis"); 
   }
   else 
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   return rval; 
}

int ChipsPlot::get_current_yaxis(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      if (cur_yall)   cid->set_yaxis_id("all"); 
         else if (cur_y)   cid->set_yaxis_id(cur_x->get_id());
            else  cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"axis");
   }
   else 
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
   return rval; 
}

int ChipsPlot::limits(ChipsId* cid, int dim, double xmin, double xmax, 
   double ymin, double ymax)
{
   int rval = ERR_STATUS_GOOD;
   bool transformed = true;

   //unlike other limits calls  skip ratio check to allow for pick_limits  
   if ((dim & X_AXIS) && (axes_x.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
   else if ((dim & Y_AXIS) && (axes_y.size() < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
/*   else if (xmin > xmax)
   {
      // adjust to ensure xmin is less than xmax
      tmp = xmax;
      xmax = xmin;
      xmin = tmp;
   }
   else if (ymin > ymax)
   {
      // adjust to ensure ymin is less than ymax
      tmp = ymax;
      ymax = ymin;
      ymin = tmp;
   }
*/
   else if (xmin == xmax)
   {
      // adjust to ensure range is not 1 point
      if (xmin != 0)
      {
         xmin *= 0.95;
         xmax *= 1.05;
      }
      else
      {
         xmin = -1.0;
         xmax = 1.0;
      }
   }
   else if (ymin == ymax)
   {
      // adjust to ensure range is not 1 point
      if (ymin != 0)
      {
         ymin *= 0.95;
         ymax *= 1.05;
      }
      else
      {
         ymin = -1.0;
         ymax = 1.0;
      }
   }

   if (cid && (rval == ERR_STATUS_GOOD))
   {
      ChipsAxis* tmpAxisPtr;
      list<ChipsAxis*>::iterator jj;
      string tmpId = "";
      string altId = "";

      string ds_id = cid->get_coord_id ();
      if (ds_id.length () > 0)
      {
         ChipsDataSystem *ds = dynamic_cast <ChipsDataSystem *> (this->coord->get_coord (DATA,cid->get_coord_id ()));
         if (ds)
            transformed = ds->get_transform (ds_id);
      }

      if ((dim & X_AXIS) == X_AXIS)
      {
         tmpId = cid->get_xaxis_id();
         if (dim == XY_AXIS)
            altId = cid->get_yaxis_id();
      }
      else
         tmpId = cid->get_xaxis_id();

      if (chips_strcmp(tmpId.c_str(), "all") == 0)
      {
         for(axes_map::iterator am=axes_x.begin(); !rval && am!=axes_x.end(); ++am)
            rval = am->second->validate_limits (xmin, xmax, ymin, ymax, XY_AXIS, transformed);

         for(axes_map::iterator am=axes_x.begin(); !rval && am!=axes_x.end(); ++am)
            if (am->second->get_scalable())
            {
               // explicit limits commands turn off axis automin/max
               rval = bindings->update_all_nodes(XY_AXIS,xmin,xmax,transformed,AUTOLIM_OFF,1,true,&ymin,&ymax);
            }
      }
      else if ((dim & X_AXIS) == X_AXIS)  // handle X_AXIS or XY_AXIS
      {
         if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
         {
            rval = tmpAxisPtr->validate_limits (xmin, xmax, ymin, ymax, XY_AXIS,transformed);
            if (!rval)
            {
               if (dim == X_AXIS)
               {
                  // only setting limits of X axis, don't set altId 
                  altId = "";
               }
               else
               {
                  // setting both X and Y axis, get id for Y axis 
                  ChipsAxis *altax = get_yaxis (altId);
                  if (altax)   altId = altax->get_id ();
               }
               rval = bindings->update_connected_nodes(tmpAxisPtr, xmin, xmax,
                     transformed, AUTOLIM_OFF, 0, 1, &ymin, &ymax, altId); 
               cmd_manager->update_mask(STATE_CHNG_XLIMS); 
            }
         }
         else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
      }
      else if ((dim & Y_AXIS) == Y_AXIS)   // handle Y_AXIS  
      {
         if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
         {
            if ((rval = tmpAxisPtr->validate_limits (xmin, xmax, ymin, ymax, XY_AXIS, transformed)) == ERR_STATUS_GOOD)
            {
               rval = bindings->update_connected_nodes(tmpAxisPtr, ymin, ymax,
                     transformed, AUTOLIM_OFF, 0, 1, &xmin, &xmax, "");
               cmd_manager->update_mask(STATE_CHNG_YLIMS); 
            }
         }
         else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
      }
      else 
         cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                              "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
   }
   return rval; 
}

int ChipsPlot::limits(ChipsId* cid, int dim, double min, double max)
{
   int rval = ERR_STATUS_GOOD;
   bool transformed = true;

   if ((dim & X_AXIS) && (axes_x.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
   else if ((dim & Y_AXIS) && (axes_y.size() < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
   else if (dim == XY_AXIS && check_axes_for_ratios(cid))
      cmd_manager->add_err(rval=ERR_XY_LIM_RATIO,id,0); 
/*   else if (min > max) 
   {
      // adjust to ensure min is less than max
      double tmp = max;
      max = min; 
      min = tmp;
   }
*/
   else if (min == max)
   {
      // adjust to ensure range is not 1 point
      if (min != 0)
      {
         min *= 0.95; 
         max *= 1.05; 
      }
      else 
      {
         min = -1.0;
         max = 1.0; 
      }
   }

   if (cid)
   {
      string ds_id = cid->get_coord_id ();
      if (ds_id.length () > 0)
      {
         ChipsDataSystem *ds = dynamic_cast <ChipsDataSystem *> (this->coord->get_coord (DATA,cid->get_coord_id ()));
         if (ds)
            transformed = ds->get_transform (ds_id);
      }
   }

   if (cid && (rval == ERR_STATUS_GOOD))
   {
      string tmpId;
      ChipsAxis* tmpAxisPtr;
      list<ChipsAxis*>::iterator jj;
      cmd_manager->open_block();
      switch(dim)
      {
         case X_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               for(axes_map::iterator am=axes_y.begin(); !rval && am!=axes_y.end(); ++am)
                  rval = am->second->validate_limits (min, max, 0, 0, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_all_nodes(dim, min, max, transformed);
                  cmd_manager->update_mask(STATE_CHNG_XLIMS); 
               }
            }
	    else 
            {
               if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
               {
                  rval = tmpAxisPtr->validate_limits (min, max, 0, 0, dim, transformed);
                  if (!rval)
                  {
                     rval = bindings->update_connected_nodes(tmpAxisPtr, min, max, transformed);
                     cmd_manager->update_mask(STATE_CHNG_XLIMS);
                  }
               }
	       else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
	    }
	 };
         break;
         case Y_AXIS:
	 {
            tmpId=cid->get_yaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall))
            {
               for(axes_map::iterator am=axes_y.begin(); !rval && am!=axes_y.end(); ++am)
                  rval = am->second->validate_limits (0, 0, min, max, dim, transformed);
               rval = bindings->update_all_nodes(dim, min, max, transformed);
               cmd_manager->update_mask(STATE_CHNG_YLIMS);
            }
            else 
            { 
               if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
               {
                  rval = tmpAxisPtr->validate_limits (0, 0, min, max, dim, transformed);
                  if (!rval)
                  {
                     rval = bindings->update_connected_nodes(tmpAxisPtr, min, max, transformed); 
                     cmd_manager->update_mask(STATE_CHNG_YLIMS);
                  }
               }
	       else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
            }
         }
         break;
         case XY_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               for(axes_map::iterator am=axes_x.begin(); !rval && am!=axes_x.end(); ++am)
                  rval = am->second->validate_limits (min, max, min, max, dim, transformed);

               if (!rval)
               {
                  rval = bindings->update_all_nodes(X_AXIS, min, max, transformed);
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
            }
            else 
            {
               if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
               {
                  rval = tmpAxisPtr->validate_limits (min, max, min, max, dim, transformed);
                  if (!rval)
                  {
                     list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
                     for(jj=tl.begin(); jj != tl.end(); ++jj)
                        rval=bindings->adjust_axis_minmax((*jj), min,max, transformed, AUTOLIM_OFF,1);
                     cmd_manager->update_mask(STATE_CHNG_XLIMS);
                  }
               }
	       else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
            }
            tmpId=cid->get_yaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall))
            {
               for(axes_map::iterator am=axes_y.begin(); !rval && am!=axes_y.end(); ++am)
                  rval = am->second->validate_limits (min, max, min, max, dim, transformed);

               if (!rval)
               {
                  rval = bindings->update_all_nodes(Y_AXIS, min, max, transformed);
                  cmd_manager->update_mask(STATE_CHNG_YLIMS);
               }
            }
            else 
            {       
               if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
	       {
                  rval = tmpAxisPtr->validate_limits (min, max, min, max, dim, transformed);
                  if (!rval)
                  {
                     list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
                     for(jj=tl.begin(); jj != tl.end(); ++jj)
                        rval=bindings->adjust_axis_minmax((*jj), min,max, transformed, AUTOLIM_OFF,1);
                     cmd_manager->update_mask(STATE_CHNG_YLIMS);
                  }
               }
	       else cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
            }

	 }
         break;
         default:
            cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                              "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
         break; 
      }
      cmd_manager->close_block();
   }
   return rval; 
}


int ChipsPlot::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE,
   V_AUTOLIMIT_TYPE)
{
   int rval = ERR_NO_ID_OBJ;
   bool transformed=true;
   
   if ((dim & X_AXIS) && (axes_x.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
   else if ((dim & Y_AXIS) && (axes_y.size() < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
   else if (dim == XY_AXIS && check_axes_for_ratios(cid))
         cmd_manager->add_err(rval=ERR_XY_LIM_RATIO,id,0);
   else if (cid)
   {

      string ds_id = cid->get_coord_id ();
      if (ds_id.length () > 0)
      {
         ChipsDataSystem *ds = dynamic_cast <ChipsDataSystem *> (this->coord->get_coord (DATA,cid->get_coord_id ()));
         if (ds)
            transformed = ds->get_transform (ds_id);
      }

      list<ChipsAxis*>::iterator ii;
      string tmpId;
      ChipsAxis* tmpAxisPtr;

      cmd_manager->open_block();

      switch(dim)
      {
         case X_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               // WLM add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;

               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (!rval)
                  rval = tmpAxisPtr->validate_limits (bmn, bmx, 0, 0, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, bmx,
                                                       transformed, AUTOLIM_ON); 
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
            }
	 };
         break;
         case Y_AXIS:
	 {
            tmpId=cid->get_yaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall))
            {
               // WLM add later 
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, true);
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (0, 0, bmn, bmx, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, bmx,
                     transformed, AUTOLIM_ON); 
                  cmd_manager->update_mask(STATE_CHNG_YLIMS); 
               }
            }
         }
         break;
         case XY_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);
               if (!rval)
               {
                  rval = tmpAxisPtr->validate_limits (bmn, bmx, 0, 0, X_AXIS, transformed);
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
               if (!rval)
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, bmx, 
                     transformed,AUTOLIM_ON); 
            }
            tmpId=cid->get_yaxis_id();
            if (((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall)) && !rval)
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if (((tmpAxisPtr=get_yaxis(tmpId.c_str()))) && !rval)
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (0, 0, bmn, bmx, Y_AXIS, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, bmx,
                     transformed, AUTOLIM_ON); 
                  cmd_manager->update_mask(STATE_CHNG_YLIMS);
               }
            }
	 }
         break;
         default:
            cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                                 "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
         break; 
      }
      cmd_manager->close_block();
   }
   return rval; 
}

int ChipsPlot::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE)
{
   int rval = ERR_NO_ID_OBJ;
   bool transformed = true;
   if ((dim & X_AXIS) && (axes_x.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
   else if ((dim & Y_AXIS) && (axes_y.size() < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
   else if (dim == XY_AXIS && check_axes_for_ratios(cid))
      cmd_manager->add_err(rval=ERR_XY_LIM_RATIO,id,0);
   else if (cid)
   {
      list<ChipsAxis*>::iterator ii;
      string tmpId;
      ChipsAxis* tmpAxisPtr;

      string ds_id = cid->get_coord_id ();
      if (ds_id.length () > 0)
      {
         ChipsDataSystem *ds = dynamic_cast <ChipsDataSystem *> (this->coord->get_coord (DATA,cid->get_coord_id ()));
         if (ds)
            transformed = ds->get_transform (ds_id);
      }

      cmd_manager->open_block();
      switch(dim)
      {
         case X_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               // WLM add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (min == bmx)
               {
                  // adjust to ensure range is not 1 point
                  if (min != 0)
                  {
                     min *= 0.95;
                     bmx *= 1.05;
                  }
                  else
                  {
                     min = -1.0;
                     bmx = 1.0;
                  }
               }
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (min, bmx, 0, 0, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, min, bmx,
                     transformed, AUTOMAX_ON); 
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
            }
	 };
         break;
         case Y_AXIS:
	 {
            tmpId=cid->get_yaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall))
            {
              // WLM add later 
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

	       if (min == bmx)
               {
                  // adjust to ensure range is not 1 point
                  if (min != 0)
                  {
                     min *= 0.95;
                     bmx *= 1.05;
                  }
                  else
                  {
                     min = -1.0;
                     bmx = 1.0;
                  }
               }
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (0, 0, min, bmx, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, min, bmx, 
                     transformed, AUTOMAX_ON); 
                  cmd_manager->update_mask(STATE_CHNG_YLIMS);
               }
            }
         }
         break;
         case XY_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (min == bmx)
               {
                  // adjust to ensure range is not 1 point
                  if (min != 0)
                  {
                     min *= 0.95;
                     bmx *= 1.05;
                  }
                  else
                  {
                     min = -1.0;
                     bmx = 1.0;
                  }
               }
               for(ii=tl.begin(); !rval && ii != tl.end() && !rval; ++ii)
               {
                  rval = tmpAxisPtr->validate_limits (min, bmx, 0, 0, X_AXIS, transformed);
                  if (!rval)
                  {
                     if ((rval=bindings->adjust_axis_minmax((*ii), min,bmx, transformed, AUTOMAX_ON,1)) == ERR_STATUS_GOOD)
                     {
                        // reset log min value
                        if (bmx < min)
                        {
                           if (bmx > 0)   blm = bmx;
                           else           blm = 0;
                        }
                        else if (min > 0)
                        {
                           if (min > 0)   blm = min;
                           else           blm = 0;
                        }
                        (*ii)->adjust_logmin(blm);
                        cmd_manager->update_mask(STATE_CHNG_XLIMS);
                     }
                  }
               }
            }
            tmpId=cid->get_yaxis_id();
            if (((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall)) && !rval)
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if (((tmpAxisPtr=get_yaxis(tmpId.c_str()))) && !rval)
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (min == bmx)
               {
                  // adjust to ensure range is not 1 point
                  if (min != 0)
                  {
                     min *= 0.95;
                     bmx *= 1.05;
                  }
                  else
                  {
                     min = -1.0;
                     bmx = 1.0;
                  }
               }
               for(ii=tl.begin(); ii != tl.end() && !rval; ++ii)
               {
                  rval = (*ii)->validate_limits (0, 0, min, bmx, Y_AXIS, transformed);
                  if (!rval)
                  {
                     if ((rval=bindings->adjust_axis_minmax((*ii), min,bmx, transformed, AUTOMAX_ON,1)) == ERR_STATUS_GOOD)
                     {
                        // reset log min value
                        if (bmx < min)
                        {
                           if (bmx > 0)   blm = bmx;
                           else           blm = 0;
                        }
                        else if (min > 0)
                        {
                           if (min > 0)   blm = min;
                           else           blm = 0;
                        }
                        (*ii)->adjust_logmin(blm);
                        cmd_manager->update_mask(STATE_CHNG_YLIMS);
                     }
                  }
               }
            }
	 }
         break;
         default:
            cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                                 "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
         break; 
      }
      cmd_manager->close_block();
   }
   return rval; 
}

int ChipsPlot::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE,
   double max)
{
   int rval = ERR_NO_ID_OBJ;
   bool transformed = true;

   if ((dim & X_AXIS) && (axes_x.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"x");
   else if ((dim & Y_AXIS) && (axes_y.size () < 3))
      cmd_manager->add_err(rval=ERR_LIMITS_NO_AXES,id,0,"y");
   else if (dim == XY_AXIS && check_axes_for_ratios(cid))
      cmd_manager->add_err(rval=ERR_XY_LIM_RATIO,id,0);
   else if (cid)
   {
      list<ChipsAxis*>::iterator ii;
      string tmpId;
      ChipsAxis* tmpAxisPtr;

      string ds_id = cid->get_coord_id ();
      if (ds_id.length () > 0)
      {
         ChipsDataSystem *ds = dynamic_cast <ChipsDataSystem *> (this->coord->get_coord (DATA,cid->get_coord_id ()));
         if (ds)
            transformed = ds->get_transform (ds_id);
      }

      cmd_manager->open_block();

      switch(dim)
      {
         case X_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall)) 
            {
               // WLM add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (bmn == max)
               {
                  // adjust to ensure range is not 1 point
                  if (bmn != 0)
                  {
                     bmn *= 0.95;
                     max *= 1.05;
                  }
                  else
                  {
                     bmn = -1.0;
                     max = 1.0;
                  }
               }
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (bmn, bmx, 0, 0, X_AXIS, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, max, 
                     transformed, AUTOMIN_ON); 
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
            }
	 };
         break;
         case Y_AXIS:
	 {
            tmpId=cid->get_yaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall)) 
            {
              // WLM add later 
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (bmn == max)
               {
                  // adjust to ensure range is not 1 point
                  if (bmn != 0)
                  {
                     bmn *= 0.95;
                     max *= 1.05;
                  }
                  else
                  {
                     bmn = -1.0;
                     max = 1.0;
                  }
               }
               if (!rval)
                  rval = tmpAxisPtr->validate_limits (0, 0, bmn, bmx, dim, transformed);
               if (!rval)
               {
                  rval = bindings->update_connected_nodes(tmpAxisPtr, bmn, max,
                     transformed, AUTOMIN_ON); 
                  cmd_manager->update_mask(STATE_CHNG_XLIMS);
               }
            }
         }
         break;
         case XY_AXIS:
         {
            tmpId=cid->get_xaxis_id();
            if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_xall))
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
            {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (bmn == max)
               {
                  // adjust to ensure range is not 1 point
                  if (bmn != 0)
                  {
                     bmn *= 0.95;
                     max *= 1.05;
                  }
                  else
                  {
                     bmn = -1.0;
                     max = 1.0;
                  }
               }
               for(ii=tl.begin(); !rval && ii != tl.end() && !rval; ++ii)
               {
                  rval = (*ii)->validate_limits (bmn, bmx, 0, 0, X_AXIS, transformed);
                  if (!rval)
                     if ((rval=bindings->adjust_axis_minmax((*ii), bmn,max, transformed, AUTOMIN_ON, 1)) == ERR_STATUS_GOOD)
                     {
                        // reset log min value
                        (*ii)->adjust_logmin(blm);
                        cmd_manager->update_mask(STATE_CHNG_XLIMS);
                     }
               }
            }
            tmpId=cid->get_yaxis_id();
            if (((chips_strcmp(tmpId.c_str(), "all") == 0) ||
                ((tmpId.length () == 0) && cur_yall)) & !rval)
            {
               // WLM - add later
               rval = ERR_FEATURE_DNE; 
            }
            else if (((tmpAxisPtr=get_yaxis(tmpId.c_str()))) && !rval)
	    {
               list <ChipsAxis*> tl = bindings->get_axes(tmpAxisPtr);
               double bmn = 0, blm = 0, bmx = 0;
               rval = ERR_NO_DATA_FOR_AUTO;
               rval = bindings->find_axis_min_max(tmpAxisPtr, tmpAxisPtr->get_orientation(), &bmn, &bmx, &blm, transformed);

               if (bmn == max)
               {
                  // adjust to ensure range is not 1 point
                  if (bmn != 0)
                  {
                     bmn *= 0.95;
                     max *= 1.05;
                  }
                  else
                  {
                     bmn = -1.0;
                     max = 1.0;
                  }
               }
               for(ii=tl.begin(); !rval && ii != tl.end() && !rval; ++ii)
               {
                  rval = (*ii)->validate_limits (0, 0, bmn, bmx, Y_AXIS, transformed);
                  if (!rval)
                     if ((rval=bindings->adjust_axis_minmax((*ii), bmn,max, transformed, AUTOMIN_ON,1)) == ERR_STATUS_GOOD)
                     {
                        // reset log min value
                        (*ii)->adjust_logmin(blm);
                        cmd_manager->update_mask(STATE_CHNG_YLIMS);
                     }
               }
            }
	 }
         break;
         default:
            cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                                 "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
         break; 
      }
      cmd_manager->close_block();
   }
   return rval; 
}

int ChipsPlot::zoom (ChipsId* cid, float scale)
{
   int rval = ERR_STATUS_GOOD;
   // make sure the scale is greater than zero
   if (scale <=0)
   {
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"zoom scale",scale, 0.0);
      return rval;
   }

   // make sure the chips id exists
   if (cid == NULL)
   {
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
      return rval;
   }

   // retrieve the desired axes to zoom into to 
   string xaxis_id = cid->get_xaxis_id ();
   string yaxis_id = cid->get_yaxis_id ();
   ChipsAxis *ax = this->get_xaxis (xaxis_id);
   ChipsAxis *ay = this->get_yaxis (yaxis_id);
   ChipsDataSystem *ds;
   ChipsVector min_coord, max_coord, tmp_coord;

   if (!ax)
   {
      if ((chips_strcmp(xaxis_id.c_str(), "all") == 0) || ((xaxis_id.length () == 0) && cur_xall))
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
      else if (xaxis_id.length () != 0)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,xaxis_id.c_str());
      return rval;
   }

   if (!ay)
   {
      if ((chips_strcmp(yaxis_id.c_str(), "all") == 0) || ((yaxis_id.length () == 0) && cur_yall))
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
      else if (yaxis_id.length () != 0)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,yaxis_id.c_str());
      return rval;
   }
      

   ds = this->get_datasystem (ax, ay);

   if (!ds)
   {
      stringstream error_message;
      error_message << "The current axes, " << ax->get_id () << " and " << ay->get_id () << ", do not create a data system. Zoom cannot continue.";
      cmd_manager->add_err (error_message.str (), this->id, 0);
      return ERR_STATUS_BAD;
   }

   min_coord = ds->get_lower_bounds ();
   max_coord = ds->get_upper_bounds ();

   if (ax)
   {
      min_coord[0] = ax->get_absolute_min (false);
      max_coord[0] = ax->get_absolute_max (false);

      // convert to linear scale
      if (ax->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[0] = log10 (min_coord[0]);
         max_coord[0] = log10 (max_coord[0]);
      }
   }
 
   if (ay)
   {
      min_coord[1] = ay->get_absolute_min (false);
      max_coord[1] = ay->get_absolute_max (false);

      // convert to linear scale
      if (ay->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[1] = log10 (min_coord[1]);
         max_coord[1] = log10 (max_coord[1]);
      }
   }

   scale = 1/scale;  // invert the scale
   tmp_coord = max_coord * ((1-scale)*.5) + min_coord * ((1+scale)*.5);
   max_coord = max_coord * ((1+scale)*.5) + min_coord * ((1-scale)*.5);
   min_coord = tmp_coord;

   min_coord[2] = 1;
   max_coord[2] = 1;

   // convert back to log scale
   if (ax && ax->get_scaletype () == V_SCALE_LOG)
   {
      min_coord[0] = pow (10,min_coord[0]);
      max_coord[0] = pow (10,max_coord[0]);
   }

   if (ay && ay->get_scaletype () == V_SCALE_LOG)
   {
      min_coord[1] = pow (10,min_coord[1]);
      max_coord[1] = pow (10,max_coord[1]);
   }


   ds->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, ds->get_id_no_wcs (), DATA, "");
   ds->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, ds->get_id_no_wcs (), DATA, "");


   rval = ax->validate_limits (min_coord[0], max_coord[0], min_coord[1], max_coord[1], XY_AXIS,true);
   
   if (!rval)
   {
      rval = bindings->update_connected_nodes(ax, min_coord[0], max_coord[0],
            true, AUTOLIM_DEF, 0, 1, &min_coord[1], &max_coord[1], ay->get_id ());
   }
   
   return rval;
}

int ChipsPlot::panto (ChipsId* cid, double x, double y)
{
   int rval = ERR_STATUS_GOOD;

   // make sure the chips id exists
   if (cid == NULL)
   {
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
      return rval;
   }

   // retrieve the desired axes to zoom into to
   string xaxis_id = cid->get_xaxis_id ();
   string yaxis_id = cid->get_yaxis_id ();
   ChipsAxis *ax = this->get_xaxis (xaxis_id);
   ChipsAxis *ay = this->get_yaxis (yaxis_id);
   ChipsDataSystem *ds;
   ChipsVector min_coord(3), max_coord(3), width;
   ChipsVector coord (x,y);


   if (!ax)
   {
      if ((chips_strcmp(xaxis_id.c_str(), "all") == 0) || ((xaxis_id.length () == 0) && cur_xall))
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
      else if (xaxis_id.length () != 0)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,xaxis_id.c_str());
      return rval;
   }

   if (!ay)
   {
      if ((chips_strcmp(yaxis_id.c_str(), "all") == 0) || ((yaxis_id.length () == 0) && cur_yall))
         cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
      else if (yaxis_id.length () != 0)
         cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,yaxis_id.c_str());
      return rval;
   }

   ds = this->get_datasystem (ax, ay);

   rval = ax->validate_limits (coord, true);

   if (rval == ERR_STATUS_GOOD)
   {

      ds->convert_cs_to_cs (&coord, &coord, 1, DATA, "", DATA, ds->get_id_no_wcs ());

      if (!ds)
      {
         stringstream error_message;
         error_message << "The current axes, " << ax->get_id () << " and " << ay->get_id () << ", do not create a data system. Pan cannot continue.";
         cmd_manager->add_err (error_message.str (), this->id, 0);
         return ERR_STATUS_BAD;
      }
   
      min_coord = ds->get_lower_bounds ();
      max_coord = ds->get_upper_bounds ();
 
      if (ax)
      {
         min_coord[0] = ax->get_absolute_min (false);
         max_coord[0] = ax->get_absolute_max (false);

         if (ax->get_scaletype () == V_SCALE_LOG)
         {
            min_coord[0] = log10 (min_coord[0]);
            max_coord[0] = log10 (max_coord[0]);

            coord[0] = log10 (coord[0]);
         }

      }

      if (ay)
      {
         min_coord[1] = ay->get_absolute_min (false);
         max_coord[1] = ay->get_absolute_max (false);

         if (ay->get_scaletype () == V_SCALE_LOG)
         {
            min_coord[1] = log10 (min_coord[1]);
            max_coord[1] = log10 (max_coord[1]);

            coord[1] = log10 (coord[1]);
         }

      }

      width = (max_coord - min_coord)*.5;

      max_coord = coord + width;
      min_coord = coord - width;

      // convert back to log scale
      if (ax && ax->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[0] = pow (10,min_coord[0]);
         max_coord[0] = pow (10,max_coord[0]);
      }

      if (ay && ay->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[1] = pow (10,min_coord[1]);
         max_coord[1] = pow (10,max_coord[1]);
      }

      ds->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, ds->get_id_no_wcs (), DATA, "");
      ds->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, ds->get_id_no_wcs (), DATA, "");

      rval = bindings->update_connected_nodes(ax, min_coord[0], max_coord[0],               true, AUTOLIM_DEF, 0, 1, &min_coord[1], &max_coord[1], ay->get_id ());

   }

   return rval;

}


int ChipsPlot::pan (ChipsId* cid, double x1, double y1, double x2, double y2)
{
   int rval = ERR_STATUS_GOOD;

   // make sure the chips id exists
   if (cid == NULL)
   {
      cmd_manager->add_err(rval=ERR_NO_ID_OBJ,id,0);
      return rval;
   }

   // retrieve the desired axes to zoom into to
   string xaxis_id = cid->get_xaxis_id ();
   string yaxis_id = cid->get_yaxis_id ();
   ChipsAxis *ax = this->get_xaxis (xaxis_id);
   ChipsAxis *ay = this->get_yaxis (yaxis_id);
   ChipsDataSystem *ds;
   ChipsVector min_coord(3), max_coord(3), dist;
   ChipsVector coord1 (x1,y1);
   ChipsVector coord2 (x2,y2);

   if (ax)
   {
      if ((rval = ax->validate_limits (coord1, true)) == ERR_STATUS_GOOD)
         rval = ax->validate_limits (coord2, true);
   }


   if (rval == ERR_STATUS_GOOD)
   {
      if (!ax)
      {
         if ((chips_strcmp(xaxis_id.c_str(), "all") == 0) || ((xaxis_id.length () == 0) && cur_xall))
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
         else if (xaxis_id.length () != 0)
            cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,xaxis_id.c_str());
         return rval;
      }

      if (!ay)
      {
         if ((chips_strcmp(yaxis_id.c_str(), "all") == 0) || ((yaxis_id.length () == 0) && cur_yall))
            cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"axis");
         else if (yaxis_id.length () != 0)
            cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,yaxis_id.c_str());
         return rval;
      }

      ds = this->get_datasystem (ax, ay);

      if (!ds)
      {
         stringstream error_message;
         error_message << "The current axes, " << ax->get_id () << " and " << ay->get_id () << ", do not create a data system. Pan cannot continue.";
         cmd_manager->add_err (error_message.str (), this->id, 0);
         return ERR_STATUS_BAD;
      }

      ds->convert_cs_to_cs (&coord1, &coord1, 1, DATA, "", DATA, ds->get_id_no_wcs ());
      ds->convert_cs_to_cs (&coord2, &coord2, 1, DATA, "", DATA, ds->get_id_no_wcs ());

      min_coord = ds->get_lower_bounds ();
      max_coord = ds->get_upper_bounds ();

      if (ax)
      {
         min_coord[0] = ax->get_absolute_min (false);
         max_coord[0] = ax->get_absolute_max (false);

         if (ax->get_scaletype () == V_SCALE_LOG)
         {
            min_coord[0] = log10 (min_coord[0]);
            max_coord[0] = log10 (max_coord[0]);

            coord1[0] = log10 (coord1[0]);
            coord2[0] = log10 (coord2[0]);
         }
      }

      if (ay)
      {
         min_coord[1] = ay->get_absolute_min (false);
         max_coord[1] = ay->get_absolute_max (false);

         if (ay->get_scaletype () == V_SCALE_LOG)
         {
            min_coord[1] = log10 (min_coord[1]);
            max_coord[1] = log10 (max_coord[1]);

            coord1[1] = log10 (coord1[1]);
            coord2[1] = log10 (coord2[1]);
         }
      }

      dist = (coord1 - coord2);

      max_coord = max_coord + dist;
      min_coord = min_coord + dist;


      // convert back to log scale
      if (ax && ax->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[0] = pow (10,min_coord[0]);
         max_coord[0] = pow (10,max_coord[0]);
      }

      if (ay && ay->get_scaletype () == V_SCALE_LOG)
      {
         min_coord[1] = pow (10,min_coord[1]);
         max_coord[1] = pow (10,max_coord[1]);
      }

      ds->convert_cs_to_cs (&min_coord, &min_coord, 1, DATA, ds->get_id_no_wcs (
), DATA, "");
      ds->convert_cs_to_cs (&max_coord, &max_coord, 1, DATA, ds->get_id_no_wcs (
), DATA, "");

      rval = bindings->update_connected_nodes(ax, min_coord[0], max_coord[0],
            true, AUTOLIM_DEF, 0, 1, &min_coord[1], &max_coord[1], ay->get_id ()
);
   }

   return rval;
}

int ChipsPlot::format_axis_point (const ChipsId *cid, double pnt, string &formatted_pnt)
{
   int rval = ERR_STATUS_BAD;
   if (cid)
   {
      string aid = cid->get_axis_id();
      string xid = cid->get_xaxis_id();
      string yid = cid->get_yaxis_id();

      // precedence-
      //   1) single specified x_axis id or y_axis id
      //   2) a specified axis id
      if ((aid.length()==0) && (xid.length() != 0) && (yid.length() == 0))
         rval = format_axis_point(xid, pnt, formatted_pnt);
      else if ((aid.length()==0) && (xid.length() == 0) && (yid.length() != 0))
         rval = format_axis_point(yid, pnt, formatted_pnt);
      else
      {
         int usedim = cid->get_axis_dim();
         if (aid.length () == 0)
         { 
            if ((usedim == X_AXIS) && cur_x)
               aid = cur_x->get_id ();
            else if ((usedim == Y_AXIS) && cur_y)
               aid = cur_y->get_id ();
         }
          
         rval = format_axis_point(aid, pnt, formatted_pnt);
      }
   }
   return rval;

}


int ChipsPlot::format_axis_point (const string &axis_id, double pnt, string &formatted_pnt)
{
   int rval = ERR_STATUS_GOOD;
   ChipsAxis *ax;

   ax = get_xaxis (axis_id);
   if (!ax)
      ax = get_yaxis (axis_id);

   if (!ax)
      cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,axis_id.c_str());

   if (!rval)
      rval = ax->format_axis_point (pnt, formatted_pnt);

   return rval;

}


int ChipsPlot::set_axis_mode_property (ChipsAxis *ax, void *value, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   if (!ax)   return ERR_STATUS_BAD;
   
   if ((in_property == SET_AXIS_PAD) ||
       (in_property == SET_AXIS_MAJ_MODE) ||
       (in_property == SET_AXIS_MIN_MODE) ||
       (in_property == SET_AXIS_MAJOR_NUM) ||
       (in_property == SET_AXIS_MINOR_NUM) ||
       (in_property == SET_AXIS_TICKFORMAT)||
       (in_property == SET_AXIS_MAJORINT) ||
       (in_property == SET_AXIS_AUTOMAX) ||
       (in_property == SET_AXIS_AUTOMIN) ||
       (in_property == SET_AXIS_MINORINT))
   {
      list <ChipsAxis*> tl = bindings->get_axes(ax);
      list<ChipsAxis*>::iterator jj;
      for(jj=tl.begin(); jj != tl.end(); ++jj)
      {
         rval = (*jj)->set_property(value, in_property);
      }
      bindings->update_connected_node_properties(ax);
   }
   else 
      rval = ERR_STATUS_BAD;
   return rval;
}

int ChipsPlot::set_dimension_property(ChipsId* cid, int dim, int in_property, bool rec)
{
   int rval = ERR_NO_ID_OBJ;

   if (cid)
   {
      list<ChipsAxis*>::iterator jj;
      string tmpId;
      ChipsAxis* tmpAxisPtr;

      // don't allow log scaling for axes with data aspect ratios set
      if (in_property == SET_PLOT_LOGSCALE)
      {
         bool rat_set = false; 
         axes_map::iterator am;
         if ((dim & X_AXIS) == X_AXIS)
         {
            tmpId=cid->get_xaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)
            {
               if (cur_xall)  tmpId= "all";
               else if (cur_x) tmpId = cur_x->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
	       for(am=axes_x.begin(); am!=axes_x.end() && !rat_set; ++am)
                  rat_set = bindings->aspect_ratio_set((*am).second);
            else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
               rat_set = bindings->aspect_ratio_set(tmpAxisPtr); 
         }
         if (!rat_set && ((dim & Y_AXIS) == Y_AXIS))
         {
            tmpId=cid->get_yaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)
            {
               if (cur_yall)  tmpId= "all";
               else if (cur_y) tmpId = cur_y->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
               for(am=axes_y.begin(); am!=axes_y.end() && !rat_set; ++am)
                  rat_set = bindings->aspect_ratio_set((*am).second);
            else if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
               rat_set = bindings->aspect_ratio_set(tmpAxisPtr);
         }
         if (rat_set)
         {
            cmd_manager->add_err(rval=ERR_LOG_DASP, id, 0);
            return rval; 
         }
      }
      if (rec)  cmd_manager->open_block ();

      switch(dim)
      {
         case X_AXIS:
         {
            // precedence-
            //   1) single specified x_axis id 
            //   2) a specified axis id
            //   3) current axis  
            tmpId=cid->get_xaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)   
            {
               if (cur_xall)  tmpId= "all";
               else if (cur_x) tmpId = cur_x->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
            {
               int ival; 
               list <ChipsAxis *>axis_list;
	       for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
                  axis_list.push_back (am->second);

               switch(in_property)
               {
                  case SET_PLOT_LOGSCALE:
                     ival = V_SCALE_LOG;  
                     rval = bindings->update_property (axis_list, 
                           static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_PLOT_LINEARSCALE:
                     ival = V_SCALE_LINEAR;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_AXIS_REVERSE:
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_REVERSE, rec); 
                  break;
                  default:
                  break; 
	       }
            }
            else 
            {
               if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
               {
                  int ival; 
                  switch(in_property)
                  {
                     case SET_PLOT_LOGSCALE:
                        ival = V_SCALE_LOG;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_PLOT_LINEARSCALE:
                        ival = V_SCALE_LINEAR;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);

                     break;
                     case SET_AXIS_REVERSE:
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_REVERSE, rec); 
                     break;
                     default:
                     break; 
                  }
               }
            }
            if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_XLIMS); 
         }
         break;
	 case Y_AXIS:
         {
            // precedence-
            //   1) single specified y_axis id
            //   2) a specified axis id
            //   3) current axis  
            tmpId=cid->get_yaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)   
            {
               if (cur_yall)  tmpId= "all";
               else if (cur_y) tmpId = cur_y->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
            {
               int ival;
               list <ChipsAxis *>axis_list;
               for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
                  axis_list.push_back (am->second);

               switch(in_property)
               {
                  case SET_PLOT_LOGSCALE:
                     ival = V_SCALE_LOG;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_PLOT_LINEARSCALE:
                     ival = V_SCALE_LINEAR;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_AXIS_REVERSE:
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                  break;
                  default:
                  break; 
               }
            }
            else 
            { 
               if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
               {
                  int ival; 
                  switch(in_property)
                  {
                     case SET_PLOT_LOGSCALE:
                        ival = V_SCALE_LOG;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_PLOT_LINEARSCALE:
                        ival = V_SCALE_LINEAR;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_AXIS_REVERSE:
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                     break;
                     default:
                     break; 
                  }
               }
            }
            if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_YLIMS); 
         }
	 break;
	 case XY_AXIS:
	 {
            // precedence-
            //   1) single specified y_axis id
            //   2) a specified axis id
            //   3) current axis
            tmpId=cid->get_xaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)
            {
               if (cur_xall)  tmpId= "all";
               else if (cur_x) tmpId = cur_x->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
            {
               int ival;
               list <ChipsAxis *>axis_list;
               for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
                  axis_list.push_back (am->second);
               switch(in_property)
               {
                  case SET_PLOT_LOGSCALE:
                     ival = V_SCALE_LOG;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_PLOT_LINEARSCALE:
                     ival = V_SCALE_LINEAR;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_AXIS_REVERSE:
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                  break;
                  default:
                  break; 
               }
            }
            else 
            {
               if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
               {
                  int ival;
                  switch(in_property)
                  {
                     case SET_PLOT_LOGSCALE:
                        ival = V_SCALE_LOG;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_PLOT_LINEARSCALE:
                        ival = V_SCALE_LINEAR;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_AXIS_REVERSE:
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                     break;
                     default:
                     break; 
                  }
               }
            }
	   
            // precedence-
            //   1) single specified y_axis id
            //   2) a specified axis id
            //   3) current axis
            tmpId=cid->get_yaxis_id();
            if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
            if (tmpId.length() == 0)
            {
               if (cur_yall)  tmpId= "all";
               else if (cur_y) tmpId = cur_y->get_id();
            }
            if (chips_strcmp(tmpId.c_str(), "all") == 0)
            {
               int ival;
               list <ChipsAxis *>axis_list;
               for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
                  axis_list.push_back (am->second);
               switch(in_property)
               {
                  case SET_PLOT_LOGSCALE:
                     ival = V_SCALE_LOG;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_PLOT_LINEARSCALE:
                     ival = V_SCALE_LINEAR;  
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                  break;
                  case SET_AXIS_REVERSE:
                     rval = bindings->update_property (axis_list,
                        static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                  break;
                  default:
                  break; 
               }
            }
            else 
            {
               if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
               {
                  int ival;
                  switch(in_property)
                  {
                     case SET_PLOT_LOGSCALE:
                        ival = V_SCALE_LOG;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_PLOT_LINEARSCALE:
                        ival = V_SCALE_LINEAR;  
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_SCALE, rec);
                     break;
                     case SET_AXIS_REVERSE:
                        rval = bindings->update_property (tmpAxisPtr,
                                 static_cast<void*>(&ival), SET_AXIS_REVERSE, rec);
                     break;
                     default:
                     break; 
                  }
               }
            }
            if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_LIMS); 
         }
         break;
	 default:
            cmd_manager->add_err(rval=ERR_INT_SUBSET,id,0,"dimension",dim,
                                 "X_AXIS|Y_AXIS|XY_AXIS|RA_AXIS|DEC_AXIS|RADEC_AXIS");
         break; 
      }
      if (rec)  cmd_manager->close_block ();
   }
   if (cid && !rval)
   {
      list <string> axis_ids;
      map <string, ChipsAxis* >::iterator axis_iter;
      string tmpId;

      if ((dim & X_AXIS) == X_AXIS)
      {
         tmpId=cid->get_xaxis_id();
         if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
         if (tmpId.length() == 0)
         {
            if (cur_xall)  tmpId= "all";
            else if (cur_x) tmpId = cur_x->get_id();
         }
         if (chips_strcmp(tmpId.c_str(), "all") == 0)
         {
            for (axis_iter=axes_x.begin (); axis_iter!=axes_x.end (); axis_iter++)
               axis_ids.push_back (axis_iter->first);
            this->update_property (axis_ids, SET_AXIS_SCALE);
         }
         else
            this->update_property (tmpId, SET_AXIS_SCALE);
      }
      if ((dim & Y_AXIS) == Y_AXIS)
      {
         axis_ids.clear ();
         tmpId=cid->get_yaxis_id();
         if (tmpId.length() == 0)   tmpId=cid->get_axis_id();
         if (tmpId.length() == 0)
         {
            if (cur_xall)  tmpId= "all";
            else if (cur_y) tmpId = cur_y->get_id();
         }
         if (chips_strcmp(tmpId.c_str(), "all") == 0)
         {
            for (axis_iter=axes_y.begin (); axis_iter!=axes_y.end (); axis_iter++)
               axis_ids.push_back (axis_iter->first);
            this->update_property (axis_ids, SET_AXIS_SCALE);
         }
         else
            this->update_property (tmpId, SET_AXIS_SCALE);

      }
   }
   return rval; 
}

int ChipsPlot::add_legend(ChipsId* cid, const char** curve_ids,
   const char** curve_labels, const string& lgid)
{
   int rval = ERR_STATUS_GOOD;
   if (legend == NULL)
   {
      legend = new_legend(curve_ids, curve_labels, lgid);
      if (legend)
         cid->set_legend_id(legend->get_id());
      else
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"legend"); 
   }
   else 
      cmd_manager->add_err(rval=ERR_PLOT_HAS_LEGEND,id,0);
   return rval;
}

int ChipsPlot::delete_legend(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   string lgid;

   if (cid)  lgid = cid->get_legend_id();
   if (legend)
   {
      if ((lgid == "all") || (legend->get_id() == lgid))
      {
         delete legend;
         legend = NULL;
         cid->set_legend_id("");
         rval = ERR_STATUS_GOOD;
      } 
   }
   else cmd_manager->add_err(rval=ERR_OBJ_DNE,id,0,"legend");
   return rval;
}

int ChipsPlot::add_coordinate_system (ChipsAxis *ax, ChipsAxis *ay, Transform *transform, bool create_data_coord )
{
   int rval = ERR_STATUS_GOOD;
   bool ds_created = false;

   ChipsTransformFactory tf;
   ChipsVector data_min = ChipsVector(-1,-1);
   ChipsVector data_max = ChipsVector(1,1);
   ChipsDataSystem *ds = NULL;
   string data_sys_id;
   string axid = "";
   string ayid = "";


   if ((ax == NULL) && (ay == NULL)) 
   {
      cmd_manager->add_err(ERR_OBJ_DNE,id,0,"x/y axis");
      return ERR_OBJ_DNE;
   }
   else if (ax && ay)
   {
      if (ax->get_algorithm_type () != ay->get_algorithm_type ())
      {
         return ERR_STATUS_BAD;
      }
   }

   // Get data system info from axis
   if (ax)
   {
      data_min[0] = ax->get_absolute_min ();
      data_max[0] = ax->get_absolute_max ();
      axid = ax->get_id ();
   }

   if (ay)
   {
      data_min[1] = ay->get_absolute_min ();
      data_max[1] = ay->get_absolute_max ();
      ayid = ay->get_id ();
   }


   if (ax && ay && create_data_coord)
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;
      ChipsAxisAlgorithmFactory::TransformType alg_type = ax->get_algorithm_type ();

      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         if ((ds_iter->first.first == axid) ||
             (ds_iter->first.second == ayid))
         {
            if ((ds_iter->first.first == axid) && (ds_iter->first.second == ""))
            {
               ds = ds_iter->second;
               break;
            }
            else if ((ds_iter->first.first == "") && (ds_iter->first.second == ayid))
            {
               ds = ds_iter->second;
               break;
            }
            else if (alg_type == ChipsAxisAlgorithmFactory::WCSTAN)
            {
               // wcs axes can only have one pair between ra and dec
               rval = ERR_STATUS_BAD;
               break;
            }
         }
      }

      if (ds && !rval)
      {
         string ds_transform_id = ds->get_default_transform_id ();
         Transform *local_transform = ds->get_transform (ds_transform_id);

         // if there is no transform passed in, create a temporary
         // transform. this accounts for a default transform being put
         // on the data system
         if (transform == NULL)
            transform = ChipsTransformFactory::new_transform (ChipsTransformFactory::LINEAR2D);
         

         // make sure the transform being passed in is the same as the one
         // already attached to the data system
         if (local_transform && !tf.compare (local_transform, transform))
            rval = ERR_STATUS_BAD;
         else
         {
            // remove the data system as it will be added back 
            // at the bottom of this routine
            data_systems.erase (ds_iter);
            ds->modify_limits (data_min, data_max, ds_transform_id);
         }
      }
      else if (!rval)
      {
         ds = new ChipsDataSystem (coord, data_min, data_max,
                                axid, ayid);
         ds_created = true;
      }
   }
   else if (create_data_coord)
   {
      ds = new ChipsDataSystem (coord, data_min, data_max,
                                axid, ayid); 
      ds_created = true;
   }
   else if (coord)
   {
      // Find the data system
      ds = (dynamic_cast <ChipsPlotSystem *> (coord))->get_data_coord (DATA, axid, ayid);
   }

   // Create and add data system
   if (!rval)
   {
      if (!ds)
         cmd_manager->add_err(rval=ERR_COORD_DS_NO_ADD,id,0);
      else
      {

         // tell the axis about the data coord they created
         if (ax)
            rval = ax->set_data_coord_id (ds->get_id ());

         if (!rval && ay)
            rval = ay->set_data_coord_id (ds->get_id ());

         if (!rval)
         {
            bool lcl_add = false;

            // Add the data system to each of the axis
            pair <string, string> p;
            p.first = axid;
            p.second = ayid;
            data_systems[p] = ds;

            if (ax)
            {
               ds->set_log_flag (ax->get_scaletype(), X_AXIS);
               ds->set_xaxis_id (axid);
            }
            if (ay)
            {
               ds->set_log_flag (ay->get_scaletype(), Y_AXIS);
               ds->set_yaxis_id (ayid);
            }

            if (!transform)
            {
               transform = ChipsTransformFactory::new_transform (ChipsTransformFactory::LINEAR2D);
               lcl_add = true;
            }
            if (transform)
            {
               ds->add_transform (transform);
               ds->modify_limits (data_min, data_max, "");
               if (lcl_add)   delete transform;
            }
         }
      }
   }

   if (ds_created)
   {
      if (rval != ERR_STATUS_GOOD)
         delete ds;
      else 
         cmd_manager->update_mask(STATE_CHNG_COORD);
   }

   return rval;
}

int ChipsPlot::update_current_coord (ChipsId *cid)
{
   ChipsAxis *ax = cur_x;
   ChipsAxis *ay = cur_y; 
   string csid = "";
   int rval = ERR_STATUS_GOOD;
   COORD_SYS coord_sys = COORD_NONE;

   // Get the last x axis in the list
   if (cur_xall && (!axes_x.empty ())){
      axes_map::iterator ii = axes_x.end();
      ii--;
      ax = (*ii).second;
   }

   // Get the last y axis in the list
   if (cur_yall && (axes_y.size () > 0)){
      axes_map::iterator ii = axes_y.end();
      ii--;
      ay = (*ii).second;
   }

   if ((ax == top) || (ax == bottom) || (ay == left) || (ay == right))
      return ERR_STATUS_GOOD;

   if (ax || ay){
      map < pair <string, string>, ChipsDataSystem *>::iterator iter;
      ChipsDataSystem *tmp_coord = NULL;
      pair <string, string> axis_pair ("", "");
      if (ax)
         axis_pair.first = ax->get_id ();
      if (ay)
         axis_pair.second = ay->get_id ();
      iter = data_systems.find (axis_pair);
      if (iter != data_systems.end ())
      {
         tmp_coord = iter->second;
         csid = tmp_coord->get_id ();
         coord_sys = DATA;
      }

      if (cid){
	 cid->set_coord_sys (DATA);
	 cid->set_coord_id (csid);
      }
   }
   else if (cid){
      cid->set_coord_sys (PLOT_NORM);
      cid->set_coord_id (coord->get_id ());
   }

   cmd_manager->update_mask(STATE_CHNG_COORD);

   return rval;
}


ChipsCoordinate *ChipsPlot::find_coordinate_system (ChipsId &cid, int &err_status)
{
   COORD_SYS coord_sys = cid.get_coord_sys ();
   ChipsCoordinate *outCoord = static_cast <ChipsCoordinate *>(0);
   err_status = ERR_STATUS_GOOD;
   string coord_id;
   
   if (coord_sys == COORD_NONE)
   {
      err_status = this->update_current_coord (&cid);
      coord_sys = cid.get_coord_sys ();
   }
   coord_id = cid.get_coord_id ();

   if ((coord_sys == DATA) && !err_status)
   {
      // If coordinate id is not set look at the axes
      if (coord_id.length () == 0)
      {
         string ax_id = cid.get_xaxis_id ();
         string ay_id = cid.get_yaxis_id ();
         // Don't allow all axes to be set. This is consistent with curves
         if ((ax_id == "all") || (ay_id == "all"))
            err_status = ERR_OBJ_INVALID_ALL;
         else if ((err_status = this->update_current_coord (&cid)) == 
                                                            ERR_STATUS_GOOD)
         {
            if (cid.get_coord_sys () != DATA)
            {
               cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));
               err_status = ERR_COORD_DNE;
            }
            else 
               coord_id = cid.get_coord_id ();

            if  (coord_id.length () == 0)
            {
               stringstream error_message;
               if ((ax_id.length () == 0) && cur_x)
                  ax_id = cur_x->get_id ();
               if ((ay_id.length () == 0) && cur_y)
                  ay_id = cur_y->get_id ();
               if ((ax_id.length () > 0) && (ay_id.length () > 0))
                  error_message << "Could not find a valid coordinate system between axes, " << ax_id << " and " << ay_id <<"." << endl;
               else
                  error_message << "Could not find a valid coordinate system.\n";

               cmd_manager->add_err (error_message.str (), this->id, 0);
               err_status = ERR_STATUS_BAD;
            }
         }
      }
   }
   else if (coord_sys == PLOT_NORM)
   {
      cid.set_coord_id (coord->get_id ());
   }

   if (!err_status)
   {
      // Current don't allow window coords stop at frame norm. This
      // has to do to the fact objects are not currently storing
      // window coordinate systems but rather always frame coords.
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) || (coord_sys == PIXEL_REAL))
         outCoord = coord;
      else
         outCoord = coord->get_coord (coord_sys, coord_id);
      if (outCoord == NULL)
      {
         if (coord_sys == DATA)
            cmd_manager->add_err(ERR_COORD_DS_DNE,id,0,
                                  v_coord_sys_to_string(coord_sys), coord_id.c_str());
         else
            cmd_manager->add_err(ERR_COORD_DNE,id,0,v_coord_sys_to_string(coord_sys));

         err_status = ERR_COORD_DNE;
      }
   }
   return outCoord;
}

string ChipsPlot::get_current_xaxis_id()
{
   string xid = "\0";
   if (cur_x)   xid = cur_x->get_id();
   return xid; 
}

string ChipsPlot::get_current_yaxis_id()
{
   string yid = "\0";
   if (cur_y)   yid = cur_y->get_id();
   return yid; 
}

char* ChipsPlot::info(ChipsId* cid, V_OBJECT_TYPE obj, bool current_obj)
{
   char* ptr = NULL;
   string temp, val;
   map <int, string>chron;
   char buffer[80], buffer2[80];
   int  infdepth = theDefaults->get_infodepth();
   int  infcur   = theDefaults->get_infocurrent();
   int  infcoord = theDefaults->get_infocoord();

   if ((obj == chips_axis) || (obj <= chips_plot))
   {
      if (axes_x_exist())
      {
         if (bottom && top && left && right)
         {
            temp.append("      Border bottom [");
            temp.append(bottom->get_id());
            temp.append("]  top [");
            temp.append(top->get_id());
            temp.append("]  left [");
            temp.append(left->get_id());
            temp.append("]  right [");
            temp.append(right->get_id());
            temp.append("]\n");
         }
         for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
	    if (((*am).second != top) && ((*am).second != bottom))
	    {
               int rev = (infcur && current_obj &&
                          ((cur_x == (*am).second) || cur_xall));
               val = "      X Axis [";
               if (rev)  val.append("\x1b[7m");
               val.append((*am).second->get_id());
               if (rev)  val.append("\x1b[m");
               val.append("]");
               if (infdepth || infcoord)
               {
                  if (rev)
                     sprintf(buffer,"%-34s ", val.c_str());
                  else
                     sprintf(buffer,"%-27s ", val.c_str());
                  val = buffer;
                  if (infdepth)
                  {
                     sprintf(buffer,"d %1d", (*am).second->get_depth());
                     sprintf(buffer2,"%-9s", buffer);
                     val.append(buffer2);
                  }
                  if (infcoord)
                  {
                     COORD_SYS csys= (*am).second->get_coord_system();
                     if (csys == DATA)
                     {
                        string ds = (*am).second->get_data_system_id();
                        sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                           ds.c_str());
                     }
                     else
                        sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
                     val.append(buffer);
                  }
               }
               val.append("\n");
               chron.insert(pair<int, string>((*am).second->get_ref_cnt(),val));
	    }
      }
      if (axes_y_exist())
         for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
	    if (((*am).second != left) && ((*am).second != right))
	    {
               int rev = (infcur && current_obj &&
                          ((cur_y == (*am).second) || cur_yall));
               val = "      Y Axis [";
               if (rev)  val.append("\x1b[7m");
               val.append((*am).second->get_id());
               if (rev)  val.append("\x1b[m");
               val.append("]");
               if (infdepth || infcoord)
               {
                  if (rev)
                     sprintf(buffer,"%-34s ", val.c_str());
                  else
                     sprintf(buffer,"%-27s ", val.c_str());
                  val = buffer;
                  if (infdepth)
                  {
                     sprintf(buffer,"d %1d", (*am).second->get_depth());
                     sprintf(buffer2,"%-9s", buffer);
                     val.append(buffer2);
                  }
                  if (infcoord)
                  {
                     COORD_SYS csys= (*am).second->get_coord_system();
                     if (csys == DATA)
                     {
                        string ds = (*am).second->get_data_system_id();
                        sprintf(buffer,"cs %s:%s",v_coord_sys_to_enum_string(csys),
                           ds.c_str());
                     }
                     else
                        sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
                     val.append(buffer);
                  }
               }
               val.append("\n");
               chron.insert(pair<int, string>((*am).second->get_ref_cnt(),val));
	    }
   }

   if (!curves.empty() && ((obj == chips_curve) || (obj <= chips_plot)))
   {
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
      {
         int rev = (infcur && current_obj &&
                          ((currcurv == (*cc).first)||( currcurv == "all")));
         val = "      Curve [";
         if (rev)  val.append("\x1b[7m");
         val.append((*cc).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infcoord || infdepth)
         {
            if (rev)
               sprintf(buffer,"%-34s ", val.c_str());
            else
               sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*cc).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*cc).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*cc).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s",v_coord_sys_to_enum_string(csys),
                     ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
               val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), val));
      }
   }
   if (contours_exist() && ((obj == chips_contour) || (obj <= chips_plot)))
   {
      for(contours_map::iterator cc=contours.begin(); cc!=contours.end(); ++cc)
      {
         int rev = (infcur && current_obj &&
                          ((currcont == (*cc).first)||( currcont == "all")));
         val = "      Contour [";
         if (rev)  val.append("\x1b[7m");
         val.append((*cc).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infcoord || infdepth)
         {
            if (rev)
               sprintf(buffer,"%-34s ", val.c_str());
            else
               sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*cc).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*cc).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*cc).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s",v_coord_sys_to_enum_string(csys),
                     ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
               val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), val));
      }
   }
   if (histograms_exist() && ((obj == chips_histogram) || (obj <= chips_plot)))
   {
      for(histograms_map::iterator hh=histograms.begin(); hh!=histograms.end(); ++hh)
      {
         int rev = (infcur && current_obj &&
                          ((currhist == (*hh).first)||(currhist == "all")));
         val = "      Histogram [";
         if (rev)  val.append("\x1b[7m");
         val.append((*hh).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infcoord || infdepth)
         {
            if (rev)
               sprintf(buffer,"%-34s ", val.c_str());
            else
               sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*hh).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*hh).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*hh).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s",v_coord_sys_to_enum_string(csys),
                     ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
               val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*hh).second->get_ref_cnt(), val));
      }
   }
   if (!images.empty() && ((obj == chips_image) || (obj <= chips_plot)))
   {
      // list all images in frame
      for(map <string, ChipsImage*>::iterator ii=images.begin(); ii!=images.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currimage == (*ii).first) || ( currimage == "all")));
         val = "      Image [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }
   if (!labels.empty() && ((obj == chips_label) || (obj <= chips_plot)))
   {
      // list all labels in frame
      for(map <string, ChipsLabel*>::iterator ii=labels.begin(); ii!=labels.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currlabel == (*ii).first) || ( currlabel == "all")));
         val = "      Label [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }
   if (!lines.empty() && ((obj == chips_line) || (obj <= chips_plot)))
   {
      // list all lines in frame
      for(map <string, ChipsLine*>::iterator ii=lines.begin(); ii!=lines.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currline == (*ii).first) || ( currline == "all")));
         val = "      Line [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }

   if (!cbars.empty() && ((obj == chips_colorbar) || (obj <= chips_plot)))
   {
      // list all colorbars in plot
      for(map <string, ChipsColorBar*>::iterator ii=cbars.begin(); ii!=cbars.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currcbar == (*ii).first) || ( currcbar == "all")));
         val = "      ColorBar [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }
   if (!points.empty() && ((obj == chips_point) || (obj <= chips_plot)))
   {
      // list all points in plot
      for(map <string, ChipsPoint*>::iterator ii=points.begin(); ii!=points.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currpoint == (*ii).first) || ( currpoint == "all")));
         val = "      Point [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }
   if (!regions.empty() && ((obj == chips_region) || (obj <= chips_plot)))
   {
      // list all regions in plot
      for(map <string, ChipsRegion*>::iterator ii=regions.begin(); ii!=regions.end(); ++ii)
      {
         int rev = (infcur && current_obj &&
                   ((currregion == (*ii).first) || ( currregion == "all")));
         val = "      Region [";
         if (rev)  val.append("\x1b[7m");
         val.append((*ii).first);
         if (rev)  val.append("\x1b[m");
         val.append("]");
         if (infdepth || infcoord)
         {
            if (rev)  sprintf(buffer,"%-34s ", val.c_str());
               else   sprintf(buffer,"%-27s ", val.c_str());
            val = buffer;
            if (infdepth)
            {
               sprintf(buffer,"d %1d", (*ii).second->get_depth());
               sprintf(buffer2,"%-9s", buffer);
               val.append(buffer2);
            }
            if (infcoord)
            {
               COORD_SYS csys= (*ii).second->get_coord_system();
               if (csys == DATA)
               {
                  string ds = (*ii).second->get_data_system_id();
                  sprintf(buffer,"cs %s:%s", v_coord_sys_to_enum_string(csys),
                          ds.c_str());
               }
               else
                  sprintf(buffer,"cs %s", v_coord_sys_to_enum_string(csys));
	       val.append(buffer);
            }
         }
         val.append("\n");
         chron.insert(pair<int, string>((*ii).second->get_ref_cnt(), val));
      }
   }

   // add in the plot objects list- ordered chronologically
   map <int, string>::iterator i2;
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      temp.append((*i2).second);
   chron.clear();

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}

char* ChipsPlot::list_current()
{
   char* ptr = NULL;
   string temp,val;
   map <int, string>chron;

   if (axes_x_exist())
   {
      if (cur_xall)
      {
	 for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
         {
            temp = "      X Axis [";
            temp.append((*am).second->get_id());
            temp.append("]\n");
            chron.insert(pair<int, string>((*am).second->get_ref_cnt(),temp));
         }
      }
      else if (cur_x)
      {
         temp = "      X Axis [";
         temp.append(cur_x->get_id());
         temp.append("]\n");
         chron.insert(pair<int, string>(cur_x->get_ref_cnt(),temp));
      }
   }
   if (axes_y_exist())
   {
      if (cur_yall)
      {
         for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
         {
            temp = "      Y Axis [";
            temp.append((*am).second->get_id());
            temp.append("]\n");
            chron.insert(pair<int, string>((*am).second->get_ref_cnt(),temp));
         }
      }
      else if (cur_y)
      {
         temp = "      Y Axis [";
         temp.append(cur_y->get_id());
         temp.append("]\n");
         chron.insert(pair<int, string>(cur_y->get_ref_cnt(),temp));
      }
   }
   if (curves_exist())
   {
      if (chips_strcmp(currcurv.c_str(), "all") == 0)
      {
	 for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
         {
            temp = "      Curve [";
            temp.append((*cc).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*cc).second->get_ref_cnt(),temp));
         }
      }
      else 
      {
         ChipsCurve* cc;
         if ((cc = get_curve()) != NULL)
         {
            temp = "      Curve [";
            temp.append(currcurv);
            temp.append("]\n");
            chron.insert(pair<int, string>(cc->get_ref_cnt(),temp));
	 }
      }
   }
   if (contours_exist ())
   {
      if (chips_strcmp(currcont.c_str(), "all") == 0)
      {
         for(contours_map::iterator cc=contours.begin();cc!=contours.end();++cc)
         {
            temp = "      Contour [";
            temp.append((*cc).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*cc).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsContour* cc; 
         if ((cc = get_contour()) != NULL)
         {
            temp = "      Contour [";
            temp.append(currcont);
            temp.append("]\n");
            chron.insert(pair<int, string>(cc->get_ref_cnt(),temp));
	 }
      }
   }
   if (histograms_exist())
   {
      if (chips_strcmp(currhist.c_str(), "all") == 0)
      {
         for(histograms_map::iterator hh=histograms.begin();hh!=histograms.end();++hh)

         {
            temp = "      Histogram [";
            temp.append((*hh).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*hh).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsHistogram* hh; 
         if ((hh = get_histogram()) != NULL)
         {
            temp = "      Histogram [";
            temp.append(currhist);
            temp.append("]\n");
            chron.insert(pair<int, string>(hh->get_ref_cnt(),temp));
	 }
      }
   }
   if (currimage != "")
   {
      if (chips_strcmp(currimage.c_str(), "all") == 0)
      {
         map <string, ChipsImage* >::iterator ii;
         for(ii=images.begin(); ii != images.end(); ++ii)
         {
            temp = "      Image [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsImage* ci; 
         if ((ci = get_image()) != NULL)
         {
            temp = "      Image [";
            temp.append(currimage);
            temp.append("]\n");
            chron.insert(pair<int, string>(ci->get_ref_cnt(),temp));
	 }
      }
   }
   if (currlabel != "")
   {
      if (chips_strcmp(currlabel.c_str(), "all") == 0)
      {
         map <string, ChipsLabel* >::iterator ii;
         for(ii=labels.begin(); ii != labels.end(); ++ii)
         {
            temp = "      Label [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsLabel* ll; 
         if ((ll = get_label()) != NULL)
         {
            temp = "      Label [";
            temp.append(currlabel);
            temp.append("]\n");
            chron.insert(pair<int, string>(ll->get_ref_cnt(),temp));
	 }
      }
   }
   if (currline!= "")
   {
      if (chips_strcmp(currline.c_str(), "all") == 0)
      {
         map <string, ChipsLine* >::iterator ii;
         for(ii=lines.begin(); ii != lines.end(); ++ii)
         {
            temp = "      Line [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsLine* ll; 
         if ((ll = get_line()) != NULL)
         {
            temp = "      Line [";
            temp.append(currline);
            temp.append("]\n");
            chron.insert(pair<int, string>(ll->get_ref_cnt(),temp));
	 }
      }
   }
   if (currcbar!= "")
   {
      if (chips_strcmp(currcbar.c_str(), "all") == 0)
      {
         map <string, ChipsColorBar* >::iterator ii;
         for(ii=cbars.begin(); ii != cbars.end(); ++ii)
         {
            temp = "      ColorBar [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsColorBar* ll; 
         if ((ll = get_colorbar()) != NULL)
         {
            temp = "      ColorBar [";
            temp.append(currcbar);
            temp.append("]\n");
            chron.insert(pair<int, string>(ll->get_ref_cnt(),temp));
	 }
      }
   }
   if (currpoint!= "")
   {
      if (chips_strcmp(currpoint.c_str(), "all") == 0)
      {
         map <string, ChipsPoint* >::iterator ii;
         for(ii=points.begin(); ii != points.end(); ++ii)
         {
            temp = "      Point [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsPoint* ll; 
         if ((ll = get_point()) != NULL)
         {
            temp = "      Point [";
            temp.append(currpoint);
            temp.append("]\n");
            chron.insert(pair<int, string>(ll->get_ref_cnt(),temp));
	 }
      }
   }
   if (currregion!= "")
   {
      if (chips_strcmp(currregion.c_str(), "all") == 0)
      {
         map <string, ChipsRegion* >::iterator ii;
         for(ii=regions.begin(); ii != regions.end(); ++ii)
         {
            temp = "      Region [";
            temp.append((*ii).first);
            temp.append("]\n");
            chron.insert(pair<int, string>((*ii).second->get_ref_cnt(),temp));
         }
      }
      else
      {
         ChipsRegion* ll; 
         if ((ll = get_region()) != NULL)
         {
            temp = "      Region [";
            temp.append(currregion);
            temp.append("]\n");
            chron.insert(pair<int, string>(ll->get_ref_cnt(),temp));
	 }
      }
   }

   // add in the plot objects list- ordered chronologically
   val = "";
   map <int, string>::iterator i2;
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      val.append((*i2).second);
   chron.clear();

   int ll = val.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, val.c_str());
   }
   return ptr;
}

ChipsCurve* ChipsPlot::new_curve(long cnt, double* x, double* y, const string &transform_id, const string& cvid)
{
   string tmpid;
   if (cvid.length())
      tmpid = cvid;
   else
      tmpid = create_curve_id();
   return new ChipsCurve(this, cnt, x, y, transform_id, tmpid);
}

ChipsCurve* ChipsPlot::new_curve(long cnt, double* x, double* y,
   double* ydown, double* yup, double* xdown, double* xup, double* ylims, 
   double* xlims, const string &transform_id, const string& cvid)
{
   string tmpid;
   if (cvid.length())
      tmpid = cvid;
   else
      tmpid = create_curve_id();
   return new ChipsCurve(this, cnt, x, y, ydown, yup, xdown, xup, ylims, xlims,
                         transform_id, tmpid);
}

ChipsContour* ChipsPlot::new_contour(long cnt, double* x, double* y,
   double* z,  unsigned int count, double *levels, string transform_id,
   const string& ctid)
{
   string tmpid;
   if (ctid.length())
      tmpid = ctid;
   else
      tmpid = create_contour_id();
   return new ChipsContour(this, cnt, x, y, z, count, levels, transform_id, tmpid);
}

ChipsHistogram* ChipsPlot::new_histogram(long cnt, double* lo, double* hi, 
   double* counts, double* yd, double* yu, const string& hid)
{
   string tmpid;
   ChipsHistogram* ch = NULL; 

   if (hid.length())
      tmpid = hid;
   else
      tmpid = create_histogram_id();
   if ((ch = new ChipsHistogram(this, cnt, lo, hi, counts, yd, yu, tmpid)) 
       != NULL)   ch->initialize();
   return ch; 
}

ChipsAxis* ChipsPlot::new_axis(ChipsCoordinate *inCoord, COORD_SYS sys,
                              double inMinX, double inMinY, double inMaxX,
                              double inMaxY, int inOrientation,
                              ChipsAxisAlgorithmFactory::TransformType type,
                              const string& aid)
{
   ChipsAxis *ax = NULL;
   string tmpid;

   if (aid.length())
      tmpid = aid;
   else
      tmpid = create_axis_id(inOrientation);
   ax =  new ChipsAxis(this, inCoord, sys, inMinX, inMinY, inMaxX, inMaxY, inOrientation, type, tmpid);
   return ax;
}


ChipsAxis* ChipsPlot::new_axis(ChipsCoordinate *inCoord, COORD_SYS sys,
			      double inMinX, double inMinY, double inMaxX,
                              double inMaxY, int inOrientation,
			      double inMin,double inMax, 
                              ChipsAxisAlgorithmFactory::TransformType type,
                              const string& aid)
{
   ChipsAxis *ax = NULL;
   string tmpid;

   if (aid.length())
      tmpid = aid;
   else
      tmpid = create_axis_id(inOrientation);
   ax =  new ChipsAxis(this, inCoord, sys, inMinX, inMinY, inMaxX, inMaxY, inOrientation, inMin, inMax, type, tmpid);
   return ax;
}

ChipsAxis* ChipsPlot::new_axis(ChipsStateStream& in)
{
   return new ChipsAxis(this, coord, coord_system, in);
}

ChipsCurve* ChipsPlot::new_curve(ChipsStateStream& in)
{
   return new ChipsCurve(this, in);
}

ChipsContour* ChipsPlot::new_contour(ChipsStateStream& in)
{
   return new ChipsContour(this, in);
}

ChipsHistogram* ChipsPlot::new_histogram(ChipsStateStream& in)
{
   return new ChipsHistogram(this, in);
}

ChipsLegend* ChipsPlot::new_legend(const char** cids, const char** clabs,
   const string& lgid)
{
   string tmpid;

   if (lgid.length())
      tmpid = lgid;
   else
      tmpid = create_legend_id();
   return new ChipsLegend(this, cids, clabs, tmpid);
}

char* ChipsPlot::create_curve_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_curvestem();
      else            stem = "crv";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++curve_cnt);

   // increment count til next unused curve id
   while(get_curve(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++curve_cnt);
   }
   return temp_id;
}

char* ChipsPlot::create_contour_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_contourstem();
      else            stem = "ctr";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++contour_cnt);

   // increment count til next unused contour id
   while(get_contour(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++contour_cnt);
   }
   return temp_id;
}

char* ChipsPlot::create_histogram_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_histogramstem();
      else            stem = "hst";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++histogram_cnt);

   // increment count til next unused histogram id
   while(get_histogram(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++histogram_cnt);
   }
   return temp_id;
}

char* ChipsPlot::create_axis_id(int dim)
{
   static char temp_id[80];
   string stem;

   if (dim == X_AXIS)
   {
      if (theDefaults)   stem = theDefaults->get_axisxstem();
         else            stem = "ax";
      sprintf(temp_id, "%s%1d", stem.c_str(), ++axis_x_cnt);

      // increment count til next unused axis id
      while(get_xaxis(temp_id) || get_yaxis(temp_id))
      {
         sprintf(temp_id, "%s%1d", stem.c_str(), ++axis_x_cnt);
      }
   }
   else 
   {
      if (theDefaults)   stem = theDefaults->get_axisystem();
         else            stem = "ay";
      sprintf(temp_id, "%s%1d", stem.c_str(), ++axis_y_cnt);

      // increment count til next unused axis id
      while(get_xaxis(temp_id) || get_yaxis(temp_id))
      {
         sprintf(temp_id, "%s%1d", stem.c_str(), ++axis_y_cnt);
      }
   }
   return temp_id;
}

char* ChipsPlot::create_legend_id()
{
   static char temp_id[80];

   ++legend_cnt;
   if (theDefaults)
   {
      string stem = theDefaults->get_legendstem();
      sprintf(temp_id, "%s%1d", stem.c_str(), legend_cnt);
   }
   else
      sprintf(temp_id, "key%1d", legend_cnt);
   return temp_id;
}

bool ChipsPlot::images_exist ()
{
   return (!images.empty());
}

bool ChipsPlot::curves_exist ()
{
   return (!curves.empty ());
}

bool ChipsPlot::axes_x_exist ()
{
   return (!axes_x.empty ());
}

bool ChipsPlot::axes_y_exist ()
{
   return (!axes_y.empty ());
}

bool ChipsPlot::histograms_exist ()
{
   return (!histograms.empty ());
}

bool ChipsPlot::contours_exist ()
{
   return (!contours.empty ());
}

int ChipsPlot::reposition(float x1, float y1, float x2, float y2)
{
   int rval = ERR_STATUS_GOOD;

   if (x2 <= x1)
      cmd_manager->add_err(rval=ERR_FLT_V1_GT_V2,id,0,"x2",x2,"x1",x1);
   else if (y2 <= y1)
      cmd_manager->add_err(rval=ERR_FLT_V1_GT_V2,id,0,"y2",y2,"y1",y1);
   else if ((x1 < 0) || (x2 > 1 + FLT_EPSILON) || (y1 < 0) || (y2 > 1 + FLT_EPSILON))
      cmd_manager->add_err(rval=ERR_EDGE_BOUNDARY,id,0,"plot"); 
   else 
   {
      left_margin = x1;
      right_margin = 1 - x2; 
      bottom_margin = y1;
      top_margin = 1 - y2; 
      adjust_plot_anchors(); 
      recalculate_position(); 
   }
   return rval;
}

int ChipsPlot::move(COORD_SYS coord_sys, const string &coord_id, double xx,
                   double yy, int mode, bool rec)
{
   ChipsVector tmp_n (xx, yy);
   ChipsVector tmp_org (0.0, 0.0);
   int rval = ERR_STATUS_GOOD;
   ChipsMoveCommand* cmd = NULL;

   if (!coord)
   {
      cmd_manager->add_err(ERR_COORD_OBJ_DNE, id, 0);
      return ERR_COORD_OBJ_DNE;
   }
   if (coord_sys == COORD_NONE)  coord_sys = coord_system;
   coord->convert_cs_to_cs(&tmp_n, &tmp_n, 1, coord_sys, coord_id, 
			    FRAME_NORM, "");
   if (rec)
      cmd = new ChipsMoveCommand(this, left_margin, bottom_margin, xx, yy, 
         coord_sys, coord_id, mode, ((ChipsFrame *)parent)->get_current_plot());
   if (mode == V_MOVE_DELTA)
   {
      rval = coord->convert_cs_to_cs(&tmp_org, &tmp_org, 1, coord_sys,
                                      coord_id, FRAME_NORM, "");
      tmp_n -= tmp_org;
      rval = reposition(bx1+tmp_n[0], by1+tmp_n[1], bx2+tmp_n[0], by2+tmp_n[1]);
   }
   else
   {
      tmp_org[0] =  1 - right_margin + tmp_n[0] - left_margin;
      tmp_org[1] =  1 - top_margin  + tmp_n[1] - bottom_margin;
      rval = reposition(tmp_n[0],tmp_n[1], tmp_org[0], tmp_org[1]);
   }
   if (rval == ERR_STATUS_GOOD)   
   {
      // update plot property mask for margins 
      cmd_manager->update_mask(STATE_CHNG_PROP_PLT);
      if (rec)
      {
         if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
   }
   else if (rec)
      delete(cmd);
   return rval;
}

int ChipsPlot::set_plot_lmargin(float margin_sz)
{
   int rval = ERR_STATUS_GOOD; 
   if ((margin_sz >=0) && (margin_sz < (1 - right_margin)))
   {
      left_margin = margin_sz;
      if (coord)  adjust_plot_anchors(); 
      recalculate_position(); 
   }
   else 
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"left",margin_sz,
                           "right",right_margin);
   return rval; 
}

int ChipsPlot::set_plot_rmargin(float margin_sz)
{
   int rval = ERR_STATUS_GOOD; 
   if ((margin_sz >=0) && (margin_sz < (1 - left_margin)))
   {
      right_margin = margin_sz;
      if (coord)   adjust_plot_anchors(); 
      recalculate_position(); 
   }
   else 
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"right",margin_sz,
                           "left",left_margin);
   return rval; 
}

int ChipsPlot::set_plot_tmargin(float margin_sz)
{
   int rval = ERR_STATUS_GOOD; 
   if ((margin_sz >=0) && (margin_sz < (1 - bottom_margin)))
   {
      top_margin = margin_sz;
      if (coord)   adjust_plot_anchors(); 
      recalculate_position(); 
   }
   else 
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"top",margin_sz,
                           "bottom",bottom_margin);
   return rval; 
}

int ChipsPlot::set_plot_bmargin(float margin_sz)
{
   int rval = ERR_STATUS_GOOD; 
   if ((margin_sz >=0) && (margin_sz < (1 - top_margin)))
   {
      bottom_margin = margin_sz;
      if (coord)   adjust_plot_anchors(); 
      recalculate_position(); 
   }
   else 
      cmd_manager->add_err(rval=ERR_MARGIN_SIZE,id,0,"bottom",margin_sz,
                           "top",top_margin);
   return rval; 
}


int ChipsPlot::find_axis_min_max(ChipsAxis* axis, int dim,
    double* min, double* max, double* lmin, double* tmin, 
    double* tmax, bool init, ChipsAxis* altax)
{

   if ((axis == top) || (axis == bottom) || (axis == left) || (axis == right))
      return ERR_STATUS_GOOD;

   if (axis && min && max && lmin)
   {
      ChipsDataSystem *dsCoord = axis->get_data_system ();

      if (!dsCoord)
         return ERR_STATUS_BAD;

      string image_id = dsCoord->get_id_no_wcs ();

      ChipsVector tmp_min(3);
      ChipsVector tmp_lmin(3);
      ChipsVector tmp_max(3);

      ChipsVector tmp_tmin(3); // transformed
      ChipsVector tmp_tmax(3); // transformed

      int index = (dim == X_AXIS) ? 0 : 1;

      ChipsCurve* cv; 
      ChipsContour* ct; 
      ChipsHistogram* ch; 
      ChipsImage* im; 
      string axid, axidalt, axidx, axidy;
      axid = axis->get_id();
      if (altax)  axidalt = altax->get_id(); 

      // iterate through curves to find potential axes limits 
      for(curves_map::iterator cm = curves.begin(); cm != curves.end(); ++cm)
      {
         cv = (*cm).second;
         axidx = cv->get_xaxis();
         axidy = cv->get_yaxis();

         // ignore objects not attached to requested axis
         if ((axidx != axid) && (axidy != axid))
            continue;
         // if an alternate axis was provided, we only want to deal with objects
         // using the requested and alt axes (ie. in the given coord sys)
         if (altax && ((axidx == axid && axidy != axidalt) || 
             (axidy == axid && axidx != axidalt)))
            continue; 

         tmp_tmin[0] = cv->get_xmin(true);
         tmp_lmin[0] = cv->get_xlmin();
         tmp_tmax[0] = cv->get_xmax(true);
         tmp_tmin[1] = cv->get_ymin(true);
         tmp_lmin[1] = cv->get_ylmin();
         tmp_tmax[1] = cv->get_ymax(true);

         dsCoord->convert_cs_to_cs (&tmp_tmin, &tmp_min, 1, DATA, "", DATA, image_id);
         dsCoord->convert_cs_to_cs (&tmp_tmax, &tmp_max, 1, DATA, "", DATA, image_id);
            
         if (init)
         {
            *min  = tmp_min[index];
            *tmin = tmp_tmin[index];
            if (tmp_lmin[index] > 0)  *lmin = tmp_lmin[index]; 
               else *lmin = 0;  
            *max  = tmp_max[index]; 
            *tmax = tmp_tmax[index];
            init = false; 
         }
         else 
         {
            if (tmp_min[index] < *min)
            {
               *min = tmp_min[index];  
               *tmin = tmp_tmin[index];
            }
            if ((tmp_lmin[index] > 0) && ((tmp_lmin[index] < *lmin) || (*lmin == 0)))
                  *lmin = tmp_lmin[index];  
            if (tmp_max[index] > *max)
            {
               *max = tmp_max[index];  
               *tmax = tmp_tmax[index];
            }
         }
      }

      for(histograms_map::iterator hm = histograms.begin(); hm != histograms.end(); ++hm)
      {
         ch = (*hm).second;
         axidx = ch->get_xaxis();
         axidy = ch->get_yaxis();

         // ignore objects not attached to requested axis
         if ((axidx != axid) && (axidy != axid))
            continue;
         // if an alternate axis was provided, we only want to deal with objects
         // using the requested and alt axes (ie. in the given coord sys)
         if (altax && ((axidx == axid && axidy != axidalt) || 
             (axidy == axid && axidx != axidalt)))
            continue; 

         tmp_tmin[0] = ch->get_xmin(true);
         tmp_lmin[0] = ch->get_xlmin();
         tmp_tmax[0] = ch->get_xmax(true);
         tmp_tmin[1] = ch->get_ymin(true);
         tmp_lmin[1] = ch->get_ylmin();
         tmp_tmax[1] = ch->get_ymax(true);

         dsCoord->convert_cs_to_cs (&tmp_tmin, &tmp_min, 1, DATA, "", DATA, image_id);
         dsCoord->convert_cs_to_cs (&tmp_tmax, &tmp_max, 1, DATA, "", DATA, image_id);

         if (init)
         {
            *min  = tmp_min[index];
            *tmin = tmp_tmin[index];
            if (tmp_lmin[index] > 0)  *lmin = tmp_lmin[index];
               else *lmin = 0;
            *max  = tmp_max[index];
            *tmax = tmp_tmax[index];
            init = false;
         }
         else
         {
            if (tmp_min[index] < *min)
            {
               *min = tmp_min[index];
               *tmin = tmp_tmin[index];
            }
            if ((tmp_lmin[index] > 0) && ((tmp_lmin[index] < *lmin) || (*lmin == 0)))
                  *lmin = tmp_lmin[index];
            if (tmp_max[index] > *max)
            {
               *max = tmp_max[index];
               *tmax = tmp_tmax[index];
            }
         }
      }

      for(contours_map::iterator cm=contours.begin();cm!=contours.end();++cm)
      {
         ct = (*cm).second;
         axidx = ct->get_xaxis();
         axidy = ct->get_yaxis();

         // ignore objects not attached to requested axis
         if ((axidx != axid) && (axidy != axid))
            continue;
         // if an alternate axis was provided, we only want to deal with objects
         // using the requested and alt axes (ie. in the given coord sys)
         if (altax && ((axidx == axid && axidy != axidalt) || 
             (axidy == axid && axidx != axidalt)))
            continue; 

         tmp_tmin[0] = ct->get_xmin(true);
         tmp_tmax[0] = ct->get_xmax(true);
         tmp_tmin[1] = ct->get_ymin(true);
         tmp_tmax[1] = ct->get_ymax(true);

         dsCoord->convert_cs_to_cs (&tmp_tmin, &tmp_min, 1, DATA, "", DATA, image_id);
         dsCoord->convert_cs_to_cs (&tmp_tmax, &tmp_max, 1, DATA, "", DATA, image_id);
         tmp_lmin = tmp_min;

         if (init)
         {
            *min  = tmp_min[index];
            *tmin = tmp_tmin[index];
            if (tmp_lmin[index] > 0)  *lmin = tmp_lmin[index];
               else *lmin = 0;
            *max  = tmp_max[index];
            *tmax = tmp_tmax[index];
            init = false;
         }
         else
         {
            if (tmp_min[index] < *min)
            {
               *min = tmp_min[index];
               *tmin = tmp_tmin[index];
            }
            if ((tmp_lmin[index] > 0) && ((tmp_lmin[index] < *lmin) || (*lmin == 0)))
                  *lmin = tmp_lmin[index];
            if (tmp_max[index] > *max)
            {
               *max = tmp_max[index];
               *tmax = tmp_tmax[index]; 
            }
         }
      }

      for(images_map::iterator ii=images.begin();ii!=images.end();++ii)
      {
         im = (*ii).second;
         axidx = im->get_xaxis();
         axidy = im->get_yaxis();

         // ignore objects not attached to requested axis
         if ((axidx != axid) && (axidy != axid))
            continue;
         // if an alternate axis was provided, we only want to deal with objects
         // using the requested and alt axes (ie. in the given coord sys)
         if (altax && ((axidx == axid && axidy != axidalt) || 
             (axidy == axid && axidx != axidalt)))
            continue; 

         tmp_tmin[0] = im->get_xmin(true);
         tmp_tmax[0] = im->get_xmax(true);
         tmp_tmin[1] = im->get_ymin(true);
         tmp_tmax[1] = im->get_ymax(true);

         dsCoord->convert_cs_to_cs (&tmp_tmin, &tmp_min, 1, DATA, "", DATA, image_id);
         dsCoord->convert_cs_to_cs (&tmp_tmax, &tmp_max, 1, DATA, "", DATA, image_id);
         tmp_lmin = tmp_min;

         if (init)
         {
            *min  = tmp_min[index];
            *tmin = tmp_tmin[index];
            if (tmp_lmin[index] > 0)  *lmin = tmp_lmin[index];
               else *lmin = 0;
            *max  = tmp_max[index];
            *tmax = tmp_tmax[index];
            init = false;
         }
         else
         {
            if (tmp_min[index] < *min)
            {
               *min = tmp_min[index];
               *tmin = tmp_tmin[index];
            }
            if ((tmp_lmin[index] > 0) && ((tmp_lmin[index] < *lmin) || (*lmin == 0)))
                  *lmin = tmp_lmin[index];
            if (tmp_max[index] > *max)
            {
               *max = tmp_max[index];
               *tmax = tmp_tmax[index];
            }
         }
      }
   }

   if (!init) return  ERR_STATUS_GOOD;
   else 
   {
      cmd_manager->add_err(ERR_NO_DATA_FOR_AUTO,id,0);
      return ERR_NO_DATA_FOR_AUTO;
   }
}

void ChipsPlot::cover_box_axes(ChipsAxis *ax, bool rec)
{
   float tol = 0.001;   // tolerance factor
   int ival = 0, oval = 0;
   bool changed[] = {false, false};

   if (axis_style == chips_open)   return; 
   if (!ax || ax->get_coord_system () == DATA)  return; 

   int dim = ax->get_orientation ();
   if (dim == X_AXIS)
   {
      // check whether bottom plot box should be displayed
      if ((ax != bottom) && ax->is_visible() &&
          ((ax->get_norm_pos() >= (bottom->get_norm_pos() - tol)) &&
          (ax->get_norm_pos() <= (bottom->get_norm_pos() + tol)))) {

         bottom->get_property (static_cast<void*>(&oval), SET_AXIS_VISIBLE);
         if (oval)
         {
           changed[1] = true;
           bottom->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec);
         }
      }
      // check whether top plot box should be displayed
      if ((ax != top) && ax->is_visible() &&
          ((ax->get_norm_pos() >= (top->get_norm_pos() - tol)) &&
          (ax->get_norm_pos() <= (top->get_norm_pos() + tol))))
      {
         top->get_property (static_cast<void*>(&oval), SET_AXIS_VISIBLE);
         if (oval)
         {
           changed[1] = true;
           top->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec);
         }
      }
   }
   else
   {
      // check whether left plot box should be displayed
      if ((ax != left) && ax->is_visible() &&
          ((ax->get_norm_pos() >= (left->get_norm_pos() - tol)) &&
          (ax->get_norm_pos() <= (left->get_norm_pos() + tol))))
      {
         left->get_property (static_cast<void*>(&oval), SET_AXIS_VISIBLE);
         if (oval)
         {
           changed[1] = true;
           left->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec);
         }
      }
      // check whether right plot box should be displayed
      if ((ax != right) && ax->is_visible() &&
          ((ax->get_norm_pos() >= (right->get_norm_pos() - tol)) &&
          (ax->get_norm_pos() <= (right->get_norm_pos() + tol))))
      {
         right->get_property (static_cast<void*>(&oval), SET_AXIS_VISIBLE);
         if (oval)
         {
           changed[1] = true;
           right->set_property(static_cast<void*>(&ival), SET_AXIS_VISIBLE, rec);
         }
      }
   }
   if (changed[0] || changed[1])
      adjust_corners(rec); // recompute corner positions based on used axes
}

void ChipsPlot::cover_box_axes(int dim, bool rec)
{
   if ((dim == X_AXIS) || (dim == XY_AXIS))
      for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
	 cover_box_axes (am->second, rec);
   if ((dim == Y_AXIS) || (dim == XY_AXIS))
      for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
	 cover_box_axes (am->second, rec);
}

void ChipsPlot::remove_box_axes(ChipsAxis *bx)
{
#ifdef WLM_COMMENT_OUT_4NOW
   // no longer needed w/ bind graph?
   if (bx && bindings)
   {
      if ((bx == bottom) || (bx == top))
      {
         bindings->remove_axis(top);
         bindings->remove_axis(bottom);
         bindings->add_binding(bottom, top);
      }
      else if ((bx == left) || (bx == right))
      {
         bindings->remove_axis(left);
         bindings->remove_axis(right);
         bindings->add_binding(left, right);
      }
   }
#endif
}

int ChipsPlot::adjust_default_colors()
{
   if (title && (title->get_color() == "default")) title->change_color();

   // force update of 'default' colors 
   for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
      (*am).second->adjust_default_colors();
   for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
      (*am).second->adjust_default_colors();
   for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
   {
      if ((*cc).second->get_linecolor() == "default")
         (*cc).second->change_linecolor();
      if ((*cc).second->get_symbolcolor() == "default")  
         (*cc).second->change_symbolcolor();
      if ((*cc).second->get_errcolor() == "default")  
         (*cc).second->change_errcolor();
   }
   for(contours_map::iterator cc=contours.begin(); cc != contours.end(); ++cc)
      if ((*cc).second->get_color() == "default")
          (*cc).second->change_color();
   for(histograms_map::iterator hh=histograms.begin(); hh != histograms.end(); ++hh)
   {
      if ((*hh).second->get_linecolor() == "default")
         (*hh).second->change_linecolor();
      if ((*hh).second->get_symbolcolor() == "default")
         (*hh).second->change_symbolcolor();
      if ((*hh).second->get_errcolor() == "default")
         (*hh).second->change_errcolor();
      if ((*hh).second->get_fillcolor() == "default")
         (*hh).second->change_fillcolor();
   }
   for (labels_map::iterator lb =labels.begin(); lb != labels.end(); ++lb)
      if ((*lb).second->get_color() == "default")  (*lb).second->change_color();
   for(lines_map::iterator ln=lines.begin(); ln != lines.end(); ++ln)
      if ((*ln).second->get_color() == "default")
          (*ln).second->change_color();
   for(cbars_map::iterator ib=cbars.begin(); ib != cbars.end(); ++ib)
      (*ib).second->adjust_default_colors(); 
   for(points_map::iterator pt=points.begin(); pt != points.end(); ++pt)
      if ((*pt).second->get_color() == "default")
          (*pt).second->change_color();
   for(regions_map::iterator rr=regions.begin(); rr != regions.end(); ++rr)
   {
      if ((*rr).second->get_color() == "default")
          (*rr).second->change_color();
      if ((*rr).second->get_fillcolor() == "default")
          (*rr).second->change_fillcolor();
   }
   return ERR_STATUS_GOOD;
}

int ChipsPlot::update_print_scale (double scale)
{
   int rval = ERR_STATUS_GOOD;

   // currently only images get scales updated for printing
   map<string, ChipsPlot* >::iterator pp;
   for (images_map::iterator ii=images.begin(); ii != images.end(); ++ii)
      (*ii).second->update_print_scale (scale);
   return rval;
}

int ChipsPlot::redraw()
{
   return ERR_FEATURE_DNE;
}

int ChipsPlot::change_axis_style()
{
   return ERR_FEATURE_DNE; 
}

int ChipsPlot::change_visible ()
{
   return ERR_FEATURE_DNE;
}

int ChipsPlot::show_current ()
{
   return ERR_FEATURE_DNE;
}

int ChipsPlot::hide_current ()
{
   return ERR_FEATURE_DNE;
}

// TEMPORARY- since contours currently only support regularly gridded data
int ChipsPlot::prescreen_contour(unsigned long num_pts, double* xx, 
   double*yy, double* zz)
{
   bool valid = false;
   unsigned int rr = 1;
   double x_spacing, y_spacing, xmax, xmin, ymax, ymin;
   int    x_gridsize, y_gridsize;

   if (!xx || !yy || !zz || (num_pts == 0))
      return ERR_STATUS_BAD;

   // check for nans 
   for (unsigned long rr = 0; rr < num_pts; rr++)
   {
      if (!finite(xx[rr]) || !finite(yy[rr]) || !finite(zz[rr]))
      {
         cmd_manager->add_err(ERR_CONTOUR_NAN,id,0);
         return ERR_CONTOUR_NAN; 
      } 
   } 

   // check if data forms a regular grid - N x M array where all N axis entries
   // are equi-distant and all M axis entries are equi-distant and no missing
   // points in the already sorted data array.
   xmin = xx[0];
   xmax = xx[num_pts -1];
   ymin = yy[0];
   ymax = yy[num_pts -1];

   // need to handle either row or column precedence...
   if ((x_spacing = xx[1] - xx[0]) > 0)
   {
      x_gridsize = (int)((xmax - xmin) / x_spacing + 1.5);

      if ((static_cast <unsigned int>(x_gridsize) < num_pts) &&
          ((y_spacing = yy[x_gridsize] - yy[0]) > 0))
      {
         y_gridsize = (int)((ymax - ymin) / y_spacing + 1.5);
         // make sure we have the right number of points
         if ((static_cast <unsigned int> (x_gridsize * y_gridsize)) == num_pts)
         {
            int rw = 0;
            rr = 0;
            valid = true;
            double dist = yy[0];
            // verify y positions are regular
            while (valid && (rr < num_pts))
            {
               valid = approx_equal(yy[rr], dist, .001);
               if (rr && (((rr+1)%x_gridsize) == 0) && ((rr+1) < num_pts))
               {
                  rw++;
                  dist = yy[rr+1] - yy[rr];
                  if (approx_equal(dist, y_spacing, .001))
                     dist = yy[rr+1];
                  else
                     dist = yy[rr] + y_spacing;
               }
               rr++;
            }
            rr = 1;
            dist = xx[0];
            // verify x positions are regular
            while (valid && (rr < num_pts))
            {
               valid =  approx_equal((xx[rr] - dist), x_spacing, .001);
               if (((rr+1) % x_gridsize) == 0) dist = xx[0] - x_spacing;
                  else dist = xx[rr];
               rr++;
            }
         }
      }
   }
   else if ((y_spacing = yy[1] - yy[0]) > 0)
   {
      y_gridsize = (int)((ymax - ymin) / y_spacing + 1.5);

      if ((static_cast <unsigned int> (y_gridsize) < num_pts) &&
          ((x_spacing = xx[y_gridsize] - xx[0]) > 0))
      {
         x_gridsize = (int)((xmax - xmin) / x_spacing + 1.5);

         // make sure we have the right number of points
         if ((static_cast <unsigned int> (y_gridsize * x_gridsize)) == num_pts)
         {
            int rw = 0;
            rr = 0;
            valid = true;
            double dist = xx[0];
            // verify x positions are regular
            while (valid && (rr < num_pts))
            {
               valid = approx_equal(xx[rr], dist, .001);
               // adjust expected distance value if new row
               if (rr && ((rr+1)%y_gridsize == 0) && ((rr+1) < num_pts))
               {
                  rw++;
                  dist = xx[rr+1] - xx[rr];
                  if (approx_equal(dist, x_spacing, .001))
                     dist = xx[rr+1];
                  else
                     dist = xx[rr] + x_spacing;
               }
               rr++;
            }
            rr = 1;
            dist = yy[0];
            // verify y positions are regular
            while (valid && (rr < num_pts))
            {
               valid = approx_equal((yy[rr] - dist), y_spacing, .001);
               if (((rr+1) % y_gridsize) == 0) dist = yy[0] - y_spacing;
                  else dist = yy[rr];
               rr++;
            }
         }
      }
   }
   if (!valid) 
   {
      cmd_manager->add_err(ERR_CNTR_REG_GRID,id,0);
      return ERR_CNTR_REG_GRID;
   }
   else   return ERR_STATUS_GOOD;
}

void ChipsPlot::serialize(ChipsStateStream& out)
{
   this->serialize (out, false);
}

void ChipsPlot::serialize(ChipsStateStream& out, bool write_coord)
{
   int sz;
   out.write_string(&id);
   sz = (int) ref_cnt;
   out.write_int(&sz);
   sz = (int) coord_system;
   out.write_int(&sz);
   out.write_string(&data_system_id);
   out.write_float(&left_margin);
   out.write_float(&right_margin);
   out.write_float(&top_margin);
   out.write_float(&bottom_margin);
   if (out.get_minor_version() > 0)
   {
     out.write_float(&bx1); 
     out.write_float(&by1); 
     out.write_float(&bx2); 
     out.write_float(&by2); 
     out.write_float(&aspect_h); 
     out.write_float(&aspect_w); 
     out.write_bool(&aspect_use); 
     out.write_string(&aspect_ratio); 
   }
   if (out.get_minor_version() > 13)
   {
      out.write_float(&title_xpos);
      out.write_float(&title_ypos);
   }
   sz = (int) corner_style; 
   out.write_int(&sz);
   out.write_int(&curve_cnt);
   out.write_int(&contour_cnt);
   out.write_int(&histogram_cnt);
   if (out.get_minor_version() > 0)  
      out.write_int(&image_cnt); 
   out.write_int(&label_cnt); 
   out.write_int(&line_cnt); 
   out.write_int(&point_cnt); 
   out.write_int(&region_cnt); 

   out.write_int(&axis_x_cnt);
   out.write_int(&axis_y_cnt);
   sz = (int) axis_style; 
   out.write_int(&sz);
   out.write_bool (&visible);

   // Write out the coordinate system first so it can be added back
   // during deserialize
   out.write_bool (&write_coord);
   out.write_bool (&offscreen);

   // write title info
   sz = (title != NULL);
   out.write_int(&sz);

   if (out.get_minor_version() > 5)
   {
      string tmp_id;
      sz = deleted_data_systems.size ();
      out.write_int (&sz);
      map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
      for (ds_iter=deleted_data_systems.begin (); ds_iter!=deleted_data_systems.end (); ds_iter++)
      {
         tmp_id = ds_iter->first.first;
         out.write_string (&tmp_id);
         tmp_id = ds_iter->first.second;
         out.write_string (&tmp_id);
         ds_iter->second->serialize (out);
      }
   }


   if (write_coord)
      coord->serialize (out);
   if (title)
      title->serialize(out);
   bottom->serialize(out);
   top->serialize(out);
   left->serialize(out);
   right->serialize(out);
   serialize_axes(out);
   serialize_objects(out);
}

void ChipsPlot::serialize_ascii(ChipsStateStream& out, int keep)
{
   char buf[256];
   string tmp = "", pos = "", att = "";
   string tmpid = theDefaults->get_plotstem();

   if (left_margin != theDefaults->get_plotleftmargin() ||
       right_margin != theDefaults->get_plotrightmargin() ||
       top_margin != theDefaults->get_plottopmargin() ||
       bottom_margin != theDefaults->get_plotbottommargin())
   {
      sprintf(buf,"%3.2f,%3.2f,%3.2f,%3.2f,", left_margin, bottom_margin,
      1 - right_margin, 1 - top_margin);
      pos = buf;
   }
   if (theDefaults->get_cornerstyle() != corner_style)
   {
      att = "corner.style=";
      att.append(get_corner_style());
   }
   if (theDefaults->get_axisstyle() != axis_style)
   {
      if (att.length()) att.append(" ");
      att = "style=";
      att.append(v_axisstyle_to_string(axis_style));
   }

   if (id.find(tmpid) != 0) 
   {
      if (att.length())   att.append(" ");
      att.append("id=");
      att.append(id);
   }

   if (title) tmp = title->get_text();
   if (tmp.length())
   {
      // handle title and title related attributes
      tmp = title->get_color();
      if (tmp != theDefaults->get_plottitlecolor())
      {
         att.append(" title.color=");
         att.append(tmp);
      }
      int itmp = title->get_size();
      if (itmp != theDefaults->get_plottitlesize())
      {
         sprintf(buf," title.size=%d", itmp);
         att.append(buf);
      }
      itmp = title->get_depth();
      int dep = theDefaults->get_labeldepth();
      if (itmp != dep &&
          (dep != V_USE_DEFAULT_DEPTH || itmp != get_depth()))
      {
         sprintf(buf, " title.depth=%d", itmp);
         att.append(buf);
      }
      float ftmp = get_title_xpos();
      if (ftmp != theDefaults->get_plottitlexpos())
      {
         sprintf(buf, " title.xpos=%2.1f", ftmp);
         att.append(buf);
      }
      ftmp = get_title_ypos();
      if (ftmp != theDefaults->get_plottitleypos())
      {
         sprintf(buf, " title.ypos=%2.1f", ftmp);
         att.append(buf);
      }
      ftmp = title->get_angle();
      if (ftmp != theDefaults->get_plottitleangle())
      {
         sprintf(buf, " title.angle=%2.1f", ftmp);
         att.append(buf);
      }
      ftmp = title->get_horizalignment();
      if (ftmp != theDefaults->get_plottitlehalign())
      {
         sprintf(buf, " title.halign=%2.1f", ftmp);
         att.append(buf);
      }
      ftmp = title->get_vertalignment();
      if (ftmp != theDefaults->get_plottitlevalign())
      {
         sprintf(buf, " title.valign=%2.1f", ftmp);
         att.append(buf);
      }
      tmp = title->get_font();
      if (tmp != theDefaults->get_plottitlefont())
      {
         att.append(" title.font=");
         att.append(tmp);
      }
      V_FONTSTYLE_TYPE vftmp = title->get_fontstyle();
      const char *cp = v_fontstyle_to_string(vftmp);
      if (vftmp != theDefaults->get_plottitlefontstyle() && cp)
      {
         att.append(" title.fontstyle=");
         att.append(cp);
      }
      if (att.length())   att.append(" ");
      att.append("title.label=\\\"");
      att.append(title->get_text());
      att.append("\\\"");
   }
   
   if (att == "" && pos == "")  out << "add_plot()" << endl;
   else   out << "add_plot(" << pos << "\"" << att <<"\")"<<endl;
   serialize_objects_ascii(out);
}


void ChipsPlot::serialize_axes (ChipsStateStream& out)
{
   int sz;
   ChipsAxis* aa; 
   sz = axes_x.size() -2;   // omit borders
   out.write_int(&sz);
   out.write_int(&sz);
   map <int, string>chron;
   map <int, string>::iterator i2;
   map <unsigned int, ChipsAxis *>::iterator jj;
   for (axes_map::iterator ii = axes_x.begin(); ii != axes_x.end(); ii++)
   {
      aa = (*ii).second;
      if (!((aa == top) || (aa == bottom)))
        chron.insert(pair<int, string>(aa->get_ref_cnt(), aa->get_id()));
   }
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      (axes_x[(*i2).second])->serialize(out);

   sz = daxes_x.size();
   out.write_int(&sz);

   for (map <unsigned int, ChipsAxis *>::iterator ii = daxes_x.begin(); ii != daxes_x.end(); ii++)
      ii->second->serialize (out);

   
   out.write_checkpoint();
   sz = axes_y.size() -2;   // omit borders
   out.write_int(&sz);
   out.write_int(&sz);
   chron.clear();
   for (axes_map::iterator ii = axes_y.begin(); ii != axes_y.end(); ii++)
   {
      aa = (*ii).second;
      if (!((aa == left) || (aa == right)))
        chron.insert(pair<int, string>(aa->get_ref_cnt(), aa->get_id()));
   }
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      (axes_y[(*i2).second])->serialize(out);

   sz = daxes_y.size();
   out.write_int(&sz);

   for (map <unsigned int, ChipsAxis *>::iterator ii = daxes_y.begin(); ii != daxes_y.end(); ii++)
      ii->second->serialize (out);
   out.write_checkpoint();
}

void ChipsPlot::serialize_objects (ChipsStateStream& out)
{

   int sz;
   map <int, string>chron;
   map <int, string>::iterator i2;

   // write out images
   if ((sz = images.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      for(images_map::iterator cc = images.begin(); cc != images.end(); ++cc)
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), (*cc).first));
      for(i2 = chron.begin(); i2 != chron.end(); i2++)
         (images[(*i2).second])->serialize(out);
   }
   else
   {
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }
   sz = dimages.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsImage *>::iterator ii = dimages.begin(); ii != dimages.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currimage);
   out.write_checkpoint();

   // write out curves
   if ((sz = curves.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), (*cc).first));
      for(i2 = chron.begin(); i2 != chron.end(); i2++)
         (curves[(*i2).second])->serialize(out);
   }
   else
   {
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dcurves.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsCurve *>::iterator ii = dcurves.begin(); ii != dcurves.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currcurv);
   out.write_checkpoint();

   // write out contours
   if ((sz = contours.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      for(contours_map::iterator cc=contours.begin(); cc!=contours.end(); ++cc)
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), (*cc).first));
      for(i2 = chron.begin(); i2 != chron.end(); i2++)
         (contours[(*i2).second])->serialize(out);
   }
   else
   {
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }
   sz = dcontours.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsContour *>::iterator ii = dcontours.begin(); ii != dcontours.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currcont);
   out.write_checkpoint();

   // write out histograms
   if ((sz = histograms.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      for(histograms_map::iterator cc=histograms.begin(); cc!=histograms.end(); ++cc)
         chron.insert(pair<int, string>((*cc).second->get_ref_cnt(), (*cc).first));
      for(i2 = chron.begin(); i2 != chron.end(); i2++)
         (histograms[(*i2).second])->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dhistograms.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsHistogram *>::iterator ii = dhistograms.begin(); ii != dhistograms.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currhist);
   out.write_checkpoint();

   // write out labels
   if ((sz = labels.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      map<string, ChipsLabel*>::iterator li;
         for(li = labels.begin(); li != labels.end(); ++li)
	       (*li).second->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dlabels.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsLabel *>::iterator ii = dlabels.begin(); ii != dlabels.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currlabel);
   out.write_checkpoint();

   // write out lines
   if ((sz = lines.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      map<string, ChipsLine*>::iterator li;
         for(li = lines.begin(); li != lines.end(); ++li)
	       (*li).second->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dlines.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsLine*>::iterator ii = dlines.begin(); ii != dlines.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currline);
   out.write_checkpoint();

   // write out cbars
   if ((sz = cbars.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      map<string, ChipsColorBar*>::iterator li;
         for(li = cbars.begin(); li != cbars.end(); ++li)
	       (*li).second->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dcbars.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsColorBar*>::iterator ii = dcbars.begin(); ii != dcbars.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currcbar);
   out.write_checkpoint();

   // write out points
   if ((sz = points.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      map<string, ChipsPoint*>::iterator li;
         for(li = points.begin(); li != points.end(); ++li)
	       (*li).second->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dpoints.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsPoint*>::iterator ii = dpoints.begin(); ii != dpoints.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currpoint);
   out.write_checkpoint();

   // write out regions
   if ((sz = regions.size()) != 0)
   {
      out.write_int(&sz);
      out.write_int(&sz);
      chron.clear();
      map<string, ChipsRegion*>::iterator ri;
         for(ri = regions.begin(); ri != regions.end(); ++ri)
	       (*ri).second->serialize(out);
   }
   else{
      sz = 0;
      out.write_int(&sz);
      out.write_int(&sz);
   }

   sz = dregions.size();
   out.write_int(&sz);
   for (map <unsigned int, ChipsRegion*>::iterator ii = dregions.begin(); ii != dregions.end(); ii++)
      ii->second->serialize (out);

   out.write_string(&currregion);
   out.write_checkpoint();
}

void ChipsPlot::serialize_objects_ascii(ChipsStateStream& out)
{
   map <int, ChipsObject*>chron;
   map <int, ChipsObject*>pending_chron;
   ChipsAxis* aa;
   for (axes_map::iterator ii = axes_x.begin(); ii != axes_x.end(); ii++)
   {
      aa = (*ii).second;
      if (!((aa == top) || (aa == bottom)))
        chron.insert(pair<int, ChipsObject*>(aa->get_ref_cnt(), aa));
   }
   for (axes_map::iterator ii = axes_y.begin(); ii != axes_y.end(); ii++)
   {
      aa = (*ii).second;
      if (!((aa == left) || (aa == right)))
        chron.insert(pair<int, ChipsObject*>(aa->get_ref_cnt(), aa));
   }

   for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
      chron.insert(pair<int, ChipsObject*>((*cc).second->get_ref_cnt(), (*cc).second));
   for(histograms_map::iterator hi = histograms.begin(); hi != histograms.end(); ++hi)
      chron.insert(pair<int, ChipsObject*>((*hi).second->get_ref_cnt(), (*hi).second));
   for(contours_map::iterator ci = contours.begin(); ci != contours.end(); ++ci)
      chron.insert(pair<int, ChipsObject*>((*ci).second->get_ref_cnt(), (*ci).second));
   for(images_map::iterator im = images.begin(); im != images.end(); ++im)
      chron.insert(pair<int, ChipsObject*>((*im).second->get_ref_cnt(), (*im).second));
   for(map <string, ChipsLine*>::iterator ii=lines.begin(); ii!=lines.end(); ++ii)
      chron.insert(pair<int, ChipsObject*>((*ii).second->get_ref_cnt(), (*ii).second));
   for(map <string, ChipsLabel*>::iterator lb=labels.begin(); lb!=labels.end(); ++lb)
      chron.insert(pair<int, ChipsObject*>((*lb).second->get_ref_cnt(), (*lb).second));
   for(map <string, ChipsPoint*>::iterator pp=points.begin(); pp!=points.end(); ++pp)
      chron.insert(pair<int, ChipsObject*>((*pp).second->get_ref_cnt(), (*pp).second));
   for(map <string, ChipsRegion*>::iterator rr=regions.begin(); rr!=regions.end(); ++rr)
      chron.insert(pair<int, ChipsObject*>((*rr).second->get_ref_cnt(), (*rr).second));
   for(map <string, ChipsColorBar*>::iterator cb=cbars.begin(); cb!=cbars.end(); ++cb)
      chron.insert(pair<int, ChipsObject*>((*cb).second->get_ref_cnt(), (*cb).second));

   map <int, ChipsObject*>::iterator i2;
   map <int, ChipsObject*>::iterator hold_i2;
   unsigned int bflag = 0;
   ChipsAxis *tmpaxis, *last_x = NULL, *last_y = NULL;
   string atts_x, atts_y, by1_atts, by2_atts, bx1_atts, bx2_atts;
   bool wrote_x = false, wrote_y = false, def_x = false, def_y = false, 
        bwrite = false, defaxes = false;
   int dim;
   bool xaxis_added = false;
   bool yaxis_added = false;
   bool use_pending = false;

   for(i2 = chron.begin(); i2 != chron.end() || use_pending; i2++)
   {
      if (use_pending)
      {
         i2 = pending_chron.begin ();
         use_pending = false;
      }
      else if (i2==pending_chron.end ())
      {
         i2 = hold_i2;
         pending_chron.clear ();
         continue;
      }

      if ((*i2).second->get_object_type() == chips_axis)
      {              
         tmpaxis = (ChipsAxis*)(*i2).second;
         dim = tmpaxis->get_orientation();
         if (dim == X_AXIS)
         {
            xaxis_added = true;
            if (last_x && last_y && !wrote_x && !wrote_y)
            {
               // have unwritten x and y- got new
               ChipsAxis *ax1 = last_x, *ax2 = last_y;  
               combine_axes_ascii_serialize(out, ax1, ax2);
               wrote_y = wrote_x = def_x = def_y = true;

            } 
            else if (last_x && !wrote_x)
            {
               // already have x axis- need to write it out 
               last_x->serialize_ascii(out);
               def_x = true;
            }
            last_x = tmpaxis; 
            wrote_x = false;  
            if (!def_x)  atts_x = last_x->build_set_axis_string();
         }
         else 
         {
            yaxis_added = true;
            if (last_x && last_y && !wrote_x && !wrote_y)
            {
               // have unwritten x and y- got new y
               ChipsAxis *ax1 = last_x, *ax2 = last_y;
               combine_axes_ascii_serialize(out, ax1, ax2);
               wrote_x = def_x = def_y = true;
            }
            else if (last_y && !wrote_y)
            {
               // already have y axis- need to write it out 
               last_y->serialize_ascii(out);
               def_y = true;
            }
            last_y = tmpaxis;
            wrote_y = false; 
            if (!def_y)  atts_y = last_y->build_set_axis_string();
         }


         // we must have two axes added before plot data can be added
         // this is to work around the fact that axes are created after
         // plot data
         if (xaxis_added && yaxis_added)
         {
            if (pending_chron.size () > 0)
               use_pending = true;
         }
      }
      else if ((*i2).second->get_object_type() == chips_curve ||
               (*i2).second->get_object_type() == chips_histogram ||
               (*i2).second->get_object_type() == chips_image ||
               (*i2).second->get_object_type() == chips_contour)
      {
         // we must have two axes added before plot data can be added
         // this is to work around the fact that axes are created after
         // plot data
         if (!(xaxis_added && yaxis_added))
         {
            pending_chron.insert(pair<const int, ChipsObject*>(*i2));
            continue;
         }


         // first handle any unwritten axes 
         if (last_x && last_y && !wrote_x && !wrote_y)
         {
            if (def_x || def_y || last_x->get_orig_pos() != 0 || 
                last_y->get_orig_pos() != 0)
            {
               // have unwritten axis pair but not defaults 
               ChipsAxis *ax1 = last_x, *ax2 = last_y;
               combine_axes_ascii_serialize(out, ax1, ax2);
            }
            else 
               defaxes = true;
            wrote_x = wrote_y = def_x = def_y = true;
         }
         else if (last_x && !wrote_x)
         {
            // have an unwritten x axis 
            last_x->serialize_ascii(out);
            def_x = wrote_x = true; 
         }
         else if (last_y && !wrote_y)
         {
            // have an unwritten x axis
            last_y->serialize_ascii(out);
            def_y = wrote_y = true;
         }
         if (!bwrite)
         {
            // if the border axes properties differ from the axes properties
            // we need to write out set_axis() calls for the borders
            by1_atts = left->build_set_axis_string(),
            by2_atts = right->build_set_axis_string(),
            bx1_atts = bottom->build_set_axis_string(),
            bx2_atts = top->build_set_axis_string();
            bwrite = true;
            // only write out border attributes if they contain elements not
            // in their corresponding bound axes
            if (check_for_matching_attribute_list(by1_atts, atts_y))
               bflag |= 0x01;
            if (check_for_matching_attribute_list(by2_atts, atts_y))
               bflag |= 0x02;
            if (check_for_matching_attribute_list(bx1_atts, atts_x))
               bflag |= 0x04;
            if (check_for_matching_attribute_list(bx2_atts, atts_x))
               bflag |= 0x08;
            if ((bflag & 0x0f) == 0x0f)
            {
               int s1 = by1_atts.find(',');
               int e1 = by1_atts.length() -s1,
                   e2 = by2_atts.length() -s1,
                   e3 = bx1_atts.length() -s1,
                   e4 = bx2_atts.length() -s1;
               if (by1_atts.substr(s1, e1) == by2_atts.substr(s1, e2) &&
                   bx1_atts.substr(s1, e3) == bx2_atts.substr(s1, e4) &&
                   bx1_atts.substr(s1, e3) == by1_atts.substr(s1, e1))
               {
                  // all axes are the same- write them as a single entry
                  out << "# set the attriutes of all border axes" << endl;
                  out << "set_axis(\"all\""<<by1_atts.substr(s1, e1)<<endl;
                  bflag = 0;
               }
            }
         }
         // change axes to get the correct data system if necessary
         if ((*i2).second->get_coord_system()  == DATA)
         {
            string txid, tyid;
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>((*i2).second->get_chips_coord());
            if (ds_coord) {
               txid = ds_coord->get_xaxis_id(); 
               tyid = ds_coord->get_yaxis_id(); 
               if (last_x && last_x->get_id() != txid) {
                  out << "current_axis(\"" << txid<<"\")"<<endl;
                  last_x = get_xaxis (txid);  
               } 
               if (last_y && last_y->get_id() != tyid) {
                  out << "current_axis(\"" << tyid<<"\")"<<endl;
                  last_y = get_yaxis (tyid);  
               } 
            }
         } 
         // write out the object
         (*i2).second->serialize_ascii(out);
      }
      else
      {
         // general objects don't add axes so we must explicitly add them
         if (last_x && last_y && !wrote_x && !wrote_y)
         {
            ChipsAxis *ax1 = last_x, *ax2 = last_y;  
            combine_axes_ascii_serialize(out, ax1, ax2);
            if (last_x)    def_x = wrote_x = true;
            if (last_y)    def_y = wrote_y = true;
         }
         else if (last_x && !wrote_x)
         {
            // have an unwritten x axis
            last_x->serialize_ascii(out);
            def_x = wrote_x = true;
         }
         else if (last_y && !wrote_y)
         {
            // have an unwritten y axis
            last_y->serialize_ascii(out);
            def_y = wrote_y = true;
         }
         // change axes to get the correct data system if necessary
         if ((*i2).second->get_coord_system()  == DATA)
         {
            string txid, tyid;
            ChipsDataSystem *ds_coord = dynamic_cast <ChipsDataSystem *>((*i2).second->get_chips_coord());
            if (ds_coord) {
               txid = ds_coord->get_xaxis_id(); 
               tyid = ds_coord->get_yaxis_id(); 
               if (last_x && last_x->get_id() != txid) {
                  out << "current_axis(\"" << txid<<"\")"<<endl;
                  last_x = get_xaxis (txid);  
               } 
               if (last_y && last_y->get_id() != tyid) {
                  out << "current_axis(\"" << tyid<<"\")"<<endl;
                  last_y = get_yaxis (tyid);  
               } 
            }
         } 
         (*i2).second->serialize_ascii(out);
      }

      // if we're going to use pending list hold off the current iterator
      if (use_pending)
         hold_i2 = i2;
   }

   // if we've gone through all objects but haven't added our axes, do so
   if (last_x && last_y && !wrote_x && !wrote_y) 
   {
      ChipsAxis *ax1 = last_x, *ax2 = last_y;
      combine_axes_ascii_serialize(out, ax1, ax2);
      wrote_x = wrote_y = def_x = def_y = true;
   }
   else if (last_x && !wrote_x) 
   {
      last_x->serialize_ascii(out);
      wrote_x = def_x = true; 
   }
   else if (last_y && !wrote_y) 
   {
      last_y->serialize_ascii(out);
      wrote_y = def_y = true; 
   }

   // set plot aspect ratio
   if (aspect_ratio.length())
      out << "set_plot_aspect_ratio('"<<aspect_ratio<<"')"<<endl;

   // explicitly set limits
   double xmn, xmx, ymn, ymx;
   int aminx=0, aminy=0, amaxx=0, amaxy=0, bndx=0, bndy=0;
   int delobj = dcurves.size() + dcontours.size() + dimages.size() +
       dhistograms.size();
   ChipsDataSystem* cds = get_datasystem(last_x, last_y);
   int objcnt = 0;
   if (cds) objcnt = cds->get_num_objs(chips_curve) + 
      cds->get_num_objs(chips_contour) + cds->get_num_objs(chips_image) +
      cds->get_num_objs(chips_histogram);

   if (last_x)
   {
      last_x->get_range(xmn, xmx);
      amaxx = last_x->get_automax_used();
      aminx = last_x->get_automin_used();
   }
   if (last_y)
   {
      last_y->get_range(ymn, ymx);
      amaxy = last_y->get_automax_used();
      aminy = last_y->get_automin_used();
   }
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator iter;
      ChipsDataSystem *ds = NULL;

      for (iter=data_systems.begin (); iter!=data_systems.end (); iter++)  
      {
         ds = iter->second;
         string tmp = ds->get_aspect_ratio();
         if (tmp.length())
         {
            out << "# current_axis(\"" << iter->first.first<<"\")"<<endl;
            out << "# current_axis(\"" << iter->first.second<<"\")"<<endl;
            out << "# set_data_aspect_ratio(\""<<tmp<<"\")"<<endl;
         }
      }
   }

   // check bindings- if the axis is bound (either a true bind or a data
   // aspect bind, we should wait til after the bind table is written out
   // to set our limits
   if (bindings)
   {
      list <ChipsAxis*>::iterator ai;
      string aid;
      if (last_x)
      {
         list <ChipsAxis*> axes = bindings->get_connections(last_x);
         for (ai = axes.begin(); ai != axes.end(); ai++)
         {
            aid = (*ai)->get_id();
            if (aid != "bx1" && aid != "bx2" && aid != "by1" && aid != "by2" &&
                (*ai) != last_x)
            {
               bndx = 1;
               if ((*ai)->get_orientation() != X_AXIS) bndy=1;
            }
         }
      }
      if (last_y)
      {
         list <ChipsAxis*> axes = bindings->get_connections(last_y);
         for (ai = axes.begin(); ai != axes.end(); ai++)
         {
            if (aid != "bx1" && aid != "bx2" && aid != "by1" && aid != "by2" &&
                (*ai) != last_y)
            {
               bndy = 1;
               if ((*ai)->get_orientation() != Y_AXIS) bndy=1;
            }
         }
      }
   }


   if (last_x && last_y && aminx && aminy && amaxx && amaxy && !delobj && !bndx && !bndy)
   {
      // if we only added one object and autolimits were used- we can skip the
      // limits call, otherwise set them to AUTO- consider only the current
      // data system 
      if (objcnt > 1)
      {
         out << "limits(XY_AXIS, AUTO, AUTO)"<<endl;
         last_x->set_wrote_limits(1);
         last_y->set_wrote_limits(1);
      }
   }
   else if (last_x && last_y &&
       xmn >= ymn - FLT_EPSILON  && xmn <= ymn + FLT_EPSILON &&
       xmx >= ymx - FLT_EPSILON && xmx <= ymx + FLT_EPSILON &&!bndx && !bndy)
   {
      out << "limits("<<xmn<<", "<<xmx<<"," <<xmn<<", "<<xmx<<")"<<endl;
      last_x->set_wrote_limits(1);
      last_y->set_wrote_limits(1);
   }
   else if (bndx && bndy && !aminx && !aminy && !amaxx && !amaxy && last_x && last_y)
   {
      // data aspect- set x & y together to avoid centering issues
      out << "limits("<<xmn<<", "<<xmx<<", "<<ymn<<", "<<ymx<<")"<<endl;
      last_x->set_wrote_limits(1);
      last_y->set_wrote_limits(1);
   }
   else
   {
      if (last_x && !delobj && objcnt && !bndx)
      {
         if (aminx&&amaxx)  out << "limits(X_AXIS, AUTO, AUTO)" << endl;
         else if (aminx)    out << "limits(X_AXIS, AUTO, "<<xmx<<")"<<endl;
         else if (amaxx)    out << "limits(X_AXIS, "<<xmn<<", AUTO)"<<endl;
         else               out << "limits(X_AXIS,  "<<xmn<<", "<<xmx<<")"<<endl;
         last_x->set_wrote_limits(1);
      }
      else if (last_x && !bndx)
      {
         out << "limits(X_AXIS, "<<xmn<<", "<<xmx<<")"<<endl;
         last_x->set_wrote_limits(1);
      }
      if (last_y && !delobj && objcnt && !bndy)
      {
         if (aminy&&amaxy)  out << "limits(Y_AXIS, AUTO, AUTO)" << endl;
         else if (aminy)    out << "limits(Y_AXIS, AUTO, "<<ymx<<")"<<endl;
         else if (amaxy)    out << "limits(Y_AXIS, "<<ymn<<", AUTO)"<<endl;
         else               out << "limits(Y_AXIS, "<<ymn<<", "<<ymx<<")"<<endl;
         last_y->set_wrote_limits(1);
      }
      else if (last_y && !bndy)
      {
         out << "limits(Y_AXIS, "<<ymn<<", "<<ymx<<")"<<endl;
         last_y->set_wrote_limits(1);
      }
   } 

   // if we didn't update the borders via an 'all' above, do it now 
   if ((bflag & 0x01) == 0x01) out << by1_atts << endl;
   if ((bflag & 0x02) == 0x02) out << by2_atts << endl;
   if ((bflag & 0x04) == 0x04) out << bx1_atts << endl;
   if ((bflag & 0x08) == 0x08) out << bx2_atts << endl;
 
   // if axis was automatically created- set attribute values
   if (defaxes && atts_x.length())   out << atts_x << endl;
   if (defaxes && atts_y.length())   out << atts_y << endl;

   chron.clear();
}

void ChipsPlot::deserialize(ChipsStateStream& in)
{
   int len;
   string cname;
   double tmpd;
   in.read_string(&id);
   in.read_int(&len);
   ref_cnt = len;
   in.read_int(&len);
   coord_system = (COORD_SYS)len;
   in.read_string(&data_system_id);

   if (in.get_major_version() < 3 || 
      (in.get_major_version() == 3 && in.get_minor_version() < 16))  
   {
      in.read_double(&tmpd);
      left_margin = tmpd;
      in.read_double(&tmpd);
      right_margin = tmpd;
      in.read_double(&tmpd);
      top_margin = tmpd;
      in.read_double(&tmpd);
      bottom_margin = tmpd;
   }
   else
   {
      in.read_float(&left_margin);
      in.read_float(&right_margin);
      in.read_float(&top_margin);
      in.read_float(&bottom_margin);
   }
   if (in.get_minor_version() > 0)
   {
     in.read_float(&bx1); 
     in.read_float(&by1); 
     in.read_float(&bx2); 
     in.read_float(&by2); 
     in.read_float(&aspect_h); 
     in.read_float(&aspect_w); 
     in.read_bool(&aspect_use); 
     in.read_string(&aspect_ratio); 
   }
   if (in.get_minor_version() > 13)
   {
      in.read_float(&title_xpos);
      in.read_float(&title_ypos);
   }
   in.read_int(&len);
   corner_style = (V_CORNERSTYLE_TYPE)len; 
   in.read_int(&curve_cnt);
   in.read_int(&contour_cnt);
   in.read_int(&histogram_cnt);
   if (in.get_minor_version() > 0)  
      in.read_int(&image_cnt);
   in.read_int(&label_cnt);
   in.read_int(&line_cnt);
   in.read_int(&point_cnt);
   in.read_int(&region_cnt);
   in.read_int(&axis_x_cnt);
   in.read_int(&axis_y_cnt);
   in.read_int(&len);
   axis_style = (V_AXISSTYLE_TYPE)len;
   in.read_bool (&visible);
   in.read_bool (&deserialize_coord);
   in.read_bool (&offscreen);
   in.read_int(&len);
   deserialize_title = len;  // deserialize title in derived Plot constuctor?

   if (in.get_minor_version() > 5)
   {
      string tmp_id;
      in.read_int (&len);
      pair <string, string> p;
      ChipsDataSystem *ds;
      for (int ii=0; ii<len; ii++)
      {
         in.read_string (&tmp_id);
         p.first = tmp_id;
         in.read_string (&tmp_id);
         p.second = tmp_id;
         ds = new ChipsDataSystem (in, NULL);
         deleted_data_systems[p] = ds;
      }
   }

}

int ChipsPlot::deserialize_axis(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   int asz, asize, rr;
   ChipsAxis* ax;
   ChipsAxis* ay; 

   // handle x axes
   in.read_int(&asz);
   in.read_int(&asize);
   if (asz != asize)   
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < asize; rr++)
   {
      ax = dynamic_cast <ChipsAxis *> (deserialize_object (chips_axis, in));
      if (ax)
      {
         add_coordinate_system (ax, NULL, NULL, false);

         // add a coordinate system for every y axis. this should be
         // none since the yaxes are added after the xaxes. but incase
         // things get moved around, i leave this in as its cost is minimal
         for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
            if (((*am).second != left) && ((*am).second != right))
               add_coordinate_system (ax, (*am).second, NULL, false);

         ax->set_axis_num (rr);
         cur_x = ax;
         cur_xall = false;
         axes_x.insert(pair<string, ChipsAxis*>(ax->get_id(), ax));
      }

   }
   in.read_int(&asize);
   for (rr = 0; rr < asize; rr++)
   {
      ax = dynamic_cast <ChipsAxis *> (deserialize_object (chips_axis, in));
      if (ax)
      {
         bool locked = cmd_manager->cmd_lock ();
         ax->force_hide ();
         ax->reset_position ();
         ax->set_axis_num (rr);
         cmd_manager->set_cmd_lock (true);
         rval = this->dereference_axis (ax);
         cmd_manager->set_cmd_lock (locked);
      }
   }


   // checkpoint to avoid issues with reading corrupt data
   if (in.read_checkpoint() != ERR_STATUS_GOOD)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   // load in y axes
   in.read_int(&asz);
   in.read_int(&asize);
   if (asz != asize)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < asize; rr++)
   {
      ax = dynamic_cast <ChipsAxis *> (deserialize_object (chips_axis, in));
      if (ax)
      {
         // add a coordinate system for each x axis
         add_coordinate_system (NULL, ax, NULL, false);
         for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
            if (((*am).second != bottom) && ((*am).second != top))
               add_coordinate_system ((*am).second, ax, NULL, false);

         ax->set_axis_num (rr);
         cur_y = ax;
         cur_yall = false;
         axes_y.insert(pair<string, ChipsAxis*>(ax->get_id(), ax));
      }
   }

   // if this is an older state file go through and adjust the axes so
   // that they have the correct algorithm
   if (in.get_minor_version() < 7)
   {
      map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter = data_systems.begin ();
      ChipsDataSystem *ds;
      Transform *transform;
      double abs_min, abs_max;
      for (; ds_iter != data_systems.end (); ds_iter++)
      {
         ds = ds_iter->second;
         ax = get_xaxis (ds_iter->first.first);
         ay = get_yaxis (ds_iter->first.second);
          
         transform = ds->get_transform (ds->get_default_transform_id ());
         if (transform)
         {
            abs_min = ax->get_absolute_min ();
            abs_max = ax->get_absolute_max ();
            ax->set_algorithm_type (transform);
            ax->adjust_minmax (abs_min, abs_max, AUTOLIM_DEF, false, true);

            abs_min = ay->get_absolute_min ();
            abs_max = ay->get_absolute_max ();
            ay->set_algorithm_type (transform);
            ay->adjust_minmax (abs_min, abs_max, AUTOLIM_DEF, false, true);

            // make sure the axes get updated
            list<string> exclude_axes;
            update_data_coord_positions (ax, ay, exclude_axes);
         }
      }
   }

   in.read_int(&asize);
   for (rr = 0; rr < asize; rr++)
   {
      ax = dynamic_cast <ChipsAxis *> (deserialize_object (chips_axis, in));
      if (ax)
      {
         bool locked = cmd_manager->cmd_lock ();
         ax->force_hide ();
         ax->set_axis_num (rr);
         cmd_manager->set_cmd_lock (true);
         rval = this->dereference_axis (ax);
         cmd_manager->set_cmd_lock (locked);
      }
   }

   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;
   list <string> exclude_list; // empty list needed for update_data_coord_pos
   for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
   {
      ax = axes_x[ds_iter->first.first];
      ay = axes_y[ds_iter->first.second];
      update_data_coord_positions (ax, ay, exclude_list);
   }

   // create coordinates for the border axes
   if (in.get_minor_version() < 8)
   {
      add_coordinate_system (top, right, NULL);
      add_coordinate_system (bottom, left, NULL);
      top->set_border_axis (true);
      bottom->set_border_axis (true);
      left->set_border_axis (true);
      right->set_border_axis (true);

      top->disable_interaction ();
      bottom->disable_interaction ();
      left->disable_interaction ();
      right->disable_interaction ();
      

   }

   ChipsId blank; 
   update_current_coord(&blank);
   // checkpoint to avoid issues with reading corrupt data
   rval = in.read_checkpoint();

   return rval; 
}

int ChipsPlot::deserialize_image(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   int isz, isize, rr;
   images_map::iterator iter;
   ChipsImage* img; 
   in.read_int(&isz);
   in.read_int(&isize);
   if (isz != isize)
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < isize; rr++)
   {
      img = dynamic_cast <ChipsImage *> (deserialize_object (chips_image, in));
      if (img)
      {
         img->set_image_num (rr);
         images.insert(images_map::value_type (img->get_id(), img));
      }
   }
   in.read_int(&isize);
   for (rr = 0; rr < isize; rr++)
   {
      img = dynamic_cast <ChipsImage *> (deserialize_object(chips_image, in));
      if (img)
      {
         img->set_image_num (rr);
         dimages[img->get_ref_cnt ()] = img;
         img->force_hide ();
         if (img->get_chips_coord ())
         {
            img->get_chips_coord ()->remove(img);
            img->set_chips_coord (NULL);
         }
      }
   }
   in.read_string(&currimage);
   rval = in.read_checkpoint(); 
   return rval; 
}

int ChipsPlot::deserialize_curve(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   int csz, csize, rr;
   curves_map::iterator iter;
   ChipsCurve* curv; 
   in.read_int(&csz);
   in.read_int(&csize);
   if (csz != csize)
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < csize; rr++)
   {
      curv = dynamic_cast <ChipsCurve *> (deserialize_object (chips_curve, in));
      if (curv)
      {
         curv->set_curve_num (rr);
         curves.insert(curves_map::value_type (curv->get_id(), curv));
      }
   }
   in.read_int(&csize);
   for (rr = 0; rr < csize; rr++)
   {
      curv = dynamic_cast <ChipsCurve *> (deserialize_object (chips_curve, in));
      if (curv)
      {
         curv->set_curve_num (rr);
         dcurves[curv->get_ref_cnt ()] = curv;
         curv->force_hide ();
         if (curv->get_chips_coord ())
         {
            curv->get_chips_coord ()->remove (curv);
            curv->set_chips_coord (NULL);
         }
      }
   }
   in.read_string(&currcurv);
   rval = in.read_checkpoint(); 
   return rval; 
}


int ChipsPlot::deserialize_contour(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   int csz, csize, rr;
   contours_map::iterator iter;
   ChipsContour* cntr; 
   in.read_int(&csz);
   in.read_int(&csize);
   if (csz != csize)
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < csize; rr++)
   {
      cntr = dynamic_cast <ChipsContour *> (deserialize_object (chips_contour, in));
      if (cntr)
      {
         cntr->set_contour_num (rr);
         contours.insert(contours_map::value_type (cntr->get_id(), cntr));
      }
   }
   in.read_int(&csize);
   for (rr = 0; rr < csize; rr++)
   {
      cntr = dynamic_cast <ChipsContour *> (deserialize_object (chips_contour, in));
      if (cntr)
      {
         cntr->set_contour_num (rr);
         dcontours[cntr->get_ref_cnt ()] = cntr;
         cntr->force_hide ();
         if (cntr->get_chips_coord ())
         {
            cntr->get_chips_coord ()->remove (cntr);
            cntr->set_chips_coord (NULL);
         }
      }
   }
   in.read_string(&currcont);
   rval = in.read_checkpoint();
   return rval; 
}


int ChipsPlot::deserialize_histogram(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   int hsz, hsize, rr;
   histograms_map::iterator iter;
   ChipsHistogram* hst; 
   in.read_int(&hsz);
   in.read_int(&hsize);
   if (hsz != hsize) 
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   for (rr = 0; rr < hsize; rr++)
   {
      hst = dynamic_cast <ChipsHistogram *> (deserialize_object (chips_histogram, in));
      if (hst)
      {
         hst->set_histogram_num (rr);
         histograms.insert(histograms_map::value_type (hst->get_id(), hst));
      }
   }
   in.read_int(&hsize);
   for (rr = 0; rr < hsize; rr++)
   {
      hst = dynamic_cast <ChipsHistogram *> (deserialize_object (chips_histogram, in));
      if (hst)
      {
         hst->set_histogram_num (rr);
         dhistograms[hst->get_ref_cnt ()] = hst;
         hst->force_hide ();
         if (hst->get_chips_coord ())
         {
            hst->get_chips_coord ()->remove (hst);
            hst->set_chips_coord (NULL);
         }
      }
   }
   in.read_string(&currhist);
   rval = in.read_checkpoint();
   return rval; 
}

ChipsObject *ChipsPlot::deserialize_object (V_OBJECT_TYPE co_type,
                                                    ChipsStateStream& in)
{
   ChipsObject *co = static_cast <ChipsObject *>(0);
   switch (co_type)
   {
      case chips_axis:
      {
         ChipsAxis *ax = NULL;
         if ((ax = new_axis(in)) != NULL)
         {
            ax->reset_position ();
            co = ax;
         }
      }
      break;
      case chips_image:
      {
         ChipsImage *img = static_cast <ChipsImage *>(0);
         if ((img = new_image(in)) != NULL)
         {
            string tmp_ds_id;
            ChipsCoordinate *tmp_coord=NULL;
            COORD_SYS tmp_coord_sys;

            img->get_coord_system (tmp_coord_sys, tmp_ds_id);
            if (coord)
               tmp_coord = coord->get_coord (tmp_coord_sys, tmp_ds_id);
            if (tmp_coord){
               img->set_chips_coord (tmp_coord);
               img->set_coord_system (DATA, tmp_coord->get_id ());
               tmp_coord->insert(img);
               // Updates the range for the image
               img->reset_position();
            }
            // updates the plot area of the image
            img->set_plot_location(bx1, by1, bx2, by2);
            img->reset_position();
            co = img;
         }
      }
      break;
      case chips_curve:
      {
         ChipsCurve *curv = static_cast <ChipsCurve *>(0);
         if ((curv = new_curve(in)) != NULL)
         {
            string tmp_ds_id;
            ChipsCoordinate *tmp_coord=NULL;
            COORD_SYS tmp_coord_sys;

            curv->get_coord_system (tmp_coord_sys, tmp_ds_id);
            if (coord)
               tmp_coord = coord->get_coord (tmp_coord_sys, tmp_ds_id);
            if (tmp_coord){
               curv->set_chips_coord (tmp_coord);
               curv->set_coord_system (DATA, tmp_coord->get_id ());
               tmp_coord->insert (curv);
               // Updates the range for the curve
               curv->reset_position ();
            }
            // updates the plot area of the curve
            curv->set_plot_location(bx1, by1, bx2, by2);
            curv->reset_position ();
            co = curv;
         }
      }
      break;
      case chips_contour:
      {
         ChipsContour *cntr = static_cast <ChipsContour *>(0);
         if ((cntr = new_contour(in)) != NULL)
         {
            string tmp_ds_id;
            ChipsDataSystem *tmp_coord=NULL;
            COORD_SYS tmp_coord_sys;

            cntr->get_coord_system (tmp_coord_sys, tmp_ds_id);
            if (coord)
               tmp_coord = dynamic_cast <ChipsDataSystem *>(coord->get_coord (tmp_coord_sys, tmp_ds_id));
            if (tmp_coord){

               if (in.get_minor_version() == 0)
               {
                  Transform *transform;
                  string transform_id = "";
                  cntr->get_transform (&transform);

                  if (transform)
                  {
                     ChipsVector min_bound;
                     ChipsVector max_bound;
                     transform_id = tmp_coord->add_transform (transform);
                     
                     // if this is the default transform then the limits
                     // originally created with this coord system are 
                     // invalid, reapply the limits
                     if (transform_id == tmp_coord->get_default_transform_id ())
                     {
                        min_bound = tmp_coord->get_lower_bounds ();
                        max_bound = tmp_coord->get_upper_bounds ();

                        tmp_coord->modify_limits (min_bound, max_bound, transform_id);
                     }
                  }

                  cntr->set_transform_id (transform_id);
               }

               cntr->set_chips_coord (tmp_coord);
               cntr->set_coord_system (DATA, tmp_coord->get_id ());
               tmp_coord->insert (cntr);
               // Updates the range for the curve
               cntr->reset_position ();
            }
   
            // updates the plot area of the contour
            cntr->set_plot_location(bx1, by1, bx2, by2);
            cntr->reset_position ();
            co = cntr;
         }
   
      }
      break;
      case chips_histogram:
      {
         ChipsHistogram *hst = static_cast <ChipsHistogram *>(0);
         if ((hst = new_histogram(in)) != NULL)
         {
            string tmp_ds_id;
            ChipsCoordinate *tmp_coord= NULL;
            COORD_SYS tmp_coord_sys;

            hst->get_coord_system (tmp_coord_sys, tmp_ds_id);
            if (coord)
               tmp_coord = coord->get_coord (DATA, tmp_ds_id);
            if (tmp_coord){
               hst->set_chips_coord (tmp_coord);
               hst->set_coord_system (DATA, tmp_coord->get_id ());
               tmp_coord->insert (hst);
               // Updates the range for the curve
               hst->reset_position ();
            }

            // updates the plot area of the contour
            hst->set_plot_location(bx1, by1, bx2, by2);
            hst->reset_position ();
            co = hst;
         }
      }
      break;
      default:
      break;
   }

   // this is necessary because the coordinate system is added after
   // the object is created. since visiblity is currently dependent on
   // the coordinate system the re-assement of visiblity needs to take place
   co->change_visible ();

   return co;
}


int ChipsPlot::deserialize_annotations(ChipsStateStream& in)
{
   int ival, ival2, rval = ERR_STATUS_GOOD;

   in.read_int(&ival);     //  get label size
   in.read_int(&ival2);
   if (ival == ival2)
   {
      ChipsLabel* lab;
      for (int rr=0; rr< ival2; rr++)
         if ((lab = new_label(in)) != NULL)
         {
            currlabel = lab->get_id(); 
            labels.insert(pair<string, ChipsLabel*>(currlabel, lab));
         }
         in.read_int(&ival);     //  get label size
      for (int rr=0; rr< ival; rr++)
         if ((lab = new_label(in)) != NULL)
         {
            dlabels[lab->get_ref_cnt ()] = lab;
            if (lab->get_chips_coord ())
            {
               lab->get_chips_coord ()->remove (lab);
               lab->set_chips_coord (NULL);
            }
            lab->force_hide ();
         }
   }
   in.read_string(&currlabel);
   if (in.read_checkpoint() != ERR_STATUS_GOOD)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   in.read_int(&ival);     //  get line size
   in.read_int(&ival2);
   if (ival == ival2)
   {
      ChipsLine* lin;
      for (int rr=0; rr< ival2; rr++)
         if ((lin = new_line(in)) != NULL)
         {
            currline = lin->get_id(); 
            lines.insert(pair<string, ChipsLine*>(currline, lin));
         }
         in.read_int(&ival);     //  get line size
      for (int rr=0; rr< ival; rr++)
         if ((lin = new_line(in)) != NULL)
         {
            dlines[lin->get_ref_cnt ()] = lin;
            if (lin->get_chips_coord ())
            {
               lin->get_chips_coord ()->remove(lin);
               lin->set_chips_coord (NULL);
            }
            lin->force_hide ();
         }
   }
   in.read_string(&currline);
   if (in.read_checkpoint() != ERR_STATUS_GOOD)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   if (in.get_minor_version() > 0)
   {
      in.read_int(&ival);     //  get colorbar size
      in.read_int(&ival2);
      if (ival == ival2)
      {
         ChipsColorBar* ib;
         for (int rr=0; rr< ival2; rr++)
            if ((ib = new_colorbar(in)) != NULL)
            {
               currcbar = ib->get_id();
               cbars.insert(pair<string, ChipsColorBar*>(currcbar, ib));
            }
         in.read_int(&ival);     //  get colorbar size
         for (int rr=0; rr< ival; rr++)
            if ((ib = new_colorbar(in)) != NULL)
            {
               dcbars[ib->get_ref_cnt ()] = ib;
               if (ib->get_chips_coord())
               {
                  ib->get_chips_coord()->remove(ib);
                  ib->set_chips_coord(NULL);
               }
               ib->force_hide();
            }
      }
      in.read_string(&currcbar);
      if (in.read_checkpoint() != ERR_STATUS_GOOD)  
      {
         cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
         return ERR_STATEFILE_COR;
      }
   }
   in.read_int(&ival);     //  get point size
   in.read_int(&ival2);
   if (ival == ival2)
   {
      ChipsPoint* pnt;
      for (int rr=0; rr< ival2; rr++)
         if ((pnt = new_point(in)) != NULL)
         {
            currpoint = pnt->get_id();
            points.insert(pair<string, ChipsPoint*>(currpoint, pnt));
         }
      in.read_int(&ival);     //  get point size
      for (int rr=0; rr< ival; rr++)
         if ((pnt = new_point(in)) != NULL)
         {
            dpoints[pnt->get_ref_cnt ()] = pnt;
            if (pnt->get_chips_coord ())
            {
               pnt->get_chips_coord ()->remove (pnt);
               pnt->set_chips_coord (NULL);
            }
            pnt->force_hide ();
         }
   }
   in.read_string(&currpoint);
   if (in.read_checkpoint() != ERR_STATUS_GOOD)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   in.read_int(&ival);     //  get region map size
   in.read_int(&ival2);
   if (ival == ival2)
   {
      ChipsRegion* reg;
      for (int rr=0; rr< ival2; rr++)
         if ((reg = new_polygon(in)) != NULL)
         {
            currregion = reg->get_id();
            regions.insert(pair<string, ChipsRegion*>(currregion, reg));
         }

      in.read_int(&ival);     //  get region map size
      for (int rr=0; rr< ival; rr++)
         if ((reg = new_polygon(in)) != NULL)
         {
            dregions[reg->get_ref_cnt()] = reg;
            if (reg->get_chips_coord ())
            {
               reg->get_chips_coord ()->remove (reg);
               reg->set_chips_coord (NULL);
            }
            reg->force_hide ();
         }
   }
   in.read_string(&currregion);
   if (in.read_checkpoint() != ERR_STATUS_GOOD)  
   {
      cmd_manager->add_err(ERR_STATEFILE_COR,id,0);
      return ERR_STATEFILE_COR;
   }
   recalculate_position();
   return rval; 
}





ChipsObject* ChipsPlot::get_object(const ChipsId* cid, V_OBJECT_TYPE otype,
                                   bool include_deref)
{
   ChipsObject* obj = NULL;
   int obj_ref = 0;
   string my_id;
   if (cid)
   {
      switch(otype)
      {
         case chips_curve:      
         {
            ChipsCurve *cv = NULL;
            my_id = cid->get_curve_id();
            obj_ref = cid->get_curve_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all") != 0) || 
                (!(chips_strcmp(my_id.c_str(), "all")) && curves_exist()))
               cv = get_curve(my_id);
            if (cv && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (cv->get_ref_cnt () == obj_ref))))
               obj = cv;
            if (include_deref && !obj && (dcurves.count (obj_ref) != 0))
               obj = dcurves[obj_ref];
         }
	 break; 
         case chips_contour:   
         {
            ChipsContour *cc = NULL;
            my_id = cid->get_contour_id();
            obj_ref = cid->get_contour_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all") != 0) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && contours_exist()))
               cc = get_contour(my_id);
            if (cc && ((obj_ref==0) || 
                       ((obj_ref > 0) && (cc->get_ref_cnt () == obj_ref))))
               obj = cc;
            if (include_deref && !obj && (dcontours.count (obj_ref) != 0))
               obj = dcontours[obj_ref];
         }
	 break; 
         case chips_histogram:
         {
            ChipsHistogram *ch = NULL;
            my_id = cid->get_histogram_id();
            obj_ref = cid->get_histogram_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all") != 0) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && histograms_exist()))
               ch = get_histogram(my_id);
            if (ch && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (ch->get_ref_cnt () == obj_ref))))
               obj = ch;
            if (include_deref && !obj && (dhistograms.count (obj_ref) != 0))
               obj = dhistograms[obj_ref];
	 }
         break; 
	 case chips_image:
	 {
            ChipsImage *ci = NULL;
            my_id = cid->get_image_id();
            obj_ref = cid->get_image_ref_id();
           if ((chips_strcmp(my_id.c_str(), "all") != 0) ||
               ((!chips_strcmp(my_id.c_str(), "all")) && images.size() == 1))
               ci = get_image(my_id);
            if (ci && ((obj_ref == 0) || ((obj_ref > 0) && (ci->get_ref_cnt () == obj_ref))))
               obj = ci;
            if (include_deref && !obj && (dimages.count (obj_ref) != 0))
               obj = dimages[obj_ref];
	 }
	 break; 
         case chips_label:
         {
            ChipsLabel *cl = NULL;
            my_id = cid->get_label_id();
            obj_ref = cid->get_label_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all")) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && labels.size() == 1))
               cl = get_label(my_id);
            if (cl && ((obj_ref == 0) || ((obj_ref > 0) && (cl->get_ref_cnt () == obj_ref))))
               obj = cl;
            if (include_deref && !obj && (dlabels.count (obj_ref) != 0))
               obj = dlabels[obj_ref];
	 }
         break; 
         case chips_line:
         {
            ChipsLine *cl = NULL;
            my_id = cid->get_line_id();
            obj_ref = cid->get_line_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all")) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && lines.size() == 1))
               cl = get_line(my_id);
            if (cl && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (cl->get_ref_cnt () == obj_ref))))
               obj = cl;
            if (include_deref && !obj && (dlines.count (obj_ref) != 0))
               obj = dlines[obj_ref];
	 }
         break; 
         case chips_colorbar_label:
         case chips_colorbar:
         {
            ChipsColorBar *cl = NULL;
            my_id = cid->get_colorbar_id();
            obj_ref = cid->get_colorbar_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all")) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && cbars.size() == 1))
               cl = get_colorbar(my_id);
            if (cl && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (cl->get_ref_cnt () == obj_ref))))
               obj = cl;
            if (include_deref && !obj && (dcbars.count (obj_ref) != 0))
               obj = dcbars[obj_ref];
	 }
         break; 
         case chips_point:
         {
            ChipsPoint *cl = NULL;
            my_id = cid->get_point_id();
            obj_ref = cid->get_point_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all")) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && points.size() == 1))
               cl = get_point(my_id);
            if (cl && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (cl->get_ref_cnt () == obj_ref))))
               obj = cl;
            if (include_deref && !obj && (dpoints.count (obj_ref) != 0))
               obj = dpoints[obj_ref];
	 }
         break; 
         case chips_region:
         {
            ChipsRegion *cl = NULL;
            my_id = cid->get_region_id();
            obj_ref = cid->get_region_ref_id();
	    if ((chips_strcmp(my_id.c_str(), "all")) || 
                ((!chips_strcmp(my_id.c_str(), "all")) && regions.size() == 1))
               cl = get_region(my_id);
            if (cl && ((obj_ref == 0) || (
                       (obj_ref > 0) && (cl->get_ref_cnt () == obj_ref))))
               obj = cl;
            if (include_deref && !obj && (dregions.count (obj_ref) != 0))
               obj = dregions[obj_ref];
	 }
         break; 
         case chips_axis:
	 {
            ChipsAxis *ca = NULL;
            my_id = cid->get_axis_id();
            obj_ref = cid->get_axis_ref_id();
	    if (my_id != "all")
               if ((obj = get_xaxis(my_id)) == NULL) ca = get_yaxis(my_id);
            if (ca && ((obj_ref == 0) || 
                       ((obj_ref > 0) && (ca->get_ref_cnt () == obj_ref))))
               obj = ca;
            if (include_deref && !obj)
            {
               if (daxes_x.count (obj_ref) != 0)
                  obj = daxes_x[obj_ref];
               else if (daxes_y.count (obj_ref) != 0)
                  obj = daxes_y[obj_ref];
            }
         }
	 break; 
         default:
	 break; 
      }
   }
   return obj; 
}

int ChipsPlot::reference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_GOOD;
   bool cmd_locked = cmd_manager->cmd_lock ();
   ChipsId cid;
   string coord_id;
   COORD_SYS coord_sys;
   ChipsCoordinate *tmp_coord;

   if (!co)   return ERR_STATUS_BAD;

   cmd_manager->set_cmd_lock (true);

   co->get_coord_system (coord_sys, coord_id);

   co->enable_interaction ();

   // chips axis is handled below
   if (coord && (co->get_object_type () != chips_axis))
   {
      // This is temporarily hard coded until a better method of searching
      // down the coordinate system tree is implemented. Otherwise objects
      // in window system must be kept in the frame coord
      if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
          (coord_sys == PIXEL_REAL))
         tmp_coord = coord->get_coord(FRAME_NORM, "");
      else
         tmp_coord = coord->get_coord (coord_sys, coord_id);
      if (tmp_coord){
         co->set_chips_coord (tmp_coord);
         tmp_coord->insert (co);

         // this is necessary because if the coordinate system was never
         // set (ie after deserialize) then this is first time the ranges 
         // will be calculated
         co->reset_position ();
      }
      else
         rval = ERR_COORD_OBJ_DNE;
   }
   if (co->is_visible ())
      co->show ();

   switch (co->get_object_type ())
   {
      case chips_axis:
      {
         ChipsAxis *ax = static_cast <ChipsAxis *>(co);
         ChipsAxis *tmp_ax;
         int ax_orient = ax->get_orientation ();

         // This is necessary if the axis was loaded from a state files
         if (coord)
         {
            // This is temporarily hard coded until a better method of searching
            // down the coordinate system tree is implemented. Otherwise objects
            // in window system must be kept in the frame coord
            if ((coord_sys == WINDOW_NORM) || (coord_sys == PIXEL) ||
                (coord_sys == PIXEL_REAL))
               tmp_coord = coord->get_coord(FRAME_NORM, "");
            else
               tmp_coord = coord->get_coord (coord_sys, coord_id);

            ax->post_initialize (tmp_coord, coord_sys);
            ax->reset_position ();
         }
       
         if (ax_orient == X_AXIS)
         {
            ChipsAxis *tmp_yaxis;
            string csid = "";
            axes_x[ax->get_id ()] = ax;
            daxes_x.erase (ax->get_ref_cnt ());

            // recreate the associated coordinate systems
            for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am)
            {
               tmp_yaxis = am->second;
               if ((tmp_yaxis != left) && (tmp_yaxis != right))
               {
                  map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
                  pair <string, string> p(ax->get_id (), tmp_yaxis->get_id ());
                  ds_iter = deleted_data_systems.find (p);
                  if (ds_iter != deleted_data_systems.end ())
                  {
                     data_systems[p] = ds_iter->second;
                     tmp_yaxis->set_data_coord_id (ds_iter->second->get_id ());
                     ax->set_data_coord_id (ds_iter->second->get_id ());
                     if (coord) coord->add_coord_sys (ds_iter->second);
                     deleted_data_systems.erase (ds_iter);
                  }
               }
            }
            rval = update_current_coord(&cid);

            if (change_cnt)
              axis_x_cnt++;

            // If a current string is passed in use it. It is possible that if a
            // delete all occured and was undone that all is current. Otherwise
            // use referenced frame
            if (curr.length() && (tmp_ax = get_xaxis(curr)))
            {
               cur_x = tmp_ax;
               cur_xall = false;
            }
            else if (curr.length() && (!chips_strcmp (curr.c_str(), "all")))
            {
               cur_x = NULL;
               cur_xall = true;
            }
            else
            { 
               cur_x = ax;
               cur_xall = false;
            }
         }
         else 
         {
            ChipsAxis *tmp_xaxis;
            string csid = "";
            axes_y[ax->get_id ()] = ax;
            daxes_y.erase (ax->get_ref_cnt ());

            // recreate the associated coordinate systems
            for(axes_map::iterator am = axes_x.begin(); am != axes_x.end(); ++am)
            {
               tmp_xaxis = am->second;
               if ((tmp_xaxis != top) && (tmp_xaxis != bottom))
               {
                  map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
                  pair <string, string> p(tmp_xaxis->get_id (), ax->get_id ());
                  ds_iter = deleted_data_systems.find (p);
                  if (ds_iter != deleted_data_systems.end ())
                  {
                     data_systems[p] = ds_iter->second;
                     tmp_xaxis->set_data_coord_id (ds_iter->second->get_id ());
                     ax->set_data_coord_id (ds_iter->second->get_id ());
                     if (coord) coord->add_coord_sys (ds_iter->second);
                     deleted_data_systems.erase (ds_iter);
                  }
               }
            }

            rval = update_current_coord(&cid);

            if (change_cnt)
              axis_y_cnt++;

            // If a current string is passed in use it. It is possible that
            // if a delete all occured then undone that all is current. Otherwise
            // use referenced frame
            if (curr.length() && (tmp_ax = get_yaxis(curr)))
            {
               cur_y = tmp_ax;
               cur_yall = false;
            }
            else if (curr.length() && (!chips_strcmp (curr.c_str(), "all")))
            {
               cur_y = NULL;
               cur_yall = true;
            }
            else
            {
               cur_y = ax;
               cur_yall = false;
            }
         }
      }
      break;
      case chips_curve:
      {
         ChipsCurve *cc = static_cast <ChipsCurve *>(co);
         curves[cc->get_id ()] = cc;
         dcurves.erase (cc->get_ref_cnt ());
         if (change_cnt)
            curve_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_curve(curr)))
            currcurv = curr;
         else
            currcurv = co->get_id ();

      }
      break;
      case chips_contour:
      {
         ChipsContour *cc = static_cast <ChipsContour *>(co);
         contours[cc->get_id ()] = cc;
         dcontours.erase (cc->get_ref_cnt ());
         if (change_cnt)
            contour_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_contour(curr)))
            currcont = curr;
         else
            currcont = co->get_id ();

      }
      break;
      case chips_image:
      {
         ChipsImage *cc = static_cast <ChipsImage *>(co);
         images[cc->get_id ()] = cc;
         dimages.erase (cc->get_ref_cnt ());
         if (change_cnt)
            image_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_image(curr)))
            currimage = curr;
         else
            currimage = co->get_id ();
      }
      break;
      case chips_label:
      {
         ChipsLabel *cc = static_cast <ChipsLabel *>(co);
         labels[cc->get_id ()] = cc;
         dlabels.erase (cc->get_ref_cnt ());
         if (change_cnt)
            label_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_label(curr)))
            currlabel = curr;
         else
            currlabel = co->get_id ();
      }
      break;
      case chips_line:
      {
         ChipsLine *cc = static_cast <ChipsLine *>(co);
         lines[cc->get_id ()] = cc;
         dlines.erase (cc->get_ref_cnt ());
         if (change_cnt)
            line_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_line(curr)))
            currline = curr;
         else
            currline = co->get_id ();
      }
      break;
      case chips_colorbar:
      {
         ChipsColorBar *cc = static_cast <ChipsColorBar*>(co);
         cbars[cc->get_id ()] = cc;
         dcbars.erase (cc->get_ref_cnt ());
         if (change_cnt)
            cbar_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_colorbar(curr)))
            currcbar = curr;
         else
            currcbar = co->get_id ();
      }
      break;
      case chips_point:
      {
         ChipsPoint *cc = static_cast <ChipsPoint*>(co);
         points[cc->get_id ()] = cc;
         dpoints.erase (cc->get_ref_cnt ());
         if (change_cnt)
            point_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_point(curr)))
            currpoint = curr;
         else
            currpoint = co->get_id ();
      }
      break;
      case chips_region:
      {
         ChipsRegion *cc = static_cast <ChipsRegion *>(co);
         regions[cc->get_id ()] = cc;
         dregions.erase (cc->get_ref_cnt ());
         if (change_cnt)
            region_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_region(curr)))
            currregion = curr;
         else
            currregion = co->get_id ();
      }
      break;
      case chips_histogram:
      {
         ChipsHistogram *ch = static_cast <ChipsHistogram *>(co);
         histograms[ch->get_id ()] = ch;
         dhistograms.erase (ch->get_ref_cnt ());
         if (change_cnt)
            histogram_cnt++;

         // If a current string is passed in use it. It is possible that
         // if a delete all occured then undone that all is current. Otherwise
         // use referenced frame
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_histogram(curr)))
            currhist = curr;
         else
            currhist = co->get_id ();
      }
      break;
      default: rval = ERR_STATUS_BAD;
   }

   cmd_manager->set_cmd_lock (cmd_locked);
   return rval;
}

int ChipsPlot::dereference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_GOOD;
   bool cmd_locked = cmd_manager->cmd_lock ();
   ChipsId cid;

   if (!co)  return ERR_STATUS_BAD;

   co->disable_interaction ();

   cmd_manager->set_cmd_lock (true);

   switch (co->get_object_type ())
   {
      case chips_axis:
      {
         ChipsAxis *ax = dynamic_cast <ChipsAxis *> (co);
         int ax_orient = ax->get_orientation ();
         ChipsAxis *tmp_ax;

         cid.set_axis_id (co->get_id ());
         rval = this->delete_axis (&cid, false);

         // if current has a length overide what ever delete axis
         // set as current
         if (!rval && (ax_orient == X_AXIS))
         {
            if (curr.length() && (tmp_ax = get_xaxis(curr)))
            {
               cur_x = tmp_ax;
               cur_xall = false;
            }
            else if (curr.length() && (!chips_strcmp (curr.c_str(), "all")))
            {
               cur_x = NULL;
               cur_xall = true;
            }
            if (change_cnt)
               axis_x_cnt--;
         }
         else if (!rval)
         {
            if (curr.length() && (tmp_ax = get_yaxis(curr)))
            {
               cur_y = tmp_ax;
               cur_yall = false;
            }
            else if (curr.length() && (!chips_strcmp (curr.c_str(), "all")))
            {
               cur_y = NULL;
               cur_yall = true;
            }
            if (change_cnt)
               axis_y_cnt--;
         }
      }
      break;
      case chips_curve:
      {
         cid.set_curve_id (co->get_id ());
         rval = this->delete_curve (&cid, false);

         // if current has a length overide what ever delete curve
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_curve(curr))
)
            currcurv = curr;

         if (change_cnt)
            curve_cnt--;
      }
      break;
      case chips_contour:
      {
         cid.set_contour_id (co->get_id ());
         rval = this->delete_contour (&cid, false);

         // if current has a length overide what ever delete contour
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_contour(curr)))
            currcont = curr;

         if (change_cnt)
            contour_cnt--;
      }
      break;
      case chips_image:
      {
         cid.set_image_id (co->get_id ());
         rval = this->delete_image(&cid, false);

         // if current has a length overide what ever delete image
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_image(curr)))
            currimage = curr;

         if (change_cnt)
            image_cnt--;
      }
      break;
      case chips_label:
      {
         cid.set_label_id (co->get_id ());
         rval = this->delete_label(&cid, false);

         // if current has a length overide what ever delete label
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_label(curr)))
            currlabel = curr;

         if (change_cnt)
            label_cnt--;
      }
      break;
      case chips_line:
      {
         cid.set_line_id (co->get_id ());
         rval = this->delete_line(&cid, false);
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_line(curr)))
            currline = curr;

         if (change_cnt)
            line_cnt--;
      }
      break;
      case chips_colorbar:
      {
         cid.set_colorbar_id (co->get_id ());
         rval = this->delete_colorbar(&cid, false);
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_colorbar(curr)))
            currcbar = curr;

         if (change_cnt)
            cbar_cnt--;
      }
      break;
      case chips_point:
      {
         cid.set_point_id (co->get_id ());
         rval = this->delete_point(&cid, false);
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_point(curr)))
            currpoint = curr;

         if (change_cnt)
            point_cnt--;
      }
      break;
      case chips_region:
      {
         cid.set_region_id (co->get_id ());
         rval = this->delete_region(&cid, false);
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_region(curr)))
            currregion = curr;

         if (change_cnt)
            region_cnt--;
      }
      break;
      case chips_histogram:
      {
         cid.set_histogram_id (co->get_id ());
         rval = this->delete_histogram (&cid, false);

         // if current has a length overide what ever delete histogram
         // set as current
         if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_histogram(curr)))
            currhist = curr;

         if (change_cnt)
            histogram_cnt--;
      }
      break;
      default: rval = ERR_STATUS_BAD;
   }

   cmd_manager->set_cmd_lock (cmd_locked);
   return rval;

}

int ChipsPlot::dereference_axis (ChipsAxis *ax)
{
   int rval = ERR_STATUS_GOOD;
   ChipsDataSystem *tmp_coord;
   list <string> data_coords;
   ChipsAxis *tmp_ax = NULL;
   int dim;

   // hide the axes without changing the visibility
   ax->force_hide ();
   if (ax->get_chips_coord ())
   {
      ax->get_chips_coord ()->remove (ax);
      ax->set_chips_coord (NULL);
   }

   dim = ax->get_orientation ();
   if (dim == X_AXIS)
   {
      daxes_x[ax->get_ref_cnt ()] = ax;
      // delete coord from opposite axis
   }
   else
   {
      daxes_y[ax->get_ref_cnt ()] = ax;
      // delete coord from opposite axis
   }
   // delete coord for this axis
   // loop through all data systems and find ones related to this axis
   map < pair <string,string>, ChipsDataSystem* >::iterator ds_iter;
   map < pair <string,string>, ChipsDataSystem* >::iterator tmp_iter;
   string axis_id = ax->get_id ();
   ds_iter=data_systems.begin ();
   while (ds_iter!=data_systems.end ())
   {
      if ((ds_iter->first.first == axis_id) ||
          (ds_iter->first.second == axis_id))
      {
         string tmp_coord_id;

         tmp_coord = ds_iter->second;
         tmp_coord_id = tmp_coord->get_id ();
         if (coord)
            coord->remove_coord_sys (tmp_coord_id);
         deleted_data_systems[ds_iter->first] = tmp_coord;

         ax->remove_data_coord_id (tmp_coord_id);
         if (dim == Y_AXIS)
            tmp_ax = this->get_xaxis (ds_iter->first.first);
         else
            tmp_ax = this->get_yaxis (ds_iter->first.second);

         if (tmp_ax)
            tmp_ax->remove_data_coord_id (tmp_coord_id);
             

         tmp_iter = ds_iter++;
         data_systems.erase (tmp_iter);
      }
      else
         ds_iter++;
      
   }

   if (!cmd_manager->cmd_lock ())
   {
      cmd_manager->open_block ();

      string curr = (cur_xall) ? "all" :
                    (cur_x)    ? cur_x->get_id () : "";
      ChipsDeleteCommand *del_cmd = new ChipsDeleteCommand(ax, curr);
      ChipsBindCommand *bind_cmd;
      list<ChipsAxis *> ax_list;
      list<ChipsAxis *>::iterator ax_iter;
      double* dp = NULL;
      int*    ip = NULL;
      string* sp = NULL;
       // Get all the bound axes and an unbind command for each
      if (bindings)
      {
         ax_list = bindings->get_axes (ax);
         for (ax_iter= ax_list.begin (); ax_iter != ax_list.end ();
                                                       ax_iter++)
         {
            // only handle directly bound axes 
            if ((*ax_iter)->get_ref_cnt () != ax->get_ref_cnt () &&
                bindings->check_binding(ax, (*ax_iter)))
            {
               bindings->compile_synch_data((*ax_iter),&dp, &ip, &sp);
               // This isn't the best approach another approach
               // should be looked into
               bind_cmd = new ChipsBindCommand(get_parent (),
                            ax->get_parent ()->get_id (),
                            ax->get_id (),
                            (*ax_iter)->get_parent()->get_id (),
                            (*ax_iter)->get_id (),
                            dp, ip, sp, false);
               if ((rval = cmd_manager->add_cmd(bind_cmd)) !=
                                                ERR_STATUS_GOOD)
                  delete bind_cmd;
               if (dp) delete []dp;
               if (ip) delete []ip;
               if (sp) delete []sp;
            }
         }
      }
      if ((rval = cmd_manager->add_cmd(del_cmd)) != ERR_STATUS_GOOD)
         delete del_cmd;
      cmd_manager->close_block ();
   }
   if (bindings)  bindings->remove_axis(ax);

   return rval;
}

int ChipsPlot::clear_cmd (ChipsObject *co)
{
   int rval = ERR_STATUS_GOOD;
   if (!co)
      return ERR_STATUS_BAD;

   switch (co->get_object_type ())
   {
      case chips_curve:
      {
         map <unsigned int, ChipsCurve *>::iterator ii;
         ii = dcurves.find (co->get_ref_cnt ());
         if (ii != dcurves.end ())
         {
            ii->second->object_delete ();
            dcurves.erase (ii);
         }
      }
      break;
      case chips_contour:
      {
         map <unsigned int, ChipsContour *>::iterator ii;
         ii = dcontours.find (co->get_ref_cnt ());
         if (ii != dcontours.end ())
         {
            ii->second->object_delete ();
            dcontours.erase (ii);
         }
      }
      break;
      case chips_histogram:
      {
         map <unsigned int, ChipsHistogram *>::iterator ii;
         ii = dhistograms.find (co->get_ref_cnt ());
         if (ii != dhistograms.end ())
         {
            ii->second->object_delete ();
            dhistograms.erase (ii);
         }
      }
      break;
      case chips_image:
      {
         map <unsigned int, ChipsImage *>::iterator ii;
         ii = dimages.find (co->get_ref_cnt ());
         if (ii != dimages.end ())
         {
            ii->second->object_delete ();
            dimages.erase (ii);
         }
      }
      break;
      case chips_label:
      {
         map <unsigned int, ChipsLabel *>::iterator ii;
         ii = dlabels.find (co->get_ref_cnt ());
         if (ii != dlabels.end ())
         {
            ii->second->object_delete ();
            dlabels.erase (ii);
         }
      }
      break;
      case chips_line:
      {
         map <unsigned int, ChipsLine *>::iterator ii;
         ii = dlines.find (co->get_ref_cnt ());
         if (ii != dlines.end ())
         {
            ii->second->object_delete ();
            dlines.erase (ii);
         }
      }
      break;
      case chips_colorbar:
      {
         map <unsigned int, ChipsColorBar *>::iterator ii;
         ii = dcbars.find (co->get_ref_cnt ());
         if (ii != dcbars.end ())
         {
            ii->second->object_delete ();
            dcbars.erase (ii);
         }
      }
      break;
      case chips_point:
      {
         map <unsigned int, ChipsPoint *>::iterator ii;
         ii = dpoints.find (co->get_ref_cnt ());
         if (ii != dpoints.end ())
         {
            ii->second->object_delete ();
            dpoints.erase (ii);
         }
      }
      break;
      case chips_region:
      {
         map <unsigned int, ChipsRegion *>::iterator ii;
         ii = dregions.find (co->get_ref_cnt ());
         if (ii != dregions.end ())
         {
            ii->second->object_delete ();
            dregions.erase (ii);
         }
      }
      break;
      case chips_axis:
      {
         ChipsAxis *ax = dynamic_cast <ChipsAxis *>(co);
         map <unsigned int, ChipsAxis *>::iterator ii;
         int dim = ax->get_orientation ();

         if (dim == X_AXIS) 
         {
            ii = daxes_x.find (co->get_ref_cnt ());
            if (ii != daxes_x.end ())
            {
               map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
               map <pair <string, string>, ChipsDataSystem*>::iterator tmp_iter;
               ds_iter = deleted_data_systems.begin ();
               while (ds_iter!=deleted_data_systems.end ())
               {
                  if (ds_iter->first.first == ii->second->get_id ())
                  {
                     delete ds_iter->second;

                     tmp_iter = ds_iter++;
                     deleted_data_systems.erase (tmp_iter);
                  }
                  else
                     ds_iter++;
               }
               ii->second->object_delete ();
               daxes_x.erase (ii);
            }
         }
         else
         {
            ii = daxes_y.find (co->get_ref_cnt ());
            if (ii != daxes_y.end ())
            {
string axid,ayid;
ChipsDataSystem *ds;
               map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
               map <pair <string, string>, ChipsDataSystem*>::iterator tmp_iter;
               ds_iter = deleted_data_systems.begin ();
               while (ds_iter!=deleted_data_systems.end ())
               {
                  ds = ds_iter->second;
                  axid=ds_iter->first.first;
                  ayid=ds_iter->first.second;
                  if (ds_iter->first.second == ii->second->get_id ())
                  {
                     delete ds_iter->second;

                     tmp_iter = ds_iter++;
                     deleted_data_systems.erase (tmp_iter);
                  }
                  else
                     ds_iter++;
               }
               ii->second->object_delete ();
               daxes_y.erase (ii);
            }
         }
      }
      break;
      default:
        rval = ERR_STATUS_BAD;
      break;
   }

   return rval;
}

int ChipsPlot::show ()
{
   int rval = ERR_STATUS_GOOD;
   visible = true;
   rval = this->change_visible ();
   return rval;
}

int ChipsPlot::hide ()
{
   int rval = ERR_STATUS_GOOD;

   visible = false;
   this->change_visible ();
   return rval;
}

void ChipsPlot::move_offscreen(bool rec)
{
   int vis = 1;
   set_property(NULL, static_cast<void*>(&vis), SET_PLOT_OFFSCREEN, rec);
}

void ChipsPlot::move_onscreen(bool rec)
{
   int vis = 0;
   set_property(NULL, static_cast<void*>(&vis), SET_PLOT_OFFSCREEN, rec);
}

int ChipsPlot::change_objects_visibility ()
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsCurve*>::iterator curve_ii;
   map<string, ChipsContour*>::iterator contour_ii;
   map<string, ChipsHistogram*>::iterator histogram_ii;
   map<string, ChipsAxis*>::iterator xaxis_ii;
   map<string, ChipsAxis*>::iterator yaxis_ii;
   map<string, ChipsImage*>::iterator image_ii;
   map<string, ChipsLabel*>::iterator label_ii;
   map<string, ChipsLine*>::iterator line_ii;
   map<string, ChipsColorBar*>::iterator cbar_ii;
   map<string, ChipsPoint*>::iterator point_ii;
   map<string, ChipsRegion*>::iterator region_ii;

   for (curve_ii =curves.begin(); !rval && curve_ii != curves.end(); ++curve_ii)
      rval = curve_ii->second->change_visible();

   contour_ii =contours.begin();
   for (; !rval && contour_ii != contours.end(); ++contour_ii)
      rval = contour_ii->second->change_visible();

   histogram_ii =histograms.begin();
   for (; !rval && histogram_ii != histograms.end(); ++histogram_ii)
      rval = histogram_ii->second->change_visible();

   for (xaxis_ii =axes_x.begin(); !rval && xaxis_ii != axes_x.end(); ++xaxis_ii)
      rval = xaxis_ii->second->change_visible();

   for (yaxis_ii =axes_y.begin(); !rval && yaxis_ii != axes_y.end(); ++yaxis_ii)
      rval = yaxis_ii->second->change_visible();

   for (image_ii =images.begin(); !rval && image_ii != images.end(); ++image_ii)
      rval = image_ii->second->change_visible();

   for (label_ii =labels.begin(); !rval && label_ii != labels.end(); ++label_ii)
      rval = label_ii->second->change_visible();

   for (line_ii =lines.begin(); !rval && line_ii != lines.end(); ++line_ii)
      rval = line_ii->second->change_visible();

   for (cbar_ii =cbars.begin(); !rval && cbar_ii != cbars.end(); ++cbar_ii)
      rval = cbar_ii->second->change_visible ();

   for (point_ii =points.begin(); !rval && point_ii != points.end(); ++point_ii)
      rval = point_ii->second->change_visible ();

   for (region_ii =regions.begin(); !rval && region_ii != regions.end(); ++region_ii)
      rval = region_ii->second->change_visible ();
   return rval;
}

int ChipsPlot::adjust_corner_widths(ChipsAxis* changed, bool rec)
{
   if (!changed)  return ERR_OBJ_DNE;
   if (!(left && right && top && bottom))  return ERR_STATUS_BAD;
   if (changed->is_visible())
   {
      double npos      = changed->get_norm_pos(),
             bottom_lo = bottom->get_norm_pos() - 0.001, 
             top_lo    = top->get_norm_pos() - 0.001, 
             right_lo  = right->get_norm_pos() - 0.001, 
             left_lo   = left->get_norm_pos() - 0.001, 
             bottom_hi = bottom->get_norm_pos() + 0.001, 
             top_hi    = top->get_norm_pos() + 0.001, 
             right_hi  = right->get_norm_pos() + 0.001, 
             left_hi   = left->get_norm_pos() + 0.001;
      int    loc = 0; 
      int    dim = changed->get_orientation();
      if (dim == X_AXIS)
      {
         if ((npos >= bottom_lo) && (npos <= bottom_hi))
            loc = 1;
         else if ((npos >= top_lo) && (npos <= top_hi))
            loc = 2;
      }
      else
      {
         if ((npos >= left_lo) && (npos <= left_hi))
            loc = 3;
         else if ((npos >= right_lo) && (npos <= right_hi))
            loc = 4;
      }
      ChipsAxis *ax;
      if ((loc == 3) || (loc == 4))
      {
         for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
         {
	    ax = (*am).second;
            if (ax->is_visible())
            { 
               npos = ax->get_norm_pos(); 
               // check whether bottom axis/box corners needs adjusting
               if (((npos >= bottom_lo) && (npos <= bottom_hi)) ||
                   ((npos >= top_lo) && (npos <= top_hi)))
               {
                  // use set_property so we can use undo/redo
                  double dval = changed->get_thickness();
                  if (loc == 3) 
                     ax->set_property(static_cast<void*>(&dval), 
                                         SET_AXIS_LLCORNWID, rec);
                  else 
                     ax->set_property(static_cast<void*>(&dval),
                                         SET_AXIS_URCORNWID, rec);
               }
            }
         }
      }
      else if ((loc == 1) || (loc == 2))
      {
         for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
         {
	    ax = (*am).second;
            if (ax->is_visible())
            { 
               npos = ax->get_norm_pos();
               // check whether bottom axis/box corners needs adjusting
               if (((npos >= left_lo) && (npos <= left_hi)) ||
                   ((npos >= right_lo) && (npos <= right_hi)))
               {
                  double dval = changed->get_thickness();
                  if (loc == 1)
                     ax->set_property(static_cast<void*>(&dval), 
                                         SET_AXIS_LLCORNWID, rec);
                  else
                     ax->set_property(static_cast<void*>(&dval), 
                                         SET_AXIS_URCORNWID, rec);
               }
            }
         }
      }
   }  // end if (visible)
   return ERR_STATUS_GOOD; 
}

int ChipsPlot::adjust_corner_style()
{
   double bottom_lo = bottom->get_norm_pos() - 0.001, 
          top_lo    = top->get_norm_pos() - 0.001, 
          right_lo  = right->get_norm_pos() - 0.001, 
          left_lo   = left->get_norm_pos() - 0.001, 
          bottom_hi = bottom->get_norm_pos() + 0.001, 
          top_hi    = top->get_norm_pos() + 0.001, 
          right_hi  = right->get_norm_pos() + 0.001, 
          left_hi   = left->get_norm_pos() + 0.001;
   double npos;
   ChipsAxis *ax;

   for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
   {
      ax = (*am).second;
      npos = ax->get_norm_pos(); 
      // change corners of axes that are on top/bottom edge of plot
      if (((npos >= bottom_lo) && (npos <= bottom_hi)) ||
          ((npos >= top_lo) && (npos <= top_hi)))
      {
         ax->set_ll_corner(corner_style);
         ax->set_ur_corner(corner_style);
      }
   }
   for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
   {
      ax = (*am).second;
      npos = ax->get_norm_pos();
      // change corners of axes that are on left/right edge of plot
      if (((npos >= left_lo) && (npos <= left_hi)) ||
          ((npos >= right_lo) && (npos <= right_hi)))
      {
         ax->set_ll_corner(corner_style);
         ax->set_ur_corner(corner_style);
      }
   }
   return ERR_STATUS_GOOD; 
}

int ChipsPlot::adjust_corners(bool rec)
{
   if (!(left && right && top && bottom))  return ERR_STATUS_BAD; 

   double bottom_lo = bottom->get_norm_pos() - 0.001,
          top_lo    = top->get_norm_pos() - 0.001,
          right_lo  = right->get_norm_pos() - 0.001,
          left_lo   = left->get_norm_pos() - 0.001,
          bottom_hi = bottom->get_norm_pos() + 0.001,
          top_hi    = top->get_norm_pos() + 0.001,
          right_hi  = right->get_norm_pos() + 0.001,
          left_hi   = left->get_norm_pos() + 0.001;
   double npos;
   ChipsAxis *ax, *use_b = NULL, *use_t = NULL, *use_l = NULL, *use_r = NULL; 

   for(axes_map::iterator am=axes_x.begin(); am != axes_x.end(); ++am)
   {
      ax = (*am).second;
      npos = ax->get_norm_pos();
      if ((npos >= bottom_lo) && (npos <= bottom_hi) && ax->is_visible()) 
         use_b = ax; 
      else if ((npos >= top_lo) && (npos <= top_hi) && ax->is_visible()) 
         use_t = ax; 
   }
   for(axes_map::iterator am=axes_y.begin(); am != axes_y.end(); ++am)
   {
      ax = (*am).second;
      npos = ax->get_norm_pos();
      if ((npos >= right_lo) && (npos <= right_hi) && ax->is_visible()) 
         use_r = ax; 
      else if ((npos >= left_lo) && (npos <= left_hi) && ax->is_visible())
         use_l = ax; 
   }
   // set corner withs of the axes/borders we are using
   if (use_b)  adjust_corner_widths(use_b, rec);
   if (use_l)  adjust_corner_widths(use_l, rec);
   if (use_t)  adjust_corner_widths(use_t, rec);
   if (use_r)  adjust_corner_widths(use_r, rec);
   return ERR_STATUS_GOOD;
}

int ChipsPlot::update_object_positions()
{
   int rval = ERR_STATUS_GOOD;
   map <string, ChipsCurve *>::iterator curve_iter;
   map <string, ChipsContour *>::iterator contour_iter;
   map <string, ChipsHistogram *>::iterator histogram_iter;
   map <string, ChipsImage *>::iterator image_iter;
   map <string, ChipsAxis *>::iterator axis_iter;

   for (curve_iter = curves.begin (); !rval && curve_iter != curves.end (); ++curve_iter)
      rval = curve_iter->second->set_plot_location(bx1, by1, bx2, by2);

   for (contour_iter = contours.begin (); !rval && contour_iter != contours.end (); ++contour_iter)
      rval = contour_iter->second->set_plot_location(bx1, by1, bx2, by2);

   for (histogram_iter = histograms.begin(); !rval && histogram_iter != histograms.end(); ++histogram_iter)
      rval = histogram_iter->second->set_plot_location(bx1, by1, bx2, by2);

   for (image_iter = images.begin(); !rval && image_iter != images.end(); ++image_iter)
      rval = image_iter->second->set_plot_location(bx1, by1, bx2, by2);

   for (axis_iter = axes_x.begin(); !rval && axis_iter != axes_x.end(); ++axis_iter)
   {
      rval = axis_iter->second->set_plot_location(bx1, by1, bx2, by2);
      if (!rval)  axis_iter->second->set_grid_position(by1, by2);
   }

   for (axis_iter = axes_y.begin(); !rval && axis_iter != axes_y.end(); ++axis_iter)
   {
      rval = axis_iter->second->set_plot_location(bx1, by1, bx2, by2);
      if (!rval)  axis_iter->second->set_grid_position(bx1, bx2);
   }
   update_title_position();
   return rval;
}

int ChipsPlot::reset_object_positions ()
{
   int rval = ERR_STATUS_GOOD;
   map <string, ChipsCurve *>::iterator curve_iter;
   map <string, ChipsImage *>::iterator image_iter;
   map <string, ChipsContour *>::iterator contour_iter;
   map <string, ChipsHistogram *>::iterator histogram_iter;
   map <string, ChipsAxis *>::iterator axis_iter;

   for (image_iter = images.begin (); !rval && image_iter != images.end (); ++image_iter)
      rval = image_iter->second->reset_position();

   for (curve_iter = curves.begin (); !rval && curve_iter != curves.end (); ++curve_iter)
      rval = curve_iter->second->reset_position();

   for (contour_iter = contours.begin (); !rval && contour_iter != contours.end (); ++contour_iter)
      rval = contour_iter->second->reset_position();

   for (histogram_iter = histograms.begin(); !rval && histogram_iter != histograms.end(); ++histogram_iter)
      rval = histogram_iter->second->reset_position();

   for (axis_iter = axes_x.begin(); !rval && axis_iter != axes_x.end(); ++axis_iter)
      rval = axis_iter->second->reset_position();

   for (axis_iter = axes_y.begin(); !rval && axis_iter != axes_y.end(); ++axis_iter)
      rval = axis_iter->second->reset_position ();

   return rval;
}

int ChipsPlot::update_title_position (bool redrawFlag)
{
   int rval = ERR_STATUS_GOOD;

   ChipsVector tmpNormVal(0,0),tmpPixVal(0,0);
   tmpNormVal[1]=1-top_margin;

   coord->convert_cs_to_cs(&tmpNormVal,&tmpPixVal,1,FRAME_NORM,"",PIXEL,"");
   // if plot aspect ratio is on- use the plot's aspect ceiling
   if (py1)  tmpPixVal[1] = py1; 
   tmpPixVal[1] +=title_ypos;
   coord->convert_cs_to_cs(&tmpPixVal,&tmpNormVal,1,PIXEL,"",FRAME_NORM,"");

   title_ypos_norm=tmpNormVal[1];

   tmpNormVal[0]=title_xpos;

   coord->convert_cs_to_cs(&tmpNormVal,&tmpPixVal,1,PLOT_NORM,"",FRAME_NORM,"");
   title_xpos_norm=tmpPixVal[0];

   if ( redrawFlag)  title->set_position(ChipsVector(title_xpos_norm, title_ypos_norm)); 
   return rval;
}

int ChipsPlot::load_depth_table(ChipsDepthControl* tab)
{
   if (!tab)         return ERR_STATUS_BAD;   // no table provided
   if (depth_table)  return ERR_STATUS_BAD;   // table already set
   depth_table = tab;
   return ERR_STATUS_GOOD; 
}

void ChipsPlot::add_border_axes()
{
   ChipsId tmp_cid;
   int rval = ERR_STATUS_GOOD;
   tmp_cid.set_coord_sys(PLOT_NORM);
   bottom = new_axis(coord,PLOT_NORM,0.0, 0.0, 1.0, 0.0, X_AXIS, 0.0, 1.0,ChipsAxisAlgorithmFactory::CARTESIAN, "bx1");
   top = new_axis(coord,PLOT_NORM,0.0, 1.0, 1.0, 1.0, X_AXIS, 0.0, 1.0,ChipsAxisAlgorithmFactory::CARTESIAN, "bx2");
   cmd_manager->open_block ();
   if (bottom)
   {
      rval =  handle_new_xaxis(bottom, false, &tmp_cid);
      add_object_to_depth_table(bottom);
      bottom->set_border_axis (true);
      bottom->disable_interaction ();
   }
   if (top)
   {
      rval =  handle_new_xaxis(top, false, &tmp_cid);
      add_object_to_depth_table(top);
      top->set_border_axis (true);
      top->disable_interaction ();
   }
   cmd_manager->close_block ();
   if (bindings)   bindings->add_binding(bottom, top);

   left = new_axis(coord,PLOT_NORM,0.0, 0.0, 0.0, 1.0, Y_AXIS, 0.0, 1.0, ChipsAxisAlgorithmFactory::CARTESIAN, "by1");
   right = new_axis(coord,PLOT_NORM,1.0, 0.0, 1.0, 1.0, Y_AXIS, 0.0, 1.0, ChipsAxisAlgorithmFactory::CARTESIAN, "by2");
   cmd_manager->open_block ();
   if (left)
   {
      rval =  handle_new_yaxis(left, false, &tmp_cid);
      add_object_to_depth_table(left);
      left->set_border_axis (true);
      left->disable_interaction ();
   }
   if (right)
   {
      rval =  handle_new_yaxis(right, false, &tmp_cid);
      add_object_to_depth_table(right);
      right->set_border_axis (true);
      right->disable_interaction ();
   }

   // create coordinate systems for the axes -- these coordinate
   // systems are for internal use only 
   add_coordinate_system (top, right, NULL);
   add_coordinate_system (bottom, left, NULL);

   cmd_manager->close_block ();
   if (bindings)   bindings->add_binding(left, right);
   bottom->set_ticklabel_visible(false);
   bottom->set_major_grid_visible(false);
   bottom->set_minor_grid_visible(false);
   top->set_ticklabel_visible(false);
   top->set_major_grid_visible(false);
   top->set_minor_grid_visible(false);
   left->set_ticklabel_visible(false);
   left->set_major_grid_visible(false);
   left->set_minor_grid_visible(false);
   right->set_ticklabel_visible(false);
   right->set_major_grid_visible(false);
   right->set_minor_grid_visible(false);
   set_axis_style(axis_style);
   set_title_halign(title_halign); 
   adjust_corner_widths(left); 
   adjust_corner_widths(right); 
   adjust_corner_widths(top); 
   adjust_corner_widths(bottom); 

   this->recalculate_position();

   cur_x = NULL;
   cur_y = NULL;
}

int ChipsPlot::get_depth()
{
   if (depth_table)
      return depth_table->get_default_depth();
   else if (theDefaults)
      return theDefaults->get_defaultdepth();
   else
      return V_DEFAULT_DEPTH;
}

int ChipsPlot::rearrange_object(ChipsId* cid, V_OBJECT_TYPE obj, int manner, bool rec)
{
   string oid;
   ChipsObject* obp;
   ChipsDepthArrangeCommand* cmd = NULL;
   int rval = ERR_STATUS_GOOD;
   int moved = ERR_STATUS_GOOD;

   switch(obj)
   {
      case chips_curve:
      {
         oid = cid->get_curve_id();
         if (oid.length() == 0)  oid = currcurv;
         if ((obp = get_curve(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"curves");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"curve",oid.c_str());
            return rval;
         }
         currcurv = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
           cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
              manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_image:
      {
         oid = cid->get_image_id();
         if (oid.length() == 0)  oid = currimage;
         if ((obp = get_image(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"images");
            else 
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"image",oid.c_str());
            return rval;
         }
         currimage = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs, manner,
               obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_label:
      {
         oid = cid->get_label_id();
         if (oid.length() == 0)  oid = currlabel;
         if ((obp = get_label(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"labels");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
            return rval;
         }
         currlabel = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand
                                        (this, position, num_objs,
                                         manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_line:
      {
         oid = cid->get_line_id();
         if (oid.length() == 0)  oid = currline;
         if ((obp = get_line(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"lines");
            else 
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",oid.c_str());
            return rval;
         }
         currline = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
               manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_colorbar:
      {
         oid = cid->get_colorbar_id();
         if (oid.length() == 0)  oid = currcbar;
         if ((obp = get_colorbar(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"colorbars");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "colorbar", oid.c_str());
            return rval;
         }
         currcbar = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
                                         manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_point:
      {
         oid = cid->get_point_id();
         if (oid.length() == 0)  oid = currpoint;
         if ((obp = get_point(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"points");
            else 
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id,0,"point",oid.c_str());
            return rval;
         }
         currpoint = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
               manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_region:
      {
         oid = cid->get_region_id();
         if (oid.length() == 0)  oid = currregion;
         if ((obp = get_region(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"regions");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id,0,"region",oid.c_str());
            return rval;  
         }
         currregion = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth(obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
               manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_contour:
      {
         oid = cid->get_contour_id();
         if (oid.length() == 0)  oid = currcont;
         if ((obp = get_contour(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all")) 
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"contours");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"contour",oid.c_str());
            return rval;
         }
         currcont = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
               manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_histogram:
      {
         oid = cid->get_histogram_id();
         if (oid.length() == 0)  oid = currhist;
         if ((obp = get_histogram(oid)) == NULL)
         {
            if (!chips_strcmp(oid.c_str(), "all"))
               cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"histograms");
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"histogram",oid.c_str());
            return rval;
         }
         currhist = oid;
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (), obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth());
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs,
               manner, obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      case chips_axis:
      {
         oid = cid->get_axis_id();
         if (!oid.length())   oid = cid->get_xaxis_id();
         if (!oid.length())   oid = cid->get_yaxis_id();
         if (!oid.length())   
         {
            cmd_manager->add_err(ERR_SPECIFY_AXIS,id,0);
            return ERR_SPECIFY_AXIS; 
         }
         if (!chips_strcmp(oid.c_str(), "all")) 
         {
            cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"axes");
            return ERR_OBJ_INVALID_ALL;
         }
         if (axes_x.count(oid))
         {
            obp = axes_x[oid];
            if (obp != top && obp != bottom)  // border axes can't be current
            {
               cur_x = axes_x[oid];
               cur_xall = false;
            }
         }
         else if (axes_y.count(oid))
         {
            obp = axes_y[oid];
            if (obp != left && obp != right)  // border axes can't be current
            {
               cur_y = axes_y[oid];
               cur_yall = false; 
            }
         }
         else
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"axis", oid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            int position =  depth_table->position_in_depth (obp->get_ref_cnt (),
 obp->get_depth());
            int num_objs = depth_table->total_objects_in_depth (obp->get_depth()
);
            cmd = new ChipsDepthArrangeCommand(this, position, num_objs, manner,
                                               obp->get_object_type (), cid);
         }
         moved = depth_table->rearrange_object(obp, manner);
      }
      break;
      default:
         return ERR_FEATURE_DNE;
      break;
   }
   if (rec && cmd)
   {
      if (moved != ERR_STATUS_GOOD)
      {
         // shuffle was a no-op   ie. shuffle front when already at front
         // store no-up to preserved undo/redo count 
         cmd->make_no_op(); 
      }
      if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }
   }
   return rval;
}


int ChipsPlot::adjust_border_visibility()
{
   ChipsAxis* ax;
   int sho_t = 1, sho_b = 1, sho_l = 1, sho_r = 1;
   bool vis;
   float normpos,        // normalized position of axis 
         tol = 0.001;    // tolerance factor

   for(axes_map::iterator am=axes_x.begin(); am!=axes_x.end(); ++am)
   {
      ax = (*am).second;  
      normpos = ax->get_norm_pos(); 
      vis = ax->is_visible();
      if ((ax != bottom) && (normpos >= (bottom->get_norm_pos() - tol)) &&
          (normpos <= (bottom->get_norm_pos() + tol)) && vis)
         sho_b = 0;
      else if ((ax != top) && (normpos >= (top->get_norm_pos() - tol)) &&
          (normpos <= (top->get_norm_pos() + tol)) && vis)
         sho_t = 0;
   }
   for(axes_map::iterator am=axes_y.begin(); am!=axes_y.end(); ++am)
   {
      ax = (*am).second;  
      normpos = ax->get_norm_pos(); 
      vis = ax->is_visible();
      if ((ax != left) && (normpos >= (left->get_norm_pos() - tol)) &&
          (normpos <= (left->get_norm_pos() + tol)) && vis)
         sho_l = 0;
      else if ((ax != right) && (normpos >= (right->get_norm_pos() - tol)) &&
          (normpos <= (right->get_norm_pos() + tol)) && vis)
         sho_r = 0;
   }
   if (sho_b != bottom->is_visible())
      bottom->set_property(static_cast<void*>(&sho_b),SET_AXIS_VISIBLE);
   if (sho_t != top->is_visible())
      top->set_property(static_cast<void*>(&sho_t),SET_AXIS_VISIBLE);
   if (sho_l != left->is_visible())
      left->set_property(static_cast<void*>(&sho_l),SET_AXIS_VISIBLE);
   if (sho_r != right->is_visible())
      right->set_property(static_cast<void*>(&sho_r),SET_AXIS_VISIBLE);
   adjust_corners(); // recompute corner positions based on used axes
   return ERR_STATUS_GOOD;
}

int ChipsPlot::calculate_limits ( ChipsAxis *ax[], double co_min[],
                          double co_max[], double co_lmin[],
                          bool axis_created[], int num_axes,
                          bool single_object)
{
   int rval = ERR_STATUS_GOOD;

   bool updated = false;

   double *axis_min = new double[num_axes];
   double *axis_max = new double[num_axes];
   double *log_min = new double[num_axes];
   bool   *automin = new bool[num_axes];
   bool   *automax = new bool[num_axes];


   double object_min, object_max, object_logmin;


   for (int jj=0; jj<num_axes; jj++)
   {
      if (!ax[jj])
         continue;

//      axis_reversed = ax[jj]->get_reverse ();

      log_min[jj] = ax[jj]->get_log_minimum();
      automin[jj] = ax[jj]->get_automin();
      automax[jj] = ax[jj]->get_automax(),

      ax[jj]->get_linear_range(axis_min[jj], axis_max[jj]);
      object_min = axis_min[jj];
      object_max = axis_max[jj];
      object_logmin = log_min[jj];

      if (single_object && !bindings->check_binding(ax[jj], NULL, true))
      {
         object_min = co_min[jj];
         object_max = co_max[jj];
         object_logmin = co_lmin[jj];
         if (object_max == object_min)
         {
            if (object_min != 0)
            {
               object_min *= 0.95;
               object_max *= 1.05;
            }
            else
            {
               object_min = -1.0;
               object_max = 1.0;
            }
         }
      }
      // when axis' automin/max is set, use the actual data min/max
      else if (automin[jj] || automax[jj])
         bindings->find_axis_min_max(ax[jj], ax[jj]->get_orientation(), &object_min, &object_max,
            &object_logmin, true);

      if (automin[jj] || axis_created[jj])
      {
         axis_min[jj] = object_min;
         log_min[jj] = object_logmin;
      }
      if (automax[jj] || axis_created[jj]) axis_max[jj] = object_max;

      if (axis_created[jj])
      {
         ax[jj]->set_automin_used();
         ax[jj]->set_automax_used();
      }

   }

   double xmin=0;
   double ymin=0;
   double xmax=0;
   double ymax=0;
   ChipsAxis *xaxis;
   ChipsAxis *yaxis;

   for (int jj=0; jj<num_axes; jj++)
   {
      if (ax[jj]->get_orientation () == X_AXIS)
      {
         xaxis = ax[jj];
         xmin = axis_min[jj];
         xmax = axis_max[jj];
      }
      else
      {
         yaxis = ax[jj];
         ymin = axis_min[jj];
         ymax = axis_max[jj];
      }
   }

   for (int jj=0; jj<num_axes; jj++)
   {
      bool rec = !axis_created[jj];
      // update the limits for the current axis
      if (automin[jj] || axis_created[jj])
      {
         bindings->adjust_axis_min_or_max (ax[jj], axis_min[jj], 0, rec);
         bindings->adjust_axis_min_or_max (ax[jj], log_min[jj], 2, rec);
         updated = true;
      }
      if (automax[jj] || axis_created[jj])
      {
         bindings->adjust_axis_min_or_max (ax[jj], axis_max[jj], 1, rec);
         updated = true;
      }
   }

   if (updated == true)
      cmd_manager->update_mask(STATE_CHNG_XLIMS);

   update_data_coord_limits (xaxis, xmin, xmax, yaxis, ymin, ymax, true);


//---- check connected ?? ----

   bindings->reset_access_list(this);
   for (int jj=0; jj<num_axes; jj++)
   {
      if (!ax[jj])
         continue;
      rval = bindings->update_connected_nodes(ax[jj], ax[jj]->get_absolute_min (), ax[jj]->get_absolute_max (), true,
               AUTOLIM_DEF, 1);
   }
   bindings->reset_access_list();


   delete []axis_min;
   delete []axis_max;
   delete []log_min;
   delete []automin; 
   delete []automax; 

   return rval;
}

int ChipsPlot::update_property (const string &axis_id, int property)
{
   list <string> full_list;
   int rval = ERR_STATUS_GOOD;

   full_list.push_back (axis_id);

   rval = this->update_property (full_list, property);

   return rval;
}

int ChipsPlot::update_property (const list <string> &axis_ids, int property)
{
   list <string> full_list = axis_ids;
   list <string>::iterator axis_iter; 
   list <string> exclude_list;
   ChipsAxis *ax;
   int rval = ERR_STATUS_GOOD;
   map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
   int value;
   int orient;

   // only support SET_AXIS_SCALE for now
   if ((property != SET_AXIS_SCALE) && (property != SET_AXIS_REVERSE))
      return ERR_STATUS_GOOD;

   // make sure there are no duplicates
   full_list.unique ();

   // update the coordinates for each of the axes
   for (axis_iter = full_list.begin (); axis_iter != full_list.end (); axis_iter++)
   {
      if ((ax = get_xaxis (*axis_iter)) != NULL)
         orient = ax->get_orientation ();
      else if ((ax = get_yaxis (*axis_iter)) != NULL)
         orient = ax->get_orientation ();
      else
      {
         rval = ERR_STATUS_BAD;
         break;
      }

      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         ax->get_property ((void *)(&value),property);
         if (orient == X_AXIS)
         {
            // update the scale flag
            if ((property == SET_AXIS_SCALE) && (ds_iter->first.first == *axis_iter))
               ds_iter->second->set_log_flag (value==V_SCALE_LOG, orient);

            // update the limits
            update_data_coord_positions (ax, NULL, exclude_list);

         }
         else if (orient == Y_AXIS)
         {
            // update the scale flag
            if ((property == SET_AXIS_SCALE) && (ds_iter->first.second == *axis_iter))
               ds_iter->second->set_log_flag (value==V_SCALE_LOG, orient);

            // update the limits
            update_data_coord_positions (NULL, ax, exclude_list);
         }

      }
   }

   return rval;
}


int ChipsPlot::update_limits (const string &axis_id)
{
   list <string> full_list;
   int rval = ERR_STATUS_GOOD;

   full_list.push_back (axis_id);

   rval = this->update_limits (full_list);
   cmd_manager->update_mask(STATE_CHNG_LIMS); 

   return rval;
}
   

int ChipsPlot::update_limits (const list <string> &axis_ids)
{
   list <string> full_list = axis_ids;
   list <string> used_xids;
   list <string> used_yids;
   list <string>::iterator axis_iter;
   list <string>::iterator used_iter;
   ChipsAxis *ax;
   int rval = ERR_STATUS_GOOD;

   // make sure there are no duplicates
   full_list.unique ();

   // update the coordinates for each of the axes
   for (axis_iter = full_list.begin (); axis_iter != full_list.end (); axis_iter++)
   {
      if ((ax = get_xaxis (*axis_iter)) != NULL)
      {
         update_data_coord_positions (ax, NULL, used_yids);
         used_xids.push_back (*axis_iter);
      }
      else if ((ax = get_yaxis (*axis_iter)) != NULL)
      {
         update_data_coord_positions (NULL,ax,used_xids);
         used_yids.push_back (*axis_iter);
      }
      else
         rval = ERR_STATUS_BAD;
   }
   if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_LIMS); 

   return rval;

}

int ChipsPlot::add_object_to_depth_table(ChipsObject* obj, bool rec)
{
   // add entry to depth table and set up undo/redo command
   if (depth_table)
   {
      if (rec)
      {
         ChipsDepthUpdateCommand* cmd =
            new ChipsDepthUpdateCommand(this, obj, 0);
         if (cmd_manager->add_cmd(cmd) != ERR_STATUS_GOOD)
            delete cmd;
      }
      depth_table->add_object(obj);
      return ERR_STATUS_GOOD;
   }
   return ERR_STATUS_BAD;
}


int ChipsPlot::remove_object_from_depth_table(ChipsObject* obj)
{
   // remove entry from depth table and set up undo/redo command
   if (depth_table)
   {
      // store of position in depth so that we can get reinsert the object
      // in the right position during an undo
      int posn = depth_table->total_objects_in_depth(obj->get_depth()) -
         depth_table->position_in_depth(obj->get_ref_cnt(), obj->get_depth());
      ChipsDepthUpdateCommand* cmd = new ChipsDepthUpdateCommand(this, obj,
         1, posn);
      cmd_manager->add_cmd(cmd);
      depth_table->remove_object(obj);
      return ERR_STATUS_GOOD;
   }
   return ERR_STATUS_BAD;
}

int ChipsPlot::update_depth_table(V_OBJECT_TYPE obj, string oid, int how, int pos)
{
   // this routine is used for updating the depth table during undo/redo of
   // object set depth calls

   if (!depth_table)  return ERR_STATUS_BAD;   // no depth table to update
   ChipsObject* cobj = NULL;
   switch (obj)
   {
      case chips_image:
         cobj = get_image(oid); 
      break; 
      case chips_label:
         cobj = get_label(oid); 
      break; 
      case chips_line:
         cobj = get_line(oid); 
      break; 
      case chips_colorbar:
         cobj = get_colorbar(oid); 
      break; 
      case chips_point:
         cobj = get_point(oid); 
      break; 
      case chips_region:
         cobj = get_region(oid); 
      break; 
      case chips_curve:
         cobj = get_curve(oid);
      break;
      case chips_contour:
         cobj = get_contour(oid);
      break;
      case chips_histogram:
         cobj = get_histogram(oid);
      break;
      case chips_axis:
         if ((cobj = get_xaxis(oid)) == NULL)
            cobj = get_yaxis(oid);
      break;
//      case chips_label: // plot title
//         cobj = get_title();
//      break; 
      default:
         return ERR_STATUS_BAD;
      break;
   }
   if (cobj)  // object found- update the depth table accordingly
   {
      if (how) 
         depth_table->remove_object(cobj);
      else
      {
         depth_table->add_object(cobj);
         // reposition object to its previous location
         for (int rr=0; rr < pos; rr++)
            depth_table->rearrange_object(cobj, chips_backward);
      }
   }
   return ERR_STATUS_GOOD;
}



void ChipsPlot::reset_title_defaults()
{
   if (!title)  return;

   this->title->unselect ();
   
   double tmp_x, tmp_y; 
   if (theDefaults)
   {
      tmp_x        = theDefaults->get_plottitlexpos();
      tmp_y        = theDefaults->get_plottitleypos();
      title_angle  = theDefaults->get_plottitleangle();
      title_halign = theDefaults->get_plottitlehalign();
      title_valign = theDefaults->get_plottitlevalign();
      title_size   = theDefaults->get_plottitlesize();
      title_color  = theDefaults->get_plottitlecolor();
      title_font   = theDefaults->get_plottitlefont();
      title_fontstyle = theDefaults->get_plottitlefontstyle();
   }
   else
   {
      tmp_x = 0.45;
      tmp_y = 15;
      title_angle = 0.0;
      title_halign = title_halign = AUTO_ALIGN;
      title_size = 12;
      title_color = "white";
      title_font = "helvetica";
      title_fontstyle = chips_normal;
   }

   if (tmp_x != title_xpos)
      set_property(NULL, static_cast<void*>(&tmp_x), SET_PLOT_TITLE_X); 
   if (tmp_y != title_ypos)
      set_property(NULL, static_cast<void*>(&tmp_y), SET_PLOT_TITLE_Y); 
   if (title_angle != get_title_angle())
      set_property(NULL, static_cast<void*>(&title_angle),SET_PLOT_TITLE_ANGLE);
   if (title_size != get_title_size())
      set_property(NULL, static_cast<void*>(&title_size), SET_PLOT_TITLE_SIZE);
   if (title_font != get_title_font())
      set_property(NULL, (void*)title_font.c_str(), SET_PLOT_TITLE_FONT);
   if (title_color != get_title_color())
      set_property(NULL, (void*)title_color.c_str(), SET_PLOT_TITLE_COLOR);
   if (title_fontstyle != title->get_fontstyle())
   {
      string tmpstr =  v_fontstyle_to_string(title_fontstyle);
      set_property(NULL, (void*)tmpstr.c_str(), SET_PLOT_TITLE_FONTSTYLE);
   }
   if (get_title_text() != "")
   {
      string tmpstr = "";
      set_property(NULL, (void*)tmpstr.c_str(), SET_PLOT_TITLE);
   }
   if (title_halign != get_title_halign())
      set_property(NULL, static_cast<void*>(&title_halign), SET_PLOT_TITLE_HALIGN);
   if (title_valign != get_title_valign())
      set_property(NULL, static_cast<void*>(&title_valign), SET_PLOT_TITLE_VALIGN);
}


void ChipsPlot::initialize_title()
{
   if (title) 
   {
      title->set_fontstyle(title_fontstyle);
      title->set_size(title_size);
      title->set_font(title_font);
      title->set_color(title_color);
      title->set_angle(title_angle);
      set_title_halign(title_halign);
      set_title_valign(title_valign);

      if (title_depth==V_USE_DEFAULT_DEPTH) title_depth = get_depth ();
      set_title_depth(title_depth);
      if (theDefaults)
      {
         string tmptitle = theDefaults->get_plottitlelabel(); 
         if (tmptitle != "")   title->set_text(tmptitle); 
      }
   }
}


int ChipsPlot::clear()
{
   ChipsId cid;
   cid.set_curve_id("all");
   cid.set_contour_id("all");
   cid.set_histogram_id("all");
   cid.set_image_id("all");
   cid.set_label_id("all");
   cid.set_line_id("all");
   cid.set_point_id("all");
   cid.set_colorbar_id("all");
   cid.set_region_id("all");

   delete_curve (&cid, 0);
   delete_contour (&cid, 0);
   delete_histogram (&cid, 0);
   delete_image(&cid, 0);
   delete_label(&cid, 0);
   delete_line(&cid, 0);
   delete_point(&cid, 0);
   delete_colorbar(&cid, 0);
   delete_region(&cid, 0);
   cid.set_xaxis_id("all");
   delete_xaxis (&cid, 0);
   ChipsId cid2;
   cid2.set_yaxis_id("all");
   delete_yaxis (&cid2, 0);

   // reset default title properties
   if (title)    reset_title_defaults();

   // reset border axes properties
   if (bottom)
   {
      bottom->reset_values();
      bottom->adjust_minmax (0,1);
   }
   if (top)
   {
      top->reset_values();
      top->adjust_minmax (0,1);
   }
   if (left)
   {
      left->reset_values();
      left->adjust_minmax (0,1);
   }
   if (right)
   {
      right->reset_values();
      right->adjust_minmax (0,1);
   }

   return ERR_STATUS_GOOD; 
}


int ChipsPlot::remove_plot_from_depth_table()
{
   if (depth_table)
   {
      if (title)   remove_object_from_depth_table(title); 
      for(images_map::iterator ll = images.begin(); ll != images.end(); ++ll)
         remove_object_from_depth_table((*ll).second);
      for(labels_map::iterator ll = labels.begin(); ll != labels.end(); ++ll)
         remove_object_from_depth_table((*ll).second);
      for(lines_map::iterator ll = lines.begin(); ll != lines.end(); ++ll)
         remove_object_from_depth_table((*ll).second);
      for(curves_map::iterator cc = curves.begin(); cc != curves.end(); ++cc)
         remove_object_from_depth_table((*cc).second);
      for(contours_map::iterator ct=contours.begin();ct!=contours.end(); ++ct)
         remove_object_from_depth_table((*ct).second);
      for(histograms_map::iterator ch=histograms.begin();ch!=histograms.end(); ++ch)
	 remove_object_from_depth_table((*ch).second);
      for(axes_map::iterator ax=axes_x.begin(); ax!=axes_x.end(); ++ax)
	 remove_object_from_depth_table((*ax).second);
      for(axes_map::iterator ay=axes_y.begin(); ay!=axes_y.end(); ++ay)
	 remove_object_from_depth_table((*ay).second);
   }
   return ERR_STATUS_GOOD;
}


int ChipsPlot::delete_label(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsLabel* clab;
   string lbid = cid->get_label_id();

   if (lbid.length() == 0)   lbid = currlabel;

   if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      map<string, ChipsLabel* >::iterator ii;

      if (labels.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"labels");
         return rval;
      }

      // delete all labels in frame
      if (!force_del)
         cmd_manager->open_block ();

      for(ii=labels.begin(); ii != labels.end(); ++ii)
      {
         clab = (*ii).second;
         clab->unselect ();
         if (depth_table)
         {
	    if (!force_del)   remove_object_from_depth_table(clab);
            else              depth_table->remove_object(clab);
         }
	 if (force_del)
            clab->object_delete ();
	 else
         {
            // this hides the label without changing its visibility
            clab->force_hide ();
            dlabels [clab->get_ref_cnt ()] = clab;
	    if (clab->get_chips_coord ())
            {
               clab->get_chips_coord ()->remove (clab);
               clab->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(clab, currlabel);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }

      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         // If force deleteing all labels also remove previously deleted labels 
         map <unsigned int, ChipsLabel*>::iterator  dlabel_ii;
         for (dlabel_ii = dlabels.begin (); dlabel_ii != dlabels.end (); dlabel_ii++)
            dlabel_ii->second->object_delete ();
         dlabels.clear ();
      }

      labels.clear();
      currlabel = "";
      cid->set_label_id("");
      rval = ERR_STATUS_GOOD;
      label_cnt = 0;
   }
   else if (labels.count(lbid))
   {
      map<string, ChipsLabel* >::iterator ii;
      ii = labels.find(lbid);
      clab = (*ii).second;
      clab->unselect ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(clab);
         else               depth_table->remove_object(clab);
      }
      labels.erase(ii);
      if (force_del)
         clab->object_delete ();
      else
      {
         // this hides the label without changing its visibility
         clab->force_hide ();
         dlabels [clab->get_ref_cnt ()] = clab;
         if (clab->get_chips_coord ())
         {
            clab->get_chips_coord ()->remove (clab);
            clab->set_chips_coord (NULL);
         }
         coord->remove (clab);
         if (!cmd_manager->cmd_lock ())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(clab, currlabel);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_label_id("");
      if (!labels.empty())
      {
         map<string, ChipsLabel* >::reverse_iterator ri = labels.rbegin();
         currlabel = (*ri).first;
      }
      else   currlabel = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",lbid.c_str());
   return rval;
}


int ChipsPlot::add_label(ChipsId* cid, double xpos, double ypos, string text)
{
   int rval = ERR_STATUS_GOOD;
   ChipsLabel* tmplab = NULL; 
   string tid = theDefaults->get_labelid();
   bool label_created = false;

   if (tid.length() && labels.count(tid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,tid.c_str(), "label");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !label_created)
         tmplab = new_label(tmp_coord, tmpId.get_coord_sys (), xpos, ypos, text, tid);

      if (tmplab && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if tid is empty that means the label_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmplab,(tid.length()==0),currlabel);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
         // new label created- add to list and update current
         add_object_to_depth_table(tmplab);
         tmplab->set_label_num (label_cnt);
         currlabel = tmplab->get_id();
         labels.insert(pair<string, ChipsLabel*>(currlabel, tmplab));
         cid->set_label_id(currlabel);
      }
      else if (!rval && !label_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"label");
   }
   return rval;
}


ChipsLabel* ChipsPlot::new_label(ChipsCoordinate *inCoord, COORD_SYS sys,
      double xpos, double ypos, string text, string tid)
{
   string tmpid;
   if (tid.length())
      tmpid = tid;
   else
      tmpid = create_label_id();

   return new ChipsLabel(this, tmpid, inCoord, sys, xpos, ypos, text);
}


ChipsLabel* ChipsPlot::new_label(ChipsStateStream& in)
{
   return new ChipsLabel(this, coord, coord_system, in);
}


char* ChipsPlot::create_label_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_labelstem();
      else            stem = "lbl";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++label_cnt);

   // increment count til next unused label id
   while(get_label(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++label_cnt);
   }
   return temp_id;
}


ChipsLabel* ChipsPlot::get_label(const string& val)
{
   string clid = val;
   if (!val.length())  clid = currlabel;
   if ((chips_strcmp(clid.c_str(), "all") == 0) && labels.size() == 1)
   {
      map<string, ChipsLabel*>::iterator ii = labels.begin();
      return ((*ii).second);
   }
   else if ((chips_strcmp(clid.c_str(), "all") != 0) && labels.count(clid))
      return labels[clid];
   else
      return NULL;
}


int ChipsPlot::set_label_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string lbid = cid->get_label_id();
   if (lbid.length() == 0)  lbid = currlabel;

   if (labels.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"labels");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      map<string, ChipsLabel*>::iterator ii;
      for(ii=labels.begin(); ii != labels.end(); ++ii)
         rval |= (*ii).second->set_property(in_val, in_property);
   }
   else
   {
      if (labels.count(lbid) == 0) 
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"label",lbid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ChipsLabel* lab = labels[lbid];
      rval = lab->set_property(in_val, in_property);
   }
   if (rval == ERR_STATUS_GOOD)
   {
      currlabel = lbid;
      cid->set_label_id(currlabel);
   }
   return rval;
}


int ChipsPlot::get_label_property(ChipsId* cid, void* out_val, int property)
{
   string lbid = cid->get_label_id();
   if (lbid.length() == 0)  lbid = currlabel;
   if (labels.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"labels");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      if (labels.size() == 1)
      {
         map<string, ChipsLabel*>::iterator ii = labels.begin();
         return ((*ii).second)->get_property(out_val, property);
      }
      cmd_manager->add_err(ERR_INVALID_GET_ALL);
      return ERR_INVALID_GET_ALL;
   }
   else if (labels.count(lbid) == 0)  
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"label",lbid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   return (labels[lbid])->get_property(out_val, property);
}

int ChipsPlot::move_label(ChipsId* cid, double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD;
   string oid = cid->get_label_id();
   string  cd_id = cid->get_coord_id ();
   COORD_SYS  cd_sys = cid->get_coord_sys ();

   if (oid.length() == 0)  oid = currlabel;
   if (labels.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"labels");
      return ERR_OBJ_LIST_EMPTY;
   }
   map <string, ChipsLabel*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts, rval = ERR_STATUS_GOOD;
      for(ii=labels.begin(); ii != labels.end(); ++ii)
        if ((sts |= (*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
             ERR_STATUS_GOOD)   rval = sts;
   }
   else if (labels.count(oid))
      rval = labels[oid]->move(cd_sys, cd_id, xx, yy, mode);
   else
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"label",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   currlabel = oid;
   return rval;
}

int ChipsPlot::set_current_label(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_label_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!labels.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "label", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currlabel, lid, chips_label);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currlabel = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currlabel,
                                                      lid, chips_label);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currlabel = "all";
      }
   }
   return rval;
}



int ChipsPlot::delete_line(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsLine* clab;
   string lbid = cid->get_line_id();

   if (lbid.length() == 0)   lbid = currline;

   if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      map<string, ChipsLine* >::iterator ii;

      if (lines.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"lines");
         return rval;
      }

      // delete all lines in plot
      if (!force_del)
         cmd_manager->open_block ();

      for(ii=lines.begin(); ii != lines.end(); ++ii)
      {
         clab = (*ii).second;
         clab->unselect ();
         if (depth_table)
         {
	    if (!force_del)   remove_object_from_depth_table(clab);
            else              depth_table->remove_object(clab);
         }
	 if (force_del)
            clab->object_delete ();
	 else
         {
            // this hides the line without changing its visibility
            clab->force_hide ();
            dlines [clab->get_ref_cnt ()] = clab;
	    if (clab->get_chips_coord ())
            {
               clab->get_chips_coord ()->remove (clab);
               clab->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(clab, currline);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }

      if (!force_del)
         cmd_manager->close_block ();
      else
      {
         // If force deleteing all lines also remove previously deleted lines 
         map <unsigned int, ChipsLine*>::iterator  dline_ii;
         for (dline_ii = dlines.begin (); dline_ii != dlines.end (); dline_ii++)
            dline_ii->second->object_delete ();
         dlines.clear ();
      }

      lines.clear();
      currline = "";
      cid->set_line_id("");
      rval = ERR_STATUS_GOOD;
      line_cnt = 0;
   }
   else if (lines.count(lbid))
   {
      map<string, ChipsLine* >::iterator ii;
      ii = lines.find(lbid);
      clab = (*ii).second;
      clab->unselect ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(clab);
         else              depth_table->remove_object(clab);
      }
      lines.erase(ii);
      if (force_del)
         clab->object_delete ();
      else
      {
         // this hides the line without changing its visibility
         clab->force_hide ();
         dlines [clab->get_ref_cnt ()] = clab;
         if (clab->get_chips_coord ())
         {
            clab->get_chips_coord ()->remove (clab);
            clab->set_chips_coord (NULL);
         }
         coord->remove (clab);
         if (!cmd_manager->cmd_lock ())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(clab, currline);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_line_id("");
      if (!lines.empty())
      {
         map<string, ChipsLine* >::reverse_iterator ri = lines.rbegin();
         currline = (*ri).first;
      }
      else   currline = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",lbid.c_str());
   return rval;
}


int ChipsPlot::add_line(ChipsId* cid, double startx, double starty,
                        double endx,  double endy)
{
   int rval = ERR_STATUS_GOOD;
   ChipsLine* tmplin = NULL; 
   string tid = theDefaults->get_lineid();
   bool line_created = false;

   if (tid.length() && lines.count(tid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,tid.c_str(), "line");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !line_created)
         tmplin = new_line(tmp_coord, tmpId.get_coord_sys(), startx,
	 starty, endx, endy, tid);

      if (tmplin && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if tid is empty that means the line_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmplin,(tid.length()==0),currline);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
         // new line created- add to list and update current
         add_object_to_depth_table(tmplin);
         tmplin->set_line_num (line_cnt);
         currline = tmplin->get_id();
         lines.insert(pair<string, ChipsLine*>(currline, tmplin));
         cid->set_line_id(currline);
      }
      else if (!rval && !line_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"line");
   }
   return rval;
}

int ChipsPlot::add_line(ChipsId* cid, double* xx, double* yy, 
   unsigned int count)
{
   int rval = ERR_STATUS_GOOD;
   ChipsLine* tmplin = NULL; 
   string tid = theDefaults->get_lineid();
   bool line_created = false;

   if (tid.length() && lines.count(tid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,tid.c_str(), "line");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !line_created)
         tmplin = new_line(tmp_coord, tmpId.get_coord_sys(), xx, yy, count, tid);
      if (tmplin && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if tid is empty that means the line_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmplin,(tid.length()==0),currline);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
         // new line created- add to list and update current
         add_object_to_depth_table(tmplin);
         tmplin->set_line_num (line_cnt);
         currline = tmplin->get_id();
         lines.insert(pair<string, ChipsLine*>(currline, tmplin));
         cid->set_line_id(currline);
      }
      else if (!rval && !line_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"line");
   }
   return rval;
}


ChipsLine* ChipsPlot::new_line(ChipsCoordinate *inCoord, COORD_SYS sys,
   double startx, double starty, double endx, double endy, string lnid)
{
   string tmpid;
   if (lnid.length())
      tmpid = lnid;
   else
      tmpid = create_line_id();
   return new ChipsLine(this, inCoord, sys, startx, starty, endx, endy, tmpid);
}


ChipsLine* ChipsPlot::new_line(ChipsCoordinate *inCoord, COORD_SYS sys,
   double* xx, double* yy, unsigned int count, string lnid)
{
   string tmpid;
   if (lnid.length())
      tmpid = lnid;
   else
      tmpid = create_line_id();
   return new ChipsLine(this, inCoord, sys, xx, yy, count, tmpid);
}

ChipsLine* ChipsPlot::new_line(ChipsStateStream& in)
{
   return new ChipsLine(this, coord, coord_system, in);
}


char* ChipsPlot::create_line_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_linestem();
      else            stem = "lin";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++line_cnt);

   // increment count til next unused line id
   while(get_line(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++line_cnt);
   }
   return temp_id;
}


ChipsLine* ChipsPlot::get_line(const string& val)
{
   string clid = val;
   if (!val.length())  clid = currline;
   if ((chips_strcmp(clid.c_str(), "all") == 0) && lines.size() == 1)
   {
      map<string, ChipsLine*>::iterator ii = lines.begin();
      return ((*ii).second);
   }
   else if ((chips_strcmp(clid.c_str(), "all") != 0) && lines.count(clid))
      return lines[clid];
   else
      return NULL;
}


int ChipsPlot::set_line_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string lbid = cid->get_line_id();
   if (lbid.length() == 0)  lbid = currline;

   if (lines.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"lines");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      map<string, ChipsLine*>::iterator ii;
      for(ii=lines.begin(); ii != lines.end(); ++ii)
         rval |= (*ii).second->set_property(in_val, in_property);
   }
   else
   {
      if (lines.count(lbid) == 0) 
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"line",lbid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ChipsLine* lab = lines[lbid];
      rval = lab->set_property(in_val, in_property);
   }
   if (rval == ERR_STATUS_GOOD)
   {
      currline = lbid;
      cid->set_line_id(currline);
   }
   return rval;
}


int ChipsPlot::get_line_property(ChipsId* cid, void* out_val, int property)
{
   string lbid = cid->get_line_id();
   if (lbid.length() == 0)  lbid = currline;
   if (lines.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"lines");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(lbid.c_str(), "all") == 0)
   {
      if (lines.size() == 1)
      {
         map<string, ChipsLine*>::iterator ii = lines.begin();
         return ((*ii).second)->get_property(out_val, property);
      }
      cmd_manager->add_err(ERR_INVALID_GET_ALL);
      return ERR_INVALID_GET_ALL;
   }
   else if (lines.count(lbid) == 0)  
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"line",lbid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   return (lines[lbid])->get_property(out_val, property);
}

int ChipsPlot::move_line(ChipsId* cid, double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD;
   string  oid = cid->get_line_id();
   string  cd_id = cid->get_coord_id ();
   COORD_SYS  cd_sys = cid->get_coord_sys ();

   if (oid.length() == 0)  oid = currline;
   if (lines.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"lines");
      return ERR_OBJ_LIST_EMPTY;
   }
   map <string, ChipsLine*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts, rval = ERR_STATUS_GOOD;
      for(ii=lines.begin(); ii != lines.end(); ++ii)
        if ((sts |= (*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
             ERR_STATUS_GOOD)   rval = sts;
   }
   else if (lines.count(oid))
      rval = lines[oid]->move(cd_sys, cd_id, xx, yy, mode);
   else
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"line",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   currline = oid;
   return rval;
}

int ChipsPlot::set_current_line(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_line_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!lines.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "line", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }

         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currline, lid, chips_line);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currline = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currline,
                                                      lid, chips_line);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currline = "all";
      }
   }
   return rval;
}


int ChipsPlot::add_point(ChipsId* cid, double xx, double yy)
{
   int rval = ERR_STATUS_GOOD;
   ChipsPoint* tmppoint = NULL;
   string ptid = theDefaults->get_pointid();
   bool point_created = false;

   if (ptid.length() && points.count(ptid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,ptid.c_str(), "point");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !point_created)
         tmppoint = new_point(tmp_coord, tmpId.get_coord_sys (), xx, yy, ptid);

      if (tmppoint && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if ptid is empty that means the point_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmppoint,(ptid.length()==0),currpoint);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         // new point created- add to list and update current
         add_object_to_depth_table(tmppoint);
         tmppoint->set_point_num (point_cnt);
         currpoint = tmppoint->get_id();
         points.insert(pair<string, ChipsPoint*>(currpoint, tmppoint));
         cid->set_point_id(currpoint);
      }
      else if (!rval && !point_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"point");
   }
   return rval;
}

int ChipsPlot::add_point(ChipsId* cid, double xx, double yy, double zz)
{
   int rval = ERR_STATUS_GOOD;
   ChipsPoint* tmppoint = NULL;
   string ptid = theDefaults->get_pointid();
   bool point_created = false;

   if (ptid.length() && points.count(ptid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,ptid.c_str(), "point");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !point_created)
         tmppoint = new_point(tmp_coord, tmpId.get_coord_sys (), xx, yy, zz,ptid);

      if (tmppoint && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if ptid is empty that means the point_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmppoint,(ptid.length()==0),currpoint);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         // new point created- add to list and update current
         add_object_to_depth_table(tmppoint);
         tmppoint->set_point_num (point_cnt);
         currpoint = tmppoint->get_id();
         points.insert(pair<string, ChipsPoint*>(currpoint, tmppoint));
         cid->set_point_id(currpoint);
      }
      else if (!rval && !point_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"point");
   }
   return rval;
}


int ChipsPlot::delete_point(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsPoint* cpoint;
   string ptid = cid->get_point_id();
   if (ptid.length() == 0)   ptid = currpoint;

   if (chips_strcmp(ptid.c_str(), "all") == 0)
   {
      map<string, ChipsPoint*>::iterator ii;

      if (points.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"points");
         return rval;
      }

      // delete all points in frame
      if (!force_del)
         cmd_manager->open_block ();
      for(ii=points.begin(); ii != points.end(); ++ii)
      {
         cpoint = (*ii).second;
         cpoint->unselect ();
         if (depth_table)
         {
            if (!force_del)  remove_object_from_depth_table(cpoint);
            else             depth_table->remove_object(cpoint);
         }
	 if (force_del)
            cpoint->object_delete ();
	 else{
            // this hides the point without changing its visibility
            cpoint->force_hide ();
            dpoints[cpoint->get_ref_cnt ()] = cpoint;
            if (cpoint->get_chips_coord ())
            {
               cpoint->get_chips_coord ()->remove (cpoint);
               cpoint->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ()){
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cpoint, currpoint);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }
      if (!force_del) 
         cmd_manager->close_block ();
      else
      {
         map <unsigned int, ChipsPoint*>::iterator  dpoint_ii;
         // If force deleteing all points also remove previously delete points
         for (dpoint_ii = dpoints.begin (); dpoint_ii != dpoints.end (); dpoint_ii++)
            dpoint_ii->second->object_delete ();
         dpoints.clear ();

      }
      points.clear();
      currpoint = "";
      cid->set_point_id("");
      rval = ERR_STATUS_GOOD;
      point_cnt = 0;
   }
   else if (points.count(ptid))
   {
      map<string, ChipsPoint*>::iterator ii;
      ii = points.find(ptid);
      cpoint = (*ii).second;
      cpoint->unselect ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(cpoint);
         else              depth_table->remove_object(cpoint);
      }
      points.erase(ii);
      if (force_del)
         cpoint->object_delete ();
      else
      {
         // this hides the region without changing its visibility
         cpoint->force_hide ();
         dpoints[cpoint->get_ref_cnt ()] = cpoint;
         if (cpoint->get_chips_coord ())
         {
            cpoint->get_chips_coord ()->remove (cpoint);
            cpoint->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock ())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cpoint, currpoint);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_point_id("");
      if (!points.empty())
      {
         map<string, ChipsPoint* >::reverse_iterator ri = points.rbegin();
         currpoint = (*ri).first;
      }
      else   currpoint = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",ptid.c_str());
   return rval;
}

int ChipsPlot::add_colorbar(ChipsId* cid, double xx, double yy,
   string obj_id, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   ChipsColorBar* tmpcbar = NULL;
   string ibid = theDefaults->get_colorbarid();
   bool cbar_created = false;

   if (ibid.length() && cbars.count(ibid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,ibid.c_str(), "colorbar");
   else
   {
      // Find the coordinate system
      ChipsId tmpId (*cid);
      ChipsCoordinate *tmp_coord;
      unsigned int cmap_id = static_cast <unsigned short>(-1);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)   tmpId.set_coord_id ("");

      // if the coordsys is not set default to plot norm
      if (tmpId.get_coord_sys () == COORD_NONE)
         tmpId.set_coord_sys (PLOT_NORM);

      // This updates the chips id for the appropriate coord sys
      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (obj == chips_image)
      {
         ChipsImage *mapobj = NULL;
         mapobj = get_image(obj_id);

         if (!mapobj && obj_id != "")
            cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"colorbar"); 
         else if (mapobj)
            cmap_id = mapobj->get_colormap_id ();
      } 
/*      else if (obj == chips_contour)
      {
         ChipsContour *mapobj = NULL;
         mapobj = get_contour(obj_id);
         if (!mapobj && obj_id != "")
            cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"colorbar");
         else if (mapobj)
            cmap_id = mapobj->get_colormap_id ();
      } 
*/
      else if (obj == chips_not_object)
      {
         // first try getting the id as an image
         ChipsImage *img_obj = get_image(obj_id); 
         ChipsContour *ctr_obj = NULL;
         if (img_obj)
            cmap_id = img_obj->get_colormap_id ();
/*         else if ((ctr_obj = get_contour (obj_id)) != NULL)
            cmap_id = ctr_obj->get_colormap_id ();
*/
         if (!img_obj && !ctr_obj && (obj_id != ""))
            cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"colorbar");

      } 
      else  
      {
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"colorbar");
         return rval; 
      }   

      if (!rval && !cbar_created)
      {
         if (cmap_id != static_cast <unsigned short> (-1))
            tmpcbar = new_colorbar(tmp_coord, tmpId.get_coord_sys (), xx, yy, cmap_id, ibid);
         else
            tmpcbar = new_colorbar(tmp_coord, tmpId.get_coord_sys (), xx, yy, ibid);
      }
      if (tmpcbar && !rval)
      {
         if (!cmd_manager->cmd_lock ()){
            // if ibid is empty that means the ibr_cnt changed
            ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpcbar,(ibid.length()==0),currcbar);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }

         // new colorbar created- add to list and update current
         add_object_to_depth_table(tmpcbar);
         tmpcbar->set_cbar_num (cbar_cnt);
         currcbar = tmpcbar->get_id();
         cbars.insert(pair<string, ChipsColorBar*>(currcbar, tmpcbar));
         cid->set_colorbar_id(currcbar);
      }
      else if (!rval && !cbar_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"colorbar");
   }
   return rval;
}

int ChipsPlot::delete_colorbar(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsColorBar* ccbar;
   string ibid = cid->get_colorbar_id();
   if (ibid.length() == 0)   ibid = currcbar;

   if (chips_strcmp(ibid.c_str(), "all") == 0)
   {
      map<string, ChipsColorBar*>::iterator ii;

      if (cbars.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"colorbars");
         return rval;
      }

      // delete all colorbars in plot
      if (!force_del)
         cmd_manager->open_block ();
      for(ii=cbars.begin(); ii != cbars.end(); ++ii)
      {
         ccbar = (*ii).second;
         ccbar->disable_interaction ();
         if (depth_table)
         {
            if (!force_del)   remove_object_from_depth_table(ccbar);
            else              depth_table->remove_object(ccbar);
         }
	 if (force_del)
            ccbar->object_delete ();
	 else{
            // this hides the point without changing its visibility
            ccbar->force_hide ();
            dcbars[ccbar->get_ref_cnt ()] = ccbar;
            if (ccbar->get_chips_coord ())
            {
               ccbar->get_chips_coord ()->remove(ccbar);
               ccbar->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ())
            {
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ccbar, currcbar);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }
      if (!force_del) 
         cmd_manager->close_block ();
      else
      {
         map <unsigned int, ChipsColorBar*>::iterator  dcbar_ii;
         // If force deleteing all colorbars also remove previously delete icolorbars
         for (dcbar_ii = dcbars.begin (); dcbar_ii != dcbars.end (); dcbar_ii++)
            dcbar_ii->second->object_delete ();
         dcbars.clear ();

      }
      cbars.clear();
      currcbar = "";
      cid->set_colorbar_id("");
      rval = ERR_STATUS_GOOD;
      cbar_cnt = 0;
   }
   else if (cbars.count(ibid))
   {
      map<string, ChipsColorBar*>::iterator ii;
      ii = cbars.find(ibid);
      ccbar = (*ii).second;
      ccbar->disable_interaction ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(ccbar);
         else              depth_table->remove_object(ccbar);
      }
      cbars.erase(ii);
      if (force_del)
         ccbar->object_delete ();
      else
      {
         // this hides the colorbar without changing its visibility
         ccbar->force_hide ();
         dcbars[ccbar->get_ref_cnt ()] = ccbar;
         if (ccbar->get_chips_coord ())
         {
            ccbar->get_chips_coord ()->remove(ccbar);
            ccbar->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock ())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(ccbar, currcbar);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_colorbar_id("");
      if (!cbars.empty())
      {
         map<string, ChipsColorBar* >::reverse_iterator ri = cbars.rbegin();
         currcbar = (*ri).first;
      }
      else   currcbar = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbar",ibid.c_str());
   return rval;
}


int ChipsPlot::add_region(ChipsId* cid, double* xx, double* yy, 
   unsigned int count)
{
   int rval = ERR_STATUS_GOOD, rct = region_cnt;
   ChipsRegion* tmpreg = NULL;
   string rid = theDefaults->get_regionid();
   bool region_created = false;

   if (rid.length() && regions.count(rid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,rid.c_str(),"region");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)  tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !region_created)
         tmpreg = new_polygon(tmp_coord, tmpId.get_coord_sys (), xx, yy, count, rid);

      if (tmpreg && !rval)
      {
         if ((rval = tmpreg->get_status()) == ERR_STATUS_GOOD)
         {
            // if rid is empty that means the region_cnt changed
            if (!cmd_manager->cmd_lock())
            {
               ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpreg,(rid.length()==0),currregion);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
            // new region created- add to list and update current
            add_object_to_depth_table(tmpreg);
            tmpreg->set_region_num(region_cnt);
            currregion = tmpreg->get_id();
            regions.insert(pair<string, ChipsRegion*>(currregion, tmpreg));
            cid->set_region_id(currregion);
         }
         else if (region_cnt == rct + 1) 
            region_cnt = rct;  // region creation failed- reset id count 
      }
      else if (!(rval || region_created))
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"region");
   }
   return rval;
}

int ChipsPlot::add_region(ChipsId* cid, int sides, double xx, double yy, 
   double radius)
{
   int rval = ERR_STATUS_GOOD, rct = region_cnt;
   ChipsRegion* tmpreg = NULL;
   string rid = theDefaults->get_regionid();
   bool region_created = false;

   if (rid.length() && regions.count(rid))
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,rid.c_str(),"region");
   else
   {
      ChipsCoordinate *tmp_coord = NULL;
      ChipsId tmpId (*cid);

      // coord id should only be used for data coordinates
      if (tmpId.get_coord_sys () != DATA)
         tmpId.set_coord_id ("");

      tmp_coord = this->find_coordinate_system (tmpId, rval);

      if (!rval && !region_created)
         tmpreg = new_polygon(tmp_coord, tmpId.get_coord_sys (), sides, xx, yy, radius, rid);

      if (tmpreg && !rval)
      {
         if ((rval = tmpreg->get_status()) == ERR_STATUS_GOOD)
         {
            if (!cmd_manager->cmd_lock ())
            {
               // if rid is empty that means the region_cnt changed
               ChipsCreateCommand *cmd = new ChipsCreateCommand(tmpreg,
	          (rid.length()==0),currregion);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
            // new region created- add to list and update current
            add_object_to_depth_table(tmpreg);
            tmpreg->set_region_num(region_cnt);
            currregion = tmpreg->get_id();
            regions.insert(pair<string, ChipsRegion*>(currregion, tmpreg));
            cid->set_region_id(currregion);
         }
         else
         {
            if (region_cnt == rct + 1)  region_cnt = rct;  // reset id count
            tmpreg->object_delete ();
            tmpreg = NULL;
         }
      }
      else if (!rval && !region_created)
         cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"region");
   }
   return rval;
}

int ChipsPlot::delete_region(ChipsId* cid, bool force_del)
{
   int rval = ERR_STATUS_GOOD;
   ChipsRegion* cregion;
   string rid = cid->get_region_id();

   if (rid.length() == 0)   rid = currregion;

   if (chips_strcmp(rid.c_str(), "all") == 0)
   {
      map<string, ChipsRegion* >::iterator ii;

      if (regions.empty())
      {
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"regions");
         return rval;
      }

      // delete all regions in frame
      if (!force_del)
         cmd_manager->open_block ();
      for(ii = regions.begin(); ii != regions.end(); ++ii)
      {
         cregion = (*ii).second;
         cregion->unselect ();
         if (depth_table)
         {
            if (!force_del)   remove_object_from_depth_table(cregion);
            else              depth_table->remove_object(cregion);
         }
         if (force_del)
            cregion->object_delete ();
         else 
         {
            // this hides the region without changing its visibility
            cregion->force_hide();
            dregions[cregion->get_ref_cnt ()] = cregion;
	    if (cregion->get_chips_coord ())
            {
               cregion->get_chips_coord ()->remove (cregion);
               cregion->set_chips_coord (NULL);
            }
            if (!cmd_manager->cmd_lock ())
            {
               ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cregion,currregion);
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  delete cmd;
            }
         }
      }
      if (!force_del) 
         cmd_manager->close_block();
      else 
      {
         // If force deleteing all regions also remove previously delete regions
         map <unsigned int, ChipsRegion*>::iterator dregion_ii;
         for (dregion_ii = dregions.begin (); dregion_ii != dregions.end (); dregion_ii++)
            dregion_ii->second->object_delete ();
         dregions.clear ();
      }
      regions.clear();
      currregion = "";
      cid->set_region_id("");
      rval = ERR_STATUS_GOOD;
      region_cnt = 0;
   }
   else if (regions.count(rid))
   {
      map<string, ChipsRegion* >::iterator ii;
      ii = regions.find(rid);
      cregion = (*ii).second;
      cregion->unselect ();
      if (depth_table)
      {
         if (!force_del)   remove_object_from_depth_table(cregion);
         else              depth_table->remove_object(cregion);
      }
      regions.erase(ii);
      if (force_del)
         cregion->object_delete ();
      else
      {
         // this hides the region without changing its visibility
         cregion->force_hide();
         dregions[cregion->get_ref_cnt ()] = cregion;
         if (cregion->get_chips_coord ())
         {
            cregion->get_chips_coord ()->remove (cregion);
            cregion->set_chips_coord (NULL);
         }
         if (!cmd_manager->cmd_lock())
         {
            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cregion,currregion);
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               delete cmd;
         }
      }
      cid->set_region_id("");
      if (!regions.empty())
      {
         map<string, ChipsRegion* >::reverse_iterator ri = regions.rbegin();
         currregion = (*ri).first;
      }
      else   currregion = "";
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",rid.c_str());
   return rval;
}

int ChipsPlot::move_colorbar(ChipsId* cid, double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD;
   string oid = cid->get_colorbar_id();
   string  cd_id = cid->get_coord_id ();
   COORD_SYS  cd_sys = cid->get_coord_sys ();

   if (oid.length() == 0)  oid = currcbar;
   if (cbars.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
      return ERR_OBJ_LIST_EMPTY;
   }
   map <string, ChipsColorBar*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts, rval = ERR_STATUS_GOOD;
      for(ii=cbars.begin(); ii != cbars.end(); ++ii)
        if ((sts |= (*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
             ERR_STATUS_GOOD)   rval = sts;
   }
   else if (cbars.count(oid))
      rval = cbars[oid]->move(cd_sys, cd_id, xx, yy, mode);
   else
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"colorbars",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   currcbar = oid;
   return rval;
}

int ChipsPlot::move_point(ChipsId* cid, double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD;
   string oid = cid->get_point_id();
   string  cd_id = cid->get_coord_id ();
   COORD_SYS  cd_sys = cid->get_coord_sys ();

   if (oid.length() == 0)  oid = currpoint;
   if (points.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"points");
      return ERR_OBJ_LIST_EMPTY;
   }
   map <string, ChipsPoint*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts, rval = ERR_STATUS_GOOD;
      for(ii=points.begin(); ii != points.end(); ++ii)
        if ((sts |= (*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
             ERR_STATUS_GOOD)   rval = sts;
   }
   else if (points.count(oid))
      rval = points[oid]->move(cd_sys, cd_id, xx, yy, mode);
   else
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"point",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   currpoint = oid;
   return rval;
}

int ChipsPlot::move_region(ChipsId* cid, double xx, double yy, int mode)
{
   int rval = ERR_STATUS_GOOD;
   string oid = cid->get_region_id();
   string  cd_id = cid->get_coord_id ();
   COORD_SYS  cd_sys = cid->get_coord_sys ();

   if (oid.length() == 0)  oid = currregion;
   if (regions.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"regions");
      return ERR_OBJ_LIST_EMPTY;
   }
   map <string, ChipsRegion*>::iterator ii;
   if (chips_strcmp(oid.c_str(), "all") == 0)
   {
      int sts, rval = ERR_STATUS_GOOD;
      for(ii=regions.begin(); ii != regions.end(); ++ii)
        if ((sts |= (*ii).second->move(cd_sys, cd_id, xx, yy, mode)) !=
             ERR_STATUS_GOOD)   rval = sts;
   }
   else if (regions.count(oid))
      rval = regions[oid]->move(cd_sys, cd_id, xx, yy, mode);
   else
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"region",oid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   currregion = oid;
   return rval;
}

int ChipsPlot::set_current_colorbar(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_colorbar_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!cbars.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "colorbar", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currcbar, lid, chips_colorbar);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcbar = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currcbar,
                                                     lid, chips_colorbar);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currcbar = "all";
      }
   }
   return rval;
}

int ChipsPlot::set_current_point(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_point_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!points.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "point", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currpoint, lid, chips_point);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currpoint = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currpoint,
                                                      lid, chips_point);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currpoint = "all";
      }
   }
   return rval;
}

int ChipsPlot::set_current_region(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_region_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!regions.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "region", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currregion, lid, chips_region);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currregion = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currregion,
                                                      lid, chips_region);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currregion = "all";
      }
   }
   return rval;
}

ChipsImage* ChipsPlot::new_image(ChipsStateStream& in)
{
   return new ChipsImage(this, in);
}

ChipsImage* ChipsPlot::new_image(string fname, string transform_id, string iid)
{
   string tmpid;
   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();
   return new ChipsImage(this, fname, transform_id, tmpid);
}

ChipsImage* ChipsPlot::new_image( string fname, float posx1, float posy1, 
    float posx2, float posy2, string transform_id, string iid)
{
   string tmpid;
   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();
   return new ChipsImage(this,fname,posx1,posy1,posx2,posy2,transform_id,tmpid);
}

ChipsImage* ChipsPlot::new_image( double* rvals, double *gvals, double *bvals,
    double *avals, int xaxis, int yaxis, string transform_id, string iid)
{
   string tmpid;
   if (iid.length())
      tmpid = iid;
   else
      tmpid = create_image_id();
   return new ChipsImage(this, xaxis, yaxis, rvals, gvals, bvals, avals, 
                          transform_id, tmpid);
}

char* ChipsPlot::create_image_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_imagestem();
   else            stem = "img";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++image_cnt);

   // increment count til next unused image id
   while(get_image(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++image_cnt);
   }
   return temp_id;
}

ChipsImage* ChipsPlot::get_image(const string& val)
{
   string imd = val;
   if (!val.length())  imd = currimage;
   if ((chips_strcmp(imd.c_str(), "all") == 0) && images.size() == 1)
   {
      map<string, ChipsImage*>::iterator ii = images.begin();
      return ((*ii).second);
   }
   else if ((chips_strcmp(imd.c_str(), "all") != 0) && images.count(imd))
      return images[imd];
   return NULL;
}

int ChipsPlot::set_current_image(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (cid)
   {
      string lid = cid->get_image_id();

      if (chips_strcmp(lid.c_str(), "all") != 0)
      {
	 if (!images.count(lid))
         {
            cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "image", lid.c_str());
            return ERR_OBJ_VAL_DNE;
         }
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
               currimage, lid, chips_image);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currimage = lid; 
      }
      else
      {
         if (rec)
         {
            ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this, currimage,
                                                      lid, chips_image);
            if ((rval = cmd_manager->add_cmd(cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
         currimage = "all";
      }
   }
   return rval;
}

ChipsColorBar* ChipsPlot::new_colorbar(ChipsCoordinate *inCoord,
   COORD_SYS sys,  double xpos, double ypos, string tid)
{
   string tmpid;
   if (tid.length())
      tmpid = tid;
   else
       tmpid = create_cbar_id();
   return new ChipsColorBar(this, inCoord, sys, xpos, ypos, tmpid);
}

ChipsColorBar* ChipsPlot::new_colorbar(ChipsCoordinate *inCoord,
   COORD_SYS sys,  double xpos, double ypos, unsigned short cmap_id, string tid)
{
   string tmpid;
   if (tid.length())
      tmpid = tid;
   else
       tmpid = create_cbar_id();
   return new ChipsColorBar(this, inCoord, sys, xpos, ypos, cmap_id, tmpid);
}

ChipsColorBar* ChipsPlot::new_colorbar(ChipsStateStream& in)
{
   return new ChipsColorBar(this, coord, coord_system, in);
}

ChipsPoint* ChipsPlot::new_point(ChipsCoordinate *inCoord, COORD_SYS sys,
   double xx, double yy, string ptid)
{
   string tmpid;
   if (ptid.length())
      tmpid = ptid;
   else
       tmpid = create_point_id();
   return new ChipsPoint(this, inCoord, sys, xx, yy, tmpid);
}

ChipsPoint* ChipsPlot::new_point(ChipsCoordinate *inCoord, COORD_SYS sys,
   double xx, double yy, double zz, string ptid)
{
   string tmpid;
   if (ptid.length())
      tmpid = ptid;
   else
       tmpid = create_point_id();
   return new ChipsPoint(this, inCoord, sys, xx, yy, zz, tmpid);
}

ChipsPoint* ChipsPlot::new_point(ChipsStateStream& in)
{
   return new ChipsPoint(this, coord, coord_system, in);
}

char* ChipsPlot::create_cbar_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_colorbarstem();
   else            stem = "cbar";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++cbar_cnt);

   // increment count til next unused colorbar id
   while(get_colorbar(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++cbar_cnt);
   }
   return temp_id;
}

ChipsColorBar* ChipsPlot::get_colorbar(const string& val)
{
   string ibid=val;
   if (!val.length())  ibid = currcbar;
   if ((chips_strcmp(ibid.c_str(), "all") == 0) && cbars.size() == 1)
   {
      map<string, ChipsColorBar*>::iterator ii = cbars.begin();
      return ((*ii).second);
   }
   else if (cbars.count(ibid))
      return cbars[ibid];
   else
      return NULL;
}

char* ChipsPlot::create_point_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_pointstem();
   else            stem = "pnt";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++point_cnt);

   // increment count til next unused point id
   while(get_point(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++point_cnt);
   }
   return temp_id;
}

ChipsPoint* ChipsPlot::get_point(const string& val)
{
   string ptid=val;
   if (!val.length())  ptid = currpoint;
   if ((chips_strcmp(ptid.c_str(), "all") == 0) && points.size() == 1)
   {
      map<string, ChipsPoint*>::iterator ii = points.begin();
      return ((*ii).second);
   }
   else if (points.count(ptid))
      return points[ptid];
   else
      return NULL;
}

ChipsRegion* ChipsPlot::new_polygon(ChipsCoordinate *inCoord, COORD_SYS sys,
   double* xx, double* yy, unsigned int count, string rid)
{
   string tmpid;
   if (rid.length())
      tmpid = rid;
   else
      tmpid = create_region_id();
   return new ChipsPolygon(this, inCoord, sys, xx, yy, count, tmpid);
}

ChipsRegion* ChipsPlot::new_polygon(ChipsCoordinate *inCoord, COORD_SYS sys,
   int sides, double xx, double yy, double radius, string rid)
{
   string tmpid;
   if (rid.length())
      tmpid = rid;
   else
      tmpid = create_region_id();
   return new ChipsPolygon(this, inCoord, sys, sides, xx, yy, radius, tmpid);
}

ChipsRegion* ChipsPlot::new_polygon(ChipsStateStream& in)
{
   return new ChipsPolygon(this, coord, coord_system, in);
}

char* ChipsPlot::create_region_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_regionstem();
   else            stem = "reg";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++region_cnt);

   // increment count til next unused region id
   while(get_region(temp_id))
   {
      sprintf(temp_id, "%s%1d", stem.c_str(), ++region_cnt);
   }
   return temp_id;
}

ChipsRegion* ChipsPlot::get_region(const string& val)
{
   string rid = val;
   if (!val.length())  rid = currregion;
   if ((chips_strcmp(rid.c_str(), "all") == 0) && regions.size() == 1)
   {
      map<string, ChipsRegion*>::iterator ii = regions.begin();
      return ((*ii).second);
   }
   else if (regions.count(rid))
      return regions[rid];
   else
      return NULL;
}

int ChipsPlot::set_colorbar_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string ibid = cid->get_colorbar_id();
   if (ibid.length() == 0)  ibid = currcbar;

   if (cbars.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(ibid.c_str(), "all") == 0)
   {
      map<string, ChipsColorBar*>::iterator ii;
      for(ii=cbars.begin(); ii != cbars.end(); ++ii)
         rval |= (*ii).second->set_property(in_val, in_property);
   }
   else
   {
      if (cbars.count(ibid) == 0)
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"colorbar",ibid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ChipsColorBar* ib = cbars[ibid];
      rval = ib->set_property(in_val, in_property);
   }
   if (rval == ERR_STATUS_GOOD)
   {
      currcbar = ibid;
      cid->set_colorbar_id(currcbar);
   }
   return rval;
}

int ChipsPlot::get_colorbar_property(ChipsId* cid, void* out_val, int property)
{
   string ibid = cid->get_colorbar_id();
   if (!ibid.length())  ibid = currcbar;
   if (cbars.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"colorbars");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(ibid.c_str(), "all") == 0)
   {
      if (cbars.size() == 1)
      {
         map<string, ChipsColorBar*>::iterator ii = cbars.begin();
         return ((*ii).second)->get_property(out_val, property);
      }
      cmd_manager->add_err(ERR_INVALID_GET_ALL);
      return ERR_INVALID_GET_ALL;
   }
   else if (cbars.count(ibid) == 0)
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"colorbar",ibid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   return (cbars[ibid])->get_property(out_val, property);
}

int ChipsPlot::set_point_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string ptid = cid->get_point_id();
   if (ptid.length() == 0)  ptid = currpoint;

   if (points.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"points");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(ptid.c_str(), "all") == 0)
   {
      map<string, ChipsPoint*>::iterator ii;
      for(ii=points.begin(); ii != points.end(); ++ii)
         rval |= (*ii).second->set_property(in_val, in_property);
   }
   else
   {
      if (points.count(ptid) == 0)
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"point",ptid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ChipsPoint* pnt = points[ptid];
      rval = pnt->set_property(in_val, in_property);
   }
   if (rval == ERR_STATUS_GOOD)
   {
      currpoint = ptid;
      cid->set_point_id(currpoint);
   }
   return rval;
}

int ChipsPlot::get_point_property(ChipsId* cid, void* out_val, int property)
{
   string ptid = cid->get_point_id();
   if (!ptid.length())  ptid = currpoint;
   if (points.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"points");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(ptid.c_str(), "all") == 0)
   {
      if (points.size() == 1)
      {
         map<string, ChipsPoint*>::iterator ii = points.begin();
         return ((*ii).second)->get_property(out_val, property);
      }
      cmd_manager->add_err(ERR_INVALID_GET_ALL);
      return ERR_INVALID_GET_ALL;
   }
   else if (points.count(ptid) == 0)
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"point",ptid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   return (points[ptid])->get_property(out_val, property);
}

int ChipsPlot::set_region_property(ChipsId* cid, void* in_val, int in_property)
{
   int rval = ERR_STATUS_GOOD;
   string rid = cid->get_region_id();

   if (rid.length() == 0)  rid = currregion;
   if (regions.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"regions");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(rid.c_str(), "all") == 0)
   {
      map<string, ChipsRegion*>::iterator ii;
      rval = ERR_STATUS_GOOD;
      for(ii=regions.begin(); ii != regions.end(); ++ii)
         rval |= (*ii).second->set_property(in_val, in_property);
   }
   else
   {
      if (regions.count(rid) == 0)
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"region",rid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      ChipsRegion* reg = regions[rid];
      rval = reg->set_property(in_val, in_property);
   }
   if (rval == ERR_STATUS_GOOD)
   {
      currregion = rid;
      cid->set_region_id(currregion);
   }
   return rval;
}

int ChipsPlot::get_region_property(ChipsId* cid, void* out_val, int property)
{
   string rid = cid->get_region_id();
   if (rid.length() == 0)  rid = currregion;
   if (regions.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"regions");
      return ERR_OBJ_LIST_EMPTY;
   }
   else if (chips_strcmp(rid.c_str(), "all") == 0)
   {
      if (regions.size() == 1)
      {
         map<string, ChipsRegion*>::iterator ii = regions.begin();
         return ((*ii).second)->get_property(out_val, property);
      }
      cmd_manager->add_err(ERR_INVALID_GET_ALL);
      return ERR_INVALID_GET_ALL;
   }
   else if (!regions.count(rid))
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"region",rid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   return  (regions[rid])->get_property(out_val, property);
}

void ChipsPlot::adjust_plot_anchors()
{
   bx1 = left_margin;
   by1 = bottom_margin;
   bx2 = 1.0 - right_margin;
   by2 = 1.0 - top_margin; 

   if (coord)
   {
      ((ChipsPlotSystem *)coord)->set_min_pos(ChipsVector(bx1, by1));
      ((ChipsPlotSystem *)coord)->set_max_pos(ChipsVector(bx2, by2));

      ChipsVector in0(bx1, by1);
      ChipsVector in1(bx2, by2);
      ChipsVector out0 (0.0, 0.0);
      ChipsVector out1 (0.0, 0.0);
      coord->convert_cs_to_cs(&in0, &out0, 1, FRAME_NORM, "", PIXEL, "");
      coord->convert_cs_to_cs(&in1, &out1, 1, FRAME_NORM, "", PIXEL, "");
      px0 = out0[0];
      py0 = out0[1];
      px1 = out1[0];
      py1 = out1[1]; 
   }
}

int ChipsPlot::recalculate_position()
{
   ChipsVector in0(bx1, by1);
   ChipsVector in1(bx2, by2);
   ChipsVector out0 (0.0, 0.0);
   ChipsVector out1 (0.0, 0.0);

   if (!coord)  return ERR_STATUS_BAD; 
   
   coord->convert_cs_to_cs(&in0, &out0, 1, FRAME_NORM, "", PIXEL, "");
   coord->convert_cs_to_cs(&in1, &out1, 1, FRAME_NORM, "", PIXEL, "");
   double ht   = (out1[1] - out0[1]),
          wt   = (out1[0] - out0[0]), 
          ctrx = (out1[0] + out0[0])/2.0,
          ctry = (out1[1] + out0[1])/2.0;
   double distxo, distyo, distxn, distyn, deltax, deltay;

   // don't try to recalculate position if window dimensions are too small
   if (ht < 5 || wt < 5)  return ERR_STATUS_BAD;

   distxn = out1[0] - out0[0];
   distyn = out1[1] - out0[1];
   if (px0 == 0 && px1 == 0)    distxo = distxn;
      else                      distxo = px1 - px0;
   if (py0 == 0 && py1 == 0)    distyo = distyn;
      else                      distyo = py1 - py0;
   deltax = distxn - distxo;
   deltay = distyn - distyo; 

   // store new corner pixel locations
   px0 = out0[0];
   py0 = out0[1];
   px1 = out1[0];
   py1 = out1[1]; 

   // handle plot aspect ratio 
   if (aspect_use && aspect_w && aspect_h)
   {
      double factor, half;
      if (wt < 0.0) wt *= -1.0;
      if (ht < 0.0) ht *= -1.0;
      wt /= aspect_w;
      ht /= aspect_h;
      if (wt < ht) factor = wt;
         else factor = ht;
      half = factor * aspect_w / 2.0;
      out0[0] = ctrx - half;
      out1[0] = ctrx + half;
      half = factor * aspect_h / 2.0;
      out0[1] = ctry - half;
      out1[1] = ctry + half;
      coord->convert_cs_to_cs(&out0, &in0, 1, PIXEL, "", FRAME_NORM, "");
      coord->convert_cs_to_cs(&out1, &in1, 1, PIXEL, "", FRAME_NORM, "");
      bx1 = in0[0];
      by1 = in0[1];
      bx2 = in1[0];
      by2 = in1[1];
   
      ((ChipsPlotSystem *)coord)->modify_position(bx1, bx2, X_AXIS);
      ((ChipsPlotSystem *)coord)->modify_position(by1, by2, Y_AXIS);
   }

   // handle data aspect 
   double ln, lx, mm,nn;

   coord->convert_cs_to_cs(&in0,&out0,1,FRAME_NORM,"",PIXEL,"");
   coord->convert_cs_to_cs(&in1,&out1,1,FRAME_NORM,"",PIXEL,"");
   px0 = out0[0];
   py0 = out0[1];
   px1 = out1[0];
   py1 = out1[1];

   ChipsAxis *tmpax, *tmpay;
   list <string> data_coords;
   ChipsDataSystem *tmp_coord;
   ChipsId  chid; 
   string axid; 

   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;

   for(axes_map::iterator am = axes_y.begin(); am != axes_y.end(); ++am) 
   {
      tmpay = (*am).second;
      // loop through all data systems and find ones related to this axis
      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         if (ds_iter->first.second != tmpay->get_id ())
               continue;
         tmp_coord = ds_iter->second;
         if (tmp_coord->use_ratio() && !aspect_use)
         {
            axid = tmp_coord->get_xaxis_id(); 
            tmpax = get_xaxis(axid); 
            double ph = tmp_coord->get_unit_pix_h(); 
            double pw = tmp_coord->get_unit_pix_w(); 
            tmpax->get_range(mm,nn);
            tmpay->get_range(ln,lx);
            list<ChipsAxis*>::iterator jj;

            ChipsVector min_pnt(mm, ln);
            ChipsVector max_pnt(nn, lx);
            string ds_id = tmp_coord->get_id_no_wcs ();

            tmp_coord->convert_cs_to_cs (&min_pnt, &min_pnt, 1, DATA, "",  DATA, ds_id);
            tmp_coord->convert_cs_to_cs (&max_pnt, &max_pnt, 1, DATA, "",  DATA, ds_id);

            min_pnt[0] = min_pnt[0]-deltax/pw/2;
            min_pnt[1] = min_pnt[1]-deltay/ph/2;
            max_pnt[0] = max_pnt[0]+deltax/pw/2;
            max_pnt[1] = max_pnt[1]+deltay/ph/2;

//            if (deltay == 0 && deltax != 0)
            if ( flt_equal(deltay, 0) &&
                !flt_equal (deltax, 0))
            {
               if (ph) bindings->update_connected_nodes(tmpay, min_pnt[1],
                          max_pnt[1], false, AUTOLIM_OFF,1,false);
               if (pw) bindings->update_connected_nodes(tmpax,min_pnt[0],
                          max_pnt[0], false, AUTOLIM_OFF,1,false);
            }
//            else if (deltax == 0 && deltay != 0)
            else if (flt_equal (deltax, 0) &&
                !flt_equal (deltay, 0))
            {
               if (pw) bindings->update_connected_nodes(tmpax,min_pnt[0],
                          max_pnt[0], false, AUTOLIM_OFF,1,false);
               if (ph) bindings->update_connected_nodes(tmpay, min_pnt[1],
                          max_pnt[1], false, AUTOLIM_OFF,1,false);
            }
            else
            {
               if (ph)  bindings->update_connected_nodes(tmpay, min_pnt[1],
                          max_pnt[1], false, AUTOLIM_OFF,1,false); 
               if (pw)  bindings->update_connected_nodes(tmpax,min_pnt[0],
                          max_pnt[0], false, AUTOLIM_OFF,1,false);
	    }
	 }
      }
   }
   ((ChipsPlotSystem *)coord)->modify_position(bx1, bx2, X_AXIS);
   ((ChipsPlotSystem *)coord)->modify_position(by1, by2, Y_AXIS);
   update_object_positions();

   return 0;
}

bool ChipsPlot::check_axes_for_ratios(ChipsId* cid)
{
   string tmpId;
   list <string> data_coords;
   list <string>::iterator ds_ii;
   ChipsAxis* tmpAxisPtr; 
   if (cid)  tmpId = cid->get_xaxis_id();
   if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
       ((tmpId.length () == 0) && cur_xall))
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;

      // loop through all data systems and find ones related to this axis
      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
         if (ds_iter->second->use_ratio ())
            return true;
   }
   else if ((tmpAxisPtr=get_xaxis(tmpId.c_str())))
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;
      string axis_id = tmpAxisPtr->get_id ();
      // loop through all data systems and find ones related to this axis
      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
         if ((ds_iter->first.first == axis_id) &&
             (ds_iter->second->use_ratio ()))
               return true;
   }
   if (cid)  tmpId = cid->get_yaxis_id();
   if ((chips_strcmp(tmpId.c_str(), "all") == 0) ||
       ((tmpId.length () == 0) && cur_yall))
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;

      // loop through all data systems
      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
         if (ds_iter->second->use_ratio ())
            return true;
   }
   else if ((tmpAxisPtr=get_yaxis(tmpId.c_str())))
   {
      map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;
      string axis_id = tmpAxisPtr->get_id ();
      // loop through all data systems and find ones related to this axis
      for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
         if ((ds_iter->first.second == axis_id) &&
             (ds_iter->second->use_ratio ()))
               return true;
   }
   return false; 
}

double ChipsPlot::get_plot_width()
{
   if (px1 > px0)   return (px1 - px0); 
   return (px0 - px1); 
}

double ChipsPlot::get_plot_height()
{
   if (py1 > py0)   return (py1 - py0); 
   return (py0 - py1); 
}

double ChipsPlot::get_scale_factor(int dim)
{
   double deltax = px1 - px0,     // plot width in pixels
          deltay = py1 - py0;     // plot height in pixels
   if (dim == X_AXIS)
      return (deltax != 0)? deltay/deltax : 1.0;
   else 
      return (deltay != 0)? deltax/deltay : 1.0;
}

int ChipsPlot::update_data_coord_limits (ChipsAxis *xaxis, double xmin, double xmax,
                                      ChipsAxis *yaxis, double ymin, double ymax,
                                      bool transformed)
{
   int rval = ERR_STATUS_GOOD;
   list <ChipsAxis *> xaxis_list;
   list <ChipsAxis *> yaxis_list;
   list <ChipsDataSystem *> ds_list;
   map <pair <string, string>,  ChipsDataSystem *>::iterator ds_iter;
   ChipsDataSystem *ds;
   ChipsAxis *tmp_xaxis;
   ChipsAxis *tmp_yaxis;

   if (!xaxis && !yaxis)
      return ERR_STATUS_BAD;


   if (xaxis && yaxis)
   {
      string xaxis_id = xaxis->get_id ();
      string yaxis_id = yaxis->get_id ();
      pair <string, string> p(xaxis_id, yaxis_id);

      ds_list.push_back (this->get_datasystem (xaxis, yaxis));
      xaxis_list.push_back (xaxis);
      yaxis_list.push_back (yaxis);
   }
   // find all y axis associated with this xaxis
   else if (xaxis)
   {
      string xaxis_id = xaxis->get_id ();
      xaxis_list.push_back (xaxis);
      for (ds_iter = data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         // create a list of y axes and data systems
         if (ds_iter->first.first == xaxis_id)
         {
            yaxis_list.push_back (get_yaxis (ds_iter->first.second));
            ds_list.push_back (ds_iter->second);
         }
      }
   }
   // find all x axis associated with this yaxis
   else if (yaxis)
   {
      string yaxis_id = yaxis->get_id ();
      yaxis_list.push_back (yaxis);
      for (ds_iter = data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         // create a list of y axes and data systems
         if (ds_iter->first.second == yaxis_id)
         {
            xaxis_list.push_back (get_xaxis (ds_iter->first.first));
            ds_list.push_back (ds_iter->second);
         }
      }
   }

   list <ChipsDataSystem *>::iterator ds_list_iter;
   list <ChipsAxis *>::iterator xaxis_iter = xaxis_list.begin ();
   list <ChipsAxis *>::iterator yaxis_iter = yaxis_list.begin ();

   ChipsVector min (xmin, ymin);
   ChipsVector max (xmax, ymax);
   string transform_id = "";

   if (!transformed)
      transform_id = "none";

   // loop through all data systems found and update them
   for (ds_list_iter = ds_list.begin (); ds_list_iter != ds_list.end (); ds_list_iter++)
   {
      ds = *ds_list_iter;

      tmp_xaxis = *xaxis_iter;
      tmp_yaxis = *yaxis_iter;

      min = ds->get_lower_bounds (transform_id);
      max = ds->get_upper_bounds (transform_id);
      
      if (tmp_xaxis && (tmp_xaxis == xaxis))
      {
         min[0] = xmin;
         max[0] = xmax;
      }
	
      if (tmp_yaxis && (tmp_yaxis == yaxis))
      {
         min[1] = ymin;
         max[1] = ymax;
      }

      ds->modify_limits (min, max, transform_id);

      if (tmp_xaxis)
      {
         tmp_xaxis->update_tick_limits ();
         tmp_xaxis->transform_tick_limits ();
      }
      if (tmp_yaxis)
      {
         tmp_yaxis->update_tick_limits ();
         tmp_yaxis->transform_tick_limits ();
      }

      if (xaxis_list.size () > 1)
         xaxis_iter++;
      if (yaxis_list.size () > 1)
         yaxis_iter++;
   }
   return rval;
}

int ChipsPlot::update_data_coord_positions (ChipsAxis *xaxis, ChipsAxis *yaxis, list<string> &exclude_axes)
{
   // *** update_data_coord_limits must have been called before this function ***

   int rval = ERR_STATUS_GOOD;
   list <ChipsAxis *> xaxis_list;
   list <ChipsAxis *> yaxis_list;
   list <ChipsDataSystem *> ds_list;
   map <pair <string, string>,  ChipsDataSystem *>::iterator ds_iter;
   ChipsDataSystem *ds;
   ChipsAxis *tmp_xaxis;
   ChipsAxis *tmp_yaxis;

   if (!xaxis && !yaxis)
      return ERR_STATUS_BAD;

   if (xaxis && yaxis)
   {
      pair <string, string> p(xaxis->get_id (), yaxis->get_id ());

      ds_list.push_back (this->data_systems[p]);
      xaxis_list.push_back (xaxis);
      yaxis_list.push_back (yaxis);
   }
   // find all y axis associated with this xaxis
   else if (xaxis)
   {
      string axis_id = xaxis->get_id ();

      xaxis_list.push_back (xaxis);
      for (ds_iter = data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         // create a list of y axes and data systems
         if (ds_iter->first.first == axis_id)
         {
            list<string>::iterator ex_iter = find (exclude_axes.begin (), exclude_axes.end (), ds_iter->first.second);

            if (ex_iter == exclude_axes.end ())
            {
               yaxis_list.push_back (get_yaxis (ds_iter->first.second));
               ds_list.push_back (ds_iter->second);
            }
         }
      }
   }
   // find all x axis associated with this yaxis
   else if (yaxis)
   {
      string axis_id = yaxis->get_id ();

      yaxis_list.push_back (yaxis);
      for (ds_iter = data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
      {
         // create a list of y axes and data systems
         if (ds_iter->first.second == axis_id)
         {
            list<string>::iterator ex_iter = find (exclude_axes.begin (), exclude_axes.end (), ds_iter->first.first);

            if (ex_iter == exclude_axes.end ())
            {
               xaxis_list.push_back (get_xaxis (ds_iter->first.first));
               ds_list.push_back (ds_iter->second);
            }
         }
      }
   }

   list <ChipsDataSystem *>::iterator ds_list_iter;
   list <ChipsAxis *>::iterator xaxis_iter = xaxis_list.begin ();
   list <ChipsAxis *>::iterator yaxis_iter = yaxis_list.begin ();
   double xmin = 0;
   double xmax = 1;
   double ymin = 0;
   double ymax = 1; 
   // loop through all data systems found and update them
   for (ds_list_iter = ds_list.begin (); ds_list_iter != ds_list.end (); ds_list_iter++)
   {
      ds = *ds_list_iter;

      tmp_xaxis = *xaxis_iter;
      tmp_yaxis = *yaxis_iter;

/*      if (tmp_xaxis)
      {
         xmin = tmp_xaxis->get_absolute_min ();
         xmax = tmp_xaxis->get_absolute_max ();
         transform_id = transformed ? "" : "none";
      }

      if (tmp_yaxis)
      {
         ymin = tmp_yaxis->get_absolute_min ();
         ymax = tmp_yaxis->get_absolute_max ();
         transform_id = transformed ? "" : "none";
      }
*/
      ChipsVector min (xmin, ymin);
      ChipsVector max (xmax, ymax);

/*      ds->modify_limits (min, max, "none");
*/

      if (tmp_xaxis)
      {
         tmp_xaxis->update_tick_limits ();
         tmp_xaxis->transform_tick_limits ();
         min[0] = tmp_xaxis->get_min (false);
         max[0] = tmp_xaxis->get_max (false);
      }
      if (tmp_yaxis)
      {
         tmp_yaxis->update_tick_limits ();
         tmp_yaxis->transform_tick_limits ();
         min[1] = tmp_yaxis->get_min (false);
         max[1] = tmp_yaxis->get_max (false);
      }

      ds->modify_limits (min, max, "none");


      // update the tick positions after all the limits have been sorted out
      if (tmp_xaxis)
      {
         tmp_xaxis->transform_tick_limits ();
         tmp_xaxis->update_tick_positions ();
      }
      if (tmp_yaxis)
      {
         tmp_yaxis->transform_tick_limits ();
         tmp_yaxis->update_tick_positions ();
      }

      ds->update_obj_positions ();
   
      if (xaxis_list.size () > 1)
         xaxis_iter++;
      if (yaxis_list.size () > 1)
         yaxis_iter++;
   }
   return rval;
}

void ChipsPlot::remove_data_coord (ChipsAxis *ax, ChipsAxis *ay)
{
   string xaxis_id = "", yaxis_id = "";
   pair <string, string> axis_pair;
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter; 
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter_tmp;

   if ((ax == NULL) && (ay == NULL))  return;

   if (ax)
      xaxis_id = ax->get_id ();
   if (ay)
      yaxis_id = ay->get_id ();
  
   axis_pair.first = xaxis_id;
   axis_pair.second = yaxis_id;

   if (ax)
   {
      // loop through all data systems and remove ones that are associated
      // with this axis
      ds_iter = data_systems.begin ();
      axis_pair.first = "";
      while (ds_iter != data_systems.end ())
      {
         if (ds_iter->first.first == xaxis_id)
         {
            bool add_single = true;
            ds_iter_tmp = ds_iter++;

            // if there's a second axis add it back as a single axis 
            // coordinate system
            if (ds_iter_tmp->first.second.length () > 0)
            {
               map <pair <string, string>, ChipsDataSystem *>::iterator iter;
               for (iter=data_systems.begin (); iter!=data_systems.end (); iter++)
               {
                  if (ds_iter_tmp->first.second == iter->first.second)
                     if (iter->first.first != xaxis_id)
                     {
                        add_single = false;
                        break;
                     }
               }
            
               if (add_single)
               {
                  axis_pair.second = ds_iter_tmp->first.second;
                  ds_iter_tmp->second->set_xaxis_id ("");
                  data_systems[axis_pair] = ds_iter_tmp->second;
               }
               else
               {
                  ChipsAxis *tmp_ax = this->get_yaxis (ds_iter_tmp->first.second);
                  tmp_ax->remove_data_coord_id (ds_iter_tmp->second->get_id ());
               }
            }   
            data_systems.erase (ds_iter_tmp);
            ax->remove_data_coord_id (ds_iter_tmp->second->get_id ());
         }
         else
            ds_iter++;
      }
   }

   axis_pair.first = xaxis_id;
   axis_pair.second = yaxis_id;

   if (ay)
   {
      // loop through all data systems and remove ones that are associated
      // with this axis
      ds_iter = data_systems.begin ();
      axis_pair.second = "";
      while (ds_iter != data_systems.end ())
      {
         if (ds_iter->first.second == yaxis_id)
         {
            bool add_single = true;
            ds_iter_tmp = ds_iter++;

            if (ds_iter_tmp->first.first.length () > 0)
            {
               map <pair <string, string>, ChipsDataSystem *>::iterator iter;
               for (iter=data_systems.begin (); iter!=data_systems.end (); iter++)
               {
                  if (ds_iter_tmp->first.first == iter->first.first)
                     if (iter->first.second != yaxis_id)
                     {
                        add_single = false;
                        break;
                     }
               }

               if (add_single)
               {
                  axis_pair.first = ds_iter_tmp->first.first;
                  ds_iter_tmp->second->set_yaxis_id ("");
                  data_systems[axis_pair] = ds_iter_tmp->second;
               }
               else
               {
                  ChipsAxis *tmp_ax = this->get_xaxis (ds_iter_tmp->first.second);
                  tmp_ax->remove_data_coord_id (ds_iter_tmp->second->get_id ());
               }

            }
            data_systems.erase (ds_iter_tmp);
            ay->remove_data_coord_id (ds_iter_tmp->second->get_id ());
         }
         else
            ds_iter++;
      }
   }
}

bool ChipsPlot::check_for_data_sys_objs (ChipsAxis *ax)
{
   if (!ax)  return false;
  
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter;
   string axis_id = ax->get_id ();

   // loop through all data systems and find ones related to this axis
   for (ds_iter=data_systems.begin (); ds_iter!=data_systems.end (); ds_iter++)
   {
      if ((ds_iter->first.first == axis_id) ||
          (ds_iter->first.second == axis_id))
         if (ds_iter->second->get_num_objs () > 0)
            return true;
   }
   return false;
}

ChipsDataSystem* ChipsPlot::get_datasystem(ChipsAxis *ax, ChipsAxis *ay)
{
   string xid = "";
   string yid = "";
   if (ax)
      xid = ax->get_id ();
   if (ay)
      yid = ay->get_id ();

   pair <string, string> axis_pair (xid, yid);
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter = this->data_systems.find (axis_pair);

   if (ds_iter != data_systems.end ())
      return ds_iter->second;

   return NULL;
}

void ChipsPlot::log_updates()
{
   if (bindings)
   {
      bindings->record_updates (true);
      bindings->reset_access_list();
   }
}

void ChipsPlot::handle_updates()
{
   if (bindings)
   {
      bindings->record_updates (false);

      if (bindings->get_updates_recorded())
      {
         list <string> axis_ids;
         bindings->get_recorded_axes(id, axis_ids);
         bindings->reset_access_list();
         update_limits (axis_ids);
      }
   }
}

int ChipsPlot::get_xaxis_cnt()
{
   int sz = axes_x.size () -2;
   if (sz < 0) sz = 0;
   return sz;
}

int ChipsPlot::get_yaxis_cnt()
{
   int sz = axes_y.size () -2;
   if (sz < 0) sz = 0;
   return sz;
}

int ChipsPlot::serialize_currency_ascii(ChipsStateStream& out, int obcnt)
{
   int plid = 0;
   if (obcnt < 2) plid++; // only 1 plot- don't need to set its currency
   if (cur_xall && cur_yall)
   {
      if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
      out << "current_axis(\"all\")" << endl;
   }
   else if (cur_xall && !cur_yall && cur_y)
   {
      if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
      out << "current_axis(\"all\")" << endl;
      out << "current_axis(\""<< cur_y->get_id()<<"\")"<<endl;
   }
   else if (!cur_xall && cur_yall && cur_x)
   {
      if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
      out << "current_axis(\"all\")" << endl;
      out << "current_axis(\""<< cur_x->get_id()<<"\")"<<endl;
   }
   else
   {
      if (axes_x.size() > 3)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out<<"current_axis(\""<< cur_x->get_id()<<"\")"<<endl;
      }
      if (axes_y.size() > 3)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out<<"current_axis(\""<< cur_y->get_id()<<"\")"<<endl;
      }
   }
   if (curves.size() > 1)
   {
      map<string, ChipsCurve* >::reverse_iterator ri = curves.rbegin();
      if ((*ri).first != currcurv)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_curve(\"" << currcurv << "\")" <<endl;
      }
   }
   if (contours.size()>1)
   {
      map<string, ChipsContour* >::reverse_iterator ri = contours.rbegin();
      if ((*ri).first != currcont)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_contour(\"" << currcont << "\")" <<endl;
      }
   }
   if (histograms.size()>1)
   {
      map<string, ChipsHistogram* >::reverse_iterator ri = histograms.rbegin();
      if ((*ri).first != currhist)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_histogram(\"" <<currhist<< "\")" <<endl;
      }
   }
   if (images.size()>1)
   {
      map<string, ChipsImage* >::reverse_iterator ri = images.rbegin();
      if ((*ri).first != currimage)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_image(\"" << currimage << "\")" <<endl;
      }
   }
   if (lines.size()>1)
   {
      map<string, ChipsLine* >::reverse_iterator ri = lines.rbegin();
      if ((*ri).first != currline)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_line(\"" << currline << "\")" <<endl;
      }
   }
   if (labels.size()>1)
   {
      map<string, ChipsLabel* >::reverse_iterator ri = labels.rbegin();
      if ((*ri).first != currlabel)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_label(\"" << currlabel << "\")" <<endl;
      }
   }
   if (points.size()>1)
   {
      map<string, ChipsPoint* >::reverse_iterator ri = points.rbegin();
      if ((*ri).first != currpoint)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_point(\"" << currpoint << "\")" <<endl;
      }
   }
   if (regions.size()>1)
   {
      map<string, ChipsRegion* >::reverse_iterator ri = regions.rbegin();
      if ((*ri).first != currregion)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_region(\"" << currregion << "\")" <<endl;
      }
   }
   if (cbars.size()>1)
   {
      map<string, ChipsColorBar* >::reverse_iterator ri = cbars.rbegin();
      if ((*ri).first != currcbar)
      {
         if (plid++ == 0)  out << "current_plot(\""<<id<<"\")" << endl;
         out << "current_colorbar(\"" <<currcbar << "\")" <<endl;
      }
   }
   return (plid > 0);
}

void ChipsPlot::combine_axes_ascii_serialize(ChipsStateStream& out, ChipsAxis* &ax1, ChipsAxis* &ax2)
{
   if (ax1 && ax2)
   {
      double d1, d2, d3, d4, o1, o2;
      string b1, b2;
      ax1->get_linear_range(d1, d2);
      b1 = ax1->build_axis_property_string();
      o1 = ax1->get_orig_pos();
      ax2->get_linear_range(d3, d4);
      b2 = ax2->build_axis_property_string();
      o2 = ax2->get_orig_pos();
      if (d1 == d3 && d2 == d4 && b1 == b2 && o1 == o2)
      {
         if (b1.length()) b1.append(" ");
         b1.append("x.id=");
         b1.append(ax1->get_id());
         b1.append(" y.id=");
         b1.append(ax2->get_id());
         out<<"add_axis(XY_AXIS,"<<o1<<","<<d1<<","<<d2<<",\""<<b1<<"\")"<<endl;
         ax1 = ax2 = NULL;
      }
   }
   if (ax1)
   {
      ax1->serialize_ascii(out);
      ax1 = NULL;
   }
   if (ax2)
   {
      ax2->serialize_ascii(out);
      ax2 = NULL;
   }
}


int ChipsPlot::check_for_matching_attribute_list(string bstr, string astr)
{
   // returns 0 if bstr attributes are all contained in astr 
   int needed = 0, ps1= bstr.find(','),  ps2 = bstr.length();
   
   if ((ps1 > 0) && (ps2 > 0) && astr.length())
   {
      char *where, *buff = new char[ps2];
      string sb = bstr.substr(ps1+2, ps2-ps1-4);
      strcpy(buff, sb.c_str());
      where = strtok(buff, " ");
      while(where)
      {
         if (astr.find(where) == string::npos) needed++;
         where = strtok(NULL, " ");
      }
      delete []buff; 
   }
   return needed; 
}

// setup the current axes for the specified transform if possible
// returns a flag indicating what axes need to be created if any
int ChipsPlot::setup_current_axes (Transform *transform, int &out_create)
{
   bool x_compatible = true;
   bool y_compatible = true;
   int rval = ERR_STATUS_GOOD;
   stringstream error_message;

   out_create = 0;

   // if current x is set to all but we only have a single axis, use it
   if ((!cur_x) && cur_xall && axes_x.size () == 3)
   {
      cur_x = axes_x.begin()->second;
      cur_xall = false;
   }
   // if current y is set to all but we only have a single axis, use it
   if ((!cur_y) && cur_yall && axes_y.size() == 3)
   {
      cur_y = axes_y.begin()->second;
      cur_yall = false;
   }

   // look for a pair of axes which match the transform type
   if (cur_x)
      x_compatible = cur_x->check_transform_compatibility (transform);

   if (cur_y)
      y_compatible = cur_y->check_transform_compatibility (transform);

   // if either x or y has data associated then there is an error
   // otherwise convert the axis to the appropriate type
   if (!x_compatible && !y_compatible)
   {
      if (check_for_data_sys_objs (cur_x) || check_for_data_sys_objs (cur_y))
      {
         error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", are not a valid type and already have data associated with them.";
         cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
      }
      else
      {
         string xid = cur_x->get_id ();
         string yid = cur_y->get_id ();
         pair<string, string> axis_pair(xid, yid);

         cur_x->set_algorithm_type (transform);
         cur_y->set_algorithm_type (transform);

         // update the coordinate system with the new transform

         ChipsDataSystem *ds = data_systems[axis_pair];
         ds->set_transform (ds->get_default_transform_id (), transform);

         // Need to delete any coordinate systems alread created by these two axes
         remove_data_coord (cur_x, cur_y);

         // add back the x/y axis coordinate system since it was delete too
         data_systems[axis_pair] = ds;
         cur_x->set_data_coord_id (ds->get_id ());
         cur_y->set_data_coord_id (ds->get_id ());

      }
   }

   // for now force error out if the user is trying to setup coordinate
   // system between to non-compatible axes
   else if (!x_compatible)
   {
      if (check_for_data_sys_objs (cur_x))
      {
         error_message << "Cannot add object because axis, " << cur_x->get_id () << ", is not a valid type and already has data associated with it.";
         cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);

      }
      else if (cur_y)
      {
         error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinates system.";
         cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
      }
      else
      {
         // if the axis is NOT a single axis for its data system than error out
         pair<string, string> axis_pair(cur_x->get_id (), "");
         map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter = data_systems.find (axis_pair);

         if (ds_iter==data_systems.end ())
         {
            error_message << "Cannot add object because axis, " << cur_x->get_id () << ", is not a valid type and other axes associated with it.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {

            cur_x->set_algorithm_type (transform);

            // update the coordinate system with the new transform
            ChipsDataSystem *ds = ds_iter->second;
            ds->set_transform (ds->get_default_transform_id (), transform);
         }
      }
   }
   // for now force error out if the user is trying to setup coordinate
   // system between to non-compatible axes
   else if (!y_compatible)
   {
      // if either x or y has data associated then there is an error
      if (check_for_data_sys_objs (cur_y))
      {
         error_message << "Cannot add object because axis, " << cur_y->get_id () << ", is not a valid type and already has data associated with it.";
         cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
      }
      else if (cur_x)
      {
         error_message << "Cannot add object because axes, " << cur_x->get_id () << " and " << cur_y->get_id () << ", do not create a valid coordinates system.";
         cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
      }

      else
      {

         // if the axis is NOT a single axis for its data system than error out
         pair<string, string> axis_pair("", cur_y->get_id ());
         map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter = data_systems.find (axis_pair);

         if (ds_iter==data_systems.end ())
         {
            error_message << "Cannot add object because axis, " << cur_x->get_id () <<
", is not a valid type and other axes associated with it.";
            cmd_manager->add_err(error_message.str (), rval=ERR_STATUS_BAD,id,0);
         }
         else
         {

            cur_y->set_algorithm_type (transform);

            // update the coordinate system with the new transform
            ChipsDataSystem *ds = ds_iter->second;
            ds->set_transform (ds->get_default_transform_id (), transform);
         }
      }
   }

   if (!rval)
   {
      if (!(cur_x || cur_xall) && !(cur_y || cur_yall))
         out_create = 3;
      else if (!(cur_x || cur_xall))  // no x axis exists so create one
         out_create = 1;
      if (!(cur_y || cur_yall))  // no y axis exists so create one
         out_create = 2;
   }


   return rval;
}





// creates axes if none exists for a new plot object 
// returns a flag indicating what axes were created if any
int ChipsPlot::create_axes (Transform *transform, int &out_created, double *min, double *max)
{

   ChipsId tcid;
   double tmp_min[]={0,0};
   double tmp_max[]={1,1};
   int rval;

   if ((rval=this->setup_current_axes (transform, out_created)) != ERR_STATUS_GOOD)
      return rval;


   if (min)
   {
      tmp_min[0] = min[0];
      tmp_min[1] = min[1];
   }
   if (max)
   {
      tmp_max[0] = max[0];
      tmp_max[1] = max[1];
   }

   
   if (!(cur_x || cur_xall) && !(cur_y || cur_yall))
   {
      rval = add_axis (&tcid, XY_AXIS, 0, tmp_min[0], tmp_max[0], tmp_min[1], tmp_max[1], transform);
      out_created = 3;
   }
   else if (!(cur_x || cur_xall))  // no x axis exists so create one
   {
      rval = add_axis (&tcid, X_AXIS, 0, tmp_min[0], tmp_max[1], transform);
      out_created = 1;
   }
   if (!(cur_y || cur_yall))  // no y axis exists so create one
   {
      rval = add_axis (&tcid, Y_AXIS, 0, tmp_min[0], tmp_max[1], transform);
      out_created = 2;
   }

   return rval;
}

int ChipsPlot::update_axis_coordinates (ChipsAxis *ax, list<string> &ds_ids)
{ 
   int rval = ERR_STATUS_GOOD;
   if (!ax)
      return ERR_STATUS_BAD;

   if (ds_ids.size () == 0)
      return ERR_STATUS_GOOD;

   if ((ax == top) || (ax == bottom) || (ax == left) || (ax == right))
      return ERR_STATUS_GOOD;

   string axis_id = ax->get_id ();
   list<string>::iterator ds_id_iter = ds_ids.begin ();
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter = data_systems.begin ();
   map <pair <string, string>, ChipsDataSystem *>::iterator tmp_ds_iter;
   ChipsDataSystem *ds;

   // if its an xaxis
   if (axes_x.count(axis_id))
   {

      // remove this axis data systems
      while (ds_iter!=data_systems.end ())
      {
         if (ds_iter->first.first == axis_id)
         {
            // if the xaxis is the only id remove the data system
            ds = ds_iter->second;
            if (ds->get_xaxis_id () == axis_id)
            {
               if (ds_iter->first.second.length () != 0)
               {
                  pair <string, string> tmp_id ("", ds->get_yaxis_id ());
                  data_systems[tmp_id] = ds;
               }   

               ds->set_xaxis_id ("");
            }
            tmp_ds_iter = ds_iter;
            ds_iter++;
            data_systems.erase (tmp_ds_iter);
         }
         else
            ds_iter++;
      }

      // add in the new id
      for (; ds_id_iter != ds_ids.end (); ds_id_iter++)
      {
         ds = dynamic_cast <ChipsDataSystem *>(this->coord->get_coord (DATA, *ds_id_iter));
         if (ds)
         {
            string ds_xaxis_id = ds->get_xaxis_id ();
            string ds_yaxis_id = ds->get_yaxis_id ();
            pair <string, string> tmp_id (ds_xaxis_id, ds_yaxis_id);
            
            // if the axis is being added to an empty coordinate system
            // simply add the id back to the data system
            if (ds_xaxis_id.length () == 0)
            {
               if (data_systems.count (tmp_id) != 0)
                  data_systems.erase (tmp_id);
               tmp_id.first = axis_id;
               data_systems[tmp_id] = ds;
               ds->set_xaxis_id (axis_id);
            }
            else 
            {
               tmp_id.first = axis_id;
               data_systems[tmp_id] = ds;

               if (ds_xaxis_id != axis_id)
                  ax->set_shared_ds (true);
            }
         }
      }
   }
   // if its an yaxis
   else if (axes_y.count(axis_id))
   {

      // remove this axis data systems
      while (ds_iter!=data_systems.end ())
      {
         if (ds_iter->first.second == axis_id)
         {
            // if the yaxis is the only id remove the data system
            ds = ds_iter->second;
            if (ds->get_yaxis_id () == axis_id)
            {
               if (ds_iter->first.first.length () != 0)
               {
                  pair <string, string> tmp_id (ds->get_xaxis_id (), "");
                  data_systems[tmp_id] = ds;
               }

               ds->set_yaxis_id ("");
            }
            tmp_ds_iter = ds_iter;
            ds_iter++;
            data_systems.erase (tmp_ds_iter);
         }
         else
            ds_iter++;
      }

      // add in the new id
      for (; ds_id_iter != ds_ids.end (); ds_id_iter++)
      {
         ds = dynamic_cast <ChipsDataSystem *>(this->coord->get_coord (DATA, *ds_id_iter));
         if (ds)
         {
            string ds_xaxis_id = ds->get_xaxis_id ();
            string ds_yaxis_id = ds->get_yaxis_id ();
            pair <string, string> tmp_id (ds_xaxis_id, ds_yaxis_id);

            // if the axis is being added to an empty coordinate system
            // simply add the id back to the data system
            if (ds_yaxis_id.length () == 0)
            {
               if (data_systems.count (tmp_id) != 0)
                  data_systems.erase (tmp_id);
               tmp_id.second = axis_id;
               data_systems[tmp_id] = ds;
               ds->set_yaxis_id (axis_id);
            }
            else
            {
               tmp_id.second = axis_id;
               data_systems[tmp_id] = ds;
               if (ds_yaxis_id != axis_id)
                  ax->set_shared_ds (true);
            }
         }
      }
   }

   return rval;

}

void ChipsPlot::clone_data_systems (ChipsAxis *ax)
{
   if (!ax)
      return;

   if ((ax == top) || (ax == bottom) || (ax == left) || (ax == right))
      return;

   string axis_id = ax->get_id ();
   map <pair <string, string>, ChipsDataSystem *>::iterator ds_iter = data_systems.begin ();
   ChipsDataSystem *ds;
   ChipsAxis *opp_axis;
   string opp_axis_id;

   // if its an xaxis
   if (axes_x.count(axis_id))
   {
      for (; ds_iter != data_systems.end (); ds_iter++)
      {
         if (ds_iter->first.first == axis_id)
         {
            ax->remove_data_coord_id (ds_iter->second->get_id ());
            ds = new ChipsDataSystem (NULL);
            *ds = *ds_iter->second;
            ds->set_parent (this->coord);
            ds->new_id ();
            this->coord->add_coord_sys (ds);
            ds_iter->second = ds;
            ds->set_xaxis_id (axis_id);
            ax->set_data_coord_id (ds->get_id ());
     
            opp_axis_id = ds->get_yaxis_id ();
            if (opp_axis_id.length () > 0)
            {
               opp_axis = this->get_yaxis (opp_axis_id);

               if (opp_axis)
                  opp_axis->set_data_coord_id (ds->get_id ());
            }
         }
      }
   }
   // if its an yaxis
   else if (axes_y.count(axis_id))
   {
      for (; ds_iter != data_systems.end (); ds_iter++)
      {
         if (ds_iter->first.second == axis_id)
         {
            ax->remove_data_coord_id (ds_iter->second->get_id ());
            ds = new ChipsDataSystem (NULL);
            *ds = *ds_iter->second;
            ds->set_parent (this->coord);
            ds->new_id ();
            this->coord->add_coord_sys (ds);
            ds_iter->second = ds;
            ds->set_yaxis_id (axis_id);
            ax->set_data_coord_id (ds->get_id ());

            opp_axis_id = ds->get_xaxis_id ();
            if (opp_axis_id.length () > 0)
            {
               opp_axis = this->get_xaxis (opp_axis_id);

               if (opp_axis)
                  opp_axis->set_data_coord_id (ds->get_id ());
            }

         }
      }
   }
}


char *ChipsPlot::info_coordinate ()
{
   char* ptr = NULL;
   string temp;

   if (!data_systems.empty())
   {
      // Run through objects in frame system
      map <pair <string, string>, ChipsDataSystem*>::iterator ds_iter;
      ChipsDataSystem *ds;
      string xaxis_id;
      string yaxis_id;
      ChipsAxis *ax;
      ChipsAxis *ay;
      for (ds_iter = data_systems.begin (); ds_iter != data_systems.end (); ++ds_iter)
      {

         ds = ds_iter->second;
         xaxis_id = ds_iter->first.first;
         yaxis_id = ds_iter->first.second;
         ax = this->get_xaxis (xaxis_id);
         ay = this->get_yaxis (yaxis_id);
        
         if (((xaxis_id.length () != 0) || (yaxis_id.length () != 0)) && 
             ((ax != bottom) && (ax != top) && (ay != left) && (ay != right)))
         {
            temp.append("        Coord Sys ID [");
            temp.append(ds->get_id());
            temp.append("]\n");
            temp.append("        X Axis [");
            temp.append(xaxis_id);
            temp.append("]\n");
            temp.append("        Y Axis [");
            temp.append(yaxis_id);
            temp.append("]\n");
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


int ChipsPlot::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj, bool vis)
{
   int  rval = ERR_STATUS_GOOD;

   string tid;
   switch(obj)
   {
      case chips_point:
      {
         tid = cid->get_point_id();
         if (tid.length() == 0)  tid = currpoint;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(points_map::iterator ii=points.begin(); ii != points.end(); ++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (points.count(tid))
         {
            if (vis)  points[tid]->force_show();
               else   points[tid]->force_hide(); 
         }
      }
      break;
      case chips_contour:
      {
         tid = cid->get_contour_id();
         if (tid.length() == 0)  tid = currcont;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(contours_map::iterator ii=contours.begin(); ii != contours.end(); ++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (contours.count(tid))
         {
            if (vis)  contours[tid]->force_show();
               else   contours[tid]->force_hide();
         }
      }
      break;
      case chips_curve:
      {
         tid = cid->get_curve_id();
         if (tid.length() == 0)  tid = currcurv;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(curves_map::iterator ii=curves.begin(); ii != curves.end(); ++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (curves.count(tid))
         {
            if (vis)  curves[tid]->force_show();
               else   curves[tid]->force_hide();
         }
      }
      break;
      case chips_histogram:
      {
         tid = cid->get_histogram_id();
         if (tid.length() == 0)  tid = currhist;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(histograms_map::iterator ii=histograms.begin(); ii != histograms.end(); ++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (histograms.count(tid))
         {
            if (vis)  histograms[tid]->force_show();
               else   histograms[tid]->force_hide();
         }
      }
      break;
      case chips_image:
      {
         tid = cid->get_image_id();
         if (tid.length() == 0)  tid = currimage;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(images_map::iterator ii=images.begin();ii!=images.end();++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (images.count(tid))
         {
            if (vis)  images[tid]->force_show();
               else   images[tid]->force_hide();
         }
      }
      break;
      case chips_label:
      {
         tid = cid->get_label_id();
         if (tid.length() == 0)  tid = currlabel;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(labels_map::iterator ii=labels.begin();ii!=labels.end();++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (labels.count(tid))
         {
            if (vis)  labels[tid]->force_show();
               else   labels[tid]->force_hide();
         }
      }
      break;
      case chips_line:
      {
         tid = cid->get_line_id();
         if (tid.length() == 0)  tid = currline;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(lines_map::iterator ii=lines.begin();ii!=lines.end();++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (lines.count(tid))
         {
            if (vis)  lines[tid]->force_show();
               else   lines[tid]->force_hide();
         }
      }
      break;
      case chips_colorbar:
      {
         tid = cid->get_colorbar_id();
         if (tid.length() == 0)  tid = currcbar;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(cbars_map::iterator ii=cbars.begin();ii!=cbars.end();++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (cbars.count(tid))
         {
            if (vis)  cbars[tid]->force_show();
               else   cbars[tid]->force_hide();
         }
      }
      break;
      case chips_region:
      {
         tid = cid->get_region_id();
         if (tid.length() == 0)  tid = currregion;
         if (chips_strcmp(tid.c_str(), "all") == 0)
         {
            for(regions_map::iterator ii=regions.begin();ii!=regions.end();++ii)
               if (vis)  (*ii).second->force_show();
                  else   (*ii).second->force_hide();
         }
         else if (regions.count(tid))
         {
            if (vis)  regions[tid]->force_show();
               else   regions[tid]->force_hide();
         }
      }
      break;
      case chips_axis: 
      {
         axes_map::iterator am; 
         ChipsAxis *tax;
         tid = cid->get_axis_id();
         if (tid.length())
         {
            if (chips_strcmp(tid.c_str(), "all") == 0)
            {
               if (axes_x_exist())
                  for(am=axes_x.begin(); am != axes_x.end(); ++am)
                     if (vis)  (*am).second->force_show();
                        else   (*am).second->force_hide();
               if (axes_y_exist())
                  for(am=axes_y.begin(); am != axes_y.end(); ++am)
                     if (vis)  (*am).second->force_show();
                        else   (*am).second->force_hide();
            }
            else if ((tax = get_xaxis(tid)) != NULL)
            {
               if (vis)  tax->force_show();
                  else   tax->force_hide();
            }    
            else if ((tax = get_yaxis(tid)) != NULL)
            {
               if (vis)  tax->force_show();
                  else   tax->force_hide();
            }    
         }
         else
         {
            string txid = cid->get_xaxis_id();
            string tyid = cid->get_yaxis_id();
        
            if (axes_x_exist())
            {
               if (chips_strcmp(tid.c_str(), "all") == 0)
               {
                  for(am=axes_x.begin(); am != axes_x.end(); ++am)
                     if (vis)  (*am).second->force_show();
                        else   (*am).second->force_hide();
               } 
               else if ((tax = get_xaxis(tid)) != NULL)
               {
                  if (vis)  tax->force_show();
                     else   tax->force_hide(); 
               }
            } 
            if (axes_y_exist())
            {
               if (chips_strcmp(tid.c_str(), "all") == 0)
               {
                  for(am=axes_y.begin(); am != axes_y.end(); ++am)
                     if (vis)  (*am).second->force_show();
                        else   (*am).second->force_hide();
               }
               else if ((tax = get_yaxis(tid)) != NULL)
               {
                  if (vis)  tax->force_show();
                     else   tax->force_hide();
               }
            }
         }
      }
      break; 
      default:   // do nothing
      break;
   }
   return rval;
}

void ChipsPlot::disable_all_interaction ()
{
   this->disable_interaction ();
   this->disable_object_interaction ("all", chips_label);
   this->disable_object_interaction ("all", chips_line);
   this->disable_object_interaction ("all", chips_point);
   this->disable_object_interaction ("all", chips_axis);
   this->disable_object_interaction ("all", chips_region);
   this->disable_object_interaction ("all", chips_colorbar);
   this->disable_object_interaction ("all", chips_curve);
   this->disable_object_interaction ("all", chips_histogram);
   this->disable_object_interaction ("all", chips_contour);
   this->disable_object_interaction ("all", chips_image);
}

void ChipsPlot::enable_all_interaction ()
{
   this->enable_interaction ();
   this->enable_object_interaction ("all", chips_label);
   this->enable_object_interaction ("all", chips_line);
   this->enable_object_interaction ("all", chips_point);
   this->enable_object_interaction ("all", chips_axis);
   this->enable_object_interaction ("all", chips_region);
   this->enable_object_interaction ("all", chips_colorbar);
   this->enable_object_interaction ("all", chips_curve);
   this->enable_object_interaction ("all", chips_histogram);
   this->enable_object_interaction ("all", chips_contour);
   this->enable_object_interaction ("all", chips_image);
}


void ChipsPlot::disable_object_interaction(const string &oid, V_OBJECT_TYPE obj)
{
   switch(obj)
   {
      case chips_point:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(points_map::iterator ii=points.begin(); ii != points.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (points.count(oid))
            points[oid]->disable_interaction();
      }
      break;
      case chips_label:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(labels_map::iterator ii=labels.begin(); ii != labels.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (labels.count(oid))
            labels[oid]->disable_interaction();
      }
      break;
      case chips_line:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(lines_map::iterator ii=lines.begin(); ii != lines.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (lines.count(oid))
            lines[oid]->disable_interaction();
      }
      break;
      case chips_region:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(regions_map::iterator ii=regions.begin(); ii != regions.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (regions.count(oid))
            regions[oid]->disable_interaction();
      }
      break;
      case chips_axis:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(axes_map::iterator ii=axes_x.begin(); ii != axes_x.end(); ++ii)
               (*ii).second->disable_interaction();
            for(axes_map::iterator ii=axes_y.begin(); ii != axes_y.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (axes_x.count(oid))
            axes_x[oid]->disable_interaction();
         else if (axes_y.count(oid))
            axes_y[oid]->disable_interaction();

      }
      break;
      case chips_colorbar:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(cbars_map::iterator ii=cbars.begin(); ii != cbars.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (cbars.count(oid))
            cbars[oid]->disable_interaction();
      }
      break;
      case chips_curve:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(curves_map::iterator ii=curves.begin(); ii != curves.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (curves.count(oid))
            curves[oid]->disable_interaction();
      }
      break;
      case chips_histogram:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(histograms_map::iterator ii=histograms.begin(); ii != histograms.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (histograms.count(oid))
            histograms[oid]->disable_interaction();
      }
      break;
      case chips_contour:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(contours_map::iterator ii=contours.begin(); ii != contours.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (contours.count(oid))
            contours[oid]->disable_interaction();
      }
      break;
      case chips_image:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(images_map::iterator ii=images.begin(); ii != images.end(); ++ii)
               (*ii).second->disable_interaction();
         }
         else if (images.count(oid))
            images[oid]->disable_interaction();
      }
      default:
      break;
   }
}

void ChipsPlot::enable_object_interaction(const string &oid, V_OBJECT_TYPE obj)
{
   switch(obj)
   {
      case chips_point:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(points_map::iterator ii=points.begin(); ii != points.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (points.count(oid))
            points[oid]->enable_interaction();
      }
      break;
      case chips_label:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(labels_map::iterator ii=labels.begin(); ii != labels.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (labels.count(oid))
            labels[oid]->enable_interaction();
      }
      break;
      case chips_line:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(lines_map::iterator ii=lines.begin(); ii != lines.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (lines.count(oid))
            lines[oid]->enable_interaction();
      }
      break;
      case chips_region:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(regions_map::iterator ii=regions.begin(); ii != regions.end(); ++ii)
               ii->second->enable_interaction ();
         }
         else if (regions.count(oid))
            regions[oid]->enable_interaction();
      }
      break;
      case chips_axis:
      {
         ChipsAxis *ax;
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(axes_map::iterator ii=axes_x.begin(); ii != axes_x.end(); ++ii)
               if (!ii->second->get_border_axis ())
                  ii->second->enable_interaction();
            for(axes_map::iterator ii=axes_y.begin(); ii != axes_y.end(); ++ii)
               if (!ii->second->get_border_axis ())
                  ii->second->enable_interaction();
         }
         else
         {
            if (axes_x.count(oid))   
               ax = axes_x[oid];
            else if (axes_y.count(oid))
               ax = axes_y[oid];

            if (ax && !ax->get_border_axis ())
               ax->enable_interaction ();
         }

      }
      break;
      case chips_colorbar:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(cbars_map::iterator ii=cbars.begin(); ii != cbars.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (cbars.count(oid))
            cbars[oid]->enable_interaction();
      }
      break;
      case chips_curve:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(curves_map::iterator ii=curves.begin(); ii != curves.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (curves.count(oid))
            curves[oid]->enable_interaction();
      }
      break;
      case chips_histogram:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(histograms_map::iterator ii=histograms.begin(); ii != histograms.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (histograms.count(oid))
            histograms[oid]->enable_interaction();
      }
      break;
      case chips_contour:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(contours_map::iterator ii=contours.begin(); ii != contours.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (contours.count(oid))
            contours[oid]->enable_interaction();
      }
      break;
      case chips_image:
      {
         if (chips_strcmp(oid.c_str(), "all") == 0)
         {
            for(images_map::iterator ii=images.begin(); ii != images.end(); ++ii)
               (*ii).second->enable_interaction();
         }
         else if (images.count(oid))
            images[oid]->enable_interaction();
      }
      default:
      break;
   }
}




