// 
//  Copyright (C) 2007,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsRemote.hh"
#include "ChipsClient.hh"
#include "ChipsMessages.hh"
#include "ChipsDefinesP.hh"
#include "ChipsTransformFactory.hh"
#include <stdio.h>
#include <sstream>

#include <sys/types.h>
#include <sys/stat.h>

long ChipsRemote::serverIndx=1;

ChipsRemote::ChipsRemote(bool use_server, bool load_prefs, string id)
{
   server_pid = ""; 
   startup_prefs = load_prefs;
   errverb = -1;
   offscreen_flag = false;
   if (use_server && ((client = new ChipsClient(load_prefs)) != NULL))
      create_server(load_prefs, id);
}

ChipsRemote::ChipsRemote(const string& inServerId, bool load_prefs)
{
   server_pid = ""; 
   startup_prefs = load_prefs;
   errverb = -1;
   offscreen_flag = false;
   if ((client = new ChipsClient(load_prefs)) != NULL)
      connect (inServerId);
};

ChipsRemote::ChipsRemote( bool load_prefs)
{
   server_pid = ""; 
   startup_prefs = load_prefs;
   errverb = -1;
   offscreen_flag = false;
   if ((client = new ChipsClient(load_prefs)) != NULL)
      client->create_client();
};
  
ChipsRemote::~ChipsRemote()
{
   if (client) delete client; 
}

int ChipsRemote::set_preference_autoload(bool toggle_flag)
{
   startup_prefs = toggle_flag;

   return ERR_STATUS_GOOD;
}

int ChipsRemote::load_preferences(string in_file)
{
   int rval; 
   set_id_spec(); 
   string filepath = in_file;
   if (in_file.length()) 
   {
      // since the server can't resolve paths for the client, we need
      // to add the path to file, if it is not provided.
      if (in_file.find("/") != 0)
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
            filepath.append("/");
            filepath.append(in_file);
         }
      }
   }
   else
   {
      filepath = getenv("HOME");
      if (filepath.length()) filepath.append("/.chips.rc");
      else   return ERR_RESFLE_PATH; 
   }
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_LOAD_PREFS_B, (void*)filepath.c_str());
   return rval; 
}


int ChipsRemote::save_preferences(string in_file)
{
   set_id_spec();
   string filepath = in_file;
   // since the server can't resolve paths for the client, we need
   // to add the path to file, if it is not provided.
   if (in_file.length())
   {
      if (in_file.find("/") != 0) 
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
            filepath.append("/");
            filepath.append(in_file);
         }
      }
   }
   else
   {
      filepath = getenv("HOME");
      if (filepath.length()) filepath.append("/.chips.rc");
      else   return ERR_RESFLE_PATH; 
   }

   // server needs to get umask of client to set the output file permission
   mode_t mv;
   umask(mv = umask(0));
   int  imv = (int)mv;

   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SAVE_PREFS, (void*)filepath.c_str(),
                          static_cast<void*>(&imv));
   return rval; 
}

int ChipsRemote::set_preference(string pref, string value)
{
   set_id_spec();
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_PREFERENCE, (void*)pref.c_str(),
                          (void*)value.c_str());
   return rval; 
}



char* ChipsRemote::get_preference(string pref)
{
   char* buf = NULL;
   set_id_spec();
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt, rval;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_GET_PREFERENCE, &vp, &typ, &len, &cnt,
                      (void*)pref.c_str());
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         buf = (char*) (char*)vp[0];
      free(vp);
      free(typ); 
      free(len); 
   }
   return buf;
}

int ChipsRemote::set_offscreen_mode (bool flag)
{
   offscreen_flag = flag;

   return ERR_STATUS_GOOD;
   int verbose = 0;
   int rval = get_error_verbosity (verbose);

   // currently if there's no server then the error message wont
   // be displayed. Display the error explicitly
   if (rval == ERR_NO_SOCKET)
      errmsg = "chips ERROR: The requested functionality is currently not implemented";
   return ERR_FEATURE_DNE;
   
}



int ChipsRemote::set_line_stipple(V_LINE_TYPE line, int val)
{
   set_id_spec();
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_LINE_STIPPLE, 
                          static_cast<void*>(&line), static_cast<void*>(&val));
   return rval; 
}


int ChipsRemote::get_line_stipple(V_LINE_TYPE line, int &out_val)
{
   int cnt, rval = ERR_STATUS_GOOD;
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   set_id_spec();
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_LINE_STIPPLE,&vp,&typ,&len,&cnt,
                              static_cast<void*>(&line));
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((V_LINE_TYPE*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}


int ChipsRemote::load_state(string fname)
{
   int rval = ERR_FILE_NOT_SUPPLIED;
   set_id_spec(); 

   string filepath = fname;
   // since the server can't resolve paths for the client, we need
   // to add the path to file, if it is not provided.
   if (fname.length() && fname.find("/") != 0)
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(fname);
      }
   }
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_LOAD_STATE, (void*)filepath.c_str());
   return rval;
}


int ChipsRemote::save_state(string fname, string mode)
{
   int rval = ERR_FILE_NOT_SUPPLIED;
   set_id_spec(); 

   if (fname.length()) 
   {
      string filepath = fname;
      // since the server can't resolve paths for the client, we need
      // to add the path to file, if it is not provided.
      if (fname.find("/") != 0)
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
            filepath.append("/");
            filepath.append(fname);
         }
      }

      // server needs to get umask of client to set the output file permission
      mode_t mv;
      umask(mv = umask(0));
      int  imv = (int)mv;
      int  omode = 0;

      if (!chips_strcmp(mode.c_str(),"binary") ||
          !chips_strcmp(mode.c_str(),"bin"))
         omode = 0;
      else if (!chips_strcmp(mode.c_str(),"ascii") ||
          !chips_strcmp(mode.c_str(),"python"))
         omode = 1;
      else if (!chips_strcmp(mode.c_str(),"template"))
         omode = 2;
      else
      {
         char buff[256];
         sprintf(buff, "The specified mode value (%s) must be either [binary|ascii|template]", mode.c_str());
         rval = ERR_NORM_DATA_RANGE;
         this->create_server ();
         client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
         return rval;
      }

      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_SAVE_STATE_B, (void*)filepath.c_str(),
                             static_cast<void*>(&imv), 
                             static_cast<void*>(&omode));
   }
   return rval;
}


int ChipsRemote::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE start,
   string property,  string value)  
{
   int rval = ERR_STATUS_BAD;

   set_id_spec(cid);
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CASCADE, static_cast<void*>(&start),
                          (void*)property.c_str(), (void*)value.c_str());
   return rval;
}


int ChipsRemote::move(ChipsId* cid, V_OBJECT_TYPE obj, double xx, double yy, 
   int mode)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_MOVE, static_cast<void*>(&obj),
                          static_cast<void*>(&xx), static_cast<void*>(&yy),
                          static_cast<void*>(&mode));
   return rval; 
}


int ChipsRemote::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj,
   int& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_OBJ_COUNT,&vp,&typ,&len,&cnt,
                              static_cast<void*>(&obj));
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT) &&
          (rval == ERR_STATUS_GOOD))
      out_val = ((int*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}


int ChipsRemote::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_OBJ_COORD,&vp,&typ,&len,&cnt,
                              static_cast<void*>(&obj));
      if ((cnt == 2) && typ && vp && (typ[0] == MSG_DATATYPE_INT) &&
          (typ[1] == MSG_DATATYPE_CHAR) && (rval == ERR_STATUS_GOOD))
      {
         out_val = ((COORD_SYS*)vp[0])[0];
         out_id = (char*) vp[1]; 
      }
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}


int ChipsRemote::add_window(ChipsId* cid, string attlist)
{
   int rval = ERR_STATUS_BAD;   

   set_id_spec(cid);
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      if (attlist.length())
         rval = client->send(id_spec, MSG_ADD_WINDOW_B, (void*)attlist.c_str());
      else 
         rval = client->send(id_spec, MSG_ADD_WINDOW_A);
   }
   return rval; 
}


int ChipsRemote::add_window(ChipsId* cid, float ww, float hh, 
   string format, string attlist)
{
   int rval = ERR_STATUS_BAD;

   set_id_spec(cid);
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ADD_WINDOW_C, static_cast<void*>(&ww),
                          static_cast<void*>(&hh), (void*)format.c_str(),
                          (void*)attlist.c_str());
   return rval;
}


int ChipsRemote::delete_window(ChipsId* cid)
{
   int rval = ERR_STATUS_BAD;

   set_id_spec(cid);
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_DEL_WINDOW);
   return rval;
}


int ChipsRemote::load_font(string fontname, string fontpath, 
   V_FONTSTYLE_TYPE style)
{
   int rval = ERR_FEATURE_DNE;
   set_id_spec();

   if (client)
   {
      // since the server can't resolve paths for the client, we need
      // to add the path to file, if it is not provided.
      string filepath = fontpath;
      if (fontpath.length() && fontpath.find("/") != 0)
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
            filepath.append("/");
            filepath.append(fontpath);
         }
      }
      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_LOAD_FONT, (void*)fontname.c_str(),  
                           (void*)filepath.c_str(), static_cast<void*>(&style));
   }
   return rval; 
}

char* ChipsRemote::info(ChipsId* cid, V_OBJECT_TYPE obj, bool descend)
{
   char* buf = NULL;
   set_id_spec(cid, 1);
   void** vp = NULL; 
   char *typ = NULL; 
   int *len = NULL; 
   int cnt, desc = descend;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_INFO_REQUEST, &vp, &typ, &len, &cnt, 
                       static_cast<void*>(&obj), static_cast<void*>(&desc));
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         buf = (char*) vp[0]; 
      free(vp);
      free(typ); 
      free(len); 
   }
   return buf; 
}

char* ChipsRemote::info_coordinate(ChipsId* cid)
{
   char* buf = NULL;
   set_id_spec(cid, 1);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_INFO_COORD_REQUEST, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
            buf = (char*) vp[0];
      free(vp); 
      free(typ); 
      free(len); 
   }
   return buf;
}

char* ChipsRemote::info_depth(ChipsId* cid, int depth)
{
   char* buf = NULL;
   set_id_spec(cid, 1);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_INFO_DEPTH_REQUEST, &vp, &typ, &len, &cnt,
                       static_cast<void*>(&depth));
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
            buf = (char*) vp[0];
      free(vp);
      free(typ);
      free(len);
   }
   return buf;
}

char* ChipsRemote::info_bound_axes(ChipsId* cid)
{
   char* buf = NULL;
   set_id_spec(cid, 1);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_INFO_BOUND_REQUEST, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         buf = (char*) vp[0];
      free(vp);
      free(typ);
      free(len);
   }
   return buf;
}


char* ChipsRemote::info_current()
{
   char* buf = NULL;
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt;
   set_id_spec();
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      client->get_data(id_spec, MSG_CURRENT_REQUEST, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
            buf = (char*) vp[0];
      free(vp); 
      free(typ); 
      free(len); 
   }
   return buf;
}



char* ChipsRemote::get_current_frame(ChipsId* cid)
{
   cerr <<"Client/Server for this feature not implemented yet" << endl;
   return NULL;
}


int ChipsRemote::set_current_window(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_WIN);
   return rval; 
}


int ChipsRemote::set_current_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_FRM);
   return rval; 
}


int ChipsRemote::set_current_image(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval =  client->send(id_spec, MSG_SET_CURRENT_IMG);
   return rval; 
}

int ChipsRemote::set_current_colorbar(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_CBAR);
   return rval; 
}

int ChipsRemote::set_current_plot(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_PLT);
   return rval; 
}


int ChipsRemote::set_current_curve(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_CRV);
   return rval; 
}

int ChipsRemote::set_current_contour(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_CNTR);
   return rval; 
}


int ChipsRemote::set_current_histogram(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_HST);
   return rval; 
}


int ChipsRemote::set_current_axis(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_AXS);
   return rval; 
}


int ChipsRemote::set_current_line(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_LINE);
   return rval; 
}


int ChipsRemote::set_current_label(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_LAB);
   return rval; 
}


int ChipsRemote::set_current_point(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_PNT);
   return rval; 
}


int ChipsRemote::set_current_legend(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_LGD);
   return rval;
}


int ChipsRemote::set_current_region(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CURRENT_REGN);
   return rval; 
}



int ChipsRemote::add_frame(ChipsId* cid, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      if (attlist.length())
         rval = client->send(id_spec, MSG_ADD_FRAME_B, (void*)attlist.c_str());
      else
         rval = client->send(id_spec, MSG_ADD_FRAME_A);
   }
   return rval; 
}


int ChipsRemote::add_frame(ChipsId* cid, float x1, float y1, 
                          float x2, float y2, string attlist)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD) 
   {
      set_id_spec(cid);
      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_ADD_FRAME_C, static_cast<void*>(&x1),
                             static_cast<void*>(&y1), static_cast<void*>(&x2),
                             static_cast<void*>(&y2), (void*)attlist.c_str());
   }
   return rval;
}


int ChipsRemote::delete_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_DEL_FRAME);
   return rval; 
}


int ChipsRemote::display_depth(ChipsId* cid, int in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec,MSG_DISPLAY_DEPTH,static_cast<void*>(&in_val));
   return rval; 
}

int ChipsRemote::hide_depth(ChipsId* cid, int in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_HIDE_DEPTH, static_cast<void*>(&in_val));
   return rval; 
}

int ChipsRemote::collapse_depths(ChipsId* cid, int scnt, int* srcs, int dest)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_COLLAPSE_DEPTH,static_cast<void*>(&dest),
                          static_cast<void*>(&scnt), static_cast<void*>(srcs));
   return rval; 
}

int ChipsRemote::shuffle(ChipsId* cid, V_OBJECT_TYPE obj, V_SHUFFLE_TYPE way)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_REARRANGE_DEPTH,static_cast<void*>(&obj),
                          static_cast<void*>(&way));
   return rval; 
}

int ChipsRemote::shuffle_forward(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   V_SHUFFLE_TYPE manner = chips_forward;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_REARRANGE_DEPTH,static_cast<void*>(&obj),
                          static_cast<void*>(&manner));
   return rval; 
}

int ChipsRemote::shuffle_backward(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   V_SHUFFLE_TYPE manner = chips_backward;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_REARRANGE_DEPTH,static_cast<void*>(&obj),
                          static_cast<void*>(&manner));
   return rval; 
}

int ChipsRemote::shuffle_front(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   V_SHUFFLE_TYPE manner = chips_front;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval =client->send(id_spec, MSG_REARRANGE_DEPTH, static_cast<void*>(&obj),
                          static_cast<void*>(&manner));
   return rval; 
}

int ChipsRemote::shuffle_back(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   V_SHUFFLE_TYPE manner = chips_back;
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval =  client->send(id_spec, MSG_REARRANGE_DEPTH, static_cast<void*>(&obj),
                          static_cast<void*>(&manner));
   return rval; 
}

int ChipsRemote::set_default_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval =  client->send(id_spec, MSG_SET_DEF_DEPTH, 
                           static_cast<void*>(&in_size));
   return rval; 
}


int ChipsRemote::get_default_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_DEF_DEPTH, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((int*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::set_framemode(ChipsId* cid, int mode)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAMEMODE, static_cast<void*>(&mode));
   return rval; 
}


int ChipsRemote::tile(ChipsId* cid, unsigned int cols, unsigned int rows)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_TILED_FRAME, static_cast<void*>(&cols), 
                          static_cast<void*>(&rows));
   return rval; 
}


int ChipsRemote::row_tile(ChipsId* cid, unsigned int rows)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ROW_TILED_FRAME, static_cast<void*>(&rows));
   return rval;
}


int ChipsRemote::col_tile(ChipsId* cid, unsigned int cols)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_COL_TILED_FRAME, static_cast<void*>(&cols));
   return rval; 
}



int ChipsRemote::prev_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_PREV_FRAME);
   return rval; 
}


int ChipsRemote::next_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_NEXT_FRAME);
   return rval; 
}


int ChipsRemote::reposition_frame(ChipsId* cid, float x1,
                                  float y1, float x2, float y2)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD)  
   {
      set_id_spec(cid);
      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_REPOSITION_FRAME,
                             static_cast<void*>(&x1), static_cast<void*>(&y1),
                             static_cast<void*>(&x2),static_cast<void*>(&y2));
   }
   return rval;
}


int ChipsRemote::reposition_plot(ChipsId* cid, float x1,
                                 float y1, float x2, float y2)
{
   int rval;
   if ((rval = normalized_area_check(x1, y1, x2, y2)) == ERR_STATUS_GOOD)
   {
      set_id_spec(cid);
      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_REPOSITION_PLOT,
                             static_cast<void*>(&x1), static_cast<void*>(&y1),
                             static_cast<void*>(&x2),static_cast<void*>(&y2));
   }
   return rval;
}


int ChipsRemote::adjust_grid_gaps(ChipsId* cid, float xgap, float ygap)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
    rval = client->send(id_spec, MSG_GRID_GAP, 
                        static_cast<void*>(&xgap), static_cast<void*>(&ygap));
   return rval; 
}


int ChipsRemote::adjust_grid_xgap(ChipsId* cid, float xgap)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_XGAP, static_cast<void*>(&xgap));
   return rval; 
}

int ChipsRemote::adjust_grid_ygap(ChipsId* cid, float ygap)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_YGAP, static_cast<void*>(&ygap));
   return rval; 
}


int ChipsRemote::adjust_grid_xrelsize(ChipsId* cid, int col, float ratio)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_XRELSIZE, static_cast<void*>(&col), 
                          static_cast<void*>(&ratio));
   return rval; 
}


int ChipsRemote::adjust_grid_yrelsize(ChipsId* cid, int row, float ratio)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_YRELSIZE, static_cast<void*>(&row), 
                          static_cast<void*>(&ratio));
   return rval; 
}

int ChipsRemote::adjust_grid_xrelsizes(ChipsId* cid, int fcnt, float* sizes)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_XRELSIZES, static_cast<void*>(&fcnt),
                          static_cast<void*>(sizes));
   return rval; 
}


int ChipsRemote::adjust_grid_yrelsizes(ChipsId* cid, int fcnt, float* sizes)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_YRELSIZES, static_cast<void*>(&fcnt),
                          static_cast<void*>(sizes));
   return rval; 
}


int ChipsRemote::grid_objects(ChipsId* cid, int cols, int rows, float xgap,
   float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_OBJS, static_cast<void*>(&cols), 
                          static_cast<void*>(&rows),static_cast<void*>(&xgap),
                          static_cast<void*>(&ygap),static_cast<void*>(&usage),
                          static_cast<void*>(&lcnt), (void*)lids);
   return rval; 
}


int ChipsRemote::row_grid_objects(ChipsId* cid, int cnt, int* vals,
    float xgap, float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   int ort = 0; 
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_OBJS_B, static_cast<void*>(&ort), 
                          static_cast<void*>(&cnt), static_cast<void*>(vals), 
                          static_cast<void*>(&xgap), static_cast<void*>(&ygap), 
                          static_cast<void*>(&usage), static_cast<void*>(&lcnt),
                          (void*)lids);
   return rval; 
}

int ChipsRemote::col_grid_objects(ChipsId* cid, int cnt, int* vals,
    float xgap, float ygap, int usage, int lcnt, string* lids)
{
   set_id_spec(cid);
   int ort = 1; 
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_GRID_OBJS_B, static_cast<void*>(&ort),
                          static_cast<void*>(&cnt), static_cast<void*>(vals),
                          static_cast<void*>(&xgap), static_cast<void*>(&ygap),
                          static_cast<void*>(&usage), static_cast<void*>(&lcnt),
                          (void*)lids);
   return rval;
}


int ChipsRemote::swap_object_positions(ChipsId* cid, string lid2)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SWAP_OBJ_POS, (void*)lid2.c_str());
   return rval; 
}


int ChipsRemote::view_single_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SINGLE_FRAME);
   return rval; 
}


int ChipsRemote::view_placed_frame(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_PLACED_FRAME);
   return rval;
}


int ChipsRemote::set_frame_bgcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_COLOR,(void*)in_color.c_str());
   return rval; 
}


int ChipsRemote::get_frame_bgcolor(ChipsId* cid, string& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_FRAME_COLOR, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_val = (char*) vp[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::set_frame_border_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_BCOLOR,(void*)in_color.c_str());
   return rval;
}


int ChipsRemote::get_frame_border_color(ChipsId* cid, string& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_FRAME_BCOLOR,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_val = (char*) vp[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::set_frame_transparency(ChipsId* cid, bool seethrough)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_TRANS, 
                          static_cast<void*>(&seethrough));
   return rval; 
}


int ChipsRemote::get_frame_transparency(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval =client->get_data(id_spec, MSG_GET_FRAME_TRANS,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = (((int*)vp[0])[0] != 0);
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::display_frame_border(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_BVIS, static_cast<void*>(&iv));
   return rval;
}


int ChipsRemote::hide_frame_border(ChipsId* cid)
{
   int iv = 0; 
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_BVIS, static_cast<void*>(&iv));
   return rval; 
}


int ChipsRemote::get_frame_border_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_FRAME_BVIS,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = (((int*)vp[0])[0] != 0);
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}

int ChipsRemote::display_frame(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_VIS, static_cast<void*>(&iv));
   return rval;
}


int ChipsRemote::hide_frame(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_VIS, static_cast<void*>(&iv));
   return rval; 
}


int ChipsRemote::get_frame_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_FRAME_VIS,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = (((int*)vp[0])[0] != 0);
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}


int ChipsRemote::set_frame_scale(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_FRAME_SCALE,
                          static_cast<void*>(&val));
   return rval; 
}


int ChipsRemote::get_frame_scale(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_FRAME_SCALE,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = (((int*)vp[0])[0] != 0);
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::blink_frames(ChipsId* cid, string fid2, float duration)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval =  client->send(id_spec, MSG_BLINK_FRAMES, (void*)fid2.c_str(),
                           static_cast<void*>(&duration));
   return rval; 
}


int ChipsRemote::stop_blink(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_STOP_BLINK);
   return rval; 
}


int ChipsRemote::arrange_frames(ChipsId* cid, int cnt, string* fids)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ARRANGE_FRAMES, static_cast<void*>(&cnt),
                          (void*)fids);
   return rval; 
}


int ChipsRemote::add_plot(ChipsId* cid, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ADD_PLOT, (void*)attlist.c_str());
   return rval; 
}


int ChipsRemote::add_plot(ChipsId* cid,
   float px1, float py1, float px2, float py2, string attlist)
{
   int rval;
   if ((rval = normalized_area_check(px1, py1, px2, py2)) == ERR_STATUS_GOOD)  
   {
      set_id_spec(cid);
      rval = this->create_server();
      if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
         rval = client->send(id_spec, MSG_ADD_PLOT_B, static_cast<void*>(&px1),
                             static_cast<void*>(&py1), static_cast<void*>(&px2),
                             static_cast<void*>(&py2), (void*)attlist.c_str());
   }
   return rval;
}


int ChipsRemote::delete_plot(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_DEL_PLOT);
   return rval; 
}


int ChipsRemote::clear_plot(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_CLEAR_PLOT);
   return rval; 
}


int ChipsRemote::add_curve(ChipsId* cid, long cnt, double* x, double* y, 
   Transform *transform, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ADD_CURVE_A, static_cast<void*>(&cnt), 
                          static_cast<void*>(x), static_cast<void*>(y), 
                          static_cast<void*>(transform), (void*)attlist.c_str());
   return rval; 
}


int ChipsRemote::add_curve(ChipsId* cid, long cnt, double* x, double*y, 
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_ADD_CURVE_B, static_cast<void*>(&cnt), 
                          static_cast<void*>(x), static_cast<void*>(y), 
                          static_cast<void*>(ydown), static_cast<void*>(yup),
                          static_cast<void*>(xdown), static_cast<void*>(xup),
                          static_cast<void*>(ylims), static_cast<void*>(xlims),
                          static_cast<void*>(transform), (void*)attlist.c_str());
   return rval; 
}


int ChipsRemote::delete_curve(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_DEL_CURVE);
   return rval;
}


int ChipsRemote::set_curve_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_DEPTH, 
                          static_cast<void*>(&in_size));
   return rval; 
}


int ChipsRemote::get_curve_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec, MSG_GET_CV_DEPTH, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((int*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::display_curve(ChipsId* cid)
{
   int iv = 1;
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_VISIBLE,static_cast<void*>(&iv));
   return rval; 
}


int ChipsRemote::hide_curve(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_VISIBLE, static_cast<void*>(&iv));
   return rval; 
}


int ChipsRemote::get_curve_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_CV_VISIBLE,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = (((int*)vp[0])[0] != 0);
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}


int ChipsRemote::set_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_SYMBOL, static_cast<void*>(&val));
   return rval; 
}


int ChipsRemote::get_curve_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_CV_SYMBOL,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((V_SYMBOL_TYPE*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::set_curve_symbolangle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_SYM_ANGLE, 
                          static_cast<void*>(&in_val));
   return rval; 
}


int ChipsRemote::get_curve_symbolangle(ChipsId* cid, float& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval=client->get_data(id_spec, MSG_GET_CV_SYM_ANGLE, &vp, &typ, &len, &cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
         out_val = ((float*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}

int ChipsRemote::set_curve_linestyle(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SET_CV_LINE_STYLE,static_cast<void*>(&val));
   return rval; 
}


int ChipsRemote::get_curve_linestyle(ChipsId* cid, V_LINE_TYPE& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      rval = client->get_data(id_spec,MSG_GET_CV_LINE_STYLE,&vp,&typ,&len,&cnt);
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((V_LINE_TYPE*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]); 
      free(vp); 
      free(typ); 
      free(len); 
   }
   return rval;
}


int ChipsRemote::set_curve_symbolfill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
   {
      if (in_fill)
         rval = client->send(id_spec, MSG_SET_CV_SYM_FILL);
      else 
         rval = client->send(id_spec, MSG_SET_CV_SYM_UNFILL);
   }
   return rval;
}


int ChipsRemote::get_curve_symbolfill(ChipsId* cid, bool& out_val)
{
   int cnt, rval = ERR_STATUS_BAD;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_SYM_FILL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_symbolsize(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_SYM_SIZE, 
                          static_cast<void*>(&in_size));
}


int ChipsRemote::get_curve_symbolsize(ChipsId* cid, int& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_SYM_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_symbolcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_SYM_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_curve_symbolcolor(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_CV_SYM_COLOR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_linethickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_LINE_SIZE, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_curve_linethickness(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_LINE_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_linecolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_LINE_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_curve_linecolor(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_CV_LINE_COLOR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_errthickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_SIZE, 
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_curve_errthickness(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_ERR_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_curve_errcaplength(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_CAPLEN,
                          static_cast<void*>(&in_size));
}


int ChipsRemote::get_curve_errcaplength(ChipsId* cid, int& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_ERR_CAPLEN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_curve_errcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_curve_errcolor(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_ERR_COLOR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::get_curve_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      outx1 = ((double*)vp[0])[0];
      outx2 = ((double*)vp[0])[1];
      outy1 = ((double*)vp[0])[2];
      outy2 = ((double*)vp[0])[3];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_curve_xrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_XRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_curve_yrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_YRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_curve_errup(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_UP, static_cast<void*>(&val));
}


int ChipsRemote::get_curve_errup(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_ERR_UP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_errdown(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_DOWN, static_cast<void*>(&val));
}


int ChipsRemote::get_curve_errdown(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_ERR_DOWN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_errleft(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_LEFT, 
                          static_cast<void*>(&val));
}


int ChipsRemote::get_curve_errleft(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_ERR_LEFT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_errright(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return  client->send(id_spec, MSG_SET_CV_ERR_RIGHT,
                        static_cast<void*>(&val));
}


int ChipsRemote::get_curve_errright(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_ERR_RIGHT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_limitoverride(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_LIMOVRIDE, static_cast<void*>(&val));
}


int ChipsRemote::get_curve_limitoverride(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CV_LIMOVRIDE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_curve_limitlength(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_LIMLENGTH,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_curve_limitlength(ChipsId* cid, double& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_LIMLENGTH,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}



int ChipsRemote::set_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_BADDATA, static_cast<void*>(&val));
}


int ChipsRemote::get_curve_baddatamode(ChipsId* cid, V_BADDATAMODE_TYPE& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_BADDATA,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_BADDATAMODE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_curve_errstyle(ChipsId* cid, string in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CV_ERR_STYLE, (void*)in_style.c_str());
}


int ChipsRemote::get_curve_errstyle(ChipsId* cid, string& out_style)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CV_ERR_STYLE,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_style = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::limits(ChipsId* cid, int dim, V_OBJECT_TYPE obj, ChipsId* ccid)
{
   set_id_spec(cid);
   string oid;
   if (!ccid)  return  ERR_NO_ID_OBJ;
   if (obj == chips_curve)          oid = ccid->get_curve_id();
   else if (obj == chips_contour)   oid = ccid->get_contour_id();
   else if (obj == chips_histogram) oid = ccid->get_histogram_id();
   else if (obj == chips_image)     oid = ccid->get_image_id();
   else  return ERR_LIMIT_OBJ; 
   int rval = this->create_server(); 
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LIMITS_CID, static_cast<void*>(&dim),
      static_cast<void*>(&obj), (void*)(ccid->get_window_id()),
      (void*)(ccid->get_frame_id()), (void*)(ccid->get_plot_id()), 
      (void*)oid.c_str());
}

int ChipsRemote::limits(ChipsId* cid, int dim, V_OBJECT_TYPE obj, string oid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LIMITS_OID, static_cast<void*>(&dim),
      static_cast<void*>(&obj), (void*)oid.c_str());
}

int ChipsRemote::limits(ChipsId* cid, double xmin, double xmax, double ymin, double ymax)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LIMITS_XX_YY,
                   static_cast<void*>(&xmin), static_cast<void*>(&xmax),
                   static_cast<void*>(&ymin), static_cast<void*>(&ymax));
   
}

int ChipsRemote::limits(ChipsId* cid, int dim, double min, double max)
{
   int msgid, rval;
   set_id_spec(cid);

   if (dim == X_AXIS)             msgid = MSG_SET_LIMITS_X;
      else if (dim == Y_AXIS)     msgid = MSG_SET_LIMITS_Y;
         else if (dim == XY_AXIS) msgid = MSG_SET_LIMITS_XY; 
            else return ERR_AXIS_INVALID_DIM;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, msgid, static_cast<void*>(&min),
                       static_cast<void*>(&max));
}


int ChipsRemote::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE amin,
   V_AUTOLIMIT_TYPE amax)
{
   char buff[256];
   int msgid, rval, lval = ERR_STATUS_GOOD;
   set_id_spec(cid);
   buff[0] = '\0';
   if (dim == X_AXIS)             msgid = MSG_SET_LIMITS_X_AA;
      else if (dim == Y_AXIS)     msgid = MSG_SET_LIMITS_Y_AA;
         else if (dim == XY_AXIS) msgid = MSG_SET_LIMITS_XY_AA; 
            else 
            {
               sprintf(buff, "The axis dimension (%1d) must be one of [X_AXIS|Y_AXIS|XY_AXIS|1|2|3]", dim);
               lval = ERR_AXIS_INVALID_DIM; 
            }
   if (chips_strcmp(amin.c_str(), "auto")) 
   {
      lval = ERR_INVALID_NUMERIC_ARG; 
      sprintf(buff, "The limits maximum value (%s) must be numeric, the keyword AUTO, or the string 'auto'.", amin.c_str());
   }
   else if (chips_strcmp(amax.c_str(), "auto")) 
   {
      lval = ERR_INVALID_NUMERIC_ARG; 
      sprintf(buff, "The limits maximum value (%s) must be numeric, the keyword AUTO, or the string 'auto'.", amax.c_str());
   }

   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   
   if (strlen(buff))
   {
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
      return lval;
   }
   else  
      return client->send(id_spec, msgid);
}


int ChipsRemote::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE amax)
{
   char buff[256];
   int msgid, rval, lval = ERR_STATUS_GOOD;
   set_id_spec(cid);
   buff[0] = '\0';
   if (dim == X_AXIS)             msgid = MSG_SET_LIMITS_X_MA;
      else if (dim == Y_AXIS)     msgid = MSG_SET_LIMITS_Y_MA;
         else if (dim == XY_AXIS) msgid = MSG_SET_LIMITS_XY_MA;
            else 
            {
               sprintf(buff, "The axis dimension (%1d) must be one of [X_AXIS|Y_AXIS|XY_AXIS|1|2|3]", dim);
               lval = ERR_AXIS_INVALID_DIM; 
            }
   if (chips_strcmp(amax.c_str(), "auto"))
   {
      lval = ERR_INVALID_NUMERIC_ARG;
      sprintf(buff, "The limits maximum value (%s) must be numeric, the keyword AUTO, or the string 'auto'.", amax.c_str());
   }
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (strlen(buff))
   {
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
      return lval;
   }
   else
      return client->send(id_spec, msgid, static_cast<void*>(&min));
}


int ChipsRemote::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE amin,
   double max)
{
   char buff[256];
   int msgid, rval, lval = ERR_STATUS_GOOD;
   set_id_spec(cid);
   buff[0] = '\0';
   if (dim == X_AXIS)             msgid = MSG_SET_LIMITS_X_AM;
      else if (dim == Y_AXIS)     msgid = MSG_SET_LIMITS_Y_AM;
         else if (dim == XY_AXIS) msgid = MSG_SET_LIMITS_XY_AM;
            else 
            {
               sprintf(buff, "The axis dimension (%1d) must be one of [X_AXIS|Y_AXIS|XY_AXIS|1|2|3]", dim);
               lval = ERR_AXIS_INVALID_DIM; 
            }
   if (chips_strcmp(amin.c_str(), "auto"))
   {
      lval = ERR_INVALID_NUMERIC_ARG;
      sprintf(buff, "The limits minimum value (%s) must be numeric, the keyword AUTO, or the string 'auto'.", amin.c_str());
   }
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (strlen(buff))
   {
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
      return lval;
   }
   else
      return client->send(id_spec, msgid, static_cast<void*>(&max));
}

int ChipsRemote::zoom(ChipsId* cid, float scale)
{
   int rval;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ZOOM_PLOT, static_cast<void*>(&scale));
}

int ChipsRemote::zoom(ChipsId* cid)
{
   int rval;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ZOOM_PLOT_GUI);
}

int ChipsRemote::panto(ChipsId* cid, double x, double y)
{
   int rval;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_PANTO_PLOT, static_cast<void*>(&x), static_cast<void*>(&y));
}

int ChipsRemote::pan(ChipsId* cid)
{
   int rval;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_PAN_PLOT);
}



int ChipsRemote::pick_limits(ChipsId* cid, int dim)
{
   int rval;
   set_id_spec(cid);
   if (!(dim == X_AXIS || dim == Y_AXIS || dim == XY_AXIS)) 
      return ERR_AXIS_INVALID_DIM;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_PICK_LIMITS, static_cast<void*>(&dim));
}


int ChipsRemote::set_plot_title(ChipsId* cid, string in_title)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return  client->send(id_spec, MSG_SET_PLOT_TITLE, (void*)in_title.c_str());
}


int ChipsRemote::get_plot_title(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_PLOT_TITLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_plot_title_color(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_TITLE_COLOR, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_TITLE_ANGLE,static_cast<void*>(&in_val));
}


int ChipsRemote::get_plot_title_angle(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_TITLE_ANGLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_HALIGN,
		       static_cast<void*>(&in_val));
}


int ChipsRemote::get_plot_title_halign(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_TITLE_HALIGN,&vp,&typ,&len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_plot_title_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_VALIGN,
		       static_cast<void*>(&in_val));
}


int ChipsRemote::get_plot_title_valign(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_TITLE_VALIGN,&vp,&typ,&len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_FONT, (void*)font.c_str());
}


int ChipsRemote::get_plot_title_font(ChipsId* cid, string& out_font)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TITLE_FONT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_font = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_FNTSTYL, (void*)style.c_str());
}


int ChipsRemote::get_plot_title_fontstyle(ChipsId* cid, string& out_style)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_TITLE_FNTSTYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_style = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_SIZE, static_cast<void*>(&sz));
}


int ChipsRemote::get_plot_title_size(ChipsId* cid, int& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval= client->get_data(id_spec, MSG_GET_TITLE_SIZE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_xpos(ChipsId* cid, float xpos_norm)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return  client->send(id_spec, MSG_SET_TITLE_X_POS,
                        static_cast<void*>(&xpos_norm));
}


int ChipsRemote::get_plot_title_xpos(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_TITLE_X_POS, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_ypos(ChipsId* cid, float ypos_norm)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TITLE_Y_POS,
                       static_cast<void*>(&ypos_norm));
}


int ChipsRemote::get_plot_title_ypos(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_TITLE_Y_POS, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_title_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_plot_title_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_PLOT_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_axisstyle(ChipsId* cid, V_AXISSTYLE_TYPE style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_AX_STYLE,
                       static_cast<void*>(&style));
}


int ChipsRemote::set_plot_axisstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   V_AXISSTYLE_TYPE astyle;
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval = string_to_v_axisstyle(style.c_str(),&astyle)) == ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_PLOT_AX_STYLE, 
                          static_cast<void*>(&astyle));
   else 
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   return rval;
}

int ChipsRemote::get_plot_axisstyle(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_PLOT_AX_STYLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_cornerstyle(ChipsId* cid, string style)
{
   V_CORNERSTYLE_TYPE cstyle;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval = string_to_v_cornerstyle(style.c_str(),&cstyle)) == ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_PLOT_CORNER,
		          static_cast<void*>(&cstyle));
   else 
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   return rval;
}


int ChipsRemote::set_plot_cornerstyle(ChipsId* cid, V_CORNERSTYLE_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_CORNER, static_cast<void*>(&val));
}


int ChipsRemote::get_plot_cornerstyle(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_PLOT_CORNER,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_plot_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      outx1 = ((double*)vp[0])[0];
      outx2 = ((double*)vp[0])[1];
      outy1 = ((double*)vp[0])[2];
      outy2 = ((double*)vp[0])[3];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_plot_xrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_XRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_plot_yrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_YRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_plot_xlabel(ChipsId* cid, string in_label)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_XLABEL, (void*)in_label.c_str());
}


int ChipsRemote::set_plot_ylabel(ChipsId* cid, string in_label)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_YLABEL, (void*)in_label.c_str());
}


int ChipsRemote::display_plot(ChipsId* cid)
{
   int iv = 1; 
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::hide_plot(ChipsId* cid)
{
   int iv = 0;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_plot_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_data_aspect_ratio(ChipsId* cid, string ratio)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_DATA_ASPRATIO, (void*)ratio.c_str());
}

int ChipsRemote::get_data_aspect_ratio(ChipsId* cid, string& out_ratio)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_DATA_ASPRATIO, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
       out_ratio = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_plot_aspect_ratio(ChipsId* cid, string ratio)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_ASPRATIO, (void*)ratio.c_str());
}


int ChipsRemote::get_plot_aspect_ratio(ChipsId* cid, string& out_ratio)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_PLOT_ASPRATIO, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_ratio = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::get_plot_aspect_width(ChipsId* cid, float& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_ASPW,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::get_plot_aspect_height(ChipsId* cid, float& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_ASPH,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::display_axis(ChipsId* cid)
{
   int rval, iv = 1; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_VISIBLE,static_cast<void*>(&iv));
}


int ChipsRemote::hide_axis(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_axis_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_leftmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_LMARGIN, 
                       static_cast<void*>(&margin_sz));
}


int ChipsRemote::get_plot_leftmargin(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_LMARGIN, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_rightmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_RMARGIN,
                       static_cast<void*>(&margin_sz));
}


int ChipsRemote::get_plot_rightmargin(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_PLOT_RMARGIN, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_topmargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_TMARGIN,
                          static_cast<void*>(&margin_sz));
}


int ChipsRemote::get_plot_topmargin(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_TMARGIN, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_plot_bottommargin(ChipsId* cid, float margin_sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_PLOT_BMARGIN,
                       static_cast<void*>(&margin_sz));
}


int ChipsRemote::get_plot_bottommargin(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_PLOT_BMARGIN, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::reverse_axes(ChipsId* cid, int dim)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_REVERSE_AXIS, static_cast<void*>(&dim));
}


int ChipsRemote::set_axis_scalable(ChipsId* cid, bool scalable)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_SCALABLE, 
                          static_cast<void*>(&scalable));
}


int ChipsRemote::display_major_ticks(ChipsId* cid)
{
   int rval, ival = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_VIS,static_cast<void*>(&ival));
}


int ChipsRemote::hide_major_ticks(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_VIS, static_cast<void*>(&iv));
}


int ChipsRemote::get_major_ticks_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_TICKPOS, 
                       static_cast<void*>(&arraysize),
                       static_cast<void*>(pos), (void*)vals);
}


int ChipsRemote::set_axis_major_mode(ChipsId* cid, string in_mode)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_AXIS_MAJ_MODE,(void*)in_mode.c_str());
}


int ChipsRemote::get_axis_major_mode(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_MODE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_THK, 
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_axis_major_thickness(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_THK,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_STYL, 
                       static_cast<void*>(&in_style));
}


int ChipsRemote::set_axis_major_style(ChipsId* cid, string style)
{
   V_TICKSTYLE_TYPE tstyle; 
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD) 
      rval = client->send(id_spec, MSG_SET_AXIS_MAJ_STYL, 
                          static_cast<void*>(&tstyle));
   else 
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   return rval;
}


int ChipsRemote::get_axis_major_style(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_STYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_ticklabel_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSTCKLB_STYL, 
                          static_cast<void*>(&in_style));
}


int ChipsRemote::set_axis_ticklabel_style(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   V_TICKSTYLE_TYPE tstyle;
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
   {
      if (tstyle == chips_centered) 
      {
         rval = ERR_BAD_TICKLABSTYLE;
         client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                      (void*)"default"); 
      }
      else 
         rval = client->send(id_spec, MSG_SET_AXSTCKLB_STYL, 
                             static_cast<void*>(&tstyle));
   }
   else 
   {
      rval = ERR_BAD_TICKLABSTYLE;
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   }
   return rval;
}


int ChipsRemote::get_axis_ticklabel_style(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSTCKLB_STYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::display_minor_ticks(ChipsId* cid)
{
   int rval, iv = 1; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_VIS, static_cast<void*>(&iv));
}


int ChipsRemote::hide_minor_ticks(ChipsId* cid)
{
   int rval, iv = 0; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_VIS, static_cast<void*>(&iv));
}


int ChipsRemote::get_minor_ticks_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_minor_mode(ChipsId* cid, string in_mode)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_AXIS_MIN_MODE,(void*)in_mode.c_str());
}


int ChipsRemote::get_axis_minor_mode(ChipsId* cid, string& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_MODE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_THK, 
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_axis_minor_thickness(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_THK,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_STYL, 
                       static_cast<void*>(&in_style));
}



int ChipsRemote::set_axis_minor_style(ChipsId* cid, string style)
{
   V_TICKSTYLE_TYPE tstyle;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval = string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_AXIS_MIN_STYL,
                          static_cast<void*>(&tstyle));
   else 
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   return rval;
}


int ChipsRemote::get_axis_minor_style(ChipsId* cid, string& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_STYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_parallel_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_PAR_OFF, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_parallel_offset(ChipsId* cid, float& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_PAR_OFF,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_perpendicular_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_PERP_OFF,  
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_perpendicular_offset(ChipsId* cid, float& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_PERP_OFF,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_automin(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_AUTOMIN, 
                          static_cast<void*>(&val));
}


int ChipsRemote::get_axis_automin(ChipsId* cid, bool& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_AXIS_AUTOMIN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_automax(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_AUTOMAX, 
                          static_cast<void*>(&val));
}


int ChipsRemote::get_axis_automax(ChipsId* cid, bool& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_AXIS_AUTOMAX,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_pad(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_PAD,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_pad(ChipsId* cid, double& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_PAD,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_axis_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_AXIS_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_color(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_COLOR, (void*)in_val.c_str());
}

int ChipsRemote::get_axis_ticklabel_color(ChipsId* cid, string& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_COLOR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_SIZE, static_cast<void*>(&sz));
}

int ChipsRemote::get_axis_ticklabel_size(ChipsId* cid, int& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_offset(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_TCKLB_OFFSET, static_cast<void*>(&sz));
}

int ChipsRemote::get_axis_ticklabel_offset(ChipsId* cid, int& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_OFFSET,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_ANGLE,
		       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_ticklabel_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_ANGLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_HALIGN,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_axis_ticklabel_halign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_HALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_ticklabel_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_VALIGN,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_ticklabel_valign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_VALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_ticklabel_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_FONT, (void*)font.c_str());
}


int ChipsRemote::get_axis_ticklabel_font(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_FONT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_ticklabel_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_FNTSTYL, (void*)style.c_str());
}


int ChipsRemote::get_axis_ticklabel_fontstyle(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_FNTSTYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_label_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_axis_label_color(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_COLOR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_label_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_VIS, static_cast<void*>(&vis));
}

int ChipsRemote::get_axis_label_visible(ChipsId* cid, bool& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_label_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_FONT, (void*)font.c_str());
}

int ChipsRemote::get_axis_label_font(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_FONT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_axis_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_FNTSTYL, (void*)style.c_str());
}

int ChipsRemote::get_axis_label_fontstyle(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_FNTSTYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_label_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_SIZE, static_cast<void*>(&sz));
}


int ChipsRemote::get_axis_label_size(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_ANGLE,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_label_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_ANGLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_xlabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_XLAB_ANG,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_xlabel_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_XLAB_ANG,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_axis_ylabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_YLAB_ANG,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_ylabel_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_YLAB_ANG,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_axis_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_HALIGN,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_label_halign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len =NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_HALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSLB_VALIGN,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_axis_label_valign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSLB_VALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::linear_scale(ChipsId* cid, int dim)
{
   if (!(dim == X_AXIS || dim == Y_AXIS || dim == XY_AXIS)) 
      return ERR_AXIS_INVALID_DIM;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINEARSCALE, static_cast<void*>(&dim));
}


int ChipsRemote::log_scale(ChipsId* cid, int dim)
{
   if (!(dim == X_AXIS || dim == Y_AXIS || dim == XY_AXIS)) 
      return ERR_AXIS_INVALID_DIM;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LOGSCALE, static_cast<void*>(&dim));
}


int ChipsRemote::get_axis_scale(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_SCALE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_label_color(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_COLOR, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}



int ChipsRemote::set_label_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_TEXT, (void*)in_text.c_str());
}


int ChipsRemote::get_label_text(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_TEXT, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_size(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_SIZE, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_label_size(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_SIZE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_label_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_label_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::display_label(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::hide_label(ChipsId* cid)
{
   int rval, iv = 0; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_label_visible(ChipsId* cid, bool& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_LAB_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_ANGLE, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_label_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_ANGLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_HALIGN, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_label_halign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_HALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_VALIGN, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_label_valign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_VALIGN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_font(ChipsId* cid, string in_font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_FONT, (void*)in_font.c_str());
}


int ChipsRemote::get_label_font(ChipsId* cid, string& out_font)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_FONT, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_font = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAB_STYLE, (void*)style.c_str());
}


int ChipsRemote::get_label_fontstyle(ChipsId* cid, string& out_style)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LAB_STYLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_style = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval; 
   if (xdim < 1)
   { 
      char buff[256];
      sprintf(buff, "The x dim value (%1d) must be greater than 0", (int)xdim);
      rval = ERR_INT_MIN;
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff); 
   }
   else if (ydim < 1)
   { 
      char buff[256];
      sprintf(buff, "The y dim value (%1d) must be greater than 0", (int)ydim);
      rval = ERR_INT_MIN;
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff); 
   }
   else 
      rval = client->send(id_spec, MSG_ADD_CONTOUR_B, static_cast<void*>(dat),
                       static_cast<void*>(&xdim), static_cast<void*>(&ydim),
                       static_cast<void*>(&numlev), static_cast<void*>(levels),
		       static_cast<void*>(trans), (void*)attlist.c_str());
   return rval;
}


int ChipsRemote::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int numlev, double* levels, Transform *trans, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_CONTOUR, static_cast<void*>(&cnt),
                       static_cast<void*>(x), static_cast<void*>(y),
                       static_cast<void*>(z), static_cast<void*>(&numlev),
                       static_cast<void*>(levels), static_cast<void*>(trans), 
                       (void*)attlist.c_str());
}


int ChipsRemote::delete_contour(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_CONTOUR);
}


int ChipsRemote::display_contour(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_VISIBLE,static_cast<void*>(&iv));
}


int ChipsRemote::hide_contour(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_contour_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}




int ChipsRemote::set_contour_mode(ChipsId* cid, string md)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_MODE, (void*)md.c_str());
}


int ChipsRemote::get_contour_mode(ChipsId* cid, string& out_md)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_CNTR_MODE,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_md = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_contour_style(ChipsId* cid, string styl)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_STYLE, (void*)styl.c_str());
}


int ChipsRemote::get_contour_style(ChipsId* cid, string& out_styl)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_STYLE,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_styl = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_contour_color(ChipsId* cid, string inc)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_COLOR, (void*)inc.c_str());
}


int ChipsRemote::get_contour_color(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_COLOR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_contour_thickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_THICK, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_contour_thickness(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CNTR_THICK,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::get_contour_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      outx1 = ((double*)vp[0])[0];
      outx2 = ((double*)vp[0])[1];
      outy1 = ((double*)vp[0])[2];
      outy2 = ((double*)vp[0])[3];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::get_contour_xrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_XRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::get_contour_yrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_YRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_contour_zrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_ZRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_contour_levels(ChipsId* cid, unsigned int arraysize,
   double* pos)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_LEVELS, 
                          static_cast<void*>(&arraysize),
                          static_cast<void*>(pos));
}

int ChipsRemote::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_LEVELS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && len && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      for (int ii=0; ii<*len; ii++)
         pos.push_back (((double*)(vp[0]))[ii]);
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}



int ChipsRemote::set_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_ALG, static_cast<void*>(&val));
}


int ChipsRemote::get_contour_algorithm(ChipsId* cid, V_CONTOUR_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int cnt, rval;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_ALG,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_CONTOUR_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_contour_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_INT,
                          static_cast<void*>(&in_interval));
}


int ChipsRemote::get_contour_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_INT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}



int ChipsRemote::set_contour_numlevels(ChipsId* cid, int in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_NUMLEV,
                          static_cast<void*>(&in_val));
}


int ChipsRemote::get_contour_numlevels(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CNTR_NUMLEV,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_contour_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CNTR_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_contour_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CNTR_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_contour_transform (ChipsId *cid, Transform *trans)
{
   return ERR_FEATURE_DNE;  /* DEPRECATED */
}


int ChipsRemote::get_contour_transform (ChipsId *cid, Transform **trans)
{
   return ERR_FEATURE_DNE;  /* DEPRECATED */
}

int ChipsRemote::apply_contour_transform (ChipsId *cid)
{
   return ERR_FEATURE_DNE;  /* DEPRECATED */
}


int ChipsRemote::apply_contour_transform_inverse (ChipsId *cid)
{
   return ERR_FEATURE_DNE;  /* DEPRECATED */
}


int ChipsRemote::add_histogram(ChipsId* cid, long cnt, double* bin_low,
   double* bin_high, double* counts, double* yd, double* yu, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_HISTOGRAM, static_cast<void*>(&cnt),
                       static_cast<void*>(counts), 
                       static_cast<void*>(bin_low),static_cast<void*>(bin_high),
                       static_cast<void*>(yd), static_cast<void*>(yu),  
                       (void*)attlist.c_str());
}


int ChipsRemote::delete_histogram(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_HISTOGRAM);
}


int ChipsRemote::display_histogram(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_VISIBLE,static_cast<void*>(&iv));
}


int ChipsRemote::hide_histogram(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_histogram_visible(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_SYMBOL, static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_symboltype(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_SYMBOL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_SYMBOL_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_symbolangle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_SYM_ANGLE, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_histogram_symbolangle(ChipsId* cid, float& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_HST_SYM_ANGLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_linestyle(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_LINE_STYL, 
                          static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_linestyle(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid); 
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_LINE_STYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_symbolfill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (in_fill)
      rval = client->send(id_spec, MSG_SET_HST_SYM_FILL);
   else
      rval = client->send(id_spec, MSG_SET_HST_SYM_UNFIL);
   return rval;
}


int ChipsRemote::get_histogram_symbolfill(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_SYM_FILL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_symbolsize(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_SYM_SIZE,
                          static_cast<void*>(&in_size));
}


int ChipsRemote::get_histogram_symbolsize(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_SYM_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_symbolcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   this->create_server();
   return client->send(id_spec, MSG_SET_HST_SYM_COLOR,
                          (void*)in_color.c_str());
}


int ChipsRemote::get_histogram_symbolcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_HST_SYM_COLOR,&vp,&typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_linethickness(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_LINE_SIZE, 
                          static_cast<void*>(&in_size));
}


int ChipsRemote::get_histogram_linethickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_LINE_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_linecolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_LINE_COLR,
                          (void*)in_color.c_str());
}


int ChipsRemote::get_histogram_linecolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_HST_LINE_COLR,&vp,&typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_errthickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_SIZE, 
                          static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_histogram_errthickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_ERR_SIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_histogram_errcaplength(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_CPLEN,
                          static_cast<void*>(&in_size));
}


int ChipsRemote::get_histogram_errcaplength(ChipsId* cid, int& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_ERR_CPLEN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_histogram_errcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_histogram_errcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int rval, cnt;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_HST_ERR_COLOR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_errup(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_UP, static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_errup(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_ERR_UP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_errdown(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_DOWN, static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_errdown(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_ERR_DOWN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_dropline(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_DROPLINE, static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_dropline(ChipsId* cid, bool& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_DROPLINE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_histogram_errstyle(ChipsId* cid, string in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_ERR_STYLE, (void*)in_style.c_str());
}


int ChipsRemote::get_histogram_errstyle(ChipsId* cid, string& out_style)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_HST_ERR_STYLE,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_style = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_histogram_fillopacity(ChipsId* cid, double val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_FILL_OPAC,
                          static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_fillopacity(ChipsId* cid, double& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_FILL_OPAC,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_histogram_fillcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_FILL_COLR, (void*)in_color.c_str());
}


int ChipsRemote::get_histogram_fillcolor(ChipsId* cid, string& out_color)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_HST_FILL_COLR,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_color = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_FILL_STYL,static_cast<void*>(&val));
}


int ChipsRemote::get_histogram_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_FILL_STYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_FILL_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_histogram_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      outx1 = ((double*)vp[0])[0];
      outx2 = ((double*)vp[0])[1];
      outy1 = ((double*)vp[0])[2];
      outy2 = ((double*)vp[0])[3];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_histogram_xrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_XRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_histogram_yrange(ChipsId* cid, double& out1, double& out2)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_HST_YRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_histogram_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_HST_DEPTH,static_cast<void*>(&in_size));
}


int ChipsRemote::get_histogram_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_HST_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::add_axis(ChipsId* cid, int dim, float pos,
   float minx, float maxx, float miny, float maxy, 
   Transform *transform, string aid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_AXIS, static_cast<void*>(&dim), 
                       static_cast<void*>(&pos), static_cast<void*>(&minx), 
                       static_cast<void*>(&maxx), static_cast<void*>(&miny),
                       static_cast<void*>(&maxy), (void*)transform,
                       (void*)aid.c_str());
}

int ChipsRemote::delete_axis(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_AXIS);
}


int ChipsRemote::bind_axes(ChipsId* cid, string lid, string aid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_AXIS_BIND, (void*)lid.c_str(),
                    (void*)aid.c_str());
}

int ChipsRemote::unbind_axes(ChipsId* cid, string lid, string aid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_AXIS_UNBIND, (void*)lid.c_str(),
                          (void*)aid.c_str());
}


int ChipsRemote::set_axis_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_COLOR,(void*)in_color.c_str());
}


int ChipsRemote::get_axis_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_COLOR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_CLR, (void*)in_color.c_str());
}


int ChipsRemote::get_axis_major_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_CLR, (void*)in_color.c_str());
}


int ChipsRemote::get_axis_minor_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_TEXT,(void*)in_text.c_str());
}


int ChipsRemote::get_axis_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_TEXT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_xlabel_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_XLAB_TXT,(void*)in_text.c_str());
}


int ChipsRemote::get_axis_xlabel_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_XLAB_TXT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_axis_ylabel_text(ChipsId* cid, string in_text)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_YLAB_TXT,(void*)in_text.c_str());
}


int ChipsRemote::get_axis_ylabel_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_YLAB_TXT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_axis_tickformat(ChipsId* cid, string in_format)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_AXIS_TICKFMT,(void*)in_format.c_str());
}


int ChipsRemote::get_axis_tickformat(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_TICKFMT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_THICK, 
                          static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_axis_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_THICK,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_tickstyle(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_TICKSTYL, 
                          static_cast<void*>(&in_style));
}


int ChipsRemote::set_axis_tickstyle(ChipsId* cid, string in_style)
{
   V_TICKSTYLE_TYPE tstyle;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval=string_to_v_tickstyle(in_style.c_str(),&tstyle))==ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_AXIS_TICKSTYL,
                          static_cast<void*>(&tstyle));
   else 
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default"); 
   return rval;
}


int ChipsRemote::get_axis_tickstyle(ChipsId* cid, string& out_val)
{
   // tickstyle is actually majortick style- use that method
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_TICKSTYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*)vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_num(ChipsId* cid, int in_num)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_NUM, 
                       static_cast<void*>(&in_num));
}


int ChipsRemote::get_axis_major_num(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_NUM,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_num(ChipsId* cid, int in_num)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_NUM,
                       static_cast<void*>(&in_num));
}


int ChipsRemote::get_axis_minor_num(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_NUM,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_length(ChipsId* cid, int in_len)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_LEN, 
                       static_cast<void*>(&in_len));
}


int ChipsRemote::get_axis_major_length(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_LEN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_length(ChipsId* cid, int in_len)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_LEN, 
                          static_cast<void*>(&in_len));
}


int ChipsRemote::get_axis_minor_length(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_LEN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_major_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MAJ_INT,
                          static_cast<void*>(&in_interval));
}


int ChipsRemote::get_axis_major_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MAJ_INT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minor_interval(ChipsId* cid, double in_interval)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXIS_MIN_INT,
                       static_cast<void*>(&in_interval));
}


int ChipsRemote::get_axis_minor_interval(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_MIN_INT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_ticklabel_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_TCKLB_VIS, static_cast<void*>(&vis));
}


int ChipsRemote::get_axis_ticklabel_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_TCKLB_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_majorgrid_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MJ_VIS, static_cast<void*>(&vis));
}


int ChipsRemote::get_axis_majorgrid_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MJ_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_majorgrid_thickness(ChipsId* cid, float in_thick)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MJ_SZ, 
                          static_cast<void*>(&in_thick));
}


int ChipsRemote::get_axis_majorgrid_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MJ_SZ,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MJ_STL, 
                       static_cast<void*>(&style));
}


int ChipsRemote::get_axis_majorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MJ_STL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_majorgrid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MJ_CLR,(void*)in_color.c_str());
}


int ChipsRemote::get_axis_majorgrid_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MJ_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minorgrid_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MN_VIS, static_cast<void*>(&vis));
}


int ChipsRemote::get_axis_minorgrid_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MN_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minorgrid_thickness(ChipsId* cid, float in_thick)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MN_SZ, 
                          static_cast<void*>(&in_thick));
}


int ChipsRemote::get_axis_minorgrid_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MN_SZ,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MN_STL, 
                          static_cast<void*>(&style));
}


int ChipsRemote::get_axis_minorgrid_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MN_STL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_axis_minorgrid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_AXSG_MN_CLR,(void*)in_color.c_str());
}


int ChipsRemote::get_axis_minorgrid_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXSG_MN_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::get_axis_range(ChipsId* cid, double& out1, double& out2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_AXIS_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::add_image(ChipsId* cid, string filename, Transform *transform, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_IMAGE, (void*)filename.c_str(),
                       (void*)transform, (void*)attlist.c_str());
}

int ChipsRemote::add_image(ChipsId* cid,
   string filename, float ix1, float iy1, float ix2, float iy2, 
   Transform *transform, string attlist)
{
   cerr <<"Client/Server for this feature not implemented yet" << endl;
   return ERR_FEATURE_DNE;
}


int ChipsRemote::add_image(ChipsId* cid, double* rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, 
   Transform *transform, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_IMAGE_C, (void*)rvals, (void*)gvals,
                     (void*)bvals,(void*)avals, static_cast<void*>(&xaxis), 
                     static_cast<void*>(&yaxis), (void*)transform,
                     (void*)attlist.c_str());
}


int ChipsRemote::delete_image(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_IMAGE);
}

int ChipsRemote::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_IMAGE_CHANNEL, (void*)vals,
                  static_cast<void*>(&xaxis), static_cast<void*>(&yaxis),
                  static_cast<void*>(&channel));
}

int ChipsRemote::remove_image_channel(ChipsId* cid, int channel)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_IMAGE_CHANNEL, static_cast<void*>(&channel));
}


int ChipsRemote::set_image_invert_colormap(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_INVCMAP, static_cast<void*>(&val));
}


int ChipsRemote::get_image_invert_colormap(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_INVCMAP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_image_colormap_interpolate(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_CMAPINTRP, static_cast<void*>(&val));
}

int ChipsRemote::get_image_colormap_interpolate(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_CMAPINTRP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_image_colormap_size(ChipsId* cid, int val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_CMAPSIZE, static_cast<void*>(&val));
}

int ChipsRemote::get_image_colormap_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_CMAPSIZE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_image_colormap(ChipsId* cid, V_COLORMAP_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return  client->send(id_spec, MSG_SET_IMG_CMAP, static_cast<void*>(&val));
}

int ChipsRemote::get_image_colormap(ChipsId* cid, V_COLORMAP_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_CMAP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_COLORMAP_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return  client->send(id_spec, MSG_SET_IMG_INTERP, static_cast<void*>(&val));
}

int ChipsRemote::get_image_interpolation(ChipsId* cid, V_INTERPOLATION_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_INTERP,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_INTERPOLATION_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_image_alpha(ChipsId* cid, double min, double max)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_ALPHA, 
                      static_cast<void*>(&min), static_cast<void*>(&max));
}

int ChipsRemote::get_image_alpha(ChipsId* cid, double& out1, double& out2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_ALPHA,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_image_threshold(ChipsId* cid, double min, double max)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_THRESH, 
                      static_cast<void*>(&min), static_cast<void*>(&max));
}

int ChipsRemote::get_image_threshold(ChipsId* cid, double& out1, double& out2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_THRESH,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_image_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_DEPTH, 
                       static_cast<void*>(&in_size));
}

int ChipsRemote::get_image_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_IMG_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_image_scale_channels(ChipsId* cid, bool val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_SCL_CHNNL, static_cast<void*>(&val));
}

int ChipsRemote::get_image_scale_channels(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_SCL_CHNNL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::display_image(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_VISIBLE,static_cast<void*>(&iv));
}

int ChipsRemote::hide_image(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_IMG_VISIBLE, static_cast<void*>(&iv));
}

int ChipsRemote::get_image_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::get_image_range(ChipsId* cid, double& outx1, double& outx2,
   double& outy1, double& outy2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_RANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      outx1 = ((double*)vp[0])[0];
      outx2 = ((double*)vp[0])[1];
      outy1 = ((double*)vp[0])[2];
      outy2 = ((double*)vp[0])[3];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_image_xrange(ChipsId* cid, double& out1, double& out2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_XRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::get_image_yrange(ChipsId* cid, double& out1, double& out2)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_IMG_YRANGE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
   {
      out1 = ((double*)vp[0])[0];
      out2 = ((double*)vp[0])[1];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::add_colorbar(ChipsId* cid, double xpos, 
   double ypos, string obj_id, V_OBJECT_TYPE obj, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_CBAR,
               static_cast<void*>(&xpos), static_cast<void*>(&ypos), 
               (void*)(obj_id.c_str ()), static_cast<void*>(&obj), 
               (void*)attlist.c_str());
}

int ChipsRemote::delete_colorbar(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_CBAR);
}

int ChipsRemote::add_label(ChipsId* cid, double xpos, 
   double ypos, string text, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_LABEL,
               static_cast<void*>(&xpos), static_cast<void*>(&ypos), 
               (void*)text.c_str(), (void*)attlist.c_str());
}

int ChipsRemote::set_colorbar_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_DEPTH, 
                       static_cast<void*>(&in_size));
}

int ChipsRemote::get_colorbar_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_width(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_WIDTH,
                       static_cast<void*>(&in_size));
}

int ChipsRemote::get_colorbar_width(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_WIDTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_length(ChipsId* cid, float in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LENGTH,
                       static_cast<void*>(&in_size));
}

int ChipsRemote::get_colorbar_length(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_LENGTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::display_colorbar_border(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_BVIS, static_cast<void*>(&iv));
}


int ChipsRemote::hide_colorbar_border(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_BVIS, static_cast<void*>(&iv));
}


int ChipsRemote::get_colorbar_border_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_BVIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_HALIGN,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_halign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_HALIGN,&vp,&typ,&len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_VALIGN,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_valign(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_VALIGN,&vp,&typ,&len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_label_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_ANG,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_label_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_LAB_ANG,&vp,&typ,&len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_label_text(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_TXT, (void*)in_val.c_str());
}

int ChipsRemote::get_colorbar_label_text(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_TXT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_colorbar_label_color(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_CLR, (void*)in_val.c_str());
}

int ChipsRemote::get_colorbar_label_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_label_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_SZ, static_cast<void*>(&sz));
}

int ChipsRemote::get_colorbar_label_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_SZ,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_color(ChipsId* cid, string in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_CLR, (void*)in_val.c_str());
}

int ChipsRemote::get_colorbar_ticklabel_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_CLR,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_size(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_SZ, static_cast<void*>(&sz));
}

int ChipsRemote::get_colorbar_ticklabel_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_SZ,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::display_colorbar(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_VISIBLE,static_cast<void*>(&iv));
}

int ChipsRemote::hide_colorbar(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_VISIBLE, static_cast<void*>(&iv));
}

int ChipsRemote::get_colorbar_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_orientation(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_ORIENT,
                       static_cast<void*>(&in_size));
}

int ChipsRemote::get_colorbar_orientation(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_ORIENT, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_label_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_VIS, static_cast<void*>(&vis));
}

int ChipsRemote::get_colorbar_label_visible(ChipsId* cid, bool& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_label_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_FONT, (void*)font.c_str());
}

int ChipsRemote::get_colorbar_label_font(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_FONT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_label_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_FSTY, (void*)style.c_str());
}

int ChipsRemote::get_colorbar_label_fontstyle(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_FSTY,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_HAL,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_ticklabel_halign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_HAL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_VAL,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_ticklabel_valign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_VAL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_ANG,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_ticklabel_angle(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_ANG,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_label_halign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_HAL,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_label_halign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_HAL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_label_valign(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_VAL,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_label_valign(ChipsId* cid, float& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_VAL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_VIS, static_cast<void*>(&vis));
}

int ChipsRemote::get_colorbar_ticklabel_visible(ChipsId* cid, bool& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_font(ChipsId* cid, string font)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_FONT, (void*)font.c_str());
}

int ChipsRemote::get_colorbar_ticklabel_font(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_FONT,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_fontstyle(ChipsId* cid, string style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TL_FSTY, (void*)style.c_str());
}

int ChipsRemote::get_colorbar_ticklabel_fontstyle(ChipsId* cid, string& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_FSTY,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_colorbar_label_location(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_LAB_LOC,
                       static_cast<void*>(&in_style));
}

int ChipsRemote::set_colorbar_label_location(ChipsId* cid, string style)
{
   V_TICKSTYLE_TYPE tstyle = chips_inside;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;

   rval = string_to_v_tickstyle(style.c_str(),&tstyle);
   if (tstyle == chips_centered)  rval = ERR_BAD_TICKSTYLE;
   if (rval == ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_CBAR_LAB_LOC,
                          static_cast<void*>(&tstyle));
   else
   {
      string errm = "The label location value (" + style + 
         ") must be one of [inside, outside]";
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*) errm.c_str());
   }
   return rval;
}

int ChipsRemote::get_colorbar_label_location(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_LAB_LOC,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}
int ChipsRemote::set_colorbar_tick_style(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_STY,
                       static_cast<void*>(&in_style));
}

int ChipsRemote::set_colorbar_tick_style(ChipsId* cid, string style)
{
   V_TICKSTYLE_TYPE tstyle;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if ((rval=string_to_v_tickstyle(style.c_str(),&tstyle)) == ERR_STATUS_GOOD)
      rval = client->send(id_spec, MSG_SET_CBAR_TK_STY,
                          static_cast<void*>(&tstyle));
   else
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"default");
   return rval;
}

int ChipsRemote::get_colorbar_tick_style(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TK_STY,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_mode(ChipsId* cid, string style)
{
   string val = "";
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (!chips_strcmp(style.c_str(),"none") || !chips_strcmp(style.c_str(),"0"))
      val = "none"; 
   else if (!chips_strcmp(style.c_str(),"ends") || !chips_strcmp(style.c_str(),"1"))
      val = "ends"; 
   else if (!chips_strcmp(style.c_str(),"nice") || !chips_strcmp(style.c_str(),"2"))
      val = "nice";
   if (val.length()) 
      rval = client->send(id_spec,MSG_SET_CBAR_TK_MODE,(void*)val.c_str());
   else
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)"The colorbar tickmode style most be one of [none|ends|nice]");
   return rval;
}

int ChipsRemote::get_colorbar_tick_mode(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TK_MODE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_location(ChipsId* cid, string in_style)
{
   V_TICKSTYLE_TYPE tstyle;
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=string_to_v_tickstyle(in_style.c_str(),&tstyle);
   if (tstyle == chips_inside || tstyle == chips_outside)  
      rval = client->send(id_spec, MSG_SET_CBAR_TK_LOC,
                          static_cast<void*>(&tstyle));
   else
   {
      char buff[256];
      sprintf(buff, "The specified location value (%s) must be either [inside|outside]", in_style.c_str());
      rval = ERR_BAD_TICKSTYLE;
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
   }
   return rval;
}

int ChipsRemote::set_colorbar_tick_location(ChipsId* cid, int in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_LOC,
                       static_cast<void*>(&in_style));
}

int ChipsRemote::get_colorbar_tick_location(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TK_LOC,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_visible(ChipsId* cid, bool vis)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_VIS, static_cast<void*>(&vis));
}

int ChipsRemote::get_colorbar_tick_visible(ChipsId* cid, bool& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TK_VIS,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_CLR,(void*)in_color.c_str());
}


int ChipsRemote::get_colorbar_tick_color(ChipsId* cid, string& out_color)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_TK_CLR, &vp, &typ, &len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_color = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_length(ChipsId* cid, int in_length)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_LEN, static_cast<void*>(&in_length));
}

int ChipsRemote::get_colorbar_tick_length(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TK_LEN,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_colorbar_parallel_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_PAR_OFF,
                       static_cast<void*>(&in_val));
}

int ChipsRemote::get_colorbar_parallel_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_PAR_OFF,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_perpendicular_offset(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_PERP_OFF,
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_colorbar_perpendicular_offset(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_PERP_OFF,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_ticklabel_offset(ChipsId* cid, int sz)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec,MSG_SET_CBAR_TL_OFF, static_cast<void*>(&sz));
}

int ChipsRemote::get_colorbar_ticklabel_offset(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_CBAR_TL_OFF,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_colorbar_tick_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_CBAR_TK_THK,
                       static_cast<void*>(&in_thickness));
}

int ChipsRemote::get_colorbar_tick_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_CBAR_TK_THK,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::delete_label(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_LABEL);
}


int ChipsRemote::add_line(ChipsId* cid,
   double startx, double starty, double endx, double endy, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_LINE,
                  static_cast<void*>(&startx), static_cast<void*>(&starty),
                  static_cast<void*>(&endx), static_cast<void*>(&endy), 
                  (void*)attlist.c_str());
}


int ChipsRemote::add_line(ChipsId* cid,
   double* xx, double* yy, unsigned int count, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_LINE_B,
                  static_cast<void*>(xx), static_cast<void*>(yy),
                  static_cast<void*>(&count), (void*)attlist.c_str());
}


int ChipsRemote::delete_line(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_LINE);
}


int ChipsRemote::set_line_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_COLOR,(void*)in_color.c_str());
}


int ChipsRemote::get_line_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LINE_COLOR, &vp, &typ, &len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_line_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_line_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LINE_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_line_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_THICK, 
                          static_cast<void*>(&in_thickness));
}

int ChipsRemote::get_line_thickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid); 
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_LINE_THICK,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_line_style(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_STYLE, 
                          static_cast<void*>(&in_style));
}


int ChipsRemote::get_line_style(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_LINE_STYLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_EXTEND,
                          static_cast<void*>(&in_val));
}


int ChipsRemote::get_line_extend(ChipsId* cid, V_LINE_EXTENSION_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_LINE_EXTEND,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_EXTENSION_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::display_line(ChipsId* cid)
{
   int rval, iv = 1; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::hide_line(ChipsId* cid)
{
   int rval, iv = 0; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LINE_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_line_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_LINE_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::add_point(ChipsId* cid,
   double xx, double yy, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_POINT,
                    static_cast<void*>(&xx), static_cast<void*>(&yy),
                    (void*)attlist.c_str());
}


int ChipsRemote::add_point(ChipsId* cid,
   double xx, double  yy, double  zz, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_POINT_B,
                    static_cast<void*>(&xx), static_cast<void*>(&yy),
                    static_cast<void*>(&zz), (void*)attlist.c_str());
}


int ChipsRemote::delete_point(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_POINT);
}



int ChipsRemote::set_point_size(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_SIZE, 
                        static_cast<void*>(&in_size));
}


int ChipsRemote::get_point_size(ChipsId* cid, int& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_POINT_SIZE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_point_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_ANGLE, 
                       static_cast<void*>(&in_val));
}


int ChipsRemote::get_point_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_POINT_ANGLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_point_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_COLOR,(void*)in_color.c_str());
}


int ChipsRemote::get_point_color(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec, MSG_GET_POINT_COLOR, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_point_fill(ChipsId* cid, bool in_fill)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_FILL, 
                          static_cast<void*>(&in_fill));
}


int ChipsRemote::get_point_fill(ChipsId* cid, bool& out_fill)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_POINT_FILL, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_fill = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_point_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_point_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_POINT_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_point_symbol(ChipsId* cid, V_SYMBOL_TYPE in_symbol)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_SYMBOL, 
                       static_cast<void*>(&in_symbol));
}


int ChipsRemote::get_point_symbol(ChipsId* cid, V_SYMBOL_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_POINT_SYMBOL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_SYMBOL_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::display_point(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_VISIBLE, 
                       static_cast<void*>(&iv));
}


int ChipsRemote::hide_point(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_POINT_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::get_point_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_POINT_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_window_title(ChipsId* cid, string title)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_WIN_TITLE, (void*)title.c_str());
}


int ChipsRemote::get_window_title(ChipsId* cid, string& out_title)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_WIN_TITLE, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         out_title = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_redraw(bool val)
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REDRAW, static_cast<void*>(&val));
}


int ChipsRemote::set_window_redraw(ChipsId* cid,  bool draw)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (draw)   return client->send(id_spec, MSG_SET_WIN_DRAW_ON);
      else     return client->send(id_spec, MSG_SET_WIN_DRAW_OFF);
}


int ChipsRemote::get_window_redraw(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_WIN_REDRAW,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::set_window_bgcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_WINDOW_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::get_window_bgcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_WINDOW_COLOR, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}


int ChipsRemote::iconify_window(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ICONIFY_WIN);
}


int ChipsRemote::deiconify_window(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEICONIFY_WIN);
}


int ChipsRemote::display_grid(ChipsId* cid, int scale)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SHOW_GRID, static_cast<void*>(&scale));
}


int ChipsRemote::hide_grid(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_HIDE_GRID);
}


int ChipsRemote::set_grid_linestyle(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_GRID_STYLE, 
                       static_cast<void*>(&in_style));
}


int ChipsRemote::set_grid_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_GRID_COLOR, (void*)in_color.c_str());
}


int ChipsRemote::set_grid_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_GRID_THICK, 
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::set_grid_scale(ChipsId* cid, int in_scale)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_GRID_SCALE,
                       static_cast<void*>(&in_scale));
}


int ChipsRemote::set_layout_linestyle(ChipsId* cid, V_LINE_TYPE in_style)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAYOUT_STYLE, 
                       static_cast<void*>(&in_style));
}


int ChipsRemote::set_layout_color(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAYOUT_COLOR,(void*)in_color.c_str());
}


int ChipsRemote::set_layout_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LAYOUT_THICK, 
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::display_frame_layout(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SHOW_LAYOUT);
}


int ChipsRemote::hide_frame_layout(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_HIDE_LAYOUT);
}


int ChipsRemote::add_legend(ChipsId* cid, const char** curve_ids,
   const char** curve_labels, string lgid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_LEGEND, (void*)curve_ids,
                       (void*)curve_labels, (void*)lgid.c_str());
}

int ChipsRemote::delete_legend(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_LEGEND);
}

int ChipsRemote::hide_legend(ChipsId* cid)
{
   int rval, iv = 2; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_VIS, static_cast<void*>(&iv));
}

int ChipsRemote::display_legend(ChipsId* cid)
{
   int rval, iv = 1; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_VIS, static_cast<void*>(&iv));
}

int ChipsRemote::set_legend_color(ChipsId* cid, string in_clr)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_COLOR, (void*)in_clr.c_str());
}

int ChipsRemote::set_legend_borders(ChipsId* cid, bool showborders)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_BORDER, 
                       static_cast<void*>(&showborders));
}

int ChipsRemote::set_legend_thickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_THICK, 
                       static_cast<void*>(&in_thickness));
}

int ChipsRemote::set_legend_transparency(ChipsId* cid, bool seethrough)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_LEGEND_TRANS, static_cast<void*>(&seethrough));
}


int ChipsRemote::add_region(ChipsId* cid,
   double* xx, double* yy, unsigned int count, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_REGION,
                  static_cast<void*>(xx), static_cast<void*>(yy),
                  static_cast<void*>(&count), (void*)attlist.c_str());
}


int ChipsRemote::add_region(ChipsId* cid,
   int sides, double xx, double yy, double radius, string attlist)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_ADD_REGION_B, static_cast<void*>(&sides),
                  static_cast<void*>(&xx), static_cast<void*>(&yy),
                  static_cast<void*>(&radius), (void*)attlist.c_str());
}


int ChipsRemote::delete_region(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_REGION);
}


int ChipsRemote::display_region(ChipsId* cid)
{
   int rval, iv = 1;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_VISIBLE, static_cast<void*>(&iv));
}


int ChipsRemote::hide_region(ChipsId* cid)
{
   int rval, iv = 0;
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_VISIBLE, static_cast<void*>(&iv));
}

int ChipsRemote::get_region_visible(ChipsId* cid, bool& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_REGN_VISIBLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = (((int*)vp[0])[0] != 0);
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_depth(ChipsId* cid, int in_size)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_DEPTH, 
                       static_cast<void*>(&in_size));
}


int ChipsRemote::get_region_depth(ChipsId* cid, int& out_val)
{
   int rval, cnt, *len = NULL;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_REGN_DEPTH, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((int*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::set_region_angle(ChipsId* cid, float in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_ANGLE,static_cast<void*>(&in_val));
}


int ChipsRemote::get_region_angle(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_REGN_ANGLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_edgethickness(ChipsId* cid, float in_thickness)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_THICK,
                       static_cast<void*>(&in_thickness));
}


int ChipsRemote::get_region_edgethickness(ChipsId* cid, float& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_REGN_THICK,&vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_FLOAT))
      out_val = ((float*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_edgestyle(ChipsId* cid, V_LINE_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_STYLE, static_cast<void*>(&val));
}


int ChipsRemote::get_region_edgestyle(ChipsId* cid, V_LINE_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_REGN_STYLE,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_LINE_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_fillstyle(ChipsId* cid, V_FILL_TYPE val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_FILSTYL, static_cast<void*>(&val));
}


int ChipsRemote::get_region_fillstyle(ChipsId* cid, V_FILL_TYPE& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_REGN_FILSTYL,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
      out_val = ((V_FILL_TYPE*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_edgecolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_EDGCOL, (void*)in_color.c_str());
}


int ChipsRemote::get_region_edgecolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_REGN_EDGCOL, &vp, &typ, &len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::set_region_fillcolor(ChipsId* cid, string in_color)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_FILCOL, (void*)in_color.c_str());
}


int ChipsRemote::get_region_fillcolor(ChipsId* cid, string& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_REGN_FILCOL, &vp, &typ, &len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::set_region_opacity(ChipsId* cid, double in_val)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_REGN_OPACITY,
		       static_cast<void*>(&in_val));
}


int ChipsRemote::get_region_opacity(ChipsId* cid, double& out_val)
{
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_REGN_OPACITY,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::pick(ChipsId* cid, int count, unsigned int mask)
{
   int rval, ocnt; 
   set_id_spec(cid);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   CHIPS_PICK_EVT_P plist = NULL; 
   lock(); 
   if (count == 0)  count--;  // treat 0 and negative as loop until [esc] 
   for (int rr = count; rr--; )
   {
      rval = get_pick(cid, ocnt, &plist, 1, mask); 
      if (ocnt && plist)
      {
         cout << "(";
         if (plist->x_str.length () > 0)
            cout << plist->x_str;
         else 
            cout << plist->x;
         cout << ", ";
         if (plist->x_str.length () > 0)
            cout << plist->y_str;
         else
            cout << plist->y;
         cout << ")" << endl;
         delete []plist; 
      }
      else 
         rr = 0; 
   }
   unlock();
   return rval; 
}


int ChipsRemote::get_pick(ChipsId* cid, int& outcount, 
   CHIPS_PICK_EVT_P* picklist, int count, unsigned int mask)
{
   int cnt, rval;
   set_id_spec(cid);
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   int x_str_pos = 0;
   int y_str_pos = 0;
   
   outcount = 0;
   *picklist = static_cast <CHIPS_PICK_EVT_P>(0);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_GET_PICK, &vp, &typ, &len, &cnt,
                      static_cast<void*>(&count), static_cast<void*>(&mask));
   if ((cnt == 7) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
   {
      // set count
      outcount = ((int*)vp[0])[0];

      if (outcount > 0)
      {
         *picklist = new CHIPS_PICK_EVT[outcount]; 
	 
         for (int rr = outcount; rr--; )
         {
            ((*picklist)[rr]).x = ((double*)vp[1])[rr]; 
            ((*picklist)[rr]).y = ((double*)vp[2])[rr]; 
            ((*picklist)[rr]).event = ((int*)vp[3])[rr]; 
            ((*picklist)[rr]).key = ((char*)vp[4])[rr]; 
         }

         for (int rr=0; rr < outcount; rr++)
         {
            ((*picklist)[rr]).x_str = &((char*)vp[5])[x_str_pos];
            ((*picklist)[rr]).y_str = &((char*)vp[6])[y_str_pos];
            x_str_pos += ((*picklist)[rr]).x_str.length ()+1;
            y_str_pos += ((*picklist)[rr]).y_str.length ()+1;
         }
      }
   }
   free(vp);
   free(typ);
   free(len);
   return rval;
}


int ChipsRemote::print_frame(ChipsId* cid, string outfile)
{
   set_id_spec(cid);
   string filepath = outfile;
   // since the server can't resolve paths for the client, we need
   // to add the path to outfile, if it is not provided.
   if (outfile.length() && (outfile.find('/') != 0))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
	 filepath.append("/");
	 filepath.append(outfile);
      }
   }
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_PRINT_FRAME, (void*)filepath.c_str());
}


int ChipsRemote::print_window(ChipsId* cid, string outfile, string prefs)
{
   set_id_spec(cid);
   string filepath = outfile;
   // since the server can't resolve paths for the client, we need
   // to add the path to outfile, if it is not provided.
   if (outfile.length())
   {
      if (outfile.find("/") != 0)
      {
         char pathbuf[1024];
         if (getcwd(pathbuf, 1024))
         {
            filepath = pathbuf;
	    filepath.append("/");
	    filepath.append(outfile);
         }
      }
   }
   else
   {
      // if we have a print dialog we need to seed default path
      int pos = prefs.find("printdialog=");
      if (pos != -1) 
      {
         pos+=12;
	 if (tolower(prefs[pos]) == 'y' || prefs[pos] == '1' ||
             tolower(prefs[pos]) == 't' || 
	     (tolower(prefs[pos]) == 'o' && tolower(prefs[pos+1]) == 'n'))
         {
	    char pathbuf[1024];
            if (getcwd(pathbuf, 1024))  filepath = pathbuf;
	 }
      }
   }

   // server needs to get umask of client to set the output file permission
   mode_t mv;
   umask(mv = umask(0));
   int  rval, imv = (int)mv; 
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_PRINT_WINDOW, (void*)filepath.c_str(),
                       (void*)prefs.c_str(), static_cast<void*>(&imv));
}

int ChipsRemote::print_image(ChipsId* cid, string outfile, string prefs)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   if (!outfile.length())
   {
      char buff[] = "print_image requires an output filename.";
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                          (void*)buff);
      return MSG_ADD_CLIENT_ERR; 
   }
   string filepath = outfile;
   // since the server can't resolve paths for the client, we need
   // to add the path to outfile, if it is not provided.
   if (outfile.find("/") != 0)
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(outfile);
      }
   }
   // server needs to get umask of client to set the output file permission
   mode_t mv;
   umask(mv = umask(0));
   int  imv = (int)mv;
   return client->send(id_spec, MSG_PRINT_IMAGE, (void*)filepath.c_str(),
                       (void*)prefs.c_str(), static_cast<void*>(&imv));
}

int ChipsRemote::delete_all_windows(ChipsId* cid)
{
   set_id_spec(cid, 1);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_ALL_WINS);
}


int ChipsRemote::delete_all_frames(ChipsId* cid)
{
   set_id_spec(cid);
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_ALL_FRAMES);
}


int ChipsRemote::delete_all_plots(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_ALL_PLOTS);
}


int ChipsRemote::delete_all_images(ChipsId* cid)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_DEL_ALL_IMAGES);
}


int ChipsRemote::normalized_area_check(float x1, float y1, float x2, float y2)
{
   char buff[256];
   int rval = ERR_STATUS_GOOD; 
   if (x1 < 0)  
   {
      sprintf(buff, "The x1 value (%4.3f) must be between 0.0 and 1.0", x1);
      rval = ERR_NORM_DATA_RANGE;
   }
   else if (x1 > 1)
   {
      sprintf(buff, "The x1 value (%4.3f) must be between 0.0 and 1.0", x1);
      rval = ERR_NORM_DATA_RANGE;
   }
   else if (x2 <= x1)
   {
      sprintf(buff, "The x2 value (%4.3f) must be greater than the x1 value (%4.3f)", x2, x1);
      rval = ERR_NORM_XMAX_LE_XMIN;
   }
   else if (x2 > 1)
   {
      sprintf(buff, "The x2 value (%4.3f) must be between 0.0 and 1.0", x2);
      rval = ERR_NORM_DATA_RANGE;
   }
   else if (y1 < 0)  
   {
      sprintf(buff, "The y1 value (%4.3f) must be between 0.0 and 1.0", y1);
      rval = ERR_NORM_DATA_RANGE;
   }
   else if (y1 > 1)
   {
      sprintf(buff, "The y1 value (%4.3f) must be between 0.0 and 1.0", y1);
      rval = ERR_NORM_DATA_RANGE;
   }
   else if (y2 <= y1)
   {
      sprintf(buff, "The y2 value (%4.3f) must be greater than the y1 value (%4.3f)", y2, y1);
      rval = ERR_NORM_YMAX_LE_YMIN;
   }
   else if (y2 > 1)
   {
      sprintf(buff, "The y2 value (%4.3f) must be between 0.0 and 1.0", y2);
      rval = ERR_NORM_DATA_RANGE; 
   }

   if (rval != ERR_STATUS_GOOD) // send any errors to the server 
   {
      rval = this->create_server();
      if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
      client->send(id_spec, MSG_ADD_CLIENT_ERR, static_cast<void*>(&rval),
                   (void*)buff);
   }
   return rval; 
}


int ChipsRemote::undo(int depth)
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_UNDO,  static_cast<void*>(&depth));
}


int ChipsRemote::undo(string tag)
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_UNDO_TAG, (void*)tag.c_str());
}


int ChipsRemote::redo(int depth)
{
   set_id_spec();
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_REDO,  static_cast<void*>(&depth));
}


int ChipsRemote::redo(string tag)
{
   set_id_spec();
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_REDO_TAG, (void*)tag.c_str());
}


int ChipsRemote::set_undo_tag(string tag)
{
   set_id_spec();
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_SET_UNDO_TAG, (void*)tag.c_str());
}


int ChipsRemote::clear_undo_tag(string tag)
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_CLEAR_UNDO_TAG, (void*)tag.c_str());
}


int ChipsRemote::open_undo_block()
{
   set_id_spec();
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_OPEN_UNDO_BLK);
}


int ChipsRemote::close_undo_block()
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_CLOSE_UNDO_BLK);
}


int ChipsRemote::open_undo_buffer()
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_OPEN_UNDO_BUFF);
}


int ChipsRemote::close_undo_buffer()
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_CLOSE_UNDO_BUFF);
}


int ChipsRemote::discard_undo_buffer()
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_CLEAR_UNDO_BUFF);
}


void ChipsRemote::set_id_spec(ChipsId* supplied, bool use_all)
{
   if (!supplied)
   {
      id_spec = &default_ids; 
      if (use_all)
         id_spec->set_all();
      else 
         id_spec->reset_window_id(""); 
   }
   else 
      id_spec = supplied;
}


int ChipsRemote::convert_coordinate (double *in_val, double *out_val, int count,
                               ChipsId *cid1, ChipsId *cid2, int axis)
{
   if (cid2 == NULL)
      return ERR_STATUS_BAD;

   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;

   string win = cid2->get_window_id ();
   string frm = cid2->get_frame_id ();
   string lyr = cid2->get_plot_id ();
   string coord_id = cid2->get_coord_id ();
   int coord_sys = (int)cid2->get_coord_sys ();
   set_id_spec (cid1);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_CONVERT_COORD_A,&vp,&typ,&len,&cnt,
			static_cast <void *> (in_val), 
			static_cast <void *> (&count), 
			(void *)win.c_str (), (void *)frm.c_str (), 
			(void *)lyr.c_str (), (void *)coord_id.c_str (), 
			static_cast <void *> (&coord_sys), 
			static_cast <void *> (&axis));

   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE)){
      for (int ii=0; ii<count; ii++)
         out_val[ii] = ((double*)vp[0])[ii];
   }
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);
   return rval;
}

int ChipsRemote::convert_coordinate (double in_val, double &out_val,
                               ChipsId *cid1, ChipsId *cid2, int axis)
{
   if (cid2 == NULL)  return ERR_STATUS_BAD;

   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   string win = cid2->get_window_id ();
   string frm = cid2->get_frame_id ();
   string lyr = cid2->get_plot_id ();
   string coord_id = cid2->get_coord_id ();
   int coord_sys = (int)cid2->get_coord_sys ();
   set_id_spec (cid1);
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec, MSG_CONVERT_COORD_B,&vp,&typ,&len,&cnt,
			static_cast <void *> (&in_val),
                        (void *)win.c_str (), (void *)frm.c_str (),
                        (void *)lyr.c_str (), (void *)coord_id.c_str (),
                        static_cast <void *> (&coord_sys),
                        static_cast <void *> (&axis));
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_DOUBLE))
      out_val = ((double*)vp[0])[0];
   if (vp && vp[0]) free(vp[0]);
   free(vp);
   free(typ);
   free(len);

   return rval;
}


int ChipsRemote::destroy_server()
{ 
   int rval = disconnect();
   client->send(id_spec, MSG_QUIT);
   return rval;
}

int ChipsRemote::lock()
{ 
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_LOCK);
}

int ChipsRemote::unlock()
{ 
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_UNLOCK);
}

int ChipsRemote::check_server_exists(const string& inServerId)
{ 
   return client->does_server_exist(inServerId);
}


int ChipsRemote::connect(const string& inServerId, string myid)
{ 
   string server_id = "";
   string new_server_id = inServerId;
   int rval = ChipsRemote::get_server_id (server_id);

   // if chips is connected to the specified server just return
   if ((rval == ERR_STATUS_GOOD) && (new_server_id == server_id))
   {
      errverb = -1;   // reset local error verbosity- only for creating server
      return rval;
   }
   if (!client->does_server_exist (new_server_id))
   {
      int verbose = 0;
      rval = get_error_verbosity (verbose);

      // currently if there's no server then the error message wont
      // be displayed. Display the error explicitly
      if (rval == ERR_NO_SOCKET)
         errmsg = "chips ERROR: The server " + new_server_id + " does not exist or is currently not accepting connections. Connection not established\n";
      return ERR_SOCK_SERVERID;
   }

   // if connected to a server disconnect from it
   if (server_id != "")  this->disconnect ();

   if ((rval = client->connect_server(new_server_id)) == ERR_STATUS_GOOD)
   {
      if (myid.length())
      {
         set_id_spec();
         client->send(id_spec, MSG_CMDLNE_CONN, (void*)myid.c_str());
      }
      errverb = -1;   // reset local error verbosity- only for creating server
   }
   return rval;
}

int ChipsRemote::disconnect()
{ 
   return client->disconnect_server();
}

int ChipsRemote::create_server(string id)
{
   return this->create_server(this->startup_prefs, id);
}

int ChipsRemote::create_server(bool load_prefs, string id)
{ 
   int rval = ERR_STATUS_GOOD;

   string server_id = "";
   rval = ChipsRemote::get_server_id (server_id);

   // if the chips is connected disconnect from that server
   if ((rval == ERR_STATUS_GOOD) && (server_id != ""))
      return ERR_SOCK_CONN;

   set_id_spec();

   if (id.length() == 0)
   {
      ostringstream tmpStr1;
      tmpStr1<<"ChipsServer_"<<getpid()<<"_"<<serverIndx++;
      server_id=tmpStr1.str();
   }
   else if (check_server_exists(id))
   {
      errmsg = "chips ERROR: Unable to create server- the requested id (";
      errmsg.append(id);
      errmsg.append(") is already in use\n");
      return ERR_DUP_SRVRID;
   }
   else   server_id = id;

   string command = "";

   if (offscreen_flag)
      command="chipsServer_os -s ";
   else 
      command="chipsServer -s ";
   command += server_id;
   if (load_prefs)   command += " -d";
   command += " &";     // launch as a background process

   // loop through and attempt to launch server and connect multiple
   // times so avoid hang/failures where the original intialization of
   // the chipsServer may have failed.
   int tries = 5;
   while(tries-- > 0)
   {
      launch_server(command); 
      int check = 10;
      while (check-- > 0)
      {
         if ((rval = client->connect_server(server_id, 5)) == ERR_STATUS_GOOD)
         {
            check = 0;
            tries = 0;
         }
      }
   }

   if (rval == ERR_STATUS_GOOD)
   {
      if (errverb != -1)
      {
         // verbosity was set locally, update the created server
         set_error_verbosity(errverb); 
         errverb = -1; 
      }
      // set default working directory for gui to cwd
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
         client->send(id_spec, MSG_SET_CWD, (void*)pathbuf);
   }
   else if (rval == ERR_SOCK_UNAVAIL)  
      errmsg = "chips ERROR: Unable to connect to the server\n"; 
   return  rval;
}


int ChipsRemote::get_server_id(string& out_val)
{
   int cnt, rval = ERR_STATUS_GOOD;
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;

   ChipsId cid; // dont use the id_spec as this function is called by many 
                // other functions and the set_id_spec will reset the id_spec

   rval = client->get_data(&cid,MSG_GET_SERVER_ID,&vp,&typ,&len,&cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

list <string> ChipsRemote::list_servers()
{
   if (client)
   {
      return client->list_servers();
   }
   else
   {
      list <string> slist;
      slist.clear();
      return slist;
   }
}



const char* ChipsRemote::get_errors(int mode)
{
   static string localerr; 
   const char* buf = NULL;
   set_id_spec();
   void** vp = NULL; 
   char *typ = NULL; 
   int *len = NULL; 
   int sts, cnt;
   sts = client->get_data(id_spec, MSG_GET_ERRS, &vp, &typ, &len, &cnt,
                    static_cast<void*>(&mode));
   if (sts == ERR_STATUS_GOOD)
   {
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
         buf = (char*) vp[0]; 
      free(vp);
      free(typ); 
      free(len); 
   }
   else
   {
      // failed to get data, we probably haven't connected yet so check locally 
      localerr = errmsg;
      buf = localerr.c_str(); 
      errmsg.clear(); 
   }
   return buf; 
}


int ChipsRemote::set_error_verbosity(int val)
{
   int rval = ERR_STATUS_GOOD;
   set_id_spec();
   rval = client->send(id_spec, MSG_SET_ERR_VERB, static_cast<void*>(&val));

   if (rval == ERR_NO_SOCKET)
   {
      // client is not connected to a server yet- locally store verbosity 
      errverb = val; 
   }
   return client->send(id_spec, MSG_SET_ERR_VERB, static_cast<void*>(&val));
}


int ChipsRemote::clear_errors()
{
   set_id_spec();
   return client->send(id_spec, MSG_CLEAR_ERRS);
}


int ChipsRemote::get_error_verbosity(int &out_val)
{
   int cnt, rval = ERR_STATUS_GOOD;
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   set_id_spec();

   // Don't automatically connect to server. If not connected just return

   rval = client->get_data(id_spec,MSG_GET_ERR_VERB,&vp,&typ,&len,&cnt);
   if (rval == ERR_STATUS_GOOD)
   {
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((int*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   else
   {
      // not connected to the server yet, return local err value 
      out_val = errverb; 
   }
   return rval;
}

int ChipsRemote::get_error_count(int mode, int &out_val)
{
   int cnt, rval = ERR_STATUS_GOOD;
   void** vp = NULL;
   char *typ = NULL;
   int *len = NULL;
   set_id_spec();
   out_val = 0;

   // Don't automatically connect to server. If not connected just return

   rval = client->get_data(id_spec,MSG_GET_ERR_CNT,&vp,&typ,&len,&cnt,
                           static_cast<void*>(&mode));
   if (rval == ERR_STATUS_GOOD)
   {
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((V_LINE_TYPE*)vp[0])[0];

      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}

int ChipsRemote::get_error_types(int &out_val)
{
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   set_id_spec();
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval = client->get_data(id_spec,MSG_GET_ERR_TYPES,&vp,&typ,&len,&cnt);
   if (rval == ERR_STATUS_GOOD)
   {
      if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_INT))
         out_val = ((int*)vp[0])[0];
      if (vp && vp[0]) free(vp[0]);
      free(vp);
      free(typ);
      free(len);
   }
   return rval;
}

int ChipsRemote::load_fill(string in_file, int slot)
{
   set_id_spec(); 
   string filepath = in_file;
   
   // since the server can't resolve paths for the client, we need
   // to add the path to file, if it is not provided.
   if (in_file.length() && in_file.find("/") != 0)
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(in_file);
      }
   }
   
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_LOAD_FILL, (void*)filepath.c_str(),
                       static_cast<void*>(&slot));
}

int ChipsRemote::load_colormap(long len, double* r, double* g, double* b,
	double* alpha, int slot)
{
   set_id_spec();
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_LOAD_CMAP_ARRAY, 
		       static_cast<void*>(&len),
		       static_cast<void*>(&slot),
		       static_cast<void*>(r), 
		       static_cast<void*>(g), 
		       static_cast<void*>(b),
                       static_cast<void*>(alpha));
}
int ChipsRemote::load_colormap(string in_file, int slot)
{
   set_id_spec();
   string filepath = in_file;

   // since the server can't resolve paths for the client, we need
   // to add the path to file, if it is not provided.
   if (in_file.length() && in_file.find("/") != 0)
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         filepath.append("/");
         filepath.append(in_file);
      }
   }
   int rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_LOAD_CMAP, (void*)filepath.c_str(),
                       static_cast<void*>(&slot));
}

int ChipsRemote::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj)
{
   set_id_spec(cid);
   int rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_HILITE_OBJ,static_cast<void*>(&obj));
   return rval;
}

int ChipsRemote::show_gui(string fname)
{
   int rval = ERR_FILE_NOT_SUPPLIED;
   set_id_spec();

   string filepath = fname;
   // since the server can't resolve paths for the client, we need
   // to add the path to file, if it is not provided.
   if ((fname.length() && fname.find("/") != 0) || (fname.length() == 0))
   {
      char pathbuf[1024];
      if (getcwd(pathbuf, 1024))
      {
         filepath = pathbuf;
         if (fname.length())
         {
            filepath.append("/");
            filepath.append(fname);
         }
      }
   }
   rval = this->create_server();
   if (rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)
      rval = client->send(id_spec, MSG_SHOW_GUI, (void*)filepath.c_str());
   return rval;
}

int ChipsRemote::hide_gui()
{
   set_id_spec();
   int rval =this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   return client->send(id_spec, MSG_HIDE_GUI);
}

int ChipsRemote::get_server_cwd(string& out_val)
{
   set_id_spec();
   void** vp = NULL;
   char *typ = NULL;
   int rval, cnt, *len = NULL;
   rval = this->create_server();
   if (!(rval == ERR_STATUS_GOOD || rval == ERR_SOCK_CONN)) return rval;
   rval=client->get_data(id_spec,MSG_GET_CWD, &vp, &typ, &len, &cnt);
   if ((cnt == 1) && typ && vp && (typ[0] == MSG_DATATYPE_CHAR))
      out_val = (char*) vp[0];
   if (vp && vp[0]) free(vp[0]); 
   free(vp); 
   free(typ); 
   free(len); 
   return rval;
}

int ChipsRemote::launch_server(string cmd)
{
   FILE *fd;

   if (server_pid != "")
   {
      size_t ps = server_pid.find('\n'); 
      // check if server is already running... 
      string check = "ps -p ";
      if (ps != string::npos)
         check.append(server_pid.substr(0,ps));
      else
         check.append(server_pid);
      check.append(" > /dev/null; echo $?");

      if ((fd = popen(check.c_str(),"r")) != NULL)
      {
         char buf[256];
         memset(buf,0,sizeof(buf));
         fgets(buf, sizeof(buf), fd);
         pclose(fd);
         if (buf[0] == '0')  
         {
            return 1;    // found server  
         }
         else 
         {
            server_pid = ""; 
         }
      }
   } 

   string pidcmd = cmd;
   pidcmd.append(" echo \"$!\"");
   if ((fd = popen(pidcmd.c_str(),"r")) != NULL)
   {
      char buf[256];
      memset(buf,0,sizeof(buf));
      fgets(buf, sizeof(buf), fd);
      pclose(fd);
      server_pid = buf;
   }
   return 0;
} 

