// 
//  Copyright (C) 2004-2009,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <unistd.h>
#include <sys/wait.h>

#include "ChipsServer.hh"
#include "ChipsSystem.hh"
#include "ChipsId.hh"
#include "ChipsMessages.hh"

#include "Transform.hh"
#include "ChipsTransformFactory.hh"

#ifndef HEADLESS
#include "gui/ChipsGuiMain.hh"
#endif

#include "SCSServer.hh"
#include "SCSMessage.hh"

#include <stdio.h>
#ifdef HAVE_POLL_H
#  include <poll.h>
#elif defined HAVE_SYS_POLL_H
#  include <sys/poll.h>
#elif defined HAVE_SELECT_H
#  include <select.h>
#elif defined HAVE_SYS_SELECT_H
#  include <sys/select.h>
#endif

#ifndef HEADLESS
#include <gtk/gtkgl.h>
#include <gdk/gdkglconfig.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <limits.h>
#include <pwd.h>
#include <sstream>
#include <vector>

// define static members
pid_t  ChipsServer::pid = 0;
double ChipsServer::buffer[1024];
string ChipsServer::serverId="";
string ChipsServer::serverCwd="";
string ChipsServer::xvfb_pid="";
long  ChipsServer::lockFlag=0;
long   ChipsServer::lockClient=-1;
long   ChipsServer::currClientId;
int    ChipsServer::cmdlineConn = 0; 
ChipsSystem* ChipsServer::chipsSys = NULL;

#ifndef HEADLESS
ChipsGuiMain* chipsGui = NULL;
#endif

SCSServer chipsServer;
guint timeoutId;
//guint32 timeoutMs=50;
guint32 timeoutMs=2;
extern int errno;

// constructor 
ChipsServer::ChipsServer(bool load_prefs)
{
   if (!chipsSys)  chipsSys = new ChipsSystem(load_prefs); 
}


// destructor 
ChipsServer::~ChipsServer()
{
   if (chipsSys){
      delete chipsSys;
      chipsSys = NULL;
   }
   clean_xvfb();
   clean_gui();
}


// signal handler for SIGPIPE detected by server
void ChipsServer::sig_handler(int vv)
{
   // Shut down the scs server to remove the socket file
   chipsServer.Shutdown();
   if (chipsSys){
      delete chipsSys;
      chipsSys = NULL;
   }
   clean_xvfb();
   clean_gui();
}

int ChipsServer::clean_xvfb()
{
   if (xvfb_pid.length())
   {
      pclose(popen(xvfb_pid.c_str(), "r"));
      xvfb_pid = "";
   }
   return 0;
}

int ChipsServer::clean_gui()
{
#ifndef HEADLESS
   if (chipsGui)
   {
      delete chipsGui;
      chipsGui = NULL;
   }
#endif
   return 0;
}

// method to install server signal handlers
int ChipsServer::handle_server_signals()
{
   int rval = ERR_STATUS_GOOD;
   struct sigaction act;
   struct sigaction oact;

   act.sa_flags   = 0;
   sigset_t set;
   sigemptyset(&set);
   act.sa_mask = set;
  
   act.sa_handler = &ChipsServer::sig_handler;

   // handle broken pipe
   sigaction(SIGPIPE, &act, &oact);
   // handle terminating signals 
   sigaction(SIGTERM, &act, &oact);
   sigaction(SIGINT, &act, &oact);
   sigaction(SIGQUIT, &act, &oact);
   sigaction(SIGHUP, &act, &oact);

   return (rval); 
}


int ChipsServer::server_process(gpointer data)
{
   static int resetcnt = 1;
   static int connected = 1000;
   static int missed_cnt = 0; 
   gtk_timeout_remove(timeoutId);

   SCSMessage tmpMessage;
   ChipsServer* me = (ChipsServer*)data;
   int connCnt = chipsServer.GetConnectionsCount(); 

   if (resetcnt)
   {
      // first pass- check to see if we should adjust # of connection attempts
      char* tmpstr=getenv("CHIPS_SERVER_RETRIES");
      if (tmpstr)
      {
         int cnts = atoi(tmpstr);
         if (cnts > 0 && cnts < 1000000)
            connected = cnts;
      }
      resetcnt = 0;
   }

   if (connCnt == 0)
   {
      // after connecting if the count drops to 0 (for 2 cycles)  OR
      // after n cycles without connecting  - exit the server
      if (connected-- == 0)
      {
         chipsServer.Shutdown();
	 if (chipsSys)
         {
            delete chipsSys;
            chipsSys = NULL;
         }
         clean_xvfb();
         clean_gui();
         gtk_exit(0);
      }
   }
#ifndef HEADLESS
   else if (connCnt == 1)
   {
      // only have a single connection- is the gui shell still active
      if (chipsGui && chipsGui->get_visible() == 0 && (cmdlineConn == 1)) 
      {
         clean_gui(); 
         chipsServer.Shutdown();
         if (chipsSys)
         {
            delete chipsSys;
            chipsSys = NULL;
         }
         clean_xvfb();
         clean_gui();
         gtk_exit(0);
      }
   }
   // we've got a full connection- we shoould exit if we lose the connection
   if (connCnt > 1 && connected > 1)  connected = 1;
#else
   else if (connCnt > 0) connected = 1;
#endif

   if (lockFlag > 0)
   {
      // To avoid a permanent server lock if a client dies abnormally, track
      // failed receipts. After a set number of failures, consider the lock as
      // stale and release it. 
      missed_cnt++; 
      if ( (chipsServer.ReceiveMessage(0,currClientId,tmpMessage)) <= 0 )
      {
         if (missed_cnt == 75000)  
         {
            lockFlag--;
            missed_cnt = 0;
         }
         timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, me);
         return 1;
      }
   }
   else
   {
      if ( (currClientId = chipsServer.ReceiveMessage(0,tmpMessage)) <= 0 )
      {
          timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, me);
          return 1;
      }
   }

   if (tmpMessage.Type() == SCSMessage::PROTOCOL_CLOSE)
   {
      if (lockFlag)  lockFlag=0;
      timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, me);
      return 1;
   }

      unsigned int buff[1000];         // buffer for message header (id,size)
      unsigned int msglen = 0,         // specified message length w/o header
                   bytes_in_buffer;    // number of bytes in pipe
      int          rval = ERR_STATUS_GOOD;  // return status of request
      char*        pos = (char*)buff;  // pointer to message buffer
      double       pbuf[PIPE_BUF];     // copy buffer for short messages
      double       rbuf[PIPE_BUF];     // return buffer for short messages
      int          overflow = 0,       // indicator for long messages
                   respond = 1;        // need to send response back to client
      char*        rp = (char*)rbuf;   // pointer to return message buffer
      ChipsId*     cid = new ChipsId();

      // clear stale lock check
      missed_cnt = 0;

      // read the message id and message length (fixed length of 8 bytes)
      buff[0] = buff[1] = 0;

      if(! tmpMessage.Size() )
	{
	  timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, me);
	  return 0;
	};

      memcpy(pos, tmpMessage.Data(),8);
      msglen = buff[1] - 8;
      bytes_in_buffer = msglen;

      // ensure we don't overflow our buffer
      overflow = (msglen > PIPE_BUF);

      // make sure the message length is appropriate for the data -
      // Note: To eliminate the need of allocating and copying the raw data
      // to a buffer, some messages are handled as multiple message segments.
      // Because these exceptions may fail the msglen <= bytes_in_buffer test,
      // an explicit check is performed.
      if (((msglen <= bytes_in_buffer) || (buff[0] == MSG_ADD_CURVE_A) ||
          (buff[0] == MSG_ADD_CURVE_B) || (buff[0] == MSG_ADD_IMAGE_C) || 
          (buff[0] == MSG_ADD_IMAGE_CHANNEL) || (buff[0] == MSG_ADD_CONTOUR_B) || 
          (buff[0] == MSG_ADD_CONTOUR) || (buff[0] == MSG_ADD_HISTOGRAM) ||
          (buff[0] == MSG_ADD_LINE_B) || (buff[0] == MSG_ADD_REGION) ||
           overflow) && (rval == ERR_STATUS_GOOD))
      {
         char *bf = (char*)pbuf;     // buffer pointer
         int   bytes_left = msglen;  // number of bytes left to read from pipe
         long  ndx = 0;              // index into message buffer

         if (overflow)   // too large for preset buffer- allocate new buffer
            bf = (char*) malloc(msglen*sizeof(char));

	 memcpy(bf,tmpMessage.Data()+8,bytes_left);
	 tmpMessage.Clear();

         // populate the ChipsId structure
         if ((ndx = unpack_it(bf, cid)) > 0)
         {
            // handle message based on id
            switch(buff[0])
            {
	       case MSG_QUIT:
                  clean_gui();
	          chipsServer.Shutdown();
		  if (chipsSys){
      		     delete chipsSys;
      		     chipsSys = NULL;
   		  }
                  clean_xvfb();
                  clean_gui();
	          gtk_exit(0);
	       break;
	       case MSG_LOCK:
                  lock(currClientId);
	       break;
	       case MSG_UNLOCK:
                  unlock(currClientId);
	       break;
               case MSG_CLEAR_ERRS:
                  chipsSys->clear_errors();
               break; 
               case MSG_OPEN_UNDO_BLK:
                  chipsSys->open_undo_block();
               break;
               case MSG_CLOSE_UNDO_BLK:
                  chipsSys->close_undo_block();
               break; 
               case MSG_OPEN_UNDO_BUFF:
                  chipsSys->open_undo_buffer();
               break;
               case MSG_CLOSE_UNDO_BUFF:
                  chipsSys->close_undo_buffer();
               break; 
               case MSG_CLEAR_UNDO_BUFF:
                  chipsSys->discard_undo_buffer();
               break; 
               case MSG_ADD_WINDOW_A:
                  if ((rval = chipsSys->add_window(cid)) == ERR_STATUS_GOOD)
                     cid->reset_frame_id("");
               break;
               case MSG_SET_WIN_TITLE:
                  rval = chipsSys->set_window_title(cid, &bf[ndx]);
               break;
               case MSG_SET_CWD:
               {
#ifndef HEADLESS
                  serverCwd = (char*)&bf[ndx];
                  if (serverCwd.length() && chipsGui)
                     chipsGui->set_working_dir(serverCwd);
#endif
               }
               break; 
	       case MSG_UNDO_TAG:
               {
                  string new_tag = (char*)&bf[ndx];
                  rval = chipsSys->undo(new_tag);
               }
               break;
               case MSG_REDO_TAG:
               {
                  string new_tag = (char*)&bf[ndx];
                  rval = chipsSys->redo(new_tag);
               }
               break;
               case MSG_SET_UNDO_TAG:
               {
                  string new_tag = (char*)&bf[ndx];
                  rval = chipsSys->set_undo_tag(new_tag);
               }
               break;
               case MSG_CLEAR_UNDO_TAG:
               {
                  string old_tag = (char*)&bf[ndx];
                  rval = chipsSys->clear_undo_tag(old_tag);
               }
               break;
	       case MSG_UNDO:
      	       {
       	          int *depth = (int*)&bf[NEXT_INT_POS(ndx)];
 	          rval = chipsSys->undo(*depth);
 	       }
       	       break;
    	       case MSG_REDO:
      	       {
       	          int *depth = (int*)&bf[NEXT_INT_POS(ndx)];
	          rval = chipsSys->redo(*depth);
  	       }
       	       break;
               case MSG_SET_WINDOW_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_window_bgcolor(cid, new_color);
               }
               break;
               case MSG_PRINT_WINDOW:
               {
                  string outf, prefs;
		  int* md = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  mode_t mode = *md;
                  outf = (char*)&bf[ndx];
                  prefs = (char*)&bf[ndx + outf.length() + 1];
                  chipsSys->set_umask(mode);
                  rval = chipsSys->print_window(cid, outf, prefs);
               }
               break;
               case MSG_PRINT_IMAGE:
               {
                  string outf, prefs;
                  int* md = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  mode_t mode = *md;
                  outf = (char*)&bf[ndx];
                  prefs = (char*)&bf[ndx + outf.length() + 1];
                  chipsSys->set_umask(mode);
                  rval = chipsSys->print_image(cid, outf, prefs);
               }
               break;
               case MSG_PRINT_FRAME:
               {
                  string outf = (char*)&bf[ndx];
                  rval = chipsSys->print_frame(cid, outf);
               }
               break;
               case MSG_SET_PLOT_TITLE:
               {
                  string new_title = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_title(cid, new_title);
               }
               break;
               case MSG_SET_PLOT_ASPRATIO:
               {
                  string new_title = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_aspect_ratio(cid, new_title);
               }
               break;
	       case MSG_SET_DATA_ASPRATIO:
               {
                  string ratio = (char*)&bf[ndx];
                  rval = chipsSys->set_data_aspect_ratio(cid, ratio);
               }
               break;
               case MSG_SET_IMG_INVCMAP:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_image_invert_colormap(cid, *bval);
               }
               break;
               case MSG_SET_IMG_CMAPINTRP:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_image_colormap_interpolate(cid, *bval);
               }
               break;
               case MSG_SET_IMG_SCL_CHNNL:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_image_scale_channels(cid, *bval);
               }
               break;
               case MSG_SET_IMG_CMAPSIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_image_colormap_size(cid, *ival);
               }
               break;
               case MSG_SET_IMG_CMAP:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_COLORMAP_TYPE typ = (V_COLORMAP_TYPE)*style;
                  rval = chipsSys->set_image_colormap(cid, typ);
               }
               break;
               case MSG_SET_IMG_INTERP:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_INTERPOLATION_TYPE typ = (V_INTERPOLATION_TYPE)*style;
                  rval = chipsSys->set_image_interpolation(cid, typ);
               }
               break;
               case MSG_SET_PLOT_CORNER:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_CORNERSTYLE_TYPE typ = (V_CORNERSTYLE_TYPE)*style;
                  rval = chipsSys->set_plot_cornerstyle(cid, typ);
               }
               break;
               case MSG_DISPLAY_DEPTH:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->display_depth(cid, *ival);
               }
               break;
               case MSG_HIDE_DEPTH:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->hide_depth(cid, *ival);
               }
               break;
               case MSG_COLLAPSE_DEPTH:
               {
                  int *dest, *scnt, *vls = NULL;
                  dest = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  scnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if ((*scnt) > 0)
                  {
                     if ((vls = new int[*scnt]) != NULL)
                        for (int rr = 0; rr < (*scnt); rr++)      
                        {
                           vls[rr] = *(int*)&bf[NEXT_INT_POS(ndx)];
                           ndx = NEXT_INT_POS(ndx) + sizeof(int);
                        }
                  } 
                  rval =  chipsSys->collapse_depths(cid, *scnt, vls, *dest);
                  if (vls)  delete[] vls; 
               }
               break;
               case MSG_SET_CNTR_MODE:
               {
                  string new_mode = (char*)&bf[ndx];
                  rval = chipsSys->set_contour_mode(cid, new_mode);
               }
               break;
               case MSG_SET_CNTR_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_contour_color(cid, new_color);
               }
               break;
               case MSG_SET_CNTR_STYLE:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_contour_style(cid, new_style);
               }
               break;
               case MSG_SET_CNTR_NUMLEV:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_contour_numlevels(cid, *ival);
               }
               break;
               case MSG_SET_AXSLB_SIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_label_size(cid, *ival);
               }
               break;
               case MSG_SET_TCKLB_SIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_size(cid, *ival);
               }
               break;
               case MSG_SET_CBAR_TL_OFF:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_offset(cid, *ival);
               }
               break;
               case MSG_SET_TCKLB_OFFSET:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_offset(cid, *ival);
               }
               break;
               case MSG_SET_PLOT_XLABEL:
               {
                  string new_title = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_xlabel(cid, new_title);
               }
               break;
               case MSG_SET_PLOT_YLABEL:
               {
                  string new_title = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_ylabel(cid, new_title);
               }
               break;
               case MSG_SET_AXIS_AUTOMIN:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_automin(cid, *bval);
               }
               break;
               case MSG_SET_AXSG_MJ_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_majorgrid_visible(cid,*bval);
               }
               break;
               case MSG_SET_TCKLB_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_visible(cid,*bval);
               }
               break;
               case MSG_SET_AXSLB_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_label_visible(cid,*bval);
               }
               break;
               case MSG_SET_CBAR_LAB_VIS:
               {
                  bool *bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_visible(cid, *bval);
               }
               break;
               case MSG_SET_AXSG_MN_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_minorgrid_visible(cid,*bval);
               }
               break;
               case MSG_SET_CBAR_TK_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_colorbar_tick_visible(cid,*bval);
               }
               break;
               case MSG_SET_CBAR_TL_VIS:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_visible(cid,*bval);
               }
               break;
               case MSG_SET_AXIS_AUTOMAX:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_axis_automax(cid, *bval);
               }
               break;
               case MSG_ICONIFY_WIN:
                  rval = chipsSys->iconify_window(cid);
               break;
               case MSG_DEICONIFY_WIN:
                  rval = chipsSys->deiconify_window(cid);
               break;
               case MSG_SET_CURRENT_WIN:
                  rval = chipsSys->set_current_window(cid);
               break;
               case MSG_SET_CURRENT_FRM:
                  rval = chipsSys->set_current_frame(cid);
               break;
               case MSG_SET_CURRENT_IMG:
                  rval = chipsSys->set_current_image(cid);
               break;
               case MSG_SET_CURRENT_CBAR:
                  rval = chipsSys->set_current_colorbar(cid);
               break;
               case MSG_SET_CURRENT_PLT:
                  rval = chipsSys->set_current_plot(cid);
               break;
               case MSG_SET_CURRENT_CRV:
                  rval = chipsSys->set_current_curve(cid);
               break;
               case MSG_SET_CURRENT_CNTR:
                  rval = chipsSys->set_current_contour(cid);
               break;
               case MSG_SET_CURRENT_HST:
                  rval = chipsSys->set_current_histogram(cid);
               break;
               case MSG_SET_CURRENT_AXS:
                  rval = chipsSys->set_current_axis(cid);
               break;
               case MSG_SET_CURRENT_LAB:
                  rval = chipsSys->set_current_label(cid);
               break;
               case MSG_SET_CURRENT_LINE:
                  rval = chipsSys->set_current_line(cid);
               break;
               case MSG_SET_CURRENT_PNT:
                  rval = chipsSys->set_current_point(cid);
               break;
               case MSG_SET_CURRENT_LGD:
                  rval = chipsSys->set_current_legend(cid);
               break;
	       case MSG_SET_CURRENT_REGN:
                  rval = chipsSys->set_current_region(cid);
               break;
               case MSG_ADD_FRAME_A:
                  rval = chipsSys->add_frame(cid);
               break;
               case MSG_DEL_WINDOW:
                  rval = chipsSys->delete_window(cid);
               break;
               case MSG_DEL_LEGEND:
                  rval = chipsSys->delete_legend(cid);
               break;
               case MSG_DEL_ALL_WINS:
                  rval = chipsSys->delete_all_windows(cid);
               break;
               case MSG_DEL_FRAME:
                  rval = chipsSys->delete_frame(cid);
               break;
               case MSG_DEL_ALL_FRAMES:
                  rval = chipsSys->delete_all_frames(cid);
               break;
               case MSG_DEL_PLOT:
                  rval = chipsSys->delete_plot(cid);
               break;
               case MSG_CLEAR_PLOT:
                  rval = chipsSys->clear_plot(cid);
               break;
               case MSG_DEL_CURVE:
                  rval = chipsSys->delete_curve(cid);
               break;
               case MSG_DEL_CONTOUR:
                  rval = chipsSys->delete_contour(cid);
               break;
               case MSG_DEL_HISTOGRAM:
                  rval = chipsSys->delete_histogram(cid);
               break;
               case MSG_DEL_AXIS:
                  rval = chipsSys->delete_axis(cid);
               break;
               case MSG_DEL_LINE:
                  rval = chipsSys->delete_line(cid);
               break;
               case MSG_DEL_REGION:
                  rval = chipsSys->delete_region(cid);
               break;
               case MSG_DEL_POINT:
                  rval = chipsSys->delete_point(cid);
               break;
               case MSG_DEL_ALL_PLOTS:
                  rval = chipsSys->delete_all_plots(cid);
               break;
               case MSG_DEL_IMAGE:
                  rval = chipsSys->delete_image(cid);
               break;
               case MSG_DEL_IMAGE_CHANNEL:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->remove_image_channel(cid, *ival);
               }
               break;
               case MSG_DEL_ALL_IMAGES:
                  rval = chipsSys->delete_all_images(cid);
               break;
               case MSG_DEL_CBAR:
                  rval = chipsSys->delete_colorbar(cid);
               break;
               case MSG_DEL_LABEL:
                  rval = chipsSys->delete_label(cid);
               break;
               case MSG_SINGLE_FRAME:
                  rval = chipsSys->view_single_frame(cid);
               break;
               case MSG_PREV_FRAME:
                  rval = chipsSys->prev_frame(cid);
               break;
               case MSG_NEXT_FRAME:
                  rval = chipsSys->next_frame(cid);
               break;
               case MSG_PLACED_FRAME:   // set frames as placed
                  rval = chipsSys->view_placed_frame(cid);
               break;
               case MSG_SET_FRAME_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_frame_bgcolor(cid, new_color);
               }
               break;
               case MSG_SET_FRAME_BCOLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_frame_border_color(cid, new_color);
               }
               break;
               case MSG_LOAD_FONT:
               {
                  string fname, fpath;
                  V_FONTSTYLE_TYPE *styl = 
                     (V_FONTSTYLE_TYPE *)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  fname = (char*)&bf[ndx];
                  fpath = (char*)&bf[ndx + fname.length() + 1];
                  rval = chipsSys->load_font(fname, fpath, *styl);
               }
               break; 
               case MSG_AXIS_BIND:
               {
                  string lid, aid;
                  lid = (char*)&bf[ndx];
                  aid = (char*)&bf[ndx + lid.length() + 1];
                  rval = chipsSys->bind_axes(cid, lid, aid);
               }
               break;
               case MSG_AXIS_UNBIND:
               {
                  string lid, aid;
                  lid = (char*)&bf[ndx];
                  aid = (char*)&bf[ndx + lid.length() + 1];
                  rval = chipsSys->unbind_axes(cid, lid, aid);
               }
               break;
               case MSG_BLINK_FRAMES:
               {
                  float* time = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  string f2 = (char*)&bf[ndx];
                  rval = chipsSys->blink_frames(cid, f2, *time);
               }
               break;
               case MSG_STOP_BLINK:
                  rval = chipsSys->stop_blink(cid);
               break;
               case MSG_ARRANGE_FRAMES:
               {
                  int *cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
   
                  if (*cnt > 0)
                  {
                     string* c1 =  new string[*cnt];
                     if (c1)
                     {
                        int rr;

                        char* cp = &bf[ndx];
                        c1[0] = cp++;
                        for (rr = 1; rr < *cnt; rr++)
                        {
                           while(*cp != '\0') cp++;
                           c1[rr] = ++cp;
                        }
                        rval = chipsSys->arrange_frames(cid, *cnt, c1);
                        delete[] c1;
                     }
                  }
               }
               break;
               case MSG_SHOW_GRID:
               {
                  int *scale = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->display_grid(cid, *scale);
               }
               break;
               case MSG_HIDE_GRID:
                  rval = chipsSys->hide_grid(cid);
               break;
               case MSG_SET_GRID_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_grid_color(cid, new_color);
               }
               break;
               case MSG_SET_GRID_THICK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_grid_thickness(cid,*thick);
               }
               break;
               case MSG_SET_AXSG_MJ_SZ:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_majorgrid_thickness(cid,*thick);
               }
               break;
               case MSG_SET_AXSG_MJ_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_majorgrid_color(cid, new_color);
               }
               break;
               case MSG_SET_AXSG_MN_SZ:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_minorgrid_thickness(cid,*thick);
               }
               break;
               case MSG_SET_AXSG_MN_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_minorgrid_color(cid, new_color);
               }
               break;
               case MSG_SET_CBAR_TK_THK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_tick_thickness(cid,*thick);
               }
               break;
               case MSG_SET_AXIS_MAJ_THK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_major_thickness(cid,*thick);
               }
               break;
               case MSG_SET_AXIS_MIN_THK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_minor_thickness(cid,*thick);
               }
               break;
               case MSG_SET_GRID_SCALE:
               {
                  int *scale = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_grid_scale(cid,*scale);
               }
               break;
               case MSG_SET_GRID_STYLE:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_grid_linestyle(cid,typ);
               }
               break;
               case MSG_SET_AXSG_MJ_STL:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_axis_majorgrid_style(cid,typ);
               }
               break;
               case MSG_SET_AXSG_MN_STL:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_axis_minorgrid_style(cid,typ);
               }
               break;
               case MSG_SET_LAYOUT_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_layout_color(cid, new_color);
               }
               break;
               case MSG_SET_TITLE_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_title_color(cid, new_color);
               }
               break;
               case MSG_SET_TITLE_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_title_font(cid, new_font);
               }
               break;
               case MSG_SET_CBAR_TL_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_ticklabel_font(cid, new_font);
               }
               break;
               case MSG_SET_CBAR_LAB_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_label_font(cid, new_font);
               }
               break;
               case MSG_SET_LAB_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_label_font(cid, new_font);
               }
               break;
	       case MSG_GET_PLOT_ASPRATIO:  // fallthrough intended
	       case MSG_GET_DATA_ASPRATIO:  // fallthrough intended
               case MSG_GET_AXIS_COLOR:     // fallthrough intended
               case MSG_GET_CBAR_TL_CLR:    // fallthrough intended
               case MSG_GET_AXIS_MAJ_CLR:   // fallthrough intended
               case MSG_GET_AXIS_MIN_CLR:   // fallthrough intended
               case MSG_GET_AXIS_TEXT:      // fallthrough intended
               case MSG_GET_AXIS_XLAB_TXT:  // fallthrough intended
               case MSG_GET_AXIS_YLAB_TXT:  // fallthrough intended
               case MSG_GET_TCKLB_COLOR:    // fallthrough intended
               case MSG_GET_AXSLB_COLOR:    // fallthrough intended
               case MSG_GET_AXSG_MJ_CLR:    // fallthrough intended
               case MSG_GET_AXSG_MN_CLR:    // fallthrough intended
               case MSG_GET_AXIS_TICKFMT:   // fallthrough intended
               case MSG_GET_CBAR_TK_MODE:   // fallthrough intended
               case MSG_GET_AXIS_MAJ_MODE:  // fallthrough intended
               case MSG_GET_AXIS_MIN_MODE:  // fallthrough intended
               case MSG_GET_CBAR_LAB_LOC:   // fallthrough intended
               case MSG_GET_AXIS_TICKSTYL:  // fallthrough intended
               case MSG_GET_CBAR_TK_STY:    // fallthrough intended
               case MSG_GET_CBAR_TK_LOC:    // fallthrough intended
               case MSG_GET_CBAR_TK_CLR:    // fallthrough intended
               case MSG_GET_AXIS_MAJ_STYL:  // fallthrough intended
               case MSG_GET_AXIS_MIN_STYL:  // fallthrough intended
               case MSG_GET_TCKLB_FONT:     // fallthrough intended
               case MSG_GET_TCKLB_FNTSTYL:  // fallthrough intended
               case MSG_GET_AXSLB_FONT:     // fallthrough intended
               case MSG_GET_AXSLB_FNTSTYL:  // fallthrough intended
               case MSG_GET_AXSTCKLB_STYL:  // fallthrough intended
               case MSG_GET_WIN_TITLE:      // fallthrough intended
               case MSG_GET_WINDOW_COLOR:   // fallthrough intended
               case MSG_GET_FRAME_COLOR:    // fallthrough intended
               case MSG_GET_FRAME_BCOLOR:   // fallthrough intended
               case MSG_GET_TITLE_FNTSTYL:  // fallthrough intended:
               case MSG_GET_TITLE_FONT:     // fallthrough intended:
               case MSG_GET_TITLE_COLOR:    // fallthrough intended:
               case MSG_GET_PLOT_TITLE:     // fallthrough intended:
               case MSG_GET_PLOT_CORNER:    // fallthrough intended:
               case MSG_GET_PLOT_AX_STYLE:  // fallthrough intended:
               case MSG_GET_CV_LINE_COLOR:  // fallthrough intended
               case MSG_GET_CV_ERR_COLOR:   // fallthrough intended
               case MSG_GET_CV_ERR_STYLE:   // fallthrough intended
               case MSG_GET_CV_SYM_COLOR:   // fallthrough intended
               case MSG_GET_CNTR_COLOR:     // fallthrough intended
               case MSG_GET_CNTR_STYLE:     // fallthrough intended
               case MSG_GET_CNTR_MODE:      // fallthrough intended
               case MSG_GET_HST_LINE_COLR:  // fallthrough intended
               case MSG_GET_HST_ERR_COLOR:  // fallthrough intended
               case MSG_GET_HST_ERR_STYLE:  // fallthrough intended
               case MSG_GET_HST_SYM_COLOR:  // fallthrough intended
               case MSG_GET_HST_FILL_COLR:  // fallthrough intended
               case MSG_GET_POINT_COLOR:    // fallthrough intended
               case MSG_GET_REGN_EDGCOL:    // fallthrough intended
               case MSG_GET_REGN_FILCOL:    // fallthrough intended
               case MSG_GET_LINE_COLOR:     // fallthrough intended
               case MSG_GET_LAB_TEXT:       // fallthrough intended:
               case MSG_GET_CBAR_LAB_TXT:   // fallthrough intended
               case MSG_GET_LAB_COLOR:      // fallthrough intended
               case MSG_GET_CBAR_LAB_CLR:   // fallthrough intended
               case MSG_GET_CBAR_TL_FSTY:   // fallthrough intended
               case MSG_GET_CBAR_LAB_FSTY:  // fallthrough intended
               case MSG_GET_LAB_STYLE:      // fallthrough intended
               case MSG_GET_SERVER_ID:      // fallthrough intended
               case MSG_GET_CBAR_TL_FONT:   // fallthrough intended
               case MSG_GET_CBAR_LAB_FONT:  // fallthrough intended
               case MSG_GET_LAB_FONT:       // fallthrough intended
               case MSG_GET_CWD:
               {
                  string resp = "";
                  switch (buff[0])
                  {
                     case MSG_GET_LAB_TEXT: 
                        rval = chipsSys->get_label_text(cid, resp);
                     break;
                     case MSG_GET_CBAR_LAB_TXT:
                        rval = chipsSys->get_colorbar_label_text(cid, resp);
                     break; 
                     case MSG_GET_CBAR_LAB_FONT:
                        rval = chipsSys->get_colorbar_label_font(cid, resp);
                     break; 
                     case MSG_GET_CBAR_TL_FONT:
                        rval = chipsSys->get_colorbar_ticklabel_font(cid, resp);
                     break; 
                     case MSG_GET_LAB_FONT:
                        rval = chipsSys->get_label_font(cid, resp);
                     break; 
                     case MSG_GET_CBAR_LAB_FSTY:
                        rval = chipsSys->get_colorbar_label_fontstyle(cid,resp);
                     break; 
                     case MSG_GET_CBAR_TL_FSTY:
                        rval = chipsSys->get_colorbar_ticklabel_fontstyle(cid,resp);
                     break; 
                     case MSG_GET_LAB_STYLE:
                        rval = chipsSys->get_label_fontstyle(cid, resp);
                     break; 
                     case MSG_GET_LAB_COLOR:
                        rval = chipsSys->get_label_color(cid, resp);
                     break; 
                     case MSG_GET_CBAR_LAB_CLR:
                        rval = chipsSys->get_colorbar_label_color(cid, resp);
                     break; 
                     case MSG_GET_FRAME_COLOR:
                        rval = chipsSys->get_frame_bgcolor(cid, resp);
                     break; 
                     case MSG_GET_FRAME_BCOLOR:
                        rval = chipsSys->get_frame_border_color(cid, resp);
                     break; 
                     case MSG_GET_WINDOW_COLOR:
                        rval = chipsSys->get_window_bgcolor(cid, resp);
                     break; 
                     case MSG_GET_CV_LINE_COLOR:
                        rval = chipsSys->get_curve_linecolor(cid, resp);
                     break; 
                     case MSG_GET_CV_SYM_COLOR:
                        rval = chipsSys->get_curve_symbolcolor(cid, resp);
                     break; 
                     case MSG_GET_CV_ERR_COLOR:
                        rval = chipsSys->get_curve_errcolor(cid, resp);
                     break; 
                     case MSG_GET_CV_ERR_STYLE:
                        rval = chipsSys->get_curve_errstyle(cid, resp);
                     break; 
                     case MSG_GET_CNTR_COLOR:
                        rval = chipsSys->get_contour_color(cid, resp);
                     break; 
                     case MSG_GET_CNTR_STYLE:
                        rval = chipsSys->get_contour_style(cid, resp);
                     break; 
                     case MSG_GET_CNTR_MODE:
                        rval = chipsSys->get_contour_mode(cid, resp);
                     break; 
                     case MSG_GET_HST_LINE_COLR:
                        rval = chipsSys->get_histogram_linecolor(cid, resp);
                     break; 
                     case MSG_GET_HST_SYM_COLOR:
                        rval = chipsSys->get_histogram_symbolcolor(cid, resp);
                     break; 
                     case MSG_GET_HST_ERR_COLOR:
                        rval = chipsSys->get_histogram_errcolor(cid, resp);
                     break; 
                     case MSG_GET_HST_ERR_STYLE:
                        rval = chipsSys->get_histogram_errstyle(cid, resp);
                     break; 
                     case MSG_GET_HST_FILL_COLR:
                        rval = chipsSys->get_histogram_fillcolor(cid, resp);
                     break;
                     case MSG_GET_LINE_COLOR:
                        rval = chipsSys->get_line_color(cid, resp);
                     break; 
                     case MSG_GET_REGN_EDGCOL:
                        rval = chipsSys->get_region_color(cid, resp);
                     break;
                     case MSG_GET_REGN_FILCOL:
                        rval = chipsSys->get_region_fillcolor(cid, resp);
                     break;
                     case MSG_GET_POINT_COLOR:
                        rval = chipsSys->get_point_color(cid, resp);
                     break; 
                     case MSG_GET_PLOT_AX_STYLE:
                        rval = chipsSys->get_plot_axisstyle(cid, resp);
                     break; 
                     case MSG_GET_PLOT_TITLE:
                        rval = chipsSys->get_plot_title(cid, resp);
                     break; 
                     case MSG_GET_PLOT_ASPRATIO:
                        rval = chipsSys->get_plot_aspect_ratio(cid, resp);
                     break; 
                     case MSG_GET_DATA_ASPRATIO:
                        rval = chipsSys->get_data_aspect_ratio(cid, resp);
                     break; 
                     case MSG_GET_PLOT_CORNER:
                        rval = chipsSys->get_plot_cornerstyle(cid, resp);
                     break; 
                     case MSG_GET_TITLE_COLOR:
                        rval = chipsSys->get_plot_title_color(cid, resp);
                     break; 
                     case MSG_GET_TITLE_FONT:
                        rval = chipsSys->get_plot_title_font(cid, resp);
                     break; 
                     case MSG_GET_TITLE_FNTSTYL:
                        rval = chipsSys->get_plot_title_fontstyle(cid, resp);
                     break; 
                     case MSG_GET_WIN_TITLE:
                        rval = chipsSys->get_window_title(cid, resp);
                     break; 
                     case MSG_GET_AXIS_COLOR:   
                        rval = chipsSys->get_axis_color(cid, resp);
                     break;
                     case MSG_GET_CBAR_TL_CLR:
                        rval = chipsSys->get_colorbar_ticklabel_color(cid,resp);
                     break;
                     case MSG_GET_AXIS_MAJ_CLR:
                        rval = chipsSys->get_axis_major_color(cid, resp);
                     break;
                     case MSG_GET_AXIS_MIN_CLR: 
                        rval = chipsSys->get_axis_minor_color(cid, resp);
                     break;
                     case MSG_GET_AXIS_TEXT:    
                        rval = chipsSys->get_axis_text(cid, resp);
                     break;
                     case MSG_GET_AXIS_XLAB_TXT:
                        rval = chipsSys->get_axis_xlabel_text(cid, resp);
                     break;
                     case MSG_GET_AXIS_YLAB_TXT:
                        rval = chipsSys->get_axis_ylabel_text(cid, resp);
                     break;
                     case MSG_GET_TCKLB_COLOR:  
                        rval = chipsSys->get_axis_ticklabel_color(cid, resp);
                     break;
                     case MSG_GET_AXSLB_COLOR:  
                        rval = chipsSys->get_axis_label_color(cid, resp);
                     break;
                     case MSG_GET_AXSG_MJ_CLR:  
                        rval = chipsSys->get_axis_majorgrid_color(cid, resp);
                     break;
                     case MSG_GET_AXSG_MN_CLR:   
                        rval = chipsSys->get_axis_minorgrid_color(cid, resp);
                     break;
                     case MSG_GET_AXIS_TICKFMT:  
                        rval = chipsSys->get_axis_tickformat(cid, resp);
                     break;
                     case MSG_GET_CBAR_TK_MODE: 
                        rval = chipsSys->get_colorbar_mode(cid, resp);
                     break;
                     case MSG_GET_AXIS_MAJ_MODE: 
                        rval = chipsSys->get_axis_major_mode(cid, resp);
                     break;
                     case MSG_GET_AXIS_MIN_MODE: 
                        rval = chipsSys->get_axis_minor_mode(cid, resp);
                     break;
                     case MSG_GET_CBAR_LAB_LOC: 
                        rval = chipsSys->get_colorbar_label_location(cid,resp);
                     break;
                     case MSG_GET_AXIS_TICKSTYL: 
                        rval = chipsSys->get_axis_tickstyle(cid, resp);
                     break;
                     case MSG_GET_CBAR_TK_STY: 
                        rval = chipsSys->get_colorbar_tickstyle(cid, resp);
                     break;
                     case MSG_GET_CBAR_TK_LOC:
                        rval = chipsSys->get_colorbar_tick_location(cid, resp);
                     break;
                     case MSG_GET_CBAR_TK_CLR:
                        rval = chipsSys->get_colorbar_tick_color(cid, resp);
                     break;
                     case MSG_GET_AXIS_MAJ_STYL: 
                        rval = chipsSys->get_axis_major_style(cid, resp);
                     break;
                     case MSG_GET_AXIS_MIN_STYL: 
                        rval = chipsSys->get_axis_minor_style(cid, resp);
                     break;
                     case MSG_GET_TCKLB_FONT:    
                        rval = chipsSys->get_axis_ticklabel_font(cid, resp);
                     break;
                     case MSG_GET_TCKLB_FNTSTYL: 
                        rval=chipsSys->get_axis_ticklabel_fontstyle(cid, resp);
                     break;
                     case MSG_GET_AXSLB_FONT:    
                        rval = chipsSys->get_axis_label_font(cid, resp);
                     break;
                     case MSG_GET_AXSLB_FNTSTYL: 
                        rval = chipsSys->get_axis_label_fontstyle(cid, resp);
                     break;
                     case MSG_GET_AXSTCKLB_STYL:
                        rval = chipsSys->get_axis_ticklabel_style(cid, resp);
                     break; 
                     case MSG_GET_CWD:
#ifndef HEADLESS
                        resp = serverCwd;  
                        rval = ERR_STATUS_GOOD; 
#endif
                     break; 
		     case MSG_GET_SERVER_ID:
		        resp = serverId;
			rval = ERR_STATUS_GOOD; 
		     break;
                     default:
                        rval = ERR_MESSAGE_UNKNOWN;
                     break;
                  }
                  if (rval == ERR_STATUS_GOOD)
                  {
                     long ln = (resp.length())/4 + 5;
                     int* ival;

                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = resp.length() + 1;
                        cp = (char*)&ival[4];
                        if (resp.length())   strcpy(cp, resp.c_str());
                        else                 cp[0] = '\0'; 
                        ival[1] =  strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
                  }
               }
               break;
	       case MSG_GET_IMG_INVCMAP:  // fallthrough intended
               case MSG_GET_IMG_CMAPSIZE: // fallthrough intended
               case MSG_GET_IMG_CMAPINTRP:// fallthrough intended
               case MSG_GET_IMG_SCL_CHNNL:// fallthrough intended
               case MSG_GET_AXIS_MAJ_NUM: // fallthrough intended 
               case MSG_GET_AXIS_MIN_NUM: // fallthrough intended 
               case MSG_GET_AXIS_MAJ_LEN: // fallthrough intended 
               case MSG_GET_AXIS_MIN_LEN: // fallthrough intended 
               case MSG_GET_TCKLB_SIZE:   // fallthrough intended 
               case MSG_GET_CBAR_TL_OFF:  // fallthrough intended 
               case MSG_GET_CBAR_TK_LEN:  // fallthrough intended
               case MSG_GET_TCKLB_OFFSET: // fallthrough intended 
               case MSG_GET_REGN_VISIBLE: // fallthrough intended
               case MSG_GET_PLOT_VISIBLE: // fallthrough intended 
               case MSG_GET_AXIS_AUTOMAX: // fallthrough intended 
               case MSG_GET_AXIS_AUTOMIN: // fallthrough intended 
               case MSG_GET_AXIS_VISIBLE: // fallthrough intended 
               case MSG_GET_AXIS_MAJ_VIS: // fallthrough intended 
               case MSG_GET_AXIS_MIN_VIS: // fallthrough intended 
               case MSG_GET_POINT_VISIBLE:// fallthrough intended 
               case MSG_GET_LINE_VISIBLE: // fallthrough intended 
               case MSG_GET_CBAR_TL_VIS:  // fallthrough intended 
               case MSG_GET_CBAR_TK_VIS:  // fallthrough intended
               case MSG_GET_CBAR_LAB_VIS: // fallthrough intended 
               case MSG_GET_LAB_VISIBLE:  // fallthrough intended 
               case MSG_GET_IMG_VISIBLE:  // fallthrough intended 
               case MSG_GET_CBAR_VISIBLE: // fallthrough intended 
               case MSG_GET_CBAR_ORIENT:  // fallthrough intended 
               case MSG_GET_CV_VISIBLE:   // fallthrough intended 
               case MSG_GET_CNTR_VISIBLE: // fallthrough intended 
               case MSG_GET_HST_VISIBLE:  // fallthrough intended 
               case MSG_GET_TCKLB_VIS:    // fallthrough intended 
               case MSG_GET_AXSLB_VIS:    // fallthrough intended 
               case MSG_GET_CNTR_NUMLEV:  // fallthrough intended 
               case MSG_GET_SCALE:        // fallthrough intended 
               case MSG_GET_AXSLB_SIZE:   // fallthrough intended 
               case MSG_GET_AXSG_MJ_STL:  // fallthrough intended 
               case MSG_GET_AXSG_MN_STL:  // fallthrough intended 
               case MSG_GET_AXSG_MJ_VIS:  // fallthrough intended 
               case MSG_GET_AXSG_MN_VIS:  // fallthrough intended 
               case MSG_GET_CV_BADDATA:   // fallthrough intended 
               case MSG_GET_CV_LINE_STYLE:// fallthrough intended 
               case MSG_GET_CV_SYMBOL:    // fallthrough intended 
               case MSG_GET_CV_ERR_CAPLEN:// fallthrough intended 
               case MSG_GET_CV_SYM_SIZE:  // fallthrough intended 
               case MSG_GET_CV_SYM_FILL:  // fallthrough intended 
               case MSG_GET_CV_LIMOVRIDE: // fallthrough intended 
               case MSG_GET_CV_ERR_UP:    // fallthrough intended 
               case MSG_GET_CV_ERR_DOWN:  // fallthrough intended 
               case MSG_GET_CV_ERR_LEFT:  // fallthrough intended 
               case MSG_GET_CV_ERR_RIGHT: // fallthrough intended 
               case MSG_GET_CNTR_ALG:     // fallthrough intended 
               case MSG_GET_HST_LINE_STYL:// fallthrough intended 
               case MSG_GET_HST_DROPLINE: // fallthrough intended 
               case MSG_GET_HST_ERR_UP:   // fallthrough intended 
               case MSG_GET_HST_ERR_DOWN: // fallthrough intended 
               case MSG_GET_HST_ERR_CPLEN:// fallthrough intended 
               case MSG_GET_HST_SYMBOL:   // fallthrough intended 
               case MSG_GET_HST_SYM_SIZE: // fallthrough intended 
               case MSG_GET_HST_SYM_FILL: // fallthrough intended 
               case MSG_GET_HST_FILL_STYL:// fallthrough intended
               case MSG_GET_TITLE_SIZE:   // fallthrough intended 
               case MSG_GET_POINT_SYMBOL: // fallthrough intended 
               case MSG_GET_POINT_FILL:   // fallthrough intended 
               case MSG_GET_POINT_SIZE:   // fallthrough intended 
               case MSG_GET_REGN_STYLE:   // fallthrough intended
               case MSG_GET_REGN_FILSTYL: // fallthrough intended
               case MSG_GET_LINE_STYLE:   // fallthrough intended 
               case MSG_GET_LINE_EXTEND:  // fallthrough intended
               case MSG_GET_FRAME_SCALE:  // fallthrough intended 
               case MSG_GET_FRAME_TRANS:  // fallthrough intended 
               case MSG_GET_FRAME_BVIS:   // fallthrough intended 
               case MSG_GET_CBAR_BVIS:    // fallthrough intended 
               case MSG_GET_FRAME_VIS:    // fallthrough intended 
               case MSG_GET_WIN_REDRAW:   // fallthrough intended 
               case MSG_GET_DEF_DEPTH:    // fallthrough intended 
               case MSG_GET_HST_DEPTH:    // fallthrough intended 
               case MSG_GET_CNTR_DEPTH:   // fallthrough intended 
               case MSG_GET_IMG_DEPTH:    // fallthrough intended 
               case MSG_GET_CBAR_DEPTH:   // fallthrough intended 
               case MSG_GET_CV_DEPTH:     // fallthrough intended 
               case MSG_GET_AXIS_DEPTH:   // fallthrough intended 
               case MSG_GET_PLOT_DEPTH:   // fallthrough intended 
               case MSG_GET_POINT_DEPTH:  // fallthrough intended 
               case MSG_GET_LINE_DEPTH:   // fallthrough intended 
               case MSG_GET_REGN_DEPTH:   // fallthrough intended 
               case MSG_GET_LAB_DEPTH:    // fallthrough intended 
               case MSG_GET_IMG_CMAP:     // fallthrough intended:
               case MSG_GET_IMG_INTERP:   // fallthrough intended:
               case MSG_GET_CBAR_LAB_SZ:  // fallthrough intended:
               case MSG_GET_CBAR_TL_SZ:   // fallthrough intended:
               case MSG_GET_LAB_SIZE:
               {
                  int resp;
                  int* ival;
                  bool bval;
                  V_SYMBOL_TYPE stype;
                  V_FILL_TYPE ftype;
                  V_LINE_TYPE ltype;
                  V_LINE_EXTENSION_TYPE letype;
                  V_INTERPOLATION_TYPE itptype;
                  V_COLORMAP_TYPE cmaptype;

                  switch(buff[0])
                  {
                     case MSG_GET_LAB_SIZE:
                        rval = chipsSys->get_label_size(cid, resp);
                     break; 
                     case MSG_GET_CBAR_LAB_SZ:
                        rval = chipsSys->get_colorbar_label_size(cid, resp);
                     break; 
                     case MSG_GET_CBAR_TL_SZ:
                        rval = chipsSys->get_colorbar_ticklabel_size(cid, resp);
                     break; 
                     case MSG_GET_CV_DEPTH:
                        rval = chipsSys->get_curve_depth(cid, resp);
                     break; 
                     case MSG_GET_IMG_DEPTH:
                        rval = chipsSys->get_image_depth(cid, resp);
                     break; 
                     case MSG_GET_CBAR_DEPTH:
                        rval = chipsSys->get_colorbar_depth(cid, resp);
                     break; 
                     case MSG_GET_CNTR_DEPTH:
                        rval = chipsSys->get_contour_depth(cid, resp);
                     break; 
                     case MSG_GET_DEF_DEPTH:
                        rval = chipsSys->get_default_depth(cid, resp);
                     break; 
                     case MSG_GET_HST_DEPTH:
                        rval = chipsSys->get_histogram_depth(cid, resp);
                     break; 
                     case MSG_GET_HST_FILL_STYL:
                        rval = chipsSys->get_histogram_fillstyle(cid, ftype);
                        resp = (int)ftype;
                     break;
                     case MSG_GET_IMG_CMAPSIZE:
                        rval = chipsSys->get_image_colormap_size(cid,resp); 
                     break;
                     case MSG_GET_IMG_INTERP:
                        rval = chipsSys->get_image_interpolation(cid, itptype);
		        resp = (int)itptype;
                     break; 
                     case MSG_GET_IMG_CMAP:
                        rval = chipsSys->get_image_colormap(cid, cmaptype);
                        resp = (int)cmaptype;
                     break; 
                     case MSG_GET_AXIS_DEPTH:
                        rval = chipsSys->get_axis_depth(cid, resp);
                     break; 
                     case MSG_GET_PLOT_DEPTH:
                        rval = chipsSys->get_plot_title_depth(cid, resp);
                     break; 
                     case MSG_GET_LINE_DEPTH:
                        rval = chipsSys->get_line_depth(cid, resp);
                     break; 
                     case MSG_GET_LAB_DEPTH:
                        rval = chipsSys->get_label_depth(cid, resp);
                     break; 
                     case MSG_GET_POINT_DEPTH:
                        rval = chipsSys->get_point_depth(cid, resp);
                     break; 
                     case MSG_GET_REGN_DEPTH:
                        rval = chipsSys->get_region_depth(cid, resp);
                     break; 
                     case MSG_GET_LINE_STYLE:
                        rval = chipsSys->get_line_style(cid, ltype);
                        resp = (int)ltype;
                     break; 
                     case MSG_GET_LINE_EXTEND:
                        rval = chipsSys->get_line_extend(cid, letype);
                        resp = (int)letype;
                     break; 
                     case MSG_GET_REGN_STYLE:
                        rval = chipsSys->get_region_style(cid, ltype);
                        resp = (int)ltype;
                     break;
                     case MSG_GET_REGN_FILSTYL:
                        rval = chipsSys->get_region_fillstyle(cid, ftype);
                        resp = (int)ftype;
                     break;
                     case MSG_GET_POINT_SIZE:
                        rval = chipsSys->get_point_size(cid, resp);
                     break; 
                     case MSG_GET_POINT_FILL:
                        rval = chipsSys->get_point_fill(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_POINT_SYMBOL:
                        rval = chipsSys->get_point_symbol(cid, stype);
                        resp = (int)stype;
                     break; 
                     case MSG_GET_TITLE_SIZE:
                        rval = chipsSys->get_plot_title_size(cid, resp);
                     break; 
                     case MSG_GET_CV_SYMBOL:
                        rval = chipsSys->get_curve_symboltype(cid, stype);
                        resp = (int)stype;
                     break; 
                     case MSG_GET_CV_SYM_SIZE:
                        rval = chipsSys->get_curve_symbolsize(cid, resp);
                     break; 
                     case MSG_GET_CV_ERR_CAPLEN:
                        rval = chipsSys->get_curve_errcaplength(cid, resp);
                     break; 
                     case MSG_GET_CV_BADDATA:
		     {
			V_BADDATAMODE_TYPE bdm; 
                        rval = chipsSys->get_curve_baddatamode(cid, bdm);
                        resp = (int)bdm;
		     }
                     break;
                     case MSG_GET_CV_LINE_STYLE:
                        rval = chipsSys->get_curve_linestyle(cid, ltype);
                        resp = (int)ltype;
                     break; 
                     case MSG_GET_CV_SYM_FILL:
                        rval = chipsSys->get_curve_symbolfill(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CV_LIMOVRIDE:
                        rval = chipsSys->get_curve_limitoverride(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CV_ERR_UP:
                        rval = chipsSys->get_curve_errup(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CV_ERR_DOWN:
                        rval = chipsSys->get_curve_errdown(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CV_ERR_LEFT:
                        rval = chipsSys->get_curve_errleft(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CV_ERR_RIGHT:
                        rval = chipsSys->get_curve_errright(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CNTR_ALG:
		     {
			V_CONTOUR_TYPE alg; 
                        rval = chipsSys->get_contour_algorithm(cid, alg);
                        resp = (int)alg;
		     }
                     break; 
                     case MSG_GET_HST_SYMBOL:
                        rval = chipsSys->get_histogram_symboltype(cid, stype);
                        resp = (int)stype;
                     break; 
                     case MSG_GET_HST_SYM_FILL:
                        rval = chipsSys->get_histogram_symbolfill(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_HST_LINE_STYL:
                        rval = chipsSys->get_histogram_linestyle(cid, ltype);
                        resp = (int)ltype;
                     break; 
                     case MSG_GET_HST_SYM_SIZE:
                        rval = chipsSys->get_histogram_symbolsize(cid,resp);
                     break; 
                     case MSG_GET_HST_ERR_CPLEN:
                        rval = chipsSys->get_histogram_errcaplength(cid, resp);
                     break; 
                     case MSG_GET_HST_DROPLINE:
                        rval = chipsSys->get_histogram_dropline(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_HST_ERR_UP:
                        rval = chipsSys->get_histogram_errup(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_HST_ERR_DOWN:
                        rval = chipsSys->get_histogram_errdown(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_AXIS_MAJ_NUM: 
                        rval = chipsSys->get_axis_major_num(cid, resp);
                     break;
                     case MSG_GET_AXIS_MIN_NUM: 
                        rval = chipsSys->get_axis_minor_num(cid, resp);
                     break;
                     case MSG_GET_AXIS_MAJ_LEN: 
                        rval = chipsSys->get_axis_major_length(cid, resp);
                     break;
                     case MSG_GET_AXIS_MIN_LEN: 
                        rval = chipsSys->get_axis_minor_length(cid, resp);
                     break;
                     case MSG_GET_AXSLB_SIZE:  
                        rval = chipsSys->get_axis_label_size(cid, resp);
                     break;
                     case MSG_GET_SCALE:  
                        rval = chipsSys->get_axis_scale(cid, resp);
                     break;
                     case MSG_GET_CNTR_NUMLEV:  
                        rval = chipsSys->get_contour_numlevels(cid, resp);
                     break;
                     case MSG_GET_TCKLB_SIZE:  
                        rval = chipsSys->get_axis_ticklabel_size(cid, resp);
                     break;
                     case MSG_GET_CBAR_TL_OFF:  
                        rval= chipsSys->get_colorbar_ticklabel_offset(cid,resp);
                     break;
                     case MSG_GET_CBAR_TK_LEN:
                        rval= chipsSys->get_colorbar_tick_length(cid,resp);
                     break;
                     case MSG_GET_TCKLB_OFFSET:  
                        rval = chipsSys->get_axis_ticklabel_offset(cid, resp);
                     break;
                     case MSG_GET_AXSG_MJ_STL: 
                        rval = chipsSys->get_axis_majorgrid_style(cid, ltype);
                        resp = (int)ltype;
                     break;
                     case MSG_GET_AXSG_MN_STL: 
                        rval = chipsSys->get_axis_minorgrid_style(cid, ltype);
                        resp = (int)ltype;
                     break;
                     case MSG_GET_CBAR_ORIENT: 
                        rval = chipsSys->get_colorbar_orientation(cid, resp);
                     break;
                     case MSG_GET_REGN_VISIBLE:
                        rval = chipsSys->get_region_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_PLOT_VISIBLE: 
                        rval = chipsSys->get_plot_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXIS_AUTOMIN: 
                        rval = chipsSys->get_axis_automin(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_IMG_INVCMAP: 
                        rval = chipsSys->get_image_invert_colormap(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_IMG_CMAPINTRP:
                        rval = chipsSys->get_image_colormap_interpolate(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_IMG_SCL_CHNNL:
                        rval = chipsSys->get_image_scale_channels(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXIS_AUTOMAX: 
                        rval = chipsSys->get_axis_automax(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXIS_VISIBLE: 
                        rval = chipsSys->get_axis_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_LINE_VISIBLE: 
                        rval = chipsSys->get_line_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_POINT_VISIBLE: 
                        rval = chipsSys->get_point_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_CBAR_TL_VIS:
                        rval=chipsSys->get_colorbar_ticklabel_visible(cid,bval);
                        resp = bval;
                     break;
                     case MSG_GET_CBAR_TK_VIS:
                        rval=chipsSys->get_colorbar_tick_visible(cid,bval);
                        resp = bval;
                     break;
                     case MSG_GET_CBAR_LAB_VIS:
                        rval = chipsSys->get_colorbar_label_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_LAB_VISIBLE: 
                        rval = chipsSys->get_label_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_CV_VISIBLE: 
                        rval = chipsSys->get_curve_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_IMG_VISIBLE: 
                        rval = chipsSys->get_image_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_CBAR_VISIBLE: 
                        rval = chipsSys->get_colorbar_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_CNTR_VISIBLE: 
                        rval = chipsSys->get_contour_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_HST_VISIBLE: 
                        rval = chipsSys->get_histogram_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXIS_MAJ_VIS: 
                        rval = chipsSys->get_major_ticks_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXIS_MIN_VIS: 
                        rval = chipsSys->get_minor_ticks_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_TCKLB_VIS: 
                        rval = chipsSys->get_axis_ticklabel_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXSLB_VIS: 
                        rval = chipsSys->get_axis_label_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXSG_MJ_VIS: 
                        rval = chipsSys->get_axis_majorgrid_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_AXSG_MN_VIS:
                        rval = chipsSys->get_axis_minorgrid_visible(cid, bval);
                        resp = bval;
                     break;
                     case MSG_GET_FRAME_SCALE:
                        rval = chipsSys->get_frame_scale(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_CBAR_BVIS:
                        rval = chipsSys->get_colorbar_border_visible(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_FRAME_BVIS:
                        rval = chipsSys->get_frame_border_visible(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_FRAME_VIS:
                        rval = chipsSys->get_frame_visible(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_FRAME_TRANS:
                        rval = chipsSys->get_frame_transparency(cid, bval);
                        resp = bval;
                     break; 
                     case MSG_GET_WIN_REDRAW:
                        rval = chipsSys->get_window_redraw(cid, bval);
                        resp = bval;
                     break; 
                     default:
                        rval = ERR_MESSAGE_UNKNOWN;
                     break;
                  }
		  if (rval == ERR_STATUS_GOOD)
		  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_AXIS_MAJ_INT:  // fallthrough intended
               case MSG_GET_AXIS_MIN_INT:  // fallthrough intended
               case MSG_GET_REGN_OPACITY:  // fallthrough intended
               case MSG_GET_HST_FILL_OPAC: // fallthrough intended
               case MSG_GET_CNTR_INT:      // fallthrough intended
               case MSG_GET_CV_LIMLENGTH:  // fallthrough intended
               case MSG_GET_AXIS_PAD:
               {
                  double resp;
                  int* ival;

                  if (buff[0] == MSG_GET_AXIS_MAJ_INT)
                     rval = chipsSys->get_axis_major_interval(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_MIN_INT)
                     rval = chipsSys->get_axis_minor_interval(cid, resp);
                  else if (buff[0] == MSG_GET_CNTR_INT)
                     rval = chipsSys->get_contour_interval(cid, resp);
                  else if (buff[0] == MSG_GET_REGN_OPACITY)
                     rval = chipsSys->get_region_opacity(cid, resp);
                  else if (buff[0] == MSG_GET_HST_FILL_OPAC)
                     rval = chipsSys->get_histogram_fillopacity(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_PAD)
                     rval = chipsSys->get_axis_pad(cid, resp);
                  else if (buff[0] == MSG_GET_CV_LIMLENGTH)
                     rval = chipsSys->get_curve_limitlength(cid, resp);

                  if (rval == ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(6*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = 1;
                        double* dp = (double*)&ival[4];
                        dp[0] = resp;
                        ival[1] = 6 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_IMG_XRANGE:   // fallthrough intended
               case MSG_GET_IMG_YRANGE:   // fallthrough intended
	       case MSG_GET_IMG_ALPHA:    // fallthrough intended
	       case MSG_GET_IMG_THRESH:   // fallthrough intended
               case MSG_GET_PLOT_XRANGE:  // fallthrough intended 
               case MSG_GET_PLOT_YRANGE:  // fallthrough intended 
               case MSG_GET_AXIS_RANGE:   // fallthrough intended 
               case MSG_GET_CNTR_XRANGE:  // fallthrough intended 
               case MSG_GET_CNTR_YRANGE:  // fallthrough intended 
               case MSG_GET_CNTR_ZRANGE:  // fallthrough intended 
               case MSG_GET_HST_XRANGE:   // fallthrough intended 
               case MSG_GET_HST_YRANGE:   // fallthrough intended 
               case MSG_GET_CV_XRANGE:    // fallthrough intended 
               case MSG_GET_CV_YRANGE:
               {
                  double resp1;
                  double resp2;
                  int* ival;

                  if (buff[0] == MSG_GET_CV_XRANGE)
                     rval = chipsSys->get_curve_xrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_CV_YRANGE)
                     rval = chipsSys->get_curve_yrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_HST_XRANGE)
                     rval = chipsSys->get_histogram_xrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_HST_YRANGE)
                     rval = chipsSys->get_histogram_yrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_CNTR_XRANGE)
                     rval = chipsSys->get_contour_xrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_CNTR_YRANGE)
                     rval = chipsSys->get_contour_yrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_CNTR_ZRANGE)
                     rval = chipsSys->get_contour_zrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_PLOT_XRANGE)
                     rval = chipsSys->get_plot_xrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_PLOT_YRANGE)
                     rval = chipsSys->get_plot_yrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_AXIS_RANGE)
                     rval = chipsSys->get_axis_range(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_IMG_ALPHA)
                     rval = chipsSys->get_image_alpha(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_IMG_THRESH)
                     rval = chipsSys->get_image_threshold(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_IMG_XRANGE)
                     rval = chipsSys->get_image_xrange(cid, resp1, resp2);
                  else if (buff[0] == MSG_GET_IMG_YRANGE)
                     rval = chipsSys->get_image_yrange(cid, resp1, resp2);

                  if (rval == ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(8*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = 2;
                        double* dp = (double*)&ival[4];
                        dp[0] = resp1;
                        dp[1] = resp2;
                        ival[1] = 8 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break; 
               case MSG_GET_CV_RANGE:     // fallthrough intended
               case MSG_GET_CNTR_RANGE:   // fallthrough intended
               case MSG_GET_HST_RANGE:    // fallthrough intended
               case MSG_GET_IMG_RANGE:    // fallthrough intended
               case MSG_GET_PLOT_RANGE:
               {
                  double x1, x2, y1, y2;

                  if (buff[0] == MSG_GET_CV_RANGE)
                     rval = chipsSys->get_curve_range(cid, x1, x2, y1, y2);
                  else if (buff[0] == MSG_GET_HST_RANGE)
                     rval = chipsSys->get_histogram_range(cid, x1, x2, y1, y2);
                  else if (buff[0] == MSG_GET_IMG_RANGE)
                     rval = chipsSys->get_image_range(cid, x1, x2, y1, y2);
                  else if (buff[0] == MSG_GET_CNTR_RANGE)
                     rval = chipsSys->get_contour_range(cid, x1, x2, y1, y2);
                  else if (buff[0] == MSG_GET_PLOT_RANGE)
                     rval = chipsSys->get_plot_range(cid, x1, x2, y1, y2);
                  if (rval == ERR_STATUS_GOOD)
                  {
                     int* ival;
                     if ((ival = (int*)malloc(12*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = 4;
                        double* dp = (double*)&ival[4];
                        dp[0] = x1;
                        dp[1] = x2;
                        dp[2] = y1;
                        dp[3] = y2;
                        ival[1] = 12 *sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_LAB_VALIGN:    // fallthrough intended
               case MSG_GET_LAB_HALIGN:    // fallthrough intended
               case MSG_GET_CBAR_LAB_HAL:  // fallthrough intended
               case MSG_GET_CBAR_LAB_VAL:  // fallthrough intended
               case MSG_GET_TCKLB_VALIGN:  // fallthrough intended
               case MSG_GET_TCKLB_HALIGN:  // fallthrough intended
               case MSG_GET_AXSLB_VALIGN:  // fallthrough intended
               case MSG_GET_AXSLB_HALIGN:  // fallthrough intended
               case MSG_GET_TITLE_VALIGN:  // fallthrough intended
               case MSG_GET_TITLE_HALIGN:  // fallthrough intended
               case MSG_GET_CBAR_HALIGN:   // fallthrough intended
               case MSG_GET_CBAR_VALIGN:   // fallthrough intended
               case MSG_GET_AXIS_PAR_OFF:  // fallthrough intended
               case MSG_GET_AXIS_PERP_OFF: // fallthrough intended
               case MSG_GET_CBAR_PAR_OFF:  // fallthrough intended
               case MSG_GET_CBAR_PERP_OFF: // fallthrough intended
               case MSG_GET_LINE_THICK:    // fallthrough intended 
               case MSG_GET_TCKLB_ANGLE:   // fallthrough intended
               case MSG_GET_AXSLB_ANGLE:   // fallthrough intended
               case MSG_GET_AXIS_XLAB_ANG: // fallthrough intended
               case MSG_GET_AXIS_YLAB_ANG: // fallthrough intended
               case MSG_GET_TITLE_X_POS:   // fallthrough intended
               case MSG_GET_TITLE_Y_POS:   // fallthrough intended
               case MSG_GET_TITLE_ANGLE:   // fallthrough intended
               case MSG_GET_CV_SYM_ANGLE:  // fallthrough intended
               case MSG_GET_HST_SYM_ANGLE: // fallthrough intended
               case MSG_GET_POINT_ANGLE:   // fallthrough intended
               case MSG_GET_REGN_ANGLE:    // fallthrough intended
               case MSG_GET_PLOT_LMARGIN:  // fallthrough intended
               case MSG_GET_PLOT_BMARGIN:  // fallthrough intended
               case MSG_GET_PLOT_RMARGIN:  // fallthrough intended
               case MSG_GET_PLOT_TMARGIN:  // fallthrough intended
               case MSG_GET_CV_LINE_SIZE:  // fallthrough intended 
               case MSG_GET_HST_LINE_SIZE: // fallthrough intended 
               case MSG_GET_CV_ERR_SIZE:   // fallthrough intended 
               case MSG_GET_HST_ERR_SIZE:  // fallthrough intended 
               case MSG_GET_CNTR_THICK:    // fallthrough intended
               case MSG_GET_REGN_THICK:    // fallthrough intended
               case MSG_GET_AXIS_THICK:    // fallthrough intended
               case MSG_GET_CBAR_TK_THK:   // fallthrough intended
               case MSG_GET_AXIS_MAJ_THK:  // fallthrough intended
               case MSG_GET_AXIS_MIN_THK:  // fallthrough intended
               case MSG_GET_AXSG_MJ_SZ:    // fallthrough intended
               case MSG_GET_AXSG_MN_SZ:    // fallthrough intended
	       case MSG_GET_PLOT_ASPW:     // fallthrough intended
	       case MSG_GET_PLOT_ASPH:     // fallthrough intended
	       case MSG_GET_CBAR_TL_ANG:   // fallthrough intended
	       case MSG_GET_CBAR_TL_HAL:   // fallthrough intended
	       case MSG_GET_CBAR_TL_VAL:   // fallthrough intended
	       case MSG_GET_CBAR_LAB_ANG:  // fallthrough intended
               case MSG_GET_CBAR_LENGTH:   // fallthrough intended
               case MSG_GET_CBAR_WIDTH:    // fallthrough intended
               case MSG_GET_LAB_ANGLE:
               {
                  float resp;
                  int* ival;

                  if (buff[0] == MSG_GET_LAB_ANGLE)
                     rval = chipsSys->get_label_angle(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_LAB_ANG)
                     rval = chipsSys->get_colorbar_label_angle(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_TL_ANG)
                     rval = chipsSys->get_colorbar_ticklabel_angle(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_TL_HAL)
                     rval = chipsSys->get_colorbar_ticklabel_halign(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_TL_VAL)
                     rval = chipsSys->get_colorbar_ticklabel_valign(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_ASPW)
                     rval = chipsSys->get_plot_aspect_width(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_ASPH)
                     rval = chipsSys->get_plot_aspect_height(cid, resp);
                  else if (buff[0] == MSG_GET_CV_SYM_ANGLE)
                     rval = chipsSys->get_curve_symbolangle(cid, resp);
                  else if (buff[0] == MSG_GET_HST_SYM_ANGLE)
                     rval = chipsSys->get_histogram_symbolangle(cid, resp);
                  else if (buff[0] == MSG_GET_POINT_ANGLE)
                     rval = chipsSys->get_point_angle(cid, resp);
                  else if (buff[0] == MSG_GET_REGN_ANGLE)
                     rval = chipsSys->get_region_angle(cid, resp);
                  else if (buff[0] == MSG_GET_TITLE_ANGLE)
                     rval = chipsSys->get_plot_title_angle(cid, resp);
                  else if (buff[0] == MSG_GET_TITLE_X_POS)
                     rval = chipsSys->get_plot_title_xpos(cid, resp);
                  else if (buff[0] == MSG_GET_TITLE_Y_POS)
                     rval = chipsSys->get_plot_title_ypos(cid, resp);
                  else if (buff[0] == MSG_GET_LINE_THICK)
                     rval = chipsSys->get_line_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_HST_ERR_SIZE)
                     rval = chipsSys->get_histogram_errthickness(cid, resp);
                  else if (buff[0] == MSG_GET_CV_LINE_SIZE)
                     rval = chipsSys->get_curve_linethickness(cid, resp);
                  else if (buff[0] == MSG_GET_CV_ERR_SIZE)
                     rval = chipsSys->get_curve_errthickness(cid, resp);
                  else if (buff[0] == MSG_GET_HST_LINE_SIZE)
                     rval = chipsSys->get_histogram_linethickness(cid, resp);
                  else if (buff[0] == MSG_GET_REGN_THICK)
                     rval = chipsSys->get_region_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_CNTR_THICK)
                     rval = chipsSys->get_contour_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_TK_THK)
                     rval = chipsSys->get_colorbar_tick_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_THICK)
                     rval = chipsSys->get_axis_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_MAJ_THK)
                     rval = chipsSys->get_axis_major_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_MIN_THK)
                     rval = chipsSys->get_axis_minor_thickness(cid, resp);
                  else if (buff[0] == MSG_GET_AXSG_MJ_SZ)
                     rval = chipsSys->get_axis_majorgrid_thickness(cid,resp);
                  else if (buff[0] == MSG_GET_AXSG_MN_SZ)
                     rval = chipsSys->get_axis_minorgrid_thickness(cid,resp);
                  else if (buff[0] == MSG_GET_AXIS_PAR_OFF)
                     rval = chipsSys->get_axis_parallel_offset(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_PERP_OFF)
                     rval = chipsSys->get_axis_perpendicular_offset(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_PAR_OFF)
                     rval = chipsSys->get_colorbar_parallel_offset(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_PERP_OFF)
                     rval = chipsSys->get_colorbar_perpendicular_offset(cid, resp);
                  else if (buff[0] == MSG_GET_AXSLB_ANGLE)
                     rval = chipsSys->get_axis_label_angle(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_XLAB_ANG)
                     rval = chipsSys->get_axis_xlabel_angle(cid, resp);
                  else if (buff[0] == MSG_GET_AXIS_YLAB_ANG)
                     rval = chipsSys->get_axis_ylabel_angle(cid, resp);
                  else if (buff[0] == MSG_GET_TCKLB_ANGLE)
                     rval = chipsSys->get_axis_ticklabel_angle(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_LMARGIN)
                     rval = chipsSys->get_plot_leftmargin(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_RMARGIN)
                     rval = chipsSys->get_plot_rightmargin(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_TMARGIN)
                     rval = chipsSys->get_plot_topmargin(cid, resp);
                  else if (buff[0] == MSG_GET_PLOT_BMARGIN)
                     rval = chipsSys->get_plot_bottommargin(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_HALIGN)
                     rval = chipsSys->get_colorbar_halign(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_VALIGN)
                     rval = chipsSys->get_colorbar_valign(cid, resp);
                  else if (buff[0] == MSG_GET_TITLE_HALIGN)
                     rval = chipsSys->get_plot_title_halign(cid, resp);
                  else if (buff[0] == MSG_GET_TITLE_VALIGN)
                     rval = chipsSys->get_plot_title_valign(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_LAB_HAL)
                     rval = chipsSys->get_colorbar_label_halign(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_LAB_VAL)
                     rval = chipsSys->get_colorbar_label_valign(cid, resp);
                  else if (buff[0] == MSG_GET_LAB_HALIGN)
                     rval = chipsSys->get_label_halign(cid, resp);
                  else if (buff[0] == MSG_GET_LAB_VALIGN)
                     rval = chipsSys->get_label_valign(cid, resp);
                  else if (buff[0] == MSG_GET_TCKLB_HALIGN)
                     rval = chipsSys->get_axis_ticklabel_halign(cid, resp);
                  else if (buff[0] == MSG_GET_TCKLB_VALIGN)
                     rval = chipsSys->get_axis_ticklabel_valign(cid, resp);
                  else if (buff[0] == MSG_GET_AXSLB_HALIGN)
                     rval = chipsSys->get_axis_label_halign(cid, resp);
                  else if (buff[0] == MSG_GET_AXSLB_VALIGN)
                     rval = chipsSys->get_axis_label_valign(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_WIDTH)
                     rval = chipsSys->get_colorbar_width(cid, resp);
                  else if (buff[0] == MSG_GET_CBAR_LENGTH)
                     rval = chipsSys->get_colorbar_length(cid, resp);

                  if (rval == ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_FLOAT;
                        cp[3] = 0;
                        ival[3] = 1;
                        float* fp = (float*)&ival[4];
                        fp[0] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_OBJ_COUNT:
               {
                  int resp;
                  int* ival;
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if ((rval = chipsSys->get_object_count(cid, obs, resp)) ==
                     ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_HILITE_OBJ:   
               {
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rval = chipsSys->highlight_object(cid, obs);
               }
               break; 
               case MSG_GET_OBJ_COORD:   
               {
                  COORD_SYS resp;
		  string sid;
                  int* ival;
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if ((rval = chipsSys->get_object_coordinfo(cid, obs, resp, sid)) ==
                     ERR_STATUS_GOOD)
                  {
                     int ln = sid.length()/4 + 9;
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 2;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = MSG_DATATYPE_CHAR;
                        ival[3] = 1;
			ival[4] = sid.length()+1;
			ival[5] = 0;   // pad data start to double boundary
                        ival[6] = (int)resp;
			ival[7] = 0;   // pad data start to double boundary
			cp = (char*)&ival[8];
                        ival[1] = 8 *sizeof(int) + sid.length() + 1;
			if (sid.length())   strcpy(cp, sid.c_str());
			else                cp[0] = '\0';
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break; 

               case MSG_GET_LINE_STIPPLE:
               {
                  int resp;
                  int* ival;
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE obs = (V_LINE_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if ((rval = chipsSys->get_line_stipple(obs, resp)) ==
                     ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_SET_LINE_STIPPLE:
               {
                  int *v1 , *v2;
                  v1 = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  v2 = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*v1;
                  rval = chipsSys->set_line_stipple(typ, *v2);
               }
               break;
               case MSG_GET_ERR_VERB:
               {
                  int resp;
                  int* ival;
                  if ((rval = chipsSys->get_error_verbosity(resp)) ==
                     ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_SET_ERR_VERB:
               {
                  int *v1;
                  v1 = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rval = chipsSys->set_error_verbosity(*v1);
               }
               break;
               case MSG_GET_ERR_CNT:
               {
                  int resp;
                  int* ival;
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE obs = (V_LINE_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if ((rval = chipsSys->get_error_count(obs, resp)) ==
                     ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_ERR_TYPES:
               {
                  int resp;
                  int* ival;
                  if ((rval = chipsSys->get_error_types(resp)) ==
                     ERR_STATUS_GOOD)
                  {
                     if ((ival = (int*)malloc(5*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_INT;
                        cp[3] = 0;
                        ival[3] = 1;
                        ival[4] = resp;
                        ival[1] = 5 *sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
               break;
               case MSG_GET_CNTR_LEVELS:
               {
                  vector<double> cntr_levels;
                  if ((rval = chipsSys->get_contour_levels(cid, cntr_levels)) == ERR_STATUS_GOOD)
                  {
                     int* ival;
                     if ((ival = (int*)malloc(cntr_levels.size ()*sizeof(double) + 4*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = cntr_levels.size ();
                        double* dp = (double*)&ival[4];
                        for (unsigned int ii=0; ii<cntr_levels.size(); ii++)
                           dp[ii] = cntr_levels[ii];
                        // set the size to be the size of the data array plus the size of
                        // the header
                        ival[1] = cntr_levels.size() *sizeof(double) + 4*sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }

                  
               }
               break;
               case MSG_SET_TITLE_FNTSTYL:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_plot_title_fontstyle(cid, new_style);
               }
               break; 
               case MSG_SET_CBAR_TL_FSTY:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_ticklabel_fontstyle(cid, new_style);
               }
               break;
               case MSG_SET_CBAR_LAB_FSTY:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_label_fontstyle(cid, new_style);
               }
               break;
               case MSG_SET_LAB_STYLE:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_label_fontstyle(cid, new_style);
               }
               break;
               case MSG_SET_TCKLB_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_ticklabel_color(cid, new_color);
               }
               break;
               case MSG_SET_AXSLB_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_label_color(cid, new_color);
               }
               break;
               case MSG_SET_TCKLB_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_ticklabel_font(cid, new_font);
               }
               break;
               case MSG_SET_TCKLB_FNTSTYL:
               {
                  string new_style = (char*)&bf[ndx];
                  rval=chipsSys->set_axis_ticklabel_fontstyle(cid, new_style);
               }
               break;
               case MSG_SET_AXSLB_FONT:
               {
                  string new_font = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_label_font(cid, new_font);
               }
               break;
               case MSG_SET_AXSLB_FNTSTYL:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_label_fontstyle(cid, new_style);
               }
               break;
               case MSG_SET_LAYOUT_THICK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_layout_thickness(cid,*thick);
               }
               break;
               case MSG_SET_LAYOUT_STYLE:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_layout_linestyle(cid,typ);
               }
               break;
               case MSG_SHOW_LAYOUT:
                  rval = chipsSys->display_frame_layout(cid);
               break;
               case MSG_HIDE_LAYOUT:
                  rval = chipsSys->hide_frame_layout(cid);
               break;
               case MSG_ADD_LEGEND:   // add legend to plot
               {
                  int *cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
   
                  if (*cnt > 0)
                  {
                     const char** c1 =
                        (const char**)malloc((*cnt+1)*sizeof(char*));
                     const char** c2 = 
                        (const char**)malloc((*cnt+1)*sizeof(char*));
                     int rr;
                     if (c1 && c2)
                     {
                        char* cp = &bf[ndx];
                        c1[0] = cp++;
                        for (rr = 1; rr < *cnt; rr++)
                        {
                           while(*cp != '\0') cp++;
                           c1[rr] = ++cp;
                        }
                        while(*cp != '\0') cp++;
                        c2[0] = ++cp;
                        for (rr = 1; rr < *cnt; rr++)
                        {
                           while(*cp != '\0') cp++;
                           c2[rr] = ++cp;
                        }
                        while(*cp != '\0') cp++;
                        cp++;
                        c1[*cnt] = NULL;
                        c2[*cnt] = NULL;

                        rval = chipsSys->add_legend(cid, c1, c2, cp);
                        free(c1);
                        free(c2);
                     }
                  }
               }
               break;
               case MSG_ADD_IMAGE:   // add image
	       {
                  string atts, fnm = (char*)&bf[ndx];
                  int ln = fnm.length();
                  atts = (char*)&bf[ndx + ln + 1];
                  int *trans_exist;
                  Transform *trans = static_cast <Transform *>(0);

                  // find transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform,
                                                           serial_transform_size);

                     delete []serial_transform;

                  }

                  rval = chipsSys->add_image(cid, fnm, trans, atts);
                  if (trans)
                     delete trans;
	       }
               break;
               case MSG_ADD_LABEL:   // add label
               {
                  double *x1, *y1;
                  char* cp; 
                  string txt = "", tid = ""; 
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  cp = &bf[ndx];
                  if (cp != NULL) 
                  {
                     txt = cp;
                     cp = &bf[ndx + strlen(cp) + 1];
                     if (cp != NULL) tid = cp;
                  }
                  rval = chipsSys->add_label(cid, *x1, *y1, txt, tid);
               }
               break;
               case MSG_ADD_CBAR:   // add colorbar
               {
                  double *x1, *y1;
                  char* cp;
                  string txt = "";
                  string tid = "";
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  cp = &bf[ndx];
                  if (cp != NULL)
                  {
                     txt = cp;
                     cp = &bf[ndx + strlen(cp) + 1];
                     if (cp != NULL) tid = cp;
                  }
                  rval = chipsSys->add_colorbar(cid, *x1, *y1, txt, obs, tid);
               }
               break;
               case MSG_LOAD_PREFS_A:
                  rval = chipsSys->load_preferences();
               break;
               case MSG_LOAD_PREFS_B:
                  rval = chipsSys->load_preferences(&bf[ndx]);
               break;
               case MSG_CMDLNE_CONN:
               {
                  string myconn = &bf[ndx]; 
                  if (myconn == "guicmdline")
                     cmdlineConn = 1;
               }
               break; 
               case MSG_SAVE_PREFS:
               {
                  int* md = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  mode_t mode = *md;
                  chipsSys->set_umask(mode);
                  rval = chipsSys->save_preferences(&bf[ndx]);
               }
               break;
	       case MSG_LOAD_STATE:
                  rval = chipsSys->load_state(&bf[ndx]);
               break;
               case MSG_SAVE_STATE:
               {
                  int* md = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  mode_t mode = *md;
                  chipsSys->set_umask(mode);
                  rval = chipsSys->save_state(&bf[ndx]);
               }
               break;
               case MSG_SAVE_STATE_B:
               {
                  int *md, *otyp;
                  md = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  mode_t mode = *md;
                  chipsSys->set_umask(mode);
                  otyp = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);

                  rval = chipsSys->save_state(&bf[ndx], *otyp);
               }
               break;
               case MSG_ADD_WINDOW_B:   // new window w/ id
                  if ((rval = chipsSys->add_window(cid,  &bf[ndx])) == 
                      ERR_STATUS_GOOD)
                     cid->reset_frame_id("");
               break;
               case MSG_ADD_PLOT:
                  rval = chipsSys->add_plot(cid, &bf[ndx]);
               break;
               case MSG_ADD_FRAME_B:   // new frame w/ id
                  rval = chipsSys->add_frame(cid, &bf[ndx]);
               break;
               case MSG_ADD_FRAME_C:
               {
                  float *x1, *y1, *x2, *y2;
                  char  *fid;
   
                  x1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  x2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  fid = &bf[ndx];
                  rval = chipsSys->add_frame(cid, *x1, *y1, *x2, *y2, fid);
               }
               break;
               case MSG_ADD_AXIS:
               {
                  float *ps, *minx, *maxx, *miny, *maxy;
                  int *dim;
                  char  *aid;
                  int *trans_exist;
                  Transform *trans = static_cast <Transform *>(0);
   
                  dim = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ps = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  minx = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  maxx = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  miny = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  maxy = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);


                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size]
;
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform,
                                                           serial_transform_size
);

                     delete []serial_transform;

                  }

                  aid = &bf[ndx];
                  rval = chipsSys->add_axis(cid,*dim,*ps,*minx,*maxx,*miny,*maxy,trans,aid);
               }
               break;
               case MSG_SET_REGN_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_region(cid);
                  else
                     rval = chipsSys->hide_region(cid);
               }
               break;
               case MSG_SET_PLOT_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_plot(cid);
                  else
                     rval = chipsSys->hide_plot(cid);
               }
               break;
               case MSG_SET_AXIS_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_axis(cid);
                  else
                     rval = chipsSys->hide_axis(cid);
               }
               break;
               case MSG_SET_CV_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_curve(cid);
                  else
                     rval = chipsSys->hide_curve(cid);
               }
               break;
               case MSG_SET_IMG_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_image(cid);
                  else
                     rval = chipsSys->hide_image(cid);
               }
               break;
               case MSG_SET_CBAR_ORIENT:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_orientation(cid, *vis);
               }
               break;
               case MSG_SET_CBAR_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_colorbar(cid);
                  else
                     rval = chipsSys->hide_colorbar(cid);
               }
               break;
               case MSG_SET_CNTR_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_contour(cid);
                  else
                     rval = chipsSys->hide_contour(cid);
               }
               break;
               case MSG_SET_HST_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_histogram(cid);
                  else
                     rval = chipsSys->hide_histogram(cid);
               }
               break;
               case MSG_REVERSE_AXIS:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->reverse_axes(cid, *ival);
               }
               break;
               case MSG_SET_WIN_DRAW_ON:
                  rval = chipsSys->set_window_redraw(cid, true); 
               break;
               case MSG_SET_WIN_DRAW_OFF:
                  rval = chipsSys->set_window_redraw(cid, false); 
               break;
               case MSG_SET_REDRAW:
               {
                  int *ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_redraw((bool)*ival);
               }
               break;
               case MSG_PICK:
               {
		  unsigned int* uval; 
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  uval = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->pick(cid, *ival, *uval);
               }
               break;
               case MSG_SET_LEGEND_VIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_legend(cid);
                  else
                     rval = chipsSys->hide_legend(cid);
               }
               break;
               case MSG_SET_FRAME_TRANS:
               {
                  bool *vis = (bool*)&bf[NEXT_BOOL_POS(ndx)]; 
                  rval = chipsSys->set_frame_transparency(cid, *vis);
               }
               break;
               case MSG_SET_CBAR_BVIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_border_visible(cid, *vis);
               }
               break;
               case MSG_SET_FRAME_BVIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_frame_border(cid);
                  else
                     rval = chipsSys->hide_frame_border(cid);
               }
               break;
               case MSG_SET_FRAME_VIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_frame(cid);
                  else
                     rval = chipsSys->hide_frame(cid);
               }
               break;
               case MSG_SET_FRAME_SCALE:
               {
                  bool *vis = (bool*)&bf[NEXT_BOOL_POS(ndx)]; 
                  rval = chipsSys->set_frame_scale(cid, *vis);
               }
               break;
               case MSG_SET_LEGEND_TRANS:
               {
                  bool *vis = (bool*)&bf[NEXT_BOOL_POS(ndx)]; 
                  rval = chipsSys->set_legend_transparency(cid, *vis);
               }
               break;
               case MSG_SET_LEGEND_BORDER:
               {
                  bool *vis = (bool*)&bf[NEXT_BOOL_POS(ndx)]; 
                  rval = chipsSys->set_legend_borders(cid, *vis);
               }
               break;
               case MSG_SET_AXIS_SCALABLE:
               {
                  bool *vis = (bool*)&bf[NEXT_BOOL_POS(ndx)]; 
                  rval = chipsSys->set_axis_scalable(cid, *vis);
               }
               break;
               case MSG_SET_CBAR_TK_STY:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_tickstyle(cid, *vis);
               }
               break;
               case MSG_SET_CBAR_TK_LOC:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_tick_location(cid, *vis);
               }
               break;
               case MSG_SET_AXIS_MAJ_STYL:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_major_style(cid, *vis);
               }
               break;
               case MSG_SET_AXIS_MIN_STYL:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_minor_style(cid, *vis);
               }
               break;
               case MSG_SET_AXIS_MAJ_VIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_major_ticks(cid);
                  else
                     rval = chipsSys->hide_major_ticks(cid);
               }
               break;
               case MSG_SET_AXIS_MIN_VIS:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_minor_ticks(cid);
                  else
                     rval = chipsSys->hide_minor_ticks(cid);
               }
               break;
               case MSG_ADD_LINE:
               {
                  double *x1, *y1, *x2, *y2;
                  char  *lid;
   
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  x2 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y2 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  lid = &bf[ndx];
                  rval = chipsSys->add_line(cid,*x1,*y1,*x2,*y2,lid);
               }
               break;
               case MSG_ADD_LINE_B:
               {
                  unsigned int *cnt;
                  double *xx, *yy;
                  char  *lid;
  
                  cnt = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
		  lid = &bf[ndx];
                  ndx += (strlen(lid) + 1);
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*cnt)* sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->add_line(cid, xx, yy, *cnt, lid);
               }
               break;
               case MSG_ADD_REGION:
               {
                  unsigned int *cnt;
                  double *xx, *yy;
                  char  *lid;

                  cnt = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
		  lid = &bf[ndx];
                  ndx += (strlen(lid) + 1);
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*cnt)* sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->add_region(cid, xx, yy, *cnt, lid);
               }
               break;
               case MSG_ADD_REGION_B:
               {
                  int *side;
                  double *xx, *yy, *rad;
                  char  *lid;

                  side = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) +sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rad =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  lid = &bf[ndx];
                  rval = chipsSys->add_region(cid, *side, *xx, *yy, *rad, lid);
               }
               break;
               case MSG_SET_LINE_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_line_color(cid, new_color);
               }
               break;
               case MSG_SET_LINE_THICK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_line_thickness(cid,*thick);
               }
               break;
               case MSG_SET_LINE_STYLE:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_line_style(cid, typ);
               }
               break;
               case MSG_SET_LINE_EXTEND:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_EXTENSION_TYPE typ = (V_LINE_EXTENSION_TYPE)*style;
                  rval = chipsSys->set_line_extend(cid, typ);
               }
               break;
               case MSG_SET_AXIS_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_color(cid, new_color);
               }
               break;
               case MSG_SET_CBAR_TK_MODE:
               {
                  string new_mode = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_mode(cid, new_mode);
               }
               break;
               case MSG_SET_AXIS_MAJ_MODE:
               {
                  string new_mode = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_major_mode(cid, new_mode);
               }
               break;
               case MSG_SET_AXIS_MIN_MODE:
               {
                  string new_mode = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_minor_mode(cid, new_mode);
               }
               break;
               case MSG_SET_CBAR_TL_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_ticklabel_color(cid, new_color);
               }
               break;
               case MSG_SET_AXIS_MAJ_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_major_color(cid, new_color);
               }
               break;
               case MSG_SET_AXIS_MIN_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_minor_color(cid, new_color);
               }
               break;
               case MSG_SET_CBAR_TK_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_tick_color(cid, new_color);
               }
               break;
               case MSG_SET_CBAR_LAB_TXT:
               {
                  string new_text = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_label_text(cid, new_text);
               }
               break;
               case MSG_SET_AXIS_TEXT:
               {
                  string new_text = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_text(cid, new_text);
               }
               break;
               case MSG_SET_AXIS_XLAB_TXT:
               {
                  string new_text = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_xlabel_text(cid, new_text);
               }
               break;
               case MSG_SET_AXIS_YLAB_TXT:
               {
                  string new_text = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_ylabel_text(cid, new_text);
               }
               break;
               case MSG_SET_AXIS_TICKFMT:
               {
                  string new_format = (char*)&bf[ndx];
                  rval = chipsSys->set_axis_tickformat(cid, new_format);
               }
               break;
               case MSG_SET_AXIS_THICK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_thickness(cid,*thick);
               }
               break;
               case MSG_SET_CBAR_LAB_LOC:
               {
                  int *styl = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_location(cid,*styl);
               }
               break;
               case MSG_SET_AXIS_TICKSTYL:
               {
                  int *styl = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_tickstyle(cid,*styl);
               }
               break;
               case MSG_SET_AXSTCKLB_STYL:
               {
                  int *styl = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_style(cid,*styl);
               }
               break;
               case MSG_SET_AXIS_MAJ_NUM:
               {
                  int *num = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_major_num(cid,*num);
               }
               break;
               case MSG_SET_AXIS_MIN_NUM:
               {
                  int *num = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_minor_num(cid,*num);
               }
               break;
               case MSG_SET_AXIS_MAJ_LEN:
               {
                  int *major = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_major_length(cid,*major);
               }
               break;
               case MSG_SET_AXIS_MIN_LEN:
               {
                  int *minor = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_minor_length(cid,*minor);
               }
               break;
               case MSG_SET_CBAR_TK_LEN:
               {
                  int *len = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_tick_length(cid,*len);
               }
               break;
               case MSG_SET_AXIS_MAJ_INT:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_axis_major_interval(cid, *dval);
               }
               break;
               case MSG_SET_AXIS_MIN_INT:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_axis_minor_interval(cid, *dval);
               }
               break;
               case MSG_SET_CNTR_INT:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_contour_interval(cid, *dval);
               }
               break;
               case MSG_SET_AXIS_PAD:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_axis_pad(cid, *dval);
               }
               break;
               case MSG_SET_REGN_OPACITY:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_region_opacity(cid, *dval);
               }
               break;
               case MSG_ZOOM_PLOT:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->zoom(cid, *fval);
               }
               break;
               case MSG_ZOOM_PLOT_GUI:
               {
                  rval = chipsSys->zoom(cid);
               }
               break;
               case MSG_PANTO_PLOT:
               {
                  double *dmax, *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->panto(cid, *dmin, *dmax);
               }
               break;
               case MSG_PAN_PLOT:
               {
                  rval = chipsSys->pan(cid);
               }
               break;
               case MSG_SET_LIMITS_X:
               {
                  double *dmax, *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, X_AXIS, *dmin, *dmax);
               }
               break;
               case MSG_SET_LIMITS_Y:
               {
                  double *dmax, *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, Y_AXIS, *dmin, *dmax);
               }
               break;
               case MSG_SET_LIMITS_XY:
               {
                  double *dmax, *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, XY_AXIS, *dmin, *dmax);
               }
               break;
               case MSG_SET_LIMITS_XX_YY:
               {
                  double *dxmax, *dxmin;
                  double *dymax, *dymin;

                  dxmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dxmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dymin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  dymax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);

                  rval = chipsSys->limits(cid, *dxmin, *dxmax, *dymin, *dymax);
               }
               break;
               case MSG_SET_LIMITS_X_AA:
               {
                  rval = chipsSys->limits(cid, X_AXIS, AUTO, AUTO);
               }
               break;
               case MSG_SET_LIMITS_Y_AA:
               {
                  rval = chipsSys->limits(cid, Y_AXIS, AUTO, AUTO);
               }
               break;
               case MSG_SET_LIMITS_XY_AA:
               {
                  rval = chipsSys->limits(cid, XY_AXIS, AUTO, AUTO);
               }
               break;
               case MSG_SET_LIMITS_X_MA:
               {
                  double *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->limits(cid, X_AXIS, *dmin, AUTO);
               }
               break;
               case MSG_SET_LIMITS_Y_MA:
               {
                  double *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->limits(cid, Y_AXIS, *dmin, AUTO);
               }
               break;
               case MSG_SET_LIMITS_XY_MA:
               {
                  double *dmin = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->limits(cid, XY_AXIS, *dmin, AUTO);
               }
               break;
               case MSG_SET_LIMITS_X_AM:
               {
                  double *dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, X_AXIS, AUTO, *dmax);
               }
               break;
               case MSG_SET_LIMITS_Y_AM:
               {
                  double *dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, Y_AXIS, AUTO, *dmax);
               }
               break;
               case MSG_SET_LIMITS_XY_AM:
               {
                  double *dmax = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  rval = chipsSys->limits(cid, XY_AXIS, AUTO, *dmax);
               }
               break;
               case MSG_SET_LIMITS_OID:
               {
                  int *ival, *ival2;
                  ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ival2 = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  string oid = (char*)&bf[ndx];
                  V_OBJECT_TYPE ob = (V_OBJECT_TYPE)*ival2;  
                  rval = chipsSys->limits(cid, *ival, ob, oid);
               }
               break;
               case MSG_SET_LIMITS_CID:
               {
                  ChipsId ccid;
                  int *ival, *ival2;
                  ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ival2 = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  string oid = (char*)&bf[ndx];
                  V_OBJECT_TYPE ob = (V_OBJECT_TYPE)*ival2;
                  ccid.set_window_id(oid);
                  ndx += oid.length() +1;  
                  oid = (char*)&bf[ndx];
                  ccid.set_frame_id(oid);
                  ndx += oid.length() +1;  
                  oid = (char*)&bf[ndx];
                  ccid.set_plot_id(oid);
                  ndx += oid.length() +1;  
                  oid = (char*)&bf[ndx];
                  if (ob == chips_curve)  
                     ccid.set_curve_id(oid);
                  else if (ob == chips_contour)  
                     ccid.set_contour_id(oid);
                  else if (ob == chips_histogram)  
                     ccid.set_histogram_id(oid);
                  rval = chipsSys->limits(cid, *ival, ob, &ccid);
               }
               break;
               case MSG_SET_LINEARSCALE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->linear_scale(cid, *ival);
               }
               break;
               case MSG_SET_LOGSCALE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->log_scale(cid, *ival);
               }
               break;
               case MSG_PICK_LIMITS:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->pick_limits(cid, *ival);
               }
               break;
               case MSG_REPOSITION_FRAME:
               {
                  float *x1, *y1, *x2, *y2;
   
                  x1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  x2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  rval = chipsSys->reposition_frame(cid, *x1, *y1, *x2, *y2);
               }
               break;
               case MSG_REPOSITION_PLOT:
               {
                  float *x1, *y1, *x2, *y2;

                  x1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  x2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  rval = chipsSys->reposition_plot(cid, *x1, *y1, *x2, *y2);
               }
               break;
               case MSG_SWAP_OBJ_POS:
               {
                  string f2 = (char*)&bf[ndx];
                  rval = chipsSys->swap_object_positions(cid, f2);
               }
               break;
               case MSG_SET_IMG_ALPHA:
               {
                  double *min, *max;
                  min =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  max =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_image_alpha(cid, *min, *max);
               }
               break;
               case MSG_SET_IMG_THRESH:
               {
                  double *min, *max;
                  min =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  max =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_image_threshold(cid, *min, *max);
               }
               break;
               case MSG_GRID_GAP:
               {
                  float *xx, *yy;
                  xx =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  yy =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->adjust_grid_gaps(cid, *xx, *yy);
               }
               break;
               case MSG_GRID_XRELSIZE:
               {
                  int *col; 
                  float  *ratio = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  col = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->adjust_grid_xrelsize(cid, *col, *ratio);
               }
               break;
               case MSG_GRID_YRELSIZE:
               {
                  int *col; 
                  float  *ratio = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  col = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->adjust_grid_yrelsize(cid, *col, *ratio);
               }
               break;
               case MSG_GRID_XRELSIZES:
               {
                  float *ps;
                  int *cnt;
                  cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ps =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->adjust_grid_xrelsizes(cid, *cnt, ps);
               }
               break;
               case MSG_GRID_YRELSIZES:
               {
                  float *ps;
                  int *cnt;
                  cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ps =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->adjust_grid_yrelsizes(cid, *cnt, ps);
               }
               break;
               case MSG_GRID_XGAP:
               {
                  float *xx;
                  xx = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->adjust_grid_xgap(cid, *xx);
               }
               break;
               case MSG_GRID_YGAP:
               {
                  float *yy;
                  yy =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->adjust_grid_ygap(cid, *yy);
               }
               break;
               case MSG_GRID_OBJS:
               {
                  string* c1 = NULL;
                  int rr, *rv, *cv, *us, *cnt;
                  float *xx, *yy;

                  xx =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  yy =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  cv = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rv = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  us = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*cnt > 0)
                  {
                     c1 =  new string[*cnt];
                     if (c1)
                     {
                        char* cp = &bf[ndx];
                        c1[0] = cp++;
                        for (rr = 1; rr < *cnt; rr++)
                        {
                           while(*cp != '\0') cp++;
                           c1[rr] = ++cp;
                        }
                     }
                  }
                  rval = chipsSys->grid_objects(cid, *cv, *rv, *xx, *yy,
                                                *us, *cnt, c1);
                  if (c1)    delete[] c1;
               }
               break;
               case MSG_GRID_OBJS_B:
               {
                  string* c1 = NULL; 
                  int *ornt, *cv, *us, *lcnt, *vls = NULL;
                  float *xx, *yy;
                  int rr;

                  xx =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  yy =  (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  ornt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  cv = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  us = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*cv > 0)
                  {
                     vls = (int*)malloc((*cv)*sizeof(int));
                     for (rr = 0; rr < *cv; rr++)
                     {
                        vls[rr] = *(int*)&bf[NEXT_INT_POS(ndx)];
                        ndx = NEXT_INT_POS(ndx) + sizeof(int);
                     }
                  } 
                  lcnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*lcnt > 0)
                  {
                     c1 =  new string[*lcnt];
                     if (c1)
                     {
                        int qq;

                        char* cp = &bf[ndx];
                        c1[0] = cp++;
                        for (qq = 1; qq < *lcnt; qq++)
                        {
                           while(*cp != '\0') cp++;
                           c1[qq] = ++cp;
                        }
                     }
                  }
                  if (*ornt == 1)
                     rval = chipsSys->col_grid_objects(cid, *cv, vls, *xx,
                                                        *yy, *us, *lcnt, c1);
                  else 
                     rval = chipsSys->row_grid_objects(cid, *cv, vls, *xx,
                                                        *yy, *us, *lcnt, c1);
                  if (vls)  free(vls);
                  if (c1)  delete[] c1;
               }
               break;
               case MSG_ADD_PLOT_B:
               {
                  float *x1, *y1, *x2, *y2;
                  char  *plid;
   
                  x1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y1 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  x2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  y2 = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_FLOAT_POS(ndx) + sizeof(float);
                  plid = &bf[ndx];
                  rval = chipsSys->add_plot(cid, *x1, *y1, *x2, *y2, plid);
               }
               break;
               case MSG_SET_PREFERENCE:
               {
                  string val, pref = (char*)&bf[ndx];
                  int ln = pref.length();
                  val = (char*)&bf[ndx + ln + 1];
                  rval = chipsSys->set_preference(pref, val);
               }
               break;
               case MSG_MOVE:
               {
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  double *x1, *y1;
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  int *md; 
                  md = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->move(cid, obs, *x1, *y1, *md);
               }
               break;
               case MSG_REARRANGE_DEPTH:
               {
                  int *manner, *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  manner =  (int*)&bf[NEXT_INT_POS(ndx)];
                  V_SHUFFLE_TYPE way = (V_SHUFFLE_TYPE)*manner; 
                  switch(way)
                  {
                     case chips_forward:
                        rval = chipsSys->shuffle_forward(cid, obs);
                     break;
                     case chips_backward:
                        rval = chipsSys->shuffle_backward(cid, obs);
                     break;
                     case chips_front:
                        rval = chipsSys->shuffle_front(cid, obs);
                     break;
                     case chips_back:
                        rval = chipsSys->shuffle_back(cid, obs);
                     break;
                     default:
		     { 
                        char buf[256];
			sprintf(buf, "The shuffle type value (%1d) must be one of [chips_forward|chips_backward|chips_front|chips_back]\n", (int)way);
                        chipsSys->add_client_error(rval=ERR_INT_SUBSET, buf);
		     }
                     break; 
                  } 
               }
	       break; 
               case MSG_SET_CASCADE:
               {
                  int *obj = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*obj;
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  string val, prop = (char*)&bf[ndx];
                  int ln = prop.length();
                  val = (char*)&bf[ndx + ln + 1];
                  rval = chipsSys->set_cascading_property(cid, obs, prop, val);
               }
               break;
               case MSG_SET_CBAR_WIDTH:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_width(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_LENGTH:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_length(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_TL_ANG: 
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_angle(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_TL_HAL:  
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_halign(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_TL_VAL:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_valign(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_LAB_ANG:   // change colorbar label angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_angle(cid, *fval);
               }
               break;
               case MSG_SET_LAB_ANGLE:   // change label angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_label_angle(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_LAB_HAL:  
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_halign(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_LAB_VAL:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_valign(cid, *fval);
               }
               break;
               case MSG_SET_LAB_HALIGN:   // change label horiz alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_label_halign(cid, *fval);
               }
               break;
               case MSG_SET_LAB_VALIGN:   // change label vert alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_label_valign(cid, *fval);
               }
               break;
               case MSG_SET_AXSLB_ANGLE:   // change axis label angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_label_angle(cid, *fval);
               }
               break; 
               case MSG_SET_AXIS_XLAB_ANG:   // change x axis label angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_xlabel_angle(cid, *fval);
               }
               break;
               case MSG_SET_AXIS_YLAB_ANG:   // change y axis label angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_ylabel_angle(cid, *fval);
               }
               break;
               case MSG_SET_AXSLB_HALIGN:  // change axis label horiz alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_label_halign(cid, *fval);
               }
               break; 
               case MSG_SET_AXSLB_VALIGN:  // change axis label vert alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_label_valign(cid, *fval);
               }
               break; 
               case MSG_SET_TCKLB_ANGLE:   // change axis ticklabel angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_angle(cid, *fval);
               }
               break; 
               case MSG_SET_TCKLB_HALIGN:  // change axis ticklabel horiz align
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_halign(cid, *fval);
               }
               break; 
               case MSG_SET_TCKLB_VALIGN:  // change axis ticklabel vert align
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_ticklabel_valign(cid, *fval);
               }
               break; 
               case MSG_SET_TITLE_ANGLE:   // change plot title angle
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_title_angle(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_HALIGN:  // change colorbar horiz alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_halign(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_VALIGN:  // change colorbar vertical alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_valign(cid, *fval);
               }
               break;
               case MSG_SET_TITLE_HALIGN:  // change plot title horiz alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_title_halign(cid, *fval);
               }
               break;
               case MSG_SET_TITLE_VALIGN:  // change plot title vert alignment
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_title_valign(cid, *fval);
               }
               break;
               case MSG_SET_TITLE_X_POS:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_title_xpos(cid, *fval);
               }
               break;
               case MSG_SET_TITLE_Y_POS:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_title_ypos(cid, *fval);
               }
               break;
               case MSG_SET_AXIS_PAR_OFF: // change axis parallel offset
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_parallel_offset(cid, *fval);
               }
               break;
               case MSG_SET_AXIS_PERP_OFF: // change axis perpenmdicular offset
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_axis_perpendicular_offset(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_PAR_OFF: // change axis parallel offset
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_parallel_offset(cid, *fval);
               }
               break;
               case MSG_SET_CBAR_PERP_OFF: // change axis perpenmdicular offset
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_colorbar_perpendicular_offset(cid, *fval);
               }
               break;
               case MSG_SET_PLOT_LMARGIN:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_leftmargin(cid, *fval);
               }
               break;
               case MSG_SET_PLOT_RMARGIN:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_rightmargin(cid, *fval);
               }
               break;
               case MSG_SET_PLOT_TMARGIN:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_topmargin(cid, *fval);
               }
               break;
               case MSG_SET_PLOT_BMARGIN:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_plot_bottommargin(cid, *fval);
               }
               break;
               case MSG_SET_LAB_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_label(cid);
                  else
                     rval = chipsSys->hide_label(cid);
               }
               break;
	       case MSG_SET_LINE_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_line(cid);
                  else
                     rval = chipsSys->hide_line(cid);
               }
               break;
               case MSG_SET_POINT_VISIBLE:
               {
                  int *vis = (int*)&bf[NEXT_INT_POS(ndx)];
                  if (*vis == 1)
                     rval = chipsSys->display_point(cid);
                  else
                     rval = chipsSys->hide_point(cid);
               }
               break;
               case MSG_SET_TITLE_SIZE:   // change title size
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_plot_title_size(cid, *ival);
               }
               break;
               case MSG_SET_CBAR_TL_SZ:  
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_ticklabel_size(cid, *ival);
               }
               break;
               case MSG_SET_CBAR_LAB_SZ:   // change colorbar label size
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_label_size(cid, *ival);
               }
               break;
               case MSG_SET_LAB_SIZE:   // change label size
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_label_size(cid, *ival);
               }
               break;
               case MSG_SET_LAB_DEPTH:   // change label depth
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_label_depth(cid, *ival);
               }
               break;
               case MSG_SET_LAB_TEXT:
                  rval = chipsSys->set_label_text(cid, &bf[ndx]);
               break;
               case MSG_SET_CBAR_LAB_CLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_colorbar_label_color(cid, new_color);
               }
               break;
               case MSG_SET_LAB_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_label_color(cid, new_color);
               }
               break;
               case MSG_ADD_POINT:
               {
                  double *x1, *y1;
                  char  *tid;
   
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  tid = &bf[ndx];
                  rval = chipsSys->add_point(cid, *x1, *y1, tid);
               }
               break;
               case MSG_ADD_POINT_B:
               {
                  double *x1, *y1, *z1;
                  char  *tid;
   
                  x1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  y1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  z1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  tid = &bf[ndx];
                  rval = chipsSys->add_point(cid, *x1, *y1, *z1, tid);
               }
               break;
               case MSG_SET_POINT_SIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_point_size(cid, *ival);
               }
               break;
               case MSG_SET_POINT_ANGLE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_point_angle(cid, *fval);
               }
               break;
               case MSG_SET_POINT_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_point_depth(cid, *ival);
               }
               break;
               case MSG_SET_LINE_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_line_depth(cid, *ival);
               }
               break;
               case MSG_SET_AXIS_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_axis_depth(cid, *ival);
               }
               break;
               case MSG_SET_CV_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_depth(cid, *ival);
               }
	       break; 
               case MSG_SET_IMG_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_image_depth(cid, *ival);
               }
               break;
               case MSG_SET_CBAR_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_colorbar_depth(cid, *ival);
               }
               break;
               case MSG_SET_CNTR_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_contour_depth(cid, *ival);
               }
               break;
               case MSG_SET_DEF_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_default_depth(cid, *ival);
               }
               break;
               case MSG_SET_HST_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_histogram_depth(cid, *ival);
               }
               break;
               case MSG_SET_HST_FILL_STYL:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_FILL_TYPE typ = (V_FILL_TYPE)*style;
                  rval = chipsSys->set_histogram_fillstyle(cid, typ);
               }
               break;
               case MSG_SET_PLOT_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_plot_title_depth(cid, *ival);
               }
               break;
               case MSG_SET_CV_SYM_ANGLE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_curve_symbolangle(cid, *fval);
               }
               break;
               case MSG_SET_HST_SYM_ANGLE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_histogram_symbolangle(cid, *fval);
               }
               break;
               case MSG_SET_POINT_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_point_color(cid, new_color);
               }
               break;
               case MSG_SET_LEGEND_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_legend_color(cid, new_color);
               }
               break;
               case MSG_SET_LEGEND_THICK:
               {
                  float *thick = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_legend_thickness(cid,*thick);
               }
               break;
               case MSG_SET_POINT_FILL:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_point_fill(cid, *bval);
               }
               break;
               case MSG_SET_REGN_DEPTH:   
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_region_depth(cid, *ival);
               }
               break;
               case MSG_SET_REGN_ANGLE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_region_angle(cid, *fval);
               }
               break;
               case MSG_SET_REGN_THICK:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_region_thickness(cid, *fval);
               }
               break;
               case MSG_SET_REGN_STYLE:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_region_style(cid, typ);
               }
               break;
               case MSG_SET_REGN_FILSTYL:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_FILL_TYPE typ = (V_FILL_TYPE)*style;
                  rval = chipsSys->set_region_fillstyle(cid, typ);
               }
               break;
               case MSG_SET_REGN_EDGCOL:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_region_color(cid, new_color);
               }
               break;
               case MSG_SET_REGN_FILCOL:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_region_fillcolor(cid, new_color);
               }
               break;
               case MSG_SET_PLOT_AX_STYLE:
               {
                  int *ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_AXISSTYLE_TYPE styl = (V_AXISSTYLE_TYPE)*ival;
                  rval = chipsSys->set_plot_axisstyle(cid,styl);
               }
               break;
               case MSG_SET_POINT_SYMBOL:
               {
                  int *sym = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_SYMBOL_TYPE typ = (V_SYMBOL_TYPE)*sym;
                  rval = chipsSys->set_point_symbol(cid,typ);
               }
               break;
               case MSG_SET_CV_BADDATA:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_BADDATAMODE_TYPE typ = (V_BADDATAMODE_TYPE)*style;
                  rval = chipsSys->set_curve_baddatamode(cid, typ);
               }
               break;
               case MSG_SET_CV_LINE_STYLE:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_curve_linestyle(cid, typ);
               }
               break;
               case MSG_SET_CV_SYMBOL:
               {
                  int *sym = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_SYMBOL_TYPE typ = (V_SYMBOL_TYPE)*sym;
                  rval = chipsSys->set_curve_symboltype(cid, typ);
               }
               break;
               case MSG_SET_CV_SYM_FILL:
                  rval = chipsSys->set_curve_symbolfill(cid, true);
               break;
               case MSG_SET_CV_SYM_UNFILL:
                  rval = chipsSys->set_curve_symbolfill(cid, false);
               break;
               case MSG_SET_CV_SYM_SIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_symbolsize(cid, (int)*ival);
               }
               break;
               case MSG_SET_CV_ERR_CAPLEN:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_errcaplength(cid, (int)*ival);
               }
               break;
               case MSG_SET_CV_LINE_SIZE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_curve_linethickness(cid, (float)*fval);
               }
               break;
               case MSG_SET_CV_ERR_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_curve_errcolor(cid, new_color);
               }
               break;
               case MSG_SET_CV_ERR_UP:
               {
                  bool* bval = (bool*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_errup(cid, (bool)*bval);
               }
               break; 
               case MSG_SET_CV_ERR_DOWN:
               {
                  bool* bval = (bool*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_errdown(cid, (bool)*bval);
               }
               break; 
               case MSG_SET_CV_ERR_LEFT:
               {
                  bool* bval = (bool*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_errleft(cid, (bool)*bval);
               }
               break; 
               case MSG_SET_CV_ERR_RIGHT:
               {
                  bool* bval = (bool*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_errright(cid, (bool)*bval);
               }
               break;
               case MSG_SET_CV_LIMOVRIDE:
               {
                  bool* bval = (bool*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_curve_limitoverride(cid, (bool)*bval);
               }
               break;
               case MSG_SET_CV_LIMLENGTH:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_curve_limitlength(cid, *dval);
               }
               break;
               case MSG_SET_CV_LINE_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_curve_linecolor(cid, new_color);
               }
               break;
               case MSG_SET_CV_SYM_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_curve_symbolcolor(cid, new_color);
               }
               break;
               case MSG_SET_CV_ERR_SIZE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_curve_errthickness(cid, (float)*fval);
               }
               break;
               case MSG_SET_CV_ERR_STYLE:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_curve_errstyle(cid, new_style);
               }
               break;
               case MSG_SET_CNTR_THICK:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_contour_thickness(cid, *fval);
               }
               break;
               case MSG_SET_CNTR_ALG:
               {
                  int *sym = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_CONTOUR_TYPE typ = (V_CONTOUR_TYPE)*sym;
                  rval = chipsSys->set_contour_algorithm(cid, typ);
               }
               break;
               case MSG_SET_HST_LINE_STYL:
               {
                  int *style = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_LINE_TYPE typ = (V_LINE_TYPE)*style;
                  rval = chipsSys->set_histogram_linestyle(cid, typ);
               }
               break;
               case MSG_SET_HST_SYMBOL:
               {
                  int *sym = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_SYMBOL_TYPE typ = (V_SYMBOL_TYPE)*sym;
                  rval = chipsSys->set_histogram_symboltype(cid, typ);
               }
               break;
               case MSG_SET_HST_SYM_FILL:
                  rval = chipsSys->set_histogram_symbolfill(cid, true);
               break;
               case MSG_SET_HST_SYM_UNFIL:
                  rval = chipsSys->set_histogram_symbolfill(cid, false);
               break;
               case MSG_SET_HST_SYM_SIZE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_histogram_symbolsize(cid, (int)*ival);
               }
               break;
               case MSG_SET_HST_LINE_SIZE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_histogram_linethickness(cid, *fval);
               }
               break;
               case MSG_SET_HST_ERR_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_histogram_errcolor(cid, new_color);
               }
               break;
               case MSG_SET_HST_ERR_CPLEN:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_histogram_errcaplength(cid, (int)*ival);
               }
               break;
               case MSG_SET_HST_ERR_UP:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_histogram_errup(cid, *bval);
               }
               break;
               case MSG_SET_HST_ERR_DOWN:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_histogram_errdown(cid, *bval);
               }
               break;
               case MSG_SET_HST_DROPLINE:
               {
                  bool* bval = (bool*)&bf[NEXT_BOOL_POS(ndx)];
                  rval = chipsSys->set_histogram_dropline(cid, *bval);
               }
               break;
               case MSG_SET_HST_LINE_COLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_histogram_linecolor(cid, new_color);
               }
               break;
               case MSG_SET_HST_SYM_COLOR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_histogram_symbolcolor(cid, new_color);
               }
               break;
               case MSG_SET_HST_ERR_SIZE:
               {
                  float* fval = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  rval = chipsSys->set_histogram_errthickness(cid,(float)*fval);
               }
               break;
               case MSG_SET_HST_ERR_STYLE:
               {
                  string new_style = (char*)&bf[ndx];
                  rval = chipsSys->set_histogram_errstyle(cid, new_style);
               }
               break;
               case MSG_SET_HST_FILL_COLR:
               {
                  string new_color = (char*)&bf[ndx];
                  rval = chipsSys->set_histogram_fillcolor(cid, new_color);
               }
               break;
               case MSG_SET_HST_FILL_OPAC:
               {
                  double* dval = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_histogram_fillopacity(cid, *dval);
               }
               break;
               case MSG_SET_FRAMEMODE:
               {
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->set_framemode(cid, (int)*ival);
               }
               break;
               case MSG_TILED_FRAME:   // tile n x m
               {
                  unsigned int *rv, *cv;
   
                  cv = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);
                  rv = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->tile(cid, *cv, *rv);
               }
               break;
               case MSG_ROW_TILED_FRAME:   // tile frames using row priority
               {
                  unsigned int *rv;
   
                  rv = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->row_tile(cid, *rv);
               }
               break;
               case MSG_COL_TILED_FRAME:   // tile frames using column priority
               {
                  unsigned int *cv;
   
                  cv = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  rval = chipsSys->col_tile(cid, *cv);
               }
               break;
               case MSG_ADD_WINDOW_C:   // add window specifying dimensions
               {
                  float *height, *width;
                  char  *format, *winid;
   
                  height = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  width = (float*)&bf[NEXT_FLOAT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  format = &bf[ndx];
                  winid = &bf[ndx + strlen(format) + 1];
                  if ((rval = chipsSys->add_window(cid,*height,*width,format,
                       winid)) == ERR_STATUS_GOOD)
                     cid->reset_frame_id("");
               }
               break;
               case MSG_INFO_REQUEST:
               {
                  int *iv, *iob = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  iv = (int*)&bf[NEXT_INT_POS(ndx)];
                  V_OBJECT_TYPE obs = (V_OBJECT_TYPE)*iob;
                  char* cval = chipsSys->info(cid, obs, *iv);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
   
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
		     free(cval); 
                  }
               }
               break;
	       case MSG_INFO_COORD_REQUEST:
               {
                  char* cval = chipsSys->info_coordinate(cid);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
  
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
		     free(cval);
                  }
               }
               break;
               case MSG_INFO_DEPTH_REQUEST:
               {
                  int* dpth = (int*)&bf[NEXT_INT_POS(ndx)];
                  char* cval = chipsSys->info_depth(cid, *dpth);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
 
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                     free(cval);
                  }
               }
               break;
               case MSG_INFO_BOUND_REQUEST:
               {
                  char* cval = chipsSys->info_bound_axes(cid);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
 
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                     free(cval);
                  }
               }
               break;
               case MSG_CURRENT_REQUEST:
               {
                  char* cval = chipsSys->info_current();
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
   
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
		     free(cval);
                  }
               }
               break;
               case MSG_ADD_CLIENT_ERR:
               {
                  int *errcode = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rval = chipsSys->add_client_error(*errcode, &bf[ndx]);
               }
               break; 
	       case MSG_LOAD_FILL:
               {
                  int *slot = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rval = chipsSys->load_fill(&bf[ndx], *slot);
               }
               break;
	       case MSG_LOAD_CMAP_ARRAY:
	       {
		  double *rr, *gg, *bb, *alpha;
		  int *slot;
		  long *len;
		  //int ndx2; is this neccesary?
		  
		  //Length has to be before data - can't calc offsets otherwise
		  len = (long*)&bf[NEXT_INT_POS(ndx)];
		  ndx = NEXT_INT_POS(ndx) + sizeof(long);
		  
		  slot = (int*)&bf[NEXT_INT_POS(ndx)];
		  ndx = NEXT_INT_POS(ndx) + sizeof(int);
		  
		  rr = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*len)*sizeof(double);

		  gg = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*len)*sizeof(double);
		  
		  bb = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*len)*sizeof(double);
		  
                  alpha = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + (*len)*sizeof(double);

		  rval = chipsSys->load_colormap(*len, rr, gg, bb, alpha, *slot);
		  //ASDFQWER
	       }
	       break;
               case MSG_LOAD_CMAP:
               {
                  int *slot = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  rval = chipsSys->load_colormap(&bf[ndx], *slot);
               }
               break; 
               case MSG_GET_ERRS:
               {
                  int *mode = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  char* cval = chipsSys->get_errors(*mode);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
   
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival);
                     }
		     free(cval);
                  }
               }
               break;
               case MSG_GET_PREFERENCE:
               {
                  string pref = (char*)&bf[ndx];
                  char* cval = chipsSys->get_preference(pref);
                  rval = ERR_STATUS_BAD;
                  if (cval)
                  {
                     int ln = strlen(cval)/4 + 5;
                     int* ival;
   
                     if ((ival = (int*)malloc(ln*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_CHAR;
                        cp[3] = 0;
                        ival[3] = strlen(cval) + 1;
                        cp = (char*)&ival[4];
                        strcpy(cp, cval);
                        ival[1] = strlen(cp) + 1 + 4*sizeof(int);
                        cp = (char*)ival;
			tmpMessage.Set(cp,ival[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
		        free(ival); 
                     }
		     free(cval); 
                  }
               }
               break;
               case MSG_GET_PICK:
               {
                  unsigned int* uval;
                  int* ival = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  uval = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  int sc, *se = NULL;
                  double *sx = NULL, *sy = NULL;
                  string *sxstr = NULL, *systr = NULL;
                  char *sk = NULL;
                  int str_size = 0;
		  vector<CHIPS_PICK_EVT> pl;
                  rval = chipsSys->get_pick(cid, pl, *ival, *uval);
                  sc = pl.size ();
                  if ((rval == ERR_STATUS_GOOD) && (sc > 0))
                  {
		     sx = new double[sc]; 
		     sy = new double[sc]; 
                     sxstr = new string[sc];
                     systr = new string[sc];
		     se = new int[sc]; 
		     sk = new char[sc]; 

		     if (sx && sy && sxstr && systr && se && sk)
		     {
			int rr; 
		        for (rr = sc; rr--; )
			{
			   sx[rr] = pl[rr].x;  
			   sy[rr] = pl[rr].y;  
                           sxstr[rr] = pl[rr].x_str;
                           systr[rr] = pl[rr].y_str;
                           str_size += (pl[rr].y_str.length () + pl[rr].x_str.length () + 2);
			   se[rr] = pl[rr].event;  
			   sk[rr] = pl[rr].key;  
			}
                        pl.clear ();
		     }
		     else rval = ERR_STATUS_BAD; 
		  }
                  if (rval == ERR_STATUS_GOOD) 
		  {
                     if (sc > 0)
                     {
                        // pack up pick data 
                        int rr, *ival;
                        int sz = 15*sizeof(int) + 2*sc*sizeof(double) + sc*sizeof(int) +
                                 sc*sizeof(char) + str_size*sizeof(char);
		        if (sc%2)  sz+= sizeof(int);  //double align keys array 
                        // increase size for possible double alignment padding
                        if ((ival = (int*)malloc(sz*2)) != NULL) 
                        {
                           char* cp = (char*)&ival[2];
                           ival[0] = MSG_INFO_RESPONSE;
                           ival[1] = sz; 
                           short* sp = (short*)&ival[2];
                           double* dp = (double*)&ival[14];
                           sp[0] = 7;
                           cp[2] = MSG_DATATYPE_INT;
                           cp[3] = MSG_DATATYPE_DOUBLE;
                           cp[4] = MSG_DATATYPE_DOUBLE;
                           cp[5] = MSG_DATATYPE_INT;
                           cp[6] = MSG_DATATYPE_CHAR;
                           cp[7] = MSG_DATATYPE_CHAR;
                           cp[8] = MSG_DATATYPE_CHAR;
                           for (int ii=9; ii<12; ii++)
                              cp[ii] = 0;

                           ival[5] = 1;
                           ival[6] = sc; 
                           ival[7] = sc; 
                           ival[8] = sc; 
                           ival[9] = sc;
                           ival[10] = 0;
                           ival[11] = 0; 
                           ival[12] = sc;
                           ival[13] = 0;
                           for (rr = 0; rr < sc; rr++)   dp[rr] = sx[rr];
                           for (rr = 0; rr < sc; rr++)   dp[sc+rr] = sy[rr];
                           int* ip = (int*)&dp[2*sc];
                           for (rr = 0; rr < sc; rr++)   ip[rr] = se[rr];
                           char* kp = (char*)&ip[sc];
			   if ((sc % 2) == 1)
                           {
                              kp+= sizeof(int); // double align
                              ival[1] += sizeof(int);
                           }
                           for (rr = 0; rr < sc; rr++, kp++)   *kp = sk[rr];
                           if ((sc % 8) != 0)
                           {
                              kp+= (8-(sc%8))*sizeof(char); // double align
                              ival[1] += (8-(sc%8))*sizeof(char);
                           }
                           int strsize = 0;
                           for (rr = 0; rr < sc; rr++)   
                           {
                              strcpy (&kp[strsize], sxstr[rr].c_str ());
                              strsize += (sxstr[rr].length () + 1);
                              
                           }
                           kp += strsize;
                           ival[10] = strsize;
                           if ((strsize % 8) != 0)
                           {
                              kp+= (8-(strsize%8))*sizeof(char); // double align
                              ival[1] += (8-(strsize%8))*sizeof(char);
                           }
                           strsize = 0;
                           for (rr = 0; rr < sc; rr++)
                           {
                              strcpy (&kp[strsize], systr[rr].c_str ());
                              strsize += (systr[rr].length () + 1);
                           }
                           ival[11] = strsize;
			   cp = (char*)ival;
			   tmpMessage.Set(cp,ival[1]);
			   chipsServer.SendMessage(-1,currClientId,tmpMessage);
                           respond = 0;
                           free(ival);
                        }
		     }
		     else if (sc == 0)  
		     {
		        // pick was cancelled before events were selected
			int erval[2] = {0, 8};
			tmpMessage.Set((char*)erval,erval[1]);
			chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                     }
                  }
		  // cleanup any memory we allocated
                  if (sx) delete[] sx; 
                  if (sy) delete[] sy; 
                  if (se) delete[] se; 
                  if (sk) delete[] sk; 
               }
               break;
               case MSG_SET_AXIS_TICKPOS:
               {
                  double *ps;
                  unsigned int *cnt;
                  string* labs = NULL;
                  int rr; 

                  cnt = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ps =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  for (rr = *cnt; rr--; )
                     ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);
                  int *scnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*scnt > 0)
                  {
                     labs =  new string[*scnt];
                     if (labs)
                     {
                        int rr;
                        char* cp = &bf[ndx];
                        labs[0] = cp;
                        while(*cp != '\0') cp++;
                        for (rr = 1; rr < *scnt; rr++)
                        {
                           labs[rr] = ++cp;
                           while(*cp != '\0') cp++;
                        }
                     }
                  }
                  rval = chipsSys->set_arbitrary_tick_positions(cid, *cnt, ps,
                                                                 labs);
                  if (labs)  delete[] labs;
               }
               break;
               case MSG_SET_CNTR_LEVELS:
               {
                  double *ps;
                  unsigned int *cnt;
 
                  cnt = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  ps =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  rval = chipsSys->set_contour_levels(cid, *cnt, ps);
               }
               break;
               case MSG_ADD_CURVE_A:
               {
                  double *xx, *yy;
                  long *cnt;
                  int   ndx2;
                  char  *cv;
                  int *trans_exist;
                  Transform *trans=NULL;

                  // check for transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform,
                                                           serial_transform_size);
                     delete []serial_transform;
                  }
   
                  cv = &bf[ndx];
                  ndx += (strlen(cv) + 1);
		  cnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx2 = NEXT_DOUBLE_POS(ndx) + (*cnt)* sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];

                  rval = chipsSys->add_curve(cid, *cnt, xx, yy, trans, cv);
               }        
          
               break;
	    case MSG_ADD_CURVE_B:
               {
                  double *xx, *yy, *xu = NULL, *xd = NULL, *yu = NULL, 
                         *yd = NULL, *yl = NULL, *xl = NULL;
                  long *cnt;
                  char col_mask; 
                  int   ndx2;
                  char  *cv;
                  int *trans_exist;
                  Transform *trans=NULL;

                  // check for transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform,
                                                           serial_transform_size);
                     delete []serial_transform;
                  }

  
                  cv = &bf[ndx];
                  ndx += (strlen(cv) + 1);
                  col_mask = bf[NEXT_CHAR_POS(ndx)];
                  ndx = NEXT_CHAR_POS(ndx) + sizeof(char);
                  cnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx2 = NEXT_DOUBLE_POS(ndx) + (*cnt)* sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  if ((col_mask & 0x01) == 0x01) 
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     xu =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }
                  if ((col_mask & 0x02) == 0x02) 
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     xd =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }
                  if ((col_mask & 0x04) == 0x04) 
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     yu =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }
                  if ((col_mask & 0x08) == 0x08) 
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     yd =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }
                  if ((col_mask & 0x010) == 0x010)
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     yl =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }
                  if ((col_mask & 0x020) == 0x020)
                  {
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                     xl =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  }

                  rval = chipsSys->add_curve(cid, *cnt, xx, yy, xu, xd, 
                                              yu, yd, yl, xl, trans, cv);
               }
               break;
               case MSG_ADD_HISTOGRAM:
               {
                  double *counts, *lo = NULL, *hi = NULL, 
                         *yu = NULL, *yd = NULL;
                  long *cnt;
                  char col_mask;
                  int   ndx2;
                  char  *ch;
 
                  ch = &bf[ndx];
                  ndx += (strlen(ch) + 1);
                  col_mask = bf[NEXT_CHAR_POS(ndx)];
                  ndx = NEXT_CHAR_POS(ndx) + sizeof(char);
                  cnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  counts =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx2 = NEXT_DOUBLE_POS(ndx) + (*cnt)* sizeof(double);
                  if ((col_mask & 0x02) == 0x02)
                  {
                     lo =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  }
                  if ((col_mask & 0x04) == 0x04)
                  {
                     hi =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  }
                  if ((col_mask & 0x08) == 0x08)
                  {
                     yu =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  }
                  if ((col_mask & 0x10) == 0x10)
                  {
                     yd =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                     ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  }
                  rval = chipsSys->add_histogram(cid, *cnt, lo, hi,
                                                  counts, yu, yd, ch);
               }
               break;
               case MSG_ADD_CONTOUR:
               {
                  double *xx, *yy, *zz, *ll;
                  long *cnt;
                  unsigned int* levs;
                  int ndx2; 
                  char  *cv;
                  int *trans_exist;
                  Transform *trans = static_cast <Transform *>(0);
 
                  cv = &bf[ndx];
                  ndx += (strlen(cv) + 1);

                  // check for transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;
       
                     trans = transform_factory.deserialize (serial_transform, 
                                                           serial_transform_size);
                     delete []serial_transform;
                  }

                  cnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  levs = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx2 = ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);

                  if ((*levs) != 0)
                  {
                     ll =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx2 = NEXT_DOUBLE_POS(ndx) + (*levs)* sizeof(double);
                  }
                  else ll = NULL; 

                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  yy =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  ndx2 = NEXT_DOUBLE_POS(ndx2) + (*cnt)* sizeof(double);
                  zz =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
                  rval = chipsSys->add_contour(cid, *cnt, xx, yy, zz, *levs,
                                                ll, trans, cv);
                  if (trans)
                     delete trans;
               }
               break;
               case MSG_ADD_CONTOUR_B:
               {
                  double *xx, *ll;
                  long *xcnt, *ycnt;
                  unsigned int* levs;
                  int ndx2; 
                  char  *cv;
                  int *trans_exist;
                  Transform *trans = static_cast <Transform *>(0);

                  cv = &bf[ndx];
                  ndx += (strlen(cv) + 1);

                  // find transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform, 
                                                           serial_transform_size);

                     delete []serial_transform;

                  }
 
                  // x,y counts 
                  xcnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  ycnt = (long*)&bf[NEXT_LONG_POS(ndx)];
                  ndx = NEXT_LONG_POS(ndx) + sizeof(long);
                  levs = (unsigned int*)&bf[NEXT_INT_POS(ndx)];
                  ndx2 = ndx = NEXT_INT_POS(ndx) + sizeof(unsigned int);
                  if ((*levs) != 0)
                  {
                     ll =  (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx2 = NEXT_DOUBLE_POS(ndx) + (*levs)* sizeof(double);
                  }
                  else ll = NULL; 
                  xx =  (double*)&bf[NEXT_DOUBLE_POS(ndx2)];
		  long cnt = *xcnt * *ycnt; 
                  ndx2 = NEXT_DOUBLE_POS(ndx2) + (cnt)* sizeof(double);
                  rval = chipsSys->add_contour(cid, xx, *xcnt, *ycnt,
                                                *levs, ll, trans, cv);
                  if (trans)
                     delete trans;
               }
               break;
  
               case MSG_ADD_IMAGE_C:
               {
                  int *xaxis, *yaxis;
                  double *dat1 = NULL;
                  double *dat2 = NULL;
                  double *dat3 = NULL;
                  double *dat4 = NULL;
                  char *iid;
                  int *trans_exist;
                  Transform *trans = static_cast <Transform *>(0);

                  int *image_cnt;
   
                  iid = &bf[ndx];
                  ndx += (strlen(iid) + 1);

                  // find transform
                  trans_exist = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  if (*trans_exist)
                  {
                     ChipsTransformFactory transform_factory;
                     unsigned char *serial_transform;
                     int serial_transform_size = *trans_exist;
                     serial_transform = new unsigned char[serial_transform_size];
                     memcpy (serial_transform, &bf[ndx], serial_transform_size);
                     ndx += serial_transform_size;

                     trans = transform_factory.deserialize (serial_transform,
                                                           serial_transform_size);

                     delete []serial_transform;

                  }


                  xaxis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  yaxis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  image_cnt = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);

                  if ((*image_cnt & 1) == 1)
                  {
                     dat1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double)*(*xaxis)*(*yaxis);
                  }

                  if ((*image_cnt & 2) == 2)
                  {
                     dat2 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double)*(*xaxis)*(*
yaxis);
                  }

                  if ((*image_cnt & 4) == 4)
                  {
                     dat3 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double)*(*xaxis)*(*
yaxis);
                  }

                  if ((*image_cnt & 8) == 8)
                  {
                     dat4 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                     ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double)*(*xaxis)*(*
yaxis);
                  }
                  rval = chipsSys->add_image(cid, dat1, dat2, dat3, dat4, *xaxis, *yaxis, trans, iid);
                  if (trans)
                     delete trans;
               }
               break;
               case MSG_ADD_IMAGE_CHANNEL:
               {
                  int *xaxis, *yaxis;
                  int *channel;
                  double *dat = NULL;

                  xaxis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  yaxis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  channel = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);

                  dat = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double)*(*xaxis)*(*yaxis);
                  rval = chipsSys->set_image_channel(cid, dat, *xaxis, *yaxis, *channel);
               }
               break;
/* DEPRECATED
	       case MSG_ADD_TRANSFORM:
               {
                  int *row, *col;
		  double *dat;
                  double **matrix;
                  char *iid;

                  iid = &bf[ndx];
                  ndx += (strlen(iid) + 1);
                  row = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  col = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  dat = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
		  matrix = new double* [*row];
		  for (int ii=0; ii<*row; ii++){
		     matrix[ii] = new double[*col];
		     for (int jj=0; jj<*col; jj++)
		        matrix[ii][jj] = dat[ii*(*col)+jj];
		  }
                  rval = chipsSys->add_data_transform(*row, *col, matrix, iid);
		  for (int ii=0; ii<*row; ii++)
		     delete [](matrix[ii]);
		  delete [] (matrix);
               }
               break;
	       case MSG_REMOVE_TRANSFORM:
               {
                  char *iid;

                  iid = &bf[ndx];
                  ndx += (strlen(iid) + 1);
		  rval = chipsSys->remove_data_transform(iid);
	       }
	       break;
	       case MSG_APPLY_TRANS_A:
               {
                  char *iid;

                  iid = &bf[ndx];
                  rval = chipsSys->apply_data_transform(cid, iid);
               }
               break;
	       case MSG_APPLY_TRANS_B:
               {
                  int *row, *col;
                  double *dat;
                  double **matrix;
                  char *iid;

                  iid = &bf[ndx];
                  ndx += (strlen(iid) + 1);
                  row = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  col = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  dat = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  matrix = new double* [*row];
                  for (int ii=0; ii<*row; ii++){
                     matrix[ii] = new double[*col];
                     for (int jj=0; jj<*col; jj++)
                        matrix[ii][jj] = dat[ii*(*col)+jj];
                  }
                  rval = chipsSys->apply_data_transform(cid, *row, *col, matrix, iid);
                  for (int ii=0; ii<*row; ii++)
                     delete [](matrix[ii]);
                  delete [] (matrix);
               }
               break;
	       case MSG_INVERT_TRANS:
	       {
                  rval = chipsSys->invert_data_transform(cid);
	       }
	       break;
*/
	       case MSG_CONVERT_COORD_A:
	       {
		  int *count, *coord_sys, *axis;
                  double *dat1, *dat2;
                  char *win, *frm, *lyr, *coord_id;
		  ChipsId cid2;

                  win = &bf[ndx];
                  ndx += (strlen(win) + 1);
	          frm = &bf[ndx];
                  ndx += (strlen(frm) + 1);
		  lyr = &bf[ndx];
                  ndx += (strlen(lyr) + 1);
		  coord_id = &bf[ndx];
                  ndx += (strlen(coord_id) + 1);

                  count = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  coord_sys = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
		  axis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);

                  dat1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double) * *count;

	          cid2.set_window_id (win);
		  cid2.set_frame_id (frm);
		  cid2.set_coord_sys (static_cast <COORD_SYS>(*coord_sys));
		  cid2.set_coord_id (coord_id);

		  dat2 = new double [*count];

		  if ((rval = chipsSys->convert_coordinate (dat1, dat2, *count,
					 cid, &cid2, *axis)) == ERR_STATUS_GOOD)

                  {
		     int *ival;
		     int arr_size = 4 + 2*(*count);
                     if ((ival = (int*)malloc(arr_size*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = 2;
                        double* dp = (double*)&ival[4];
			for (int ii =0; ii<*count; ii++)
                           dp[ii] = dat2[ii];
                        ival[1] = arr_size*sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }

		  delete []dat2;

	       }
	       break;
	       case MSG_CONVERT_COORD_B:
	       {
		  int *coord_sys, *axis;
                  double *dat1;
	          double resp;
                  char *win, *frm, *lyr, *coord_id;
                  ChipsId cid2;

                  win = &bf[ndx];
                  ndx += (strlen(win) + 1);
                  frm = &bf[ndx];
                  ndx += (strlen(frm) + 1);
                  lyr = &bf[ndx];
                  ndx += (strlen(lyr) + 1);
                  coord_id = &bf[ndx];
                  ndx += (strlen(coord_id) + 1);

                  coord_sys = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);
                  axis = (int*)&bf[NEXT_INT_POS(ndx)];
                  ndx = NEXT_INT_POS(ndx) + sizeof(int);

                  dat1 = (double*)&bf[NEXT_DOUBLE_POS(ndx)];
                  ndx = NEXT_DOUBLE_POS(ndx) + sizeof(double);

                  cid2.set_window_id (win);
                  cid2.set_frame_id (frm);
                  cid2.set_coord_sys (static_cast <COORD_SYS>(*coord_sys));
                  cid2.set_coord_id (coord_id);

                  if ((rval = chipsSys->convert_coordinate (*dat1, resp, cid,
                                              &cid2, *axis)) == ERR_STATUS_GOOD)
                  {
		     int *ival;
		     if ((ival = (int*)malloc(6*sizeof(int))) != NULL)
                     {
                        char* cp = (char*)&ival[2];
                        ival[0] = MSG_INFO_RESPONSE;
                        short* sp = (short*)&ival[2];
                        sp[0] = 1;
                        cp[2] = MSG_DATATYPE_DOUBLE;
                        cp[3] = 0;
                        ival[3] = 2;
                        double* dp = (double*)&ival[4];
                        dp[0] = resp;
                        ival[1] = 6 *sizeof(int);
                        cp = (char*)ival;
                        tmpMessage.Set(cp,ival[1]);
                        chipsServer.SendMessage(-1,currClientId,tmpMessage);
                        respond = 0;
                        free(ival);
                     }
                  }
               }
	       break;
#ifndef HEADLESS
               case MSG_SHOW_GUI:
               {
                  string wd = &bf[ndx]; 
                  if (chipsGui == NULL)
                  {
                     chipsGui = new ChipsGuiMain(me, chipsSys);
                  }
                  chipsGui->show_dialog(wd);
               }
               break;
               case MSG_HIDE_GUI:
                  if (chipsGui != NULL)    chipsGui->hide_dialog();
               break;
#endif
               default:
                  rval = ERR_MESSAGE_UNKNOWN;
               break;
            }
         }
         else   rval =  ERR_SERVER_UNPACK;
   
         if (overflow)
         {
            free(bf);   // cleanup any allocated buffer
         }
      };

      // check if we need to send a reponse back to the client
      if (respond)
      {
         if (rval != ERR_STATUS_GOOD)
         {
            int* ip = (int*)rp;
            ip[0] = rval;
            ip[1] = 8;
	    tmpMessage.Set(rp,8);
	    chipsServer.SendMessage(-1,currClientId,tmpMessage);
         }
         else
         {
            int* lp = (int*)buffer;
            char* cp = (char*)buffer;
            lp++;
            unsigned int ln;
            pack_it(cid, rval);
            ln = *lp;
	    tmpMessage.Set(cp,ln);
	    chipsServer.SendMessage(-1,currClientId,tmpMessage);
         }
      }
      delete cid;
      timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, me);
      return 1;
}


// pack message id and ChipsId object
int ChipsServer::pack_it(ChipsId* cid,  int msgid)
{
   int *intp, *msg_sz, idlen, pos;
   char* cp = (char*)buffer; 
   string id;
   int *csys;

   if (!cid) return ERR_NO_ID_OBJ; 

   // set message id and storage for size of message
   intp = (int*)buffer;
   *intp = msgid;
   msg_sz = ++intp;
   pos = 2*sizeof(int);

   // set window id 
   id = cid->get_window_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set frame id 
   id = cid->get_frame_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set plot id 
   id = cid->get_plot_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set image id 
   id = cid->get_image_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set coord sys id 
   id = cid->get_coord_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set curve id 
   id = cid->get_curve_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set contour id
   id = cid->get_contour_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set histogram id
   id = cid->get_histogram_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set axis id 
   id = cid->get_axis_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set xaxis id 
   id = cid->get_xaxis_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set yaxis id 
   id = cid->get_yaxis_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set lbid id 
   id = cid->get_label_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set lnid id 
   id = cid->get_line_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set ptid id 
   id = cid->get_point_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set region id 
   id = cid->get_region_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   // set colorbar id
   id = cid->get_colorbar_id();
   if ((idlen = id.length()) != 0)
   {
      strcpy(&cp[pos], id.c_str());
      pos += idlen;
   }
   cp[pos++] = '\0';
   csys = (int*)&cp[NEXT_INT_POS(pos)];
   *csys = (int)cid->get_coord_sys ();
   pos = NEXT_INT_POS(pos) + sizeof(int);
   // axis dim
   csys = (int*)&cp[NEXT_INT_POS(pos)];
   *csys = (int)cid->get_axis_dim ();
   pos = NEXT_INT_POS(pos) + sizeof(int);

   *msg_sz = pos;   // set message size

   return ERR_STATUS_GOOD; 
}


int ChipsServer::unpack_it(const char* buf, ChipsId* cid)
{
   int rval = 0;   // return current index in char* buffer

   if (buf && cid)
   {
      const char* val = buf;
      int ln, pos = 0;
      int *ival;

      if ((ln=strlen(&val[pos]))!=0) {cid->set_window_id(&val[pos]);pos+=ln;}
      else  cid->set_window_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_frame_id(&val[pos]);pos+=ln;}
      else  cid->set_frame_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_plot_id(&val[pos]);pos+=ln;}
      else  cid->set_plot_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_image_id(&val[pos]);pos+=ln;}
      else  cid->set_image_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_coord_id(&val[pos]);pos+=ln;}
      else  cid->set_coord_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_curve_id(&val[pos]);pos+=ln;}
      else  cid->set_curve_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_contour_id(&val[pos]);pos+=ln;}
      else  cid->set_contour_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_histogram_id(&val[pos]);pos+=ln;}
      else  cid->set_histogram_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_axis_id(&val[pos]);pos+=ln;}
      else  cid->set_axis_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_xaxis_id(&val[pos]);pos+=ln;}
      else  cid->set_xaxis_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_yaxis_id(&val[pos]);pos+=ln;}
      else  cid->set_yaxis_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_label_id(&val[pos]);pos+=ln;}
      else  cid->set_label_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_line_id(&val[pos]);pos+=ln;}
      else  cid->set_line_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_point_id(&val[pos]);pos+=ln;}
      else  cid->set_point_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_region_id(&val[pos]);pos+=ln;}
      else  cid->set_region_id("");
      pos++;
      if ((ln=strlen(&val[pos]))!=0) {cid->set_colorbar_id(&val[pos]);pos+=ln;}
      else  cid->set_colorbar_id("");
      pos++;
      ival = (int*)&val[NEXT_INT_POS(pos)];
      cid->set_coord_sys ((COORD_SYS)(*ival));
      pos = NEXT_INT_POS(pos) + sizeof(int);
      // axis dim
      ival = (int*)&val[NEXT_INT_POS(pos)];
      cid->set_axis_dim((*ival));
      pos = NEXT_INT_POS(pos) + sizeof(int);
      rval = pos; 
   }
   return rval; 
}




// method to create a server process 
int ChipsServer::create_server(const string& inServerId)
{

   if (!serverId.empty())  return ERR_SOCK_CONN;

   serverId=inServerId;
   string tmpServerSocket="/tmp/.chips";

   errno = 0; 
   struct passwd *pw = getpwuid(getuid());
   if (pw && pw->pw_name[0] != '\0')
   {
      tmpServerSocket.append(".");
      tmpServerSocket.append(pw->pw_name);
   }


   // check that the temp directory exists- create it if not
   if (access(tmpServerSocket.c_str(), W_OK) != 0)
   {
      mkdir(tmpServerSocket.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
      if (access(tmpServerSocket.c_str(), W_OK) != 0)  tmpServerSocket = "/tmp";
   }
   tmpServerSocket.append("/");

   tmpServerSocket+=inServerId;

      
   setsid();
   chdir("/");
   umask(0);


   if ( chipsServer.Start(10,tmpServerSocket) <=0 )
   {
      cout<<"chips ERROR: FAILED TO START SERVER"<<endl;
      if (chipsSys){
         delete chipsSys;
         chipsSys = NULL;
      }
      exit(-1);
   }

   // server process 
   int ac = 1;
   char *a0  =  (char*)"ChipsServer";
   char *a1  =  (char*)"\0";
   char **argv;
   argv = (char**)malloc(2*sizeof(char*));
   argv[0] = a0;
   argv[1] = a1;
	   
   handle_server_signals(); 

#ifndef HEADLESS	   

   if (!gtk_init_check(&ac, &argv))
   {
      char cmd[256];
      sprintf(cmd,"sh chips_xvfb.sh %d", getpid());
      FILE *fd;
      if ((fd = popen(cmd,"r")) != NULL)
      {
         char buf[256]; 
         memset(buf,0,sizeof(buf));
         fgets(buf, sizeof(buf), fd);
         pclose(fd);
         xvfb_pid = buf;
      }
      if (xvfb_pid.length())
      {
         string clean = "ps "+xvfb_pid+"> /dev/null 2>&1 && kill "+xvfb_pid;
         xvfb_pid = clean;
         sprintf(cmd,"DISPLAY=:%d", getpid());
         putenv(cmd);
         if (!gtk_init_check(&ac, &argv))  
         {
            cout << "chips ERROR: Failed to open DISPLAY or Xvfb session- exiting chipsServer"<< endl;
            clean_xvfb();
            exit(1);
         }
      }
      else 
      {
         cout << "chips ERROR: Failed to open DISPLAY or Xvfb session- exiting chipsServer"<< endl;
         exit(1);
      }
   }

   // load gtkrc files
   string rclocal, rcname = getenv("HOME");
   if (rcname.length()) rcname.append("/");
      rcname.append(".gtkrc");
   rclocal = rcname + ".chips";
   if (access(rclocal.c_str(), R_OK) == 0)
      gtk_rc_parse(rclocal.c_str());
   else
   {
      rclocal = rcname + ".ciao";
      if (access(rclocal.c_str(), R_OK) == 0)
         gtk_rc_parse(rclocal.c_str());
      else
         gtk_rc_parse(rcname.c_str());
   }
	   
   // initialize gtkglext lib to allow use of private colormap
   // via GTK_GL_INSTALL_COLORMAP env var 
   gtk_gl_init (&ac, &argv);
#endif
	   
   free(argv); 
   timeoutId=gtk_timeout_add(timeoutMs,&ChipsServer::server_process, this);

   // instantiate a copy of the gui now so we can handle gui related events
#ifndef HEADLESS
   chipsGui = new ChipsGuiMain(this, chipsSys);
#endif
	   
   gtk_main();

   return (ERR_STATUS_GOOD);
}

int ChipsServer::remove_gui()
{
//   if (chipsGui && (chipsGui->get_visible() == 0))
#ifndef HEADLESS
   if (chipsGui) 
   {
      //chipsGui->disconnect_from_server();
      delete chipsGui;
      chipsGui = NULL;
   }
#endif
   return (ERR_STATUS_GOOD);
}

void ChipsServer::lock(long myid)
{
   if (lockFlag == 0)
   {
      lockFlag++;
      lockClient = myid;
   }
   else if (lockClient == myid)
   {
      // nested lock
      lockFlag++;
   }
}

void ChipsServer::unlock(long myid)
{
   if (lockFlag > 0 && lockClient == myid)
   {
      lockFlag--;
   }
   if (lockFlag == 0)  lockClient = -1;
}

int ChipsServer::is_locked()
{
   return (lockFlag>0);
}


