// 
//  Copyright (C) 2004-2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "ChipsWindow.hh"
#include "ChipsDefaults.hh"
#include "ChipsEngine.hh"
#include "ChipsCmdManager.hh"
#include "ChipsWindowSystem.hh"
#include "ChipsDeleteCommand.hh"
#include "ChipsZoomCommand.hh"
#include "ChipsPanCommand.hh"
#include "ChipsFontList.hh"

#include <stdio.h>
#include <string>
#include <sstream>
#include <unistd.h>
#include <algorithm>

#ifndef HEADLESS
#include "ChipsGtkWindow.hh"
#include "gtk/gtk.h"
#endif

extern ChipsDefaults* theDefaults; 
extern ChipsCmdManager* cmd_manager;
extern ChipsFontList* theFonts;

// comparator for frames
bool comp_frame_order (list<ChipsFrame *>::const_reference x,
                       list<ChipsFrame *>::const_reference y)
  { if (!x) return true;
     else if (!y) return false;
    return (x->get_frame_num()<y->get_frame_num ()); 
  }




ChipsWindow::ChipsWindow(ChipsEngine* en, string val):ChipsObject(NULL, val)
{
   id = val;
   eng = en;
   initialize();
   set_window_size();  // set up window size based on defaults 
}

ChipsWindow::ChipsWindow(ChipsEngine* en, ChipsStateStream& in):ChipsObject(NULL, "")
{
   eng = en;
   initialize(false);
   deserialize(in);
   set_window_size();  // set up window size based on defaults 
}


ChipsWindow::ChipsWindow(ChipsEngine* en, float ww, float hh, string format, 
   string val):ChipsObject(NULL, val)
{
   id = val;
   eng = en;
   initialize();
   
   // adjust height 
   raw_width = ww;
   raw_height = hh;
 
   // setup units 
   if (format.length())  
      string_to_v_units(format.c_str(), &units);
   set_window_size();
}


ChipsWindow::~ChipsWindow()
{
#ifndef HEADLESS
   delete gui_window;
#endif
}


void ChipsWindow::object_delete ()
{
   ChipsId cid;
   cid.set_frame_id("all");
   delete_frame(&cid, true);

   frame_cnt = 0;

   delete coord;
   delete (this);
}


void ChipsWindow::initialize(bool create_coord)
{
   activated = drawready = print_mode = false; 
   frame_cnt = frame_ord = 0;
   string tmp_color, tmp_geom = "";
   current = "";
   framemode = timer = duration = 0;
   prefix = ""; 
   title = id;
   focus = smoothing = export_clobber = false;
   export_clobber = false;
   export_font = true; 
   export_scheme = chips_rgb; 
   tile_rows = tile_cols = 1; 
   if (theDefaults)
   {
      string temp; 
      prefix = theDefaults->get_windowprefix(); 
      focus = theDefaults->get_windowfocus(); 
      tmp_geom = theDefaults->get_windowgeom(); 
      redraw_on = theDefaults->get_windowdisplay(); 
      raw_width = theDefaults->get_windowwidth();
      raw_height = theDefaults->get_windowheight();
      scale_width = theDefaults->get_windowscalewidth ();
      scale_height = theDefaults->get_windowscaleheight ();
      smoothing = theDefaults->get_windowsmoothing();
      update_grid_properties();
      vis_layout = theDefaults->show_layout();
      layout_thickness = theDefaults->get_layoutthickness();
      layout_linestyle = theDefaults->get_layoutlinestyle();
      temp = theDefaults->get_layoutcolor();
      if (set_layout_color(temp))  set_layout_color("yellow");
      units = theDefaults->get_windowunits(); 
      tmp_color = theDefaults->get_windowcolor();
      set_color(tmp_color);
      tmp_color = theDefaults->get_defaultfgdisp();
      if (color.valid_color(tmp_color) == ERR_STATUS_GOOD)  
         fg_display = tmp_color; 
      else
         fg_display = "white"; 
      tmp_color = theDefaults->get_defaultbgdisp();
      if (color.valid_color(tmp_color) == ERR_STATUS_GOOD)  
         bg_display = tmp_color; 
      else
         bg_display = "black"; 
      export_clobber = theDefaults->get_exportclobber(); 
      export_font = theDefaults->get_exportfonts(); 
      export_orient = theDefaults->get_exportorient(); 
      export_scheme = theDefaults->get_exportcolorscheme(); 
      leftmargin = theDefaults->get_pageleftmargin();
      rightmargin = theDefaults->get_pagerightmargin();
      topmargin = theDefaults->get_pagetopmargin();
      bottommargin = theDefaults->get_pagebottommargin();
      page_width = theDefaults->get_pagewidth();
      page_height = theDefaults->get_pageheight();
      page_units = theDefaults->get_pageunits();
      if (page_units == chips_nounit)
	 page_units = units;
      page_scale_width = theDefaults->get_pagescalewidth ();
      page_scale_height = theDefaults->get_pagescaleheight ();
      fit_to_page = theDefaults->get_fittopage();
      keep_aspect = theDefaults->get_pageaspect();
      export_dpi = theDefaults->get_exportdpi ();
      printmethod = theDefaults->get_printmethod ();
   }
   else
   {
      redraw_on = true; 
      raw_height = raw_width = V_DEF_WINDOW_DIM;
      scale_height = scale_width = V_DEF_WINDOW_SCALE;
      units = chips_pixels; 
      vis_layout = vis_grid = false;
      set_grid_color("blue");
      grid_linestyle = chips_longdash;
      layout_linestyle = chips_shortdash;
      gscale = V_GRID_SCALE_HALF_INCH;
      layout_thickness = grid_thickness = 1;
      set_layout_color("yellow");
      set_color("white");
      fg_display = "white"; 
      bg_display = "black";
      leftmargin = rightmargin = topmargin = bottommargin = 0;
      page_width = page_height = 0;
      page_units = chips_pixels;
      page_scale_width = 0;
      page_scale_height = 0;
      fit_to_page = false;
      keep_aspect = true; 
      export_dpi = 72;
      printmethod = "framebuffer";
   }
   object_type = chips_window;
   color.set_default_foreground(fg_display); 
   color.set_default_background(bg_display); 
   layout_linestyle = chips_solid;
   visible = redraw_on;
   if (create_coord)
   {
      coord = new ChipsWindowSystem (eng); 
      data_system_id = coord->get_id ();
   }
      

   initialize_gui(tmp_geom);
}

void ChipsWindow::initialize_gui(string geom)
{
#ifndef HEADLESS
   gui_window = this->new_gui_window (geom, focus);
   this->set_title (title);
#endif
}


void ChipsWindow::set_window_size()
{
   float conv = conversion_factor (units); 

   width = (int)(raw_width * conv * scale_width);
   height = (int)(raw_height * conv * scale_height);
   if ((width < 50) || (width > 2000))   width = (int)(V_DEF_WINDOW_DIM*scale_width);
   if ((height < 50) || (height > 2000))   height = (int)(V_DEF_WINDOW_DIM*scale_height);
   ((ChipsWindowSystem *)coord)->set_dim ((int)(raw_width*conv), (int)(raw_height*conv));
   ((ChipsWindowSystem *)coord)->set_scale (scale_width, scale_height);

#ifndef HEADLESS
   gui_window->set_default_size (width, height);
#endif

   if (activated)
      this->render_grids ();
}



void ChipsWindow::set_window_size(int inPixelWidth,int inPixelHeight)
{
   float conv = conversion_factor (units);
   width=inPixelWidth;
   height=inPixelHeight;

   raw_width = static_cast<float>(width / (conv * scale_width));
   raw_height = static_cast<float>(height / (conv * scale_height));

   ((ChipsWindowSystem *)coord)->set_dim ((int)(raw_width*conv), (int)(raw_height*conv));

   if (activated)
      this->render_grids ();

};

float ChipsWindow::conversion_factor(V_UNITS_TYPE &val)
{
  float conv;

   switch(val)
   {
      case chips_cm:   conv = V_CM2PIX_CONV;   break;
      case chips_mm:   conv = V_MM2PIX_CONV;   break;
      case chips_inch: conv = V_INCH2PIX_CONV; break;
      case chips_pixels: conv = 1.0;           break;
      default:
         val = chips_pixels;
         conv = 1.0;   // use pixels for unknown formats;
      break;
   }
   return conv;
};

void ChipsWindow::clear() 
{
   ChipsId cid;
   cid.set_frame_id("all");
   delete_frame(&cid);
   vis_grid = vis_layout = false; 
}


int ChipsWindow::set_title(string val)
{
#ifndef HEADLESS
   title = val; 
  
   string my_txt = prefix;
   my_txt.append(val); 
   gui_window->set_title(my_txt);
#endif
   return ERR_STATUS_GOOD; 
}

int ChipsWindow::set_title_suffix(string val)
{
#ifndef HEADLESS
   string my_txt = prefix + title + val;
   gui_window->set_title(my_txt);
#endif
   return ERR_STATUS_GOOD;
}



int ChipsWindow::set_current(const string &val)
{
   string prev_curr_frame = current;
   map<string, ChipsFrame*>::iterator ii;
   int rval = ERR_STATUS_GOOD;


   if (chips_strcmp(val.c_str(), "all") == 0)
   {
      current = "all"; 
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         ii->second->enable_picklimits (true);

      rval = ERR_STATUS_GOOD; 
   }
   else if (frame.count(val))
   {
      current = val; 
      frame[val]->enable_picklimits (true);
      rval = ERR_STATUS_GOOD; 
   } 
   else if (val.length () == 0)
   {
      current = val;
      rval = ERR_STATUS_GOOD;
   }
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", val.c_str());

   if (!rval && (prev_curr_frame != current))
   {
      if (chips_strcmp(prev_curr_frame.c_str(), "all") == 0)
      {
         for(ii=frame.begin(); ii != frame.end(); ++ii)
            if (ii->first != current)
               ii->second->enable_picklimits (true);
      }
      else if (frame.count (prev_curr_frame))
         frame[prev_curr_frame]->enable_picklimits (false);
   }
   if (rval == ERR_STATUS_GOOD)  cmd_manager->update_mask(STATE_CHNG_CURRENCY); 
   return rval;

}


// highlight objects
int ChipsWindow::highlight_object(ChipsId* cid, V_OBJECT_TYPE obj, bool vis)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();
   ChipsFrame* tmp = NULL; 
   map<string, ChipsFrame*>::iterator ii;

   // can't hightlight during  blink
   if (duration)
   {
      cmd_manager->add_err(rval=ERR_BLINK_ACTIVE,id);
      return rval;
   }
   // make sure we have a frame to work on 
   if (!tmpid.length()) tmpid = current;
   if (((tmp = get_frame(tmpid)) == NULL) && (chips_strcmp(tmpid.c_str(), "all") != 0))
   {
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"frame",tmpid.c_str());
      return rval;  
   } 

   if (tmp)
   {
      // single frame 
      if (obj == chips_frame)
      {
         if (vis)  tmp->force_show(); 
            else   tmp->force_hide(); 
      }
      else
         rval = tmp->highlight_object(cid, obj, vis);  
   }
   else 
   {
      // highlight all frames
      if (obj == chips_frame)
      {
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            if (vis)  ((*ii).second)->force_show(); 
               else   ((*ii).second)->force_hide(); 
         }
      }
      else 
      {
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            ((*ii).second)->highlight_object(cid, obj, vis);  
         }
      }
   }
   redraw();
   return rval;
}


// window level method of moving objects
int ChipsWindow::move_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj,
   double xx, double yy, int mode)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();

   if (!tmpid.length()) tmpid = current;

   if (obj == chips_frame)
   {
      COORD_SYS cd_sys = cid->get_coord_sys ();

      if (cd_sys == FRAME_NORM)
         cmd_manager->add_err(rval=ERR_INV_MOVE,"",0,"frame",
                                 "frame normalized");
      else if (cd_sys == DATA)
         cmd_manager->add_err(rval=ERR_INV_MOVE,"",0,"frame", "data");
      else if (cd_sys == PLOT_NORM)
         cmd_manager->add_err(rval=ERR_INV_MOVE,"",0,"frame",
                                 "plot normalized");

   }

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         // move all frames
         if (obj == chips_frame)
         {

            if (!frame.empty())
            {
               map<string, ChipsFrame*>::iterator ii;
               COORD_SYS cd_sys = cid->get_coord_sys ();
               string cd_id = cid->get_coord_id ();

               for(ii=frame.begin(); ii != frame.end(); ++ii)
               {
                  if ((sts=(*ii).second->move(cd_sys,cd_id,xx,yy,mode)) ==
                      ERR_STATUS_GOOD)
                     pass++;
                  else if (!err_type_object_DNE(sts))
                     failval = sts;
               }
               cmd_manager->clear_errs(CLEAR_DNE_ERRS);
               if (failval)
                  rval = failval;
               else if (!pass)
                  cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
	       update_display();
            }
            else
               cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"frames");
         }
         else
         {
            map<string, ChipsFrame*>::iterator ii;
            for(ii=frame.begin(); ii != frame.end(); ++ii)
            {
               if ((sts = (*ii).second->move_object(cid, obj, subobj, xx, yy, mode)) ==
                      ERR_STATUS_GOOD)
                  pass++;
               else if (!err_type_object_DNE(sts))
                  failval = sts;
            }
            cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
               rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
         }
      }
      else
      {
         // move this frame
         ChipsFrame* tmp = get_frame(tmpid);

         if (tmp)
            if (obj == chips_frame)
            {
               COORD_SYS cd_sys = cid->get_coord_sys ();
               string cd_id = cid->get_coord_id ();
                  rval = tmp->move(cd_sys, cd_id, xx, yy, mode);
               if (rval == ERR_STATUS_GOOD)
               {
		  if (framemode != 2)
		     tmp->set_display_location(framemode, 0, 0, width, height);
		  else
		     update_display();
               }
             }
            else 
               rval = tmp->move_object(cid, obj, subobj, xx, yy, mode);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "frame", tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) set_current ( tmpid);
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::resize_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, double *bounds)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();
   ChipsVector lower_left;
   ChipsVector upper_right;
   ChipsFrame* cf;


   if (!tmpid.length()) tmpid = current;

   if (tmpid.length())
   {

      if (obj == chips_frame)
      {
         COORD_SYS cd_sys = cid->get_coord_sys ();
         string cd_id = cid->get_coord_id ();

         lower_left = ChipsVector (bounds[0], bounds[1]);
         upper_right = ChipsVector (bounds[2], bounds[3]);

         rval = coord->convert_cs_to_cs (&lower_left, &lower_left, 1,
                  cd_sys, cd_id, WINDOW_NORM, "");
         rval = coord->convert_cs_to_cs (&upper_right, &upper_right, 1,
                  cd_sys, cd_id, WINDOW_NORM, "");

         rval = this->reposition_frame (cid, lower_left[0], lower_left[1], 
                                   upper_right[0], upper_right[1]);

      }
      else if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsFrame*>::iterator ii;
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            cf = (*ii).second;
            if ((sts = cf->resize_object(cid, obj, subobj, bounds)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      }
      else 
      {
         // move this frame
         cf = get_frame(tmpid);

         if (cf)
            rval = cf->resize_object(cid, obj, subobj, bounds);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "frame", tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) set_current ( tmpid);
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::rotate_object(ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, float angle)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();
   ChipsFrame* cf;


   if (!tmpid.length()) tmpid = current;

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsFrame*>::iterator ii;
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            cf = (*ii).second;
            if ((sts = cf->rotate_object(cid, obj, subobj, angle)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      }
      else
      {
         // move this frame
         cf = get_frame(tmpid);
         if (cf)
            rval = cf->rotate_object(cid, obj, subobj, angle);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "frame", tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) set_current ( tmpid);
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::select_object(const ChipsId* cid, V_OBJECT_TYPE obj, V_SUBOBJECT_TYPE subobj, bool selected)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();
   ChipsFrame* cf;


   if (!tmpid.length()) tmpid = current;

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         int sts, pass = 0, failval = 0;
         map<string, ChipsFrame*>::iterator ii;
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            cf = (*ii).second;
            if (obj == chips_frame)
            {
               pass++;
               if (selected) cf->select ();
                 else cf->unselect ();
            }
            else if ((sts = cf->select_object(cid, obj, subobj, selected)) ==
                   ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
         }
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      }
      else
      {
         // move this frame
         cf = get_frame(tmpid);
         if (cf)
         {
            if (obj == chips_frame)
            {
               if (selected) cf->select ();
                 else cf->unselect ();
            }
            else
               rval = cf->select_object(cid, obj, subobj, selected);
         }
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, "", 0, "frame", tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) set_current ( tmpid);
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::get_annotation_position (const ChipsId* cid,
                                          V_OBJECT_TYPE obj, ChipsVector &pos)
{
   list<ChipsFrame*> cframes;
   ChipsFrame *cf = NULL;
   int  rval = ERR_STATUS_GOOD;

   if ((rval = get_frames (cid, cframes, false)) == ERR_STATUS_GOOD)
   {
      // run the command
      cf = *(cframes.begin());
      cf->get_annotation_position (cid, obj, pos);
   }

   return rval;

}

int ChipsWindow::change_annotation_coordinate (const ChipsId* cid,
                                          V_OBJECT_TYPE obj,
                                          COORD_SYS coordsys,
                                          const string& coordid)
{
   list<ChipsFrame*> cframes;
   ChipsFrame *cf = NULL;
   int  rval = ERR_STATUS_GOOD;

   if ((rval = get_frames (cid, cframes, true)) == ERR_STATUS_GOOD)
   {
      list<ChipsFrame*>::iterator ii;
      int rr, pass = 0, failval = 0;
      string fid = cid->get_frame_id ();

      // run the command
      for(ii=cframes.begin(); ii != cframes.end(); ++ii)
      {
         cf = *ii;
         rr = cf->change_annotation_coordinate(cid, obj, coordsys, coordid);
         if (rr == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(rr))
            failval = rr;
      }

      // check all passed
      if (frame.size () == cframes.size())
      {
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
         if (failval)
            rval = failval;
         else if (!pass)
            cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      }

      // update currency
      if (fid.length () > 0)
         current = fid;
   }

   return rval;
}



int ChipsWindow::get_object_count(ChipsId* cid, V_OBJECT_TYPE obj,
   int& out_val)
{
   int rval = ERR_STATUS_GOOD;

   if (!frame.empty())
   {
      string fid = cid->get_frame_id();
      if (fid.length() == 0)  fid = current;
      if (chips_strcmp(fid.c_str(), "all") != 0)
      {
         ChipsFrame* cf = get_frame(fid);   // get frame
         if (cf)
         {
            if (obj == chips_frame)
               out_val = 1;
            else
               cf->get_object_count(cid, obj, out_val);
         }
      }
      else
      {
         if (obj == chips_frame)
            out_val = frame.size();
         else
         {
            map<string, ChipsFrame*>::iterator ii;
            int cnt;
            for(ii=frame.begin(); ii != frame.end(); ++ii)
            {
               cnt = 0;
               (*ii).second->get_object_count(cid, obj, cnt);
               out_val += cnt;
            }
         }
      }
   }
   return rval;
}


int ChipsWindow::get_object_coordinfo(ChipsId* cid, V_OBJECT_TYPE obj,
   COORD_SYS& out_val, string& out_id)
{
   int rval = ERR_STATUS_GOOD; 
   if (!frame.empty())
   {
      string fid = cid->get_frame_id();
      if (fid.length() == 0)  fid = current;
      if ((chips_strcmp(fid.c_str(), "all") != 0) || (frame.size() == 1))
      {
         // get frame
         ChipsFrame* cf = get_frame(fid);
         if (cf)
         {
            if (obj == chips_frame)
            {
               out_val = cf->get_coord_system();
               out_id = (out_val == DATA) ? cf->get_data_system_id() : "";
               rval = ERR_STATUS_GOOD;
            }
            else
               rval = cf->get_object_coordinfo(cid, obj, out_val, out_id);
         }
         else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",fid.c_str());
      }
      else
         cmd_manager->add_err(rval = ERR_OBJ_INVALID_ALL,id,0,"frames");
   }
   else   cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"frames");
   return rval;
}

int ChipsWindow::get_object_depth_info (const ChipsId* cid, V_OBJECT_TYPE otype,
                            int &out_depth_num, int &out_shuffle_num)
{
   int rval = ERR_STATUS_GOOD;

   if (otype == chips_window)
   {
      out_depth_num = 0;
      out_shuffle_num = 0;
   }
   else if (!frame.empty())
   {
      string fid = cid->get_frame_id();
      if (fid.length() == 0)  fid = current;
      if ((chips_strcmp(fid.c_str(), "all") != 0) || (frame.size() == 1))
      {
         // get frame
         ChipsFrame* cf = get_frame(fid);
         if (cf)
            rval = cf->get_object_depth_info(cid, otype, out_depth_num, out_shuffle_num);
         else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",fid.c_str());
      }
      else
         cmd_manager->add_err(rval = ERR_OBJ_INVALID_ALL,id,0,"frames");
   }
   else   cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"frames");
   return rval;
}





// window level method of setting current objects
int ChipsWindow::set_current(ChipsId* cid, V_OBJECT_TYPE what, bool rec)
{
   int  rval = ERR_STATUS_GOOD;
   string tmpid = cid->get_frame_id();

   // if we are setting something other than frame and a layout spec
   // frame id was not provided, use current frame val 
   if ((tmpid.length() == 0) && (what != chips_frame))
      tmpid = current;

   if (tmpid.length())
   {
      if (chips_strcmp(tmpid.c_str(), "all") == 0)
      {
         if (what == chips_frame)
         {
            if (!frame.empty())
            {
               if (rec)
               {
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
                     current, tmpid, chips_frame);
                  if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
               }
               set_current ( "all");
            }
            else   cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"frames");
         }
         else
         {
            map<string, ChipsFrame*>::iterator ii;
            int sts, pass = 0, failval = 0;
            // if a set_current call passed for any entry within an 'all'
            // we don't generate an error. However, if no entries pass we
            // generate errors- either the specific errors or a DNE failure
            // if no entries exist
            for(ii=frame.begin(); ii != frame.end(); ++ii)
            {
               if ((sts=(*ii).second->set_current(cid,what,rec)) == ERR_STATUS_GOOD)
                  pass++;
               else if (!err_type_object_DNE(sts))
                     failval = sts;
            }
            cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
               rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
            else  
               set_current ( "all");
         }
      }
      else
      {
         ChipsFrame* tmp = get_frame(tmpid);

         if (tmp)
            if (what == chips_frame)
            {
               if (rec)
               {
                  ChipsCurrentCommand *cmd = new ChipsCurrentCommand(this,
                     current, tmpid, chips_frame);
                  if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
                  {
                     if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                     delete cmd;
                  }
               }
               set_current ( tmpid); 
            }
            else 
               rval = tmp->set_current(cid, what, rec);
         else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", tmpid.c_str());
      }
      if (rval == ERR_STATUS_GOOD) set_current ( tmpid);
   }
   else  // no id to set current to 
      cmd_manager->add_err(rval=ERR_NO_OBJ_ID, id, 0, "frame");
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}



int ChipsWindow::set_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = color.set_background_color(in_color)) == ERR_STATUS_GOOD)
      rval = change_color();
   return rval;
}


ChipsObject* ChipsWindow::get_object(const ChipsId* cid, V_OBJECT_TYPE otype, 
                                  bool include_deref)
{
   ChipsObject* obj = NULL;
   string fid = cid->get_frame_id();
   int obj_ref = cid->get_frame_ref_id ();
   if ((chips_strcmp(fid.c_str(), "all") != 0) || (frame.size() == 1))
   {
      ChipsFrame* fr = get_frame(fid);
      if (fr && (otype != chips_frame))
         obj = fr->get_object(cid, otype, include_deref);
      else if (fr)
      {
         // if obj_ref is zero then accept otherwise confirm the reference count
         if ((obj_ref == 0) || ((obj_ref > 0) && (fr->get_ref_cnt () == obj_ref)))
            obj = fr;
      }
   }

   if (include_deref && !obj)
   {
      ChipsFrame *fr = NULL;
      if (dframe.count (obj_ref) != 0)
         fr = dframe[obj_ref];
      if (fr && (otype != chips_frame))
         obj = fr->get_object(cid, otype, include_deref);
      else
         obj = fr;
   }

   return obj;
}

int ChipsWindow::add_frame(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf = NULL;
   string val = theDefaults->get_frameid(); 

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE; 
   }
   if (frame.count(val))  
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,val.c_str(), "frame");
   else 
      cf = new_frame(val);

   if (cf)
   {
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      // new frame created- add to list and update current

      if (!cmd_manager->cmd_lock ()){
         // if val is empty that means the frame_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(cf,(val.length()==0), current);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            delete cmd;
      }

      cf->set_frame_num (++frame_ord);

      insert_frame (cf);
      set_current ( cf->get_id());
      cid->reset_frame_id(current); 

      // Don't reset the coordinate info if coord_sys is a window coord
      if ((tmp_coord == PIXEL) || (tmp_coord == WINDOW_NORM) || 
          (tmp_coord == PIXEL_REAL))
      {
         cid->set_coord_sys (tmp_coord);
         cid->set_coord_id (tmp_coord_id);
      }
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"frame");

   if (!rval)
   {
      // Dont keep any commands that might result in the update
      bool cmd_lock = cmd_manager->cmd_lock ();
      cmd_manager->set_cmd_lock (true);
      if (framemode != 2)
         cf->set_display_location(framemode, 0, 0, width, height);
      else
         update_display();
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   return rval;
}


int ChipsWindow::add_frame(ChipsId* cid, float xmin, float ymin, float xmax,
    float ymax)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf = NULL;
   string val = theDefaults->get_frameid();

   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE; 
   }
   if (frame.count(val))  
      cmd_manager->add_err(rval=ERR_ID_IN_USE,id,0,val.c_str(), "frame");
   else   
      cf = new_frame(xmin, ymin, xmax, ymax, val);

   if (cf)
   {
      // new frame created- add to list and update current 
      if (!cmd_manager->cmd_lock ())
      {
         // if val is empty that means the frame_cnt changed
         ChipsCreateCommand *cmd = new ChipsCreateCommand(cf,(val.length()==0), current);
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            delete cmd;
      }

      cf->set_frame_num (++frame_ord);
      insert_frame (cf);
      set_current ( cf->get_id());
      cid->reset_frame_id(current); 
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_NOT_CREATED,id,0,"frame");

   if (!rval)
   {
      // Don't keep any commands that might be a result of the update
      bool cmd_lock = cmd_manager->cmd_lock ();
      cmd_manager->set_cmd_lock (true);
      if (framemode != 2)
         cf->set_display_location(framemode, 0, 0, width, height);
      update_display();
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   return rval;
}


int ChipsWindow::delete_frame(ChipsId* cid, bool force_del)
{
   string fid;
   ChipsFrame* cf;
   int rval = ERR_STATUS_GOOD;

   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   // if we're force deleting w/ nothing to delete then we're done
   if (force_del && frame.empty() && dframe.empty())  return rval;  
   
   if (frame.empty())
   {
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"frames");
      return rval;
   }

   fid = cid->get_frame_id();

   if (fid.length() == 0)  fid = current;

   if (chips_strcmp(fid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      bool opened_block = false;

      if (!force_del){
         cmd_manager->open_block ();
         opened_block = true;
      }

      // delete all frames in window 
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cf = (*ii).second;
         cf->unselect ();

         if (force_del)
	    cf->object_delete ();
         else 
	 {
            cf->hide ();
            dframe [cf->get_ref_cnt ()] = cf;

            ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cf,"all");
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }

            // remove the frame coord sys from the system
            ChipsCoordinate *frame_coord = cf->get_chips_coord ();
            if (frame_coord)
               rval = coord->remove_coord_sys (frame_coord->get_id ());
         }
      }
      if (!force_del && opened_block) cmd_manager->close_block ();
      frame.clear();
      frame_order.clear ();
      current="";
      cid->reset_frame_id("");

      // If force deleteing all frames also remove previously delete frames
      if (force_del)
      {
         map <unsigned int, ChipsFrame *>::iterator jj;
         for (jj = dframe.begin (); jj != dframe.end (); jj++)
            jj->second->object_delete ();
         dframe.clear ();
      }
      rval = ERR_STATUS_GOOD;
   }
   else if (frame.count(fid))
   {
      map<string, ChipsFrame*>::iterator ii;
      ii = frame.find(fid);
      cf = (*ii).second;
      cf->unselect();
      erase_frame (fid);
      if (force_del)
         cf->object_delete ();
      else
      {
         cf->hide ();
         dframe [cf->get_ref_cnt ()] = cf;
         ChipsDeleteCommand *cmd = new ChipsDeleteCommand(cf,cf->get_id ());
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }

         // remove the frame coord sys from the system
         ChipsCoordinate *frame_coord = cf->get_chips_coord ();
         if (frame_coord)
            rval = coord->remove_coord_sys (frame_coord->get_id ());
      }
      cid->reset_frame_id("");
      rval = ERR_STATUS_GOOD;
   }
   else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",fid.c_str());
   if (rval == ERR_STATUS_GOOD)
   {
      if (!frame.empty())
      {
         map<string, ChipsFrame* >::reverse_iterator ri = frame.rbegin();
         set_current ( (*ri).first);
      }
      else   set_current ( "");

   }
   return rval;
}


void ChipsWindow::show_layout()
{
   if (!vis_layout)
   {
      vis_layout = true; 
      render_layout(); 
   }
}


void ChipsWindow::hide_layout()
{
   if (vis_layout)
   {
      vis_layout = false; 
      render_layout(); 
   }
}


int ChipsWindow::set_layout_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;

   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_DBL_MINMAX, id, 0, "layout thickness", val,
                                  V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      layout_thickness = val;
      if (activated)   change_layout_thickness();
   }
   return (rval);
}


int ChipsWindow::set_layout_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = layout_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (activated)   rval = change_layout_color();
   return rval;
}


int ChipsWindow::set_layout_linestyle(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   // need to determine style of line
   if (valid_line_type(val))
   {
      layout_linestyle = val;
      if (activated)   change_layout_linestyle();
   }
   else
      cmd_manager->add_err(rval=ERR_LINETYPE_DNE,id,0,val);
   return rval;
}


void ChipsWindow::iconify()
{
#ifndef HEADLESS
   if (gui_window)   gui_window->iconify();
#endif
}


void ChipsWindow::deiconify()
{
#ifndef HEADLESS
   if (gui_window)   gui_window->deiconify();
#endif
}


bool ChipsWindow::iconic()
{
   return false; 
}


char* ChipsWindow::info(ChipsId* cid, V_OBJECT_TYPE obj, bool descend, 
   bool current_obj, bool full)
{
   ChipsFrame *cf;
   string temp, fid;
   char* ptr = NULL;

   if (!cid)  return ptr;
   fid = cid->get_frame_id();
   if (fid.length() == 0)  fid = current;

   temp.clear();

   if (!chips_strcmp(fid.c_str(), "all"))
   {
      map<string, ChipsFrame*>::iterator ii;
      map <int, string>chron;
      // remap so that we are ordered chronologically (or by arrange order)
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         chron.insert(pair<int, string>((*ii).second->get_frame_num(), 
                                        (*ii).first));
      map <int, string>::iterator i2;
      for(i2 = chron.begin(); i2 != chron.end(); i2++)
      {
	 int rev = 0;
         cf = frame[(*i2).second]; 
         if (obj <= chips_frame)
         {
            rev = (theDefaults->get_infocurrent() && current_obj &&
                   ((current == cf->get_id()) || ( current == "all")));
            temp.append("  Frame [");
            if (rev)  temp.append("\x1b[7m");
            temp.append(cf->get_id());
            if (rev)  temp.append("\x1b[m");
            temp.append("]\n");
	 } 
         if (descend || (obj > chips_frame))
         {
            char* lays = NULL;

            if ((lays = cf->info(cid, obj, descend, rev, full)) != NULL)
            {
               temp.append(lays);
               free(lays);
            }
         }
      }
      chron.clear();
   }
   else if (fid != "")
   {
      if ((cf = get_frame(fid)) != NULL)
      {
	 int rev = 0;
         if (obj <= chips_frame)
         {
            rev = (theDefaults->get_infocurrent() && current_obj &&
                   ((current == cf->get_id()) || ( current == "all")));
            temp.append("  Frame [");
            if (rev)  temp.append("\x1b[7m");
            temp.append(cf->get_id());
            if (rev)  temp.append("\x1b[m");
            temp.append("]\n");
         }
         if (descend || (obj > chips_frame)) 
         {
            char* lays = NULL;

            if ((lays = cf->info(cid, obj, descend, rev, full)) != NULL)
            {
               temp.append(lays);
               free(lays);
            }
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}

char* ChipsWindow::info_coordinate(ChipsId* cid)
{
   string temp, fid;
   char* ptr = NULL;

   if (!cid)  return ptr;
   fid = cid->get_frame_id();
   if (fid.length() == 0)  fid = current;

   temp.clear();

   if (!chips_strcmp(fid.c_str(), "all"))
   {
      map<string, ChipsFrame*>::iterator ii;

      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         char* lays = NULL;

	 /* Make sure a coordinate system exists */
         if ((lays = (*ii).second->info_coordinate(cid)) != NULL)
         {
	    temp.append("  Frame [");
            temp.append((*ii).first);
            temp.append("]\n");
            temp.append(lays);
            free(lays);
         }
      }
   }
   else if (fid != "")
   {
      if (frame.count(fid))
      {
         char* lays = NULL;

         // Make sure the coordinate system exists 
         if ((lays = frame[fid]->info_coordinate(cid)) != NULL)
         {
            temp.append("  Frame [");
            temp.append(fid);
            temp.append("]\n");
            temp.append(lays);
            free(lays);
         }
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsWindow::info_depth(ChipsId* cid, int dep)
{
   string temp, fid;
   char *ptr = NULL, *lays;

   if (!cid)  return ptr;
   fid = cid->get_frame_id();
   if (fid.length() == 0)  fid = current;

   temp.clear();

   if (chips_strcmp(fid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((lays = (*ii).second->info_depth(cid, dep)) != NULL)
         {
            temp.append("  Frame [");
            temp.append((*ii).first);
            temp.append("]\n");
            temp.append(lays);
            delete[] lays;
         }
      }
   }
   else if (fid != "")
   {
      if (frame.count(fid))
      {
         if ((lays = frame[fid]->info_depth(cid, dep)) != NULL)
         {
            temp.append("  Frame [");
            temp.append(fid);
            temp.append("]\n");
            temp.append(lays);
            delete[] lays;
         }
      }
   }
   int ll = temp.length();
   if (ll++)
   {
      ptr = new char[ll];
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsWindow::info_bound_axes(ChipsId* cid)
{
   string temp, fid;
   char* ptr = NULL;
   char* lays = NULL;

   if (!cid)  return ptr;

   fid = cid->get_frame_id();
   if (fid.length() == 0)  fid = current;

   temp.clear();

   if (chips_strcmp(fid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;

      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((lays = (*ii).second->info_bound_axes(cid)) != NULL)
         {
	    temp.append("  Frame [");
            temp.append((*ii).first);
            temp.append("]\n");
            temp.append(lays);
            delete [] lays;
         }
      }
   }
   else if (fid != "")
   {
      if (frame.count(fid))
      {
         if ((lays = frame[fid]->info_bound_axes(cid)) != NULL)
         {
	    temp.append("  Frame [");
            temp.append(fid);
            temp.append("]\n");
            temp.append(lays);
            delete [] lays;
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = new char[ll];
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}


char* ChipsWindow::list_current()
{
   string temp;
   char* ptr = NULL;

   temp.clear();

   if (chips_strcmp(current.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      map <int, string>chron;
      char* lays = NULL; 
      ChipsFrame* cf;
      // remap so that we are ordered chronologically (or by arrange order)
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         chron.insert(pair<int, string>((*ii).second->get_frame_num(),
                                        (*ii).first));
      for(map <int, string>::iterator i2=chron.begin();i2!= chron.end();i2++)
      {
         cf = frame[(*i2).second];
         temp.append("  Frame [");
         temp.append(cf->get_id());
         temp.append("]\n");
         if ((lays = cf->list_current()) != NULL)
         {
            temp.append(lays);
            free(lays);
         } 
      }
   }
   else if (current != "")
   {
      if (frame.count(current))
      {
         char* lays = NULL;
         temp.append("  Frame [");
         temp.append(current);
         temp.append("]\n");

         if ((lays = frame[current]->list_current()) != NULL)
         {
            temp.append(lays);
            free(lays);
         }
      }
   }

   int ll = temp.length();
   if (ll++)
   {
      ptr = (char*)malloc(ll*sizeof(char));
      strcpy(ptr, temp.c_str());
   }
   return ptr;
}

bool ChipsWindow::is_current ()
{
   bool rval = false;
   if (eng)
   {
      string current_window = eng->current_window ();
      rval = ((current_window == this->id) || (chips_strcmp(current_window.c_str (), "all")==0));
   }
   return rval;
}


ChipsFrame* ChipsWindow::get_frame(string val)
{
   string fid = val;
   if (!val.length())  fid = current;
   if ((chips_strcmp(fid.c_str(), "all") == 0) && frame.size() == 1)
   {
      map<string, ChipsFrame*>::iterator ii = frame.begin();
      return ((*ii).second); 
   }
   else if ((chips_strcmp(fid.c_str(), "all") != 0) && frame.count(fid))
      return frame[fid];
   else
      return NULL;
}


void ChipsWindow::activate()
{
   if (redraw_blocked() || !visible)
      return;
#ifndef HEADLESS    
   gui_window->show_window ();
   //Added to fix clear()->undo() issue.  Sets up RenderWindow before widgets are attached.
   this->show ();
#endif
  
}


void ChipsWindow::update_display()
{
   ChipsFrame *cf = NULL;
   map<string, ChipsFrame*>::iterator ii;
   switch(framemode)
   {
      case 0: // placed
      {
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            cf = (*ii).second;
            cf->set_display_location(framemode, 0, 0, width, height); 
            cf->force_show ();
         }
      }
      break; 
      case 1: // single 
      {
         for(ii=frame.begin(); ii != frame.end(); ++ii)
         {
            cf = (*ii).second;
            if (current == cf->get_id())
	    {
               cf->set_display_location(framemode, 0, 0, width, height);
               cf->force_show ();
            }
            else
               cf->force_hide ();
         }
      }
      break;
      case 2: // tiled
      {
         if (tile_rows && tile_cols)
         {
            int *rc = (int*) malloc((tile_rows+1)*sizeof(int));
            int *cc = (int*) malloc((tile_cols+1)*sizeof(int));
            if (rc && cc) 
            {
               int zz, yy = 0;
               for (zz=tile_rows+1;zz--;) rc[tile_rows-zz] = zz* height/tile_rows;
               for (zz=tile_cols+1;zz--;) cc[zz] = zz* width/tile_cols; 

               zz = 0; 
 
               for(ii=frame.begin(); ii != frame.end(); ++ii)
               {
                  cf = (*ii).second;
                  if ((yy < tile_rows) && (zz < tile_cols))
                  {
                     cf->set_display_location(framemode, cc[zz], rc[yy+1], 
                                 cc[zz+1], rc[yy]); 
                     cf->force_show ();
                  }
                  else
                     cf->force_hide ();
                  zz++;
                  if (zz == tile_cols) { zz = 0; yy++; } 
               }
               free(rc); 
               free(cc);
            }
         }
      }
      break; 
      default:
      break; 
   }
}


int ChipsWindow::set_framemode(int mode, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if ((mode > -1) && (mode < 3))
   {
      ChipsTileCommand *cmd = NULL;
      if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, 
                               tile_rows, current, mode, tile_cols, tile_rows,
                               current, eng->current_window());
      framemode = mode; 
      if (rec) 
      {
         if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
         {
            if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
            delete cmd;
         }
      }
      update_display();
   }
   else
      cmd_manager->add_err(rval=ERR_INT_MINMAX,id,0,"frame mode",mode,0,2);
   return rval; 
}


int ChipsWindow::prev_frame(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (framemode == 1)
   {
      ChipsFrame *cf = NULL;
      string prev;
      int rr = 0;
      ChipsTileCommand *cmd = NULL;

      map<string, ChipsFrame*>::iterator ii;

      for(ii=frame.begin(); ((rr > -1) && (ii != frame.end())); ++ii)
      {
         cf = (*ii).second;
         if (current == cf->get_id())
         {
           if (rr) rr = -1; 
              else rr++; 
         }
         else
         {
            prev = cf->get_id(); 
            rr++;
         }
      }

      if (prev.length())
      {
         if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, 
                                tile_rows, current, framemode, tile_cols, 
                                tile_rows, prev, eng->current_window());
         if (set_current((char*)prev.c_str()) == ERR_STATUS_GOOD) 
         {
            update_display();
            cid->reset_frame_id(current); 
            if (rec)
            {
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
         else
            delete cmd;
      }
      else  cmd_manager->add_err(rval=ERR_FRAME_NO_PREV,id,0);
   }
   else  cmd_manager->add_err(rval=ERR_FRAME_INVALID_MODE,id,0,"prev_frame",framemode);
   return rval;
}


int ChipsWindow::next_frame(ChipsId* cid, bool rec)
{
   int rval = ERR_STATUS_GOOD;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (framemode == 1)
   {
      ChipsFrame *cf = NULL;
      string next;
      int rr = 0;
      ChipsTileCommand *cmd = NULL;

      map<string, ChipsFrame*>::iterator ii;

      for(ii=frame.begin(); ((rr < 2) && (ii != frame.end())); ++ii)
      {
         cf = (*ii).second;

         
         if (current == cf->get_id()) 
         {
             rr = 1;
         }
         else
         {
            if (rr == 1)
            { 
               next = cf->get_id();  
               rr = 2; 
            }
         }
      }
      if (rr == 1)  // last frame is 'current'  
      {
         ii=frame.begin();
         cf = (*ii).second;
         next = cf->get_id();
      } 

      if (next.length())
      {
         if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, 
                                tile_rows, current, framemode, tile_cols, 
                                tile_rows, next, eng->current_window());
         if (set_current((char*)next.c_str()) == ERR_STATUS_GOOD)
         {
            update_display();
            cid->reset_frame_id(current); 
            if (rec)
            {
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
         else
            delete cmd;
      }
      else  cmd_manager->add_err(rval=ERR_FRAME_NO_NEXT,id,0);
   }
   else  cmd_manager->add_err(rval=ERR_FRAME_INVALID_MODE,id,0,"next_frame",framemode);

   return rval;
}


int ChipsWindow::tile(unsigned int cols, unsigned int rows, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
      cmd_manager->add_err(rval=ERR_BLINK_ACTIVE,id);
   if (rows > V_MAX_TILE_ROWS || rows == 0)
      cmd_manager->add_err(rval=ERR_ROW_TILE_BOUND,id,0,rows,1, V_MAX_TILE_ROWS);
   if (cols > V_MAX_TILE_COLS || cols == 0)
      cmd_manager->add_err(rval=ERR_COL_TILE_BOUND,id,0,cols,1, V_MAX_TILE_COLS);
   if (rval != ERR_STATUS_GOOD)  return rval; 

   ChipsTileCommand *cmd = NULL;
   if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, tile_rows, current,
                                        2, cols, rows, current, eng->current_window());
   tile_rows = rows; 
   tile_cols = cols; 
   set_framemode(2, false);  
   if (rec)
   {
      if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
      {
         if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
         delete cmd;
      }
   }
   return rval;
}


int ChipsWindow::row_tile(unsigned int rows, bool rec)
{
   int rval = ERR_STATUS_GOOD; 
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (rows > V_MAX_TILE_ROWS)
      cmd_manager->add_err(rval=ERR_ROW_TILE_BOUND,id,0,rows,1,V_MAX_TILE_ROWS);
   else if (rows)
   {
      // row priority- calculate number of columns
      int sz = frame.size();
      if (sz > 0) 
      {
         ChipsTileCommand *cmd = NULL;
         int cols = sz / rows; 
         if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, 
                                              tile_rows, current, 2, cols, rows,
                                              current, eng->current_window());
         tile_rows = rows;
         tile_cols = cols;
         if (sz > (tile_rows * tile_cols)) tile_cols++;
         set_framemode(2, false);
         if (rec)
         {
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
      }
      else  cmd_manager->add_err(rval=ERR_NO_ROW_TILES,id,0); 
   }
   else 
      cmd_manager->add_err(rval=ERR_ROW_TILE_BOUND,id,0,rows,1,V_MAX_TILE_ROWS);
   return rval; 
}

int ChipsWindow::col_tile(unsigned int cols, bool rec)
{
   int rval = ERR_STATUS_GOOD; 

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (cols > V_MAX_TILE_COLS)
      cmd_manager->add_err(rval=ERR_COL_TILE_BOUND,id,0,cols,1,V_MAX_TILE_COLS);
   else if (cols)
   {
      // col priority- calculate number of columns
      int sz = frame.size();
      if (sz > 0)
      {
         ChipsTileCommand *cmd = NULL;
         int rows = sz / cols; 
         if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols, 
                                              tile_rows, current, 2, cols, rows,
                                              current, eng->current_window());
         tile_cols = cols;
         tile_rows = rows;
         if (sz > (tile_rows * tile_cols)) tile_rows++;
         set_framemode(2, false);
         if (rec)
         {
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
      }
      else  cmd_manager->add_err(rval=ERR_NO_COL_TILES,id,0);
   }
   else 
      cmd_manager->add_err(rval=ERR_COL_TILE_BOUND,id,0,cols,1,V_MAX_TILE_COLS);
   return rval; 
}


int ChipsWindow::reposition_frame(const ChipsId* cid, float px1, float py1, 
   float px2, float py2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   string frid = cid->get_frame_id();;
   if (frid.length() == 0)  frid = current;
   if (frame.empty()) 
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"frames");
   else if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cf = (*ii).second;
         if ((sts = cf->reposition(px1, py1, px2, py2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
   }
   else
   {
      if (frame.count(frid) == 0)
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
      cf = frame[frid];
      if ((rval = cf->reposition(px1, py1, px2, py2)) == ERR_STATUS_GOOD)
      {
         if (framemode != 2)
            cf->set_display_location(framemode, 0, 0, width, height);
      }
   }
   return rval;
}


int ChipsWindow::single_frame(string fid, bool rec)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid;
   ChipsTileCommand *cmd = NULL;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (fid.length())  frid = fid;
   else   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      // we can switch into single frame mode if a single frame is
      // selected as current
      if (frame.size() == 1) 
      {
         map<string, ChipsFrame*>::iterator ii = frame.begin();
         cf = (*ii).second;
         if (cf)
         {
            if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols,
                                    tile_rows, current, 1, tile_cols, tile_rows,
                                    current, eng->current_window());
            set_current ( cf->get_id()); 
            set_framemode(1, false);
            if (rec)
            {
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
      }
      else 
         rval = ERR_MULT_SINGLE_FRM;
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         if (rec)  cmd = new ChipsTileCommand(this, framemode, tile_cols,
                            tile_rows, current, 1, tile_cols, tile_rows,
                            current, eng->current_window());
         set_current ( cf->get_id());
         set_framemode(1, rec);
         if (rec)
         {
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::blink_frames(ChipsId* cid, string fid2, float interval)
{
   int rval = ERR_STATUS_GOOD;

#ifndef HEADLESS

   if (interval < 0)
   {
      cmd_manager->add_err(rval=ERR_FLT_MIN,id,0,"interval",interval,0);
      return ERR_NO_FLT_NEGATIVE;
   }
   if (duration == 0)
   {
      frame_1 = cid->get_frame_id();
      frame_2 = fid2;

      ChipsFrame *cf1, *cf2; 
      // make sure both frames exist before setting blink timer
      if ((cf1 = get_frame(frame_1)) != NULL  && 
          ( cf2 = get_frame(frame_2)) != NULL)
      { 
         // need to use two different frames to blink
         if (cf1 == cf2)
         {
            string estr= "The blink operation requires two different frames. Only one (";    
            estr.append(cf1->get_id()); 
            estr.append(") was provided.");
            cmd_manager->add_err(estr.c_str(), ERR_FRAME_NO_NEXT,id,0);
            return ERR_FRAME_NO_NEXT;
         }
        
         lastframe = current;
         prev_mode = framemode; 

         // since the duration variable is used as a flag to prevent
         // certain commands from executing while 'blinking', we need
         // to set the framemode before adjusting the duration value
         set_framemode(1); 
         duration = (int)(interval * 1000.0);
         timer = gtk_timeout_add(duration, (GtkFunction)ChipsWindow::blink,
                                 (gpointer)this);
      }
   }
   else
      cmd_manager->add_err(rval=ERR_BLINK_ACTIVE,id);
#endif
   return rval; 
}


int ChipsWindow::stop_blink(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;

#ifndef HEADLESS


   if (timer)
   {
       gtk_timeout_remove(timer);
       duration = 0;
       // return to old frame mode 
       cid->set_frame_id(lastframe.c_str());
       rval = set_framemode(prev_mode);
       redraw();
   }
#endif
   return rval; 
}


int ChipsWindow::set_cascading_property(ChipsId* cid, V_OBJECT_TYPE obj)
{
   int sts, pass = 0, failval = 0, rval = ERR_STATUS_GOOD; 
   ChipsFrame* cf = NULL;

   if (obj == chips_window)
   {
      // set window properties
      int mask = theDefaults->get_allpropmask();
 
      if ((mask & PREF_ALL_BGCOLOR) == PREF_ALL_BGCOLOR)
      {
         // set background (window) color 
	 string clr = theDefaults->get_allbgcolor();
         sts = set_property(NULL, (void*)clr.c_str(), SET_WINDOW_COLOR);
         if (sts == ERR_STATUS_GOOD)    pass++;
            else if (!err_type_object_DNE(sts))   failval = sts;
      }
      else if ((mask & PREF_ALL_COLOR) == PREF_ALL_COLOR)
      {
         // set foreground (grids, layout, etc...) colors  
      }

      // set all objects in heirarchy 
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->set_cascading_property(cid, obj)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
   }
   else 
   {
      string frid = cid->get_frame_id();
      if (frid.length() == 0)   frid = current;
      if (frid.length() && (chips_strcmp(frid.c_str(), "all") != 0))
      {
         if ((cf = get_frame(frid)) != NULL)
         {
            if ((sts = cf->set_cascading_property(cid, obj)) == ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts; 
         }
         else
            cmd_manager->add_err(failval=ERR_OBJ_VAL_DNE, id, 0, "frame",
                                 frid.c_str());
      }
      else 
      {
         map<string, ChipsFrame*>::iterator ii;
         for(ii=frame.begin(); ii != frame.end(); ++ii)
            if ((sts = (*ii).second->set_cascading_property(cid, obj)) ==
                ERR_STATUS_GOOD)
               pass++;
            else if (!err_type_object_DNE(sts))
               failval = sts;
      }
   }
   cmd_manager->clear_errs(CLEAR_DNE_ERRS);
   if (failval)
      rval = failval;
   else if (!pass)
      cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
   return rval; 
}


int ChipsWindow::set_property(ChipsId* cid, void* in_val, int in_property, 
   bool rec)
{
   int rval = ERR_STATUS_GOOD, set_win_prop = true;
   ChipsFrame *cf;
   ChipsSetCommand *cmd = NULL;
   string frid;

   switch(in_property)
   {
      case SET_GRID_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         rval = set_grid_thickness(*fval); 
      }
      break; 
      case SET_GRID_SCALE:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_grid_scale(*ival); 
      }
      break; 
      case SET_GRID_COLOR:
      {
         string clr = (const char*)in_val;
         rval = set_grid_color(clr);
      }
      break; 
      case SET_WINDOW_TITLE:
      {
         string val = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_title(), val, eng->current_window());
         if ((rval = set_title(val)) == ERR_STATUS_GOOD)
         {
            if (rec)
            {
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
         else
            delete cmd;
      }
      break;
      case SET_WINDOW_COLOR:
      {
         string clr = (const char*)in_val;
         if (rec)  cmd = new ChipsSetCommand(this, in_property,
            get_color(), clr, eng->current_window());
         if ((rval = set_color(clr)) == ERR_STATUS_GOOD)
         {
            if (rec)
            {
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
         }
      }
      break;
      case SET_WINDOW_REDRAW:
      {
         bool* bval = static_cast<bool*>(in_val);
	 if (rec)  cmd = new ChipsSetCommand(this, in_property,
	    redraw_on, *bval, eng->current_window());
         set_redraw(*bval);
         if (redraw_on)
         {
            if (!visible)
            {
	       bool bv = true;
               rval = set_property(NULL, static_cast<void*>(&bv), 
	                           SET_WINDOW_VISIBLE, true);
            }
         }
         if (rec)
         {
            if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
            {
               if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
               delete cmd;
            }
         }
      }
      break; 
      case SET_WINDOW_VISIBLE:
      {
         bool* bval = static_cast<bool*>(in_val);
         // since this command is only called internally, we don't have
         // to worry about adding an entry to the undo/redo stack if the
         // value hasn't  changed
         if (*bval != visible)
         {
            if (rec)
            {
               cmd = new ChipsSetCommand(this, in_property, visible, *bval,
                                         eng->current_window());
               if ((rval=cmd_manager->add_cmd (cmd)) != ERR_STATUS_GOOD)
               {
                  if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                  delete cmd;
               }
            }
            if ((visible = *bval) == true)  
               rval = show();
            else
               rval = hide();
         }
      }
      break; 
      case SET_EXPORT_FONTS:
      {
         bool* bval = static_cast<bool*>(in_val);
         rval = set_export_font(*bval); 
      }
      break;
      case SET_EXPORT_ORIENT:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_export_orient(*ival); 
      }
      break;
      case SET_GRID_LINESTYLE:  
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_grid_linestyle((V_LINE_TYPE)*ival);
      }
      break; 
      case SET_LAYOUT_THICKNESS:
      {
         float* fval = static_cast<float*>(in_val);
         rval = set_layout_thickness(*fval); 
      }
      break; 
      case SET_LAYOUT_COLOR:     
      {
         string clr = (const char*)in_val;
         rval = set_layout_color(clr);
      }
      break; 
      case SET_LAYOUT_LINESTYLE:
      {
         int* ival = static_cast<int*>(in_val);
         rval = set_layout_linestyle((V_LINE_TYPE)*ival);
      }
      break; 
      case SET_PAGE_LMARGIN:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_lmargin(*fval);
      }
      break;
      case SET_PAGE_RMARGIN:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_rmargin(*fval);
      }
      break;
      case SET_PAGE_TMARGIN:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_tmargin(*fval);
      }
      break;
      case SET_PAGE_BMARGIN:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_bmargin(*fval);
      }
      break;
      case SET_PAGE_WIDTH:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_width(*fval);
      }
      break;
      case SET_PAGE_HEIGHT:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_height(*fval);
      }
      break;
      case SET_PAGE_SIZE:
      {
	 string clr = (const char*)in_val;
         rval = set_page_size(clr);
      }
      break;
      case SET_PAGE_SCALE_WIDTH:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_scale_width(*fval);
      }
      break;
      case SET_PAGE_SCALE_HEIGHT:
      {
	 float* fval = static_cast<float*>(in_val);
         rval = set_page_scale_height(*fval);
      }
      break;
      case SET_PAGE_UNITS:
      {
	 string clr = (const char*)in_val;
         rval = set_page_units(clr);
      }
      break;
      case SET_PAGE_FIT:
      {
	 bool* bval = static_cast<bool*>(in_val);
         rval = set_fit_to_page(*bval);
      }
      break;
      case SET_PAGE_KEEP_ASPECT:
      {
	 bool* bval = static_cast<bool*>(in_val);
         rval = set_page_keep_aspect(*bval);
      }
      break;
      default:
      set_win_prop = false; 
      if (cid)  
      {
         frid = cid->get_frame_id();
         if (frid.length() == 0)   frid = current;
         if (chips_strcmp(frid.c_str(), "all") == 0)
         {
            map<string, ChipsFrame*>::iterator ii;
            int sts, pass = 0, failval = 0; 
            // if a set_property call passed for any entry within an 'all'
            // we don't generate an error. However, if no entries pass we
            // generate errors- either the specific errors or a DNE failure
            // if no entries exist 
            ChipsId* tcid;
            for(ii=frame.begin(); ii != frame.end(); ++ii)
               if ((cf = (*ii).second) != NULL)
               {
                  tcid = new ChipsId(*cid);
                  if ((sts = cf->set_property(tcid, in_val, in_property)) ==
                      ERR_STATUS_GOOD)  
                     pass++;
                  else if (!err_type_object_DNE(sts))  
                     failval = sts; 
                  delete tcid;
               }
	    cmd_manager->clear_errs(CLEAR_DNE_ERRS);
            if (failval)
	       rval = failval;
            else if (!pass)
               cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
            else
               set_current ( "all");
         }
         else if (frid != "")
         {
            if ((cf = get_frame(frid)) != NULL)
            {
               if ((rval = cf->set_property(cid, in_val, in_property)) ==
                    ERR_STATUS_GOOD)
               {
		  if (rec && (current != frid))
		  {
		     ChipsCurrentCommand *cmd2= new ChipsCurrentCommand(this,
		        current, frid, chips_frame);
		     if ((rval=cmd_manager->add_cmd(cmd2))!= ERR_STATUS_GOOD)
                     {
                        if (rval == ERR_CMD_LOCKED) rval = ERR_STATUS_GOOD;
                        delete cmd2;
                     }
		  }
		  set_current ( frid); 
               }
            }
            else  
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame",
	                            frid.c_str());
         }
         else
         {
            // request to update the property of an item which does
            // not exist or a request for an unknown property update
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,id,0,"frames");
         }
      }
      break;
   } 

   // need to add frame id for errors
   if (cmd_manager && rval)   cmd_manager->add_window_id_to_errs(id);

   if (rval == ERR_STATUS_GOOD && set_win_prop)
      cmd_manager->update_mask(STATE_CHNG_PROP_WIN);

   return rval;
}


int ChipsWindow::get_property(ChipsId* cid, void* out_val, int out_property)
{
   int rval = ERR_STATUS_GOOD;
   string frid;

//WLM   cmd_manager->clear_errs();

   switch(out_property)
   {
      case SET_WINDOW_REDRAW:
      {
         int *ival = (int*)out_val;
         *ival = get_redraw();
      }
      break;
      case SET_WINDOW_COLOR:
      {
         string *sval =  (string*)out_val;
         *sval = get_color();
      }
      break;
      case SET_WINDOW_TITLE:
      {
         string *sval =  (string*)out_val;
         *sval = get_title();
      }
      break;
      case SET_EXPORT_FONTS:
      {
         int *ival = (int*)out_val;
         *ival = get_export_font(); 
      }
      break;
      case SET_EXPORT_ORIENT:
      {
         int *ival = (int*)out_val;
         *ival = get_export_orient(); 
      }
      break;
      case SET_PAGE_LMARGIN:
      {
	 float *fval = (float*)out_val;
         *fval = get_page_lmargin();
      }
      break;
      case SET_PAGE_RMARGIN:
      {
         float *fval = (float*)out_val;
         *fval = get_page_rmargin();
      }
      break;
      case SET_PAGE_TMARGIN:
      {
         float *fval = (float*)out_val;
         *fval = get_page_tmargin();
      }
      break;
      case SET_PAGE_BMARGIN:
      {
         float *fval = (float*)out_val;
         *fval = get_page_bmargin();
      }
      break;
      case SET_PAGE_WIDTH:
      {
         float *fval = (float*)out_val;
         *fval = get_page_width();
      }
      break;
      case SET_PAGE_HEIGHT:
      {
         float *fval = (float*)out_val;
         *fval = get_page_height();
      }
      break;
      case SET_PAGE_SIZE:
      {
         string *sval = (string*)out_val;
         *sval = get_page_size();
      }
      break;
      case SET_PAGE_SCALE_WIDTH:
      {
         float *fval = (float*)out_val;
         *fval = get_page_scale_width();
      }
      break;
      case SET_PAGE_SCALE_HEIGHT:
      {
         float *fval = (float*)out_val;
         *fval = get_page_scale_height();
      }
      break;
      case SET_PAGE_SCALE_UNITS:
      {
         string *sval = (string*)out_val;
         *sval = get_page_units();
      }
      break;
      case SET_PAGE_FIT:
      {
         int *ival = (int*)out_val;
         *ival = (int)get_fit_to_page();
      }
      break;
      case SET_PAGE_KEEP_ASPECT:
      {
         int *ival = (int*)out_val;
         *ival = (int)get_page_keep_aspect();
      }
      break;
      default:
         frid = cid->get_frame_id();
         if (frid.length() == 0)   frid = current;

	 if (frame.empty())
            cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"frames");
         else if (chips_strcmp(frid.c_str(), "all") == 0)
         {
            ChipsFrame* cf = get_frame(frid); 
            if (frame.size() == 1 && (cf != NULL))
               rval = frame[cf->get_id()]->get_property(cid, out_val, out_property);
            else 
               cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"frames");
         }
         else if (frame.count(frid))
            rval = frame[frid]->get_property(cid, out_val, out_property);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
      break;
   }
   return rval;
}


int ChipsWindow::adjust_grid_gaps(ChipsId* cid, float xg, float yg)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

//WLM   cmd_manager->clear_errs();

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->adjust_grid_gaps(xg, yg)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_gaps(xg, yg);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::adjust_grid_xgap(ChipsId* cid, float xg)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

//WLM   cmd_manager->clear_errs();

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->adjust_grid_xgap(xg)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_xgap(xg);
         set_current ( cf->get_id());
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::adjust_grid_ygap(ChipsId* cid, float yg)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   cmd_manager->clear_errs();

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->adjust_grid_ygap(yg)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_ygap(yg);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::adjust_grid_xrelsize(ChipsId* cid, int col, float ratio)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->adjust_grid_xrelsize(col,ratio))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_xrelsize(col, ratio);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}

int ChipsWindow::adjust_grid_yrelsize(ChipsId* cid, int row, float ratio)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->adjust_grid_yrelsize(row,ratio))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_yrelsize(row, ratio);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::adjust_grid_xrelsizes(ChipsId* cid, int cnt, float* ratios)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->adjust_grid_xrelsizes(cnt,ratios))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_xrelsizes(cnt, ratios);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::adjust_grid_yrelsizes(ChipsId* cid, int cnt, float* ratios)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid;

   frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->adjust_grid_xrelsizes(cnt,ratios))==ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->adjust_grid_yrelsizes(cnt, ratios);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::grid_objects(ChipsId* cid, int cols, int rows, 
   float xg, float yg, int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid;
   cmd_manager->clear_errs();
   frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->grid_objects(cols,rows,xg,yg,usage,lcnt,lids))==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->grid_objects(cols, rows, xg, yg, usage, lcnt, lids);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::grid_objects(ChipsId* cid, int orient, int cnt, int* vals,
   float xg, float yg, int usage, int lcnt, string* lids)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid;
   cmd_manager->clear_errs();
   frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=cf->grid_objects(orient,cnt,vals,xg,yg,usage,lcnt,lids)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->grid_objects(orient, cnt, vals, xg, yg, usage, lcnt, lids);
         set_current ( cf->get_id());
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::swap_object_positions(ChipsId* cid, string id2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid, id1;
   cmd_manager->clear_errs();
   frid = cid->get_frame_id();
   id1 = cid->get_plot_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = cf->swap_object_positions(id1, id2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->swap_object_positions(id1, id2);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::add_plot(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->add_plot(cid);
         set_current ( frid);
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);
      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)   rval = cf->add_plot(cid);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_plot(ChipsId* cid, float px1,float py1, float px2,
    float py2)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_plot(cid, px1, py1, px2, py2)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->add_plot(cid, px1, py1, px2, py2);
         set_current ( frid);
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if (frame.count(current))
         rval = frame[current]->add_plot(cid, px1, py1, px2, py2);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_plot(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
//WLM   cmd_manager->clear_errs();

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string pid = cid->get_plot_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_plot_id(pid);
         if ((sts = (*ii).second->delete_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_plot(cid);
         set_current ( cf->get_id());
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::clear_plot(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();

   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string pid = cid->get_plot_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_plot_id(pid);
         if ((sts = (*ii).second->clear_plot(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->clear_plot(cid);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::reposition_plot(ChipsId* cid, float px1, float py1,
   float px2, float py2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();;
   if (frid.length() == 0)  frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts = (*ii).second->reposition_plot(cid, px1, py1, px2, py2)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->reposition_plot(cid, px1, py1, px2, py2);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::limits(ChipsId* cid, int dim, double min, double max)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   cmd_manager->clear_errs();
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;

      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->limits(cid, dim, min, max)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else 
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->limits(cid, dim, min, max);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval; 
}


int ChipsWindow::limits(ChipsId* cid, int dim, double xmin, double xmax,
   double ymin, double ymax)
{
   int rval = ERR_STATUS_GOOD;
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   cmd_manager->clear_errs();
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->limits(cid, dim, xmin, xmax, ymin, ymax)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      ChipsFrame* cf; 
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->limits(cid, dim, xmin, xmax, ymin, ymax);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   V_AUTOLIMIT_TYPE automax)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->limits(cid, dim, automin, automax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->limits(cid, dim, automin, automax);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::limits(ChipsId* cid, int dim, double min,
   V_AUTOLIMIT_TYPE automax)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid = cid->get_frame_id();
   cmd_manager->clear_errs();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->limits(cid, dim, min, automax)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->limits(cid, dim, min, automax);
         set_current ( cf->get_id());
      }
      else  
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::limits(ChipsId* cid, int dim, V_AUTOLIMIT_TYPE automin,
   double max)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->limits(cid, dim, automin, max)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->limits(cid, dim, automin, max);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}

int ChipsWindow::panto(ChipsId* cid, double x, double y)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->panto(cid, x, y)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->panto(cid, x, y);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}



int ChipsWindow::zoom(ChipsId* cid, float scale)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->zoom(cid, scale)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->zoom(cid, scale);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}

int ChipsWindow::pan(ChipsId* cid, double x1, double y1, double x2, double y2)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((sts=(*ii).second->pan(cid, x1, y1, x2, y2)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->pan(cid, x1, y1, x2, y2);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;


   return rval;
}






int ChipsWindow::add_curve(ChipsId* cid, long cnt, double* x, double* y,
   double* ydown, double* yup, double* xdown, double* xup, double* ylims,
   double* xlims, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         if ((sts = (*ii).second->add_curve(&tmpid, cnt, x, y, ydown, yup, 
              xdown, xup, ylims, xlims, transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup, ylims,
                              xlims, transform);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());

   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
      {
         if ((rval = cf->add_plot(cid)) == ERR_STATUS_GOOD)
            rval = cf->add_curve(cid, cnt, x, y, ydown, yup, xdown, xup, ylims,
                                 xlims, transform);
      }
      cmd_manager->set_cmd_lock (cmd_lock); 
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);


   return rval;
}


int ChipsWindow::delete_curve(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string cvid = cid->get_curve_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_curve_id(cvid);
         if ((sts = (*ii).second->delete_curve(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_curve(cid);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_contour(ChipsId* cid, long cnt, double* x, double* y,
   double* z, unsigned int numlev, double* levels, Transform *trans)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         if ((sts = (*ii).second->add_contour(&tmpid, cnt, x, y, z, numlev,
              levels, trans)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
         set_current ( cf->get_id());
      }
      else
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
      {
         if ((rval = cf->add_plot(cid)) == ERR_STATUS_GOOD)
            rval = cf->add_contour(cid, cnt, x, y, z, numlev, levels, trans);
      }
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_contour(ChipsId* cid, double* dat, long xdim, long ydim,
   unsigned int numlev, double* levels, Transform *trans)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         if ((sts = (*ii).second->add_contour(&tmpid, dat, xdim, ydim, numlev,
              levels, trans)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_contour(cid, dat, xdim, ydim, numlev, levels, trans);
         set_current ( cf->get_id());
      }
      else   
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
      {
         if ((rval = cf->add_plot(cid)) == ERR_STATUS_GOOD)
            rval = cf->add_contour(cid, dat, xdim, ydim, numlev, levels, trans);
      }
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_contour(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ctid = cid->get_contour_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_contour_id(ctid);
         if ((sts = (*ii).second->delete_contour(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_contour(cid);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::set_contour_levels(ChipsId* cid,
   unsigned int arraysize, double* pos)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   cmd_manager->clear_errs();
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         rval |= (*ii).second->set_contour_levels(cid, arraysize, pos);
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->set_contour_levels(cid, arraysize, pos);
         set_current ( cf->get_id());
      }
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}

int ChipsWindow::get_contour_levels(ChipsId* cid, vector<double> &pos)
{
   int rval = ERR_STATUS_GOOD;
   cmd_manager->clear_errs();
   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (frame.empty())
      cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"frames");
   else if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      if (frame.size() == 1)
      {
         ChipsFrame* cf = get_frame(frid);
         rval = cf->get_contour_levels(cid, pos);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_INVALID_ALL,id,0,"frames");
   }
   else if (frame.count(frid))
      rval = frame[frid]->get_contour_levels(cid, pos);
   else
      cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   return rval;
}

int ChipsWindow::add_histogram(ChipsId* cid, long cnt, double* bin_low,
   double* bin_high, double* counts, double* yd, double* yu)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         ChipsId tmpid = *cid;
         if ((sts = (*ii).second->add_histogram(&tmpid, cnt, bin_low, 
              bin_high, counts, yd, yu)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_histogram(cid, cnt, bin_low, bin_high, counts,
                                  yd, yu);
         set_current ( cf->get_id());
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
      {
         if ((rval = cf->add_plot(cid)) == ERR_STATUS_GOOD)  rval = 
            cf->add_histogram(cid, cnt, bin_low, bin_high, counts, 
                              yd, yu); 
      }
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::delete_histogram(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string hid = cid->get_histogram_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_histogram_id(hid);
         if ((sts = (*ii).second->delete_histogram(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_histogram(cid);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}
  


int ChipsWindow::add_axis(ChipsId* cid, int dim, float pos,
   float minx, float maxx, float miny, float maxy, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform)) == 
              ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);
         set_current ( frid);
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str()); 
   }
   else
   {
      bool cmd_lock = cmd_manager->cmd_lock ();

      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);


      cmd_manager->set_cmd_lock (true);
      if (frame.count(current))
         rval = frame[current]->add_axis(cid, dim, pos, minx, maxx, miny, maxy, transform);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_axis(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string aid = cid->get_axis_id();
      string axid = cid->get_xaxis_id();
      string ayid = cid->get_yaxis_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_axis_id(aid);
         cid->set_xaxis_id(axid);
         cid->set_yaxis_id(ayid);
         if ((sts = (*ii).second->delete_axis(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"plots");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->delete_axis(cid);
         set_current ( frid);
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::set_arbitrary_tick_positions(ChipsId* cid, 
   unsigned int arraysize, double* pos, string* vals)
{
   int rval = ERR_STATUS_GOOD;
//WLM   cmd_manager->clear_errs();
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts=(*ii).second->set_arbitrary_tick_positions(cid,arraysize,
             pos,vals)) == ERR_STATUS_GOOD)
           pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->set_arbitrary_tick_positions(cid, arraysize, pos, vals);
         set_current ( frid);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::bind_axes(ChipsId* cid, const string lid, const string aid)
{
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (!chips_strcmp(frid.c_str(), "all") && (frame.size() != 1))  
   {
      cmd_manager->add_err(ERR_OBJ_INVALID_ALL,id,0,"frames");
      return ERR_OBJ_INVALID_ALL;
   }
   if (frid == "")
   {
       cmd_manager->add_err(ERR_NO_CUR_FRAME,id);
       return ERR_NO_CUR_FRAME; 
   }
   if (frame.count(frid) == 0)              
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   set_current ( frid);
   int rval;
   ChipsFrame *cf = get_frame(frid); 
   if ((rval = cf->bind_axes(cid, lid, aid)) != ERR_STATUS_GOOD)   
      cmd_manager->add_window_id_to_errs(id);
   return rval; 
}


int ChipsWindow::unbind_axes(ChipsId* cid, const string lid, const string aid)
{
   string frid = cid->get_frame_id();
//WLM   cmd_manager->clear_errs();
   if (frid.length() == 0)   frid = current;
   if (!chips_strcmp(frid.c_str(), "all") && (frame.size() != 1))  
   {
      cmd_manager->add_err(ERR_OBJ_INVALID_ALL,id,0,"frames");
      return ERR_OBJ_INVALID_ALL;
   }
   if (frid == "")
   {
       cmd_manager->add_err(ERR_NO_CUR_FRAME,id);
       return ERR_NO_CUR_FRAME; 
   }
   if (frame.count(frid) == 0)
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   set_current ( frid);
   int rval;
   ChipsFrame* cf = get_frame(frid); 
   if ((rval = cf->unbind_axes(cid, lid, aid)) != ERR_STATUS_GOOD)
      cmd_manager->add_window_id_to_errs(id);
   return rval; 
}

int ChipsWindow::format_axis_point (const ChipsId* cid, double pnt, string &formatted_pnt)
{
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (!chips_strcmp(frid.c_str(), "all") && (frame.size() != 1))
   {
      cmd_manager->add_err(ERR_OBJ_INVALID_ALL,id,0,"frames");
      return ERR_OBJ_INVALID_ALL;
   }
   if (frid == "")
   {
       cmd_manager->add_err(ERR_NO_CUR_FRAME,id);
       return ERR_NO_CUR_FRAME;
   }
   if (frame.count(frid) == 0)
   {
      cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
      return ERR_OBJ_VAL_DNE;
   }
   int rval;
   ChipsFrame* cf = get_frame(frid);
   if ((rval = cf->format_axis_point(cid, pnt, formatted_pnt)) != ERR_STATUS_GOOD)
      cmd_manager->add_window_id_to_errs(id);
   return rval;
}



int ChipsWindow::add_image(ChipsId* cid, string filename, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_image(cid, filename, transform)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid) == 0)   
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE, id, 0, "frame", frid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
      rval = frame[frid]->add_image(cid, filename, transform);
      set_current ( frid);
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if (frame.count(current))
         rval = frame[current]->add_image(cid, filename, transform);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_image(ChipsId* cid,
   string filename, float ix1, float iy1, float ix2, float iy2, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         rval = (*ii).second->add_image(cid, filename, ix1, iy1, ix2, iy2, transform);
      {
         if ((sts = (*ii).second->add_image(cid, filename, ix1, iy1, ix2, iy2, transform))
             == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if (frame.count(frid))
      {
         rval = frame[frid]->add_image(cid, filename, ix1, iy1, ix2, iy2, transform);
         set_current ( frid);
      }
      else  cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if (frame.count(current))
         rval = frame[current]->add_image(cid, filename, ix1, iy1, ix2, iy2, transform);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_image(ChipsId* cid, double *rvals, double *gvals, double *bvals, double *avals, int xaxis, int yaxis, Transform *transform)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      rval = add_frame(cid);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
      {
         rval = cf->add_image(cid, rvals, gvals, bvals, avals, xaxis, yaxis, transform);
      }
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_image(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((cf = (*ii).second) != NULL)
            rval = cf->delete_image(cid);
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_image(cid);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::set_image_channel(ChipsId* cid, double* vals, int xaxis, int yaxis, int channel)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((cf = (*ii).second) != NULL)
            rval = cf->set_image_channel(cid, vals, xaxis, yaxis, channel);
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->set_image_channel(cid, vals, xaxis, yaxis, channel);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str ());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::remove_image_channel(ChipsId* cid, int channel)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         if ((cf = (*ii).second) != NULL)
            rval = cf->remove_image_channel(cid, channel);
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->remove_image_channel(cid, channel);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::print_image(ChipsId* cid, const string &outfile)
{
   int rval = ERR_STATUS_GOOD;
   int use_format;
   string format = theDefaults->get_exportformat();
   string checkfile;
   ChipsFrame *cf;

   const char* ext[] = { ".ps", ".eps", ".pdf", ".png", ".jpg", ".jpeg"};
   bool add_ext = true;

   // must specify file for cmyk components
   if (outfile.length() == 0)
   {
      cmd_manager->add_err(ERR_FILE_NOT_SUPPLIED);
      return ERR_FILE_NOT_SUPPLIED;
   }

   // check the extension of the outfile to and use that as the format
   // if it's a proper extensions otherwise refer to the preferences
   string::size_type ext_pos = outfile.find_last_of (".", outfile.length ());

   if ((ext_pos != string::npos) && (ext_pos+1 != outfile.length ()))
   {
      string outfile_ext = outfile.substr (ext_pos, outfile.length ());
      for (int ii=0; ii<6; ii++)
      {
         if (outfile_ext == ext[ii])
         {
            add_ext = false;
            use_format = ii;
         }
      }
   }

   if (add_ext)
   {
      if (format.length() == 0)
         use_format = 3;          // default format (png)
      else if (chips_strcmp(format.c_str(), "ps") == 0)
         use_format = 0;          // postscript format
      else if (chips_strcmp(format.c_str(), "eps") == 0)
         use_format = 1;          // encapsulated postscript format
      else if (chips_strcmp(format.c_str(),"pdf") == 0)
         use_format = 2;          // pdf format
      else if (chips_strcmp(format.c_str(),"png") == 0)
         use_format = 3;          // png format
      else if (chips_strcmp(format.c_str(),"jpg") == 0)
         use_format = 4;          // jpg format
      else if (chips_strcmp(format.c_str(),"jpeg") == 0)
         use_format = 5;          // jpg format
      else{
         cmd_manager->add_err(ERR_STR_SUBSET,id,0,"export format",
                           export_format.c_str(),"png|jpg|jpeg");
         return ERR_STR_SUBSET;  // unsupported format
      }
   }

   if (outfile.length())
   {
      // check for directory write permission
      string fpath = outfile;
      int spos = fpath.rfind('/');
      if ((spos >= 0) && (spos < fpath.size()))
      {
         // found a '/' in path- remove filename portion
         fpath.erase(spos);
         if (access(fpath.c_str(), W_OK) != 0)
         {
            cmd_manager->add_err(ERR_DIR_WRITE_PERM,id,0,fpath.c_str());
            return ERR_DIR_WRITE_PERM;
         }
      }

      // check if file already exists
      checkfile = outfile;
      if (add_ext)
         checkfile.append(ext[use_format]);
      if (!access(checkfile.c_str(), F_OK))
      {
         if (!theDefaults->get_exportclobber())
         {
            cmd_manager->add_err(ERR_FILE_CLOBBER_OFF,"",0,checkfile.c_str());
            return ERR_FILE_CLOBBER_OFF;
         }
         else if (access(checkfile.c_str(), W_OK) != 0)
         {
            cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
            return ERR_ACCESS_PERMISSION;
         }
      }
   }


   if (duration)
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0 && frame.size() != 1)
   {
      cmd_manager->add_err (rval=ERR_OBJ_INVALID_ALL,id,0,"frames");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->print_image(cid, outfile, use_format);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_label(ChipsId* cid, double xpos, 
   double ypos, string text)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_label(cid, xpos, ypos, text)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_label(cid, xpos, ypos, text);
         set_current ( cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
     
      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord); 
      cid->set_coord_id (tmp_coord_id);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_label(cid, xpos, ypos, text);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_label(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
//WLM   cmd_manager->clear_errs();
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string lid = cid->get_label_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_label_id(lid);
         if ((sts = (*ii).second->delete_label(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ( "all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_label(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_line(ChipsId* cid, double startx, double starty,
     double endx, double endy)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_line(cid, startx, starty, endx, endy)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_line(cid, startx, starty, endx, endy);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      
      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);


      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_line(cid, startx, starty, endx, endy);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_line(ChipsId* cid, double* xx, double* yy, 
   			  unsigned int count)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_line(cid, xx, yy, count)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_line(cid, xx, yy, count);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();

      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);


      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_line(cid, xx, yy, count);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_line(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
//WLM   cmd_manager->clear_errs();
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string lid = cid->get_line_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_line_id(lid);
         if ((sts = (*ii).second->delete_line(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_line(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::add_colorbar(ChipsId* cid, double xpos, double ypos,
   string obj_id, V_OBJECT_TYPE obj)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_colorbar(cid,xpos,ypos,obj_id,obj)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_colorbar(cid,xpos,ypos,obj_id,obj);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      
      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_colorbar(cid,xpos,ypos,obj_id,obj);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}

int ChipsWindow::delete_colorbar(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ptid = cid->get_colorbar_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_colorbar_id(ptid);
         if ((sts = (*ii).second->delete_colorbar(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_colorbar(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_point(ChipsId* cid, double xx, double yy)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_point(cid, xx, yy)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_point(cid, xx, yy);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      
      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);

      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_point(cid, xx, yy);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_point(ChipsId* cid, double xx, double yy, double zz)
{
  cmd_manager->add_err(ERR_FEATURE_DNE);
  return ERR_FEATURE_DNE;
}


int ChipsWindow::delete_point(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
//WLM   cmd_manager->clear_errs();
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string ptid = cid->get_point_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_point_id(ptid);
         if ((sts = (*ii).second->delete_point(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_point(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_legend(ChipsId* cid, const char** curve_ids,
   const char** curve_labels, string lgid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;
   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_legend(cid, curve_ids, curve_labels, 
             lgid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
      if (rval == ERR_STATUS_GOOD)  set_current ("all"); 
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_legend(cid, curve_ids, curve_labels, lgid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_legend(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         rval = (*ii).second->delete_legend(cid);
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_legend(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::convert_coordinate (ChipsVector in_val[], ChipsVector out_val[],
                               int num_vectors, ChipsId *cid1, ChipsId *cid2, int axis)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string frid;
   string frid2;
   int frame_split = 0;

   if (num_vectors == 0)
      return ERR_STATUS_GOOD;

   if (!cid1 || !cid2)
   {
      cmd_manager->add_err(ERR_NO_ID_OBJ,id);
      return ERR_NO_ID_OBJ;   // ChipsId object not specified
   }
   ChipsId id_in (*cid1);
   ChipsId id_out (*cid2);

   // The coord id should not be used for anything but data coordinates
   if (id_in.get_coord_sys () != DATA)
       id_in.set_coord_id ("");
   if (id_out.get_coord_sys () != DATA)
       id_out.set_coord_id ("");
 
   // If the coord sys is not a window coordinate find the
   // current frame
   if ((cid1->get_coord_sys () != WINDOW_NORM) && 
       (cid1->get_coord_sys () != PIXEL) &&
       (cid1->get_coord_sys () != PIXEL_REAL)){
      frid = cid1->get_frame_id();
      if (frid.length() == 0)   frid = current;

      if (chips_strcmp(frid.c_str(), "all") == 0)
      {
         if (frame.size() != 1)
         {
            cmd_manager->add_err(ERR_USED_ALL_AS_ID);
            return ERR_USED_ALL_AS_ID;
         }
         ChipsFrame* cf = get_frame(frid);
         frid = cf->get_id(); 
      }
      // This will be used to determine if the coordinates
      // are coming from two different frames
      frame_split++;
   }

   // If the coord sys is not a window coordinate find the
   // current frame
   if ((cid2->get_coord_sys () != WINDOW_NORM) &&
       (cid2->get_coord_sys () != PIXEL) &&
       (cid2->get_coord_sys () != PIXEL_REAL)){

      frid2 = cid2->get_frame_id();
      if (frid2.length() == 0)   frid2 = current;

      if (chips_strcmp(frid2.c_str(), "all") == 0)
      {
         if (frame.size() != 1)
         {
            cmd_manager->add_err(ERR_USED_ALL_AS_ID);
            return ERR_USED_ALL_AS_ID;
         }
         ChipsFrame* cf = get_frame(frid2);
         frid2 = cf->get_id();
      }
      // Set the default frame to this one
      if (frame_split != 1)
         frid = frid2;

      // This will be used to determine if the coordinates
      // are coming from two different frames
      frame_split++;
   }


   // If frame split is 2 and frame 1 doesn't equal frame 2
   // then the coordinates are coming from two different frames
   if ((frame_split == 2) && (frid != frid2)){

      // Convert the coordinate back to window norm
      COORD_SYS tmp_coord = id_out.get_coord_sys ();
      id_out.set_coord_sys (WINDOW_NORM);
      if (frid != ""){
         if ((cf = get_frame(frid)) != NULL)
            rval = cf->convert_coordinate (in_val, out_val, num_vectors,  &id_in, &id_out, axis);
	 else 
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
      }
      else   cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
      
      // Convert the coordinate from window norm to the desired system
      if (rval == ERR_STATUS_GOOD){
	 id_in.set_coord_sys (WINDOW_NORM);
	 id_out.set_coord_sys (tmp_coord);

         if (frid2 != ""){
            if ((cf = get_frame(frid2)) != NULL)
               rval = cf->convert_coordinate (out_val, out_val, num_vectors, &id_in, &id_out, axis);
            else
               cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
         }
         else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
      }
   }
   // If frame split is zero then the coordinates are all window based
   else if (frame_split == 0){

      // Handle single values by creating a vector and populating
      // the correct axis
      if ((axis == X_AXIS) || (axis == Y_AXIS)){
         ChipsVector *tmp_in = new ChipsVector[num_vectors];
         ChipsVector *tmp_out = new ChipsVector[num_vectors];
         int pos = (axis == X_AXIS) ? 0 : 1;

         for (int ii=num_vectors; ii--;)
         {
            tmp_in[ii] = ChipsVector(3);
            tmp_in[ii][pos] = in_val[ii][0];
            tmp_out[ii] = ChipsVector(3);
         }

         // Convert the coordinate
         if (!rval)
            rval = coord->convert_cs_to_cs (tmp_in, tmp_out, num_vectors,
                        id_in.get_coord_sys (), id_in.get_coord_id (),
                        id_out.get_coord_sys (), id_out.get_coord_id ());

         for (int ii=num_vectors; ii--;)
            out_val[ii][0] = tmp_out[ii][pos];

         delete []tmp_out;
         delete []tmp_in;

      }
      else if (axis == XY_AXIS)
      {
         // Convert the coordinate
         if (in_val[0].get_size () == 3)
            rval = coord->convert_cs_to_cs (in_val, out_val, num_vectors,
                        id_in.get_coord_sys (), id_in.get_coord_id (),
                        id_out.get_coord_sys (), id_out.get_coord_id ());
         else
	    cmd_manager->add_err(rval=ERR_BAD_COORD_DIM_SZ);

      }
      else cmd_manager->add_err(rval = ERR_INVALID_AXIS);
   }
   // Otherwise the coordinates are coming from the same frame
   else {
      if (frid != ""){
         if ((cf = get_frame(frid)) != NULL)
            rval = cf->convert_coordinate (in_val, out_val, num_vectors, &id_in, &id_out, axis);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
      }
      else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   }
   return rval;
}

int ChipsWindow::get_current_coord (ChipsId *cid)
{
   ChipsFrame *cf;
   int rval = ERR_STATUS_GOOD;
   string oid = cid->get_frame_id ();
   if (oid.length () == 0) oid = current;

   // If the length is zero then the current coordinate is here
   if (oid.length () == 0){
      cid->set_coord_sys (PIXEL);
      cid->set_coord_id ("");
   }
   // Find the current coordinate of the frame
   else{
      
      if ((chips_strcmp(oid.c_str(), "all") != 0) || (frame.size() == 1))
      {
         if ((cf = get_frame(oid)) != NULL)
            rval = cf->get_current_coord (cid);
      }
      else 
      {
         cmd_manager->add_err(ERR_USED_ALL_AS_ID);
         return ERR_USED_ALL_AS_ID;
      }
   }
   return rval;
}

int ChipsWindow::get_coord_object_count(ChipsId *cid, int objtype)
{
   ChipsFrame *cf;
   int err, cnt = 0;
   string oid = cid->get_frame_id ();
   if (oid.length () == 0) oid = current;

   if ((chips_strcmp(oid.c_str(), "all") != 0) || (frame.size() == 1))
   {
      if ((cf = get_frame(oid)) != NULL)
      {
         ChipsId tmp_cid (*cid);
         ChipsCoordinate* ccoord = cf->find_coordinate_system(tmp_cid, err);
         if (ccoord)   cnt = ccoord->get_num_objs(objtype);
      }
   }
   return cnt;
}

bool ChipsWindow::check_coordinate_exists (ChipsId *cid)
{
   bool rval = true;
   ChipsFrame *cf;

   if (!cid)
      return false;   // ChipsId object not specified

   ChipsId tmp_cid (*cid);

   // The coord id should not be used for anything but data coordinates
   if (tmp_cid.get_coord_sys () != DATA)
       tmp_cid.set_coord_id ("");

   // If the coord sys is not a window coordinate find the
   // current frame
   if ((cid->get_coord_sys () == WINDOW_NORM) ||
       (cid->get_coord_sys () == PIXEL) ||
       (cid->get_coord_sys () == PIXEL_REAL))
      rval = true;
   else
   {
      string frid = cid->get_frame_id();
      if (frid.length() == 0)   frid = current;

      cf = get_frame(frid);

      if (cf)
      {
         int err;
         rval = (cf->find_coordinate_system (tmp_cid, err) != NULL);
      }
      else
         rval = false;

   }

   return rval;
}


int ChipsWindow::add_region(ChipsId* cid, double* xx, double* yy, 
   			  unsigned int count)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      int sts, pass = 0, failval = 0;
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_region(cid, xx, yy, count)) ==
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_region(cid, xx, yy, count);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();

      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);


      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_region(cid, xx, yy, count);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::add_region(ChipsId* cid, int sides, double xx, double yy, 
   			    double radius)
{
   int rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;

   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->add_region(cid, sides, xx, yy, radius)) == 
             ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->add_region(cid, sides, xx, yy, radius);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else
   {
      // don't have a frame and none specified so create a new one
      bool cmd_lock = cmd_manager->cmd_lock ();
      
      // add frame will reset the coord info -- save it off
      COORD_SYS tmp_coord = cid->get_coord_sys ();
      string tmp_coord_id = cid->get_coord_id ();
      rval = add_frame(cid);

      cid->set_coord_sys (tmp_coord);
      cid->set_coord_id (tmp_coord_id);


      cmd_manager->set_cmd_lock (true);
      if ((cf = get_frame(current)) != NULL)
         rval = cf->add_region(cid, sides, xx, yy, radius);
      cmd_manager->set_cmd_lock (cmd_lock);
   }
   if (rval)  cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::delete_region(ChipsId* cid)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }
   string frid = cid->get_frame_id();
   if (frid.length() == 0)   frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      string rid = cid->get_region_id();
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         cid->set_region_id(rid);
         if ((sts = (*ii).second->delete_region(cid)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      else
         set_current ("all");
   }
   else if (frid != "")
   {
      ChipsFrame *cf;
      if ((cf = get_frame(frid)) != NULL)
      {
         rval = cf->delete_region(cid);
         set_current (cf->get_id());
      }
      else cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   if (rval)   cmd_manager->add_window_id_to_errs(id);
   return rval;
}


int ChipsWindow::arrange_frames(int cnt, string* fids)
{
   int rval = ERR_STATUS_GOOD;
   if (duration) 
   {
      cmd_manager->add_err(ERR_BLINK_ACTIVE,id);
      return ERR_BLINK_ACTIVE;
   }

   // inefficient - needs to be updated !!!!!!!
   if (cnt == (int)frame.size())
   {
      int cc;
      int* check =(int*) calloc(cnt, sizeof(int)); 
      map<string, ChipsFrame*>::iterator ii;
 
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         ChipsFrame *cf = (*ii).second;
         string val = cf->get_id();

         for (cc = cnt; cc--; )
            if (val == fids[cc])  
            {
               check[cc] = 1;
               cc = 0; 
            } 
      }
      for (cc = cnt; cc--; ) if (!check[cc]) rval += 1; 

      if (rval == 0)
      {
         // all entries in arrangment list map 1-1 to actual frames
         // reorder list 
         for (cc = 0; cc < cnt; cc++)
         {
            ChipsFrame *cf = get_frame(fids[cc]);
	    cf->set_frame_num (cc);
            frame_ord = cnt; 
         }
         frame_order.sort (comp_frame_order);

	 re_order_frames ();
      } 
   }
   else
      cmd_manager->add_err(rval=ERR_REORDER_LIST_LEN,id,0,cnt,(int)frame.size());
   return rval;
}


int ChipsWindow::get_order(list<string> &out_order)
{
   if (frame.size() != 0)
   {
      list<ChipsFrame*>::iterator ii;

      for(ii=frame_order.begin(); ii != frame_order.end(); ++ii)
         out_order.push_back ((*ii)->get_id ());

   }
   return ERR_STATUS_GOOD; 
}


int ChipsWindow::display_grid(int val)
{
   int rval;
   vis_grid = true; 
   this->change_grid_visible ();
   if (activated) rval = render_grids();
   return rval; 
}


int ChipsWindow::set_grid_scale(int val)
{
   int rval = ERR_STATUS_GOOD; 

   if (val != gscale)
   {
      switch(val)
      {
         case V_GRID_SCALE_HALF_INCH:   // fallthrough intended
         case V_GRID_SCALE_QTR_INCH:    // fallthrough intended
         case V_GRID_SCALE_CENTIMETER:  // fallthrough intended
         case V_GRID_SCALE_NORMALIZED:
            gscale = val;
         break; 
         case V_GRID_SCALE_DEFAULT:
         break; 
         default:
            cmd_manager->add_err(rval=ERR_STR_SUBSET,id,0,"grid scale", 
                                 val, "0:4");
         break; 
      }
   }

   if (!rval && activated)   rval = render_grids();

   return (rval); 
}


int ChipsWindow::hide_grid()
{
   int rval = ERR_STATUS_GOOD; 
   if (vis_grid)
   {
      // call virtual to hide grids  
      vis_grid = false; 
      rval = change_grid_visible();      
   }
   return (rval);
}


int ChipsWindow::set_grid_color(string in_color)
{
   int rval = ERR_STATUS_GOOD;
   if ((rval = grid_color.set_foreground_color(in_color)) == ERR_STATUS_GOOD)
      if (activated)   rval = change_grid_color();
   return rval;
}


int ChipsWindow::set_grid_thickness(float val)
{
   int rval = ERR_STATUS_GOOD;
   if (grid_thickness == val)  return rval; 
   if ((val < V_THICK_MIN_VAL) || (val > V_THICK_MAX_VAL))
      cmd_manager->add_err(rval=ERR_FLT_MINMAX,id,0,"grid thickness", val,
                           V_THICK_MIN_VAL, (float)V_THICK_MAX_VAL);
   else
   {
      grid_thickness = val; 
      if (activated)   rval = change_grid_thickness(); 
   }
   return (rval);
}


int ChipsWindow::set_grid_linestyle(V_LINE_TYPE val)
{
   int rval = ERR_STATUS_GOOD;

   if (val != grid_linestyle)
   {
      if (valid_line_type(val))
      {
         grid_linestyle = val;
         if (activated)   rval = change_grid_linestyle();
      }
      else
         cmd_manager->add_err(rval=ERR_LINETYPE_DNE,id,0,val);
   }
   else  rval = ERR_STATUS_GOOD;  // style already set to value

   return rval;
}

int ChipsWindow::print(string outfile, string format, V_PRINTFILE_TYPE scheme,
   string* wrotename)
{
   int rval = 0, use_format; 
   string checkfile = outfile;
   const char* ext[] = { ".ps", ".eps", ".pdf", ".png", ".jpg", ".jpeg"};
   bool add_ext = true;
   list<string> font_names;
   
   print_mode = true;


   // check the extension of the outfile to and use that as the format
   // if it's a proper extensions otherwise refer to the preferences
   string::size_type ext_pos = string::npos;

   if (outfile.length () != 0)
      ext_pos = outfile.find_last_of (".", outfile.length ());

   if ((ext_pos != string::npos) && (ext_pos+1 != outfile.length ()))
   {
      string outfile_ext = outfile.substr (ext_pos, outfile.length ());
      for (int ii=0; ii<6; ii++)
      {
         if (outfile_ext == ext[ii])
         {
            add_ext = false;
            use_format = ii;
         }
      }
   }

   if (add_ext)
   {
      if (format.length() == 0)
         use_format = 0;          // default format (postscript)
      else if (chips_strcmp(format.c_str(), "ps") == 0)
         use_format = 0;          // postscript format
      else if (chips_strcmp(format.c_str(), "eps") == 0)
         use_format = 1;          // encapsulated postscript format
      else if (chips_strcmp(format.c_str(),"pdf") == 0)
         use_format = 2;          // pdf format
      else if (chips_strcmp(format.c_str(),"png") == 0)
         use_format = 3;          // png format
      else if (chips_strcmp(format.c_str(),"jpg") == 0)
         use_format = 4;          // jpg format
      else if (chips_strcmp(format.c_str(),"jpeg") == 0)
         use_format = 5;          // jpg format
      else{
         print_mode = false;
         cmd_manager->add_err(ERR_STR_SUBSET,id,0,"export format",
                           export_format.c_str(),"ps|eps|png|pdf|jpg|jpeg");
         return ERR_STR_SUBSET;  // unsupported format
      }
   }

   // must specify file for cmyk components
   if ((outfile.length() == 0) && (scheme == chips_cmyksep)){
      print_mode = false;
      cmd_manager->add_err(ERR_FILE_NOT_SUPPLIED);
      return ERR_FILE_NOT_SUPPLIED;  
   }

   // if the format is pdf produce warning if that preloaded fonts
   // may not appear correct
   if (theFonts && (use_format == 2))
   {
     theFonts->get_font_names (font_names);

     // if there are more than 4 fonts than the user must have loaded
     // user defined fonts 
     if (font_names.size () > 4)
        cmd_manager->add_err ("User defined fonts may not appear correct in pdf files.", "", 1);
   }
 
   // cmyk only supported for ps, eps 
   if (((scheme == chips_cmyksep) || (scheme == chips_cmyk)) && 
       ((use_format != 0) && (use_format != 1) && (use_format != 2))){
      print_mode = false;
      cmd_manager->add_err(ERR_CMYK_NONPS);
      return ERR_CMYK_NONPS;
   }   

   if ((scheme != chips_cmyksep) && outfile.length())
   { 
      // check for directory write permission
      string fpath = outfile;
      unsigned int spos = fpath.rfind('/');
      if ((spos >= 0) && (spos < fpath.size())) 
      {
	 // found a '/' in path- remove filename portion
         fpath.erase(spos); 
         if (access(fpath.c_str(), W_OK) != 0)
         {
            print_mode = false;
            cmd_manager->add_err(ERR_DIR_WRITE_PERM,id,0,fpath.c_str());
            return ERR_DIR_WRITE_PERM;
         }
      }
      // check if file already exists
      if (add_ext)
         checkfile.append(ext[use_format]);  
      if (!access(checkfile.c_str(), F_OK))
      {
	 if (!export_clobber)
         {
            print_mode = false;
            cmd_manager->add_err(ERR_FILE_CLOBBER_OFF,"",0,checkfile.c_str());
            return ERR_FILE_CLOBBER_OFF;
         }
         else if (access(checkfile.c_str(), W_OK) != 0)
         {
            print_mode = false;
            cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
            return ERR_ACCESS_PERMISSION;
         }
      }
   }

   switch(scheme)
   {
      case chips_rgb:
         rval = do_print(outfile, use_format, 6, add_ext); 
         if (rval == ERR_STATUS_GOOD && wrotename) 
         {
            *wrotename = outfile; 
            if (add_ext) (*wrotename).append(ext[use_format]);
         }
      break;
      case chips_cmyk:
         rval = do_print(outfile, use_format, 0, add_ext);   // cymk composite
         if (rval == ERR_STATUS_GOOD && wrotename) 
         {
            *wrotename = outfile; 
            if (add_ext) (*wrotename).append(ext[use_format]);
         }
      break;
      case chips_grayscale:
         rval = do_print(outfile, use_format, 5, add_ext); 
         if (rval == ERR_STATUS_GOOD && wrotename) 
         {
            *wrotename = outfile; 
            if (add_ext) (*wrotename).append(ext[use_format]);
         }
      break; 
      case chips_cmyksep:
      {
         string file_c, file_y, file_m, file_k;
         string basefile = "";
         // remove the extensions
         if (!add_ext)
            basefile = outfile.substr (0, ext_pos);     
         else
            basefile = outfile; 

         file_c = file_y = file_m = file_k = basefile;
         file_c.append("_c");
         checkfile = file_c; 
         if (add_ext)
            checkfile.append(ext[use_format]);
         if (!access(checkfile.c_str(), F_OK))
         {
            if (!export_clobber)
            {
               print_mode = false;
               cmd_manager->add_err(rval=ERR_CMYK_EXISTS,id,0,"cyan",
                                  checkfile.c_str());
               return rval;
            }
            else if (access(checkfile.c_str(), W_OK) != 0)
            {
               print_mode = false;
               cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
               return ERR_ACCESS_PERMISSION;
            }
         }
         
         file_m.append("_m");
         checkfile = file_m; 

         if (add_ext)
            checkfile.append(ext[use_format]);
         if (!access(checkfile.c_str(), F_OK))
         {
            if (!export_clobber)
            {
               print_mode = false;
               cmd_manager->add_err(rval=ERR_CMYK_EXISTS,id,0,"magenta",
                                    checkfile.c_str());
               return rval;
            }
            else if (access(checkfile.c_str(), W_OK) != 0)
            {
               print_mode = false;
               cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
               return ERR_ACCESS_PERMISSION;
            }
         }

         file_y.append("_y");
         checkfile = file_y;

         if (add_ext)
            checkfile.append(ext[use_format]);
         if (!access(checkfile.c_str(), F_OK))
         {
            if (!export_clobber)
            {
               print_mode = false;
               cmd_manager->add_err(rval=ERR_CMYK_EXISTS,id,0,"yellow",
                                       checkfile.c_str());
               return rval;
            }
            else if (access(checkfile.c_str(), W_OK) != 0)
            {
               print_mode = false;
               cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
               return ERR_ACCESS_PERMISSION;
            }
         }

         file_k.append("_k");
         checkfile = file_k;

         if (add_ext)
            checkfile.append(ext[use_format]);
         if (!access(checkfile.c_str(), F_OK))
         {
            if (!export_clobber)
            {
               print_mode = false;
               cmd_manager->add_err(rval=ERR_CMYK_EXISTS,id,0,"black",
                                       checkfile.c_str());
               return rval;
            }
            else if (access(checkfile.c_str(), W_OK) != 0)
            {
               print_mode = false;
               cmd_manager->add_err(ERR_ACCESS_PERMISSION,"",0,checkfile.c_str());
               return ERR_ACCESS_PERMISSION;
            }
         }
         
         // component cmyk printing
         rval |= do_print(file_c, use_format, 1, add_ext);   // cyan
         rval |= do_print(file_m, use_format, 2, add_ext);   // magenta
         rval |= do_print(file_y, use_format, 3, add_ext);   // yellow
         rval |= do_print(file_k, use_format, 4, add_ext);   // black
      }
      break;
      case chips_nofiletype:   // default value not privided
         cmd_manager->add_err(rval=ERR_COLORSCHEME_DNE);
      break; 
      default:
         cmd_manager->add_err(rval=ERR_COLORSCHEME_DNE);
      break; 
   }
   print_mode = false;
   return rval; 
}

int ChipsWindow::do_print(string outfile, int fmt, int colorSpace, bool add_ext)
{
   cmd_manager->add_err(ERR_FEATURE_DNE,id);
   return ERR_FEATURE_DNE; 
}

int ChipsWindow::render_grids()
{
   cmd_manager->add_err(ERR_FEATURE_DNE,id);
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_grid_color()
{
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_grid_thickness()
{
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::render_layout()
{
   cmd_manager->add_err(ERR_FEATURE_DNE,id);
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_layout_color()
{
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_layout_thickness()
{
   return ERR_FEATURE_DNE; 
}

int ChipsWindow::change_layout_linestyle()
{
   return ERR_FEATURE_DNE; 
}

int ChipsWindow::change_grid_linestyle()
{
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_order()
{
   cmd_manager->add_err(ERR_FEATURE_DNE,id);
   return ERR_FEATURE_DNE; 
}


int ChipsWindow::change_grid_visible()
{
   return ERR_FEATURE_DNE;
}


int ChipsWindow::change_color()
{
   return ERR_FEATURE_DNE;
}

int ChipsWindow::change_size()
{
   return ERR_FEATURE_DNE;
}

int ChipsWindow::change_redraw()
{
   return ERR_FEATURE_DNE;
}


int ChipsWindow::redraw()
{
   return ERR_FEATURE_DNE; 
}

int ChipsWindow::show()
{
#ifndef HEADLESS
   visible = true;

   if (redraw_blocked())  //just have return here
       return ERR_STATUS_GOOD;
   gui_window->show_window ();

   if (coord)
   {
      int old_vis = coord->is_visible(0);
      coord->set_visible (true);
      if (!old_vis)
      {
         // coord visibility changed- remap all items which got reset
         map<string, ChipsFrame*>::iterator ii;
         for(ii = frame.begin(); ii != frame.end(); ++ii)
	    ii->second->force_show();
      }
   }
#endif

   return ERR_STATUS_GOOD;
}

int ChipsWindow::hide()
{
#ifndef HEADLESS
   if (!visible)
      return ERR_STATUS_GOOD;


   visible = false;
   this->remap_hide ();
   gui_window->hide_window ();

   this->drawready = false;
   if (coord)
      coord->set_visible (false);
#endif

   return ERR_STATUS_GOOD;

}

int ChipsWindow::update_print_scale (double scale)
{
   int rval = ERR_STATUS_GOOD;
   map<string, ChipsFrame*>::iterator ii;

   // now go through frames
   for(ii=frame.begin(); ii != frame.end(); ++ii)
      (*ii).second->update_print_scale(scale);
  
   return rval;
}


int ChipsWindow::adjust_default_colors(string newbg, string newfg, 
   string& oldbg, string& oldfg)
{
   int rval = ERR_STATUS_GOOD; 

   oldbg = color.get_background_color();
   oldfg = color.get_foreground_color();

   if ((newbg != oldbg) || (newfg != oldfg))
   {
      map<string, ChipsFrame*>::iterator ii;

      if ((rval = color.set_default_background(newbg)) == ERR_STATUS_GOOD)
         if ((rval = color.set_default_foreground(newfg)) == ERR_STATUS_GOOD)
         {
            // now go through frames
            for(ii=frame.begin(); ii != frame.end(); ++ii)
               (*ii).second->adjust_default_colors();
            change_color(); 
         }
   }
   return rval; 
}


int ChipsWindow::set_default_colors_ps(string newbg, string newfg)
{
   int rval = ERR_STATUS_BAD;
   string defbg, deffg;

   defbg = theDefaults->get_defaultbgps();
   deffg = theDefaults->get_defaultfgps();

   if (newbg.length())
      rval = color.set_default_background(newbg);
   if (rval != ERR_STATUS_GOOD) 
      rval = color.set_default_background(defbg);
   if (rval != ERR_STATUS_GOOD) 
      rval = color.set_default_background("white");

   rval = ERR_STATUS_BAD;
   if (newfg.length())
      rval = color.set_default_foreground(newfg);
   if (rval != ERR_STATUS_GOOD) 
      rval = color.set_default_foreground(deffg);
   if (rval != ERR_STATUS_GOOD) 
      rval = color.set_default_foreground("black");

   if (rval == ERR_STATUS_GOOD)
   {
      map<string, ChipsFrame*>::iterator ii;

      // now go through frames
      for(ii=frame.begin(); ii != frame.end(); ++ii)
          (*ii).second->adjust_default_colors();
      change_color();
   }
   return rval;
}


int ChipsWindow::set_default_colors_display(string newbg, string newfg)
{
   int rval = ERR_STATUS_BAD;

   // set display default foreground and background colors-
   // precedence- 1) user specified colors
   //             2) colors window previously used as defauls
   //             3) default colors from preferences file
   //             4) black background, white foreground 
   if (newbg.length())  
   {
      rval = color.set_default_background(newbg);
      if (rval == ERR_STATUS_GOOD)  bg_display = newbg;
   }
   if (rval != ERR_STATUS_GOOD)
      if ((rval = color.set_default_background(bg_display)) != ERR_STATUS_GOOD)
      {
         string defbg = theDefaults->get_defaultbgdisp();
         if ((rval = color.set_default_background(defbg)) != ERR_STATUS_GOOD)
            rval = color.set_default_background("black");
      }

   rval = ERR_STATUS_BAD;
   if (newfg.length())  
   {
      rval = color.set_default_foreground(newfg);
      if (rval == ERR_STATUS_GOOD)  fg_display = newfg;
   }
   if (rval != ERR_STATUS_GOOD)
      if ((rval = color.set_default_foreground(fg_display)) != ERR_STATUS_GOOD)
      {
         string deffg = theDefaults->get_defaultfgdisp();
         if ((rval = color.set_default_foreground(deffg)) != ERR_STATUS_GOOD)
            rval = color.set_default_foreground("white");
      }

   if (rval == ERR_STATUS_GOOD)
   {
      map<string, ChipsFrame*>::iterator ii;

      // reset colors in frames 
      for(ii=frame.begin(); ii != frame.end(); ++ii)
          (*ii).second->adjust_default_colors();
      change_color();  // update window colors in middle layer
   }
   return rval;
}


int ChipsWindow::set_export_orient(int val)
{
   int rval = ERR_STATUS_GOOD;
   switch(val)
   {
      case 0:  // portrait    - fallthrough intended 
      case 1:  // landscape 
         export_orient = val;
      break;
      default: 
         rval = ERR_EXPORT_ORIENT; 
      break;
   }
   return rval; 
}


int ChipsWindow::set_export_font(bool val)
{
   export_font = val;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_export_clobber(bool val)
{
   export_clobber = val;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_export_dpi(int val)
{
   export_dpi = val;
   return ERR_STATUS_GOOD;
}


int ChipsWindow::set_page_lmargin(float margin_sz)
{
   if (margin_sz < 0)
      return ERR_DATA_RANGE;
   leftmargin = margin_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_rmargin(float margin_sz)
{
   if (margin_sz < 0)
      return ERR_DATA_RANGE;
   rightmargin = margin_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_tmargin(float margin_sz)
{
   if (margin_sz < 0)
      return ERR_DATA_RANGE;
   topmargin = margin_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_bmargin(float margin_sz)
{
   if (margin_sz < 0)
      return ERR_DATA_RANGE;
   bottommargin = margin_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_width(float page_sz)
{
   if (page_sz < 0)
      return ERR_DATA_RANGE;
   page_width = page_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_height(float page_sz)
{
   if (page_sz < 0)
      return ERR_DATA_RANGE;
   page_height = page_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_size(string page_sz)
{
   return string_to_pagesize (page_sz.c_str (), page_width, page_height, page_units);
}

string ChipsWindow::get_page_size()
{
   return v_pagesize_to_string (page_width, page_height, page_units);
}

int ChipsWindow::set_page_units (string val)
{
   return c2_string_to_units(val.c_str (), &page_units);
}
  
string ChipsWindow::get_page_units ()
{
   return v_units_to_string(page_units);
}
int ChipsWindow::set_page_scale_width(float scale_sz)
{
   if (scale_sz < 0)
      return ERR_DATA_RANGE;
   page_scale_width = scale_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_scale_height(float scale_sz)
{
   if (scale_sz < 0)
      return ERR_DATA_RANGE;
   page_scale_height = scale_sz;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_fit_to_page(bool flag)
{
   fit_to_page = flag;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_page_keep_aspect(bool flag)
{
   keep_aspect = flag;
   return ERR_STATUS_GOOD;
}

int ChipsWindow::set_printmethod(const string &method)
{
   printmethod = method;
   return ERR_STATUS_GOOD;
}


void ChipsWindow::toggle_blink_frame()
{
   if (blinkframe)
   {
      set_current(frame_1.c_str());
      update_display();
      redraw();
      blinkframe = 0; 
   }
   else
   {
      set_current(frame_2.c_str());
      update_display();
      redraw();
      blinkframe = 1; 
   }
}

#ifndef HEADLESS
int ChipsWindow::blink(void *data)
{
   ChipsWindow *vw = static_cast<ChipsWindow *>(data);

   if (vw)
   {
      vw->toggle_blink_frame();
   }
   return 1; 
}
#endif


ChipsFrame* ChipsWindow::new_frame(string fid)
{
   ChipsFrame* cf = NULL;

   if (fid.length() == 0)
   {
      string tmpid = create_frame_id();
      cf =   new ChipsFrame(this, tmpid);
   }
   else
      cf =   new ChipsFrame(this, fid);

   return cf;
}


ChipsFrame* ChipsWindow::new_frame(ChipsStateStream& in)
{
   return new ChipsFrame(this, in);

}


ChipsFrame* ChipsWindow::new_frame(float xmin, float ymin, float xmax,
      float ymax, string fid)
{
   ChipsFrame* cf = NULL;

   if (fid.length() == 0)
   {
      string tmpid = create_frame_id();
      cf =   new ChipsFrame(this, xmin,xmax,ymin,ymax,tmpid);
   }
   else
      cf =   new ChipsFrame(this, xmin,xmax,ymin,ymax,fid);
   return cf;
}


char* ChipsWindow::create_frame_id()
{
   static char temp_id[80];
   string stem;

   if (theDefaults)   stem = theDefaults->get_framestem();
      else            stem = "frm";
   sprintf(temp_id, "%s%1d", stem.c_str(), ++frame_cnt);

   // increment count til next unused frame id 
   while(get_frame(temp_id))
      sprintf(temp_id, "%s%1d", stem.c_str(), ++frame_cnt);
   return temp_id;
}

int ChipsWindow::reference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_BAD;
   if (!co)  return rval;

   if (co->get_object_type () == chips_frame)
   {
      ChipsFrame *cf = static_cast <ChipsFrame *>(co);

//      if (!redraw_blocked ())
         co->show ();
      insert_frame (cf);
      dframe.erase (co->get_ref_cnt ());

      // insert the coordinate system back into the system
      ChipsCoordinate *frame_coord = co->get_chips_coord ();
      if (frame_coord)
         rval = coord->add_coord_sys (frame_coord);


      if (change_cnt)
         frame_cnt++;

      // If a current string is passed in use it. It is possible that
      // if a delete all occured then undone that all is current. Otherwise
      // use referenced frame
      if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_frame(curr)))
         set_current (curr);
      else
         set_current (co->get_id ());
      rval = ERR_STATUS_GOOD;
   }
   else if (co->get_object_type () == chips_window)
   {
      if (visible)
         show ();

      rval = eng->reference_object (co, change_cnt, curr);
   }
   return rval;
}

int ChipsWindow::dereference_object (ChipsObject *co, bool change_cnt, const string &curr)
{
   int rval = ERR_STATUS_BAD;

   if (!co)  return ERR_STATUS_BAD;
   if (co->get_object_type() == chips_frame)
   {
      map<string, ChipsFrame* >::iterator ii = frame.find (co->get_id ());
      map<string, ChipsFrame* >::reverse_iterator ri;

      if (ii == frame.end())   return ERR_STATUS_BAD;

      // remove the frame from the window
      erase_frame (co->get_id());
      dframe[co->get_ref_cnt ()] = dynamic_cast <ChipsFrame *>(co);
   
      // remove the frame coord sys from the system
      ChipsCoordinate *frame_coord = co->get_chips_coord ();
      if (frame_coord)
         rval = coord->remove_coord_sys (frame_coord->get_id ());

      // this hides the frame without changing its visibility
      co->force_hide ();

      if (curr.length() && (!chips_strcmp (curr.c_str(), "all") || get_frame(curr)))
         set_current (curr);
      else if (!frame.empty()) 
      {
	 ri = frame.rbegin();
         set_current((*ri).first);
      }
      else
         set_current ("");

      if (change_cnt)
         frame_cnt--;

      rval = ERR_STATUS_GOOD;
   }
   else if (co->get_object_type() == chips_window)
   {
      //this hides the window without changing its visiblity
      this->force_hide ();

      rval = eng->dereference_object (co, change_cnt, curr);
   }
   return rval;
}


int ChipsWindow::clear_cmd (ChipsObject *co)
{
   int rval = ERR_STATUS_BAD;

   if (!co)   return ERR_STATUS_BAD;
   if (co->get_object_type() == chips_frame)
   {
      map <unsigned int, ChipsFrame *>::iterator ii;
      ii = dframe.find (co->get_ref_cnt ());
      if (ii != dframe.end ())
      {
         ii->second->object_delete ();
         dframe.erase (ii);
      }
   }
   else if (co->get_object_type() == chips_window)
      rval = eng->clear_cmd (co);
   return rval;
}



void ChipsWindow::serialize(ChipsStateStream& out)
{
   out.write_string(&id);
   int sz;
   sz = (int) ref_cnt;
   out.write_int(&sz);
   out.write_float(&raw_width);
   out.write_float(&raw_height);
   out.write_int(&width);
   out.write_int(&height);
   out.write_int(&framemode);
   out.write_int(&tile_rows);
   out.write_int(&tile_cols);
   out.write_bool(&redraw_on);
   out.write_bool(&smoothing);
   out.write_bool(&visible);
   if (out.get_minor_version() > 0)
      out.write_string(&title);
   if (out.get_minor_version() > 2)
      out.write_string(&prefix);
   out.write_int(&frame_cnt);
   string col = color.get_color();
   sz = (int)units;
   out.write_int(&sz);
   out.write_string(&col);

   // Serializes the entire coordinate tree
   coord->serialize (out);

   out.write_checkpoint();
   sz = frame.size();
   out.write_int(&sz);
   out.write_int(&sz);

   map <int, string>chron;
   map <int, string>::iterator i2;
   map<string, ChipsFrame*>::iterator ii; 
   // remap so that we are ordered chronologically (or by arrange order)
   for(ii=frame.begin(); ii != frame.end(); ++ii)
      chron.insert(pair<int,string>((*ii).second->get_frame_num(),(*ii).first));
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
      frame[(*i2).second]->serialize(out);
   out.write_string(&current);

   sz = dframe.size();
   out.write_int(&sz);
   map<unsigned int, ChipsFrame*>::iterator jj;
   for(jj = dframe.begin(); jj != dframe.end(); jj++)
      jj->second->serialize(out, true);
}

void ChipsWindow::serialize_ascii(ChipsStateStream& out, int keep)
{
   static char buf[1024];
   string att, stem = theDefaults->get_windowstem();
   string color1 = get_color();
   bool wrotewin = false;
   int gotatts;
   bool same_dims = (theDefaults->get_windowwidth() == raw_width &&
                     theDefaults->get_windowheight() == raw_height &&
                     theDefaults->get_windowunits() == units);

   if (scale_width != theDefaults->get_windowscalewidth())
   {
      sprintf(buf,"scalewidth=%2.1f", scale_width);
      att.append(buf);
   }
   if (scale_height != theDefaults->get_windowscaleheight())
   {
      if (att.length())  att.append(" ");
      sprintf(buf,"scaleheight=%2.1f", scale_height);
      att.append(buf);
   }
   if (color1 != theDefaults->get_windowcolor())
   {
      if (att.length())  att.append(" ");
      att.append("bgcolor=");
      att.append(color1);
   }
   if (prefix != theDefaults->get_windowprefix())
   {
   if (att.length())  att.append(" ");
      att.append("prefix=\\\"");
      att.append(prefix);
      att.append("\\\"");
   }
   if (redraw_on != theDefaults->get_windowdisplay())
   {
      if (att.length())  att.append(" ");
      if (redraw_on)  att.append("display=true");
         else         att.append("display=false");
   }
   if ((gotatts = att.length()) != 0)  att.append(" ");
   att.append("id=");
   att.append(id);
   if (id.find(stem) != 0)  gotatts = 1; // window id was set
   if (same_dims)
   {
      if (gotatts)
      {
         // explicitly call add_window to set properties
         out << "add_window(\""<<att<<"\")"<<endl;
         wrotewin = true;
      }
      else if (frame.size() < 1 || keep)
      {
         // no frames- explicitly call add_window
         out << "add_window(\""<<att<<"\")"<<endl;
         wrotewin = true;
      }
   }
   else
   {
      string myunits = v_units_to_string(units);
      sprintf(buf, "add_window(%3.2f,%3.2f,\"%s\",\"%s\")",
        raw_width,raw_height,myunits.c_str(),att.c_str());
      out << buf << endl;
      wrotewin = true;
   }
   int simplify = 0;
   map <int, string>chron;
   map <int, string>::iterator i2;
   map<string, ChipsFrame*>::iterator ii;
   string last_id;

   // set tile here- so axes limits set correctly
   if (framemode == 2)
   {
      if (!wrotewin)
      {
         out << "add_window(\""<<att<<"\")"<<endl;
      }
      out << "tile("<< tile_cols <<", "<< tile_rows<<")"<<endl;
   }

   // remap so that we are ordered chronologically (or by arrange order)
   for(ii=frame.begin(); ii != frame.end(); ++ii)
      chron.insert(pair<int,string>((*ii).second->get_frame_num(),(*ii).first));
   for(i2 = chron.begin(); i2 != chron.end(); i2++)
   {
      frame[(*i2).second]->serialize_ascii(out, simplify++);
      last_id = (*i2).second;
   }

   // add currency info
   if (frame.size() > 1 && current != last_id)
      out << "current_frame(\""<<current<<"\")"<<endl;
}


int ChipsWindow::serialize_object(ChipsObject *co, ChipsStateStream& out)
{
   if (co)
      co->serialize(out);
   return ERR_STATUS_GOOD;
}


void ChipsWindow::deserialize(ChipsStateStream& in)
{
   int iu;
   string cname;
   in.read_string(&id);
   in.read_int(&iu);
   ref_cnt = iu;
   title = id;
   in.read_float(&raw_width);
   in.read_float(&raw_height);
   in.read_int(&width);
   in.read_int(&height);
   in.read_int(&framemode);
   in.read_int(&tile_rows);
   in.read_int(&tile_cols);
   in.read_bool(&redraw_on);
   in.read_bool(&smoothing);
   in.read_bool(&visible);
   if (in.get_minor_version() > 0)
      in.read_string(&title);
   if (in.get_minor_version() > 2)
      in.read_string(&prefix);
   set_title(title); 
   in.read_int(&frame_cnt);
   in.read_int(&iu);
   units = (V_UNITS_TYPE)iu;
   in.read_string(&cname);
   set_color(cname);

   if (coord)
      delete coord;

   // Deserializes the entire coordinate tree
   coord = new ChipsWindowSystem (in, eng);
   data_system_id = coord->get_id ();
}

int ChipsWindow::deserialize_frames(ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   if (in.get_minor_version() == 0)   set_title(id);
   ChipsFrame *cf = NULL; 
   int fcnt;


   // checkpoint to avoid issues with reading corrupt data
   if ((rval = in.read_checkpoint()) == ERR_STATUS_GOOD) 
   {
      int ival;

      in.read_int(&ival);
      in.read_int(&fcnt);
 
      if (ival == fcnt)
      {
         // iterate through frames 
         for (int rr=0; ((rr< fcnt) && (rval == ERR_STATUS_GOOD)); rr++)
         {
	    cf = new_frame(in);
            if ((cf != NULL) && (cf->get_chips_coord () != NULL))
            {
               cf->set_frame_num(rr);
	       insert_frame (cf);
               set_current (cf->get_id ());

	       if (framemode != 2)
                  cf->set_display_location(framemode, 0, 0, width, height);
               else
                  update_display();
               // add plots
               if ((rval = cf->deserialize_plot(in)) == ERR_STATUS_GOOD)
               {
                  // insert other objects such as labels, and regions
                  rval = cf->deserialize_objects(in); 
               }
               if (!cf->is_visible ())
                  cf->force_hide ();
            }
	    else
	       rval = ERR_STATEFILE_COR;
         }
         frame_order.sort (comp_frame_order);
         in.read_string(&current); 

         // deserialize deleted frames
         in.read_int(&ival);
         for(int jj = 0; jj < ival; jj++)
         {
            cf = new_frame(in);
            if ((cf != NULL) && (cf->get_chips_coord () != NULL))
            {
               dframe[cf->get_ref_cnt ()] = cf;

               if (framemode != 2)
                  cf->set_display_location(framemode, 0, 0, width, height);
               else
                  update_display();
               // add plots
               if ((rval = cf->deserialize_plot(in)) == ERR_STATUS_GOOD)
               {
                  // insert other objects such as labels, and regions
                  rval = cf->deserialize_objects(in);
               }
               if (!cf->is_visible ())
                  cf->hide ();
            }
            else
               rval = ERR_STATEFILE_COR;
         }
      }
      else  rval = ERR_STATEFILE_COR;
   }
   return rval; 
}

ChipsObject *ChipsWindow::deserialize_object (V_OBJECT_TYPE co_type, 
                                                       ChipsStateStream& in)
{
   int rval = ERR_STATUS_GOOD;
   ChipsObject *co = static_cast <ChipsObject *> (0);
   if (co_type == chips_frame)
   {
      ChipsFrame *cf;
      cf = new_frame(in);
      if ((cf != NULL) && (cf->get_chips_coord () != NULL))
      {
          if (framemode != 2)
            cf->set_display_location(framemode, 0, 0, width, height);
         else
            update_display();
         // add plots
         if ((rval = cf->deserialize_plot(in)) == ERR_STATUS_GOOD)
         {
            // insert other objects such as labels, and regions
            rval = cf->deserialize_objects(in);
         }
      }
      if (!rval)
         co = cf;
   }

   return co;

}


int ChipsWindow::update_all ()
{
   int ret_val = ERR_STATUS_GOOD;
   ret_val = change_color ();
   if (ret_val != ERR_STATUS_GOOD)
      return ret_val;
//   ret_val = change_size ();
   return ret_val;
}

void ChipsWindow::set_redraw(bool val)
{ 
   bool change = (redraw_on != val);
   redraw_on = val;
   if (change) change_redraw ();
}

int ChipsWindow::re_order_frames ()
{
   return ERR_FEATURE_DNE;
}


void ChipsWindow::remap_hide ()
{
   return;
}

void ChipsWindow::remap_show ()
{
   return;
}


bool ChipsWindow::redraw_blocked()
{
   if (eng)  // if the engine is available, check for forced redraw lock 
      return eng->redraw_blocked(); 
   else   
      return false; 
}


void ChipsWindow:: set_redraw_ready()   
{ 
   if (!drawready)
   {
      drawready=true;
      // call change redraw here to set up our back buffer in case we
      // get a redraw off command before an EXPOSE event 
      change_redraw();
   }
}


int ChipsWindow::display_depth(ChipsId* cid, int dep)
{
   int rval = ERR_STATUS_GOOD;
   string frid = cid->get_frame_id();;
   if (frid.length() == 0)  frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->display_depth(dep)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
   }
   else if (frid != "")
   {
      ChipsFrame *cf;
      if ((cf = get_frame(frid)) != NULL)
         rval = cf->display_depth(dep);
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::hide_depth(ChipsId* cid, int dep)
{
   int rval = ERR_STATUS_GOOD;
   string frid = cid->get_frame_id();;
   if (frid.length() == 0)  frid = current;

   if (chips_strcmp(frid.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->hide_depth(dep)) == ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
   }
   else if (frid != "")
   {
      ChipsFrame *cf;
      if ((cf = get_frame(frid)) != NULL)
         rval = cf->hide_depth(dep);
      else 
         cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,id,0,"frame",frid.c_str());
   }
   else  cmd_manager->add_err(rval=ERR_NO_CUR_FRAME,id);
   return rval;
}


int ChipsWindow::collapse_depths(ChipsId* cid, int* sources, int srccnt, 
   int dest)
{
   if (frame.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"frames");
      return ERR_OBJ_LIST_EMPTY;
   }
   string fid = cid->get_frame_id();
   if (id.length() == 0)  fid = current;
   if (chips_strcmp(fid.c_str(), "all") != 0)
   {
      ChipsFrame* cf = get_frame(fid);
      if (cf)
         return cf->collapse_depths(sources, srccnt, dest);
      else
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"frame",fid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
   }
   else
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0, rval = ERR_STATUS_GOOD;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->collapse_depths(sources, srccnt, dest)) 
             != ERR_STATUS_GOOD)
          pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      return rval; 
   }
}


int ChipsWindow::rearrange_depth(ChipsId* cid, V_OBJECT_TYPE obj, int manner)
{
   if (frame.empty())
   {
      cmd_manager->add_err(ERR_OBJ_LIST_EMPTY,id,0,"frames");
      return ERR_OBJ_LIST_EMPTY;
   }
   string fid = cid->get_frame_id();
   if (id.length() == 0)  fid = current;
   if (chips_strcmp(fid.c_str(), "all") != 0)
   {
      ChipsFrame* cf = get_frame(fid);
      if (cf)
         return cf->rearrange_depth(cid, obj, manner);
      else
      {
         cmd_manager->add_err(ERR_OBJ_VAL_DNE,id,0,"frame",fid.c_str());
         return ERR_OBJ_VAL_DNE;
      }
   }
   else
   {
      map<string, ChipsFrame*>::iterator ii;
      int sts, pass = 0, failval = 0, rval = ERR_STATUS_GOOD;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
      {
         if ((sts = (*ii).second->rearrange_depth(cid, obj, manner))
             != ERR_STATUS_GOOD)
            pass++;
         else if (!err_type_object_DNE(sts))
            failval = sts;
      }
      cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      if (failval)
         rval = failval;
      else if (!pass)
         cmd_manager->add_err(rval=ERR_ALL_TASK_FAIL,"",0,"frames");
      return rval;
   }
}

void ChipsWindow::disable_all_interaction ()
{
   this->disable_interaction ();
   this->disable_frame_interaction ("all");
}

void ChipsWindow::enable_all_interaction ()
{
   this->enable_interaction ();
   this->enable_frame_interaction ("all");
}


void ChipsWindow::disable_frame_interaction (const string &frame_id)
{
   if (chips_strcmp(frame_id.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         ii->second->disable_all_interaction ();
   }
   else if (frame.count (frame_id))
      frame[frame_id]->disable_all_interaction ();
}

void ChipsWindow::enable_frame_interaction (const string &frame_id)
{
   if (chips_strcmp(frame_id.c_str(), "all") == 0)
   {
      map<string, ChipsFrame*>::iterator ii;
      for(ii=frame.begin(); ii != frame.end(); ++ii)
         ii->second->enable_all_interaction ();
   }
   else if (frame.count (frame_id))
      frame[frame_id]->enable_all_interaction ();
}

void ChipsWindow::insert_frame (ChipsFrame *cf)
{
   frame.insert(pair<string, ChipsFrame*>(cf->get_id (), cf));
   frame_order.push_back (cf);
   frame_order.sort (comp_frame_order);

}

void ChipsWindow::erase_frame (const string &fid)
{
   if (frame.count (fid) > 0)
   {
      ChipsFrame *cf = frame[fid];
      frame_order.remove (cf);
      frame.erase(fid);
   }
}

int ChipsWindow::get_frames (const ChipsId *cid, list <ChipsFrame *>&out_frames, bool allow_all)
{
   int  rval = ERR_STATUS_GOOD;
   ChipsFrame *cf;
   string id = cid->get_frame_id();
   if (id.length() == 0)  id = current;

   if (rval == ERR_STATUS_GOOD)
   {
      if (frame.empty())
         cmd_manager->add_err(rval=ERR_OBJ_LIST_EMPTY,"",0,"frames");
      else if (id.length() && (chips_strcmp(id.c_str(), "all") != 0))
      {
         // get frame
         if ((cf = get_frame(id)) != NULL)
            out_frames.push_back (cf);
         else
            cmd_manager->add_err(rval=ERR_OBJ_VAL_DNE,"",0,"frame",id.c_str());
      }
      else if (allow_all)
      {
         map<string, ChipsFrame*>::iterator ii;

         for(ii=frame.begin(); ii != frame.end(); ++ii)
            out_frames.push_back (ii->second);
         cmd_manager->clear_errs(CLEAR_DNE_ERRS);
      }
      else
         cmd_manager->add_err(rval = ERR_INVALID_ALL_USE);
   }
   else
      cmd_manager->add_err(rval, id);
   return rval;
}

void ChipsWindow::update_grid_properties()
{
   if (theDefaults)
   {
      gscale = theDefaults->get_gridscale();
      float tmpth = theDefaults->get_gridthickness();
      if (tmpth != grid_thickness) 
      {
         grid_thickness = tmpth;
         change_grid_thickness();
      } 
      V_LINE_TYPE tmpst = theDefaults->get_gridlinestyle();
      if (tmpst != grid_linestyle)  
      {
         grid_linestyle = tmpst;
         change_grid_linestyle();
      }   
      vis_grid = theDefaults->get_showgrid();
      string temp = theDefaults->get_gridcolor();
      if (set_grid_color(temp))  set_grid_color("blue");
   }
}


#ifndef HEADLESS
ChipsGtkWindow *ChipsWindow::new_gui_window (string geom, bool focus)
{
   return new ChipsGtkWindow(this, geom, focus);
}
#endif



